# ----------------------------------------------------------------------------
# - warn-verifier-module                                                     -
# - wild cruncher: verifier modules                                    arion -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       warn_verifier_modules
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Arion database verifier modules

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from silkan_common.scmn_utils import first
from wcr_core.common.wcmn_verifier_module import WcmnHierarchyVerificationError
from wcr_core.db.whdb_verifier import WhdbVerifierModuleP
from wcr_core.modules.system.db.wphdb_host_instance import WphdbHostInstance
from wplg_arion.warn_arion_connection_point import \
    WarnArionXPoint
from wplg_arion.warn_distrib_service_type import \
    WARN_ARION_DISTRIBUTION_SERVICE_CODE, WARN_ARION_DISTRIBSVC_LEID_PARAM


class WarnXPointLeidVerifier(WhdbVerifierModuleP):
    """
    Verify that all Arion connection points have a logical equipment ID.
    """

    ERROR_ID = 'hierarchy.system.arion.xpoint-leids'

    ERROR_MESSAGE = 'Connection point without Arion logical equipment ID on {}'

    DESCRIPTION = 'All Arion connection points must be given a ' \
                  'logical equipment ID.'

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        """
        :type instance: WphdbHostInstance
        """
        errors = []

        xpnts = [x for x in instance.getConnectionPoints()
                 if isinstance(x, WarnArionXPoint)]
        """:type: list[WarnArionXPoint]"""

        for xpnt in xpnts:
            if not xpnt.getLogicalEquipmentId():
                errors.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID,
                    self.ERROR_MESSAGE.format(instance.getName()),
                    [instance], self))
        return errors

    def description(self):
        return WarnXPointLeidVerifier.DESCRIPTION


class WarnXPointNumericLeidVerifier(WhdbVerifierModuleP):
    """
    Verify that all Arion connection points have a numeric logical equipment
    ID.
    """

    ERROR_ID = 'hierarchy.system.arion.xpoint-valid-leids'

    ERROR_MESSAGE = 'Invalid connection point Arion logical equipment ID on {}'

    DESCRIPTION = 'All Arion connection points must be given a ' \
                  'valid integer logical equipment ID.'

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        """
        :type instance: WphdbHostInstance
        """
        errors = []

        xpnts = [x for x in instance.getConnectionPoints()
                 if isinstance(x, WarnArionXPoint)]
        """:type: list[WarnArionXPoint]"""

        for xpnt in xpnts:
            leid = xpnt.getLogicalEquipmentId()
            if leid is None or leid == '':
                continue
            try:
                nleid = int(leid)
                assert nleid >= 0
            except:
                errors.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID,
                    self.ERROR_MESSAGE.format(instance.getName()),
                    [instance], self))
        return errors

    def description(self):
        return WarnXPointNumericLeidVerifier.DESCRIPTION


class WarnUniqueXPointLeidVerifier(WhdbVerifierModuleP):
    """
    Verify that all logical Arion equipment IDs for a host are unique.
    """

    ERROR_ID = 'hierarchy.system.arion.unique-leids'

    ERROR_MESSAGE = 'Logical equipment ID {} not unique on {}'

    DESCRIPTION = 'Arion logical equipment IDs must be unique for every host.'

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        """
        :type instance: WphdbHostInstance
        """
        errors = []

        xpnts = [x for x in instance.getConnectionPoints()
                 if isinstance(x, WarnArionXPoint)]
        """:type: list[WarnArionXPoint]"""
        leids = set()
        """:type: set[str]"""

        for xpnt in xpnts:
            leid = xpnt.getLogicalEquipmentId()
            if leid in leids:
                errors.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID,
                    self.ERROR_MESSAGE.format(leid, instance.getName()),
                    [instance], self))
            leids.add(leid)
        return errors

    def description(self):
        return WarnUniqueXPointLeidVerifier.DESCRIPTION


class WarnDistribServiceSocketLeidVerifier(WhdbVerifierModuleP):
    """
    Verify that all a matching connection point exists for all Arion
    distribution services.
    """

    ERROR_ID = 'hierarchy.system.arion.distrib-service-leid'

    ERROR_MESSAGE = 'No logical equipment ID for Arion distribution service ' \
                    '{d} on {i}'

    DESCRIPTION = 'Every Arion distribution service socket must be assigned ' \
                  'a logical equipment ID.'

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        """
        :type instance: WhdbInstanceP
        """
        errors = []
        if isinstance(instance, WphdbHostInstance):
            for process in instance.getProcesses():
                for dsvc in process.getDistributionServiceSockets():
                    if dsvc.getDistributionService() != \
                            WARN_ARION_DISTRIBUTION_SERVICE_CODE:
                        continue

                    if not dsvc.getSocketParameterValue(
                            WARN_ARION_DISTRIBSVC_LEID_PARAM):
                        errors.append(WcmnHierarchyVerificationError(
                            self.ERROR_ID,
                            self.ERROR_MESSAGE.format(
                                d=dsvc.getSocketLabel(), i=instance.getName()),
                            [instance], self))
        return errors

    def description(self):
        return WarnDistribServiceSocketLeidVerifier.DESCRIPTION


class WarnDistribServiceSocketXPointVerifier(WhdbVerifierModuleP):
    """
    Verify that all a matching connection point exists for all Arion
    distribution services.
    """

    ERROR_ID = 'hierarchy.system.arion.distrib-service-xpoint'

    ERROR_MESSAGE = 'No connection point for Arion distribution service {d} ' \
                    'with logical equipment ID {l} on {i}'

    DESCRIPTION = 'For every Arion distribution service instance, a ' \
                  'corresponding connection point must exist for the host.'

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        """
        :type instance: WhdbInstanceP
        """
        errors = []

        if isinstance(instance, WphdbHostInstance):
            xpnts = [x for x in instance.getConnectionPoints()
                     if isinstance(x, WarnArionXPoint)]
            """:type: list[WarnArionXPoint]"""

            for process in instance.getProcesses():
                for dsvc in process.getDistributionServiceSockets():
                    if dsvc.getDistributionService() != \
                            WARN_ARION_DISTRIBUTION_SERVICE_CODE:
                        continue

                    leid = dsvc.getSocketParameterValue(
                        WARN_ARION_DISTRIBSVC_LEID_PARAM)
                    """:type: dict[str, object]"""
                    if not leid:
                        continue

                    xpoint = first(x for x in xpnts
                                   if x.getLogicalEquipmentId() == leid)

                    if not xpoint:
                        errors.append(WcmnHierarchyVerificationError(
                            self.ERROR_ID,
                            self.ERROR_MESSAGE.format(
                                d=dsvc.getSocketLabel(), l=leid,
                                i=instance.getName()),
                            [instance], self))
        return errors

    def description(self):
        return WarnDistribServiceSocketXPointVerifier.DESCRIPTION
