# ----------------------------------------------------------------------------
# - scmn-cfg-table                                                           -
# - silkan python library: config table                                      -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from collections import Sized, Container

from .scmn_cfg_table_record import *


# ----------------------------------------------------------------------------
# - table class                                                             -
# ----------------------------------------------------------------------------

class ScmnConfigTable(Sized, Container):
    def __init__(self):
        #: the record table
        self.__rtbl = {}

    def addRecord(self, rec):
        """
        Add a record to the table.

        :param rec: Record to be added
        :type rec: ScmnConfigTableRecord
        """
        name = rec.getName()
        if not name: return
        assert (name not in self.__rtbl)
        self.__rtbl[name] = rec

    def newRecord(self, name, info=None, rval=None, text=None):
        """
        Create and add a new record to the table.

        :param name: Record name
        :type name: str
        :param info: Record info
        :type info: str
        :param rval: Record value
        :param text: Record text
        :type text: str
        """
        rec = ScmnConfigTableRecord(name=name, info=info, rval=rval, text=text)
        self.addRecord(rec)

    def isRecord(self, name):
        """
        Whether this table contains a record with the given name.

        :param name: Name of the record to check
        :type name: str
        :return: True iff this table contains a record with the given name
        :rtype: bool
        """
        # noinspection PyTypeChecker
        return name in self

    def getRecord(self, name):
        """
        Get a record for a given name.

        :param name: Name of the record to retrieve
        :type name: str
        :return: Record with the given name, if any
        :rtype: ScmnConfigTableRecord
        """
        return self[name]

    def getRecordValue(self, name):
        """
        Get the value for the record with the given name.

        :param name: Name of the record to retrieve the value of
        :type name: str
        :return: Value of the record with the given name, if any
        """
        record = self[name]
        return record.getValue() if record is not None else None

    def setRecordValue(self, name, value, info=None, text=None):
        """
        Set the value for the record with the given name.
        Will create a new record if it doesn't exist yet.

        :param name: Record name
        :type name: str
        :param info: Record info
        :type info: str
        :param value: Record value
        :param text: Record text
        :type text: str
        """
        record = self[name]
        if record is not None:
            self.__rtbl[name] = ScmnConfigTableRecord(
                name, rval=value, info=record.getInfo(), text=record.getText())
        else:
            self.__rtbl[name] = ScmnConfigTableRecord(
                name, rval=value, info=info, text=text)

    def __len__(self):
        return len(self.__rtbl)

    def __getitem__(self, item):
        """
        Get a record for a given name.

        :param item: Name of the record to retrieve
        :type item: str
        :return: Record with the given name, if any
        :rtype: ScmnConfigTableRecord
        """
        return self.__rtbl.get(item)

    def __contains__(self, item):
        """
        Whether this table contains a record with the given name.

        :param item: Name of the record to check
        :type item: str
        :return: True iff this table contains a record with the given name
        :rtype: bool
        """
        return item in self.__rtbl

    def __iter__(self):
        return self.__rtbl.items().__iter__()
