# ----------------------------------------------------------------------------
# - scmn-file-path                                                           -
# - silkan python library: file path functions                               -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2012-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       scmn_file_path
   :platform:     Unix, Windows
   :synopsis:     Silkan Python Library: path management

.. moduleauthor:: Amaury Darsch <amaury.darsch@silkan.com>
"""

import os
import re
import shutil

from silkan_common.scmn_utils import first


SCMN_VALID_PATH_REGEX = re.compile(r'[^A-Za-z0-9/_\-\\]')


def scmn_path_base(path):
    """
    Extract the file path base name.

    :param path: the file path to process
    :type path: str
    :return: File basename
    :rtype: str
    """
    # check for none
    if not path: return None
    # get base name
    return os.path.basename(path)


def scmn_path_extension(path):
    """
    Extract the file extension from a path.

    :param path: the file path to process
    :type path: str
    :return: File extension
    :rtype: str
    """
    if not path:
        return None
    _, ext = os.path.splitext(path)
    return ext.lstrip('.') if ext else None


def scmn_path_base_extension(path):
    """
    Extract the filename and the extension from the given path.

    :param path: the file path to process
    :type path: str
    :return: File basename and extension
    :rtype: (str, str)
    """
    if not path:
        return None, None

    fileName = os.path.basename(path)
    name, ext = os.path.splitext(fileName)
    return name, ext.lstrip('.') if ext else None


def scmn_path_home():
    """
    Get the home directory path

    :return: Home directory
    :rtype: str
    """
    # check for posix
    if os.name == 'posix':
        return os.path.expanduser('~')
    # check for wonderland
    if os.name == 'nt':
        return os.environ.get('USERPROFILE')


def scmn_path_join(*parts):
    """
    Join multiple names into a single path

    :param parts: Path parts to join
    :type parts: Iterable[str]
    :return: Joined path
    :rtype: str
    """
    return os.path.join(*[p for p in parts if p is not None])


def scmn_path_mkabs(path, bdir=None):
    """
    Make the given path absolute if it is a relative path, using the given
    base directory or the current working directory if no base directory is
    given.

    :param path: Absolute or relative path
    :type path: str
    :param bdir: Base directory
    :type bdir: str
    :return: Absolute path
    :rtype: str
    """
    if os.path.isabs(path):
        return path
    elif bdir:
        return os.path.normpath(os.path.join(bdir, path))
    else:
        return os.path.abspath(path)


def scmn_path_normalize(path):
    """
    Normalize a path for the undelying os.

    :return: Normalized path
    :rtype: str
    """
    if path is None:
        return None
    return os.path.normpath(path)


def scmn_path_split(path):
    """
    Split a path into its components.

    :param path: Path to split
    :type path: str
    :return: Elements of the given path
    :rtype: list[str]
    """
    return os.path.split(path)


def scmn_isfile(path):
    """
    Check if a file exists

    :param path: the directory path
    :type path: str
    :return: True iff a file exists at the given path
    :rtype: bool
    """
    # check for none
    if path is None:
        return False
    # validate by os
    return os.path.isfile(path)


def scmn_exists(path):
    """
    Check if a path exists

    :param path: the directory path
    :type path: str
    :return: True iff a file exists at the given path
    :rtype: bool
    """
    # check for none
    if path is None:
        return False
    # validate by os
    return os.path.exists(path)


def scmn_isdir(path):
    """
    Check if a directory exists

    :param path: the directory path
    :type path: str
    :return: True iff a directory exists at the given path
    :rtype: bool
    """
    # check for none
    if path is None:
        return False
    # validate by os
    return os.path.isdir(path)


def scmn_mkdir(path):
    """
    Create a directory by path

    :param path: the directory path
    :type path: str
    """
    # check for none
    if path is None:
        return
    # check for existence
    if os.path.isdir(path):
        return
    # create hierarchically
    os.makedirs(path)


def scmn_rmdir(path, onerror=None):
    """
    Remove a directory tree by path

    :param path: the directory path
    :type path: str
    :param onerror: function to call on errors
    :type onerror: () -> _
    """
    # check for none
    if path is None:
        return
    # check for existence
    if not os.path.isdir(path):
        return
    # remove hierarchically
    shutil.rmtree(path, onerror=onerror)


def scmn_rm(path):
    """
    Remove a file by path

    :param path: the file path
    :type path: str
    """
    # check for none
    if path is None:
        return
    # check for existence
    if not os.path.isfile(path):
        return
    # remove
    os.remove(path)


def scmn_expshenv(path):
    """
    Expand tilde and environment variables in the given path.

    :param path: Path possibly containing ~ and ${...}
    :type path: str
    :return: Path with ~ and ${...} expanded
    :rtype: str
    """
    if path is None:
        return None
    path = os.path.expanduser(path)
    if path is not None and '$' in path:
        path = re.sub(r'\${(\w+)}', lambda x: os.environ.get(x.group(1), ''),
                      path)
    return path


def scmn_path_dir_base(filepath):
    """
    Get the directory and filename part of the given path.

    :param filepath: File path
    :type filepath: str
    :return: Directory and filename of the path
    :rtype: (str, str)
    """
    return os.path.dirname(filepath), os.path.basename(filepath)


def scmn_path_dir(filepath):
    """
    Get the directory part of the given path.

    :param filepath: File path
    :type filepath: str
    :return: Directory of the path
    :rtype: str
    """
    if not filepath:
        return filepath
    return os.path.dirname(filepath)


def scmn_listdir(dirpath, fullpath=False, pattern=None):
    """
    List the files/directories in a given directory.

    :param dirpath: Directory to list entries from
    :type dirpath: str
    :param fullpath: Whether to return the full paths of the entries (`True`),
                     or just their names (`False`)
    :type fullpath: bool
    :param pattern: Pattern to match the entry names against, only matching
                    entries will be returned
    :type pattern: str
    :return: List of directory entries that match the pattern (if any), either
             as paths or file names
    :rtype: list[str]
    """
    if not scmn_isdir(dirpath):
        return []
    try:
        entries = sorted([e for e in os.listdir(dirpath)
                          if not e.startswith('.')])
    except FileNotFoundError:
        return []

    if pattern:
        m = re.compile(pattern + '$')
        entries = [e for e in entries if m.match(e) is not None]

    if fullpath:
        return [os.path.join(dirpath, f) for f in entries]
    else:
        return entries


def scmn_isdirempty(dirpath):
    """
    Whether a given directory is non-existent or empty.

    :param dirpath: Directory to check
    :type dirpath: str
    :return: True iff the directory is empty, or does not exist
    :rtype: bool
    """
    if not scmn_isdir(dirpath):
        return True

    entry = first(e for e in os.listdir(dirpath) if e not in ['.', '..'])

    return entry is None


def scmn_ispath(path):
    """
    Whether the given path is valid, i.e. syntactically correct.
    This function does not test whether a file/directory with the given path
    exists.

    :param path: Path to check for validity
    :type path: str
    :return: True iff the given path is valid
    :rtype: bool
    """
    return path.startswith('/') if path else False


def scmn_abspath(path, base=None):
    """
    Get the absolute version of the given path.

    :param path: Path
    :type path: str
    :param base: Path prefix to turn relative path into absolute path,
                 rather than current working directory
    :type base: str
    :return: Absolute path
    :rtype: str
    """
    if os.path.isabs(path):
        return path
    if base:
        path = os.path.join(base, path)
    return os.path.abspath(path) if path else ''


def scmn_relpath(path, base):
    """
    Get the relative the given path to the given base.

    :param path: Path
    :type path: str
    :param base: Path from where the relative path will start
    :type base: str
    :return: Absolute path
    :rtype: str
    """
    abspath = scmn_abspath(path)
    absbase = scmn_abspath(base)

    if abspath is '':
        return ''
    elif absbase is '':
        return abspath
    return os.path.relpath(abspath, base)


def scmn_filesize(path):
    """
    Get the filesize/length of the given file.

    :param path: Path
    :type path: str
    :return: Filesize
    :rtype: int
    """
    return os.path.getsize(path) if path else 0


def scmn_path_rel2file(file, *relPaths):
    _relPaths = [x for r in relPaths for x in r.split('/')]
    return os.path.join(scmn_path_dir(file), *_relPaths)
