# ----------------------------------------------------------------------------
# - scmn-geometry                                                            -
# - silkan python library: geometry objects                                  -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       scmn_geometry
   :platform:     Unix, Windows
   :synopsis:     Silkan Python Library - common geometry objects

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod, ABCMeta


class ScmnPointP(metaclass=ABCMeta):
    """
    Abstract base class for two-dimensional geometric points.
    """

    @abstractmethod
    def getX(self):
        """
        Get the x coordinate.

        :return: X coordinate
        :rtype: float
        """
        raise NotImplementedError()

    @abstractmethod
    def getY(self):
        """
        Get the y coordinate.

        :return: Y coordinate
        :rtype: float
        """
        raise NotImplementedError()

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError()

    def __str__(self):
        return '({}, {})'.format(self.getX(), self.getY())


# noinspection PyAbstractClass
class ScmnPointMP(ScmnPointP):
    """
    Abstract base class for mutable two-dimensional geometric points.
    """

    @abstractmethod
    def setXY(self, x, y):
        """
        Set the x coordinate.

        :param x: New x coordinate
        :type x: float
        :param y: New x coordinate
        :type y: float
        """
        raise NotImplementedError()

    @abstractmethod
    def setX(self, x):
        """
        Set the x coordinate.

        :param x: New x coordinate
        :type x: float
        """
        raise NotImplementedError()

    @abstractmethod
    def setY(self, y):
        """
        Set the y coordinate.

        :param y: New x coordinate
        :type y: float
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class ScmnPoint(ScmnPointP):
    """
    Default implementation for two-dimensional geometric points.
    """

    def __init__(self, x, y):
        """
        :param x: the x coordinate
        :param y: the y coordinate
        """
        self.__x = x
        self.__y = y

    def getX(self):
        return self.__x

    def getY(self):
        return self.__y


class ScmnRectangleP(metaclass=ABCMeta):
    """
    The ScmnRectangle is a two-dimensional rectangle. The rectangle is
    defined by the top-left coordinate and a wight/height tuple.
    """

    @property
    @abstractmethod
    def x(self):
        """
        Get the rectangle's upper-left corner's x position.

        :return: X-position of the rectangle
        :rtype: float
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def y(self):
        """
        Get the rectangle's upper-left corner's y position.

        :return: Y-position of the rectangle
        :rtype: float
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def width(self):
        """
        Get the rectangle's width.

        :return: Width of the rectangle
        :rtype: float
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def height(self):
        """
        Get the rectangle's height.

        :return: Height of the rectangle
        :rtype: float
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def left(self):
        """
        Get the rectangle's left-most X-coordinate.

        :return: Left x-coordinate
        :rtype: float
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def right(self):
        """
        Get the rectangle's right-most X-coordinate.

        :return: Right x-coordinate
        :rtype: float
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def top(self):
        """
        Get the rectangle's top-most Y-coordinate.

        :return: Top y-coordinate
        :rtype: float
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def bottom(self):
        """
        Get the rectangle's bottom-most Y-coordinate.

        :return: Bottom y-coordinate
        :rtype: float
        """
        raise NotImplementedError()

    @abstractmethod
    def united(self, other):
        """
        Return the union of this rectangle with the given rectangle.

        :param other: Rectangle to create union with
        :type other: ScmnRectangle
        :return: Union rectangle
        :rtype: ScmnRectangle
        """
        raise NotImplementedError()

    def __str__(self):
        return '({} / {}) - ({} / {})'.format(
            self.x, self.y, self.width + self.x, self.height + self.y)


class ScmnRectangle(ScmnRectangleP):
    """
    The ScmnRectangle is a two-dimensioanl rectangle. The rectangle is
    defined by the top-left coordinate and a wight/height tuple.
    """

    def __init__(self, x=0, y=0, w=0, h=0):
        """
        :param x: the top left x coordinate
        :type x: float
        :param y: the top left y coordinate
        :type y: float
        :param w: the rectangle width
        :type w: float
        :param h: the rectangle height
        :type h: float
        """
        # reference position
        self.__x = x if w >= 0 else x + w
        self.__y = y if h >= 0 else y + h
        # width and height
        self.__w = abs(w)
        self.__h = abs(h)

    @property
    def x(self):
        """
        Get the rectangle's upper-left corner's x position.

        :return: X-position of the rectangle
        :rtype: float
        """
        return self.__x

    @property
    def y(self):
        """
        Get the rectangle's upper-left corner's y position.

        :return: Y-position of the rectangle
        :rtype: float
        """
        return self.__y

    @property
    def width(self):
        """
        Get the rectangle's width.

        :return: Width of the rectangle
        :rtype: float
        """
        return self.__w

    @property
    def height(self):
        """
        Get the rectangle's height.

        :return: Height of the rectangle
        :rtype: float
        """
        return self.__h

    @property
    def left(self):
        """
        Get the rectangle's left-most X-coordinate.

        :return: Left x-coordinate
        :rtype: float
        """
        return self.__x

    @property
    def right(self):
        """
        Get the rectangle's right-most X-coordinate.

        :return: Right x-coordinate
        :rtype: float
        """
        return self.__x + self.__w

    @property
    def top(self):
        """
        Get the rectangle's top-most Y-coordinate.

        :return: Top y-coordinate
        :rtype: float
        """
        return self.__y

    @property
    def bottom(self):
        """
        Get the rectangle's bottom-most Y-coordinate.

        :return: Bottom y-coordinate
        :rtype: float
        """
        return self.__y + self.__h

    def united(self, other):
        """
        Return the union of this rectangle with the given rectangle.

        :param other: Rectangle to create union with
        :type other: ScmnRectangle
        :return: Union rectangle
        :rtype: ScmnRectangle
        """
        if other is None:
            return self

        x = min(self.left, other.left)
        y = min(self.top, other.top)
        w = max(self.right, other.right) - x
        h = max(self.bottom, other.bottom) - y

        return ScmnRectangle(x, y, w, h)

    def __str__(self):
        return '({} / {}) - ({} / {})'.format(
            self.x, self.y, self.width + self.x, self.height + self.y)
