# ----------------------------------------------------------------------------
# - scmn-graph                                                               -
# - silkan python library: graph object and manipulators                     -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 amaury darsch                                    -
# - author    (c) 2014-2014 jean-paul fiorini                                -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from abc import ABCMeta, abstractmethod


# ----------------------------------------------------------------------------
# - exception section                                                        -
# ----------------------------------------------------------------------------

class ScmnGraphError(RuntimeError):
    """
    Class for different types of errors while running graphs
    """
    pass


class ScmnGraphAdditionError(ScmnGraphError):
    """
    Error occuring during the addition of nodes or edges to a graph
    """
    pass


class ScmnGraphSearchError(ScmnGraphError):
    """
    Error occuring during the search of nodes or edges in the graph
    """
    pass


class ScmnGraphExecutionError(ScmnGraphError):
    """
    Error occuring during the execution of either detectCycle or
    detectCycleNode or topologicalSort
    """
    pass


# ----------------------------------------------------------------------------
# - interface section                                                        -
# ----------------------------------------------------------------------------

class ScmnGraphP(metaclass=ABCMeta):
    @abstractmethod
    def nodes(self):
        """
        Return a list of all the nodes.
        """
        #CO COMMENT: It's often nice to use verbs to express what a function
        # does, e.g.: getNode()
        #CO TODO: We should document all input types and the return-type:
        # :return: Decription of the return value
        # :rtype: type-of-return-value
        raise NotImplementedError()

    @abstractmethod
    def neighbours(self, node):
        """
        Return all the neighbours of one node
        """
        #CO TODO: We should document all input types and the return-type:
        # :param node: Description of node parameter
        # :type node: type-of-node-parameter
        # :return: Decription of the return value
        # :rtype: type-of-return-value
        raise NotImplementedError()

    @abstractmethod
    def edges(self):
        """
        Return all edges
        """
        raise NotImplementedError()

    @abstractmethod
    def hasNode(self, node):
        """
        check if node exists in the graph
        """
        raise NotImplementedError()

    @abstractmethod
    def addNode(self, node, attr=None):
        """
        add a node to the graph
        """
        raise NotImplementedError()

    @abstractmethod
    def delNode(self, node):
        """
        delete a node from the graph, including the edges attached
        to it
        """
        raise NotImplementedError()

    @abstractmethod
    def addEdge(self, u, v, attr=None):
        """
        add an edge to the graph
        """
        raise NotImplementedError()

    @abstractmethod
    def hasEdge(self, u, v):
        """
        Check if an edge exists in the graph
        """
        raise NotImplementedError()

    @abstractmethod
    def getEdge(self, u, v):
        """
        Return the edge searched
        """
        raise NotImplementedError()

    @abstractmethod
    def delEdge(self, u, v):
        """
        delete an edge from the graph
        """
        #CO COMMENT: Why not use the full word, "deleteEdge"?
        raise NotImplementedError()


class ScmnDirectedP(metaclass=ABCMeta):
    @abstractmethod
    def incidents(self, node):
        """
        Return all nodes that are incident to the given node
        """
        #CO COMMENT: Could be called "getIncidentNodesFor"
        raise NotImplementedError()

    @abstractmethod
    def detectCycleNode(self):
        """
        detect all the nodes in the graph that cycle on themselves
        and returns a list containing those nodes
        """
        #CO COMMENT: detectCycleNodeS
        raise NotImplementedError()

    @abstractmethod
    def detectCycle(self):
        """
        detect all the cycles in the graph and returns a list
        of several lists containing the nodes forming the cycles
        """
        #CO COMMENT: detectCycleS
        raise NotImplementedError()

    @abstractmethod
    def iteration(self, node, checkList):
        """
        goes through the graph until all
        paths are exhausted
        """
        raise NotImplementedError()

    @abstractmethod
    def checkDupli(self, nlst):
        """
        check for duplicates of nodes
        in the tempList. returns True
        or False
        """
        #CO COMMENT: Why not "checkForDuplicates"?
        raise NotImplementedError()

    @abstractmethod
    def loop(self, checkList, node):
        """
        Main loop of detectCycle. returns
        either a cycle or an empty list when nothing
        is found
        """
        #CO COMMENT: Not a very descriptive name
        raise NotImplementedError()

    @abstractmethod
    def chooseNode(self, checkList):
        """
        choose a node with several branches
        for the continuation of the tests
        """
        raise NotImplementedError()

    @abstractmethod
    def topologicalSort(self):
        """
        Does a topological sort on the digraph
        and returns a list of sets
        """
        raise NotImplementedError()

    #CO COMMENT: Not abstract?
    def topologicalSortFlat(self):
        """
        works the same as topologicalSort(). It just
        returns a list with the nodes contrary
        to topologicalSort() which returns a list of sets
        with different topological levels
        """
        raise NotImplementedError()

    #CO COMMENT: Not abstract?
    def nextNodes(self, node):
        """
        Return all nodes that follow directly the given node
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class ScmnDigraphP(ScmnGraphP, ScmnDirectedP):
    """
    Protocol for Directed graphs
    """
    pass


# ----------------------------------------------------------------------------
# - Class Section                                                            -
# ----------------------------------------------------------------------------

class ScmnEdge:
    #CO COMMENT: namedtuples are also great for this
    """
    an edge is a tuple composed of two nodes first one
    is the origin, second one is the destination
    """

    def __init__(self, u, v, attr):
        """
        Creates an edge

        FIXME: Parameter names
        FIXME: Indentation

        :param __value: contains the value of the edge,\
        or the nodes linked by the edge
        :type __value: tuple
        :param __attr: contains the attributes of said edge
        :type __attr: list
        """
        self.__value = u, v
        self.__attr = attr
        # CO COMMENT: Should not use the original list,
        # which may be altered externally

    #CO COMMENT: @property might be useful here
    def getVal(self):
        return self.__value

    def getAttr(self):
        return self.__attr
        # CO COMMENT: Should not return the list,
        # but an iterator or a copy, unless mutability is explicitely desired

    def setAttr(self, attr):
        self.__attr = attr
        # CO COMMENT: Should not use the original list,
        # which may be altered externally


class ScmnDigraph(ScmnDigraphP):
    """
    This class is able to create a directed graph
    via the addition or subtraction of nodes and edges
    methods : __init__, nodes, neighbours, incidents, next
    edges, getEdge, addNode, delNode, addEdge, delEdge, detectCycleNode,
    detectCycle, iteration, checkDupli, loop, chooseNode, topologicalSort,
    topologicalSortFlat
    """

    def __init__(self):
        """
        Create a new directed graph

        FIXME: Parameter names
        FIXME: Indentation

        :param nodeNeigh: contains all nodes as keys,
        each key contains the neighbours of said node
        :type nodeNeigh: dictionary #CO COMMENT: Type descriptor for a dict is:
                         dict[keyType, valueType]
        :param nodeInci: contains all nodes as keys,
        each key contains the directly incident nodes of said node
        :type nodeInci: dictionary
        :param nodeAttr: contains all nodes as keys,
        each key contains the attributes of said node
        :type nodeAttr: dictionary
        :param edges: contains all the edges of the graph
        :type edges: list #CO COMMENT: Type descriptor for a list is:
                     list[elementType]
        :param nodeCheck: contains all nodes as keys,
        each key contains a boolean (useful for detecting cycles)
        :type nodeCheck: dictionary
        :param nodeNext: contains all nodes as keys,
        each key contains the direct nodes next to the key
        :type nodeNext: dictionary
        """
        self.__nodeNeigh = {}
        """:type: dict[keyType, valueType]"""
        # CO COMMENT: Type descriptions for fields are given as comments
        # beneath the variable declaration
        self.__nodeInci = {}
        self.__nodeAttr = {}
        self.__edges = []
        self.__nodeCheck = {}
        self.__nodeNext = {}

    def nodes(self):
        """
        Return a list of all the nodes.
        """
        # CO COMMENT: Don't need to duplicate the doc-string,
        # or you'll run the risk that they get out-of-sync
        return list(self.__nodeNeigh.keys())

    def neighbours(self, node):
        """
        Return all the neighbours of one node
        """
        return self.__nodeNeigh[node]
        # CO COMMENT: Have to be careful: Is it guaranteed that nodes are
        # properly hashable?

    def incidents(self, node):
        """
        Return all nodes that are directly incident to the given node
        """

        return self.__nodeInci[node]

    def nextNodes(self, node):
        """
        Return all nodes that follow directly the given node
        """
        return self.__nodeNext[node]

    def edges(self):
        """
        Return all edges
        """
        return self.__edges

    def hasEdge(self, u, v):
        """
        Check if a edge exists in the graph
        """
        for each in self.__edges:
            if u == each.getVal()[0] and v == each.getVal()[1]:
                return True
        return False
        # CO COMMENT: return any(u == e.getVal()[0] and v == e.getVal()[1]
        #                        for e in self.__edges)

    def getEdge(self, u, v):
        """
        Return the edge searched
        """
        edge = None

        for each in self.__edges:
            if u == each.getVal()[0] and v == each.getVal()[1]:
                edge = each
        if edge in self.__edges:  # CO COMMENT: How could this ever be false?
            return edge
        else:
            return None
        # CO COMMENT: return next((e for e in self.__edges
        #                          if u == each.getVal()[0] and
        #                             v == each.getVal()[1]), None)

    def hasNode(self, node):
        """
        Check if a node exists
        """
        if node in self.nodes():
            return True
        return False
        #CO COMMENT: return node in self.nodes()

    def addNode(self, node, attr=None):
        """
        add a node to the Digraph
        """
        if attr is None:
            attr = []

        if node not in self.__nodeNeigh:
            self.__nodeNeigh[node] = []
            self.__nodeInci[node] = []
            self.__nodeAttr[node] = attr
            self.__nodeCheck[node] = False
            self.__nodeNext[node] = []
        else:
            raise ScmnGraphAdditionError("Node %s already \
            in digraph" % node)
            #CO COMMENT: We should use format rather than %:
            # 'Node {n} is already in digraph'.format(n=node)
            # Also, ensure that node has a proper implementation of __str__

    def delNode(self, node):
        """
        delete a node from the digraph, including the edges attached
        to it
        """
        # First delete the edges
        for each in self.__edges[:]:
            u = each.getVal()[0]
            v = each.getVal()[1]
            if node == u or node == v:
                self.delEdge(u, v)
                self.delEdge(v, u)
        # Then delete the node from the tables
        if node in self.__nodeNeigh.keys():
            # CO COMMENT: Could it happen that the node is not in the dicts,
            #  i.e. multiple deletion?
            del self.__nodeNeigh[node]
            del self.__nodeInci[node]
            del self.__nodeAttr[node]
            del self.__nodeCheck[node]
            del self.__nodeNext[node]

    def addEdge(self, u, v, attr=None):
        """
        add an edge to the Digraph
        """
        if attr is None:
            attr = []

        for n in [u, v]:
            if n not in self.__nodeNeigh:
                raise ScmnGraphAdditionError("%s is missing from the\
                __nodeNeigh table" % n)  # CO COMMENT: String format
            if n not in self.__nodeInci:
                raise ScmnGraphAdditionError("%s is missing from the\
                __nodeInci table" % n)  # CO COMMENT: String format

        if v in self.__nodeNeigh[u] and u in self.__nodeNeigh[v] \
                and u in self.__nodeInci[v]:
            # CO COMMENT: String format
            raise ScmnGraphAdditionError(
                "Edge (%s, %s) already in digraph" % (u, v))
        else:
            self.__edges.append(ScmnEdge(u, v, attr))
            self.__nodeNeigh[u].append(v)
            self.__nodeNeigh[v].append(u)
            self.__nodeInci[v].append(u)
            self.__nodeNext[u].append(v)

    def delEdge(self, u, v):
        """
        delete an edge from the graph
        """
        edge = None
        for each in self.__edges:
            if u == each.getVal()[0] and v == each.getVal()[1]:
                edge = each
                #CO COMMENT: break?

        if edge in self.__edges:
            # CO COMMENT: Why the separate check? Why not run this code right
            # when the if-statement above is true?
            self.__edges.remove(edge)
            self.__nodeNeigh[u].remove(v)
            self.__nodeNeigh[v].remove(u)
            self.__nodeInci[v].remove(u)
            self.__nodeNext[u].remove(v)

    def detectCycleNode(self):
        """
        detect all the nodes in the graph that cycle on themselves
        and returns a list containing those nodes
        """
        result = []
        for node in self.__nodeInci.keys():
            if node in self.__nodeInci[node]:
                result.append(node)

        return result
        # CO COMMENT: return [n for n in self.__nodeInci.keys()
        #                     if n in self.__nodeInci[n]]

    def detectCycle(self):
        """
        detect all the cycles in the graph and returns a list
        of several lists containing the nodes forming the cycles
        """
        checkList = []
        resultCheck = []
        result = []

        # creating a checkList containing all nodes with at least 2 branches
        # or nodes at the extremity of the graph

        for each in self.__nodeNeigh.keys():
            if len(self.__nodeInci[each]) >= 2 or \
                    (len(self.__nodeInci[each]) == 1 and
                             len(self.__nodeNeigh[each]) == 1):
                checkList.append(each)

        #CO COMMENT: checkList = [n for n in self.__nodeNeigh.keys() if ...]

        # Main loop

        for num, _ in enumerate(self.__nodeNeigh.keys()):
            #CO COMMENT: for _ in self.__nodeNeigh:
            # choosing a node from the checkList
            # which have not exhausted all his paths
            nodeChosen = self.chooseNode(checkList)
            # if there is no node without
            # exhausted paths we return the result
            if not nodeChosen:
                # We reset the values of check
                # for the node incidents to False for
                # following detections
                for node in self.__nodeNeigh.keys():
                    # CO COMMENT: keys() is superflous
                    self.__nodeCheck[node] = False
                return result
            else:
                # the function loop will either return an empty
                # list if no cycles are found or a list containing
                # a cycle, we evaluate if its set is already in resultCheck
                # so we don't have duplicates
                i = self.loop(checkList, nodeChosen)
                if not i:
                    continue
                if set(i) not in resultCheck:
                    resultCheck.append(set(i))
                    # Reversing the list to get the correct order for nodes
                    i.reverse()
                    result.append(i)

    def iteration(self, node, checkList):
        """
        goes through the graph until all
        paths are exhausted
        """
        # First condition block before the loop
        if node in checkList:
            nodeChosen = False
            # We're looking for a not already checked path
            # with the use of check with incident nodes
            # if there is one we take the path and then
            # set it to True (meaning already checked)
            # if there is none we break the loop by yielding
            # an empty list that will be treated by other
            # methods above
            for nodeInci in self.__nodeInci[node]:
                if not self.__nodeCheck[nodeInci]:
                    self.__nodeCheck[nodeInci] = True
                    nodeChosen = nodeInci
                    temp = nodeChosen
                    break
            if not nodeChosen:
                yield []
        else:
            # in case of a "normal" node we just check
            # that there is a node incident to it
            # and store the node inside temp for later use
            temp2 = self.__nodeInci[node]
            if not temp2:
                yield []
            temp = temp2[0]

        for num, _ in enumerate(self.__nodeNeigh.keys()):
            yield temp
            # second block of test equivalent in use
            # as the first one
            if temp in checkList:
                nodeChosen = False
                for nodeInci in self.__nodeInci[temp]:
                    if not self.__nodeCheck[nodeInci]:
                        self.__nodeCheck[nodeInci] = True
                        nodeChosen = nodeInci
                        temp = nodeChosen
                        break
                if not nodeChosen:
                    yield []
            else:
                temp2 = self.__nodeInci[temp]
                if not temp2:
                    yield []
                temp = temp2[0]

    def checkDupli(self, nlst):
        """
        check for duplicates of nodes
        in the tempList. returns True
        or False
        """
        lastItem = nlst[len(nlst) - 1]
        if nlst.count(lastItem) >= 2:
            return True
        else:
            return False
        #CO COMMENT: return list.count(lastItem) >= 2

    def loop(self, checkList, node):
        """
        Main loop of detectCycle. returns
        either a cycle or an empty list when nothing
        is found
        """
        # We use a tempList to store in memory
        # the nodes we go through
        # we create and use an iterator with the iteration method
        # if the iterator returns an empty list (if not i: case)
        # loop will return an empty list because there is nothing to
        # return, all paths have been exhausted.
        # In the "normal" case we just add the node we've read
        # into the tempList and check if there is a duplicate
        # of the last node added (if there is one it forms a cycle)
        # which we return as a slice of tempList
        tempList = [node]
        itera = self.iteration(node, checkList)
        for i in itera:
            if not i:
                return []
            else:
                tempList.append(i)
                if self.checkDupli(tempList):
                    return tempList[tempList.index(i): (len(tempList) - 1)]
                else:
                    continue

    def chooseNode(self, checkList):
        """
        choose a node with several branches for the continuation of the tests
        """
        for node in checkList:
            for nodeInci in self.__nodeInci[node]:
                if (nodeInci in self.__nodeInci) and \
                   (not self.__nodeCheck[nodeInci]): return node
        return None

    def topologicalSort(self):
        """
        Does a topological sort on the digraph
        and returns a list of sets
        """
        checkList = []
        result = []

        # First checking if the graph given is acyclic
        if len(self.detectCycleNode()) != 0:
            # CO COMMENT: if detectCycleNode returns a list, you can simply
            # check: if not self.detectCycleNode():
            raise ScmnGraphExecutionError("This graph possess cycles\
            so a topological sorting is impossible")
            # CO COMMENT: String format
        else:
            if len(self.detectCycle()) != 0:
                raise ScmnGraphExecutionError("This graph possess \
                cycles so a topological sorting is impossible")
                # CO COMMENT: String format

        # Creating the first group for the topological sort
        for each in self.__nodeNeigh.keys():
            if len(self.__nodeInci[each]) == 0:
                checkList.append(each)

        # CO COMMENT: checkList = [n for n in self.__nodeNeigh
        #                          if not self.__nodeInci.get(n)]

        # Adding the first group to the result
        result.append(checkList)
        # Adding a group each iteration of num
        # which would be equivalent to the maximum
        # number of topological levels
        for num, _ in enumerate(self.__nodeNeigh.keys()):
            #CO COMMENT: for num, _ in enumerate(self.__nodeNeigh):
            result.append([])
            # num is equivalent to the precedent group (n-1)
            # before doing anything checking if it is empty
            if len(result[num]) != 0:  # CO COMMENT: if result[num]
                for node in result[num]:
                    # running through the nodes of the precedent group
                    # and checking if they have a nodeNext
                    if len(self.__nodeNext[node]) != 0:
                        # CO COMMENT: if self.__nodeNext[node]
                        # if they have, running through for different
                        # nodeNext and checking if they
                        # have all their incident
                        # nodes in precedent topological groups
                        for n in self.__nodeNext[node]:
                            check = True
                            for i in self.__nodeInci[n]:
                                found = False
                                for lst in result[:-1]:
                                    if i in lst:
                                        found = True
                                        break
                                #: CO COMMENT: Use 'else' branch
                                if not found:
                                    check = False
                                    break
                            if check:
                                # if they have we add the node
                                # to the next topological group
                                result[num + 1].append(n)

        # removing the empty topological groups
        for num in range(result.count([])):
            result.remove([])
        #CO COMMENT: result = [x for x in result if x]

        # creating sets to remove duplicates of nodes inside
        # the topological groups
        for num in range(len(result)):
            result[num] = set(result[num])
        #CO COMMENT: result = [set(x) for x in result]
        # EVEN BETTER, COMBINE STEPS: result = [set(x) for x in result if x]
        return result

    def topologicalSortFlat(self):
        """
        works the same as topologicalSort(). It just
        returns a list with the nodes contrary
        to topologicalSort() which returns a list of sets
        with different topological levels
        """
        result = []
        tempList = self.topologicalSort()
        for num in range(len(tempList)):
            #CO COMMENT: for l in tempList: for i in l: result.append(i)
            for i in tempList[num]:
                result.append(i)

        # EVEN BETTER: result = [i for l in tempList for i in l]

        return result
