# ----------------------------------------------------------------------------
# - scmn-log                                                                 -
# - silkan python library: log                                               -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod
from collections import deque
from datetime import datetime
from threading import RLock
from itertools import count, takewhile, dropwhile


# ----------------------------------------------------------------------------
# - message section                                                          -
# ----------------------------------------------------------------------------

# the message types
SCMN_LOG_INFO, SCMN_LOG_WARN, SCMN_LOG_ERROR = range(1, 4)

SCMN_LOG_CATNAMES = dict({
    SCMN_LOG_INFO:'info',
    SCMN_LOG_WARN:'warning',
    SCMN_LOG_ERROR:'error'
})


SCMN_LOG_CATNAMES_MAXLEN = max(
    [len(n) for n in SCMN_LOG_CATNAMES.values()])


class ScmnLogMessage:
    """
    The ScmnLogMessage class is a base class which hold a message to be
    logged. The message is define by type, and an optional level.
    """

    #: Global message counter
    __COUNTER = count(1)

    def __init__(self, mtxt, mcmp, mcat=None, mlvl=None, trce=None):
        """
        Create a new message by type and level

        :param mtxt: the message to log
        :type mtxt: str
        :param mcmp: component the message originated from
        :type mcmp: str
        :param mcat: the message type
        :type mcat: int
        :param mlvl: the message level
        :type mlvl: int
        :param trce: Stack trace, if any
        :type trce: str
        """
        #: Text of the message
        self.__text = mtxt
        #: Originating component
        self.__component = mcmp
        #: Message category/classification
        self.__category = mcat if mcat is not None else SCMN_LOG_INFO
        #: Message level for info messages
        self.__level = mlvl if mlvl is not None else 0
        #: Stack trace
        self.__trace = trce

        #: Time stamp
        self.__timestamp = datetime.now()
        #: Unique counter value
        self.__counter = next(ScmnLogMessage.__COUNTER)

    @property
    def category(self):
        """
        Get the message type
        """
        return self.__category

    @property
    def level(self):
        """
        Get the message level
        """
        return self.__level

    @property
    def text(self):
        """
        Get the message text
        """
        return self.__text

    @property
    def component(self):
        return self.__component

    @property
    def trace(self):
        return self.__trace

    @property
    def counter(self):
        """
        Get the message's counter value.

        :return: Counter value
        :rtype: int
        """
        return self.__counter

    def formatted(self):
        """
        Format a message as a string.

        :return: Formatted message
        :rtype; str
        """
        # noinspection PyArgumentList
        return '[%7s] %s - %s: %s' % (
            SCMN_LOG_CATNAMES.get(self.__category, ''),
            self.__timestamp.strftime('%x %X'),
            self.__component,
            self.__text
        )


# ----------------------------------------------------------------------------
# - logger section                                                           -
# ----------------------------------------------------------------------------

class ScmnMessageTargetP(metaclass=ABCMeta):
    @abstractmethod
    def addMessage(self, mtxt, mcmp=None, mcat=None, mlvl=None, trce=None):
        """
        Add a new message to the message target.

        TODO: Remove parameter inter-dependencies (mcat-mlvl)!

        :param mtxt: Message text
        :type mtxt: str
        :param mcmp: Component issuing the message
        :type mcmp: str
        :param mcat: Message classification
        :type mcat: int
        :param mlvl: Message level (only if info message)
        :type mlvl: int
        :param trce: Stack trace, if any
        :type trce: str
        """
        raise NotImplementedError()


class ScmnMessageTargetMux(ScmnMessageTargetP):
    """
    Log message multiplexer.
    """

    def __init__(self, *targets):
        """
        :param targets: Log targets
        :type targets: Iterable[ScmnMessageTargetP]
        """
        self.__targets = targets

    def addMessage(self, mtxt, mcmp=None, mcat=None, mlvl=None, trce=None):
        for target in self.__targets:
            target.addMessage(mtxt, mcmp, mcat, mlvl, trce)


class ScmnMessageSourceP(metaclass=ABCMeta):
    @abstractmethod
    def messageCount(self):
        """
        Get the number of messages the source can currently provide.

        :return: Number of messages
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def lastMessage(self):
        """
        Get the last message added to the source.

        :return: The last message added, if any
        :rtype: ScmnLogMessage
        """
        raise NotImplementedError()

    @abstractmethod
    def lastMessageCounter(self):
        """
        Get the counter value of the most recent message, or 0 if no messages
        are available.

        :return: Counter value of most recent message
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def messages(self, category=None, minLevel=None, since=None):
        """
        Get all messages currently held by the source.

        :param category: Minimum category level
        :type category: int
        :param minLevel: Minimum message level within the category
        :type minLevel: int
        :param since: Minimum counter value
        :type since: int
        :return: All messages on file
        :rtype: Iterable[ScmnLoggerMessage]
        """
        raise NotImplementedError()

    def addMessageListener(self, listener):
        """
        Add a listener to be notified when messages are added.

        :param listener: Listener to be added
        :type listener: ScmnLogListener
        """
        raise NotImplementedError()

    def removeMessageListener(self, listener):
        """
        Remove a listener.

        :param listener: Listener to be removed
        :type listener: ScmnLogListener
        """
        raise NotImplementedError()


class ScmnLogListener(metaclass=ABCMeta):
    @abstractmethod
    def messageAdded(self, msg):
        """
        Callback function used to notify the listener about new messages.

        :param msg: Message that has been added
        :type msg: ScmnLogMessage
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class ScmnLoggerP(ScmnMessageTargetP, ScmnMessageSourceP):
    """
    Protocol for loggers that serve both as message target and message source.
    """
    pass


class ScmnLogger(ScmnLoggerP):
    """
    The ScmnLogger class a system logger designed to hold a series of
    message in a circular fifo (aka ring buffer).
    """

    #: Default buffer size (number of messages)
    DEFAULT_BUFFER_SIZE = 10000

    def __init__(self, logSize=None, logToStdout=False):
        """
        :param logSize: Number of log entries to store
        :type logSize: int
        :param logToStdout: Log to std out?
        :type logToStdout: bool
        """

        #: Maximum number of messages
        self.__size = logSize if logSize else self.DEFAULT_BUFFER_SIZE
        #: Log to std-out?
        self.__logStdout = logToStdout

        #: FIFO queue for messages
        self.__fifo = deque([], self.__size)
        #: Lock for adding messages
        self.__lock = RLock()
        #: Listeners to be notified when messages are added
        self.__listeners = []

    def addMessageListener(self, listener):
        self.__listeners.append(listener)

    def removeMessageListener(self, listener):
        self.__listeners.remove(listener)

    def addMessage(self, mtxt, mcmp=None, mcat=None, mlvl=None, trce=None):
        if self._isPrintTrace() and (mcat == SCMN_LOG_ERROR or trce):
            print('{c}: {m}'.format(c=mcmp, m=mtxt))
            if trce:
                print(trce)

        with self.__lock:
            msg = ScmnLogMessage(mtxt=mtxt, mcmp=mcmp, mcat=mcat, mlvl=mlvl,
                                   trce=trce)
            self.__fifo.append(msg)

        for listener in self.__listeners:
            listener.messageAdded(msg)

    def messageCount(self):
        return len(self.__fifo)

    def lastMessage(self):
        return self.__fifo[-1] if self.__fifo else None

    def lastMessageCounter(self):
        return self.__fifo[-1].counter if self.__fifo else 0

    def messages(self, category=None, minLevel=None, since=None):
        if category is None and minLevel is None and since is None:
            return list(self.__fifo)

        l = minLevel or 0
        c = category or 0

        msgs = self.__fifo

        if since:
            if self.lastMessageCounter() - since > 100:
                msgs = dropwhile(lambda m: m.counter <= since, msgs)
            else:
                msgs = reversed(list(takewhile(lambda m: m.counter > since,
                                               reversed(msgs))))

        return list([m for m in msgs
                     if ((m.category == c and m.level >= l) or
                         m.category > c)])

    def _isPrintTrace(self):
        return bool(self.__logStdout)


# ----------------------------------------------------------------------------
# - logger helpers section                                                   -
# ----------------------------------------------------------------------------

def scmn_try_log(logr, mtxt, mcmp=None, mcat=None, mlvl=None, trce=None):
    """
    Try to log the message if the logger exist, otherwise return.

    :param logr: Logger
    :type logr: ScmnMessageTargetP
    :param mtxt: the message to log
    :type mtxt: str
    :param mcmp: component the message originated from
    :type mcmp: str
    :param mcat: the message type
    :type mcat: int
    :param mlvl: the message level
    :type mlvl: int
    :param trce: Stack trace, if any
    :type trce: str
    """
    if logr is not None:
        logr.addMessage(mtxt=mtxt, mcmp=mcmp, mcat=mcat, mlvl=mlvl, trce=trce)
    return
