# ----------------------------------------------------------------------------
# - scmn-mime-type                                                           -
# - silkan python library: mime type base classes                            -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2013-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       scmn_mime_type
   :platform:     Unix, Windows
   :synopsis:     Silkan Python Library - mime type base classes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
import re
from abc import abstractmethod, ABCMeta
from collections import OrderedDict
from xml.etree import ElementTree

from .scmn_file_path import scmn_path_extension, scmn_path_base
from .scmn_file_path import scmn_isfile, scmn_filesize
from .scmn_file_path import scmn_expshenv


# Generic payload types
SCMN_PAYLOAD_NONE = 0
SCMN_PAYLOAD_TEXT = 1
SCMN_PAYLOAD_XML = 2


class ScmnMimeTypeP(metaclass=ABCMeta):
    """
    Protocol for MIME types.

    TODO: Introduce protocol for file types (MIME-type(s) + extension(s))?
    """
    @property
    @abstractmethod
    def mimeType(self):
        """
        Get the MIME type identifier.

        :return: MIME type identifier
        :rtype: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def description(self):
        """
        Get the description for the MIME type.

        :return: Description for the MIME type
        :rtype: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def extension(self):
        """
        Get the MIME type's primary file extension.

        :return: Primary file extension for the MIME type
        :rtype: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def auxExtensions(self):
        """
        Get the MIME type's auxiliary file extensions.

        :return: Auxiliary file extensions for the MIME type
        :rtype: Iterable[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def isExtension(self, path):
        """
        Whether the given file matches the MIME type's extensions.

        :param path: File path to check for extension match
        :type path: str
        :return: True iff the given path matches the MIME type's extensions
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def isFileMimeType(self, path):
        """
        Whether the file at the given path is of this MIME type.

        :param path: File to test for MIME-type
        :type path: str
        :return: True iff the file is of this MIME type
        :rtype: bool
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def payloadTypes(self):
        """
        Get the payload types that can be generated from content of this
        MIME type.

        :return: Payload types that can be generated
        :rtype: int
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def readers(self):
        """
        Get the readers registered with this MIME type.

        :return: Readers registered with this MIME type
        :rtype: Iterable[WcmnReader]
        """
        raise NotImplementedError()

    def __str__(self):
        return self.mimeType

    def __repr__(self):
        return self.mimeType


# noinspection PyAbstractClass
class ScmnMimeTypeBase(ScmnMimeTypeP):
    """
    Base class for MIME type implementation.
    """

    def __init__(self, mimeType, description):
        """
        :param mimeType: MIME type identifier
        :type mimeType: str
        :param description: Description for the MIME type
        :type description: str
        """
        self.__mimeType = mimeType
        self.__descr = description

        #: Relevance-level
        self.__relevance = mimeType.count('+')

        self.__readers = []
        self.__types = SCMN_PAYLOAD_NONE

    @property
    def _relevance(self):
        return self.__relevance

    @property
    def mimeType(self):
        return self.__mimeType

    @property
    def description(self):
        return self.__descr

    @property
    def payloadTypes(self):
        return self.__types

    @property
    def readers(self):
        return iter(self.__readers)

    def addReader(self, reader):
        """
        Register a reader with this MIME type.

        :param reader: Reader to register
        :type reader: WcmnReader
        """
        self.__readers.append(reader)
        self.__types |= reader.payloadTypesRead

    def isExtension(self, path):
        fext = scmn_path_extension(path)
        mext = self.extension
        auxs = self.auxExtensions

        return fext and mext and (fext == mext or fext in auxs)


class ScmnMimeType(ScmnMimeTypeBase):
    """
    Default MIME type implementation.
    """

    def __init__(self, mimeType, description, extensions=None, pattern=None):
        """
        :param mimeType: MIME type identifier
        :type mimeType: str
        :param description: Description for the MIME type
        :type description: str
        :param extensions: File extensions (primary and auxiliary)
        :type extensions: list[str]
        :param pattern: File name pattern
        :type pattern: str
        """
        super().__init__(mimeType, description)

        self.__ext, *self.__auxExt = extensions if extensions else [None]
        self.__pattern = re.compile(pattern, re.I) if pattern else None

    @property
    def extension(self):
        return self.__ext

    @property
    def auxExtensions(self):
        return self.__auxExt

    def isFileMimeType(self, path):
        if self.isExtension(path):
            return True

        pattern = self.__pattern
        base = scmn_path_base(path)
        return bool(base and pattern and pattern.match(base))


class ScmnXmlMimeType(ScmnMimeTypeBase):
    XML_EXT = 'xml'

    def __init__(self, mimeType, description, topNode,
                 extension=None, auxExtensions=None):
        """
        :param mimeType: MIME type identifier
        :type mimeType: str
        :param description: Description for the MIME type
        :type description: str
        :param topNode: Top-level node name
        :type topNode: str
        :param auxExtensions: Auxiliary file extensions
        :type auxExtensions: Iterable[str]
        """
        super().__init__(mimeType, description)

        self.__extension = extension or self.XML_EXT

        self.__auxExt = list(auxExtensions) if auxExtensions else []
        if self.XML_EXT not in self.__extension[-3:] and \
                self.XML_EXT not in self.__auxExt:
            self.__auxExt.append(self.XML_EXT)
        self.__topNode = topNode.lower()

    @property
    def extension(self):
        return self.__extension

    @property
    def auxExtensions(self):
        return self.__auxExt

    def isFileMimeType(self, path):
        if not self.isExtension(path):
            return False

        if not self.__topNode:
            return True

        if not scmn_isfile(path) or scmn_filesize(path) == 0:
            return True

        try:
            with open(path, 'r') as file:
                for _, elem in ElementTree.iterparse(file, events=('start',)):
                    etag = elem.tag.lower()
                    if '}' in etag:
                        return etag.split('}')[-1] == self.__topNode
                    else:
                        return etag == self.__topNode
        except:
            pass
        return False


global SCMN_MIMES
SCMN_MIMES = OrderedDict()
""":type: dict[str, ScmnMimeTypeBase]"""

global SCMN_MIME_EXTS
SCMN_MIME_EXTS = {}
""":type: dict[str, list[ScmnMimeTypeP]]"""

#: Global registry for MIME type co-variance
global SCMN_COVARIANT_MIMES
SCMN_COVARIANT_MIMES = {}
""":type: dict[str, set[str]]"""

#: Global registry for MIME type contra-variance
global SCMN_CONTRAVARIANT_MIMES
SCMN_CONTRAVARIANT_MIMES = {}
""":type: dict[str, set[str]]"""


def scmn_register_mime_type(mimeType, coVariants=None, contraVariants=None):
    """
    Register a MIME type with the global MIME-type registry.

    :param mimeType: MIME-type to register
    :type mimeType: ScmnMimeTypeP
    :param coVariants: Co-variant MIME types
    :type coVariants: Iterable[str]
    :param contraVariants: Contra-variant MIME types
    :type contraVariants: Iterable[str]
    """
    SCMN_MIMES[mimeType.mimeType] = mimeType

    SCMN_MIME_EXTS.setdefault(mimeType.extension, []).append(mimeType)
    for ext in mimeType.auxExtensions:
        SCMN_MIME_EXTS.setdefault(ext, []).append(mimeType)

    cos = SCMN_COVARIANT_MIMES
    contras = SCMN_CONTRAVARIANT_MIMES

    if coVariants:
        for co in coVariants:
            cos.setdefault(mimeType.mimeType, set()).add(co)
            contras.setdefault(co, set()).add(mimeType.mimeType)

    if contraVariants:
        for co in contraVariants:
            contras.setdefault(mimeType.mimeType, set()).add(co)
            cos.setdefault(co, set()).add(mimeType.mimeType)


def scmn_add_reader(mime, reader):
    """
    Add a reader to the MIME-type with the given identifier.

    :param mime: MIME-type identifier
    :type mime: str
    :param reader: File reader to add
    :type reader: WcmnReader
    """
    mimeType = SCMN_MIMES.get(mime)
    if mimeType:
        mimeType.addReader(reader)


def scmn_mime_type(mime):
    """
    Get the MIME-type for the given identifier.

    :param mime: MIME-type identifier
    :type mime: str
    :return: MIME-type for the given identifier
    :rtype: ScmnMimeTypeP
    """
    return SCMN_MIMES.get(mime)


def scmn_mime_types(path):
    """
    Get the possible MIME-types for the file at the given path.

    The returned MIME-types are sorted by descending relevance.

    :param path: File to get possible MIME types for
    :type path: str
    :return: Possible MIME types for the given file
    :rtype: list[ScmnMimeTypeP]
    """
    if not path:
        return []
    path = scmn_expshenv(path)
    mimes = [mime for mime in SCMN_MIMES.values()
             if mime.isFileMimeType(path)]
    # noinspection PyProtectedMember
    mimes.sort(key=lambda x: x._relevance, reverse=True)
    return mimes


def scmn_contravariant_mime_types(mimes):
    """
    Get the possible contra-variant MIME-types for the file at the given path.

    :param mimes: Mime-types to get possible contra-variant MIME types for
    :type mimes: Iterable[ScmnMimeTypeP]
    :return: Possible contra-variant MIME types for the given file
    :rtype: set[ScmnMimeTypeP]
    """
    return set([SCMN_MIMES[mime] for m in mimes
                for mime in SCMN_CONTRAVARIANT_MIMES.get(m.mimeType) or []])


def scmn_covariant_mime_types(mimes):
    """
    Get the possible co-variant MIME-types for the file at the given path.

    :param mimes: Mime-types to get possible co-variant MIME types for
    :type mimes: Iterable[ScmnMimeTypeP]
    :return: Possible co-variant MIME types for the given file
    :rtype: set[ScmnMimeTypeP]
    """
    return set([SCMN_MIMES[mime] for m in mimes
                for mime in SCMN_COVARIANT_MIMES.get(m.mimeType) or []])


def scmn_payload_mime_types(payloadTypes):
    """
    Get all MIME types that can produce the given payload type.

    :param payloadTypes: Requested payload type(s)
    :type payloadTypes: int
    :return: MIME types that can produce the given payload type
    :rtype: set[ScmnMimeTypeP]
    """
    return set([m for m in SCMN_MIMES.values()
                if m.payloadTypes & payloadTypes])


# Register 'Unknown' MIME type
SCMN_MIME_ID_UNKNOWN = 'application/unknown'
SCMN_MIME_TYPE_UNKNOWN = ScmnMimeType(
    SCMN_MIME_ID_UNKNOWN, 'Unkwown', extensions=[])

scmn_register_mime_type(SCMN_MIME_TYPE_UNKNOWN)


# Register plain-text MIME type
SCMN_MIME_ID_TEXT = 'text/plain'
SCMN_MIME_TYPE_TEXT = ScmnMimeType(
    SCMN_MIME_ID_TEXT, 'Plain text', extensions=['txt'])

scmn_register_mime_type(SCMN_MIME_TYPE_TEXT)


# Register XML MIME type
SCMN_MIME_ID_XML = 'application/xml'
SCMN_MIME_TYPE_XML = ScmnMimeType(
    SCMN_MIME_ID_XML, 'XML Document', extensions=['xml'])

scmn_register_mime_type(SCMN_MIME_TYPE_XML, coVariants=[SCMN_MIME_ID_TEXT])


# Register C-Code MIME type
SCMN_MIME_ID_C = 'text/x-c'
SCMN_MIME_TYPE_C = ScmnMimeType(
    SCMN_MIME_ID_C, 'C Code', extensions=['c', 'h'])

scmn_register_mime_type(SCMN_MIME_TYPE_C, coVariants=[SCMN_MIME_ID_TEXT])


# Register C++-Code MIME type
SCMN_MIME_ID_CPP = 'text/x-c++'
SCMN_MIME_TYPE_CPP = ScmnMimeType(
    SCMN_MIME_ID_CPP, 'C++ Code', extensions=['cpp', 'cxx', 'hpp', 'hxx'])

scmn_register_mime_type(SCMN_MIME_TYPE_CPP, coVariants=[SCMN_MIME_ID_TEXT])

# Register Makefile MIME type
SCMN_MIME_ID_MAKEFILE = 'text/x-makefile'
SCMN_MIME_TYPE_MAKEFILE = ScmnMimeType(
    SCMN_MIME_ID_MAKEFILE, 'Makefile',
    extensions=['mak'], pattern='^Makefile$')

scmn_register_mime_type(
    SCMN_MIME_TYPE_MAKEFILE, coVariants=[SCMN_MIME_ID_TEXT])
