# ----------------------------------------------------------------------------
# - scmn-process                                                             -
# - silkan python library: process handling and command execution            -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       scmn_process
   :platform:     Unix, Darwin, Windows
   :synopsis:     Silkan Python Library - process handling / command execution

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import ABCMeta, abstractmethod
from threading import Thread
import platform
import subprocess
import select


def scmn_execute(command, strm=None, env=None):
    """
    Execute a given command.

    :param command: Command to execute
    :type command: list[str]
    :param strm: Output/error stream handler
    :type strm: ScmnProcStreamHandlerP
    :param env: Environment
    :type env: dict[str, str]
    :return: Return code as returned by command
    :rtype: int
    """
    # noinspection PyUnusedLocal
    rcode = 0
    try:
        if strm is None:
            proc = subprocess.Popen(command, env=env)
            proc.wait()
            rcode = proc.returncode
        else:
            strm.handleControlMsg('Executing ' + ' '.join(command),
                                  ScmnProcStreamHandlerP.CTRL_START)
            if scmn_has_pipe_select():
                rcode = _scmn_select_exec(command, strm, env)
            else:
                rcode = _scmn_threaded_exec(command, strm, env)

            if rcode == 0:
                strm.handleControlMsg(command[0] + ' finished successfully',
                                      ScmnProcStreamHandlerP.CTRL_FINISHED)
            else:
                strm.handleControlMsg(
                    '{c} exited with code {r}'.format(c=command[0], r=rcode),
                    ScmnProcStreamHandlerP.CTRL_FAILED)
    except Exception as e:
        strm.handleControlMsg(
            'Failed to execute {c}: {e}'.format(c=command[0], e=e),
            ScmnProcStreamHandlerP.CTRL_FAILED)
        raise ScmnProcLaunchError(command, exception=e)

    if rcode != 0:
        raise ScmnProcExecError(command, rcode)


def _scmn_threaded_exec(command, strm, env=None):
    def startThread(stream, handler):
        def run():
            while True:
                data = stream.readline()
                if data:
                    handler(data.decode('utf-8').rstrip())
                elif proc.returncode is not None:
                    break

        t = Thread(target=run)
        t.setDaemon(True)
        t.start()
        return t

    PIPE = subprocess.PIPE
    proc = subprocess.Popen(command, env=env,
                            bufsize=1, stdout=PIPE, stderr=PIPE)
    tout = startThread(proc.stdout, strm.handleOutput)
    terr = startThread(proc.stderr, strm.handleError)

    proc.wait()
    tout.join()
    terr.join()

    return proc.returncode


def _scmn_select_exec(command, strm, env=None):
    PIPE = subprocess.PIPE
    proc = subprocess.Popen(command, env=env,
                            bufsize=1, stdout=PIPE, stderr=PIPE)
    proc.stdout.handler = strm.handleOutput
    proc.stderr.handler = strm.handleError
    rlst = [proc.stdout, proc.stderr]
    while True:
        slctd, *_ = select.select(rlst, [], [])
        for chnl in slctd:
            line = chnl.readline().decode('utf-8').rstrip()
            if line:
                chnl.handler(line)
        if proc.poll() is not None:
            break

    for chnl in rlst:
        for line in chnl.readlines():
            chnl.handler(line.decode('utf-8').rstrip())

    proc.wait()
    return proc.returncode


class ScmnProcStreamHandlerP(metaclass=ABCMeta):
    """
    Protocol for stream output/error handlers.
    """

    #: Message type for 'execution started' messages
    CTRL_START = 1
    #: Message type for 'execution finished' messages
    CTRL_FINISHED = 2
    #: Message type for 'execution failed' messages
    CTRL_FAILED = 4

    @abstractmethod
    def handleControlMsg(self, text, msgType):
        """
        Handle a control message.

        :param text: Message text to handle
        :type text: str
        :param msgType: Message type
        :type msgType: int
        """
        raise NotImplementedError()

    @abstractmethod
    def handleOutput(self, text):
        """
        Handle output text.

        :param text: Output text to handle
        :type text: str
        """
        raise NotImplementedError()

    @abstractmethod
    def handleError(self, text):
        """
        Handle error text.

        :param text: Error text to handle
        :type text: str
        """
        raise NotImplementedError()


class ScmnProcAccumStreamHandler(ScmnProcStreamHandlerP):
    """
    Stream output/error handler that accumulates all output/errors.
    """

    def __init__(self):
        #: Output accumulator
        self.__output = ''
        #: Error accumulator
        self.__errors = ''

    @property
    def output(self):
        """
        Get the accumulated output.

        :return: Accumulated output
        :rtype: str
        """
        return self.__output

    @property
    def error(self):
        """
        Get the accumulated errors.

        :return: Accumulated errors
        :rtype: str
        """
        return self.__errors

    def handleControlMsg(self, text, msgType):
        pass

    def handleOutput(self, text):
        self.__output += text

    def handleError(self, text):
        self.__errors += text


class ScmnProcError(Exception):
    def __init__(self, command):
        self.__command = command

    @property
    def command(self):
        if isinstance(self.__command, str):
            return self.__command
        else:
            return ' '.join(self.__command)


class ScmnProcExecError(ScmnProcError):
    def __init__(self, command, retcode, errors=None):
        super().__init__(command)
        self.__retcode = retcode
        self.__errors = errors

    @property
    def retcode(self):
        return self.__retcode

    @property
    def errors(self):
        return self.__errors or []

    def __str__(self):
        return 'Command returned #{r} ({c})'.format(
            r=self.retcode, c=self.command)


class ScmnProcLaunchError(ScmnProcError):
    def __init__(self, command, exception=None):
        super().__init__(command)
        self.__exception = exception

    @property
    def exception(self):
        return self.__exception

    def __str__(self):
        return 'Command failed to launch ({c}): {e}'.format(
            c=self.command, e=self.exception)


def scmn_has_pipe_select():
    """
    Whether the host platform supports select() on pipes.

    :return: True iff select() on pipes is supported
    :rtype: bool
    """
    pf = platform.system().lower()
    return pf == 'darwin' or pf == 'linux'
