# ----------------------------------------------------------------------------
# - scmn-trigonometry                                                        -
# - silkan python library: trigonometric functions                           -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       scmn_trigonometry
   :platform:     Unix, Darwin, Windows
   :synopsis:     Silkan Python Library - trigonometric functions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
import math


#: Full circle in radians
SCMN_FULL_CIRCLE_RAD = 2.0 * math.pi

#: Half circle in radians
SCMN_HALF_CIRCLE_RAD = math.pi

#: Quarter circle in radians
SCMN_QUARTER_CIRCLE_RAD = math.pi / 2.0


def _len(a, b):
    """
    Length of line between (0, 0) and (a, b).

    :type a: float
    :type b: float
    :rtype: float
    """
    return math.sqrt(pow(a, 2.0) + pow(b, 2.0))


def scmn_getSlope(xdiff, ydiff):
    """
    Get the slope for the given slope components.

    :param xdiff: X component
    :type xdiff: float
    :param ydiff: Y component
    :type ydiff: float
    :return: Slope
    :rtype: float
    """
    if xdiff != 0:
        slope = math.atan(ydiff / xdiff)
        if xdiff < 0:
            slope -= SCMN_HALF_CIRCLE_RAD
    elif ydiff < 0:
        slope = -SCMN_QUARTER_CIRCLE_RAD
    else:
        slope = SCMN_QUARTER_CIRCLE_RAD
    return slope


def scmn_pointLineDist(x, y, sx, sy, ex, ey):
    """
    Get the distance between the point (x, y) and the line through (sx, sy)
    and (ex, ey).

    :type x: float
    :type y: float
    :type sx: float
    :type sy: float
    :type ex: float
    :type ey: float
    :rtype: float
    """
    fx, fy = scmn_perpendicularFoot(x, y, sx, sy, ex, ey)
    sign = 1 if ((ex - sx) * (y - sy) - (ey - sy) * (x - sx)) < 0 else -1
    return sign * _len(x - fx, y - fy) / _len(ex - sx, ey - sy)


def scmn_perpendicularFoot(x, y, sx, sy, ex, ey):
    """
    Get the coordinates of the foot point of the perpendicular to the
    line (sx, sy) - (ex, ey) and the given point (x, y).

    :type x: float
    :type y: float
    :type sx: float
    :type sy: float
    :type ex: float
    :type ey: float
    :rtype: (float, float)
    """
    dx = ex - sx
    dy = ey - sy

    t = (dx * (x - sx) + dy * (y - sy)) / (pow(dx, 2.0) + pow(dy, 2.0))
    return sx + t * dx, sy + t * dy


def scmn_radToDeg(rad):
    """
    Convert radians to degrees.

    :param rad: Radian value
    :type rad: float
    :return: Converted value in degrees
    :rtype: float
    """
    return rad * 360 / SCMN_FULL_CIRCLE_RAD


def scmn_degToRad(deg):
    """
    Convert degrees to radians.

    :param deg: Degree value
    :type deg: float
    :return: Converted value in radians
    :rtype: float
    """
    return deg * SCMN_FULL_CIRCLE_RAD / 360
