# ----------------------------------------------------------------------------
# - scmn-xml-utils                                                           -
# - silkan python library: xml utility functions                             -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       scmn_utils
   :platform:     Unix, Windows
   :synopsis:     Silkan Python Library - xml utility functions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from xml.etree import ElementTree


def scmn_defaultXmlRootF(tree):
    """
    Default root-node function. Returns default root-node of tree.

    :param tree: ElementTree to return root node for
    :type tree: ElementTree.ElementTree
    :return: Root node
    :rtype: ElementTree.Element
    """
    return tree.getroot()


def scmn_readXmlTree(path, edsc=None, rootF=None, elemHandlers=None):
    """
    Read an XML file into a generic data structure.

    Which tags/attributes are being read can be controlled by the `edsc`-
    parameter. If empty or `None`, everything will be read. If a value is
    provided, it must be a dictionary of tags (`str`) to lists of attribute-
    names (`list[str]`). Only tags with an entry in the dictionary will be
    read. If the dictionary-value for a tag is `None`, all attributes will
    be read, otherwise only those named.

    The recursive structure returned represents the XML as tuples of
     (tag, attributes, text-content, children).

    :param path: Path of XML-file to read
    :type path: str
    :param edsc: Description of arguments to read
    :type edsc: dict[str, list[str] | None] | None
    :param rootF: Root-element function
    :type rootF: (ElementTree.ElementTree) -> ElementTree.Element
    :param elemHandlers: Optional iterable of element handlers
    :type elemHandlers: Iterable[(ElementTree.ElementTree,
                                 str, dict[str, list[str] | None],
                                 (ElementTree.ElementTree) ->
                                  ElementTree.Element,
                                 Iterable) ->
                                (str, dict[str, str], str, list)]
    :return: Read XML data
    :rtype: (str, dict[str, str], str, list) | None
    """
    def readElement(elem):
        """
        :type elem: ElementTree.Element
        :rtype: (str, dict[str, str], str, list) | None
        """
        if elemHandlers:
            for hndl in elemHandlers:
                rslt = hndl(elem, path, edsc, rootF, elemHandlers)
                if rslt is not None:
                    return rslt

        tag = elem.tag.split('}')[-1].lower()
        if edsc:
            if tag not in edsc:
                return None
            desc = edsc.get(tag)
            if desc is None:
                attrs = list(elem.items())
            else:
                attrs = [(k, elem.get(k)) for k in desc or []
                         if elem.get(k) is not None]
        else:
            attrs = list(elem.items())
        attrs.sort()

        children = list(filter(None, (readElement(e) for e in elem)))
        children.sort()
        text = elem.text.strip() if elem.text else ''

        if attrs or text or children:
            return tag, attrs, text, children
        else:
            return None

    tree = ElementTree.ElementTree(file=path)
    root = rootF(tree) if rootF else tree.getroot()

    if root is None:
        return None

    return readElement(root)
