# ----------------------------------------------------------------------------
# - wmaob-top-builder                                                        -
# - wild cruncher:aob scheduling group builder                               -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2014 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
..  module::      wmaob_group_builder
   :platform:     Unix, Windows
   :synopsis:     Scheduling group builder environment

.. moduleauthor:: Amaury Darsch <amaury.darsch@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from wcr_backend.modules.model.aob.wmaob_wmi_multi_builder import \
    WmaobMultipleBuilder
from wcr_core.modules.model.db.wmhdb_scheduling_group import WmhdbGroupBlock
from wcr_backend.modules.model.fog.wmfog_builder import WmfogBuilder
from wcr_backend.modules.model.aob.wmaob_utils import wmaob_format_model_names,\
    get_expre_param
from wplg_cold.wcld_builder import *
from wplg_cold.wcld_execution import *
from wcr_core.modules.model.db.wmhdb_instance import WmhdbBlockInstance
from wcr_backend.modules.model.aob.wmaob_utils import wmaob_C_format_name
import os


# ----------------------------------------------------------------------------
# - class section                                                           -
# ----------------------------------------------------------------------------

class WmaobGroupBuilder:
    """
    The WmaobGroupBuilder is the scheduling group factory builder. The class
    is designed to operate with a scheduling group. In order to operate
    properly, the scheduling group has to be flatten. This scheduling group
    builder operates only with matlab/scilab/octave text model and functional
    operator blocks.
    """

    def __init__(self, pctx, cgns, prof):
        """
        Create a builder by context and profile

        :param pctx: Project context context
        :type pctx: WcorProjectContextP
        :param cgns: Code generators
        :type cgns: Iterable[WcgnCodeGeneratorP]
        :param prof: Profile name
        :type prof: str | None
        """
        # save the project/profile
        self.prof = prof
        #: Code generators
        self.__codegens = cgns
        # create a fog builder
        self.fogb = WmfogBuilder(pctx, prof)
        # create a multi generator
        self.mbld = WmaobMultipleBuilder(pctx, cgns, prof)
        self.path = pctx.getProjectWorkspace().getProjectPath()
        self.script = pctx.getConfigurationValue('WCR_SCRIPT_PATH')
        self.cold = pctx.getConfigurationValue('WCR_CML_HOME')
        self.expfname = 'expreparam'

    def setCallback(self, cb, fs):
        self.mbld.addCallbacks(cb, fs)

    def generate(self, inst, strm=None):
        """
        Generate the workspace code for an instance. This is the entry point
        in the builder. If the instance is not supported, nothing is
        produced.

        :param inst: the model instance to process
        :param strm: Process output/error stream handler
        :type strm: ScmnProcStreamHandlerP
        """
        # check for a scheduling group
        if not isinstance(inst, WmhdbGroupBlock):
            return
        # check for valid scheduling group
        if inst.getSchedulingGroup() == 'none':
            return
        # blocks renaming to remove the special character of the C language
        wmaob_format_model_names(inst)
        self.processParams(inst, inst.getSchedulingGroup())
        # flatten the scheduling group
        flat = inst.flatten()
        # loop in the instances and generate
        for li in flat.getModelInstances():
            for cgen in self.__codegens:
                if cgen.canGenerate(li):
                    cgen.generate(li, strm)
                    break
        # loop in the functions and generate
        for lf in flat.getOperatorInstances():
            self.fogb.generate(lf)
        # generate the scheduling group wmi interface
        self.mbld.generate(flat, inst.getSchedulingGroup())

    def processParams(self, model, shedgroup, inst='', dic=None,
                       formatname=False):
        """
        This function changes parameter blocks that are expressions into
        simple variables. The expressions computation is transfer to an
        external matlab file, that will be process during the code generation.

        :param model: the model instance to process
        :param inst: the name of the upper instance use as prefix to
                     distinguish from the upper blocks
        :type inst: str
        :param dic: the dictionnary to store the variable and its associatexd
                    expression
        :param formatname: set to true to process special chars
        """
        prof = self.prof
        if prof is None:
            prof = "default"
        expf = os.path.join(self.path, "script", self.expfname+'.m')

        vstr = self._convertModelParam(model, inst, formatname, dic)

        if vstr:
            # write the created variable into the matlab file
            if not os.path.exists(os.path.dirname(expf)):
                os.mkdir(os.path.dirname(expf))
            if not os.path.exists(os.path.join(self.path, prof, shedgroup)):
                os.mkdir(os.path.join(self.path, prof, shedgroup))
            file = open(expf, 'a+')
            file.write(vstr)
            file.close()
            # if expressions has been found, used cold
            cold = WcldColdExecute(os.path.join(self.cold, "bin", "cold"))
            mfile = expf
            cfile = os.path.join(os.path.join(self.path, prof, shedgroup),
                                 self.expfname+".cpp")
            refcldpath = list()
            refcldpath.append(self.script)
            command = cold.getFunctionCommand(mfile, cfile, None,
                                              func=self.expfname,
                                              rpth=refcldpath, glob=True)
            if cold.execute(command):
                self.mbld.addCallbacks(self.expfname, "ParamExpre")

    def _convertModelParam(self, model, inst, formatname, dic):
        """
        This function changes parameter blocks that are expressions into
        simple variables. Return the string of the script to initialise them

        :return: vstr str
        """
        # the variable string to be inserted into the expression file
        vstr = ""
        # dig into the model
        for li in model.getBlockAndModelInstances():
            if isinstance(li, WmhdbBlockInstance):
                if formatname:
                    name = wmaob_C_format_name(li.getName())
                else:
                    name = li.getName()
                vstr += self._convertModelParam(li.getBlock(), inst+name+'_',
                                                formatname, dic)
        # do the work, ie modify the expression and change them into variables
        for i in model.getOperatorInstances():
            if formatname:
                blkname = wmaob_C_format_name(i.getName())
            else:
                blkname = i.getName()
            expre = get_expre_param(i)
            if expre:
                params = i.getParameters()
                for p in params:
                    if p.getName() in expre:
                        vstr += "\nglobal "+inst+blkname+"_"+p.getName()+\
                               ";\n"
                        vstr += inst+blkname+"_"+p.getName()+"="+\
                               p.getValue()+";\n"
                        if dic is not None:
                            dic[inst+blkname+"_"+p.getName()]=p.getValue()
                        p.setValue(inst+blkname+"_"+p.getName())
        return vstr

    def restoreParams(self, model, ref, inst=''):
        """
        restore the inital parameters
        """
        for li in model.getBlockAndModelInstances():
            if isinstance(li, WmhdbBlockInstance):
                self.restoreParams(li.getBlock(), ref, inst+li.getName()+'_')

        for i in model.getOperatorInstances():
            params = i.getParameters()
            for p in params:
                if p.getValue() in ref.keys():
                    p.setValue(ref[p.getValue()])

    def cleanFiles(self):
        """
        Delete the matlab script (and the associated cpp file) that
        computes expressions
        """
        prof = self.prof
        if prof is None:
            prof = "default"
        expf = os.path.join(self.path, "script", self.expfname+'.m')
        if os.path.exists(expf):
            os.remove(expf)
        cfile = os.path.join(os.path.join(self.path, 'default', prof),
                             self.expfname+".cpp")
        if os.path.exists(cfile):
            os.remove(cfile)
