# ----------------------------------------------------------------------------
# - wmaob-top-builder                                                        -
# - wild cruncher:aob tophm builder                                          -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2014 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
..  module::      wmaob_top_builder
   :platform:     Unix, Windows
   :synopsis:     Top hierarchy builder environment

.. moduleauthor:: Amaury Darsch <amaury.darsch@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from wcr_backend.modules.model.aob.wmaob_group_builder import WmaobGroupBuilder
from wcr_backend.modules.model.aob.wmaob_data_flow import WmaobDataFlow
from wcr_core.modules.model.db.wmhdb_top_block import WmhdbTopHierarchy
import os
import subprocess


# ----------------------------------------------------------------------------
# - class section                                                           -
# ----------------------------------------------------------------------------

class WmaobTopBuilder:
    """
    The WmaobTopBuilder is the top hierarchy factory builder. The class is
    designed to operate with a top hierarchy model which contains scheduling
    groups. When the generate method is called, the set of scheduling group
    is generated and the scheduling group builder is called.
    """

    def __init__(self, pctx, cgns, prof):
        """
        Create a builder by context and profile

        :param pctx: Project context context
        :type pctx: WcorProjectContextP
        :param cgns: Code generators
        :type cgns: Iterable[WcgnCodeGeneratorP]
        :param prof: Profile name
        :type prof: str | None
        """
        #: The project context
        self.__pctx = pctx
        #: Code generators
        self.__codeGens = cgns
        #: Profile
        self.__prof = prof
        # create a data flow evaluator
        self.__flow = WmaobDataFlow(pctx)

    def generate(self, inst, strm=None):
        """
        Generate the workspace code for an instance. This is the entry point
        in the builder. If the instance is not supported, nothing is
        produced.

        :param inst: the instance to process
        :param strm: Process output/error stream handler
        :type strm: ScmnProcStreamHandlerP
        """
        # check for top hierarchy
        if not isinstance(inst, WmhdbTopHierarchy): return
        # perform the data flow evaluation
        self.__flow.evaluateBlock(inst)
        # get the set of scheduling group
        sset = inst.getSchedulingGroupSet()
        # Callbacks are supported only if there is one scheduling group.
        # In the case of multiple scheduling group we don't know how
        # to handle global variables
        if len(sset) > 1:
            for v in inst.getCallbackDict().values():
                if v != '':
                    strm.handleError("Exception : Can't generate (Callback "
                                     "function for multiple Scheduling Group)")
                    return
        # loop in the set
        for name in sset:
            # do not process none group
            if name == 'none': continue
            # create a scheduling group
            sg = inst.createSchedulingGroup(name)
            # create a schduling group builder
            sbld = WmaobGroupBuilder(self.__pctx, self.__codeGens,
                                     self.__prof)
            if inst.getCallbackFunction('InitFcn') is not "" \
                    and inst.getCallbackFunction('InitFcn') is not None:
                sbld.setCallback('InitFcn', "Initcallback")
            # call builder for generation
            sbld.generate(sg, strm=strm)
            # generate the function cold_get_uagt_by_name for the cold
            # generated files
            self._generateCldVarInterface(name)

    def _generateCldVarInterface(self, sched, strm=None):
        """
        This function generates and executes the command calling the utils
        that creates the cold variable acces function

        :param strm: Process output/error stream handler
        :type strm: WcliProcStreamHandler or WguiCommandOutput
        :return:
        """
        if not self.__pctx.getConfigurationValue('WCR_CML_HOME'):
            if strm: strm.handleWarning("WARNING : no cold path set")
            return
        cmd = [os.path.join(self.__pctx.getConfigurationValue("WCR_CML_HOME"),
                            'bin', 'cruncher_utils')]
        prof = self.__prof
        if prof is None:
            prof = "default"
        path = os.path.join(self.__pctx.getProjectWorkspace().getProjectPath(),
                            prof, sched)
        if not (os.path.exists(os.path.join(path, "CldGlobals.vars"))
                or os.path.exists(os.path.join(path, "CldLocals.vars"))):
            return
        cmd.append('MATLAB')
        cmd.append(os.path.join(path, "CldLocals.vars"))
        cmd.append(os.path.join(path, "CldGlobals.vars"))
        cmd.append(os.path.join(path, "CldWorkspace.h"))
        cmd.append(os.path.join(path, "Cldvaracces.cpp"))
        subprocess.call(cmd)
