# ----------------------------------------------------------------------------
# - wmaob-wmi-code-generator                                                 -
# - wild cruncher:aob wmi code generator                                     -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2014 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - author    (c) 2014-2014 piv                                              -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
..  module::      wmaob_wmi_code_generator
   :platform:     Unix, Windows
   :synopsis:     Mockup interface code generator

.. moduleauthor:: Amaury Darsch <amaury.darsch@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------
from silkan_common.scmn_mime_type import SCMN_MIME_ID_C

from wcr_core.modules.model.db.wmhdb_rti_type import WmhdbRtiType
import os


# ----------------------------------------------------------------------------
# - private section                                                          -
# ----------------------------------------------------------------------------

def prv_aob_map_rti_sid(rti):
    t = rti.getType()
    if t == WmhdbRtiType.F64: return 'd_f64'
    if t == WmhdbRtiType.F32: return 'd_f32'
    if t == WmhdbRtiType.S64: return 'd_s64'
    if t == WmhdbRtiType.U64: return 'd_u64'
    if t == WmhdbRtiType.S32: return 'd_s32'
    if t == WmhdbRtiType.U32: return 'd_u32'
    if t == WmhdbRtiType.S16: return 'd_s16'
    if t == WmhdbRtiType.U16: return 'd_u16'
    if t == WmhdbRtiType.S08: return 'd_s08'
    if t == WmhdbRtiType.U08: return 'd_u08'
    if t == WmhdbRtiType.BLN: return 'd_bln'
    return None


def prv_aob_map_rti_ssid(sid, rti):
    ext = prv_aob_map_rti_sid(rti)
    if ext is None: return None
    return sid + '.' + ext


def prv_aob_map_rti_dsid(sid, rti):
    ext = prv_aob_map_rti_sid(rti)
    if ext is None: return None
    return sid + '->' + ext


class WmaobMockupOperator:
    """
    The WmaobMockupOperator class is self contain object for an operator call.
    The class holds the operator and function name in association with a list
    of input and output wires.
    """

    def __init__(self, name, inam, flgs=None):
        """
        Initialize a mockup operator by name and function name

        :param name: the operator name
        :param inam: the function name
        :param flgs: the operator flags
        """
        # save the operator and function name
        self._name = name
        self._inam = inam
        self._flgs = flgs
        # initialize the input/output list
        self._ilst = list()
        self._olst = list()

    def isignore(self):
        """
        Check if an operator can be safely ignored
        """
        if self._flgs is None: return False
        return 'port' in self._flgs

    def isxbus(self):
        """
        Check if an operator is a bus
        """
        if self._flgs is None: return False
        return 'bus' in self._flgs

    def isinput(self):
        """
        Check if an operator is an input type
        """
        if self._flgs is None: return False
        return 'input' in self._flgs

    def isoutput(self):
        """
        Check if an operator is an output type
        """
        if self._flgs is None: return False
        return 'output' in self._flgs

    def getMangleOperatorName(self):
        """
        Get an operator mangled name as a mean to avoid conflict with
        C keyword.

        :return: the variable name associated with the operator
        """
        return 'op_' + self._inam

    def _getOutputListCode(self):
        # check for null list
        if len(self._olst) == 0: return None
        return '{' + ','.join('&S' + str(cid) for cid in self._olst) + '};'

    def _getInputListCode(self):
        # check for null list
        if len(self._ilst) == 0: return None
        return '{' + ','.join('D' + str(cid) for cid in self._ilst) + '};'

    def addInputConnector(self, snam):
        """
        Add an input signal by name

        :param snam: the signal name to add
        """
        self._ilst.append(snam)

    def addOutputConnector(self, snam):
        """
        Add an output signal by name

        :param snam: the signal name to add
        """
        self._olst.append(snam)

    def signature(self):
        """
        Create an operator signature to be included in a header
        """
        # safely ignore some operators
        if self.isignore() or self.isxbus(): return None
        # generate signature
        return '// function {inam}[{name}]\n' \
               't_wcr_wco_status wcr_wco_initialize_{inam}' \
               '(t_wcr_wmi_rtctx*, t_wcr_wco_opctx*, t_wcr_rti_U64, ' \
               't_wcr_rti_U64, t_wcr_rti_uagt**, t_wcr_rti_uagt**);'.format(
                   inam=self._inam, name=self._name)

    def declaration(self):
        """
        Create an operator declaration
        """
        # safely ignore some operators
        if self.isignore() or self.isxbus(): return None
        # generate declaration
        return 'static t_wcr_wco_opctx ' + self.getMangleOperatorName() + ';\n'

    def getPortIdDeclaration(self):
        """
        Generate a port id declaration
        """
        # ignore non ibus/obus operator
        if not self.isxbus(): return None
        result = ''
        # check for input bus
        if self.isinput():
            for sid in self._olst:
                result = 'static long ID' + str(sid) + ' = -1;\n'
            # check for output bus
        if self.isoutput():
            for did in self._ilst:
                result = 'static long ID' + str(did) + ' = -1;\n'
        return result or None

    def getPortTypeDeclaration(self):
        """
        Generate a port type declaration
        """
        # ignore non ibus/obus operator
        if not self.isxbus(): return None
        result = ''
        # check for an input bus
        if self.isinput():
            for sid in self._olst:
                result = '  t_wcr_rti_udef T' + str(sid) + ';\n'
        # check for an output bus
        if self.isoutput():
            for did in self._ilst:
                result = '  t_wcr_rti_udef T' + str(did) + ';\n'
        return result or None

    def getPortConnect(self):
        """
        Generate a port connect statement
        """
        # ignore non ibus/obus operator
        if not self.isxbus(): return None
        result = ''
        # check for an input bus
        if self.isinput():
            for sid in self._olst:
                result = '  wcr_rti_udef_set_default (&T{sid});\n' \
                         '  ID{sid} = wcr_wmi_new_port  (ctx, "{inam}",\n' \
                         '    &T{sid}, False);\n' \
                         '  if (ID{sid} < 0) return WCR_WMI_FAILURE;' \
                         '\n\n'.format(sid=sid, inam=self._inam)
        # check for an output bus
        if self.isoutput():
            for did in self._ilst:
                result = '  wcr_rti_udef_set_default (&T{did});\n' \
                         '  ID{did} = wcr_wmi_new_port  (ctx, "{inam}",\n' \
                         '    &T{did}, True);\n' \
                         '  if (ID{did} < 0) return WCR_WMI_FAILURE;' \
                         '\n\n'.format(did=did, inam=self._inam)
        return result or None

    def getInputPortRead(self):
        """
        Generate a call to the input port read
        """
        # ignore non ibus operator
        if not (self.isxbus() and self.isinput()): return None
        # loop in the connector - normally only one
        result = ''
        for sid in self._olst:
            result += '  result = wcr_wmi_get_port_value (ctx, ID{sid}' \
                      ', &S{sid});\n' \
                      '  if (result != WCR_WMI_SUCCESS) ' \
                      'return WCR_WMI_FAILURE;\n'.format(sid=sid)
        return result or None

    def getOutputPortWrite(self):
        """
        Generate a call to the output port read
        """
        # ignore non obus operator
        if not (self.isxbus() and self.isoutput()): return None
        # loop in the connector - normally only one
        result = ''
        for did in self._ilst:
            result += '  result = wcr_wmi_set_port_value (ctx, ID{did}' \
                      ', D{did});\n' \
                      '  if (result != WCR_WMI_SUCCESS) ' \
                      'return WCR_WMI_FAILURE;\n'.format(did=did)
        return result or None

    def getInitializeCall(self):
        """
        Create an operator initialize call
        """
        # safely ignore some operators
        if self.isignore() or self.isxbus(): return None
        # generate call
        result = '  // operator ' + self._inam + '[' + self._name + ']\n'
        # update output list
        text = self._getOutputListCode()
        if text is None:
            result += '    t_wcr_rti_uagt**    out = WCR_RTI_NILP;\n'
        else:
            result += '    t_wcr_rti_uagt*   out[] = ' + text + '\n'
        # update input list
        text = self._getInputListCode()
        if text is None:
            result += '    t_wcr_rti_uagt**     in = WCR_RTI_NILP;\n'
        else:
            result += '    t_wcr_rti_uagt*    in[] = ' + text + '\n'
        # update result
        result += '    t_wcr_wco_status status = \n' \
                  '      wcr_wco_initialize_{inam} (ctx, &{mopn}, ' \
                  '{olen}, {ilen}, out, in);\n'.format(
                      inam=self._inam, mopn=self.getMangleOperatorName(),
                      olen=len(self._olst), ilen=len(self._ilst))
        return result

    def getResetCall(self):
        """
        Create an operator reset call
        """
        # safely ignore some operators
        if self.isignore() or self.isxbus(): return None
        # generate call
        result = 'status = wcr_wco_tplte_rset (&sg_exe, &' + \
                 self.getMangleOperatorName() + ');'
        return result

    def getUpdateCall(self):
        """
        Create an operator update call
        """
        # safely ignore some operators
        if self.isignore() or self.isxbus(): return None
        # generate call
        result = 'status = wcr_wco_tplte_updt (&sg_exe, &' + \
                 self.getMangleOperatorName() + ');'
        return result

    def getComputeCall(self):
        """
        Create an operator compute call
        """
        # safely ignore some operators
        if self.isignore() or self.isxbus(): return None
        # generate call
        result = 'status = wcr_wco_tplte_comp (&sg_exe, &' + \
                 self.getMangleOperatorName() + ');\n'
        # test status
        result += '  if (status != WCR_WCO_SUCCESS) ' + \
                  'result = WCR_WMI_FAILURE;\n'
        return result

    def getFreeCall(self):
        """
        Create an operator free call
        """
        # safely ignore some operators
        if self.isignore() or self.isxbus(): return None
        # generate call
        result = 'wcr_wco_tplte_free (&' + self.getMangleOperatorName() + ');'
        return result


class WmaobMockupCall:
    """
    The WmaobMockupCall class is self contain object for a function call.
    The class hold the function name to call, the argument list and the
    return value. The argument list is stored in the form of (type, name)
    tuple value.
    """

    def __init__(self, name, args=None, rtyp=None):
        """
        Create a mockup call by name without argument

        :param name: the function name
        """
        self._name = name
        self._args = args
        self._rtyp = rtyp
        # initialize the input/output list
        self._ilst = dict()
        self._olst = dict()

    def setTypes(self, args, rtyp=None):
        """
        Set the arguments and output types at once

        :param args: the argument list types
        :param rtyp: the return types
        """
        self._args = args
        self._rtyp = rtyp

    def addInputConnector(self, name, sid):
        """
        Add an input connector/signal tuple

        :param name: the connector name to add
        :param sid:  the connector signal id
        """
        self._ilst[name] = 'D' + str(sid)

    def addOutputConnector(self, name, sid):
        """
        Add an output connector/signal tuple

        :param name: the connector name to add
        :param sid:  the connector signal id
        """
        self._olst[name] = 'S' + str(sid)

    def signature(self):
        """
        Create a function signature to be included in a header
        """
        # compute the return type
        result = '// mockup ' + self._name + '\n'
        if (self._rtyp is not None) and (len(self._rtyp) == 1):
            result = result + self._rtyp[0][1].getRtiTypeStr()
        else:
            result += 'void'
            # add the name
        result += ' ' + self._name
        # add the argument list
        if self._args is None:
            # check outputs
            if (self._rtyp is not None) and (len(self._rtyp) > 1):
                text = None
                for arg in self._rtyp:
                    tval = arg[1].getRtiTypeStr() + '* ' + arg[0]
                    if text is None:
                        text = '(' + tval
                    else:
                        text = text + ', ' + tval
                result += text + ');'
            else:
                # set no inputs
                result += '();'
        else:
            # add inputs
            text = None
            for arg in self._args:
                tval = arg[1].getRtiTypeStr() + ' ' + arg[0]
                if text is None:
                    text = '(' + tval
                else:
                    text += ',' + tval
            # add outputs
            if (self._rtyp is not None) and (len(self._rtyp) > 1):
                for arg in self._rtyp:
                    tval = arg[1].getRtiTypeStr() + '* ' + arg[0]
                    if text is None:
                        text = '(' + tval
                    else:
                        text += ', ' + tval
            result += text + ');'
            # here it is
        return result

    def getPortIdDeclaration(self):
        """
        Generate a port id declaration
        """
        return None

    def getPortTypeDeclaration(self):
        """
        Generate a port type declaration
        """
        return None

    def getPortConnect(self):
        """
        Generate a port connect statement
        """
        return None

    def getInputPortRead(self):
        """
        Generate a call to the input port read
        """
        return None

    def getInitializeCall(self):
        """
        Create an operator initialize call
        """
        return None

    def getResetCall(self):
        """
        Create a function reset call
        """
        return None

    def getOutputPortWrite(self):
        """
        Generate a call to the output port read
        """
        return None

    def getUpdateCall(self):
        """
        Create a function update call
        """
        return None

    def getComputeCall(self):
        """
        Create a function compute call
        """
        result = ''
        if (self._olst is not None) and (len(self._olst) == 1):
            for sid in self._olst.values():
                result = prv_aob_map_rti_ssid(sid, self._rtyp[0][1])
            result += ' = '
            # add the function name
        result += self._name
        # add the arguments
        text = None
        if (self._args is not None) and (len(self._args) > 0):
            for arg in self._args:
                sid = self._ilst[arg[0]]
                if text is None:
                    text = '(' + prv_aob_map_rti_dsid(sid, arg[1])
                else:
                    text += ', ' + prv_aob_map_rti_dsid(sid, arg[1])
        if (self._rtyp is not None) and (len(self._rtyp) > 1):
            for arg in self._rtyp:
                sid = self._olst[arg[0]]
                if text is None:
                    text = '(&' + prv_aob_map_rti_ssid(sid, arg[1])
                else:
                    text += ', &' + prv_aob_map_rti_ssid(sid, arg[1])
        if text is None:
            text = '();'
        else:
            text += ');'
        result += text + '\n'
        # here it is
        return result

    def getFreeCall(self):
        """
        Create a function free call
        """
        return None


# ----------------------------------------------------------------------------
# - class section                                                           -
# ----------------------------------------------------------------------------

class WmaobMockupGenerator:
    """
    The WmaobMockupGenerator a wmi code generator which is used for the
    production of a mockup interface. The class enables the interface functions
    as soon as some parameters are defined in it. For example, if the port list
    to register is defined, the interface will enable automatically the
    implementation of the initialize function. The execute function is always
    implemented with a serial call of registrered functions.
    """

    def __init__(self, pctx, prof="default"):
        """
        Create a mockup generator by context and profile

        :param pctx: the application context
        :type pctx:  WcorProjectContextP
        :param prof: the profile name
        """
        # save the project/profile
        self.pnam = pctx.getProjectName()
        self.prof = prof
        # the wire set with its source
        self.wset = set()
        self.sset = set()
        # the executing list
        self.elst = []
        # the operator flag
        self._oflg = False

        self.dir = pctx.getProjectWorkspace().getDirectoryPath(prof=prof)
        # the external functions dictionary : keys will be used to distinguish
        # the different type of callback function : Init, Stop, Start ...
        self.externfunc = dict(list())

    def getMedia(self):
        return SCMN_MIME_ID_C

    def addCallback(self, fname, fset):
        """
        Add the element fname to the externfunc dictionary at fset key.
        
        :param fname: function name to add
        :param fset: function set to add the function
        """
        for key in self.externfunc.keys():
            if fname in self.externfunc[key]:
                raise Exception("Multiple declaration of extern function : "
                                +fname)
        if fset in self.externfunc:
            self.externfunc[fset].append([fname])
        else:
            self.externfunc[fset] = [fname]

    def addWireId(self, wid, sid, did, wnm, ret, red, rxd, ryd):
        """
        Add a wire id in the wire set with its source and destination id

        :param wid: the wire id to add
        :param sid: the source id to add
        :param did: the destination id to add
        :param wnm: the optional wire name
        :param ret: rti enumeration type
        :param red: rti enumeration dimension
        :param rxd: rti x dimension
        :param ryd: rti y dimension
        """
        if (wnm is not None) and (len(wnm) == 0): wnm = None
        if rxd is None: rxd = 0
        if ryd is None: ryd = 0
        self.wset.add((wid, sid, did, wnm))
        self.sset.add((sid, ret, red, rxd, ryd))

    def addExecuteCall(self, name):
        """
        Add a new function call to the execute list

        :param name: the function to add
        """
        # create a mockup call
        mc = WmaobMockupCall(name)
        # add it to the call chain
        self.elst.append(mc)
        # here it is
        return mc

    def addExecuteOperator(self, name, inam, flgs=None):
        """
        Add a new operator to the execute list

        :param name: the operator name
        :param inam: the function name
        :param flgs: the operator flags
        """
        # create a mockup operator
        mo = WmaobMockupOperator(name, inam, flgs)
        # add it to the operator list
        self.elst.append(mo)
        self._oflg = True
        # here it is
        return mo

    def generate(self):
        """
        Generate the wmi interface code as a text string. The resulting string
        can be used to feed the workspace.
        """
        # get the copyright
        result = self._copyright() + '\n'
        # add the header
        text = self._header()
        if text is not None: result = result + text + '\n'
        # add the signature
        text = self._signature()
        if text is not None: result = result + text + '\n'
        # add the wire list
        text = self._wirelist()
        if text is not None: result = result + text + '\n'
        # add the port id list
        text = self._portidlist()
        if text is not None: result = result + text + '\n'
        # add the operator declaration
        text = self._operatorlist()
        if text is not None: result = result + text + '\n'
        # add the initialize code
        text = self._initializeCode()
        if text is not None: result = result + text + '\n'
        # add the reset code
        text = self._resetCode()
        if text is not None: result = result + text + '\n'
        # add the connect code
        text = self._connectCode()
        if text is not None: result = result + text + '\n'
        # add the execute code
        text = self._executeCode()
        if text is not None: result = result + text + '\n'
        # add the finalize code
        text = self._finalizeCode()
        if text is not None: result = result + text + '\n'
        # here it is
        return result

    def _copyright(self):
        """
        Generate a copyright header for the mockup interface
        """
        return '// -------------------------------------------------------' \
               '-----\n' \
               '// - {pnam}\n' \
               '// - wild cruncher - mockup interface implementation      ' \
               '    -\n' \
               '// -------------------------------------------------------' \
               '-----\n' \
               '// - author    (c) 2012-2014 the wild team                ' \
               '    -\n' \
               '// - copyright (c) 2012-2014 silkan / wild systems        ' \
               '    -\n' \
               '// -------------------------------------------------------' \
               '-----\n'.format(pnam=self.pnam)

    def _header(self):
        """
        Generate the header calls for the mockup interface
        """
        head = '#include "wcr_rti_sysos.h"\n' \
               '#include "wcr_wra_rtlib.h"\n' \
               '#include "lib_msc_allxx.h"\n' \
               '#include "lib_mth_allxx.h"\n' \
               '#include "lib_dsp_allxx.h"\n'
        if self.externfunc:
            head += "#include <stdio.h>\n"
            head += "extern void cold_init(size_t s);\n"
            head += "extern void cold_terminate(void);\n"
        for key in self.externfunc.keys():
            for p in self.externfunc[key]:
                head += "extern void "+p+"(void);\n"
        return head

    def _signature(self):
        """
        Generate the function signature for the mockup interface
        """
        result = ''
        for f in self.elst:
            text = f.signature()
            if text is not None:
                result += text + '\n'
        if result == '': return None
        result = '// function call declaration\n' + result
        return result

    def _wirelist(self):
        """
        Generate the wire list code for the mockup interface
        """
        # check for null set
        if len(self.wset) == 0: return
        # preparse the result string
        result = '// source list mockup interface\n'
        # loop in the destination set
        for sid,ret,red,rxd,ryd in self.sset:
            # declare source type
            result += 'static t_wcr_rti_udef  T{sid}'.format(sid=sid)
            result += ' = {'
            result += '{ret},{red},{rxd},{ryd}'.format(ret=ret,red=red,
                                                       rxd=rxd,ryd=ryd)
            result += '};\n'
            # declare source value
            result += 'static t_wcr_rti_uagt  S{sid};\n'.format(sid=sid)
        result += '// wire list mockup interface\n'
        # loop in the wire set
        for wid,sid,did,wnm in self.wset:
            # declare wire id
            result += 'static t_wcr_rti_uagt* W{wid} = WCR_RTI_NILP;'.format(
                wid=wid)
            if wnm is not None: result += ' // ' + wnm
            result += '\n' \
                      'static t_wcr_rti_uagt* D{did} = WCR_RTI_NILP;\n'.format(
                          did=did)
        return result

    def _portidlist(self):
        """
        Generate the port id list
        """
        # check for null set
        if len(self.elst) == 0: return None
        # preparse the result string
        result = ''
        for f in self.elst:
            text = f.getPortIdDeclaration()
            if text is not None: result += text
            # check for empty result
        if result == '': return None
        result = '// port id declaration\n' + result
        return result

    def _porttypelist(self):
        """
        Generate the port type list
        """
        # check for null set
        if len(self.elst) == 0: return None
        # preparse the result string
        result = ''
        for f in self.elst:
            text = f.getPortTypeDeclaration()
            if text is not None: result = result + text
            # check for empty result
        if result == '': return None
        return result

    def _operatorlist(self):
        """
        Generate the operator list for the mockup interface
        """
        # check for null set
        if len(self.elst) == 0: return None
        # preparse the result string
        result = ''
        # loop in the execute set
        for mo in self.elst:
            # check for a valid mockup
            if not isinstance(mo, WmaobMockupOperator): continue
            # get the declaration
            text = mo.declaration()
            if text is not None:
                result += text
        if result == '': return None
        result = \
            '// operator execution context\n' + \
            'static t_wcr_wco_exctx sg_exe;\n' + \
            '// operator declaration\n' + result
        return result

    def _initializeCode(self):
        """
        Generate the initialize code for the mockup interface
        """
        # check for null set
        if len(self.elst) == 0: return None
        # prepare function header
        result = '// initialize mockup interface\n' \
                 't_wcr_wmi_status wcr_wmi_initialize (' \
                 't_wcr_wmi_rtctx* ctx, void* ioh) {\n'
        # loop in the wire list
        text = ''
        for wid,sid,did,wnm in self.wset:
            text += '  W' + str(wid) + ' = &S' + str(sid) + ';\n'
            text += '  D' + str(did) + ' = &S' + str(sid) + ';\n'
        if text != '':
            result += '  // wire and destination initialization\n'
            result += text + '\n'
        if self.externfunc:
            result += '  //allocate cold memory\n' \
                      '  cold_init(1000000000);\n'
        if "Initcallback" in self.externfunc:
            result += '  // Initcallback execution\n'
            for nf in self.externfunc["Initcallback"]:
                result += '  '+nf+'();\n'
        if "ParamExpre" in self.externfunc:
            result += '  // Parameters with expretion calculating\n'
            for nf in self.externfunc["ParamExpre"]:
                result += '  '+nf+'();\n'
        # loop in the source list for type initialization
        text = ''
        for sid,ret,red,rxd,ryd in self.sset:
            text += '  if (wcr_rti_uagt_set_type (&S' + str(sid) + \
                    ', &T' + str(sid) + \
                    ') == False)\n    return WCR_WMI_FAILURE;\n'
        if text != '':
            result += '  // type initialization\n'
            result += text + '\n'

        # loop in the execute set
        tval = ''
        for m in self.elst:
            text = m.getInitializeCall()
            if text is not None:
                tval += '  {\n'
                tval += '  ' + text
                tval += '    if (status != WCR_WCO_SUCCESS) {\n'
                tval += '      wcr_rti_write (WCR_RTI_NILP, ' \
                        '"fail to initialize'
                tval += ' : ' + str(m.getMangleOperatorName()) + '");\n'
                tval += '      return WCR_WMI_FAILURE;\n'
                tval += '    }\n'
                tval += '  }\n'
        if tval != '':
            result += '  // initialize wco execution context\n' \
                      '  if (wcr_wco_exctx_init (&sg_exe) != ' \
                      'WCR_WCO_SUCCESS)\n' \
                      '    return WCR_WMI_FAILURE;\n\n' \
                      '  // operator initialization\n'
            result += tval + '\n'
            # finish function
        result += '  // success\n' \
                  '  return WCR_WMI_SUCCESS;\n' \
                  '}\n'
        return result

    def _connectCode(self):
        """
        Generate the connect code for the mockup interface
        """
        # check for null set
        if len(self.elst) == 0: return None
        # get the port type list and check for none
        text = self._porttypelist()
        if text is None: return None

        # prepare function header
        result = \
            '// connect mockup interface\n'
        result += 't_wcr_wmi_status wcr_wmi_connect (t_wcr_wmi_rtctx* ctx, ' \
                  'void* ioh) {\n'
        # add the port type
        result = result + text + '\n'

        # add the port connect statements
        result += '  // connect the ports\n'
        for f in self.elst:
            text = f.getPortConnect()
            if text is not None: result = result + text

        # finish function
        result += '  // success\n' \
                  '  return WCR_WMI_SUCCESS;\n' \
                  '}\n'
        return result

    def _executeCode(self):
        """
        Generate the execute code for the mockup interface
        """
        # prepare function header
        result = \
            '// execute mockup interface\n'
        result = result + \
                 't_wcr_wmi_status wcr_wmi_execute (t_wcr_wmi_rtctx* ctx, ' + \
                 'void* ioh) {\n'
        # declare wco execution context
        if self._oflg:
            result += '  t_wcr_wco_status status;\n'
            # declare wmi status
        result += '  t_wcr_wmi_status result = WCR_WMI_SUCCESS;\n\n'

        # add wco context initialization
        if self._oflg:
            result += '  // update wco context\n' \
                      '  if (wcr_wra_set_wco (&sg_exe, ctx) == False) ' \
                      'return WCR_WMI_FAILURE;\n\n'

        # add the operator read call
        itxt = ''
        for f in self.elst:
            text = f.getInputPortRead()
            if text is not None: itxt = itxt + text
        if itxt != '':
            result += '  // update input ports\n' + itxt + '\n'

        # add the update calls
        utxt = ''
        for f in self.elst:
            text = f.getUpdateCall()
            if text is not None:
                utxt = utxt + '  ' + text + '\n'
                utxt = utxt + \
                       '  if (status != WCR_WCO_SUCCESS) ' + \
                       'return WCR_WMI_FAILURE;\n'
        if utxt != '':
            result += '  // update operators\n' + utxt + '\n'

        # add the compute calls
        ctxt = ''
        for f in self.elst:
            text = f.getComputeCall()
            if text is not None:
                ctxt = ctxt + '  ' + text + '\n'
        if ctxt != '':
            result += '  // operator compute and function calls\n' + ctxt

        # add the operator read call
        otxt = ''
        for f in self.elst:
            text = f.getOutputPortWrite()
            if text is not None: otxt = otxt + text
        if otxt != '':
            result += '  // update output ports\n' + otxt + '\n'

        # finish function
        result += '  // success\n' \
                  '  return result;\n' \
                  '}\n'
        return result

    def _resetCode(self):
        """
        Generate the reset code for the mockup interface
        """
        # prepare function header
        result = \
            '// reset mockup interface\n'
        result = result + \
                 't_wcr_wmi_status wcr_wmi_reset (t_wcr_wmi_rtctx* ctx, ' + \
                 'void* ioh) {\n'
        # declare wco execution context
        if self._oflg:
            result += '  t_wcr_wco_status status;\n'
            # declare wmi status
        result += '  t_wcr_wmi_status result = WCR_WMI_SUCCESS;\n\n'

        # add wco context initialization
        if self._oflg:
            result += '  // update wco context\n' \
                      '  if (wcr_wra_set_wco (&sg_exe, ctx) == False) ' \
                      'return WCR_WMI_FAILURE;\n\n'

        # add the reset calls
        utxt = ''
        for f in self.elst:
            text = f.getResetCall()
            if text is not None:
                utxt = utxt + '  ' + text + '\n'
                utxt = utxt + \
                       '  if (status != WCR_WCO_SUCCESS) ' + \
                       'return WCR_WMI_FAILURE;\n'
        if utxt != '':
            result += '  // reset operators\n' + utxt + '\n'

        # finish function
        result += '  // success\n' \
                  '  return result;\n' \
                  '}\n'
        return result

    def _finalizeCode(self):
        """
        Generate the finalize code for the mockup interface
        """
        # prepare function header
        result = '// finalize mockup interface\n'
        result += 't_wcr_wmi_status wcr_wmi_finalize (t_wcr_wmi_rtctx* ctx, ' \
                  'void* ioh) {\n'

        # add the free calls
        ftxt = ''
        for f in self.elst:
            text = f.getFreeCall()
            if text is not None:
                ftxt = ftxt + '  ' + text + '\n'
        if ftxt != '':
            result += '  // free operator\n' + ftxt + '\n'

        if self.externfunc:
            result += '  // free cold memory\n' \
                      '  cold_terminate();\n'

        # finish function
        result += '  // success\n' \
                  '  return WCR_WMI_SUCCESS;\n' \
                  '}\n'
        return result
