# ----------------------------------------------------------------------------
# - wmaob-wmi-makefile-generator                                             -
# - wild cruncher:aob wmi makefile generator                                 -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2014 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
..  module::   wmaob_wmi_makefile_generator
:platform:     Unix, Windows
:synopsis:     Mockup interface makefile generator

.. moduleauthor:: Amaury Darsch <amaury.darsch@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from silkan_common.scmn_file_path import *
from wcr_backend.modules.model.aob.wmaob_utils import *


# ----------------------------------------------------------------------------
# - class section                                                           -
# ----------------------------------------------------------------------------

class WmaobMakefileGenerator:
    """
    The WmaobMakefileGenerator a makefile code generator which is used for the
    production of a mockup interface. The class generates a makefile which is
    used to compile a scheduling group into a dynamic library. The generator
    is designed to operate with a template which is automically copied in the
    workspace if set. By default, the code is generated to operate in
    optimized mode.
    """

    def __init__(self, pctx, prof):
        """
        Create a makefile generator by context and profile

        :param pctx: Project context context
        :type pctx: WcorProjectContextP
        :param prof: Profile name
        :type prof: str | None
        """
        # get the workspace
        self.wspc = pctx.getProjectWorkspace()
        # save the profile
        self.prof = prof
        # the code template
        self.tmpl = wmaob_get_template()
        # the default mode
        self.mode = None

    def setTemplate(self, name):
        """
        Set the template name

        :param name: the template name to set
        """
        self.tmpl = name

    def setMode(self, mode):
        """
        Set the makefile mode

        :param mode: the mode to set
        """
        self.mode = mode

    def generate(self, sgrp):
        """
        Generate the makefile code in the workspace for a scheduling group

        :param sgrp: the scheduling group name
        """
        # generate the template if any
        self._generateTemplate(sgrp)
        # generate a specific file if any
        self._generateSpecific(sgrp)

    def _generateTemplate(self, sgrp):
        """
        Generate the template in the workspace

        :param sgrp: the scheduling group

        :param sgrp: the scheduling group
        """
        # check for none
        if self.tmpl is None or sgrp == 'none': return
        # build template path and expand
        path = scmn_path_join('${WCR_HOME}', 'mak', self.tmpl)
        path = scmn_expshenv(path)
        # import it in the workspace as a makefile
        self.wspc.importFile(path, 'Makefile', self.prof, sgrp)

    def _generateSpecific(self, sgrp):
        """
        Generate a specific scheduling group makefile

        :param sgrp: the scheduling group
        """
        # check for none
        if sgrp == 'none': return
        # generate a copyright
        result = self._copyright() + '\n'
        # add the project variables
        text = self._project(sgrp)
        if text is not None:
            result += text
            # import it in the workspace as a text
        self.wspc.importText('project.mak', self.prof, sgrp, result)

    def _copyright(self):
        """
        Generate a makefile header
        """
        result = '# --------------------------------------------------------' \
                 '------\n'
        result += '# - Makefile' \
                  '                                                   -\n'
        result += '# - single scheduling group makefile' \
                  '                           -\n'
        result += '# -------------------------------------------------------' \
                  '-------\n'
        result += '# - author    (c) 2012-2014 the wild team' \
                  '                      -\n'
        result += '# - copyright (c) 2012-2014 silkan / wild systems' \
                  '              -\n'
        result += '# -------------------------------------------------------' \
                  '-------\n'
        return result

    def _project(self, sgrp):
        result = 'SGRDBGPROJECT  = ' + sgrp + '\n'
        result += 'SGROPTPROJECT  = ' + sgrp + '\n'
        return result
