# ----------------------------------------------------------------------------
# - wmaob-wmi-multi-builder                                                  -
# - wild cruncher:aob wmi multiple instance builder                          -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2014 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
..  module::      wmaob_wmi_multi_builder
   :platform:     Unix, Windows
   :synopsis:     Mockup interface multiple instance builder environment

.. moduleauthor:: amaury darsch <amaury.darsch@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from silkan_common.scmn_graph import ScmnDigraph
from silkan_common.scmn_mime_type import scmn_mime_type
from wcr_backend.modules.model.aob.wmaob_wmi_code_generator import \
    WmaobMockupGenerator
from wcr_backend.modules.model.aob.wmaob_wmi_makefile_generator import \
    WmaobMakefileGenerator
from wcr_core.modules.model.db.wmhdb_instance import WmhdbInstance, \
    WmhdbModelInstance
from wcr_core.modules.model.db.wmhdb_function import WmhdbFunction
from wcr_core.modules.model.db.wmhdb_operator import WMHDB_DELAY_FLAG
from wcr_core.modules.model.db.wmhdb_connection_point import WmhdbConnector


# ----------------------------------------------------------------------------
# - local util functions                                                      -
# ----------------------------------------------------------------------------

def wmaob_builder_graph(mflat):
    """
    Build a direct graph with functions and instances as nodes and
    wire as edges

    :rtyp : ScmnGraph
    :return: the digraph
    :type  mflat: WmhdbBlock
    :param mflat: the flat model to be tranformed into a direct graph
    """
    result = ScmnDigraph()
    # create function nodes
    for inst in mflat.getInstances():
        # add a new node for the function
        result.addNode(inst)
    # create edges
    for wire in mflat.getConnections():
        # ignore egdes from input and output function
        so = wire.getStartObject()
        eo = wire.getEndObject()
        if not result.hasNode(so) or not result.hasNode(eo):
            message = 'edge found for undefined nodes {}@{} {}@{}'.format(
                so.getName(), id(so), eo.getName(), id(eo))
            raise Exception(message)
        if not result.hasEdge(so, eo):
            result.addEdge(so, eo)
    return result


def wmaob_schedule(mflat):
    """
    The scheduling of a flatten block is basically a topoligcal sort of a
    graph with instances or functions as nodes and wires as edges. The
    nodes that insert a delay into the garph are removed from the graph
    and can be scheduled in any order at the end.

    :rtype : list
    :return: the list of sorted Functions.
    """
    if any(mflat.getBlockInstances()):
        raise ValueError('Provided instance is not flat')
    graph = wmaob_builder_graph(mflat)
    delay = []

    # remove delay nodes from the graph
    for node in graph.nodes():
        if (isinstance(node, WmhdbFunction) and
                node.getOperator().hasFlag(WMHDB_DELAY_FLAG)):
            graph.delNode(node)
            delay.append(node)

    # check for cycles
    cycle = graph.detectCycle()
    if cycle:
        # create a cycle error message
        message = "Cycle found in model: "
        raise Exception(message)
        # and now do the topological sort
    result = graph.topologicalSortFlat()
    result.extend(delay)
    return result


# ----------------------------------------------------------------------------
# - class section                                                            -
# ----------------------------------------------------------------------------

class WmaobMultipleBuilder:
    """
    The WmaobMultipleBuilder is a wmi multiple instances/functions builder.
    It is a little bit more complex that the Simple Builder since a topological
    sorting is needed to generate the wmi interface.
    """

    def __init__(self, pctx, cgns, prof):
        """
        Create a builder by context and profile

        :param pctx: Project context context
        :type pctx: WcorProjectContextP
        :param cgns: Code generators
        :type cgns: Iterable[WcgnCodeGeneratorP]
        :param prof: Profile name
        :type prof: str | None
        """
        # get the workspace
        self.wspc = pctx.getProjectWorkspace()
        #: Code generators
        self.__codegens = cgns
        # save the profile
        self.prof = prof
        # create a wmi code generator
        self.wmig = WmaobMockupGenerator(pctx, prof)
        # create a wmi makefile generator
        self.makg = WmaobMakefileGenerator(pctx, prof)

    def addBlockWiring(self, inst):
        """
        Add the group wires to the mockup generator

        :param inst: the instance/model to process
        """
        # check for none
        if inst is None: return
        # loop in the wire
        for w in inst.getConnections():
            # do not process none
            if w is None: continue
            # get the source/destination connectors
            sc = w.getStartConnector()
            dc = w.getEndConnector()
            wn = w.getName()
            if (sc is None) or (dc is None): continue
            if sc.getDirection() != WmhdbConnector.Output: continue
            if dc.getDirection() != WmhdbConnector.Input: continue
            # get the inferred type - must be present
            it = w.getInferredRtiType()
            et = it.getRtiEnumType()
            ed = it.getRtiEnumDim()
            xd = it.getXdim()
            yd = it.getYdim()
            if it is None:
                sn = sc.getName()
                dn = dc.getName()
                mesg = 'no type associated with wire {wn} [{sn},{dn}]'
                raise Exception(mesg.format(wn=wn, sn=sn, dn=dn))
            # add by id in the code generator
            self.wmig.addWireId(id(w), id(sc), id(dc), wn, et, ed, xd, yd)

    def addInstanceCall(self, inst):
        """
        Add an instance to the mockup generator.

        :param inst: the instance to add
        """
        # check for none
        if inst is None:
            return
        if not isinstance(inst, WmhdbModelInstance):
            return
        for cgen in self.__codegens:
            if cgen.canGenerate(inst):
                # get the entrey point from the code generator
                name = inst.getName()
                sgrp = inst.getSchedulingGroup()
                epnt = cgen.getEntryPoint(name, sgrp)
                # create an execute call
                mc = self.wmig.addExecuteCall(epnt)
                # get the input/output type list
                args = cgen.getInputTypeList(name, sgrp)
                rtyp = cgen.getOutputTypeList(name, sgrp)
                # set the argument list at call
                mc.setTypes(args, rtyp)
                # add the input, output port to the operator
                ic = inst.getInputConnectors()
                for c in ic:
                    mc.addInputConnector(c.getPortReference(), id(c))
                oc = inst.getOutputConnectors()
                for c in oc:
                    mc.addOutputConnector(c.getPortReference(), id(c))
                break

    def addFunctionCall(self, func):
        """
        Add a function to the mockup generator.

        :param func: the function to add
        """
        # check for none
        if func is None: return
        # get the operator and and function name
        name = func.getOperator().getCode()
        inam = func.getName()
        flgs = func.getOperator().getFlags()
        if 'bus' in flgs:
            pnam = func.getParameterValue('name')
            if pnam:
                inam = pnam
                # add an operator to the scheduling list
        mo = self.wmig.addExecuteOperator(name, inam, flgs)
        # add the input, output port to the operator
        oc = func.getOutputConnectors()
        for c in oc: mo.addOutputConnector(id(c))
        ic = func.getInputConnectors()
        for c in ic: mo.addInputConnector(id(c))

    def addScheduledCall(self, inst):
        """
        Generate the scheduling call for the calling instance

        :param inst: the instance/model to process
        """
        if any(inst.getBlockInstances()):
            raise ValueError('Provided instance is not flat')
        schd = wmaob_schedule(inst)
        for node in schd:
            if isinstance(node, WmhdbInstance):
                self.addInstanceCall(node)
            elif isinstance(node, WmhdbFunction):
                self.addFunctionCall(node)
            else:
                raise Exception()

    def addCallbacks(self, func, fset):
        self.wmig.addCallback(func, fset)

    def generate(self, inst, sgrp):
        """
        Generate the workspace code for a multiple instance flat model.

        :param inst: the instance/model to process
        """
        # check for flat model
        if any(inst.getBlockInstances()):
            raise ValueError('Provided instance is not flat')
        # generate the block wiring
        self.addBlockWiring(inst)
        # generate the scheduling calls
        self.addScheduledCall(inst)
        # call the generator
        text = self.wmig.generate()
        # import it into the workspace
        inam = inst.getName()
        name = inam + '.' + scmn_mime_type(self.wmig.getMedia()).extension
        self.wspc.importText(name, self.prof, sgrp, text)
        # generate the scheduling group makefile
        self.makg.generate(sgrp)
        return
