# ----------------------------------------------------------------------------
# - wmfog-function-code-gen                                                  -
# - wild cruncher: fog function code producer                                -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 amaury darsch                                    -
# - author    (c) 2012-2013 piv                                              -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
..  module::      wmfog_function_code_gen
   :platform:     Unix, Windows
   :synopsis:     fog function code producer

.. moduleauthor:: piv <pierre.villalon@silkan.com>
"""

LINE_LENGTH = 78

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from silkan_common.scmn_log import scmn_try_log, SCMN_LOG_INFO
from wcr_backend.modules.model.fog.wmfog_param_code import WmfogParamCode
# noinspection PyUnresolvedReferences
from wcr_backend.modules.model.fog.wmfog_parameter import *
from wcr_backend.modules.model.fog.wmfog_c_base import *
import os


# ----------------------------------------------------------------------------
# - local util functions                                                     -
# ----------------------------------------------------------------------------

def hds_to_header(name):
    v = name.split('.')
    result = 'lib_' + v[1] + '_allxx.h'
    return result


def hds_to_oper(name):
    """
    Convert a qualifed name to an operator name
    """
    v = name.split('.')
    return v[-1]


def name_to_param_func(name):
    """
    Convert a qualifed name to a parameter evaulation function
    """
    o = hds_to_oper(name)
    s = 'wmfog_get_' + o + '_implicit_params'
    return eval(s)


def wmfog_get_params_names(l):
    """
    :return: the list of names separated by a coma
    :rtype: [str]
    :param l: the list of parameters to evaluate
    :type  l: [WmhdbParameter]
    """
    result = ['"' + x.getName() + '"' for x in l]
    return result


def wmfog_get_params_values(l):
    """
    :return: the list of names separated by a coma
    :rtype: [str]
    :param l: the list of parameters to evaluate
    :type  l: [WmhdbParameter]
    """
    return ['"' + x.getValue() + '"' for x in l]


# ----------------------------------------------------------------------------
# - class section                                                           -
# ----------------------------------------------------------------------------

class WmfogFuncCode:
    """
    The WmfogFuncCode can produce C code for any operator. It produces
    code from function description as WmhdbFunction objects.
    """

    def __init__(self, pctx):
        # get the logger
        self.__pctx = pctx
        # create a parameter coder
        self.pcode = WmfogParamCode(pctx)

    def getImplicitParams(self, func):
        """
        :return: the list of implicit parameters
        :rtype: [WmhdbParameter]
        """
        op = func.getOperator()
        name = op.getCode() if op is not None else ''
        try:
            oper = name_to_param_func(name)
            result = oper(func)
        except:
            result = []
        return result

    def getParamsUagt(self, l, indent=0):
        """
        :return: the C style string to intialize uagt statically
        :rtype: str
        :param l: the list of parameters to evaluate
        :type  l: [WmhdbParameter]
        :param indent: an optional indentention value
        :type  indent: int
        """
        sp = " " * indent
        result = sp + "{\n"
        for p in l:
            result += sp + self.pcode.initUagt(p) + ",\n"
        result += sp + "}"
        return result

    def getParamsValue(self, l, indent=0):
        """
        :return: the C style string to intialize the value of a list of
                 evaluated parameters
        :rtype: str
        :param l: the list of parameters to evaluate
        :type  l: [WmhdbParameter]
        :param indent: an optional indentention value
        :type  indent: int
        """
        sp = " " * indent
        result = ""
        for p in l:
            result += sp + self.pcode.initEvaluated(p) + "\n"
        return result

    def getParamDcl(self, func):
        """
        Declare the parameter objects

        :param func: the function for which code is needed
        :type  func: WhdbOperatorInstanceP
        """
        # noinspection PyUnresolvedReferences
        params = [x for x in func.getParameters()]
        params.extend(self.getImplicitParams(func))
        nbp = len(params)
        pdta = '  t_wcr_rti_uagt  pdta[{nbp}] =\n'.format(nbp=nbp)
        return '  // The number of parameters\n' \
               '  t_wcr_rti_U64   nbp = {nbp};\n' \
               '  t_wcr_wco_param prm[nbp];\n' \
               '  // The parameter values\n' \
               '{pvls}' \
               '  // The parameters names\n' \
               '  const t_wcr_rti_STR       pnms[{nbp}] = {arr};\n' \
               '{pdta}{uagt};\n'.format(
                   nbp=nbp, pvls=self.getParamsValue(params, 2),
                   arr=wmfog_init_c_array(wmfog_get_params_names(params)),
                   pdta=pdta, uagt=self.getParamsUagt(params, len(pdta)))

    # noinspection PyUnusedLocal
    def getParamIni(self, func):
        """
        Initialize the parameter objects

        :param func: the function for which code is needed
        :type  func: WhdbOperatorInstanceP
        """
        return '  // Init the  operator parameters\n' \
               '  for (idx = 0; idx < nbp; idx ++) {\n' \
               '    prm[idx].p_nam = pnms[idx];\n' \
               '    prm[idx].p_val = &pdta[idx];\n' \
               '  }\n'

    # noinspection PyUnusedLocal
    def getXconfIni(self, func):
        """
        Update the xconf object

        :param func: the function for which code is needed
        :type  func: WhdbOperatorInstanceP
        """
        # cnf initialization
        return '  // cnf initialization\n' \
               '  if ((result = wcr_wco_xconf_init (&cnf, nbp, nbo, nbi)) ' \
               '!= WCR_WCO_SUCCESS)\n' \
               '    return result;\n' \
               '  // parameter binding\n' \
               '  for (idx = 0; idx < nbp; idx ++)\n' \
               '    if ((result = wcr_wco_xconf_add_pdta ' \
               '(&cnf, idx, &(prm[idx]))) != WCR_WCO_SUCCESS)\n' \
               '      return result;\n' \
               '  // output binding\n' \
               '  for (idx = 0; idx < nbo; idx ++)\n' \
               '    if ((result = wcr_wco_xconf_bind_odta ' \
               '(&cnf, idx, out[idx])) != WCR_WCO_SUCCESS)\n' \
               '      return result;\n' \
               '  // input binding\n' \
               '  for (idx = 0; idx < nbi; idx ++)\n' \
               '    if ((result = wcr_wco_xconf_bind_idta ' \
               '(&cnf, idx, in[idx])) != WCR_WCO_SUCCESS)\n' \
               '      return result;\n'

    def getHeader(self, func):
        """
        Get the header of the file, i.e some C comments

        :param func: the function for which code is needed
        """
        # get the function unique name
        uname = func.getName()
        tmp1 = '// - {}.c'.format(uname)
        tmp2 = '// - wild cruncher - {i} [{f}]'.format(
            i=uname,f=func.getOperator().getCode())
        tmp3 = '// - author    (c) 2012-2014 the wild team'
        tmp4 = '// - copyright (c) 2012-2014 silkan / wild systems'

        return '// '.ljust(LINE_LENGTH, '-') + '\n' + \
               tmp1.ljust(LINE_LENGTH - 1, ' ') + '-\n' + \
               tmp2.ljust(LINE_LENGTH - 1, ' ') + '-\n' + \
               '// '.ljust(LINE_LENGTH, '-') + '\n' + \
               tmp3.ljust(LINE_LENGTH - 1, ' ') + '-\n' + \
               tmp4.ljust(LINE_LENGTH - 1, ' ') + '-\n' + \
               '// '.ljust(LINE_LENGTH, '-') + '\n'

    # noinspection PyUnusedLocal
    def getInclude(self, func):
        """
        Get the header of the file, i.e some C comments

        :param func: the function for which code is needed
        """
        result = '#include "wcr_wmi_rtctx.h"\n' \
                 '#include "wcr_rti_types.h"\n'
        try:
            head = hds_to_header(func.getOperator().getCode())
            result = result + '#include "' + head + '"\n'
        except:
            pass
        return result

    def getInitFunc(self, func):
        """
        Get the initialize C function for the  wsh func

        :param func: the function for which code is needed
        :type  func: WhdbOperatorInstanceP
        """
        try:
            oper = hds_to_oper(func.getOperator().getCode())
        except KeyError:
            raise KeyError('Could not find operator ' +
                           func.getOperator().getCode())

        return '// the operator initialization\n' \
               't_wcr_wco_status wcr_wco_initialize_{name}' \
               '(t_wcr_wmi_rtctx* ctx, t_wcr_wco_opctx* ope,' \
               ' t_wcr_rti_U64 nbo, t_wcr_rti_U64 nbi,' \
               ' t_wcr_rti_uagt** out, t_wcr_rti_uagt** in) {{\n' \
               '  t_wcr_wco_xconf cnf;\n' \
               '  t_wcr_rti_U64  idx;\n' \
               '  t_wcr_wco_status result = WCR_WCO_SUCCESS;\n' \
               '{pdcl}\n' \
               '{pini}\n' \
               '{xcnf}\n' \
               '  // initializate ope struct and return\n' \
               '  return wcr_wco_{oper}_init (ope, &cnf);\n' \
               '}}\n'.format(
                   name=func.getName(),
                   pdcl=self.getParamDcl(func),
                   pini=self.getParamIni(func),
                   xcnf=self.getXconfIni(func),
                   oper=oper)

    def getCode(self, func):
        """
        Get the code for a given function.

        :param func: the function for which code is needed
        :type  func: WhdbOperatorInstanceP
        """
        msg = 'Generating code for ' + func.getName()
        scmn_try_log(self.__pctx, mtxt=msg, mcmp='fog',
                        mcat=SCMN_LOG_INFO, mlvl=1)
        return '{hdr}\n' \
               '{inc}\n' \
               '{ini}\n'.format(
                   hdr=self.getHeader(func),
                   inc=self.getInclude(func),
                   ini=self.getInitFunc(func))
