# ----------------------------------------------------------------------------
# - wmfog-param-code                                                         -
# - wild cruncher: fog C code generator for parameters                       -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 piv                                              -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
..  module::      wmfog_param_code
   :platform:     Unix, Windows
   :synopsis:     Produce C code for the WmhdbParameter

.. moduleauthor:: piv <pierre.villalon@silkan.com>
"""
# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from wcr_core.modules.model.db.wmhdb_rti_type import WmhdbRtiType
from wcr_backend.modules.model.fog.wmfog_c_base import *


# ----------------------------------------------------------------------------
# - local util functions                                                     -
# ----------------------------------------------------------------------------

def wmfog_is_evaluated(p):
    """
    :return: True if teh parameter is evaluated
    :rtype: bool
    :param p: the parameter to test
    :type  p: WmhdbParameter
    """
    return p.rtiType is not None


# ----------------------------------------------------------------------------
# - class section                                                            -
# ----------------------------------------------------------------------------

class WmfogParamCode:
    """
    The WmfogFuncCode can produce C code for any parameter. It produces
    C code to declare and initialize WmhdbParameter objects.
    """

    def __init__(self, pctx):
        # get the logger
        self.__pctx = pctx

    def parameterStringToC(self, p):
        """
        :param p: the parameter to be intialized
        :type  p: WmhdbParameter
        """
        t = p.rtiType
        result = None
        if t.getDimension() == WmhdbRtiType.Scalar:
            if t.getType() == WmhdbRtiType.STR:
                result = '"' + p.getValue() + '"'
            else:
                result = p.getValue()
        elif t.getDimension() == WmhdbRtiType.Vector:
            values = p.getValue()
            # remove the vector/matrix delimiter
            values = values.replace("[", "")
            values = values.replace("]", "")
            l = values.split(';')
            result = "{" + ", ".join(l) + "}"
        elif t.getDimension() == WmhdbRtiType.Matrix:
            mesg = 'unsupported matrix type for C mapping'
            raise Exception(mesg)
        return result

    def initUdef(self, p):
        """
        :return: the C style string to intialize udef statically
        :rtype: str
        :param p: the parameter to be intialized
        :type  p: WmhdbParameter
        """
        t = p.rtiType
        etyp = t.getRtiEnumType()
        edim = t.getRtiEnumDim()
        xdim = str(t.getXdim()) if t.getXdim() else "0"
        ydim = str(t.getYdim()) if t.getYdim() else "0"
        return wmfog_init_c_struct([etyp, edim, xdim, ydim])

    def initUval(self, p):
        """
        :return: the C style string to intialize the unified value statically
        :rtype: str
        :param p: the parameter to be intialized
        :type  p: WmhdbParameter
        """
        result = self.getUagtField(p)
        result += " = "
        result += self.getParamDeclName(p)
        return result

    def initUagt(self, p):
        """
        :return: the C style string to intialize uagt statically
        :rtype: str
        :param p: the parameter to be intialized
        :type  p: WmhdbParameter
        """
        udef = self.initUdef(p)
        uval = self.initUval(p)
        return wmfog_init_c_struct([udef, uval])

    def initEvaluated(self, p):
        """
        :return: the C style string to intialize an evaluated parameter value
        :rtype: str
        :param p: the parameter to be intialized
        :type  p: WmhdbParameter
        """
        result = ""
        t = p.rtiType
        if t:
            result += t.getRtiTypeStr() + " " + self.getParamDeclName(p)
            if t.getDimension() != WmhdbRtiType.Scalar: result += "[]"
            result += " = " + self.parameterStringToC(p) + ";"
        else:
            raise ValueError("parameter " + self.getParamDeclName(p) +
                             " with no type")
        return result

    def getUagtField(self, p):
        """
        :return: the string to be use to initialize the value field of the uagt
        :return: str
        :param p: the parameter to be intialized
        :type  p: WmhdbParameter
        """
        t = p.rtiType
        result = "."
        if t.getDimension() == WmhdbRtiType.Scalar:
            result += "d_"
        else:
            result += "p_"
        result += WmhdbRtiType.getTypeName(t.getType()).lower()
        return result

    def getParamDeclName(self, p):
        """
        :return: the string to be use to declare a parameter value
        :return: str
        :param p: the parameter to be declared
        :type  p: WmhdbParameter
        """
        return "param_" + p.getName()
