# ----------------------------------------------------------------------------
# - wmrun_execute                                                            -
# - wild cruncher:wrt runtime execution environment                          -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
..  module::      wmrun_execute
   :platform:     Unix, Windows
   :synopsis:     Runtime execution environment

.. moduleauthor:: Amaury Darsch <amaury.darsch@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from silkan_common.scmn_process import scmn_execute, ScmnProcError
from silkan_common.scmn_file_path import *
from silkan_common.scmn_os_environment import *
from wcr_core.modules.model.wmmod_sgroup_execution import \
    WmmodSgroupExecutionFactoryP, WmmodSgroupExecutionP


# ----------------------------------------------------------------------------
# - public section                                                           -
# ----------------------------------------------------------------------------

def wmrun_get_sgrun_path():
    """
    Get the default scheduling group path
    """
    return scmn_path_join('${WCR_HOME}', 'bin', 'wcr-wmi-sgrun')


def wmrun_get_sgctl_path():
    """
    Get the default scheduling group control executable path
    """
    return scmn_path_join('${WCR_HOME}', 'bin', 'wcr-wmi-sgctl')


def wmrun_stopBus(pctx, strm):
    """
    Send a stop signal to the bus.

    :param pctx: the application context
    :type pctx: WcorProjectContextP
    :param strm: Process output/error stream handler
    :type strm: ScmnProcStreamHandlerP
    :return: True if the signal was sent succesfully
    :rtype: bool
    """
    path = pctx.getConfigurationValue('WCR_SGCTL_PATH') or \
           wmrun_get_sgctl_path()
    cmd = [scmn_expshenv(path), '-H']
    lib = scmn_path_join(scmn_expshenv('${WCR_HOME}'), 'lib')
    env = scmn_set_env_ldpath(lib)
    try:
        scmn_execute(cmd, env=env, strm=strm)
        return True
    except ScmnProcError:
        return False


# ----------------------------------------------------------------------------
# - class section                                                            -
# ----------------------------------------------------------------------------

class WmrunExecute:
    """
    The WmrunExecute is the runtime execute environment. The class provides
    the support for the execution of a scheduling group.
    """

    def __init__(self, pctx, prof=None):
        """
        Create an execute context

        :param pctx: the application context
        :type pctx: WcorProjectContextP
        :param prof: the project profile
        """
        # get the workspace/config table
        self.__context = pctx
        # save the project/profile/streams
        self.pnam = pctx.getProjectName()
        self.prof = prof or 'default'
        # set the runtime path
        path = pctx.getConfigurationValue('WCR_SGRUN_PATH') or \
               wmrun_get_sgrun_path()
        self.path = scmn_path_normalize(scmn_expshenv(path))
        # the bus name
        self.bnam = None
        """:type: str | None"""
        # default bus period
        self.time = None
        """:type: float | None"""
        # max number of bus cycles
        self.cnum = None
        """:type: int | None"""
        # the debug flag
        self.dbug = False

    def isValidRti(self):
        """
        Check if the runtime path is valid.

        :return: True iff the runtime path is valid
        :rtype: bool
        """
        return scmn_isfile(self.path)

    def getExecuteEnvironment(self, sgrp):
        """
        Create the runtime environment by scheduling group.

        :param sgrp: Scheduling group
        :param sgrp: str
        :return: Execution environment for the given scheduling group
        :rtype: {str:str}
        """
        # get the default environment
        lib = scmn_path_join(scmn_expshenv('${WCR_HOME}'), 'lib')
        env = scmn_set_env_ldpath(lib)
        # check for valid scheduling group
        if not sgrp: return env
        # check for valid workspace
        path = self.__context.getProjectWorkspace().getDirectoryPath(
            self.prof, sgrp)
        # update the environment
        return scmn_set_env_ldpath(path, env)

    def getExecuteCommand(self, sgrp):
        """
        Create the runtime command by scheduling group.

        :param sgrp: Scheduling group
        :type sgrp: str
        :return: Execution command for the given scheduling group
        :rtype: [str]
        """
        # check for valid scheduling group
        if not sgrp: return [self.path]
        # map the scheduling group command
        ccmd = [self.path, '-l', sgrp]
        # add the bus name
        if self.bnam is not None:
            ccmd.append('-b')
            ccmd.append(self.bnam)
            # add the bus period
        if self.time is not None:
            ccmd.append('-t')
            ccmd.append(str(self.time))
            # add the maximum cycle
        if self.cnum is not None:
            ccmd.append('-c')
            ccmd.append(str(self.cnum))
            # here it is
        return ccmd

    def execute(self, sgrp, strm=None):
        """
        Execute a runtime command by scheduling group.

        :param sgrp: Scheduling group
        :type sgrp: str
        :param strm: Process output/error stream handler
        :type strm: ScmnProcStreamHandlerP
        """
        # get the execute environment
        env = self.getExecuteEnvironment(sgrp)
        # get the execute command
        ccmd = self.getExecuteCommand(sgrp)
        # execute
        scmn_execute(ccmd, env=env, strm=strm)

    def setBusName(self, name):
        """
        Set the bus by name.

        :param name: Bus name to set
        :type name: str
        """
        self.bnam = name

    def setBusPeriod(self, bper):
        """
        Set the bus period.

        :param bper: the bus period to set
        :type bper: float
        """
        self.time = bper

    def setMaximumCycleCount(self, cnum):
        """
        Set the execution maximum cycle count.

        :param cnum: the cycle count to set
        :type cnum: int
        """
        self.cnum = cnum


class WmrunExecution(WmmodSgroupExecutionP):
    def __init__(self, exct, sgrp, pctx):
        self.__exct = exct
        self.__sgrp = sgrp
        self.__pctx = pctx

    def start(self, strm):
        self.__exct.execute(self.__sgrp, strm)

    def stop(self, strm):
        wmrun_stopBus(self.__pctx, strm)


class WmrunExecutionFactory(WmmodSgroupExecutionFactoryP):
    def createExecution(self, sgroup, busName, busPeriod, cycles, pctx,
                        prof=None):
        execute = WmrunExecute(pctx, prof)
        execute.setBusName(busName)
        execute.setBusPeriod(busPeriod)
        execute.setMaximumCycleCount(cycles)
        return WmrunExecution(execute, sgroup, pctx)
