# ----------------------------------------------------------------------------
# - wsaob-drs-builder                                                        -
# - wild cruncher:sfs xml drsbuilder                                         -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
..  module::      wsaob_drs_builder
   :platform:     Unix, Windows
   :synopsis:     Simulator drs builder

.. moduleauthor:: Amaury Darsch <amaury.darsch@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from xml.dom.minidom import Document
from wcr_core.modules.simulation.db.wshdb_binding_point_provider import \
    bindingPointToStr
from wcr_core.modules.simulation.db.wshdb_block import WshdbBlock


# ----------------------------------------------------------------------------
# - class section                                                           -
# ----------------------------------------------------------------------------

class WsaobDrsBuilder:
    """
    The WsaobDrsBuilder is the xml drs builder in the wild cruncher
    simulator environment. A drs is build by taking all busses and
    listing the bindings with the simulation variables.
    """

    def __init__(self, pctx, prof='default'):
        """
        Create a builder by context and profile

        :param pctx: the context
        :type pctx: WsmodSimulationProjectContextP
        :param prof: the profile name
        """
        self.__pctx = pctx
        # save the profile
        self.__prof = prof

    # noinspection PyUnusedLocal
    def addBusInRootNode(self, root, bus, xdoc):
        """
        Add a bus connection list in the drs

        :param root: the root node
        :param bus: the bus to process
        :type bus: WshdbBusInstance
        """
        # get the bus name and drs node
        bnam = bus.getBusName()
        # create a bus node
        bnod = xdoc.createElement("logicalBus")
        bnod.setAttribute("name", bnam)
        root.appendChild(bnod)
        # get the bus binding tables
        tlvl = self.__pctx.getProjectDB().getTopHierarchy()
        """:type: WshdbTopBlock"""
        prds = {k: b for k, b in
                self.__pctx.getBindingPointProvider().getProducersForBus(bnam)}
        csms = {k: b for k, b in
                self.__pctx.getBindingPointProvider().getConsumersForBus(bnam)}
        # iterate with the producers
        for k in prds:
            b = prds[k]
            # create a variable element
            snod = xdoc.createElement("scalar")
            # recreate name from binding table's producer entry
            vname = bindingPointToStr(b)
            snod.setAttribute("name", vname)
            snod.setAttribute("type", "F64")
            # add the source alias
            anod = xdoc.createElement("alias")
            anod.setAttribute("name", b.portname)
            anod.setAttribute("instance", b.instance.getName())
            snod.appendChild(anod)
            # get all consumers bound to the producer
            clst = [c for c, p in tlvl.getBindings().items() if p == k]
            # iterate in all consumer
            for c in clst:
                a = csms[c]
                anod = xdoc.createElement("alias")
                anod.setAttribute("name", a.portname)
                anod.setAttribute("instance", a.instance.getName())
                snod.appendChild(anod)
            # add the node to root
            bnod.appendChild(snod)

    # noinspection PyUnusedLocal
    def generate(self, inst, strm=None):
        """
        Generate the drs in the workspace

        :param inst: the instance to process
        :param strm: Process output/error stream handler
        :type  strm: ScmnProcStreamHandlerP
        """
        # check for ssh block
        if not isinstance(inst, WshdbBlock): return
        xdoc = Document()
        # add the root node by name
        root = xdoc.createElement("simulator")
        root.setAttribute("name", inst.getName())
        # loop in the bus
        bhlp = self.__pctx.getBindingPointProvider()
        for bus in inst.getBusInstances():
            self.addBusInRootNode(root, bus, xdoc)
            # bind root and get result
        xdoc.appendChild(root)
        text = xdoc.toprettyxml(indent="  ")
        # save in in the workspace
        name = inst.getName()
        if name is None or len(name) == 0: name = 'default'
        path = name + '.xml'
        self.__pctx.getProjectWorkspace().importText(
            path, self.__prof, None, text)
