# ----------------------------------------------------------------------------
# - wpaob-config-generator                                                   -
# - wild cruncher: tutor configuration file generator                        -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 samuel longchamps                                -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
..  module::      wpaob_config_generator
   :platform:     Unix, Windows
   :synopsis:     Tutor 2 config. file generator from system mapper project

.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
from xml.dom.minidom import Document

from wcr_core.modules.system.db.wphdb_block import WphdbBlockP
from wcr_core.modules.system.db.wphdb_connection_points import \
    WphdbEthernetXPoint
from wcr_core.modules.system.db.wphdb_host_instance import WphdbHostInstance
from wcr_core.modules.system.db.wphdb_phybus_instance import \
    WphdbPhyBusInstance
from wcr_core.modules.system.wpmod_distrib_service_default_types import \
    WPMOD_ETH_DISTSERV_PARAMCODE_INTERFACE, WPMOD_ETH_DISTSERV_CODE, \
    WPMOD_ETHERNET_DISTRIBUTION_SERVICE_TYPE, wpmod_retrieve_conpt_interface
from wcr_core.modules.system.wpmod_project_context import \
    wpmod_retrieve_lbus_obj
from wcr_core.modules.system.wpmod_project_context_api import \
    NoSimulationLoadedException
from wcr_core.modules.simulation.db.wshdb_binding_point_provider import \
    WshdbBindingPointProvider, bindingPointToStr, WshdbBindingPoint


class WpaobTutorConfigGenerator:
    """
    Generation of Tutor 2 configuration file from system mapper database in
    XML format.
    """

    ERROR_MESSAGE = 'Cannot generate configuration file: '

    DEFAULT_PROFILE_NAME = 'default'

    DEFAULT_GENERATED_FILETITLE = 'generated_tutor_config_{}'

    def __init__(self, pctx, prof=None):
        """
        Create generator by context and profile

        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param prof: Profile name
        :type prof: str | None
        """
        self.__pctx = pctx
        assert self.__pctx
        self.__prof = prof or self.DEFAULT_PROFILE_NAME
        self.__defaultCnt = 0

    # noinspection PyUnusedLocal
    def _genCheckValidParams(self, block, strm):
        """
        Apply checks to make sure the data is apporiately set for generation

        :param block: the block to generate from
        :type block: WphdbBlockP | None
        :param strm: Process output/error stream handler
        :type  strm: ScmnProcStreamHandlerP | None
        """
        if not isinstance(block, WphdbBlockP):
            raise ValueError(self.ERROR_MESSAGE +
                             "Block is not a system mapper implementation or "
                             "doesn't exist")

        hosts = [inst for inst in block.getInstances()
                 if isinstance(inst, WphdbHostInstance)]
        """:type: list[WphdbHostInstance]"""
        processes = [proc for hst in hosts for proc in hst.getProcesses()]
        """:type: list[WphdbProcessP]"""

        if len(processes) == 0:
            raise Exception(self.ERROR_MESSAGE +
                            'The project contains no process')

        hasModelSrv = False
        hasService = False
        hasDistServ = False
        for proc in processes:
            if hasModelSrv and hasService and hasDistServ:
                break
            if not hasModelSrv and len(proc.getModelMappings()) > 0:
                hasModelSrv = True
            if not hasService and len(proc.getServiceMappings()) > 0:
                hasService = True
            if not hasDistServ and \
                    len(proc.getDistributionServiceSockets()) > 0:
                hasDistServ = True

        if not (hasModelSrv or hasService or hasDistServ):
            raise Exception(self.ERROR_MESSAGE + "No service defined")

        if hasModelSrv:
            if self.__pctx.getLogicalSimulation() is None:
                raise NoSimulationLoadedException(self.ERROR_MESSAGE +
                                                  "No simulation loaded")

            mapMng = self.__pctx.getMappingManager()
            mapMng.rescanMappings()
            modelCnt = len(list(mapMng.getLogicalModels()))
            if modelCnt == 0:
                raise Exception(self.ERROR_MESSAGE +
                                'No models in the loaded in the simulation')

    # noinspection PyUnusedLocal
    def generate(self, block=None, strm=None, header=True):
        """
        Generate the configuration file as an XML file
        If no block is specified, the top block from the project context is
        used.

        :param block: the block to generate from
        :type block: WphdbBlockP | None
        :param strm: Process output/error stream handler
        :type  strm: ScmnProcStreamHandlerP | None
        :param header: True to write the XML header, False to skip it.
        :type header: bool
        """
        block = block or self.__pctx.getProjectDB().getTopHierarchy()
        self._genCheckValidParams(block, strm)

        hosts = [inst for inst in block.getInstances()
                 if isinstance(inst, WphdbHostInstance)]
        """:type: list[WphdbHostInstance]"""

        if len(hosts) == 0:
            return

        self.__defaultCnt = 0
        for hst in hosts:
            processes = hst.getProcesses()
            for proc in processes:
                self._generateProcess(proc, header)

    def _generateProcess(self, proc, header=True):
        """
        :param proc: Process to create a file from
        :type proc: WphdbProcessP
        :param header: True to write the XML header, False to skip it.
        :type header: bool
        """
        xdoc = Document()
        declaration = xdoc.toxml()

        # tup
        tup = xdoc.createElement('tup')
        xdoc.appendChild(tup)
        # - osTime
        tup.appendChild(xdoc.createElement('osTime'))
        # - instances
        inst = xdoc.createElement('instances')
        tup.appendChild(inst)
        # -- wcrService
        for modelMap in proc.getModelMappings():
            inst.appendChild(self._createWcrServiceElem(xdoc, modelMap))
        # -- service
        for srvMap in proc.getServiceMappings():
            # TODO: Service
            pass
        # -- distribService
        for distSrvSck in proc.getDistributionServiceSockets():
            inst.appendChild(self._createDistribServiceElem(
                xdoc, distSrvSck))

        text = xdoc.toprettyxml(indent="  ")
        if not header:
            text = text[len(declaration) + 1:]
        self._exportTextToFile(self._getProcessXMLFilename(proc), text)

    def _createDistribServiceElem(self, xdoc, distSrvSck):
        """
        :param xdoc: XML document used for creation
        :type xdoc: Document
        :param distSrvSck: Distribution service socket
        :type distSrvSck: WphdbDistributionServiceSocketP
        :return: XML element with adequate parameters
        :rtype: Element
        """
        key = distSrvSck.getDistributionService()
        name = distSrvSck.getSocketLabel()

        module = self.__pctx.getModule()
        """:type: WpmodSystemApplicationModuleP"""
        distSrvType = module.getDistributionServiceType(key)

        if distSrvType is not None:
            if distSrvType.getCode() == WPMOD_ETH_DISTSERV_CODE:
                return self._createEthDistribServiceElem(xdoc, distSrvSck)

        errorMsg = "Distribution service code is invalid ('{}')".format(key)
        raise Exception(self.ERROR_MESSAGE +
                        "Error while generating distributed "
                        "service instance '{}' - {}".format(name, errorMsg))

    # noinspection PyMethodMayBeStatic
    def _createEthDistribServiceElem(self, xdoc, distSrvSck):
        """
        :param xdoc: XML document used for creation
        :type xdoc: Document
        :param distSrvSck: Distribution service socket
        :type distSrvSck: WphdbDistributionServiceSocketP
        :return: XML element with adequate parameters
        :rtype: Element
        """
        distSrvType = WPMOD_ETHERNET_DISTRIBUTION_SERVICE_TYPE

        # -- ethDistribService
        distServ = xdoc.createElement('ethDistribService')
        distServ.setAttribute('name', distSrvSck.getSocketLabel())
        # --- parameters
        params = xdoc.createElement('parameters')
        distServ.appendChild(params)
        # ---- (parameters values)
        conPt = None
        for prm in distSrvType.getSocketParameters():
            paramName = prm.getName()
            paramVal = distSrvSck.getSocketParameterValue(paramName)
            if paramVal and isinstance(paramVal, str):
                paramElem = xdoc.createElement(paramName)
                params.appendChild(paramElem)

                # Interface
                if paramName == WPMOD_ETH_DISTSERV_PARAMCODE_INTERFACE:
                    conPt = wpmod_retrieve_conpt_interface(
                        distSrvSck.host, paramVal)
                    paramVal = conPt.getInterface() \
                        if isinstance(conPt, WphdbEthernetXPoint) else None

                    if not paramVal:
                        raise Exception(
                            self.ERROR_MESSAGE +
                            "Cannot retrieve interface linked to '{}', verify "
                            "that the connection point exist and has defined "
                            "parameters".format(paramName))

                cfgFilePathText = xdoc.createTextNode(paramVal)
                paramElem.appendChild(cfgFilePathText)
        # --- configFiles
        cfgFiles = xdoc.createElement('configFiles')
        distServ.appendChild(cfgFiles)
        # ---- configFilePath
        cfgFilePath = xdoc.createElement('configFilePath')
        cfgFiles.appendChild(cfgFilePath)
        cfgFilePathText = xdoc.createTextNode(
            self._createEthDistribSrvConfigFile(distSrvSck, conPt))
        cfgFilePath.appendChild(cfgFilePathText)

        return distServ

    def _createEthDistribSrvConfigFile(self, distSrvSck, conPt, header=True):
        """
        Create the ethernet distribution configuration file related to the
        passed distribution service

        :param distSrvSck: Distribution service socket
        :type distSrvSck: WphdbDistributionServiceSocketP
        :param conPt: Ethernet connection point
        :type conPt: WphdbEthernetXPoint
        :param header: True to write the XML header, False to skip it.
        :type header: bool
        :return: Filename of the exported file
        :rtype: str
        """
        filename = distSrvSck.host.getName() + '_' + \
            distSrvSck.getSocketLabel() + '_cfg.xml'

        phyBus = self._getConnectedPhysicalBus(distSrvSck.host, conPt)
        assert phyBus

        # Find ethernet variables
        ethVars = self._retrieveEthDistribSrvVariables(
            self._getProcessModels(distSrvSck.process),
            self._getLogBuses(phyBus))

        xdoc = Document()
        declaration = xdoc.toxml()
        # -- service
        srv = xdoc.createElement('service')
        xdoc.appendChild(srv)
        # --- variables
        varbls = xdoc.createElement('variables')
        srv.appendChild(varbls)
        # ---- ethVariable
        for ev in ethVars:
            varbls.appendChild(ev.toXmlElement(xdoc))
        # --- properties
        srv.appendChild(xdoc.createElement('properties'))

        text = xdoc.toprettyxml(indent="  ")
        if not header:
            text = text[len(declaration) + 1:]
        self._exportTextToFile(filename, text)

        return filename

    def _getConnectedPhysicalBus(self, host, conPt):
        """
        Get the physical bus linked to the host's connection point
        There must be only one physical bus connected, otherwise an exception
        is raised.

        :param host: Host instance
        :type host: WphdbHostInstance
        :param conPt: Connection point which must belong to the host
        :type conPt: WhdbConnectionPointP
        :return: The linked physical bus object, None if not found
        :rtype: WphdbPhyBusInstance | None
        """
        # Get the list of connections linked to the distributed service's
        # linked connection point
        topBlk = self.__pctx.getProjectDB().getTopHierarchy()
        connsForConPt = list(topBlk.getConnectionsForObject(host, conPt))

        # Function returning true if the connection's object selected is a
        # connected physical bus, knowing the connection passed has the host
        # either as start or end object
        conPhyBusValF = lambda con, start:\
            (con.getStartConnectionPoint() if start else
             con.getEndConnectionPoint()) is not conPt and \
            isinstance(con.getStartObject() if start else con.getEndObject(),
                       WphdbPhyBusInstance)
        connPhyBus = \
            [con.getStartObject() for con in connsForConPt
             if conPhyBusValF(con, True)] + \
            [con.getEndObject() for con in connsForConPt
             if conPhyBusValF(con, False)]
        """:type: list[WphdbPhyBusInstance]"""

        if len(connPhyBus) > 1:
            raise ValueError(
                "Multiple physical buses are connected to the connection point"
                " '{}'".format(conPt.getName()))

        # Knowing there is only one physical bus, we get the object directly
        return connPhyBus[0] if len(connPhyBus) != 0 else None

    def _getLogBuses(self, phyBus):
        """
        Get all the logical buses objects bound to a physical bus

        :param phyBus: Physical bus linked to the distributed service's
            connection point
        :type phyBus: WphdbPhyBusInstance
        :return: List of logical buses bound to the passed physical bus
        :rtype: list[WshdbBusInstance]
        """
        logBusBoundToPhyBusDict = wpmod_retrieve_lbus_obj(
            [phyBus], [self.__pctx.getLogicalSimulation()])
        boundLogBuses = logBusBoundToPhyBusDict.get(phyBus.getUUID(), [])
        return boundLogBuses

    def _getProcessModels(self, process):
        """
        Retrieve all models bound to a process

        :param process: Process from which is retrieved bound models
        :type process: WphdbProcessP
        :return: List of models bound to the passed process
        :rtype: list[WshdbModelInstance]
        """
        retMdlsList = []
        """:type: list[WshdbModelInstance]"""

        modelMngr = self.__pctx.getMappingManager()
        for mdlMp in process.getModelMappings():
            retMdl = modelMngr.getModelByUUID(mdlMp.uuid)
            if retMdl is not None:
                retMdlsList.append(retMdl)
        return retMdlsList

    def _retrieveEthDistribSrvVariables(self, procMdls, logBuses):
        """
        Retrieve distributed service variables by trying to match bindings
        (same a variables) found in logical buses to which the distributed
        service's connection point is linked by the intermediate of a physical
        bus with models present in the distributed service's process.

        Simply put:
        ->  If a model doesn't consume nor produce a variable, that is
            it is not referenced in any binding, no mention of it in the
            distributed service configuration file is necessary.
        ->  If it is produced by a model seen by one of the logical buses, it
            is set as a producer (even if another visible model consumes it).
        ->  If it is only consumed, then it is set as a consumer.

        IMPORTANT NOTE:
            Variables' flag 'producer' has an inverse meaning to the one
            present in the software. Therefore, consumers write 'true' and
            producers write 'false'.

        :param procMdls: List of models bound to the process which contains
            the distributed service
        :type procMdls: list[WshdbModelInstance]
        :param logBuses: List of logical buses containing bindings (variables)
        :type logBuses: list[WshdbBusInstance]
        :return: A list of ethernet variables ready to be exported
        :rtype: list[ethDistSrvVariable]
        """
        ethVars = []
        """:type: list[ethDistSrvVariable]"""

        lsim = self.__pctx.getLogicalSimulation()
        simPrjBndMngr = WshdbBindingPointProvider(lsim)
        simPrjBindings = lsim.getBindings()

        allBoundProds = {}
        """:type: dict[(UUID, UUID, str), WshdbBindingPoint]"""

        # Iterate through the logical buses and try matching bindings with the
        # models linked with the distributed service's process
        for lb in logBuses:
            busName = lb.getBusName()
            assert busName

            prodsOfLogBus = list(simPrjBndMngr.getProducersForInstance(lb))
            for prod in prodsOfLogBus:
                allBoundProds[prod[0]] = prod[1]
                boundMdl = prod[1].instance
                assert boundMdl

                for pcMdl in procMdls:
                    if pcMdl.getUUID() == boundMdl.getUUID():
                        ethVars.append(ethDistSrvVariable(
                            name=bindingPointToStr(prod[1]),
                            busName=busName,
                            producer=False))
                        break

            cnsmsOfLogBus = list(simPrjBndMngr.getConsumersForInstance(lb))
            for cons in cnsmsOfLogBus:
                for pcMdl in procMdls:
                    boundMdl = cons[1].instance
                    if pcMdl.getUUID() == boundMdl.getUUID():
                        # Because the variable is named after the producer,
                        # we retrieve the producer using the projet bindings
                        prod = simPrjBindings.get(cons[0])
                        assert prod

                        pdObj = allBoundProds.get(prod)
                        """:type: WshdbBindingPoint"""
                        assert pdObj

                        # If it's already a producer model locally in the
                        # process, we don't duplicate the entry since producers
                        # win over consumers
                        if pdObj.instance in procMdls:
                            continue

                        ethVars.append(ethDistSrvVariable(
                            name=bindingPointToStr(pdObj),
                            busName=busName,
                            producer=True))

        return ethVars

    def _createWcrServiceElem(self, xdoc, modelMap):
        """
        :param xdoc: XML document used for creation
        :type xdoc: Document
        :param modelMap: Model mapping of a wild cruncher service
        :type modelMap: WphdbModelMappingP
        :return: XML element with adequate parameters
        :rtype: Element
        """
        modelSim = self.__pctx.getMappingManager().getModelByUUID(
            modelMap.uuid)
        if modelSim is None:
            raise Exception(self.ERROR_MESSAGE +
                            "Model '{}' cannot be found (Simulation might not "
                            "be loaded correctly)".format(
                                modelMap.getModelName()))
        # -- wcrService
        wcrService = xdoc.createElement('wcrService')
        wcrService.setAttribute('name', modelMap.getModelName())
        # --- parameters
        params = xdoc.createElement('parameters')
        wcrService.appendChild(params)
        # --- lib
        lib = xdoc.createElement('lib')
        libText = xdoc.createTextNode(self._getEntryPtSchedGrp(
            modelSim, modelMap.getModelName()))
        lib.appendChild(libText)
        params.appendChild(lib)
        # -- configFiles
        wcrService.appendChild(xdoc.createElement('configFiles'))

        return wcrService

    # noinspection PyMethodMayBeStatic
    def _getEntryPtSchedGrp(self, modelSim, mName):
        """
        Returns a valid entry-point scheduling group, raises exception in case
        it is not defined or empty, or if the simulation model doesn't have
        a model defined

        :param modelSim: Simulation model
        :type modelSim: WshdbModelInstance
        :param mName: Model name as information in case of exception
            (has no impact on normal output)
        :type mName: str
        :return: The model entry-point scheduling group's name
        :rtype: str
        """
        model = modelSim.getModel()
        entryPt = model.getEntryPoint() if model else None
        if not entryPt:
            raise Exception(self.ERROR_MESSAGE +
                            "Model '{}' has incorrectly defined entry"
                            " point".format(mName))
        return entryPt

    def _getProcessXMLFilename(self, proc):
        """
        Return a valid XML filename from a process.
        If no valid name is found for the process, default value using
        DEFAULT_GENERATED_FILETITLE as file title is returned.

        :param proc: Process to create a file from
        :type proc: WphdbProcessP
        :return: Filename with title and extension
        :rtype: str
        """
        filetitle = proc.getName()
        if filetitle is None or len(filetitle) == 0:
            filetitle = self.DEFAULT_GENERATED_FILETITLE.format(
                self.__defaultCnt)
            self.__defaultCnt += 1
        return filetitle + '.xml'

    def _exportTextToFile(self, filename, text):
        """
        Export the text to a file of passed filename in the default workspace
        directory

        :param filename: Name of the file to be exported, including extension
        :type filename: str
        :param text: Text to be written in the file
        :type text: str
        """
        self.__pctx.getProjectWorkspace().importText(
            filename, self.__prof, None, text)


class ethDistSrvVariable:
    """
    Ethernet distribution service variable structure
    (member variables are made public intentionally)
    """

    CODE_TITLE = 'ethVariable'
    CODE_BUSNAME = 'busName'
    CODE_NAME = 'name'
    CODE_PRODUCER = 'producer'

    def __init__(self, name, busName, producer):
        """
        :param name: Name of the variable
        :type name: str
        :param busName: Name of the logical bus on which it is bound
        :type busName: str
        :param producer: True for producer, False for consumer
        :type producer: bool
        """
        self.name = name
        self.busName = busName
        self.producer = producer

    def toXmlElement(self, xdoc):
        """
        :param xdoc: Document which will own the element
        :type xdoc: Document
        :return: XML tree element
        :rtype: Element
        """
        ethVar = xdoc.createElement(self.CODE_TITLE)
        ethVar.setAttribute(self.CODE_BUSNAME, self.busName)
        ethVar.setAttribute(self.CODE_NAME, self.name)
        ethVar.setAttribute(self.CODE_PRODUCER,
                            'true' if self.producer else 'false')
        return ethVar