# ----------------------------------------------------------------------------
# - wbke-make-execute                                                        -
# - wild cruncher:wwm workspace makefile management                          -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wbke_make_execute
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher workspace makefile management

.. moduleauthor:: Amaury Darsch <amaury.darsch@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from silkan_common.scmn_file_path import *
from silkan_common.scmn_os_environment import *
from silkan_common.scmn_process import scmn_execute


# ----------------------------------------------------------------------------
# - workspace class                                                          -
# ----------------------------------------------------------------------------

class WbkeMakeExecute:
    """
    The BasWspMakeExecute is the makefile execute command class. The class is
    designed to run specific make file rule which are supported by the wcr
    environment in the context of an automatic build.
    """

    def __init__(self, pctx, prof='default'):
        """
        :param pctx: Application context
        :type pctx: WcorProjectContextP
        :param prof: Profile
        :type prof: str
        """
        #: Workspace
        self.__wspc = pctx.getProjectWorkspace()
        #: Profile
        self.__prof = prof
        #: make environment variables
        self.__env = None
        """:type: dict[str, str] | None"""
        #: the make command path
        self.__make = scmn_get_make()

    def addEnvironment(self, name, sval):
        """
        Add a make file environment value

        :param name: the environment name
        :param sval: the environment value
        """
        # check for none
        if (name is None) or (sval is None): return
        # check environment
        if self.__env is None: self.__env = {}
        # set name/value pair
        self.__env[name] = sval

    def execute(self, rule, sgrp, strm=None):
        """
        Execute a makefile command in a subprocess and the result string

        :param rule: Makefile rule to execute
        :type rule: str
        :param sgrp: Scheduling group to run rule for
        :type sgrp: str
        :param strm: Process output/error stream handler
        :type strm: ScmnProcStreamHandlerP
        """
        ccmd = self.getSchedulingMakeCommand(sgrp, rule)
        scmn_execute(ccmd, strm=strm)

    def getMakeCommand(self, pdir, rule):
        """
        Create a makefile command by path and rule name.

        :param pdir: the makefile directory path
        :type pdir: str
        :param rule: the rule to execute
        :type rule: str
        """
        # map the command
        ccmd = [self.__make, '-C', pdir]
        # add the environment variables
        if self.__env is not None:
            for name, value in self.__env.items():
                varg = name + '=' + value
                ccmd.append(varg)
        # check for a rule
        if rule is not None: ccmd.append(rule)
        return ccmd

    def getSchedulingMakeCommand(self, sgrp, rule):
        """
        Create a makefile command by scheduling group and rule name.

        :param sgrp: the scheduling group
        :type sgrp: str
        :param rule: the rule to execute
        :type rule: str
        """
        # get the directory path
        pdir = self.__wspc.getDirectoryPath(self.__prof, sgrp)
        # get the make command
        return self.getMakeCommand(pdir, rule)

    def doAllRule(self, sgrp, strm=None):
        """
        Execute a all rule in a makefile.

        :param sgrp: Scheduling group to run rule for
        :type sgrp: str
        :param strm: Process output/error stream handler
        :type strm: ScmnProcStreamHandlerP
        """
        self.execute('all', sgrp, strm)

    def doCleanRule(self, sgrp, strm=None):
        """
        Execute a clean rule in a makefile.

        :param sgrp: Scheduling group to run rule for
        :type sgrp: str
        :param strm: Process output/error stream handler
        :type strm: ScmnProcStreamHandlerP
        """
        self.execute('clean', sgrp, strm)

    def doCleanProject(self, strm=None):
        """
        Clean the complete project

        :param strm: Process output/error stream handler
        :type strm: ScmnProcStreamHandlerP
        """
        def onerror(func, cdir, info):
            strm.handleError(
                '{f} failed on {d} / {i}'.format(f=func, d=cdir, i=info))

        # get the project path
        path = self.__wspc.getProjectPath()
        if not scmn_isdir(path): return
        # remove the whole tree
        scmn_rmdir(path, onerror if strm is not None else None)
