# ----------------------------------------------------------------------------
# - wact-action-api                                                          -
# - wild cruncher: action api                                  core / action -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wact_action_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Action API

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod

WACT_ACTION_EVENT_COUNT = 6

WACT_ACTION_NO_EVENT, \
    WACT_ACTION_ENABLED_CHANGE_EVENT, \
    WACT_ACTION_EXECUTING_EVENT, \
    WACT_ACTION_EXECUTION_SUCCEEDED_EVENT, \
    WACT_ACTION_EXECUTION_CANCELLED_EVENT, \
    WACT_ACTION_EXECUTION_FAILED_EVENT = \
    (2 ** i for i in range(WACT_ACTION_EVENT_COUNT))

WACT_ACTION_EXECUTED_EVENT = \
    WACT_ACTION_EXECUTION_SUCCEEDED_EVENT | \
    WACT_ACTION_EXECUTION_CANCELLED_EVENT | \
    WACT_ACTION_EXECUTION_FAILED_EVENT


#: Code prefix for internal actions
WACT_WCR_ACTION_CODE = 'wcr'

#: Action flag indicating pot. database verification exceptions on execution
WACT_ACTION_W_DB_VERIFY = 'db-verify'

#: Action flag indicating potential database changes on execution
WACT_ACTION_W_DB_MODIFY = 'db-modify'

#: Action flag indicating potential workspace changes on execution
WACT_ACTION_W_WS_MODIFY = 'ws-modify'


class WactActionObserverP(metaclass=ABCMeta):
    """
    Protocol for action listeners.
    """

    @abstractmethod
    def actionNotify(self, actx, action, event, *args, **kwargs):
        """
        Callback function to indicate a change in the action's state.

        :param action: Action that triggered the notification
        :type action: WactActionP
        :param event: Event code
        :type event: int
        """
        raise NotImplementedError()


class WactObservableP(metaclass=ABCMeta):
    """
    Protocol for listenable actions.
    """

    @abstractmethod
    def registerActionObserver(self, observer):
        """
        Add an action listener.

        :param observer: Action listener to add
        :type observer: WactActionObserverP
        """
        raise NotImplementedError()

    @abstractmethod
    def unregisterActionObserver(self, observer):
        """
        Remove an action listener.

        :param observer: Action listener to remove
        :type observer: WactActionObserverP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WactActionP(WactObservableP):
    """
    Protocol for actions.
    """

    @abstractmethod
    def getName(self):
        """
        Get the name of the action.

        :return: Name of the action
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getDescription(self):
        """
        Get the description for the action.

        :return: Description for the action
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getParameters(self):
        """
        Get the action parameters.

        :return: Action parameters
        :rtype: Iterable[WactActionParameterP]
        """
        raise NotImplementedError()

    @abstractmethod
    def isEnabled(self, actx):
        """
        Whether the action can be executed at this time.

        :param actx: Action execution context
        :type: WactActionExecutionContextP
        :return: True iff the action can be currently executed
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def execute(self, actx, **paramValuePairs):
        """
        Execute the action with the given parameter values.

        :param actx: Action execution context
        :type actx: WactActionExecutionContextP
        :param paramValuePairs: Parameter values
        :type paramValuePairs: dict[str, object]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WactStateSelectionActionP(WactActionP):
    """
    Protocol for actions that set a single property.
    """

    @abstractmethod
    def getPossibleValues(self):
        """
        Get the possible values to choose from.

        :return: Possible values
        :rtype: Iterable[object]
        """
        raise NotImplementedError()

    @abstractmethod
    def getCurrentValue(self, actx):
        """
        Get the current value for the property, if any.

        :return: Current value for the property, or `None`
        :rtype: object
        :param actx: Action execution context
        :type actx: WactActionExecutionContextP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WactStateToggleActionP(WactStateSelectionActionP):
    """
    Protocol for actions that set a single property.
    """

    pass
