# ----------------------------------------------------------------------------
# - wact-action-parameter-type-api                                           -
# - wild cruncher: action parameter type api                   core / action -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wact_action_parameter_type_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Action parameter type API

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod


class WactActionParameterTypeP(metaclass=ABCMeta):
    """
    Protocol for action parameter types.
    """

    @abstractmethod
    def getName(self):
        """
        Get the name of the action parameter type.

        :return: Name of the action parameter type
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getDescription(self):
        """
        Get the description for the action parameter type.

        :return: Description for the action parameter type
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def validate(self, value):
        """
        Validate the given value.

        :param value: Value to validate
        :type value: object
        :return: `None` if validation was succesful, a list of warnings/errors
                 otherwise [(IsError, message)]
        :rtype: list[(bool, str)] | None
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WactStringableActionParameterTypeP(WactActionParameterTypeP):
    """
    Protocol for action parameter types that can be converted into/fro
    string.
    """

    @abstractmethod
    def parse(self, inStr):
        """
        Parse a given string into a value for this parameter.
        Will raise an exception if the string is not parseable.

        :param inStr: String to parse
        :type inStr: str
        :return: Parameter value
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def format(self, obj):
        """
        Format a given parameter value into a string.

        The return string is parseable into a parameter value.

        :param obj: Value to format
        :type obj: object
        :return: String-formatted version of the value
        :rtype: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WactActionPathParameterTypeP(WactStringableActionParameterTypeP):
    """
    Path parameter type.
    """
    pass


# noinspection PyAbstractClass
class WactActionDirectoryParameterTypeP(WactActionPathParameterTypeP):
    """
    Directory parameter type.
    """
    pass


# noinspection PyAbstractClass
class WactActionFileParameterTypeP(WactActionPathParameterTypeP):
    """
    File parameter type.
    """
    pass


# noinspection PyAbstractClass
class WactActionMimeFileParameterTypeP(WactActionFileParameterTypeP):
    """
    MIME-type based file parameter type.
    """

    @property
    @abstractmethod
    def mimeTypes(self):
        raise NotImplementedError()
