# ----------------------------------------------------------------------------
# - wact-action-parameter-types                                              -
# - wild cruncher: action parameter types                      core / action -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wact_action_parameter_types
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Action parameter types

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from silkan_common.scmn_file_path import scmn_isdir, scmn_isfile, \
    scmn_isdirempty, SCMN_VALID_PATH_REGEX
from silkan_common.scmn_mime_type import scmn_payload_mime_types
from silkan_common.scmn_utils import concatWithoutDuplicates
from wcr_core.action.wact_action_parameter_type_api import *
from wcr_core.common.wcmn_file_handler import WcmnFile
from wcr_core.db.whdb_block import WHDB_TOPLEVEL_BLOCK_CODE
from wcr_core.db_api.whdb_block_api import WhdbBlockP
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP, WhdbBlockInstanceP


class WactActionStringParameterType(WactStringableActionParameterTypeP):
    """
    String parameter type.
    """

    def getName(self):
        return 'string'

    def getDescription(self):
        return 'A string'

    def parse(self, inStr):
        return inStr

    def format(self, obj):
        if isinstance(obj, str):
            return obj
        else:
            raise ValueError('Object is not a string: ' + str(obj))

    def validate(self, value):
        return None


class WactActionIntParameterType(WactStringableActionParameterTypeP):
    """
    Integer paramater type.
    """

    def getName(self):
        return 'int'

    def getDescription(self):
        return 'An integer'

    def parse(self, inStr):
        return int(inStr)

    def format(self, obj):
        if isinstance(obj, int):
            return str(obj)
        else:
            raise ValueError('Object is not a integer value: ' + str(obj))

    def validate(self, value):
        return None


class WactActionBoolParameterType(WactStringableActionParameterTypeP):
    """
    Boolean parameter type.
    """

    def getName(self):
        return 'bool'

    def getDescription(self):
        return 'A boolean'

    def parse(self, inStr):
        return inStr.lower().startswith('t')

    def format(self, obj):
        if isinstance(obj, bool):
            return str(obj)
        else:
            raise ValueError('Object is not a boolean value: {}.'.format(obj))

    def validate(self, value):
        return None


class WactActionPathParameterType(WactActionPathParameterTypeP):
    """
    Path parameter type.
    """

    def getName(self):
        return 'path'

    def getDescription(self):
        return 'A path'

    def parse(self, inStr):
        if SCMN_VALID_PATH_REGEX.search(inStr):
            raise ValueError('Invalid input for path: {}.'.format(inStr))
        return inStr

    def format(self, obj):
        return obj

    def validate(self, value):
        return None


class WactActionDirectoryParameterType(WactActionPathParameterType,
                                       WactActionDirectoryParameterTypeP):
    """
    Directory parameter type.
    """

    def getName(self):
        return 'directory'

    def getDescription(self):
        return 'A directory'

    def validate(self, value):
        if scmn_isfile(value):
            return [(True, 'File exists at {}.'.format(value))]
        return None


class WactActionExistingDirectoryParameterType(
        WactActionDirectoryParameterType):
    """
    Existing directory parameter type.
    """

    def name(self):
        return 'existing-directory'

    def description(self):
        return 'An existing directory'

    def validate(self, value):
        if not scmn_isdir(value):
            return [(True, 'Directory {} does not exist.'.format(value))]
        return None


class WactActionEmptyDirectoryParameterType(WactActionDirectoryParameterType):
    """
    Empty directory parameter type.
    """

    def __init__(self, notEmptyMsg=None):
        self.__notEmptyMsg = notEmptyMsg

    def name(self):
        return 'empty-directory'

    def description(self):
        return 'An empty directory'

    def validate(self, value):
        if not scmn_isdirempty(value):
            return [(False, 'Directory {} is not empty.'.format(value))]
        return None


class WactActionFileParameterType(WactActionPathParameterType,
                                  WactActionFileParameterTypeP):
    """
    File parameter type.
    """

    def getName(self):
        return 'file'

    def getDescription(self):
        return 'A file'

    def parse(self, inStr):
        if SCMN_VALID_PATH_REGEX.search(inStr):
            raise ValueError('Invalid input for path: {}'.format(inStr))
        return WcmnFile(inStr)

    def format(self, obj):
        """
        :type obj: WcmnFile
        """
        if isinstance(obj, WcmnFile):
            return obj.path
        else:
            raise ValueError('Object is not a file: ' + str(obj))


# noinspection PyAbstractClass
class _WactActionMimeFileParameterTypeBase(WactActionFileParameterType,
                                           WactActionMimeFileParameterTypeP):
    """
    MIME-based file parameter type.
    """

    def getName(self):
        return ','.join(m.mimeType for m in self.mimeTypes) + '-file'

    def getDescription(self):
        return 'A {} file'.format(
            ', '.join(m.mimeType for m in self.mimeTypes))

    def validate(self, value):
        """
        :type value: WcmnFile
        """
        fmimes = set(value.mimeTypes)
        if fmimes.isdisjoint(set(self.mimeTypes)):
            return [
                (True, 'Invalid MIME-type for {} ({} vs. {}).'.format(
                    value, fmimes, self.mimeTypes))]
        return None


# noinspection PyAbstractClass
class WactActionSaveFileParameterTypeP(WactActionMimeFileParameterTypeP):
    pass


class WactActionSaveFileByMimeParameterType(
        _WactActionMimeFileParameterTypeBase,
        WactActionSaveFileParameterTypeP):
    """
    MIME-type based file parameter type.

    Only files of the given MIME-types are considered valid values.
    """

    def __init__(self, *mimeTypes):
        """
        :param mimeTypes: Possible MIME-types
        :type mimeTypes: list[ScmnMimeType]
        """
        self.__mimeTypes = mimeTypes

    @property
    def mimeTypes(self):
        return self.__mimeTypes[:]

    def validate(self, value):
        """
        :type value: WcmnFile
        :rtype: bool
        """
        if not any(m.isExtension(value.path) for m in self.mimeTypes):
            return [
                (True, 'Invalid MIME-type for {} ({}).'.format(
                    value, self.mimeTypes))]
        return None


# noinspection PyAbstractClass
class WactActionOpenFileParameterTypeP(WactActionMimeFileParameterTypeP):
    pass


class WactActionOpenFileByPayloadParameterType(
        _WactActionMimeFileParameterTypeBase,
        WactActionOpenFileParameterTypeP):
    """
    Payload-type based file parameter type.

    Only files that can provide one of the given payload types are
    considered valid values.
    """

    def __init__(self, payloadTypes):
        """
        :param payloadTypes: Payload-types
        :type payloadTypes: int
        """
        self.__payloadTypes = payloadTypes

    @property
    def mimeTypes(self):
        return scmn_payload_mime_types(self.__payloadTypes)


class WactActionOpenFileByHandlerParameterType(
        _WactActionMimeFileParameterTypeBase,
        WactActionOpenFileParameterTypeP):
    """
    File-handler based file parameter type.

    Only files that can be handled by one of the given file handlers are
    considered valid values.
    """

    def __init__(self, *handlers):
        """
        :param mimeTypes: File-handlers
        :type mimeTypes: list[WcmnFileHandler]
        """
        self.__fileHandlers = handlers

    @property
    def mimeTypes(self):
        return concatWithoutDuplicates(
            *[h.handledMimeTypes for h in self.__fileHandlers])


class WactActionDbTopBlockType(WactActionParameterTypeP):
    """
    Database top-level block parameter type.
    """

    def getName(self):
        return 'top-block'

    def getDescription(self):
        return 'A top-level block'

    def validate(self, value):
        if not isinstance(value, WhdbBlockP):
            return [
                (True, '{} is not a block.'.format(value.__class__.__name__))]
        if value.getEntityType() != WHDB_TOPLEVEL_BLOCK_CODE:
            return [
                (True, '{} is not a top-level block.'.format(value))]
        return None


class WactActionDbInstanceType(WactActionParameterTypeP):
    """
    Database instance parameter type.
    """

    def getName(self):
        return 'instance'

    def getDescription(self):
        return 'An instance'

    def validate(self, value):
        if not isinstance(value, WhdbInstanceP):
            return [
                (True, '{} is not an instance.'.format(value))]
        return None


class WactActionDbHierarchyPathType(WactActionParameterTypeP):
    """
    Database hierarchy path parameter type.
    """

    def getName(self):
        return 'db-hierarchy-path'

    def getDescription(self):
        return 'A DB hierarchy path'

    def validate(self, value):
        if not (isinstance(value, list) or isinstance(value, tuple)):
            return [
                (True, '{} is not a path.'.format(value.__class__.__name__))]
        if any(not isinstance(e, WhdbBlockInstanceP) for e in value):
            return [
                (True, '{} is not a valid path.'.format(value))]
        return None
