# ----------------------------------------------------------------------------
# - wact-verify-db-action                                                    -
# - wild cruncher: database verify action                     core / actions -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wact_verify_db_action
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Database verify action

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.action.wact_action import WactActionBase
from wcr_core.project.wprj_project_verifier import WprjProjectVerifier


class WactDatabaseVerificationException(Exception):
    def __init__(self, result, verifier):
        """
        :param result: Project verification result
        :type result: WprjVerificationResult
        :param verifier: Project verifier
        :type verifier: WprjProjectVerifier
        """
        self.__result = result
        self.__verifier = verifier

    def getResult(self):
        """
        Get the result leading to the exception.

        :return: Project verification result
        :rtype: WprjVerificationResult
        """
        return self.__result

    def getVerifier(self):
        """
        Get the project verifier that produced the result.

        :return: Project verifier
        :rtype: WprjProjectVerifier
        """
        return self.__verifier

    def __str__(self):
        return 'WactDatabaseVerificationException ({})'.format(
            self.__result.allErrors)

    def __repr__(self):
        return str(self)


class _WactVerifyDatabaseAction(WactActionBase):
    """
    Project action to display the database verification dialog.
    """

    def __init__(self):
        super().__init__('Verify Database', 'Verify Database for Consistency')

    def _isProtocolException(self, e):
        return isinstance(e, WactDatabaseVerificationException)

    def isEnabled(self, actx):
        """
        :type actx: WactActionExecutionContextP
        """
        if not super().isEnabled(actx):
            return False

        return actx.getProjectContext() is not None

    def _doExecute(self, actx, paramValuePairs):
        """
        :type actx: WactActionExecutionContextP
        """
        actx.statusMessage('Verifying database consistency ...')
        wact_verifyProjectDatabase(actx.getProjectContext())
        actx.statusMessage('Succesfully verified database consistency')


def wact_verifyProjectDatabase(pctx):
    """
    Verify the given project.

    Raises an exception if the project contains errors.

    :param pctx: Project context
    :type pctx: WcorProjectContextP
    """
    pdb = pctx.getProjectDB()
    module = pctx.getModule()
    vrfr = WprjProjectVerifier(module.createHierarchyVerifier(pctx),
                               module.getProjectVerifierModules())
    result = vrfr.verifyProjectDatabase(pdb)
    if result.hasErrors():
        raise WactDatabaseVerificationException(result, vrfr)


#: Database verification action
WACT_VERIFY_DATABASE_ACTION = _WactVerifyDatabaseAction()
