# ----------------------------------------------------------------------------
# - wcmn-file-handler                                                        -
# - wild cruncher: file handler                                core / common -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcmn_file_handler
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher - file handler base classes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from functools import reduce
from abc import ABCMeta
import operator
import traceback

from silkan_common.scmn_mime_type import *
from silkan_common.scmn_file_path import scmn_expshenv, scmn_isfile
from silkan_common.scmn_log import SCMN_LOG_INFO


class WcmnFile(object):
    """
    Representation of a file (path and MIME type).
    """

    def __init__(self, path):
        """
        :param path: Path of the file
        :type path: str
        """
        #: Path of the file
        self.__path = scmn_expshenv(path)
        #: MIME types
        self.__mimeTypes = frozenset(scmn_mime_types(self.__path))

    @property
    def mimeTypes(self):
        """
        Get the file's MIME types.

        :return: MIME types of the file
        :rtype: Iterable[ScmnMimeTypeP]
        """
        return self.__mimeTypes

    @property
    def path(self):
        """
        Get the file's path.

        :return: Path of the file
        :rtype: str
        """
        return self.__path

    def __eq__(self, other):
        return isinstance(other, WcmnFile) and other.path == self.path

    def __str__(self):
        return self.__path


class WcmnReadPayloadHandler(metaclass=ABCMeta):
    """
    Protocol for import payload handlers.
    """

    @property
    @abstractmethod
    def handledPayloadTypes(self):
        """
        Get the payload types handled by the handler.

        :return: Payload types handled by the handler
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def handlePayload(self, payloadType, mimeType, payload):
        """
        Handle payload of a given type.

        :param payloadType: Type of the payload
        :type payloadType: int
        :param mimeType: MIME-type of the payload
        :type mimeType: str
        :param payload: Payload or payload function
        :type payload: object
        """
        raise NotImplementedError()


class WcmnReader(metaclass=ABCMeta):
    """
    Protocol for content readers.

    Content readers read content of specified MIME types, either from files
    or from strings, and call payload handlers on the read data.
    """

    @property
    @abstractmethod
    def name(self):
        """
        Get the reader's name.

        :return: Name of the reader
        :rtype: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def readableMimeTypes(self):
        """
        Get the MIME types read by the reader.

        :return: MIME types read by the reader
        :rtype: Iterable[ScmnMimeTypeP]
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def payloadTypesRead(self):
        """
        Get the payload types the reader can read.

        :return: Payload types read
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def readsPayloadType(self, payloadType):
        """
        Whether the reader can read the given payload type.

        :return: True iff the reader can read the given payload type
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def readText(self, text, payloadHandler, xctx, types=0):
        """
        Read data from the given text string, and call the given payload
        handler with the generated payload.
        The reader will only call the payload handler for either the types
        provided, if given, or the types indicated by the payload handler.
        The payload handler will be called once per payload item.

        :param text: Text to read payload from
        :type text: str
        :param payloadHandler: Callback function to call with the payload
        :type payloadHandler: WcmnReadPayloadHandler
        :param xctx: Context
        :type xctx: WcorProjectContextP
        :return: Types read
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def readFile(self, file, payloadHandler, xctx, types=0):
        """
        Read data from the given file, and call the given payload
        handler with the generated payload.
        The reader will only call the payload handler for either the types
        provided, if given, or the types indicated by the payload handler.
        The payload handler will be called once per payload item.

        :param file: File to read payload from
        :type file: WcmnFile
        :param payloadHandler: Callback function to call with the payload
        :type payloadHandler: WcmnReadPayloadHandler
        :param xctx: context
        :type xctx: WcorContextP
        :return: Types read
        :rtype: int
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WcmnReaderBase(WcmnReader):
    """
    Abstract base class for connent readers.
    """

    def __init__(self, mimeTypes, payloadTypes):
        """
        :param mimeTypes: MIME types read by this reader
        :type mimeTypes: Iterable[ScmnMimeTypeP]
        :param payloadTypes: Payload types generated by this reader
        :type payloadTypes: int
        """
        self.__mimeTypes = frozenset(mimeTypes)
        self.__payloadTypes = payloadTypes

    @property
    def readableMimeTypes(self):
        return self.__mimeTypes

    def readsPayloadType(self, contentType):
        return self.__payloadTypes & contentType

    @property
    def payloadTypesRead(self):
        return self.__payloadTypes

    def readText(self, text, payloadHandler, xctx, types=None):
        return SCMN_PAYLOAD_NONE

    def readFile(self, file, payloadHandler, xctx, types=None):
        return SCMN_PAYLOAD_NONE


# noinspection PyAbstractClass
class WcmnFileReaderBase(WcmnReaderBase):
    def __init__(self, mimeTypes, payloadTypes):
        """
        :param mimeTypes: MIME types read by this reader
        :type mimeTypes: Iterable[ScmnMimeTypeP]
        :param payloadTypes: Payload types generated by this reader
        :type payloadTypes: int
        """
        super().__init__(mimeTypes, payloadTypes)


class WcmnPlainTextFileReader(WcmnFileReaderBase):
    """
    Content reader for non-interpreted plain-text content.
    """

    #: MIME types read
    MIME_TYPES = [SCMN_MIME_TYPE_TEXT]

    def __init__(self):
        super().__init__(self.MIME_TYPES, SCMN_PAYLOAD_TEXT)

    @property
    def name(self):
        return 'Plaintext'

    def readFile(self, file, handler, xctx, types=None):
        types = types if types is not None else handler.handledPayloadTypes

        if types & SCMN_PAYLOAD_TEXT:
            handler.handlePayload(SCMN_PAYLOAD_TEXT,
                                  SCMN_MIME_ID_TEXT,
                                  self.__readPlainText(file.path))
            return SCMN_PAYLOAD_TEXT
        return SCMN_PAYLOAD_NONE

    def readText(self, text, handler, xctx, types=None):
        types = types if types is not None else handler.handledPayloadTypes

        if types & SCMN_PAYLOAD_TEXT:
            handler.handlePayload(SCMN_PAYLOAD_TEXT,
                                  SCMN_MIME_ID_TEXT,
                                  text)
            return SCMN_PAYLOAD_TEXT
        return SCMN_PAYLOAD_NONE

    def __readPlainText(self, path):
        with open(path, 'r') as f:
            return f.read()


class WcmnFileHandler(metaclass=ABCMeta):
    """
    Protocol for file handlers.

    File handlers handle files with exposing the way how the files are handled.
    """

    @abstractmethod
    def canHandleFile(self, file):
        """
        Whether the file handler can handle the given file.

        :param file: File to test for ability to handle
        :type file: WcmnFile
        :return: True iff the handler can handle the given file
        :rtype: bool
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def handledMimeTypes(self):
        """
        Get the MIME types handled by this handler.

        :return: MIME types handled by this handler
        :rtype: Iterable[ScmnMimeTypeP]
        """
        raise NotImplementedError()

    @abstractmethod
    def handleFile(self, file, xctx):
        """
        Handle the given file.

        :param file: File to handle
        :type file: WcmnFile
        :param xctx: Global Context
        :type xctx: WcorContextP
        :return: Whether the file could be handled
        :rtype: bool
        """
        raise NotImplementedError()

    def __str__(self):
        return self.__class__.__name__


class WcmnDefaultFileHandler(WcmnFileHandler):
    """
    Default file handler implementation.

    The reader uses the provided payload handler to determine which MIME
    types to read and which payload types to generate.
    """

    def __init__(self, payloadHandler, contentTypes=0):
        """
        :param payloadHandler: Handler for read payloads
        :type payloadHandler: WcmnReadPayloadHandler
        :param contentTypes: Content types to limit reading to
                             (o/w as specified by payload handler)
        :type contentTypes: int
        """
        #: Payload handler
        self.__payloadHandler = payloadHandler
        #: Content types to read
        self.__types = contentTypes or payloadHandler.handledPayloadTypes

    def _types(self, file):
        """
        Get all payload types that could theoretically read from
        the given file.

        :param file: File to test for payload types
        :type file: WcmnFile
        :return: Payload types that can be read from the file
        :rtype: int
        """
        readers = [r for m in file.mimeTypes for r in m.readers]
        return reduce(operator.ior, (r.payloadTypesRead for r in readers), 0)

    def handleFile(self, file:WcmnFile, xctx):
        if not scmn_isfile(file.path):
            raise ValueError('No such file: ' + file.path)

        types = self.__types

        for mimeType in file.mimeTypes:

            for reader in mimeType.readers:
                # noinspection PyBroadException
                try:
                    types &= ~reader.readFile(file, self.__payloadHandler,
                                              xctx, types)

                except BaseException as e:
                    msg = 'Reader \'{r}\' failed to read {f}: {e}'.format(
                        r=reader.name, f=file.path, e=e)
                    xctx.addMessage(
                        msg, 'fileio', SCMN_LOG_INFO,
                        trce=traceback.format_exc())
                    traceback.print_exc()
                if types == SCMN_PAYLOAD_NONE:
                    break

        return types == SCMN_PAYLOAD_NONE

    @property
    def handledMimeTypes(self):
        return scmn_payload_mime_types(self.__types)

    def canHandleFile(self, file):
        return bool(self._types(file) & self.__types)


# noinspection PyAbstractClass
class WcmnMimeFileHandlerBase(WcmnFileHandler):
    """
    Abstract base file handler for MIME-type based file handlers.
    """

    def __init__(self, *mimeTypes):
        """
        :param mimeTypes: MIME types handled by this handler
        :type mimeTypes: Iterable[ScmnMimeTypeP]
        """
        self.__mimeTypes = tuple(mimeTypes)

    @property
    def handledMimeTypes(self):
        return self.__mimeTypes

    def canHandleFile(self, file:WcmnFile):
        if not set(file.mimeTypes).isdisjoint(self.__mimeTypes):
            return True
        covars = scmn_covariant_mime_types(file.mimeTypes)
        return not covars.isdisjoint(self.__mimeTypes)


class WcmnMimeFileHandler(WcmnMimeFileHandlerBase):
    """
    File handler that handles a given set of MIME-types and passes the
    generated payloads to the given payload handler.
    """

    def __init__(self, payloadHandler, *mimeTypes):
        """
        :param payloadHandler: Handler for read payloads
        :type payloadHandler: WcmnReadPayloadHandler
        :param mimeTypes: MIME types handled by this handler
        :type mimeTypes: Iterable[ScmnMimeTypeP]
        """
        super().__init__(*mimeTypes)

        self.__payloadHandler = payloadHandler

    def handleFile(self, file, xctx):
        if not scmn_isfile(file.path):
            raise ValueError('No such file: ' + file.path)
        types = self.__payloadHandler.handledPayloadTypes
        for mimeType in self.handledMimeTypes:
            for reader in mimeType.readers:
                # noinspection PyBroadException
                try:
                    types &= ~reader.readFile(file, self.__payloadHandler,
                                              xctx, types)
                except Exception as e:
                    msg = 'Reader \'{r}\' failed to read {f}: {e}'.format(
                        r=reader.name, f=file.path, e=e)
                    xctx.addMessage(
                        msg, 'fileio', SCMN_LOG_INFO,
                        trce=traceback.format_exc())
                    traceback.print_exc()
                if types == SCMN_PAYLOAD_NONE:
                    return True

        return types == SCMN_PAYLOAD_NONE


def wcmn_handleFile(file, gctx, *handlers):
    """
    Handle the given file with the first suitable file handler.

    :param file: File to be handled
    :type file: WcmnFile
    :param gctx: Global context
    :type gctx: WcorGlobalContextP
    :param handlers: File handlers
    :type handlers: Iterable[WcmnFileHandler]
    """
    if file is None: return
    for handler in handlers:
        if handler.canHandleFile(file):
            handler.handleFile(file, gctx)
            break


WCMN_PLAIN_TEXT_READER = WcmnPlainTextFileReader()

scmn_add_reader(SCMN_MIME_ID_TEXT, WCMN_PLAIN_TEXT_READER)
