# ----------------------------------------------------------------------------
# - wcmn_graph_layouter                                                      -
# - wild cruncher: layouter                                    core / common -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 samuel longchamps                                -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcmn_graph_layouter
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graph layouter

.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
import math
from abc import ABCMeta, abstractmethod

# Global values
WCMN_GRAPH_LAYOUTER_DEFAULT_SPACING = 200

# -- Layouter abstract base class


class WcmnGraphLayouterP(metaclass=ABCMeta):
    """
    Layouter base class
    """

    @abstractmethod
    def applyLayout(self, graph):
        """
        Organize instances positions referenced in the passed graph object
        according to layout implementation

        :param graph: Object representing a set of nodes and edges between them
        :type graph: _WcmnLayouterGraph
        """
        raise NotImplementedError()


# -- Classes for representation of layouter-specific graph, node and edge

class _WcmnLayouterGraph():
    """
    Most basic representation of a graph, that is a set of nodes linked
    together by edges.
    The class does not intend to manage the objects contained in the list,
    but rather to assist in searching and classifying them.
    """

    def __init__(self, nodes, edges):
        """
        :type nodes: list[_WcmnLayouterNode]
        :type edges: list[_WcmnLayouterEdge]
        """
        if nodes is None:
            raise ValueError('Cannot create graph: The nodes list is None')
        if edges is None:
            raise ValueError('Cannot create graph: The edges list is None')

        self.__nodes = nodes
        """:type: list[_WcmnLayouterNode]"""
        self.__edges = edges
        """:type: list[_WcmnLayouterEdge]"""

    def getNodes(self):
        """
        Get the list of nodes

        :rtype: list[_WcmnLayouterNode]
        """
        return self.__nodes

    def getEdges(self):
        """
        Get the list of edges

        :rtype: list[_WcmnLayouterEdge]
        """
        return self.__edges

    def getGraphNodeAndEdgeCount(self):
        """
        Convenient method to get the length of both nodes and edges

        :return: Tupple of int representing in order the number of nodes and
                 the number of edges
        :rtype: int, int
        """
        return len(self.__nodes), len(self.__edges)

    def flagAllNodesOriginal(self):
        """
        Flag all the nodes in the graph as having original position
        """
        for node in self.__nodes:
            node.flagPositionAsOriginal()

    def copyObject(self):
        """
        Copy the current graph object with references to the same lists.

        :rtype: _WcmnLayouterGraph
        """
        return _WcmnLayouterGraph(self.__nodes, self.__edges)

    def deepCopyObject(self):
        """
        Copy the graph object to a new one containing copies of the nodes and
        edges linking the new nodes objects together.
        Important: The nodes of the new graph objects are themselves new
                   objects, therefore the copy operation of a large graph can
                   be costly. They are not only references to the same objects.
                   For a reference copy object containing the same lists
                   see copyObject
        :return: The deep-copied graph, None if an error (miscopy) has occured
        :rtype: _WcmnLayouterGraph | None
        """
        nodesCount, edgesCount = self.getGraphNodeAndEdgeCount()

        newNodeList = []
        """:type: list[WcmnNode]"""
        for i in range(nodesCount):
            newNodeList.append(self.__nodes[i].copyObject())

        newEdgeList = []
        """:type: list[_WcmnLayouterEdge]"""
        for i in range(edgesCount):
            startObject = None
            """:type: _WcmnLayouterNode"""
            endObject = None
            """:type: _WcmnLayouterNode"""

            for n in range(nodesCount):
                if self.__edges[i].getStartNode() is self.__nodes[n]:
                    startObject = newNodeList[n]
                if self.__edges[i].getEndNode() is self.__nodes[n]:
                    endObject = newNodeList[n]
                if startObject and endObject:
                    break

            # If they both haven't been found there must be a problem with the
            # node list copied, we return None immediately
            if not (startObject and endObject):
                return None

            newEdgeList.append(_WcmnLayouterEdge(startObject, endObject))

        newGraph = _WcmnLayouterGraph(newNodeList, newEdgeList)

        newNodesCount, newEdgesCount = newGraph.getGraphNodeAndEdgeCount()

        if newNodesCount == nodesCount and newEdgesCount == edgesCount:
            return newGraph
        else:
            return None

    def getConnectedNodes(self, node, onlyDifferent=False,
                          singleOccurence=False):
        """
        Search through edges for the passed node and returns a list containing
        the nodes to which the node was found to be connected by an edge from
        the graph.
        Multiple occurences of the same node is expected: one per edge in fact.

        :param node: Node object to find connected nodes to
        :type node: _WcmnLayouterNode
        :param onlyDifferent: Flag which allows to return only nodes connected
                              to the passed node other than itself
        :type onlyDifferent: bool
        :param singleOccurence: Flag which allows to return a list which
                                contains only unique nodes
        :type singleOccurence: bool
        :rtype: list[_WcmnLayouterNode]
        """
        if not node:
            raise ValueError(
                'Cannot look for connected nodes: The node is None')

        connectedNodes = []
        """:type: list[_WcmnLayouterNode]"""

        for edge in self.__edges:
            if edge.getStartNode() is node:
                if (onlyDifferent and
                            edge.getStartNode() is edge.getEndNode()) or \
                        (singleOccurence and
                         connectedNodes.count(edge.getStartNode()) >= 1):
                    continue
                else:
                    connectedNodes.append(edge.getEndNode())
            elif edge.getEndNode() is node:
                if (onlyDifferent and
                            edge.getStartNode() is edge.getEndNode()) or \
                        (singleOccurence and
                                 connectedNodes.count(edge.getEndNode()) >= 1):
                    continue
                else:
                    connectedNodes.append(edge.getStartNode())

        return connectedNodes

    def printContent(self, nodesOnly=False):
        """
        Print the content of the graph in the console (for debug)

        :param nodesOnly: Flag which allows to print only the nodes
        :type nodesOnly: bool
        """
        print("Content of graph ID=%d" % id(self))
        nodesNum, edgesNum = self.getGraphNodeAndEdgeCount()

        print("Nodes (%d)" % nodesNum)
        for i, node in zip(range(nodesNum), self.__nodes):
            x, y = node.getPosition()
            print("   Node #%d - (%.6f, %.6f)" % (i, x, y))

        if not nodesOnly:
            print("Edges (%d)" % edgesNum)
            for i, edge in zip(range(edgesNum), self.__edges):
                startNode = edge.getStartNode()
                endNode = edge.getEndNode()
                index1 = index2 = -1
                for j, node in zip(range(nodesNum), self.__nodes):
                    if startNode is node:
                        index1 = j
                    if endNode is node:
                        index2 = j
                    if index1 != -1 and index2 != -1:
                        break

                print("   Edge #%d - (Node #%d) <--> (Node #%d)" % (
                    i, index1, index2))


class _WcmnLayouterNode(object):
    """
    Generic node used in graphs
    Note that the position values are in floating point type
    """

    def __init__(self, obj=None, initial=False, final=False, x=0, y=0):
        """
        :param x: X-axis position of the node
        :type x: float
        :param x: Y-axis position of the node
        :type y: float
        :param obj: Object to which the node relates to
        :param initial: Flag that indicates if the node is initial in the
            generic sense.
        :type initial: bool
        :param final: Flag that indicates if the node is initial in the
            generic sense.
        :type final: bool

        """
        self.__x = x or 0
        """:type: float"""
        self.__y = y or 0
        """:type: float"""
        self.__netForceX = 0
        """:type: float"""
        self.__netForceY = 0
        """:type: float"""
        self.__velocityX = 0
        """:type: float"""
        self.__velocityY = 0
        """:type: float"""
        self.__originalPosition = True
        """:type: bool"""
        self.__initial = initial
        """:type: bool"""
        self.__final = final
        """:type: bool"""
        self.__obj = obj

    # -- Getters
    def isInitial(self):
        return True if self.__initial else False

    def isFinal(self):
        return True if self.__final else False

    def getObj(self):
        return self.__obj

    def getPosition(self):
        """
        :rtype: (float, float)
        """
        return self.__x, self.__y

    def getNetForce(self):
        """
        :rtype: (float, float)
        """
        return self.__netForceX, self.__netForceY

    def getVelocity(self):
        """
        :rtype: (float, float)
        """
        return self.__velocityX, self.__velocityY

    # -- Setters

    def flagPositionAsOriginal(self):
        """
        Sets the current position as the original one
        """
        self.__originalPosition = True

    def setPosition(self, x, y):
        """
        Position setter. Note: None values are interpreted as 0.

        :type x: float | None
        :type y: float | None
        """
        x = x or 0
        y = y or 0

        if x != self.__x or y != self.__y:
            self.__originalPosition = False

        self.__x = x
        self.__y = y

    def setVelocity(self, velocityX, velocityY):
        """
        :type velocityX: float | None
        :type velocityY: float | None
        """
        self.__velocityX = velocityX or 0
        self.__velocityY = velocityY or 0

    def setNetForce(self, netForceX, netForceY):
        """
        :type netForceX: float | None
        :type netForceY: float | None
        """
        self.__netForceX = netForceX or 0
        self.__netForceY = netForceY or 0

    # -- General methods

    def hasPositionChanged(self):
        """
        return: True if the position has not been modified, False otherwise
        rtype: bool
        """
        return not self.__originalPosition

    def addNetForce(self, netForceX, netForceY):
        """
        :param netForceX: float
        :param netForceY: float
        """
        self.__netForceX += netForceX or 0
        self.__netForceY += netForceY or 0

    def translate(self, x, y):
        """
        Adds the (x,y) position values to the current ones.
        None are interpreted as 0.

        :type x: float | None
        :type y: float | None
        """
        x = x or 0
        y = y or 0

        if x != 0 or y != 0:
            self.__originalPosition = False

        self.__x += x
        self.__y += y

    def getDistance(self, node):
        """
        :return: Distance between the current and passed nodes
        :type node: _WcmnLayouterNode
        :rtype: float
        """
        if not node:
            raise ValueError("Cannot compute distance: Node is None")

        return math.sqrt(self.getSqrDistance(node))

    def getSqrDistance(self, node):
        """
        :return: Squared distance between the current and passed nodes
        :type node: _WcmnLayouterNode
        :rtype: float
        """
        if not node:
            raise ValueError("Cannot compute squared distance: Node is None")

        xc, yc = self.getPosition()
        x, y = node.getPosition()
        return (xc - x) ** 2 + (yc - y) ** 2

    def copyObject(self, coordinatesOnly=False):
        """
        Copy the current node and create a new one with the same exact memmber
        variables
        Note: Original coordinate flag is set to True since the node is a new
        object

        :param coordinatesOnly: Flag which allows to bypass the net force and
                                velocity member variables value copy
        :rtype: _WcmnLayouterNode
        """
        newNode = _WcmnLayouterNode(self.__obj, self.__initial,
                                    self.__final, *self.getPosition())
        if not coordinatesOnly:
            nFX, nFY = self.getNetForce()
            velX, velY = self.getVelocity()
            newNode.setNetForce(nFX, nFY)
            newNode.setVelocity(velX, velY)

        return newNode


class _WcmnLayouterEdge(object):
    """
    Generic edge used in graphs
    """

    def __init__(self, nodeStart, nodeEnd):
        """
        :type nodeStart: _WcmnLayouterNode
        :type nodeEnd: _WcmnLayouterNode
        """
        if not nodeStart:
            raise ValueError('Cannot create edge: starting node is None')
        if not nodeEnd:
            raise ValueError('Cannot create edge: ending node is None')

        self.__nodeStart = nodeStart
        self.__nodeEnd = nodeEnd

    def getLength(self):
        """
        Returns the the distance between the edge's two nodes' center.

        :return: Direct length of the edge
        :rtype: float
        """
        return self.__nodeStart.getDistance(self.__nodeEnd)

    def getStartNode(self):
        """
        Get the edge's start node.

        :return: Start object of the connection
        :rtype: _WcmnLayouterNode
        """
        return self.__nodeStart

    def getEndNode(self):
        """
        Get the edge's end node.

        :return: End object of the connection
        :rtype: _WcmnLayouterNode
        """
        return self.__nodeEnd

    def copyObject(self):
        """
        Copy the current object containing the same two nodes references
        :return: New edge object
        :rtype: _WcmnLayouterEdge
        """
        return _WcmnLayouterEdge(self.__nodeStart, self.__nodeEnd)

    def intersect(self, edge):
        """
        Intersection test between two edges which must have two defined nodes
        (no protection against same Node reference),
        Note: Collinear edges are NOT CURRENTLY SUPPORTED

        :return: Returns True if the edge intersects with the passed edge,
                 False otherwise
        :type edge: _WcmnLayouterEdge
        :rtype: bool
        """
        if edge is None:
            raise ValueError(
                'Cannot check intersection between edges: Edge is None')

        selfStartPos = self.getStartNode().getPosition()
        selfEndPos = self.getEndNode().getPosition()
        edgeStartPos = edge.getStartNode().getPosition()
        edgeEndPos = edge.getEndNode().getPosition()

        # Check cases where two nodes of different edge share the same position
        if selfStartPos == edgeStartPos or selfStartPos == edgeEndPos or \
           selfEndPos == edgeStartPos or selfEndPos == edgeEndPos:
            return False

        # Check cases where two edges are on top of each other
        if (selfStartPos == edgeStartPos) and (selfEndPos == edgeEndPos) or \
           (selfStartPos == edgeStartPos) and (selfEndPos == edgeEndPos):
            return False

        return self.__CCW(selfStartPos, edgeStartPos, edgeEndPos) != \
               self.__CCW(selfEndPos, edgeStartPos, edgeEndPos) and \
               self.__CCW(selfStartPos, selfEndPos, edgeStartPos) != \
               self.__CCW(selfStartPos, selfEndPos, edgeEndPos)

    def __CCW(self, pos1, pos2, pos3):
        """
        :type pos1: (float, float)
        :type pos2: (float, float)
        :type pos3: (float, float)
        :rtype: bool
        """
        return ((pos2[0] - pos1[0]) * (pos3[1] - pos1[1])) > (
            (pos2[1] - pos1[1]) * (pos3[0] - pos1[0]))
