# ----------------------------------------------------------------------------
# - wcmn_graph_layouter_circular                                             -
# - wild cruncher: layouter                                    core / common -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 samuel longchamps                                -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcmn_graph_layouter_circular
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graph layouter

.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
import math
# noinspection PyProtectedMember
# from wcr_core.common.wcmn_graph_layouter import *
from wcr_core.common.wcmn_graph_layouter import WcmnGraphLayouterP, \
    WCMN_GRAPH_LAYOUTER_DEFAULT_SPACING, _WcmnLayouterNode


class WcmnGraphLayouterCircular(WcmnGraphLayouterP):
    """
    Circular Layouter class
    """

    def __init__(self, spacing=None):
        """
        :param spacing: Positive spacing value
        :type spacing: float | None
        """
        if spacing is not None and spacing < 0:
            raise ValueError('The spacing must be positive ({spc} < 0)'.format(
                spc=spacing))

        self.__spacing = spacing or WCMN_GRAPH_LAYOUTER_DEFAULT_SPACING

    def applyLayout(self, graph):
        """
        Organize the passed graph object as a circle

        :param graph: Object representing a set of nodes and edges between them
        :type graph: _WcmnLayouterGraph
        """
        if not graph:
            raise ValueError('Cannot apply circular layout: graph is None')

        nodes = graph.getNodes()
        """:type: list[_WcmnLayouterNode]"""
        numNode = len(nodes)

        if numNode <= 1:
            return

        postProcessingEnable = False
        orderedNodes = self._reorderAsPath(graph)
        if orderedNodes:
            nodes = orderedNodes
        else:
            postProcessingEnable = True

        anglePerNode = (2 * math.pi) / numNode
        radius = self.__spacing / (2 * math.sin(anglePerNode / 2))

        for i, node in zip(range(numNode), nodes):
            node.setPosition(radius * math.cos(math.pi + anglePerNode * i),
                             radius * math.sin(math.pi + anglePerNode * i))

        if postProcessingEnable:
            self._swapPostProcessing(graph, 10)

    def _reorderAsPath(self, graph):
        """
        Create a reordered list of the nodes contained in the passed graph
        such that its first element is the initial state, the last one its
        final state and a path exists between the elements.
        If no initial or final state exist, the first node of the list
        is taken as initial and the first path order found is returned.

        :param graph: Graph containing the nodes to be ordered
        :type graph: _WcmnLayouterGraph
        :return: Organized list of nodes or None if no path from initial
                state to the final state (when applies) through all the nodes
                was found
        :rtype: list[_WcmnLayouterNode] | None
        """
        nodeNum, edgeNum = graph.getGraphNodeAndEdgeCount()
        if nodeNum < 3 or edgeNum < nodeNum - 1:
            return
        nodes = graph.getNodes()

        atLeastOneFinal = False

        def recursivePathLookupStep(connectedNodes, pathNodes):
            """
            Iterate through the passed connectedNodes and add nodes recursively
            that have not already been added in the pathNodes list up to this
            step. The base case is when pathNodes has the same length as
            the graph, therefore all the nodes are contained in it.
            If no path is found, None is returned.

            :param connectedNodes: List of nodes which were connected to the
                last node in the pathNodes
            :type connectedNodes: list[_WcmnLayouterNode]
            :param pathNodes: The path constructed up to this step
            :type pathNodes: list[_WcmnLayouterNode]
            :return: Organized list of nodes or None if no path from initial
                state to the final state through all the nodes was found
            :rtype: list[_WcmnLayouterNode] | None
            """
            if pathNodes:
                if len(pathNodes) == len(nodes):
                    return pathNodes

            newPathNodes = pathNodes.copy()

            for conNode in connectedNodes:
                if newPathNodes.count(conNode) == 0:
                    pathNodesBkp = newPathNodes.copy()

                    if len(newPathNodes) == len(nodes) - 1:
                        if not atLeastOneFinal or conNode.isFinal():
                            newPathNodes.append(conNode)
                            return newPathNodes
                    else:
                        newPathNodes.append(conNode)

                    fullPathNodes = recursivePathLookupStep(
                        graph.getConnectedNodes(conNode, True, True),
                        newPathNodes)

                    if fullPathNodes:
                        if len(fullPathNodes) == len(nodes):
                            return fullPathNodes
                    else:
                        newPathNodes = pathNodesBkp
            return None

        initialNode = None
        for node in nodes:
            if node.isInitial():
                initialNode = node
            if node.isFinal():
                atLeastOneFinal = True
            if initialNode and atLeastOneFinal:
                break

        if initialNode:
            return recursivePathLookupStep(
                graph.getConnectedNodes(initialNode, True, True),
                [initialNode])
        else:
            # Try all nodes as initial node
            for node in nodes:
                path = recursivePathLookupStep(
                    graph.getConnectedNodes(node, True, True),
                    [node])
                if path:
                    return path

            return None

    def _swapPostProcessing(self, graph, maxIteration):
        """
        Tries to order the nodes such that the less possible crossing occurs
        by swapping neighbor nodes with connected nodes when possible.

        :param graph: Object representing a set of nodes and edges between them
        :type graph: _WcmnLayouterGraph
        :param maxIteration: Maximum number of iteration to be performed by
            the post processing
        :type maxIteration: int
        :return: An ordered list of the nodes contained in the graph
        :rtype: list[_WcmnLayouterNode]
        """
        nodesList = graph.getNodes()
        """:type: list[_WcmnLayouterNode]"""

        for i in range(maxIteration):
            crossInitial = self.__getNumberCrossing(graph.getEdges())
            for node in nodesList:
                neighborNodes = self.__getNeighborNodes(nodesList, node)

                for conNode in graph.getConnectedNodes(node):
                    for neNode in neighborNodes:
                        if neNode:
                            self.__swapPositionCrossing(graph, conNode, neNode)

            crossFinal = self.__getNumberCrossing(graph.getEdges())
            if crossInitial == crossFinal:
                break

    def __swapPositionCrossing(self, graph, node1, node2):
        """
        Swap the position of the two nodes only if the effect is beneficial in
        term of minimizing the number of crossing edges

        :param graph: Graph containing the two other passed nodes
        :type graph: _WcmnLayouterGraph
        :param node1: First node to swap
        :type node1: _WcmnLayouterNode
        :param node2: First node to swap
        :type node2: _WcmnLayouterNode
        """
        crossBefore = self.__getNumberCrossing(graph.getEdges())
        self.__swapPosition(graph.getNodes(), node1, node2)
        crossAfter = self.__getNumberCrossing(graph.getEdges())

        if crossBefore <= crossAfter:
            self.__swapPosition(graph.getNodes(), node1, node2)

    def __swapPosition(self, nodes, node1, node2):
        """
        Swap the position of two nodes as well as their index in the nodes list

        :param nodes: List of nodes which contains the nodes to be swapped
        :type nodes: list[_WcmnLayouterNode]
        :param node1: First node to swap
        :type node1: _WcmnLayouterNode
        :param node2: Second node to swap
        :type node2: _WcmnLayouterNode
        """
        index1 = nodes.index(node1)
        index2 = nodes.index(node2)
        nodes[index1] = node2
        nodes[index2] = node1

        x1, y1 = node1.getPosition()
        x2, y2 = node2.getPosition()
        node1.setPosition(x2, y2)
        node2.setPosition(x1, y1)

    def __getNumberCrossing(self, edges):
        """
        :param edges: List of edges to check crossing for
        :type edges: list[_WcmnLayouterEdges]
        :return: The number of crossing edges for the passed graph
        :rtype: int
        """
        if not edges:
            return 0

        edgesNum = len(edges)
        count = 0
        for i, edge1 in zip(range(edgesNum), edges):
            for edge2 in edges[i + 1:edgesNum]:
                if edge1.intersect(edge2):
                    count += 1
        return count

    def __getNeighborNodes(self, nodesList, node):
        """
        Get the two neighbor nodes closest on the circle to the passed node
        The list must contain at least 3 nodes.

        :param nodesList: List of nodes which contains the node passed
        :type nodesList: list[_WcmnLayouterNode]
        :param node: Node for which to look for the neighbor
        :type node: _WcmnLayouterNode
        :return: A tuple containing the neighbor nodes, or None if not found
            or if the list of nodes is too small
        :rtype: (_WcmnLayouterNode, _WcmnLayouterNode) | (None, None)
        """
        numNodes = len(nodesList)
        if numNodes < 3:
            return None, None

        try:
            index = nodesList.index(node)
        except ValueError as e:
            print(e)
            return None, None

        previousIndex = index - 1
        nextIndex = index + 1

        if previousIndex < 0:
            previousIndex = numNodes - 1
        if nextIndex >= numNodes:
            nextIndex = 0

        return nodesList[previousIndex], nodesList[nextIndex]
