# ----------------------------------------------------------------------------
# - wcmn_graph_layouter_spring                                               -
# - wild cruncher: layouter                                    core / common -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 samuel longchamps                                -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcmn_graph_layouter_spring
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graph layouter

.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
# noinspection PyProtectedMember
from wcr_core.common.wcmn_graph_layouter import WcmnGraphLayouterP, \
    WCMN_GRAPH_LAYOUTER_DEFAULT_SPACING, _WcmnLayouterGraph
from wcr_core.common.wcmn_graph_layouter_grid import WcmnGraphLayouterGrid


# Global values
WCMN_GRAPH_LAYOUTER_DEFAULT_ATTRACTION = 0.04
WCMN_GRAPH_LAYOUTER_DEFAULT_REPULSION = 800
WCMN_GRAPH_LAYOUTER_DEFAULT_DAMPING = 0.85
WCMN_GRAPH_LAYOUTER_DEFAULT_MAX_INTERATION = 400
WCMN_GRAPH_LAYOUTER_DEFAULT_EQUILIBRIUM_THRESHOLD = 0.01


class WcmnGraphLayouterSpring(WcmnGraphLayouterP):
    """
    Spring Model Layouter class
    """

    def __init__(self, gainAttraction=None, gainRepulsion=None, damping=None,
                 maxIteration=None, equiThres=None, spacing=None):
        """
        :param gainAttraction: Positive weight coefficient for attraction of
                               springs
        :type gainAttraction: float | None
        :param gainRepulsion: Positive weight coefficient for repulsion of
                              springs
        :type gainRepulsion: float | None
        :param damping: Coefficient used to reduce gradually the velocity near
                        equilibrium point.
                        Domain expected is ]0, 1] (with 1 = no Damping)
        :type damping: float | None
        :param maxIteration: Positive maximum number of iteration steps for
                             the simulation
        :type maxIteration: int | None
        :param equiThres: Positive threshold of delta velocity between steps
                          that stops the simulation
        :type equiThres: float | None
        :param spacing: Positive spacing value (used for Grid Layouter)
        :type spacing: float | None
        """
        if gainAttraction is not None and gainAttraction < 0:
            raise ValueError('The attraction gain must be positive '
                             '({val} < 0)'.format(val=gainAttraction))
        if gainRepulsion is not None and gainRepulsion < 0:
            raise ValueError('The repulsion gain must be positive '
                             '({val} < 0)'.format(val=gainRepulsion))
        if damping is not None and not (0 < damping <= 1):
            raise ValueError('The damping must be between 0 and 1 '
                             '({val} is invalid)'.format(val=damping))
        if maxIteration is not None and maxIteration < 0:
            raise ValueError('The maximum number of iteration must be positive'
                             ' ({val} < 0)'.format(val=maxIteration))
        if equiThres is not None and equiThres < 0:
            raise ValueError('The equilibrium threshold must be positive '
                             '({val} < 0)'.format(val=equiThres))
        if spacing is not None and spacing < 0:
            raise ValueError('The spacing must be positive '
                             '({val} < 0)'.format(val=spacing))

        self.__gainAttraction = gainAttraction or \
                            WCMN_GRAPH_LAYOUTER_DEFAULT_ATTRACTION
        self.__gainRepulsion = gainRepulsion or \
                            WCMN_GRAPH_LAYOUTER_DEFAULT_REPULSION
        self.__damping = damping or \
                            WCMN_GRAPH_LAYOUTER_DEFAULT_DAMPING
        self.__maxIteration = maxIteration or \
                            WCMN_GRAPH_LAYOUTER_DEFAULT_MAX_INTERATION
        self.__equilibriumThreshold = equiThres or \
                            WCMN_GRAPH_LAYOUTER_DEFAULT_EQUILIBRIUM_THRESHOLD
        self.__gridLayouter = WcmnGraphLayouterGrid(
            spacing or WCMN_GRAPH_LAYOUTER_DEFAULT_SPACING)

    # -- Base class methods implementations

    def applyLayout(self, graph):
        """
        Organize the passed graph object with the spring model
        :param graph: Object representing a set of nodes and edges between them
        :type graph: _WcmnLayouterGraph
        """

        if not graph:
            raise ValueError('Cannot apply layout: graph is None')
        nodeNum, edgeNum = graph.getGraphNodeAndEdgeCount()
        if nodeNum == 0:
            raise ValueError('Cannot apply layout: graph doesn\'t contain '
                             'any node')

        # Layout as a grid for initial simulation position
        self.__gridLayouter.applyLayout(graph)
        graph.flagAllNodesOriginal()

        if nodeNum <= 2 or edgeNum == 0:
            return

        for node in graph.getNodes():
            node.setVelocity(0, 0)

        prevTotalVel = self.__equilibriumThreshold
        k = totalVel = 0

        while k in range(self.__maxIteration) and \
              abs(prevTotalVel - totalVel) >= self.__equilibriumThreshold:

            prevTotalVel = totalVel
            totalVel = 0

            for node1 in graph.getNodes():
                node1.setNetForce(0, 0)

                self.__applyRepulsionForce(node1, graph)

                if self.__applyAttractionForce(node1, graph):
                    self.__applyVelocityFromNetForce(node1)
                else:
                    # If no interconnection with another node than itself has
                    # been found, we don't apply any force on the node because
                    # it would be repulsed far away (caused by the lack of
                    # attraction force)
                    node1.setNetForce(0, 0)

            totalVel += self.__translatePositionFromVelocity(graph)

            k += 1

        chgNodes, unchgNodes = self.__classifyChangedUnchangedNodes(
            graph.getNodes())

        if len(chgNodes) > 0:
            self.__centerAround(chgNodes, 0, 0)

            if len(unchgNodes) > 0:
                self.__repositionUnchangedNodes(chgNodes, unchgNodes)

    # -- Private tool methods

    def __boundingBox(self, nodes):
        """
        Rough bounding box of generic nodes (which are dimensionless,
        therefore the bounding box considers each node as a point) for
        quick approximation

        :type nodes: list[_WcmnLayouterNode]
        :return: x, y, widht, height (x-y are coordinates of top-left point of
                 rectangle of size defined by width and height)
        :rtype: (float, float, float, float)
        """

        if not nodes:
            raise ValueError('Cannot evaluate the bounding box: nodes list is '
                             'empty or None')

        xMin = xMax = yMin = yMax = None

        isFirst = True
        for n in nodes:
            x, y = n.getPosition()

            if isFirst:
                xMin = xMax = x
                yMin = yMax = y
                isFirst = False
                continue

            xMin = min(x, xMin)
            xMax = max(x, xMax)
            yMin = min(y, yMin)
            yMax = max(y, yMax)

        return xMin, yMin, xMax - xMin, yMax - yMin

    def __centerAround(self, nodes, x, y):
        """
        Offsets the nodes to be centered around a given position

        :type nodes: list[_WcmnLayouterNode]
        :type x: float
        :type y: float
        """
        if not nodes:
            raise ValueError('Cannot center around the list of nodes: nodes '
                             'list is empty or None')

        xBB, yBB, widht, height = self.__boundingBox(nodes)
        xCenterCorrection = x - (xBB + widht / 2)
        yCenterCorrection = y - (yBB + height / 2)

        for n in nodes:
            n.translate(xCenterCorrection, yCenterCorrection)

    def __repositionUnchangedNodes(self, chgNodes, unchgNodes):
        """
        Repositions the unchanged nodes under the nodes that have changed.

        :param chgNodes: List of nodes that have changed
        :type chgNodes: list[_WcmnLayouterNode]
        :param unchgNodes: List of nodes that have not changed
        :type unchgNodes: list[_WcmnLayouterNode]
        """
        xChg, yChg, widthChg, heightChg = self.__boundingBox(chgNodes)
        remainingGraph = _WcmnLayouterGraph(unchgNodes, [])
        self.__gridLayouter.applyLayout(remainingGraph)

        xUnchg, yUnchg, widthUnchg, heightUnchg = self.__boundingBox(
            unchgNodes)
        self.__centerAround(remainingGraph.getNodes(),
                            0,
                            heightChg / 2 + heightUnchg / 2 +
                            self.__gridLayouter.getSpacing())

    def __classifyChangedUnchangedNodes(self, nodesList):
        """
        Classifies in two lists the nodes in the passed nodesList: one which
        contains the changed nodes and the other the unchanged ones.
        Note the (un)changed is with regard to their position compared to its
        initial values.
        :param nodesList: List of nodes to be classified
        :type nodesList: list[_WcmnLayouterNode]
        :return: One changed and one unchanged list of nodes
        :rtype: (list[_WcmnLayouterNode], list[_WcmnLayouterNode])
        """

        unchangedNodes = []
        """:type: list[_WcmnLayouterNode]"""
        changedNodes = []
        """:type: list[_WcmnLayouterNode]"""

        for node in nodesList:
            if node.hasPositionChanged():
                changedNodes.append(node)
            else:
                unchangedNodes.append(node)

        return changedNodes, unchangedNodes

    def __translatePositionFromVelocity(self, graph):
        """
        A new position is computed and applied on the graph's nodes based on
        their current velocity.

        :param graph: Graph containing nodes to which the position will be
                      applied to.
        :type graph: _WcmnLayouterGraph
        :return: Total velocity of the nodes which were translated
        :rtype: float
        """
        totalVel = 0
        for node in graph.getNodes():
            velX, velY = node.getVelocity()
            node.translate(velX, velY)
            totalVel += abs(velX) + abs(velY)
        return totalVel

    def __applyVelocityFromNetForce(self, node):
        """
        A new velocity is computed and applied on the passed node based on its
        net force and current velocity.
        A damping based on the __damping attribute is applied to the new
        velocity to help the algorithm converge.

        :param node: Node to which the velocity is applied
        :type node: _WcmnLayouterNode
        """
        velX, velY = node.getVelocity()
        netFX, netFY = node.getNetForce()
        node.setVelocity(self.__damping * (velX + netFX),
                         self.__damping * (velY + netFY))

    def __applyRepulsionForce(self, node, graph):
        """
        Applies a repulsion force on the passed node caused by other nodes in
        the graph.
        The repulsion force applied is based on the __gainRepulsion attribute.

        :param node: Node to which the repulsion force is applied
        :type node: _WcmnLayouterNode
        :param graph: Graph containing the passed nodes.
        :type graph: _WcmnLayouterGraph
        """
        x1, y1 = node.getPosition()
        for otherNode in graph.getNodes():
            if node is otherNode:
                continue

            x2, y2 = otherNode.getPosition()
            sqrDist = node.getSqrDistance(otherNode)

            if sqrDist > 0:
                node.addNetForce(self.__gainRepulsion * (x1 - x2) / sqrDist,
                                 self.__gainRepulsion * (y1 - y2) / sqrDist)

    def __applyAttractionForce(self, node, graph):
        """
        Applies an attraction force on the passed node caused by linked nodes
        present in the passed graph.
        The attraction force applied is based on the __gainAttraction attribute
        and is limited by a total factor of two when multiple connections
        between two nodes occur.

        :param node: Node to which the attraction force is applied
        :type node: _WcmnLayouterNode
        :param graph: Graph containing the passed nodes.
        :type graph: _WcmnLayouterGraph
        :return: True if interconnections between the passed nodes and other
                 nodes in the graph allows the application of Force,
                 False otherwise.
        :rtype: bool
        """
        x1, y1 = node.getPosition()

        connectedNodes = graph.getConnectedNodes(node, True)
        """:type: list[_WcmnLayouterNode]"""
        alreadyConnectedNodes = []
        """:type: list[_WcmnLayouterNode]"""

        for conNode in connectedNodes:
            connectionToNodeNum = alreadyConnectedNodes.count(conNode)

            weightFactor = 2 ** connectionToNodeNum

            adjustedGainAttraction = self.__gainAttraction / weightFactor

            x2, y2 = conNode.getPosition()
            node.addNetForce(adjustedGainAttraction * (x2 - x1),
                             adjustedGainAttraction * (y2 - y1))

            alreadyConnectedNodes.append(conNode)

        return len(connectedNodes) > 0
