# ----------------------------------------------------------------------------
# - wcmn-verifier-module                                                     -
# - wild cruncher: verification protocols                      core / common -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcmn_verifier_module
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher database verification protocols

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>

TODO: This module should be in a 'common'-package.
"""
from abc import ABCMeta, abstractmethod


class WcmnVerificationError(object):
    """
    Base class for errorr found during a verification run.
    """

    def __init__(self, errorId, errorMsg, reVerifyF=None, autoFixF=None):
        """
        :param errorId: Identifier for the error type
        :type errorId: str
        :param errorMsg: Error message
        :type errorMsg: str
        :param reVerifyF: Function to re-verify, i.e. test whether the error
                          still persists, if possible (TBD/NYI)
        :type reVerifyF: () -> WcmnVerificationError
        :param autoFixF: Function to automatically correct the error,
                         if possible (TBD/NYI)
        :type autoFixF: () -> _
        """
        #: Error ID
        self.__errorId = errorId
        #: Error message
        self.__errorMsg = errorMsg
        #: Re-verification function
        self.__reVerifyF = reVerifyF
        #: Auto-fix function
        self.__autoFixF = autoFixF

    @property
    def errorId(self):
        """
        Get the error type ID.

        :return: Error type ID
        :rtype: str
        """
        return self.__errorId

    @property
    def errorMessage(self):
        """
        Get the error message.

        :return: Error message
        :rtype: str
        """
        return self.__errorMsg

    def reVerifyPossible(self):
        """
        Whether it is possible to re-verify whether the error still persists.

        :return: True iff the existence of the error can be checked
        :rtype: bool
        """
        return self.__reVerifyF is not None

    def reVerify(self):
        """
        Test whether the error still persists, if possible.

        :return: Error that occurred during the re-verification, or None
        :rtype: WcmnVerificationError
        """
        return self.__reVerifyF()

    @property
    def autoFixPossible(self):
        """
        Whether it is possible to automatically fix the error.

        :return: True iff it is possible to automatically fix the error
        """
        return self.__autoFixF is not None

    def autoFix(self):
        """
        Automatically fix the error if possible.
        """
        self.__autoFixF()


class WcmnHierarchyVerificationError(WcmnVerificationError):
    """
    Single error found during a hierarchy verification run.
    """

    def __init__(self, errorId, errorMsg, offenders, offended,
                 reVerifyF=None, autoFixF=None):
        """
        :param errorId: Identifier for the error type
        :type errorId: str
        :param errorMsg: Error message
        :type errorMsg: str
        :param offenders: Offending hierarchy entities causing this error
        :type offenders: list[WhdbBaseP]
        :param offended: Verification module raising this error
        :type offended: WcmnVerifierModuleP
        :param reVerifyF: Function to re-verify, i.e. test whether the error
                          still persists, if possible (TBD/NYI)
        :type reVerifyF: () -> WcmnVerificationError
        :param autoFixF: Function to automatically correct the error,
                         if possible (TBD/NYI)
        :type autoFixF: None -> None
        """
        super().__init__(errorId, errorMsg, reVerifyF, autoFixF)

        #: Offending entities
        self.__offenders = offenders
        #: Offended verification module
        self.__offended = offended
        #: Instance-path to the model containing the offending entities
        self.__path = None

    @property
    def offendingEntities(self):
        """
        Get the entities that caused this error.

        :return: Entities that caused the error
        :rtype: list[WhdbBaseP]
        """
        return self.__offenders

    @property
    def offendedVerifierModule(self):
        """
        Get the verification module that raised this error.

        :return: Verification module that raised the error
        :rtype: WhdbVerifierModuleP
        """
        return self.__offended

    @property
    def path(self):
        """
        Get the model instance path leading to the model containing the
        offending instances.

        :return: Model instance path
        :rtype: list[WhdbBlockInstanceP]
        """
        return self.__path

    @path.setter
    def path(self, path):
        """
        Set the model instance path leading to the model containing the
        offending instances.

        :param path: Model instance path
        :type path: list[WhdbBlockInstanceP]
        """
        self.__path = path

    def __repr__(self):
        return self.errorMessage


class WcmnVerifierModuleP(metaclass=ABCMeta):
    @property
    @abstractmethod
    def description(self):
        """
        Get a description for the verification performed by this module.

        The description can be used to provide further information about
        verification errors to the user.

        :return: Module description
        :rtype: str
        """
        raise NotImplementedError()


# TODO: Move to wcr_core.db
class WcmnVerificationResult(object):
    """
    Result of a single hierarchy verification run.
    """

    def __init__(self, topModel):
        """
        :param topModel: Top-level model
        :type topModel: WhdbBlockP
        """
        #: List of all errors found
        self.__errors = []
        #: List of errors by hierarchy entity
        self.__errorsByEntity = {}
        #: List of errors by model
        self.__errorsByModel = {}
        self.__blockInstanceStack = []
        self.__p = set()
        self.__topModel = topModel

    def _pushBlockInstance(self, instance):
        """
        :param instance:
        :type instance: WhdbBlockInstanceP
        """
        self.__blockInstanceStack.append(instance)

    def _popBlockInstance(self):
        self.__blockInstanceStack.pop()

    @property
    def _currentModel(self):
        """
        :return:
        :rtype: WhdbBlockP
        """
        if self.__blockInstanceStack:
            return self.__blockInstanceStack[-1].getBlock()
        else:
            return self.__topModel

    def _addErrors(self, errors):
        """
        Add errors found during verification to the result.

        :param errors: Errors found
        :type errors: list[WcmnHierarchyVerificationError]
        """
        if not errors: return

        self.__errors.extend(errors)

        path = self.__blockInstanceStack[:]
        modelId = self._currentModel.getUUID()
        for err in errors:
            err.path = path
            for entity in err.offendingEntities:
                self.__errorsByEntity.setdefault(entity.getUUID(), []) \
                    .append(err)

            self.__errorsByModel.setdefault(modelId, []).append(err)

        for inst in self.__blockInstanceStack:
            self.__p.add(inst.getUUID())

    @property
    def allErrors(self):
        """
        Get all the errors found during the hierarchy verification run.

        :return: All errors found
        :rtype: list[WcmnHierarchyVerificationError]
        """
        return self.__errors

    @allErrors.setter
    def allErrors(self, errors):
        """
        Get all the errors found during the hierarchy verification run.

        :return: All errors found
        :rtype: list[WcmnHierarchyVerificationError]
        """
        self.__errors = errors

    def errorsForEntity(self, entity):
        """
        Get error found for the given entity.

        :param entity: Entity to provide found errors for
        :type entity: WhdbBaseP
        :return: Errors found for the given entity
        :rtype: list[WcmnHierarchyVerificationError]
        """
        return self.__errorsByEntity.get(entity.getUUID())

    def hasDeepErrors(self, inst):
        """
        Whether any of the elements in the given model block instance's model
        contains any errors, recursively.<s

        :param inst: Model block instance to check for deep errors
        :type inst: WhdbBlockInstanceP
        :return: True iff any of the elements in the instance's model contain
                 errors, recursively
        :rtype: bool
        """
        return inst.getUUID() in self.__p

    def hasErrors(self):
        """
        Whether the result set contains any errors.

        :return: True iff the result set contains any errors
        :rtype: bool
        """
        return bool(self.__errors)
