# ----------------------------------------------------------------------------
# - wctl-controller-mixins                                                   -
# - wild cruncher: controller mixins                       core / controller -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wctl_controller_mixins
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Controller mixins

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
# noinspection PyAbstractClass
from wcr_core.controller.wctl_controller_api import *


_WCTL_EMPTY_TUPLE = tuple()


# noinspection PyAbstractClass
class WctlInstanceControllerDefaultXPointMixin(WctlInstanceControllerP):
    """
    Mixin to provide default connection point functionalities for
    instance controllers.
    """

    @property
    @abstractmethod
    def factory(self):
        """
        :rtyp: WctlControllerFactoryP
        """
        raise NotImplementedError()

    def canCreateConnectionPoint(self, instance, **kwargs):
        """
        :type instance: WhdbInstanceP
        :type kwargs: dict[str, object]
        """
        return instance.canCreateConnectionPoint(**kwargs)

    def createConnectionPoint(self, instance, valuePrv, **kwargs):
        """
        :type instance: WhdbInstanceP
        :type valuePrv: WctlInstanceControllerValueProviderP
        :type kwargs: dict[str, object]
        """
        assert self.canCreateConnectionPoint(instance, **kwargs)

        if 'fcng' in kwargs:
            kwargs['gvty'] = kwargs['fcng']

        return instance.createConnectionPoint(**kwargs)

    def canRemoveConnectionPoint(self, instance, xpoint):
        """
        :type instance: WhdbInstanceP
        :type xpoint: WhdbConnectionPointP
        """
        return instance.canRemoveConnectionPoint(xpoint)

    def removeConnectionPoint(self, instance, xpoint, block):
        """
        :type instance: WhdbInstanceP
        :type xpoint: WhdbConnectionPointP
        :type block: WhdbBlockP
        """
        conns = list(block.getConnectionsForObject(instance, xpoint))

        dels = set()
        upds = set()

        if conns:
            ctrl = self.factory.getConnectionController()
            for conn in conns:
                _dls, _ups = ctrl.deleteItem(conn, block)

                if _dls:
                    dels.update(_dls)
                if _ups:
                    upds.update(_ups)

        instance.removeConnectionPoint(xpoint)
        upds.add(instance)

        return dels, upds


# noinspection PyAbstractClass
class WctlInstanceControllerDefaultDeleteMixin(WctlInstanceControllerP):
    """
    Mixin to provide default deletion functionality for instance controllers.
    """

    @property
    @abstractmethod
    def factory(self):
        """
        :rtyp: WctlControllerFactoryP
        """
        raise NotImplementedError()

    def deleteItem(self, item, block):
        """
        :type item: WhdbInstanceP
        :type block: WhdbBlock
        """
        upds = []
        dels = []

        ctrl = self.factory.getConnectionController()
        conns = list(block.getConnectionsForObject(item))

        for conn in conns:
            _dels, _upds = ctrl.deleteItem(conn, block)
            if _dels:
                dels.extend(_dels)
            if _upds:
                upds.extend(_upds)

        block.remove(item)
        dels.append(item)

        return dels, upds


# noinspection PyAbstractClass
class WctlConnectionControllerDefaultCreateMixin(WctlConnectionControllerP):
    """
    Mixin to provide default connection functionalities for connection
    controllers.
    """

    def canStartConnectionAt(self, block, instance, xpoint=None):
        """
        :type block: WhdbBlock
        :type instance: WhdbInstanceP
        :type xpoint: WhdbConnectionPointP | None
        """
        return instance and xpoint and xpoint in instance.getConnectionPoints()

    def canCreateConnection(self, block, sinst, einst, sxpnt=None, expnt=None,
                            **kwargs):
        """
        :type block: WhdbBlock
        :type sinst: WhdbInstanceP
        :type einst: WhdbInstanceP
        :type sxpnt: WhdbConnectionPointP | None
        :type expnt: WhdbConnectionPointP | None
        """
        noNones = sinst and einst and sxpnt and expnt
        if not noNones:
            return False

        if sxpnt not in sinst.getConnectionPoints():
            return False
        if expnt not in einst.getConnectionPoints():
            return False

        return True

    def createConnection(self, block, sinst, einst, sxpnt=None, expnt=None,
                         **kwargs):
        """
        :type block: WhdbBlock
        :type sinst: WhdbInstanceP
        :type einst: WhdbInstanceP
        :type sxpnt: WhdbConnectionPointP | None
        :type expnt: WhdbConnectionPointP | None
        """
        assert self.canCreateConnection(
            block, sinst, einst, sxpnt, expnt, **kwargs)

        return \
            block.createConnection(
                so=sinst, eo=einst, sc=sxpnt, ec=expnt, **kwargs), \
            (sinst, einst)

    def hasEndPortMultipleConnection(self, block, StartPt, EndPt):
        state = False
        points = []
        for conn in block.getConnections():
            connEndPt = conn.getEndConnectionPoint()
            connStartPt = conn.getStartConnectionPoint()
            if EndPt.getUUID() == connEndPt.getUUID():
                state = True
                points.append(EndPt.getName())
            if StartPt.getUUID() == connStartPt.getUUID():
                state = True
                points.append(StartPt.getName())
            if EndPt.getUUID() == connStartPt.getUUID():
                state = True
                points.append(EndPt.getName())
            if StartPt.getUUID() == connEndPt.getUUID():
                state = True
                points.append(StartPt.getName())

        return state, points


# noinspection PyAbstractClass
class WctlConnectionControllerDefaultDeleteMixin(WctlConnectionControllerP):
    """
    Mixin to provide default deletion functionalities for connection
    controllers.
    """

    def deleteItem(self, item, block):
        """
        :type item: WhdbBaseP
        :type block: WhdbBlock
        """
        block.remove(item)

        return (item,), None


# noinspection PyAbstractClass
class WctlShapeControllerDefaultCreateMixin(WctlShapeControllerP):
    """
    Mixin to provide default shape controller functionalities.
    """

    def createShape(self, shapeType, block, **kwargs):
        """
        :type shapeType: str
        :type block: WhdbBlockP
        :type kwargs: dict[str, object]
        """
        return block.createShape(shapeType, **kwargs)


# noinspection PyAbstractClass
class WctlShapeControllerDefaultDeleteMixin(WctlShapeControllerP):
    """
    Mixin to provide default deletion functionality for shape controllers.
    """

    def deleteItem(self, item, block):
        """
        :type item: WhdbShapeP
        :type block: WhdbBlock
        """
        block.remove(item)
        return [item], []
