# ----------------------------------------------------------------------------
# - whdb-block                                                               -
# - wild cruncher: block object                                    core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general hierarchy blocks

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db_api.whdb_block_api import WhdbBlockP, WHDB_TOPLEVEL_BLOCK_CODE
from wcr_core.db_api.whdb_button_api import WhdbButtonP
from wcr_core.db_api.whdb_connection_api import WhdbConnectionP
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP
from wcr_core.db_api.whdb_operator_api import WhdbOperatorP
from wcr_core.db_api.whdb_shape_api import WhdbShapeP
from wcr_core.dbstore_api.wdbs_instance_api import WdbsOperatorInstanceP, \
    WdbsModelInstanceP, WdbsBlockInstanceP
from .whdb_connection import WHDB_KEEP_PATH_KEY
from .whdb_operator import WHDB_PORT_OPERATOR_FLAG
from .whdb_protocols import *


# noinspection PyAbstractClass
# noinspection PyMethodMayBeStatic
class WhdbBlock(WhdbBase, WhdbBlockP, WhdbTypedEntity, WhdbNamedEntity):
    META = WdbsMetaDescription()

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbBase.__init__(self, block, wreg)
        WhdbNamedEntity.__init__(self, block)

        self.__rawBlock = block
        self.__wrapperRegistry = wreg
        self.isModified = False

    @property
    def rawElement(self):
        """
        :rtype: WdbsBlockP
        """
        return self.__rawBlock

    def getTypeName(self):
        return 'block'

    def _addNewRawInstanceMeta(self, mdct, ipos, zpos, width, height, rotn,
                               vflp, hflp):
        mdct[WHDB_POSITION_KEY] = ipos or (0, 0)
        mdct[WHDB_Z_VALUE_KEY] = zpos
        if width is not None and height is not None:
            mdct[WHDB_DIMENSION_KEY] = (width, height)
        mdct[WHDB_ROTATION_KEY] = rotn or 0
        mdct[WHDB_HFLIP_KEY] = hflp or False
        mdct[WHDB_VFLIP_KEY] = vflp or False

    def _createRawBlockInstance(self, name, info, btyp=None, **kwargs):
        self.isModified = True
        return self.__rawBlock.createBlockInstance(
            name=name, info=info, btyp=btyp, **kwargs)

    def createBlockInstance(self, name=None, info=None,
                            ipos=None, zpos=None, width=None, height=None,
                            rotn=None, vflp=None, hflp=None, btyp=None,
                            **kwargs):
        if btyp == WHDB_TOPLEVEL_BLOCK_CODE:
            raise ValueError('Cannot add top-level block')
        self._addNewRawInstanceMeta(
            kwargs, ipos, zpos, width, height, rotn, vflp, hflp)
        inst = self._createRawBlockInstance(
            name=name, info=info, btyp=btyp, **kwargs)
        wrap = self.__wrapperRegistry.wrap(inst)
        """:type: WhdbBlockInstanceP"""
        if not (width and height):
            inst.setMeta(wrap.getDefaultDimension(), WHDB_DIMENSION_KEY)
        self.isModified = True
        return wrap

    def getBlockInstances(self):
        """
        :rtype: Iterable[WhdbBlockInstanceP]
        """
        return (self.__wrapperRegistry.wrap(i) for i
                in self.__rawBlock.getInstances()
                if isinstance(i, WdbsBlockInstanceP))

    def _createRawModelInstance(self, mime, name, info, mtyp=None, **kwargs):
        inst = self.__rawBlock.createModelInstance(
            name=name, info=info, mtyp=mtyp, **kwargs)
        inst.getModel().setMeta(mime, WHDB_MEDIA_KEY)
        self.isModified = True
        return inst

    def createModelInstance(self, mime, name=None, info=None,
                            ipos=None, zpos=None, width=None, height=None,
                            rotn=None, vflp=None, hflp=None, mtyp=None,
                            **kwargs):
        self._addNewRawInstanceMeta(
            kwargs, ipos, zpos, width, height, rotn, vflp, hflp)
        inst = self._createRawModelInstance(
            mime=mime, name=name, info=info, mtyp=mtyp, **kwargs)
        wrap = self.__wrapperRegistry.wrap(inst)
        """:type: WhdbModelInstanceP"""
        if not (width and height):
            inst.setMeta(wrap.getDefaultDimension(), WHDB_DIMENSION_KEY)
        self.isModified = True
        return wrap

    def getModelInstances(self):
        return (self.__wrapperRegistry.wrap(i) for i
                in self.__rawBlock.getInstances()
                if isinstance(i, WdbsModelInstanceP))

    def _createRawOperatorInstance(self, oper, name, info, **kwargs):
        self.isModified = True
        return self.__rawBlock.createOperatorInstance(
            oper.getCode() if isinstance(oper, WhdbOperatorP) else oper,
            name=name, info=info, **kwargs)

    def createOperatorInstance(self, oper, name=None, info=None,
                               ipos=None, zpos=None, width=None, height=None,
                               rotn=None, vflp=None, hflp=None, prms=None,
                               **kwargs):
        self.isModified = True
        self._addNewRawInstanceMeta(
            kwargs, ipos, zpos, width, height, rotn, vflp, hflp)
        inst = self._createRawOperatorInstance(
            oper, name, info, **kwargs)
        wrap = self.__wrapperRegistry.wrap(inst)
        """:type: WhdbOperatorInstanceP"""
        if not (width and height):
            inst.setMeta(wrap.getDefaultDimension(), WHDB_DIMENSION_KEY)
        return wrap

    def getOperatorInstances(self):
        return (self.__wrapperRegistry.wrap(i) for i
                in self.__rawBlock.getInstances()
                if isinstance(i, WdbsOperatorInstanceP))

    def getPorts(self):
        # noinspection PyUnresolvedReferences
        return (i for i in self.getOperatorInstances()
                if i.getOperator().hasFlag(WHDB_PORT_OPERATOR_FLAG))

    def getInstances(self):
        return (self.__wrapperRegistry.wrap(i) for i
                in self.__rawBlock.getInstances())

    def removeInstance(self, inst):
        raw = self.__wrapperRegistry.unwrap(inst)
        """:type: WdbsInstanceP"""
        self.__rawBlock.removeInstance(raw)
        self.isModified = True

    def _createRawConnection(self, so, sc, eo, ec, name, info, path, kppt,
                             zval, **kwargs):
        kwargs[WHDB_KEEP_PATH_KEY] = kppt
        kwargs[WHDB_Z_VALUE_KEY] = zval
        self.isModified = True
        return self.__rawBlock.createConnection(
            so, sc, eo, ec, name=name, info=info, path=path, **kwargs)

    def createConnection(self, so, sc, eo, ec, name=None, info=None, path=None,
                         kppt=None, zval=None, **kwargs):
        wreg = self.__wrapperRegistry
        conn = self._createRawConnection(
            wreg.unwrap(so), wreg.unwrap(sc), wreg.unwrap(eo), wreg.unwrap(ec),
            name, info, path, kppt, zval, **kwargs)
        conn = wreg.wrap(conn)
        """:type: WhdbConnectionP"""
        if not path:
            conn.getPath().resetPath()
        self.isModified = True
        return conn

    def getConnections(self):
        """
        :rtype: Iterable[WhdbConnectionP]
        """
        wreg = self.__wrapperRegistry
        return (wreg.wrap(c) for c in self.__rawBlock.getConnections())

    def removeConnection(self, cntn):
        raw = self.__wrapperRegistry.unwrap(cntn)
        """:type: WdbsConnectionP"""
        self.isModified = True
        self.__rawBlock.removeConnection(raw)

    def getConnectionsForObject(self, ix, cx=None):
        return (w for w in self.getConnections() if w.isConnected(ix, cx))

    def _createRawShape(self, styp, **kwargs):
        self.isModified = True
        return self.__rawBlock.createShape(styp, **kwargs)

    def createShape(self, styp, **kwargs):
        self.isModified = True
        return self.__wrapperRegistry.wrap(
            self._createRawShape(styp, **kwargs))

    def getShapes(self):
        return (self.__wrapperRegistry.wrap(s) for s
                in self.__rawBlock.getShapes())

    def removeShape(self, shape):
        self.isModified = True
        raw = self.__wrapperRegistry.unwrap(shape)
        """:type: WdbsShapeP"""
        self.__rawBlock.removeShape(raw)

    def _createRawButton(self, title, command, parameters, bpos, zpos, bbox,
                         **kwargs):
        kwargs[WHDB_Z_VALUE_KEY] = zpos
        kwargs[WHDB_BBOX_KEY] = bbox
        kwargs[WHDB_POSITION_KEY] = bpos
        self.isModified = True
        return self.__rawBlock.createButton(
            title=title, command=command, params=parameters, **kwargs)

    def createButton(self, title=None, command=None, parameters=None,
                     bpos=None, zpos=None, bbox=None, **kwargs):
        btn = self._createRawButton(title, command, parameters, bpos, zpos,
                                    bbox, **kwargs)
        self.isModified = True
        return self.__wrapperRegistry.wrap(btn)

    def getButtons(self):
        return (self.__wrapperRegistry.wrap(b) for b
                in self.__rawBlock.getButtons())

    def removeButton(self, element):
        raw = self.__wrapperRegistry.unwrap(element)
        """:type: WdbsButtonP"""
        self.__rawBlock.removeButton(raw)
        self.isModified = True

    def remove(self, entity):
        if entity is None: return
        if isinstance(entity, WhdbInstanceP):
            self.removeInstance(entity)
        if isinstance(entity, WhdbConnectionP):
            self.removeConnection(entity)
        if isinstance(entity, WhdbShapeP):
            self.removeShape(entity)
        if isinstance(entity, WhdbButtonP):
            self.removeButton(entity)

    def getPortNames(self):
        """
        Get a list of port names for this block.

        :return: List of port names
        :rtype: dict[int, Iterable[(str,str)]]
        """
        return {}

    def createProxy(self):
        """
        Create a proxy for this block.

        :return: Block proxy
        :rtype: WhdbHierarchyBlockProxy
        """
        from .whdb_block_proxy import WhdbHierarchyBlockProxy

        return WhdbHierarchyBlockProxy(self)
