# ----------------------------------------------------------------------------
# - whdb-block-flattener                                                     -
# - wild cruncher:gsh block flattener                              core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_block_flattener
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher block flattener

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod, ABCMeta
from collections import defaultdict

from wcr_core.dbstore.mutable.wdbm_block import WdbmBlock
from wcr_core.dbstore.mutable.wdbm_context import WdbmContext
from wcr_core.dbstore.wdbs_cloning import wdbs_cloneBlock
from wcr_core.dbstore.wdbs_util import wdbs_getOtherEndByPoint, \
    wdbs_getOtherEndByObj, wdbs_getConnections
from wcr_core.dbstore_api.wdbs_instance_api import WdbsOperatorInstanceP, \
    WdbsBlockInstanceP
from wcr_core.dbstore.mutable.wdbm_connection import WdbmConnection
from .whdb_connection import WHDB_CONNECTION_ALIASES_KEY


def _isSameConnection(w1, w2):
    """
    Whether 2 connections connect the same objects/connection points.

    :type w1: WdbsConnectionP
    :type w2: WdbsConnectionP
    :rtype: bool
    """
    (so1, sc1) = (w1.getStartObject(), w1.getStartXPoint())
    (do1, dc1) = (w1.getEndObject(), w1.getEndXPoint())
    (so2, sc2) = (w2.getStartObject(), w2.getStartXPoint())
    (do2, dc2) = (w2.getEndObject(), w2.getEndXPoint())
    return so1 is so2 and sc1 is sc2 and do1 is do2 and dc1 is dc2


def _findConnection(w, l):
    """
    :return: the index of the wire in the list or -1 if not present
    :rtype : WdbsConnectionP
    :param w: the wire to be found
    :type  w: WdbsConnectionP
    :param l: the list to walk
    :type  l: Iterable[WdbsConnectionP]
    """
    for idx, wire in enumerate(l):
        if _isSameConnection(wire, w):
            return wire
    return None


def _mergeConnections(dst, src):
    """
    merge a set two list of wires using the compare
    wire function
    :param dst: the destination list of wires
    :param src: the source list of wires
    """
    for wire in src:
        if not _findConnection(wire, dst):
            dst.append(wire)


def _getObjectConnector(wire, obj):
    """
    Get the connector belonging to the given object.

    :type wire: WdbsConnectionP
    :return: The object's connector
    :rtype: WdbsConnectionPointP
    """
    if wire.getStartObject() is obj:
        return wire.getStartXPoint()
    elif wire.getEndObject() is obj:
        return wire.getEndXPoint()
    else: raise Exception ('wire not connected to the object')


def _getXPointsToPortsMapping(inst, xpntToPortF):
    """
    :return:
    :rtype: dict[WdbsConnectionPointP, WdbsOperatorInstanceP]
    :param inst: the instance to be processed
    :type inst: WdbsBlockInstanceP
    """
    if inst is not None:
        imdl = inst.getBlock()
        if imdl is not None:
            return {c: xpntToPortF(c, imdl)
                    for c in inst.getConnectionPoints()}
    return {}


def _getXPointsToConnectionsMapping(block, inst):
    """
    :param block: the block that the instance belongs to
    :type block: WdbsBlockP
    :param inst: the instance to be processed
    :type inst: WdbsInstanceP
    :return: the dictionary between the instance's connectors ports and its
             list of connected wires for the given instance inside the given
            block
    :rtype: dict[WdbsConnectionPointP, Iterable[WdbsConnectionP]]
    """
    return {c: wdbs_getConnections(block, inst, c) for c
            in inst.getConnectionPoints()}


def _getInnerPortsToConnectionsMapping(inst, xpntToPortF):
    """
    :param inst: the instance to be processed
    :type inst: WdbsBlockInstanceP
    :return: the dictionary between the instance's function ports and its list
             of connected wires for the given instance
    :rtype: dict[WdbsOperatorInstanceP, Iterable[WdbsConnectionP]]
    """
    def getPortAndCnctns(xpnt):
        port = xpntToPortF(xpnt, block)
        return (port, wdbs_getConnections(block, port)) if port else None

    block = inst.getBlock()

    return dict(filter(None,
                       map(getPortAndCnctns, inst.getConnectionPoints())))


class _BorderBreaker():
    """
    The _WdbmBorderBreaker object  holds the mapping information that
    allows to cross the boder between instances and blocks.
    """

    def __init__(self, block, portTestF, xpntToPortF):
        """
        :param block: the block that will be flatten
        :type  block: WdbsBlockP
        """
        self.__portTestF = portTestF
        self.__xpntToPortF = xpntToPortF
        #: the dictionary between the function object and its associated
        #: (obj, connector)
        self.portsToXpnts = {}
        #: the list of hierarchical instance of that block
        self.__hinst = []
        #: the dictionary between the connector at the
        #: instance level and the list of connected pairs
        #: (obj, connector). Most of the time the pair is inside the instance
        #: execept in the case of crossing wire. In this last case the pointed
        #: object is outside the instance.
        self.inside = {}
        #: the dictionary between the connector at the
        #: instance level and the list of related obj/conn/xpnt-tuples at same
        #: block level
        self.outside = {}
        # update the internal structure for the given block
        for inst in block.getInstances():
            if isinstance(inst, WdbsBlockInstanceP):
                self.update(block, inst)
                self.__hinst.append(inst)
        l = []
        for k, v in self.inside.items():
            for (w, do, dc) in v:
                l.append(do.getName())
        l = []
        for k, v in self.outside.items():
            for (w, do, dc) in v:
                l.append(do.getName())

    def getInsts(self):
        """
        Return the list of hierachical instances of that border breaker

        :return: the list of hierachical instances of that border breaker
        :rtype: [WdbsBlockInstanceP]
        """
        return self.__hinst

    def update(self, block, inst):
        """
        do the update for a given blok inst

        :param block: the block that the instance belongs to
        :type  block: WdbsBlockP
        :param inst: the instance to be processed
        :type  inst: WdbsBlockInstanceP
        """
        xpntsToPorts = _getXPointsToPortsMapping(inst, self.__xpntToPortF)
        portsToConns = _getInnerPortsToConnectionsMapping(inst,
                                                          self.__xpntToPortF)
        # upadte the dfunc
        self.portsToXpnts.update({v: k for k, v in xpntsToPorts.items()})
        for xpnt, port in xpntsToPorts.items():
            # update the inside field
            inside = []
            for wire in portsToConns[port]:
                # fetch the destination object and connector
                (do, dc) = wdbs_getOtherEndByObj(wire, port)
                if isinstance(do, WdbsOperatorInstanceP):
                    if not self.__portTestF(do):
                        inside.append((wire, do, dc))
                    else:
                        # crossing : wire in this case the inside will point
                        # to an outside object which is ok
                        wires = wdbs_getConnections(
                            block, inst, self.portsToXpnts[do])
                        for w in wires:
                            # fetch the destination object and connector
                            (do, dc) = wdbs_getOtherEndByPoint(w, self.portsToXpnts[do])
                            inside.append((w, do, dc))
                else:
                    inside.append((wire, do, dc))
            self.inside[xpnt] = inside
            # update the outside fields
            outside = []
            wires = wdbs_getConnections(block, inst, xpnt)
            for wire in wires:
                (do, dc) = wdbs_getOtherEndByPoint(wire, xpnt)
                outside.append((wire, do, dc))
            self.outside[xpnt] = outside


class _DeferrenceManager():
    """
    The _DeferrenceManager hold the unsolved wires. When a wire can not be
    mapped into the final model it is recorded in that object. It will be
    solved later at a higher level of the hierachy. The _DeferrenceManager
    object is also able to solve the recorded objects.
    """

    def __init__(self, portTestF, cnctnMergeF):
        #: the deferred connection, i.e the ones that are not fully
        #: defined at a given level of the hierachy. Information from upper
        #: level will be used during the bottom up part of the recursive
        #: process to clearly defined them. The are store as a dict between
        #: the object to be solved and a list of conected wires
        self.deferred = defaultdict(set)

        self.__portTestF = portTestF
        self.__cnctnMergeF = cnctnMergeF

    def addDeferred(self, obj, wire):
        """
        A deferred entry is an entry that can not be resolved at a given level
        ie, the obj would be replaced later during the flattening process. This
        happen if the obj is a model instance or a function port of an upper
        level.

        :param obj: the object
        :param wire: the wire connected to that object
        :type  wire: WdbsConnectionP
        """
        self.deferred[obj].add(wire)

    def popDeferred(self, obj):
        """
        :return: the list of deferred wire
        :rtype: Iterable[WdbsConnectionP]
        :param obj: the object
        """
        try:
            return list(self.deferred.pop(obj))
        except KeyError:
            return []

    def objToDefer(self, obj):
        """
        Object that implies to defer a wire are the hierarchical instances and
        the function ports.

        :return: true if obj implies a deferring
        :param obj: the object
        """
        return isinstance(obj, WdbsBlockInstanceP) or self.__portTestF(obj)

    def shouldBeDeferred(self, wire):
        """
        :type wire: WdbsConnectionP
        :return: true if the wire has a deferred extremity
        """
        result = False
        for obj in [wire.getStartObject(), wire.getEndObject()]:
            if self.objToDefer(obj):
                self.addDeferred(obj, wire)
                result = True
        return result

    def solveInst(self, bbrk):
        """
        Solve all the deferred entries for the given border breaker.

        :return: the list of wires to be added
        :param bbrk: the border breaker to be used
        """
        # TOOD: result = {(obj, xpnt, obj, xpnt):{name, info, aliases}}
        result = []
        flag = True
        memco = list()
        memcross = list()
        while flag:
            flag = False
            for inst in bbrk.getInsts():
                for wire in self.popDeferred(inst):
                    sc = _getObjectConnector(wire, inst)
                    do, dc = wdbs_getOtherEndByPoint(wire, sc)
                    if sc not in bbrk.inside:
                        # there must be a wire inside the instance for a given
                        # connector
                        raise Exception('Wire for connector ' + str(sc) +
                                        ' and instance ' + inst.getName() + " "
                                        + str(inst) + ' not found')
                    newsrc = bbrk.inside[sc]
                    for sw, so, sc in newsrc:
                        # empty  wire loop verification
                        if (so, sc, do, dc) not in memco:
                            # record every single wire we find
                            # for the first time
                            memco.append((so, sc, do, dc))
                        elif (so, sc, do, dc) in memcross:
                            raise Exception("Empty wire loop found "
                                            "with wire from " +
                                            so.getName() + " connector " +
                                            str(sc) + " and " + do.getName() +
                                            " connector " + str(dc))
                        elif isinstance(so, WdbsBlockInstanceP) \
                                and isinstance(do, WdbsBlockInstanceP):
                            # wire with same source and destination can be only
                            # to wire crossing each other or solving but it can
                            # append only one time per wire
                            memcross.append((so, sc, do, dc))
                        # Check that the wire has a different source and dest
                        # this might happen when an instance has a loop back on
                        # itself.
                        if (so, sc) == (do, dc): continue
                        # check that types can be merged
                        aliases = self.getAliases(wire)
                        aliases.extend(self.getAliases(sw))
                        if sw.getName():
                            aliases.append(sw.getName())
                        w = WdbmConnection(
                            wire.context,
                            so, sc, do, dc, wire.getName(), wire.getInfo(),
                            **{WHDB_CONNECTION_ALIASES_KEY: aliases})
                        self.__cnctnMergeF(w, sw, None, wire)
                        if self.shouldBeDeferred(w):
                            # one of the extremity of the wire can be a port
                            # or another instances that cannot be resolved now.
                            # The deferred object might be an already processed
                            # instance so we need to loop over all the inst of
                            # that hierarchy level again.
                            flag = True
                        else:
                            # the wire is fully known but can be an alias of
                            # an already added wire.
                            res = _findConnection(w, result)
                            if not res:
                                # this is a fully known wire that is not an
                                # alias of another one
                                result.append(w)
                            else:
                                # we found an alias : We are in the case where
                                # two wires of the same hierarchical level are
                                # connected because of a crossing wire of a
                                # model instance. We need to figure out the
                                # right name, the name has to be the one of
                                # the wire which source is not an instance
                                name = res.getName()
                                newa = self.getAliases(w)
                                aliases = self.getAliases(res)
                                if isinstance(so, WdbsOperatorInstanceP):
                                    # the name of that wire should be the one
                                    # of the flattened wire
                                    name = w.getName()
                                    if w.getName():
                                        aliases.remove(name)
                                else:
                                    # we need to add an alias
                                    newa.append(w.getName())
                                res.setName(name)
                                for n in newa:
                                    if n and n != name and (n not in aliases):
                                        aliases.append(n)
                                res.setMeta(aliases,
                                            WHDB_CONNECTION_ALIASES_KEY)
        return result

    def getAliases(self, wire):
        aliases = wire.getMeta(WHDB_CONNECTION_ALIASES_KEY)
        return aliases if aliases else []

    def solvePort(self, port, wires, brkr):
        """
        solve the deferred entry for the given functio port (i.e. input and
        outputs functions) and its associated wires

        :return: the list of wires to be added, may contain double
        :rtype: list[WdbsConnectionP]
        :param port: the port to be processed
        :type  port: WdbsOperatorInstanceP
        :param wires: the set of wires to be processed
        :type  wires: list[WdbsConnectionP]
        :param brkr: the mapping to be used
        :type  brkr: _BorderBreaker
        """
        result = []
        # Add the deferred entry associated to that port
        wires.extend(self.popDeferred(port))
        # all the wires connected to a port have to be resolved
        for index, wire in enumerate(wires):
            # get the two ends of the wire
            so = port
            (do, dc) = wdbs_getOtherEndByObj(wire, port)
            # there must be an associated object at the upper level for the
            # port
            newsrc = brkr.outside[brkr.portsToXpnts[so]]
            # get the associated object at the upper level for the other end
            # of the wire  if any
            # noinspection PyTypeChecker
            newdst = (brkr.outside[brkr.portsToXpnts[do]]
                      if do in brkr.portsToXpnts else [(None, do, dc)])
            # the current wire, the source wires and the destination wire have
            # to be merged.
            for sw, so, sc in newsrc:
                for dw, do, dc in newdst:
                    # check that types can be merged
                    # build the aliases
                    aliases = self.getAliases(sw)
                    aliases.extend(self.getAliases(wire))
                    if wire.getName():
                        aliases.append(wire.getName())
                    if dw is not None:
                        aliases.extend(self.getAliases(dw))
                        if dw.getName():
                            aliases.append(dw.getName())
                            # build the new name
                    name = sw.getName()
                    if index != 0:
                        name = '{n}_{i}'.format(n=name, i=index)
                    w = WdbmConnection(
                        wire.context,
                        so, sc, do, dc, name, sw.getInfo(),
                        **{WHDB_CONNECTION_ALIASES_KEY: aliases})
                    self.__cnctnMergeF(w, sw, dw, wire)
                    # one of the extremity of the wire can be a port
                    # or another instances that cannot be resolved now
                    if not self.shouldBeDeferred(w):
                        result.append(w)
        return result


class _BlockFlatteningContext(object):
    def __init__(self, block, portF, deferred, meta):
        """
        :param block: Block to flatten into
        :type block: _WhdbFlatBlockStore
        :param portF: Port-test function
        :type portF: (WdbsInstanceP) -> bool
        :param deferred: Deferrence mananger
        :type deferred: _DeferrenceManager
        :param meta: Additional meta-data for instances
        :type meta: dict[str, object]
        """
        #: the block to flatten into
        self.block = block
        #: the deferrence manager
        self.deferred = deferred
        # additional meta-data
        self.meta = meta
        # the already processed instance
        self.__doneInst = []
        #: the port test function
        self.__portF = portF

    def isPort(self, obj):
        """
        Whether the given object is a port.

        :param obj: Object to test for port
        :type obj: WdbsInstanceP
        :return: True iff the given object is a port
        :rtype: bool
        """
        return self.__portF(obj)

    def update(self, insts):
        """
        Update the list of done instances and check the deferred
        entries.

        :param insts: the last instances processed
        :type  insts:[WdbsBlockInstanceP]
        """
        l = list(self.deferred.deferred.keys())
        # There can't be any entries for an instance of that level and entries
        # for instances of lower levels in the hierachy are filtered out
        # because they will be solved by the deferred port mechanism.
        for inst in l:
            if inst in self.__doneInst:
                self.deferred.popDeferred(inst)
            if inst in insts:
                raise Exception(
                    'update failure: {}; there is a deferred entry for an '
                    'instance of that level'.format(self.__class__.__name__))
        self.__doneInst.extend(insts)


class WhdbBlockFlattenerP(metaclass=ABCMeta):
    @abstractmethod
    def flatten(self, srcBlock, **meta):
        """
        :param srcBlock: Block to be flattened
        :type srcBlock: WdbsBlockP
        :param meta: Additional meta-data for elements in flattened block
        :type meta: dict[str, object]
        :return: Flattened block
        :rtype: WdbsBlockP
        """
        raise NotImplementedError()


class WhdbBlockFlattener(WhdbBlockFlattenerP):
    """
    Base flattener implementation.
    """

    def flatten(self, srcBlock, **meta):
        block = _WhdbFlatBlockStore(srcBlock)
        wdbs_cloneBlock(srcBlock, block)

        portTstF = self.isPort
        """:type: (WdbsInstanceP) -> bool"""
        context = _BlockFlatteningContext(
            block, portTstF, self.createDeferrenceHandler(), meta)

        self.flattenBlock(block, context)

        if len(context.deferred.deferred) != 0:
            raise Exception(
                'Cannot flatten block: {}; there is some unresolved '
                'entries'.format(self.__class__.__name__))

        return block

    def createDeferrenceHandler(self):
        """
        Create a deference handler.

        :rtype: _DeferrenceManager
        """
        return _DeferrenceManager(self.isPort, self.combineConnectionMetaData)

    def createBorderBreaker(self, block):
        """
        Create a border breaker.

        :param block: Block
        :type block: WdbsBlockP
        :rtype: _BorderBreaker
        """
        return _BorderBreaker(block, self.isPort, self.getPortForXPoint)

    @abstractmethod
    def isPort(self, obj):
        """
        Whether the given object is a port.

        :param obj: Object to test for port
        :type obj: WdbsInstanceP
        :return: True iff the given object is a port
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def getPortForXPoint(self, xpnt, inst):
        """
        Get the port linked to the given connection point of the given
        instance.

        :param xpnt: Connetion point
        :type xpnt: WdbsConnectionPointP
        :param inst: Instance
        :type inst: WdbsBlockInstanceP
        :return: Port linked to the given connection point
        :rtype: WdbsOperatorInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def combineConnectionMetaData(self, cnctn, *cnctns):
        """
        Combine meta-data from the given connections into the given
        new connection.

        :param cnctn: New connection to combine meta-data for
        :type cnctn: WdbsConnectionP
        :param cnctns: Connections to combine meta-data from
        :type cnctns: Iterable[WdbsConnectionP]
        """
        raise NotImplementedError()

    def addMeta(self, entity, context):
        """
        Add additional meta-data to the given entity.

        :param entity: Entity
        :type entity: WdbsEntityP
        :type context: _BlockFlatteningContext
        """
        for k, v in context.meta.items():
            entity.setMeta(v, k)

    def flattenBlock(self, block, context, pbrk=None, prefix=None):
        """
        Flatten a hierarchical block.

        :param block: the block to flatten
        :type  block: WdbsBlockP
        :type context: _BlockFlatteningContext
        :param pbrk: the parent mapping
        :param pbrk: _WdbmBorderBreaker
        :param prefix: the prefix to be added to names
        :type  prefix: str
        """
        # do noting if none
        if block is None:
            return
        # first flatten wires
        self.flattenConnections(block, prefix)
        # create a new border breaker for this block
        bbrk = self.createBorderBreaker(block)
        # flatten the instances
        for inst in [i for i in block.getInstances()
                     if not isinstance(i, WdbsOperatorInstanceP)]:
            self.flattenInstance(inst, bbrk, prefix, context)

        # remove the instances that have been inlined (which also removes the
        # wires as a side effect).
        cdel = set()
        for inst in bbrk.getInsts():
            block.removeInstance(inst)
            cdel.update(c for c in block.getConnections()
                        if c.getStartObject() is inst or
                        c.getEndObject() is inst)
        for conn in cdel:
            block.removeConnection(conn)

        # the wires to be added at the end. The list is here to track and
        # remove duplicated wires
        awires = []

        # solve the deferred entries for that level of the hierarchy
        _mergeConnections(awires, context.deferred.solveInst(bbrk))
        context.update(bbrk.getInsts())

        # flatten the function of the block and append the created wire
        fwires = self.flattenFunctions(block, pbrk, prefix, context)
        _mergeConnections(awires, fwires)

        # add the remaining block internal wires
        for conn in block.getConnections():
            if conn not in context.block.getConnections():
                context.block.addConnection(conn)
        # add the locally computed wires
        for conn in awires:
            context.block.addConnection(conn)

    def flattenFunctions(self, block, pbrk, prefix, context):
        """
        flatten the function of the blok into the block (i.e. self).

        :return: the list of created wire
        :param block: the blok that holds the function to flatten
        :type  block: WdbsBlockP
        :param pbrk: the parent mapping
        :param pbrk: _WmhdbBorderBreaker
        :param prefix: the prefix to be added to names
        :type  prefix: str
        """
        result = []
        # check if we are self processing
        if block is not context.block:
            # get the function list
            for func in [i for i in block.getInstances()
                         if isinstance(i, WdbsOperatorInstanceP)]:
                # add all the functions, but not the ports
                if not context.isPort(func):
                    self.addMeta(func, context)
                    if prefix is not None:
                        func.setName(prefix + '_' + func.getName())
                    context.block.addInstance(func)
                else:
                    # the set of wires connected to port function, should be
                    # removed from the global list and process by the deferred
                    # object manager
                    fwires = [c for c in block.getConnections()
                              if c.getStartObject() is func or
                              c.getEndObject() is func]
                    for conn in fwires:
                        block.removeConnection(conn)
                    result.extend(
                        context.deferred.solvePort(func, fwires, pbrk))
        return result

    def flattenConnections(self, block, prefix):
        """
        flatten the wires which is just a renaming

        :type  block: WdbsBlockP
        :param prefix: the prefix to be added to names
        :type  prefix: str
        """
        # do noting if none or if prefix empty
        if block is None or not prefix:
            return
        for connection in block.getConnections():
            name = connection.getName()
            if name:
                connection.setName(prefix + '_' + name)

    def flattenInstance(self, inst, pbrk, pref, context):
        """
        Flatten an instance  block. This is equivalent to flatten the
        associated model block

        :param inst: the instance to flatten
        :type  inst: WdbsInstanceP
        :param pbrk: the parent border breaker
        :type  pbrk: _BorderBreaker
        :param pref: the prefix of the model to get here
        :type  pref: str
        """
        # check for an instance hierarchical block
        if isinstance(inst, WdbsBlockInstanceP):
            if pref is None:
                p = inst.getName()
            else:
                p = pref + '__' + inst.getName()
            # flatten the associated model
            self.flattenBlock(inst.getBlock(), context, pbrk, p)
        # Not a hierarchical block : simply add the instance if not at the
        # top level
        else:
            self.addMeta(inst, context)
            if pref is not None:
                inst.setName('{}_{}'.format(pref, inst.getName()))
            context.block.addInstance(inst)


WHDB_FLATTENED_BLOCK_CODE = 'wcr.core.block-flattened'


class _WhdbFlatBlockStore(WdbmBlock):
    def __init__(self, srcBlock, **kwargs):
        super().__init__(
            WdbmContext(srcBlock.context), WHDB_FLATTENED_BLOCK_CODE,
            srcBlock.getName(), srcBlock.getInfo(), srcBlock.getAllMeta(),
            **kwargs)

    def addInstance(self, inst):
        # noinspection PyUnresolvedReferences
        instances = self._WdbmBlock__instances
        if inst not in instances:
            instances.append(inst)

    def addConnection(self, conn):
        # noinspection PyUnresolvedReferences
        conns = self._WdbmBlock__connections
        if conn not in conns:
            conns.append(conn)
