# ----------------------------------------------------------------------------
# - whdb-block-proxy                                                         -
# - wild cruncher: block proxies                                   core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_block_proxy
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher - general hierarchy block proxies

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.db.whdb_protocols import WhdbObservableMixin
from wcr_core.db_api.whdb_block_api import WhdbEntityContainerP, WhdbBlockP
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP
from wcr_core.db_api.whdb_observer_api import WhdbObserverP, WhdbObservableP, \
    WHDB_ENTITY_RESET_EVENT


# noinspection PyAbstractClass
class WhdbBlockProxy(WhdbBlockP):
    """
    Abstract base class for block proxies.
    """

    @property
    @abstractmethod
    def proxiedElement(self):
        """
        Get the proxied element.

        :return: The proxied element
        :rtype: WhdbBlockP
        """
        raise NotImplementedError()

    def __call__(self):
        return self.proxiedElement

    @property
    def rawElement(self):
        return self.proxiedElement.rawElement

    @property
    def wrapperRegistry(self):
        return self.proxiedElement.wrapperRegistry

    def getUUID(self):
        return self.proxiedElement.getUUID()

    def getEntityType(self):
        return self.proxiedElement.getEntityType()

    def getTypeName(self):
        return self.proxiedElement.getTypeName()

    def getName(self):
        return self.proxiedElement.getName()

    def getInfo(self):
        return self.proxiedElement.getInfo()

    def setName(self, name):
        self.proxiedElement.setName(name)

    def setInfo(self, info):
        self.proxiedElement.setInfo(info)

    def getInstances(self):
        return self.proxiedElement.getInstances()

    def getBlockInstances(self):
        return self.proxiedElement.getBlockInstances()

    def getModelInstances(self):
        return self.proxiedElement.getModelInstances()

    def getOperatorInstances(self):
        return self.proxiedElement.getOperatorInstances()

    def getPorts(self):
        return self.proxiedElement.getPorts()

    def getConnections(self):
        return self.proxiedElement.getConnections()

    def getShapes(self):
        return self.proxiedElement.getShapes()

    def getButtons(self):
        return self.proxiedElement.getButtons()

    def createBlockInstance(self, *args, **kwargs):
        return self.proxiedElement.createBlockInstance(*args, **kwargs)

    def createModelInstance(self, *args, **kwargs):
        return self.proxiedElement.createModelInstance(*args, **kwargs)

    def createOperatorInstance(self, *args, **kwargs):
        return self.proxiedElement.createOperatorInstance(*args, **kwargs)

    def createConnection(self, *args, **kwargs):
        return self.proxiedElement.createConnection(*args, **kwargs)

    def createShape(self, *args, **kwargs):
        return self.proxiedElement.createShape(*args, **kwargs)

    def createButton(self, *args, **kwargs):
        return self.proxiedElement.createButton(*args, **kwargs)

    def remove(self, entity):
        self.proxiedElement.remove(entity)

    def removeInstance(self, inst):
        self.proxiedElement.removeInstance(inst)

    def removeConnection(self, cntn):
        self.proxiedElement.removeConnection(cntn)

    def removeShape(self, shape):
        self.proxiedElement.removeShape(shape)

    def removeButton(self, element):
        self.proxiedElement.removeButton(element)

    def getConnectionsForObject(self, ix, cx=None):
        return self.proxiedElement.getConnectionsForObject(ix, cx)


class WhdbHierarchyBlockProxy(WhdbBlockProxy, WhdbObservableMixin,
                              WhdbObserverP):
    """
    Stack-based proxy class for the top-level block.
    """

    def __init__(self, tphm):
        """
        :param tphm: Actual top block of the associated model hierarchy
        :type tphm: WhdbBlockP
        """
        WhdbObservableMixin.__init__(self)

        self.__topHierarchy = tphm

        #: The block model stack; the top is the "current" model
        self.__stack = []

        tphm.registerObserver(self)

    def pushInstanceBlock(self, inst):
        """
        Push an instance into the model stack

        :param inst: the instance to use for push
        :type inst: WhdbBlockInstanceP
        """
        if not isinstance(inst, WhdbBlockInstanceP):
            return

        self.proxiedElement.unregisterObserver(self)
        self.__stack.append(inst)
        self.proxiedElement.registerObserver(self)
        self._notifyObservers(WHDB_ENTITY_RESET_EVENT)

    def popModelBlock(self):
        """
        Pop a model block from the model stack

        :return: True iff an element was popped of the stack
        :rtype: bool
        """
        if self.__stack:
            self.proxiedElement.unregisterObserver(self)
            self.__stack.pop()
            self.proxiedElement.registerObserver(self)
            self._notifyObservers(WHDB_ENTITY_RESET_EVENT)
            return True
        else:
            return False

    def popTopModelBlock(self):
        """
        Remove all model blocks from the model stack

        :return: True iff there were any elements on the stack
        :rtype: bool
        """
        r = bool(self.__stack)
        self.proxiedElement.unregisterObserver(self)
        self.__stack = []
        self.proxiedElement.registerObserver(self)
        self._notifyObservers(WHDB_ENTITY_RESET_EVENT)
        return r

    def instanceStack(self):
        """
        Get a copy of the current instance stack.

        :return: Current instance stack
        :rtype: list[WmhdbBlockInstance]
        """
        return self.__stack[:]

    @property
    def proxiedElement(self):
        """
        :return: The current block model delegate
        :rtype: WmhdbBlockP
        """
        if self.__stack:
            return self.__stack[-1].getBlock()
        else:
            return self.__topHierarchy

    def dbNotify(self, obj, event, *args, **kwargs):
        self._notifyObservers(event, *args, **kwargs)


class WhdbEntityContainerProxy(WhdbEntityContainerP, WhdbObservableMixin,
                               WhdbObserverP):
    """
    Simple delegate proxy class for entity containers (incl. blocks).
    """

    def wrapperRegistry(self):
        pass

    _EMPTY = tuple()

    def __init__(self, delegate=None):
        """
        :param delegate: Actual top block of the associated model hierarchy
        :type delegate: WhdbEntityContainerP
        """
        WhdbObservableMixin.__init__(self)

        self.__delegate = delegate

        if isinstance(self.__delegate, WhdbObservableP):
            # noinspection PyUnresolvedReferences
            self.__delegate.registerObserver(self)

    def release(self):
        dlgt = self.__delegate
        if dlgt:
            dlgt.unregisterObserver(self)
        self.__delegate = None
        super().release()

    def setDelegate(self, delegate):
        if isinstance(self.__delegate, WhdbObservableP):
            # noinspection PyUnresolvedReferences
            self.__delegate.unregisterObserver(self)
        self.__delegate = delegate
        if isinstance(self.__delegate, WhdbObservableP):
            # noinspection PyUnresolvedReferences
            self.__delegate.registerObserver(self)

        self._notifyObservers(WHDB_ENTITY_RESET_EVENT)

    def getBlockInstances(self):
        return self.__delegate.getBlockInstances() \
            if self.__delegate else self._EMPTY

    def getButtons(self):
        return self.__delegate.getButtons() \
            if self.__delegate else self._EMPTY

    def getPorts(self):
        return self.__delegate.getPorts() \
            if self.__delegate else self._EMPTY

    def getModelInstances(self):
        return self.__delegate.getModelInstances() \
            if self.__delegate else self._EMPTY

    def getInstances(self):
        return self.__delegate.getInstances() \
            if self.__delegate else self._EMPTY

    def getOperatorInstances(self):
        return self.__delegate.getOperatorInstances() \
            if self.__delegate else self._EMPTY

    def getShapes(self):
        return self.__delegate.getShapes() \
            if self.__delegate else self._EMPTY

    def getConnections(self):
        return self.__delegate.getConnections() \
            if self.__delegate else self._EMPTY

    def dbNotify(self, obj, event, *args, **kwargs):
        self._notifyObservers(event, *args, **kwargs)
