# ----------------------------------------------------------------------------
# - whdb-connection                                                          -
# - wild cruncher: connections                                     core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_connection
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher - general hierarchy connections

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db_api.whdb_connection_api import WhdbConnectionP
from .whdb_protocols import *

WHDB_KEEP_PATH_KEY = 'keep_path'

WHDB_CONNECTION_ALIASES_KEY = 'aliases'


class WhdbConnection(WhdbBase, WhdbNamedEntity, WhdbHasZValue,
                     WhdbConnectionP):
    """
    Connections between instances.
    """

    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_KEEP_PATH_KEY, WDBS_BOOL_META),
        WdbsMetaField(WHDB_Z_VALUE_KEY, WDBS_FLOAT_META))

    def __init__(self, conn, wreg, pathF):
        """
        :param conn: Raw connection
        :type conn: WdbsConnectionP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        :param pathF: Path function
        :type pathF: (WhdbConnectionP) -> WhdbConnectionPathP
        """
        WhdbBase.__init__(self, conn, wreg)
        WhdbNamedEntity.__init__(self, conn)

        self.__rawConnection = conn
        self.__wrapperRegistry = wreg

        self.__path = pathF(self)
        """:type: WhdbConnectionPathP"""

    def getTypeName(self):
        return 'connection'

    # noinspection PyMethodMayBeStatic
    def getConnectionTypeDescription(self):
        return ''

    def getStartObject(self):
        return self.__wrapperRegistry.wrap(
            self.__rawConnection.getStartObject())

    def getStartConnectionPoint(self):
        return self.__wrapperRegistry.wrap(
            self.__rawConnection.getStartXPoint())

    def getEndObject(self):
        return self.__wrapperRegistry.wrap(
            self.__rawConnection.getEndObject())

    def getEndConnectionPoint(self):
        return self.__wrapperRegistry.wrap(
            self.__rawConnection.getEndXPoint())

    def isConnected(self, ix, cx=None):
        raw = self.__rawConnection
        obj = self.__wrapperRegistry.unwrap(ix)
        pnt = self.__wrapperRegistry.unwrap(cx) if cx else None
        if raw.getStartObject() is obj:
            if cx is None:
                if self._isSrcXPointValid():
                    return True
            else:
                if raw.getStartXPoint() is pnt:
                    return True
        if raw.getEndObject() is obj:
            if cx is None:
                if self._isDestXPointValid():
                    return True
            else:
                if raw.getEndXPoint() is pnt:
                    return True
        return False

    def isValid(self):
        return self._isSrcXPointValid() and self._isDestXPointValid()

    def _isSrcXPointValid(self):
        conn = self.__rawConnection.getStartXPoint()
        obj = self.__rawConnection.getStartObject()
        return conn in obj.getConnectionPoints()

    def _isDestXPointValid(self):
        conn = self.__rawConnection.getEndXPoint()
        obj = self.__rawConnection.getEndObject()
        return conn in obj.getConnectionPoints()

    def setKeepPath(self, kpth=True):
        self.__rawConnection.setMeta(kpth, WHDB_KEEP_PATH_KEY)

    def isKeepPath(self):
        return self.__rawConnection.getMeta(WHDB_KEEP_PATH_KEY) or False

    def getPath(self):
        return self.__path


def whdb_getOtherObject(cnctn, obj):
    """
    Get the other object connected to the given connection.

    :param cnctn: Connection
    :type cnctn: WhdbConnectionP
    :param obj: Connected object
    :type obj: WhdbInstanceP
    :return: Other connected object
    :rtype: (WhdbInstanceP, WhdbConnectionPointP)
    """
    if cnctn.getStartObject() is obj:
        return cnctn.getEndObject(), cnctn.getEndConnectionPoint()
    else:
        return cnctn.getStartObject(), cnctn.getStartConnectionPoint()
