# ----------------------------------------------------------------------------
# - whdb-connection-path-arc                                                 -
# - wild cruncher: arc-shaped connection path                      core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_connection_path_arc
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher arc-shaped connection path

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
import math

from silkan_common.scmn_trigonometry import scmn_perpendicularFoot, \
    scmn_getSlope, SCMN_QUARTER_CIRCLE_RAD
from wcr_core.db_api.whdb_connection_path_api import WhdbConnectionPathP, \
    WhdbConnectionPathSegmentP


def _len(a, b):
    """
    :type a: float
    :type b: float
    :rtype: float
    """
    return math.sqrt(pow(a, 2.0) + pow(b, 2.0))


class WhdbArcConnectionPath(WhdbConnectionPathP,
                            WhdbConnectionPathSegmentP):
    """
    Arc-shaped connection path.
    """

    def __init__(self, cnctn):
        """
        :param cnctn: Connection
        :type cnctn: WhdbConnectionP
        """
        self.__connection = cnctn

    def _getPath(self):
        """
        Get the raw path.

        :return: Raw path
        :rtype: WdbsPathP
        """
        relm = self.__connection.rawElement
        """:type: WdbsConnectionP"""
        path = relm.getPath()
        if not path:
            self.resetPath()
            path = relm.getPath()
        return path

    def getArcWidth(self):
        """
        Get the width of the arc, relative to the length of the direct
        distance between the centers of the two connected items.

        :return: Arc width
        :rtype: float
        """
        pnts = self._getPath().getPoints()
        return pnts[0].getX() if pnts else 0.0

    def setArcWidth(self, width):
        """
        Set the width of the arc, relative to the length of the direct
        distance between the centers of the two connected items.

        :param width: New arc width
        :type width: float
        """
        pnts = self._getPath().getPoints()
        pnts[0].setX(width)

    def resetPathSrcEnd(self):
        pass

    def resetPathDestEnd(self):
        pass

    def relocatePath(self):
        pass

    def resetPath(self):
        relm = self.__connection.rawElement
        """:type: WdbsConnectionP"""
        relm.setPath([(0.0, 0.0)])

    def translate(self, dx, dy):
        pass

    def getPathSegment(self, x, y, pad=5, mthr=3):
        return self

    # noinspection PyProtectedMember
    def move(self, x, y):
        # Centers of transition start/end state
        sx, sy = self.__connection.getStartObject().getPosition()
        ex, ey = self.__connection.getEndObject().getPosition()

        # Slope components of line between state centers ("center-line")
        dx = ex - sx
        dy = ey - sy
        dxdy = _len(dx, dy)

        # Transition start/end points on state shapes
        asx, asy, aex, aey, _, _, rpos = \
            whdb_getArcEndPoints(self.__connection)

        # Foot point of perpendicular on center-line through (x, y) ("point")
        fx, fy = scmn_perpendicularFoot(x, y, sx, sy, ex, ey)

        # Foot point of perpendicular on center-line through bezier point
        cx = sx + dx * rpos
        cy = sy + dy * rpos

        # Adjusted point for difference between the two foot points; (px, py)
        # is on perpendicular through bezier point
        px = x + cx - fx
        py = y + cy - fy

        # Bezier point for (px, py) being on bezier line
        mx = (px - 0.25 * asx - 0.25 * aex) * 2
        my = (py - 0.25 * asy - 0.25 * aey) * 2

        # Relative length of perpendicular between center-line and new
        # bezier point; 2 approximations due to instability for differently
        # sized states
        alen1 = 2 * _len(x - fx, y - fy) / dxdy
        alen2 = _len(mx - fx, my - fy) / dxdy
        alen = min(alen1, alen2)
        sign = ((ex - sx) * (y - sy) - (ey - sy) * (x - sx)) < 0

        self.setArcWidth(alen if sign else -alen)
        self.__connection.setKeepPath()

    def finish(self):
        pass

    def __str__(self):
        return 'Arc-Path'


def whdb_getArcEndPoint(sx, sy, ex, ey, objw, objh, arcl, rpos,
                        rvrs=False):
    """
    Get the arc end-point on an item.

    :param sx: X-coordinate of the center of the item
    :type sx: float
    :param sy: Y-coordinate of the center of the item
    :type sy: float
    :param ex: X-coordinate of the center of the other item
    :type ex: float
    :param ey: Y-coordinate of the center of the other item
    :type ey: float
    :param objw: Width of the item
    :type objw: float
    :param objh: Height of the item
    :type objh: float
    :param arcl: Width of the arc
    :type arcl: float
    :param rpos: Relative position of bezier-point foot-point on line
                 connecting item centers
    :type rpos: float
    :param rvrs: Whether the item is the end-point of the connection
    :type rvrs: bool
    :return: Coordinates of arc end-point in absolute coordinates
    :rtype: (float, float)
    """
    cwdth = ex - sx
    chght = ey - sy
    clen = math.sqrt(pow(cwdth, 2.0) + pow(chght, 2.0))
    alen = arcl * clen
    dlen = math.sqrt(pow(clen * rpos, 2.0) + pow(alen, 2.0))
    _a = scmn_getSlope(cwdth, chght)

    if not dlen:
        angl = SCMN_QUARTER_CIRCLE_RAD
    elif rvrs:
        angl = _a + math.asin(alen / dlen)
    else:
        angl = _a - math.asin(alen / dlen)
    asx = sx + math.cos(angl) * objw / 2
    asy = sy + math.sin(angl) * objh / 2
    return asx, asy


def whdb_getArcEndPoints(cnctn):
    """
    Get the end-points of the arc of the given connection.
    The end-points will lie on the outlines of the start/end-elements of the
    connection, and be adjusted for the width of the arc.

    :param cnctn: Connection
    :type cnctn: WhdbConnectionP
    :return: Arc end-points in absolute coordinates, the coordinates of the
             bezier point, and the relative location of the foot-point of the
             bezier point on the line connecting the two centers:
             sx, sy, ex, ey, px, py, pos
    :rtype: (float, float, float, float, float, float, float)
    """
    path = cnctn.getPath()
    """:type: WhdbArcConnectionPath"""
    arcl = path.getArcWidth()

    # Centers of transition start/end state
    sx, sy = cnctn.getStartObject().getPosition()
    ex, ey = cnctn.getEndObject().getPosition()

    # Slope components of line between state centers ("center-line")
    xdif = ex - sx
    ydif = ey - sy

    # Dimensions of start/end states
    sw, sh = cnctn.getStartObject().getDimension()
    ew, eh = cnctn.getEndObject().getDimension()

    # Relative position of foot point of perpendicular to center-line
    # through bezier point
    llen = _len(xdif, ydif)
    rpos = (llen - ew / 2 + sw / 2) / (2 * llen) if llen else 0

    # Transition start/end points on state shapes
    asx, asy = whdb_getArcEndPoint(
        sx, sy, ex, ey, sw, sh, arcl, rpos, False)
    aex, aey = whdb_getArcEndPoint(
        ex, ey, sx, sy, ew, eh, arcl, rpos, True)

    # Bezier point
    px = sx + xdif * rpos + ydif * arcl
    py = sy + ydif * rpos - xdif * arcl

    return asx, asy, aex, aey, px, py, rpos
