# ----------------------------------------------------------------------------
# - whdb_connection_path_builder_manhattan_simple                            -
# - wild cruncher: simple manhattan path builder                   core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_connection_path_builder_manhattan_simple
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher simple manhattan path builder

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from .whdb_connection_path_builder import WhdbConnectionPathBuilder, \
    WHDB_DEFAULT_MIN_LINE_LENGTH
from .whdb_connection_utils import *


class WhdbConnectionSimpleManhattanPathBuilder(WhdbConnectionPathBuilder):
    """
    Simple manhattan path builder.
    """

    def __init__(self, minLineLength=None):
        self.__minLineLength = minLineLength or WHDB_DEFAULT_MIN_LINE_LENGTH

    @staticmethod
    def _movePoint(point, drct, length):
        """
        Create a copy of the given `point`, moved by `length` in the direction
        `dir`.

        :param point: Source point
        :type point: (float, float)
        :param drct: Direction of move
        :type drct: (int, int)
        :param length: Amount to move by
        :type length: float
        :return: A point based on the source point, moved as specified.
        :rtype: (float, float)
        """
        return point[X] + length * drct[X], point[Y] + length * drct[Y]

    @staticmethod
    def _oppositeDir(drct):
        """
        Opposite direction of the given direction.

        :param drct: Direction
        :type drct: (int, int)
        :return: Opposite of the given direction
        :rtype: (int, int)
        """
        return -drct[X], -drct[Y]

    @staticmethod
    def _coordInDirDim(point, drct):
        """
        Coordinate of the point in the dimension affected by the given
        direction.

        :param point: Point
        :type point: (float, float)
        :param drct: Direction
        :type drct: (int, int)
        :return: Coordinate
        :rtype: float
        """
        return point[X] if drct[X] != 0 else point[Y]

    @staticmethod
    def _coordNotDirDim(point, drct):
        """
        Coordinate of the point in the dimension not affected by the given
        direction.

        :param point: Point
        :type point: (float, float)
        :param drct: Direction
        :type drct: (int, int)
        :return: Coordinate
        :rtype: float
        """
        return point[X] if drct[X] == 0 else point[Y]

    def createPath(self, sobj, scon, eobj, econ):
        # noinspection PyShadowingNames
        def appendPoints(ps, ps1, ps2):
            if ps[0] is not None: ps1.append(ps[0])
            if ps[1] is not None: ps2.append(ps[1])
            ps1.extend(reversed(ps2))
            return ps1

        p1 = sobj.getEffectiveConnectionPointPosition(scon)
        p2 = eobj.getEffectiveConnectionPointPosition(econ)

        ps1 = [p1]
        ps2 = [p2]

        dir1 = whdb_dirFromGravity(
            sobj.getEffectiveConnectiontPointGravity(scon))
        dir2 = whdb_dirFromGravity(
            eobj.getEffectiveConnectiontPointGravity(econ))

        ps = self._handleSimpleConnectionIfAny(p1, dir1, p2, dir2)
        if ps is not None:
            return appendPoints(ps, ps1, ps2)

        dirs1, dirs2 = self._getMacroDirections(p1, dir1, p2, dir2)

        if dir1 not in dirs1:
            p1 = self._movePoint(p1, dir1, self.__minLineLength)
            # noinspection PyTypeChecker
            ps1.append(p1)
            dir1 = next(d for d in dirs1
                        if whdb_xOrY(d) != whdb_xOrY(dir1))
        if dir2 not in dirs2:
            p2 = self._movePoint(p2, dir2, self.__minLineLength)
            # noinspection PyTypeChecker
            ps2.append(p2)
            dir2 = next(d for d in dirs2
                        if whdb_xOrY(d) != whdb_xOrY(dir2))

        ps = self._handleSimpleConnectionIfAny(p1, dir1, p2, dir2)
        if ps is not None:
            return appendPoints(ps, ps1, ps2)

        return [ps1[0], ps2[0]]

    def _handleSimpleConnectionIfAny(self, p1, dir1, p2, dir2):
        """
        Find a "simple" way to connect two rays into a path, if it exists.
        A ray is described by a starting point and a direction.
        A connection is simple if it can be made by using no more than two
        additional points in order to create the path.

        Each ray would usually represent the last open segment of one of two
        path segments which will eventually connect and form a single path.

        If no simple solution can be found, the return value will be `None`.

        :param p1: Starting point of first ray
        :type p1: (float, float)
        :param dir1: Direction of first ray
        :type dir1: (int, int)
        :param p2: Starting point of second ray
        :type p2: (float, float)
        :param dir2: Direction of second ray
        :type dir2: (int, int)
        :return: Tuple holding the points representing the next points in the
                 two respective path segments associated with the two rays.
                 Either tuple element may be `None`, indicating that no point
                 needs to be added to the respective path segment for the
                 proposed solution. The entire return value may also be `None`,
                 indicating that no simple connection could be found.
        :rtype: ((float, float), (float, float))
        """
        if dir1 == dir2:
            v1 = self._coordInDirDim(p1, dir1)
            v2 = self._coordInDirDim(p2, dir2)
            a = dir1[X] + dir1[Y]

            if a > 0:
                v = max(v1, v2) + self.__minLineLength
            else:
                v = min(v1, v2) - self.__minLineLength
            return self._movePoint(p1, dir1, abs(v1 - v)), \
                   self._movePoint(p2, dir2, abs(v2 - v))
        elif dir1 == self._oppositeDir(dir2):
            v1 = self._coordInDirDim(p1, dir1)
            v2 = self._coordInDirDim(p2, dir2)
            sgn1 = dir1[X] + dir1[Y]
            sgn2 = dir2[X] + dir2[Y]

            if (v1 < v2 and sgn1 > 0) or (v2 < v1 and sgn2 > 0):
                if self._coordNotDirDim(p1, dir1) == \
                        self._coordNotDirDim(p2, dir2):
                    return None, None
                else:
                    vd = abs((v1 + v2) / 2 - v1)
                    return self._movePoint(p1, dir1, vd), \
                           self._movePoint(p2, dir2, vd)
        else:
            v11 = self._coordInDirDim(p1, dir1)
            v12 = self._coordInDirDim(p2, dir1)
            v22 = self._coordInDirDim(p2, dir2)
            v21 = self._coordInDirDim(p1, dir2)
            sgn1 = dir1[X] + dir1[Y]
            sgn2 = dir2[X] + dir2[Y]

            d1ok = (v11 < v12 and sgn1 > 0) or (v11 > v12 and sgn1 < 0)
            d2ok = (v22 < v21 and sgn2 > 0) or (v22 > v21 and sgn2 < 0)

            if d1ok and d2ok:
                x = p1[X] if dir1[X] == 0 else p2[X]
                y = p1[Y] if dir1[Y] == 0 else p2[Y]
                return (x, y), None

    @staticmethod
    def _getMacroDirections(p1, dir1, p2, dir2) -> (list, list):
        """
        Given two rays, returns the general directions to be taken in order
        for the two associated paths to join.

        :param p1: Starting point of first ray
        :type p1: (float, float)
        :param dir1: Direction of first ray
        :type dir1: (int, int)
        :param p2: Starting point of second ray
        :type p2: (float, float)
        :param dir2: Direction of second ray
        :type dir2: (int, int)
        :return: 2-Tuple of lists of possible directions for ray 1 and ray 2
                 respectively.
        :rtype: (list[(int, int)], list[(int, int)])
        """
        if dir1 == dir2:
            return [dir1], [dir2]

        dirs1 = []
        dirs2 = []

        if p1[X] < p2[X]:
            dirs1.append(E)
            dirs2.append(W)
        elif p1[X] > p2[X]:
            dirs1.append(W)
            dirs2.append(E)
        else:
            dirs1.append(E)
            dirs2.append(E)

        if p1[Y] < p2[Y]:
            dirs1.append(S)
            dirs2.append(N)
        elif p1[Y] > p2[Y]:
            dirs1.append(N)
            dirs2.append(S)
        else:
            dirs1.append(S)
            dirs2.append(S)

        return dirs1, dirs2

    def simplifyPath(self, path, remf, pntf=None):
        return path
