# ----------------------------------------------------------------------------
# - whdb-connection-path-manhattan                                           -
# - wild cruncher: manhattan-style connectiont path                core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_connection_path_manhattan
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher manhattan-style connectiont path

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db_api.whdb_connection_path_api import WhdbConnectionPathP, \
    WhdbConnectionPathSegmentP
from .whdb_connection_utils import X, whdb_xOrY, whdb_dirFromGravity
from .whdb_connection_path_builder_manhattan import \
    WhdbConnectionManhattanPathBuilder


class WhdbManhattanConnectionPath(WhdbConnectionPathP):
    DEFAULT_PATH_BUILDER = WhdbConnectionManhattanPathBuilder()

    def __init__(self, cnctn):
        """
        :param cnctn: Connection
        :type cnctn: WhdbConnectionP
        """
        #: Connection
        self.__connection = cnctn

    def _getPath(self):
        """
        Get the raw path.

        :return: Raw path
        :rtype: WdbsPathP
        """
        relm = self.__connection.rawElement
        """:type: WdbsConnectionP"""
        path = relm.getPath()
        if not path:
            self.resetPath()
            path = relm.getPath()
        return path

    def getPoints(self):
        """
        Get the points of the manhattan-style path.

        :return: Points
        :rtype: list[ScmnPointMP]
        """
        return self._getPath().getPoints()

    def simplify(self):
        """
        Simplify the connection path. See `WhdbPathBuilder.simplifyPath()`
        for details.
        """
        path = self._getPath()
        if path is not None:
            self._getPathBuilder().simplifyPath(path, path.removePoint)

    def resetPathSrcEnd(self):
        self._resetPartialPath(self.__connection.getStartObject(),
                               self.__connection.getStartConnectionPoint(),
                               0, 1)

    def resetPathDestEnd(self):
        self._resetPartialPath(self.__connection.getEndObject(),
                               self.__connection.getEndConnectionPoint(),
                               -1, -2)

    def relocatePath(self):
        cnctn = self.__connection
        psc = cnctn.getStartObject().getEffectiveConnectionPointPosition(
            self.__connection.getStartConnectionPoint())
        pdc = cnctn.getEndObject().getEffectiveConnectionPointPosition(
            self.__connection.getEndConnectionPoint())
        pnts = self.getPoints()

        xdf1 = psc[0] - pnts[0].getX()
        ydf1 = psc[1] - pnts[0].getY()
        xdf2 = pdc[0] - pnts[-1].getX()
        ydf2 = pdc[1] - pnts[-1].getY()

        if xdf1 == xdf2 and ydf1 == ydf2:
            for p in pnts:
                p.setXY(p.getX() + xdf1, p.getY() + ydf1)

            pnts[-1].setXY(pdc[0], pdc[1])
        elif len(pnts) == 2 or not self.__connection.isKeepPath():
            self.resetPath()
        else:
            self.resetPathSrcEnd()
            self.resetPathDestEnd()

    def resetPath(self):
        self.__connection.setKeepPath(False)
        self._rebuildPath()
        self.simplify()

    def translate(self, dx, dy):
        for p in self._getPath().getPoints():
            x, y = p
            p.setXY(x + dx, y + dy)

    def _splitFirstSegment(self):
        """
        Split the first path segment into 3 to allow moving of the last of
        the newly created segments.
        """
        pnts = self.getPoints()
        ps = pnts[0]
        pe = pnts[1]
        return self._splitSegment(ps, pe, 1)

    def _splitLastSegment(self):
        """
        Split the last path segment into 3 to allow moving of the first of
        the newly created segments.
        """
        pnts = self.getPoints()
        ps = pnts[-1]
        pe = pnts[-2]
        return self._splitSegment(ps, pe, -1)

    def getPathSegment(self, x, y, pad=5, mthr=3):
        pnts = self.getPoints()

        if _whdb_isPosOnLine(x, y, pnts[0], pnts[1], pad):
            # Event on first path segment, might require split
            ps = pnts[0]
            pe = pnts[1]
            if len(pnts) == 2:
                # Only start- and endpoint in path, i.e. direct connection
                # noinspection PyTypeChecker
                return self._createPathSegment(None, None, mthr)
            elif _whdb_isLineSplittable(ps, pe):
                # Segment is long enough, split
                # noinspection PyTypeChecker
                return self._createPathSegment(None, pe, mthr)
        elif _whdb_isPosOnLine(x, y, pnts[-2], pnts[-1], pad):
            # Event on last path segment, might require split
            ps = pnts[-2]
            pe = pnts[-1]
            if _whdb_isLineSplittable(ps, pe):
                # Segment is long enough, split
                # noinspection PyTypeChecker
                return self._createPathSegment(ps, None, mthr)
        else:
            # Event on intermediate segment
            pe = None
            for p in pnts[1:-1]:
                ps = pe
                pe = p
                if ps is None: continue

                if _whdb_isPosOnLine(x, y, ps, pe, pad):
                    # Event on segment between points ps and pe
                    return self._createPathSegment(ps, pe, mthr)

    def _createPathSegment(self, ps, pe, mthr):
        """
        :param ps: First point of the segment
        :type ps: ScmnPointMP
        :param pe: Second point of the segment
        :type pe: ScmnPointMP
        :param mthr: Movement threshold for first movement.
        :type mthr: float
        :return: Created connection segment
        :rtype: _WhdbManhattanConnectionPathSegment
        """
        return _WhdbManhattanConnectionPathSegment(ps, pe, self, mthr)

    def _getPathBuilder(self):
        """
        :return: Path builder
        :rtype: WhdbConnectionPathBuilder
        """
        return self._defaultPathBuilder

    @property
    def _defaultPathBuilder(self):
        """
        :return: Default path builder
        :rtype: WhdbConnectionPathBuilder
        """
        if self.DEFAULT_PATH_BUILDER is None:
            self.DEFAULT_PATH_BUILDER = \
                WhdbConnectionManhattanPathBuilder()
        return self.DEFAULT_PATH_BUILDER

    def _resetPartialPath(self, obj, conn, idx1, idx2):
        """
        :param obj: Object at the affected path end
        :type obj: WhdbInstanceP
        :param conn: Connection point at the affected path end
        :type conn: WhdbConnectionPointP
        :param idx1: Index of first point at the affected path end
        :type idx1: int
        :param idx2: Index of second point at the affected path end
        :type idx2: int
        """
        path = self._getPath()
        points = path.getPoints()
        if len(points) <= 2 or not self.__connection.isKeepPath():
            self._rebuildPath()
        else:
            p = obj.getEffectiveConnectionPointPosition(conn)
            points[idx1].setXY(p[0], p[1])

            gvty = obj.getEffectiveConnectiontPointGravity(conn)
            if whdb_xOrY(whdb_dirFromGravity(gvty)) == X:
                points[idx2].setY(p[1])
            else:
                points[idx2].setX(p[0])

    def _rebuildPath(self):
        """
        Create a manhattan path between two connection points
        """
        cnctn = self.__connection
        plst = self._getPathBuilder().createPath(
            cnctn.getStartObject(), cnctn.getStartConnectionPoint(),
            cnctn.getEndObject(), cnctn.getEndConnectionPoint())
        self._getPathBuilder().simplifyPath(plst[:], plst.remove)
        self._setPath(plst)

    def _splitSegment(self, ps, pe, posf):
        """
        :param ps: Start point
        :type ps: ScmnPointP
        :param pe: End point
        :type pe: ScmnPointP
        :param posf: Position factor
        :type posf: int
        """
        if _whdb_dirBetweenPoints(ps, pe) == 'V':
            diff = 20 if pe.getY() > ps.getY() else -20

            self._getPath().addPoint(
                ps.getX(), ps.getY() + diff, posf * 1)
            return self._getPath().addPoint(
                ps.getX(), ps.getY() + diff, posf * 2)
        else:
            diff = 20 if pe.getX() > ps.getX() else -20

            self._getPath().addPoint(
                ps.getX() + diff, ps.getY(), posf * 1)
            return self._getPath().addPoint(
                ps.getX() + diff, ps.getY(), posf * 2)

    def _setPath(self, plst):
        """
        Set the path to the given point list.

        :param plst: Point list
        :type plst: Iterable[(float, float)]
        """
        relm = self.__connection.rawElement
        """:type: WdbsConnectionP"""
        relm.setPath(plst)

    @property
    def connection(self):
        """
        Get the connection the path belongs to.

        :return: Connection
        :rtype: WhdbConnectionP
        """
        return self.__connection

    def __str__(self):
        return 'Manhattan-Path [' + \
               ', '.join('({}, {})'.format(x, y)
                         for x, y in self.getPoints()) + ']'


class _WhdbManhattanConnectionPathSegment(WhdbConnectionPathSegmentP):
    """
    Private class to encapsulate the functionality required for moving
    segments (i.e. lines) of a connection path.
    """

    def __init__(self, ps, pe, conn, mthr):
        """
        :param ps: First point of the segment
        :type ps: ScmnPointMP
        :param pe: Second point of the segment
        :type pe: ScmnPointMP
        :param conn: Associated path
        :type conn: WhdbManhattanConnectionPath
        :param mthr: Movement threshold for first movement.
        :type mthr: float
        """
        self.__ps = ps
        self.__pe = pe
        self.__path = conn
        self.__mthr = mthr
        self.__dir = None
        self.__mvd = False

    # noinspection PyProtectedMember
    def move(self, x, y):
        if self.__ps is None:
            # Editing first segment, need to split
            self.__ps = self.__path._splitFirstSegment()
        if self.__pe is None:
            # Editing last segment, need to split
            self.__pe = self.__path._splitLastSegment()

        if self.__dir is None:
            self.__dir = _whdb_dirBetweenPoints(self.__ps, self.__pe)

        if self.__dir == 'V':
            if self.__mvd or abs(x - self.__ps.getX()) > self.__mthr:
                self.__ps.setX(x)
                self.__pe.setX(x)
                self.__mvd = True
        else:
            if self.__mvd or abs(y - self.__ps.getY()) > self.__mthr:
                self.__ps.setY(y)
                self.__pe.setY(y)
                self.__mvd = True

        self.__path.connection.setKeepPath()

    def finish(self):
        self.__path.simplify()


def _whdb_dirBetweenPoints(ps, pe):
    """
    Direction of the line connecting 2 points.
    Intended for horizontal or vertical lines only.

    :param ps: First point
    :type ps: ScmnPointP
    :param pe: Second point
    :type pe: ScmnPointP
    :return: 'H' for horizontal lines, 'V' for vertical lines.
    :rtype: String
    """
    dx = pe.getX() - ps.getX()
    dy = pe.getY() - ps.getY()
    return 'H' if abs(dx) > abs(dy) else 'V'


def _whdb_isPosOnLine(x, y, ps, pe, pad=5):
    """
    Whether a pair of coordinates lies on a line between 2 points.

    :param x: X coordinate
    :type x: float
    :param y: Y coordinate
    :type y: float
    :param ps: First point
    :type ps: ScmnPointP
    :param pe: Second point
    :type pe: ScmnPointP
    :param pad: Padding around the line to enlarge 'hit'-area
    :type pad: float
    :return: True iff the coordinates lie on the line.
    :rtype: boolean
    """
    gx = x + pad >= min(pe.getX(), ps.getX())
    lx = x - pad <= max(pe.getX(), ps.getX())
    gy = y + pad >= min(pe.getY(), ps.getY())
    ly = y - pad <= max(pe.getY(), ps.getY())
    return gx and lx and gy and ly


def _whdb_isLineSplittable(ps, pe):
    """
    Whether a line is long enough to be split into 3.

    :param ps: First point
    :type ps: ScmnPointP
    :param pe: Second point
    :type pe: ScmnPointP
    :return: True iff the line is long enough to be split.
    :rtype: boolean
    """
    return abs(ps.getX() - pe.getX()) > 20 or \
           abs(ps.getY() - pe.getY()) > 20
