# ----------------------------------------------------------------------------
# - whdb-connection-point                                                    -
# - wild cruncher: connection points                               core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_connection_point
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general hierarchy connection points

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db_api.whdb_connection_point_api import WhdbConnectionPointP
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaDescription, \
    WdbsMetaField
from wcr_core.dbstore.wdbs_meta_type import WDBS_STRING_META, \
    WDBS_RELATIVE_POINT_META, WdbsAtomicMetaTypeBase

from wcr_core.dbstore_api.wdbs_connection_point_api import \
    WdbsConnectionPointP, WDBS_XPNT_IN
from .whdb_protocols import WhdbBase, WhdbTypedEntity
from .whdb_connection_point_manager import WHDB_XPOINT_PORT, \
    WHDB_XPOINT_POSITION, WHDB_XPOINT_NAME, WHDB_XPOINT_FACING, \
    WHDB_XPOINT_DIRECTION


WHDB_DEFAULT_XPOINT_CODE = 'wcr.core.xpoint-default'

WHDB_XPNT_SHOW_TARGET = 'showPort'


class _WhdbGravityMetaType(WdbsAtomicMetaTypeBase):
    """
    Connection-point gravity meta-data type.
    """

    #: Connection point gravity values
    # noinspection PyArgumentList
    North, East, South, West = range(1, 5)

    # Translation connection point gravity model -> serialization
    _XPNT_GRAVITY_EXT = {
        North: 'north', East: 'east', West: 'west', South: 'south'}
    # Translation connection point gravity serialization -> model
    _XPNT_GRAVITY_INT = {v: k for k, v in _XPNT_GRAVITY_EXT.items()}

    def __init__(self):
        super().__init__()

    def getName(self):
        return 'gravity'

    def toString(self, metaV):
        return self._XPNT_GRAVITY_EXT.get(metaV)

    def fromString(self, inStr):
        return self._XPNT_GRAVITY_INT.get(inStr)


class WhdbConnectionPoint(WhdbBase, WhdbConnectionPointP, WhdbTypedEntity):
    """
    The WhdbConnectionPointP class is the silkan hierarchy connection point
    object (i.e. "connection point").
    A connection point is defined by name with an additional port id. A
    direction attribute is used to qualify the connection point.
    """

    #: Connection point meta-data
    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_XPOINT_NAME, WDBS_STRING_META),
        WdbsMetaField(WHDB_XPOINT_PORT, WDBS_STRING_META),
        WdbsMetaField(WHDB_XPOINT_POSITION, WDBS_RELATIVE_POINT_META),
        WdbsMetaField(WHDB_XPOINT_FACING, _WhdbGravityMetaType()))

    def __init__(self, xpoint, wreg):
        """
        :param xpoint: Raw connection point
        :type xpoint: WdbsConnectionPointP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        super().__init__(xpoint, wreg)

        self.__rawXPoint = xpoint

    @property
    def rawElement(self):
        return self.__rawXPoint

    def getTypeName(self):
        return 'connection point'

    def getName(self):
        """
        Get the connector's name.

        :return: The connector's name
        :rtype: str
        """
        return self.__rawXPoint.getMeta(WHDB_XPOINT_NAME) or ''

    def getDirection(self):
        """
        Get the connection point type (input or output).

        :return: The connection point's type
        :rtype: int
        """
        return self.__rawXPoint.getDirection()

    def getPortReference(self):
        """
        Get the connector's port reference, if any.

        :return: The connector's port reference, or `None`
        :rtype: str
        """
        return self.__rawXPoint.getMeta(WHDB_XPOINT_PORT)

    def setPortReference(self, port):
        """
        Set the connector's port reference.

        :param port: New port reference for the connector
        :type port: str
        """
        return self.__rawXPoint.setMeta(port, WHDB_XPOINT_PORT)

    def getGravity(self):
        """
        Get the connection point gravity.

        :return: The gravity of the connection point
        :rtype: int
        """
        gvty = self.__rawXPoint.getMeta(WHDB_XPOINT_FACING)
        if gvty is None:
            gvty = self.getConnectionPointDefaultGravity(self.getDirection())
        return gvty

    def setGravity(self, gvty):
        """
        Get the connection point gravity.

        :param gvty: New gravity value
        :type gvty: int
        """
        return self.__rawXPoint.setMeta(gvty, WHDB_XPOINT_FACING)

    def setDirection(self, direction):
        """
        Get the connection point direction.

        :param direction: New direction value
        :type direction: int
        """
        return self.__rawXPoint.setDirection(direction)

    def getPosition(self):
        """
        Get the connection point's position.

        :return: The position of the connection point
        :rtype: (float, float)
        """
        cpos = self.__rawXPoint.getMeta(WHDB_XPOINT_POSITION)
        return cpos if cpos else (0.0, 0.0)

    # noinspection PyShadowingNames
    def setPosition(self, x, y):
        """
        Set the connection point's position.

        :param x: The new relative X-position of the connection point
        :type x: float
        :param y: The new relative Y-position of the connection point
        :type y: float
        """
        self.__rawXPoint.setMeta((x, y), WHDB_XPOINT_POSITION)

    def getDescription(self, **kwargs):
        """
        Get the formatted name for the given connection point.

        :return: Description of the connection point
        :rtype: str
        """
        return self.getTypeName()

    @staticmethod
    def getConnectionPointDefaultGravity(cdir):
        """
        Get the default gravity for a connection point.

        :param cdir: Connection point type (input/output)
        :type cdir: int
        :return: Default gravity for the given connection point
        :rtype: int
        """
        if cdir == WDBS_XPNT_IN:
            return WhdbConnectionPointP.West
        else:
            return WhdbConnectionPointP.East

    def __str__(self):
        return 'XPoint {{cdir: {d}, {m}}}'.format(
            d=self.__rawXPoint.getDirection(),
            m=', '.join('{}: {}'.format(k, v) for k, v
                        in self.__rawXPoint.getAllMeta().items()))

    def __repr__(self):
        return '{c} {{cdir: {d}, {m}}}'.format(
            c=self.__class__.__name__, d=self.__rawXPoint.getDirection(),
            m=', '.join('{}: {}'.format(k, v) for k, v
                        in self.__rawXPoint.getAllMeta().items()))
