# ----------------------------------------------------------------------------
# - whdb-connection-point-manager                                            -
# - wild cruncher: connection point manager                        core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_connection_point_manager
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher connection point manager

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod
from collections import namedtuple


#: Connection point direction key
WHDB_XPOINT_DIRECTION = 'cdir'

#: Connection point type key
WHDB_XPOINT_TYPE = 'ctyp'

#: Connection point name key
WHDB_XPOINT_NAME = 'name'

#: Connection point port-name key
WHDB_XPOINT_PORT = 'port'

#: Connection point position key
WHDB_XPOINT_POSITION = 'cpos'

#: Connection point facing/gravity key
WHDB_XPOINT_FACING = 'gvty'

#: Connection point key
WHDB_XPOINT = 'xpnt'


#: Variable dynamic connection points code
WHDB_DYN_XPOINT_VAR = 'var'

#: Optional dynamic connection points code
WHDB_DYN_XPOINT_OPT = 'opt'

WHDB_DYN_XPOINT_NAME_F, WHDB_DYN_XPOINT_TYPE_F, WHDB_DYN_XPOINT_ARGS_F = \
    range(3)


#: Connection point description tuple
class WhdbXPointDesc(object):
    def __init__(self, ctyp, **kwargs):
        if 'gvty' in kwargs and 'fcng' not in kwargs:
            kwargs['fcng'] = kwargs['gvty']
        elif 'gvty' not in kwargs and 'fcng' in kwargs:
            kwargs['gvty'] = kwargs['fcng']

        self.__ctyp = ctyp
        self.__args = kwargs

    @property
    def ctyp(self):
        """
        :return: Connection point type
        :rtype: str
        """
        return self.__ctyp

    @property
    def kwargs(self):
        """
        :return: Connection point data
        :rtype: dict[str, objcet]
        """
        return self.__args

    def __getitem__(self, item):
        return self.__args.get(item)

    def get(self, item):
        return self.__args.get(item)


#: Dynamic connection point description tuple
WhdbDynamicXPointDesc = namedtuple(
    'WhdbDynamicXPointDesc',
    [WHDB_XPOINT_NAME, WHDB_XPOINT_TYPE, 'kwargs'])


class WhdbConnectionPointDescriptor(metaclass=ABCMeta):
    @abstractmethod
    def getDynamicXPoints(self):
        """
        Get the descriptions of the possible dynamic connection points.

        :return: Dynamic connection point descriptions
        :rtype: Iterable[WhdbDynamicXPointDesc]
        """
        raise NotImplementedError()

    @abstractmethod
    def canBatchCreateXPoint(self, xpnts, **kwargs):
        """
        Whether a new connection point with the given characteristics can be
        created.

        :param xpnts: Existing connection points
        :type xpnts: Iterable[WhdbXPointDesc]
        :param kwargs: Characteristics of new connection point
        :type kwargs: dict[str, object]
        :return: True iff the connection point can be created
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def canBatchRemoveXPoint(self, xpnts, **kwargs):
        """
        Whether a connection point with the given characteristics can be
        deleted.

        :param xpnts: Existing connection points
        :type xpnts: Iterable[WhdbXPointDesc]
        :param kwargs: Characteristics of the connection point to be deleted
        :type kwargs: dict[str, object]
        :return: True iff the connection point can be deleted
        :rtype: bool
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbConnectionPointManagerP(WhdbConnectionPointDescriptor):
    @abstractmethod
    def getInitialXPointSet(self):
        """
        Get the descriptions for the initial connection points

        :return: Initial connection points descriptions
        :rtype: Iterable[WhdbXPointDesc]
        """
        raise NotImplementedError()

    @abstractmethod
    def hasXPoint(self, clst, **kwargs):
        """
        Whether a connection point with the given characteristics exists in the
        given list.

        :param clst: Connection point list
        :type clst: list[WhdbConnectionPointP]
        :param kwargs: Characteristics to look for
        :type kwargs: dict[str, object]
        :return: True iff a connection with the given characteristics was found
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def getXPoints(self, clst, **kwargs):
        """
        Get all connection points matching the requested characteristics from
        the given list.

        :param clst: Connection point list
        :type clst: list[WhdbConnectionPointP]
        :param kwargs: Characteristics to look for
        :type kwargs: dict[str, object]
        :return: Connection points matching the requested characteristics
        :rtype: list[WhdbConnectionPointP]
        """
        raise NotImplementedError()

    @abstractmethod
    def matchesXPoint(self, xpnt, **kwargs):
        """
        Whether the given connection point matches the given characteristics.

        :param xpnt: Connection point to test
        :type xpnt: WhdbConnectionPointP
        :param kwargs: Characteristics to verify
        :type kwargs: dict[str, object]
        :return: True iff the connection point matches the characteristics
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def canInstanceCreateXPoint(self, clst, **kwargs):
        """
        Whether a connection point with the given characteristics can be
        created.

        :param clst: Existing connection points
        :type clst: list[WhdbConnectionPointP]
        :param kwargs: Characteristics of new connection point
        :type kwargs: dict[str, object]
        :return: True iff the connection point can be created
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def canInstanceRemoveXPoint(self, clst, xpnt):
        """
        Whether the given connection point can be deleted.

        :param clst: Existing connection points
        :type clst: list[WhdbConnectionPointP]
        :param xpnt: Connection point to test
        :type xpnt: dict[str, object]
        :return: True iff the connection point can be deleted
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def createInstanceXPoint(self, clst, connF, **kwargs):
        """
        Create a new connection point.

        This function creates a suitable connection point description, the
        actual creation is handled by the given `connF`.

        :param clst: Exisitng connection points
        :type clst: list[WhdbConnectionPointP]
        :param connF: Connection point creation function
        :type connF: (WhdbXPointDesc) -> WhdbConnectionPointP
        :param kwargs: Connection point data
        :type kwargs: dict[str, object]
        :return: Newly created connection point
        :rtype: WhdbConnectionPointP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeInstanceXPoint(self, clst, xpnt):
        """
        Remove a connection point.

        :param clst: Existing connection points to remove from
        :type clst: list[WhdbConnectionPointP]
        :param xpnt: Connection point to be removed
        :type xpnt: WhdbConnectionPointP
        """
        raise NotImplementedError()

    @abstractmethod
    def setInstanceXPoints(self, clst, connF, data):
        """
        Set an instances connection points from the given data in batch mode.

        :param clst: Exisiting connection points
        :type clst: list[WhdbConnectionPointP]
        :param connF: Connection point creation function
        :type connF: (WhdbXPointDesc) -> WhdbConnectionPointP
        :param data: Connection point descriptions
        :type data: list[WhdbXPointDesc]
        """
        raise NotImplementedError()

    @abstractmethod
    def areValidXPoints(self, clst):
        """
        Whether the given connection point list is valid.

        :param clst: Connection point list to validate
        :type clst: list[WhdbConnectionPointP]
        :return: True iff the connection point list is valid
                 (w/ error messages if list is invalid)
        :rtype: (bool, Iterable[str] | None)
        """
        raise NotImplementedError()

    @abstractmethod
    def resetXPointPositions(self, clst):
        """
        Reset the positions of the connection points in the given list.

        :param clst: Connection point list
        :type clst: list[WhdbConnectionPointP]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbConnectionPointManagerBase(WhdbConnectionPointManagerP):
    # Default fields to use for connection point/description matches
    _MATCH_FIELDS = {
        WHDB_XPOINT_NAME: lambda x: x.getName(),
        WHDB_XPOINT_DIRECTION: lambda x: x.getDirection(),
        WHDB_XPOINT_TYPE: lambda x: x.getEntityType(),
        WHDB_XPOINT_PORT: lambda x: x.getPortReference()}

    def hasXPoint(self, clst, **kwargs):
        return any(self.matchesXPoint(xpnt, **kwargs) for xpnt in clst)

    def getXPoints(self, clst, **kwargs):
        return [xpnt for xpnt in clst if self.matchesXPoint(xpnt, **kwargs)]

    def matchesXPoint(self, xpnt, **kwargs):
        for key, valf in self._MATCH_FIELDS.items():
            if kwargs.get(key) and valf(xpnt) != kwargs.get(key):
                return False
        return True


class WhdbFixedConnectionPointManager(WhdbConnectionPointManagerBase):
    """
    Connection point manager for fixed connection point sets.
    """

    def __init__(self, inls):
        """
        :param inls: Initial connection points points
        :type inls: Iterable[WhdbXPointDesc]
        """
        self.__initialXPoints = inls

    def getInitialXPointSet(self):
        return list(self.__initialXPoints)

    def getDynamicXPoints(self):
        return []

    def canBatchCreateXPoint(self, xpnts, **kwargs):
        return False

    def canInstanceCreateXPoint(self, clst, **kwargs):
        return False

    def canBatchRemoveXPoint(self, xpnts, **kwargs):
        return False

    def canInstanceRemoveXPoint(self, clst, xpnt):
        return False

    def createInstanceXPoint(self, clst, connF, **kwargs):
        raise Exception('Cannot create new connection points')

    def removeInstanceXPoint(self, clst, xpnt):
        raise Exception('Cannot remove connection points')

    def setInstanceXPoints(self, clst, connF, data):
        raise Exception('Cannot set connection points')

    def areValidXPoints(self, clst):
        vld = set(x.getName() for x in clst) == \
            set(x[WHDB_XPOINT_NAME] for x in self.__initialXPoints)

        return vld, ['Invalid connection points'] if not vld else None

    def resetXPointPositions(self, clst):
        for xpnt, xdef in zip(clst, self.__initialXPoints):
            assert xpnt.getName() == xdef[WHDB_XPOINT_NAME]
            xpnt.setPosition(*xdef[WHDB_XPOINT_POSITION])
            xpnt.setGravity(xdef[WHDB_XPOINT_FACING])
