# ----------------------------------------------------------------------------
# - whdb-instance                                                            -
# - wild cruncher: instances                                        core /db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_instance
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general hierarchy instances

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from copy import deepcopy
from itertools import groupby
import math

from silkan_common.scmn_geometry import ScmnRectangle
from wcr_core.common.wcmn_parameter_description import \
    WcmnRowParameterDescription
from wcr_core.db_api.whdb_connection_point_api import WhdbConnectionPointP
from wcr_core.db_api.whdb_instance_api import *
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from .whdb_protocols import *
from .whdb_connection_point_manager import WHDB_XPOINT_TYPE, \
    WHDB_XPOINT_DIRECTION
from .whdb_operator import whdb_setXPointPositions


#: Default icon size
WHDB_DEFAULT_ICON_SIZE = 50


# noinspection PyAbstractClass
class WhdbInstance(WhdbBase, WhdbNamedEntity,
                   WhdbHasZValue, WhdbHasLocation, WhdbHasDimension,
                   WhdbCanFlipRotate, WhdbHasCustomIcon, WhdbInstanceP):
    """
    Base class for instances.
    """
    META = WdbsMetaDescription(
        WhdbHasZValue.META,
        WhdbHasLocation.META,
        WhdbHasDimension.META,
        WhdbCanFlipRotate.META,
        WhdbHasCustomIcon.META)

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsInstanceP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbBase.__init__(self, inst, wreg)
        WhdbNamedEntity.__init__(self, inst)

        self.__rawInstance = inst
        self.__wrapperRegistry = wreg

    @property
    def boundingBox(self):
        # noinspection PyShadowingNames
        x, y = self.getPosition()
        w, h = self.getDimension()
        return ScmnRectangle(x - w / 2, y - h / 2, w, h)

    def _xpointMatches(self, xpnt, **kwargs):
        """
        Test whether a raw connection point matches a set of characteristics.

        :param xpnt: Raw connection point to test
        :type xpnt: WdbsConnectionPointP
        :param kwargs: Characteristics to test
        :type kwargs: dict[str, object]
        :return: True if the connection point matches all characteristics
        :rtype: bool
        """
        if WHDB_XPOINT_TYPE in kwargs:
            if not kwargs[WHDB_XPOINT_TYPE] == xpnt.getEntityType():
                return False
            del kwargs[WHDB_XPOINT_TYPE]
        if WHDB_XPOINT_DIRECTION in kwargs:
            if not kwargs[WHDB_XPOINT_DIRECTION] == xpnt.getDirection():
                return False
            del kwargs[WHDB_XPOINT_DIRECTION]

        for k, v in kwargs.items():
            if xpnt.getMeta(k) != v:
                return False

        return True

    def getConnectionPointCount(self, **kwargs):
        return len([c for c in self.__rawInstance.getConnectionPoints()
                    if self._xpointMatches(c, **kwargs)])

    def getConnectionPoints(self, **kwargs):
        """
        :type kwargs: dict[str, object]
        :rtype: Iterable[WhdbConnectionPointP]
        """
        return (self.__wrapperRegistry.wrap(c) for c
                in self.__rawInstance.getConnectionPoints()
                if self._xpointMatches(c, **kwargs))

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def getNewConnectionPointDirection(self, **kwargs):
        return WDBS_XPNT_BI

    @abstractmethod
    def _createRawConnectionPoint(self, ctyp, **kwargs):
        """
        Create a raw connection point.

        :param ctyp: Connection point type
        :type ctyp: str
        :param kwargs: Additional parameters
        :type kwargs: dict[str, object]
        :return: Newly created connection point
        :rtype: WdbsConnectionPointP
        """
        raise NotImplementedError()

    def createConnectionPoint(self, ctyp=None, **kwargs):
        return self.__wrapperRegistry.wrap(
            self._createRawConnectionPoint(ctyp, **kwargs))

    def removeConnectionPoint(self, xpnt):
        # noinspection PyTypeChecker
        self.__rawInstance.removeConnectionPoint(
            self.__wrapperRegistry.unwrap(xpnt))

    def clearConnectionPoints(self):
        self.__rawInstance.clearConnectionPoints()

    def getEffectiveConnectionPointPosition(self, xpnt):
        w, h = self.getDimension()
        ix, iy = self.getPosition()
        cx, cy = xpnt.getPosition()
        # noinspection PyShadowingNames
        x = cx * w / 2
        y = cy * h / 2

        m = self._getMatrix()
        if m is not None:
            m11, m12, m21, m22 = m
            return ix + x * m11 + y * m12, \
                   iy + x * m21 + y * m22
        else:
            return ix + x, iy + y

    def getEffectiveConnectiontPointGravity(self, xpnt):
        # check for none
        if xpnt is None: return None
        # get the original gravity
        gvty = xpnt.getGravity()

        r = self.getRotation()
        if self.getRotation() != 0:
            g = gvty - 1
            if r < math.pi / 4:
                pass
            elif r < 3 * math.pi / 4:
                g = (g + 1) % 4
            elif r < 5 * math.pi / 4:
                g = (g + 2) % 4
            else:
                g = (g + 3) % 4
            gvty = g + 1
        if self.isFlippedHorizontally():
            if gvty == WhdbConnectionPointP.East:
                gvty = WhdbConnectionPointP.West
            elif gvty == WhdbConnectionPointP.West:
                gvty = WhdbConnectionPointP.East
        if self.isFlippedVertically():
            if gvty == WhdbConnectionPointP.North:
                gvty = WhdbConnectionPointP.South
            elif gvty == WhdbConnectionPointP.South:
                gvty = WhdbConnectionPointP.North

        return gvty

    def getEffectiveDimension(self):
        w, h = self.getDimension()
        if whdb_isFlipRotation(self.getRotation()):
            return h, w
        else:
            return w, h

    # noinspection PyMethodMayBeStatic
    def getConnectionPointDescription(self, xpnt, **kwargs):
        return xpnt.getDescription(**kwargs)

    def setConnectionPointOrder(self, orderF):
        wreg = self.__wrapperRegistry
        # noinspection PyTypeChecker
        self.__rawInstance.setConnectionPointOrder(
            orderF, (wreg.wrap, wreg.unwrap))

    def getLabelAlignment(self):
        return WHDB_HALIGN_CENTER

    def getDefaultDimension(self):
        return whdb_getIconSize(self)

    def _getMatrix(self):
        """
        Get the rotation/flipping matrix for the current settings.
        Returns `None` if no rotation/flipping is set.

        :return: Rotation/flip matrix (m11, m12, m21, m22)
        :rtype: (float, float, float, float)
        """
        return whdb_getRotFlipMatrix(self.getRotation(),
                                     self.isFlippedVertically(),
                                     self.isFlippedHorizontally())


# noinspection PyAbstractClass
class _WhdbDirectInstance(WhdbInstance):
    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsInstanceP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        super().__init__(inst, wreg)

        self.__rawInstance = inst
        self.__wrapperRegistry = wreg

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        xpnt = self.__rawInstance.createConnectionPoint(
            ctyp=ctyp, **kwargs)
        return xpnt

    def canCreateConnectionPoint(self, **kwargs):
        return True

    def canRemoveConnectionPoint(self, xpnt):
        return True

    def clearConnectionPoints(self):
        self.__rawInstance.clearConnectionPoints()

    def resetConnectionPointPositions(self):
        whdb_setXPointPositions(self.getConnectionPoints())

    def __repr__(self):
        return '<{} ({}) @{}>'.format(
            self.getName(), self.__class__.__name__, id(self))


# noinspection PyAbstractClass
class WhdbBlockInstance(_WhdbDirectInstance, WhdbBlockInstanceP):
    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsBlockInstanceP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        super().__init__(inst, wreg)

        self.__rawInstance = inst
        self.__wrapperRegistry = wreg

    def getTypeName(self):
        return self.getBlock().getTypeName() + '-instance'

    def getBlock(self):
        """
        :rtype: WhdbBlock
        """
        return self.__wrapperRegistry.wrap(self.__rawInstance.getBlock())


# noinspection PyAbstractClass
class WhdbModelInstance(_WhdbDirectInstance, WhdbModelInstanceP):
    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsModelInstanceP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        super().__init__(inst, wreg)

        self.__rawInstance = inst
        self.__wrapperRegistry = wreg

    def getMedia(self):
        return self.getModel().getMedia()

    def getTypeName(self):
        return 'model-instance'

    def getModel(self):
        """
        :rtype: WhdbModelP
        """
        return self.__wrapperRegistry.wrap(self.__rawInstance.getModel())


# noinspection PyAbstractClass
class WhdbOperatorInstance(WhdbInstance, WhdbOperatorInstanceP):
    def __init__(self, operator, inst, wreg):
        """
        :param operator: Operator
        :type operator: WhdbOperatorP
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        super().__init__(inst, wreg)

        self.__rawInstance = inst
        self.__wrapperRegistry = wreg
        self.__operator = operator

        self.populateParameters(operator.getParameters(), wreg, inst)

    def populateParameters(self, parameters, wreg, inst):
        self.__parameters = {}
        for prm in parameters:
            if isinstance(prm, WcmnRowParameterDescription):
                for row in range(int(prm.getRow())):
                    key = 'channel{}'.format(row+1)
                    self.__parameters[key] = {prm.getName(): deepcopy(wreg.wrap(
                        inst.getOrCreateParameter(prm.getName()),
                        prm))
                        for prm in prm.getParameters()}
            else:
                self.__parameters[prm.getName()] = wreg.wrap(
                    inst.getOrCreateParameter(prm.getName()), prm)
    @property
    def parameters(self):
        return self.__parameters

    def getTypeName(self):
        return 'operator-instance'

    def getOperator(self):
        return self.__operator

    def canCreateConnectionPoint(self, **kwargs):
        return self.__operator.canAddConnectionPoint(
            self.getConnectionPoints(), **kwargs)

    def canRemoveConnectionPoint(self, xpnt):
        return self.__operator.canRemoveConnectionPoint(
            xpnt, self.getConnectionPoints())

    def _mkConnF(self, xpntDesc):
        """
        :type xpntDesc: WhdbXPointDesc
        :rtype: WdbsConnectionPointP
        """
        return self.__rawInstance.createConnectionPoint(
            ctyp=xpntDesc.ctyp, **xpntDesc.kwargs)

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        mkConnF = self._mkConnF
        """:type: (WhdpXPointDesc) -> WdbsConnectionPointP"""
        return self.__operator.addConnectionPoint(
            mkConnF, self.getConnectionPoints(), ctyp=ctyp, **kwargs)

    def getParameters(self):
        return iter(self.__parameters.values())

    def getParameter(self, name):
        return self.__parameters.get(name)

    def getParameterValue(self, name):
        param = self.getParameter(name)
        return param.getValue() if param else None

    def resetConnectionPointPositions(self):
        self.getOperator().resetConnectionPointPositions(
            self.getConnectionPoints())

    def getDefaultDimension(self):
        _, dim = self.getOperator().getShape()
        return dim or super().getDefaultDimension()

    def _getStrParamValue(self, name, default):
        """
        :type name: str
        :type default: str
        :rtype: str
        """
        value = self.getParameterValue(name)

        if value is not None and value != '':
            return value
        else:
            return default

    def getSymbol(self):
        paramF = self._getStrParamValue
        """:type: (str, str) -> str"""
        return self.getOperator().getSymbol(paramF)

    def getType(self):
        """:type: (str, str) -> str"""
        return self.getOperator().getType()

    def setGuiData(self, new_operator):
        """:type: new_operator: WhdbOperatorInstance"""
        x, y = new_operator.getPosition()
        self.setPosition(x, y)
        w, h = new_operator.getDimension()
        self.setDimension(w, h)
        self.setUUIDsource(new_operator.getUUID())
        hflip = new_operator.isFlippedHorizontally()
        self.setFlippedHorizontally(hflip)
        vflip = new_operator.isFlippedVertically()
        self.setFlippedVertically(vflip)
        self.setRotation(new_operator.getRotation())
        self.setZ(new_operator.getZ())

def whdb_isFlipRotation(rotation):
    """
    Whether the given rotation cause a width/height flip.

    :param rotation: Rotation
    :type rotation: float
    :return: True iff width and height are flipped under the given rotation
    :rtype: bool
    """
    return math.pi / 4 <= rotation < 3 * math.pi / 4 \
           or 5 * math.pi / 4 <= rotation


def whdb_getRotFlipMatrix(rotn, vflp, hflp):
    """
    Get the rotation/flipping matrix for the current settings.
    Returns `None` if no rotation/flipping is set.

    :param rotn: Rotation
    :type rotn: float
    :param vflp: Vertically flipped
    :type vflp: bool
    :param hflp: Horizontally flipped
    :type hflp: bool
    :return: Rotation/flip matrix (m11, m12, m21, m22)
    :rtype: (float, float, float, float)
    """
    if rotn != 0 or vflp or hflp:
        m = [1, 0, 0, 1]
        if rotn != 0:
            tpi = 2 * math.pi
            if rotn > tpi: rotn %= tpi
            if rotn < 0: rotn = -(rotn % tpi)
            if rotn < math.pi / 4:
                m = [1, 0, 0, 1]
            elif rotn < 3 * math.pi / 4:
                m = [0, -1, 1, 0]
            elif rotn < 5 * math.pi / 4:
                m = [-1, 0, 0, -1]
            else:
                m = [0, 1, -1, 0]
        if hflp:
            m[0] = -m[0]
            m[1] = -m[1]
        if vflp:
            m[2] = -m[2]
            m[3] = -m[3]
        return m
    else:
        return None


# noinspection PyShadowingNames
def whdb_getIconSize(inst):
    """
    Get the default icon size for a connection point carrying instance.

    :param inst: Instance with connection points to create box for
    :type inst: WhdbInstanceP
    :return: Icon box for the given instance
    :rtype: (float, float)
    """
    gvts = {g: len(list(i)) for g, i
            in groupby(inst.getConnectionPoints(), lambda x: x.getGravity())}

    ns = max(gvts.get(WhdbConnectionPointP.North, 0),
             gvts.get(WhdbConnectionPointP.South, 0))
    ew = max(gvts.get(WhdbConnectionPointP.East, 0),
             gvts.get(WhdbConnectionPointP.West, 0))

    w = max((ns + 1) * 20, WHDB_DEFAULT_ICON_SIZE)
    h = max((ew + 1) * 20, WHDB_DEFAULT_ICON_SIZE)

    if whdb_isFlipRotation(inst.getRotation()):
        return h, w
    else:
        return w, h
