# ----------------------------------------------------------------------------
# - whdb-meta-provider                                                       -
# - wild cruncher: default meta-data provider                      core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_meta_provider
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher silkan default hierarchy meta-data provider

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaDescription
from wcr_core.dbstore_api.wdbs_block_api import WdbsBlockP, WdbsContainerP, \
    WdbsEntityTargetP
from wcr_core.dbstore_api.wdbs_connection_api import WdbsConnectionP
from wcr_core.dbstore_api.wdbs_context_api import WdbsMetaProviderP
from wcr_core.dbstore_api.wdbs_geometry_api import WdbsPathP
from wcr_core.dbstore_api.wdbs_instance_api import WdbsOperatorInstanceP, \
    WdbsBlockInstanceP, WdbsModelInstanceP
from wcr_core.dbstore_api.wdbs_buttons_api import WdbsButtonP
from wcr_core.dbstore_api.wdbs_meta_description_api import WdbsMetaDescriptionP
from wcr_core.dbstore_api.wdbs_model_api import WdbsModelP
from wcr_core.dbstore_api.wdbs_parameter_api import WdbsParameterP
from wcr_core.dbstore_api.wdbs_shape_api import WdbsShapeP
from wcr_core.dbstore_api.wdbs_connection_point_api import WdbsConnectionPointP

from .whdb_shape import whdb_get_shape_for_name
from .whdb_block import WhdbBlock
from .whdb_connection import WhdbConnection
from .whdb_instance import WhdbBlockInstance, WhdbModelInstance, \
    WhdbOperatorInstance
from .whdb_button import WhdbButton
from .whdb_model import WhdbModel
from .whdb_connection_point import WhdbConnectionPoint


class WhdbMetaProvider(WdbsMetaProviderP):
    _EMPTY_META_DESC = WdbsMetaDescription()

    def __init__(self):
        self.__cache = {}

    # noinspection PyUnusedLocal
    def getBlockMeta(self, btyp=None):
        return WhdbBlock.META

    # noinspection PyUnusedLocal
    def getModelMeta(self, mtyp=None):
        return WhdbModel.META

    # noinspection PyUnusedLocal
    def getBlockInstanceMeta(self):
        return WhdbBlockInstance.META

    # noinspection PyUnusedLocal
    def getModelInstanceMeta(self):
        return WhdbModelInstance.META

    # noinspection PyUnusedLocal
    def getOperatorInstanceMeta(self, oper=None):
        return WhdbOperatorInstance.META

    # noinspection PyUnusedLocal
    def getConnectionMeta(self):
        return WhdbConnection.META

    # noinspection PyUnusedLocal
    def getConnectionPointMeta(self, xtyp=None):
        return WhdbConnectionPoint.META

    # noinspection PyUnusedLocal
    def getButtonMeta(self):
        return WhdbButton.META

    # noinspection PyUnusedLocal
    def getParameterMeta(self):
        return self._EMPTY_META_DESC

    # noinspection PyUnusedLocal
    def getPathMeta(self):
        return self._EMPTY_META_DESC

    # noinspection PyUnusedLocal
    def getShapeMeta(self, styp=None):
        shape = whdb_get_shape_for_name(styp)
        return shape.META if shape else None

    def _getMetaDescriptionFromCache(self, clazz, *args):
        """
        Get the meta-data description for the given class/item.

        :param clazz: Class to obtain meta-data description for
        :type clazz: class
        :param args: Arguments for sub-type detection functions
        :type args: list[object]
        :return: Meta-data description
        :rtype: WdbsMetaDescriptionP
        """
        if len(args):
            meta = self.__cache.get((clazz, tuple(args)))
            if meta:
                return meta
        else:
            meta = self.__cache.get(clazz)
            if meta:
                return meta

        if issubclass(clazz, WdbsBlockP):
            meta = self.getBlockMeta(*args)
        elif issubclass(clazz, WdbsModelP):
            meta = self.getModelMeta(*args)
        elif issubclass(clazz, WdbsBlockInstanceP):
            meta = self.getBlockInstanceMeta()
        elif issubclass(clazz, WdbsModelInstanceP):
            meta = self.getModelInstanceMeta()
        elif issubclass(clazz, WdbsOperatorInstanceP):
            meta = self.getOperatorInstanceMeta(*args)
        elif issubclass(clazz, WdbsConnectionP):
            meta = self.getConnectionMeta()
        elif issubclass(clazz, WdbsConnectionPointP):
            meta = self.getConnectionPointMeta(*args)
        elif issubclass(clazz, WdbsButtonP):
            meta = self.getButtonMeta()
        elif issubclass(clazz, WdbsParameterP):
            meta = self.getParameterMeta()
        elif issubclass(clazz, WdbsPathP):
            meta = self.getPathMeta()
        elif issubclass(clazz, WdbsShapeP):
            meta = self.getShapeMeta(*args)
        elif issubclass(clazz, WdbsContainerP):
            meta = WdbsMetaDescription()
        elif issubclass(clazz, WdbsEntityTargetP):
            meta = WdbsMetaDescription()
        else:
            raise ValueError(
                'Could not find meta-data for {}'.format(clazz))

        if len(args):
            self.__cache[(clazz, tuple(args))] = meta
        else:
            self.__cache[clazz] = meta

        return meta

    def getMetaDescription(self, clazz, *args):
        """
        Get the meta-data description for the given class/item.

        :param clazz: Class to obtain meta-data description for
        :type clazz: class
        :param args: Arguments for sub-type detection functions
        :param kwargs: Keyword arguments for sub-type detection functions
        :return: Meta-data description
        :rtype: WdbsMetaDescriptionP
        """
        return self._getMetaDescriptionFromCache(clazz, *args)
