# ----------------------------------------------------------------------------
# - whdb-operator                                                            -
# - wild cruncher: operators                                       core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_operator
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general hierarchy operators

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import OrderedDict
import re

from wcr_core.common.wcmn_parameter_description import \
    WcmnParameterDescription, WcmnRowParameterDescription
from wcr_core.db_api.whdb_connection_point_api import WhdbConnectionPointP
from wcr_core.db_api.whdb_operator_api import WhdbOperatorP
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_IN
from wcr_core.db_api.whdb_protocols_api import WHDB_HALIGN_CENTER
from wcr_core.db.whdb_connection_point import WhdbConnectionPoint
from .whdb_connection_point_manager import WhdbConnectionPointManagerP


WHDB_INPUT_OPERATOR_FLAG = 'input'

WHDB_OUTPUT_OPERATOR_FLAG = 'output'

WHDB_PORT_OPERATOR_FLAG = 'port'

WHDB_NO_UI_OPERATOR_FLAG = 'no-ui'

_WHDB_SYMBOL_VAR = re.compile('\$\((\w+)(?:\|([^}]*))?\)')


def whdb_makeConnPositions(count):
    """
    Create relative connection point positions for a set of `count` connection
    points.

    :param count: Number of connection points
    :type count: int
    :return: Relative y-positions between -1.0 and 1.0
    :rtype: Iterable[float]
    """
    return [(2.0 * i + 1.0) / count - 1.0 for i in range(count)]


def whdb_getXPointDefaultGravity(xpos, ypos):
    """
    Get the default gravity for a given connection point position.

    :param xpos: Connection point X-position
    :type xpos: float
    :param xpos: Connection point Y-position
    :type xpos: float
    :return: The connector's default gravity
    :rtype: int
    """
    if xpos > 0 and xpos > abs(ypos):
        return WhdbConnectionPointP.East
    elif xpos < 0 and abs(xpos) > abs(ypos):
        return WhdbConnectionPointP.West
    elif ypos < 0 and abs(ypos) > abs(xpos):
        return WhdbConnectionPointP.North
    else:
        return WhdbConnectionPointP.South


def whdb_getXYbyDirection(direction, posCount, xposCpt):
    gravityPosition = {'West': -1,
                       'North': -1,
                       'South': 1,
                       'East': 1}
    # switching from north/south to west/east makes
    # the axis y and x to be switched
    if direction in ('North', 'South'):
        ypos = gravityPosition[direction]
        xpos = posCount[xposCpt[direction]]
    elif direction in ('West', 'East'):
        xpos = gravityPosition[direction]
        ypos = posCount[xposCpt[direction]]
    else:
        raise ValueError('Invalid Direction value: {}, valid Directions '
                         'are [North, East, South, West]')
    return xpos, ypos

def whdb_transformDirection2Gravity(bs, count):
    """
    Get the Direction info from meta-data and transform it
    into the gravity info, then delete the Direction field to not overload
    meta-data with unnecessary information .

    :param bs: list of all info in points
    :type bs: [{}, {}, ...] or [WhhdbHardwareXPoint, WhhdbHardwareXPoint, ...]
    :param: count: Number of connection points
    :type: int
    """
    points = []
    if not bs:
        return count, points
    xposCpt = {'West': 0,
               'North': 0,
               'South': 0,
               'East': 0}

    xposCount = {'West': 0,
                 'North': 0,
                 'South': 0,
                 'East': 0}
    # Repartition of points position in x or y axis
    # depending of number of point
    xposPosition = {'West': [],
                    'North': [],
                    'South': [],
                    'East': []}
    gravityInt = {getattr(WhdbConnectionPointP, gvty): gvty for gvty in
                  ('West', 'North', 'South', 'East')}
    for pdata in bs:
        if isinstance(pdata, dict):
            direction = pdata.get('Direction')
        elif isinstance(pdata, WhdbConnectionPoint):
            direction = gravityInt[pdata.getGravity()]
        xposCount[direction] += 1
    # fill the points position for each direction
    for direction in xposPosition:
        xposPosition[direction].extend(
            whdb_makeConnPositions(xposCount[direction]))
    for pdata in bs:
        if isinstance(pdata, dict):
            direction = pdata.get('Direction')
            del pdata['Direction']
        elif isinstance(pdata, WhdbConnectionPoint):
            direction = gravityInt[pdata.getGravity()]
        posCount = xposPosition[direction]
        xpos, ypos = whdb_getXYbyDirection(direction, posCount, xposCpt)
        gravity = getattr(WhdbConnectionPointP, direction, 0)
        points.append((xpos, ypos, gravity))
        xposCpt[direction] += 1
        count -= 1
    return count, points


def whdb_makeConnXYPositions(cdir, count, bs=None):
    """
    Create relative connection point positions for a set of `count` connection
    points.

    :param count: Number of connection points
    :type count: int
    :param: bs: a list of dictionaries with points info like bool BackBone
    :type: list[{}, {}]
    :return: Relative y-positions between -1.0 and 1.0
    :rtype: list[(float, float, int)]

    """
    count, points = whdb_transformDirection2Gravity(bs, count)
    ypss = whdb_makeConnPositions(count)
    xpos = 1.0
    for ypos in ypss:
        points.append((xpos, ypos, whdb_getXPointDefaultGravity(xpos, ypos)))
    return points


def whdb_setXPointPositions(clst, cpos=None):
    """
    Set the positions for the connection points in the given list to the
    positions specified in the parameter `cpos`, or to the general default
    positions.

    :param clst: List of connection points to set positions on
    :type clst: Iterable[WhdbConnectionPointP]
    :param cpos: Optional list with default positions for the connection points
    :type cpos: Iterable[(float, float)]
    """
    if not isinstance(clst, list):
        clst = list(clst)
    if not cpos:
        xpos = [-1.0 if c.getDirection() == WDBS_XPNT_IN else 1.0
                for c in clst]
        cpos = zip(xpos, whdb_makeConnPositions(len(clst)))

    for conn, pos in zip(clst, cpos):
        conn.setPosition(*pos)
        conn.setGravity(whdb_getXPointDefaultGravity(*pos))


def _whdb_getPointListDimensions(points):
    if not points: return None

    xs, ys = list(zip(*points))
    minx = min(xs)
    maxx = max(xs)
    miny = min(ys)
    maxy = max(ys)

    return maxx - minx, maxy - miny


# noinspection PyMethodMayBeStatic
class WhdbOperator(WhdbOperatorP):
    """
    The WhdbOperator class is the silkan hierarchical operator object. An
    operator is defined by name and info. It is the operator name which
    defines its functionality. The operator contains an icon representation
    as well as a collection of connection points.
    """

    def __init__(self, code, xmng, name=None, info=None,
                 meta=None, parameters=None, flags=None,
                 icons=None, shape=None, symbol=None, labelAlign=None,
                 _id=None):
        """
        Create a new operator by name and info

        :param code: Operator code
        :type code: str
        :param xmng: Connection point manager
        :type xmng: WhdbConnectionPointManagerP
        :param name: the operator name
        :type name: str
        :param info: the operator info
        :type info: str
        :param meta: Operator meta data
        :type meta: dict[str, object]
        :param parameters: Operator parameters
                           (name, label, info, dflt-value, flags, kwargs)
        :type parameters: Iterable[(str, str, str, object, int,
                                   dict[str, object])]
        :param flags: Flags
        :type flags: set[str]
        :param icons: Operator icon paths
        :type icons: dict[str, str]
        :param shape: Operator shape
        :type shape: Iterable[(float, float)]
        :param symbol: Text symbol
        :type symbol: str
        :param labelAlign: Alignment for labels
        :type labelAlign: int
        """
        # TODO: Separate alignment enum
        align = labelAlign if labelAlign is not None else WHDB_HALIGN_CENTER

        #: Operator code
        self.__code = code
        #: Connection point set descriptions
        self.__xpntMngr = xmng
        #: the operator name
        self.__name = name or code
        #: the operator info
        self.__info = info
        #: operator id
        self.__id = _id
        #: the operators parameters
        self.__parameters = []
        #: Meta-data
        self.__meta = meta
        #: operator flags (delay/port/bus/input/output)
        self.__flags = flags or set()
        #: Operator icon paths {str:str}
        self.__icons = icons or {}
        #: Operator shape, if any
        self.__shape = shape
        #: Operator text symbol
        self.__symbol = symbol
        #: label alignment
        self.__labelAlign = align

        self.__symbolRepl = symbol and _WHDB_SYMBOL_VAR.findall(symbol)

        #: Bounding box
        self.__dimensions = _whdb_getPointListDimensions(shape)
        from wcr_core.modules.hardware.plugin.whplg_library_reader import \
            WhplgXmlParameter
        if isinstance(parameters, WhplgXmlParameter):
            # row parameters
            if parameters.ptype == 'channel':
                self.populateRowParameters(parameters)
            if parameters.ptype == 'simple':
                self.populateColParameters(parameters.prms)
        elif parameters:
            self.populateColParameters(parameters)

    def populateColParameters(self, parameters):
        self.__parameters = []
        for name, label, info, value, flags, kwargs in parameters:
            if kwargs:
                # noinspection PyTypeChecker
                self.__parameters.append(
                    WcmnParameterDescription(
                        name, label, info, value, flags, **kwargs))
            else:
                # noinspection PyTypeChecker
                self.__parameters.append(
                    WcmnParameterDescription(
                        name, label, info, value, flags))

    def populateRowParameters(self, parameters):
        self.__parameters = []
        for rowNber, param in parameters.prms:
            params = []
            for name, label, info, value, _, kwargs in param:
                params.append(WcmnParameterDescription(
                    name, label, info, value, **kwargs))
            self.__parameters.append(
                WcmnRowParameterDescription(rowNber, params))

    def getTypeName(self):
        return self.getName() + '-operator'

    def getCode(self):
        return self.__code

    def getName(self):
        return self.__name

    def getInfo(self):
        return self.__info

    def getId(self):
        return self.__id

    def getXPointDescriptor(self):
        return self.__xpntMngr

    def getFlags(self):
        return set(self.__flags)

    def getMeta(self, key):
        if self.__meta is None:
            return None
        return self.__meta.get(key)

    def getIcon(self, ifmt):
        return self.__icons.get(ifmt)

    def getShape(self):
        return self.__shape, self.__dimensions

    def getSymbol(self, valueF):
        if self.__symbolRepl:
            def repl(m):
                code = m.group(1)
                default = m.group(2) or ''
                return valueF(code, default)

            return _WHDB_SYMBOL_VAR.sub(repl, self.__symbol)
        else:
            return self.__symbol

    def getLabelAlignment(self):
        return self.__labelAlign

    def createInitialConnectionPoints(self):
        return self.__xpntMngr.getInitialXPointSet()

    def resetConnectionPointPositions(self, clst):
        self.__xpntMngr.resetXPointPositions(clst)

    def getParameters(self):
        return iter(self.__parameters)

    def hasFlag(self, flag, *_):
        return flag in self.__flags

    def canAddConnectionPoint(self, clst, **kwargs):
        return self.__xpntMngr.canInstanceCreateXPoint(clst, **kwargs)

    def addConnectionPoint(self, connF, clst, **kwargs):
        return self.__xpntMngr.createInstanceXPoint(clst, connF, **kwargs)

    def canRemoveConnectionPoint(self, xpnt, clst):
        return self.__xpntMngr.canInstanceRemoveXPoint(clst, xpnt)

    def removeConnectionPoint(self, clst, xpnt):
        if self.canRemoveConnectionPoint(xpnt, clst):
            clst.remove(xpnt)

    def __str__(self):
        return '{c}-Operator'.format(c=self.getCode())

    def __repr__(self):
        return 'WhdbOperator {{code: {c}}}'.format(c=self.getCode())


class WhdbOperatorRegistry(object):
    """
    Registry for operators.
    """

    def __init__(self):
        #: Registered operators
        self.__items = OrderedDict()

    def __getitem__(self, item):
        try:
            return self.__items[item]
        except KeyError:
            raise KeyError('COULD NOT FIND OPERATOR {}'.format(item))

    def __contains__(self, item):
        if not item:
            return False
        return item in self.__items

    def __iter__(self):
        return iter(self.__items.values())

    def _clear(self):
        self.__items.clear()

    def addOperator(self, operator):
        """
        Add an operator to the registry.

        :param operator: Operator to add
        :type operator: WhdbOperatorP
        """
        code = operator.getCode()
        self.__items[code] = operator
        shortCode = code.rsplit('.', 1)[-1]
        if shortCode != code:
            if shortCode not in self.__items:
                self.__items[shortCode] = operator

    def addOperators(self, operators):
        """
        Add a list of operators to the registry.

        :param operators: Operators to add
        :type operators: Iterable[WhdbOperatorP]
        """
        for operator in operators:
            self.addOperator(operator)

    def get(self, item):
        """
        Get the operator with the given code.

        :param item: Operator code
        :type item: str
        :return: Operator with the given code
        :rtype: WhdbOperatorP
        """
        return self.__items.get(item)

    def __str__(self):
        return str([k for k in self.__items.keys() if '.' in k])
