# ----------------------------------------------------------------------------
# - whdb-parameter                                                           -
# - wild cruncher: parameters                                      core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_parameter
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general hierarchy parameters

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db_api.whdb_parameter_api import WhdbParameterP

from wcr_core.dbstore.wdbs_meta_type import WdbsRecordMetaType, \
    WDBS_STRING_META, WdbsDictionaryMetaType


WHDB_PARAMETER_VALUE_KEY = 'value'


#: Process distribution service parameter meta-data type
WHDB_PARAMETER_META_TYPE = WdbsRecordMetaType(
    (WHDB_PARAMETER_VALUE_KEY, WDBS_STRING_META)
)

WHDB_PARAMETER_SET_META_TYPE = WdbsDictionaryMetaType(
    WDBS_STRING_META, WHDB_PARAMETER_META_TYPE)


class WhdbParameter(WhdbParameterP):
    """
    The WhdbParameterP class is the generic silkan hierarchy parameter
    object. A parameter is imply defined by a name, information string
    and value. One specificity of the object is to provide a copy mechanism.
    """

    def __init__(self, param, desc):
        """
        :param param: Raw parameter
        :type param: WdbsParameterP
        :param desc: Parameter description
        :type desc: WcmnParameterDescriptionP
        """
        self.__description = desc
        self.__rawParameter = param
        self.__default = desc.getDefaultValue(fnid=param.getUUID())

        if self.__rawParameter.getValue() is None and \
                self.__description.hasDynamicDefaultValue():
            self.__rawParameter.setValue(self.__default)

    def getDescription(self):
        """
        Get the parameter's description.

        :return: Description of the parameter
        :rtype: WcmnParameterDescriptionP
        """
        return self.__description

    def getName(self):
        """
        Get the parameter name.

        :return: Name of the parameter
        :rtype: str
        """
        return self.getDescription().getName()

    def getInfo(self):
        """
        Get the parameter info.

        :return: Info for the parameter
        :rtype: str
        """
        return self.getDescription().getInfo()

    def getValue(self):
        """
        Get the parameter value.

        :return: Value of the parameter
        :rtype: str
        """
        val = self.__rawParameter.getValue()
        return val if val is not None else self.__default

    def setValue(self, val):
        """
        Set the parameter value.

        :param val: The new parameter value
        :type val: str
        """
        if val is None and self.__description.hasDynamicDefaultValue():
            val = self.__default
        return self.__rawParameter.setValue(val)

    def getDictValue(self, k):
        """
        Get the parameter dict value.

        :return: Value of the dict parameter
        """
        return self.getDescription().getDictValue(k)

    def __str__(self):
        return '{}: {}'.format(self.getName(), self.getValue())

    def __repr__(self):
        return str(self)

    def __lt__(self, other):
        return self.getName() < other.getName()