# ----------------------------------------------------------------------------
# - whdb-path-mananger                                                       -
# - wild cruncher: uuid resolver & path manager                    core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_path_manager
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - UUID resolver & path manager

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from weakref import ref

from wcr_core.db.whdb_tree_utils import whdb_scan_tree, whdb_resolve_path


class WhdbPathManager(object):
    DELETED = (None, None)

    NOTFOUND = (None, None)

    def __init__(self, tblk):
        """
        :param tblk: Top-level block
        :type tblk: WhdbBlock
        """
        self.__topBlock = tblk

        self.__paths = {}
        """:type: dict[UUID, (weakref[WhdbBaseP], weakref[tuple])]"""

        self.__uuidPaths = {}
        """:type: dict[UUID, tuple[UUID]]"""

    def addPath(self, entity, path):
        """
        Add an entry to the path manager.

        :param entity: Referenced entity to be added
        :type entity: WhdbBaseP
        :param path: Path to entity
        :type path: Iterable[WhdbBlockInstanceP]
        """
        if entity is None or path is None:
            return
        uuid = entity.getUUID()
        if not isinstance(path, list) and not isinstance(path, tuple):
            path = list(path)
        self.__paths[uuid] = (ref(entity), tuple(ref(i) for i in path))
        self.__uuidPaths[uuid] = tuple(p.getUUID() for p in path)

    def addUuidPath(self, entity, path):
        """
        Add an entry to the path manager.

        :param entity: Referenced entity to be added
        :type entity: WhdbBaseP
        :param path: Path to entity
        :type path: Iterable[UUID]
        """
        uuid = entity.getUUID()
        if uuid not in self.__paths:
            self.__uuidPaths[uuid] = tuple(path)

    def updatePath(self, entity, path):
        """
        Update an entity's path, if it exists.

        :param entity: Entity to be updated
        :type entity: WhdbBaseP
        :param path: Path to entity
        :type path: Iterable[WhdbBlockInstanceP]
        """
        uuid = entity.getUUID()
        ntry = self.__paths.get(uuid)
        if ntry is None:
            return
        ntty, _ = ntry
        if ntty is None or ntty() is None:
            return
        self.__paths[uuid] = (ntty, [ref(i) for i in path])

    def removePath(self, uuid):
        """
        Remove the path for the given UUID.

        The UUID will be marked as non-existent, i.e. future queries for it
        will not trigger scanning.

        :param uuid: UUID to be removed
        :type uuid: UUID
        """
        if uuid in self.__paths:
            self.__paths[uuid] = self.DELETED
        if uuid in self.__uuidPaths:
            self.__uuidPaths[uuid] = self.DELETED

    def getEntityPath(self, uuid, upth=None, validate=True, scan=False):
        """
        The the entity and path for the given UUID.

        If the entity/path are unknown, the project will be scanned for it.

        If `validate` is set to `True`, any cached value will be validated
        against the current state of the project. If it is found to be invalid,
        the project will be scanned for the correct value (unless `scan` is
        `False`).

        :param uuid: UUID to return entity/path for
        :type uuid: UUID
        :return: Entity and path for the given UUID, if any
        :rtype: (WhdbBaseP, list[WhdbBlockInstanceP]) | (None, None)
        """
        if uuid not in self.__paths:
            # Not yet seen, try to resolve
            return self._resolveEntityPath(uuid, upth=upth, scan=scan)

        ntry = self.__paths[uuid]
        if ntry is self.DELETED:
            # Previously deleted, return DELETED
            return self.DELETED
        if ntry is self.NOTFOUND:
            # Previously not found w/o scan, scan if requested
            return self._scanAndUpdate(uuid) if scan else self.NOTFOUND

        nref, pref = ntry
        ntty = nref()
        if ntty is None:
            # Entity has been gc'ed, try to restore from DBS
            return self._resolveEntityPath(uuid, upth, scan=scan)

        path = [e() for e in pref]
        if validate:
            if any(e is None for e in path):
                # Path element has been gc'ed
                return self._resolveEntityPath(uuid, upth=upth, scan=scan)
            if not self._validate(uuid, ntty, path):
                # Entity no longer at path, maybe deleted?
                # Entity does not exist at path
                if scan:
                    # Rescan, may have been moved
                    return self._resolveEntityPath(uuid, upth=upth, scan=scan)
                else:
                    # No scan, return NOTFOUND
                    return self._markNotFound(uuid)

        return ntty, path

    def _resolveEntityPath(self, uuid, upth=None, scan=False):
        if upth is not None:
            if not isinstance(upth, tuple):
                upth = tuple(upth)
            self.__paths[uuid] = upth
        else:
            upth = self.__uuidPaths.get(uuid)
        if upth is self.DELETED:
            # UUID mark as deleted, return DELETED
            return self.DELETED
        elif upth:
            # Resolve UUID path
            try:
                epth = whdb_resolve_path(upth + (uuid,), self.__topBlock)
                path = epth[:-1]
                entity = epth[-1]
                self.__paths[uuid] = ref(entity), tuple(ref(p) for p in path)
                return entity, path
            except ValueError:
                rslt = self._markNotFound(uuid)
                # Element not found along path
                if scan:
                    # Rescan, may have been moved
                    return self._scanAndUpdate(uuid)
                else:
                    # No scan, return NOTFOUND
                    return rslt
        else:
            # No UUID path available, need to scan
            return self._scanAndUpdate(uuid)

    def _markDeleted(self, uuid):
        self.__paths[uuid] = self.DELETED
        if uuid in self.__uuidPaths:
            del self.__uuidPaths[uuid]
        return self.DELETED

    def _markNotFound(self, uuid):
        self.__paths[uuid] = self.NOTFOUND
        if uuid in self.__uuidPaths:
            del self.__uuidPaths[uuid]
        return self.NOTFOUND

    def _validate(self, uuid, entity, path):
        """
        Validate that the given entity and UUID match, and that the entity
        exists at the given path.

        :type uuid: UUID
        :type entity: WhdbBaseP
        :type path: list[WhdbBlockInstanceP]
        :return: True if the entity exists at the given path
        :rtype: bool
        """
        if entity.getUUID() != uuid:
            return False
        blck = self.__topBlock
        for p in path:
            if p not in blck.getInstances():
                return False
            blck = p.getBlock()
        return entity in blck.getInstances()

    def _scanAndUpdate(self, uuid):
        """
        Scan the project for the entity with the given UUID.

        If found, create an entry in the path dictionary.

        If found, the entity/path will be returned.

        :param uuid: UUID to scan for
        :type uuid: UUID
        :return: Entity and path for the given UUID, if any.
        :rtype: (WhdbBaseP, list[WhdbBlockInstanceP]) | (None, None)
        """
        rslt = next(
            whdb_scan_tree(lambda x: x.getUUID() == uuid, self.__topBlock),
            None)

        if rslt is not None:
            ntty, path = rslt
            self.__paths[uuid] = (ref(ntty), [ref(p) for p in path])
            return rslt
        else:
            return None, None
