# ----------------------------------------------------------------------------
# - whdb-protocols                                                           -
# - wild cruncher: general protocols                               core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_procotols
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general hierarchy general protocols

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from weakref import ref

from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_core.db_api.whdb_observer_api import *
from wcr_core.db_api.whdb_protocols_api import *
from wcr_core.dbstore.wdbs_meta_description import *
from wcr_core.dbstore.wdbs_meta_type import WDBS_STRING_META, WDBS_FLOAT_META,\
    WDBS_DIMENSION_META, WDBS_POINT_META, WDBS_BOOL_META
from wcr_core.dbstore_api.wdbs_entity_api import WdbsEntityP
from wcr_core.dbstore_api.wdbs_meta_description_api import *
from wcr_core.dbstore_api.wdbs_observer_api import WdbsObserverP


#: Mime type key
WHDB_MEDIA_KEY = 'mime'

#: Positition key
WHDB_POSITION_KEY = 'pos'

#: Z-Value key
WHDB_Z_VALUE_KEY = 'zpos'

#: Dimension key
WHDB_DIMENSION_KEY = 'dim'

#: Vertical flip state key
WHDB_VFLIP_KEY = 'vflip'

#: Horizontal flip state key
WHDB_HFLIP_KEY = 'hflip'

#: Rotation key
WHDB_ROTATION_KEY = 'rotation'

#: Label alignment key
WHDB_LABEL_ALIGN_KEY = 'lalgn'

#: Bounding box key
WHDB_BBOX_KEY = 'bbox'

#: Custom icon key
WHDB_ICON_PATH_KEY = 'icon-path'

#: Custom icon key
WHDB_ICON_TEXT_KEY = 'icon-text'

#: Hidden bounding box
WHDB_HIDE_BBOX_KEY = 'hide-bbox'


class WhdbObservableMixin(WhdbObservableMixinP, WcmnReleasableP):
    __slots__ = tuple()

    slots = ('__observers',)

    def __init__(self):
        self.__observers = []
        """:type: list[ReferenceType[WhdbObserverP]]"""

    def release(self):
        self._notifyObservers(WHDB_DELETE_EVENT)
        self.__observers = []

    def _hasObservers(self):
        """
        Whether the entity has any observers currently.

        :return: True iff the instance has observers
        :rtype: bool
        """
        return bool(self.__observers)

    def _notifyObservers(self, event, *args, **kwargs):
        dltd = None
        for oref in self.__observers:
            obs = oref()
            """:type: WhdbObserverP | None"""
            if obs is not None:
                # noinspection PyTypeChecker
                obs.dbNotify(self, event, *args, **kwargs)
            else:
                if dltd is None:
                    dltd = []
                dltd.append(oref)
        if dltd:
            for oref in dltd:
                self.__observers.remove(oref)

    def registerObserver(self, observer, event=None):
        self.__observers.append(ref(observer))

    def unregisterObserver(self, observer):
        oidx = None
        for i in range(len(self.__observers)):
            obs = self.__observers[i]()
            if obs is observer:
                oidx = i
                break
        if oidx is not None:
            del self.__observers[oidx]


class WhdbEntityObservableMixin(WhdbObservableMixin, WdbsObserverP):
    __slots__ = tuple()

    @property
    @abstractmethod
    def wrapperRegistry(self):
        """
        :rtype: WhdbWrapperRegistryP
        """
        raise NotImplementedError()

    def __init__(self):
        WhdbObservableMixin.__init__(self)

    def dbNotify(self, obj, event, *args, **kwargs):
        if not self._hasObservers():
            return

        wreg = self.wrapperRegistry

        if event & WDBS_CHILD_EVENT:
            self._notifyObservers(
                event, *[wreg.wrap(c) for c in args], **kwargs)
        else:
            if args:
                assert not isinstance(args[0], WdbsEntityP)
            self._notifyObservers(event, *args, **kwargs)


# noinspection PyAbstractClass
class WhdbBase(WhdbBaseP, WhdbEntityObservableMixin, WdbsObserverP):
    def __init__(self, rawElement, wreg):
        """
        :param rawElement: Raw element
        :type rawElement: WdbsEntityP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbEntityObservableMixin.__init__(self)

        self.__rawElement = rawElement
        self.__wrapperRegistry = wreg

        rawElement.registerObserver(self)

    def release(self):
        self.__rawElement.unregisterObserver(self)
        super().release()

    @property
    def wrapperRegistry(self):
        return self.__wrapperRegistry

    @property
    def rawElement(self):
        return self.__rawElement

    def getUUID(self):
        return self.__rawElement.getUUID()


# noinspection PyAbstractClass
class WhdbTypedEntity(WhdbTypedEntityP):
    @property
    @abstractmethod
    def rawElement(self):
        """
        :rtype: WdbsTypedEntityP
        """
        raise NotImplementedError()

    def getEntityType(self):
        return self.rawElement.getEntityType()


# noinspection PyAbstractClass
class WhdbNamedEntity(WhdbNamedEntityP):
    # TODO: Use abstract rawElement
    def __init__(self, rawElement):
        """
        :param rawElement: Raw element
        :type rawElement: WdbsNamedEntityP
        """
        self.__rawElement = rawElement

    def getName(self):
        return self.__rawElement.getName()

    def getInfo(self):
        return self.__rawElement.getInfo()

    def getId(self):
        return self.__rawElement.getId()

    def setName(self, name):
        self.__rawElement.setName(name)

    def setInfo(self, info):
        self.__rawElement.setInfo(info)

    def setId(self, _id):
        self.__rawElement.setId(_id)


class WhdbHasMedia(WhdbHasMediaP):
    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_MEDIA_KEY, WDBS_STRING_META))

    @property
    @abstractmethod
    def rawElement(self):
        """
        :rtype: WdbsEntityP
        """
        raise NotImplementedError()

    def getMedia(self):
        return self.rawElement.getMeta(WHDB_MEDIA_KEY)


WHDB_DEFAULT_POSITION = (0.0, 0.0)


class WhdbHasZValue(WhdbHasZValueP):
    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_Z_VALUE_KEY, WDBS_FLOAT_META))

    @property
    @abstractmethod
    def rawElement(self):
        """
        :rtype: WdbsEntityP
        """
        raise NotImplementedError()

    def getZ(self):
        return self.rawElement.getMeta(WHDB_Z_VALUE_KEY)

    def setZ(self, zval):
        self.rawElement.setMeta(zval, WHDB_Z_VALUE_KEY)


class WhdbHasLocation(WhdbHasLocationP):
    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_POSITION_KEY, WDBS_POINT_META))

    @property
    @abstractmethod
    def rawElement(self):
        """
        :rtype: WdbsEntityP
        """
        raise NotImplementedError()

    def getPosition(self):
        pos = self.rawElement.getMeta(WHDB_POSITION_KEY)
        return pos if pos else WHDB_DEFAULT_POSITION

    # noinspection PyShadowingNames
    def setPosition(self, x, y):
        self.rawElement.setMeta((x, y), WHDB_POSITION_KEY)


WHDB_DEFAULT_DIMENSION = (0.0, 0.0)


# noinspection PyAbstractClass
class WhdbHasDimension(WhdbHasDimensionP):
    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_DIMENSION_KEY, WDBS_DIMENSION_META))

    @property
    @abstractmethod
    def rawElement(self):
        """
        :rtype: WdbsEntityP
        """
        raise NotImplementedError()

    def getDimension(self):
        dim = self.rawElement.getMeta(WHDB_DIMENSION_KEY)
        return dim if dim else WHDB_DEFAULT_DIMENSION

    def setDimension(self, w, h):
        self.rawElement.setMeta((w, h), WHDB_DIMENSION_KEY)

    def unsetDimension(self):
        self.rawElement.setMeta(None, WHDB_DIMENSION_KEY)


class WhdbCanFlipRotate(WhdbCanFlipRotateP):
    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_ROTATION_KEY, WDBS_FLOAT_META),
        WdbsMetaField(WHDB_HFLIP_KEY, WDBS_BOOL_META),
        WdbsMetaField(WHDB_VFLIP_KEY, WDBS_BOOL_META))

    @property
    @abstractmethod
    def rawElement(self):
        """
        :rtype: WdbsEntityP
        """
        raise NotImplementedError()

    def isFlippedVertically(self):
        return self.rawElement.getMeta(WHDB_VFLIP_KEY) or False

    def setFlippedVertically(self, flip):
        self.rawElement.setMeta(flip, WHDB_VFLIP_KEY)

    def isFlippedHorizontally(self):
        return self.rawElement.getMeta(WHDB_HFLIP_KEY) or False

    def setFlippedHorizontally(self, flip):
        self.rawElement.setMeta(flip, WHDB_HFLIP_KEY)

    def getRotation(self):
        return self.rawElement.getMeta(WHDB_ROTATION_KEY) or 0.0

    def setRotation(self, rotation):
        self.rawElement.setMeta(rotation, WHDB_ROTATION_KEY)


class WhdbHasCustomIcon(WhdbHasCustomIconP):
    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_ICON_PATH_KEY, WDBS_STRING_META),
        WdbsMetaField(WHDB_ICON_TEXT_KEY, WDBS_STRING_META),
        WdbsMetaField(WHDB_HIDE_BBOX_KEY, WDBS_BOOL_META))

    @property
    @abstractmethod
    def rawElement(self):
        """
        :rtype: WdbsEntityP
        """
        raise NotImplementedError()

    def getCustomIcon(self):
        return self.rawElement.getMeta(WHDB_ICON_PATH_KEY)

    def setCustomIcon(self, rkey):
        self.rawElement.setMeta(rkey, WHDB_ICON_PATH_KEY)

    def getCustomText(self):
        return self.rawElement.getMeta(WHDB_ICON_TEXT_KEY)

    def setCustomText(self, text):
        self.rawElement.setMeta(text, WHDB_ICON_TEXT_KEY)

    def isBoundingBoxHidden(self):
        return self.rawElement.getMeta(WHDB_HIDE_BBOX_KEY) or False

    def setBoundingBoxHidden(self, hidden):
        self.rawElement.setMeta(hidden, WHDB_HIDE_BBOX_KEY)
