# ----------------------------------------------------------------------------
# - whdb-shape                                                               -
# - wild cruncher:shapes                                           core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_shape
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general hierarchy shapes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from wcr_core.db_api.whdb_shape_api import *
from silkan_common.scmn_geometry import ScmnRectangle
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaField, \
    WdbsMetaDescription
from wcr_core.dbstore.wdbs_meta_type import *
from wcr_core.dbstore_api.wdbs_meta_description_api import *
from .whdb_protocols import WhdbBase, WHDB_Z_VALUE_KEY, WHDB_BBOX_KEY, \
    WHDB_POSITION_KEY, WhdbHasZValue


# noinspection PyAbstractClass
class WhdbShape(WhdbBase, WhdbHasZValue, WhdbShapeP):
    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_Z_VALUE_KEY, WDBS_FLOAT_META))

    def __init__(self, shape, wreg):
        """
        :param shape: Raw shape
        :type shape: WdbsShapeP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        super().__init__(shape, wreg)

        self.__rawShape = shape

    def getTypeName(self):
        return self.shapeName + '-shape'


WHDB_LINE_COLOUR_KEY = 'line-colour'

WHDB_LINE_WIDTH_KEY = 'line-width'

WHDB_LINE_STYLE_KEY = 'line-style'


# noinspection PyAbstractClass
class WhdbLineBasedShapeMixin(WhdbLineBasedShapeP):
    """
    Base class for line-based shapes.

    A shape can be considered line-based if it consist of one or more lines,
    or if it has an outline.

    A line is characterized by its colour, width an style. If a line has a
    width of 0, it is usually considered invisible; the line's colour and style
    are without effect in this case.

    The line colour is specified as RGB hex string ('0xRRGGBB').

    The line width is specified in pixels.

    The line style is specified as int-value referring to one of the options
    provided by `WHDB_LINE_STYLE`.
    """

    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_LINE_COLOUR_KEY, WDBS_STRING_NONE_META),
        WdbsMetaField(WHDB_LINE_WIDTH_KEY, WDBS_FLOAT_NONE_META),
        WdbsMetaField(WHDB_LINE_STYLE_KEY, WDBS_INTEGER_NONE_META))

    def __init__(self, shape):
        """
        :param shape: Raw shape
        :type shape: WdbsShapeP
        """
        self.__rawShape = shape

    @property
    def lineColour(self):
        lc = self.__rawShape.getMeta(WHDB_LINE_COLOUR_KEY)
        return lc if lc else self.DEFAULT_LINE_COLOUR

    @lineColour.setter
    def lineColour(self, colour):
        self.__rawShape.setMeta(colour, WHDB_LINE_COLOUR_KEY)

    @lineColour.deleter
    def lineColour(self):
        self.__rawShape.setMeta(self.DEFAULT_LINE_COLOUR,
                                WHDB_LINE_COLOUR_KEY)

    @property
    def lineWidth(self):
        lw = self.__rawShape.getMeta(WHDB_LINE_WIDTH_KEY)
        return lw if lw is not None else self._defaultLineWidth

    @lineWidth.setter
    def lineWidth(self, width):
        self.__rawShape.setMeta(width, WHDB_LINE_WIDTH_KEY)

    @lineWidth.deleter
    def lineWidth(self):
        self.__rawShape.setMeta(self.DEFAULT_LINE_WIDTH,
                                WHDB_LINE_WIDTH_KEY)

    @property
    def lineStyle(self):
        ls = self.__rawShape.getMeta(WHDB_LINE_STYLE_KEY)
        return ls if ls is not None else self.DEFAULT_LINE_STYLE

    @lineStyle.setter
    def lineStyle(self, style):
        self.__rawShape.setMeta(style, WHDB_LINE_STYLE_KEY)

    @lineStyle.deleter
    def lineStyle(self):
        self.__rawShape.setMeta(self.DEFAULT_LINE_STYLE,
                                WHDB_LINE_STYLE_KEY)

    @property
    def _defaultLineWidth(self):
        return self.DEFAULT_LINE_WIDTH


WHDB_FILL_COLOUR_KEY = 'fill-colour'

WHDB_FILL_OPACITY_KEY = 'fill-opacity'


# noinspection PyAbstractClass
class WhdbAreaBasedShapeMixin(WhdbAreaBasedShapeP):
    """
    Base class for area based shapes.

    Area-based shapes are characterized by a fill colour and a fill opacity.
    If the fill colour is None, the fill opacity has usually no effect.

    The fill colour is specified as a RGB hex string ('0xRRGGBB').

    The fill opacity is specified as a float value between 0.0 (fully
    transparent) and 1.0 (fully opaque).
    """

    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_FILL_COLOUR_KEY, WDBS_STRING_NONE_META),
        WdbsMetaField(WHDB_FILL_OPACITY_KEY, WDBS_FLOAT_NONE_META))

    def __init__(self, shape):
        """
        :param shape: Raw shape
        :type shape: WdbsShapeP
        """
        self.__rawShape = shape

    @property
    def fillColour(self):
        return self.__rawShape.getMeta(WHDB_FILL_COLOUR_KEY)

    @fillColour.setter
    def fillColour(self, colour):
        self.__rawShape.setMeta(colour, WHDB_FILL_COLOUR_KEY)

    @fillColour.deleter
    def fillColour(self):
        self.__rawShape.setMeta(None, WHDB_FILL_COLOUR_KEY)

    @property
    def fillOpacity(self):
        fo = self.__rawShape.getMeta(WHDB_FILL_OPACITY_KEY)
        return fo if fo is not None else self.DEFAULT_FILL_OPACITY

    @fillOpacity.setter
    def fillOpacity(self, opacity):
        self.__rawShape.setMeta(opacity, WHDB_FILL_OPACITY_KEY)

    @fillOpacity.deleter
    def fillOpacity(self):
        self.__rawShape.setMeta(self.DEFAULT_FILL_OPACITY,
                                WHDB_FILL_OPACITY_KEY)

    @property
    @abstractmethod
    def width(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def height(self):
        raise NotImplementedError()


WHDB_TEXT_COLOUR_KEY = 'text-colour'

WHDB_FONT_SIZE_KEY = 'font-size'

WHDB_TEXT_HALIGN_KEY = 'text-halign'

WHDB_TEXT_VALIGN_KEY = 'text-valign'


# noinspection PyAbstractClass
class WhdbTextBasedShapeMixin(WhdbTextBasedShapeP):
    """
    Base class for text-based shapes.

    Text-based shapes are characterized by a text colour and a font size.

    The text colour is specified as RGB hex string ('0xRRGGBB').

    The font size is specified in pixels as float value.
    """

    META = WdbsMetaDescription(
        WdbsMetaField(WHDB_TEXT_COLOUR_KEY, WDBS_STRING_NONE_META),
        WdbsMetaField(WHDB_FONT_SIZE_KEY, WDBS_FLOAT_NONE_META),
        WdbsMetaField(WHDB_TEXT_HALIGN_KEY, WDBS_INTEGER_META),
        WdbsMetaField(WHDB_TEXT_VALIGN_KEY, WDBS_INTEGER_META))

    def __init__(self, shape):
        """
        :param shape: Raw shape
        :type shape: WdbsShapeP
        """
        self.__rawShape = shape

    @property
    def textColour(self):
        tc = self.__rawShape.getMeta(WHDB_TEXT_COLOUR_KEY)
        return tc if tc else self.DEFAULT_TEXT_COLOUR

    @textColour.setter
    def textColour(self, colour):
        self.__rawShape.setMeta(colour, WHDB_TEXT_COLOUR_KEY)

    @textColour.deleter
    def textColour(self):
        self.__rawShape.setMeta(self.DEFAULT_TEXT_COLOUR,
                                WHDB_TEXT_COLOUR_KEY)

    @property
    def fontSize(self):
        fs = self.__rawShape.getMeta(WHDB_FONT_SIZE_KEY)
        return fs if fs is not None else self.DEFAULT_TEXT_FONT_SIZE

    @fontSize.setter
    def fontSize(self, size):
        self.__rawShape.setMeta(size, WHDB_FONT_SIZE_KEY)

    @fontSize.deleter
    def fontSize(self):
        self.__rawShape.setMeta(self.DEFAULT_TEXT_FONT_SIZE,
                                WHDB_FONT_SIZE_KEY)

    @property
    def horizontalAlignment(self):
        ha = self.__rawShape.getMeta(WHDB_TEXT_HALIGN_KEY)
        return ha if ha is not None else self.DEFAULT_TEXT_H_ALIGNMENT

    @horizontalAlignment.setter
    def horizontalAlignment(self, alignment):
        self.__rawShape.setMeta(alignment, WHDB_TEXT_HALIGN_KEY)

    @horizontalAlignment.deleter
    def horizontalAlignment(self):
        self.__rawShape.setMeta(self.DEFAULT_TEXT_H_ALIGNMENT,
                                WHDB_TEXT_HALIGN_KEY)

    @property
    def verticalAlignment(self):
        va = self.__rawShape.getMeta(WHDB_TEXT_VALIGN_KEY)
        return va if va is not None else self.DEFAULT_TEXT_V_ALIGNMENT

    @verticalAlignment.setter
    def verticalAlignment(self, alignment):
        self.__rawShape.setMeta(alignment, WHDB_TEXT_VALIGN_KEY)

    @verticalAlignment.deleter
    def verticalAlignment(self):
        self.__rawShape.setMeta(self.DEFAULT_TEXT_V_ALIGNMENT,
                                WHDB_TEXT_VALIGN_KEY)


WHDB_LINE_POINTA_KEY = 'line-pointa'

WHDB_LINE_POINTB_KEY = 'line-pointb'


# noinspection PyAbstractClass,PyAttributeOutsideInit
class WhdbLineShape(WhdbShape, WhdbLineBasedShapeMixin, WhdbLineShapeP):
    """
    A line shape.

    A line shape is defined by its two end points.
    It's location (see :py:meth:`WhdbShapeP.location`) is defined as the
    center between the two end points.
    """

    NAME = 'line'

    META = WdbsMetaDescription(
        WhdbShape.META, WhdbLineBasedShapeMixin.META,
        WdbsMetaField(WHDB_LINE_POINTA_KEY, WDBS_POINT_META),
        WdbsMetaField(WHDB_LINE_POINTB_KEY, WDBS_POINT_META))

    def __init__(self, shape, wreg):
        """
        :param shape: Raw shape
        :type shape: WdbsShapeP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbShape.__init__(self, shape, wreg)
        WhdbLineBasedShapeMixin.__init__(self, shape)

        self.__rawShape = shape

    @property
    def pointA(self):
        return self.__rawShape.getMeta(WHDB_LINE_POINTA_KEY)

    @pointA.setter
    def pointA(self, point):
        self.__rawShape.setMeta(point, WHDB_LINE_POINTA_KEY)

    @property
    def pointB(self):
        return self.__rawShape.getMeta(WHDB_LINE_POINTB_KEY)

    @pointB.setter
    def pointB(self, point):
        self.__rawShape.setMeta(point, WHDB_LINE_POINTB_KEY)

    @property
    def pointARel(self):
        # noinspection PyShadowingNames
        x, y = self.pointA
        cx, cy = self.location
        return x - cx, y - cy

    @pointARel.setter
    def pointARel(self, point):
        # noinspection PyShadowingNames
        x, y = point
        cx, cy = self.location
        self.pointA = x + cx, y + cy

    @property
    def pointBRel(self):
        # noinspection PyShadowingNames
        x, y = self.pointB
        cx, cy = self.location
        return x - cx, y - cy

    @pointBRel.setter
    def pointBRel(self, point):
        # noinspection PyShadowingNames
        x, y = point
        cx, cy = self.location
        self.pointB = x + cx, y + cy

    @property
    def location(self):
        # Calculate and return the center
        x1, y1 = self.pointA
        x2, y2 = self.pointB
        return x1 + (x2 - x1) / 2, y1 + (y2 - y1) / 2

    @location.setter
    def location(self, point):
        # Set the two end points from the new center point.
        #
        # Calculate the difference to the previous center point
        cx, cy = self.location
        dx = point[0] - cx
        dy = point[1] - cy
        # Move the first end point by the calculated difference
        self.pointA = self._move(self.pointA, dx, dy)
        # Move the second end point by the calculated difference
        self.pointB = self._move(self.pointB, dx, dy)

    @property
    def boundingBox(self):
        x1, y1 = self.pointA
        x2, y2 = self.pointB

        return ScmnRectangle(min(x1, x2), min(y1, y2),
                             abs(x2 - x1), abs(y2 - y1))

    @property
    def shapeName(self):
        return self.NAME

    def acceptVisitor(self, visitor, *args, **kwargs):
        return visitor.visitLine(self, *args, **kwargs)

    @staticmethod
    def _move(point, dx, dy):
        return point[0] + dx, point[1] + dy


# noinspection PyAbstractClass,PyAttributeOutsideInit
class WhdbArrowShape(WhdbLineShape, WhdbArrowShapeP):
    """
    An arrow shape.

    An arrow shape is defined by its tip and end point ('tail').
    It's location (see WhdbShapeP.location) is defined as the center
    between the two points.
    """

    NAME = 'arrow'

    META = WdbsMetaDescription(
        WhdbShape.META, WhdbLineBasedShapeMixin.META,
        WdbsMetaField(WHDB_LINE_POINTA_KEY, WDBS_POINT_META),
        WdbsMetaField(WHDB_LINE_POINTB_KEY, WDBS_POINT_META))

    def __init__(self, shape, wreg):
        """
        :param shape: Raw shape
        :type shape: WdbsShapeP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        super().__init__(shape, wreg)

    @property
    def tip(self):
        return self.pointA

    @tip.setter
    def tip(self, point):
        self.pointA = point

    @property
    def endPoint(self):
        return self.pointB

    @endPoint.setter
    def endPoint(self, point):
        self.pointB = point

    @property
    def tipRel(self):
        return self.pointARel

    @property
    def endPointRel(self):
        return self.pointBRel

    @property
    def shapeName(self):
        return self.NAME

    def acceptVisitor(self, visitor, *args, **kwargs):
        return visitor.visitArrow(self, *args, **kwargs)


WHDB_AREA_WIDTH_KEY = 'area-width'

WHDB_AREA_HEIGHT_KEY = 'area-height'


# noinspection PyAbstractClass,PyAttributeOutsideInit
class WhdbRectangleShape(WhdbShape, WhdbLineBasedShapeMixin,
                         WhdbAreaBasedShapeMixin, WhdbRectangleShapeP):
    """
    A rectangle shape.

    A rectangle is defined by its upper-left corner (or center),
    width and height.
    A rectangle's location is defined as its center.
    """

    NAME = 'rectangle'

    META = WdbsMetaDescription(
        WhdbShape.META, WhdbLineBasedShapeMixin.META,
        WhdbAreaBasedShapeMixin.META,
        WdbsMetaField(WHDB_POSITION_KEY, WDBS_POINT_META),
        WdbsMetaField(WHDB_AREA_WIDTH_KEY, WDBS_FLOAT_META),
        WdbsMetaField(WHDB_AREA_HEIGHT_KEY, WDBS_FLOAT_META))

    def __init__(self, shape, wreg):
        """
        :param shape: Raw shape
        :type shape: WdbsShapeP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbShape.__init__(self, shape, wreg)
        WhdbLineBasedShapeMixin.__init__(self, shape)
        WhdbAreaBasedShapeMixin.__init__(self, shape)

        self.__rawShape = shape

    @property
    def corner(self):
        return self.__rawShape.getMeta(WHDB_POSITION_KEY)

    @corner.setter
    def corner(self, point):
        self.__rawShape.setMeta(point, WHDB_POSITION_KEY)

    @property
    def width(self):
        """
        :rtype: float
        """
        return self.__rawShape.getMeta(WHDB_AREA_WIDTH_KEY)

    @width.setter
    def width(self, width):
        self.__rawShape.setMeta(width, WHDB_AREA_WIDTH_KEY)

    @property
    def height(self):
        """
        :rtype: float
        """
        return self.__rawShape.getMeta(WHDB_AREA_HEIGHT_KEY)

    @height.setter
    def height(self, height):
        self.__rawShape.setMeta(height, WHDB_AREA_HEIGHT_KEY)

    @property
    def location(self):
        # noinspection PyShadowingNames
        x, y = self.corner
        return x + self.width / 2, y + self.height / 2

    @location.setter
    def location(self, point):
        # noinspection PyShadowingNames
        x = point[0] - self.width / 2
        y = point[1] - self.height / 2
        self.corner = x, y

    @property
    def boundingBox(self):
        # noinspection PyShadowingNames
        x, y = self.corner
        return ScmnRectangle(x, y, self.width, self.height)

    @property
    def shapeName(self):
        return self.NAME

    def acceptVisitor(self, visitor, *args, **kwargs):
        return visitor.visitRectangle(self, *args, **kwargs)


# noinspection PyAttributeOutsideInit
class WhdbOvalShape(WhdbShape, WhdbLineBasedShapeMixin,
                    WhdbAreaBasedShapeMixin, WhdbOvalShapeP):
    """
    An oval shape.

    An oval is defined by its center, width and height.
    An oval's location is its center.
    """

    NAME = 'oval'

    META = WdbsMetaDescription(
        WhdbShape.META, WhdbLineBasedShapeMixin.META,
        WhdbAreaBasedShapeMixin.META,
        WdbsMetaField(WHDB_POSITION_KEY, WDBS_POINT_META),
        WdbsMetaField(WHDB_AREA_WIDTH_KEY, WDBS_FLOAT_META),
        WdbsMetaField(WHDB_AREA_HEIGHT_KEY, WDBS_FLOAT_META))

    def __init__(self, shape, wreg):
        """
        :param shape: Raw shape
        :type shape: WdbsShapeP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbShape.__init__(self, shape, wreg)
        WhdbLineBasedShapeMixin.__init__(self, shape)
        WhdbAreaBasedShapeMixin.__init__(self, shape)

        self.__rawShape = shape

    @property
    def center(self):
        return self.__rawShape.getMeta(WHDB_POSITION_KEY)

    @center.setter
    def center(self, point):
        self.__rawShape.setMeta(point, WHDB_POSITION_KEY)

    @property
    def width(self):
        """
        :rtype: float
        """
        return self.__rawShape.getMeta(WHDB_AREA_WIDTH_KEY)

    @width.setter
    def width(self, width):
        self.__rawShape.setMeta(width, WHDB_AREA_WIDTH_KEY)

    @property
    def height(self):
        """
        :rtype: float
        """
        return self.__rawShape.getMeta(WHDB_AREA_HEIGHT_KEY)

    @height.setter
    def height(self, height):
        self.__rawShape.setMeta(height, WHDB_AREA_HEIGHT_KEY)

    @property
    def location(self):
        return self.center

    @location.setter
    def location(self, point):
        self.center = point

    @property
    def boundingBox(self):
        # noinspection PyShadowingNames
        x, y = self.center
        return ScmnRectangle(x - self.width / 2, y - self.height / 2,
                             self.width, self.height)

    @property
    def shapeName(self):
        return self.NAME

    def acceptVisitor(self, visitor, *args, **kwargs):
        return visitor.visitOval(self, *args, **kwargs)


WHDB_TEXT_KEY = 'text'


# noinspection PyAbstractClass,PyAttributeOutsideInit
class WhdbTextShape(WhdbShape, WhdbTextBasedShapeMixin,
                    WhdbLineBasedShapeMixin, WhdbAreaBasedShapeMixin,
                    WhdbTextShapeP):
    """
    A text shape.

    A text is defined by the text contents and its location.
    A text's location is the upper-left corner of its bounding box.
    """

    NAME = 'text'

    META = WdbsMetaDescription(
        WhdbShape.META,
        WhdbLineBasedShapeMixin.META,
        WhdbAreaBasedShapeMixin.META,
        WhdbTextBasedShapeMixin.META,
        WdbsMetaField(WHDB_TEXT_KEY, WDBS_TEXT_META),
        WdbsMetaField(WHDB_POSITION_KEY, WDBS_POINT_META),
        WdbsMetaField(WHDB_BBOX_KEY, WDBS_RECTANGLE_META))

    def __init__(self, shape, wreg):
        """
        :param shape: Raw shape
        :type shape: WdbsShapeP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbShape.__init__(self, shape, wreg)
        WhdbTextBasedShapeMixin.__init__(self, shape)
        WhdbLineBasedShapeMixin.__init__(self, shape)
        WhdbAreaBasedShapeMixin.__init__(self, shape)

        self.__rawShape = shape

    @property
    def location(self):
        return self.__rawShape.getMeta(WHDB_POSITION_KEY)

    @location.setter
    def location(self, point):
        self.__rawShape.setMeta(point, WHDB_POSITION_KEY)

    @property
    def text(self):
        return self.__rawShape.getMeta(WHDB_TEXT_KEY)

    @text.setter
    def text(self, txt):
        self.__rawShape.setMeta(txt, WHDB_TEXT_KEY)

    def setText(self, text):
        self.text = text

    @property
    def shapeName(self):
        return self.NAME

    def acceptVisitor(self, visitor, *args, **kwargs):
        return visitor.visitText(self, *args, **kwargs)

    @property
    def boundingBox(self):
        """
        :rtype: ScmnRectangle
        """
        bbox = self.__rawShape.getMeta(WHDB_BBOX_KEY)
        if bbox:
            # noinspection PyArgumentList
            return ScmnRectangle(*bbox)
        else:
            return None

    # noinspection PyShadowingNames
    def setBoundingBox(self, x, y, width, height):
        self.__rawShape.setMeta((x, y, width, height), WHDB_BBOX_KEY)

    @property
    def width(self):
        bbox = self.boundingBox
        return bbox.width if bbox is not None else 0

    @property
    def height(self):
        bbox = self.boundingBox
        return bbox.height if bbox is not None else 0

    @property
    def _defaultLineWidth(self):
        return 0


_WHDB_SHAPES = [WhdbLineShape, WhdbArrowShape,
                WhdbRectangleShape, WhdbOvalShape,
                WhdbTextShape]

_WHDB_SHAPES_BY_NAME = {s.NAME: s for s in _WHDB_SHAPES}
_WHDB_SHAPES_BY_NAME['oval'] = WhdbOvalShape


def whdb_get_shape_for_name(name):
    """
    Get the shape class for the given shape name.

    :param name: Shape name to return shape class for
    :type name: str
    :return: Shape class for the given name
    :rtype: class
    """
    return _WHDB_SHAPES_BY_NAME.get(name)
