# ----------------------------------------------------------------------------
# - whdb-tree-utils                                                          -
# - wild cruncher: db tree utility functions                       core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_tree_utils
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - DB tree utility functions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore_api.wdbs_instance_api import WdbsBlockInstanceP


def whdb_scan_tree(rawPredicate, block):
    """
    Scan the block recursively for entities whose rawElements match the given
    predicate.

    :param rawPredicate: Test predicate
    :type rawPredicate: (WdbsEntityP) -> bool
    :param block: Block to scan for elements
    :type block: WhdbBlock
    :return: Entity and path for the given UUID, if any.
    :rtype: Iterator[(WhdbBaseP, tuple[WhdbBlockInstanceP])]
    """
    # noinspection PyShadowingNames
    def scan(blck, path):
        """
        :type blck: WdbsBlockP
        :type path: tuple[WdbsBlockInstanceP]
        :rtype: Iterable[(WhdbBaseP, tuple[WhdbBlockInstanceP])]
        """
        for inst in blck.getInstances():
            if rawPredicate(inst):
                w = block.wrapperRegistry
                yield w.wrap(inst), [w.wrap(p) for p in path]
        blks = (i for i in blck.getInstances()
                if isinstance(i, WdbsBlockInstanceP))
        """:type: Iterable[WdbsBlockInstanceP]"""
        for inst in blks:
            _blk = inst.getBlock()
            for x in scan(_blk, path + (inst,)):
                yield x

    return scan(block.rawElement, tuple())


def whdb_validate_path(path, block):
    """
    Check whether the given path of (block-)instances exists in the given
    block.

    :param path: Path of (block-)instances
    :type path: Iterable[WhdbBlockInstanceP]
    :param block: Block to test for existence of path
    :type block: WhdbBlockP
    :return: True iff the path exists
    :rtype: bool
    """
    for p in path:
        if p not in block.getInstances():
            return False
        block = p.getBlock()
    return True


def whdb_resolve_path(uuids, block):
    """
    Resolve the given UUID-path.

    :param uuids: UUID-path
    :type uuids: Iterable[UUID]
    :param block: Block to resolve the UUID path in
    :type block: WhdbBlockP
    :return: Instance-path corresponding to the given UUID-path
    :rtype: list[WhdbInstanceP]
    """
    if not isinstance(uuids, list) and not isinstance(uuids, tuple):
        uuids = list(uuids)

    # noinspection PyShadowingNames
    def resolve(uuid):
        inst = next(
            (i for i in block.getInstances() if i.getUUID() == uuid), None)
        if not inst:
            raise ValueError(
                'Could not find instance with UUID {}'.format(uuid))
        path.append(inst)
        return inst

    path = []
    if not uuids:
        return []
    for uuid in uuids[:-1]:
        inst = resolve(uuid)
        block = inst.getBlock()
    resolve(uuids[-1])
    assert len(path) == len(uuids)
    return path


def whdb_find_by_uuid(uuid, block):
    """
    Find the instance with the given UUID in the given block, recursively.

    :param uuid: UUID to look for
    :type uuid: UUID
    :param block: Block to look in for the given UUID
    :type block: WhdbBlock
    :return: Instance with the given UUID and its path, if any
    :rtype: (WhdbBaseP, tuple[WhdbBlockInstanceP]) | (None, None)
    """
    return next(
        (whdb_scan_tree(lambda x: x.getUUID() == uuid, block)),
        (None, None))
