# ----------------------------------------------------------------------------
# - whdb-utils                                                               -
# - wild cruncher: utility functions                               core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_utils
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general hierarchy utility functions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
import re

from silkan_common.scmn_mime_type import SCMN_MIMES, SCMN_MIME_EXTS
from silkan_common.scmn_file_path import scmn_path_extension, scmn_expshenv
from wcr_core.common.wcmn_file_handler import WcmnFile, WcmnReadPayloadHandler
from wcr_core.common.wcmn_mime_type import WCMN_PAYLOAD_ENTRY_POINTS, \
    WCMN_PAYLOAD_MODEL_PORTS
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP, \
    WhdbOperatorInstanceP


_NAME_PATTERN = re.compile('(.*?)\s*\(\d+\)\s*')


def whdb_ensureUniqueName(inst, block):
    """
    Ensure that the given instance has a unique name, adjusting it as needed.

    :param inst: Instance to generate name for
    :type inst: WhdbNamedEntity
    :param block: Model block of the function
    :type block: WhdbBlockP
    """

    def adjName(orig):
        inms = set()
        for l in [block.getInstances(), block.getConnections()]:
            inms.update(i.getName() for i in l)
        m = _NAME_PATTERN.search(orig)
        base = m.group(1) if m else orig
        # noinspection PyShadowingNames
        name = base
        icnt = 0

        while name in inms:
            icnt += 1
            # noinspection PyShadowingNames
            name = '{o}_{c}'.format(o=base, c=icnt)

        if inst.getName() != name:
            inst.setName(name)

    name = inst.getName()
    if not name:
        if isinstance(inst, WhdbOperatorInstanceP):
            inst.setName(whdb_getNextFunctionName(inst.getOperator(),
                                                  block))
        elif isinstance(inst, WhdbInstanceP):
            inst.setName(whdb_getNextInstanceName(block))
    else:
        adjName(name)


def whdb_getNextFunctionName(operator, block):
    """
    Generate a name for the given name function.

    :param operator: Function to generate name for
    :type operator: WhdbOperatorP
    :param block: Model block of the function
    :type block: WhdbBlockP
    :return: New function name
    :rtype: str
    """
    return whdb_getNextName(operator.getName(), block.getOperatorInstances())


def whdb_getNextInstanceName(block):
    """
    Generate a new name for a model instance.

    :param block: Model block of the instance
    :type block: WhdbBlockP
    :return: New instance name
    :rtype: str
    """
    return whdb_getNextName('instance', block.getInstances())


def whdb_getNextName(prefix, items):
    """
    Generate the next item name for the given name prefix.

    :param prefix: Name prefix
    :type prefix: str
    :param items: List of items to find item name with highest number in
    :type items: Iterable[WhdbInstanceP]
    :return: Next item name
    :rtype: str
    """
    # try to guess if the prefix is already index
    # so that we can continue indexing
    isAlreadyIndexed = False
    splittedPrefix = prefix.split('_')
    if len(splittedPrefix) > 0:
        lastPrefix = splittedPrefix[-1]
        allButLastPrefix = '_'.join(splittedPrefix[:-1])
        try:
            int(lastPrefix)
            isAlreadyIndexed = True
        except ValueError:
            pass
        if isAlreadyIndexed:
            prefix = allButLastPrefix
    pattern = re.compile(re.escape(prefix) + '(?:\s*_(\d+)\s*)?$')
    maxID = None
    for item in items:
        if isinstance(item, str):
            name = item
        else:
            name = item.getName()
        if not name:
            continue
        m = pattern.match(name)
        if m:
            maxID = max(int(m.group(1) or 0), maxID or 0)
    if maxID is None:
        return prefix
    else:
        return '{}_{}'.format(prefix, maxID + 1)


def whdb_getEntryPoints(path, mtyp):
    """
    Get the entry points for the given type from the model.

    :param path: Model path
    :type path: str
    :param mtyp: Mime type
    :type mtyp: str
    :return: Entry points
    :rtype: set[str]
    """
    mime = SCMN_MIMES.get(mtyp)

    if mime and path:
        for reader in mime.readers:
            if reader.readsPayloadType(WCMN_PAYLOAD_ENTRY_POINTS):
                handler = _WhdbEntryPointPayloadHandler()
                tpsr = reader.readFile(WcmnFile(path), handler, None)
                if tpsr & WCMN_PAYLOAD_ENTRY_POINTS:
                    return handler.entryPoints or {}
    return {}


def whdb_getPortNamesForModel(model):
    """
    Get the port names for the given type from the model.

    :param model: Model to get port names from
    :type model: WhdbModelP
    :return: Port names and displayed names
    :rtype: dict[int, list[(str, str)]]
    """
    path = model.getPath()
    epnt = model.getEntryPoint()
    mime = model.getMedia()

    return whdb_getPortNamesForEntryPoint(path, epnt, mime)


def whdb_getPortNamesForEntryPoint(path, epnt, mime=None):
    """
    Get the port names for the given type from the model.

    :param path: Model file path
    :type path: str
    :param epnt: Model entry point name
    :type epnt: str
    :param mime: Model MIME-type (auto-decected if not provided)
    :type mime: ScmnMimeTypeP | str
    :return: Port names and displayed names
    :rtype: dict[int, list[(str, str)]]
    """
    if not path:
        raise ValueError('No model path specified')

    path = scmn_expshenv(path)

    if isinstance(mime, str):
        mime = SCMN_MIMES.get(mime)
    if not mime:
        pext = scmn_path_extension(path)
        mims = SCMN_MIME_EXTS.get(pext)
        mime = mims[0] if mims else []
        if not mime:
            raise ValueError('Cannot determine MIME-type for ' + path)

    if mime and path and epnt:
        for reader in mime.readers:
            if reader.readsPayloadType(WCMN_PAYLOAD_MODEL_PORTS):
                handler = _WhdbPortNamePayloadHandler(epnt)
                tpsr = reader.readFile(WcmnFile(path), handler, None)
                if tpsr & WCMN_PAYLOAD_MODEL_PORTS:
                    if handler.names is not None:
                        return handler.names
                    else:
                        raise Exception('Could not read ports for ' + epnt)
    else:
        raise ValueError(
            'Cannot extract port-names, missing ' +
            ', '.join(n for e, n
                      in [(mime, 'MIME-type'), (path, 'path'),
                          (epnt, 'entry-point')] if not e))
    raise Exception('No reader found for ' + mime.mimeType)


class _WhdbPortNamePayloadHandler(WcmnReadPayloadHandler):
    def __init__(self, epnt):
        """
        :param epnt: Entry point (function/scheduling group) to retrieve port
                     names for
        :type epnt: str
        """
        self.__epnt = epnt

        self.names = None

    def handlePayload(self, payloadType, mimeType, payloadF):
        self.names = payloadF(self.__epnt)

    @property
    def handledPayloadTypes(self):
        return WCMN_PAYLOAD_MODEL_PORTS


class _WhdbEntryPointPayloadHandler(WcmnReadPayloadHandler):
    def __init__(self):
        self.entryPoints = None

    def handlePayload(self, payloadType, mimeType, payload):
        self.entryPoints = payload

    @property
    def handledPayloadTypes(self):
        return WCMN_PAYLOAD_ENTRY_POINTS


def whdb_printBlock(block, depth=0):
    """
    Recursively print a model block.

    :param block: Block to print
    :type block: WhdbBlockP
    :param depth: Current depth
    :type depth: int
    """
    spc = ' ' * (2 * depth + 2)
    for k, m in block.rawElement.getMetas():
        print(spc + 'B.{k}: {m}'.format(k=k, m=m))
    ilst = list(block.getBlockInstances())
    if ilst:
        print(spc + 'BLOCK INSTANCES:')
        for inst in ilst:
            print(spc + '  @{id} {n} ({cs})'.format(
                id=inst.getUUID(), n=inst.getName(),
                cs=', '.join(['{n}.{p}@{id}'.format(n=c.getName(),
                                                    p=c.getPortReference(),
                                                    id=c.getUUID())
                              for c in inst.getConnectionPoints()])))
            for k, m in inst.rawElement.getMetas():
                print(spc + '    I.{k}: {m}'.format(k=k, m=m))
            whdb_printBlock(inst.getBlock(), depth + 1)
    ilst = list(block.getModelInstances())
    if ilst:
        print(spc + 'MODEL INSTANCES:')
        for inst in ilst:
            print(spc + '  @{id} {n} {m} ({cs})'.format(
                id=inst.getUUID(), n=inst.getName(), m=inst.getMedia(),
                cs=', '.join(['{n}.{p}@{id}'.format(n=c.getName(),
                                                    p=c.getPortReference(),
                                                    id=c.getUUID())
                              for c in inst.getConnectionPoints()])))
            for k, m in inst.rawElement.getMetas():
                print(spc + '    I.{k}: {m}'.format(k=k, m=m))
            for k, m in inst.getModel().rawElement.getMetas():
                print(spc + '    M.{k}: {m}'.format(k=k, m=m))
    flst = list(block.getOperatorInstances())
    if flst:
        print(spc + 'OPERATOR INSTANCES:')
        for func in flst:
            print(spc + '  @{id} {n} {o} ({cs})'.format(
                id=func.getUUID(), n=func.getName(),
                o=func.getOperator().getCode(),
                cs=', '.join([c.getName() + '@' + str(c.getUUID())
                              for c in func.getConnectionPoints()])))
            for k, m in func.rawElement.getMetas():
                print(spc + '    {k}: {m}'.format(k=k, m=m))
    connections = list(block.getConnections())
    if connections:
        print(spc + 'CONNECTIONS:')
        for conn in connections:
            print(spc + '  @{id} {n} {son}@{soi}:{scn}@{sci} - '
                        '{eon}@{eoi}:{ecn}@{eci}'.format(
                            id=conn.getUUID(), n=conn.getName(),
                            son=conn.getStartObject().getName(),
                            soi=conn.getStartObject().getUUID(),
                            scn=conn.getStartConnectionPoint().getName(),
                            sci=conn.getStartConnectionPoint().getUUID(),
                            eon=conn.getEndObject().getName(),
                            eoi=conn.getEndObject().getUUID(),
                            ecn=conn.getEndConnectionPoint().getName(),
                            eci=conn.getEndConnectionPoint().getUUID()))
            for k, m in conn.rawElement.getMetas():
                print(spc + '    {k}: {m}'.format(k=k, m=m))
