# ----------------------------------------------------------------------------
# - whdb_verifier                                                            -
# - wild cruncher:hierarchy verifier                                core /db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_verifier
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher hierarchy verifier

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.common.wcmn_verifier_module import WcmnVerifierModuleP, \
    WcmnHierarchyVerificationError, WcmnVerificationResult
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP, \
    WhdbOperatorInstanceP
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_IN, \
    WDBS_XPNT_BI, WDBS_XPNT_OUT
from .whdb_connection_utils import whdb_getConnectionDescription
from wcr_frontend import isNameValid

# noinspection PyAbstractClass
class WhdbVerifierModuleP(WcmnVerifierModuleP):
    @abstractmethod
    def verifyBlock(self, block, topLevel=False):
        """
        Verify a block.

        :param topLevel: Whether the block is a top-level block
        :type topLevel: bool
        :param block: Block to verify
        :type block: WhdbBlockP
        :return: List of verification errors, if any
        :rtype: list[WcmnHierarchyVerificationError]
        """
        raise NotImplementedError()

    @abstractmethod
    def verifyInstance(self, instance):
        """
        Verify an instance.

        :param instance: Instance to verify
        :type instance: WhdbInstanceP
        :return: List of verification errors, if any
        :rtype: list[WcmnHierarchyVerificationError]
        """
        raise NotImplementedError()

    @abstractmethod
    def verifyConnection(self, cnctn):
        """
        Verify a connection.

        :param cnctn: Wire instance to verify
        :type cnctn: WhdbConnectionP
        :return: List of verification errors, if any
        :rtype: list[WcmnHierarchyVerificationError]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbUniqueNameVerifierModule(WhdbVerifierModuleP):
    """
    Verify that allinstance names are unique within a block.
    """

    ERROR_ID = 'hierarchy.default.duplicate-names'

    ERROR_MESSAGE = 'Duplicate entity name \'{name}\''

    DESCRIPTION = 'Every instance name in a block must be unique.'

    def verifyBlock(self, block, topLevel=False):
        names = {}
        errors = []

        for itm in block.getInstances():
            names.setdefault(itm.getName(), []).append(itm)

        for name, entities in names.items():
            if len(entities) > 1:
                errors.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID, self.ERROR_MESSAGE.format(name=name),
                    entities, self))
        return errors

    def verifyInstance(self, instance):
        pass

    def verifyConnection(self, wire):
        pass

    def description(self):
        return WhdbUniqueNameVerifierModule.DESCRIPTION


    # noinspection PyAbstractClass
class WhdbValidNameVerifierModule(WhdbVerifierModuleP):
    """
    Verify that allinstance names are respect name convention.
    """

    ERROR_ID = 'hierarchy.default.valid-names'

    ERROR_MESSAGE = 'Entity name \'{name}\' have at least one ' \
                    'non-alphanumeric or "_" characters'

    DESCRIPTION = 'Every name can only contain alphanumeric '\
                  'or "_" characters'

    def verifyBlock(self, block, topLevel=False):
        names = {}
        errors = []

        for itm in block.getInstances():
            if not isNameValid(itm.getName()):
                errors.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID,
                    self.ERROR_MESSAGE.format(name=itm.getName()),
                    block.getInstances(), self))
            names.setdefault(itm.getName(), []).append(itm)
        return errors

    def verifyInstance(self, instance):
        names = {}
        errors = []

        for card in getattr(instance, 'getPCICards()', []):
            if not isNameValid(card):
                errors.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID,
                    self.ERROR_MESSAGE.format(name=card),
                    instance, self))
            names.setdefault(card, []).append(card)
        return errors

    def verifyConnection(self, wire):
        pass

    def description(self):
        return WhdbValidNameVerifierModule.DESCRIPTION


# noinspection PyAbstractClass
class WhdbHasNameVerifierModule(WhdbVerifierModuleP):
    """
    Verify that every instance has a non-empty name.
    """

    ERROR_ID = 'hierarchy.default.missing-name'

    ERROR_MESSAGE = 'Entity without name'

    DESCRIPTION = 'Every instance must have a name'

    def _verify(self, entity):
        if not entity.getName():
            return [WcmnHierarchyVerificationError(
                self.ERROR_ID, self.ERROR_MESSAGE, [entity], self)]
        else:
            return []

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyInstance(self, instance):
        return self._verify(instance)

    def verifyConnection(self, wire):
        pass

    def description(self):
        return WhdbHasNameVerifierModule.DESCRIPTION


class WhdbTopLevelHasNameVerifierModule(WhdbVerifierModuleP):
    """
    Verify that the top-level block has a non-empty name.
    """

    ERROR_ID = 'hierarchy.default.missing-name-tophm'

    ERROR_MESSAGE = 'Hierarchy top-level without name'

    DESCRIPTION = 'The top-level block in the hierarchy must have a name'

    def verifyBlock(self, block, topLevel=False):
        if topLevel and not block.getName():
            return [WcmnHierarchyVerificationError(
                self.ERROR_ID, self.ERROR_MESSAGE, [block], self)]

        return []

    def verifyInstance(self, instance):
        pass

    def verifyConnection(self, wire):
        pass

    def description(self):
        return WhdbHasNameVerifierModule.DESCRIPTION


class WhdbInputConnectorCountVerifierModule(WhdbVerifierModuleP):
    """
    Verify that operator instances with variable connection point counts have
    the correct number of connection points.
    """

    ERROR_ID = 'hierarchy.default.connector-count'

    ERROR_MESSAGE = 'Instance {f} has an incorrect number of {t} connection ' \
                    'points ({c}), should be between {min} and {max}'

    DESCRIPTION = 'Operator-instances that allow for variable numbers of ' \
                  'connection points must have a valid number of connection ' \
                  'points as specified by their connection point count range.'

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyInstance(self, instance):
        if not isinstance(instance, WhdbOperatorInstanceP):
            return None

        errors = []

        xdsc = instance.getOperator().getXPointDescriptor()

        vld, errs = xdsc.areValidXPoints(instance.getConnectionPoints())
        if not vld:
            errors.append(WcmnHierarchyVerificationError(
                self.ERROR_ID, '; '.join(errs), [instance], self))

        return errors

    def verifyConnection(self, wire):
        pass

    def description(self):
        return WhdbInputConnectorCountVerifierModule.DESCRIPTION


class WhdbConnectionPointsVerifierModule(WhdbVerifierModuleP):
    """
    Verify that connections are attached to the correct connection points.
    """

    ERROR_ID = 'hierarchy.default.connection-points'

    CONN_MISSING_MSG = '{t}-connection-point/object is missing for ' \
                       'connection {w}'

    INV_CONN_MSG = '{t}-connection point for connection {w} does not belong ' \
                   'to {t}-object'

    CONN_TYPE_MSG = '{t}-connection point for connection {w} has the wrong ' \
                    'type'

    DESCRIPTION = 'A connection has to be attached to a valid ' \
                  'object/connection point pair on either end. The ' \
                  'start-connection-point has to be an output or ' \
                  'bi-directional the end-connector an input or bi-directional'

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyInstance(self, instance):
        pass

    def verifyConnection(self, wire):
        msgs = []
        wireName = whdb_getConnectionDescription(wire)

        for o, c, t, n in [(wire.getStartObject(),
                            wire.getStartConnectionPoint(),
                            WDBS_XPNT_OUT, 'Start'),
                           (wire.getEndObject(),
                            wire.getEndConnectionPoint(),
                            WDBS_XPNT_IN, 'End')]:
            if o is None or c is None:
                msgs.append(self.CONN_MISSING_MSG.format(t=n, w=wireName))
            elif c not in o.getConnectionPoints():
                msgs.append(self.INV_CONN_MSG.format(t=n, w=wireName))
            elif c.getDirection() != t and c.getDirection() != WDBS_XPNT_BI:
                msgs.append(self.CONN_TYPE_MSG.format(t=n, w=wireName))

        return [
            WcmnHierarchyVerificationError(self.ERROR_ID, m, [wire], self)
            for m in msgs]

    def description(self):
        return WhdbConnectionPointsVerifierModule.DESCRIPTION


#: List of all hierarchy verification modules to be enabled by default
WHDB_VERIFIER_MODULES = [WhdbUniqueNameVerifierModule(),
                         WhdbValidNameVerifierModule(),
                         WhdbHasNameVerifierModule(),
                         WhdbTopLevelHasNameVerifierModule(),
                         WhdbInputConnectorCountVerifierModule(),
                         WhdbConnectionPointsVerifierModule()]


class WhdbHierarchyVerifier(object):
    """
    Hierarchy verifier.
    """

    def __init__(self, modules=None):
        """
        :param modules: Additional verification modules to use.
        :type modules: Iterable[WhdbVerifierModuleP] | None
        """
        if modules:
            mods = list(WHDB_VERIFIER_MODULES)
            mods.extend(modules)
        else:
            mods = WHDB_VERIFIER_MODULES
        self.__modules = mods

    def verifyBlock(self, block, recursive=True, topLevel=True):
        """
        Verify a block.

        :param block: Block to verify
        :type block: WhdbBlockP
        :param recursive: Whether to recursively verify the block's children
        :type recursive: bool
        :param topLevel: Whether the block is a top-level block
        :type topLevel: bool
        :return: Verification result
        :rtype: WcmnVerificationResult
        """
        result = WcmnVerificationResult(block)

        self._verifyBlock(block, result, recursive, topLevel)

        return result

    def _verifyBlock(self, block, result, recursive, topLevel=False):
        """
        Verify a block, and add all errors found to the given result object.

        :param block: Block to verify
        :type block: WhdbBlockP
        :param result: Result object to add found errors to
        :type result: WcmnVerificationResult
        :param recursive: Whether to recursively verify the block's children
        :type recursive: bool
        :param topLevel: Whether the block is a top-level block
        :type topLevel: bool
        """
        for module in self.__modules:
            # noinspection PyProtectedMember
            result._addErrors(module.verifyBlock(block, topLevel))

        for inst in block.getInstances():
            self._verifyInstance(inst, result, recursive)
        for wire in block.getConnections():
            self._verifyConnection(wire, result)

    # noinspection PyProtectedMember
    def _verifyInstance(self, instance, result, recursive):
        """
        Verify an instance, and add all errors found to the given result
        object.

        If `recursive` is set, and the instance is a block instance,
        the block will be verified recursively as well.

        :param instance: Instance to verify
        :type instance: WhdbInstanceP
        :param result: Result object to add found errors to
        :type result: WcmnVerificationResult
        :param recursive: Whether to recursively verify the block's children
        :type recursive: bool
        """
        for module in self.__modules:
            result._addErrors(module.verifyInstance(instance))

        if recursive and isinstance(instance, WhdbBlockInstanceP):
            result._pushBlockInstance(instance)
            self._verifyBlock(instance.getBlock(), result, recursive)
            result._popBlockInstance()

    def _verifyConnection(self, cnctn, result):
        """
        Verify a connection, and add all errors found to the given result
        object.

        :param cnctn: Wire instance to verify
        :type cnctn: WhdbConnectionP
        :param result: Result object to add found errors to
        :type result: WcmnVerificationResult
        """
        for module in self.__modules:
            # noinspection PyProtectedMember
            result._addErrors(module.verifyConnection(cnctn))
