# ----------------------------------------------------------------------------
# - whdb-wrapper-registry                                                    -
# - wild cruncher: wrapper registry                                core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_wrapper_registry
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general hierarchy wrapper registry

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod

from wcr_core.db.whdb_parameter import WhdbParameter
from wcr_core.dbstore_api.wdbs_block_api import WdbsBlockP
from wcr_core.dbstore_api.wdbs_connection_api import WdbsConnectionP
from wcr_core.dbstore_api.wdbs_instance_api import WdbsOperatorInstanceP, \
    WdbsBlockInstanceP, WdbsModelInstanceP
from wcr_core.dbstore_api.wdbs_buttons_api import WdbsButtonP
from wcr_core.dbstore_api.wdbs_model_api import WdbsModelP
from wcr_core.dbstore_api.wdbs_parameter_api import WdbsParameterP
from wcr_core.dbstore_api.wdbs_shape_api import WdbsShapeP
from wcr_core.dbstore_api.wdbs_connection_point_api import WdbsConnectionPointP
from .whdb_block import WhdbBlock
from .whdb_instance import WhdbModelInstance
from .whdb_connection import WhdbConnection
from .whdb_connection_path_manhattan import \
    WhdbManhattanConnectionPath
from .whdb_instance import WhdbOperatorInstance, WhdbBlockInstance
from .whdb_button import WhdbButton
from .whdb_model import WhdbModel
from .whdb_operator import WhdbOperator
from .whdb_shape import whdb_get_shape_for_name
from .whdb_connection_point import WhdbConnectionPoint
from .whdb_connection_point_manager import \
    WhdbFixedConnectionPointManager, WhdbXPointDesc, WHDB_XPOINT_NAME, \
    WHDB_XPOINT_DIRECTION


class WhdbWrapperRegistryP(metaclass=ABCMeta):
    """
    Registry connection raw elements (WdbsEntityP) to their HDS wrappers
    (WhdbBaseP).
    """

    @abstractmethod
    def wrap(self, obj):
        """
        Get a HDS wrapper for the given raw element.

        :param obj: Raw element to wrap
        :type obj: WdbsEntityP
        :return: Wrapped element
        :rtype: WhdbBaseP
        """
        raise NotImplementedError()

    @abstractmethod
    def unwrap(self, wrapper):
        """
        Extract the wrapped raw object from the given wrapper.

        :param wrapper: Wrapper to extract raw object from
        :type wrapper: WhdbBaseP
        :return: Raw element
        :rtype: WdbsEntityP
        """
        raise NotImplementedError()


class WhdbWrapperRegistry(WhdbWrapperRegistryP):
    """
    Managers HDS-HSB wrappers.
    """

    def __init__(self):
        self.__registry = {}

    def wrap(self, obj, *args):
        """
        Get a HDS-wrapper for the given HSB element.

        :param obj: HSB element to return wrapper for
        :type obj: WdbsEntityP
        """
        if obj is None:
            return None
        oid = obj.getUUID()
        wrapper = self.__registry.get(oid)
        if wrapper is None or getattr(wrapper, '_wrapped_obj') is not obj:
            wrapper = self.createWrapper(obj, *args)
            if wrapper is None:
                raise ValueError('Unknown database entity: {}'.format(obj))
            self.__registry[oid] = wrapper
            setattr(wrapper, '_wrapped_obj', obj)
        return wrapper

    def unwrap(self, wrapper):
        """
        Get the wrapped HSB element from a HDS-wrapper.

        :param wrapper: HDS wrapper
        :type wrapper: WhdbBaseP
        """
        return getattr(wrapper, '_wrapped_obj')

    def createWrapper(self, obj, *args):
        """
        Get a HDS-wrapper for the given HSB element.

        :param obj: HSB element to return wrapper for
        :type obj: WdbsEntityP
        """
        if isinstance(obj, WdbsConnectionP):
            return WhdbConnection(obj, self, WhdbManhattanConnectionPath)
        elif isinstance(obj, WdbsConnectionPointP):
            return WhdbConnectionPoint(obj, self)
        elif isinstance(obj, WdbsOperatorInstanceP):
            cdsc = []
            # noinspection PyTypeChecker
            for xpnt in obj.getConnectionPoints():
                cdsc.append(WhdbXPointDesc(
                    xpnt.getEntityType(),
                    **{WHDB_XPOINT_NAME: xpnt.getUUID(),
                       WHDB_XPOINT_DIRECTION: xpnt.getDirection()}))

            xmng = WhdbFixedConnectionPointManager(cdsc)
            oper = WhdbOperator(obj.getName(), xmng)
            return WhdbOperatorInstance(oper, obj, self)
        elif isinstance(obj, WdbsBlockInstanceP):
            return WhdbBlockInstance(obj, self)
        elif isinstance(obj, WdbsModelInstanceP):
            return WhdbModelInstance(obj, self)
        elif isinstance(obj, WdbsBlockP):
            return WhdbBlock(obj, self)
        elif isinstance(obj, WdbsModelP):
            return WhdbModel(obj, self)
        elif isinstance(obj, WdbsButtonP):
            return WhdbButton(obj, self)
        elif isinstance(obj, WdbsShapeP):
            shape = whdb_get_shape_for_name(obj.getEntityType())
            if shape is not None:
                return shape(obj, self)
        elif isinstance(obj, WdbsParameterP):
            return WhdbParameter(obj, *args)

        raise ValueError('Unknown database entity: {}'.format(obj))
