# ----------------------------------------------------------------------------
# - whdb-block-api                                                           -
# - wild cruncher: block api                                   core / db-api -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_block_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Block API

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod
from wcr_core.db_api.whdb_protocols_api import WhdbBaseP, WhdbTypedEntityP, \
    WhdbNamedEntityP


WHDB_DEFAULT_BLOCK_CODE = 'wcr.core.block-default'

WHDB_TOPLEVEL_BLOCK_CODE = 'wcr.core.block-top'


class WhdbEntityContainerP(metaclass=ABCMeta):
    """
    Abstract base class for blocks.
    """

    @abstractmethod
    def getInstances(self):
        """
        Get the instances.

        :return: The instances
        :rtype: Iterable[WhdbInstanceP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getBlockInstances(self):
        """
        Get the block instances.

        :return: The block instances
        :rtype: Iterable[WhdbBlockInstanceP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getModelInstances(self):
        """
        Get the model instances.

        :return: The model instances
        :rtype: Iterable[WhdbModelInstanceP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getOperatorInstances(self):
        """
        Get the operator instances.

        :return: The operator instances
        :rtype: Iterable[WhdbOperatorInstanceP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getPorts(self):
        """
        Get the port instances.

        :return: The port instances
        :rtype: Iterable[WhdbOperatorInstanceP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnections(self):
        """
        Get the connections between elements in the block.

        :return: List of connections
        :rtype: Iterable[WhdbConnectionP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getShapes(self):
        """
        Get the block's shapes.

        :return: The block's shapes
        :rtype: Iterable[WhdbShapeP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getButtons(self):
        """
        Get the block's list of buttons.

        :return: The block's buttons
        :rtype: Iterable[WhdbButtonP]
        """
        raise NotImplementedError()


class WhdbEntityTargetP(metaclass=ABCMeta):
    @abstractmethod
    def createBlockInstance(self, name=None, info=None,
                            ipos=None, zpos=None, width=None, height=None,
                            rotn=None, vflp=None, hflp=None, btyp=None,
                            **kwargs):
        """
        Add a new instance for the given mime type.

        :param name: Name of the instance
        :type name: str
        :param info: Info string for the instance
        :type info: str
        :param ipos: Position
        :type ipos: (float, float)
        :param zpos: Z-value
        :type zpos: int
        :param width: Width
        :type width: float
        :param height: Height
        :type height: float
        :param rotn: Rotation
        :type rotn: float
        :param vflp: Vertically flipped
        :type vflp: bool
        :param hflp: Horizontally flipped
        :type hflp: bool
        :param btyp: Block type
        :type btyp: str | None
        :return: The created model instance
        :rtype: WhdbBlockInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def createModelInstance(self, mime, name=None, info=None,
                            ipos=None, zpos=None, width=None, height=None,
                            rotn=None, vflp=None, hflp=None, mtyp=None,
                            **kwargs):
        """
        Add a new instance for the given mime type.

        :param mime: Model MIME type
        :type mime: str
        :param name: Name of the instance
        :type name: str
        :param info: Info string for the instance
        :type info: str
        :param ipos: Position
        :type ipos: (float, float)
        :param zpos: Z-value
        :type zpos: int
        :param width: Width
        :type width: float
        :param height: Height
        :type height: float
        :param rotn: Rotation
        :type rotn: float
        :param vflp: Vertically flipped
        :type vflp: bool
        :param hflp: Horizontally flipped
        :type hflp: bool
        :param mtyp: Model type
        :type mtyp: str | None
        :return: The created model instance
        :rtype: WhdbModelInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def createOperatorInstance(self, oper, name=None, info=None,
                               ipos=None, zpos=None, width=None, height=None,
                               rotn=None, vflp=None, hflp=None, prms=None,
                               **kwargs):
        """
        Add a new instance for the given operator.

        :param oper: the operator
        :type oper: WhdbOperatorP
        :param name: Name of the instance
        :type name: str
        :param info: Info string for the instance
        :type info: str
        :param ipos: Position
        :type ipos: (float, float)
        :param zpos: Z-value
        :type zpos: int
        :param width: Width
        :type width: float
        :param height: Height
        :type height: float
        :param rotn: Rotation
        :type rotn: float
        :param vflp: Vertically flipped
        :type vflp: bool
        :param hflp: Horizontally flipped
        :type hflp: bool
        :param prms: Parameter values
        :type prms: Iterable[(str, object)]
        :return: The newly created function
        :rtype: WhdbOperatorInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def createConnection(self, so, sc, eo, ec, name=None, info=None, path=None,
                         kppt=None, zval=None, **kwargs):
        """
        Create a new wire by objects and connection points

        :param so: the starting object
        :type so: WhdbInstanceP
        :param sc: the starting connection point
        :type sc: WhdbConnectionPointP
        :param eo: the ending object
        :type eo: WhdbInstanceP
        :param ec: the ending connection point
        :type ec: WhdbConnectionPointP
        :param name: the wire name
        :type name: str
        :param info: the wire information
        :type info: str
        :param path: the path for the wire
        :type path: Iterable[(float, float)]
        :param kppt: whether to attempt to maintain the path
        :type kppt: bool
        :param zval: Z-value
        :type zval: int
        :param aliases: Alias names
        :type aliases: Iterable[str]
        :return: The newly created wire
        :rtype: WhdbConnectionP
        """
        raise NotImplementedError()

    @abstractmethod
    def createShape(self, styp, **kwargs):
        """
        Create a new shape of the given type.

        :param styp: Shape type (WhdbLine, ...)
        :type styp: class
        :param kwargs: Constructor arguments for the new shape
        :return: The newly created shape
        :rtype: WhdbShapeP
        """
        raise NotImplementedError()

    @abstractmethod
    def createButton(self, title=None, command=None, parameters=None,
                     bpos=None, zpos=None, bbox=None, **kwargs):
        """
        Creae a new button.

        :param title: Button title
        :type title: str
        :param command: Command associated with button
        :type command: str
        :param parameters: Command parameter values
        :type parameters: dict[str, object]
        :param bpos: Position
        :type bpos: (float, float)
        :param zpos: Z value
        :type zpos: fleat
        :param bbox: Bounding box
        :type bbox: (float, float, float, float)
        :return: The newly created button
        :rtype: WhdbButtonP
        """
        raise NotImplementedError()


# TODO: Split into adding funcs and rest
class WhdbEntityContainerMP(metaclass=ABCMeta):
    """
    Abstract base class for blocks.
    """

    @abstractmethod
    def remove(self, entity):
        """
        Remove an object in this model block. The object can be
        an instance, a function or a wire.

        :param entity: the element to remove
        :type entity: WhdbBaseP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeInstance(self, inst):
        """
        Remove an instance from the block. All wires connected to that
        instance are removed as well.

        :param inst: the instance to remove
        :type inst: WhdbInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeConnection(self, cntn):
        """
        Remove a wire from the block.

        :param cntn: the wire to remove
        :type cntn: WhdbConnectionP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeShape(self, shape):
        """
        Add a shape from the model block, if the shape is in the list
        of the block's shapes.

        :param shape: Shape to be removed
        :type shape: WhdbShapeP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeButton(self, element):
        """
        Remove a button from the block.

        :param element: Element to remove
        :type element: WhdbButtonP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbBlockP(WhdbEntityContainerMP, WhdbEntityTargetP,
                 WhdbEntityContainerP, WhdbTypedEntityP,
                 WhdbNamedEntityP, WhdbBaseP):
    @property
    @abstractmethod
    def wrapperRegistry(self):
        raise NotImplementedError()

    @abstractmethod
    def getConnectionsForObject(self, ix, cx=None):
        raise NotImplementedError()
