# ----------------------------------------------------------------------------
# - whdb-connection-point-api                                                -
# - wild cruncher: connection point api                        core / db-api -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_connection_point_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Connection point API

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod
from wcr_core.db_api.whdb_protocols_api import WhdbTypedEntityP, WhdbBaseP


# noinspection PyAbstractClass,PyClassHasNoInit
class WhdbConnectionPointP(WhdbTypedEntityP, WhdbBaseP):
    """
    The WhdbConnectionPointP class is the silkan hierarchy connection point
    object (i.e. "connection point").
    A connection point is defined by name with an additional port id. A
    direction attribute is used to qualify the connection point.
    """

    #: Connection point gravity values
    North, East, South, West = range(1, 5)

    @abstractmethod
    def getName(self):
        """
        Get the connector's name.

        :return: The connector's name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getDirection(self):
        """
        Get the connection point type (input or output).

        :return: The connection point's type
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getPortReference(self):
        """
        Get the connector's port reference, if any.

        :return: The connector's port reference, or `None`
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setPortReference(self, port):
        """
        Set the connector's port reference.

        :param port: New port reference for the connector
        :type port: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getGravity(self):
        """
        Get the connection point gravity.

        :return: The gravity of the connection point
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def setGravity(self, gvty):
        """
        Get the connection point gravity.

        :param gvty: New gravity value
        :type gvty: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getPosition(self):
        """
        Get the connection point's position.

        :return: The position of the connection point
        :rtype: (float, float)
        """
        raise NotImplementedError()

    # noinspection PyShadowingNames
    @abstractmethod
    def setPosition(self, x, y):
        """
        Set the connection point's position.

        :param x: The new relative X-position of the connection point
        :type x: float
        :param y: The new relative Y-position of the connection point
        :type y: float
        """
        raise NotImplementedError()

    @abstractmethod
    def getDescription(self, **kwargs):
        """
        Get the formatted name for the given connection point.

        :return: Description of the connection point
        :rtype: str
        """
        raise NotImplementedError()

    @staticmethod
    def getConnectionPointDefaultGravity(cdir):
        """
        Get the default gravity for a connection point.

        :param cdir: Connection point type (input/output)
        :type cdir: int
        :return: Default gravity for the given connection point
        :rtype: int
        """
        raise NotImplementedError()
