# ----------------------------------------------------------------------------
# - whdb-instance-api                                                        -
# - wild cruncher: instance api                                core / db-api -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_instance_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Instance API

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db_api.whdb_parameter_api import WhdbHasParametersP

from wcr_core.db_api.whdb_protocols_api import *


# noinspection PyAbstractClass
class WhdbInstanceP(WhdbNamedEntityP, WhdbHasZValueP, WhdbHasLocationP,
                    WhdbHasDimensionP, WhdbCanFlipRotateP, WhdbHasCustomIconP,
                    WhdbBaseP):
    """
    Protocol for instances.
    """

    @abstractmethod
    def getConnectionPointCount(self, **kwargs):
        """
        Get the number of connection points.

        :param kwargs: Selector arguments
        :type kwargs: dict[str, object]
        :return: Number of connection points for the given direction,
                 or total number if no direction was given
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectionPoints(self, **kwargs):
        """
        Get the instance's connection points.

        :param kwargs: Selector arguments
        :type kwargs: dict[str, object]
        :return: Connection points for the given direction,
                 or all if no direction was given
        :rtype: Iterable[WhdbConnectionPointP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getNewConnectionPointDirection(self, **kwargs):
        """
        Get the default direction for new connection points with the under
        `kwargs` given specifications.

        :return: Default direction
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def canCreateConnectionPoint(self, **kwargs):
        """
        Whether a connection point in the given direction may be added.

        :param kwargs: Additional parameters
        :type kwargs: dict[str, object]
        :return: True iff a connection point in the given direction may be
                 added
        :rtype: Bool
        """
        raise NotImplementedError()

    @abstractmethod
    def createConnectionPoint(self, ctyp=None, **kwargs):
        """
        Create a new connection point.

        :param ctyp: Connection point type
        :type ctyp: str | None
        :param kwargs: Additional parameters
        :return: Newly created connection point
        :rtype: WhdbConnectionPointP
        """
        raise NotImplementedError()

    @abstractmethod
    def canRemoveConnectionPoint(self, xpnt):
        """
        Whether a given connection point may be deleted.

        :param xpnt: Connection point in question
        :type xpnt: WhdbConnectionPointP
        :return: True iff the given connection point may deleted
        :rtype: Bool
        """
        raise NotImplementedError()

    @abstractmethod
    def removeConnectionPoint(self, xpnt):
        """
        Delete the given connection point.

        :param xpnt: Connection point to be deleted
        :type xpnt: WhdbConnectionPointP
        """
        raise NotImplementedError()

    @abstractmethod
    def clearConnectionPoints(self):
        """
        Delete all connection points.
        """
        raise NotImplementedError()

    @abstractmethod
    def getEffectiveConnectionPointPosition(self, xpnt):
        """
        Get the effective position for the given connection point, taking
        rotation and flipping into consideration.

        :param xpnt: Connection point
        :type xpnt: WhdbConnectionPointP
        :return: Effective connection point position
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @abstractmethod
    def getEffectiveConnectiontPointGravity(self, xpnt):
        """
        Get the effective gravity for the given connection point, taking
        rotation and flipping into consideration.

        :param xpnt: Connection point
        :type xpnt: WhdbConnectionPointP
        :return: Effective connection point gravity
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getEffectiveDimension(self):
        """
        Get the effective dimension for the instance, taking
        rotation and flipping into consideration.

        :return: Effective dimension
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectionPointDescription(self, xpnt, **kwargs):
        """
        Get the description for the given connection point.

        :param xpnt: Connectino point
        :type xpnt: WhdbConnectionPointP
        :return: Description for the given connection point
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def resetConnectionPointPositions(self):
        """
        Reset the instance's connection points to their original positions.
        """
        raise NotImplementedError()

    @abstractmethod
    def setConnectionPointOrder(self, orderF):
        """
        Reorder the list of connection points based on the given ordering
        function.

        :param orderF: Ordering function
        :type orderF: (WhdbConnectionPointP) -> int
        """
        raise NotImplementedError()

    @abstractmethod
    def getLabelAlignment(self):
        """
        Get the horizontal alignment for any labels for the entity.

        Return value should be one of :py:data:`WHDB_HALIGN_LEFT`,
        :py:data:`WHDB_HALIGN_CENTER` or :py:data:`WHDB_HALIGN_RIGHT`.

        :return: Horizontal alignmen for labels
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getDefaultDimension(self):
        """
        Get the default dimension for the instance.

        :return: Default dimension
        :rtype: (float, float)
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbBlockInstanceP(WhdbInstanceP):
    @abstractmethod
    def getBlock(self):
        """
        Get the block represented by this instance.

        :return: Represented block
        :rtype: WhdbBlock
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbModelInstanceP(WhdbHasMediaP, WhdbInstanceP):
    @abstractmethod
    def getModel(self):
        """
        Get the model represented by this instance.

        :return: Represented model
        :rtype: WhdbModelP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbOperatorInstanceP(WhdbInstanceP, WhdbHasParametersP):
    @abstractmethod
    def getOperator(self):
        """
        Get the operator.

        :return: Operator
        :rtype: WhdbOperatorP
        """
        raise NotImplementedError()

    @abstractmethod
    def getSymbol(self):
        """
        Get the operator instance's text symbol, if any.

        :return: Text symbol for the operator instance, or `None`
        :rtype: str
        """
        raise NotImplementedError()
