# ----------------------------------------------------------------------------
# - whdb-operator-api                                                        -
# - wild cruncher: operator api                                core / db-api -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_operator_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Operator API

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.db_api.whdb_protocols_api import WhdbHasFlagsP


class WhdbOperatorP(WhdbHasFlagsP):
    """
    Protocol for operators.
    """

    @abstractmethod
    def getTypeName(self):
        """
        Get the type name.

        :return: Type name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getCode(self):
        """
        Get the operator's code.

        :return: Code of the operator
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getName(self):
        """
        Get the operator name.

        :return: Name of the operator
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getInfo(self):
        """
        Get the operator info.

        :return: Info for the operator
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getXPointDescriptor(self):
        """
        Get the connection-point descriptor for the operator.

        :return: Connection point descriptor
        :rtype: WhdbConnectionPointDescriptor
        """
        raise NotImplementedError()

    @abstractmethod
    def getFlags(self):
        """
        Get the operator flags.

        :return: Flags for the operator
        :rtype: sit[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def getMeta(self, key):
        """
        Get the meta-data associated with the operator for the given key.

        :param key: Meta-data key
        :type key: str
        :return: Meta-data for the given key, or `None`
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def getIcon(self, ifmt):
        """
        Get the operator's icon path for the given format.

        :param ifmt: Icon format
        :type ifmt: str
        :return: Icon path for the given format, or `None`
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getShape(self):
        """
        Get the operator's shape icon, if any.

        :return: Shape icon for the operator and its dimensions, or `None`
                 respectively
        :rtype: Iterable[(float, float)], (float, float)
        """
        raise NotImplementedError()

    @abstractmethod
    def getSymbol(self, valueF):
        """
        Get the operator's text symbol, if any

        :param valueF: Parameter value function
        :type valueF: (str, str) -> str
        :return: Text symbol for the operator, or `None`
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getLabelAlignment(self):
        """
        Get the label alignment for the operator.

        :return: Label alignment
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def createInitialConnectionPoints(self):
        """
        Get the operator connection point list.

        :return: Connection point descriptions list
        :rtype: Iterable[WhdbXPointDesc]
        """
        raise NotImplementedError()

    @abstractmethod
    def resetConnectionPointPositions(self, clst):
        """
        Reset the connection points in the given list to their default
        positions.

        :param clst: Connection point list
        :type clst: Iterable[WhdbConnectionPointP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getParameters(self):
        """
        Get an initialized copy of the operators parameters.

        :return: Copy of the operator's parameter list
        :rtype: Iterable[WcmnParameterDescriptionP]
        """
        raise NotImplementedError()

    @abstractmethod
    def hasFlag(self, flag, *_):
        """
        Whether the operator carries the given flag.

        :param flag: Flag to test
        :type flag: str
        :return: True iff the operator carries the given flag
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def canAddConnectionPoint(self, clst, **kwargs):
        """
        Whether a connection point of the given direction can be added to the
        given connection point list.

        :param clst: Connection point list
        :type clst: Iterable[WhdbConnectionPointP]
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: True iff a connection point of the given direction can be
                 added
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def addConnectionPoint(self, connF, clst, **kwargs):
        """
        Add a connection point with the given specifics to the given
        connection point list.

        :param clst: Connection point list
        :type clst: Iterable[WhdbConnectionPointP]
        :param connF: Connection point creation function
        :type connF: (str, int, dict[str, object]) -> WdbsConnectionPointP
        :param name: Connection point name
        :type name: str
        :return: Newly created connection point
        :rtype: WdbsConnectionPointP
        """
        raise NotImplementedError()

    @abstractmethod
    def canRemoveConnectionPoint(self, xpnt, clst):
        """
        Whether the given connection point can be removed from the connection
        point list.

        :param clst: Connection point list
        :type clst: Iterable[WhdbConnectionPointP]
        :param xpnt: Connection point to be removed
        :type xpnt: WhdbConnectionPointP
        :return: True iff a connection point of the given direction can be
                 removed
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def removeConnectionPoint(self, clst, xpnt):
        """
        Remove the given connection point from the connection point list.

        :param clst: Connection point list
        :type clst: Iterable[WhdbConnectionPointP]
        :param xpnt: Connection point to be removed
        :type xpnt: WhdbConnectionPointP
        """
        raise NotImplementedError()
