# ----------------------------------------------------------------------------
# - whdb-protocols-api                                                       -
# - wild cruncher: protocols api                               core / db-api -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_protocols_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Protocols API

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod
from wcr_core.db_api.whdb_observer_api import WhdbObservableP


#: Horizontal alignment
WHDB_HALIGN_LEFT, WHDB_HALIGN_CENTER, WHDB_HALIGN_RIGHT = [-1, 0, 1]

#: Vertical alignment
WHDB_VALIGN_TOP, WHDB_VALIGN_MIDDLE, WHDB_VALIGN_BOTTOM = [-1, 0, 1]


class WhdbHasIdP(metaclass=ABCMeta):
    @abstractmethod
    def getUUID(self):
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbBaseP(WhdbHasIdP, WhdbObservableP):
    """
    Base class for all HDS classes.
    """

    @property
    @abstractmethod
    def rawElement(self):
        """
        Get the raw element represented by this HDS object

        :return: Raw element
        :rtype: WdbsEntityP
        """
        raise NotImplementedError()

    @abstractmethod
    def getTypeName(self):
        """
        Get the name of the object's type.

        :return: Type name
        :rtype: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbObservableMixinP(WhdbObservableP):
    @abstractmethod
    def _notifyObservers(self, event, *args, **kwargs):
        """
        Notify all observers of this entity of a given event.

        :param event: Event code
        :type event: int
        :param args: Events-specific arguments
        :param kwargs: Event-specific keyword-arguments
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbTypedEntityP(WhdbBaseP):
    """
    Protocol for entities with a name/description.
    """

    @abstractmethod
    def getEntityType(self):
        """
        Get the object's type.

        :return: Type
        :rtype: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbNamedEntityP(WhdbBaseP):
    """
    Protocol for entities with a name/description.
    """

    @abstractmethod
    def getName(self):
        """
        Get the object's name.

        :return: Name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getInfo(self):
        """
        Get the object's description.

        :return: Description
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setName(self, name):
        """
        Set the object's name.

        :param name: New name
        :type name: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setInfo(self, info):
        """
        Set the object's description.

        :param info: New description
        :type info: str
        """
        raise NotImplementedError()


class WhdbHasMediaP(metaclass=ABCMeta):
    """
    Protocol for entities that carry a media/mime-type.
    """

    @abstractmethod
    def getMedia(self):
        """
        Get the entity's media/mime-type

        :return: Mime type
        :rtype: str
        """
        raise NotImplementedError()


class WhdbHasZValueP(metaclass=ABCMeta):
    """
    Protocol for entities that have a location.
    """

    @abstractmethod
    def getZ(self):
        """
        Get the entity's Z-value (depth in stack, items with greater Z-value
        are positioned on top of items with lower Z-value).

        :return: The Z-value of the entity
        :rtype: float
        """
        raise NotImplementedError()

    @abstractmethod
    def setZ(self, zval):
        """
        Set the entity's Z-value.

        :param zval: The new Z-value of the entity
        :type zval: float
        """
        raise NotImplementedError()


class WhdbHasLocationP(metaclass=ABCMeta):
    """
    Protocol for entities that have a location.
    """

    @abstractmethod
    def getPosition(self):
        """
        Get the entity's location.

        :return: The coordinates of the entity
        :rtype: (float, float)
        """
        raise NotImplementedError()

    # noinspection PyShadowingNames
    @abstractmethod
    def setPosition(self, x, y):
        """
        Set the entity's location.

        :param x: The new x-coordinate
        :type x: float
        :param y: The new y-coordinate
        :type y: float
        """
        raise NotImplementedError()


class WhdbHasDimensionP(metaclass=ABCMeta):
    """
    Protocol for entities that have a dimension.
    """

    @abstractmethod
    def rawElement(self):
        """
        :rtype: WdbsEntityP
        """
        raise NotImplementedError()

    @abstractmethod
    def getDimension(self):
        """
        Get the entity's dimensions.

        :return: The dimensions of the entity
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @abstractmethod
    def setDimension(self, w, h):
        """
        Set the entity's dimensions.

        :param w: The new width
        :type w: float
        :param w: The new height
        :type w: float
        """
        raise NotImplementedError()

    @abstractmethod
    def unsetDimension(self):
        """
        Unset the entity's dimensions.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def boundingBox(self):
        """
        Get the approximate bounding box for the entity.

        :return: The entities bounding box
        :rtype: ScmnRectangle
        """
        raise NotImplementedError()


class WhdbCanFlipRotateP(metaclass=ABCMeta):
    """
    Protocol for entities that can be flipped or rotated.
    """

    @abstractmethod
    def isFlippedVertically(self):
        """
        Whether the entity is flipped vertically.

        :return: True iff the entity is flipped vertically
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def setFlippedVertically(self, flip):
        """
        Set the entity's vertical flip setting.

        :param flip: Whether the entity is flipped vertically
        :type flip: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def isFlippedHorizontally(self):
        """
        Whether the entity is flipped horizontally.

        :return: True iff the entity is flipped horizontally
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def setFlippedHorizontally(self, flip):
        """
        Set the entity's horizontal flip setting.

        :param flip: Whether the entity is flipped horizontally
        :type flip: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def getRotation(self):
        """
        Get the entity's rotation in radians.

        :return: The rotation of the entity in radians
        :rtype: float
        """
        raise NotImplementedError()

    @abstractmethod
    def setRotation(self, rotation):
        """
        Set the entity's rotation.

        :param rotation: The new rotation of the entity in radians
        :type rotation: float
        """
        raise NotImplementedError()


class WhdbHasCustomIconP(metaclass=ABCMeta):
    """
    Protocol for entities that can be assigned a custom icon.
    """

    @abstractmethod
    def getCustomIcon(self):
        """
        Retrieve the custom icon resource key.

        :return: Custom icon resource key
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setCustomIcon(self, rkey):
        """
        Set the custom icon resource key.

        :param rkey: Custom icon resource key
        :type rkey: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getCustomText(self):
        """
        Retrieve the custom text.

        :return: Custom text
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setCustomText(self, text):
        """
        Set the custom text.

        :param text: Custom text
        :type text: str
        """
        raise NotImplementedError()

    @abstractmethod
    def isBoundingBoxHidden(self):
        """
        Whether the bounding box should not be displayed.

        :return: Whether the bounding box is not to be shown
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def setBoundingBoxHidden(self, hidden):
        """
        Set whether the bounding box should not be displayed.

        :param hidden: Whether the bounding box is not to be shown
        :type hidden: bool
        """
        raise NotImplementedError()


class WhdbHasFlagsP(metaclass=ABCMeta):
    """
    Protocol for entities that carry flags.
    """

    @abstractmethod
    def hasFlag(self, flag, recurse=True):
        """
        Whether the entity carries the given flag.

        If the entity has child entities, and the `recurse` option is given,
        then the child entities will be queried for the flag recursively
        as well. If any of the children carries the flag, `True` will be
        returned.

        :param flag: Flag to look for
        :type flag: str
        :param recurse: Whether to query child entities for flag, if any
        :param recurse: bool
        :return: `True` iff entity (or any of its children) carries the given
                 flag
        :rtype: bool
        """
        raise NotImplementedError()
