# ----------------------------------------------------------------------------
# - whdb-shape-api                                                           -
# - wild cruncher: shape api                                   core / db-api -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_shape_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Shape API

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod

from wcr_core.db_api.whdb_protocols_api import WhdbBaseP
from wcr_core.db_api.whdb_protocols_api import WhdbHasZValueP, \
    WHDB_HALIGN_LEFT, WHDB_VALIGN_TOP


def enum(name:str, *enums):
    """
    Create an enum type of the given options.

    :param name: Name of the enum
    :param enums: Enum options as iterable of strings
    :return: Enum type for the given options
    """
    # noinspection PyArgumentList
    return type(name, (),
                dict(map(lambda k, v: (k, v), enums,
                         range(1, len(enums) + 1))))


#: Available line styles
WHDB_LINE_STYLE = enum('WhdbLineStyle', 'SOLID', 'DOTTED', 'DASHED')


class WhdbShapeVisitorP(metaclass=ABCMeta):
    @abstractmethod
    def visitLine(self, line, *args, **kwargs):
        """
        Callback function used when the visited instance is a line.

        :param line: Line to be visited
        :type line: WhdbLineShape
        """
        raise NotImplementedError()

    @abstractmethod
    def visitArrow(self, arrow, *args, **kwargs):
        """
        Callback function used when the visited instance is an arrow.

        :param arrow: Arrow to be visited
        :type arrow: WhdbArrowShape
        """
        raise NotImplementedError()

    @abstractmethod
    def visitRectangle(self, rectangle, *args, **kwargs):
        """
        Callback function used when the visited instance is a rectangle.

        :param rectangle: Rectangle to be visited
        :type rectangle: WhdbRectangleShape
        """
        raise NotImplementedError()

    @abstractmethod
    def visitOval(self, oval, *args, **kwargs):
        """
        Callback function used when the visited instance is an oval.

        :param oval: Oval to be visited
        :type oval: WhdbOvalShape
        """
        raise NotImplementedError()

    @abstractmethod
    def visitText(self, text, *args, **kwargs):
        """
        Callback function used when the visited instance is a text shape.

        :param text: Text shape to be visited
        :type text: WhdbTextShape
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbShapeP(WhdbHasZValueP, WhdbBaseP):
    @property
    @abstractmethod
    def location(self):
        """
        The shape's location property.
        A shape's location usually refers to its center,
        but not necessarily so.

        :return: The shape's location.
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @location.setter
    @abstractmethod
    def location(self, point):
        """
        Set the shape's location.
        A shape's location usually refers to its center,
        but not necessarily so.

        :param point: The shape's location as (x, y)-coordinates
        :type point: (float, float)
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def shapeName(self):
        """
        Get the name of the shape type.

        :return: Name of the shape type
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    @abstractmethod
    def acceptVisitor(self, visitor, *args, **kwargs):
        """
        Accept a visitor.

        :param visitor: Visitor to be accepted
        :type visitor: wcr_core.db_api.whdb_shape_api.WhdbShapeVisitorP
        :return: Value returned by the called visitor function
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def boundingBox(self):
        """
        Get the shape's approximate bounding box.

        :return: The shape's bounding box
        :rtype: ScmnRectangle
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbLineBasedShapeP(WhdbShapeP):
    """
    Base class for line-based shapes.

    A shape can be considered line-based if it consist of one or more lines,
    or if it has an outline.

    A line is characterized by its colour, width an style. If a line has a
    width of 0, it is usually considered invisible; the line's colour and style
    are without effect in this case.

    The line colour is specified as RGB hex string ('0xRRGGBB').

    The line width is specified in pixels.

    The line style is specified as int-value referring to one of the options
    provided by `WHDB_LINE_STYLE`.
    """
    #: Default line colour
    DEFAULT_LINE_COLOUR = '#000000'
    #: Default line width
    DEFAULT_LINE_WIDTH = 1.0
    # noinspection PyUnresolvedReferences
    #: Default line style
    DEFAULT_LINE_STYLE = WHDB_LINE_STYLE.SOLID

    @property
    @abstractmethod
    def lineColour(self):
        """
        The line colour property.

        :return: Line colour as hex string
        :rtype: str
        """
        raise NotImplementedError()

    @lineColour.setter
    @abstractmethod
    def lineColour(self, colour):
        """
        Set the line colour.

        :param colour: Line colour as hex string
        :type colour: str
        """
        raise NotImplementedError()

    @lineColour.deleter
    @abstractmethod
    def lineColour(self):
        """
        Unset the line colour, i.e. set it to the global default value.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def lineWidth(self):
        """
        The line width property.

        :return: Line width
        :rtype: int
        """
        raise NotImplementedError()

    @lineWidth.setter
    @abstractmethod
    def lineWidth(self, width):
        """
        Set the line width.

        :param width: Line width
        :rtype: int
        """
        raise NotImplementedError()

    @lineWidth.deleter
    @abstractmethod
    def lineWidth(self):
        """
        Unset the line width, i.e. set it to the global default value.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def lineStyle(self):
        """
        The line style property.

        :return: Line style (see `WHDB_LINE_STYLE`)
        :rtype: int
        """
        raise NotImplementedError()

    @lineStyle.setter
    @abstractmethod
    def lineStyle(self, style):
        """
        Set the line style.

        :param style: Line style (see `WHDB_LINE_STYLE`)
        :type style: int
        """
        raise NotImplementedError()

    @lineStyle.deleter
    @abstractmethod
    def lineStyle(self):
        """
        Unset the line style, i.e. set it to the global default value.
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbAreaBasedShapeP(WhdbShapeP):
    """
    Base class for area based shapes.

    Area-based shapes are characterized by a fill colour and a fill opacity.
    If the fill colour is None, the fill opacity has usually no effect.

    The fill colour is specified as a RGB hex string ('0xRRGGBB').

    The fill opacity is specified as a float value between 0.0 (fully
    transparent) and 1.0 (fully opaque).
    """
    #: Default fill opacity.
    DEFAULT_FILL_OPACITY = 0.5

    @property
    @abstractmethod
    def fillColour(self):
        """
        The area's fill colour property.

        :return: Fill colour as hex string
        :rtype: str
        """
        raise NotImplementedError()

    @fillColour.setter
    @abstractmethod
    def fillColour(self, colour):
        """
        Set the area's fill colour.

        :param colour: Fill colour as hex string
        :type colour: str
        """
        raise NotImplementedError()

    @fillColour.deleter
    @abstractmethod
    def fillColour(self):
        """
        Unset the fill colour, i.e. set it to None.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def fillOpacity(self):
        """
        The area's fill opacity property.

        :return: Fill opacity between 0.0 and 1.0
        :rtype: float
        """
        raise NotImplementedError()

    @fillOpacity.setter
    @abstractmethod
    def fillOpacity(self, opacity):
        """
        Set the area's fill opacity.

        :param opacity: Fill opacity between 0.0 and 1.0
        :type opacity: float
        """
        raise NotImplementedError()

    @fillOpacity.deleter
    @abstractmethod
    def fillOpacity(self):
        """
        Unset the fill opacity, i.e. set it to the global default value.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def width(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def height(self):
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbTextBasedShapeP(WhdbShapeP):
    """
    Base class for text-based shapes.

    Text-based shapes are characterized by a text colour and a font size.

    The text colour is specified as RGB hex string ('0xRRGGBB').

    The font size is specified in pixels as float value.
    """
    #: Default text colour
    DEFAULT_TEXT_COLOUR = '#000000'
    #: Default text size
    DEFAULT_TEXT_FONT_SIZE = 15
    #: Default horizontal text alignment
    DEFAULT_TEXT_H_ALIGNMENT = WHDB_HALIGN_LEFT
    #: Default vertical text alignment
    DEFAULT_TEXT_V_ALIGNMENT = WHDB_VALIGN_TOP

    @property
    @abstractmethod
    def textColour(self):
        """
        The text colour property.

        :return: Text colour as hex string
        :rtype: str
        """
        raise NotImplementedError()

    @textColour.setter
    @abstractmethod
    def textColour(self, colour):
        """
        Set the text colour.

        :parameter colour: Text colour as hex string
        :type colour: str
        """
        raise NotImplementedError()

    @textColour.deleter
    @abstractmethod
    def textColour(self):
        """
        Unset the font colour, i.e. set it to the global default value.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def fontSize(self):
        """
        The font size property.

        :return: Font size
        :rtype: float
        """
        raise NotImplementedError()

    @fontSize.setter
    @abstractmethod
    def fontSize(self, size):
        """
        Set the font size.

        :param size: Font size
        :type size: float
        """
        raise NotImplementedError()

    @fontSize.deleter
    @abstractmethod
    def fontSize(self):
        """
        Unset the font size, i.e. set it to the global default value.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def horizontalAlignment(self):
        """
        The horizontal alignment property.

        :return: The text's horizontal alignment
        :rtype: int
        """
        raise NotImplementedError()

    @horizontalAlignment.setter
    @abstractmethod
    def horizontalAlignment(self, alignment):
        """
        Set the horizontal alignment.

        :param alignment: The text's new horizontal alignment
        :type alignment: int
        """
        raise NotImplementedError()

    @horizontalAlignment.deleter
    @abstractmethod
    def horizontalAlignment(self):
        """
        Unset the horizontal alignment,
        i.e. set it to the global default value.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def verticalAlignment(self):
        """
        The vertical alignment property.

        :return: The text's vertical alignment
        :rtype: int
        """
        raise NotImplementedError()

    @verticalAlignment.setter
    @abstractmethod
    def verticalAlignment(self, alignment):
        """
        Set the vertical alignment.

        :param alignment: The text's new vertical alignment
        :type alignment: int
        """
        raise NotImplementedError()

    @verticalAlignment.deleter
    @abstractmethod
    def verticalAlignment(self):
        """
        Unset the vertical alignment,
        i.e. set it to the global default value.
        """
        raise NotImplementedError()


# noinspection PyAbstractClass,PyAttributeOutsideInit
class WhdbLineShapeP(WhdbLineBasedShapeP):
    """
    A line shape.

    A line shape is defined by its two end points.
    It's location (see :py:meth:`WhdbShapeP.location`) is defined as the
    center between the two end points.
    """

    @property
    @abstractmethod
    def pointA(self):
        """
        The line's first end point property.

        :return: First end point of the line as (x, y)-coordinates
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @pointA.setter
    @abstractmethod
    def pointA(self, point):
        """
        Set the line's first end point.

        :param point: New first end point of the line as (x, y)-coordinates
        :type point: (float, float)
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def pointB(self):
        """
        The line's second end point property.

        :return: second end point of the line as (x, y)-coordinates
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @pointB.setter
    @abstractmethod
    def pointB(self, point):
        """
        Set the line's second end point.

        :param point: New second end point of the line as (x, y)-coordinates
        :type point: (float, float)
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def pointARel(self):
        """
        The line's first end point in relative coordinates from the center.

        :return: First end point of the line in relative (x, y)-coordinates
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @pointARel.setter
    @abstractmethod
    def pointARel(self, point):
        """
        Set the line's first end point in relative coordinates from the center.

        :param point: New first end point in relative (x, y)-coordinates
        :type point: (float, float)
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def pointBRel(self):
        """
        The line's second end point in relative coordinates from the center.

        :return: Second end point of the line in relative (x, y)-coordinates
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @pointBRel.setter
    @abstractmethod
    def pointBRel(self, point):
        """
        Set the line's second end point in relative coordinates from the
        center.

        :param point: New second end point in relative (x, y)-coordinates
        :type point: (float, float)
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbArrowShapeP(WhdbLineShapeP):
    """
    An arrow shape.

    An arrow shape is defined by its tip and end point ('tail').
    It's location (see WhdbShapeP.location) is defined as the center
    between the two points.
    """

    @property
    @abstractmethod
    def tip(self):
        """
        The arrow's tip property.

        :return: Tip of the arrow as (x, y)-coordinates
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @tip.setter
    @abstractmethod
    def tip(self, point):
        """
        Set the arrow's tip.

        :param point: Tip of the arrow as (x, y)-coordinates
        :type point: (float, float)
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def endPoint(self):
        """
        The arrow's end point property.

        :return: End point of the arrow as (x, y)-coordinates
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @endPoint.setter
    @abstractmethod
    def endPoint(self, point):
        """
        Set the arrow's end point

        :param point: End point of the arrow as (x, y)-coordinates
        :type point: (float, float)
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def tipRel(self):
        """
        The arrow's tip in relative coordinates from the center.

        :return: Tip of the arrow in relative (x, y)-coordinates
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def endPointRel(self):
        """
        The arrow's end point in relative coordinates from the center.

        :return: End point of the arrow in relative (x, y)-coordinates
        :rtype: (float, float)
        """
        raise NotImplementedError()


# noinspection PyAbstractClass,PyAttributeOutsideInit
class WhdbRectangleShapeP(WhdbLineBasedShapeP, WhdbAreaBasedShapeP):
    """
    A rectangle shape.

    A rectangle is defined by its upper-left corner (or center),
    width and height.
    A rectangle's location is defined as its center.
    """

    @property
    @abstractmethod
    def corner(self):
        """
        The rectangle's upper-left corner property.

        :return: Upper-left corner as (x, y)-coordinates
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @corner.setter
    @abstractmethod
    def corner(self, point):
        """
        Set the coordinates of the upper-left corner of the rectangle.

        :param point: New upper-left corner as (x, y)-coordinates
        :type point: (float, float)
        """
        raise NotImplementedError()

    @property
    def width(self):
        """
        The rectangle's width property.

        :return: Width of the rectangle
        :rtype: float
        """
        raise NotImplementedError()

    @width.setter
    @abstractmethod
    def width(self, width):
        """
        Set the width of the rectangle.

        :param width: Width of the rectangle
        :type width: float
        """
        raise NotImplementedError()

    @property
    def height(self):
        """
        The rectangle's height property.

        :return: Height of the rectangle
        :rtype: float
        """
        raise NotImplementedError()

    @height.setter
    @abstractmethod
    def height(self, height):
        """
        Set the height of the rectangle.

        :param height: height of the rectangle
        :type height: float
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WhdbOvalShapeP(WhdbLineBasedShapeP, WhdbAreaBasedShapeP):
    """
    An oval shape.

    An oval is defined by its center, width and height.
    An oval's location is its center.
    """

    @property
    @abstractmethod
    def center(self):
        """
        The oval's center property.

        :return: Center as (x, y)-coordinates
        :rtype: (float, float)
        """
        raise NotImplementedError()

    @center.setter
    @abstractmethod
    def center(self, point):
        """
        Set the coordinates of the center of the oval.

        :param point: New center as (x, y)-coordinates
        :type point: (float, float)
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def width(self):
        """
        The oval's width property.

        :return: Width of the oval
        :rtype: float
        """
        raise NotImplementedError()

    @width.setter
    @abstractmethod
    def width(self, width):
        """
        Set the width of the oval.

        :param width: Width of the oval
        :type width: float
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def height(self):
        """
        The oval's height property.

        :return: Height of the oval
        :rtype: float
        """
        raise NotImplementedError()

    @height.setter
    @abstractmethod
    def height(self, height):
        """
        Set the height of the oval.

        :param height: height of the oval
        :type height: float
        """
        raise NotImplementedError()


# noinspection PyAbstractClass,PyAttributeOutsideInit
class WhdbTextShapeP(WhdbTextBasedShapeP, WhdbLineBasedShapeP,
                     WhdbAreaBasedShapeP):
    """
    A text shape.

    A text is defined by the text contents and its location.
    A text's location is the upper-left corner of its bounding box.
    """

    @property
    @abstractmethod
    def text(self):
        """
        The text contents property.

        :return: The text content
        :rtype: str
        """
        raise NotImplementedError()

    @text.setter
    @abstractmethod
    def text(self, txt):
        """
        Set the text contents.

        :param txt: New text contents
        :type txt: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setText(self, text):
        """
        Set the text contents.

        :param text: New text contents
        :type text: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def boundingBox(self):
        """
        :rtype: ScmnRectangle
        """
        raise NotImplementedError()

    # noinspection PyShadowingNames
    @abstractmethod
    def setBoundingBox(self, x, y, width, height):
        """
        Set the text shape's bounding box.

        :param x: X-position
        :type x: float
        :param y: Y-position
        :type y: float
        :param width: Width
        :type width: float
        :param height: Height
        :type height: float
        """
        raise NotImplementedError()
