# ----------------------------------------------------------------------------
# - wdbi-block                                                               -
# - wild cruncher:ish block models                core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher immutable block models

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.immutable.wdbi_protocols import WdbiWrapperMaker
from wcr_core.dbstore.wdbs_entity import WdbsObservableMixinP
from wcr_core.dbstore_api.wdbs_block_api import WdbsBlockP
from wcr_core.dbstore.immutable.wdbi_tree_db import *
from wcr_core.dbstore.immutable.wdbi_entity import WdbiNodeEntityBase, \
    WdbiTreepathEntityBase, WdbiEntityP, WdbiEntityMixinP
# noinspection PyProtectedMember
from wcr_core.dbstore.immutable.wdbi_instance import _WDBI_INST_CONNECTORS, \
    wdbi_createOperatorInstance, wdbi_createBlockInstance, \
    wdbi_createModelInstance
from wcr_core.dbstore.immutable.wdbi_button import wdbi_createButton, \
    WdbiButtonForTreepath
from wcr_core.dbstore.immutable.wdbi_connection import wdbi_createConnection, \
    WdbiConnectionForTreepath
from wcr_core.dbstore.immutable.wdbi_shape import wdbi_createShape, \
    WdbiShapeForTreepath
from wcr_core.dbstore.immutable.wdbi_instance import \
    wdbi_getDynamicInstanceWrapper
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_UPDATE_CORE_EVENT, \
    WDBS_CHILD_ADDED_EVENT, WDBS_CHILD_REMOVED_EVENT


_WDBI_BLOCK_INSTANCES = 0
_WDBI_BLOCK_CONNECTIONS = 1
_WDBI_BLOCK_SHAPES = 2
_WDBI_BLOCK_BUTTONS = 3
_WDBI_BLOCK_NAME = 4
_WDBI_BLOCK_INFO = 5
_WDBI_BLOCK_TYPE = 6


def wdbi_createBlock(metaPrv, btyp, name=None, info=None, uuid=None, meta=None,
                     **kwargs):
    """
    Create a new block model node.

    :param metaPrv: Meta-description provider
    :type metaPrv: WdbsMetaProviderP
    :param btyp: Block type
    :type btyp: str
    :param name: the model block name
    :type name: str
    :param info: the model block info
    :type info: str
    :param uuid: UUID for the new block (will be created if `None`)
    :type uuid: UUID | str | None
    :param meta: Meta description
    :type meta: WdbsMetaDescriptionP | None
    :return: Newly created block model node
    :rtype: WdbiDbArrayNode
    """
    assert bool(btyp)

    if meta is None:
        meta = metaPrv.getMetaDescription(WdbsBlockP, btyp)

    return WdbiDbArrayNode(
        uuid, WDBI_BLOCK_NODE, meta,
        [WdbiListNode(None, WDBI_INSTANCE_LIST_NODE),
         WdbiListNode(None, WDBI_WIRE_LIST_NODE),
         WdbiListNode(None, WDBI_SHAPE_LIST_NODE),
         WdbiListNode(None, WDBI_BUTTON_LIST_NODE),
         WdbiFieldNode(None, 0, 'name', name if name is not None else ''),
         WdbiFieldNode(None, 0, 'info', info if info is not None else ''),
         WdbiFieldNode(None, 0, 'blockType', btyp)],
        **kwargs)


# noinspection PyAbstractClass
class WdbiBlockEntityMixin(WdbiEntityMixinP, WdbiWrapperMaker, WdbsBlockP,
                           WdbsObservableMixinP):
    @abstractmethod
    def _current(self):
        """
        :rtype: WdbiDbArrayNode
        """
        raise NotImplementedError()

    def _getInstance(self, ident):
        """
        Get the instance node for the given ident.

        :param ident: Ident value
        :type ident: int
        :return: Instance node with the given ident
        :rtype: WdbiDbArrayNode
        """
        return self._current().getPath([_WDBI_BLOCK_INSTANCES, ident])

    def _updateInstance(self, ident, updateF):
        """
        Update a instance.

        :param ident: Ident value of the instance to be updated
        :type ident: int
        :param updateF: Update function
        :type updateF: (WdbiDbArrayNode) -> WdbiDbArrayNode
        """
        self._update([_WDBI_BLOCK_INSTANCES, ident],
                     lambda inst: updateF(inst))

    def _mkInstWrapper(self, inst):
        """
        Wrap an instance.

        :param inst: Instance to be wrapped
        :type inst: WdbiDbArrayNode
        :return: Wrapped instance
        :rtype: WdbiDynamicInstanceWrapper
        """
        idnt = inst.getUUID()
        return self._wrap(
            idnt, wdbi_getDynamicInstanceWrapper(inst), self._getInstance,
            self._updateInstance, idnt)

    def _getConnection(self, ident):
        """
        Get the connection node for the given ident.

        :param ident: Ident value
        :type ident: int
        :return: Connection node with the given ident
        :rtype: WdbiDbArrayNode
        """
        return self._current().getPath([_WDBI_BLOCK_CONNECTIONS, ident])

    def _updateConnection(self, ident, updateF):
        """
        Update a connection.

        :param ident: Ident value of the connection to be updated
        :type ident: int
        :param updateF: Update function
        :type updateF: (WdbiDbArrayNode) -> WdbiDbArrayNode
        """
        self._update([_WDBI_BLOCK_CONNECTIONS, ident],
                     lambda wire: updateF(wire))

    def _mkConnectionWrapper(self, conn):
        """
        Wrap a connection.

        :param conn: Connection to be wrapped
        :type conn: WdbiDbArrayNode
        :return: Wrapped connection
        :rtype: WdbiConnectionForTreepath
        """
        idnt = conn.getUUID()
        return self._wrap(
            idnt, WdbiConnectionForTreepath, self._getConnection,
            self._updateConnection, idnt, self)

    def _getShape(self, ident):
        """
        Get the shape node for the given ident.

        :param ident: Ident value
        :type ident: int
        :return: Shape node with the given ident
        :rtype: WdbiDbArrayNode
        """
        return self._current().getPath([_WDBI_BLOCK_SHAPES, ident])

    def _updateShape(self, ident, updateF):
        """
        Update a shape.

        :param ident: Ident value of the shape to be updated
        :type ident: int
        :param updateF: Update function
        :type updateF: (WdbiDbArrayNode) -> WdbiDbArrayNode
        """
        self._update([_WDBI_BLOCK_SHAPES, ident],
                     lambda shape: updateF(shape))

    def _mkShapeWrapper(self, shape):
        """
        Wrap a shape.

        :param shape: Shape to be wrapped
        :type shape: WdbiDbArrayNode
        :return: Wrapped shape
        :rtype: WdbiEntityP
        """
        idnt = shape.getUUID()
        return self._wrap(
            idnt, WdbiShapeForTreepath, self._getShape, self._updateShape,
            idnt)

    def _getButton(self, ident):
        """
        Get the button node for the given ident.

        :param ident: Ident value
        :type ident: int
        :return: Button node with the given ident
        :rtype: WdbiDbArrayNode
        """
        return self._current().getPath([_WDBI_BLOCK_BUTTONS, ident])

    def _updateButton(self, ident, updateF):
        """
        Update a button.

        :param ident: Ident value of the button to be updated
        :type ident: int
        :param updateF: Update function
        :type updateF: (WdbiDbArrayNode) -> WdbiDbArrayNode
        """
        self._update([_WDBI_BLOCK_BUTTONS, ident],
                     lambda bttn: updateF(bttn))

    def _mkButtonWrapper(self, button):
        """
        Wrap a button.

        :param button: Button to be wrapped
        :type button: WdbiDbArrayNode
        :return: Wrapped button
        :rtype: WdbiButtonForTreepath
        """
        idnt = button.getUUID()
        return self._wrap(
            idnt, WdbiButtonForTreepath, self._getButton, self._updateButton,
            idnt)

    def createWrapperForPath(self, path):
        ntyp, idnt, *rest = path
        inst = self._getInstance(idnt)
        wrapper = self._wrap(
            idnt, wdbi_getDynamicInstanceWrapper(inst), self._getInstance,
            self._updateInstance, idnt)

        if rest:
            # noinspection PyUnresolvedReferences
            return wrapper.createWrapperForPath(rest)
        else:
            return wrapper

    def getEntityType(self):
        return self._current().getPath([_WDBI_BLOCK_TYPE]).getValue()

    def getName(self):
        return self._current().getPath([_WDBI_BLOCK_NAME]).getValue()

    def setName(self, name):
        self._update([_WDBI_BLOCK_NAME], lambda f: f.setValue(name))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def getInfo(self):
        return self._current().getPath([_WDBI_BLOCK_INFO]).getValue()

    def setInfo(self, info):
        self._update([_WDBI_BLOCK_INFO], lambda f: f.setValue(info))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def createOperatorInstance(self, *args, **kwargs):
        func = wdbi_createOperatorInstance(
            self.context, *args, **kwargs)
        self._update([_WDBI_BLOCK_INSTANCES], lambda ilst: ilst.add(func))
        wrap = self._mkInstWrapper(func)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, wrap)
        return wrap

    def createBlockInstance(self, *args, **kwargs):
        func = wdbi_createBlockInstance(
            self.context, *args, **kwargs)
        self._update([_WDBI_BLOCK_INSTANCES], lambda ilst: ilst.add(func))
        wrap = self._mkInstWrapper(func)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, wrap)
        return wrap

    def createModelInstance(self, *args, **kwargs):
        func = wdbi_createModelInstance(
            self.context, *args, **kwargs)
        self._update([_WDBI_BLOCK_INSTANCES], lambda ilst: ilst.add(func))
        wrap = self._mkInstWrapper(func)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, wrap)
        return wrap

    def removeInstance(self, i):
        self._update([_WDBI_BLOCK_INSTANCES], lambda ilst: ilst.remove(i))
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, i)

    def getInstances(self):
        return [self._mkInstWrapper(i) for i
                in self._current()[_WDBI_BLOCK_INSTANCES]]

    def getInstanceCount(self):
        return len(self._current()[_WDBI_BLOCK_INSTANCES])

    def createConnection(self, sobj, sxpt, eobj, expt,
                         name=None, info=None, path=None, uuid=None, meta=None,
                         **kwargs):
        sop = self._getRelativePath(sobj)
        scp = sop + [_WDBI_INST_CONNECTORS, sxpt.getUUID()]
        eop = self._getRelativePath(eobj)
        ecp = eop + [_WDBI_INST_CONNECTORS, expt.getUUID()]
        wire = wdbi_createConnection(
            self.context,
            sop, scp, eop, ecp, name, info, path, uuid, meta, **kwargs)
        self._update([_WDBI_BLOCK_CONNECTIONS], lambda wlst: wlst.add(wire))
        wrap = self._mkConnectionWrapper(wire)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, wrap)
        return wrap

    def removeConnection(self, w):
        self._update([_WDBI_BLOCK_CONNECTIONS], lambda wlst: wlst.remove(w))
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, w)

    def getConnections(self):
        return [self._mkConnectionWrapper(w) for w
                in self._current()[_WDBI_BLOCK_CONNECTIONS]]

    def getConnectionCount(self):
        return len(self._current()[_WDBI_BLOCK_CONNECTIONS])

    def createShape(self, styp, uuid=None, meta=None, **kwargs):
        shape = wdbi_createShape(
            self.context, styp, uuid, meta, **kwargs)
        self._update([_WDBI_BLOCK_SHAPES], lambda slst: slst.add(shape))
        wrap = self._mkShapeWrapper(shape)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, wrap)
        return wrap

    def removeShape(self, shape):
        self._update([_WDBI_BLOCK_SHAPES], lambda slst: slst.remove(shape))
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, shape)

    def getShapes(self):
        return [self._mkShapeWrapper(s) for s
                in self._current()[_WDBI_BLOCK_SHAPES]]

    def getShapeCount(self):
        return len(self._current()[_WDBI_BLOCK_SHAPES])

    def createButton(self, *args, **kwargs):
        button = wdbi_createButton(self.context, *args, **kwargs)
        self._update([_WDBI_BLOCK_BUTTONS], lambda blst: blst.add(button))
        wrap = self._mkButtonWrapper(button)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, wrap)
        return wrap

    def removeButton(self, element):
        self._update([_WDBI_BLOCK_BUTTONS],
                     lambda slst: slst.remove(element))
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, element)

    def getButtons(self):
        return [self._mkButtonWrapper(b) for b
                in self._current()[_WDBI_BLOCK_BUTTONS]]

    def getButtonCount(self):
        return len(self._current()[_WDBI_BLOCK_BUTTONS])

    def _getRelativePath(self, item):
        return [_WDBI_BLOCK_INSTANCES, item.getUUID()]


class WdbiBlockForNode(WdbiNodeEntityBase, WdbiBlockEntityMixin):
    """
    Wrapper for block nodes, managing current node itself.
    """

    def __init__(self, dctx, node):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param node: Node to wrap
        :type node: WdbiDbArrayNode
        """
        WdbiNodeEntityBase.__init__(self, dctx, node)


class WdbiBlockForTreepath(WdbiTreepathEntityBase, WdbiBlockEntityMixin):
    """
    Wrapper for externally managed block node.
    """

    def __init__(self, dctx, currentF, updateF, ident):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param currentF: Function providing the current block node
        :type currentF: (int) -> WdbiDbArrayNode
        :param updateF: Function to update the current block node
        :type updateF: ((int, WdbiDbArrayNode) -> WdbiDbArrayNode) -> _
        :param ident: Ident of the wrapped node
        :type ident: int
        """
        WdbiTreepathEntityBase.__init__(self, dctx, currentF, updateF, ident)
