# ----------------------------------------------------------------------------
# - wdbi-button                                                              -
# - wild cruncher:ish interactive elements        core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_button
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher immutable interactive elements

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.immutable.wdbi_tree import WdbiDictNode, WdbiFieldNode
from wcr_core.dbstore.immutable.wdbi_parameter import wdbi_createParam, \
    WdbiEntityParametersMixin
from wcr_core.dbstore.immutable.wdbi_tree_db import WDBI_BUTTON_NODE, \
    WdbiDbArrayNode, WDBI_PARAM_DICT_NODE
from wcr_core.dbstore.immutable.wdbi_entity import WdbiTreepathEntityBase, \
    WdbiNodeEntityBase
from wcr_core.dbstore.wdbs_entity import WdbsObservableMixinP
from wcr_core.dbstore_api.wdbs_buttons_api import WdbsButtonP
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_UPDATE_CORE_EVENT


_WDBI_BUTTON_TITLE = 0
_WDBI_BUTTON_COMMAND = 1
_WDBI_BUTTON_PARAMS = 2


def wdbi_createButton(metaPrv, title=None, command=None, params=None,
                      uuid=None, meta=None, **kwargs):
    """
    Create a new button node.

    :param metaPrv: Meta-description provider
    :type metaPrv: WdbsMetaProviderP
    :param title: Button title
    :type title: str | None
    :param command: Command associated with button
    :type command: str | None
    :param bpos: Position
    :type params: Iterable[(str, object)] | None
    :param uuid: UUID
    :type uuid: UUID | str | None
    :param meta: Meta description
    :type meta: WdbsMetaDescriptionP | None
    :param kwargs: Additional meta-data
    :return: Newly created button node
    :rtype: WdbiDbArrayNode
    """
    if meta is None:
        meta = metaPrv.getMetaDescription(WdbsButtonP, command)

    node = WdbiDbArrayNode(
        uuid, WDBI_BUTTON_NODE, meta,
        [WdbiFieldNode(None, 0, 'title', title if title is not None else ''),
         WdbiFieldNode(
             None, 0, 'command', command if command is not None else ''),
         WdbiDictNode(None, WDBI_PARAM_DICT_NODE)],
        **kwargs)
    if params:
        for p, v in params:
            node = node.update(
                [_WDBI_BUTTON_PARAMS],
                lambda ps: ps.add(p, wdbi_createParam(metaPrv, p, value=v)))
    return node


# noinspection PyAbstractClass
class WdbiButtonEntityMixin(WdbsButtonP, WdbiEntityParametersMixin,
                            WdbsObservableMixinP):
    """
    Wrapper for button nodes.
    """

    def __init__(self):
        WdbiEntityParametersMixin.__init__(self, [_WDBI_BUTTON_PARAMS])

    def getTitle(self):
        # noinspection PyUnresolvedReferences
        return self._current().get(_WDBI_BUTTON_TITLE).getValue()

    def setTitle(self, title):
        self._update([_WDBI_BUTTON_TITLE], lambda f: f.setValue(title))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def getCommand(self):
        # noinspection PyUnresolvedReferences
        return self._current().get(_WDBI_BUTTON_COMMAND).getValue()

    def setCommand(self, command):
        self._update([_WDBI_BUTTON_COMMAND], lambda f: f.setValue(command))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)


class WdbiButtonForNode(WdbiNodeEntityBase, WdbiButtonEntityMixin):
    """
    Wrapper for button nodes, handling current node internally.
    """

    def __init__(self, dctx, node):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param node: Node to wrap
        :type node: WdbiDbArrayNode
        """
        WdbiButtonEntityMixin.__init__(self)
        WdbiNodeEntityBase.__init__(self, dctx, node)


class WdbiButtonForTreepath(WdbiTreepathEntityBase, WdbiButtonEntityMixin):
    """
    Wrapper for externally managed button nodes.
    """

    def __init__(self, dctx, currentF, updateF, ident):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param currentF: Function providing the current button node
        :type currentF: (int) -> WdbiDbArrayNode
        :param updateF: Function to update the current botton node
        :type updateF: ((int, WdbiDbArrayNode) -> WdbiDbArrayNode) -> None
        :param ident: Ident of the wrapped node
        :type ident: int
        """
        WdbiButtonEntityMixin.__init__(self)
        WdbiTreepathEntityBase.__init__(self, dctx, currentF, updateF, ident)
