# ----------------------------------------------------------------------------
# - wdbi-connection                                                          -
# - wild cruncher:ish connection objects          core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_connection
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher immutable connection objects

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.immutable.wdbi_geometry import wdbi_createPath, \
    WdbiPathForTreepath
from wcr_core.dbstore.immutable.wdbi_tree import WdbiFieldNode
from wcr_core.dbstore.immutable.wdbi_tree_db import WDBI_WIRE_NODE, \
    WdbiDbArrayNode
from wcr_core.dbstore.immutable.wdbi_entity import WdbiTreepathEntityBase, \
    WdbiNodeEntityBase, WdbiEntityMixinP
from wcr_core.dbstore_api.wdbs_connection_api import WdbsConnectionP
from wcr_core.dbstore.wdbs_entity import WdbsObservableMixinP
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_UPDATE_CORE_EVENT


#: Array index for connection path
_WDBI_CONNECTION_SRCOBJ = 0
_WDBI_CONNECTION_SRCXPT = 1
_WDBI_CONNECTION_DSTOBJ = 2
_WDBI_CONNECTION_DSTXPT = 3
_WDBI_CONNECTION_NAME = 4
_WDBI_CONNECTION_INFO = 5
_WDBI_CONNECTION_PATH = 6


def wdbi_createConnection(metaPrv, sopt, sxpt, eopt, expt, name=None,
                          info=None, path=None, uuid=None, meta=None,
                          **kwargs):
    """
    Create a new connection node.

    :param metaPrv: Meta-description provider
    :type metaPrv: WdbsMetaProviderP
    :param sopt: Start object relative path
    :type sopt: list[object]
    :param sxpt: Start connector relative path
    :type sxpt: list[object]
    :param eopt: End object relative path
    :type eopt: list[object]
    :param expt: End connector relative path
    :type expt: list[object]
    :param name: the connection name
    :type name: str
    :param info: the connection information
    :type info: str
    :param path: the path for the connection
    :type path: list[(float, float)]
    :param uuid: UUID
    :type uuid: UUID | str | None
    :param meta: Meta description
    :type meta: WdbsMetaDescriptionP | None
    :return: Newly created connection node
    :rtype: WdbiDbArrayNode
    """
    if meta is None:
        meta = metaPrv.getMetaDescription(WdbsConnectionP)

    return WdbiDbArrayNode(
        uuid, WDBI_WIRE_NODE, meta,
        [WdbiFieldNode(None, 0, 'srcobj', sopt),
         WdbiFieldNode(None, 0, 'srcxpt', sxpt),
         WdbiFieldNode(None, 0, 'dstobj', eopt),
         WdbiFieldNode(None, 0, 'dstxpt', expt),
         WdbiFieldNode(None, 0, 'name', name if name is not None else ''),
         WdbiFieldNode(None, 0, 'info', info if info is not None else ''),
         wdbi_createPath(metaPrv, path)],
        **kwargs)


# noinspection PyUnresolvedReferences,PyAbstractClass
class WdbiConnectionEntityMixin(WdbiEntityMixinP, WdbsConnectionP,
                                WdbsObservableMixinP):
    """
    Wrapper for connection nodes.
    """

    def __init__(self, wmkr):
        """
        :param wmkr: Wrapper-maker for connection end-points
        :type wmkr: WdbiWrapperMaker
        """
        #: Wrapper maker
        self.__wrapperMaker = wmkr

    def _getPath(self, *_):
        """
        Get the connection path node.

        :return: Connection path node
        :rtype: WdbiListNode
        """
        return self._current()[_WDBI_CONNECTION_PATH]

    def _updatePath(self, _, updateF):
        """
        Update the path node.

        :param updateF: Update function
        :type updateF: (WdbiListNode) -> WdbiListNode
        """
        self._update([_WDBI_CONNECTION_PATH], lambda path: updateF(path))

    def _mkPathWrapper(self):
        """
        Wrap the path node.

        :return: Wrapped path node
        :rtype: WdbiPathForTreepath
        """
        uuid = self._current()[_WDBI_CONNECTION_PATH].getUUID()
        return self._wrap(
            uuid, WdbiPathForTreepath, self._getPath, self._updatePath, uuid)

    def getName(self):
        return self._current().get(_WDBI_CONNECTION_NAME).getValue()

    def setName(self, name):
        self._update([_WDBI_CONNECTION_NAME], lambda f: f.setValue(name))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def getInfo(self):
        return self._current().get(_WDBI_CONNECTION_INFO).getValue()

    def setInfo(self, info):
        self._update([_WDBI_CONNECTION_INFO], lambda f: f.setValue(info))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def getStartObject(self):
        try:
            return self.__wrapperMaker.createWrapperForPath(
                self._current().get(_WDBI_CONNECTION_SRCOBJ).getValue())
        except KeyError:
            return None

    def getStartXPoint(self):
        try:
            return self.__wrapperMaker.createWrapperForPath(
                self._current().get(_WDBI_CONNECTION_SRCXPT).getValue())
        except KeyError:
            return None

    def getEndObject(self):
        try:
            return self.__wrapperMaker.createWrapperForPath(
                self._current().get(_WDBI_CONNECTION_DSTOBJ).getValue())
        except KeyError:
            return None

    def getEndXPoint(self):
        try:
            return self.__wrapperMaker.createWrapperForPath(
                self._current().get(_WDBI_CONNECTION_DSTXPT).getValue())
        except KeyError:
            return None

    def getPath(self):
        return self._mkPathWrapper()

    def setPath(self, path):
        self._update([_WDBI_CONNECTION_PATH],
                     lambda _: wdbi_createPath(self.context, path))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)


class WdbiConnectionForNode(WdbiNodeEntityBase, WdbiConnectionEntityMixin):
    """
    Wrapper for connection nodes, manages current node itself.
    """

    def __init__(self, dctx, node, wmkr):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param node: Node to wrap
        :type node: WdbiDbArrayNode
        :param wmkr: Wrapper-maker for connection end-points
        :type wmkr: WdbiWrapperMaker
        """
        WdbiNodeEntityBase.__init__(self, dctx, node)
        WdbiConnectionEntityMixin.__init__(self, wmkr)


class WdbiConnectionForTreepath(WdbiTreepathEntityBase,
                                WdbiConnectionEntityMixin):
    """
    Wrapper for externally managed connection nodes.
    """

    def __init__(self, dctx, currentF, updateF, ident, wmkr):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param currentF: Function providing the current connection node
        :type currentF: (int) -> WdbiDbArrayNode
        :param updateF: Function to update the current connection node
        :type updateF: ((int, WdbiDbArrayNode) -> WdbiDbArrayNode) -> None
        :param ident: Ident of the wrapped node
        :type ident: int
        :param wmkr: Wrapper-maker for connection end-points
        :type wmkr: WdbiWrapperMaker
        """
        WdbiTreepathEntityBase.__init__(self, dctx, currentF, updateF, ident)
        WdbiConnectionEntityMixin.__init__(self, wmkr)
