# ----------------------------------------------------------------------------
# - wdbi-model                                                               -
# - wild cruncher:ish models                      core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_model
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher immutable models

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.immutable.wdbi_tree import WdbiFieldNode
from wcr_core.dbstore.immutable.wdbi_tree_db import WdbiDbArrayNode, \
    WDBI_MODEL_NODE
from wcr_core.dbstore.immutable.wdbi_entity import WdbiNodeEntityBase, \
    WdbiTreepathEntityBase, WdbiEntityMixinP
from wcr_core.dbstore.wdbs_entity import WdbsObservableMixinP
from wcr_core.dbstore_api.wdbs_model_api import WdbsModelP
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_UPDATE_CORE_EVENT


_WDBI_MODEL_NAME = 0
_WDBI_MODEL_INFO = 1
_WDBI_MODEL_TYPE = 2
_WDBI_MODEL_PATH = 3
_WDBI_MODEL_EPNT = 4


def wdbi_createModel(metaPrv, mtyp, name=None, info=None, path=None, epnt=None,
                     uuid=None, meta=None, **kwargs):
    """
    Create a new model node.

    :param metaPrv: Meta-description provider
    :type metaPrv: WdbsMetaProviderP
    :param mtyp: Model type
    :type mtyp: str
    :param name: the object name
    :type name: str
    :param info: the object info
    :type info: str
    :param path: Path to the model file
    :type path: str
    :param epnt: Function name
    :type epnt: str
    :param uuid: UUID
    :type uuid: UUID | str | None
    :param meta: Meta description
    :type meta: WdbsMetaDescriptionP | None
    :param kwargs: Additional meta-data
    :return: Newly created text model node
    :rtype: WdbiDbArrayNode
    """
    assert bool(mtyp)

    if meta is None:
        meta = metaPrv.getMetaDescription(WdbsModelP, mtyp)

    return WdbiDbArrayNode(
        uuid, WDBI_MODEL_NODE, meta,
        [WdbiFieldNode(None, 0, 'name', name if name is not None else ''),
         WdbiFieldNode(None, 0, 'info', info if info is not None else ''),
         WdbiFieldNode(None, 0, 'modelType', mtyp),
         WdbiFieldNode(None, 0, 'path', path if path is not None else ''),
         WdbiFieldNode(None, 0, 'epoint', epnt if epnt is not None else '')],
        **kwargs)


# noinspection PyUnresolvedReferences,PyAbstractClass
class WdbiModelEntityMixin(WdbiEntityMixinP, WdbsModelP, WdbsObservableMixinP):
    """
    Wrapper for model nodes.
    """

    def getEntityType(self):
        return self._current().getPath([_WDBI_MODEL_TYPE]).getValue()

    def getName(self):
        return self._current().getPath([_WDBI_MODEL_NAME]).getValue()

    def setName(self, name):
        self._update([_WDBI_MODEL_NAME], lambda f: f.setValue(name))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def getInfo(self):
        return self._current().getPath([_WDBI_MODEL_INFO]).getValue()

    def setInfo(self, info):
        self._update([_WDBI_MODEL_INFO], lambda f: f.setValue(info))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def getPath(self):
        return self._current().getPath([_WDBI_MODEL_PATH]).getValue()

    def setPath(self, path):
        self._update([_WDBI_MODEL_PATH], lambda f: f.setValue(path))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def getEntryPoint(self):
        return self._current().getPath([_WDBI_MODEL_EPNT]).getValue()

    def setEntryPoint(self, entryPoint):
        self._update([_WDBI_MODEL_EPNT], lambda f: f.setValue(entryPoint))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)


class WdbiModelForNode(WdbiNodeEntityBase, WdbiModelEntityMixin):
    """
    Wrapper for model nodes, handling current node internally.
    """

    def __init__(self, dctx, node):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param node: Node to wrap
        :type node: WdbiDbArrayNode
        """
        WdbiNodeEntityBase.__init__(self, dctx, node)


class WdbiModelForTreepath(WdbiTreepathEntityBase, WdbiModelEntityMixin):
    """
    Wrapper for externally managed model nodes.
    """

    def __init__(self, dctx, currentF, updateF, ident):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param currentF: Function providing the current text model node
        :type currentF: (int) -> WdbiDbArrayNode
        :param updateF: Function to update the current text model node
        :type updateF: ((int, WdbiDbArrayNode) -> WdbiDbArrayNode) -> None
        :param ident: Ident of the wrapped node
        :type ident: int
        """
        WdbiTreepathEntityBase.__init__(self, dctx, currentF, updateF, ident)
