# ----------------------------------------------------------------------------
# - wdbi-parameter                                                           -
# - wild cruncher:ish parameter objects           core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_parameter
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher immutable parameter objects

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.dbstore.immutable.wdbi_tree import WdbiFieldNode
from wcr_core.dbstore.immutable.wdbi_tree_db import WDBI_PARAMETER_NODE, \
    WdbiDbFieldNode
from wcr_core.dbstore.immutable.wdbi_entity import WdbiNodeEntityBase, \
    WdbiTreepathEntityBase, WdbiEntityMixinP
from wcr_core.dbstore.wdbs_entity import WdbsObservableMixinP
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_CHILD_ADDED_EVENT, \
    WDBS_CHILD_REMOVED_EVENT, WdbsObserverP, WDBS_UPDATE_PARAM_EVENT, \
    WDBS_UPDATE_CORE_EVENT
from wcr_core.dbstore_api.wdbs_parameter_api import WdbsParameterP, \
    WdbsEntityParametersP


_WDBI_PARAM_NAME = 0
_WDBI_PARAM_VALUE = 1


def wdbi_createParam(metaPrv, name, value=None, uuid=None, meta=None,
                     **kwargs):
    """
    Create a new parameter node.

    :param metaPrv: Meta-description provider
    :type metaPrv: WdbsMetaProviderP
    :param name: Parameter name
    :type name: str
    :param value: Parameter value
    :type value: object
    :param uuid: UUID
    :type uuid: UUID | str | None
    :param meta: Meta description
    :type meta: WdbsMetaDescriptionP | None
    :return: Newly created parameter node
    :rtype: WdbiDbFieldNode
    """
    if meta is None:
        meta = metaPrv.getMetaDescription(WdbsParameterP)
    return WdbiDbFieldNode(
        uuid, WDBI_PARAMETER_NODE, meta, name, value=value, **kwargs)


# noinspection PyUnresolvedReferences,PyAbstractClass
class WdbiParameterEntityMixin(WdbiEntityMixinP, WdbsParameterP,
                               WdbsObservableMixinP):
    """
    Wrapper for parameter nodes.
    """

    def registerObserver(self, observer, event=None):
        super().registerObserver(observer, event)

    def getName(self):
        return self._current().getFieldName()

    def getValue(self):
        return self._current().getValue()

    def setValue(self, val):
        self._update([], lambda f: f.setValue(val))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT, val)


class WdbiParameterForNode(WdbiNodeEntityBase, WdbiParameterEntityMixin):
    """
    Wrapper for parameter nodes, handling current node internally.
    """

    def __init__(self, dctx, node):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param node: Node to wrap
        :type node: WdbiDbFieldNode
        """
        WdbiNodeEntityBase.__init__(self, dctx, node)


class WdbiParameterForTreepath(WdbiTreepathEntityBase,
                               WdbiParameterEntityMixin):
    """
    Wrapper for externally managed parameter nodes.
    """

    def __init__(self, dctx, currentF, updateF, ident):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param currentF: Function providing the current parameter node
        :type currentF: (int) -> WdbiDbFieldNode
        :param updateF: Function to update the current parameter node
        :type updateF: ((int, WdbiDbFieldNode) -> WdbiDbFieldNode) -> None
        :param ident: Ident of the wrapped node
        :type ident: int
        """
        WdbiTreepathEntityBase.__init__(self, dctx, currentF, updateF, ident)


# noinspection PyAbstractClass
class WdbiEntityParametersMixin(WdbiEntityMixinP, WdbsEntityParametersP,
                                WdbsObservableMixinP, WdbsObserverP):
    """
    Protocol for node wrappers for nodes with parameter children nodes.
    """

    def __init__(self, path):
        """
        :param path: Path to parameter children nodes
        :type path: list[object]
        """
        self.__path = path

    @property
    @abstractmethod
    def context(self):
        """
        :rtype: WdbiContext
        """
        raise NotImplementedError()

    def _getParameterNode(self):
        """
        Get the node holding the parameters.

        :return: Node holding the parameters
        :rtype: WdbiDictNode
        """
        if self.__path:
            return self._current().getPath(self.__path)
        else:
            return self._current()

    def _getParameter(self, ident):
        """
        Get the parameter node for the given ident.

        :param ident: Ident value
        :type ident: int
        :return: Parameter node with the given ident
        :rtype: WdbiFieldNode
        """
        if self.__path:
            return self._current().getPath(self.__path + [ident])
        else:
            return self._current().getPath([ident])

    def _updateParameter(self, ident, updateF):
        """
        Update a parameter.

        :param ident: Ident value of the parameter to be updated
        :type ident: int
        :param updateF: Update function
        :type updateF: (WdbiFieldNode) -> WdbiFieldNode
        """
        self._update(self.__path + [ident], lambda param: updateF(param))

    def _createParamWrapper(self, *args, **kwargs):
        param = WdbiParameterForTreepath(*args, **kwargs)
        param.registerObserver(self)
        return param

    def _mkParamWrapper(self, param):
        """
        Wrap a parameter node.

        :param param: Parameter node to be wrapped
        :type param: WdbiFieldNode
        :return: Wrapped parameter
        :rtype: WdbiParameterForTreepath
        """
        idnt = param.getUUID()
        return self._wrap(
            idnt, self._createParamWrapper, self._getParameter,
            self._updateParameter, param.getFieldName())

    def getParameterCount(self):
        return len(self._getParameterNode())

    def getParameters(self):
        return [self._mkParamWrapper(p) for _, p in self._getParameterNode()]

    def getParameter(self, name):
        param = self._getParameterNode().get(name)
        return self._mkParamWrapper(param) if param is not None else None

    def getOrCreateParameter(self, name):
        param = self._getParameterNode().get(name)
        if param is not None:
            wrpd = self._mkParamWrapper(param)
        else:
            param = wdbi_createParam(self.context, name)
            self._update(self.__path, lambda pdct: pdct.add(name, param))
            wrpd = self._mkParamWrapper(param)
            self._notifyObservers(WDBS_CHILD_ADDED_EVENT, wrpd)
        return wrpd

    def removeParameter(self, name):
        param = self.getParameter(name)
        if param is not None:
            param.unregisterObserver(self)
            self._update(self.__path, lambda pdct: pdct.remove(name))
            self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, param)

    def clearParameters(self):
        params = list(self.getParameters())
        for param in params:
            param.unregisterObserver(self)
        self._update(self.__path, lambda pdct: pdct.clear())
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, *params)

    def dbNotify(self, obj, event, *args, **kwargs):
        if isinstance(obj, WdbsParameterP):
            self._notifyObservers(
                WDBS_UPDATE_PARAM_EVENT, obj.getValue(), obj.getName())
