# ----------------------------------------------------------------------------
# - wdbi-tree-db                                                             -
# - wild cruncher: db nodes                       core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_tree_db
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - DB nodes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.immutable.wdbi_tree import *
from wcr_core.dbstore.immutable.wdbi_tree_meta import wdbi_createMeta

# Node types
WDBI_GENERIC_NODE = 0
WDBI_CONNECTOR_NODE = 1
WDBI_OPERATOR_INSTANCE_NODE = 2
WDBI_BLOCK_INSTANCE_NODE = 3
WDBI_MODEL_INSTANCE_NODE = 4
WDBI_BLOCK_NODE = 5
WDBI_MODEL_NODE = 6
WDBI_PARAMETER_NODE = 7
WDBI_PATH_NODE = 8
WDBI_POINT_NODE = 9
WDBI_SHAPE_NODE = 10
WDBI_WIRE_NODE = 11
WDBI_BUTTON_NODE = 12
WDBI_CONNECTOR_LIST_NODE = 101
WDBI_INSTANCE_LIST_NODE = 103
WDBI_WIRE_LIST_NODE = 104
WDBI_SHAPE_LIST_NODE = 105
WDBI_PARAM_DICT_NODE = 106
WDBI_BUTTON_LIST_NODE = 107
WDBI_META_NODE = 999


# noinspection PyAbstractClass
class WdbiDbNodeP(WdbiNodeP, WdbiNodeMixinP):
    __slots__ = tuple()

    @property
    @abstractmethod
    def metaDescription(self):
        """
        Get the meta-description for the node.

        :return: Meta-description
        :rtype: WdbsMetaDescriptionP
        """
        raise NotImplementedError()

    @abstractmethod
    def getMeta(self, *keys):
        """
        Get the meta-item for the given key sequence.

        :param keys: Key sequence
        :type keys: tuple
        :return: Meta-node at the given key sequence
        :rtype: WdbiMetaNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def setMeta(self, value, *keys):
        """
        Set the node's meta data entry for the given key sequence.

        :param keys: Key sequence
        :type keys: tuple
        :param value: Value
        :type value: object
        :return: Updated version of the node
        :rtype: WdbiDbNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def updateMeta(self, updateF, *keys):
        """
        Update the node's meta data entry for the given key sequence
        with the given update function.

        :param keys: Key sequence
        :type keys: tuple
        :param updateF: Update function
        :type updateF: (object) -> object
        :return: Updated version of the node
        :rtype: WdbiDbNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def unsetMeta(self, *keys):
        """
        Remove th meta-data entry at the given key sequence from the node.

        :param keys: Key sequence of meta node to remove
        :type keys: tuple
        :return: Updated version of the node
        :rtype: WdbiDbNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def getMetas(self, *keys):
        """
        Get all meta nodes for the given key sequence.

        :param keys: Key sequence
        :type keys: tuple
        :return: Meta-nodes at the given key sequence
        :rtype: dict[object, WdbiMetaNodeP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getAllMeta(self):
        """
        Get the node's meta-data as key-value (!) pairs.

        :return: Meta-data
        :rtype: dict[str, object]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbiDbNodeMixin(WdbiDbNodeP):
    __slots__ = tuple()

    slots = ('_WdbiDbNodeMixin__meta', '_WdbiDbNodeMixin__metaDesc')

    def __init__(self, metaDesc, **kwargs):
        """
        :param metaDesc: Meta-data description
        :type metaDesc: WdbsMetaDescriptionP
        :param kwargs: Initial meta-data
        """
        assert metaDesc is not None
        if metaDesc:
            meta = {m.name: wdbi_createMeta(m.metaType, kwargs.get(m.name))
                    for m in metaDesc.fields}
        else:
            meta = None
        assert meta is not None

        self.__meta = WdbiDictNode(None, WDBI_META_NODE, meta)
        self.__metaDesc = metaDesc

    @property
    def metaDescription(self):
        return self.__metaDesc

    def _cloneDbNodeInto(self, new, **kwargs):
        """
        Clone the node's values into the new node.

        :param new: New node to clone values into
        :type new: WdbiDbNodeMixin
        """
        meta = kwargs.get('meta')
        new.__meta = meta if meta is not None else self.__meta
        new.__metaDesc = self.__metaDesc

    def getMeta(self, *keys):
        return self.__meta.getPath(keys)

    def setMeta(self, value, *keys):
        meta = self.__meta.update(keys, lambda m: m.setFromValue(value))
        return self._clone(meta=meta)

    def updateMeta(self, updateF, *keys):
        meta = self.__meta.update(keys, lambda d: updateF(d))
        return self._clone(meta=meta)

    def unsetMeta(self, *keys):
        meta = self.__meta
        if not keys:
            for field in self.__metaDesc.fields:
                meta = meta.update((field.name,), lambda m: m.unset())
        elif len(keys) == 1:
            try:
                meta = meta.update(keys, lambda m: m.unset())
            except KeyError:
                pass
        else:
            meta = meta.update(keys[:-1], lambda m: m.removeMeta(keys[-1]))
        return self._clone(meta=meta)

    def getMetas(self, *keys):
        return {k: m for k, m in self.__meta}

    def getAllMeta(self):
        return {k: m.asValue() for k, m in self.__meta}


class WdbiDbArrayNode(WdbiArrayNode, WdbiDbNodeMixin):
    __slots__ = WdbiArrayNode.__slots__ + WdbiDbNodeMixin.slots

    def __init__(self, uuid, ntyp, mdsc, items, **kwargs):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        :param mdsc: Meta-data description
        :type mdsc: WdbsMetaDescriptionP
        :param items: Array items
        :type items: list[WdbiNodeP]
        :param kwargs: Initial meta-data
        """
        WdbiArrayNode.__init__(self, uuid, ntyp, items)
        WdbiDbNodeMixin.__init__(self, mdsc, **kwargs)

    def _clone(self, **kwargs):
        # noinspection PyArgumentList
        new = WdbiDbArrayNode.__new__(WdbiDbArrayNode)
        self._cloneArrayInto(new, **kwargs)
        self._cloneDbNodeInto(new, **kwargs)
        return new


class WdbiDbListNode(WdbiListNode, WdbiDbNodeMixin):
    __slots__ = WdbiListNode.__slots__ + WdbiDbNodeMixin.slots

    def __init__(self, uuid, ntyp, mdsc, items=None, **kwargs):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        :param mdsc: Meta-data description
        :type mdsc: WdbsMetaDescriptionP
        :param items: Initial list items
        :type items: list[WdbiNodeP] | None
        :param kwargs: Initial meta-data
        """
        WdbiListNode.__init__(self, uuid, ntyp, items)
        WdbiDbNodeMixin.__init__(self, mdsc, **kwargs)

    def _clone(self, **kwargs):
        # noinspection PyArgumentList
        new = WdbiDbListNode.__new__(WdbiDbListNode)
        self._cloneListInto(new, **kwargs)
        self._cloneDbNodeInto(new, **kwargs)
        return new


class WdbiDbDictNode(WdbiDictNode, WdbiDbNodeMixin):
    __slots__ = WdbiDictNode.__slots__ + WdbiDbNodeMixin.slots

    def __init__(self, uuid, ntyp, mdsc, items=None, **kwargs):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        :param mdsc: Meta-data description
        :type mdsc: WdbsMetaDescriptionP
        :param items: Initial dict items
        :type items: dict[object, WdbiNodeP] | list[(object, WdbiNodeP)] | None
        :param kwargs: Initial meta-data
        """
        WdbiDictNode.__init__(self, uuid, ntyp, items)
        WdbiDbNodeMixin.__init__(self, mdsc, **kwargs)

    def _clone(self, **kwargs):
        # noinspection PyArgumentList
        new = WdbiDbDictNode.__new__(WdbiDbDictNode)
        self._cloneDictInto(new, **kwargs)
        self._cloneDbNodeInto(new, **kwargs)
        return new


class WdbiDbLeafNode(WdbiLeafNode, WdbiDbNodeMixin):
    __slots__ = WdbiLeafNode.__slots__ + WdbiDbNodeMixin.slots

    def __init__(self, uuid, ntyp, mdsc, **kwargs):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        :param mdsc: Meta-data description
        :type mdsc: WdbsMetaDescriptionP
        :param kwargs: Initial meta-data
        """
        WdbiLeafNode.__init__(self, uuid, ntyp)
        WdbiDbNodeMixin.__init__(self, mdsc, **kwargs)

    def _clone(self, **kwargs):
        # noinspection PyArgumentList
        new = WdbiDbLeafNode.__new__(WdbiDbLeafNode)
        self._cloneLeafInto(new, **kwargs)
        self._cloneDbNodeInto(new, **kwargs)
        return new


class WdbiDbFieldNode(WdbiFieldNode, WdbiDbNodeMixin):
    __slots__ = WdbiFieldNode.__slots__ + WdbiDbNodeMixin.slots

    def __init__(self, uuid, ntyp, mdsc, field, value=None, **kwargs):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        :param mdsc: Meta-data description
        :type mdsc: WdbsMetaDescriptionP
        :param field: Field name
        :type field: str
        :param value: Initial field value
        :type value: object | None
        :param kwargs: Initial meta-data
        """
        WdbiFieldNode.__init__(self, uuid, ntyp, field, value=value)
        WdbiDbNodeMixin.__init__(self, mdsc, **kwargs)

    def _clone(self, **kwargs):
        # noinspection PyArgumentList
        new = WdbiDbFieldNode.__new__(WdbiDbFieldNode)
        self._cloneFieldInto(new, **kwargs)
        self._cloneDbNodeInto(new, **kwargs)
        return new
