# ----------------------------------------------------------------------------
# - wdbi-tree-meta                                                           -
# - wild cruncher: meta-data nodes                core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_tree_meta
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Meta-data nodes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from silkan_common.scmn_tuple import ScmnMTuple
from wcr_core.dbstore.immutable.wdbi_tree import *
from wcr_core.dbstore.wdbs_meta_data import *
from wcr_core.dbstore_api.wdbs_meta_type_api import *


WDBI_META_ATOMIC_CODE = 1001
WDBI_META_TUPLE_CODE = 1002
WDBI_META_RECORD_CODE = 1003
WDBI_META_LIST_CODE = 1004
WDBI_META_DICT_CODE = 1005
WDBI_META_ORDERED_DICT_CODE = 1006
WDBI_META_UNION_CODE = 1007


# noinspection PyAbstractClass
class WdbiMetaNodeP(WdbiNodeP, WdbiNodeMixinP, WdbsMetaP):
    __slots__ = tuple()

    @abstractmethod
    def setFromValue(self, *args, **kwargs):
        """
        Set the value of the meta-data instance from the given raw value(s).

        The use of the args/kwargs-arguments depends on the actual
        implementation.

        :return: Meta-node with new values set
        :rtype: WdbiMetaNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def unset(self):
        """
        Unset the mata-data instance's value.

        :return: Meta-node with all values unset
        :rtype: WdbiMetaNodeP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbiStructuralMetaP(WdbiMetaNodeP, WdbsStructuralMetaP):
    __slots__ = tuple()

    @abstractmethod
    def removeMeta(self, key):
        """
        Remove the child item with the given key.

        :param key: Key of child item to remove
        :type key: object
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbiMetaNodeMixin(WdbiMetaNodeP):

    __slots__ = tuple()


class WdbiMetaTupleNode(WdbiArrayNode, WdbiMetaNodeMixin, WdbiStructuralMetaP,
                        WdbsTupleMetaMixin):

    __slots__ = WdbiArrayNode.__slots__ + WdbsTupleMetaMixin._SLOTS_

    @classmethod
    def _clearItems(cls, items):
        """
        :type items: list[WdbiMetaNodeP] | None
        :type return: list[WdbiMetaNodeP]
        """
        # No value or key, clear entire meta structure and return
        # noinspection PyShadowingNames
        for idx in range(len(items)):
            if items[idx] is not None:
                items[idx] = items[idx].unset()
        return items

    @classmethod
    def _createItems(cls, mtyp, *args, items=None, **kwargs):
        """
        :type mtyp: WdbsTupleMetaTypeP
        :type items: list[WdbiMetaNodeP] | None
        :rtype: list[WdbiMetaNodeP]
        """
        items = items or [None for _ in range(len(mtyp))]
        types = [t for _, t in mtyp]
        """:type: list[WdbsMetaTypeP]"""

        if args or kwargs:
            def setF(v, *keys):
                nonlocal items

                if v is None and not keys:
                    # No value or key, clear entire meta structure and return
                    cls._clearItems(items)
                    return

                # Value and/or key, update meta-structure accordingly
                assert len(keys) == 1
                key = keys[0]
                if items[key] is not None:
                    items[key] = items[key].setFromValue(v)
                else:
                    items[key] = wdbi_createMeta(types[key], v)

            mtyp.setFromValue(setF, *args, **kwargs)
        else:
            cls._clearItems(items)

        # Create all missing items
        for idx in range(len(items)):
            if items[idx] is None:
                items[idx] = wdbi_createMeta(types[idx])
        return items

    def __init__(self, uuid, mtyp, *args, **kwargs):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param mtyp: Meta-data type
        :type mtyp: WdbsTupleMetaTypeP
        """
        items = self._createItems(mtyp, *args, **kwargs)
        WdbiArrayNode.__init__(self, uuid, WDBI_META_TUPLE_CODE, items)
        WdbsTupleMetaMixin.__init__(self, mtyp)

    def _clone(self, **kwargs):
        """
        :rtype: WdbiMetaTupleNode
        """
        # noinspection PyArgumentList
        new = WdbiMetaTupleNode.__new__(WdbiMetaTupleNode)
        WdbsTupleMetaMixin.__init__(new, self.getType())
        self._cloneArrayInto(new, **kwargs)
        return new

    def setFromValue(self, *args, **kwargs):
        """
        :rtype: WdbiMetaTupleNode
        """
        # noinspection PyTypeChecker
        items = self._createItems(
            self.getType(), *args, items=self._items, **kwargs)
        return self._clone(items=items)

    def unset(self):
        """
        :rtype: WdbiMetaTupleNode
        """
        # noinspection PyTypeChecker
        items = WdbiMetaTupleNode._clearItems(self._items)
        return self._clone(items=items)

    def removeMeta(self, key):
        return self.update([key], lambda m: m.unset())

    def serializationIterator(self):
        return iter(self)

    def values(self):
        return (v.asValue() for v in self)

    def __str__(self):
        return str(self.asValue()) + ' @' + str(self.getType())

    def __repr__(self):
        return str(self)


class WdbiMetaRecordNode(WdbiDictNode, WdbiMetaNodeMixin, WdbiStructuralMetaP,
                         WdbsRecordMetaMixin):
    __slots__ = WdbiDictNode.__slots__ + WdbsRecordMetaMixin._SLOTS_

    @classmethod
    def _clearItems(cls, items):
        """
        :type items: dict[object, WdbiMetaNodeP] | None
        :rtype: dict[object, WdbiMetaNodeP]
        """
        for k, i in items.items():
            if i is not None:
                items[k] = i.unset()
        return items

    @classmethod
    def _createItems(cls, mtyp, *args, items=None, **kwargs):
        """
        :type mtyp: WdbsRecordMetaTypeP
        :type items: dict[object, WdbiMetaNodeP] | None
        :rtype: dict[object, WdbiMetaNodeP]
        """
        items = items or OrderedDict((k, None) for k, _ in mtyp)
        """:type: dict[object, WdbiMetaNodeP | None]"""
        types = dict(mtyp)
        """:type: dict[object, WdbsMetaTypeP]"""

        if args or kwargs:
            def setF(v, *keys):
                nonlocal items

                if v is None and not keys:
                    # No value or key, clear entire meta structure and return
                    cls._clearItems(items)
                    return
                assert len(keys) == 1
                key = keys[0]
                if items[key] is not None:
                    items[key] = items[key].setFromValue(v)
                else:
                    items[key] = wdbi_createMeta(types[key], v)

            mtyp.setFromValue(setF, *args, **kwargs)
        else:
            cls._clearItems(items)

        for k, i in items.items():
            if i is None:
                items[k] = wdbi_createMeta(types[k])
        return items

    def __init__(self, uuid, mtyp, *args, **kwargs):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param mtyp: Meta-data type
        :type mtyp: WdbsRecordMetaTypeP
        """
        items = self._createItems(mtyp, *args, **kwargs)
        WdbiDictNode.__init__(self, uuid, WDBI_META_RECORD_CODE, items)
        WdbsRecordMetaMixin.__init__(self, mtyp)

    def _clone(self, **kwargs):
        """
        :rtype: WdbiMetaRecordNode
        """
        # noinspection PyArgumentList
        new = WdbiMetaRecordNode.__new__(WdbiMetaRecordNode)
        WdbsRecordMetaMixin.__init__(new, self.getType())
        self._cloneDictInto(new, **kwargs)
        return new

    @classmethod
    def _createContainer(cls, items=None):
        return items or OrderedDict()

    def setFromValue(self, *args, **kwargs):
        """
        :rtype: WdbiMetaRecordNode
        """
        # noinspection PyTypeChecker
        items = self._createItems(
            self.getType(), *args, items=self._items, **kwargs)
        return self._clone(items=items)

    def unset(self):
        """
        :rtype: WdbiMetaRecordNode
        """
        # noinspection PyTypeChecker
        items = WdbiMetaRecordNode._clearItems(self._items)
        return self._clone(items=items)

    def removeMeta(self, key):
        return self.update([key], lambda m: m.unset())

    def values(self):
        return (v.asValue() for _, v in self)

    def kvPairs(self):
        return ((k, v.asValue()) for k, v in self)

    def serializationIterator(self):
        return iter(self)

    def __str__(self):
        return str(self.asValue()) + ' @' + str(self.getType())

    def __repr__(self):
        return str(self)


class WdbiMetaListNode(WdbiListNode, WdbiMetaNodeMixin, WdbiStructuralMetaP,
                       WdbsListMetaMixin):
    __slots__ = WdbiListNode.__slots__ + WdbsListMetaMixin._SLOTS_

    @classmethod
    def _createItems(cls, mtyp, *args, items=None, **kwargs):
        """
        :type mtyp: WdbsListMetaTypeP
        :type items: list[WdbiMetaNodeP] | None
        :rtype: list[WdbiMetaNodeP]
        """
        items = items or []
        """:type: list[WdbiMetaNodeP]"""

        if args or kwargs:
            def setF(v, *keys):
                nonlocal items

                if v is None and not keys:
                    # No value or key, clear entire meta structure and return
                    items = []
                    return
                assert len(keys) == 1, 'Keys: {}, value: {}'.format(keys, v)
                key = keys[0]
                if key is None:
                    items.append(wdbi_createMeta(mtyp.getValueType(), v))
                else:
                    items[key].setFromValue(v)

            mtyp.setFromValue(setF, *args, **kwargs)
        else:
            items = []

        return items

    def __init__(self, uuid, mtyp, *args, **kwargs):
        """
        :parma uuid: UUID, if known
        :type uuid: UUID | str | None
        :param mtyp: Meta-data type
        :type mtyp: WdbsListMetaTypeP
        """
        items = self._createItems(mtyp, *args, **kwargs)
        WdbiListNode.__init__(self, uuid, WDBI_META_LIST_CODE, items)
        WdbsListMetaMixin.__init__(self, mtyp)

    def _clone(self, **kwargs):
        """
        :rtype: WdbiMetaListNode
        """
        # noinspection PyArgumentList
        new = WdbiMetaListNode.__new__(WdbiMetaListNode)
        WdbsListMetaMixin.__init__(new, self.getType())
        self._cloneListInto(new, **kwargs)
        return new

    def getPath(self, path):
        try:
            return super().getPath(path)
        except (KeyError, IndexError):
            return None

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except (KeyError, IndexError):
            return None

    #def get(self, key):
    #    try:
    #        return super().get(key)
    #    except (KeyError, IndexError):
    #        return None

    def setFromValue(self, *args, **kwargs):
        """
        :rtype: WdbiMetaListNode
        """
        # noinspection PyTypeChecker
        items = self._createItems(self.getType(), *args, **kwargs)
        return self._clone(items=self._createContainer(items))

    def append(self, v):
        """
        Append a meta-data value item to the list.

        :param v: Meta-data to append
        :type v: object
        :return: Updated list with given value appended
        :rtype: WdbiMetaListNode
        """
        # noinspection PyUnresolvedReferences
        return self.add(wdbi_createMeta(self.getType().getValueType(), v))

    def extend(self, vs):
        """
        Extend the list by the given meta-data value items.

        :param vs: Meta-data to append
        :type vs: Iterable[object]
        :return: Updated list with given values appended
        :rtype: WdbiMetaListNode
        """
        # noinspection PyTypeChecker,PyUnresolvedReferences
        return self.addAll(
            (wdbi_createMeta(self.getType().getValueType(), v) for v in vs))

    def unset(self):
        """
        :rtype: WdbiMetaListNode
        """
        return self._clone(items=OrderedDict())

    def removeMeta(self, item):
        item = self[item]
        if item is None:
            raise IndexError('Index out of bounds: {}'.format(item))
        return super().remove(item)

    def values(self):
        return (v.asValue() for v in self)

    def serializationIterator(self):
        return iter(self)

    def update(self, path, updateF):
        if path and path[0] is None:
            # noinspection PyUnresolvedReferences
            value = wdbi_createMeta(self.getType().getValueType())
            path = (value.getUUID(),) + path[1:]
            return self.add(value).update(path, updateF)
        return super().update(path, updateF)

    def __str__(self):
        return str(self.asValue()) + ' @' + str(self.getType())

    def __repr__(self):
        return str(self)


class WdbiMetaDictNode(WdbiDictNode, WdbiMetaNodeMixin, WdbiStructuralMetaP,
                       WdbsDictionaryMetaMixin):
    __slots__ = WdbiDictNode.__slots__ + WdbsDictionaryMetaMixin._SLOTS_

    @classmethod
    def _createItems(cls, mtyp, *args, items=None, **kwargs):
        """
        :type mtyp: WdbsDictionaryMetaTypeP
        :type items: dict[object, WdbiMetaNodeP] | None
        :rtype: dict[object, WdbiMetaNodeP]
        """
        items = items or cls._createContainer()
        """:type: list[WdbiMetaNodeP]"""

        if args or kwargs:
            def setF(v, *keys):
                nonlocal items

                if v is None and not keys:
                    # No value or key, clear entire meta structure and return
                    items = cls._createContainer()
                    return
                assert len(keys) == 1, 'Keys: {}, value: {}'.format(keys, v)
                key = keys[0]
                if key in items:
                    items[keys[0]].setFromValue(v)
                else:
                    items[key] = wdbi_createMeta(mtyp.getValueType(), v)

            mtyp.setFromValue(setF, *args, **kwargs)
        else:
            items = cls._createContainer()

        return items

    def __init__(self, uuid, mtyp, *args, **kwargs):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param mtyp: Meta-data type
        :type mtyp: WdbsDictionaryMetaTypeP
        """
        items = self._createItems(mtyp, *args, **kwargs)
        WdbiDictNode.__init__(self, uuid, WDBI_META_DICT_CODE, items)
        WdbsDictionaryMetaMixin.__init__(self, mtyp)

    def _clone(self, **kwargs):
        """
        :rtype: WdbiMetaDictNode
        """
        # noinspection PyArgumentList
        new = WdbiMetaDictNode.__new__(WdbiMetaDictNode)
        WdbsDictionaryMetaMixin.__init__(new, self.getType())
        self._cloneDictInto(new, **kwargs)
        return new

    def getPath(self, path):
        if path:
            child = self[path[0]]
            if child is None:
                return None
            return child.getPath(path[1:])
        else:
            return self

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except KeyError:
            return None

    def get(self, key):
        try:
            return super().get(key)
        except KeyError:
            return None

    def setFromValue(self, *args, **kwargs):
        """
        :rtype: WdbiMetaDictNode
        """
        # noinspection PyTypeChecker
        items = self._createItems(self.getType(), *args, **kwargs)
        return self._clone(items=items)

    def unset(self):
        """
        :rtype: WdbiMetaDictNode
        """
        # noinspection PyTypeChecker
        return self._clone(items=self._createContainer())

    def removeMeta(self, key):
        ndct = self._items
        if key in ndct:
            del ndct[key]
        return self._clone(items=ndct)

    def kvPairs(self):
        return ((k, v.asValue()) for k, v in self)

    def values(self):
        return (v.asValue() for _, v in self)

    def serializationIterator(self):
        # noinspection PyUnresolvedReferences
        ktyp = self.getType().getKeyType()
        return ((wdbi_createMeta(ktyp, k), v) for k, v in self)

    def update(self, path, updateF):
        if path and path[0] not in self:
            # noinspection PyUnresolvedReferences
            value = wdbi_createMeta(self.getType().getValueType())
            return self.add(path[0], value).update(path, updateF)
        return super().update(path, updateF)

    def __str__(self):
        return str(self.asValue()) + ' @' + str(self.getType())

    def __repr__(self):
        return str(self)


class WdbiMetaOrderedDictNode(WdbiMetaDictNode):
    __slots__ = WdbiDictNode.__slots__

    def __init__(self, uuid, mtyp, *args, **kwargs):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param mtyp: Meta-data type
        :type mtyp: WdbsOrderedDictionaryMetaTypeP
        """
        super().__init__(uuid, mtyp, *args, **kwargs)

    @classmethod
    def _createContainer(cls, items=None):
        return items or OrderedDict()

    def _clone(self, **kwargs):
        """
        :rtype: WdbiMetaOrderedDictNode
        """
        # noinspection PyArgumentList
        new = WdbiMetaOrderedDictNode.__new__(WdbiMetaOrderedDictNode)
        WdbsDictionaryMetaMixin.__init__(new, self.getType())
        self._cloneDictInto(new, **kwargs)
        return new


class WdbiMetaAtomicNode(WdbiLeafNode, WdbiMetaNodeMixin, WdbsAtomicMetaMixin):
    __slots__ = WdbiLeafNode.__slots__ + WdbsAtomicMetaMixin._SLOTS_ + \
        ('_WdbiMetaAtomicNode__value',)

    @classmethod
    def _createValue(cls, mtyp, *args):
        """
        :type mtyp: WdbsAtomicMetaTypeP
        :rtype: object
        """
        def setF(v, *keys):
            nonlocal value
            assert not keys
            value = v if v is not None else mtyp.getEmptyValue()
        value = None
        mtyp.setFromValue(setF, *args)
        return value

    def __init__(self, uuid, mtyp, *args):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param mtyp: Meta-data type
        :type mtyp: WdbsAtomicMetaTypeP
        """
        WdbiLeafNode.__init__(self, uuid, WDBI_META_ATOMIC_CODE)
        WdbsAtomicMetaMixin.__init__(self, mtyp)

        if args:
            # noinspection PyNoneFunctionAssignment
            self.__value = WdbiMetaAtomicNode._createValue(mtyp, *args)
        else:
            self.__value = mtyp.getEmptyValue()

    def _clone(self, **kwargs):
        """
        :rtype: WdbiMetaAtomicNode
        """
        # noinspection PyArgumentList
        new = WdbiMetaAtomicNode.__new__(WdbiMetaAtomicNode)
        WdbsAtomicMetaMixin.__init__(new, self.getType())
        self._cloneLeafInto(new, **kwargs)
        if 'value' in kwargs:
            new.__value = kwargs['value']
        else:
            new.__value = self.__value
        return new

    def getValue(self):
        return self.__value

    def setFromValue(self, *args, **kwargs):
        """
        :rtype: WdbiMetaAtomicNode
        """
        # noinspection PyTypeChecker
        return self._clone(
            value=self._createValue(self.getType(), *args))

    def unset(self):
        """
        :rtype: WdbiMetaAtomicNode
        """
        return self._clone(value=self.getType().getEmptyValue())

    def __str__(self):
        return str(self.asValue()) + ' @' + str(self.getType())

    def __repr__(self):
        return str(self)


class WdbiMetaUnionNode(WdbiNodeBase, WdbiMetaNodeMixin, WdbsUnionMetaMixin):

    __slots__ = WdbiNodeBase.__slots__ + WdbsUnionMetaMixin._SLOTS_ + \
        ('_WdbiMetaUnionNode__type', '_WdbiMetaUnionNode__value')

    @classmethod
    def _createItems(cls, mtyp, *args, **kwargs):
        """
        :type mtyp: WdbsUnionMetaTypeP
        :rtype: (str | None, WdbiMetaNodeP | None)
        """
        def setF(val, *keys):
            nonlocal key, value
            assert len(keys) <= 1
            key = keys[0] if keys else mtyp.getDefaultTypeId()
            if key:
                value = wdbi_createMeta(mtyp.getType(key), val)

        key = None
        value = None

        mtyp.setFromValue(setF, *args, **kwargs)

        return key, value

    def __init__(self, uuid, mtyp, *args, **kwargs):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param mtyp: Meta-data type
        :type mtyp: WdbsUnionMetaTypeP
        """
        WdbiNodeBase.__init__(self, uuid, WDBI_META_UNION_CODE)
        WdbsUnionMetaMixin.__init__(self, mtyp)

        ntyp, nval = self._createItems(mtyp, *args, **kwargs)
        self.__type = ntyp
        """:type: str | None"""
        self.__value = nval
        """:type: WdbiMetaNodeP | None"""

    def getPath(self, path):
        if path:
            if self.__type is None:
                raise KeyError('No type set')
            if path[0] == self.__type:
                return self.__value.getPath(path[1:])
            else:
                raise KeyError(
                    'Invalid type, currently set to {}'.format(self.__type))
        else:
            return self

    def _clone(self, **kwargs):
        """
        :rtype: WdbiMetaUnionNode
        """
        # noinspection PyArgumentList
        new = WdbiMetaUnionNode.__new__(WdbiMetaUnionNode)
        WdbsUnionMetaMixin.__init__(new, self.getType())
        self._cloneNodeInto(new)
        ntyp = kwargs.get('type')
        if ntyp and ntyp != self.__type:
            new.__type = ntyp
            new.__value = kwargs['value'] if 'value' in kwargs else None
        else:
            new.__type = self.__type
            nval = kwargs['value'] if 'value' in kwargs else self.__value
            new.__value = nval
        return new

    def _setType(self, ntyp):
        mtyp = self.getType()
        """:type: WdbsUnionMetaTypeP"""
        value = wdbi_createMeta(mtyp.getType(ntyp))
        return self._clone(type=ntyp, value=value)

    def update(self, path, updateF):
        if path:
            ntyp, npth = path
            if ntyp != self.__type:
                mtyp = self.getType()
                """:type: WdbsUnionMetaTypeP"""
                value = wdbi_createMeta(mtyp.getType(ntyp))
                value = value.update(npth, updateF)
                return self._clone(type=ntyp, value=value)
            else:
                value = self.__value.update(npth, updateF)
                return self._clone(value=value)
        else:
            raise Exception('Need type identifier')

    def values(self):
        if self.__type is None:
            return ()
        return self.__value.asValue(),

    def unset(self):
        """
        :rtype: WdbiMetaUnionNode
        """
        mtyp = self.getType()
        """:type: WdbsUnionMetaTypeP"""
        ntyp = mtyp.getDefaultTypeId()
        value = wdbi_createMeta(mtyp.getType(ntyp))
        return self._clone(type=ntyp, value=value)

    def getValueType(self):
        mtyp = self.getType()
        """:type: WdbsUnionMetaTypeP"""
        return mtyp.getType(self.__type) if self.__type is not None else None

    def kvPairs(self):
        if self.__type is None:
            return ()
        return [ScmnMTuple(self.__type, self.__value.asValue())]

    def getValueTypeIdentifier(self):
        return self.__type

    def __len__(self):
        return 1 if self.__type is not None else 0

    def setFromValue(self, *args, **kwargs):
        """
        :rtype: WdbiMetaUnionNode
        """
        # noinspection PyTypeChecker
        ntyp, value = self._createItems(self.getType(), *args, **kwargs)
        return self._clone(type=ntyp, value=value)

    def keys(self):
        if self.__type is None:
            return ()
        return self.__type,

    def serializationIterator(self):
        if self.__type is None:
            return []
        return [(self.__type, self.__value)]

    def value(self):
        return self.__value.asValue() if self.__value is not None else None


def wdbi_createMeta(mType, *args, **kwargs):
    """
    Create a new meta-data item for the given type.

    Args and kwargs are used for initial values.

    :param mType: Meta-data type
    :type mType: WdbsMetaTypeP
    :param args: Initial elements
    :param kwargs: Initial elements
    :return: Newly created meta-data item
    :rtype: WdbiNodeP | WdbsMetaP | WdbmMetaMixin
    """
    if isinstance(mType, WdbsAtomicMetaTypeP):
        return WdbiMetaAtomicNode(None, mType, *args)
    elif isinstance(mType, WdbsTupleMetaTypeP):
        return WdbiMetaTupleNode(None, mType, *args)
    elif isinstance(mType, WdbsListMetaTypeP):
        return WdbiMetaListNode(None, mType, *args)
    elif isinstance(mType, WdbsRecordMetaTypeP):
        return WdbiMetaRecordNode(None, mType, *args, **kwargs)
    elif isinstance(mType, WdbsOrderedDictionaryMetaTypeP):
        return WdbiMetaOrderedDictNode(None, mType, *args, **kwargs)
    elif isinstance(mType, WdbsDictionaryMetaTypeP):
        return WdbiMetaDictNode(None, mType, *args, **kwargs)
    elif isinstance(mType, WdbsUnionMetaTypeP):
        return WdbiMetaUnionNode(None, mType, *args, **kwargs)
    else:
        raise ValueError('Unknown meta-data type: {}'.format(mType))
