# ----------------------------------------------------------------------------
# - wdbm-block                                                               -
# - wild cruncher:msh block object                  core / dbstore / mutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbm_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher mutable block objects

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from wcr_core.dbstore.mutable.wdbm_connection import WdbmConnection
from wcr_core.dbstore.mutable.wdbm_button import WdbmButton
from wcr_core.dbstore.mutable.wdbm_entity import WdbmNamedEntityMixin, \
    WdbmTypedEntityBase
from wcr_core.dbstore.mutable.wdbm_instance import WdbmOperatorInstance, \
    WdbmModelInstance, WdbmBlockInstance
from wcr_core.dbstore.mutable.wdbm_shape import WdbmShape
from wcr_core.dbstore_api.wdbs_block_api import WdbsBlockP
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_CHILD_ADDED_EVENT, \
    WDBS_CHILD_REMOVED_EVENT


class WdbmBlock(WdbmTypedEntityBase, WdbmNamedEntityMixin, WdbsBlockP):
    """
    The WmhdbModelBlock class is the silkan hierarchy block. It is used
    to construct one hierarchical level by grouping instances, functions and
    wire into several list. A block is a model and therefore is defined by name
    with an optional information string.
    """

    # noinspection PyUnresolvedReferences
    def __init__(self, dctx, btyp, name=None, info=None, uuid=None,
                 meta=None, **kwargs):
        """
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param btyp: Block type
        :type btyp: str
        :param name: the block name
        :type name: str
        :param info: the block info
        :type info: str
        :param uuid: UUID
        :type uuid: UUID | str | None
        :param meta: Meta-data description
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        """
        WdbmTypedEntityBase.__init__(self, btyp, dctx, uuid, meta, **kwargs)
        WdbmNamedEntityMixin.__init__(self, name, info)

        #: the instance list
        self.__instances = []
        """:type: list[WdbmInstance]"""
        #: the connection list
        self.__connections = []
        """:type: list[WdbmConnection]"""
        #: the shape list
        self.__shapes = []
        """:type: list[WdbmShape]"""
        #: buttons list
        self.__buttons = []
        """:type: list[WdbmButton]"""

    def doDelete(self, recurse=True):
        if recurse:
            for inst in self.__instances:
                inst.doDelete(recurse=False)
            for conn in self.__connections:
                conn.doDelete(recurse=False)
            for shape in self.__shapes:
                shape.doDelete(recurse=False)
            for bttn in self.__buttons:
                bttn.doDelete(recurse=False)
        super().doDelete()

    def getInstances(self):
        return iter(self.__instances)

    def getInstanceCount(self):
        return len(self.__instances)

    def createOperatorInstance(self, *args, **kwargs):
        inst = WdbmOperatorInstance(self.context, *args, **kwargs)
        self.__instances.append(inst)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, inst)
        return inst

    def createBlockInstance(self, *args, **kwargs):
        inst = WdbmBlockInstance(self.context, *args, **kwargs)
        self.__instances.append(inst)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, inst)
        return inst

    def createModelInstance(self, *args, **kwargs):
        inst = WdbmModelInstance(self.context, *args, **kwargs)
        self.__instances.append(inst)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, inst)
        return inst

    def removeInstance(self, inst):
        self.__instances.remove(inst)
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, inst)
        inst.doDelete()

    def getConnections(self):
        return iter(self.__connections)

    def getConnectionCount(self):
        return len(self.__connections)

    def createConnection(self, sobj, sxpt, eobj, expt,
                         name=None, info=None, path=None, uuid=None, **meta):
        conn = WdbmConnection(self.context, sobj, sxpt, eobj, expt,
                              name, info, path, uuid, **meta)
        self.__connections.append(conn)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, conn)
        return conn

    def removeConnection(self, connection):
        self.__connections.remove(connection)
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, connection)
        connection.doDelete()

    def getShapes(self):
        return iter(self.__shapes)

    def getShapeCount(self):
        return len(self.__shapes)

    def createShape(self, *args, **kwargs):
        shape = WdbmShape(self.context, *args, **kwargs)
        self.__shapes.append(shape)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, shape)
        return shape

    def removeShape(self, shape):
        self.__shapes.remove(shape)
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, shape)
        shape.doDelete()

    def getButtons(self):
        return iter(self.__buttons)

    def getButtonCount(self):
        return len(self.__buttons)

    def createButton(self, *args, **kwargs):
        button = WdbmButton(self.context, *args, **kwargs)
        self.__buttons.append(button)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, button)
        return button

    def removeButton(self, button):
        self.__buttons.remove(button)
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, button)
        button.doDelete()
