# ----------------------------------------------------------------------------
# - wdbm-protocols                                                           -
# - wild cruncher:general protocol definitions      core / dbstore / mutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbm_protocols
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher mutable hierarchy model protocol definitions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.mutable.wdbm_context import WdbmContext
from wcr_core.dbstore.wdbs_entity import WdbsObservableMixinP, WdbsEntityBase
from wcr_core.dbstore.wdbs_util import wdbs_makeUUID
from wcr_core.dbstore_api.wdbs_meta_description_api import WdbsMetaDescriptionP
from wcr_core.dbstore_api.wdbs_entity_api import WdbsNamedEntityP, \
    WdbsTypedEntityP
from wcr_core.dbstore.mutable.wdbm_meta_data import *
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_UPDATE_CORE_EVENT, \
    WDBS_UPDATE_META_EVENT


# noinspection PyAbstractClass
class WdbmEntityBase(WdbsEntityBase):
    def __init__(self, dctx, uuid=None, meta=None, **kwargs):
        """
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param uuid: UUID
        :type uuid: UUID | str | None
        :param meta: Meta-data description
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Meta-data values
        """
        assert isinstance(dctx, WdbmContext)

        super().__init__(dctx)

        self.__metaDesc = meta or dctx.getMetaDescription(self.__class__)

        self.__metaData = {
            m.name: self._createMeta(m.metaType)
            for m in self.__metaDesc.fields} if self.__metaDesc else {}
        """:type: dict[str, WdbmMetaBase]"""
        for k in self.__metaData.keys():
            if k in kwargs:
                self.__metaData[k].setFromValue(kwargs[k])

        self.__uuid = wdbs_makeUUID(uuid)

    @property
    def metaDescription(self):
        return self.__metaDesc

    @property
    def metaData(self):
        return self.__metaData

    def getUUID(self):
        return self.__uuid

    def setUUID(self, uuid):
        self.__uuid = uuid

    def _createMeta(self, metaT, *args, **kwargs):
        # Reference test
        return wdbm_createMeta(metaT, *args, **kwargs)

    def _getMeta(self, *keys, create=False):
        """
        Get the meta-data item at the given path.

        Create missing items if requested.

        :param keys: Meta-data path
        :type keys: list[object]
        :param create: Whether to create missing item along the path
        :type create: bool
        :return: Meta-data item at the given path, or `None` if not found
        :rtype: WdbmMetaP | None
        """
        # noinspection PyTypeChecker


        meta = self.__metaData[keys[0]] if keys else self.__metaData
        """:type: WdbsStructuralMetaP | dict[object, WdbsMetaP]"""
        for key in keys[1:]:
            old = meta
            meta = meta.getItem(key, create)
            if meta is None:
                if not isinstance(old, WdbsDictionaryMetaP) and \
                        not isinstance(old, WdbsListMetaP):
                    raise ValueError(
                        'Could not find meta-data item {}'.format(keys))
                break
        return meta

    def getMetaCount(self, *keys):
        meta = self._getMeta(*keys)
        if isinstance(meta, WdbsStructuralMetaP):
            return len(meta)
        else:
            return 0

    def getMetaKeys(self, *keys):
        if keys:
            meta = self._getMeta(*keys)
            if isinstance(meta, WdbsStructuralMetaP):
                return meta.keys()
            else:
                return []
        else:
            return self.__metaData.keys()

    def getMetas(self, *keys):
        if keys:
            meta = self._getMeta(*keys)
            if isinstance(meta, WdbsStructuralMetaP):
                return meta.kvPairs()
            elif meta is not None:
                return [(None, meta)]
            else:
                return []
        else:
            return self.__metaData.items()

    def getMeta(self, *keys):
        meta = self._getMeta(*keys)
        return meta.asValue() if meta is not None else None

    def setMeta(self, value, *keys):
        if not keys:
            return

        meta = self._getMeta(*keys, create=True)
        if meta is not None:
            meta.setFromValue(value)
            self._notifyObservers(WDBS_UPDATE_META_EVENT, value, *keys)
        else:
            raise ValueError(
                'Could not find meta-data item {}'.format(keys))

    def removeMeta(self, *keys):
        if not keys:
            for meta in self.__metaData.values():
                meta.unset()
                self._notifyObservers(WDBS_UPDATE_META_EVENT, None)
        elif len(keys) == 1:
            meta = self.__metaData.get(keys[0])
            """:type: WdbmMetaBase | None"""
            if meta is not None:
                meta.unset()
                self._notifyObservers(WDBS_UPDATE_META_EVENT, None, *keys)
        else:
            meta = self._getMeta(*keys[:-1])
            if isinstance(meta, WdbmStructuralMetaP):
                meta.remove(keys[-1])
                self._notifyObservers(WDBS_UPDATE_META_EVENT, None, *keys)

    def getAllMeta(self):
        return {k: v.asValue() for k, v in self.__metaData.items()
                if v is not None}

    def __str__(self):
        return self.__class__.__name__ + ' ' + str(self.__metaData)


# noinspection PyAbstractClass
class WdbmTypedEntityBase(WdbmEntityBase, WdbsTypedEntityP):
    __slots__ = ()

    slots = ('__type',)

    # noinspection PyUnresolvedReferences
    def __init__(self, etyp, dctx, uuid=None, meta=None, **kwargs):
        """
        :param etyp: Entity type identifier
        :type etyp: str
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param uuid: UUID
        :type uuid: UUID | str | None
        :param meta: Meta-data description
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Meta-data values
        """
        assert etyp

        if not meta:
            meta = dctx.getMetaDescription(self.__class__, etyp)

        super().__init__(dctx, uuid, meta, **kwargs)

        self.__type = etyp

    def getEntityType(self):
        return self.__type


# noinspection PyAbstractClass
class WdbmNamedEntityMixin(WdbsNamedEntityP, WdbsObservableMixinP):
    __slots__ = ()

    slots = ('__name', '__info')

    # noinspection PyUnresolvedReferences
    def __init__(self, name=None, info=None, _id=None):
        self.__name = name if name is not None else ''
        self.__info = info if info is not None else ''
        self.__id = _id

    def getName(self):
        return self.__name

    def getInfo(self):
        return self.__info

    def getId(self):
        return self.__id

    def setName(self, name):
        # noinspection PyUnresolvedReferences
        self.__name = name
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def setInfo(self, info):
        # noinspection PyUnresolvedReferences
        self.__info = info
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT,)

    def setId(self, info):
        # noinspection PyUnresolvedReferences
        self.__id = info
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT,)
