# ----------------------------------------------------------------------------
# - wdbs-meta-data                                                           -
# - wild cruncher:mutable meta-data instances       core / dbstore / mutable -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs_meta_data
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher mutable meta-data instances

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import OrderedDict

from silkan_common.scmn_tuple import ScmnMTuple
from wcr_core.dbstore.wdbs_meta_data import *
from wcr_core.dbstore_api.wdbs_meta_data_api import *
from wcr_core.dbstore_api.wdbs_meta_type_api import *


# noinspection PyAbstractClass
class WdbmMetaP(WdbsMetaP):

    __slots__ = ()

    @abstractmethod
    def createMeta(self, metaT, *args, **kwargs):
        """
        Create a new meta-data instance for the given type.

        :param metaT: Type of the new meta-data instance
        :type metaT: WdbsMetaTypeP
        :param args: Init-args for the created meta-data instance
        :param kwargs: Init-mwargs for the created meta-data instance
        :return: Newly created meta-data instance
        :rtype: WdbmMetaP
        """
        raise NotImplementedError()

    @abstractmethod
    def setFromValue(self, *args, **kwargs):
        """
        Set the value of the meta-data instance from the given raw value(s).

        The use of the args/kwargs-arguments depends on the actual
        implementation.

        :param create: Whether to create missing meta items
        :type create: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def unset(self):
        """
        Unset the mata-data instance's value.
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbmMetaMixin(WdbmMetaP):
    """
    Abstract base-class for HSM meta-class instances.
    """

    __slots__ = ()

    def createMeta(self, metaT, *args, **kwargs):
        """
        :type metaT: WdbsMetaTypeP
        :rtype: WdbmMetaP
        """
        return wdbm_createMeta(metaT, *args, **kwargs)

    @abstractmethod
    def _setValue(self, v=None, *keys):
        """
        Call-back function for meta-types to set the entities value.

        :param v: Value to set
        :type v: object | None
        :param keys: Keys of sub-entity to set, if applicable
        :type keys: list[object]
        """
        raise NotImplementedError()

    def setFromValue(self, *args, **kwargs):
        self.getType().setFromValue(
            self._setValue, *args, **kwargs)


class WdbmAtomicMeta(WdbsAtomicMetaMixin, WdbmMetaMixin):
    """
    Atomic HSM meta-class instances.
    """

    __slots__ = ('__value',) + WdbsAtomicMetaMixin._SLOTS_

    def __init__(self, vType, value=None):
        """
        :param vType: Meta-data type
        :type vType: WdbsAtomicMetaTypeP
        :param value: Initial value
        :type value: object
        """
        WdbsAtomicMetaMixin.__init__(self, vType)

        #: Value
        self.__value = value or None

    def getValue(self):
        value = self.__value
        # noinspection PyUnresolvedReferences
        return value if value is not None else self.getType().defaultValue

    def _setValue(self, v=None, *keys):
        self.__value = v

    def unset(self):
        self.__value = None


# noinspection PyAbstractClass
class WdbmStructuralMetaP(WdbmMetaP, WdbsStructuralMetaP):
    """
    Protocol for structural DBM meta-entities.
    """

    __slots__ = ()

    @abstractmethod
    def getItem(self, key, create=False):
        """
        Get the child item for the given key.

        Create missing child items if requested.

        :param key: Child item key
        :type key: object
        :param create: Whether to create missing child items
        :type create: bool
        :return: Child item for the given key
        :rtype: WdbsMetaMP
        """
        raise NotImplementedError()

    @abstractmethod
    def remove(self, key):
        """
        Remove the child item with the given key.

        :param key: Key of child item to remove
        :type key: object
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbmStructuralMetaMixin(WdbmMetaMixin, WdbmStructuralMetaP):
    """
    Mixin for structural DBM meta-entities.
    """

    __slots__ = ()

    def _setValue(self, v=None, *keys):
        if v is None and not keys:
            self.unset()
            return
        meta = self
        for key in keys:
            meta = meta.getItem(key, create=True)
            if meta is None:
                raise KeyError(
                    'Could not find key {} in {}'.format(key, self))
        meta.setFromValue(v)


# noinspection PyAbstractClass
class WdbmMultiMetaP(WdbmMetaP, WdbsMultiMetaP):
    """
    Protocol for multi-element DBM meta-entities.
    """

    __slots__ = ()

    @abstractmethod
    def clear(self):
        """
        Clear the instance, i.e. remove all children.
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbmMultiMetaMixin(WdbmStructuralMetaMixin, WdbmMultiMetaP):
    """
    Mixin for multi-element DBM meta-entities.
    """

    __slots__ = ()


class WdbmUnionMeta(WdbsUnionMetaMixin, WdbmStructuralMetaMixin):
    """
    Union meta-data type.
    """

    __slots__ = ('__typeId', '__value') + WdbsUnionMetaMixin._SLOTS_

    def __init__(self, uType, value=None):
        """
        :param uType: Union meta-data type
        :type uType: WdbsUnionMetaTypeP
        :param value: Initial value and type identifier
        :type value: (str, object)
        """
        WdbsUnionMetaMixin.__init__(self, uType)

        #: Type identifier
        self.__typeId = None
        """:type: str | None"""
        #: Value
        self.__value = None
        """:type: WdbmMetaMixin | None"""

        self.setFromValue(value or (uType.getDefaultTypeId(), None))

    def _getValueType(self, typeId):
        # noinspection PyUnresolvedReferences
        return self.getType().getType(typeId)

    def __len__(self):
        return 1 if self.__typeId else 0

    def getValueTypeIdentifier(self):
        return self.__typeId or None

    def getValueType(self):
        return self._getValueType(self.__typeId)

    def _setValueType(self, typeId):
        if typeId:
            self.__typeId = typeId
            self.__value = wdbm_createMeta(self._getValueType(typeId))
        return self.__value

    def value(self):
        value = self.__value
        return value.asValue() if value is not None else None

    def getItem(self, key, create=False):
        if key != self.__typeId:
            if create or not self.__typeId:
                self._setValueType(key)
            else:
                raise ValueError(
                    'Incorrect key: {} (current: {})'.format(
                        key, self.__typeId))
        return self.__value

    def remove(self, key):
        assert key == self.__typeId
        self.unset()

    def unset(self):
        self.__typeId = None
        self.__value = None

    def keys(self):
        typeId = self.__typeId
        return [typeId] if typeId else []

    def values(self):
        value = self.__value
        return [value.asValue()] if value else []

    def kvPairs(self):
        typeId = self.__typeId
        value = self.__value
        if typeId:
            return [ScmnMTuple(typeId, value.asValue() if value else None)]
        else:
            return []

    def serializationIterator(self):
        typeId = self.__typeId
        return [(typeId, self.__value)] if typeId else []


class WdbmListMeta(WdbsListMetaMixin, WdbmMultiMetaMixin):
    """
    List HSM meta-class instances.
    """

    __slots__ = ('__vtyp', '__elements') + WdbsListMetaMixin._SLOTS_

    def __init__(self, lType, *elements):
        """
        :param lType: List meta-data type
        :type lType: WdbsListMetaTypeP
        :param elements: Initial elements
        :type elements: list[object]
        """
        WdbsListMetaMixin.__init__(self, lType)

        #: Value type
        self.__vtyp = lType.getValueType()
        #: List elements
        self.__elements = []

        if elements:
            self.setFromValue(*elements)

    def __len__(self):
        return len(self.__elements)

    def getItem(self, key, create=False):
        if key is None:
            value = self.createMeta(self.__vtyp)
            self.__elements.append(value)
            return value

        key = int(key)
        try:
            return self.__elements[key]
        except IndexError:
            if key > 0:
                self.__elements.extend(
                    self.createMeta(self.__vtyp)
                    for _ in range(len(self.__elements), key + 1))
                return self.__elements[key]
            else:
                return None

    def append(self, item):
        self.__elements.append(self.createMeta(self.__vtyp, item))

    def extend(self, items):
        self.__elements.extend(
            self.createMeta(self.__vtyp, item) for item in items)

    def remove(self, k):
        del self.__elements[int(k)]

    def clear(self):
        self.__elements = []

    def unset(self):
        self.clear()

    def values(self):
        return [v.asValue() for v in self.__elements]

    def serializationIterator(self):
        return iter(self.__elements)


class WdbmDictionaryMeta(WdbsDictionaryMetaMixin, WdbmMultiMetaMixin):
    """
    Dictionary/map HSM meta-class instances.
    """

    __slots__ = ('__elements',) + WdbsDictionaryMetaMixin._SLOTS_

    def __init__(self, dType, *args, **kwargs):
        """
        :param dType: Dictionary meta-data type
        :type dType: WdbsDictionaryMetaTypeP
        :param args: Initial elements
        :param kwargs: Initial elements
        """
        WdbsDictionaryMetaMixin.__init__(self, dType)

        #: Dictionary elements
        self.__elements = self._createEmptyDict()
        if args:
            self.setFromValue(*args, **kwargs)

    def __len__(self):
        return len(self.__elements)

    def _createEmptyDict(self):
        """
        :return: Empty dictionary
        :rtype: dict[object, WdbmMetaP]
        """
        return {}

    def getItem(self, key, create=False):
        item = self.__elements.get(key)

        if item is None and create:
            item = self.createMeta(self.getValueType())
            self.__elements[key] = item

        return item

    def remove(self, k):
        if k in self.__elements:
            del self.__elements[k]

    def clear(self):
        self.__elements = self._createEmptyDict()

    def unset(self):
        self.clear()

    def keys(self):
        return self.__elements.keys()

    def values(self):
        return (v.asValue() for v in self.__elements.values() if v is not None)

    def kvPairs(self):
        return ((k, v.asValue()) for k, v
                in self.__elements.items() if v is not None)

    def serializationIterator(self):
        kType = self.getKeyType()
        return ((self.createMeta(kType, k), v) for k, v
                in self.__elements.items() if v is not None)


class WdbmOrderedDictionaryMeta(WdbmDictionaryMeta):
    """
    Dictionary/map HSM meta-class instances.
    """

    __slots__ = WdbmDictionaryMeta.__slots__

    def __init__(self, dType, *args, **kwargs):
        """
        :param dType: Dictionary meta-data type
        :type dType: WdbsOrderedDictionaryMetaTypeP
        :param args: Initial elements
        :param kwargs: Initial elements
        """
        assert isinstance(dType, WdbsOrderedDictionaryMetaTypeP)
        super().__init__(dType, *args, **kwargs)

    def _createEmptyDict(self):
        return OrderedDict()


class WdbmTupleMeta(WdbsTupleMetaMixin, WdbmStructuralMetaMixin):
    """
    Tuple HSM meta-class instances.
    """

    __slots__ = ('__elements',) + WdbsTupleMetaMixin._SLOTS_

    def __init__(self, tType, *elements):
        """
        :param tType: Tuple meta-data type
        :type tType: WdbsTupleMetaTypeP
        :param elements: Initial elements
        :type elements: list[object]
        """
        WdbsTupleMetaMixin.__init__(self, tType)

        #: Tuple elements
        self.__elements = [self.createMeta(t) for (_, t) in tType]
        if elements:
            self.setFromValue(*elements)

    def __len__(self):
        return len(self.__elements)

    def getItem(self, key, create=False):
        return self.__elements[int(key)]

    def remove(self, key):
        self.__elements[int(key)].unset()

    def unset(self):
        for item in self.__elements:
            item.unset()

    def values(self):
        return (v.asValue() for v in self.__elements)

    def serializationIterator(self):
        return iter(self.__elements)


class WdbmRecordMeta(WdbsRecordMetaMixin, WdbmStructuralMetaMixin):
    """
    Record HSM meta-class instances.
    """

    __slots__ = ('__elements',) + WdbsRecordMetaMixin._SLOTS_

    def __init__(self, rType, *args, **kwargs):
        """
        :param rType: Record meta-data type
        :type rType: WdbsRecordMetaTypeP
        :param args: Initial elements
        :param kwargs: Initial elements
        """
        WdbsRecordMetaMixin.__init__(self, rType)

        #: Elements
        self.__elements = {k: self.createMeta(t) for k, t in rType}
        """:type: dict[object, WdbmMetaMixin]"""
        if args:
            self.setFromValue(*args, **kwargs)

    def __len__(self):
        return len(self.__elements)

    def getItem(self, key, create=False):
        return self.__elements[key]

    def remove(self, key):
        self.__elements[key].unset()

    def unset(self):
        for item in self.__elements.values():
            item.unset()

    def keys(self):
        return self.__elements.keys()

    def values(self):
        return (v.asValue() for v in self.__elements.values())

    def kvPairs(self):
        return ((k, v.asValue()) for k, v in self.__elements.items())

    def serializationIterator(self):
        return self.__elements.items()


def wdbm_createMeta(mType, *args, **kwargs):
    """
    Create a new meta-data item for the given type.

    Args and kwargs are used for initial values.

    :param mType: Meta-data type
    :type mType: WdbsMetaTypeP
    :param args: Initial elements
    :param kwargs: Initial elements
    :return: Newly created meta-data item
    :rtype: WdbmMetaP
    """
    if isinstance(mType, WdbsAtomicMetaTypeP):
        return WdbmAtomicMeta(mType, args[0] if args else None)
    elif isinstance(mType, WdbsTupleMetaTypeP):
        return WdbmTupleMeta(mType, *args)
    elif isinstance(mType, WdbsListMetaTypeP):
        return WdbmListMeta(mType, *args)
    elif isinstance(mType, WdbsRecordMetaTypeP):
        return WdbmRecordMeta(mType, *args, **kwargs)
    elif isinstance(mType, WdbsOrderedDictionaryMetaTypeP):
        return WdbmOrderedDictionaryMeta(mType, *args, **kwargs)
    elif isinstance(mType, WdbsDictionaryMetaTypeP):
        return WdbmDictionaryMeta(mType, *args, **kwargs)
    elif isinstance(mType, WdbsUnionMetaTypeP):
        return WdbmUnionMeta(mType, *args, **kwargs)
    else:
        raise ValueError('Unknown meta-data type: {}'.format(mType))
