# ----------------------------------------------------------------------------
# - wdbm-parameter                                                           -
# - wild cruncher:msh parameter object              core / dbstore / mutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbm_parameter
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher mutable parameter object

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.mutable.wdbm_entity import WdbmEntityBase
from wcr_core.dbstore.wdbs_entity import WdbsObservableMixinP
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_UPDATE_CORE_EVENT, \
    WDBS_CHILD_ADDED_EVENT, WDBS_CHILD_REMOVED_EVENT, WdbsObserverP, \
    WDBS_UPDATE_PARAM_EVENT
from wcr_core.dbstore_api.wdbs_parameter_api import WdbsParameterP, \
    WdbsEntityParametersP


class WdbmParameter(WdbmEntityBase, WdbsParameterP):
    def __init__(self, dctx, name, value=None, uuid=None, meta=None,
                 **kwargs):
        """
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param name: Name
        :type name: str
        :param value: Value
        :type value: object | None
        :param uuid: UUID
        :type uuid: UUID | str | None
        :param meta: Meta-data description
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Meta-data values
        """
        WdbmEntityBase.__init__(self, dctx, uuid=uuid, meta=meta, **kwargs)

        self.__name = name
        self.__value = value

    def getName(self):
        return self.__name

    def getValue(self):
        return self.__value

    def setValue(self, value):
        self.__value = value
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT, value)


# noinspection PyAbstractClass
class WdbmHasParametersMixin(WdbsEntityParametersP, WdbsObservableMixinP,
                             WdbsObserverP):
    def __init__(self, dctx, params=None):
        """
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param params: Parameter name-value pairs
        :type params: [(str, object)] | None
        """
        self.__dctx = dctx
        self.__parameters = {}
        """:type: dict[str, WdbmParameter]"""

        if params:
            for n, v in params:
                param = WdbmParameter(dctx, n, v)
                param.registerObserver(self)
                self.__parameters[n] = param

    def doDelete(self):
        for prm in self.__parameters.values():
            prm.doDelete()

    def getParameterCount(self):
        return len(self.__parameters)

    def getParameters(self):
        return self.__parameters.values()

    def getParameter(self, name):
        return self.__parameters.get(name)

    def getOrCreateParameter(self, name):
        param = self.__parameters.get(name)
        if param is None:
            param = WdbmParameter(self.__dctx, name)
            param.registerObserver(self)
            self.__parameters[name] = param
            self._notifyObservers(WDBS_CHILD_ADDED_EVENT, param)
        return param

    def removeParameter(self, name):
        param = self.__parameters.get(name)
        if param is not None:
            param.unregisterObserver(self)
            del self.__parameters[name]
            self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, param)

    def clearParameters(self):
        params = list(self.__parameters.values())
        for param in params:
            param.unregisterObserver(self)
        self.__parameters.clear()
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, *params)

    def dbNotify(self, obj, event, *args, **kwargs):
        if isinstance(obj, WdbmParameter):
            self._notifyObservers(
                WDBS_UPDATE_PARAM_EVENT, obj.getValue(), obj.getName())
