# ----------------------------------------------------------------------------
# - wdbs-cloning                                                             -
# - wild cruncher:rsh cloning functions                       core / dbstore -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs_cloning
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher raw hierarchy cloning functions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.wdbs_util import wdbs_mkConnectorList
from wcr_core.dbstore_api.wdbs_connection_api import WdbsConnectionP
from wcr_core.dbstore_api.wdbs_instance_api import WdbsBlockInstanceP, \
    WdbsModelInstanceP, WdbsOperatorInstanceP
from wcr_core.dbstore_api.wdbs_buttons_api import WdbsButtonP
from wcr_core.dbstore_api.wdbs_shape_api import WdbsShapeP


def wdbs_cloneEntities(entities, source, block, inputF, outputF):
    """
    Clone a list of entities and their connecting wires into the given block.
    For wires connected with only one end to one of the given entities, a port
    element will be created.

    :param entities: Entities to clone
    :type entities: list[WdbsEntityP]
    :param source: Source the given entities originate from
    :type source: WdbsBlockP
    :param block: Target block
    :type block: WdbsBlockP
    :param inputF: Input-port generating function
                   (target, inner object, inner connector, outer object,
                    outer connector)
    :type inputF: (WdbsBlockP, WdbsInstanceP, WdbsConnectionPointP,\
                   WdbsInstanceP, WdbsConnectionPointP, WdbsConnectionP) \
                   -> WdbsInstanceP
    :param outputF: Output-port generating function
                    (target, inner object, inner connector, outer object,
                     outer connector)
    :type outputF: (WdbsBlockP, WdbsInstanceP, WdbsConnectionPointP,
                    WdbsInstanceP, WdbsConnectionPointP, WdbsConnectionP)
                    -> WdbsInstanceP
    """
    if not entities: return

    entities = set(e for e in entities if not isinstance(e, WdbsConnectionP))

    old2new = {}
    wdbs_clone(entities, block, old2new)

    for wire in source.getConnections():
        so = wire.getStartObject()
        do = wire.getEndObject()

        ssel = so in entities
        dsel = do in entities

        if ssel and dsel:
            wdbs_cloneConnection(wire, block, old2new)
        elif ssel and not dsel and outputF is not None:
            port = outputF(block, so, wire.getStartXPoint(),
                           do, wire.getEndXPoint(), wire)
            if port is not None:
                nso = old2new[so.getUUID()]
                nsc = old2new[wire.getStartXPoint().getUUID()]
                block.createConnection(
                    nso, nsc, port, next(port.getConnectionPoints()),
                    name=wire.getName(), info=wire.getInfo(),
                    **wire.getAllMeta())
        elif not ssel and dsel and inputF is not None:
            port = inputF(block, do, wire.getEndXPoint(),
                          so, wire.getStartXPoint(), wire)
            if port is not None:
                neo = old2new[do.getUUID()]
                nec = old2new[wire.getEndXPoint().getUUID()]
                block.createConnection(
                    port, next(port.getConnectionPoints()), neo, nec,
                    name=wire.getName(), info=wire.getInfo(),
                    **wire.getAllMeta())


# noinspection PyTypeChecker
def wdbs_clone(elements, target, old2new=None):
    """
    Clone a list of entities into a given target.

    :param elements: Entities to clone
    :type elements: Iterable[WdbsEntityP]
    :param target: Target block
    :type target: WdbsBlockP
    :param old2new: Mapping from old instances' IDs to their cloned equivalent
    :type old2new: dict[UUID, WdbsEntityP]
    :return: The given target
    :rtype: WdbsBlockP
    """
    if old2new is None:
        old2new = {}

    for inst in (e for e in elements if isinstance(e, WdbsBlockInstanceP)):
        old2new[inst.getUUID()] = \
            wdbs_cloneBlockInstance(inst, target, old2new)
    for inst in (e for e in elements if isinstance(e, WdbsModelInstanceP)):
        old2new[inst.getUUID()] = \
            wdbs_cloneModelInstance(inst, target, old2new)
    for inst in (e for e in elements if isinstance(e, WdbsOperatorInstanceP)):
        old2new[inst.getUUID()] = \
            wdbs_cloneOperatorInstance(inst, target, old2new)
    for wire in (e for e in elements if isinstance(e, WdbsConnectionP)):
        old2new[wire.getUUID()] = wdbs_cloneConnection(wire, target,
                                                       old2new)
    for shape in (e for e in elements if isinstance(e, WdbsShapeP)):
        old2new[shape.getUUID()] = wdbs_cloneShape(shape, target)
    for button in (e for e in elements if isinstance(e, WdbsButtonP)):
        old2new[button.getUUID()] = wdbs_cloneButton(button, target)

    return target


def wdbs_cloneBlockInstance(inst, target, old2new=None):
    """
    Clone an instance into the given target

    :param inst: Instance to clone
    :type inst: WdbsBlockInstanceP
    :param target: Target
    :type target: WdbsBlockP
    :param old2new: Mapping from old instances' IDs to their cloned equivalent
    :type old2new: dict[UUID, WdbsEntityP]
    :return: Cloned instance
    :rtype: WdbsBlockInstanceP
    """
    block = inst.getBlock()
    cids, clst = wdbs_mkConnectorList(inst)
    meta = inst.getAllMeta()
    new = target.createBlockInstance(
        name=inst.getName(), info=inst.getInfo(), clst=clst,
        btyp=block.getEntityType(), **meta)
    if old2new is not None:
        old2new.update(zip(cids, new.getConnectionPoints()))
    newBlock = new.getBlock()
    newBlock.setName(block.getName())
    newBlock.setInfo(block.getInfo())
    for k, v in block.getAllMeta().items():
        newBlock.setMeta(v, k)
    wdbs_cloneBlock(block, newBlock, old2new)
    return new


def wdbs_cloneModelInstance(inst, target, old2new=None):
    """
    Clone an instance into the given target

    :param inst: Instance to clone
    :type inst: WdbsModelInstanceP
    :param target: Target
    :type target: WdbsBlockP
    :param old2new: Mapping from old instances' IDs to their cloned equivalent
    :type old2new: dict[UUID, WdbsEntityP]
    :return: Cloned instance
    :rtype: WdbsModelInstanceP
    """
    model = inst.getModel()
    cids, clst = wdbs_mkConnectorList(inst)
    meta = inst.getAllMeta()
    new = target.createModelInstance(
        name=inst.getName(), info=inst.getInfo(), clst=clst,
        mtyp=model.getEntityType(), **meta)
    newModel = new.getModel()
    newModel.setName(model.getName())
    newModel.setInfo(model.getInfo())
    newModel.setPath(model.getPath())
    newModel.setEntryPoint(model.getEntryPoint())
    for k, v in model.getAllMeta().items():
        newModel.setMeta(v, k)
    if old2new is not None:
        old2new.update(zip(cids, new.getConnectionPoints()))
    return new


def wdbs_cloneOperatorInstance(func, target, old2new=None):
    """
    Clone a function into the given target

    :param func: Function to clone
    :type func: WdbsOperatorInstanceP
    :param target: Target
    :type target: WdbsBlockP
    :param old2new: Mapping from old instances' IDs to their cloned equivalent
    :type old2new: dict[UUID, WdbsEntityP]
    :return: Cloned function
    :rtype: WdbsOperatorInstanceP
    """
    params = [(p.getName(), p.getValue()) for p in func.getParameters()]
    uids, clst = wdbs_mkConnectorList(func)
    meta = func.getAllMeta()
    new = target.createOperatorInstance(
        func.getOperator(), name=func.getName(), info=func.getInfo(),
        clst=clst, params=params, **meta)
    if old2new is not None:
        old2new.update(zip(uids, new.getConnectionPoints()))
    return new


def wdbs_cloneConnection(conn, target, old2new):
    """
    Clone a wire into the given target

    :param conn: Connection to clone
    :type conn: WdbsConnectionP
    :param target: Target
    :type target: WdbsBlockP
    :param old2new: Mapping from old instances' IDs to their cloned equivalent
    :type old2new: dict[UUID, WdbsEntityP]
    :return: Cloned wire
    :rtype: WdbsConnectionP
    """
    so = old2new[conn.getStartObject().getUUID()]
    """:type: WdbsInstanceP"""
    if so is None:
        raise ValueError('Could not find source object')
    sc = old2new[conn.getStartXPoint().getUUID()]
    """:type: WdbsConnectionPointP"""
    if sc is None:
        raise ValueError('Could not find source connector')
    eo = old2new[conn.getEndObject().getUUID()]
    """:type: WdbsInstanceP"""
    if eo is None:
        raise ValueError('Could not find destination object')
    ec = old2new[conn.getEndXPoint().getUUID()]
    """:type: WdbsConnectionPointP"""
    if ec is None:
        raise ValueError('Could not find destination connector')
    return target.createConnection(
        so, sc, eo, ec, conn.getName(), conn.getInfo(), **conn.getAllMeta())


def wdbs_cloneShape(shape, target):
    """
    Clone a shape into the given target

    :param shape: Shape to clone
    :type shape: WdbsShapeP
    :param target: Target
    :type target: WdbsBlockP
    :return: Cloned shape
    :rtype: WdbsShapeP
    """
    meta = shape.getAllMeta()
    return target.createShape(shape.getEntityType(), **meta)


def wdbs_cloneButton(button, target):
    """
    Clone a button into the given target

    :param button: Button to clone
    :type button: WdbsButtonP
    :param target: Target
    :type target: WdbsBlockP
    :return: Cloned button
    :rtype: WdbsButtonP
    """
    params = [(p.getName(), p.getValue()) for p in button.getParameters()]
    meta = button.getAllMeta()
    return target.createButton(
        title=button.getTitle(), command=button.getCommand(), params=params,
        **meta)


def wdbs_cloneBlock(block, target, old2new=None):
    """
    Recursively clone a model block's contents into the given target.

    :param block: Block to clone contents of
    :type block: WdbsBlockP
    :param target: Target
    :type target: WdbsBlockP
    :param old2new: Mapping from old instances' IDs to their cloned equivalent
    :type old2new: dict[UUID, WdbsEntityP]
    :return: The given target
    :rtype: WdbsBlockP
    """
    if old2new is None:
        old2new = {}

    for inst in block.getInstances():
        if isinstance(inst, WdbsBlockInstanceP):
            old2new[inst.getUUID()] = \
                wdbs_cloneBlockInstance(inst, target, old2new)
        elif isinstance(inst, WdbsModelInstanceP):
            old2new[inst.getUUID()] = \
                wdbs_cloneModelInstance(inst, target, old2new)
        elif isinstance(inst, WdbsOperatorInstanceP):
            old2new[inst.getUUID()] = \
                wdbs_cloneOperatorInstance(inst, target, old2new)
        else:
            raise ValueError('Unknown instance type: ' + str(inst))
    for wire in block.getConnections():
        old2new[wire.getUUID()] = wdbs_cloneConnection(
            wire, target, old2new)
    for shape in block.getShapes():
        old2new[shape.getUUID()] = wdbs_cloneShape(shape, target)
    for button in block.getButtons():
        old2new[button.getUUID()] = wdbs_cloneButton(button, target)

    return target
