# ----------------------------------------------------------------------------
# - wdbs-entity                                                              -
# - wild cruncher: entity base classes                        core / dbstore -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs_entity
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Entity base classes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod
from weakref import ref

from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_core.dbstore_api.wdbs_entity_api import WdbsEntityP
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_DELETE_EVENT, \
    WdbsObservableP


# noinspection PyAbstractClass
class WdbsObservableMixinP(WdbsObservableP):
    __slots__ = ()

    @abstractmethod
    def _notifyObservers(self, event, *args, **kwargs):
        raise NotImplementedError()


class WdbsObservableMixin(WdbsObservableMixinP, WcmnReleasableP):
    __slots__ = ()

    slots = ('__observers',)

    # noinspection PyUnresolvedReferences
    def __init__(self):
        self.__observers = []
        """:type: list[ReferenceType(WdbsObserverP)]"""

    def release(self):
        self.__observers = []

    def _notifyObservers(self, event, *args, **kwargs):
        dltd = None
        for oref in self.__observers:
            obs = oref()
            if obs is not None:
                obs.dbNotify(self, event, *args, **kwargs)
            else:
                if dltd is None:
                    dltd = []
                dltd.append(oref)
        if dltd:
            for oref in dltd:
                self.__observers.remove(oref)

    def registerObserver(self, observer, event=None):
        self.__observers.append(ref(observer))

    def unregisterObserver(self, observer):
        oidx = None
        for i in range(len(self.__observers)):
            obs = self.__observers[i]()
            if obs is observer:
                oidx = i
                break
        if oidx is not None:
            del self.__observers[oidx]


# noinspection PyAbstractClass
class WdbsEntityBase(WdbsEntityP, WdbsObservableMixin, WcmnReleasableP):
    __slots__ = ('__context',) + WdbsObservableMixin.slots

    def __init__(self, dctx):
        """
        :param dctx: DBS context
        :type dctx: WdbsContextP
        """
        WdbsObservableMixin.__init__(self)

        self.__context = dctx

        self.__observers = []
        """:type: list[ReferenceType(WdbsObserverP)]"""

    def release(self):
        WdbsObservableMixin.release(self)

    def doDelete(self, recurse=True):
        self._notifyObservers(WDBS_DELETE_EVENT)
        self.release()

    def delete(self):
        self.doDelete()

    @property
    def context(self):
        return self.__context
