# ----------------------------------------------------------------------------
# - wdbs_meta_data                                                           -
# - wild cruncher: meta-data instances                        core / dbstore -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs-meta-data
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher meta-data instances

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from wcr_core.dbstore_api.wdbs_meta_data_api import *
from wcr_core.dbstore_api.wdbs_meta_type_api import *


# noinspection PyAbstractClass
class WdbsMetaMixin(WdbsMetaP):
    """
    Abstract base-class for meta-data instances.
    """

    __slots__ = tuple()

    _SLOTS_ = ('_WdbsMetaMixin__type',)

    def __init__(self, mtyp):
        """
        :param mtyp: Meta-data type
        :type mtyp: WdbsMetaTypeP
        """
        self.__type = mtyp

    def getType(self):
        return self.__type

    def asValue(self):
        return self.getType().asValue(self)

    def __str__(self):
        return str(self.asValue()) + ' @' + str(self.getType())

    def __repr__(self):
        return str(self)


# noinspection PyAbstractClass
class WdbsAtomicMetaMixin(WdbsMetaMixin, WdbsAtomicMetaP):
    """
    Abstract base-class for atomic meta-data instances.
    """

    __slots__ = tuple()

    def __init__(self, metaT):
        """
        :param metaT: Meta-data type
        :type metaT: WdbsAtomicMetaTypeP
        """
        assert isinstance(metaT, WdbsAtomicMetaTypeP)
        super().__init__(metaT)

    def toString(self):
        mtyp = self.getType()
        """:type: WdbsAtomicMetaTypeP"""
        return mtyp.toString(self.getValue())


# noinspection PyAbstractClass
class WdbsStructuralMetaMixin(WdbsMetaMixin, WdbsStructuralMetaP):
    """
    Abstract base-class for structural meta-data instances.
    """

    __slots__ = tuple()

    def __init__(self, metaT):
        """
        :param metaT: Meta-data type
        :type metaT: WdbsStructuralMetaTypeP
        """
        assert isinstance(metaT, WdbsStructuralMetaTypeP)
        super().__init__(metaT)


# noinspection PyAbstractClass
class WdbsUnionMetaMixin(WdbsMetaMixin, WdbsUnionMetaP):
    """
    Abstract base-class for union meta-data instances.
    """

    __slots__ = tuple()

    def __init__(self, metaT):
        """
        :param metaT: Meta-data type
        :type metaT: WdbsUnionMetaTypeP
        """
        assert isinstance(metaT, WdbsUnionMetaTypeP)
        super().__init__(metaT)

    @abstractmethod
    def serializationIterator(self):
        """
        :return: Iterable of all type-id/value pair
        :rtype: Iterable[(str, WdbsMetaMP)]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsRecordMetaMixin(WdbsStructuralMetaMixin, WdbsRecordMetaP):
    """
    Abstract base-class for record meta-data instances.
    """

    __slots__ = tuple()

    def __init__(self, metaT):
        """
        :param metaT: Meta-data type
        :type metaT: WdbsRecordMetaTypeP
        """
        assert isinstance(metaT, WdbsRecordMetaTypeP)
        WdbsStructuralMetaMixin.__init__(self, metaT)

    def getValueType(self, key):
        mtyp = self.getType()
        """:type: WdbsRecordMetaTypeP"""
        return mtyp.getValueType(key)


# noinspection PyAbstractClass
class WdbsTupleMetaMixin(WdbsStructuralMetaMixin, WdbsTupleMetaP):
    """
    Abstract base-class for tuple meta-data instances.
    """

    __slots__ = tuple()

    def __init__(self, metaT):
        """
        :param metaT: Meta-data type
        :type metaT: WdbsTupleMetaTypeP
        """
        assert isinstance(metaT, WdbsTupleMetaTypeP)
        WdbsStructuralMetaMixin.__init__(self, metaT)

    def getValueType(self, index):
        mtyp = self.getType()
        """:type: WdbsTupleMetaTypeP"""
        return mtyp.getValueType(index)

    def keys(self):
        mtyp = self.getType()
        """:type: WdbsTupleMetaTypeP"""
        return range(len(mtyp))

    def kvPairs(self):
        return enumerate(self.values())


# noinspection PyAbstractClass
class WdbsListMetaMixin(WdbsStructuralMetaMixin, WdbsListMetaP):
    """
    Abstract base-class for list meta-data instances.
    """

    __slots__ = tuple()

    def __init__(self, metaT):
        """
        :param metaT: Meta-data type
        :type metaT: WdbsListMetaTypeP
        """
        assert isinstance(metaT, WdbsListMetaTypeP)
        WdbsStructuralMetaMixin.__init__(self, metaT)

    def getValueType(self):
        mtyp = self.getType()
        """:type: WdbsListMetaTypeP"""
        return mtyp.getValueType()

    def keys(self):
        return range(len(self.values()))

    def kvPairs(self):
        return enumerate(self.values())


# noinspection PyAbstractClass
class WdbsDictionaryMetaMixin(WdbsStructuralMetaMixin,
                              WdbsDictionaryMetaP):
    """
    Abstract base-class for dictionary/map meta-data instances.
    """

    __slots__ = tuple()

    def __init__(self, metaT):
        """
        :param metaT: Meta-data type
        :type metaT: WdbsDictionaryMetaTypeP
        """
        assert isinstance(metaT, WdbsDictionaryMetaTypeP)
        WdbsStructuralMetaMixin.__init__(self, metaT)

    def getKeyType(self):
        mtyp = self.getType()
        """:type: WdbsDictionaryMetaTypeP"""
        return mtyp.getKeyType()

    def getValueType(self):
        mtyp = self.getType()
        """:type: WdbsDictionaryMetaTypeP"""
        return mtyp.getValueType()
