# ----------------------------------------------------------------------------
# - wdbs-meta-description                                                    -
# - wild cruncher:rsh meta descriptions                       core / dbstore -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs_meta_description
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher raw hierarchy meta descriptions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from base64 import b64encode, b64decode

from silkan_common.scmn_utils import nmap, smap


#: Pre-defined meta-data types
from wcr_core.dbstore_api.wdbs_meta_description_api import WdbsMetaFieldP, \
    WdbsMetaDescriptionP

WDBS_META_STR, WDBS_META_INT, WDBS_META_FLOAT, WDBS_META_BOOL, \
    WDBS_META_PXL_POINT, WDBS_META_REL_POINT, WDBS_META_DIM, \
    WDBS_META_BINARY = range(1, 9)

_WDBS_META_FLOAT_TUPLE = (
    lambda x: [str(_x) for _x in x] if x else None,
    lambda xs: (float(xs[0]), float(xs[1])) if (xs is not None and len(xs) == 2
                                                and all(xs)) else None)

_WDBS_META_SER = {
    WDBS_META_STR: (lambda x: x, lambda x: x),
    WDBS_META_INT: (lambda x: nmap(str, x), lambda x: smap(int, x)),
    WDBS_META_FLOAT: (lambda x: nmap(str, x), lambda x: smap(float, x)),
    WDBS_META_PXL_POINT: _WDBS_META_FLOAT_TUPLE,
    WDBS_META_REL_POINT: _WDBS_META_FLOAT_TUPLE,
    WDBS_META_DIM: _WDBS_META_FLOAT_TUPLE,
    WDBS_META_BOOL: (lambda x: nmap(str, x),
                     lambda x: x.lower().startswith('t') if x else None),
    WDBS_META_BINARY: (
        lambda x: (b64encode(x if isinstance(x, bytes) else x.encode('UTF-8'))
                   if x else None).decode('UTF-8'),
        lambda x: b64decode(x.encode('UTF-8') if x else None))
}


class WdbsMetaField(WdbsMetaFieldP):
    """
    Meta-data field.
    """
    __slots__ = ('__name', '__title', '__temporary', '__type', '__rwFuncs')

    def __init__(self, name, mType, title='', temporary=False):
        """
        :param name: Field name
        :type name: str
        :param mType: Field type
        :type mType: WdbsMetaTypeP
        :param title: Field title/description
        :type title: str
        :param temporary: Field holds temporary/non-serializable data
        :type temporary: bool
        """
        #: Field name
        self.__name = name
        #: Field title/description
        self.__title = title
        #: Whether the field holds temporary/non-serializable data
        self.__temporary = temporary
        #: Meta type
        self.__type = mType

        self.__rwFuncs = {}

    @property
    def name(self):
        return self.__name

    @property
    def title(self):
        return self.__title

    @property
    def is_temporary(self):
        return self.__temporary

    def set_temporary(self, temporary):
        self.__temporary = temporary

    @property
    def metaType(self):
        return self.__type

    def addRWFuncs(self, code, version, readF, writeF=None):
        self.__rwFuncs[(code, version)] = (readF, writeF)

    def getRWFuncs(self, code, version):
        return self.__rwFuncs.get((code, version)) or (None, None)


class WdbsMetaDescription(WdbsMetaDescriptionP):
    """
    Meta-data description for an entity class.
    """

    __slots__ = ('__fields',)

    def __init__(self, *fields):
        """
        :param fields: Meta-data fields
        :type fields: list[WdbsMetaFieldP | WdbsMetaDescriptionP |
                           dict[str, WdbsMetaFieldP]]
        """
        self.__fields = {}
        """:type: dict[str, WdbsMetaField]"""
        for f in fields:
            if isinstance(f, WdbsMetaField):
                self.__fields[f.name] = f
            elif isinstance(f, WdbsMetaDescriptionP):
                for _f in f.fields:
                    self.__fields[_f.name] = _f
            else:
                # noinspection PyUnresolvedReferences
                for k, v in f.items():
                    self.__fields[k] = v

    @property
    def fields(self):
        """
        :rtype: Iterable[WdbsMetaField]
        """
        return self.__fields.values()

    def addField(self, field):
        """
        :rtype: Iterable[WdbsMetaField]
        """
        self.__fields[field.name] = field

    def __getitem__(self, item):
        """
        :type item: str
        :rtype: WdbsMetaField | None
        """
        return self.__fields.get(item)

    def __str__(self):
        return str(list(self.__fields.keys()))

    def __repr__(self):
        return str(self)
