# ----------------------------------------------------------------------------
# - wdbs_meta_type                                                           -
# - wild cruncher: meta-data types                            core / dbstore -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs-meta-type
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher meta-data types

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from base64 import b64decode, b64encode
from collections import OrderedDict
import re
from uuid import UUID

from silkan_common.scmn_geometry import ScmnRectangle
from wcr_core.common.wcmn_file_handler import WcmnFile
from silkan_common.scmn_tuple import ScmnMTuple
from wcr_core.dbstore_api.wdbs_meta_type_api import *


# noinspection PyAbstractClass
class WdbsMetaTypeBase(WdbsMetaTypeP):
    """
    Abstract base class for meta-data types.
    """

    def __str__(self):
        return self.getName()

    def __repr__(self):
        return str(self)


# noinspection PyAbstractClass
class WdbsAtomicMetaTypeBase(WdbsMetaTypeBase, WdbsAtomicMetaTypeP):
    """
    Abstract base class for meta-data type for atomic values.
    """

    def __init__(self, default=None):
        """
        :param default: Default value
        :type default: object
        """
        self.__default = default

    def getStructName(self):
        return 'atomic'

    @property
    def defaultValue(self):
        return self.__default

    def getEmptyValue(self):
        return self.defaultValue

    def setFromValue(self, setF, *args, **kwargs):
        if not args:
            setF(None)
            return
        arg = args[0]
        if isinstance(arg, str):
            setF(self.fromString(arg))
        else:
            setF(arg)

    def asValue(self, meta):
        return meta.getValue()


class WdbsAnyMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic values of unspecified type.
    """

    def __init__(self, default=None):
        """
        :param default: Default value
        :type default: object | None
        """
        super().__init__(default)

    def getName(self):
        return 'any'

    def toString(self, metaV):
        return str(metaV)

    def fromString(self, inStr):
        return None


class WdbsStringMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic string values.
    """

    def __init__(self, default=''):
        """
        :param default: Default value
        :type default: str | None
        """
        super().__init__(default)

    def getName(self):
        return 'string'

    def toString(self, metaV):
        return metaV

    def fromString(self, inStr):
        return inStr


class WdbsTextMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic text values.
    """

    def __init__(self, default=''):
        """
        :param default: Default value
        :type default: str | None
        """
        super().__init__(default)

    def getName(self):
        return 'text'

    def toString(self, metaV):
        return metaV

    def fromString(self, inStr):
        return inStr


class WdbsIntegerMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic integer values.
    """

    def __init__(self, default=0):
        """
        :param default: Default value
        :type default: int | None
        """
        super().__init__(default)

    def getName(self):
        return 'integer'

    def toString(self, metaV):
        return str(metaV) if metaV is not None else ''

    def fromString(self, inStr):
        return int(inStr)


class WdbsFloatMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic float values.
    """

    def __init__(self, default=0.0):
        """
        :param default: Default value
        :type default: float | None
        """
        super().__init__(default)

    def getName(self):
        return 'float'

    def toString(self, metaV):
        return str(metaV) if metaV is not None else ''

    def fromString(self, inStr):
        return float(inStr)


class WdbsBoolMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic boolean values.
    """

    #: Pattern to detect truth-values
    _TRUTH_PATTERN = re.compile('t(?:rue)?', re.IGNORECASE)

    def __init__(self, default=False):
        """
        :param default: Default value
        :type default: bool | None
        """
        super().__init__(default)

    def getName(self):
        return 'bool'

    def toString(self, metaV):
        return str(metaV) if metaV is not None else ''

    def fromString(self, inStr):
        return bool(self._TRUTH_PATTERN.match(inStr))


class WdbsBinaryMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic binary values.
    """

    def __init__(self, default=b''):
        """
        :param default: Default value
        :type default: bytes | None
        """
        super().__init__(default)

    def getName(self):
        return 'binary'

    def toString(self, metaV):
        if not metaV:
            return ''
        raw = metaV if isinstance(metaV, bytes) else metaV.encode('UTF-8')

        return b64encode(raw).decode('UTF-8')

    def fromString(self, inStr):
        return b64decode(inStr.encode('UTF-8')) if inStr else b''


class WdbsUuidMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic UUID values.
    """

    def __init__(self, default=None):
        """
        :param default: Default value
        :type default: UUID | None
        """
        super().__init__(default)

    def getName(self):
        return 'uuid'

    def toString(self, metaV):
        return str(metaV) if metaV is not None else ''

    def fromString(self, inStr):
        return UUID(inStr)


class WdbsReferenceMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic reference values.
    """

    def __init__(self, default=None):
        """
        :param default: Default value
        :type default: object | None
        """
        super().__init__(default)

    def getName(self):
        return 'reference'

    def toString(self, metaV):
        return str(metaV.getUUID()) if metaV is not None else ''

    def fromString(self, inStr):
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsStructuralMetaTypeBase(WdbsMetaTypeBase,
                                 WdbsStructuralMetaTypeP):
    """
    Abstract base-class for structural meta-data types.
    """
    pass


class WdbsUnionMetaType(WdbsStructuralMetaTypeBase, WdbsUnionMetaTypeP):
    """
    Abstract base class for meta-data type for union values.
    """

    def __init__(self, *idTypePairs, default=None):
        """
        :param idTypePairs: Possible sub-types [(ident, type)]
        :type idTypePairs: list[(str, WdbsMetaTypeP)]
        :param default: Identifier of default type
        :type default: str | None
        """
        self.__default = default
        self.__types = dict(idTypePairs)

    def getStructName(self):
        return 'union'

    def getName(self):
        return self.getStructName()

    def getDefaultTypeId(self):
        return self.__default

    def getType(self, typeId):
        return self.__types.get(typeId)

    def getEmptyValue(self):
        dflt = self.__default
        if dflt:
            return ScmnMTuple(dflt, self.__types[dflt].getEmptyValue())
        else:
            return None

    def setFromValue(self, setF, *args, **kwargs):
        key = None
        value = None
        alen = len(args)
        assert alen <= 2 and not kwargs
        if alen == 1:
            arg = args[0]
            if (isinstance(arg, tuple) or isinstance(arg, ScmnMTuple)) \
                    and len(arg) == 2:
                key, value = arg
            elif arg is not None:
                raise ValueError('Invalid argument for union: {}'.format(args))
        elif alen == 2:
            if isinstance(args[0], str) or args[0] is None:
                key, value = args
            else:
                raise ValueError('Invalid argument for union: {}'.format(args))

        if key:
            setF(value, key)
        else:
            setF(None)

    def asValue(self, meta):
        typeId = meta.getValueTypeIdentifier()
        return ScmnMTuple(typeId, meta.value()) if typeId else None

    def getKey(self, item, data):
        assert isinstance(data, tuple) or isinstance(data, ScmnMTuple), \
            str(type(data)) + ' ' + str(data)
        assert len(data) == 2, str(len(data)) + ' ' + str(data)
        return 1

    def __str__(self):
        return '<' + \
               ' | '.join(
                   '{t}{d}:{v}'.format(t=t, v=v,
                                       d='*' if t == self.__default else '')
                   for t, v in self.__types.items()) + '>'


class WdbsRecordMetaType(WdbsStructuralMetaTypeBase,
                         WdbsRecordMetaTypeP):
    """
    Abstract base-class for record meta-data types.
    """

    def __init__(self, *kpts):
        """
        :param kpts: Record child types
        :type kpts: list[(str, WdbsMetaTypeP) | (str, WdbsMetaTypeP, str)]
        """
        self.__types = OrderedDict(
            ((k, tuple(vs) if len(vs) == 2 else (vs[0], None))
             for k, *vs in kpts))
        """:type: dict[str, (WdbsMetaTypeP, str | None)]"""

    def getStructName(self):
        return 'record'

    def getName(self):
        return self.getStructName()

    def setFromValue(self, setF, *args, **kwargs):
        if args:
            if len(args) == 1:
                vItem = args[0]
                if vItem is None:
                    return
                elif isinstance(vItem, dict):
                    items = vItem.items()
                elif isinstance(vItem, list):
                    items = vItem
                elif isinstance(vItem, tuple) and \
                                len(vItem) == len(self):
                    items = zip(self.getKeys(), vItem)
                else:
                    raise ValueError(
                        'Invalid argument list for {} - {}: {}'.format(
                            self.getName(), self, args))
            else:
                raise ValueError(
                    'Invalid argument list for {} - {}: {}'.format(
                        self.getName(), self, args))
            for k, v in items:
                setF(v, k)

        for k, v in kwargs.items():
            setF(v, k)

    def asValue(self, meta):
        return dict(meta.kvPairs())

    def __len__(self):
        return len(self.__types)

    def __iter__(self):
        return ((k, v) for k, (v, _) in self.__types.items())

    def getKeys(self):
        return self.__types.keys()

    def getValueType(self, key):
        return self.__types[key][0]

    def getLabel(self, key):
        return self.__types[key][1]

    def getEmptyValue(self):
        return {k: v.getEmptyValue() for k, (v, _) in self.__types.items()}

    def getKey(self, item, data):
        assert isinstance(data, dict), '{} != dict'.format(type(data))
        return next(k for k, v in data.items() if v is item)

    @property
    def types(self):
        return self.__types

    def __str__(self):
        return '{' + ', '.join('{}: {}'.format(k, t)
                               for k, (t, _) in self.__types.items()) + '}'

    def __add__(self, other):
        newTypes = OrderedDict()
        newTypes.update(self.types)
        newTypes.update(other.types)
        newMeta = WdbsRecordMetaType()
        newMeta.types.update(newTypes)
        return newMeta



class WdbsTupleMetaType(WdbsStructuralMetaTypeBase, WdbsTupleMetaTypeP):
    """
    Abstract base-class for tuple meta-data types.
    """

    def __init__(self, *types):
        """
        :param types: Tuple child types
        :type types: Iterable[WdbsMetaTypeP]
        """
        self.__types = tuple(types)

    def getStructName(self):
        return 'tuple'

    def getName(self):
        return self.getStructName()

    def asValue(self, meta):
        return tuple(meta.values())

    def setFromValue(self, setF, *args, **kwargs):
        if not args:
            return

        if len(args) == 1:
            vItem = args[0]
            if vItem is None:
                setF(None)
                return
            elif (isinstance(vItem, list) or
                      isinstance(vItem, tuple) or
                      isinstance(vItem, ScmnMTuple)) and \
                            len(vItem) == len(self):
                items = vItem
            else:
                raise ValueError(
                    'Invalid argument list for {} - {}: {}'.format(
                        self.getName(), self, args))
        else:
            items = args
        for k, v in enumerate(items):
            setF(v, k)

    def __len__(self):
        return len(self.__types)

    def __iter__(self):
        return enumerate(self.__types)

    def getValueType(self, index):
        return self.__types[index]

    def getEmptyValue(self):
        return tuple(t.getEmptyValue() for t in self.__types)

    def getKey(self, item, data):
        assert isinstance(data, tuple) or isinstance(data, ScmnMTuple)
        return next(i for i, v in enumerate(data) if v is item)

    def __str__(self):
        return '(' + ', '.join(map(str, self.__types)) + ')'


# noinspection PyAbstractClass
class WdbsMultiMetaTypeBase(WdbsStructuralMetaTypeBase,
                            WdbsMultiMetaTypeP):
    """
    Abstract base-class for container meta-data types.
    """

    def __init__(self, vType):
        """
        :param vType: Container children type
        :type vType: WdbsMetaTypeP
        """
        self.__type = vType

    def getValueType(self):
        """
        Get the value-type of the child items.

        :return: Value-type of the child items
        :rtype: WdbsMetaTypeP
        """
        return self.__type


class WdbsListMetaType(WdbsMultiMetaTypeBase, WdbsListMetaTypeP):
    """
    Abstract base-class for list meta-data types.
    """

    def __init__(self, vType):
        """
        :param vType: List children type
        :type vType: WdbsMetaTypeP
        """
        super().__init__(vType)

    def getStructName(self):
        return 'list'

    def getName(self):
        return self.getStructName()

    def asValue(self, meta):
        """
        :type meta: WdbsListMetaP
        :rtype: list[object]
        """
        return list(meta.values())

    def setFromValue(self, setF, *args, **kwargs):
        """
        :type meta: WdbsListMetaP
        :rtype: WdbsListMetaP
        """
        # Clear the list
        setF(None)

        if not args:
            return

        if len(args) == 1:
            vItem = args[0]
            if isinstance(vItem, list) or isinstance(vItem, tuple):
                items = vItem
            elif hasattr(vItem, '__iter__'):
                items = vItem
            else:
                if vItem is not None:
                    setF(vItem, None)
                return
        else:
            items = args

        for item in items:
            setF(item, None)
        # TOOD: setF(items)

    def getEmptyValue(self):
        return []

    def getKey(self, item, data):
        assert isinstance(data, list)
        return next(i for i, v in enumerate(data) if v is item)

    def __str__(self):
        return '[*{}]'.format(self.getValueType())


class WdbsDictionaryMetaType(WdbsMultiMetaTypeBase,
                             WdbsDictionaryMetaTypeP):
    """
    Base-class for dictionary/map meta-data types.
    """

    def __init__(self, kType, vType):
        """
        :param kType: Map key type
        :type kType: WdbsMetaTypeP
        :param vType: Map value type
        :type vType: WdbsMetaTypeP
        """
        super().__init__(vType)

        self.__keyType = kType

    def getStructName(self):
        return 'dictionary'

    def getName(self):
        return self.getStructName()

    def getKeyType(self):
        return self.__keyType

    def asValue(self, meta):
        return dict(meta.kvPairs())

    def setFromValue(self, setF, *args, **kwargs):
        setF(None)

        if args:
            if len(args) == 1:
                vItem = args[0]
                if vItem is None:
                    items = None
                elif isinstance(vItem, dict):
                    items = vItem.items()
                elif isinstance(vItem, list):
                    items = vItem
                elif isinstance(vItem, tuple):
                    items = [vItem]
                else:
                    raise ValueError(
                        'Invalid argument list for {} - {}: {}'.format(
                            self.getName(), self, args))
            else:
                items = args
            if items:
                for k, v in items:
                    setF(v, k)

        for k, v in kwargs.items():
            setF(v, k)

    def getEmptyValue(self):
        return {}

    def getKey(self, item, data):
        assert isinstance(data, dict)
        return next(k for k, v in data.items() if v is item)

    def __str__(self):
        return '{{*{}:{}}}'.format(self.__keyType, self.getValueType())


class WdbsOrderedDictionaryMetaType(WdbsDictionaryMetaType,
                                    WdbsOrderedDictionaryMetaTypeP):
    """
    Base-class for dictionary/map meta-data types that maintain insertion
    order.
    """

    def __init__(self, kType, vType):
        """
        :param kType: Map key type
        :type kType: WdbsMetaTypeP
        :param vType: Map value type
        :type vType: WdbsMetaTypeP
        """
        super().__init__(kType, vType)

    def getStructName(self):
        return 'ordered-dictionary'

    def asValue(self, meta):
        return OrderedDict(meta.kvPairs())

    def getEmptyValue(self):
        return OrderedDict()


#: Any meta-data type singleton
WDBS_ANY_META = WdbsAnyMetaType()

#: String meta-data type singleton
WDBS_STRING_META = WdbsStringMetaType()

#: String meta-data type singleton with None as default
WDBS_STRING_NONE_META = WdbsStringMetaType(default=None)

#: Text meta-data type singleton
WDBS_TEXT_META = WdbsTextMetaType()

#: Integer meta-data type singleton
WDBS_INTEGER_META = WdbsIntegerMetaType()

#: Integer meta-data type singleton with None as default
WDBS_INTEGER_NONE_META = WdbsIntegerMetaType(default=None)

#: Float meta-data type singleton
WDBS_FLOAT_META = WdbsFloatMetaType()

#: Float meta-data type singleton with None as default
WDBS_FLOAT_NONE_META = WdbsFloatMetaType(default=None)

#: Boolean meta-data type singleton
WDBS_BOOL_META = WdbsBoolMetaType()

#: Binary meta-data type singleton
WDBS_BINARY_META = WdbsBinaryMetaType()

#: UUID meta-data type singleton
WDBS_UUID_META = WdbsUuidMetaType()

#: String meta-data type singleton
WDBS_TIME_META = WdbsStringMetaType()


class WdbsFloatTupleMetaType(WdbsRecordMetaType):
    """
    Meta-data type based on tuples of floats.
    """

    def __init__(self, name, *fields):
        """
        :param name: Meta-data type name
        :type name: str
        :param fields: Field names
        :type fields: list[str]
        """
        super().__init__(*[(f, WDBS_FLOAT_META) for f in fields])

        self.__name = name
        self.__fields = fields

    def getName(self):
        return self.__name

    def asValue(self, meta):
        """
        :type meta: WdbsRecordMetaP
        """
        # TODO: use __getitem__
        data = super().asValue(meta)
        """:type: dict[str, float]"""
        return tuple(data[f] for f in self.__fields)

    def setFromValue(self, setF, *args, **kwargs):
        if len(args) == 1:
            arg = args[0]
            if arg is None:
                setF(None)
                return
            elif isinstance(args[0], tuple) and \
                            len(args[0]) == len(self.__fields) and \
                    isinstance(args[0][0], float):
                for f, v in zip(self.__fields, arg):
                    setF(v, f)
                return

        return super().setFromValue(setF, *args, **kwargs)


#: Point meta-data type singleton
WDBS_POINT_META = WdbsFloatTupleMetaType('point', 'x', 'y')

#: Relative-point meta-data type singleton
WDBS_RELATIVE_POINT_META = WdbsFloatTupleMetaType('rel-point', 'x', 'y')

#: Dimension meta-data type singleton
WDBS_DIMENSION_META = WdbsFloatTupleMetaType('dimension', 'w', 'h')

#: Rectangle meta-data type singleton
WDBS_RECTANGLE_META = WdbsFloatTupleMetaType('rectangle',
                                             'x', 'y', 'w', 'h')


class WdbsPathMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic file values representing a file-system path.
    """

    def __init__(self):
        super().__init__(default=None)

    def getName(self):
        return 'file'

    def toString(self, metaV):
        return metaV.path if metaV else None

    def fromString(self, inStr):
        return WcmnFile(inStr) if inStr else None


#: Path meta-data type singleton
WDBS_PATH_META = WdbsPathMetaType()


class _WdbsRectangleMetaType(WdbsFloatTupleMetaType):
    """
    Meta-data type for rectangles.
    """

    def __init__(self):
        super().__init__('cmn-rectangle', 'x', 'y', 'w', 'h')

    def asValue(self, meta):
        return ScmnRectangle(*(v or 0 for v in meta.values()))

    def setFromValue(self, setF, *args, **kwargs):
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, ScmnRectangle):
                return super().setFromValue(
                    setF, (arg.x, arg.y, arg.width, arg.height))
        return super().setFromValue(setF, *args, **kwargs)

#: Rectangle meta-data type singleton
WDBS_META_RECTANGLE = _WdbsRectangleMetaType()


class WdbsFrequencyMetaType(WdbsAtomicMetaTypeBase):
    """
    Meta-data type for atomic string values.
    """

    def __init__(self, default=''):
        """
        :param default: Default value
        :type default: str | None
        """
        super().__init__(default)

    def getName(self):
        return 'frequency'

    def toString(self, metaV):
        return metaV

    def fromString(self, inStr):
        return inStr


WDBS_META_FREQUENCY = WdbsFrequencyMetaType()
