# ----------------------------------------------------------------------------
# - wdbs-util                                                                -
# - wild cruncher: utility functions                          core / dbstore -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs_util
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher utility functions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from uuid import UUID, uuid1


def wdbs_makeUUID(uuid=None):
    """
    Create an UUID, possibly based on the given (deserialized) uuid value.

    :param uuid: UUID value read from deserialization
    :type uuid: UUID | str | None
    :return: New UUID
    :rtype: UUID
    """
    if isinstance(uuid, UUID):
        return uuid
    if uuid:
        try:
            return UUID(uuid)
        except (ValueError, AttributeError):
            pass
    return uuid1()


def wdbs_getOtherEndByPoint(cnctn, point):
    """
    Get the other end (object, connectionPoint) of the given wire (aka
    connection).

    :param cnctn: Connection
    :type cnctn: WdbsConnectionP
    :param point: Connection point
    :type point: WdbsConnectionPointP
    :return: Other connected object
    :rtype: (WdbsInstanceP, WdbsConnectionPointP)
    """
    if cnctn.getStartXPoint() is point:
        return cnctn.getEndObject(), cnctn.getEndXPoint()
    else:
        return cnctn.getStartObject(), cnctn.getStartXPoint()


def wdbs_getOtherEndByObj(cnctn, obj):
    """
    Get the other object connected to the given connection

    :param cnctn: Connection
    :type cnctn: WdbsConnectionP
    :param obj: Connected object
    :type obj: WdbsInstanceP
    :return: Other connected object
    :rtype: (WdbsInstanceP, WdbsConnectionPointP)
    """
    if cnctn.getStartObject() is obj:
        return cnctn.getEndObject(), cnctn.getEndXPoint()
    else:
        return cnctn.getStartObject(), cnctn.getStartXPoint()


def wdbs_getConnections(block, obj, xpnt=None):
    """
    Get the connections connected to the given instance object.

    :param block: Block of the object
    :type block: WdbsBlockP
    :param obj: Instance to find connections for
    :type obj: WdbsInstanceP
    :param xpnt: Connection point - if given, limit connections found to it
    :type xpnt: WdbsConnectionPointP
    :return: Connections to the given instance (and possibly connection point)
    :rtype: list[WdbsConnectionP]
    """
    return [c for c in block.getConnections()
            if (c.getStartObject() is obj and
                (xpnt is None or c.getStartXPoint() is xpnt)) or
            (c.getEndObject() is obj and
             (xpnt is None or c.getEndXPoint() is xpnt))]


def wdbs_mkConnectorList(instance):
    """
    Get the data for the given instance's connectors.

    :param instance: Instance to obtain connector data for
    :type instance: WdbsInstanceP
    :return: The instance's connector data [(dir, id, meta)]
    :rtype: (list[UUID], list[(str, None, dict[str, object])])
    """

    def getArgs(xpnt):
        args = xpnt.getAllMeta()
        args['cdir'] = xpnt.getDirection()
        return args

    # noinspection PyTypeChecker
    xpts = list(instance.getConnectionPoints())
    return ([c.getUUID() for c in xpts],
            [(c.getEntityType(), None, getArgs(c)) for c in xpts])
