# ----------------------------------------------------------------------------
# - wdbs-block                                                               -
# - wild cruncher:rsh block object                        core / dbstore-api -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher raw hierarchy block objects

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.dbstore_api.wdbs_entity_api import WdbsNamedEntityP, \
    WdbsEntityP, WdbsTypedEntityP


# noinspection PyAbstractClass
class WdbsContainerP(WdbsEntityP):
    @abstractmethod
    def getInstances(self):
        """
        Get the block instances.

        :return: Instance iterator
        :rtype: Iterable[WdbsInstanceP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getInstanceCount(self):
        """
        Get the number of instances.

        :return: Number of instances
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnections(self):
        """
        Get the block's connections.

        :return: Connection iterator
        :rtype: Iterable[WdbsConnectionP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectionCount(self):
        """
        Get the number of connections in the block.

        :return: Number of connections
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getButtons(self):
        """
        Get the buttons.

        :return: Button iterator
        :rtype: Iterable[WdbsButtonP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getButtonCount(self):
        """
        Get the number of buttons.

        :return: Number of buttons
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getShapes(self):
        """
        Get the shapes.

        :return: Shape iterator
        :rtype: Iterable[WdbsShapeP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getShapeCount(self):
        """
        Get the number of shapes.

        :return: Number of shapes
        :rtype: int
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsEntityTargetP(WdbsEntityP):
    @abstractmethod
    def createBlockInstance(self, name=None, info=None, clst=None, uuid=None,
                            btyp=None, buuid=None, meta=None, bmeta=None,
                            **kwargs):
        """
        Create a new block instance.

        :param name: Instance name
        :type name: str
        :param info: Instance info string
        :type info: str
        :param clst: Instance connection points
        :type clst: Iterable[(str, UUID | None, dict[str, object])]
        :param uuid: UUID for the new entity (should only be provided when
                     deserializing; if `None`, a new UUID will be created)
        :type uuid: UUID | str | None
        :param btyp: Block type
        :type btyp: str
        :param buuid: UUID for the new entity's block (should only be provided
                      when deserializing; if `None`, a new UUID will be
                      created)
        :type buuid:  str
        :param meta: Meta-data description for the instance; will be auto-
                     determined via context meta-function if None
        :type meta: WdbsMetaDescriptionP | None
        :param bmeta: Meta-data description for the block; will be auto-
                      determined via context meta-function if None
        :type bmeta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        :return: Newly created block instance
        :rtype: WdbsBlockInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def createModelInstance(self, name=None, info=None, clst=None, uuid=None,
                            mtyp=None, muuid=None, meta=None, mmeta=None,
                            **kwargs):
        """
        Create a new model instance.

        :param name: Instance name
        :type name: str
        :param info: Instance info string
        :type info: str
        :param clst: Instance connection points
        :type clst: Iterable[(str, UUID | None, dict[str, object])]
        :param uuid: UUID for the new entity (should only be provided when
                     deserializing; if `None`, a new UUID will be created)
        :type uuid: UUID | str | None
        :param mtyp: Model type
        :type mtyp: str | None
        :param muuid: UUID for the new entity's model (should only be provided
                      when deserializing; if `None`, a new UUID will be
                      created)
        :type muuid:  str
        :param meta: Meta-data description for the instance; will be auto-
                     determined via context meta-function if None
        :type meta: WdbsMetaDescriptionP | None
        :param mmeta: Meta-data description for the model; will be auto-
                      determined via context meta-function if None
        :type mmeta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        :return: Newly created model instance
        :rtype: WdbsModelInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def createOperatorInstance(self, oper, name=None, info=None, clst=None,
                               params=None, uuid=None, meta=None, **kwargs):
        """
        Create a new operator instance (i.e. function).

        :param oper: Operator code
        :type oper: str
        :param name: Instance name
        :type name: str
        :param info: Instance info string
        :type info: str
        :param clst: Instance connection points
        :type clst: Iterable[(str, UUID | None, dict[str, object])]
        :param params: Operator parameter values [(name, value)]
        :type params: Iterable[(str, object)]
        :param uuid: UUID for the new entity (should only be provided when
                     deserializing; if `None`, a new UUID will be created)
        :type uuid: UUID | str | None
        :param meta: Meta-data description for the instance; will be auto-
                     determined via context meta-function if None
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        :return: Newly created operator instance
        :rtype: WdbsOperatorInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def createConnection(self, sobj, sxpt, eobj, expt, name=None, info=None,
                         path=None, uuid=None, meta=None, **kwargs):
        """
        Create a new connection between two instances in the block.

        :param sobj: Start object
        :type sobj: WdbsInstanceP
        :param sxpt: Start connection point
        :type sxpt: WdbsConnectionPointP
        :param eobj: End object
        :type eobj: WdbsInstanceP
        :param expt: End connection point
        :type expt: WdbsConnectionPointP
        :param name: Connection name
        :type name: str
        :param info: Connection info string
        :type info: str
        :param path: Connection path
        :type path: Iterable[(float, float)]
        :param uuid: UUID for the new entity (should only be provided when
                     deserializing; if `None`, a new UUID will be created)
        :type uuid: UUID | str | None
        :param meta: Meta-data description for the connection; will be auto-
                     determined via context meta-function if None
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        :return: Newly created connection
        :rtype: WdbsConnectionP
        """
        raise NotImplementedError()

    @abstractmethod
    def createButton(self, title=None, command=None, params=None, uuid=None,
                     meta=None, **kwargs):
        """
        Create a new button.

        :param title: Button title
        :type title: str
        :param command: Command associated with button
        :type command: str
        :param bpos: Position
        :type bpos: (float, float)
        :param params: Command parameter values
        :type params: Iterable[(str, object)]
        :param uuid: UUID for the new entity (should only be provided when
                     deserializing; if `None`, a new UUID will be created)
        :type uuid: UUID | str | None
        :param meta: Meta-data description for the button; will be auto-
                     determined via context meta-function if None
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        :return: Newly created button
        :rtype: WdbsButtonP
        """
        raise NotImplementedError()

    @abstractmethod
    def createShape(self, styp, uuid=None, meta=None, **kwargs):
        """
        Create a new shape.

        :param styp: Shape type
        :type args: str
        :param uuid: UUID for the new entity (should only be provided when
                     deserializing; if `None`, a new UUID will be created)
        :type uuid: UUID | str | None
        :param meta: Meta-data description for the shape; will be auto-
                     determined via context meta-function if None
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        :return: Newly created shape
        :rtype: WdbsShapeP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsBlockP(WdbsTypedEntityP, WdbsNamedEntityP, WdbsContainerP,
                 WdbsEntityTargetP):
    """
    Raw block object.
    """

    @abstractmethod
    def removeInstance(self, instance):
        """
        Remove an instance.

        :param instance: Instance to be removed
        :type instance: WdbsInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeConnection(self, connection):
        """
        Remove a connection.

        :param connection: Connection to be removed
        :type connection: WdbsConnectionP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeButton(self, button):
        """
        Remove a button.

        :param button: Button to be removed
        :type button: WdbsButtonP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeShape(self, shape):
        """
        Remove a shape.

        :param shape: Shape to be removed
        :type shape: WdbsShapeP
        """
        raise NotImplementedError()
