# ----------------------------------------------------------------------------
# - wdbs-instance                                                            -
# - wild cruncher:rsh instance object                     core / dbstore-api -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs_instance
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher raw hierarchy instance objects

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.dbstore_api.wdbs_entity_api import WdbsNamedEntityP
from wcr_core.dbstore_api.wdbs_parameter_api import WdbsEntityParametersP


# noinspection PyAbstractClass
class WdbsInstanceP(WdbsNamedEntityP):
    """
    Instance object.
    """

    @abstractmethod
    def getConnectionPoints(self):
        """
        Get the instance's connection points.

        :return: Connection point iterator
        :rtype: Iterator[WdbsConnectionPointP] | Iterable[WdbsConnectionPointP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectionPointCount(self):
        """
        Get the number of connection points.

        :return: Number of connection points
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def createConnectionPoint(self, ctyp, uuid=None, meta=None, **kwargs):
        """
        Create a new connection point.

        :param ctyp: Connection point type
        :type ctyp: str
        :param uuid: UUID
        :type uuid: UUID | str | None
        :param meta: Meta-data description
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        :type kwargs: dict[str, object]
        :return: New created connection point
        :rtype: WdbsConnectionPointP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeConnectionPoint(self, xpoint):
        """
        Remove a connection point.

        :param xpoint: Connection point to remove
        :type xpoint: WdbsConnectionPointP
        """
        raise NotImplementedError()

    @abstractmethod
    def clearConnectionPoints(self):
        """
        Remove all connection points.
        """
        raise NotImplementedError()

    @abstractmethod
    def setConnectionPointOrder(self, orderF, wrapFs):
        """
        Re-order the instance's connection points.

        :param orderF: Ordering function, returning the order for a given
                       (wrapped) connection point
        :type orderF: (object) -> int
        :param wrapFs: Connection point wrapping/unwrapping functions
        :type wrapFs: ((WdbsConnectionPointP) -> object,
                       (object) -> WdbsConnectionPointP)
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsBlockInstanceP(WdbsInstanceP):
    """
    Block instance object.
    """

    @abstractmethod
    def getBlock(self):
        """
        Get the block instantiated by the instance.

        :return: Instantiated block
        :rtype: WdbsBlockP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsModelInstanceP(WdbsInstanceP):
    """
    Model instance object.
    """

    @abstractmethod
    def getModel(self):
        """
        Get the model instantiated by the instance.

        :return: Instantiated model
        :rtype: WdbsModelP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsOperatorInstanceP(WdbsInstanceP, WdbsEntityParametersP):
    """
    Operator instance object (i.e. function).
    """

    @abstractmethod
    def getOperator(self):
        """
        Get the operator instantiated by the instance.

        :return: Instantiated operator
        :rtype: str
        """
        raise NotImplementedError()
