# ----------------------------------------------------------------------------
# - wdbs_meta_data                                                           -
# - wild cruncher: meta-data instances                    core / dbstore-api -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs-meta-data
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher meta-data instances

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import ABCMeta, abstractmethod


class WdbsMetaP(metaclass=ABCMeta):
    """
    Protocol for meta-data instances.
    """

    __slots__ = tuple()

    @abstractmethod
    def getType(self):
        """
        Get the meta-data type of the instance.

        :return: Meta-data type
        :rtype: WdbsMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def asValue(self):
        """
        Return the encapsulated meta-data.

        :return: Encapsulated meta-data
        :rtype: object
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsAtomicMetaP(WdbsMetaP):
    """
    Protocol for atomic meta-data instances.
    """

    __slots__ = tuple()

    @abstractmethod
    def getType(self):
        """
        :return: Meta-data type
        :rtype: WdbsAtomicMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def getValue(self):
        """
        Get the encapsulated value.

        :return: Encapsulated value
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def toString(self):
        """
        Get a string representation of the encapsulated value, suitable for
        parsing with `WdbsAtomicMetaP.fromString`.

        :return: String representation of the encapsulated value
        :rtype: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsStructuralMetaP(WdbsMetaP):
    """
    Protocol for structural meta-data instances.
    """

    __slots__ = tuple()

    @abstractmethod
    def getType(self):
        """
        :return: Meta-data type
        :rtype: WdbsStructuralMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def __len__(self):
        """
        Get the number of child items.

        :return: Number of child items
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def keys(self):
        """
        Get all keys as raw values.

        :return: Iterable of all keys as values
        :rtype: list[object]
        """
        raise NotImplementedError()

    @abstractmethod
    def values(self):
        """
        Get all raw values.

        :return: Iterable of all values
        :rtype: list[object]
        """
        raise NotImplementedError()

    @abstractmethod
    def kvPairs(self):
        """
        Get all key/raw value-pairs (as values).

        :return: Iterable of all key/value-pairs
        :rtype: dict[object, object]
        """
        raise NotImplementedError()

    @abstractmethod
    def serializationIterator(self):
        """
        Get all key/meta-data-pairs suitable for serialization.

        :return: Iterable of all key/value-pairs
        :rtype: Iterable[(object, WdbsMetaMP)]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsUnionMetaP(WdbsStructuralMetaP, WdbsMetaP):
    """
    Protocol for union meta-data instances.
    """

    __slots__ = tuple()

    @abstractmethod
    def getType(self):
        """
        :return: Meta-data type
        :rtype: WdbsUnionMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def getValueTypeIdentifier(self):
        """
        Get the current union sub-type identifier.

        :return: Type identifier
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getValueType(self):
        """
        Get the current union sub-type.

        :return: Type
        :rtype: WdbsMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def value(self):
        """
        Get the current value.

        :return: Current value
        :rtype: object | None
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsRecordMetaP(WdbsStructuralMetaP, WdbsMetaP):
    """
    Protocol for record meta-data instances.
    """

    __slots__ = tuple()

    @abstractmethod
    def getType(self):
        """
        :return: Meta-data type
        :rtype: WdbsRecordMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def getValueType(self, key):
        """
        Get the type of the child with the given key.

        :param key: Child key
        :type key: str
        :return: Meta-data type of the child with the given key
        :rtype: WdbsMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def serializationIterator(self):
        """
        :return: Iterable of all key/value-pairs
        :rtype: Iterable[(str, WdbsMetaMP)]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsTupleMetaP(WdbsStructuralMetaP, WdbsMetaP):
    """
    Protocol for tuple meta-data instances.
    """

    __slots__ = tuple()

    @abstractmethod
    def getType(self):
        """
        :return: Meta-data type
        :rtype: WdbsTupleMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def getValueType(self, index):
        """
        Get the type of the child at the given indek.

        :param index: Child index
        :type index: int
        :return: Meta-data type of the child with the given key
        :rtype: WdbsMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def serializationIterator(self):
        """
        :return: Iterable of all key/value-pairs
        :rtype: Iterable[WdbsMetaMP]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsMultiMetaP(WdbsStructuralMetaP, WdbsMetaP):
    """
    Protocol for container meta-data instances.

    All child items of a container instance are of the same type.
    """

    __slots__ = tuple()

    @abstractmethod
    def getValueType(self):
        """
        Get the value type of the children.

        :return: Meta-data type of the children
        :rtype: WdbsMultiMetaTypeP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsListMetaP(WdbsMultiMetaP):
    """
    Protocol for list meta-data instances.
    """

    __slots__ = tuple()

    @abstractmethod
    def getType(self):
        """
        :return: Meta-data type
        :rtype: WdbsListMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def serializationIterator(self):
        """
        :return: Iterable of all key/value-pairs
        :rtype: Iterable[WdbsMetaMP]
        """
        raise NotImplementedError


# noinspection PyAbstractClass
class WdbsDictionaryMetaP(WdbsMultiMetaP):
    """
    Protocol for dictionary/map meta-data instances.
    """

    __slots__ = tuple()

    @abstractmethod
    def getType(self):
        """
        :return: Meta-data type
        :rtype: WdbsDictionaryMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def getKeyType(self):
        """
        Get the type of the keys.

        :return: Meta-data type of the keys
        :rtype: WdbsMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def serializationIterator(self):
        """
        :return: Iterable of all key/value-pairs
        :rtype: Iterable[(WdbsMetaMP, WdbsMetaMP)]
        """
        raise NotImplementedError()
