# ----------------------------------------------------------------------------
# - wbhdb-connection-point-base                                              -
# - wild cruncher: connection point base                     core / bus / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2014-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wbhdb_connection_point_base
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Connection points base classes

.. moduleauthor:: Philippe lambert <lambert@silkan.com>
"""
from abc import abstractmethod

from wcr_core.db.whdb_connection_point import WhdbConnectionPoint
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_NAME
from wcr_core.dbstore.wdbs_meta_description import *
from wcr_core.dbstore.wdbs_meta_type import *
from wcr_core.modules.bus.db.wbhdb_messages_format_meta import \
    WBHDB_MESSAGE_FORMAT_TYPE, WBHDB_FILE_KEY, WBHDB_NAME_KEY


WBHDB_XPOINT_CONFIG_KEY = 'config'
WBHDB_META_CONS_MSG_KEY = 'consumed'
WBHDB_META_PROD_MSG_KEY = 'produced'

class WbhdbConnectionPointBase(WhdbConnectionPoint):

    META = WdbsMetaDescription(
        WhdbConnectionPoint.META,
        WdbsMetaField(
            WBHDB_META_PROD_MSG_KEY,
            WdbsDictionaryMetaType(WDBS_STRING_META,
                                   WBHDB_MESSAGE_FORMAT_TYPE)),
        WdbsMetaField(
            WBHDB_META_CONS_MSG_KEY,
            WdbsDictionaryMetaType(WDBS_STRING_META,
                                   WBHDB_MESSAGE_FORMAT_TYPE)))



    def __init__(self, xpoint, wreg):
        """
        :param xpoint: Raw connection point
        :type xpoint: WdbsConnectionPointP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbConnectionPoint.__init__(self, xpoint, wreg)

    def setName(self, name):
        """
        Set the connection point's name.

        :param name: New name
        :type name: str
        """
        self.rawElement.setMeta(name, WHDB_XPOINT_NAME)

    def _getConfig(self, *keys):
        """
        Get the configuration value for the given key sequence.

        :param keys: Keys
        :type keys: tuple[object]
        :return: Configuration value
        :rtype: object
        """
        return self.rawElement.getMeta(WBHDB_XPOINT_CONFIG_KEY, *keys)

    def _setConfig(self, value, *keys):
        """
        Set the configuration value for the given key sequence.

        :param value: New configuration value
        :type value: object
        :param keys: Keys
        :type keys: tuple[object]
        """
        self.rawElement.setMeta(value, WBHDB_XPOINT_CONFIG_KEY, *keys)

    @abstractmethod
    def getNamePrefix(self):
        """
        Get the prefix for connection point names.

        :return: Name prefix
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getTypeName(self):
        raise NotImplementedError()

    @abstractmethod
    def getDescription(self, **kwargs):
        raise NotImplementedError()

    def getProducedMessages(self):
        lmodk=sorted(self.rawElement.getMeta(WBHDB_META_PROD_MSG_KEY).keys())
        lpmes=[]
        for modk in lmodk:
            lmesk=sorted(self.rawElement.getMeta(WBHDB_META_PROD_MSG_KEY,
                                                 modk).\
                         keys())
            for mesk in lmesk:
                pmes = (self.rawElement.getMeta(WBHDB_META_PROD_MSG_KEY,
                                                modk, mesk, WBHDB_FILE_KEY),
                        modk,
                        self.rawElement.getMeta(WBHDB_META_PROD_MSG_KEY,
                                                modk, mesk, WBHDB_NAME_KEY),
                        )
                lpmes.append(pmes)
        return lpmes

    def getConsumedMessages(self):
        lmodk=sorted(self.rawElement.getMeta(WBHDB_META_CONS_MSG_KEY).keys())
        lcmes=[]
        for modk in lmodk:
            lmesk=sorted(self.rawElement.getMeta(WBHDB_META_CONS_MSG_KEY,
                                                 modk).\
                         keys())
            for mesk in lmesk:
                pmes = (
                    self.rawElement.getMeta(WBHDB_META_CONS_MSG_KEY,
                                            modk, mesk, WBHDB_FILE_KEY),
                    modk,
                    self.rawElement.getMeta(WBHDB_META_CONS_MSG_KEY,
                                            modk, mesk, WBHDB_NAME_KEY),
                        )
                lcmes.append(pmes)
        return lcmes

    def getAllTransitMessages(self):
        return self.getProducedMessages(), self.getConsumedMessages()

    def removeMessageBind(self, msg, prod):
        statk = WBHDB_META_PROD_MSG_KEY if prod else WBHDB_META_CONS_MSG_KEY
        self.rawElement.removeMeta(statk,
                                   msg[2],
                                   self.formatMessageKey((msg[0], msg[1])))

    def addMessageBind(self, msg, prod):
        """
        :param msg: structur to save message
                            (file, message_name, module)
        :type msg: (str, str)
        :param prod: is producer of consumer
        :type prod: boolean
        """
        newkey = self.formatMessageKey((msg[0], msg[1]))
        if prod:
            self.rawElement.setMeta({WBHDB_FILE_KEY: msg[0],
                                     WBHDB_NAME_KEY: msg[3]},
                                    WBHDB_META_PROD_MSG_KEY, msg[2], newkey)
        if not prod:
            self.rawElement.setMeta({WBHDB_FILE_KEY: msg[0],
                                     WBHDB_NAME_KEY: msg[3]},
                                    WBHDB_META_CONS_MSG_KEY, msg[2], newkey)

    def formatMessageKey(self, datamessage):
        """
        :param datamessage: dhe data used to format key
        :type (str,str)
        """
        return datamessage[0]+' - '+datamessage[1]


class WbhdbConnectionPointSimpleBase(WbhdbConnectionPointBase):
    def __init__(self, tnam, prefix, xpoint, wreg):
        """
        :param tnam: Connection point type name
        :type tnam: str
        :param prefix: Connection point name prefix
        :type prefix: str
        :param xpoint: Raw connection point
        :type xpoint: WdbsConnectionPointP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WbhdbConnectionPointBase.__init__(self, xpoint, wreg)

        #: Connection point type name
        self.__xpntTypeName = tnam
        #: Connection point name prefix
        self.__prefix = prefix

    def getNamePrefix(self):
        return self.__prefix

    def getTypeName(self):
        return '{} connection point'.format(self.__xpntTypeName)

    def getDescription(self, **kwargs):
        return '{n} ({t})'.format(n=self.getName(), t=self.__xpntTypeName)
