# ----------------------------------------------------------------------------
# - wbhdb-bus-instance                                                       -
# - wild cruncher: bus instances                             core / bus / db -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wbhdb_bus_instance
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Logical Bus instance

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_DIRECTION, \
    WHDB_XPOINT_NAME
from wcr_core.db.whdb_instance import WhdbOperatorInstance
from wcr_core.dbstore.wdbs_meta_description import *
from wcr_core.dbstore.wdbs_meta_type import *
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.bus.db.wbhdb_messages_format_meta import \
    WBHDB_MESSAGE_BASE, WBHDB_FILE_KEY, WBHDB_NAME_KEY
from wcr_core.modules.bus.db.wbhdb_standard_operators import WBHDB_BUS, \
    WBHDB_BUS_CODE
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_XPOINT_CETRAC_CODE

WPHDB_XPNT_NAME_PATTERN = 'C{id}'

WBHDB_META_BUS_KEY = 'bustype'
WBHDB_META_SUB_BUS_KEY = 'sbubustype'
WBHDB_META_EXT_MODULE_KEY="BBmessages_hosted"

WBHDB_SIMU_SUB_TYPE = {'AFDX': 'adfx',
                       'Ethernet': 'ethernet',
                       WHHDB_XPOINT_CETRAC_CODE: 'cetrac',
                       'ARINC429': 'arinc429',
                       '1553': '1553',
                       'SPI': 'spi'}

WBHDB_LOGI_SUB_TYPE = {}

BUS_TYPES = {'Simulated': ('simu', WBHDB_SIMU_SUB_TYPE),
             'Logic': ('logi', WBHDB_LOGI_SUB_TYPE)}


class WbhdbLogiBusInstance(WhdbOperatorInstance):
    """
    Operator-instance for bus instances.
    """

    META = WdbsMetaDescription(
                WhdbOperatorInstance.META,
                WdbsMetaField(
                    WBHDB_META_BUS_KEY, WDBS_STRING_META),
                WdbsMetaField(
                    WBHDB_META_SUB_BUS_KEY, WDBS_STRING_META),
                WdbsMetaField(
                    WBHDB_META_EXT_MODULE_KEY,
                    WdbsDictionaryMetaType(WDBS_STRING_META,
                                           WBHDB_MESSAGE_BASE))
        )

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        assert inst.getOperator() == WBHDB_BUS_CODE
        WhdbOperatorInstance.__init__(self, WBHDB_BUS, inst, wreg)

    def getTypes(self):
        return sorted(BUS_TYPES.keys())

    def getType(self):
        st=self.rawElement.getMeta(WBHDB_META_BUS_KEY)
        if st == "":
            return False
        for k in BUS_TYPES.keys():
            if BUS_TYPES[k][0] == st:
                return k
        return False

    def setType(self, type):
        if type in BUS_TYPES.keys():
            self.rawElement.setMeta(BUS_TYPES[type][0], WBHDB_META_BUS_KEY)
        else:
            raise Exception("Trying to set an Unknown type for Logical Bus "
                            "=> {}".format(type))

    def setSubType(self, stype):
        bustype = self.getType()
        if not stype:
            return
        if not bustype:
            raise Exception("No Type set => No sub Type can be set")
        if stype not in BUS_TYPES[bustype][1].keys():
            raise Exception("Invalide SubType :", stype)
        self.rawElement.setMeta(BUS_TYPES[bustype][1][stype],
                                WBHDB_META_SUB_BUS_KEY)

    def getSubType(self):
        st = self.rawElement.getMeta(WBHDB_META_SUB_BUS_KEY)
        if st == "":
            return False
        for k in BUS_TYPES[self.getType()][1].keys():
            if BUS_TYPES[self.getType()][1][k] == st:
                return k
        return False

    def getSubTypeforType(self, btype):
        if not btype in BUS_TYPES.keys():
            raise Exception("Unknown type given for subtype : {}".format(btype))
        styped = BUS_TYPES[btype][1]
        stl = []
        for k in styped.keys():
            stl.append(k)
        return sorted(stl)

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        if WHDB_XPOINT_DIRECTION not in kwargs:
            kwargs[WHDB_XPOINT_DIRECTION] = WDBS_XPNT_BI
        return super()._createRawConnectionPoint(ctyp, **kwargs)

    def createConnectionPoint(self, ctyp=None, **kwargs):
        return super().createConnectionPoint(
            ctyp, **_wphdb_setConnectorMeta(self, **kwargs))

    def getMessages(self):
        lmk = sorted(self.rawElement.getMeta(WBHDB_META_EXT_MODULE_KEY).keys())
        lm = []
        for k in lmk:
            mt = (self.rawElement.getMeta(WBHDB_META_EXT_MODULE_KEY,
                                          k, WBHDB_FILE_KEY),
                  self.rawElement.getMeta(WBHDB_META_EXT_MODULE_KEY,
                                        k, WBHDB_NAME_KEY))
            lm.append(mt)
        return lm

    def addMessage(self,message):
        """
        :param message: structur to save message
                        (file, messagename)
        :type message: (str, str)
        """
        newkey=self.formatMessageKey((message[0],message[1]))
        self.rawElement.setMeta({WBHDB_FILE_KEY:message[0],
                                 WBHDB_NAME_KEY:message[1]},
                                WBHDB_META_EXT_MODULE_KEY, newkey)

    def removeAllMessages(self):
        lm = self.rawElement.getMeta(WBHDB_META_EXT_MODULE_KEY).keys()
        for messagekey in lm:
            self.rawElement.removeMeta(WBHDB_META_EXT_MODULE_KEY, messagekey)

    def formatMessageKey(self, datamessage):
        """
        :param datamessage: dhe data used to format key
        :type (str,str)
        """
        return datamessage[0]+' - '+datamessage[1]


def _wphdb_setConnectorMeta(inst, **kwargs):
    """
    Ensure that all meta-data for a new connection point is set correctly.

    :param inst: Instance to which the connection point will be added
    :type inst: WhdbInstanceP
    :param kwargs: Original arguments
    :type kwargs: dict[str, object]
    :return: Updated arguments
    :rtype: dict[str, object]
    """
    kwargs[WHDB_XPOINT_DIRECTION] = WDBS_XPNT_BI
    if not kwargs.get(WHDB_XPOINT_NAME):
        kwargs[WHDB_XPOINT_NAME] = _wphdb_getNewXPointName(inst)
    return kwargs


def _wphdb_getNewXPointName(inst):
    """
    Get the name for the connection point added next to the given instance.

    :param inst: Instance to which the connection point will be added
    :type inst: WhdbInstanceP
    :return: Name for the new connection point
    :rtype: str
    """
    names = set(c.getName() for c in inst.getConnectionPoints())
    cnid = 1
    cname = WPHDB_XPNT_NAME_PATTERN.format(id=cnid)
    while cname in names:
        cnid += 1
        cname = WPHDB_XPNT_NAME_PATTERN.format(id=cnid)
    return cname
