# ----------------------------------------------------------------------------
# - whhdb-PCI-card-instance                                                  -
# - wild cruncher: PCI card instances                   core / hardware / db -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whhdb_PCI_card_instance
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - PCI Card instance

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""
from wcr_core.dbstore.wdbs_meta_description import *
from wcr_core.dbstore.wdbs_meta_type import *
from wcr_core.modules.hardware.db.whhdb_standard_operators import \
    WHHDB_HOST, WHHDB_HOST_CODE
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_DIRECTION, \
    WHDB_XPOINT_NAME
from uuid import UUID
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.hardware.db.whhdb_generic_Host_instance\
    import WhhdbGenericHostInstance
from wcr_core.modules.hardware.db.whhdb_Host_instance_meta import \
    WHHDB_HOST_META, WHHDB_META_CTYPE_KEY, WHHDB_CARD_PORTS_KEY, \
    WHHDB_META_IOCTYPE_KEY, WHHDB_CARD_TYPE_KEY, WHHDB_META_STATUS_KEY, \
    WHHDB_META_BIND_KEY, WHHDB_META_SKEW_KEY
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_XPOINT_NAMES, WHHDB_XPOINT_DEFAULT_CODE, WHHDB_MODULES_IO_NAMES
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError


class WhhdbHostInstance(WhhdbGenericHostInstance):
    """
    Operator-instance for bus instances.
    """
    WHHDB_HOST_EXTERNAL_KEY = 'External'
    WHHDB_HOST_CETRAC_KEY = 'CetraCPCI'
    TYPES = [WHHDB_HOST_EXTERNAL_KEY, WHHDB_HOST_CETRAC_KEY]
    IO_TYPES = WHHDB_MODULES_IO_NAMES
    STATUS_TYPES = ['Yes', 'No']

    META = WHHDB_HOST_META

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        assert inst.getOperator() == WHHDB_HOST_CODE
        WhhdbGenericHostInstance.__init__(self, WHHDB_HOST, inst, wreg)
        self.IOCARD_TO_REMOVE = []

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        if WHDB_XPOINT_DIRECTION not in kwargs:
            kwargs[WHDB_XPOINT_DIRECTION] = WDBS_XPNT_BI
        return super()._createRawConnectionPoint(ctyp, **kwargs)

    def getPCITypes(self):
        return super().getTypes()

    def getPCICards(self):
        """
        return every non marked card
        """
        return super().getCards()

    def setPCICard(self, card):
        """
        initialising a new card or remove from deleted if allready exist
        """
        super().setCard(card)

    def getPCICardType(self, cardname):
        return super().getCardType(cardname)

    def getPCICardPtpMaster(self, cardname):
        return super().getCardPtpMaster(cardname)

    def getPCICardStatus(self, cardname):
        return super().getCardStatus(cardname)

    def getPCICardMaxSkew(self, cardname):
        if not cardname in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            return ''
        maxSkew = self.rawElement.getMeta(WHHDB_META_CTYPE_KEY, cardname,
                                       WHHDB_META_SKEW_KEY)

        return maxSkew

    def setPCICardType(self, card, type):
        super().setCardType(card, type)

    def setPCICardPtpMaster(self, card, status):
        super().setCardPtpMaster(card, status)

    def setPCICardStatus(self, card, status):
        super().setCardStatus(card, status)

    def setPCICardMaxSkew(self, card, maxSkew):
        if card not in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            return []
        return self.rawElement.setMeta(maxSkew,
                                       WHHDB_META_CTYPE_KEY, card,
                                       WHHDB_META_SKEW_KEY)

    def replaceCard(self, hcardname, ncard):
        """
        replace the card hcardname by the ncard.
        :param hcardname: name ot the card ho will be replaced
        :type hcardname: str
        :param ncard: description of the full new card
        :type ncard: [cardname, cardtype, ptp, status [portsbinded]]
        """
        super().replaceCard(hcardname, ncard)
        self.setPCICardStatus(ncard[0], ncard[3])
        self.setPCICardMaxSkew(ncard[0], ncard[4])
        for puuid in ncard[5]:
            self.addPortUUIDToCard(ncard[0], puuid)

    def getBindedPorts(self, cardname):
        """
        Return the list of binded port to cardname
        """
        if not cardname in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            return []
        return self.rawElement.getMeta(WHHDB_META_CTYPE_KEY, cardname,
                                       WHHDB_CARD_PORTS_KEY)

    def addPortUUIDToCard(self, cardname, cuuid):
        """
        add cuuid from the list of binded port to cardname
        """
        if not cardname in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            return None
        buuid=self.getBindedPorts(cardname)
        if buuid:
            if cuuid is list:
                for elem in cuuid:
                    if not isinstance(elem, UUID):
                        raise ValueError(
                            "Every element of the list must be UUIDs")
            buuid.append(cuuid)
        else:
            if cuuid is list:
                for elem in cuuid:
                    if not isinstance(elem, UUID):
                        raise ValueError(
                            "Every element of the list must be UUIDs")
                buuid=cuuid
            else:
                if not isinstance(cuuid, UUID):
                    raise ValueError(
                        "Invalid port reference Cards "
                        "and port are binded by UUID")
                buuid=[cuuid]
        self.rawElement.setMeta(buuid, WHHDB_META_CTYPE_KEY, cardname,
                                WHHDB_CARD_PORTS_KEY)

    def removeCardPort(self, cardname, cuuid):
        """
        remove cuuid from the list of binded port to cardname
        """
        if not cardname in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            return None
        buuid = self.getBindedPorts(cardname)

        if cuuid in buuid:
            buuid.remove(cuuid)

        self.rawElement.setMeta(buuid, WHHDB_META_CTYPE_KEY, cardname,
                                WHHDB_CARD_PORTS_KEY)

    def getPCICardsDescription(self):
        """
        retrn the full description of the card stocked in metadata
        :return : [cardname, cardtyp, [bindedports], status]
        """
        fullcards=[]

        for card in self.getPCICards():
            structcards=(card,
                         self.getPCICardType(card),
                         self.getBindedPorts(card),
                         self.getPCICardPtpMaster(card),
                         self.getPCICardStatus(card),
                         self.getPCICardMaxSkew(card))
            fullcards.append(structcards)
        return fullcards

    def getCardConnectionPoints(self, card):
        """
        git every connectionpoint binded to card. If card is None return
        every connexion point of Host.
        :return : [ConnectionPoint]
        """
        if not isinstance(card, str):
            return []
            # return self.getConnectionPoints()
        cbpl = self.getBindedPorts(card)
        cpl = self.getConnectionPoints()
        res = []
        for cp in cpl:
            if cp.getUUID() in cbpl:
                res.append(cp)
        return res

    def bindFreePorts(self, card):
        """
        Bind to card every connection who isn't allready binded.
        """
        if not card:
            return None
        bcpuuidl = []
        # get port binded in regulare card
        for c in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            for cpuuid in self.getBindedPorts(c):
                bcpuuidl.append(cpuuid)
        iobcpuuidl = []
        # get port binded in io card
        for c in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            for cpuuid in self.getBindedIoPorts(c):
                iobcpuuidl.append(cpuuid)
        cpl = self.getConnectionPoints()
        for cp in cpl:
            if cp.getUUID() not in bcpuuidl and cp.getUUID() not in iobcpuuidl:
                self.addPortUUIDToCard(card,cp.getUUID())

    def unbindInexistentPort(self):
        """
        unbind every UUID of cards port list if they don't
        correspond to a connection point
        """
        cpl = self.getConnectionPoints()
        cpuuidl=[]
        for cp in cpl:
            cpuuidl.append(cp.getUUID())
        for c in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            for cpuuid in self.getBindedPorts(c):
                if not cpuuid in cpuuidl:
                    self.removeCardPort(c,cpuuid)

    def markPCICard(self, cardn):
        """
        mark cardn as card to remove
        """
        super().markCard(cardn)

    def getConnectionDistributionTypeByUUID(self, uuidl):
        """
        return type distribution of ports having UUID in uuidl
        :return : {str: int}
        """
        distrib={}
        if not isinstance(uuidl, list):
            return distrib
        # test=[x.getUUID() for x in self.getConnectionPoints()]
        for c in self.getConnectionPoints():
            if c.getUUID() in set(uuidl):
                if c.getCode() in WHHDB_XPOINT_NAMES:
                    tn = c.getCode()
                else:
                    raise Exception('Connector type is unknown !')
                if not tn in distrib.keys():
                    distrib[tn]=1
                else:
                    distrib[tn]+=1
        return distrib

    def delCardandPorts(self):
        """
        Remove card marked to remove and they binded ports
        """
        dcpl=[]
        for rmcn in self.CARD_TO_REMOVE:
            cpl=self.getCardConnectionPoints(rmcn)
            for cp in cpl:
                dcpl.append(cp)
        for crds in self.getPCICards():
            cpl = self.getCardConnectionPoints(crds)
            for cp in cpl:
                if cp in dcpl:
                    dcpl.remove(cp)
        for dcp in dcpl:
            self.removeConnectionPoint(dcp)
        self.delCards()

    def checkPortNumberConsistency(self):

        for card in self.getCards():
            portNumbers = []
            for point in self.getCardConnectionPoints(card):
                portNum = point.getPortNum()
                if portNum in portNumbers:
                    msg = "Port consistency error, " \
                          "multiple ports have the same port number " \
                          "for the same card.\n" \
                          "Please check port {}.".format(point.getName())
                    return False, msg
                portNumbers.append(portNum)
            modulesIo = self.getModuleIoFromCard(card)
            for moduleIo in modulesIo:
                pointsIo = self.getIoCardConnectionPoints(moduleIo)
                for pointIo in pointsIo:
                    portIoNum = pointIo.getPortNum()
                    if portIoNum in portNumbers:
                        msg = "Port consistency error, " \
                              "multiple ports have the same port number "\
                              "for the same card.\n" \
                              "Please check port {}.".format(point.getName())
                        return False, msg
                    portNumbers.append(portIoNum)
        return True, None

    def getIoTypes(self):
        return self.IO_TYPES

    def getIoCards(self):
        """
        return every non marked card
        """
        cards = sorted(self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys())
        for rmc in self.IOCARD_TO_REMOVE:
            cards.remove(rmc)
        return cards

    def setIoCard(self, card):
        """
        initialising a new card or remove from deleted if already exist
        """
        if not card in \
                self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            if not isNameValid(card):
                showErrorWindow("Name Error", NameFormatMessageError,
                                card)
            self.rawElement.setMeta({WHHDB_CARD_PORTS_KEY: []},
                                    WHHDB_META_IOCTYPE_KEY, card)
        else:
            if card in self.CARD_TO_REMOVE:
                self.CARD_TO_REMOVE.remove(card)

    def getModuleIoFromCard(self, cardname):
        modules = []
        for module in self.getIoCards():
            cardN = self.getIoCardBind(module)
            if cardN == cardname:
                modules.append(module)
        return modules

    def getIoCardBind(self, moduleIo):
        if not moduleIo in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            return ''
        bind = self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY, moduleIo,
                                       WHHDB_META_BIND_KEY)
        return bind

    def setIoCardBind(self, card, bind):
        if not card in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            return None
        self.rawElement.setMeta(bind,
                                WHHDB_META_IOCTYPE_KEY, card,
                                WHHDB_META_BIND_KEY)

    def setIoCardPtpMaster(self, card, status):
        return super().setCardPtpMaster(card, status)

    def replaceIoCard(self, hcardname, ncard):
        """
        replace the card hcardname by the ncard.
        :param hcardname: name ot the card ho will be replaced
        :type hcardname: str
        :param ncard: description of the full new card
        :type ncard: [cardname, cardtype, [portsbinded]]
        """
        if hcardname is not '' or hcardname is not None:
            self.rawElement.removeMeta(WHHDB_META_IOCTYPE_KEY, hcardname)
        self.setIoCard(ncard[0])
        self.setIoCardBind(ncard[0], ncard[1])
        for puuid in ncard[2]:
            self.addPortUUIDToIoCard(ncard[0], puuid)

    def getBindedIoPorts(self, cardname):
        """
        Return the list of binded port to cardname
        """
        if not cardname in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            return []
        return self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY, cardname,
                                       WHHDB_CARD_PORTS_KEY)

    def addPortUUIDToIoCard(self, cardname, cuuid):
        """
        add cuuid from the list of binded port to cardname
        """
        if not cardname in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            return None
        buuid = self.getBindedIoPorts(cardname)
        if buuid:
            if cuuid is list:
                for elem in cuuid:
                    if not isinstance(elem, UUID):
                        raise ValueError(
                            "Every element of the list must be UUIDs")
            buuid.append(cuuid)
        else:
            if cuuid is list:
                for elem in cuuid:
                    if not isinstance(elem, UUID):
                        raise ValueError(
                            "Every element of the list must be UUIDs")
                buuid = cuuid
            else:
                if not isinstance(cuuid, UUID):
                    raise ValueError(
                        "Invalid port reference Cards "
                        "and port are binded by UUID")
                buuid = [cuuid]
        self.rawElement.setMeta(buuid, WHHDB_META_IOCTYPE_KEY, cardname,
                                WHHDB_CARD_PORTS_KEY)

    def removeIoCardPort(self, cardname, cuuid):
        """
        remove cuuid from the list of binded port to cardname
        """
        if not cardname in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            return None
        buuid = self.getBindedIoPorts(cardname)

        if cuuid in buuid:
            buuid.remove(cuuid)

        self.rawElement.setMeta(buuid, WHHDB_META_IOCTYPE_KEY, cardname,
                                WHHDB_CARD_PORTS_KEY)

    def getIoCardsDescription(self):
        """
        retrn the full description of the card stocked in metadata
        :return : [cardname, cardtyp, [bindedports]]
        """
        fullcards = []
        for card in self.getIoCards():
            structcards = (card,
                           self.getIoCardBind(card),
                           self.getBindedIoPorts(card))
            fullcards.append(structcards)
        return fullcards

    def getIoCardConnectionPoints(self, card):
        """
        git every connectionpoint binded to card. If card is None return
        every connexion point of Host.
        :return : [ConnectionPoint]
        """
        if not isinstance(card, str):
            return []
        cbpl = self.getBindedIoPorts(card)
        cpl = self.getConnectionPoints()
        res = []
        for cp in cpl:
            if cp.getUUID() in cbpl:
                res.append(cp)
        return res

    def getAllIoConnectionPoints(self):
        ioCards = self.getIoCards()
        IoCardsPoints = []
        for card in ioCards:
            IoCardsPoints.extend(self.getIoCardConnectionPoints(card))

        return IoCardsPoints

    def bindFreeIoPorts(self, card):
        """
        Bind to card every connection who isn't allready binded.
        """
        if not card:
            return None
        iobcpuuidl = []
        #get port binded in io card
        for c in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            for cpuuid in self.getBindedIoPorts(c):
                iobcpuuidl.append(cpuuid)
        bcpuuidl = []
        # get port binded in regular card
        for c in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            for cpuuid in self.getBindedPorts(c):
                bcpuuidl.append(cpuuid)
        cpl = self.getConnectionPoints()
        for cp in cpl:
            if cp.getUUID() not in bcpuuidl and cp.getUUID() not in iobcpuuidl:
                self.addPortUUIDToIoCard(card, cp.getUUID())

    def getIoConnectionDistributionTypeByUUID(self, uuidl):
        """
        return type distribution of ports having UUID in uuidl
        :return : {str: int}
        """
        distrib={}
        if not isinstance(uuidl, list):
            return distrib
        for c in self.getConnectionPoints():
            if c.getUUID() in set(uuidl):
                if c.getCode() in self.IO_TYPES:
                    tn = c.getCode()
                else:
                    raise Exception('Connector type is unknown !')
                if not tn in distrib.keys():
                    distrib[tn] = 1
                else:
                    distrib[tn] += 1
        return distrib

    def unbindInexistentIoPort(self):
        """
        unbind every UUID of cards port list if they don't
        correspond to a connection point
        """
        cpl = self.getConnectionPoints()
        cpuuidl = []
        for cp in cpl:
            cpuuidl.append(cp.getUUID())
        for c in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            for cpuuid in self.getBindedIoPorts(c):
                if not cpuuid in cpuuidl:
                    self.removeIoCardPort(c, cpuuid)

    def markIoCard(self, cardn):
        """
        mark cardn as card to remove
        """
        if cardn in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            self.IOCARD_TO_REMOVE.append(cardn)

    def delIoCardandPorts(self):
        """
        Remove card marked to remove and they binded ports
        """
        dcpl = []
        for rmcn in self.IOCARD_TO_REMOVE:
            cpl = self.getIoCardConnectionPoints(rmcn)
            for cp in cpl:
                dcpl.append(cp)
        for crds in self.getIoCards():
            cpl = self.getIoCardConnectionPoints(crds)
            for cp in cpl:
                if cp in dcpl:
                    dcpl.remove(cp)
        for dcp in dcpl:
            self.removeConnectionPoint(dcp)
        self.delIoCards()

    def delIoCards(self):
        """
        cleaning metada from card marked to remove and asociated information
        """
        portl = []
        for c in self.IOCARD_TO_REMOVE:
            getCon = getattr(self, 'getCardConnectionPoints', None)
            if getCon:
                portl += getCon(c)
            self.rawElement.removeMeta(WHHDB_META_IOCTYPE_KEY, c)
        self.IOCARD_TO_REMOVE = []
