# ----------------------------------------------------------------------------
# - whhdb-connection-point-base                                              -
# - wild cruncher: connection point base                core / hardware / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2014-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wbhdb_connection_point_base
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Connection points base classes

.. moduleauthor:: Philippe lambert <lambert@silkan.com>
"""
from abc import abstractmethod

from wcr_core.db.whdb_connection_point import WhdbConnectionPoint
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_NAME
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError


WHHDB_XPOINT_CONFIG_KEY = 'config'
WHHDB_XPOINT_DISCRET_KEY = 'discrete'
WHHDB_XPOINT_FIBER_CHANNEL_KEY = 'fiberchannel'
WHHDB_XPOINT_1553_KEY = '1553'
WHHDB_XPOINT_ARINC825_KEY = 'arinc825'
WHHDB_XPOINT_AFDX_KEY = 'afdx'


class WhhdbConnectionPointBase(WhdbConnectionPoint):

    def __init__(self, xpoint, wreg):
        """
        :param xpoint: Raw connection point
        :type xpoint: WdbsConnectionPointP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbConnectionPoint.__init__(self, xpoint, wreg)

    def setName(self, name):
        """
        Set the connection point's name.

        :param name: New name
        :type name: str
        """
        if isNameValid(name):
            self.rawElement.setMeta(name, WHDB_XPOINT_NAME)
        else:
            showErrorWindow("Name Error", NameFormatMessageError, name)

    def _getConfig(self, *keys):
        """
        Get the configuration value for the given key sequence.

        :param keys: Keys
        :type keys: tuple[object]
        :return: Configuration value
        :rtype: object
        """
        return self.rawElement.getMeta(WHHDB_XPOINT_CONFIG_KEY, *keys)

    def _setConfig(self, value, *keys):
        """
        Set the configuration value for the given key sequence.

        :param value: New configuration value
        :type value: object
        :param keys: Keys
        :type keys: tuple[object]
        """
        self.rawElement.setMeta(value, WHHDB_XPOINT_CONFIG_KEY, *keys)

    def _setConfigDiscret(self, value, *keys):
        """
        Set the configuration for discret value for the given key sequence.

        :param value: New configuration value
        :type value: object
        :param keys: Keys
        :type keys: tuple[object]
        """
        self.rawElement.setMeta(value, WHHDB_XPOINT_DISCRET_KEY, *keys)

    def _getConfigDiscret(self, *keys):
        """
        Get the configuration for discret value for the given key sequence.

        :param keys: Keys
        :type keys: tuple[object]
        :return: Configuration value
        :rtype: object
        """
        return self.rawElement.getMeta(WHHDB_XPOINT_DISCRET_KEY, *keys)

    def _setConfigAfdx(self, value, *keys):
        """
        Set the configuration for AFDX value for the given key sequence.

        :param value: New configuration value
        :type value: object
        :param keys: Keys
        :type keys: tuple[object]
        """
        self.rawElement.setMeta(value, WHHDB_XPOINT_AFDX_KEY, *keys)

    def _getConfigAfdx(self, *keys):
        """
        Get the configuration for AFDX value for the given key sequence.

        :param keys: Keys
        :type keys: tuple[object]
        :return: Configuration value
        :rtype: object
        """
        return self.rawElement.getMeta(WHHDB_XPOINT_AFDX_KEY, *keys)

    def _isDiscretEmpty(self):
        """
        Check if there are any discret data.

        :return: bool
        """
        items = self.rawElement.getMeta(WHHDB_XPOINT_DISCRET_KEY)
        for value in items.values():
            if value is not '':
                return False

        return True

    def _setConfigFiberChannel(self, value, *keys):
        """
        Set the configuration for fiber channel value for the given key sequence.

        :param value: New configuration value
        :type value: object
        :param keys: Keys
        :type keys: tuple[object]
        """
        self.rawElement.setMeta(value, WHHDB_XPOINT_FIBER_CHANNEL_KEY, *keys)

    def _getConfigFiberChannel(self, *keys):
        """
        Get the configuration for fiber channel value for the given key sequence.

        :param keys: Keys
        :type keys: tuple[object]
        :return: Configuration value
        :rtype: object
        """
        return self.rawElement.getMeta(WHHDB_XPOINT_FIBER_CHANNEL_KEY, *keys)

    def _isFiberChannelEmpty(self):
        """
        Check if there are any Fiber Channel data.

        :return: bool
        """

        items = self.rawElement.getMeta(WHHDB_XPOINT_FIBER_CHANNEL_KEY)
        for value in items.values():
            if value is not '':
                return False
        return True

    def _setConfig1553(self, value, *keys):
        """
        Set the configuration for 1553 value for the given key sequence.

        :param value: New configuration value
        :type value: object
        :param keys: Keys
        :type keys: tuple[object]
        """
        self.rawElement.setMeta(value, WHHDB_XPOINT_1553_KEY, *keys)

    def _getConfig1553(self, *keys):
        """
        Get the configuration for 1553 value for the given key sequence.

        :param keys: Keys
        :type keys: tuple[object]
        :return: Configuration value
        :rtype: object
        """
        return self.rawElement.getMeta(WHHDB_XPOINT_1553_KEY, *keys)

    def _setConfigARINC825(self, value, *keys):
        """
        Set the configuration for ARINC825 value for the given key sequence.

        :param value: New configuration value
        :type value: object
        :param keys: Keys
        :type keys: tuple[object]
        """
        self.rawElement.setMeta(value, WHHDB_XPOINT_ARINC825_KEY, *keys)

    def _getConfigARINC825(self, *keys):
        """
        Get the configuration for ARINC825 value for the given key sequence.

        :param keys: Keys
        :type keys: tuple[object]
        :return: Configuration value
        :rtype: object
        """
        return self.rawElement.getMeta(WHHDB_XPOINT_ARINC825_KEY, *keys)

    @abstractmethod
    def getNamePrefix(self):
        """
        Get the prefix for connection point names.

        :return: Name prefix
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getTypeName(self):
        raise NotImplementedError()

    @abstractmethod
    def getDescription(self, **kwargs):
        raise NotImplementedError()

    def getCode(self, **kwargs):
        raise NotImplementedError()
