# ----------------------------------------------------------------------------
# - wphdb-connection-point-manager                                           -
# - wild cruncher: connection point manager               core / system / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wphdb_connection_point_manager
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher connection point manager

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db.whdb_operator import whdb_makeConnXYPositions
from wcr_core.db.whdb_connection_point_manager import \
    WhdbConnectionPointManagerBase, WHDB_XPOINT_DIRECTION, \
    WHDB_XPOINT_POSITION, WHDB_XPOINT_TYPE, WHDB_XPOINT_NAME, \
    WHDB_XPOINT_FACING, WhdbXPointDesc, WhdbFixedConnectionPointManager, \
    WHDB_DYN_XPOINT_VAR
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.hardware.db.whhdb_connection_points \
    import WHHDB_XPOINT_CETRAC_CODE, WHHDB_XPOINT_DEFAULT_CODE
import re
from wcr_core.db.whdb_operator import whdb_makeConnPositions
from operator import methodcaller
from itertools import repeat, groupby


def _whhdb_getNextXPointName(clst, cdir, prfx=None, rgxp=None):
    """
    Get the next name for a connection point of the given direction for the
    given connection point list.

    :param clst: Connection point list
    :type clst: list[WhdbConnectionPointP]
    :param cdir: Direction
    :type cdir: int
    :param prfx: Name prefix (value will be determined based on cdir if `None`)
    :type prfx: str | None
    :param rgxp: Regular expression to extract names
    :type rgxp: SRE_Pattern | None
    :return: Next name
    :rtype: str
    """
    if prfx is None:
        prfx = 'c'
    if rgxp is None:
        rgxp = re.compile(prfx + '\s*(\d+)$')
    cnmm = 0
    _clst = [c for c in clst]
    for c in _clst:
        m = rgxp.search(c.getName())
        if m:
            cnmm = max(int(m.group(1)), cnmm)
    return '{p}{c}'.format(p=prfx, c=cnmm + 1)


def _whhdb_resetXPointPositions(clst):
    """
    Reset the positions of the connection points in the given list.

    :param clst: Connection point list
    :type clst: list[WhdbConnectionPointP]
    """
    keyf = methodcaller('getDirection')
    clst = list(sorted(clst, key=keyf))
    for cdir, xpnts in groupby(clst, keyf):
        xpnts = list(xpnts)
        for xpnt, (x, y, f) in zip(
                xpnts, whdb_makeConnXYPositions(cdir, len(xpnts), clst)):
            xpnt.setPosition(x, y)
            xpnt.setGravity(f)


class WhhdbConnectionPointManager(WhdbConnectionPointManagerBase):
    """
    Connection point manager for states.
    """

    def __init__(self, xpntTypes=None):
        self.__xpntTypes = tuple(xpntTypes) if xpntTypes is not None else None

    def getInitialXPointSet(self):
        return []

    def getDynamicXPoints(self):
        return self.__xpntTypes

    def canBatchCreateXPoint(self, xpnts, **kwargs):
        return True

    def canInstanceCreateXPoint(self, clst, **kwargs):
        return True

    def canBatchRemoveXPoint(self, xpnts, **kwargs):
        return True

    def canInstanceRemoveXPoint(self, clst, xpnt):
        """
        :type clst: list[WhdbConnectionPointP]
        :type xpnt: dict[str, object]
        :rtype: bool
        """
        return xpnt in clst

    def createInstanceXPoint(self, clst, connF, **kwargs):
        assert WHDB_XPOINT_TYPE in kwargs
        assert WHDB_XPOINT_DIRECTION in kwargs

        xdsc = whhdb_createXPointDesc(
            ctyp=kwargs.get(WHDB_XPOINT_TYPE),
            cdir=kwargs.get(WHDB_XPOINT_DIRECTION),
            cpos=kwargs.get(WHDB_XPOINT_POSITION),
            fcng=kwargs.get(WHDB_XPOINT_FACING))
        xpnt = connF(xdsc)
        xpnt.setMeta(kwargs.get(WHDB_XPOINT_NAME) or str(xpnt.getUUID()),
                     WHDB_XPOINT_NAME)
        return xpnt

    def removeInstanceXPoint(self, clst, xpnt):
        clst.remove(xpnt)

    def setInstanceXPoints(self, clst, connF, data):
        raise Exception('Cannot set connection points')

    def areValidXPoints(self, clst):
        return True, None

    def resetXPointPositions(self, clst):
        if not isinstance(clst, list):
            clst = list(clst)
        for xpnt, (x, y, f) in zip(
                clst, whdb_makeConnXYPositions(None, len(clst), clst)):
            xpnt.setPosition(x, y)
            xpnt.setGravity(f)


def whhdb_createXPointDesc(ctyp, cdir, name=None, cpos=None, fcng=None,
                           xpnt=None, **kwargs):
    """
    Create a connection point description.

    :param ctyp: Connection point type
    :type ctyp: str
    :param cdir: Direction
    :type cdir: int
    :param name: Name
    :type name: str | None
    :param cpos: Position
    :type cpos: (float, float) | None
    :param fcng: Facing/gravity
    :type fcng: int | None
    :param xpnt: Referenced connection point
    :type xpnt: WhdbConnectionPointP | None
    :param kwargs: Additional arguments
    :type kwargs: dict[str, object]
    :return: Newly created connection point description
    :rtype: WhdbXPointDesc
    """
    assert 'gvty' not in kwargs

    return WhdbXPointDesc(
        ctyp=ctyp, cpos=cpos or (1.0, 0.0), cdir=cdir, name=name, fcng=fcng,
        xpnt=xpnt, **kwargs)


def whhdb_createFixedXPointManager(cns=None, pos=None, css=None, cfs=None,
                                   cds=None, cas=None, c1553s=None, carinc825s=None):
    """
    Create a connection point manager for fixed connection point sets.

    :param cns: Input connection point names, if any
    :type cns: list[str] | None
    :param pos: Input connection point positions / facings, if any
    :type pos: list[(float, float, int)] | None
    :return: Connection point manager
    :rtype: WhdbFixedConnectionPointManager
    """

    def mkXPoints(cdir, ns, ps, bs, cfs, cds, cc, c1553, carinc825):
        if not ns:
            return []
        assert len(ns) == len(bs)
        if ps:
            assert len(ps) == len(ns)
        else:
            ps = whdb_makeConnXYPositions(cdir, len(ns), bs=bs)
        return list(zip(repeat(cdir), ns,
                        [(x, y) for x, y, _ in ps], [f for _, _, f in ps],
                        bs, cfs, cds, cc, c1553, carinc825))
    xpnts = [whhdb_createXPointDesc(ctyp=conf['type'],
                                    cdir=cdir, name=name,
                                    cpos=cpos, fcng=fcng, config=conf,
                                    fiber=cf, discret=cd, afdx=ca, c1553=c1553, carinc825=carinc825)
             for cdir, name, cpos, fcng, conf, cf, cd, ca, c1553, carinc825
             in mkXPoints(WDBS_XPNT_BI, cns, pos, css, cfs, cds, cas, c1553s, carinc825s)]
    return WhdbFixedConnectionPointManager(xpnts)
