# ----------------------------------------------------------------------------
# - whhdb-card-instance                                                      -
# - wild cruncher: card instances                       core / hardware / db -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whhdb_PCI_card_instance
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Card instance

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""
from wcr_core.db.whdb_instance import WhdbOperatorInstance
from wcr_core.dbstore.wdbs_meta_description import *
from wcr_core.dbstore.wdbs_meta_type import *
from wcr_core.modules.hardware.db.whhdb_standard_operators import \
    WHHDB_HOST, WHHDB_HOST_CODE, WHHDB_RACK_IO_CODE, \
    WHHDB_CETRAC_RACK_IO_CODE
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_DIRECTION, \
    WHDB_XPOINT_NAME
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.hardware.db.whhdb_Host_instance_meta import \
    WHHDB_HOST_META, WHHDB_META_CTYPE_KEY, WHHDB_CARD_TYPE_KEY, \
    WHHDB_CARD_PORTS_KEY, WHHDB_META_PTP_MASTER_KEY
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError


class WhhdbGenericHostInstance(WhdbOperatorInstance):
    """
    Operator-instance for bus instances.
    """

    META = WHHDB_HOST_META
    TYPES = []

    PTP_GM_TYPES = ['Yes', 'No']

    def __init__(self, op, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        assert inst.getOperator() in (WHHDB_HOST_CODE, WHHDB_RACK_IO_CODE,
                                      WHHDB_CETRAC_RACK_IO_CODE)
        WhdbOperatorInstance.__init__(self, op, inst, wreg)
        self.CARD_TO_REMOVE = []

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        if WHDB_XPOINT_DIRECTION not in kwargs:
            kwargs[WHDB_XPOINT_DIRECTION] = WDBS_XPNT_BI
        return super()._createRawConnectionPoint(ctyp, **kwargs)

    def getTypes(self):
        return sorted(self.TYPES)

    def getCards(self):
        """
        return every non marked card
        """
        cards = sorted(self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys())
        for rmc in self.CARD_TO_REMOVE:
            cards.remove(rmc)
        return cards

    def setCard(self, card):
        """
        initialising a new card or remove from deleted if allready exist
        """
        if not card in \
                self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            if not isNameValid(card):
                showErrorWindow("Name Error", NameFormatMessageError,
                                card)
            self.rawElement.setMeta({WHHDB_CARD_TYPE_KEY:'',
                                     WHHDB_CARD_PORTS_KEY:[]},
                                    WHHDB_META_CTYPE_KEY, card)
        else:
            if card in self.CARD_TO_REMOVE:
                self.CARD_TO_REMOVE.remove(card)

    def getCardType(self, cardname):
        if not cardname in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            return ''
        type = self.rawElement.getMeta(WHHDB_META_CTYPE_KEY, cardname,
                                       WHHDB_CARD_TYPE_KEY)
        for val in self.TYPES:
            if val == type:
                return val
        return ''

    def getCardPtpMaster(self, cardname):
        if not cardname in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            return ''
        type = self.rawElement.getMeta(WHHDB_META_CTYPE_KEY, cardname,
                                       WHHDB_META_PTP_MASTER_KEY)
        for val in self.PTP_GM_TYPES:
            if val == type:
                return val
        return 'No'

    def setCardType(self, card, type):
        if not card in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            return None
        if type in self.TYPES:
            self.rawElement.setMeta(type, WHHDB_META_CTYPE_KEY, card,
                                    WHHDB_CARD_TYPE_KEY)

    def setCardPtpMaster(self, card, status):
        """
        set the state of ptpMaster of cardname
        """
        if not card in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            return []
        if status in self.STATUS_TYPES:
            return self.rawElement.setMeta(status,
                                           WHHDB_META_CTYPE_KEY, card,
                                           WHHDB_META_PTP_MASTER_KEY)

    def replaceCard(self, hcardname, ncard):
        """
        replace the card hcardname by the ncard.
        :param hcardname: name ot the card ho will be replaced
        :type hcardname: str
        :param ncard: description of the full new card
        :type ncard: [cardname, cardtype, status, [portsbinded]]
        """
        if hcardname is not '' or hcardname is not None:
            self.rawElement.removeMeta(WHHDB_META_CTYPE_KEY, hcardname)
        self.setCard(ncard[0])
        self.setCardType(ncard[0], ncard[1])
        self.setCardPtpMaster(ncard[0], ncard[2])

    def getCardsDescription(self):
        """
        return the full description of the card stocked in metadata
        """
        NotImplementedError

    def markCard(self, cardn):
        """
        mark cardn as card to remove
        """
        if cardn in self.rawElement.getMeta(WHHDB_META_CTYPE_KEY).keys():
            self.CARD_TO_REMOVE.append(cardn)

    def restoreCards(self):
        self.CARD_TO_REMOVE = []

    def delCards(self):
        """
        cleaning metada from card marked to remove and asociated information
        """
        portl=[]
        for c in self.CARD_TO_REMOVE:
            getCon = getattr(self, 'getCardConnectionPoints', None)
            if getCon:
                portl += getCon(c)
            self.rawElement.removeMeta(WHHDB_META_CTYPE_KEY, c)
        self.restoreCards()
