# ----------------------------------------------------------------------------
# - whhdb-RackIo-instance                                                    -
# - wild cruncher: Rack IO instances                   core / hardware / db  -
# ----------------------------------------------------------------------------
# - author    (c) 2017-2017 achraf mahti                                     -
# - copyright (c) 2017-2017 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whhdb_rack_instance
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Rack instance

.. moduleauthor:: Achraf Mahti <mahti@silkan.net>
"""
from uuid import UUID
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_DIRECTION
from wcr_core.dbstore.wdbs_meta_description import *
from wcr_core.dbstore.wdbs_meta_type import *
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.hardware.db.whhdb_generic_Host_instance \
    import WhhdbGenericHostInstance
from wcr_core.modules.hardware.db.whhdb_operator import WHHDB_OPER
from wcr_core.modules.hardware.db.whhdb_rack_instance_meta import \
    IO_TYPES, WHHDB_MAX_CARD
from wcr_core.modules.hardware.db.whhdb_standard_operators import \
    WHHDB_RACK, WHHDB_RACK_IO_CODE, WHHDB_CETRAC_RACK_IO_CODE
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError

INIT_IO_CARD_POSITION = 65792  # hexa = 10100
INCREMENT_POSITION = 4096  # hexa = 1000


class WhhdbRackInstance(WhhdbGenericHostInstance):
    """
    Operator-instance for bus instances.
    """
    STATUS_TYPES = ['Yes', 'No']
    TYPES = IO_TYPES
    META = {}
    MAX_CARD = WHHDB_MAX_CARD

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        if inst.getOperator() == WHHDB_RACK_IO_CODE:
            whhdbinstance = WHHDB_RACK
        elif inst.getOperator() == WHHDB_CETRAC_RACK_IO_CODE:
            whhdbinstance = WHHDB_OPER[WHHDB_CETRAC_RACK_IO_CODE]
        else:
            assert inst.getOperator() in (WHHDB_RACK_IO_CODE,
                                          WHHDB_CETRAC_RACK_IO_CODE)
        WhhdbGenericHostInstance.__init__(self, whhdbinstance, inst, wreg)

        self.rackSymbol = self.getSymbol()
        self.cardsPositionCpt = INIT_IO_CARD_POSITION
        self.cards = inst.cardsInst
        self.populateCards(inst.cardsOper, wreg)
        self.nb_card = len(self.getCards())
        self.updateRackSymbol()

    def populateCards(self, cards, wreg):
        """
         instanciate operator cards to populate the dict of cards in rack
        :param cards: list of card Operator and data associated
        :type [card, [data]
        :param wreg: Wrapper registry used to instantiate cards
        :return:
        """
        # iter cards by sorting over idtstatus to add them in order
        for uuid, values in sorted(cards.items(), key=lambda x: x[1][1]):
            card, [name, idtstatus, channels] = values
            cardInst = wreg.wrap(card)
            cardInst.setName(name)
            for num, channel in channels.items():
                for prmName, value in channel:
                    cardInst.setChannelParamValue(num, prmName, value)
            cardInst.setEquipement(self)
            if isinstance(uuid, str):
                uuid = UUID(uuid)
            self.setIOCard(uuid, cardInst, idtstatus)

    def updateRackSymbol(self):
        nbCard = len(self.cards)
        rackSymbol = '{} \n {}/{}'.format(self.rackSymbol, nbCard,
                                          self.MAX_CARD)
        self.setCustomText(rackSymbol)

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        if WHDB_XPOINT_DIRECTION not in kwargs:
            kwargs[WHDB_XPOINT_DIRECTION] = WDBS_XPNT_BI
        return super()._createRawConnectionPoint(ctyp, **kwargs)

    def getIOTypes(self):
        return super().getTypes()

    def getCards(self):
        """
        return every non marked card
        """
        return [c.getName() for c in self.cards.values()]

    def setIOCard(self, uuid, card, position=None):
        """
        Store a card
        """
        if len(self.cards) < self.MAX_CARD:
            if uuid in self.cards:
                showErrorWindow('Card Error',
                                'This card name already exist.', None)
                return
            position = self.newPosition(position)
            self.cards[uuid] = card
            self.setIOCardIDTStatus(uuid, position)
            self.updateRackSymbol()
        else:
            showErrorWindow('Card Error',
                            'A RackIO can have a max of 11 cards',
                            None)
            return
        return True

    def setIOCardName(self, uuid, cardname):
        self.cards[uuid].setName(cardname)

    def getIOCardName(self, uuid):
        return self.cards[uuid].getName()

    def setIOCardMeta(self, uuid, meta):
        self.cards[uuid].setChannels(meta)

    def setIOCardType(self, uuid, cardType):
        if cardType.lower() in [t.lower() for t in self.TYPES]:
            self.cards[uuid].setType(cardType)

    def getIOCardType(self, uuid):
        return self.cards[uuid].getType()

    def getIOCardChannels(self, uuid):
        return self.cards[uuid].getChannels()

    def getIOCardIDTStatus(self, uuid):
        return self.cards[uuid].getIDTStatus()

    def setIOCardIDTStatus(self, uuid, pos):
        # convert to hex if necessary
        if isinstance(pos, int):
            pos = hex(pos).replace('0x', '')
        # check unicity of the position
        if not pos in [c.getPosition() for c in self.cards.values()]:
            self.cards[uuid].setIDTStatus(pos)
        else:
            showErrorWindow('Card Error',
                            'Position duplication {}',
                            pos)

    def updateCard(self, uuid, cardName, cardType):
        """
        update the card info .
        :param hcardname: name ot the card ho will be replaced
        :type hcardname: str
        :param ncard: description of the full new card
        :type ncard: [cardname, cardtype, status]
        """
        if uuid in self.cards:
            self.cards[uuid].setName(cardName)
            self.cards[uuid].setType(cardType)

    def getIOCardsDescription(self):
        """
        return the full description of the card stocked in metadata
        :return : [cardname, cardtyp, status]
        """
        fullcards=[]

        for uuid in self.cards:
            if uuid not in self.CARD_TO_REMOVE:

                structcards = (uuid,
                               self.getIOCardName(uuid),
                               self.getIOCardType(uuid),
                               self.getIOCardIDTStatus(uuid))
                fullcards.append(structcards)
        return fullcards

    def setStatus(self, status):
        self.rawElement.setStatus(status)

    def getStatus(self):
        return self.rawElement.getStatus

    def markIOCard(self, uuid):
        """
        mark cardn as card to remove
        """
        if uuid in self.cards:
            self.CARD_TO_REMOVE.append(uuid)

    def delCards(self):
        """
        cleaning metada from card marked to remove and asociated information
        """
        for uuid in self.CARD_TO_REMOVE:
            del self.cards[uuid]
        self.updateRackSymbol()
        self.restoreCards()

    def newPosition(self, position):
        if position is None:
            self.cardsPositionCpt += INCREMENT_POSITION
        else:
            self.cardsPositionCpt = int(position, 16)
        return self.cardsPositionCpt
