# ----------------------------------------------------------------------------
# - whplg-library-reader                                                     -
# - wild cruncher: library reader                   core / hardware / plugin -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whplg_library_reader
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Hardware library reader

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""
from xml.etree.ElementTree import ElementTree
from silkan_common.scmn_file_path import scmn_path_dir, scmn_path_join
from wcr_core.common.wcmn_parameter_description import \
    WCMN_PARAM_FORMAT_STRING_FLAG
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.hardware.db.whhdb_connection_point_manager import \
    whhdb_createFixedXPointManager
from wcr_core.modules.hardware.db.whhdb_connection_point_base_meta import \
    WHHDB_PORT_SPEED_KEY, WHHDB_TYPE_KEY, WHHDB_PORT_STATUS_MASK_KEY, \
    WHHDB_PORT_NUM_KEY, WHHDB_PORT_BBONE_KEY, WHHDB_PORT_DIRECTION_KEY, \
    WHHDB_PORT_STATUS_MONITOR_KEY, WHHDB_PORT_ACTIVE_KEY
from wcr_core.modules.hardware.db.whhdb_connection_point_base import \
    WHHDB_XPOINT_CONFIG_KEY, WHHDB_XPOINT_FIBER_CHANNEL_KEY, \
    WHHDB_XPOINT_DISCRET_KEY, WHHDB_XPOINT_AFDX_KEY, WHHDB_XPOINT_1553_KEY, \
    WHHDB_XPOINT_ARINC825_KEY
from wcr_core.modules.hardware.db.whhdb_operator import WhhdbOperator
from wcr_core.plugin.wplg_library import WLIB_ICON_SVG, WLIB_ICON_PNG32

_WMPLG_PARAM_FLAGS = {
    'formatted_string': WCMN_PARAM_FORMAT_STRING_FLAG
}

_WMPLG_ICON_TYPES = {
    'png32': WLIB_ICON_PNG32,
    'svg': WLIB_ICON_SVG
}


def whplg_make_dict_from_tree(element_tree):
    """Traverse the given XML element tree to convert it into a dictionary.

    :param element_tree: An XML element tree
    :type element_tree: xml.etree.ElementTree
    :rtype: dict
    """

    def internal_iter(tree, accum):
        """Recursively iterate through the elements of the tree accumulating
        a dictionary result.

        :param tree: The XML element tree
        :type tree: xml.etree.ElementTree
        :param accum: Dictionary into which data is accumulated
        :type accum: dict
        :rtype: dict
        """
        if tree is None:
            return accum
        if tree.tag == 'meta':
            key_name = tree.tag
        else:
            key_name = '{}.{}'.format(tree.tag, tree.get('type'))
        if tree.getchildren():
            accum[key_name] = {}
            for each in tree.getchildren():
                each_name = '{}.{}'.format(each.tag, each.get('type'))
                result = internal_iter(each, {})
                if each_name in accum[key_name]:
                    if not isinstance(accum[key_name][each_name], list):
                        accum[key_name][each_name] = [
                            accum[key_name][each_name]
                        ]
                    accum[key_name][each_name].append(result[each_name])
                else:
                    accum[key_name].update(result)
        else:
            accum[key_name] = tree.text

        return accum

    return internal_iter(element_tree, {})


def whplg_read_operator(file, lcod):
    """
    Read an operator from given XML description file.

    :param file: XML description file
    :type file: str
    :param lcod: Library code
    :type lcod: str
    :return: Operator
    :rtype: WmhdbOperator
    """
    tree = ElementTree(file=file)
    node = tree.getroot()
    if node.tag != 'wcr-operator':
        raise Exception(
            'File {} does not contain a Wild Cruncher 2.0 '
            'operator specification'.format(file))
    code = lcod + '.' + node.get('code')
    title = node.get('title')

    descNode = node.find('description')
    desc = descNode.text if descNode is not None else ''

    cns = _whplg_read_connection_points(
        node.find('connectors'), 'c', WDBS_XPNT_BI)

    paramsNode = node.find('parameters')
    params = []
    paramType = None
    """:type: list[(str, str, str, object, str, int)]"""
    if paramsNode is not None:
        paramType = paramsNode.get('type')
        if paramType == 'simple':
            for paramNode in paramsNode:
                params.append(_whplg_read_param(paramNode))
        elif paramType == 'channel':
            for row in paramsNode:
                rowNumber = row.get('number')
                prm = []
                for param in row:
                    prms = _whplg_read_param(param)
                    # replace 2 last values by dict
                    ptype = param.get('type')
                    pvalues = param.get('values')
                    prms[-1] = {'values': pvalues,
                                'type': ptype}
                    prm.append(prms)
                params.append((rowNumber, prm))
    paramOp = WhplgXmlParameter(params, paramType)

    meta = {}
    for mitm in node.findall('meta'):
        children = mitm.getchildren()
        if children:
            meta['type'] = mitm.get('type')
            meta.update(whplg_make_dict_from_tree(mitm))
        else:
            meta[mitm.get('key')] = mitm.text.strip()

    symNode = node.find('symbol')
    symbol = symNode.text.strip() if symNode is not None else None

    icons = {}
    iconsNode = node.find('icons')
    if iconsNode is not None:
        for iconNode in iconsNode:
            ityp = _WMPLG_ICON_TYPES.get(iconNode.get('type'))
            if ityp is not None:
                icons[ityp] = scmn_path_join(scmn_path_dir(file),
                                             iconNode.text.strip())

    flagsNode = node.find('flags')
    if flagsNode is not None:
        flags = {f.get('name') for f in flagsNode.findall('flag')}
    else:
        flags = set()
    # noinspection PyTypeChecker
    return WhhdbOperator(
        code=code,
        xmngr=_whplg_create_xpoint_manager(cns),
        name=title,
        info=desc,
        meta=meta,
        parameters=paramOp,
        flags=flags,
        icons=icons,
        symbol=symbol)


def _whplg_read_param(paramNode):
    pname = paramNode.get('name')
    ptitle = paramNode.get('title')
    pdesc = paramNode.get('info')
    pdefault = paramNode.get('default')
    ptype = paramNode.get('type')
    pflags = sum(_WMPLG_PARAM_FLAGS[f] for f
                 in paramNode.get('flags', '').split(',') if f)
    return [pname, ptitle, pdesc, pdefault, ptype, pflags]

class WhplgXmlParameter():
    def __init__(self, prms, ptype):
        self.prms = prms
        self.ptype = ptype


def _whplg_read_connection_points(nlst, pfx, cdir):
    if nlst is None or len(nlst) == 0:
        return []

    def readXPointDef(node):
        if node.tag == 'connector':
            conf = {WHHDB_PORT_SPEED_KEY: node.get('bandwidth'),
                    WHHDB_TYPE_KEY: node.get('type'),
                    WHHDB_PORT_NUM_KEY: node.get('PortNum'),
                    WHHDB_PORT_BBONE_KEY: node.get('backbone'),
                    WHHDB_PORT_DIRECTION_KEY: node.get('Direction'),
                    WHHDB_PORT_STATUS_MASK_KEY: node.get('StatusMask'),
                    WHHDB_PORT_STATUS_MONITOR_KEY: node.get('StatusMonitor'),
                    WHHDB_PORT_ACTIVE_KEY: node.get('active')
                    }
            return 'single', {'cdir': cdir, 'name': node.get('name') or pfx,
                              WHHDB_XPOINT_CONFIG_KEY: conf,
                              WHHDB_XPOINT_DISCRET_KEY: {},
                              WHHDB_XPOINT_FIBER_CHANNEL_KEY: {},
                              WHHDB_XPOINT_AFDX_KEY: {},
                              WHHDB_XPOINT_1553_KEY: {},
                              WHHDB_XPOINT_ARINC825_KEY: {}}
        raise(ValueError, "Unknown connection tag : "+node.tag)

    return list(filter(None, map(readXPointDef, nlst)))


def _whplg_create_xpoint_manager(xpnts):
    _xpnts = []
    for xtyp, xdat in xpnts:
            _xpnts.append((xtyp, xdat))

    cns = [xpnt for xpnt in _xpnts if xpnt[1]['cdir'] == WDBS_XPNT_BI]

    tps = [xtyp for xtyp, _ in cns]

    sng = all(map(lambda x: x == 'single', tps))

    if sng:
        return whhdb_createFixedXPointManager(
            cns=[d['name'] for _, d in cns],
            css=[d[WHHDB_XPOINT_CONFIG_KEY] for _, d in cns],
            cfs=[d.get(WHHDB_XPOINT_DISCRET_KEY, {}) for _, d in cns],
            cds=[d.get(WHHDB_XPOINT_FIBER_CHANNEL_KEY, {}) for _, d in cns],
            cas=[d.get(WHHDB_XPOINT_AFDX_KEY, {}) for _, d in cns],
            c1553s=[d.get(WHHDB_XPOINT_1553_KEY, {}) for _, d in cns],
            carinc825s=[d.get(WHHDB_XPOINT_ARINC825_KEY, {}) for _, d in cns])
    else:
        raise Exception(
            'Cannot handle connection point types {}'.format(tps))
