# ----------------------------------------------------------------------------
# - wmehdb-top-block                                                         -
# - wild cruncher:wsh top hierarchy object               core / message / db -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 lambert philippe                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmehdb_top_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher top hierarchy object

.. moduleauthor:: philippe lambert <lambert@silkan.com>
"""
from wcr_core.modules.icd.db.wihdb_top_block_meta import \
    WIHDB_ICD_GLOBAL_DICTIONNARY, WIHDB_META_GLOBAL_TYPES_KEY ,\
    WIHDB_META_GLOBAL_MESSAGES_KEY, WIHDB_META_MESSAGES_BINDING_KEY
from wcr_core.db.whdb_block import WhdbBlock


GENERIC_TYPES = {'char':   [['varchar',   '0x0', 'char']],
                 'int':    [['varInt',    '0x0', 'int']],
                 'int64':  [['varInt64',  '0x0', 'int64']],
                 'float':  [['varFloat',  '0x0', 'float']],
                 'short':  [['varShort',  '0x0', 'short']]}


class WihdbTopHierarchy(WhdbBlock):
    """
    The WmhdbTopHierarchy is the top hierarchy model class. A top hierarchy
    is at the same time a model and an instance. The class provides several
    methods for creating the hierarchy.
    """
    META = WIHDB_ICD_GLOBAL_DICTIONNARY

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        super().__init__(block, wreg)

        self.iniGlobalType()

    # GLOBAL TYPE DATA
    def iniGlobalType(self):
        self.setGlobalTypes([])
        for type_, variables in GENERIC_TYPES.items():
            self.addType(type_, variables)

    def setGlobalTypes(self, dictType):
        """
        Set the global dictionary type with variables

        :param varlist: List of the variables
        :type varlist: [(str,str,[str])]
        """
        self.rawElement.setMeta(dictType, WIHDB_META_GLOBAL_TYPES_KEY)

    def getGlobalTypes(self):
        """
        return the global dictionary type with variables
        :return : [(str,str,[str])]
        """
        return self.rawElement.getMeta(WIHDB_META_GLOBAL_TYPES_KEY)

    def addType(self, type_, variables):
        """
        Add new type to the global dictionnary.
        """
        if not isinstance(type_, str):
            raise Exception("Type must be string and not {}".
                            format(type(type_).__name__))
        if not isinstance(variables, list):
            raise Exception("Variables must be list and not {}".
                            format(type(variables).__name__))
        gb = self.getGlobalTypes()
        elem = (type_, variables)
        if not self.typeExist(type_):
            gb.append(elem)
        else:
            raise Exception('Type "{}" already exist, delete the existing type '
                            'if you want to redefine it'.format(type_))
        self.setGlobalTypes(gb)

    def typeExist(self, type_):
        """
        Return boolean if type is in the global type dictionary
        :return : Bool
        """
        gb = self.getGlobalTypes()
        for v, _ in gb:
            if type_ == v:
                return True
        return False

    def getTypeVariables(self, type_):
        """
        Get the variable of type from the dictionary
        """
        types = self.œgetGlobalTypes()
        for (gbType, data) in types:
            if type_ == gbType:
                return data

    # GLOBAL MESSAGE DATA
    def getAllMessages(self):
        """
        return the global dictionary of messages
        :return : [(str,[str])]
        """
        return self.rawElement.getMeta(WIHDB_META_GLOBAL_MESSAGES_KEY)

    def setAllMessages(self, msgs):
        """
        Set the global dictionary of messages with types
        """
        self.rawElement.setMeta(msgs, WIHDB_META_GLOBAL_MESSAGES_KEY)

    def addMsgType(self, msg, type_, size):
        """
        bind a message to a type from global dictionary type
        """
        if type_ is None or msg is None or size is None:
            return
        if not isinstance(msg, str):
            raise Exception("Message must be string and not {}".
                            format(type(msg).__name__))

        messages = self.getAllMessages()
        if not self.msgExist(msg):
            messages[msg] = (type_, size)
            self.setAllMessages(messages)
        else:
            self.editMessageType(msg, type_, size)

    def msgExist(self, msg):
        """
        Return boolean if msg is in the global type dictionary
        :return : Bool
        """
        messages = self.getAllMessages()
        for v in messages:
            if msg == v:
                return True
        return False

    def editMessageType(self, msg, type_, size):
        """
        Edit the binding between a msg and variable
        """
        messages = self.getAllMessages()
        for m in messages:
            if m == msg:
                messages[msg] = (type_, size)
                self.setAllMessages(messages)
                return

    def removeMessageType(self, msg):
        """
        Edit the binding between a msg and variable
        """
        messages = self.getAllMessages()
        if msg not in messages:
            return
        del messages[msg]
        self.setAllMessages(messages)

    def getAllMessageBindings(self):
        """
        return the global dictionary of messages
        :return : [(str,[str])]
        """
        return self.rawElement.getMeta(WIHDB_META_MESSAGES_BINDING_KEY)

    def getAllMessageDictBindings(self):
        allBindings = self.getAllMessageBindings()
        messages = {}
        for glbMsg in allBindings:
            prodconsMessages = allBindings[glbMsg]
            for (modMsg, state) in prodconsMessages:
                if state == 'P':
                    messages.setdefault(glbMsg, {})['P'] = modMsg
                elif state == 'C':
                    messages.setdefault(glbMsg, {}).setdefault('C', []).append(modMsg)
                else:
                    raise Exception('State of message undefined')

        return messages

    def setAllMessageBindings(self, msgs):
        """
        Set the global dictionary of messages with types
        """
        self.rawElement.setMeta(msgs, WIHDB_META_MESSAGES_BINDING_KEY)

    def bindMsg(self, moduleMsg, glbMsg, moduleName, state, messages=None):
        """
        bind a module message to a global message from global message dictionary
        """
        if moduleMsg is None or glbMsg is None:
            return

        moduleMsg = '{}.{}'.format(moduleName, moduleMsg)
        if not messages:
            messages = self.getAllMessageBindings()

        if self.glbMsgExist(glbMsg):
            if not self.checkMsgProducerUnicity(glbMsg, moduleMsg, state):
                return
            messages[glbMsg].append((moduleMsg, state))
        else:
            messages[glbMsg] = [(moduleMsg, state)]
        self.setAllMessageBindings(messages)

    def glbMsgExist(self, msg):
        """
        Return boolean if msg is in the global type dictionary
        :return : Bool
        """
        messages = self.getAllMessageBindings()
        for v in messages:
            if msg == v:
                return True
        return False

    def checkMsgProducerUnicity(self, glbMsg, moduleMsg, state, messages=None):
        """
        Return boolean if msg is already used
        :return : Bool
        """
        if not messages:
            messages = self.getAllMessageDictBindings()
        if state == 'P':
            msgState = messages.get(glbMsg, None)
            if msgState is None:
                return True
            else:
                if msgState.get(state, None) is not None:
                    return False
                else:
                    return True
        else:
            return True

    def checkMsgProducerCoherency(self):
        """
        Return list of messages that are consumed but not produced
        :return : Bool
        """
        messages = self.getAllMessageDictBindings()
        messageWoProducer = []
        for msg, state in messages.items():
            if state.get('C') and not state.get('P'):
                messageWoProducer.append(msg)

        return messageWoProducer

    def editMessageBinding(self, moduleMsg, glbMsg):
        """
        Edit the binding between a msg and variable
        """
        messages = self.removeMessageBinding(moduleMsg, glbMsg)
        self.bindMsg(moduleMsg, glbMsg, messages=messages)

    def removeGlobalMessageBinding(self, glbMsg):
        """
        Edit the binding between a msg and variable
        """
        messages = self.getAllMessageBindings()
        del messages[glbMsg]
        self.setAllMessageBindings(messages)
        return messages

    def removeMessageBinding(self, moduleMsg, glbMsg):
        """
        Edit the binding between a msg and variable
        """
        messages = self.getAllMessageBindings()
        messages[glbMsg].remove(moduleMsg)
        self.setAllMessageBindings(messages)
        return messages

    def setSourcePath(self, path):
        if path is not None:
            self._srdfilepath = path

    def formatToSave(self, npath):
        self.setSourcePath(npath)

    def getDataByModule(self, modName):
        for mod in self.getInstances():
            if mod.getName == modName:
                return mod.getData()
        return []

    def getMessageByModule(self, modName, msg):
        for mod in self.getInstances():
            if mod.getName == modName:
                for modMsgName in mod.getData():
                    if modMsgName == msg:
                        return mod.getData()[modMsgName]
        return None

    def getGlbMsgByLocalMsg(self, module, lclMsg):
        """
        return the global message bound to module local msg
        :param module: module name
        :param lclMsg: local message
        :return:
        """
        modMsage = '{}.{}'.format(module, lclMsg)
        for glbMsg, data in self.getAllMessageBindings().items():
            for lMsg, _ in data:
                if modMsage == lMsg:
                    return glbMsg
        return None
