# ----------------------------------------------------------------------------
# - wmhdb-block                                                              -
# - wild cruncher:wsh block object                         core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher silkan hierarchy blocks

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import defaultdict
from itertools import chain

from wcr_core.db.whdb_block import WhdbBlock
from wcr_core.db_api.whdb_block_api import WHDB_DEFAULT_BLOCK_CODE
from wcr_core.db_api.whdb_model_api import WHDB_DEFAULT_MODEL_CODE
from wcr_core.db.whdb_operator import WHDB_INPUT_OPERATOR_FLAG, \
    WHDB_OUTPUT_OPERATOR_FLAG, \
    WHDB_PORT_OPERATOR_FLAG
from wcr_core.db.whdb_connection_point import WHDB_DEFAULT_XPOINT_CODE
from wcr_core.dbstore_api.wdbs_instance_api import WdbsOperatorInstanceP
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_IN, \
    WDBS_XPNT_OUT, WDBS_XPNT_NONE
from wcr_core.modules.model.db.wmhdb_block_base import WmhdbBlockP
from wcr_core.modules.model.db.wmhdb_block_proxy import WmhdbBlockProxy
from wcr_core.modules.model.db.wmhdb_operator import WMHDB_OPER
from wcr_core.modules.model.db.wmhdb_wire import WMHDB_WIRE_UTYPE_KEY
from wcr_core.modules.model.db.wmhdb_protocols import \
    WmhdbHasSchedulingGroupP, WMHDB_SGROUP_KEY
from wcr_core.db.whdb_protocols import WHDB_ROTATION_KEY, WHDB_HFLIP_KEY, \
    WHDB_VFLIP_KEY, WHDB_POSITION_KEY, WHDB_Z_VALUE_KEY, \
    WHDB_DIMENSION_KEY
from wcr_core.modules.model.db.wmhdb_standard_operators import \
    WMHDB_PORT_NAME_PARAMETER


# noinspection PyAbstractClass
class WmhdbBlock(WhdbBlock, WmhdbBlockP):
    """
    The WmhdbModelBlock class is the silkan hierarchy model block. It is used
    to construct one hierarchical level by grouping instances, functions and
    wire into several list. A block is a model and therefore is defined by name
    with an optional information string.
    """
    _PORT_FLAG = WHDB_PORT_OPERATOR_FLAG

    _PORT_NAME_PARAM = WMHDB_PORT_NAME_PARAMETER

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Raw element wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        WhdbBlock.__init__(self, block, wreg)

        self.__rawBlock = block
        self.__wrapperRegistry = wreg

    def setNameInfo(self, name, info):
        self.setName(name)
        self.setInfo(info)

    def _createRawOperatorInstance(self, oper, *args, **kwargs):
        if isinstance(oper, str):
            oper = WMHDB_OPER[oper]
        kwargs.setdefault(WMHDB_SGROUP_KEY,
                          WmhdbHasSchedulingGroupP.DEFAULT_SCHED_GROUP)
        return wmhdb_createOperatorInstance(
            self.__rawBlock, oper, *args, **kwargs)

    def getBlockAndModelInstances(self):
        return chain(self.getBlockInstances(), self.getModelInstances())

    def _createRawModelInstance(self, *args, **kwargs):
        kwargs.setdefault(WMHDB_SGROUP_KEY,
                          WmhdbHasSchedulingGroupP.DEFAULT_SCHED_GROUP)
        return super()._createRawModelInstance(*args, **kwargs)

    def _createRawBlockInstance(self, *args, **kwargs):
        kwargs.setdefault(WMHDB_SGROUP_KEY,
                          WmhdbHasSchedulingGroupP.DEFAULT_SCHED_GROUP)
        if not kwargs.get('btyp'):
            kwargs['btyp'] = WHDB_DEFAULT_BLOCK_CODE
        return super()._createRawBlockInstance(*args, **kwargs)

    def _createRawModelInstance(self, *args, **kwargs):
        kwargs.setdefault(WMHDB_SGROUP_KEY,
                          WmhdbHasSchedulingGroupP.DEFAULT_SCHED_GROUP)
        kwargs['mtyp'] = WHDB_DEFAULT_MODEL_CODE
        return super()._createRawModelInstance(*args, **kwargs)

    def _createRawConnection(self, *args, rtyp=None, **kwargs):
        conn = super()._createRawConnection(*args, **kwargs)
        conn.setMeta(rtyp, WMHDB_WIRE_UTYPE_KEY)
        return conn

    def getInputWires(self, instance):
        return (wire for conn in instance.getInputConnectors()
                for wire in self.getConnectionsForObject(instance, conn))

    def getOutputWires(self, instance):
        return (wire for conn in instance.getOutputConnectors()
                for wire in self.getConnectionsForObject(instance, conn))

    def resetWireList(self):
        # noinspection PyUnresolvedReferences
        for w in self.getConnections():
            w.getPath().resetPath()

    def _getConnectorType(self, instance, connector, typeF):
        wires = list(self.getConnectionsForObject(instance, connector))
        if len(wires) == 1:
            return typeF(wires[0])
        elif wires:
            userType = typeF(wires[0])
            if any([typeF(w) != userType for w in wires]):
                raise ValueError('Connector has wires of different types')
            return userType
        else:
            return None

    def _getConnectorUserType(self, instance, connector):
        return self._getConnectorType(
            instance, connector, lambda w: w.getUserRtiType())

    def _getConnectorInferredType(self, instance, connector):
        return self._getConnectorType(
            instance, connector, lambda w: w.getInferredRtiType())

    def getInstanceOutputConnectorsUserTypes(self, instance):
        return (self._getConnectorUserType(instance, conn)
                for conn in instance.getOutputConnectors())

    def getInstanceOutputConnectorsInferredTypes(self, instance):
        return (self._getConnectorInferredType(instance, conn)
                for conn in instance.getOutputConnectors())

    def getInstanceInputConnectorsUserTypes(self, instance):
        return (self._getConnectorUserType(instance, conn)
                for conn in instance.getInputConnectors())

    def getInstanceInputConnectorsInferredTypes(self, instance):
        return (self._getConnectorInferredType(instance, conn)
                for conn in instance.getInputConnectors())

    def getPortNames(self):
        ports = defaultdict(list)
        for inst in sorted((f for f in self.getOperatorInstances()
                            if f.getOperator().hasFlag(self._PORT_FLAG)),
                           key=lambda f: f.getPosition()):
            if inst.getOperator().hasFlag(WHDB_INPUT_OPERATOR_FLAG):
                cdir = WDBS_XPNT_IN
            elif inst.getOperator().hasFlag(WHDB_OUTPUT_OPERATOR_FLAG):
                cdir = WDBS_XPNT_OUT
            else:
                cdir = WDBS_XPNT_NONE
            name = inst.getName()
            pid = inst.getParameterValue(self._PORT_NAME_PARAM) or name
            if pid:
                ports.setdefault(cdir, []).append((pid, name))
        return ports

    def createProxy(self):
        return WmhdbBlockProxy(self)


def wmhdb_createOperatorInstance(block, oper, name=None, info=None,
                                 pos=None, zpos=None, width=None,
                                 height=None, rotation=None, vflip=None,
                                 hflip=None, sgrp=None, prms=None):
    """
    :param block: Block to create function in
    :type block: WdbsBlockP
    :param oper: Operator code
    :type oper: WmhdbOperator
    :param name: Name of the instance
    :type name: str
    :param info: Info string for the instance
    :type info: str
    :param pos: Position
    :type pos: (float, float)
    :param zpos: Z-value
    :type zpos: int
    :param width: Width
    :type width: float
    :param height: Height
    :type height: float
    :param rotation: Rotation
    :type rotation: float
    :param vflip: Vertically flipped
    :type vflip: bool
    :param hflip: Horizontally flipped
    :type hflip: bool
    :param sgrp: Scheduling group name
    :type sgrp: str
    :param prms: Parameter values
    :type prms: Iterable[(str, object)]
    :return: The newly created operator instance
    :rtype: WdbsOperatorInstanceP
    """
    if width is not None and height is not None:
        dim = (width, height)
    else:
        dim = None
    if not sgrp:
        sgrp = WmhdbHasSchedulingGroupP.DEFAULT_SCHED_GROUP
    fclst = [
        (WHDB_DEFAULT_XPOINT_CODE, None, xdsc.kwargs)
        for xdsc in oper.createInitialConnectionPoints()]

    return block.createOperatorInstance(
        oper.getCode(), name, info, fclst, prms, **{
            WHDB_POSITION_KEY: pos or (0, 0),
            WHDB_Z_VALUE_KEY: zpos,
            WHDB_DIMENSION_KEY: dim,
            WHDB_ROTATION_KEY: rotation or 0,
            WHDB_HFLIP_KEY: hflip or False,
            WHDB_VFLIP_KEY: vflip or False,
            WMHDB_SGROUP_KEY: sgrp})
