# ----------------------------------------------------------------------------
# - wmhdb-connection-point                                                   -
# - wild cruncher:wsh connector object                     core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_connection_point
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher database connector

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from wcr_core.db.whdb_connection_point import WhdbConnectionPoint, \
    WHDB_XPNT_SHOW_TARGET
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaDescription, \
    WdbsMetaField
from wcr_core.dbstore.wdbs_meta_type import WDBS_BOOL_META


WMHDB_CONNECTOR_INVPOL = 'inv_pol'

WMHDB_XPNT_SHOW_POLARITY = 'showPolarity'


class WmhdbConnector(WhdbConnectionPoint):
    """
    The WmhdbConnector class is the silkan hierarchy connector object. A
    connector is defined by name with an additional port id. A direction
    attributes is used to qualify the connector.
    """

    #: Connection point types
    # noinspection PyArgumentList
    Input, Output = range(1, 3)

    #: Connection point meta-data
    META = WdbsMetaDescription(
        WhdbConnectionPoint.META,
        WdbsMetaField(WMHDB_CONNECTOR_INVPOL, WDBS_BOOL_META))

    def __init__(self, xpoint, wreg):
        """
        :param xpoint: Raw connection point
        :type xpoint: WdbsConnectionPointP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbConnectionPoint.__init__(self, xpoint, wreg)

        self.__rawXPoint = xpoint

    def isInvertedPolarity(self):
        """
        Whether the connector's polarity is inverted.

        :return: True iff the connector's polarity is inverted.
        :rtype: bool
        """
        return self.__rawXPoint.getMeta(WMHDB_CONNECTOR_INVPOL) or False

    def setInvertedPolarity(self, invPol):
        """
        Set the polarity as inverted or regular.

        :param invPol: Whether the polarity is inverted
        :type invPol: bool
        """
        return self.__rawXPoint.setMeta(invPol, WMHDB_CONNECTOR_INVPOL)

    def getDescription(self, **kwargs):
        """
        Get the formatted name for the given connector.

        :param showPort: Whether to include the port information in the name
        :type showPort: bool
        :param showPolarity: Whether to include the polarity in the name
        :type showPolarity: bool
        :param portNameF: Function for resolving the port nam
        :type portNameF: (str) -> str
        """
        name = self.getName()
        ipol = self.isInvertedPolarity()
        showPort = bool(kwargs.get(WHDB_XPNT_SHOW_TARGET))
        showPolarity = bool(kwargs.get(WMHDB_XPNT_SHOW_POLARITY))
        portNameF = kwargs.get('portNameF')

        if showPort:
            port = self.getPortReference()
            port = portNameF(port) if portNameF else port
        else:
            port = None
        pol = ', inv. pol.' if showPolarity and ipol else ''

        if name:
            if port:
                return '{n} -> {p}{pol}'.format(
                    n=name, p=port, pol=pol)
            else:
                return '{n}{pol}'.format(
                    n=name, pol=pol)
        else:
            if port:
                return '-> {p}{pol}'.format(
                    p=port, pol=pol)
            else:
                return 'inv. pol.' if showPolarity and ipol else ''


# noinspection PyUnusedLocal
def wmhdb_getConnectorDefaultGravity(ctyp):
    """
    Get the default gravity for a connector.

    :param ctyp: Connector type (input/output)
    :type ctyp: int
    :return: Default gravity for the given connector
    :rtype: int
    """
    if ctyp == WmhdbConnector.Input:
        return WmhdbConnector.West
    else:
        return WmhdbConnector.East
