# ----------------------------------------------------------------------------
# - wmhdb-connection-point-manager                                           -
# - wild cruncher: connection point manager                 module hierarchy -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_connection_point_manager
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher connection point manager

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import OrderedDict
from operator import methodcaller
from itertools import repeat, groupby
import re

from wcr_core.db.whdb_operator import whdb_makeConnPositions, \
    whdb_makeConnXYPositions, whdb_getXPointDefaultGravity
from wcr_core.db.whdb_connection_point import WHDB_DEFAULT_XPOINT_CODE
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_OUT, \
    WDBS_XPNT_IN
from wcr_core.modules.model.db.wmhdb_connection_point import WmhdbConnector, \
    wmhdb_getConnectorDefaultGravity
from wcr_core.db.whdb_connection_point_manager import *


#: Connection point direction names
WMHDB_XPOINT_NAMES = {
    WmhdbConnector.Input: 'Input',
    WmhdbConnector.Output: 'Output'}


def _wmhdb_getNextXPointName(clst, cdir, prfx=None, rgxp=None):
    """
    Get the next name for a connection point of the given direction for the
    given connection point list.

    :param clst: Connection point list
    :type clst: list[WhdbConnectionPointP]
    :param cdir: Direction
    :type cdir: int
    :param prfx: Name prefix (value will be determined based on cdir if `None`)
    :type prfx: str | None
    :param rgxp: Regular expression to extract names
    :type rgxp: SRE_Pattern | None
    :return: Next name
    :rtype: str
    """
    if prfx is None:
        prfx = 'i' if cdir == WDBS_XPNT_IN else 'o'
    if rgxp is None:
        rgxp = re.compile(prfx + '\s*(\d+)$')
    cnmm = 0
    _clst = [c for c in clst if c.getDirection() == cdir]
    for c in _clst:
        m = rgxp.search(c.getName())
        if m:
            cnmm = max(int(m.group(1)), cnmm)
    return '{p}{c}'.format(p=prfx, c=cnmm + 1)


def _wmhdb_resetXPointPositions(clst):
    """
    Reset the positions of the connection points in the given list.

    :param clst: Connection point list
    :type clst: list[WhdbConnectionPointP]
    """
    keyf = methodcaller('getDirection')
    clst = list(sorted(clst, key=keyf))
    for cdir, xpnts in groupby(clst, keyf):
        xpnts = list(xpnts)
        for xpnt, (x, y, f) in zip(
                xpnts, whdb_makeConnXYPositions(cdir, len(xpnts))):
            xpnt.setPosition(x, y)
            xpnt.setGravity(f)


class WmhdbFixedConnectionPointManager(WhdbFixedConnectionPointManager):
    """
    Connection point manager for fixed connection point sets with optional
    connection points.
    """

    def __init__(self, inls, opts=None):
        """
        :param inls: Initial connection points points
        :type inls: Iterable[WhdbXPointDesc]
        :param opts: Optional connection points
        :type opts: Iterable[dict[str, object]]
        """
        super().__init__(inls)

        if opts:
            self.__optionals = OrderedDict()
            for opt in opts:
                self.__optionals[opt[WHDB_XPOINT_NAME]] = \
                    (opt['title'], WHDB_DYN_XPOINT_OPT, opt)
        else:
            self.__optionals = {}

    def getDynamicXPoints(self):
        return self.__optionals.values()

    def _batchGetName(self, xpnt):
        """
        Get the name of the connection point data set.

        :param xpnt: Connection point data
        :type xpnt: dict[str, object]
        :return: Name
        :rtype: str
        """
        return xpnt.get(WHDB_XPOINT_NAME)

    def _instanceGetName(self, xpnt):
        """
        Get the name of the connection point.

        :param xpnt: Connection point
        :type xpnt: WhdbConnectionPointP
        :return: Name
        :rtype: str
        """
        return xpnt.getName()

    def _getCreatableXPoints(self, nameF, xpnts, **kwargs):
        """
        Return a list of dynamic connection point descriptions that can be
        created with the given characteristics in the given connection point
        set.

        A connection point can be created if it is one of the optional
        connection points, and does not yet exists in the given list of
        connection points.

        :param nameF: Name retrieval function
        :type nameF: (object) -> str
        :param xpnts: Existing connection points
        :type xpnts: Iterable[object]
        :param kwargs: Characteristics of new connection point
        :type kwargs: dict[str, object]
        :return: List of dynamic connection point descriptions
        :rtype: list[(str, str, dict[str, object])]
        """
        if not isinstance(xpnts, list):
            xpnts = list(xpnts)
        # noinspection PyUnresolvedReferences
        mtch = [opt for _, opt in self.__optionals.items()
                if (all(opt[2].get(k) == kwargs[k]
                        for k in kwargs.keys() & opt[2].keys()) and
                    not any(nameF(x) == opt[2][WHDB_XPOINT_NAME]
                            for x in xpnts))]
        return mtch

    def canBatchCreateXPoint(self, xpnts, **kwargs):
        return bool(self._getCreatableXPoints(
            self._batchGetName, xpnts, **kwargs))

    def canInstanceCreateXPoint(self, clst, **kwargs):
        return bool(self._getCreatableXPoints(
            self._instanceGetName, clst, **kwargs))

    def createInstanceXPoint(self, clst, connF, **kwargs):
        def getValue(key):
            return mtch[0][2].get(key) or kwargs.get(key)

        mtch = self._getCreatableXPoints(self._instanceGetName, clst, **kwargs)
        assert len(mtch) == 1

        name = getValue(WHDB_XPOINT_NAME)
        cdir = getValue(WHDB_XPOINT_DIRECTION)
        cpos = getValue(WHDB_XPOINT_POSITION)
        fcng = getValue(WHDB_XPOINT_FACING)
        if not fcng:
            fcng = wmhdb_getConnectorDefaultGravity(cdir)

        xpnt = wmhdb_createXPointDesc(
            cdir=cdir, name=name, fcng=fcng, cpos=cpos)
        return connF(xpnt)

    def canBatchRemoveXPoint(self, xpnts, **kwargs):
        return kwargs.get(WHDB_XPOINT_NAME) in self.__optionals

    def canInstanceRemoveXPoint(self, clst, xpnt):
        return xpnt.getName() in self.__optionals

    def removeInstanceXPoint(self, clst, xpnt):
        clst.remove(xpnt)

    def setInstanceXPoints(self, clst, connF, data):
        # TODO
        raise Exception('Cannot set connection points')

    def areValidXPoints(self, clst):
        vld = set(x.getName() for x in clst
                  if x.getName() not in self.__optionals) == \
            set(x[WHDB_XPOINT_NAME] for x in self.getInitialXPointSet())

        return vld, ['Invalid connection points'] if not vld else None

    def resetXPointPositions(self, clst):
        _wmhdb_resetXPointPositions(clst)


class _WmhdbVarConnectionPointManager(WhdbConnectionPointManagerBase):
    """
    Connection point manager for connection point sets variable in one
    direction.
    """

    def __init__(self, cdir, prfx, fcng, fxds,
                 dflc=None, minc=None, maxc=None, opts=None):
        """
        :param cdir: Direction of variable connection points
        :type cdir: int
        :param prfx: Prefix for names of variable connection points
        :type prfx: str
        :param fcng: Facing/gravity of variable connection points
        :type fcng: int
        :param fxds: Fixed connection points
        :type fxds: list[WhdbXPointDesc]
        :param dflc: Default count of variable connection points
        :type dflc: int
        :param minc: Minimum count of variable connection points
        :type minc: int
        :param maxc: Maximum count of variable connection points
        :type maxc: int
        :param opts: Optional fixed connection points
        :type opts: Iterable[dict[str, object]]
        """

        #: Direction of variable connection points
        self.__direction = cdir
        #: Prefix for names of variable connection points
        self.__prefix = prfx
        #: Facing/gravity of variable connection points
        self.__facing = fcng
        #: Fixed connection points
        self.__fixed = fxds
        #: Minimum count of variable connection points
        self.__minCount = minc
        #: Maximum count of variable connection points
        self.__maxCount = maxc

        if opts:
            self.__optionals = OrderedDict()
            for opt in opts:
                self.__optionals[opt[WHDB_XPOINT_NAME]] = \
                    (opt['title'], WHDB_DYN_XPOINT_OPT, opt)
        else:
            self.__optionals = {}

        self.__xpos = 1.0 if fcng == WmhdbConnector.East else -1.0
        self.__nameRe = re.compile(prfx + '\s*(\d+)$')

        initCount = dflc or minc or 0

        if initCount:
            self.__initials = fxds + [
                wmhdb_createXPointDesc(
                    cdir=cdir, name='{p}{c}'.format(p=prfx, c=c),
                    cpos=(self.__xpos, ypos), fcng=fcng)
                for c, ypos in zip(range(1, 1 + initCount),
                                   whdb_makeConnPositions(initCount))]
        else:
            self.__initials = fxds

    def getInitialXPointSet(self):
        return self.__initials

    def getDynamicXPoints(self):
        xvrs = [(WMHDB_XPOINT_NAMES[self.__direction],
                 WHDB_DYN_XPOINT_VAR,
                 {WHDB_XPOINT_DIRECTION: self.__direction})]
        return xvrs + list(self.__optionals.values())

    def _batchGetName(self, xpnt):
        """
        Get the name of the connection point data set.

        :param xpnt: Connection point data
        :type xpnt: dict[str, object]
        :return: Name
        :rtype: str
        """
        return xpnt.get(WHDB_XPOINT_NAME)

    def _batchGetDirection(self, xpnt):
        """
        Get the direction of the connection point data set.

        :param xpnt: Connection point data
        :type xpnt: dict[str, object]
        :return: Direction
        :rtype: int
        """
        return xpnt.get(WHDB_XPOINT_DIRECTION)

    def _instanceGetName(self, xpnt):
        """
        Get the name of the connection point.

        :param xpnt: Connection point
        :type xpnt: WhdbConnectionPointP
        :return: Name
        :rtype: str
        """
        return xpnt.getName()

    def _instanceGetDirection(self, xpnt):
        """
        Get the direction of the connection point.

        :param xpnt: Connection point
        :type xpnt: WhdbConnectionPointP
        :return: Direction
        :rtype: int
        """
        return xpnt.getDirection()

    def _count(self, dirF, xpnts):
        """
        Count the connection points in the variable direction.

        :param dirF: Connection point direction function
        :type dirF: (object) -> int
        :param xpnts: Connection points
        :type xpnts: Iterable[object]
        :return: Count
        :rtype: int
        """
        return len([x for x in xpnts if dirF(x) == self.__direction])

    def _getCreatableXPoints(self, dirF, nameF, xpnts, **kwargs):
        """
        Return a list of dynamic connection point descriptions that can be
        created with the given characteristics in the given connection point
        set.

        A connection point can be created if
        - it is one of the optional connection points, and does not yet exists
          in the given list of connection points, or
        - it is a variable connection point and the maximum number of
          variable connection points has not yet been reached.

        :param dirF: Connection point direction function
        :type dirF: (object) -> int
        :param nameF: Name retrieval function
        :type nameF: (object) -> str
        :param xpnts: Existing connection points
        :type xpnts: Iterable[object]
        :param kwargs: Characteristics of new connection point
        :type kwargs: dict[str, object]
        :return: List of dynamic connection point descriptions
        :rtype: list[(str, str, dict[str, object])]
        """
        if not isinstance(xpnts, list):
            xpnts = list(xpnts)
        # noinspection PyUnresolvedReferences
        mtch = [opt for _, opt in self.__optionals.items()
                if (all(opt[2].get(k) == kwargs[k]
                        for k in kwargs.keys() & opt[2].keys()) and
                    not any(nameF(x) == opt[2][WHDB_XPOINT_NAME]
                            for x in xpnts))]

        if kwargs.get(WHDB_XPOINT_DIRECTION) == self.__direction:
            if self.__maxCount is None or \
                    self._count(dirF, xpnts) < self.__maxCount:
                mtch.append((WMHDB_XPOINT_NAMES[self.__direction],
                             WHDB_DYN_XPOINT_VAR,
                             {WHDB_XPOINT_DIRECTION: self.__direction}))

        return mtch

    def canBatchCreateXPoint(self, xpnts, **kwargs):
        return bool(self._getCreatableXPoints(
            self._batchGetDirection, self._batchGetName, xpnts, **kwargs))

    def canInstanceCreateXPoint(self, clst, **kwargs):
        return bool(self._getCreatableXPoints(
            self._instanceGetDirection, self._instanceGetName, clst, **kwargs))

    def _canRemoveXPoint(self, dirF, nameF, xpnt, xpnts):
        """
        Whether the given connection point may be removed.

        A connection point can be removed if
        - it is one of the optional connection points, and exists in the given
          list of connection points, or
        - it is a variable connection point and the minimum number of
          variable connection points has not yet been reached.

        :param dirF: Connection point direction function
        :type dirF: (object) -> int
        :param nameF: Name retrieval function
        :type nameF: (object) -> str
        :param xpnt: Connection point
        :type xpnt: object
        :param xpnts: Existing connection points
        :type xpnts: Iterable[object]
        :return: True iff the given connection point can be removed
        :rtype: bool
        """
        if self.__optionals:
            name = nameF(xpnt)
            if name and name in self.__optionals:
                return True

        if dirF(xpnt) != self.__direction:
            return False
        elif not self.__minCount:
            return True
        else:
            return self._count(dirF, xpnts) > self.__minCount

    def canBatchRemoveXPoint(self, xpnts, **kwargs):
        return self._canRemoveXPoint(
            self._batchGetDirection, self._batchGetName, kwargs, xpnts)

    def canInstanceRemoveXPoint(self, clst, xpnt):
        return self._canRemoveXPoint(
            self._instanceGetDirection, self._instanceGetName, xpnt, clst)

    def createInstanceXPoint(self, clst, connF, **kwargs):
        name = kwargs.get(WHDB_XPOINT_NAME)
        if name and name in self.__optionals:
            cdir = kwargs[WHDB_XPOINT_DIRECTION]
            fcng = kwargs.get(WHDB_XPOINT_FACING)
            if not fcng:
                fcng = wmhdb_getConnectorDefaultGravity(cdir)
        else:
            cdir = self.__direction
            fcng = kwargs.get(WHDB_XPOINT_FACING) or self.__facing
            if not name:
                name = _wmhdb_getNextXPointName(
                    clst, cdir, self.__prefix, self.__nameRe)

        xpnt = wmhdb_createXPointDesc(
            cdir=cdir, name=name, fcng=fcng,
            cpos=kwargs.get(WHDB_XPOINT_POSITION))
        return connF(xpnt)

    def removeInstanceXPoint(self, clst, xpnt):
        clst.remove(xpnt)

    def setInstanceXPoints(self, clst, connF, data):
        # TODO
        raise Exception('Cannot set connection points')

    def areValidXPoints(self, clst):
        def dynXpnts(xpnt):
            return xpnt.getDirection() == self.__direction and \
                not xpnt.getName() in self.__optionals

        dyns = [x for x in clst if dynXpnts(x)]

        if not ((self.__minCount is None or self.__minCount <= len(dyns)) and
                    (self.__maxCount is None or len(dyns) <= self.__maxCount)):
            return False, [
                'Must have between {mn} and {mx} {tp}-connectors, '
                'has {cn}'.format(
                    self.__minCount or 0,
                    self.__maxCount or '\u221E',
                    WMHDB_XPOINT_NAMES[self.__direction], len(dyns))]

        return True, None

    def resetXPointPositions(self, clst):
        _wmhdb_resetXPointPositions(clst)


class WmhdbVarInConnectionPointManager(_WmhdbVarConnectionPointManager):
    """
    Connection point manager for variable input connection point sets with
    optional connection points.
    """

    def __init__(self, fxds, prfx=None, minc=None, maxc=None, opts=None):
        """
        :param fxds: Fixed connection points
        :type fxds: list[WhdbXPointDesc]
        :param prfx: Prefix for names of variable connection points
        :type prfx: str | None
        :param minc: Minimum count of variable connection points
        :type minc: int | None
        :param maxc: Maximum count of variable connection points
        :type maxc: int | None
        :param opts: Optional fixed connection points
        :type opts: Iterable[dict[str, object]] | None
        """
        super().__init__(
            WmhdbConnector.Input, prfx if prfx is not None else 'i',
            WmhdbConnector.West, fxds, minc=minc, maxc=maxc, opts=opts)


class WmhdbVarOutConnectionPointManager(_WmhdbVarConnectionPointManager):
    """
    Connection point manager for variable output connection point sets with
    optional connection points.
    """

    def __init__(self, fxds, prfx=None, minc=None, maxc=None, opts=None):
        """
        :param fxds: Fixed connection points
        :type fxds: list[WhdbXPointDesc]
        :param prfx: Prefix for names of variable connection points
        :type prfx: str | None
        :param minc: Minimum count of variable connection points
        :type minc: int | None
        :param maxc: Maximum count of variable connection points
        :type maxc: int | None
        :param opts: Optional fixed connection points
        :type opts: Iterable[dict[str, object]] | None
        """
        super().__init__(
            WmhdbConnector.Output, prfx if prfx is not None else 'o',
            WmhdbConnector.East, fxds, minc=minc, maxc=maxc, opts=opts)


class WmhdbDualVarConnectionPointManager(WhdbConnectionPointManagerBase):
    """
    Connection point manager for connection point sets variable in one
    direction.
    """

    def __init__(self, iprfx, oprfx,
                 iminc=None, imaxc=None, ominc=None, omaxc=None):
        """
        :param iprfx: Prefix for names of input connection points
        :type iprfx: str
        :param oprfx: Prefix for names of output connection points
        :type oprfx: str
        :param iminc: Minimum count of input connection points
        :type iminc: int
        :param imaxc: Maximum count of input connection points
        :type imaxc: int
        :param ominc: Minimum count of output connection points
        :type ominc: int
        :param omaxc: Maximum count of output connection points
        :type omaxc: int
        """
        #: Prefix for names of input connection points
        self.__inPrefix = iprfx
        #: Minimum count of input connection points
        self.__inMinCount = iminc
        #: Maximum count of input connection points
        self.__inMaxCount = imaxc

        #: Prefix for names of output connection points
        self.__outPrefix = oprfx
        #: Minimum count of output connection points
        self.__outMinCount = ominc
        #: Maximum count of output connection points
        self.__outMaxCount = omaxc

        self.__inNameRe = re.compile(iprfx + '\s*(\d+)$')
        self.__outNameRe = re.compile(oprfx + '\s*(\d+)$')

        initials = []
        if iminc:
            initials += [
                wmhdb_createXPointDesc(
                    cdir=WDBS_XPNT_IN, name='{p}{c}'.format(p=iprfx, c=c),
                    cpos=(-1.0, ypos), fcng=WmhdbConnector.West)
                for c, ypos in zip(range(1, 1 + iminc),
                                   whdb_makeConnPositions(iminc))]
        if ominc:
            initials += [
                wmhdb_createXPointDesc(
                    cdir=WDBS_XPNT_OUT, name='{p}{c}'.format(p=oprfx, c=c),
                    cpos=(1.0, ypos), fcng=WmhdbConnector.East)
                for c, ypos in zip(range(1, 1 + ominc),
                                   whdb_makeConnPositions(ominc))]
        self.__initials = tuple(initials)

    def getInitialXPointSet(self):
        return self.__initials

    def getDynamicXPoints(self):
        return [(WMHDB_XPOINT_NAMES[WDBS_XPNT_IN],
                 WHDB_DYN_XPOINT_VAR,
                 {WHDB_XPOINT_DIRECTION: WDBS_XPNT_IN}),
                (WMHDB_XPOINT_NAMES[WDBS_XPNT_OUT],
                 WHDB_DYN_XPOINT_VAR,
                 {WHDB_XPOINT_DIRECTION: WDBS_XPNT_OUT})]

    def _batchGetDirection(self, xpnt):
        """
        Get the direction of the connection point data set.

        :param xpnt: Connection point data
        :type xpnt: dict[str, object]
        :return: Direction
        :rtype: int
        """
        return xpnt.get(WHDB_XPOINT_DIRECTION)

    def _instanceGetDirection(self, xpnt):
        """
        Get the direction of the connection point.

        :param xpnt: Connection point
        :type xpnt: WhdbConnectionPointP
        :return: Direction
        :rtype: int
        """
        return xpnt.getDirection()

    def _count(self, cdir, dirF, xpnts):
        """
        Count the connection points in the variable direction.

        :param cdir: Direction to return count for
        :type cdir: int
        :param dirF: Connection point direction function
        :type dirF: (object) -> int
        :param xpnts: Connection points
        :type xpnts: Iterable[object]
        :return: Count
        :rtype: int
        """
        return len([x for x in xpnts if dirF(x) == cdir])

    def _getCreatableXPoints(self, dirF, xpnts, **kwargs):
        """
        Return a list of dynamic connection point descriptions that can be
        created with the given characteristics in the given connection point
        set.

        A connection point can be created if
        - it is one of the optional connection points, and does not yet exists
          in the given list of connection points, or
        - it is a variable connection point and the maximum number of
          variable connection points has not yet been reached.

        :param dirF: Connection point direction function
        :type dirF: (object) -> int
        :param nameF: Name retrieval function
        :type nameF: (object) -> str
        :param xpnts: Existing connection points
        :type xpnts: Iterable[object]
        :param kwargs: Characteristics of new connection point
        :type kwargs: dict[str, object]
        :return: List of dynamic connection point descriptions
        :rtype: list[(str, str, dict[str, object])]
        """
        if not isinstance(xpnts, list):
            xpnts = list(xpnts)

        mtch = []

        cdir = kwargs.get('cdir')

        if (cdir is None or cdir == WDBS_XPNT_IN) and \
                self.__inMaxCount is None or \
                self._count(WDBS_XPNT_IN, dirF, xpnts) < self.__inMaxCount:
            mtch.append((WMHDB_XPOINT_NAMES[WDBS_XPNT_IN],
                         WHDB_DYN_XPOINT_VAR,
                         {WHDB_XPOINT_DIRECTION: WDBS_XPNT_IN}))
        if (cdir is None or cdir == WDBS_XPNT_OUT) and \
                self.__outMaxCount is None or \
                self._count(WDBS_XPNT_OUT, dirF, xpnts) < self.__outMaxCount:
            mtch.append((WMHDB_XPOINT_NAMES[WDBS_XPNT_OUT],
                         WHDB_DYN_XPOINT_VAR,
                         {WHDB_XPOINT_DIRECTION: WDBS_XPNT_OUT}))

        return mtch

    def canBatchCreateXPoint(self, xpnts, **kwargs):
        return bool(self._getCreatableXPoints(
            self._batchGetDirection, xpnts))

    def canInstanceCreateXPoint(self, clst, **kwargs):
        return bool(self._getCreatableXPoints(
            self._instanceGetDirection, clst))

    def _canRemoveXPoint(self, dirF, xpnt, xpnts):
        """
        Whether the given connection point may be removed.

        A connection point can be removed if
        - it is one of the optional connection points, and exists in the given
          list of connection points, or
        - it is a variable connection point and the minimum number of
          variable connection points has not yet been reached.

        :param dirF: Connection point direction function
        :type dirF: (object) -> int
        :param xpnt: Connection point
        :type xpnt: object
        :param xpnts: Existing connection points
        :type xpnts: Iterable[object]
        :return: True iff the given connection point can be removed
        :rtype: bool
        """
        cdir = dirF(xpnt)

        if cdir == WDBS_XPNT_IN:
            minc = self.__inMinCount or 0
        else:
            minc = self.__outMinCount or 0

        if not minc:
            return True
        else:
            return self._count(cdir, dirF, xpnts) > minc

    def canBatchRemoveXPoint(self, xpnts, **kwargs):
        return self._canRemoveXPoint(
            self._batchGetDirection, kwargs, xpnts)

    def canInstanceRemoveXPoint(self, clst, xpnt):
        return self._canRemoveXPoint(
            self._instanceGetDirection, xpnt, clst)

    def createInstanceXPoint(self, clst, connF, **kwargs):
        assert WHDB_XPOINT_DIRECTION in kwargs

        name = kwargs.get(WHDB_XPOINT_NAME)
        cdir = kwargs.get(WHDB_XPOINT_DIRECTION)
        fcng = kwargs.get(WHDB_XPOINT_FACING)

        isInput = cdir == WDBS_XPNT_IN

        if not name:
            prfx = self.__inPrefix if isInput else self.__outPrefix
            nmre = self.__inNameRe if isInput else self.__outNameRe

            name = _wmhdb_getNextXPointName(clst, cdir, prfx, nmre)
        if fcng is None:
            fcng = WmhdbConnector.West if isInput else WmhdbConnector.East

        xpnt = wmhdb_createXPointDesc(
            cdir=cdir, name=name, fcng=fcng,
            cpos=kwargs.get(WHDB_XPOINT_POSITION))
        return connF(xpnt)

    def removeInstanceXPoint(self, clst, xpnt):
        clst.remove(xpnt)

    def setInstanceXPoints(self, clst, connF, data):
        # TODO
        raise Exception('Cannot set connection points')

    def areValidXPoints(self, clst):
        inCnt = len([x for x in clst if x.getDirection() == WDBS_XPNT_IN])
        outCnt = len([x for x in clst if x.getDirection() == WDBS_XPNT_OUT])

        errs = []

        if not ((self.__inMinCount is None or self.__inMinCount <= inCnt) and
                (self.__inMaxCount is None or inCnt <= self.__inMaxCount)):
            errs.append(
                'Must have between {mn} and {mx} {tp}-connectors, '
                'has {cn}'.format(
                    self.__inMinCount or 0,
                    self.__inMaxCount or '\u221E',
                    WMHDB_XPOINT_NAMES[WDBS_XPNT_IN], inCnt))
        if not ((self.__outMinCount is None or
                 self.__outMinCount <= outCnt) and
                (self.__outMaxCount is None or outCnt <= self.__outMaxCount)):
            errs.append(
                'Must have between {mn} and {mx} {tp}-connectors, '
                'has {cn}'.format(
                    self.__outMinCount or 0,
                    self.__outMaxCount or '\u221E',
                    WMHDB_XPOINT_NAMES[WDBS_XPNT_OUT], outCnt))

        return not errs, errs

    def resetXPointPositions(self, clst):
        _wmhdb_resetXPointPositions(clst)


class WmhdbSimpleConnectionPointManager(WhdbConnectionPointManagerBase):
    """
    Simple connection point manager without any restrictions.
    """

    def __init__(self, ipfx=None, opfx=None):
        """
        :param ipfx: Name prefix for input connectors
        :type ipfx: str
        :param opfx: Name prefix for output connectors
        :type opfx: str
        """
        self.__ipfx = ipfx if ipfx is not None else 'i'
        self.__opfx = opfx if opfx is not None else 'o'

    def getInitialXPointSet(self):
        return []

    def getDynamicXPoints(self):
        return [(WMHDB_XPOINT_NAMES[WDBS_XPNT_IN],
                 WHDB_DYN_XPOINT_VAR,
                 {WHDB_XPOINT_DIRECTION: WDBS_XPNT_IN}),
                (WMHDB_XPOINT_NAMES[WDBS_XPNT_OUT],
                 WHDB_DYN_XPOINT_VAR,
                 {WHDB_XPOINT_DIRECTION: WDBS_XPNT_OUT})]

    def canBatchCreateXPoint(self, xpnts, **kwargs):
        return True

    def canInstanceCreateXPoint(self, clst, **kwargs):
        return True

    def canBatchRemoveXPoint(self, xpnts, **kwargs):
        return True

    def canInstanceRemoveXPoint(self, clst, xpnt):
        return True

    def createInstanceXPoint(self, clst, connF, **kwargs):
        cdir = kwargs.get(WHDB_XPOINT_DIRECTION)
        if not cdir:
            raise ValueError(
                'Cannot create connection point without direction')
        cpos = kwargs.get(WHDB_XPOINT_POSITION) or \
               ((-1, 0) if cdir == WDBS_XPNT_IN else (1, 0))
        fcng = kwargs.get(WHDB_XPOINT_FACING) or \
               whdb_getXPointDefaultGravity(*cpos)
        name = kwargs.get(WHDB_XPOINT_NAME)
        if not name:
            prfx = self.__ipfx if cdir == WDBS_XPNT_IN else self.__opfx
            name = _wmhdb_getNextXPointName(clst, cdir, prfx=prfx)

        xpnt = wmhdb_createXPointDesc(
            cdir=cdir, name=name, fcng=fcng,
            cpos=kwargs.get(WHDB_XPOINT_POSITION))
        return connF(xpnt)

    def removeInstanceXPoint(self, clst, xpnt):
        clst.remove(xpnt)

    def setInstanceXPoints(self, clst, connF, data):
        # TODO
        raise Exception('Cannot set connection points')

    def areValidXPoints(self, clst):
        return True, None

    def resetXPointPositions(self, clst):
        _wmhdb_resetXPointPositions(clst)


def wmhdb_createFixedXPointManager(icns=None, ocns=None, ipos=None,
                                   opos=None, opts=None):
    """
    Create a connection point manager for fixed connection point sets.

    :param icns: Input connection point names, if any
    :type icns: list[str] | None
    :param ocns: Output connection point names, if any
    :type ocns: list[str] | None
    :param ipos: Input connection point positions / facings, if any
    :type ipos: list[(float, float, int)] | None
    :param opos: Output connection point positions / facings, if any
    :type opos: list[(float, float, int)] | None
    :param opts: Optional connection points
    :type opts: Iterable[dict[str, object]] | None
    :return: Connection point manager
    :rtype: WhdbFixedConnectionPointManager
    """

    def mkXPoints(cdir, ns, ps):
        if not ns:
            return []
        if ps:
            assert len(ps) == len(ns)
        else:
            ps = whdb_makeConnXYPositions(cdir, len(ns))
        return list(zip(repeat(cdir), ns,
                        [(x, y) for x, y, _ in ps], [f for _, _, f in ps]))

    xpnts = [wmhdb_createXPointDesc(cdir=cdir, name=name,
                                    cpos=cpos, fcng=fcng)
             for cdir, name, cpos, fcng,
             in mkXPoints(WmhdbConnector.Input, icns, ipos) +
             mkXPoints(WmhdbConnector.Output, ocns, opos)]
    if opts:
        return WmhdbFixedConnectionPointManager(xpnts, opts)
    else:
        return WhdbFixedConnectionPointManager(xpnts)


def wmhdb_createXPointDesc(cdir, name=None, port=None, cpos=None, fcng=None,
                           xpnt=None, **kwargs):
    """
    Create a connection point description.

    :param cdir: Direction
    :type cdir: int | None
    :param name: Name
    :type name: str | None
    :param port: Port name
    :type port: port | None
    :param cpos: Position
    :type cpos: (float, float) | None
    :param fcng: Facing/gravity
    :type fcng: int | None
    :param xpnt: Reference connection point
    :type xpnt: WhdbConnectionPointP
    :param kwargs: Additional arguments
    :type kwargs: dict[str, object]
    :return: Connection point description
    :rtype: WhdbXPointDesc
    """
    assert WHDB_XPOINT_TYPE not in kwargs
    assert 'gvty' not in kwargs

    isin = cdir == WmhdbConnector.Input
    return WhdbXPointDesc(
        ctyp=WHDB_DEFAULT_XPOINT_CODE,
        cpos=cpos or ((-1.0, 0.0) if isin else (1.0, 0.0)),
        fcng=fcng or (WmhdbConnector.West if isin else WmhdbConnector.East),
        cdir=cdir, name=name, port=port, xpnt=xpnt, **kwargs)
