# ----------------------------------------------------------------------------
# - wmhdb-function                                                           -
# - wild cruncher:wsh model function object                core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_function
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher model function object

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from wcr_core.db.whdb_instance import WhdbOperatorInstance
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_FACING
from wcr_core.db.whdb_connection_point import WHDB_DEFAULT_XPOINT_CODE
from wcr_core.modules.model.db import wmhdb_standard_operators
from wcr_core.modules.model.db.wmhdb_instance_base import WmhdbInstanceBaseP
from wcr_core.modules.model.db.wmhdb_operator import WMHDB_OPER
from wcr_core.modules.model.db.wmhdb_connection_point import WmhdbConnector, \
    WMHDB_CONNECTOR_INVPOL


WMHDB_OP_PARAMS_KEY = 'params'


class WmhdbFunction(WhdbOperatorInstance, WmhdbInstanceBaseP):
    """
    The WcrFunction class is the Silkan hierarchical function object. A
    function is an operator model associated with a list of local parameters.
    The function object is similar to the instance object, but the duality
    model/instance connectors is merged into a single representation.
    The mime type is fixed for a function but an operator type is defined for
    it. If the operator is not defined, it is assumed to be custom.
    """
    def __init__(self, inst, wreg):
        """
        :param inst: Raw operator instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Raw element wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        WhdbOperatorInstance.__init__(
            self, WMHDB_OPER[inst.getOperator()], inst, wreg)
        WmhdbInstanceBaseP.__init__(self, inst, wreg)

        self.__rawInstance = inst
        self.__wrapperRegistry = wreg

    def getLabelAlignment(self):
        return self.getOperator().getLabelAlignment()

    def getNewConnectionPointDirection(self, **kwargs):
        g = kwargs.get(WHDB_XPOINT_FACING)
        inpt = g in [WmhdbConnector.North, WmhdbConnector.West]
        if inpt:
            if self.canCreateConnectionPoint(cdir=WmhdbConnector.Input):
                return WmhdbConnector.Input
            elif self.canCreateConnectionPoint(cdir=WmhdbConnector.Output):
                return WmhdbConnector.Output
        else:
            if self.canCreateConnectionPoint(cdir=WmhdbConnector.Output):
                return WmhdbConnector.Output
            elif self.canCreateConnectionPoint(cdir=WmhdbConnector.Input):
                return WmhdbConnector.Input
        return None

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        kwargs.setdefault(WMHDB_CONNECTOR_INVPOL, False)
        return super()._createRawConnectionPoint(
            WHDB_DEFAULT_XPOINT_CODE, **kwargs)

    def isConnectorPolarityInvertible(self):
        # noinspection PyUnresolvedReferences
        return self.getOperator().isConnectorPolarityInvertible()

    def getConnectorPolarityNames(self):
        # noinspection PyUnresolvedReferences
        return self.getOperator().getConnectorPolarityNames()

    def getRelatedConnectors(self, conn):
        return [c for c in self.getConnectionPoints()
                if c.getDirection() != conn.getDirection()]

    def resetIcon(self):
        """
        Reset the function's icon to its original size.
        """
        self.unsetDimension()

    def getPortName(self):
        """
        Get the function's port ID, if any.

        :return: The function's port ID, or `None`
        :rtype: str
        """
        return str(self.getParameterValue(
            wmhdb_standard_operators.WMHDB_PORT_NAME_PARAMETER))

    def hasFlag(self, flag, *_):
        return self.getOperator().hasFlag(flag)

    def getOutputTypes(self, inTypes, cpars):
        """
        Compute the output types for the given input types.

        :param inTypes: Input-types by connector
        :type inTypes: dict[str, WmhdbRtiType]
        :param cpars: cold variable parser
        :type cpars: WcldColdVarsParser
        :return: Output-types by connector
        :rtype: dict[str, WmhdbRtiType]
        """
        # noinspection PyUnresolvedReferences
        return self.getOperator().getOutputTypes(
            inTypes, self.getConnectionPoints(), self.getParameterValue, cpars)

    def __str__(self):
        return 'WmhdbFunction {{oper: {op}, icns: {ic}, ocns: {oc}}}'.format(
            op=self.getOperator().getCode(),
            ic=[c.getName() for c
                in self.getConnectionPoints(cdir=WmhdbConnector.Input)],
            oc=[c.getName() for c
                in self.getConnectionPoints(cdir=WmhdbConnector.Output)])
