# ----------------------------------------------------------------------------
# - wmhdb-instance                                                           -
# - wild cruncher:wsh model instance object                core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_instance
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher model instance

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------
from abc import abstractmethod

from wcr_core.db.whdb_instance import WhdbModelInstance, WhdbBlockInstance
from wcr_core.db.whdb_operator import whdb_setXPointPositions
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_FACING
from wcr_core.db.whdb_connection_point import WHDB_DEFAULT_XPOINT_CODE
from wcr_core.db_api.whdb_connection_point_api import WhdbConnectionPointP
from wcr_core.modules.model.db.wmhdb_utils import wmhdb_find_connected
from wcr_core.modules.model.db.wmhdb_instance_base import WmhdbInstanceBaseP
from wcr_core.modules.model.db.wmhdb_function import WmhdbFunction
from wcr_core.modules.model.db.wmhdb_connection_point import WmhdbConnector
from wcr_core.db.whdb_connection_point import WHDB_XPOINT_NAME


# ----------------------------------------------------------------------------
# - instance class                                                           -
# ----------------------------------------------------------------------------


# noinspection PyAbstractClass
class WmhdbInstance(WmhdbInstanceBaseP):
    """
    The WmhdbInstance class is the Silkan hierarchy model instance. An
    instance is a binding between a name and a model. The instance name is
    unique and can be optionally associated with an information string. The
    instance connectors are associated with the model connectors by port id.
    """

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsInstanceP
        :param wreg: Raw element wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        super().__init__(inst, wreg)

        self.__rawInstance = inst
        self.__wrapperRegistry = wreg

    def getConnectionPointDescription(self, conn, **kwargs):
        kwargs['portNameF'] = self.getPortName
        return conn.getDescription(**kwargs)

    def isConnectorPolarityInvertible(self):
        return False

    def getConnectorPolarityNames(self):
        return None

    # noinspection PyMethodMayBeStatic
    def getNewConnectionPointDirection(self, **kwargs):
        g = kwargs.get(WHDB_XPOINT_FACING)
        if g == WhdbConnectionPointP.North \
                or g == WhdbConnectionPointP.West:
            return WmhdbConnector.Input
        else:
            return WmhdbConnector.Output

    def clearConnectors(self):
        """
        Remove all connectors.
        """
        # noinspection PyUnresolvedReferences
        self.clearConnectionPoints()

    def setConnectorOrder(self, orderF):
        """
        Reorder the list of connectors based on the given ordering function.

        :param orderF: Ordering function
        :type orderF: (WmhdbConnector) -> object
        """
        # noinspection PyTypeChecker
        self.setConnectionPointOrder(orderF)

    @abstractmethod
    def getPortName(self, port):
        raise NotImplementedError()


class WmhdbBlockInstance(WmhdbInstance, WhdbBlockInstance):
    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsBlockInstanceP
        :param wreg: Raw element wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        WhdbBlockInstance.__init__(self, inst, wreg)
        WmhdbInstance.__init__(self, inst, wreg)

    def getTypeName(self):
        return '{}-instance'.format(self.getBlock().getTypeName())

    def getModel(self):
        """
        Get the instance model object.

        :return: Model
        :rtype: WmhdbModel
        """
        return self.getBlock()

    def getNewConnectionPointDirection(self, **kwargs):
        return WmhdbInstance.getNewConnectionPointDirection(self, **kwargs)

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        # noinspection PyProtectedMember
        xpnt = super()._createRawConnectionPoint(
            WHDB_DEFAULT_XPOINT_CODE, **kwargs)
        if not xpnt.getMeta(WHDB_XPOINT_NAME):
            xpnt.setMeta(str(xpnt.getUUID()), WHDB_XPOINT_NAME)
        return xpnt

    def getRelatedConnectors(self, conn):
        block = self.getBlock()
        cnam = conn.getPortReference()
        if not cnam:
            return []
        port = next((p for p in block.getOperatorInstances()
                     if p.getPortName() == cnam), None)
        if port is None:
            return []
        connected = wmhdb_find_connected(port, block)
        fncs = (p for p in connected if isinstance(p, WmhdbFunction))
        """:type: Iterable[WmhdbFunction]"""
        ps = set(filter(None, (p.getPortName() for p in fncs)))
        return [c for c in self.getConnectionPoints()
                if c.getPortReference() in ps]

    def hasFlag(self, flag, recurse=True):
        if not recurse:
            return False
        block = self.getBlock()
        blks = block.getBlockInstances()
        """:type: Iterable[WmhdbBlockInstance]"""
        iflgs = (i.hasFlag(flag, recurse) for i in blks)
        fflgs = (f.hasFlag(flag, recurse)
                 for f in block.getOperatorInstances())
        return any(iflgs) or any(fflgs)

    def getPortName(self, port):
        """
        Get the function name for the given port.

        :param port: Port name to look for
        :type port: str
        :return: Port name
        :rtype: str
        """
        block = self.getBlock()
        if port:
            _port = next((f.getName() for f in block.getOperatorInstances()
                          if f.getPortName() == port), None)
            if _port:
                return _port
        return 'PID ' + port if port else None

    def resetConnectionPointPositions(self):
        whdb_setXPointPositions(
            self.getConnectionPoints(cdir=WmhdbConnector.Input))
        whdb_setXPointPositions(
            self.getConnectionPoints(cdir=WmhdbConnector.Output))


class WmhdbModelInstance(WhdbModelInstance, WmhdbInstance):
    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsModelInstanceP
        :param wreg: Raw element wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        WhdbModelInstance.__init__(self, inst, wreg)
        WmhdbInstance.__init__(self, inst, wreg)

    def getTypeName(self):
        return self.getModel().getTypeName() + '-instance'

    def getNewConnectionPointDirection(self, **kwargs):
        return WmhdbInstance.getNewConnectionPointDirection(self, **kwargs)

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        # noinspection PyProtectedMember
        xpnt = super()._createRawConnectionPoint(
            WHDB_DEFAULT_XPOINT_CODE, **kwargs)
        if not xpnt.getMeta(WHDB_XPOINT_NAME):
            xpnt.setMeta(str(xpnt.getUUID()), WHDB_XPOINT_NAME)
        return xpnt

    def getRelatedConnectors(self, conn):
        return [c for c in self.getConnectionPoints()
                if c.getDirection() != conn.getDirection()]

    def hasFlag(self, flag, recurse=True):
        return False

    def getPortName(self, port):
        """
        Get the function name for the given port.

        :param port: Port name to look for
        :type port: str
        :return: Port name
        :rtype: str
        """
        return 'PID ' + port if port else None

    def resetConnectionPointPositions(self):
        whdb_setXPointPositions(
            self.getConnectionPoints(cdir=WmhdbConnector.Input))
        whdb_setXPointPositions(
            self.getConnectionPoints(cdir=WmhdbConnector.Output))
