# ----------------------------------------------------------------------------
# - wmhdb-operator                                                           -
# - wild cruncher:wsh model operator object                core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_operator
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher model function object

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------
from wcr_core.db.whdb_operator import WhdbOperatorRegistry, WhdbOperator
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_OUT
from wcr_core.modules.model.db.wmhdb_rti_type import WmhdbRtiType, \
    wmhdb_getCommonType


# ----------------------------------------------------------------------------
# - base operator class                                                      -
# ----------------------------------------------------------------------------

WMHDB_DELAY_FLAG = 'delay'

WMHDB_VECTOR_IN_FLAG = 'vector-in'

WMHDB_VECTOR_OUT_FLAG = 'vector-out'


class WmhdbOperator(WhdbOperator):
    """
    The WcrOperator class is the silkan hierarchical operator object. An
    operator is defined by name and info. It is the operator name which
    defines its functionality. The operator contains an icon representation
    as well as a collection of connectors.
    """
    # TODO: Do not extend HasMediaP

    def __init__(self, code, xmngr, name=None, info=None,
                 meta=None, parameters=None, polarities=None, flags=None,
                 icons=None, shape=None, symbol=None, labelAlign=None,
                 outTypeF=None):
        """
        Create a new operator by name and info

        :param code: Operator code
        :type code: str
        :param xmngr: Connector manager
        :type xmngr: WhdbConnectionPointManagerP
        :param name: the operator name
        :type name: str
        :param info: the operator info
        :type info: str
        :param meta: Operator meta data
        :type meta: dict[str, object]
        :param parameters: Operator parameters
                           (name, label, info, value, type, flags)
        :type parameters: list[(str, str, str, str|None, str|None, int|None)]
        :param polarities: Polarity names
        :type polarities: (str, str)
        :param flags: Flags
        :type flags: set[str]
        :param icons: Operator icon paths
        :type icons: dict[str, str]
        :param shape: Operator shape
        :type shape: list[(float, float)]
        :param symbol: Text symbol
        :type symbol: str
        :param labelAlign: Alignment for labels
        :type labelAlign: int
        :param outTypeF: Function to determine output-types, based on the
                         input types, the connection points, and a parameter-
                         value-function
        :type outTypeF: (dict[WmhdbConnector,WmhdbRtiType],
                         list[WmhdbConnector], str -> object) ->
                         dict[WmhdbConnector, WmhdbRtiType]
        """
        params = [(n, l, i, v, f, {'validType': WmhdbOperator._mkType(t)})
                  for n, l, i, v, t, f in parameters] if parameters else None
        super().__init__(code, xmngr, name, info, meta, params,
                         flags, icons, shape, symbol, labelAlign)
        #: Polarity names
        self.__polarities = polarities or []

        #: Output-types function
        self.__outTypeF = outTypeF

    @staticmethod
    def _mkType(tstr):
        if tstr == 'float' or tstr == 'float64':
            return WmhdbRtiType(WmhdbRtiType.F64)
        elif tstr == 'float32':
            return WmhdbRtiType(WmhdbRtiType.F32)
        elif tstr == 'bool':
            return WmhdbRtiType(WmhdbRtiType.BLN)
        elif tstr == 'string':
            return WmhdbRtiType(WmhdbRtiType.STR)
        elif tstr == 'vtype':
            # Value-type (currently equivalent to 'string')
            return WmhdbRtiType(WmhdbRtiType.STR)

    def isConnectorPolarityInvertible(self):
        """
        Whether the polarity of the operator's connectors can be inverted.

        :return: True iff the polarity of the connectors can be inverted
        :rtype: bool
        """
        return bool(self.getConnectorPolarityNames())

    def getConnectorPolarityNames(self):
        """
        Get the operator's connector polarity names, if any.

        :return: Polarity names of the operators (regular, inverted)
        :rtype: (str, str)
        """
        return self.__polarities

    def getOutputTypes(self, inTypes, clst, paramF, cpars):
        """
        Compute the output types for the given input types.

        :param inTypes: Input-types by connector
        :type inTypes: dict[str, WmhdbRtiType]
        :param clst: Connectors of the function
        :type clst: Iterable[WmhdbConnector]
        :param paramF: Parameter-value-function
        :type paramF: (str) -> str
        :param cpars: cold variable parser
        :type cpars: WcldColdVarsParser
        :return: Output-types by connector
        :rtype: dict[WmhdbConnector, WmhdbRtiType]
        """
        if self.__outTypeF:
            if not isinstance(clst, list):
                clst = list(clst)
            otps = self.__outTypeF(inTypes, clst, paramF, cpars)
            return {c.getName(): otps.get(c.getName()) for c in clst
                    if c.getDirection() == WDBS_XPNT_OUT}
        else:
            otyp = wmhdb_getCommonType(filter(None, inTypes.values()))
            return {c.getName(): otyp for c in clst
                    if c.getDirection() == WDBS_XPNT_OUT}

    def __str__(self):
        return '{c}-Operator'.format(c=self.getCode())

    def __repr__(self):
        return 'WmhdbOperator {{code: {c}}}'.format(c=self.getCode())


#: Global HDS function operator registry
WMHDB_OPER = WhdbOperatorRegistry()
