# ----------------------------------------------------------------------------
# - wmhdb-protocols                                                          -
# - wild cruncher:general protocol definitions             core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_protocols
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher hierarchy model protocol definitions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import ABCMeta, abstractmethod

from wcr_core.db_api.whdb_protocols_api import WhdbHasDimensionP


WMHDB_LABEL_ALIGN_KEY = 'lalgn'


# noinspection PyAbstractClass
class WmhdbHasConnectorsP(WhdbHasDimensionP):
    """
    Protocol for entities that have connectors associated with them.
    """

    @abstractmethod
    def getInputConnectors(self):
        """
        Get the instance's input connectors.

        :return: Input connectors
        :rtype: Iterable[WmhdbConnector]
        """
        raise NotImplementedError()

    @abstractmethod
    def getOutputConnectors(self):
        """
        Get the instance's output connectors.

        :return: Output connectors
        :rtype: Iterable[WmhdbConnector]
        """
        raise NotImplementedError()

    @abstractmethod
    def getRelatedConnectors(self, conn):
        """
        Get the connectors related to the given one.

        A connector B is related to connector A if B affects A or is affected
        by A.

        Usually, a connector's related connectors are of the opposite type

        :param conn: Connector to get related connectors for
        :type conn: WmhdbConnector
        :return: List of connectors affecting/affected-by the given connector
        :rtype: Iterable[WmhdbConnector]
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectorByName(self, name):
        """
        Get the operator connector by name

        TODO: Move to utility

        :param name: the connector name to be found
        :type name: str
        :return: the connector if found otherwise None
        :rtype: WmhdbConnector
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectorByPortRef(self, ref):
        """
        Get a connector by port reference

        TODO: Move to utility

        :param ref: the connector port reference
        :type ref: str
        :return: The connector with the given port reference, if any
        :rtype: WmhdbConnector
        """
        raise NotImplementedError()

    @abstractmethod
    def isConnectorPolarityInvertible(self):
        """
        Whether the polarity of the connectors of this entity can be inverted.

        :return: True iff the polarity of the connectors can be inverted
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectorPolarityNames(self):
        """
        Get the names of the connector polarities for this entity.

        :return: Names of the connector polarities
        :rtype: (str, str)
        """
        raise NotImplementedError()


WMHDB_SGROUP_KEY = 'sgrp'


class WmhdbHasSchedulingGroupP(metaclass=ABCMeta):
    """
    Protocol for entities that belong to a scheduling group.
    """

    DEFAULT_SCHED_GROUP = 'default'

    NONE_SCHED_GROUP = 'none'

    @abstractmethod
    def getSchedulingGroup(self):
        """
        Get the entity's scheduling group.

        :return: The entity's scheduling group
        :rtype: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WmhdbHasSchedulingGroupMP(WmhdbHasSchedulingGroupP):
    """
    Protocol for entities that belong to a scheduling group.
    """

    @abstractmethod
    def setSchedulingGroup(self, schedGroup):
        """
        Set the entity's scheduling group.

        :param schedGroup: The entity's new scheduling group
        :type schedGroup: str
        """
        raise NotImplementedError()
