# ----------------------------------------------------------------------------
# - wmhdb-rti-type                                                           -
# - wild cruncher:wsh real-time type                       core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_rti_type
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher real-time type object

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

import re

from wcr_core.dbstore.wdbs_meta_type import WdbsAtomicMetaTypeBase


# ----------------------------------------------------------------------------
# - rti type class                                                           -
# ----------------------------------------------------------------------------

class _WmhdbRtiTypeMetaType(WdbsAtomicMetaTypeBase):
    def __init__(self, default=None):
        super().__init__(default)

    def getName(self):
        return 'rtitype'

    def toString(self, metaV):
        return repr(metaV) if metaV else None

    def fromString(self, inStr):
        if inStr:
            return WmhdbRtiType.parseType(inStr)

        return self.defaultValue


WMHDB_RTITYPE_META_TYPE = _WmhdbRtiTypeMetaType()

WMHDB_S08_MIN = - 2 ** 7
WMHDB_S08_MAX = 2 ** 7 - 1
WMHDB_U08_MAX = 2 ** 8 - 1
WMHDB_S16_MIN = - 2 ** 15
WMHDB_S16_MAX = 2 ** 15 - 1
WMHDB_U16_MAX = 2 ** 16 - 1
WMHDB_S32_MIN = - 2 ** 31
WMHDB_S32_MAX = 2 ** 31 - 1
WMHDB_U32_MAX = 2 ** 32 - 1
WMHDB_S64_MIN = - 2 ** 63
WMHDB_S64_MAX = 2 ** 63 - 1
WMHDB_U64_MAX = 2 ** 64 - 1

WMHDB_RTI_DIM_NAMES = {1: 'SCLR', 2: 'MTR1', 3: 'MTR2'}

WMHDB_RTI_DIM_BY_NAME = {v: k for k, v in WMHDB_RTI_DIM_NAMES.items()}


class WmhdbRtiType:
    """
    The WmhdbRti class is the silkan hierarchy type definition for
    the hierarchy which is used mostly with connectors and wires. A type
    is defined by nature, like double, integer or anything else and its
    dimension which can be either scalar, vector or matrix. By default, a
    type is a double, scalar.
    """

    _TYPE_PARSING_PATTERN = re.compile(
        r"(\w+)(?:\[(\d+)(?:\s*[:,]\s*(\d+))?\])?")

    _SCALAR_TYPE_PATTERN_STR = r"^\s*({p})\s*$"

    #: Pattern to detect/extract vectors
    VECTOR_PATTERN = re.compile(r"(?:\(([^\)]*)\)|\[([^\]]*)\])")

    VECTOR_SEP_PATTERN = re.compile(r"[ ,;]+")

    #: Pattern to detect/extract strings
    STR_PATTERN = re.compile(r"(['\"])([^\1]*?)\1"), 2

    #: Pattern to detect/extract integers
    INT_PATTERN = re.compile(
        _SCALAR_TYPE_PATTERN_STR.format(p=r"-?\d+")), 1

    #: Pattern to detect/extract reals
    REAL_PATTERN = re.compile(
        _SCALAR_TYPE_PATTERN_STR.format(
            p=r"-?\d+(?:\.\d+|(?:\.\d+)?[Ee]-?\d+)")), 1

    #: Pattern to detect/extract booleans
    BOOL_PATTERN = re.compile(r"(t(?:rue)?|f(?:alse)?)$", re.IGNORECASE), 1

    #: Pattern to detect true values
    TRUE_PATTERN = re.compile(r"t(?:rue)?$", re.IGNORECASE)

    #: rti types
    NONE, F64, F32, U08, S08, U16, S16, U32, S32, U64, S64, BLN, STR, NOS = \
        range(0, 14)

    __RTI_TYPE_LIST = [
        U08, S08, U16, S16, U32, S32, U64, S64, F64, F32, BLN, STR, NOS]

    __WMHDB_RTI_TYPE_NAMES = {F64: 'F64', F32: 'F32',
                              U08: 'U08', S08: 'S08',
                              U16: 'U16', S16: 'S16',
                              U32: 'U32', S32: 'S32',
                              U64: 'U64', S64: 'S64',
                              BLN: 'BLN', STR: 'STR',
                              NOS: 'NOS'}

    __WMHDB_RTI_TYPE_BY_NAME = dict([(v, k) for k, v
                                     in __WMHDB_RTI_TYPE_NAMES.items()])

    __WMHDB_RTI_TYPE_DESCRIPTIONS = {F64: 'double (floating point 64bit)',
                                     F32: 'float (floating point 32bit)',
                                     U08: 'unsigned char (8bit)',
                                     S08: 'signed char (8bit)',
                                     U16: 'unsigned short (16bit)',
                                     S16: 'signed short (16bit)',
                                     U32: 'unsigned int (32bit)',
                                     S32: 'signed int (32bit)',
                                     U64: 'unsigned long long (64bit)',
                                     S64: 'long long (64bit)',
                                     BLN: 'boolean',
                                     STR: 'char*',
                                     NOS: 'not supported'}

    #: type dimensionalities
    Scalar, Vector, Matrix = range(1, 4)

    _BOOL_CONV = lambda s: \
        True if WmhdbRtiType.TRUE_PATTERN.match(s) else False

    _TYPE_CONVERT_FUNCTIONS = {
        F32: float, F64: float,
        S08: int, U08: int,
        S16: int, U16: int,
        S32: int, U32: int,
        S64: int, U64: int,
        BLN: _BOOL_CONV,
        STR: str, NOS: str}

    _TYPE_COMPATS = {
        F64: [],
        F32: [F64],
        U64: [F64, F32],
        S64: [F64, F32],
        U32: [F64, F32, S64, U64],
        S32: [F64, F32, S64],
        U16: [F64, F32, S64, U64, S32, U32],
        S16: [F64, F32, S64, S32],
        U08: [F64, F32, S64, U64, S32, U32, S16, U16],
        S08: [F64, F32, S64, S32, S16],
        BLN: [],
        STR: [],
        NOS: [],
    }

    # (pattern, extract-func, [(test-func, types, possible conversions)])
    _TYPE_PATTERNS = [
        (REAL_PATTERN, float, [(lambda x: True, {F64, F32}, {NOS})]),
        (INT_PATTERN, int,
         [(lambda x: WMHDB_S08_MIN <= x < 0,
           {S08, S16, S32, S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_S08_MAX >= x >= 0,
           {U08, S08, U16, S16, U32, S32, U64, S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_U08_MAX >= x >= 0,
           {U08, U16, S16, U32, S32, U64, S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_S16_MIN <= x < 0,
           {S16, S32, S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_S16_MAX >= x >= 0,
           {U16, S16, U32, S32, U64, S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_U16_MAX >= x >= 0,
           {U16, U32, S32, U64, S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_S32_MIN <= x < 0,
           {S32, S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_S32_MAX >= x >= 0,
           {U32, S32, U64, S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_U32_MAX >= x >= 0,
           {U32, U64, S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_S64_MIN <= x < 0,
           {S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_S64_MAX >= x >= 0,
           {U64, S64}, {F64, F32, NOS}),
          (lambda x: WMHDB_U64_MAX >= x >= 0,
           {U64}, {F64, F32, NOS})]),
        (BOOL_PATTERN, _BOOL_CONV, [(lambda x: True, {BLN}, {NOS})]),
        (STR_PATTERN, None, [(lambda x: True, {STR}, {NOS})]),
    ]

    def __init__(self, rtyp=None, sdim=None, xdim=None, ydim=None):
        """
        Create a rti type by type and dimension
        :param rtyp: the rti type
        :param sdim: the dimension
        :param xdim: the vector, matrix x dimension
        :param ydim: the y dimension
        """
        self.__type = rtyp or self.F64
        self.__dimensionality = sdim or self.getTypeDimension(xdim, ydim)
        self.__xDim = xdim if self.__dimensionality != self.Scalar else None
        self.__yDim = ydim if self.__dimensionality == self.Matrix else None

    @staticmethod
    def getTypeDimension(xdim, ydim):
        if not (xdim or ydim):
            return WmhdbRtiType.Scalar
        elif xdim and ydim:
            return WmhdbRtiType.Matrix
        else:
            return WmhdbRtiType.Vector

    def getType(self):
        """
        Get the rti type
        """
        return self.__type

    def getName(self):
        return WmhdbRtiType.getTypeName(self.__type)

    def getDimension(self):
        """
        Get the rti dimension
        """
        return self.__dimensionality

    def getXdim(self):
        """
        Get the x rti dimension
        """
        return self.__xDim

    def getYdim(self):
        """
        Get the y rti dimension
        """
        return self.__yDim

    def getDescription(self):
        """
        Get a textual description of the  type.

        :return: Textual description of the type
        :rtype: str
        """
        desc = self.__WMHDB_RTI_TYPE_NAMES.get(
            self.getType(), 'Unknown ({})'.format(self.getType()))
        desc += ' '
        if self.getDimension() == WmhdbRtiType.Scalar:
            desc += 'Scalar'
        elif self.getDimension() == WmhdbRtiType.Vector:
            desc += '{d}-Vector'.format(d=max(self.getYdim() or 0,
                                              self.getXdim() or 0))
        elif self.getDimension() == WmhdbRtiType.Matrix:
            desc += '{d1}x{d2} Matrix'.format(d1=self.getYdim(),
                                              d2=self.getXdim())
        return desc

    def getRtiTypeStr(self):
        """
        Get the C rti type description for this type.
        """
        return WmhdbRtiType.getRtiTypeForTypeCode(self.__type)

    def getRtiEnumType(self):
        """
        Get the C rti enumeration description for this type.
        """
        return WmhdbRtiType.getRtiEnumForTypeCode(self.__type)

    def getRtiEnumDim(self):
        """
        Get the C rti dim description for this type.
        """
        return WmhdbRtiType.getRtiEnumForDimCode(self.__dimensionality)

    def __key(self):
        return self.__type, self.__dimensionality, self.__xDim, self.__yDim

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, WmhdbRtiType) and \
            self.__key() == other.__key()

    @staticmethod
    def getRtiTypeForStrValue(value):
        """
        Determine the type for a string-encoded value, if possible.

        :param value: String-value to determine the type for
        :type value: str
        :return: the type for the given value
        :rtype: WmhdbRtiType
        """
        if not value:
            return None

        _, types = WmhdbRtiType.parse(value)
        return types[0]

    @staticmethod
    def getRtiEnumForTypeCode(code):
        typeName = WmhdbRtiType.__WMHDB_RTI_TYPE_NAMES.get(code)
        return ('WCR_RTI_' + typeName) if typeName else 'void'

    @staticmethod
    def getRtiTypeForTypeCode(code):
        typeName = WmhdbRtiType.__WMHDB_RTI_TYPE_NAMES.get(code)
        return ('t_wcr_rti_' + typeName) if typeName else 'void'

    @staticmethod
    def getRtiTypeList():
        """
        Get a list with all RTI type codes.

        :return: List with all RTI type codes
        :rtype: [int]
        """
        return WmhdbRtiType.__RTI_TYPE_LIST

    @staticmethod
    def getTypeName(code):
        """
        Get the short type name for a type code.

        :param code: Type code
        :type code: int
        :return: Name of the type with the given code
        :rtype: str
        """
        return WmhdbRtiType.__WMHDB_RTI_TYPE_NAMES.get(code)

    @staticmethod
    def getTypeForName(name):
        """
        Get the type code for a short type name.

        :param name: Name of the type with the given code
        :type name: str
        :return: Type code
        :rtype: int
        """
        return WmhdbRtiType.__WMHDB_RTI_TYPE_BY_NAME.get(name)

    @staticmethod
    def getTypeDescription(code):
        """
        Get the type description for a type code.

        :param code: Type code
        :type code: int
        :return: Description for the type with the given code
        :rtype: str
        """
        return WmhdbRtiType.__WMHDB_RTI_TYPE_DESCRIPTIONS.get(code)

    @staticmethod
    def getRtiEnumForDimCode(code):
        DimName = WMHDB_RTI_DIM_NAMES.get(code)
        return ('WCR_RTI_' + DimName) if DimName else 'void'

    @staticmethod
    def parseType(inStr):
        """
        Parse a type-description string, e.g. "F32" or "U08[4]" into a
        type object.

        :param inStr: Type description string
        :type inStr: str
        :return: Parsed type, or `None`
        :rtype: WmhdbRtiType
        """
        m = WmhdbRtiType._TYPE_PARSING_PATTERN.match(inStr)
        if not m:
            return None
        if m.group(3):
            return WmhdbRtiType(
                WmhdbRtiType.getTypeForName(m.group(1)), WmhdbRtiType.Matrix,
                int(m.group(2)), int(m.group(3)))
        elif m.group(2):
            return WmhdbRtiType(
                WmhdbRtiType.getTypeForName(m.group(1)), WmhdbRtiType.Vector,
                int(m.group(2)))
        else:
            return WmhdbRtiType(WmhdbRtiType.getTypeForName(m.group(1)))

    def __str__(self):
        return self.getDescription()

    def __repr__(self):
        tstr = WmhdbRtiType.__WMHDB_RTI_TYPE_NAMES[self.getType()]
        if self.getDimension() == WmhdbRtiType.Vector:
            return '{t}[{d}]'.format(t=tstr, d=self.getXdim())
        elif self.getDimension() == WmhdbRtiType.Matrix:
            return '{t}[{dx},{dy}]'.format(
                t=tstr, dx=self.getXdim(), dy=self.getYdim())
        else:
            return tstr

    @staticmethod
    def _parse_element(instr):
        for (pattern, grpIdx), convF, cases in WmhdbRtiType._TYPE_PATTERNS:
            m = pattern.match(instr)
            if m:
                value = m.group(grpIdx)
                if convF is not None:
                    value = convF(value)
                for testF, types, convs in cases:
                    if testF(value):
                        return value, types, convs
        return instr, {WmhdbRtiType.STR, WmhdbRtiType.NOS}, []

    @staticmethod
    def _sort_types(types):
        if not types:
            return []
        return [t for t in WmhdbRtiType.__RTI_TYPE_LIST if t in types]

    @staticmethod
    def _parse_scalar(instr, possibleTypes=None):
        value, types, convs = WmhdbRtiType._parse_element(instr)
        available = possibleTypes or WmhdbRtiType.__RTI_TYPE_LIST
        for ts, cf in [(types, False), (convs, True)]:
            ts = WmhdbRtiType._sort_types(ts.intersection(available))
            if ts:
                if ts[0] in types:
                    return value, ts
                else:
                    cf = WmhdbRtiType._TYPE_CONVERT_FUNCTIONS[ts[0]]
                    return cf(value), ts
        raise ValueError('Invalid input for given types')

    @staticmethod
    def _parse_vector(instr, possibleTypes=None):
        values = []
        types = set(possibleTypes or WmhdbRtiType.__RTI_TYPE_LIST)
        convs = set(possibleTypes or WmhdbRtiType.__RTI_TYPE_LIST)

        for element in WmhdbRtiType.VECTOR_SEP_PATTERN.split(instr):
            element = element.strip()
            if element:
                value, eTypes, eConvs = WmhdbRtiType._parse_element(element)
                types = types.intersection(eTypes)
                convs = convs.intersection(eTypes.union(eConvs))
                #aucun type commun converssible.(impossible avec le 13...?)
                if not convs:
                    return None, None
                values.append(value)
        for ts, cf in [(types, False), (convs, True)]:
            if ts:
                ts = WmhdbRtiType._sort_types(ts)
                if ts[0] in types:
                    return values, ts
                else:
                    cf = WmhdbRtiType._TYPE_CONVERT_FUNCTIONS[ts[0]]
                    return [cf(v) for v in values], ts
        raise ValueError('Invalid input for given types')

    @staticmethod
    def parse(instr, possibleTypes=None):
        """
        Parse an input string, returning the contained values and corresponding
        types.

        :param instr: String-value to parse
        :type instr: str
        :param possibleTypes: Acceptable return types
        :type possibleTypes: [int]
        :return: The parsed value and raw types
        :rtype: (object, [WmhdbRtiType])
        """
        vgrps = WmhdbRtiType.VECTOR_PATTERN.match(instr)
        if vgrps:
            contents = next(filter(None, vgrps.groups()), None)
            if not contents:
                return "", [WmhdbRtiType(WmhdbRtiType.STR)]

            vs, ts = WmhdbRtiType._parse_vector(contents, possibleTypes)

            return vs, [WmhdbRtiType(t, WmhdbRtiType.Vector, len(vs))
                        for t in ts]
        else:
            v, ts = WmhdbRtiType._parse_scalar(instr, possibleTypes)
            return v, [WmhdbRtiType(t) for t in ts]


def wmhdb_isSameDimension(*types):
    """
    Whether the given types are of the same dimension.

    :param types: Types to test for identical dimension
    :type types: [WmhdbRtiType]
    :return: True if all given types are of the same dimension
    :rtype: bool
    """
    if len(types) <= 1:
        return True

    dim = types[0].getDimension()
    xdim = types[0].getXdim()
    ydim = types[0].getYdim()

    for _typ in types[1:]:
        if _typ.getDimension() != dim or \
                _typ.getXdim() != xdim or \
                _typ.getYdim() != ydim:
            return False
    return True


# noinspection PyProtectedMember
def wmhdb_getCommonType(types):
    """
    Calculate the common ancestral type of the given types.

    :param types: Types to compute common type for
    :type types: [WmhdbRtiType]
    :return: Common type
    :rtype: WmhdbRtiType
    """
    ctyp = None

    for _typ in types:
        if not ctyp:
            ctyp = _typ
        elif _typ != ctyp:
            if not wmhdb_isSameDimension(_typ, ctyp):
                raise ValueError(
                    'Unmergeable types (dimension): '
                    '{} ({} - {}/{}) and {} ({} - {}/{})'.format(
                        ctyp, ctyp.getDimension(),
                        ctyp.getXdim(), ctyp.getYdim(),
                        _typ, _typ.getDimension(),
                        _typ.getXdim(), _typ.getYdim()))

            if _typ.getType() in WmhdbRtiType._TYPE_COMPATS[ctyp.getType()]:
                ctyp = _typ
            elif ctyp.getType() in WmhdbRtiType._TYPE_COMPATS[_typ.getType()]:
                pass
            else:
                cpts1 = WmhdbRtiType._TYPE_COMPATS[ctyp.getType()]
                cpts2 = WmhdbRtiType._TYPE_COMPATS[_typ.getType()]
                ctps = set(cpts1).union(set(cpts2))
                if not ctps:
                    raise ValueError('Unmergeable types: {} and {}'.format(
                        ctyp, _typ))
                for cmpt in reversed(cpts1):
                    if cmpt in ctps:
                        ctyp = cmpt
                        break

    return ctyp


def wmhdb_canBeConvertedInto(styp, dtyp):
    """
    Whether a source type can be converted/cast into a given destination type.

    :param styp: Source type
    :type styp: WmhdbRtiType
    :param dtyp: Destination type
    :type dtyp: WmhdbRtiType
    :return: True iff the source type can be converted into the destination
             type
    :rtype: bool
    """
    if not wmhdb_isSameDimension(styp, dtyp):
        return False
    if styp.getType() == dtyp.getType():
        return True
    # if dtyp.getType() == WmhdbRtiType.STR:
    #    return True
    if styp.getType() in [WmhdbRtiType.STR, WmhdbRtiType.BLN]:
        return False
    if dtyp.getType() in [WmhdbRtiType.STR, WmhdbRtiType.BLN]:
        return False
    return True
