# ----------------------------------------------------------------------------
# - wmhdb-standard-operators                                                 -
# - wild cruncher:standard operators                       core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_standard_operators
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher standard operators

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.common.wcmn_parameter_description import \
    WCMN_PARAM_FORMAT_STRING_FLAG
from wcr_core.db.whdb_operator import WHDB_INPUT_OPERATOR_FLAG, \
    WHDB_OUTPUT_OPERATOR_FLAG, WHDB_PORT_OPERATOR_FLAG
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_OUT, \
    WDBS_XPNT_IN
from silkan_common.scmn_utils import partition
from wcr_core.modules.model.db.wmhdb_operator import WMHDB_OPER, WmhdbOperator
from wcr_core.db_api.whdb_protocols_api import WHDB_HALIGN_LEFT, \
    WHDB_HALIGN_RIGHT
from wcr_core.modules.model.db.wmhdb_rti_type import WmhdbRtiType, \
    wmhdb_getCommonType
from wcr_core.modules.model.db.wmhdb_connection_point import WmhdbConnector
from wcr_core.modules.model.db.wmhdb_connection_point_manager import \
    wmhdb_createFixedXPointManager, WmhdbVarInConnectionPointManager, \
    WmhdbVarOutConnectionPointManager, wmhdb_createXPointDesc


WMHDB_BUS_FLAG = 'bus'

WMHDB_PORT_NAME_PARAMETER = 'pid'

WMHDB_BUS_NAME_PARAMETER = 'name'

WMHDB_BUS_TYPE_PARAMETER = 'vartype'


#: Input port operator
WMHDB_INPUT_PORT = WmhdbOperator(
    code='wcr.std.input',
    xmngr=wmhdb_createFixedXPointManager(ocns=['o']),
    name='Input',
    info='Input port',
    flags={WHDB_PORT_OPERATOR_FLAG, WHDB_INPUT_OPERATOR_FLAG, 'nofog'},
    parameters=[(WMHDB_PORT_NAME_PARAMETER, 'Port name', 'The port name',
                 'IN-{fnid}', None, WCMN_PARAM_FORMAT_STRING_FLAG)],
    shape=[(-25, -10), (15, -10), (22, -7), (25, -2), (25, 2), (22, 7),
           (15, 10), (-25, 10), (-25, -10)],
    symbol='in',
    labelAlign=WHDB_HALIGN_LEFT,
    outTypeF=lambda *_: {})

#: Output port operator
WMHDB_OUTPUT_PORT = WmhdbOperator(
    code='wcr.std.output',
    xmngr=wmhdb_createFixedXPointManager(icns=['i']),
    name='Output',
    info='Output port',
    flags={WHDB_PORT_OPERATOR_FLAG, WHDB_OUTPUT_OPERATOR_FLAG, 'nofog'},
    parameters=[(WMHDB_PORT_NAME_PARAMETER, 'Port name', 'The port name',
                 'OUT-{fnid}', None, WCMN_PARAM_FORMAT_STRING_FLAG)],
    shape=[(-25, -10), (15, -10), (22, -7), (25, -2), (25, 2), (22, 7),
           (15, 10), (-25, 10), (-25, -10)],
    symbol='out',
    labelAlign=WHDB_HALIGN_RIGHT)


# noinspection PyUnusedLocal
def _wmhdb_inBusOutType(inTypes, clst, paramF, cpars):
    """
    Output-types function for InBus-operator.

    :param inTypes: Input-types
    :type inTypes: dict[str, WmhdbRtiType]
    :param clst: Connection points
    :type clst: list[WmhdbConnector]
    :param paramF: Parameter function
    :type paramF: str -> object
    :param cpars: cold variable parser
    :type cpars: WcldColdVarsParser
    :return: Output types
    :rtype: dict[str, WmhdbRtiType]
    """
    vtyp = paramF(WMHDB_BUS_TYPE_PARAMETER)
    if not vtyp:
        return {}
    otyp = WmhdbRtiType.parseType(vtyp)
    return {c.getName(): otyp for c in clst}


#: Input bus operator
WMHDB_INPUT_BUS = WmhdbOperator(
    code='wcr.std.ibus',
    xmngr=wmhdb_createFixedXPointManager(
        ocns=['o'], opos=[(1.0, -14.0 / 15.0, WmhdbConnector.East)]),
    name='InBus',
    info='Input bus',
    flags={WMHDB_BUS_FLAG, WHDB_INPUT_OPERATOR_FLAG, 'nofog'},
    parameters=[(WMHDB_BUS_NAME_PARAMETER, 'Variable name',
                 'The variable name', None, None, None),
                (WMHDB_BUS_TYPE_PARAMETER, 'Variable type',
                 'The variable type',
                 WmhdbRtiType.getTypeName(WmhdbRtiType.F64), None, None)],
    shape=[(0, 10), (20, 10), (20, 15), (-20, 15), (-20, 10), (-2, 10),
           (-2, -15), (20, -15), (20, -13), (0, -13), (0, 10)],
    symbol='ibus',
    labelAlign=WHDB_HALIGN_LEFT,
    outTypeF=_wmhdb_inBusOutType)

#: Output bus operator
WMHDB_OUTPUT_BUS = WmhdbOperator(
    code='wcr.std.obus',
    xmngr=wmhdb_createFixedXPointManager(
        icns=['i'], ipos=[(-1.0, -14.0 / 15.0, WmhdbConnector.West)]),
    name='OutBus',
    info='Output bus',
    flags={WMHDB_BUS_FLAG, WHDB_OUTPUT_OPERATOR_FLAG, 'nofog'},
    parameters=[(WMHDB_BUS_NAME_PARAMETER, 'Variable name',
                 'The variable name', None, None, None)],
    shape=[(0, 10), (-20, 10), (-20, 15), (20, 15), (20, 10), (2, 10),
           (2, -15), (-20, -15), (-20, -13), (0, -13), (0, 10)],
    symbol='obus',
    labelAlign=WHDB_HALIGN_RIGHT)


def _wmhdb_muxInOutputType(inTypes, clst, *_):
    """
    Output-types function for MuxIn-operator.

    :param inTypes: Input-types
    :type inTypes: dict[str, WmhdbRtiType]
    :param clst: Connection points
    :type clst: list[WmhdbConnector]
    :return: Output types
    :rtype: dict[str, WmhdbRtiType]
    """
    ityp = wmhdb_getCommonType(filter(None, inTypes.values()))
    if not ityp:
        return None

    inXpnts, outXpnts = partition(
        lambda c: c.getDirection() == WDBS_XPNT_IN, clst)
    if not inXpnts:
        return None
    if len(outXpnts) != 1:
        raise ValueError(
            'Incorrect number of output connectors for MuxIn: {}'.format(
                len(inTypes)))

    if ityp.getDimension() == WmhdbRtiType.Scalar:
        otyp = WmhdbRtiType(ityp.getType(), sdim=WmhdbRtiType.Vector,
                            xdim=len(inTypes))
        return {outXpnts[0].getName(): otyp}
    elif ityp.getDimension() == WmhdbRtiType.Vector:
        otyp = WmhdbRtiType(ityp.getType(), sdim=WmhdbRtiType.Matrix,
                            xdim=ityp.getXdim(), ydim=len(inTypes))
        return {outXpnts[0].getName(): otyp}
    else:
        raise ValueError('MuxOut-Input must be scalar or vector')

#: Output bus operator
WMHDB_MUX_IN = WmhdbOperator(
    code='wcr.std.muxin',
    xmngr=WmhdbVarInConnectionPointManager(
        fxds=[wmhdb_createXPointDesc(cdir=WmhdbConnector.Output, name='o')],
        prfx='i', minc=1, maxc=None),
    name='Mux-in',
    info='Mux-in operator',
    symbol='mux-in',
    outTypeF=_wmhdb_muxInOutputType)


def _wmhdb_muxOutOutputType(inTypes, clst, *_):
    """
    Output-types function for MuxOut-operator.

    :param inTypes: Input-types
    :type inTypes: dict[str, WmhdbRtiType]
    :param clst: Connection points
    :type clst: list[WmhdbConnector]
    :return: Output types
    :rtype: dict[str, WmhdbRtiType]
    """
    if len(inTypes) != 1:
        raise ValueError(
            'Incorrect number of input types for MuxOut: {}'.format(
                len(inTypes)))

    outXpnts = [c for c in clst if c.getDirection() == WDBS_XPNT_OUT]
    if not outXpnts:
        return None

    ityp = next(iter(inTypes.values()))
    if ityp.getDimension() == WmhdbRtiType.Vector:
        otyp = WmhdbRtiType(ityp.getType(), sdim=WmhdbRtiType.Scalar)
        return {c.getName(): otyp for c in outXpnts}
    elif ityp.getDimension() == WmhdbRtiType.Matrix:
        otyp = WmhdbRtiType(ityp.getType(), sdim=WmhdbRtiType.Vector,
                            xdim=ityp.getXdim())
        return {c.getName(): otyp for c in outXpnts}
    else:
        raise ValueError(
            'MuxOut-Input must be vector or matrix, was: {}'.format(ityp))


#: Output bus operator
WMHDB_MUX_OUT = WmhdbOperator(
    code='wcr.std.muxout',
    xmngr=WmhdbVarOutConnectionPointManager(
        fxds=[wmhdb_createXPointDesc(cdir=WmhdbConnector.Input, name='i')],
        prfx='o', minc=1, maxc=None),
    name='Mux-out',
    info='Mux-out operator',
    symbol='mux-out',
    outTypeF=_wmhdb_muxOutOutputType)

WMHDB_DEFAULT_OPS = [WMHDB_INPUT_PORT, WMHDB_OUTPUT_PORT,
                     WMHDB_INPUT_BUS, WMHDB_OUTPUT_BUS,
                     WMHDB_MUX_IN, WMHDB_MUX_OUT]

WMHDB_OPER.addOperators(WMHDB_DEFAULT_OPS)
