# ----------------------------------------------------------------------------
# - wmhdb-top-block                                                          -
# - wild cruncher:wsh top hierarchy object                 core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_top_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher top hierarchy object

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db.whdb_protocols import WHDB_POSITION_KEY
from wcr_core.dbstore.mutable.wdbm_block import WdbmBlock
from wcr_core.dbstore.mutable.wdbm_context import WdbmContext
from wcr_core.dbstore.wdbs_cloning import wdbs_cloneEntities
from wcr_core.dbstore.wdbs_meta_description import *
from wcr_core.dbstore.wdbs_meta_type import *
from wcr_core.modules.model.db.wmhdb_scheduling_group import WmhdbGroupBlock, \
    WMHDB_SGROUP_BLOCK_CODE
from wcr_core.modules.model.db.wmhdb_block import WmhdbBlock, \
    wmhdb_createOperatorInstance
from wcr_core.modules.model.db.wmhdb_block_flattener import wmhdb_flatten
from wcr_core.modules.model.db.wmhdb_protocols import WMHDB_SGROUP_KEY
from wcr_core.modules.model.db.wmhdb_standard_operators import \
    WMHDB_INPUT_BUS, WMHDB_OUTPUT_BUS, WMHDB_BUS_FLAG, WMHDB_BUS_NAME_PARAMETER
from wplg_cold.wcld_builder import *
from wplg_cold.wcld_execution import *
import os


WMHDB_META_CALLBACKS_KEY = 'callbacks'


#: Available Callback-Functions (Key, Label)
WMHDB_CALLBACK_FUNCTIONS = (
    ('PreLoadFcn', 'Pre-Load'),
    ('PostLoadFcn', 'Post-Load'),
    ('InitFcn', 'Init'),
    ('StartFcn', 'Start'),
    ('PauseFcn', 'Pause'),
    ('ContinueFcn', 'Continue'),
    ('StopFcn', 'Stop'),
    ('PreSaveFcn', 'Pre-Save'),
    ('PostSaveFcn', 'Post-Save'),
    ('CloseFcn', 'Close'))


class WmhdbTopHierarchy(WmhdbBlock):
    """
    The WmhdbTopHierarchy is the top hierarchy model class. A top hierarchy
    is at the same time a model and an instance. The class provides several
    methods for creating the hierarchy.
    """
    META = WdbsMetaDescription(
        WmhdbBlock.META,
        WdbsMetaField(
            WMHDB_META_CALLBACKS_KEY,
            WdbsDictionaryMetaType(WDBS_STRING_META, WDBS_STRING_META)))

    _PORT_FLAG = WMHDB_BUS_FLAG

    _PORT_NAME_PARAM = WMHDB_BUS_NAME_PARAMETER

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        super().__init__(block, wreg)

    def flatten(self):
        """
        Flatten the top hierarchy
        """
        return wmhdb_flatten(self)

    # noinspection PyUnresolvedReferences
    def getSchedulingGroupSet(self):
        """
        Get the set of scheduling group names
        """
        # create a set to hold unique names
        result = set()
        # loop into the instances
        for i in self.getInstances():
            name = i.getSchedulingGroup()
            if name is not None: result.add(name)
        # here it is
        return result

    def createSchedulingGroup(self, name):
        """
        Create a scheduling group by name.

        :param name: the scheduling group name
        :type name: str
        :return: Newly created scheduling group block
        :rtype: WmhdbGroupBlock
        """
        # noinspection PyUnusedLocal
        def inputF(block, iobj, icon, oobj, ocon, wire):
            """
            :type block: WdbsBlockP
            :type iobj: WdbsInstanceP
            :type icon: WdbsConnectionPointP
            :type oobj: WdbsInstanceP
            :type ocon: WdbsConnectionPointP
            :type wire: WdbsConnectionP
            """
            port = wmhdb_createOperatorInstance(
                block, WMHDB_INPUT_BUS, info='output bus')
            port.setName('O' + str(id(port)))
            port.setMeta(iobj.getMeta(WMHDB_SGROUP_KEY), WMHDB_SGROUP_KEY)
            port.getOrCreateParameter('name').setValue(
                wire.getName() or 'C' + str(wire.getUUID()))
            port.setMeta(oobj.getMeta(WHDB_POSITION_KEY),
                         WHDB_POSITION_KEY)
            # noinspection PyTypeChecker
            port.getOrCreateParameter('vartype').setValue('')
            return port

        # noinspection PyUnusedLocal
        def outputF(block, iobj, icon, oobj, ocon, wire):
            """
            :type block: WdbsBlockP
            :type iobj: WdbsInstanceP
            :type icon: WdbsConnectionPointP
            :type oobj: WdbsInstanceP
            :type ocon: WdbsConnectionPointP
            :type wire: WdbsConnectionP
            """
            port = wmhdb_createOperatorInstance(
                block, WMHDB_OUTPUT_BUS, info='input bus')
            port.setName('I' + str(id(port)))
            port.setMeta(iobj.getMeta(WMHDB_SGROUP_KEY), WMHDB_SGROUP_KEY)
            port.getOrCreateParameter('name').setValue(
                wire.getName() or 'C' + str(wire.getUUID()))
            port.setMeta(oobj.getMeta(WHDB_POSITION_KEY),
                         WHDB_POSITION_KEY)
            return port

        sgrp = WmhdbGroupBlock(
            name,
            WdbmBlock(
                WdbmContext(self.rawElement.context), WMHDB_SGROUP_BLOCK_CODE,
                name, meta=WmhdbGroupBlock.META),
            self.wrapperRegistry)

        entities = []
        # noinspection PyUnresolvedReferences
        entities.extend(self.wrapperRegistry.unwrap(i)
                        for i in self.getInstances()
                        if i.getSchedulingGroup() == name)

        wdbs_cloneEntities(
            entities, self.rawElement, sgrp.rawElement, inputF, outputF)

        # here we are
        return sgrp

    def getCallbackDict(self):
        """
        Get the callback functions as a dictionnary

        :return: Callback functions dict
        :rtype: {str : str}
        """
        return self.rawElement.getMeta(WMHDB_META_CALLBACKS_KEY)
        
    def getCallbackFunction(self, fkey):
        """
        Get the callback-function name for the given key, if any.

        :param fkey: Key of callback-function to retrieve
        :type fkey: str
        :return: Callback-function name, if found
        :rtype: str | None
        """
        return self.rawElement.getMeta(WMHDB_META_CALLBACKS_KEY, fkey)

    def setCallbackFunction(self, fkey, fname):
        """
        Set the callback-function name for the given key.

        :param fkey: Key of callback-function to set
        :type fkey: str
        :param fname: Callback-function name to set
        :type fname: str | None
        """
        self.rawElement.setMeta(fname, WMHDB_META_CALLBACKS_KEY, fkey)


    def CallbackFunction(self, fkey, pctx, prof, sgrp):
        """
        Creates the .m file for the callback function fkey and
        compiles it with cold.
        :param fkey: the callbackfunction we whant to execute
        :type fkey: str
        :param pctx: Project context context
        :type pctx: WcorProjectContextP
        :param prof: Profile name
        :type prof: str | None
        :param sgrp: the scheduling group name
        :type  sgrp: str
        """
        callback = self.getCallbackFunction(fkey)
        if callback == '' or callback is None:
            return False
        path = os.path.join(pctx.getProjectWorkspace().getProjectPath(),
                            'script')
        if not os.path.exists(path):
            os.mkdir(path)
        rfkey=fkey
        if callback.find("%matlabfile") is not -1:
            path = pctx.getConfigurationValue('WCR_SCRIPT_PATH')
            rfkey = callback[:len(callback)-callback.find("%matlabfile")]
            rfkey = rfkey[rfkey.find(":")+1:]
            if rfkey.find("\n") is not -1:
                rfkey = rfkey[:len(rfkey)-rfkey.find("\n")]
        else:
            mfile = open(os.path.join(path, fkey+".m"), "w")
            mfile.write(callback)
            mfile.close()
        cold = WcldColdExecute(os.path.join(pctx.getConfigurationValue
                                            ('WCR_CML_HOME'),
                                            "bin", "cold"))
        mfile = os.path.join(path, rfkey + ".m")
        cfile = os.path.join(pctx.getProjectWorkspace().
                             getDirectoryPath(prof=prof, sgrp=sgrp),
                             rfkey+".cpp")
        refcldpath = list()
        refcldpath.append(pctx.getConfigurationValue('WCR_SCRIPT_PATH'))
        command = cold.getFunctionCommand(mfile, cfile, None, func=rfkey,
                                           rpth=refcldpath)
        if cold.execute(command):
            print(fkey + " convertion succes.")
            return True
        else:
            print("Fail on converting of " + fkey)
            return False
