# ----------------------------------------------------------------------------
# - wmhdb-type-propagation                                                   -
# - wild cruncher - model type propagation                                   -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_type_propagation
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher model type propagation

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import defaultdict

from wcr_core.db.whdb_operator import WHDB_INPUT_OPERATOR_FLAG, \
    WHDB_OUTPUT_OPERATOR_FLAG
from wcr_core.modules.model.db.wmhdb_instance_base import WmhdbInstanceBaseP
from wcr_core.modules.model.db.wmhdb_function import WmhdbFunction
from wcr_core.modules.model.db.wmhdb_instance import WmhdbModelInstance, \
    WmhdbBlockInstance
from wcr_core.modules.model.db.wmhdb_operator import WMHDB_DELAY_FLAG
from wcr_core.modules.model.db.wmhdb_rti_type import wmhdb_getCommonType, \
    wmhdb_canBeConvertedInto, WmhdbRtiType
from wcr_core.modules.model.db.wmhdb_standard_operators import \
    WMHDB_INPUT_BUS, WMHDB_INPUT_PORT, WMHDB_PORT_NAME_PARAMETER, \
    WMHDB_OUTPUT_PORT


class WmhdbElementsException(Exception):
    def __init__(self, elements, message):
        super().__init__(message)

        self.__elements = elements

    @property
    def elements(self):
        return iter(self.__elements)


class WmhdbTypePropagator(object):
    def propagateTypes(self, tblk, cpars=None, cbuild=None):
        """
        Perform type propagation for the given top-level block.

        :param tblk: Top-level block to perform type propagation on
        :type tblk: WmhdbBlock
        :param cpars: cold variable parser
        :type cpars: WcldColdVarsParser
        """

        def clear(blck):
            """
            Recursively clear inferred types from block elements.
            """
            wires = blck.getConnections()
            """:type: Iterable[WmhdbWire]"""
            for wire in wires:
                wire.setInferredRtiType(None)
            for sblk in blck.getBlockInstances():
                clear(sblk.getBlock())
        self.cbuild = cbuild
        self.cpars = cpars
        clear(tblk)
        inputBlocks = self._prescanInputBlocks(tblk)

        try:
            self._propagateTypes(tblk, inputBlocks, tuple())
        except:
            clear(tblk)
            raise

    def _propagateTypes(self, blck, ibks, path, recurse=True):
        """
        Perform type propagation for the given block.

        :param blck: Block to perform type propagation on
        :type blck: WmhdbBlock
        :param ibks: Input blocks
        :type ibks: dict[tuple, list[WmhdbBlockInstance]]
        :param path: Instance path to current block
        :type path: tuple
        :param recurse: Whether to propagate types recursively into child
                        blocks
        :type recurse: bool
        """
        def handleElementsWithOutput(elms):
            while elms:
                item = elms.pop()
                self._setOutputTypes(
                    item, inputsCompleteSet, inputsIncompleteSet, cwrs,
                    path, ibks, recurse=recurse)

        cwrs = self._getXpntToWiresDict(blck)

        inputsCompleteSet = set(list(self._getInputsForBlock(blck)))
        inputsIncompleteSet = set()

        handleElementsWithOutput(inputsCompleteSet)
        if path in ibks:
            handleElementsWithOutput(set(ibks.get(path)))

        while inputsCompleteSet or inputsIncompleteSet:
            handleElementsWithOutput(inputsCompleteSet)

            if inputsIncompleteSet:
                incmps = [i for i in inputsIncompleteSet
                          if self._isMissingOutputTypes(i, cwrs)]
                for incmp in incmps:
                    if self._deduceMissingOutputTypes(
                            incmp, inputsCompleteSet, inputsIncompleteSet,
                            cwrs):
                        break
                else:
                    if recurse:
                        raise WmhdbElementsException(
                            [(e, path) for e in inputsIncompleteSet],
                            'Cannot solve all input types for ' +
                            ', '.join(e.getName() for e
                                      in inputsIncompleteSet))
                    else:
                        return

    def _prescanInputBlocks(self, block):
        """
        Scan the block and its children for block instances that may serve as
        input due to inner loops and delays.

        :param block: Block to scan
        :type block: WmhdbBlock
        :return: Dictionary of potential input instances by path
        :rtype: dict[tuple, list[WmhdbBlockInstance]]
        """

        # noinspection PyShadowingNames
        def collectChildBlocks(blck, path, clst):
            """
            :type blck: WmhdbBlock
            :type path: tuple
            :type clst: list[(WmhdbBlock, tuple)]
            """
            chldrn = ((i.getBlock(), path + (i,))
                      for i in blck.getBlockInstances())
            """:type: Iterable[(WmhdbBlock, tuple)]"""
            clst.extend(chldrn)
            for inst in blck.getBlockInstances():
                _blck = inst.getBlock()
                """:type: WmhdbBlock"""
                collectChildBlocks(_blck, path + (inst,), clst)

        childBlocks = [(block, tuple())]
        collectChildBlocks(block, tuple(), childBlocks)

        childBlocks.reverse()

        inputBlocks = defaultdict(list)
        """:type: defaultdict[tuple, list[WmhdbBlockInstance]]"""
        _inputBlockSet = set()

        for blck, path in childBlocks:
            self._propagateTypes(blck, inputBlocks, path, recurse=False)
            if path and \
                    any(self._isInput(i) or i in _inputBlockSet
                        for i in blck.getInstances()) and\
                    any(w.getInferredRtiType()
                        for o in blck.getOperatorInstances()
                        for w in blck.getConnectionsForObject(o)
                        if o.getOperator() is WMHDB_OUTPUT_PORT):
                inputBlocks[path[:-1]].append(path[-1])
                _inputBlockSet.add(path[-1])

        return inputBlocks

    def _isInput(self, inst):
        """
        Whether the given instance is an input.

        :param inst: Instance to test
        :type inst: WhdbInstanceP
        :return: True iff the given instance is an input
        :rtype: bool
        """
        if isinstance(inst, WmhdbFunction):
            oper = inst.getOperator()
            return oper in [WMHDB_INPUT_BUS, WMHDB_INPUT_PORT] or \
                   oper.hasFlag(WHDB_INPUT_OPERATOR_FLAG) or \
                   oper.hasFlag(WMHDB_DELAY_FLAG) or \
                   len(list(inst.getInputConnectors())) == 0
        elif isinstance(inst, WmhdbInstanceBaseP):
            return len(list(inst.getInputConnectors())) == 0
        else:
            return False

    def _getXpntToWiresDict(self, blck):
        """
        Produce the connectors-to-wires lookup dictionary.

        :param blck: Block to create lookup dictionary for
        :type blck: WmhdbBlock
        :return: Connectors-to-Wires mapping
        :rtype: dict[WmhdbConnector,
                     list[(WmhdbWire, WmhdbInstance, WmhdbConnector)]]
        """
        cwrs = defaultdict(list)
        for wire in blck.getConnections():
            cwrs[wire.getStartConnectionPoint()].append((
                wire, wire.getEndObject(), wire.getEndConnectionPoint()))
            cwrs[wire.getEndConnectionPoint()].append((
                wire, wire.getStartObject(), wire.getStartConnectionPoint()))
        return cwrs

    def _getInputsForBlock(self, blck):
        """
        Get the input instances for the given block.

        :param blck: Block to retrieve input instances for
        :type blck: WmhdbBlock
        :return: Input instances for the given block
        :rtype: Iterable[WhdbInstanceP]
        """
        return (inst for inst in blck.getInstances() if self._isInput(inst))

    def _getInputTypes(self, inst, cwrs):
        """
        Get the types associated with the input connectors of the given
        instance.

        :param inst: Instance to obtain input types for
        :type inst: WmhdbHasConnectorsP
        :param cwrs: Connectors-to-Wires mapping
        :rtype cwrs: dict[WmhdbConnector,
                          list[(WmhdbWire, WmhdbInstance, WmhdbConnector)]]
        :return: Connector-to-type mapping for input connectors
        :rtype: dict[str, WmhdbRtiType]
        """
        return {c.getName(): cwrs[c][0][0].getInferredRtiType()
                for c in inst.getInputConnectors() if cwrs[c]}

    def _isMissinInputTypes(self, inst, cwrs):
        """
        Whether the given instance has input connectors without associated
        type.

        :param inst: Instance to check for untyped input connectors
        :type inst: WmhdbInstance
        :param cwrs: Connectors-to-Wires mapping
        :rtype cwrs: dict[WmhdbConnector,
                          list[(WmhdbWire, WmhdbInstance, WmhdbConnector)]]
        :return: True if the instance has untyped input connectors
        :rtype: bool
        """
        return not all(
            w.getInferredRtiType() for c in inst.getInputConnectors()
            for w, *_ in cwrs[c])

    def _isMissingOutputTypes(self, inst, cwrs):
        """
        Whether the given instance has output connectors without associated
        type.

        :param inst: Instance to check for untyped output connectors
        :type inst: WmhdbInstance
        :param cwrs: Connectors-to-Wires mapping
        :rtype cwrs: dict[WmhdbConnector,
                          list[(WmhdbWire, WmhdbInstance, WmhdbConnector)]]
        :return: True if the instance has untyped output connectors
        :rtype: bool
        """
        return not all(
            w.getInferredRtiType() for c in inst.getOutputConnectors()
            for w, *_ in cwrs[c])

    def _setOutputTypes(self, inst, completeSet, incompleteSet, cwrs, path,
                        inputBlocks, recurse):
        """
        Set the types on the outgoing wires for the given instance.

        The destination instances of the wires that originate from the given
        instance will be scheduled for having their output wires typed, as
        well.

        :param inst: Instance to perform wire-type-setting on
        :type inst: WmhdbInstance
        :param completeSet: Set of instances ready for type propagation
                            with only typed input connectors
        :type completeSet: set[WmhdbInstance]
        :param incompleteSet: Set of instances possibly ready for type
                              propagation with some untyped input connectors
        :type incompleteSet: set[WmhdbInstance]
        :param cwrs: Connectors-to-Wires mapping
        :type cwrs: dict[WmhdbConnector:
                         list[(WmhdbWire, WmhdbInstance, WmhdbConnector)]]
        :param path: Instance path to current block
        :type path: tuple
        :param inputBlocks: Input blocks
        :type inputBlocks: dict[tuple, list[WmhdbBlockInstance]]
        :param recurse: Whether to recurse into child blocks
        :type recurse: bool
        """
        # noinspection PyShadowingNames
        def outputTypesMatch():
            for xpnt in inst.getOutputConnectors():
                rtyp = otps.get(xpnt.getName())
                for wire, *_ in cwrs[xpnt]:
                    if rtyp != wire.getInferredRtiType():
                        return False
            return True

        itps = self._getInputTypes(inst, cwrs)
        if isinstance(inst, WmhdbBlockInstance):
            otps = self._getBlockInstanceOutputType(
                inst, cwrs, path, inputBlocks, recurse)
        elif isinstance(inst, WmhdbModelInstance):
            try:
                otyp = wmhdb_getCommonType(itps.values())
                if self.cbuild and self.cpars:
                    otyp = self.cbuild.getTypeOutputInfo(inst, itps)
                    otps={}
                    for c in inst.getOutputConnectors():
                        otps[c.getName()]=\
                            self.cpars.getParsedRtiType\
                                    (otyp[c.getPortReference()])
                else:
                    otps = \
                        {c.getName(): otyp for c in inst.getOutputConnectors()}
            except Exception as e:
                # traceback.print_exc()
                raise WmhdbElementsException([(inst, path)], str(e))
        elif isinstance(inst, WmhdbFunction):
            try:
                otps = inst.getOutputTypes(itps, self.cpars)
            except Exception as e:
                # traceback.print_exc()
                raise WmhdbElementsException([(inst, path)], str(e))
        else:
            raise WmhdbElementsException(
                [(inst, path)],
                'Cannot determine output types for unknown instance kind')

        if outputTypesMatch() and not self._isInput(inst):
            return

        for xpnt in inst.getOutputConnectors():
            otyp = otps.get(xpnt.getName())
            wires = list(map(lambda t: t[0], cwrs[xpnt]))
            utps = []
            msng = []
            if not wires:
                continue
            for wire in wires:
                utyp = wire.getUserRtiType()
                if utyp:
                    utps.append(utyp)
                    if otyp and not wmhdb_canBeConvertedInto(otyp, utyp):
                        raise WmhdbElementsException(
                            [(inst, path)],
                            'Incompatible types for {}: {} vs. {}'.format(
                                inst.getName(), otyp, utyp))
                ityp = utyp or otyp
                if ityp:
                    wire.setInferredRtiType(ityp)
                elif not wire.getInferredRtiType():
                    msng.append(wire)

            if msng:
                if utps:
                    utyp = wmhdb_getCommonType(utps)
                    for wire in msng:
                        wire.setInferredRtiType(utyp)
                elif recurse:
                    raise WmhdbElementsException(
                        [(inst, path)],
                        'Cannot determine type for connector {}.{}'.format(
                            inst.getName(), xpnt.getName()))

            for _, oobj, _ in cwrs[xpnt]:
                if not self._isMissinInputTypes(oobj, cwrs):
                    if self._isMissingOutputTypes(oobj, cwrs):
                        completeSet.add(oobj)
                    if oobj in incompleteSet:
                        incompleteSet.remove(oobj)
                else:
                    incompleteSet.add(oobj)

    def _getBlockInstanceOutputType(self, inst, cwrs, path, inputBlocks,
                                    recurse=True):
        """
        Perform recursive type propagation for the given block instance.

        :param inst: Block instance to perform type propagation on recursivly
        :type inst: WmhdbBlockInstance
        :param cwrs: Connectors-to-Wires mapping
        :rtype cwrs: dict[str,
                          list[(WmhdbWire, WmhdbInstance, WmhdbConnector)]]
        :param path: Instance path to current block
        :type path: tuple
        :param inputBlocks: Input blocks
        :type inputBlocks: dict[tuple, list[WmhdbBlockInstance]]
        :param recurse: Whether to recurse into child blocks
        :type recurse: bool
        :return: Block output types
        :rtype: dict[str, WmhdbRtiType]
        """
        blck = inst.getBlock()
        """:type: WmhdbBlock"""

        xtps = {c.getPortReference(): cwrs[c][0][0].getInferredRtiType()
                for c in inst.getInputConnectors()
                if c.getPortReference() and cwrs[c]}

        for port in (p for p in blck.getPorts()
                     if p.hasFlag(WHDB_INPUT_OPERATOR_FLAG)):
            pnam = port.getParameterValue(WMHDB_PORT_NAME_PARAMETER)
            if not pnam:
                raise WmhdbElementsException(
                    [(inst, path)], 'Unnamed input port')
            if pnam in xtps:
                ptyp = xtps[pnam]
                for wire in blck.getConnectionsForObject(port):
                    wire.setInferredRtiType(ptyp)

        if recurse:
            self._propagateTypes(
                blck, inputBlocks, path + (inst,), recurse=recurse)

        otps = {}
        for port in (p for p in blck.getPorts()
                     if p.hasFlag(WHDB_OUTPUT_OPERATOR_FLAG)):
            pnam = port.getParameterValue(WMHDB_PORT_NAME_PARAMETER)
            if not pnam:
                raise WmhdbElementsException(
                    [(inst, path)], 'Unnamed output port')
            xpnt = inst.getConnectorByPortRef(pnam)
            wire = next(blck.getConnectionsForObject(port), None)
            otps[xpnt.getName()] = wire.getInferredRtiType()

        return otps

    def _deduceMissingOutputTypes(self, inst, completeSet, incompleteSet,
                                  cwrs):
        """
        Set the types on the outgoing wires for an instance with untyped
        input connectors.

        The destination instances of the wires that originate from the given
        instance will be scheduled for having their output wires typed, as
        well.

        :param inst: Instance to perform wire-type-deduction on
        :type inst: WmhdbInstance
        :param completeSet: Set of instances ready for type propagation
                            with only typed input connectors
        :type completeSet: set[WmhdbInstance]
        :param incompleteSet: Set of instances possibly ready for type
                              propagation with some untyped input connectors
        :type incompleteSet: set[WmhdbInstance]
        :param cwrs: Connectors-to-Wires mapping
        :type cwrs: {WmhdbConnector:
                     [(WmhdbWire, WmhdbInstance, WmhdbConnector)]}
        :return: Whether any new wire-type was set
        :rtype: bool
        """
        if not isinstance(inst, WmhdbFunction):
            return False

        otps = inst.getOutputTypes(
            {c.getName(): (cwrs[c][0][0].getInferredRtiType()
                           if cwrs[c] else None)
             for c in inst.getInputConnectors()}, self.cpars)

        ntyp = False

        for xpnt in inst.getOutputConnectors():
            for wire, oobj, _ in cwrs[xpnt]:
                if wire.getInferredRtiType() is None:
                    wire.setInferredRtiType(
                        wire.getUserRtiType() or otps.get(xpnt.getName()))
                    if wire.getInferredRtiType() is not None:
                        ntyp = True
                    if not self._isMissinInputTypes(oobj, cwrs):
                        completeSet.add(oobj)
                        if oobj in incompleteSet:
                            incompleteSet.remove(oobj)
                    else:
                        incompleteSet.add(oobj)

        return ntyp
