# ----------------------------------------------------------------------------
# - wmhdb-utils                                                              -
# - wild cruncher: wsh util functions                                        -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 PiV                                              -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_mblock
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher wsh util functions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import OrderedDict, deque
from itertools import chain

from wcr_core.db_api.whdb_connection_api import WhdbConnectionP
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP, \
    WhdbBlockInstanceP, WhdbModelInstanceP
from wcr_core.xmlio.wxio_xml_base import WxioXmlBase
from wcr_core.db.whdb_utils import whdb_getNextFunctionName, \
    whdb_getNextInstanceName
from wcr_core.db.whdb_connection_utils import whdb_flipConnType
from wcr_core.dbstore.wdbs_cloning import wdbs_cloneEntities
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_IN, \
    WDBS_XPNT_OUT
from wcr_core.db.whdb_protocols import WHDB_POSITION_KEY
from wcr_core.modules.model.db.wmhdb_standard_operators import \
    WMHDB_INPUT_PORT, WMHDB_OUTPUT_PORT


def wmhdb_convertEntitiesToBlock(entities, top, pos):
    """
    Create a new model with the given entries, create a new instance for the
    model and add it to the parent model.

    :param entities: Entities to create a model for
    :type entities: Iterable[WmhdbEntityP]
    :param top: Model currently holding the given entities
    :type top: WhdbBlockP
    :param pos: Position for the new instance
    :type pos: float, float
    :return: New instance and wires connecting to the instance
    :rtype: (WmhdbBlockInstance, list[WmhdbWire])
    """
    if not entities: return

    # noinspection PyUnusedLocal, PyShadowingNames
    def inputF(block, iobj, icon, oobj, ocon, wire):
        oper = WMHDB_INPUT_PORT
        port = block.createOperatorInstance(oper)
        port.setName(whdb_getNextFunctionName(oper, block))
        port.setMeta(ocon.getMeta(WHDB_POSITION_KEY), WHDB_POSITION_KEY)
        ports.append((port, oobj, ocon))
        return port

    # noinspection PyUnusedLocal, PyShadowingNames
    def outputF(block, iobj, icon, oobj, ocon, wire):
        oper = WMHDB_OUTPUT_PORT
        port = block.createOperatorInstance(oper)
        port.setName(whdb_getNextFunctionName(oper, block))
        port.setMeta(ocon.getMeta(WHDB_POSITION_KEY), WHDB_POSITION_KEY)
        ports.append((port, oobj, ocon))
        return port

    ports = []
    conns = []

    instance = top.createBlockInstance(name=whdb_getNextInstanceName(top))
    instance.setPosition(*pos)

    # noinspection PyTypeChecker
    wdbs_cloneEntities(entities, top, instance.getBlock(), inputF, outputF)

    ports.sort(key=lambda t: t[0].getY())
    for port, obj, conn in ports:
        pnam = port.getPortName()
        cdir = whdb_flipConnType(conn.getDirection())
        icon = instance.createConnectionPoint(cdir=cdir, port=pnam, name=pnam)
        conns.append((obj, conn, icon))
    instance.resetConnectionPointPositions()

    wires = []
    for obj, conn, icon in conns:
        wires.append(top.createConnection(obj, conn, instance, icon))

    return instance, wires


def wmhdb_resetBlockConnectors(instance, lexical=False):
    """
    Reset an instance's connectors.

    :param instance: Instance to reset connectors for
    :type instance: WhdbBlockInstanceP
    :param lexical: Set to Tru to sort ports by names
    """
    block = instance.getBlock()
    ports = block.getPortNames()

    for cdir in (WDBS_XPNT_IN, WDBS_XPNT_OUT):
        plst = ports.get(cdir)
        if plst:
            if lexical : plst.sort(key=lambda t: int (t[0]))
            for port, _ in plst:
                instance.createConnectionPoint(cdir=cdir, port=port)
    instance.resetConnectionPointPositions()


def wmhdb_resetConnectors(inst, confirmF):
    """
    Reset the connectors of the given instances.

    :param inst: Instance to reset connectors on
    :type inst: WhdbInstanceP
    :param confirmF: Confirmation function for connector reuse --
                     True: reuse, False: clear and recreate, None: cancel
    :type confirmF: () -> bool
    """

    def getPorts(cdir):
        _ports = allPorts.get(cdir)
        if _ports:
            ports, *_ = zip(*_ports)
        else:
            ports = None

        if not ports:
            return [], set(), set()
        cprfs = [c.getPortReference() for c
                 in inst.getConnectionPoints(cdir=cdir)]

        cpfss = set(cprfs)
        if len(cpfss) != len(cprfs):
            raise Exception('Model contains duplicate port names')

        ports = list(ports) if ports else []
        portSet = set(ports)
        return ports, portSet, set(ports).intersection(cpfss)

    def orderF(conn):
        if conn.getDirection() == WDBS_XPNT_IN:
            return inPorts.index(conn.getPortReference())
        elif conn.getDirection() == WDBS_XPNT_OUT:
            return len(inPorts) + outPorts.index(conn.getPortReference())
        else:
            return len(inPorts) + len(outPorts) + 1

    if isinstance(inst, WhdbModelInstanceP):
        model = inst.getModel()
        model.resetPortNames()
        allPorts = model.getPortNames()
    elif isinstance(inst, WhdbBlockInstanceP):
        allPorts = inst.getBlock().getPortNames()

    inPorts, inPortSet, inOldPortSet = getPorts(WDBS_XPNT_IN)
    outPorts, outPortSet, outOldPortSet = getPorts(
        WDBS_XPNT_OUT)

    newInPorts = inPorts[:]
    newOutPorts = outPorts[:]

    if inPortSet == inOldPortSet and outPortSet == outOldPortSet and \
            inst.getConnectionPointCount() == len(inPorts) + len(outPorts):
        return

    if confirmF is None:
        inst.clearConnectionPoints()
    else:
        inOldConns = list(inst.getConnectionPoints(cdir=WDBS_XPNT_IN))
        outOldConns = list(inst.getConnectionPoints(cdir=WDBS_XPNT_OUT))
        if inOldConns or outOldConns:
            reuse = confirmF()
            if reuse is None:
                return
            elif reuse:
                def cleanPorts(plst, pset, olst, oset):
                    for c in olst:
                        pref = c.getPortReference()
                        if not pref or pref not in pset:
                            inst.removeConnectionPoint(c)
                    for p in oset: plst.remove(p)

                cleanPorts(newInPorts, inPortSet, inOldConns, inOldPortSet)
                cleanPorts(newOutPorts, outPortSet, outOldConns, outOldPortSet)
            else:
                inst.clearConnectionPoints()
    for port in newInPorts:
        inst.createConnectionPoint(cdir=WDBS_XPNT_IN, port=port)
    for port in newOutPorts:
        inst.createConnectionPoint(cdir=WDBS_XPNT_OUT, port=port)
    inst.setConnectionPointOrder(orderF)
    inst.resetConnectionPointPositions()


def wmhdb_find_connected(item, block):
    """
    Find database objects connected by wires to the given one, in the given
    model block.

    :param item: Database entity to find connected items for
    :type item: WhdbBaseP
    :param block: Model block level to look for connected items in
    :type block: WhdbBlockP
    :return: Set of items connected to the given one
    :rtype: set[WhdbBaseP]
    """
    from .wmhdb_protocols import WmhdbHasConnectorsP

    # noinspection PyShadowingNames
    def connected(item, conn):
        if isinstance(item, WhdbConnectionP):
            if conn == item.getStartConnectionPoint():
                return [(item.getEndObject(),
                         item.getEndConnectionPoint())]
            elif conn == item.getEndConnectionPoint():
                return [(item.getStartObject(),
                         item.getStartConnectionPoint())]
        elif isinstance(item, WmhdbHasConnectorsP):
            rs = []
            for c in item.getRelatedConnectors(conn):
                rs.extend([(w, c) for w
                           in block.getConnectionsForObject(item, c)])
            return rs
        return []

    items = set()
    items.add(item)
    if isinstance(item, WhdbConnectionP):
        todo = [(item.getStartObject(), item.getStartConnectionPoint()),
                (item.getEndObject(), item.getEndConnectionPoint())]
    elif isinstance(item, WhdbInstanceP):
        todo = [(w, c) for c in item.getConnectionPoints() for w
                in block.getConnectionsForObject(item, c)]
    else:
        todo = []

    for w, c in todo:
        items.add(w)
    while todo:
        item, conn = todo.pop()
        for i, c in connected(item, conn):
            if i not in items:
                items.add(i)
                todo.append((i, c))

    return items


# noinspection PyProtectedMember
def wmhdb_fix_names(block):
    """
    Fix missing model/instance/function-names for the given model block.

    :param block: Model block whose children's names are to be fixed
    :type block: WhdbBlockP
    """

    def fixNames(entities, maxId, prefix):
        idx = maxId
        for e in entities:
            if not e.getName():
                idx += 1
                e.setName(prefix + str(idx))

    # noinspection PyShadowingNames
    def addToDict(entity, dct, pattern):
        entityId = id(entity)
        modelName = entity.getName()
        nameId = 0
        if modelName:
            m = pattern.match(modelName)
            if m is not None:
                nameId = (int(m.group(1)))
        if entityId in dct:
            dct.move_to_end(entityId)
        else:
            dct[entityId] = entity
        return nameId

    models = OrderedDict()
    instances = {}
    funcs = {}
    togo = deque([])

    models[id(block)] = block

    togo.appendleft(block)
    maxModelId = 0
    maxInstanceId = 0
    maxFuncId = {}

    while len(togo) > 0:
        block = togo.popleft()

        for func in block.getOperatorInstances():
            oper = func.getOperator().getCode()
            nameId = addToDict(func, funcs, WxioXmlBase.NUMBERED_FUNC_NAME)
            maxFuncId[oper] = max(nameId, maxFuncId.get(oper, 0))

        for inst in chain(block.getBlockInstances(), block.getModelInstances):
            model = inst.getModel()
            nameId = addToDict(model, models,
                               WxioXmlBase.NUMBERED_MODEL_NAME)
            maxModelId = max(nameId, maxModelId)
            nameId = addToDict(inst, instances,
                               WxioXmlBase.NUMBERED_INSTANCE_NAME)
            maxInstanceId = max(nameId, maxInstanceId)

            if isinstance(inst, WhdbBlockInstanceP):
                togo.append(model)

    models = list(models.values())[1:]
    instances = instances.values()
    funcs = funcs.values()

    fixNames(models, maxModelId, 'model')
    fixNames(instances, maxInstanceId, 'minst')
    for f in funcs:
        if not f.getName():
            oper = f.getOperator().getCode()
            maxFuncId[oper] += 1
            f.setName(oper + str(maxFuncId[oper]))
