# ----------------------------------------------------------------------------
# - wmhdb-verifier                                                           -
# - wild cruncher:hierarchy verifier                       core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 clemens oertel                                   -
# - copyright (c) 2013-2013 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_verifier
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher hierarchy verifier

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db.whdb_operator import WHDB_PORT_OPERATOR_FLAG
from wcr_core.db.whdb_verifier import WhdbVerifierModuleP, \
    WhdbHierarchyVerifier
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP
from wcr_core.modules.model.db.wmhdb_block_base import WmhdbBlockP
from wcr_core.modules.model.db.wmhdb_standard_operators import \
    WMHDB_OUTPUT_PORT, WMHDB_INPUT_PORT, WMHDB_PORT_NAME_PARAMETER
from wcr_core.modules.model.db.wmhdb_connection_point import WmhdbConnector
from wcr_core.common.wcmn_verifier_module import WcmnHierarchyVerificationError


class WmhdbUniquePortIdsVerifierModule(WhdbVerifierModuleP):
    """
    Verify that all input/output ports of a model have unique identifiers.
    """

    ERROR_ID = 'hierarchy.model.duplicate-port-ids'

    ERROR_MESSAGE = 'Duplicate ID \'{pid}\' found in ports {funcs}'

    DESCRIPTION = 'The IDs of a model\'s input- and output-ports must be ' \
                  'unique.'

    def verifyBlock(self, block, topLevel=False):
        pids = {}
        errors = []

        for func in block.getOperatorInstances():
            op = func.getOperator()
            if op is WMHDB_INPUT_PORT or op is WMHDB_OUTPUT_PORT:
                pid = func.getParameterValue(WMHDB_PORT_NAME_PARAMETER)
                pids.setdefault(pid, []).append(func)

        for pid, funcs in pids.items():
            if len(funcs) > 1:
                fnms = ', '.join("'%s'" % f.getName() for f in funcs)
                errors.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID,
                    self.ERROR_MESSAGE.format(pid=pid, funcs=fnms),
                    funcs, self))
        return errors

    def verifyInstance(self, instance):
        pass

    def verifyConnection(self, wire):
        pass

    def description(self):
        return WmhdbUniquePortIdsVerifierModule.DESCRIPTION


class WmhdbInstanceToModelConnectorsVerifierModule(WhdbVerifierModuleP):
    """
    Verify that all of a model's input-/output-ports have corresponding
    instance connectors, and that all instance connectors have a corresponding
    port.
    """

    ERROR_ID = 'hierarchy.model.unconnected-model-connectors'

    ERROR_MESSAGE_M = 'Unconnected model {type} connector{s} {conns} in {name}'

    ERROR_MESSAGE_I = 'Unconnected instance {type} connector{s} {conns} in' \
                      ' {name}'

    DESCRIPTION = 'For all input- and output-ports of non-top-level model ' \
                  'blocks, the model instance representing the model must ' \
                  'have matching connectors. Also, the model instance may ' \
                  'not have any connectors without a corresponding input- ' \
                  'or output-port.'

    def _ldiff(self, la, lb):
        sb = set(lb)
        return [a for a in la if a not in sb]

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyInstance(self, instance):
        if not isinstance(instance, WhdbBlockInstanceP):
            return None

        # noinspection PyShadowingNames
        def verify(l1, l2, errMsg, cType, errors):
            diff = self._ldiff(l1, l2)
            if diff:
                errors.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID,
                    errMsg.format(type=cType,
                                  conns=', '.join("'%s'" % d for _, d in diff),
                                  name=instance.getName(),
                                  s='s'[len(diff) == 1:]),
                    [instance], self))

        # noinspection PyShadowingNames
        def verifySameLists(instConns, mConns, cType, errors):
            verify(instConns, mConns, self.ERROR_MESSAGE_I, cType, errors)
            verify(mConns, instConns, self.ERROR_MESSAGE_M, cType, errors)

        model = instance.getBlock()

        inputs = []
        outputs = []
        for func in model.getOperatorInstances():
            if func.getOperator() is WMHDB_INPUT_PORT:
                inputs.append(('IN', func.getPortName()))
            elif func.getOperator() is WMHDB_OUTPUT_PORT:
                outputs.append(('OUT', func.getPortName()))

        inConns = [
            ('IN', c.getPortReference()) for c
            in instance.getConnectionPoints(cdir=WmhdbConnector.Input)]
        outConns = [
            ('OUT', c.getPortReference()) for c
            in instance.getConnectionPoints(cdir=WmhdbConnector.Output)]

        errors = []
        verifySameLists(inConns, inputs, 'input', errors)
        verifySameLists(outConns, outputs, 'output', errors)

        return errors

    def verifyConnection(self, wire):
        pass

    def description(self):
        return WmhdbInstanceToModelConnectorsVerifierModule.DESCRIPTION


class WmhdbConnectorTypeVerifierModule(WhdbVerifierModuleP):
    """
    Verify that all of a connector's wires are of the same type.
    """

    ERROR_ID = 'hierarchy.model.connector-types'

    ERROR_MESSAGE = 'Connector {i}.{c} has multiple wires ' \
                    'of different types'

    DESCRIPTION = 'If a connector has multiple wires attached to it, all ' \
                  'the wires must be of the same sinal type'

    def _verifyConnectors(self, inst, block):
        """
        :param inst: Instance whose connectors to check
        :type inst: WhdbInstanceP
        :param block: Model block the instance is located in
        :type block: WmhdbBlockP
        :return: List of instances/connectors with faulty wires
        :rtype: Iterable[(WmhdbInstanceBaseP, WmhdbConnector,
                          Iterable[WmhdbWire])]
        """
        faulty = []
        for connector in inst.getConnectionPoints():
            wires = list(block.getConnectionsForObject(inst, connector))
            if len(wires) > 1:
                userType = wires[0].getUserRtiType()
                if any([w.getUserRtiType() != userType for w in wires]):
                    faulty.append((inst, connector, wires))

        return faulty

    def verifyBlock(self, block, topLevel=False):
        faultyWires = []
        for inst in block.getInstances():
            faultyWires.extend(self._verifyConnectors(inst, block))

        return [WcmnHierarchyVerificationError(
            self.ERROR_ID,
            self.ERROR_MESSAGE.format(i=i.getName(),
                                      c=c.getDescription()),
            ws, self) for i, c, ws in faultyWires]

    def verifyInstance(self, instance):
        pass

    def verifyConnection(self, wire):
        pass

    def description(self):
        return WmhdbConnectorTypeVerifierModule.DESCRIPTION


class WmhdbXBlockWireTypeVerifierModule(WhdbVerifierModuleP):
    """
    Verify that all the port- and the connector-type of inner block model
    instances match.
    """

    ERROR_ID = 'hierarchy.model.xblock-wire-types'

    ERROR_MESSAGE = 'Connector type {i}.{oc}, {ot}, does not match ' \
                    'corresponding port type {p}.{ic}, {it}'

    DESCRIPTION = 'The type of the wire attached to a connector of a model ' \
                  'block instance must match the type of the wire attached ' \
                  'to the corresponding port of the model model block.'

    def verifyBlock(self, block, topLevel=False):
        # noinspection PyShadowingNames
        def getPortAndConnector(cnam, model):
            port = next((p for p in model.getOperatorInstances()
                         if p.getPortName() == cnam), None)
            if port is not None:
                ic = next(port.getConnectionPoints(), None)
                if ic is not None:
                    return port, ic
            return None, None

        def getWireAndType(b, i, c):
            ws = list(b.getConnectionsForObject(i, c))
            if ws:
                return ws[0], ws[0].getUserRtiType()
            else:
                return None, None

        err = []
        for inst in block.getBlockInstances():
            subBlock = inst.getBlock()
            for oc in inst.getConnectionPoints():
                cnam = oc.getPortReference()
                if not cnam: continue
                port, ic = getPortAndConnector(cnam, subBlock)
                if port is None or ic is None: continue

                ow, ot = getWireAndType(block, inst, oc)
                iw, it = getWireAndType(subBlock, port, ic)

                if ot is not None and it is not None and ot != it:
                    msg = self.ERROR_MESSAGE
                    err.append(WcmnHierarchyVerificationError(
                        self.ERROR_ID,
                        msg.format(i=inst.getName(),
                                   oc=oc.getDescription(),
                                   it=it.getDescription(),
                                   p=port.getName(),
                                   ic=ic.getDescription(),
                                   ot=ot.getDescription()),
                        [ow, iw], self))

        return err

    def verifyInstance(self, instance):
        pass

    def verifyConnection(self, wire):
        pass

    def description(self):
        return WmhdbXBlockWireTypeVerifierModule.DESCRIPTION


class WmhdbHasSchedGroupVerifierModule(WhdbVerifierModuleP):
    """
    Verify that every model/function instance belongs to a scheduling group.
    """

    ERROR_ID = 'hierarchy.model.missing-schedgroup'

    ERROR_MESSAGE = 'Instance/Function without scheduling group ({name})'

    DESCRIPTION = 'Every model instance or function must belong to a ' \
                  'scheduling group.'

    def verifyBlock(self, block, topLevel=False):
        if not topLevel:
            return

        errors = []

        for inst in block.getInstances():
            if not inst.getSchedulingGroup():
                errors.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID,
                    self.ERROR_MESSAGE.format(name=inst.getName()),
                    [inst], self))

        return errors

    def verifyInstance(self, instance):
        pass

    def verifyConnection(self, wire):
        pass

    def description(self):
        return WmhdbHasSchedGroupVerifierModule.DESCRIPTION


class WmhdbTopLevelNoPortsVerifierModule(WhdbVerifierModuleP):
    """
    Verify that the top-level block does not contain any ports.
    """

    ERROR_ID = 'hierarchy.model.toplevel-port'

    ERROR_MESSAGE = 'Top-level block contains port: {name}'

    DESCRIPTION = 'At the top-level, ports are not allowed.'

    def verifyBlock(self, block, topLevel=False):
        def mkError(func):
            return WcmnHierarchyVerificationError(
                self.ERROR_ID,
                self.ERROR_MESSAGE.format(name=func.getName()),
                [func], self)

        if not topLevel:
            return []

        return [mkError(f) for f in block.getOperatorInstances()
                if f.getOperator().hasFlag(WHDB_PORT_OPERATOR_FLAG)]

    def verifyConnection(self, wire):
        pass

    def verifyInstance(self, instance):
        pass

    def description(self):
        return WmhdbTopLevelNoPortsVerifierModule.DESCRIPTION


#: List of all hierarchy verification modules to be enabled by default
WMHDB_VRFYM = [WmhdbHasSchedGroupVerifierModule(),
               WmhdbUniquePortIdsVerifierModule(),
               WmhdbInstanceToModelConnectorsVerifierModule(),
               WmhdbConnectorTypeVerifierModule(),
               WmhdbXBlockWireTypeVerifierModule(),
               WmhdbTopLevelNoPortsVerifierModule()]


class WmhdbHierarchyVerifier(WhdbHierarchyVerifier):
    """
    Model hierarchy verifier.
    """

    def __init__(self, modules=None):
        """
        :param modules: Additional verification modules to use.
        :type modules: Iterable[WhdbVerifierModuleP] | None
        """
        if modules:
            mods = list(WMHDB_VRFYM)
            mods.extend(modules)
        else:
            mods = WMHDB_VRFYM

        super().__init__(mods)
