# ----------------------------------------------------------------------------
# - wmhdb-wire                                                               -
# - wild cruncher:wsh model wire object                    core / model / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - author    (c) 2012-2013 amaury darsch                                    -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_wire
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher silkan hierarchy model wire

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------
from wcr_core.db.whdb_connection import WhdbConnection
from wcr_core.db.whdb_connection_path_manhattan import \
    WhdbManhattanConnectionPath
from wcr_core.db.whdb_block_flattener import WHDB_CONNECTION_ALIASES_KEY
from wcr_core.dbstore.wdbs_meta_description import *
from wcr_core.dbstore.wdbs_meta_type import WdbsListMetaType, \
    WdbsStringMetaType
from wcr_core.modules.model.db.wmhdb_rti_type import WMHDB_RTITYPE_META_TYPE


# ----------------------------------------------------------------------------
# - wire class                                                               -
# ----------------------------------------------------------------------------

WMHDB_WIRE_UTYPE_KEY = 'user_type'

WMHDB_WIRE_ITYPE_KEY = 'inferred_type'


# noinspection PyAbstractClass
class WmhdbWire(WhdbConnection):
    """
    The WmhdbWire is the silkan hierarchical wire object. A wire is defined
    by name and connect two connectors. The object bound to a connector can
    be either an instance or a function.
    """

    META = WdbsMetaDescription(
        WhdbConnection.META,
        WdbsMetaField(WMHDB_WIRE_UTYPE_KEY, WMHDB_RTITYPE_META_TYPE),
        WdbsMetaField(WMHDB_WIRE_ITYPE_KEY, WMHDB_RTITYPE_META_TYPE,
                      temporary=True),
        WdbsMetaField(WHDB_CONNECTION_ALIASES_KEY,
                      WdbsListMetaType(WdbsStringMetaType(default=None)),
                      temporary=True))

    def __init__(self, conn, wreg):
        """
        :param conn: Raw connection
        :type conn: WdbsConnectionP
        :param wreg: Raw element wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        super().__init__(conn, wreg, WhdbManhattanConnectionPath)

        self.__rawConnection = conn
        self.__wrapperRegistry = wreg

    def getConnectionTypeDescription(self):
        utyp = self.getUserRtiType()
        ityp = self.getInferredRtiType()

        wireType = utyp.getDescription() if utyp is not None else 'auto'
        if ityp is not None:
            wireType += ' ({0})'.format(ityp.getDescription())
        return wireType

    def setNameInfo(self, name, info):
        """
        Set the wire's name and info.

        :param name: The new name for the wire
        :type name: str
        :param info: The new info for the wire
        :type info: str
        """
        self.__rawConnection.setName(name)
        self.__rawConnection.setInfo(info)

    def getStartConnector(self):
        """
        Get the wire's start connector.

        :return: Start connector of the wire
        :rtype: WmhdbConnector
        """
        return self.getStartConnectionPoint()

    def getEndConnector(self):
        """
        Get the wire's end connector.

        :return: End connector of the wire
        :rtype: WmhdbConnector
        """
        return self.getEndConnectionPoint()

    def getUserRtiType(self):
        """
        Get the user-defined type for the wire.

        :return: User-defined type
        :rtype: WmhdbRtiType
        """
        return self.__rawConnection.getMeta(WMHDB_WIRE_UTYPE_KEY)

    def setUserRtiType(self, rtyp):
        """
        Set the user-defined type, if any.

        :param rtyp: The new user-defined type, or `None`
        :type rtyp: WmhdbRtiType
        """
        return self.__rawConnection.setMeta(rtyp, WMHDB_WIRE_UTYPE_KEY)

    def getInferredRtiType(self):
        """
        Get the inferred type for the wire, if any.

        :return: Inferred type, or `None`
        :rtype: WmhdbRtiType | None
        """
        return self.__rawConnection.getMeta(WMHDB_WIRE_ITYPE_KEY)

    def setInferredRtiType(self, rtyp):
        """
        Set the inferred type.

        :param rtyp: The new inferred type
        :type rtyp: WmhdbRtiType | None
        """
        return self.__rawConnection.setMeta(rtyp, WMHDB_WIRE_ITYPE_KEY)

    def getAliases(self):
        """
        Get the wire's alias names.

        :return: Alias names for the wire
        :rtype: [str]
        """
        return self.__rawConnection.getMeta(WHDB_CONNECTION_ALIASES_KEY)

    def addAlias(self, alias):
        """
        Add an alias name for the wire.

        :param alias: Alias name to be added
        :type alias: str
        """
        if not alias:
            return
        self.__rawConnection.setMeta(alias, WHDB_CONNECTION_ALIASES_KEY, -1)

    def removeAlias(self, alias):
        """
        Remove an alias name for the wire.

        :param alias: Alias name to be removed
        :type alias: str
        """
        if not alias:
            return
        aliases = self.__rawConnection.getMeta(
            WHDB_CONNECTION_ALIASES_KEY) or []
        aliases.remove(alias)
        self.__rawConnection.setMeta(aliases, WHDB_CONNECTION_ALIASES_KEY)
