# ----------------------------------------------------------------------------
# - wmhdb-xmlio                                                              -
# - wild cruncher:wsh xml serialization/deserialization     project database -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmhdb_xmlio
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher model DB xml serialization/deserialization

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from functools import partial
from xml.etree.ElementTree import ElementTree

from wcr_core.common.wcmn_file_handler import WcmnFileReaderBase
from wcr_core.common.wcmn_mime_type import WCMN_PAYLOAD_MODEL_PORTS, \
    WCMN_PAYLOAD_ENTRY_POINTS
from wcr_core.modules.model.wmmod_application_module import \
    WMMOD_MODEL_APP_MODULE
from wcr_core.xmlio.wxio_xmlio import wxio_deserializeXml
from wcr_core.project.wprj_xmlio import WPRJ_SILKAN_XML_MIME_TYPE, \
    WPRJ_SILKAN_XML_MIME_NAME


class WmhdbSilkanXmlReader(WcmnFileReaderBase):
    def __init__(self):
        super().__init__(
            [WPRJ_SILKAN_XML_MIME_TYPE],
            WCMN_PAYLOAD_MODEL_PORTS | WCMN_PAYLOAD_ENTRY_POINTS)

    @property
    def name(self):
        return 'Silkan Model-XML'

    def readText(self, text, payloadHandler, xctx, types=None):
        return super().readText(text, payloadHandler, xctx, types)

    def readFile(self, file, payloadHandler, xctx, types=None):
        typesRead = 0
        if not types:
            types = payloadHandler.handledPayloadTypes

        tree = ElementTree(file=file.path)
        hnod = tree.getroot().find('hierarchy')
        if hnod is not None and hnod.get('module') != 'model':
            return typesRead

        model = WMMOD_MODEL_APP_MODULE.createTopLevelBlock()
        wxio_deserializeXml(
            tree, 'tophm', model.rawElement)

        if types & WCMN_PAYLOAD_ENTRY_POINTS:
            typesRead |= WCMN_PAYLOAD_ENTRY_POINTS
            payloadHandler.handlePayload(
                WCMN_PAYLOAD_ENTRY_POINTS, WPRJ_SILKAN_XML_MIME_NAME,
                self._readEntryPoints(model))
        if types & WCMN_PAYLOAD_MODEL_PORTS:
            typesRead |= WCMN_PAYLOAD_MODEL_PORTS
            payloadHandler.handlePayload(
                WCMN_PAYLOAD_MODEL_PORTS, WPRJ_SILKAN_XML_MIME_NAME,
                partial(self._readModelPorts, model))
        return typesRead

    @staticmethod
    def _readEntryPoints(model):
        return iter(model.getSchedulingGroupSet())

    @staticmethod
    def _readModelPorts(model, sgrp):
        """
        Retrieve list of model ports.

        :param model: Top-level block
        :type model: WmhdbTopHierarchy
        :return: List of port names
        :rtype: {int:[(str,str)]}
        """
        if sgrp not in model.getSchedulingGroupSet():
            raise ValueError(
                'Scheduling group {} not found in model'.format(sgrp))

        return model.createSchedulingGroup(sgrp).getPortNames()
