# ----------------------------------------------------------------------------
# - wmplg-library-reader                                                     -
# - wild cruncher: library reader                      core / model / plugin -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmplg_library_reader
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Model library reader

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from importlib import import_module
from xml.etree.ElementTree import ElementTree

from silkan_common.scmn_file_path import scmn_path_dir, scmn_path_join
from wcr_core.common.wcmn_parameter_description import \
    WCMN_PARAM_FORMAT_STRING_FLAG
from wcr_core.db.whdb_operator import whdb_makeConnXYPositions
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_IN, \
    WDBS_XPNT_OUT
from wcr_core.modules.model.db.wmhdb_connection_point_manager import \
    wmhdb_createFixedXPointManager, wmhdb_createXPointDesc, \
    WmhdbVarInConnectionPointManager, WmhdbSimpleConnectionPointManager, \
    WmhdbVarOutConnectionPointManager, WmhdbDualVarConnectionPointManager
from wcr_core.modules.model.db.wmhdb_operator import WmhdbOperator
from wcr_core.plugin.wplg_library import WLIB_ICON_SVG, WLIB_ICON_PNG32


_WMPLG_PARAM_FLAGS = {
    'formatted_string': WCMN_PARAM_FORMAT_STRING_FLAG
}

_WMPLG_ICON_TYPES = {
    'png32': WLIB_ICON_PNG32,
    'svg': WLIB_ICON_SVG
}


def wmplg_read_operator(file, lcod):
    """
    Read an operator from given XML description file.

    :param file: XML description file
    :type file: str
    :param lcod: Library code
    :type lcod: str
    :return: Operator
    :rtype: WmhdbOperator
    """
    tree = ElementTree(file=file)
    node = tree.getroot()
    if node.tag != 'wcr-operator':
        raise Exception(
            'File {} does not contain a Wild Cruncher 2.0 '
            'operator specification'.format(file))
    code = lcod + '.' + node.get('code')
    title = node.get('title')

    descNode = node.find('description')
    desc = descNode.text if descNode is not None else ''

    icns = _wmplg_read_connection_points(
        node.find('in-connectors'), 'i', WDBS_XPNT_IN)
    ocns = _wmplg_read_connection_points(
        node.find('out-connectors'), 'o', WDBS_XPNT_OUT)

    polNode = node.find('polarity-names')
    pols = None
    if polNode is not None:
        regNode = polNode.find('regular-polarity')
        reg = regNode.text if regNode is not None else ''
        invNode = polNode.find('inverted-polarity')
        inv = invNode.text if invNode is not None else ''
        pols = (reg, inv)

    paramsNode = node.find('parameters')
    params = []
    """:type: list[(str, str, str, object, str, int)]"""
    if paramsNode is not None:
        for paramNode in paramsNode:
            pname = paramNode.get('name')
            ptitle = paramNode.get('title')
            pdesc = paramNode.get('description')
            pdefault = paramNode.get('default')
            ptype = paramNode.get('type')
            pflags = sum(_WMPLG_PARAM_FLAGS[f] for f
                         in paramNode.get('flags', '').split(',') if f)

            params.append((pname, ptitle, pdesc, pdefault, ptype, pflags))

    meta = {}
    for mitm in node.findall('meta'):
        meta[mitm.get('key')] = mitm.text.strip()

    symNode = node.find('symbol')
    symbol = symNode.text.strip() if symNode is not None else None

    icons = {}
    iconsNode = node.find('icons')
    if iconsNode is not None:
        for iconNode in iconsNode:
            ityp = _WMPLG_ICON_TYPES.get(iconNode.get('type'))
            if ityp is not None:
                icons[ityp] = scmn_path_join(scmn_path_dir(file),
                                             iconNode.text.strip())

    flagsNode = node.find('flags')
    if flagsNode is not None:
        flags = {f.get('name') for f in flagsNode.findall('flag')}
    else:
        flags = set()

    outTypesNode = node.find('out-types')
    outTypesF = None
    if outTypesNode is not None:
        modl = outTypesNode.get('module')
        fnct = outTypesNode.get('function')

        if modl and fnct:
            outTypesF = getattr(import_module(modl), fnct)
        elif outTypesNode.text:
            def getOutputTypes(inTypes, clst, paramF):
                outTypes = {}
                exec(outTypesCode,
                     {'INTYPES': inTypes, 'CONNECTORS': clst,
                      'OUTTYPES': outTypes, 'PARAMF': paramF})
                return outTypes

            outTypesCode = compile(outTypesNode.text, '', 'exec')
            outTypesF = getOutputTypes

    # noinspection PyTypeChecker
    return WmhdbOperator(
        code=code,
        xmngr=_wmplg_create_xpoint_manager(icns + ocns),
        name=title,
        info=desc,
        meta=meta,
        parameters=params,
        flags=flags,
        icons=icons,
        symbol=symbol,
        polarities=pols,
        outTypeF=outTypesF)


def _wmplg_read_connection_points(nlst, pfx, cdir):
    if nlst is None or len(nlst) == 0:
        return []

    def readXPointDef(node):
        if node.tag == 'conn-variable':
            minc = int(node.get('min')) if node.get('min') else 0
            maxc = int(node.get('max')) if node.get('max') else None
            return 'var', {'cdir': cdir, 'prfx': node.get('prefix') or pfx,
                           'minc': minc, 'maxc': maxc}
        elif node.tag == 'conn-sequence':
            start = int(node.get('start')) if node.get('start') else 1
            count = int(node.get('count')) if node.get('count') else 1
            return 'seq', {'cdir': cdir, 'prfx': node.get('prefix') or pfx,
                           'start': start, 'count': count}
        elif node.tag == 'conn-optional':
            return 'opt', {'cdir': cdir, 'name': node.get('name') or pfx,
                           'title': node.get('title') or pfx}
        elif node.tag == 'connector':
            return 'single', {'cdir': cdir, 'name': node.get('name') or pfx}

        return None

    return list(filter(None, map(readXPointDef, nlst)))


def _wmplg_create_xpoint_manager(xpnts):
    _xpnts = []
    opts = []
    for xtyp, xdat in xpnts:
        if xtyp == 'seq':
            start = xdat['start']
            count = xdat['count']
            for xnum in range(start, start + count):
                _xpnts.append(('single',
                               {'cdir': xdat['cdir'],
                                'name': '{}{}'.format(xdat['prfx'], xnum)}))
        elif xtyp == 'opt':
            opts.append(xdat)
        else:
            _xpnts.append((xtyp, xdat))

    icns = [xpnt for xpnt in _xpnts if xpnt[1]['cdir'] == WDBS_XPNT_IN]
    ocns = [xpnt for xpnt in _xpnts if xpnt[1]['cdir'] == WDBS_XPNT_OUT]

    itps = [xtyp for xtyp, _ in icns]
    otps = [xtyp for xtyp, _ in ocns]

    isng = all(map(lambda x: x == 'single', itps))
    osng = all(map(lambda x: x == 'single', otps))

    if isng and osng:
        return wmhdb_createFixedXPointManager(
            icns=[d['name'] for _, d in icns],
            ocns=[d['name'] for _, d in ocns],
            opts=opts)
    elif itps == ['var'] and osng:
        ocnt = len(ocns)
        if ocnt == 1:
            fxds = [wmhdb_createXPointDesc(
                cdir=WDBS_XPNT_OUT, name=ocns[0][1]['name'])]
        elif ocnt > 1:
            cpss = whdb_makeConnXYPositions(WDBS_XPNT_OUT, ocnt)
            fxds = [wmhdb_createXPointDesc(cdir=WDBS_XPNT_OUT,
                                           name=d['name'],
                                           cpos=(x, y), fcng=f)
                    for (_, d), (x, y, f) in zip(ocns, cpss)]
        else:
            fxds = []
        idat = icns[0][1]
        return WmhdbVarInConnectionPointManager(
            fxds=fxds, prfx=idat['prfx'],
            minc=idat['minc'], maxc=idat['maxc'], opts=opts)
    elif isng and otps == ['var']:
        icnt = len(icns)
        if icnt == 1:
            fxds = [wmhdb_createXPointDesc(
                cdir=WDBS_XPNT_IN, name=icns[0][1]['name'])]
        elif icnt > 1:
            cpss = whdb_makeConnXYPositions(WDBS_XPNT_IN, icnt)
            fxds = [wmhdb_createXPointDesc(cdir=WDBS_XPNT_IN,
                                           name=d['name'],
                                           cpos=(x, y), fcng=f)
                    for (_, d), (x, y, f) in zip(icns, cpss)]
        else:
            fxds = []
        odat = ocns[0][1]
        return WmhdbVarOutConnectionPointManager(
            fxds=fxds, prfx=odat['prfx'],
            minc=odat['minc'], maxc=odat['maxc'], opts=opts)
    elif itps == ['var'] and otps == ['var']:
        idat = icns[0][1]
        odat = ocns[0][1]
        if odat['minc'] or odat['maxc'] or idat['minc'] or idat['maxc']:
            return WmhdbDualVarConnectionPointManager(
                iprfx=idat['prfx'], oprfx=odat['prfx'],
                iminc=idat['minc'], imaxc=idat['maxc'],
                ominc=odat['minc'], omaxc=odat['maxc'])
        return WmhdbSimpleConnectionPointManager(
            ipfx=idat['prfx'], opfx=odat['prfx'])
    else:
        raise Exception(
            'Cannot handle connection point types IN: {}, OUT: {}'.format(
                itps, otps))
