# ----------------------------------------------------------------------------
# - wfctl-controllers                                                        -
# - wild cruncher: state machine ctrls     core / statemachine / controllers -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfctl_controllers
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - State machine controllers

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.controller.wctl_controller_mixins import *
from wcr_core.controller.wctl_default_controllers import \
    WctlSimpleControllerFactoryBase
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_OUT, \
    WDBS_XPNT_IN
from wcr_core.modules.statemachine.db.wfhdb_xbt_utils import \
    wfhdb_is_xbt_state, wfhdb_is_xbt_transition


class WfctlControllerBase(WctlControllerP):
    def __init__(self, fcty):
        """
        :param fcty: Controller factory
        :type fcty: WfctlControllerFactory
        """
        self.__factory = fcty

    @property
    def factory(self):
        return self.__factory


class WfctlInstanceController(WfctlControllerBase,
                              WctlInstanceControllerDefaultXPointMixin,
                              WctlInstanceControllerDefaultDeleteMixin,
                              WctlInstanceControllerP):
    def __init__(self, pctx, fcty):
        """
        :param pctx: Project context
        :type pctx: WfmodStatemachineProjectContextP
        :param fcty: Controller factory
        :type fcty: WfctlControllerFactory
        """
        super().__init__(fcty)

        self.__pctx = pctx

    def canCreateConnectionPoint(self, instance, **kwargs):
        return False

    def createConnectionPoint(self, instance, valuePrv, **kwargs):
        raise Exception('Cannot manually create connection points')

    def canRemoveConnectionPoint(self, instance, xpoint):
        return False

    def removeConnectionPoint(self, instance, xpoint, block):
        raise Exception('Cannot manually remove connection points')

    def deleteItem(self, item, block):
        if wfhdb_is_xbt_state(item):
            return _wfctl_deleteXbt(item, self.__pctx)
        return WctlInstanceControllerDefaultDeleteMixin.deleteItem(
            self, item, block)


class WfctlConnectionController(WfctlControllerBase,
                                WctlConnectionControllerDefaultCreateMixin,
                                WctlConnectionControllerDefaultDeleteMixin,
                                WctlConnectionControllerP):
    def __init__(self, pctx, fcty):
        """
        :param pctx: Project context
        :type pctx: WfmodStatemachineProjectContextP
        :param fcty: Controller factory
        :type fcty: WfctlControllerFactory
        """
        super().__init__(fcty)

        self.__pctx = pctx

    def canStartConnectionAt(self, block, instance, xpoint=None):
        return instance and not wfhdb_is_xbt_state(instance)

    def canCreateConnection(self, block, sinst, einst, sxpnt=None, expnt=None,
                            **kwargs):
        return sinst and einst and \
            not (wfhdb_is_xbt_state(sinst) or wfhdb_is_xbt_state(einst))

    def createConnection(self, block, sinst, einst, sxpnt=None, expnt=None,
                         **kwargs):
        """
        :type block: WhdbBlock
        :type sinst: WfhdbStateInstance | WfhdbBlockInstance
        :type einst: WfhdbStateInstance | WfhdbBlockInstance
        :type sxpnt: WhdbConnectionPointP | None
        :type expnt: WhdbConnectionPointP | None
        """
        assert self.canCreateConnection(
            block, sinst, einst, sxpnt, expnt, **kwargs)

        scon = sinst.createConnectionPoint(cdir=WDBS_XPNT_OUT)
        econ = einst.createConnectionPoint(cdir=WDBS_XPNT_IN)

        trns = [t for t in block.getConnections()
                if (t.getStartObject() is sinst and t.getEndObject() is einst)
                or (t.getStartObject() is einst and t.getEndObject() is sinst)]
        """:type: list[WfhdbTransition]"""

        connection = block.createConnection(sinst, scon, einst, econ)
        """:type: WfhdbTransition"""

        if trns:
            if sinst is einst:
                pths = [t.getPath() for t in trns]
                """:type: list[WhdbCircularConnectionPath]"""
                if pths:
                    maxRad = max(p.getRadius() for p in pths)
                    latAng = pths[-1].getAngle()
                    path = connection.getPath()
                    """:type: WhdbCircularConnectionPath"""
                    path.setRadius(maxRad + 5)
                    path.setAngle(latAng)
            else:
                # noinspection PyUnresolvedReferences
                wdths = [t.getPath().getArcWidth() *
                         (1 if t.getStartObject() is sinst else -1)
                         for t in trns]
                maxWdth = max(wdths)
                minWdth = min(wdths)
                path = connection.getPath()
                """:type: WhdbArcConnectionPath"""
                if abs(minWdth) > abs(maxWdth):
                    path.setArcWidth(maxWdth + 0.2)
                else:
                    path.setArcWidth(minWdth - 0.2)

        return connection, (sinst, einst)

    def deleteItem(self, item, block):
        """
        :type item: WfhdbTransition
        :type block: WfhdbBlock
        """
        if wfhdb_is_xbt_transition(item):
            return _wfctl_deleteXbt(item, self.__pctx)

        sobj = item.getStartObject()
        eobj = item.getEndObject()

        block.remove(item)

        sobj.removeConnectionPoint(item.getStartConnectionPoint())
        eobj.removeConnectionPoint(item.getEndConnectionPoint())

        return (item,), (sobj, eobj)


def _wfctl_deleteXbt(stateOrTrans, pctx):
    """
    Delete all XBT-elements related to the given proxy or transitino.

    :param stateOrTrans: Proxy to delete XBT for
    :type stateOrTrans: WfhdbTransition | WfhdbStateP
    :param pctx: Project context
    :type pctx: WfmodStatemachineProjectContextP
    """
    xmng = pctx.getXbtManager()
    tblk = pctx.getProjectDB().getTopHierarchy()
    """:type: WfhdbTopBlock"""
    sobj, strn, sprx, eprx, etrn, eobj, spth, epth = \
        xmng.getXbtElements(stateOrTrans)

    for path, (prxy, trns) in [(spth, (sprx, strn)), (epth, (eprx, etrn))]:
        blck = path[-1].getBlock() if path else tblk
        """:type: WhdbBlock"""
        tsob = trns.getStartObject()
        tsob.removeConnectionPoint(trns.getStartConnectionPoint())
        teob = trns.getEndObject()
        teob.removeConnectionPoint(trns.getEndConnectionPoint())

        blck.remove(trns)
        blck.remove(prxy)

    return [sprx, strn, eprx, etrn], [sobj, eobj]


class WfctlControllerFactory(WctlSimpleControllerFactoryBase):
    def __init__(self, pctx):
        """
        :param pctx: Project context
        :type pctx: WfmodStatemachineProjectContextP
        """
        super().__init__(
            instCtlr=WfctlInstanceController(pctx, self),
            connCtlr=WfctlConnectionController(pctx, self))
