# ----------------------------------------------------------------------------
# - wfhdb-block                                                              -
# - wild cruncher:fsm block                         core / statemachine / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfhdb_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher silkan state machine block

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.db.whdb_block import WhdbBlock
from wcr_core.db.whdb_protocols import WHDB_POSITION_KEY, WHDB_Z_VALUE_KEY, \
    WHDB_DIMENSION_KEY, WHDB_VFLIP_KEY, WHDB_HFLIP_KEY, \
    WHDB_ROTATION_KEY
from wcr_core.db.whdb_connection_point import WHDB_DEFAULT_XPOINT_CODE
from wcr_core.db_api.whdb_block_api import WhdbBlockP
from wcr_core.dbstore.wdbs_meta_description import *
from wcr_core.dbstore.wdbs_meta_type import WDBS_STRING_META, \
    WdbsOrderedDictionaryMetaType, WdbsRecordMetaType
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.statemachine.db.wfhdb_operator import WFHDB_OPER
from wcr_core.modules.statemachine.db.wfhdb_protocols import WfhdbHasSourceP, \
    WfhdbHasSourceMixin
from wcr_core.modules.statemachine.db.wfhdb_datamodel_support import \
    WfhdbDatamodelSupport, WFHDB_DATAMODEL_TYPE_KEY, WFHDB_DATAMODEL_KEY


WFHDB_NESTED_BLOCK_CODE = 'wcr.statemachine.block-nested'

WFHDB_FINAL_BLOCK_CODE = 'wcr.statemachine.block-final'

WFHDB_PARALLEL_BLOCK_CODE = 'wcr.statemachine.block-parallel'

WFHDB_SCRIPT_KEY = 'script'

WFHDB_BINDING_TYPE_KEY = 'binding-type'

# noinspection PyAbstractClass
class WfhdbBlockP(WhdbBlockP, WfhdbHasSourceP):
    """
    Abstract base class for state machine blocks.
    """

    @abstractmethod
    def createNestedBlockInstance(self, *args, **kwargs):
        """
        Create a new nested block instance.

        For arguments, see `WhdbBlock.createBlockInstance()`

        :param args: Arguments
        :param kwargs: Keywork-arguments
        :return: Newly created nested block instance
        :rtype: WfhdbBlockInstance
        """
        raise NotImplementedError()

    @abstractmethod
    def createParallelBlockInstance(self, *args, **kwargs):
        """
        Create a new parallel block instance.

        For arguments, see `WhdbBlock.createBlockInstance()`

        :param args: Arguments
        :param kwargs: Keywork-arguments
        :return: Newly created parallel block instance
        :rtype: WfhdbBlockInstance
        """
        raise NotImplementedError()


class WfhdbBlockBase(WhdbBlock, WfhdbBlockP, WfhdbHasSourceMixin):
    """
    Base class for state machine blocks.
    """
    META = WdbsMetaDescription(
        WhdbBlock.META,
        WfhdbHasSourceMixin.META)

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbBlock.__init__(self, block, wreg)

    def createBlockInstance(self, *args, **kwargs):
        btyp = kwargs.get('btyp')
        if not btyp:
            raise ValueError('Cannot create blocks without block type')
        if btyp not in [WFHDB_NESTED_BLOCK_CODE,
                        WFHDB_PARALLEL_BLOCK_CODE]:
            raise ValueError('Invalid block type; {}'.format(btyp))
        return super().createBlockInstance(*args, **kwargs)

    def _createRawOperatorInstance(self, oper, *args, **kwargs):
        if isinstance(oper, str):
            oper = WFHDB_OPER[oper]
        return wfhdb_createOperatorInstance(
            self.rawElement, oper, *args, **kwargs)

    def createProxy(self):
        from .wfhdb_block_proxy import WfhdbBlockProxy

        return WfhdbBlockProxy(self)

    def createNestedBlockInstance(self, *args, **kwargs):
        kwargs['btyp'] = WFHDB_NESTED_BLOCK_CODE

        return super().createBlockInstance(*args, **kwargs)

    def createParallelBlockInstance(self, *args, **kwargs):
        kwargs['btyp'] = WFHDB_PARALLEL_BLOCK_CODE

        return super().createBlockInstance(*args, **kwargs)


class WfhdbTopBlock(WfhdbDatamodelSupport, WfhdbBlockBase):
    META = WdbsMetaDescription(
        WfhdbBlockBase.META,
        WfhdbDatamodelSupport.META,
        WdbsMetaField(WFHDB_SCRIPT_KEY, WDBS_STRING_META),
        WdbsMetaField(WFHDB_DATAMODEL_TYPE_KEY, WDBS_STRING_META),
        WdbsMetaField(WFHDB_BINDING_TYPE_KEY, WDBS_STRING_META))

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WfhdbBlockBase.__init__(self, block, wreg)

    def flatten(self):
        pass

    def getTypeName(self):
        return 'top-level'

    def getScript(self):
        """
        Get the script for the top-level block.

        :return: Script text
        :rtype: str
        """
        return self.rawElement.getMeta(WFHDB_SCRIPT_KEY)

    def setScript(self, script):
        """
        Set the script for the top-level block.

        :param script: Script text
        :type script: str
        """
        self.rawElement.setMeta(script, WFHDB_SCRIPT_KEY)

    def getBindingType(self):
        """
        Get the binding type (early, late) for the state machine.

        :return: Binding type
        :rtype: str
        """
        return self.rawElement.getMeta(WFHDB_BINDING_TYPE_KEY)

    def setBindingType(self, bindingType):
        """
        Set the binding type (early, late) for the state machine.

        :param bindingType: Binding type
        :type bindingType: str
        """
        self.rawElement.setMeta(bindingType, WFHDB_BINDING_TYPE_KEY)

    def getPortNames(self):
        dmdl = self.rawElement.getMeta(WFHDB_DATAMODEL_KEY)
        # noinspection PyUnresolvedReferences
        return {WDBS_XPNT_BI: [(n, n) for n in dmdl.keys()]}


class WfhdbNestedBlock(WfhdbDatamodelSupport, WfhdbBlockBase):
    META = WdbsMetaDescription(
        WfhdbBlockBase.META,
        WfhdbDatamodelSupport.META)

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WfhdbBlockBase.__init__(self, block, wreg)

    def getTypeName(self):
        return 'nested-state'


class WfhdbParallelBlock(WfhdbDatamodelSupport, WfhdbBlockBase):
    META = WdbsMetaDescription(
        WfhdbBlockBase.META,
        WfhdbDatamodelSupport.META)

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WfhdbBlockBase.__init__(self, block, wreg)

    def getTypeName(self):
        return 'parallel'


def wfhdb_createOperatorInstance(block, oper, name=None, info=None,
                                 ipos=None, zpos=None, width=None,
                                 height=None, rotn=None, vflp=None,
                                 hflp=None, prms=None, **kwargs):
    """
    :param block: Block to create function in
    :type block: WdbsBlockP
    :param oper: Operator code
    :type oper: WhdbOperatorP
    :param name: Name of the instance
    :type name: str
    :param info: Info string for the instance
    :type info: str
    :param ipos: Position
    :type ipos: (float, float)
    :param zpos: Z-value
    :type zpos: int
    :param width: Width
    :type width: float
    :param height: Height
    :type height: float
    :param rotn: Rotation
    :type rotn: float
    :param vflp: Vertically flipped
    :type vflp: bool
    :param hflp: Horizontally flipped
    :type hflp: bool
    :param prms: Parameter values
    :type prms: Iterable[(str, object)]
    :return: The newly created operator instance
    :rtype: WdbsOperatorInstanceP
    """
    if width is not None and height is not None:
        dim = (width, height)
    else:
        dim = None
    fclst = [
        (WHDB_DEFAULT_XPOINT_CODE, None, xdsc.kwargs)
        for xdsc in oper.createInitialConnectionPoints()]

    kwargs.update({
        WHDB_POSITION_KEY: ipos or (0, 0),
        WHDB_Z_VALUE_KEY: zpos,
        WHDB_DIMENSION_KEY: dim,
        WHDB_ROTATION_KEY: rotn or 0,
        WHDB_HFLIP_KEY: hflp or False,
        WHDB_VFLIP_KEY: vflp or False})

    return block.createOperatorInstance(
        oper.getCode(), name, info, fclst, prms, **kwargs)
