# ----------------------------------------------------------------------------
# - wfhdb-instance                                                           -
# - wild cruncher:fsm instances                     core / statemachine / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfhdb_instance
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher silkan state machine instances

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db.whdb_instance import WhdbOperatorInstance, \
    WhdbBlockInstance, WHDB_DEFAULT_ICON_SIZE
from wcr_core.db.whdb_connection_point import WHDB_DEFAULT_XPOINT_CODE
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP
from wcr_core.db_api.whdb_protocols_api import WhdbNamedEntityP
from wcr_core.modules.statemachine.db.wfhdb_protocols import *
from wcr_core.modules.statemachine.db.wfhdb_standard_operators import \
    WFHDB_STATE, WFHDB_XBT_TARGET_PROXY, WFHDB_XBT_SOURCE_PROXY

from wcr_core.modules.statemachine.db.wfhdb_datamodel_support import WfhdbDatamodelSupport

# noinspection PyAbstractClass
class WfhdbInstanceP(WhdbInstanceP, WfhdbStateP, WfhdbMayBeInitialP,
                     WfhdbMayBeFinalP, WfhdbExecutableP, WhdbNamedEntityP,
                     WfhdbMayBeHistoryP):
    """
    Protocol class for state-machine instance objects.
    """
    pass


# noinspection PyAbstractClass
class WfhdbInstanceMixin(WfhdbInstanceP, WfhdbStateMixin,
                         WfhdbMayBeInitialMixin, WfhdbMayBeFinalMixin,
                         WfhdbExecutableMixin, WfhdbDatamodelSupport,
                         WfhdbMayBeHistoryMixin):
    META = WdbsMetaDescription(
        WfhdbStateMixin.META,
        WfhdbMayBeInitialMixin.META,
        WfhdbMayBeFinalMixin.META,
        WfhdbDatamodelSupport.META,
        WfhdbMayBeHistoryMixin.META,
        WfhdbExecutableMixin.META)


class WfhdbStateInstance(WhdbOperatorInstance, WfhdbInstanceMixin,
                         WfhdbHasSourceMixin):
    """
    Operator-instance for atomic states.
    """

    META = WdbsMetaDescription(
        WhdbOperatorInstance.META,
        WfhdbInstanceMixin.META,
        WfhdbHasSourceMixin.META)

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbOperatorInstance.__init__(self, WFHDB_STATE, inst, wreg)

    def getDefaultDimension(self):
        return WHDB_DEFAULT_ICON_SIZE, WHDB_DEFAULT_ICON_SIZE

    def createConnectionPoint(self, ctyp=None, **kwargs):
        return super().createConnectionPoint(
            WHDB_DEFAULT_XPOINT_CODE, **kwargs)


WFHDB_XBLOCK_TRANS_ID = 'trans-id'

WFHDB_XBLOCK_SISTER_UUID = 'sister-uuid'


class _WfhdbTransInstanceBase(WhdbOperatorInstance, WfhdbStateMixin):
    """
    Base-class for output-/input-states operator-instances.
    """

    META = WdbsMetaDescription(
        WhdbOperatorInstance.META,
        WfhdbStateMixin.META,
        WdbsMetaField(WFHDB_XBLOCK_TRANS_ID, WDBS_STRING_META),
        WdbsMetaField(WFHDB_XBLOCK_SISTER_UUID, WDBS_UUID_META))

    def __init__(self, oper, inst, wreg):
        """
        :param oper: Operator
        :type oper: WhdbOperatorP
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbOperatorInstance.__init__(self, oper, inst, wreg)

    def createConnectionPoint(self, **kwargs):
        raise Exception('Cannot create new connection points on ' +
                        self.__class__.__name__)

    def getTransId(self):
        """
        Get the transition-ID identifying both source and target proxy.

        :return: Transition-ID
        :rtype: str
        """
        return self.rawElement.getMeta(WFHDB_XBLOCK_TRANS_ID)

    def getSisterUUID(self):
        """
        Get the UUID of the sister proxy.

        :return: UUID of sister proxy
        :rtype: UUID
        """
        return self.rawElement.getMeta(WFHDB_XBLOCK_SISTER_UUID)

    def setSisterUUID(self, uuid):
        """
        Set the UUID of the sister proxy.

        :param uuid: UUID of sister proxy
        :type uuid: UUID
        """
        self.rawElement.setMeta(uuid, WFHDB_XBLOCK_SISTER_UUID)

    def getDefaultDimension(self):
        return WHDB_DEFAULT_ICON_SIZE, WHDB_DEFAULT_ICON_SIZE


class WfhdbXbtTargetProxyInstance(_WfhdbTransInstanceBase):
    """
    Operator-instance for output-states.
    """

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbOperatorInstance.__init__(
            self, WFHDB_XBT_TARGET_PROXY, inst, wreg)


class WfhdbXbtSourceProxyInstance(_WfhdbTransInstanceBase):
    """
    Operator-instance for output-states.
    """

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbOperatorInstance.__init__(
            self, WFHDB_XBT_SOURCE_PROXY, inst, wreg)


class WfhdbBlockInstance(WhdbBlockInstance, WfhdbInstanceMixin):
    """
    Block-instance for non-atomic and parallel states.
    """

    META = WdbsMetaDescription(
        WhdbBlockInstance.META,
        WfhdbInstanceMixin.META)

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsBlockInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        super().__init__(inst, wreg)

    def getDefaultDimension(self):
        return WHDB_DEFAULT_ICON_SIZE, WHDB_DEFAULT_ICON_SIZE

    def createConnectionPoint(self, ctyp=None, **kwargs):
        return super().createConnectionPoint(
            WHDB_DEFAULT_XPOINT_CODE, **kwargs)
