# ----------------------------------------------------------------------------
# - wfhdb-transition                                                         -
# - wild cruncher:fsm transition                    core / statemachine / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfhdb_transition
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher silkan state machine transition

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db.whdb_connection import WhdbConnection
from wcr_core.db.whdb_connection_path_arc import WhdbArcConnectionPath
from wcr_core.db.whdb_connection_path_circular import \
    WhdbCircularConnectionPath
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaDescription, \
    WdbsMetaField
from wcr_core.dbstore.wdbs_meta_type import WDBS_STRING_META
from wcr_core.modules.statemachine.db.wfhdb_protocols import \
    WFHDB_EXECUTABLE_META


WFHDB_TRANS_EVENT_KEY = 'event'

WFHDB_TRANS_COND_KEY = 'cond'

WFHDB_EXECUTABLE_KEY = 'exec'


class WfhdbTransition(WhdbConnection):
    """
    State machine transition.
    """

    META = WdbsMetaDescription(
        WhdbConnection.META,
        WdbsMetaField(WFHDB_TRANS_EVENT_KEY, WDBS_STRING_META),
        WdbsMetaField(WFHDB_TRANS_COND_KEY, WDBS_STRING_META),
        WdbsMetaField(WFHDB_EXECUTABLE_KEY, WFHDB_EXECUTABLE_META))

    def __init__(self, conn, wreg):
        """
        :param conn: Raw connection
        :type conn: WdbsConnectionP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        c = conn.getStartObject() is conn.getEndObject()
        pathF = WhdbCircularConnectionPath if c else WhdbArcConnectionPath
        super().__init__(conn, wreg, pathF)

    def getTriggerEvent(self):
        """
        Get the trigger event for the transition, if any.

        :return: Trigger event
        :rtype: str | None
        """
        evnt = self.rawElement.getMeta(WFHDB_TRANS_EVENT_KEY)
        return evnt.strip() if evnt is not None else None

    def setTriggerEvent(self, event):
        """
        Set the trigger event for the transition.

        :param event: Trigger event
        :type event: str | None
        """
        self.rawElement.setMeta(event or None, WFHDB_TRANS_EVENT_KEY)

    def getTriggerCondition(self):
        """
        Get the trigger condition for the transition, if any.

        :return: Trigger condition
        :rtype: str | None
        """
        cond = self.rawElement.getMeta(WFHDB_TRANS_COND_KEY)
        return cond.strip() if cond is not None else None

    def setTriggerCondition(self, cond):
        """
        Set the trigger condition for the transition.

        :param cond: Trigger condition
        :type cond: str | None
        """
        self.rawElement.setMeta(cond or None, WFHDB_TRANS_COND_KEY)

    def getExecutable(self):
        """
        Get the executable text for the transition.

        :return: Executable text
        :rtype: list[(str, object)] | None
        """
        return self.rawElement.getMeta(WFHDB_EXECUTABLE_KEY)

    def setExecutable(self, excbl):
        """
        Set the executable text for the transition.

        :param excbl: Executable text
        :type excbl: list[(str, object)] | None
        """
        self.rawElement.setMeta(excbl or None, WFHDB_EXECUTABLE_KEY)

    def __repr__(self):
        sobj = self.getStartObject()
        eobj = self.getEndObject()

        return '{} --> {}'.format(
            sobj.getName() if sobj else 'MISSING',
            eobj.getName() if eobj else 'MISSING')
