# ----------------------------------------------------------------------------
# - wfhdb-verifier                                                           -
# - wild cruncher: database verifier                core / statemachine / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfhdb_verifier
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Database verifier

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.common.wcmn_verifier_module import WcmnHierarchyVerificationError
from wcr_core.db.whdb_verifier import WhdbHierarchyVerifier, \
    WhdbVerifierModuleP
from wcr_core.modules.statemachine.db.wfhdb_block import WfhdbParallelBlock, \
    WfhdbTopBlock, WfhdbNestedBlock
from wcr_core.modules.statemachine.db.wfhdb_instance import WfhdbBlockInstance
from wcr_core.modules.statemachine.db.wfhdb_protocols import \
    WfhdbMayBeInitialP, WfhdbMayBeFinalP
from wcr_core.modules.statemachine.db.wfhdb_xbt_utils import \
    wfhdb_is_xbt_state, WfhdbXbtException


class WfhdbNoDanglingXbtVerifierModule(WhdbVerifierModuleP):
    """
    Verify that all XBTs have both proxies.
    """

    ERROR_ID = 'hierarchy.simulation.dangling-xbt'

    ERROR_MESSAGE = 'Dangling cross-block transition'

    DESCRIPTION = 'For every cross-block transitions, both a source and a ' \
                  'target proxy must exists. The dangling proxy/proxies ' \
                  'should be deleted.'

    def __init__(self, pctx):
        """
        :param pctx: Project context
        :type pctx: WfmodStatemachineProjectContextP
        """
        self.__xmng = pctx.getXbtManager()

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        if wfhdb_is_xbt_state(instance):
            try:
                sister, _ = self.__xmng.getSisterState(instance)
                return []
            except WfhdbXbtException:
                return [WcmnHierarchyVerificationError(
                    self.ERROR_ID,
                    self.ERROR_MESSAGE,
                    [instance], self)]
        return []

    def description(self):
        return WfhdbNoDanglingXbtVerifierModule.DESCRIPTION


class WfhdbXbtSameTransitionIdVerifierModule(WhdbVerifierModuleP):
    """
    Verify that all XBTs proxies have the correct transition ID.
    """

    ERROR_ID = 'hierarchy.simulation.wrong-xbt-trans-id'

    ERROR_MESSAGE = 'Cross-block transition-ID mismatch'

    DESCRIPTION = 'The two corresponding proxies of a cross-block ' \
                  'transition should have the same transition-ID.'

    def __init__(self, pctx):
        """
        :param pctx: Project context
        :type pctx: WfmodStatemachineProjectContextP
        """
        self.__xmng = pctx.getXbtManager()

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        if wfhdb_is_xbt_state(instance):
            try:
                sister, _ = self.__xmng.getSisterState(instance)
                # noinspection PyUnresolvedReferences
                if instance.getTransId() != sister.getTransId():
                    return [WcmnHierarchyVerificationError(
                        self.ERROR_ID,
                        self.ERROR_MESSAGE,
                        [instance], self)]
                return []
            except WfhdbXbtException:
                pass
        return []

    def description(self):
        return WfhdbXbtSameTransitionIdVerifierModule.DESCRIPTION


class WfhdbInitialStateVerifierModule(WhdbVerifierModuleP):
    """
    Verify that all nested states have exactly one initial state.
    """

    ERROR_ID = 'hierarchy.simulation.initial-state'

    ERROR_MESSAGE = 'Nested state without/with multiple initial state(s)'

    ERROR_MESSAGE_TOP = 'Top-level without/with multiple initial state(s)'

    DESCRIPTION = 'Every nested state, including the top-level, must have ' \
                  'exactly one initial state.'

    def verifyBlock(self, block, topLevel=False):
        """
        :type block: WfhdbBlockP
        """
        if not topLevel or not isinstance(block, WfhdbTopBlock):
            return

        initials = self._getInitials(block)

        if len(initials) == 1 or len(list(block.getInstances())) == 0:
            return []
        else:
            return [
                WcmnHierarchyVerificationError(
                    self.ERROR_ID, self.ERROR_MESSAGE_TOP,
                    initials if initials else [block], self)]

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        if not isinstance(instance, WfhdbBlockInstance):
            return
        block = instance.getBlock()
        if not isinstance(block, WfhdbNestedBlock):
            return

        initials = self._getInitials(block)

        if len(initials) == 1 or len(list(block.getInstances())) == 0:
            return []
        else:
            return [
                WcmnHierarchyVerificationError(
                    self.ERROR_ID, self.ERROR_MESSAGE,
                    initials if initials else [instance], self)]

    def description(self):
        return WfhdbInitialStateVerifierModule.DESCRIPTION

    def _getInitials(self, block):
        candidates = (s for s in block.getInstances()
                      if isinstance(s, WfhdbMayBeInitialP))
        """:type: Iterable[WfhdbMayBeInitialP | WhdbBaseP]"""
        return [s for s in candidates if s.isInitialState()]


class WfhdbNoTransFromFinalStateVerifierModule(WhdbVerifierModuleP):
    """
    Verify that no transitions originate from a final state.
    """

    ERROR_ID = 'hierarchy.simulation.trans-from-final'

    ERROR_MESSAGE = 'Transition from final state'

    DESCRIPTION = 'Transitions may not originate from a final state.'

    def verifyBlock(self, block, topLevel=False):
        pass

    def verifyConnection(self, transition):
        """
        :type transition: WfhdbTransition
        """
        sobj = transition.getStartObject()
        if not isinstance(sobj, WfhdbMayBeFinalP):
            return None
        if sobj.isFinalState():
            return [WcmnHierarchyVerificationError(
                self.ERROR_ID, self.ERROR_MESSAGE, [transition], self)]
        else:
            return []

    def verifyInstance(self, instance):
        pass

    def description(self):
        return WfhdbNoTransFromFinalStateVerifierModule.DESCRIPTION


class WfhdbParallelTransVerifierModule(WhdbVerifierModuleP):
    """
    Verify that immediate children of parallel states have no transitions.
    """

    ERROR_ID = 'hierarchy.simulation.parallel-trans'

    ERROR_MESSAGE = 'Transition from/to immediate child of parallel state'

    DESCRIPTION = 'Immediate children of parallel states may not be ' \
                  'involved in transitions.'

    def verifyBlock(self, block, topLevel=False):
        """
        :type block: WfhdbBlockP
        """
        if not isinstance(block, WfhdbParallelBlock):
            return None

        return [
            WcmnHierarchyVerificationError(
                self.ERROR_ID, self.ERROR_MESSAGE,
                [t, t.getStartObject(), t.getEndObject()], self)
            for t in block.getConnections()]

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        pass

    def description(self):
        return WfhdbParallelTransVerifierModule.DESCRIPTION


class WfhdbHierarchyVerifier(WhdbHierarchyVerifier):
    """
    State machine hierarchy verifier.
    """

    def __init__(self, pctx, modules=None):
        """
        :param pctx: Project context
        :type pctx: WfmodStatemachineProjectContextP
        :param modules: Additional verification modules to use.
        :type modules: Iterable[WhdbVerifierModuleP] | None
        """
        mods = [WfhdbNoDanglingXbtVerifierModule(pctx),
                WfhdbXbtSameTransitionIdVerifierModule(pctx),
                WfhdbInitialStateVerifierModule(),
                WfhdbNoTransFromFinalStateVerifierModule()]
        if modules:
            mods.extend(modules)

        super().__init__(mods)
