# ----------------------------------------------------------------------------
# - wfhdb-xbt-manager                                                        -
# - wild cruncher: xbt manager   core / statemachine / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfhdb_xbt_manager
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - XBT manager

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db.whdb_connection import whdb_getOtherObject
from wcr_core.db.whdb_path_manager import WhdbPathManager
from wcr_core.db.whdb_tree_utils import whdb_scan_tree
from wcr_core.dbstore_api.wdbs_instance_api import WdbsOperatorInstanceP
from wcr_core.modules.statemachine.db.wfhdb_instance import \
    WfhdbXbtTargetProxyInstance, WFHDB_XBLOCK_TRANS_ID
from wcr_core.modules.statemachine.db.wfhdb_standard_operators import \
    WFHDB_XBT_PROXY_OP_CODES
from wcr_core.modules.statemachine.db.wfhdb_transition import WfhdbTransition
from wcr_core.modules.statemachine.db.wfhdb_xbt_utils import \
    WfhdbXbtException, wfhdb_is_xbt_state


class WfhdbXbtManager(object):
    """
    Manager for XBT elements.
    """

    #: Raw predicate to test for XBT state in tree traversals
    _XBT_PREDICATE = lambda x: (isinstance(x, WdbsOperatorInstanceP) and
                                x.getOperator() in WFHDB_XBT_PROXY_OP_CODES)

    def __init__(self, tblk):
        """
        :param tblk: Top-level block
        :type tblk: WfhdbTopBlock
        """
        #: Top-level block
        self.__topBlock = tblk

        #: Path manager
        self.__pathMngr = WhdbPathManager(tblk)

        self._scanForXbts()

    def _scanForXbts(self):
        xbts = whdb_scan_tree(
            WfhdbXbtManager._XBT_PREDICATE, self.__topBlock)
        """:type: Iterable[(WhdbBaseP, tuple[WhdbBlockInstanceP])]"""
        for xbt, path in xbts:
            self.__pathMngr.addPath(xbt, path)

    def removeXbtState(self, state):
        """
        Mark an XBT-state as removed.

        :param state: XBT-state
        :type state: WfhdbXbtSourceProxyInstance | WfhdbXbtTargetProxyInstance
        """
        self.__pathMngr.removePath(state.getUUID())

    def _scanForSisterByTransID(self, state):
        """
        Scan the project for the sister-state to the given state by its
        transaction id.

        Update the state and the path manager if found.

        :param state: XBT-state
        :type state: WfhdbXbtSourceProxyInstance | WfhdbXbtTargetProxyInstance
        :return: Sister-state and path, if any
        :rtype: (WhdbBaseP, list[WhdbBlockInstanceP])
        """
        trid = state.getTransId()
        if not trid:
            raise WfhdbXbtException(
                'Invalid Cross-Block Proxy', 'Missing transition ID')

        uuid = state.getUUID()
        sister, path = next(
            whdb_scan_tree(
                lambda x: (x.getUUID() != uuid and
                           WfhdbXbtManager._XBT_PREDICATE(x) and
                           x.getMeta(WFHDB_XBLOCK_TRANS_ID) == trid),
                self.__topBlock),
            (None, None))
        if not sister:
            raise WfhdbXbtException(
                'Invalid Cross-Block Proxy',
                'Missing counter-part proxy for transition ID {}'.format(trid))

        state.setSisterUUID(sister.getUUID)
        self.__pathMngr.addPath(sister, path)
        return sister, path

    def getSisterState(self, state):
        """
        Get the sister-state for the given XBT-state.

        :param state: XBT-state
        :type state: WfhdbXbtSourceProxyInstance | WfhdbXbtTargetProxyInstance
        :return: Sister-state and path, if any
        :rtype: (WhdbBaseP, list[WhdbBlockInstanceP])
        """
        uuid = state.getSisterUUID()
        if not uuid:
            return self._scanForSisterByTransID(state)
        rslt = self.__pathMngr.getEntityPath(uuid)
        if rslt is WhdbPathManager.DELETED:
            raise WfhdbXbtException(
                'Invalid Cross-Block Proxy',
                'Missing counter-part proxy for state {}'.format(
                    state.getName()))
        if rslt[0] is None:
            return self._scanForSisterByTransID(state)
        return rslt

    def getXbtElements(self, stateOrTrans):
        """
        Get all XBT-elements related to the given XBT element.

        :param stateOrTrans: XBT element
        :type stateOrTrans: WfhdbTransition | WfhdbStateP
        :return: XBT-elements related to the given XBT element
        :rtype: (WhdbInstanceP, WfhdbTransition, WfhdbXbtTargetProxyInstance,
                 WfhdbXbtSourceProxyInstance, WfhdbTransition, WhdbInstanceP,
                 list[WhdbBlockInstanceP], list[WhdbBlockInstanceP])
        """
        # noinspection PyShadowingNames
        def getTransition(path, state):
            blka = path[-1].getBlock() if path else self.__topBlock
            trna = list(blka.getConnectionsForObject(state))
            if len(trna) != 1:
                raise WfhdbXbtException(
                    'Invalid Cross-Block Proxy',
                    'Missing or multiple transitions')
            return trna[0]

        trans = None

        if isinstance(stateOrTrans, WfhdbTransition):
            if wfhdb_is_xbt_state(stateOrTrans.getStartObject()):
                state = stateOrTrans.getStartObject()
            elif wfhdb_is_xbt_state(stateOrTrans.getEndObject()):
                state = stateOrTrans.getEndObject()
            else:
                raise ValueError('Transition does not involve an XBT state')
            trans = stateOrTrans
        elif wfhdb_is_xbt_state(stateOrTrans):
            state = stateOrTrans
        else:
            raise ValueError(
                'Entity {} is not an XBT entity'.format(stateOrTrans))

        _, path = self.__pathMngr.getEntityPath(state.getUUID())
        if path is None:
            raise WfhdbXbtException(
                'Invalid Cross-Block Proxy',
                'Cannot find path of entity with UUID {}'.format(
                    state.getUUID()))

        # noinspection PyTypeChecker
        sisState, sisPath = self.getSisterState(state)

        if not trans:
            trans = getTransition(path, state)
        sisTrans = getTransition(sisPath, sisState)

        if isinstance(state, WfhdbXbtTargetProxyInstance):
            sobj, _ = whdb_getOtherObject(trans, state)
            # noinspection PyTypeChecker
            eobj, _ = whdb_getOtherObject(sisTrans, sisState)
            return sobj, trans, state, sisState, sisTrans, eobj, path, sisPath
        else:
            # noinspection PyTypeChecker
            sobj, _ = whdb_getOtherObject(sisTrans, sisState)
            eobj, _ = whdb_getOtherObject(trans, state)
            return sobj, sisTrans, sisState, state, trans, eobj, sisPath, path

    def getXbtTransition(self, stateOrTrans):
        """
        Get the property-carrying transition for the given XBT element.

        :param stateOrTrans: XBT element
        :type stateOrTrans: WfhdbTransition | WfhdbStateP
        :return: Property-carrying transition for the given XBT element
        :rtype: WfhdbTransition
        """
        if isinstance(stateOrTrans, WfhdbTransition):
            if isinstance(stateOrTrans.getEndObject(),
                          WfhdbXbtTargetProxyInstance):
                return stateOrTrans
        return self.getXbtElements(stateOrTrans)[1]
