# ----------------------------------------------------------------------------
# - wfmod-event-manager                                                      -
# - wild cruncher: event manager                         core / statemachine -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfmod_event_manager
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher state machine event manager

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import defaultdict

from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP
from wcr_core.modules.statemachine.db.wfhdb_protocols import \
    WFHDB_EXEC_RAISE_KEY, WFHDB_EXEC_IF_KEY, \
    WFHDB_EXEC_INVOKE_KEY, WFHDB_EXEC_INVOKE_FINALIZE_KEY, \
    WFHDB_EXEC_SEND_EVENT_KEY, WFHDB_EXEC_SEND_KEY, \
    WFHDB_EXEC_IF_EXECS_KEY, WfhdbExecutableP


_WFMOD_WRITE_EVENT_IDX = 0

_WFMOD_READ_EVENT_IDX = 1


class WfmodEventManager(object):
    """
    Event manager maintaining a list of all events used in a project and
    of which items the events are used by.
    """

    def __init__(self, topBlock):
        """
        :param topBlock: Top-level block
        :type topBlock: WfhdbTopBlock
        """
        #: Top-level block
        self.__topBlock = topBlock

        #: Events
        self.__events = {}
        """:type: dict[str, (set[str], set[str])]"""

        self.refresh()

    def refresh(self):
        """
        Refresh the stored event information.
        """

        def searchExecs(item, execs):
            if not execs:
                return
            for code, data in execs:
                if code == WFHDB_EXEC_RAISE_KEY:
                    if data:
                        events[data][_WFMOD_WRITE_EVENT_IDX].add(item)
                elif code == WFHDB_EXEC_SEND_KEY:
                    e = data.get(WFHDB_EXEC_SEND_EVENT_KEY) if data else None
                    if e:
                        events[e][_WFMOD_WRITE_EVENT_IDX].add(item)
                elif code == WFHDB_EXEC_IF_KEY:
                    for brnch in data:
                        searchExecs(item, brnch.get(WFHDB_EXEC_IF_EXECS_KEY))
                elif code == WFHDB_EXEC_INVOKE_KEY:
                    searchExecs(
                        item, data.get(WFHDB_EXEC_INVOKE_FINALIZE_KEY))

        def walk(block, path=None):
            """
            :type block: WfhdbBlockBase
            """
            cnns = block.getConnections()
            """:type: Iterable[WfhdbTransition]"""
            for conn in cnns:
                tpl = (conn, path)
                e = conn.getTriggerEvent()
                if e:
                    events[e][_WFMOD_READ_EVENT_IDX].add(tpl)
                searchExecs(tpl, conn.getExecutable())
            insts = block.getInstances()
            """:type: Iterable[WfhdbInstanceP]"""
            for inst in insts:
                tpl = (inst, path)
                if isinstance(inst, WfhdbExecutableP):
                    searchExecs(tpl, inst.getOnEntryExec())
                    searchExecs(tpl, inst.getOnExitExec())
                if isinstance(inst, WhdbBlockInstanceP):
                    path = path + (inst,) if path else (inst,)
                    # noinspection PyTypeChecker
                    walk(inst.getBlock(), path)

        events = defaultdict(lambda *_: (set(), set()))
        """:type: dict[str, (set[(str, list)], set[(str, list])]"""
        walk(self.__topBlock)
        self.__events = events

    def getEvents(self):
        """
        Get all known events.

        :return: Known events
        :rtype: Iterable[str]
        """
        return self.__events.keys()

    def getEventUsers(self, event):
        """
        Get all items using the given event.

        :param event: Event
        :type event: str
        :return: Items using the given event, and their paths
                 (writers and readers)
        :rtype: (set[(WhdbBaseP, list)], set[(WhdbBaseP, list)])
        """
        usrs = self.__events.get(event)
        if not usrs:
            return set(), set()
        return set(usrs[0]), set(usrs[1])
