# ----------------------------------------------------------------------------
# - wfmod_layouter                                                           -
# - wild cruncher: state machine layouter                core / statemachine -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 samuel longchamps                                -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfmod_layouter
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher layouting tool

.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
import math
from collections import defaultdict, OrderedDict
from wcr_core.db.whdb_connection_path_arc import WhdbArcConnectionPath, \
    whdb_getArcEndPoints
from wcr_core.db.whdb_connection_path_circular import \
    WhdbCircularConnectionPath, WHDB_CIRC_DEFAULT_ANGLE
from wcr_core.db.whdb_instance import WHDB_DEFAULT_ICON_SIZE
from wcr_core.modules.statemachine.db.wfhdb_instance import WfhdbStateInstance
# noinspection PyProtectedMember
from wcr_core.common.wcmn_graph_layouter import _WcmnLayouterNode, \
    _WcmnLayouterEdge, _WcmnLayouterGraph


class WfmodLayouter(object):
    """
    State machine layouter
    """

    def __init__(self, layout):
        """
        :type layout: WcmnGraphLayouterP
        """
        if not layout:
            raise ValueError('Cannot instantiate Layouter: layout is None')

        self.__layout = layout

    @staticmethod
    def convertBlockToLayouterGraph(block, moreThanOneInstance=False,
                                    interconnectionsOnly=False):
        """
        Converts the block to a generic layouter graph by copying only the
        position data of the instances and recreating the links between them
        with edges between nodes. The method offers different flags to improve
        performance of some algorithms by skipping copy of useless data
        depending on the context of implementation of the algorithm.

        :param block: Single block containing the instances to which the layout
                is to be applied
        :type block: WhdbBlock
        :param moreThanOneInstance: When set to True, this flag allows the
                method to return None if there is less than
                two instances in the graph.
        :type moreThanOneInstance: bool
        :param interconnectionsOnly: When set to True, this flag allows to skip
                the copy of edges which connects nodes to themselves.
                Therefore, the returned graph would contains only edges
                linking two different nodes together.
        :type interconnectionsOnly: bool
        :return: Returns the graph object, None if it is empty (a graph with
                one instance and the moreThanOneInstance flag set to True
                would return None as well)
        :rtype: _WcmnLayouterGraph | None
        """
        if not block:
            raise ValueError('Cannot convert block to graph: block is None')

        instancesList = list(block.getInstances())
        """:type: list[WhdbInstanceP]"""

        if moreThanOneInstance and len(instancesList) <= 1:
            return None

        connectionsList = list(block.getConnections())
        """:type: list[WhdbConnectionP]"""

        nodes = OrderedDict((i, _WcmnLayouterNode(i,
                            i.isInitialState()
                            if isinstance(i, WfhdbStateInstance) else False,
                            i.isFinalState()
                            if isinstance(i, WfhdbStateInstance) else False,
                            *i.getPosition())) for i in instancesList)
        """:type: dict[WhdbInstanceP, _WcmnLayouterNode]"""
        edges = []
        """:type: list[_WcmnLayouterEdge]"""

        # Create edges list
        for con in connectionsList:
            startNode = nodes.get(con.getStartObject())
            endNode = nodes.get(con.getEndObject())
            if interconnectionsOnly and startNode is endNode:
                continue
            edges.append(_WcmnLayouterEdge(startNode, endNode))

        return _WcmnLayouterGraph(list(nodes.values()), edges)

    def applyLayout(self, destBlock):
        """
        Layout the states of the destination block and its child blocks using
        the passed generic layout implementation

        :param destBlock: Block whose elements should be laid out
        :type destBlock: WhdbBlock
        """

        if not destBlock:
            raise ValueError('Cannot apply layout: block is None')

        def recursiveLayoutStep(block):
            """
            Apply the currently defined layout to the block data and
            recursively to its child.
            Converts the location data already contained to generic type which
            actually holds the data and recopy the layouted positions of
            instances back to the instances

            :param block: Single block containing the instances to which the
                        layout is to be applied
            :type block: WhdbBlock
            """
            # Create the necessary generic graph objects
            graph = WfmodLayouter.convertBlockToLayouterGraph(
                block, True, True)
            if graph:
                try:
                    self.__layout.applyLayout(graph)

                    # Recopy positions to the instances and connections
                    for node in graph.getNodes():
                        node.getObj().setPosition(*node.getPosition())
                except Exception as e:
                    print(e)

            # Reset and reposition the connections
            for inst in block.getInstances():
                inst.resetConnectionPointPositions()
            placeTransitions(block)
            for inst in block.getBlockInstances():
                recursiveLayoutStep(inst.getBlock())

        def order(obj1, obj2):
            """
            Orders the two objects passed in term of name and ID

            :return: Tupple of the objects ordered in term of name or, if
                     equal, in term of ID and a value 1 if the order is the
                     same as the one provided, -1 if it differs. If both the
                     name and id of the objects are identical, the method swap
                     the two objects.
            :rtype: (WhdbInstanceP, WhdbInstanceP, int)
            """

            if obj1.getName() < obj2.getName():
                return obj1, obj2, 1
            elif obj1.getName() > obj2.getName():
                return obj2, obj1, -1
            elif id(obj1) < id(obj2):
                return obj1, obj2, 1
            else:
                return obj2, obj1, -1

        def placeTransitions(block):
            """
            Place new transitions for connections of the block

            :type block: WhdbBlock
            """
            cdct = defaultdict(list)
            for conn in block.getConnections():
                obj1, obj2, tdir = order(conn.getStartObject(),
                                         conn.getEndObject())
                cdct[(id(obj1), id(obj2))].append((conn, tdir))

            circConns = []
            for cnns in cdct.values():
                ccnt = len(cnns)
                for idx, (conn, tdir) in enumerate(cnns, start=0):
                    path = conn.getPath()
                    if isinstance(path, WhdbArcConnectionPath):
                        unit = min(1.0 / ccnt, 0.2)
                        path.setArcWidth(tdir * unit * (idx - (ccnt - 1) / 2))
                    elif isinstance(path, WhdbCircularConnectionPath):
                        unit = WHDB_DEFAULT_ICON_SIZE / max(ccnt, 8)
                        circConns.append((conn, WHDB_DEFAULT_ICON_SIZE / 4 +
                                                    (idx + 1) * unit))
            computedAngle = {}
            for cc in circConns:
                conn = cc[0]
                instance = conn.getStartObject()

                if instance in computedAngle:
                    optimalAngle = computedAngle.get(instance)
                else:
                    optimalAngle = computedAngle[instance] = \
                        self._getCircularConnectionAngle(
                            block.getConnections(), conn)

                circPath = conn.getPath()
                """:type: WhdbCircularConnectionPath"""
                if optimalAngle is not None:
                    circPath.setAngle(optimalAngle)

                conn.getPath().setRadius(cc[1])

        recursiveLayoutStep(destBlock)

    def _getCircularConnectionAngle(self, connectionList, circConn):
        """
        Compute the best possible angle, that is where there is more empty
        space on the instance, for the circular connection path and returns it.

        :param connectionList: List of other connections in a block
        :type connectionList: list[WhdbConnection]
        :param circConn: Circular connection to be relocated
        :type circConn: WhdbConnection
        :return: Optimal angle determined
        :rtype: float
        """
        if not circConn:
            raise ValueError('Cannot compute circular connection angle: '
                             'Connection is None')

        def conPtToAngle(interConn, start):
            """
            :param interConn: Connection containing the connection point
            :type interConn: WhdbConnection
            :param start: Flag to indicate if the connection point to evaluate
                          is the on the startObject or the endObject of the
                          connection.
            :type start: bool
            :return: The value in radian of the angle of the connection on
                     its connected instance (startObject if start is True,
                     endObject if start is False)
            :rtype: float
            """
            sx, sy, ex, ey, *_ = whdb_getArcEndPoints(interConn)

            x = sx if start else ex
            y = sy if start else ey

            inst = interConn.getStartObject() if start else \
                   interConn.getEndObject()
            instX, instY = inst.getPosition()
            x -= instX
            y -= instY

            return math.atan2(y, x)

        # List of angle of all other connections connected to a
        # different instances
        conPtsAngle = []
        instance = circConn.getStartObject()
        for conn in connectionList:
            startObj = conn.getStartObject()
            endObj = conn.getEndObject()

            if startObj is not endObj:
                if startObj is instance:
                    conPtsAngle.append(conPtToAngle(conn, True))
                elif endObj is instance:
                    conPtsAngle.append(conPtToAngle(conn, False))

        ptsNum = len(conPtsAngle)
        if ptsNum == 0:
            return WHDB_CIRC_DEFAULT_ANGLE
        else:
            if ptsNum == 1:
                # Special case: set at opposite angle
                optimalAngle = conPtsAngle[0] + math.pi
            else:
                conPtsAngle.sort()

                # Find angle difference between each point
                angleDiff = {}
                for i in range(0, len(conPtsAngle) - 1):
                    angleDiff[conPtsAngle[i]] = (
                        conPtsAngle[i + 1] - conPtsAngle[i])

                if len(conPtsAngle) > 1:
                    angleDiff[conPtsAngle[ptsNum - 1]] = (
                        conPtsAngle[0] - conPtsAngle[ptsNum - 1] + 2 * math.pi)

                maxAngleKey = max(angleDiff, key=angleDiff.get)
                optimalAngle = angleDiff[maxAngleKey] / 2 + maxAngleKey

            return optimalAngle
