# ----------------------------------------------------------------------------
# - wcact-import-route-binding-action                                        -
# - Dezygnr - project editor actions                                         -
# ----------------------------------------------------------------------------
# - author    (c) 2019-2019 achraf mahti                                     -
# - copyright (c) 2019-2019 silkan                                           -
# ----------------------------------------------------------------------------

"""
.. module::       wcact_import_route_binding_action
   :platform:     Unix, Darwin, Windows

.. moduleauthor:: Achraf mahti <achraf@cetrac.io>
"""

import csv, os

from wcr_core.action.wact_action_parameter import WactActionParameter
from wcr_core.action.wact_action_parameter_types import \
    WactActionOpenFileByPayloadParameterType
from wcr_core.common.wcmn_mime_type import WCMN_PAYLOAD_BLOCK_CSV
from wcr_core.db.whdb_utils import whdb_ensureUniqueName
from wcr_core.modules.vlink.db.wvhdb_standard_operators import \
    WVHDB_VLINK_CODE
from wcr_core.common.wcmn_file_handler import WcmnFile
from wcr_frontend import showErrorWindow
from wcr_core.action.wact_action import *
from wcr_frontend.modules.vlink.gui import WVGUI_CSV_VL_NAME_KEY, \
    WVGUI_CSV_VL_TYPE_KEY, WVGUI_CSV_VL_CONFIG
from wcr_frontend.modules.vlink.gui.wvgui_scene_protocol_table import *
from wcr_core.modules.vlink.db.wvhdb_Vlink_instance import *
from wcr_core.project.wprj_xmlio import WPRJ_SILKAN_CSV_MIME_TYPE
from wcr_frontend.gui.wgui_utils import wgui_getOpenFile


# noinspection PyAbstractClass
class WguiEditorActionMixin(WactActionBaseP):
    """
    Base class for project editor actions.
    """

    def isEnabled(self, actx):
        """
        :type actx: WguiActionExecutionContextP
        """
        from wcr_frontend.gui.wgui_project_editor import WguiProjectEditor
        return isinstance(actx.getEditor(), WguiProjectEditor)

    def _doExecute(self, actx, paramValuePairs):
        editor = actx.getEditor()
        from wcr_frontend.gui.wgui_project_editor import WguiProjectEditor
        if isinstance(editor, WguiProjectEditor):
            self._executeForEditor(editor, actx, paramValuePairs)

    @abstractmethod
    def _executeForEditor(self, editor, actx, paramValuePairs):
        """
        Execute the action on the given editor.

        :param actx: Action execution context
        :type actx: WactActionExecutionContextP
        :param editor: Editor to execute action for
        :type editor: WguiProjectEditor
        :param paramValuePairs: Parameter-value pairs
        """
        pass


# noinspection PyAbstractClass
class WguiEditorActionBase(WactActionBase, WguiEditorActionMixin):
    """
    Base class for project editor actions.
    """

    def __init__(self, name, description, *parameters):
        """
        :param name: Action name
        :type name: str
        :param description: Action description
        :type description: str
        :param parameters: Action parameter descriptions
        :type parameters: list[WactActionParameterP]
        """
        WactActionBase.__init__(self, name, description, *parameters)
        WguiEditorActionMixin.__init__(self)

    def isEnabled(self, actx):
        return WguiEditorActionMixin.isEnabled(self, actx)


class _WmguiEditorActionBase(WguiEditorActionBase):
    def __init__(self, name, description, *parameters):
        """
        :param name: Action name
        :type name: str
        :param description: Action description
        :type description: str
        :param parameters: Action parameter descriptions
        :type parameters: list[WactActionParameterP]
        """
        super().__init__(name, description, *parameters)

    def isEnabled(self, actx):
        if not super().isEnabled(actx):
            return False
        editor = actx.getEditor()
        """:type: WguiProjectEditor"""
        cctx = editor.cockpitSystemContext
        from wcr_core.modules.sys_map.wcmod_application_module import \
            WCMOD_SYSTEM_MAP_APP_CODE
        return cctx.code == WCMOD_SYSTEM_MAP_APP_CODE


WCACT_PRODUCER_PARAM = 'Producer'
WCACT_CONSUMER_PARAM = 'Consumer'


class _WvguiEditorImportCSVAction(_WmguiEditorActionBase):
    """
    Project editor action to import a model.
    """

    #: Action parameters
    __PARAMETER = WactActionParameter(
        'model', 'Model file', 'Model file to import',
        WactActionOpenFileByPayloadParameterType(WCMN_PAYLOAD_BLOCK_CSV))

    def __init__(self):
        super().__init__(
            'Import Vlink from CSV...',
            'Import Traffic Route from CSV File',
            self.__PARAMETER)

    def _executeForEditor(self, editor, actx, paramValuePairs):
        filename = wgui_getOpenFile("Import csv", actx.getEditor(),
                                WPRJ_SILKAN_CSV_MIME_TYPE)
        """:type: WcmnFile | None"""
        if not filename:
            return
        cctx = editor.cockpitSystemContext
        pctx = actx.getProjectContext()
        mapp = pctx.getMappingManager()
        tblk = pctx.getProjectDB().getTopHierarchy()
        routeTable = cctx.getRouteTable()
        routeTableProducer = cctx.getProducerTable()
        routeTableConsumer = cctx.getConsumerTable()
        fl = open(filename, 'r')
        from wcr_core.modules.vlink.wvmod_application_module import WVMOD_VLINK_APP_MODULE
        vlblok = tblk.getVLinkBlock()
        if not vlblok:
            vlFileName = 'TrAuto.tr.xml'
            dirname = os.path.dirname(tblk.getHardwareFile().path)
            vlfullpath = os.path.join(dirname, vlFileName)
            vlblok = WVMOD_VLINK_APP_MODULE.createTopLevelBlock()
            vlblok.setSourcePath(vlfullpath)
            vlblok.formatToSave(vlfullpath)
            tblk.setVLinkFile(WcmnFile(vlfullpath))
            tblk.setVLinkBlock(vlblok)
        else:
            vlFileName = tblk.getVLinkFiles().path

        routes, prodcons = wcact_import_routes(vlblok, fl)

        baseName = os.path.basename(vlFileName)
        wcact_add_binding_tables(tblk, mapp, prodcons, routes, baseName, routeTableProducer,
                                 routeTableConsumer, routeTable)


def wcact_import_routes(vlblok, fl):
    headers = []
    csvparamdic = {}
    csvData = csv.reader(fl)
    routes = []
    prodcons = []
    for vlp in csvData:
        if headers == []:
            headers = vlp
            continue
        for i in range(0, min(len(headers), len(vlp))):
            csvparamdic[headers[i]] = vlp[i]
        route = vlblok.createOperatorInstance(WVHDB_VLINK_CODE)
        if WVGUI_CSV_VL_NAME_KEY in csvparamdic.keys():
            name = csvparamdic[WVGUI_CSV_VL_NAME_KEY]
        else:
            name = 'default'

        route.setName(name)
        if WVGUI_CSV_VL_TYPE_KEY in csvparamdic.keys():
            VL_TYPE = csvparamdic[WVGUI_CSV_VL_TYPE_KEY]
            route.setVLType(VL_TYPE)
            prod, conss = None, None
            for hkey in csvparamdic.keys():
                if hkey != WVGUI_CSV_VL_TYPE_KEY and \
                                hkey != WVGUI_CSV_VL_NAME_KEY:
                    prmKey = WVGUI_CSV_VL_CONFIG.get(hkey, None)
                    if prmKey:
                        route.setVLParameters({prmKey: csvparamdic[hkey]})
                    elif hkey == WCACT_PRODUCER_PARAM:
                        prod = csvparamdic[hkey]
                    elif hkey == WCACT_CONSUMER_PARAM:
                        conss = csvparamdic[hkey].split(';')
                    else:
                        showErrorWindow('Import csv error',
                                        'Traffic Route key "{}" does not exist.', hkey)
            prodcons.append((prod, conss))
            whdb_ensureUniqueName(route, vlblok)
            routes.append(route)
    return routes, prodcons


def wcact_add_binding_tables(tblk, mapp, prodcons, routes, baseName, routeTableProducer,
                             routeTableConsumer, routeTable):
    for i, route in enumerate(routes):
        prod, conss = prodcons[i]
        connProducer = wcact_getPortFromName(tblk, prod)
        if not connProducer:
            showErrorWindow('Import csv error',
                            'Traffic Route {} has no producer',
                            route.getName())
        else:
            hconn = mapp.getPhyPortByUUID(connProducer.getUUIDSource())
            routeTableProducer._addEntry((prod, hconn),
                                         (route.getName(), route))
            routeTableProducer.bindPortProd((prod, hconn),
                                            (route.getName(), route))
        if not conss:
            showErrorWindow('Import csv error',
                            'Traffic Route {} has no consumer',
                            route.getName())
        else:
            for cons in conss:
                connConsumer = wcact_getPortFromName(tblk, cons)
                hconn = mapp.getPhyPortByUUID(connConsumer.getUUIDSource())
                routeTableConsumer._addEntry((cons, hconn),
                                             (route.getName(), route))
                routeTableConsumer.bindPorts((cons, hconn),
                                             (route.getName(), route))

        routeTable.modTblM.addRow(baseName,
                                  (route.getName(), route))


def wcact_getPortFromName(tblk, portName):
    if portName is None:
        return None
    eqpName, cardName, portName = portName.split('.')
    for inst in tblk.getOperatorInstances():
        if eqpName == inst.getName():
            for xpnt in inst.getConnectionPoints():
                if xpnt.getCard() == cardName and portName == xpnt.getName():
                    return xpnt


#: Logical simulation load action
WCACT_IMPORT_ROUTE_BIND_ACTION = _WvguiEditorImportCSVAction()
