# ----------------------------------------------------------------------------
# - wchdb-top-block                                                          -
# - wild cruncher:wsh top hierarchy object               core / sys_map / db -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 lambert philippe                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wchdb_top_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher top hierarchy object

.. moduleauthor:: philippe lambert <lambert@silkan.com>
"""
import os
from wcr_core.modules.sys_map.db.wchdb_block import WchdbBlock
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaDescription, \
    WdbsMetaField
from wcr_core.dbstore.wdbs_meta_type import WDBS_PATH_META, WdbsListMetaType, \
    WdbsTupleMetaType, WDBS_UUID_META, WDBS_STRING_META, WdbsDictionaryMetaType
from wcr_core.common.wcmn_file_handler import WcmnFile

from silkan_common.scmn_file_path import scmn_path_rel2file, scmn_path_mkabs, \
    scmn_isfile, scmn_path_dir, scmn_relpath

WCHDB_HARDWARE_PATH_KEY = 'Hardware-path-key'
WCHDB_VLINK_PATH_KEY = 'VLink-path-key'
WCHDB_PROD_BINDING_KEY = 'Producer-binding-key'
WCHDB_CONS_BINDING_KEY = 'Consumer-binding-key'
WCHDB_MODULE_CARD_BINDING_KEY = 'Module-Card-binding'


class WchdbTopHierarchy(WchdbBlock):
    """
    The WchdbTopHierarchy is the top hierarchy model class. A top hierarchy
    is at the same time a model and an instance. The class provides several
    methods for creating the hierarchy.
    """
    META = WdbsMetaDescription(
        WchdbBlock.META,
        WdbsMetaField(WCHDB_HARDWARE_PATH_KEY, WDBS_PATH_META),
        WdbsMetaField(WCHDB_VLINK_PATH_KEY, WdbsListMetaType(WDBS_PATH_META)),
        WdbsMetaField(WCHDB_PROD_BINDING_KEY,
                      WdbsListMetaType(
                          WdbsTupleMetaType(WDBS_UUID_META, WDBS_UUID_META))),
        WdbsMetaField(WCHDB_CONS_BINDING_KEY,
                      WdbsListMetaType(
                          WdbsTupleMetaType(WDBS_UUID_META, WDBS_UUID_META))),
        WdbsMetaField(WCHDB_MODULE_CARD_BINDING_KEY,
                      WdbsDictionaryMetaType(WDBS_STRING_META,
                                         WdbsListMetaType(
                                            WdbsTupleMetaType(WDBS_STRING_META,
                                                              WDBS_STRING_META,
                                                              WDBS_STRING_META))))

    )

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        super().__init__(block, wreg)
        self._srdfilepath=None

    def setHardwareFile(self, file):
        """
        Set the path of the file for the hardware associated with
        the system-mapper hierarchy.

        :param file: Path of the file for the hardware
        :type file: WcmnFile
        """
        self.rawElement.setMeta(file, WCHDB_HARDWARE_PATH_KEY)

    def getHardwareFile(self):
        """
        Get the path of the file for the hardware associated with
        the system-mapper hierarchy.

        :return: Path of the file for the hardware
        :rtype: WcmnFile | None
        """
        relpathhard = self.rawElement.getMeta(WCHDB_HARDWARE_PATH_KEY)
        #: no hardware file binded
        if relpathhard is None:
            return relpathhard
        abspathhard = scmn_path_mkabs(relpathhard.path,
                                      bdir=scmn_path_dir(self._srdfilepath))
        if not scmn_isfile(abspathhard):
            self.rawElement.setMeta(None, WCHDB_HARDWARE_PATH_KEY)
            return None
        return WcmnFile(abspathhard)

    def setVLinkFile(self, file):
        """
        Set the path of the file for the virtual link view associated with
        the system-mapper hierarchy.

        :param file: Path of the file for the bus
        :type file: WcmnFile
        """
        if file:
            filel = [file]
        else:
            filel = []
        self.rawElement.setMeta(filel, WCHDB_VLINK_PATH_KEY)

    def addVLinkFile(self, file):
        """
        add a file path to the list of bus file imported into the project.
        """

        add = False
        filel = self.rawElement.getMeta(WCHDB_VLINK_PATH_KEY)
        if filel:
            filepl = [x.path for x in filel]
        else:
            filepl = []
        if file.path not in filepl:
            filel.append(file)
            add = True
        self.rawElement.setMeta(filel, WCHDB_VLINK_PATH_KEY)
        return add

    def getVLinkFiles(self):
        """
        Get the path of the file for the virtual link view associated with
        the system-mapper hierarchy.

        :return: Path of the file for the bus
        :rtype: [WcmnFile]
        """
        vlfl=self.rawElement.getMeta(WCHDB_VLINK_PATH_KEY)
        #: no Vlink file binded
        if vlfl is None:
            return vlfl
        res=[]
        for vlf in vlfl:
            if vlf is None:
                continue
            abspathvl = scmn_path_mkabs(vlf.path,
                                        bdir=scmn_path_dir(self._srdfilepath))
            if not scmn_isfile(abspathvl):
                self.removeVLinkFile(vlf)
            else:
                res.append(WcmnFile(abspathvl))
        return res

    def removeVLinkFile(self, vlf):
        """
        remove the specific Traffic Route file from metadata
        """
        filel = self.rawElement.getMeta(WCHDB_VLINK_PATH_KEY)
        filel.remove(vlf)
        self.rawElement.setMeta(filel, WCHDB_VLINK_PATH_KEY)

    def addProdBind(self, pportuuid, vluuid):
        """
        add a tuple (pportuuid,moduleuuid) to the list of binding
        :param pportuuid: uuid of the host to bind as productor
        :type pportuuid: uuid
        :param vluuid: uuid of the virtual link to bind
        :type vluuid: uuid
        """
        bndl = self.rawElement.getMeta(WCHDB_PROD_BINDING_KEY)
        bndl.append((pportuuid, vluuid))
        self.rawElement.setMeta(set(bndl), WCHDB_PROD_BINDING_KEY)

    def getProdBinds(self):
        """
        return a liste of tuple (pportuuid,vluuid) to the list of binding
        :return liste of pportuuid/vluuid binding
        :rtype [(uuid,uuid)]
        """
        return self.rawElement.getMeta(WCHDB_PROD_BINDING_KEY)

    def removeProdBind(self, pportuuid, vluuid):
        """
        remove the tuple (hostuuid,vluuid) from the list of binding
        :param pportuuid: uuid of the physical port to unbind in producers
        :type pportuuid: uuid
        :param vluuid: uuid of the virtuallink to unbind
        :type vluuid: uuid
        """
        bndl = self.rawElement.getMeta(WCHDB_PROD_BINDING_KEY)
        for p,vl in bndl:
            if p == pportuuid and vl == vluuid:
                bndl.remove((pportuuid, vluuid))
        self.rawElement.setMeta(bndl, WCHDB_PROD_BINDING_KEY)

    def addConsBind(self, pportuuid, vluuid):
        """
        add a tuple (physicalportuuid,logicportuuid) to the list of binding
        :param pportuuid: uuid of the physical port to bind as consumer
        :type pportuuid: uuid
        :param vluuid: uuid of the virtual link to bind
        :type vluuid: uuid
        """
        bndl = self.rawElement.getMeta(WCHDB_CONS_BINDING_KEY)
        bndl.append((pportuuid, vluuid))
        self.rawElement.setMeta(set(bndl), WCHDB_CONS_BINDING_KEY)

    def getConsBinds(self):
        """
        return a liste of tuple (physicalportuuid,vlinkuuid) to the
        list of binding
        :return liste of physicalportuuid/vlinkuuid binding
        :rtype [(uuid,uuid)]
        """
        return self.rawElement.getMeta(WCHDB_CONS_BINDING_KEY)

    def removeConsBind(self, pportuuid, vluuid):
        """
        remove the tuple (physicalportuuid,vlinkuuid) from the
        list of binding
        :param pportuuid: uuid of the physical port to unbind
        :type pportuuid: uuid
        :param vluuid: uuid of the virtual link to unbind
        :type vluuid: uuid
        """
        bndl = self.rawElement.getMeta(WCHDB_CONS_BINDING_KEY)
        for pp,lp in bndl:
            if pp == pportuuid and lp == vluuid:
                bndl.remove((pportuuid, vluuid))
        self.rawElement.setMeta(bndl, WCHDB_CONS_BINDING_KEY)

    def checkportIsConnected(self, portUuid):
        """
            check if the given port is connected
            :param pportuuid: uuid of the physical port
            :type pportuuid: uuid
            :return True is the portUuid is found inn the connected ports list
            :rtype boolean
        """
        isConnected = False
        for conn in self.getConnections():
            portEnd = conn.getEndConnectionPoint()
            portStart = conn.getStartConnectionPoint()
            if portUuid == portEnd.getUUIDSrc() or portUuid ==portStart.getUUIDSrc():
                isConnected = True
        return isConnected

    # =====================

    def addModuleCardBind(self, eqpName, moduleName, cardName, period):
        """
        add a tuple (moduleName, cardName) to the list of binding
        :param moduleName: name of the module
        :type moduleName: str
        :param cardName:name of the card
        :type cardName: str
        :param period: the period at which the app will be requested
        :type period: str
        """
        if moduleName and cardName:
            bndl = self.rawElement.getMeta(WCHDB_MODULE_CARD_BINDING_KEY)
            bndl.setdefault(eqpName, []).append((moduleName, cardName, period))
            self.rawElement.setMeta(bndl, WCHDB_MODULE_CARD_BINDING_KEY)

    def getModuleCardBind(self):
        """
        return a liste of tuple (moduleName, cardName) to the
        list of binding
        :return liste of moduleName and cardName binding
        :rtype [(str, str)]
        """
        return self.rawElement.getMeta(WCHDB_MODULE_CARD_BINDING_KEY)

    def getEqpModuleCardBind(self, eqpName):
        """
        return a liste of tuple (moduleName, cardName) to the
        list of binding
        :return liste of moduleName and cardName binding
        :rtype [(str, str)]
        """
        bindings = self.rawElement.getMeta(WCHDB_MODULE_CARD_BINDING_KEY, eqpName)
        if not bindings:
            bindings = {}
        return bindings

    def getEqpCardFromModule(self, targetModule):
        """
        :param targetModule: name of the module
        :type targetModule: str
        return a Eqp.cardName
        :return list of moduleName and cardName binding
        :rtype str
        """
        bindings = self.rawElement.getMeta(WCHDB_MODULE_CARD_BINDING_KEY)
        for eqp in bindings:
            for moduleName, card, period in bindings[eqp]:
                if moduleName == targetModule:
                    return '{}.{}'.format(eqp, card)

    def removeModuleCardBind(self, eqpName, moduleName, CardName, period):
        """
        remove the tuple (moduleName, cardName) from the
        list of binding
        :param moduleName: name of the module
        :type moduleName: str
        :param CardName:name of the card
        :type CardName: str
        :param period: the period at which the app will be requested
        :type period: str
        """
        bndl = self.rawElement.getMeta(WCHDB_MODULE_CARD_BINDING_KEY)
        for key, (pp, lp) in bndl.items():
            if key == eqpName and pp == moduleName and lp == CardName:
                bndl[key].remove((moduleName, CardName, period))
                if len(bndl[key]) == 0:
                    del bndl[key]
        self.rawElement.setMeta(bndl, WCHDB_MODULE_CARD_BINDING_KEY)

    def resetEqpModuleCardBind(self, eqpName):
        """
        reset the module/card binding
        """
        bndl = self.rawElement.getMeta(WCHDB_MODULE_CARD_BINDING_KEY)
        if eqpName in bndl:
            del bndl[eqpName]
            self.rawElement.setMeta(bndl, WCHDB_MODULE_CARD_BINDING_KEY)

    def setSourcePath(self, path):
        if path is not None:
            self._srdfilepath = path

    def formatToSave(self, npath):
        ndir = scmn_path_dir(npath)
        vlsfl = self.getVLinkFiles()
        self.setVLinkFile(None)
        for vlf in vlsfl:
            if scmn_isfile(vlf.path):
                vlrela = scmn_relpath(vlf.path, ndir)
                self.addVLinkFile(WcmnFile(vlrela))

        hardf = self.getHardwareFile()
        if hardf:
            if scmn_isfile(hardf.path):
                hardrela = scmn_relpath(hardf.path,ndir)
                self.setHardwareFile(WcmnFile(hardrela))
        self.setSourcePath(npath)


