# ----------------------------------------------------------------------------
# - whmod_application_module                                                 -
# - wild cruncher: application module                        core / hardware -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whmod_application_module
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher hardware designer application module

.. moduleauthor:: philippe lambert <lambert@silkan.com>
"""
from wcr_core.action.wact_action_api import WACT_WCR_ACTION_CODE
from wcr_core.db.whdb_top_level_factory import whdb_createTopLevelRawBlock
from wcr_core.modules.sys_map.db.wchdb_meta_provider import \
    WchdbMetaProvider
from wcr_core.modules.sys_map.db.wchdb_wrapper_registry import \
    WchdbWrapperRegistry
from wcr_core.modules.sys_map.wcmod_project_context import \
    WcmodSystemProjectContext
from wcr_core.modules.sys_map.db.wchdb_verifier import \
    WchdbHierarchyVerifier
from wcr_core.modules.sys_map.controller.wcctl_controllers import \
    WCCTL_CONTROLLER_FACTORY
from wcr_core.modules.sys_map.db.wchdb_operator import WCHDB_OPER
from wcr_core.modules.sys_map.db.wchdb_standard_operators import \
    WCHDB_DEFAULT_OPS
from wcr_core.modules.wmod_application_module import \
    WMOD_DEFAULT_BLOCK_TYPES, WmodApplicationModuleBase, \
    WmodApplicationModuleP, WmodApplicationModuleMP
from wcr_core.common.wcmn_mime_type import WCMN_PAYLOAD_BLOCK_SYS_MAP
from wcr_core.modules.sys_map.actions.wcact_import_hardware_action \
    import WCACT_IMPORT_HARDWARE_ACTION
from wcr_core.modules.sys_map.actions.wcact_import_logic_action \
    import WCACT_IMPORT_LOGICAL_ACTION
from wcr_core.modules.sys_map.actions.wcact_reload_hardware_action \
    import WCACT_RELOAD_HARDWARE_ACTION
from wcr_core.modules.sys_map.actions.wcact_reload_logic_action \
    import WCACT_RELOAD_LOGICAL_ACTION
from wcr_core.modules.sys_map.actions.wcact_map_checker_action \
    import WCACT_CHECK_MAPPING_ACTION, WCACT_ACTION_BIND_VERIFY
from wcr_core.modules.sys_map.actions.wcact_generate_file \
    import WCACT_GENERATE_FILE_ACTION
from wcr_core.modules.sys_map.actions.wcact_action_base \
    import WCACT_ACTION_LOAD_HARD, WCACT_ACTION_UP_HARD, \
    WCACT_ACTION_UP_LOGIC, WCACT_ACTION_LOAD_LOGIC, WCACT_ACTION_IMPORT_ROUTE
from wcr_core.db_api.whdb_block_api import WHDB_TOPLEVEL_BLOCK_CODE, \
    WHDB_DEFAULT_BLOCK_CODE
from wcr_core.modules.sys_map.db.wchdb_top_block import WchdbTopHierarchy
from wcr_core.modules.wmod_block_type import WmodBlockType, \
    WMOD_SYSTEM_BLOCK_FLAG


WCMOD_SYSTEM_MAP_APP_CODE = 'System_Mapper'
WCMOD_SYSTEM_MAP_APP_NAME = 'Mapping View'

#: Section identifier for the toolbar action section
WCMOD_ACTION_SECTION_TOOLBAR = 'toolbar'

WCMOD_BLOCK_TYPES = [
    WmodBlockType(WHDB_TOPLEVEL_BLOCK_CODE, 'Top-level',
                  WchdbTopHierarchy.META, WchdbTopHierarchy,
                  WMOD_SYSTEM_BLOCK_FLAG),
]

# noinspection PyAbstractClass
class WcmodSysmapApplicationModuleP(WmodApplicationModuleP):
    pass


# noinspection PyAbstractClass
class WcmodSysmapApplicationModuleMP(WmodApplicationModuleMP,
                                     WcmodSysmapApplicationModuleP):
    pass


class _WcmodSysmapApplicationModule(WmodApplicationModuleBase,
                                    WcmodSysmapApplicationModuleMP):
    def __init__(self, operRegistry):
        super().__init__(
            code=WCMOD_SYSTEM_MAP_APP_CODE,
            name=WCMOD_SYSTEM_MAP_APP_NAME,
            blockTypes=WCMOD_BLOCK_TYPES,
            defaultOps=WCHDB_DEFAULT_OPS,
            operRegistry=operRegistry)

        #: Connection point types
        self.__xpointTypes = {}
        """:type: dict[str, WbmodConnectionPointTypeP]"""
        self._createDefaultActions()

    def _createDefaultActions(self):
        actnSection = '{}.core.{}'.format(
            WACT_WCR_ACTION_CODE, WCMOD_SYSTEM_MAP_APP_CODE)
        self.addAction(
            WCACT_IMPORT_HARDWARE_ACTION, 'sethardware', actnSection,
            WCACT_ACTION_LOAD_HARD)
        self.addAction(
            WCACT_RELOAD_HARDWARE_ACTION, 'reloadhardware', actnSection,
            WCACT_ACTION_UP_HARD)
        self.addAction(
            WCACT_IMPORT_LOGICAL_ACTION, 'setlogical', actnSection,
            WCACT_ACTION_LOAD_LOGIC)
        self.addAction(
            WCACT_RELOAD_LOGICAL_ACTION, 'reloadlogical', actnSection,
            WCACT_ACTION_UP_LOGIC)
        self.addAction(
            WCACT_CHECK_MAPPING_ACTION, 'checkbindings', actnSection,
            WCACT_ACTION_BIND_VERIFY)
        self.addAction(
            WCACT_GENERATE_FILE_ACTION, 'generateconffile', actnSection,
            WCMOD_ACTION_SECTION_TOOLBAR)

    def createHierarchyVerifier(self, pctx):
        return WchdbHierarchyVerifier(self._getDbVerifierModules())

    def getPayloadType(self):
        return WCMN_PAYLOAD_BLOCK_SYS_MAP

    def createTopLevelBlock(self, name=None, info=None):
        return wcmod_create_top_block(self, name, info)

    def createProjectContext(self, gContext, projectDB, logger=None):
        if getattr(self, 'actx', None):
            return WcmodSystemProjectContext(
                gContext, self, projectDB, self.actx, logger, WCCTL_CONTROLLER_FACTORY)
        else:
            return WcmodSystemProjectContext(
                gContext, self, projectDB, logger=logger, ctrlfct=WCCTL_CONTROLLER_FACTORY)

    def addConnectionPointType(self, xpntType):
        self.__xpointTypes[xpntType.getCode()] = xpntType

    def getConnectionPointType(self, code):
        return self.__xpointTypes.get(code)

    def getConnectionPointTypes(self):
        return self.__xpointTypes.values()

    def getFileExtention(self):
        return '.map'


def wcmod_create_top_block(module, name=None, info=None):
    blockMetas = {b.getCode(): b.getMeta() for b in module.getBlockTypes()}
    blockClzzs = {b.getCode(): b.getClass() for b in module.getBlockTypes()}
    xpntClzzs = {x.getCode(): x.getClass() for x
                 in module.getConnectionPointTypes()}
    """:type: dict[str, () -> WphdbConnectionPointBase]"""
    xpntCfgMetas = {x.getCode(): x.getConfigMeta() for x
                    in module.getConnectionPointTypes()}
    """:type: dict[str, WdbsMetaTypeP]"""
    rgty = WchdbWrapperRegistry(blockClzzs, xpntClzzs)
    tblk = whdb_createTopLevelRawBlock(
        WchdbMetaProvider(blockMetas , xpntCfgMetas), name, info)
    return rgty.createWrapper(tblk)


#: HDS Application module singleton
WCMOD_SYSTEM_MAP_APP_MODULE = _WcmodSysmapApplicationModule(WCHDB_OPER)
