# ----------------------------------------------------------------------------
# - wcmod-mapping-manager                                                    -
# - wild cruncher - sys model mapping manager                 core / sys_map -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcmod_mapping_manager
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher system mapper model mapping manager

.. moduleauthor:: Philippe lambert <lambert@silkan.com>
"""
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP
from wcr_core.modules.hardware.db.whhdb_Cetrac_switch_instance \
    import WhhdbCetracSwitchInstance
from wcr_core.modules.hardware.db.whhdb_kazymir_switch_instance import \
    WhhdbKazymirSwitchInstance
from wcr_core.modules.hardware.db.whhdb_rack_instance import WhhdbRackInstance
from wcr_core.modules.hardware.db.whhdb_Host_instance import WhhdbHostInstance
from wcr_core.modules.hardware.db.whhdb_operator_instance \
    import WhhdbOperatorInstance, WHHDB_OPER_SWITCH_TYPE
from wcr_core.db.whdb_connection import WhdbConnection, whdb_getOtherObject
from wcr_core.modules.vlink.db.wvhdb_Vlink_instance import WvhdbVLinkInstance
from wcr_core.modules.vlink.wvmod_mapping_manager import WvmodMappingManager
from wcr_core.modules.icd.wimod_application_module import \
    WIMOD_ICD_APP_CODE
from silkan_common.scmn_file_path import scmn_isfile, scmn_path_base
from wcr_core.modules.wmod_application_module import wmod_getModuleForPdb

from wcr_core.project.wprj_project import wprj_import_project
from wcr_core.common.wcmn_file_handler import WcmnFile


class WcmodMappingManager(object):
    """
    The mapping manager assists in accessing element of hardware source
    referencing elements by they're UUIDs
    """

    def __init__(self, tblk, xctx):
        """
        :param tblk: System mapper hierarchy top-level
        :type tblk: WchdbTopBlock
        """

        #: System mapper hierarchy top-level
        self.__topBlock = tblk

        self.__xctx = xctx

        #: hardware description (when loaded)
        self.__hardware = None
        #: Mapping from UUID to hardware models
        self.__models = {}

        #: Traffic Route models
        self.__vlink = []
        self.__vlinkMessageBinding = []
        #: Mapping from UUID to Traffic Route
        self.__VirtualLink = {}
        self.__globalTypes = []

        self.deletedVl = []

        #: Modules and their Virtual link Mapping managers...
        self.__modules = []
        self.__modulesByName = {}
        self.__allMessages = {}

    @property
    def virtualLink(self):
        return self.__VirtualLink

    @property
    def vlink(self):
        return self.__vlink

    def setHardware(self, lsim):
        """
        Set the hardware source.

        :param lsim: Hardware source
        :type lsim: WhhdbTopBlock | None
        """
        self.__hardware = lsim

        if lsim:
            self.__models = self._readHardware(lsim)
        else:
            self.__models = {}

    def _readHardware(self, hard):
        """
        Read the UUID-model mappings from the given Hardware.

        :param hard: Hardware design
        :type hard: WhhdbTopBlock
        :return: Mappings from UUIDs to models
        :rtype: dict[UUID, WshdbModelInstance]
        """

        def readBlock(block):
            for inst in block.getInstances():
                if isinstance(inst, WhdbBlockInstanceP):
                    readBlock(inst.getBlock())
                elif isinstance(inst, WhhdbHostInstance):
                    models[inst.getUUID()] = inst
                elif isinstance(inst, WhhdbRackInstance):
                    models[inst.getUUID()] = inst
                elif isinstance(inst, WhhdbCetracSwitchInstance):
                    models[inst.getUUID()] = inst
                elif isinstance(inst, WhhdbOperatorInstance):
                    models[inst.getUUID()] = inst
            for conn in block.getConnections():
                    models[conn.getUUID()] = conn

        models = {}

        readBlock(hard)

        return models

    def getHosts(self):
        """
        Return every host instances from hardware model
        :return: list of host instances
        :rtype: [WhhdbHostInstance]
        """
        hostl=[]
        for k in self.__models.keys():
            if isinstance(self.__models[k], WhhdbHostInstance):
                hostl.append(self.__models[k])
        return hostl

    def getRacks(self):
        """
        Return every rack instances from hardware model
        :return: list of host instances
        :rtype: [WhhdbRackInstance]
        """
        rackl = []
        for k in self.__models.keys():
            if isinstance(self.__models[k], WhhdbRackInstance):
                rackl.append(self.__models[k])
        return rackl

    def getHostByUUID(self, uuid):
        """
        Return host instances from hardware model coresponding
        to the given UUID or None
        :param uuid: uuid seek design
        :type uuid: uuid
        :return: host instances
        :rtype: WhhdbHostInstance | None
        """

        if isinstance(self.__models[uuid], WhhdbHostInstance):
            return self.__models[uuid]
        return None

    def getSwitchs(self):
        """
        Return every switches instances from hardware model
        :return: list of host instances
        :rtype: [WhhdbCetracSwitchInstance]
        """
        suuids=[]
        for k in self.__models.keys():
            if isinstance(self.__models[k], WhhdbCetracSwitchInstance):
                suuids.append(self.__models[k])
            if isinstance(self.__models[k], WhhdbOperatorInstance):
                if self.__models[k].getOperatorMetaType() == \
                        WHHDB_OPER_SWITCH_TYPE:
                    suuids.append(self.__models[k])
            if isinstance(self.__models[k], WhhdbKazymirSwitchInstance):
                suuids.append(self.__models[k])
        return suuids

    def getConections(self):
        """
        Return every connections from hardware model
        :return: list of Connection
        :rtype: [WhdbConnection]
        """
        suuids=[]
        for k in self.__models.keys():
            if isinstance(self.__models[k], WhdbConnection):
                suuids.append(self.__models[k])
        return suuids

    def getSwitchByPortUUID(self, puuid):
        """
        return the switch containing the port having given uuid
        :param puuid: uuid of the port for how we search switch
        :type puuid: uuid
        :return: Switch
        :rtype: WhhdbCetracSwitchInstance | None
        """
        for s in self.getSwitchs():
            if puuid in [sp.getUUID() for sp in s.getConnectionPoints()]:
                return s

    def getPhyPortByUUID(self, uuid):
        """
        Return phycical port instances from Hardware model corresponding
        to the given UUID or None
        :param uuid: uuid seek design
        :type uuid: uuid
        :return: Connection point
        :rtype: WhhdbConnectionPointBase | None
        """
        #searching in hosts ports
        for h in self.getHosts():
            for p in h.getConnectionPoints():
                if p.getUUID() == uuid:
                    return p
        #: searching in switchs ports
        for s in self.getSwitchs():
            for p in s.getConnectionPoints():
                if p.getUUID() == uuid:
                    return p
        #: searching in switchs ports
        for s in self.getRacks():
            for p in s.getConnectionPoints():
                if p.getUUID() == uuid:
                    return p
        return None

    def getHostFromPortUUID(self, uuid):
        """
        Return host instances from Hardware model hosting the port
        corresponding to the given UUID or None
        :param uuid: uuid seek design
        :type uuid: uuid
        :return: Connection point
        :rtype: WhhdbConnectionPointBase | None
        """
        #searching in hosts ports
        for h in self.getHosts():
            for p in h.getConnectionPoints():
                if p.getUUID() == uuid:
                    return h

    def formatName(self, inst, prefix=None):
        """
        return a formated name for an instance given.
        """
        if not prefix or not isinstance(prefix, str):
            return inst.getName()
        return prefix+'-'+inst.getName()

#######################################################################

    def resetVLink(self):
        """
        set all the Traffic Route related variables to they empty state.
        """
        self.__vlink = []
        self.__VirtualLink = {}

    def getAllVirtualLinks(self):
        """
        Return all the imported modules
        :return: all the modules and an appropriate name.
        :rtype: [(name,module)]
        """
        VLinkList = []
        for Vlinktblk, pre in self.__vlink:
            for inst in Vlinktblk.getInstances():
                if (inst.getUUID() in self.__VirtualLink.keys() and
                       inst.getUUID() not in self.deletedVl):
                    name = self.formatName(self.__VirtualLink[inst.getUUID()],
                                           pre)
                    VLinkList.append((name, inst))
        return VLinkList

    def getVlinkByUUID(self, uuid):
        """
        Return module instances from Bus model corresponding
        to the given UUID or None
        :param uuid: uuid seek design
        :type uuid: uuid
        :return: Module instances
        :rtype: WvhdbVLinkInstance | None
        """
        if uuid in self.__VirtualLink.keys():
            return self.__VirtualLink[uuid]
        return None

    def getVlinkNameByUUID(self, uuid):
        vl = self.getVlinkByUUID(uuid)
        if vl is None:
            return None
        else:
            #: allowing us to easily format the names if we want
            return vl.getName()

    def addVLinkfiletb(self, vlblk, prefix=None, hasDelVL=False):
        """
        set a the vlink source.

        :param bustblk: Bus source
        :type bustblk: WbhdbTopBlock | None
        :param deletedVl: boolean to decide if we take deleted vl
        :type deletedVl: Bool
        """
        if hasDelVL:
           for vl in self.deletedVl:
               self.removeVLink(vl)
        if vlblk not in [x[0] for x in self.__vlink]:
            if not (vlblk, prefix) in self.__vlink:
                self.__vlink.append((vlblk, prefix))
                self.__VirtualLink.update(self._readVLink(vlblk))

    def getGlobalTypes(self):
        """
        return the list of the global types defined in the icd view
        imported from Tr View
        """
        return {gb: desc for gb, desc in self.__globalTypes}

    def removeVLink(self, vlink):
        """
        Remove a vlink from the source.

        :param vlink: Virtual Link Instance
        :type vlink: WvhdbVLinkInstance
        """
        for vlblk, _ in self.__vlink:
            for inst in vlblk.getInstances():
                if inst == vlink:
                    vlblk.removeInstance(inst)
                    self.deletedVl.append(inst)

    def _readVLink(self, vlinkblk):
        """
            Read the UUID-model mappings from the given Bus.

            :param vlinkblk: Bus design
            :type vlinkblk: WbhdbTopBlock
            :return: Mappings from UUIDs to models
            :rtype: dict[UUID, WbhdbModelInstance]
        """

        def readBlock(block):
            for inst in block.getInstances():
                if isinstance(inst, WhdbBlockInstanceP):
                    readBlock(inst.getBlock())
                elif isinstance(inst, WvhdbVLinkInstance):
                    bus[inst.getUUID()] = inst
        bus = {}
        readBlock(vlinkblk)

        # geting bus files to
        icdf = vlinkblk.getICDFile()
        self.__vlinkMessageBinding = vlinkblk.getVlinkMessageBinds()
        icdmapm = WvmodMappingManager(vlinkblk, self.__xctx)

        if icdf:
            if isinstance(icdf, str):
                ifile = WcmnFile(icdf)
            else:
                ifile = icdf
            if not scmn_isfile(ifile.path):
                raise ValueError('Imported Virtual link import '
                                 'non existing ICD view')
            bprojectDB = wprj_import_project(ifile,
                                             self.__xctx.createContextWith())
            if bprojectDB is not None:
                module = wmod_getModuleForPdb(bprojectDB)
                if not (module.code == WIMOD_ICD_APP_CODE):
                    raise ValueError('Imported Virtual link import '
                                     'non invalid ICD view')
                icdtblk = bprojectDB.getTopHierarchy()

                icdtblk.setSourcePath(ifile.path)
                icdmapm.setICDfiletb(icdtblk)
                self.__allMessages = icdmapm.getAllMessages()
                self.__globalTypes = icdmapm.getGlobalTypes()
                self.addModules(icdtblk)
        return bus

    def addModules(self, tblk):
        self.__modules = [mod for mod in tblk.getInstances()]
        self.__modulesByName = {mod.getName(): mod
                                for mod in tblk.getInstances()}

    def getAllICDMessages(self):
        return self.__allMessages

    def vlcanBeBinded(self, vlink, ncon):
        """
        return true if the given vl can accept more connection than the given
        number.
        """
        dest=vlink.getDestNumber()
        if dest < 1:
            return True
        return ncon < dest

    def getModulesfromVlink(self, vlinkuuid):
        """
        return the full description of messages from imported traffic route
        """
        modules = []
        for vlinktblk, _ in self.__vlink:
            for vluuid, message in vlinktblk.getVlinkMessageBinds():
                if vluuid == vlinkuuid:
                    for key in message:
                        for messageName, prodCons in message[key]:
                            moduleName = messageName.split('.')[0]
                            if moduleName in [mod.getName()
                                              for mod in self.getAllModuleDesc()]:
                                modules.append(self.__modulesByName[moduleName])
        return set(modules)

    def getMessagesfromVlink(self, vlinkuuid):
        """
        return the full description of messages from imported traffic route
        """
        messages = {}
        for vlinktblk, _ in self.__vlink:
            for vluuid, message in vlinktblk.getVlinkMessageBinds():
                if vluuid == vlinkuuid:
                    messages.update(message)
        return messages

    def getAllModuleDesc(self):
        """
        return the list of tuple used for the application list generation
        """
        return self.__modules

    def getModuleFromName(self, moduleName):
        """
        return the module corresponding to the uuid
        """
        return self.__modulesByName[moduleName]

    def getProdModForVlink(self, vlinkuuid):
        """
        return the port description of module producing for the unique message
        binded to vlink
        """
        producersModule = None
        for _, vlUuid in self.__topBlock.getProdBinds():
            if vlinkuuid == vlUuid:
                messages = self.getMessagesfromVlink(vlinkuuid)
                for key, message in messages.items():
                    for messageName, prodCons in message:
                        if prodCons == 'P':
                            moduleName = messageName.split('.')[0]
                            if producersModule and moduleName != producersModule:
                                raise Exception('A traffic route can\'t produce 2 '
                                                'messages from 2 different modules')
                            else:
                                producersModule = moduleName
        return producersModule

    def getConsModsForVlink(self, vlinkuuid):
        """
        return the list of consumers applications description of the message
        binded to the vlink
        """
        consumersmodules=[]
        for _, vlUuid in self.__topBlock.getConsBinds():
            if vlinkuuid == vlUuid:
                messages = self.getMessagesfromVlink(vlinkuuid)
                for key, message in messages.items():
                        for messageName, prodCons in message:
                            if prodCons == 'C':
                                moduleName = messageName.split('.')[0]
                                consumersmodules.append(moduleName)

        return consumersmodules

    def getVlMessageBinding(self):
        return self.__vlinkMessageBinding

    def getMessagesOfVlink(self, vlinkuuid):
        """
        return the list of the messages bound to the vlink
        """
        allVl = self.getVlMessageBinding()
        bindedMessages = []
        for vl, messages in allVl:
            if vl == vlinkuuid:
                bindedMessages.append(messages)
        # if nothing found return empty list
        return bindedMessages

    def FileFormatName(self, op, srcfilepath):
        name = scmn_path_base(srcfilepath) + '_' + op.getName()
        name = name.replace('.', '_')
        name = name.replace('_xml', '')
        return name
