# ----------------------------------------------------------------------------
# - wpctl-controllers                                                        -
# - wild cruncher: system controllers            core / system / controllers -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - author    (c) 2014-2014 samuel longchamps                                -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpctl_controllers
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - System controllers

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
from wcr_core.controller.wctl_controller_mixins import *
from wcr_core.controller.wctl_default_controllers import \
    WctlSimpleControllerFactoryBase
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_TYPE, \
    WHDB_XPOINT_FACING
from wcr_core.db.whdb_instance import WhdbInstanceP
from wcr_core.db_api.whdb_connection_api import WhdbConnectionP
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.system.db.wphdb_phybus_instance import \
    WphdbPhyBusInstance


class WpctlControllerBase(WctlControllerP):
    def __init__(self, fcty):
        """
        :param fcty: Controller factory
        :type fcty: WpctlControllerFactory
        """
        self.__factory = fcty

    @property
    def factory(self):
        return self.__factory


class WpctlInstanceController(WpctlControllerBase,
                              WctlInstanceControllerDefaultXPointMixin,
                              WctlInstanceControllerDefaultDeleteMixin,
                              WctlInstanceControllerP):
    def __init__(self, appModule, fcty):
        """
        :param appModule: Application module
        :type appModule: WpmodSystemApplicationModuleP
        :param fcty: Controller factory
        :type fcty: WpctlControllerFactory
        """
        super().__init__(fcty)

        self.__appModule = appModule

    def createConnectionPoint(self, instance, valuePrv, **kwargs):
        """
        :type instance: WhdbInstanceP
        :type valuePrv: WctlInstanceControllerValueProviderP
        :type kwargs: dict[str, object]
        :rtype: WhdbConnectionPointP
        """
        xpntTypes = list(self.__appModule.getConnectionPointTypes())

        if WHDB_XPOINT_TYPE not in kwargs:
            ctyp = valuePrv.provideNewConnectionPointType(instance, **kwargs)

            if ctyp is None:
                return None

            ctps = (c.getCode() for c in xpntTypes)
            if ctyp not in ctps:
                raise ValueError(
                    'Invalid connection point type {}, '
                    'should be one of {}'.format(
                        ctyp, [c.getCode() for c in xpntTypes])
                )
            kwargs[WHDB_XPOINT_TYPE] = ctyp
        else:
            ctyp = kwargs[WHDB_XPOINT_TYPE]

        ctyp = next(c for c in xpntTypes if c.getCode() == ctyp)
        dfltArgs = ctyp.getDefaultCreationArguments()

        if dfltArgs:
            for k, v in dfltArgs.items():
                if k not in kwargs:
                    kwargs[k] = v

        if 'fcng' in kwargs:
            kwargs[WHDB_XPOINT_FACING] = kwargs['fcng']

        assert self.canCreateConnectionPoint(instance, **kwargs)

        return instance.createConnectionPoint(**kwargs)


class WpctlConnectionController(WpctlControllerBase,
                                WctlConnectionControllerDefaultCreateMixin,
                                WctlConnectionControllerDefaultDeleteMixin,
                                WctlConnectionControllerP):
    def __init__(self, fcty):
        """
        :param fcty: Controller factory
        :type fcty: WpctlControllerFactory
        """
        super().__init__(fcty)

    def deleteItem(self, item, block):
        """
        :type item: WhdbBaseP
        :type block: WhdbBlock
        :rtype: (Iterable[WhdbBaseP], Iterable[WhdbBaseP])
        """
        assert isinstance(item, WhdbConnectionP)

        sobj = item.getStartObject()
        eobj = item.getEndObject()

        sCanRm = sobj.canRemoveConnectionPoint(item.getStartConnectionPoint())
        eCanRm = eobj.canRemoveConnectionPoint(item.getEndConnectionPoint())

        if sCanRm and isinstance(sobj, WphdbPhyBusInstance):
            sobj.removeConnectionPoint(item.getStartConnectionPoint())
        if eCanRm and isinstance(eobj, WphdbPhyBusInstance):
            eobj.removeConnectionPoint(item.getEndConnectionPoint())
        block.remove(item)

        return (item,), (sobj, eobj)

    def canStartConnectionAt(self, block, instance, xpoint=None):
        """
        :type block: WhdbBlock
        :type instance: WhdbInstanceP
        :type xpoint: WhdbConnectionPointP | None
        :rtype: bool
        """
        if isinstance(instance, WphdbPhyBusInstance) and not xpoint:
            return True
        if not (instance and xpoint):
            return False

        return xpoint in instance.getConnectionPoints()

    def canCreateConnection(self, block, sinst, einst, sxpnt=None, expnt=None,
                            **kwargs):
        """
        :type block: WhdbBlock
        :type sinst: WhdbInstanceP
        :type einst: WhdbInstanceP
        :type sxpnt: WhdbConnectionPointP | None
        :type expnt: WhdbConnectionPointP | None
        :rtype: bool
        """
        sbus = isinstance(sinst, WphdbPhyBusInstance)
        ebus = isinstance(einst, WphdbPhyBusInstance)
        if sbus == ebus or not sinst or not einst:
            return False

        host = einst if sbus else sinst
        for hc in block.getConnectionsForObject(host):
            strObj, strConPt, endConPt = (hc.getStartObject(),
                                          hc.getStartConnectionPoint(),
                                          hc.getEndConnectionPoint())

            hostConPt = strConPt if strObj is host else endConPt
            if hostConPt is (expnt if sbus else sxpnt):
                return False

        if sbus:
            return einst and expnt and expnt in einst.getConnectionPoints()
        else:
            return sinst and sxpnt and sxpnt in sinst.getConnectionPoints()

    def createConnection(self, block, sinst, einst, sxpnt=None, expnt=None,
                         **kwargs):
        """
        :type block: WhdbBlock
        :type sinst: WhdbInstanceP
        :type einst: WhdbInstanceP
        :type sxpnt: WhdbConnectionPointP | None
        :type expnt: WhdbConnectionPointP | None
        :rtype: (WhdbConnectionP, Iterable[WhdbBaseP])
        """
        assert self.canCreateConnection(
            block, sinst, einst, sxpnt, expnt, **kwargs)

        sbus = isinstance(sinst, WphdbPhyBusInstance)
        ebus = isinstance(einst, WphdbPhyBusInstance)
        if sbus == ebus:
            raise ValueError('One of the connection end-points must be a bus')

        if sbus:
            if not expnt:
                raise ValueError('End connection point must be provided')
            sxpnt = sinst.createConnectionPoint(cdir=WDBS_XPNT_BI,
                                                ctyp=expnt.getEntityType())
        else:
            if not sxpnt:
                raise ValueError('Start connection point must be provided')
            expnt = einst.createConnectionPoint(cdir=WDBS_XPNT_BI,
                                                ctyp=sxpnt.getEntityType())

        connection = block.createConnection(sinst, sxpnt, einst, expnt)

        return connection, (sinst, einst)


class WpctlControllerFactory(WctlSimpleControllerFactoryBase):
    def __init__(self, appModule):
        """
        :param appModule: Application module
        :type appModule: WpmodSystemApplicationModuleP
        """
        super().__init__(
            instCtlr=WpctlInstanceController(appModule, self),
            connCtlr=WpctlConnectionController(self))
