# ----------------------------------------------------------------------------
# - wphdb-block                                                              -
# - wild cruncher:sys block                               core / system / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - author    (c) 2014-2014 samuel longchamps                                -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wphdb_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher silkan system mapper block
.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
from abc import abstractmethod

from wcr_core.db.whdb_block import WhdbBlock
from wcr_core.db_api.whdb_model_api import WHDB_DEFAULT_MODEL_CODE
from wcr_core.db_api.whdb_block_api import WhdbBlockP
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaDescription, \
    WdbsMetaField
from wcr_core.dbstore.wdbs_meta_type import WDBS_PATH_META
from wcr_core.modules.system.db.wphdb_standard_operator \
    import WPHDB_HOST_CODE, WPHDB_BUS_CODE


WPHDB_CLUSTER_BLOCK_CODE = 'wcr.system.block-cluster'


# noinspection PyAbstractClass
class WphdbBlockP(WhdbBlockP):
    """
    Abstract base class for system mapper blocks.
    """

    @abstractmethod
    def createClusterInstance(self, *args, **kwargs):
        """
        Create a new cluster block instance.

        For arguments, see `WhdbBlock.createBlockInstance()`

        :param args: Arguments
        :param kwargs: Keywork-arguments
        :return: Newly created cluster block instance
        :rtype: WhdbBlockInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def createHostInstance(self, *args, **kwargs):
        """
        Create a new host instance.

        For arguments, see `WhdbBlock.createOperatorInstance()`
        (note that the operator is automatically inserted).

        :param args: Arguments
        :param kwargs: Keywork-arguments
        :return: Newly created cluster block instance
        :rtype: WphdbHostInstance
        """
        raise NotImplementedError()

    @abstractmethod
    def createPhyBusInstance(self, *args, **kwargs):
        """
        Create a new bus instance.

        For arguments, see `WhdbBlock.createOperatorInstance()`
        (note that the operator is automatically inserted).

        :param args: Arguments
        :param kwargs: Keywork-arguments
        :return: Newly created bus instance
        :rtype: WphdbPhyBusInstance
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WphdbBlockBase(WhdbBlock, WphdbBlockP):
    """
    Base class for system mapper blocks.
    """
    META = WdbsMetaDescription(
        WhdbBlock.META)

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbBlock.__init__(self, block, wreg)

    def createBlockInstance(self, *args, **kwargs):
        btyp = kwargs.get('btyp')
        if btyp and btyp != WPHDB_CLUSTER_BLOCK_CODE:
            raise ValueError('Invalid block type; {}'.format(btyp))
        else:
            kwargs['btyp'] = WPHDB_CLUSTER_BLOCK_CODE
        return super().createBlockInstance(*args, **kwargs)

    def _createRawModelInstance(self, *args, **kwargs):
        kwargs['mtyp'] = WHDB_DEFAULT_MODEL_CODE
        return super()._createRawModelInstance(*args, **kwargs)

    def createProxy(self):
        from .wphdb_block_proxy import WphdbBlockProxy

        return WphdbBlockProxy(self)

    def createClusterInstance(self, *args, **kwargs):
        kwargs['btyp'] = WPHDB_CLUSTER_BLOCK_CODE

        return super().createBlockInstance(*args, **kwargs)

    def createHostInstance(self, *args, **kwargs):
        return super().createOperatorInstance(
            WPHDB_HOST_CODE, *args, **kwargs)

    def createPhyBusInstance(self, *args, **kawrgs):
        return super().createOperatorInstance(
            WPHDB_BUS_CODE, *args, **kawrgs)


WPHDB_LOGICAL_SIM_PATH = 'logical-sim-path'


class WphdbTopBlock(WphdbBlockBase):
    META = WdbsMetaDescription(
        WphdbBlockBase.META,
        WdbsMetaField(WPHDB_LOGICAL_SIM_PATH, WDBS_PATH_META))

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        super().__init__(block, wreg)

    def getTypeName(self):
        return 'top-level'

    def setLogicalSimulationFile(self, file):
        """
        Set the path of the file for the logical simulation associated with
        the system-mapper hierarchy.

        :param file: Path of the file for the logical simulation
        :type file: WcmnFile
        """
        self.rawElement.setMeta(file, WPHDB_LOGICAL_SIM_PATH)

    def getLogicalSimulationFile(self):
        """
        Get the path of the file for the logical simulation associated with
        the system-mapper hierarchy.

        :return: Path of the file for the logical simulation
        :rtype: WcmnFile
        """
        return self.rawElement.getMeta(WPHDB_LOGICAL_SIM_PATH)

    def getPortNames(self):
        return {}


class WphdbClusterBlock(WphdbBlockBase):
    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        super().__init__(block, wreg)

    def getTypeName(self):
        return 'cluster'
