# ----------------------------------------------------------------------------
# - wphdb-connection-point-base                                              -
# - wild cruncher: connection point base                  core / system / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wphdb_connection_point_base
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Connection points base classes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.db.whdb_connection_point import WhdbConnectionPoint
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_NAME


WPHDB_XPOINT_CONFIG_KEY = 'config'


class WphdbConnectionPointBase(WhdbConnectionPoint):
    def __init__(self, xpoint, wreg):
        """
        :param xpoint: Raw connection point
        :type xpoint: WdbsConnectionPointP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbConnectionPoint.__init__(self, xpoint, wreg)

    def setName(self, name):
        """
        Set the connection point's name.

        :param name: New name
        :type name: str
        """
        self.rawElement.setMeta(name, WHDB_XPOINT_NAME)

    def _getConfig(self, *keys):
        """
        Get the configuration value for the given key sequence.

        :param keys: Keys
        :type keys: tuple[object]
        :return: Configuration value
        :rtype: object
        """
        return self.rawElement.getMeta(WPHDB_XPOINT_CONFIG_KEY, *keys)

    def _setConfig(self, value, *keys):
        """
        Set the configuration value for the given key sequence.

        :param value: New configuration value
        :type value: object
        :param keys: Keys
        :type keys: tuple[object]
        """
        self.rawElement.setMeta(value, WPHDB_XPOINT_CONFIG_KEY, *keys)

    @abstractmethod
    def getNamePrefix(self):
        """
        Get the prefix for connection point names.

        :return: Name prefix
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getTypeName(self):
        raise NotImplementedError()

    @abstractmethod
    def getDescription(self, **kwargs):
        raise NotImplementedError()


class WphdbConnectionPointSimpleBase(WphdbConnectionPointBase):
    def __init__(self, tnam, prefix, xpoint, wreg):
        """
        :param tnam: Connection point type name
        :type tnam: str
        :param prefix: Connection point name prefix
        :type prefix: str
        :param xpoint: Raw connection point
        :type xpoint: WdbsConnectionPointP
        :param wreg: Raw element wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WphdbConnectionPointBase.__init__(self, xpoint, wreg)

        #: Connection point type name
        self.__xpntTypeName = tnam
        #: Connection point name prefix
        self.__prefix = prefix

    def getNamePrefix(self):
        return self.__prefix

    def getTypeName(self):
        return '{} connection point'.format(self.__xpntTypeName)

    def getDescription(self, **kwargs):
        return '{n} ({t})'.format(n=self.getName(), t=self.__xpntTypeName)
