# ----------------------------------------------------------------------------
# - wphdb-host--instance                                                     -
# - wild cruncher: host instances                         core / system / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wphdb_host_instance
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher - Host instance

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from uuid import UUID

from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_DIRECTION, \
    WHDB_XPOINT_NAME
from wcr_core.dbstore.wdbs_util import wdbs_makeUUID
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.system.db.wphdb_host_instance_api import *
from wcr_core.modules.system.db.wphdb_standard_operator import WPHDB_HOST
from wcr_core.modules.system.db.wphdb_host_meta import *


_WPHDB_XPOINT_NAME_PATTERN = re.compile(r"^(\w+)-(\d+)$")


class WphdbHostInstance(WhdbOperatorInstance):
    """
    Operator-instance for host instances.
    """

    META = WPHDB_HOST_META_DESCRIPTION

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        WhdbOperatorInstance.__init__(self, WPHDB_HOST, inst, wreg)

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        if WHDB_XPOINT_DIRECTION not in kwargs:
            kwargs[WHDB_XPOINT_DIRECTION] = WDBS_XPNT_BI
        return super()._createRawConnectionPoint(ctyp, **kwargs)

    def createConnectionPoint(self, ctyp=None, **kwargs):
        nameGiven = bool(kwargs.get(WHDB_XPOINT_NAME))

        xpnt = super().createConnectionPoint(ctyp, **kwargs)
        """:type: WphdbConnectionPointBase"""
        if not nameGiven:
            prfx = xpnt.getNamePrefix()
            mxnr = 0
            for xpnt in self.getConnectionPoints(ctyp=ctyp):
                m = _WPHDB_XPOINT_NAME_PATTERN.match(xpnt.getName())
                if m and m.group(1) == prfx:
                    mxnr = max(int(m.group(2)), mxnr)
            xpnt.setName('{p}-{n}'.format(p=prfx, n=mxnr + 1))
        return xpnt

    def addProcess(self, name=None, uuid=None):
        """
        Create a new process on the host.

        :param name: Name of the process
        :type name: str
        :param uuid: UUID of the process
        :type uuid: UUID | None
        :return: Newly created process
        :rtype: WphdbProcessP
        """
        uuid = wdbs_makeUUID(uuid)
        self.rawElement.setMeta(
            {WPHDB_HOST_PROCESS_NAME_KEY: name or 'Process {}'.format(uuid)},
            WPHDB_HOST_PROCESSES_KEY, uuid)
        return self._makeProcess(uuid)

    def removeProcess(self, process):
        """
        Destroy a process attached to the host.

        :param process: Process to destroy (or its UUID)
        :type process: WphdbProcessP | UUID
        """
        self.rawElement.removeMeta(
            WPHDB_HOST_PROCESSES_KEY, _wphdb_getUuid(process))

    def getRawProcesses(self):
        """
        Get the processes attached to the host.

        :return: Host's processes
        :rtype: dict[UUID, dict[str, object]]
        """
        return self.rawElement.getMeta(WPHDB_HOST_PROCESSES_KEY)

    def setRawProcesses(self, procs):
        """
        Set the processes attached to the host.

        :param procs: Host's processes
        :type procs: dict[UUID, dict[str, object]]
        """
        self.rawElement.setMeta(procs, WPHDB_HOST_PROCESSES_KEY)

    def getProcesses(self):
        """
        Get the processes attached to the host.

        :return: Host's processes
        :rtype: list[WphdbProcessP]
        """
        return [self._makeProcess(p) for p in
                self.rawElement.getMetaKeys(WPHDB_HOST_PROCESSES_KEY)]

    def getProcess(self, puuid):
        """
        Get the process with the given UUID.

        :param puuid: UUID of process UUID
        :type puuid: UUID
        :return: Proces with the given UUID, or `None`
        :rtype: WphdbProcessP | None
        """
        if self.rawElement.getMeta(WPHDB_HOST_PROCESSES_KEY, puuid):
            return self._makeProcess(puuid)
        else:
            return None

    def addModelMapping(self, puuid, modelUuid, name):
        """
        Add a model mapping to one of the host's processes.

        :param puuid: UUID of process to add model to
        :type puuid: UUID
        :param modelUuid: UUID of model to be added
        :type modelUuid: UUID
        :param name: Name of model to be added
        :type name: str
        :return: UUID of new model
        :rtype: UUID
        """
        self.rawElement.setMeta(
            {WPHDB_HOST_MODEL_NAME_KEY: name},
            WPHDB_HOST_PROCESSES_KEY, puuid,
            WPHDB_HOST_MODEL_MAPS_KEY, modelUuid)
        return modelUuid

    def removeModelMapping(self, puuid, muuid):
        """
        Remove a model from one of the host's processes.

        :param puuid: UUID of process to remove model from
        :type puuid: UUID
        :param muuid: UUID of model to be removed
        :type muuid: UUID
        """
        self.rawElement.removeMeta(
            WPHDB_HOST_PROCESSES_KEY, puuid, WPHDB_HOST_MODEL_MAPS_KEY, muuid)

    def getModelUuids(self, puuid):
        """
        Get all models attached to the given process.

        :param puuid: UUID of process to retrieve models for
        :type puuid: UUID
        :return: UUIDs of models attached to the given process
        :rtype: Iterable[UUID]
        """
        uuid = _wphdb_getUuid(puuid)
        return self.rawElement.getMetaKeys(
            WPHDB_HOST_PROCESSES_KEY, uuid, WPHDB_HOST_MODEL_MAPS_KEY)

    def addServiceMapping(self, puuid, service, label=None):
        """
        Add a service instance to one of the host's processes.

        :param puuid: UUID of process to add service instance to
        :type puuid: UUID
        :param service: Name of instantiated service
        :type service: str
        :param label: Name of service instance to be added
        :type label: str | None
        :return: UUID of new service
        :rtype: UUID
        """
        suuid = wdbs_makeUUID()
        if not label:
            label = '{} Instance {}'.format(service, suuid)
        self.rawElement.setMeta(
            {WPHDB_HOST_SERVICE_MAP_LABEL_KEY: label,
             WPHDB_HOST_SERVICE_KEY: service},
            WPHDB_HOST_PROCESSES_KEY, puuid,
            WPHDB_HOST_SERVICE_MAPS_KEY, suuid)
        return suuid

    def removeServiceMapping(self, puuid, suuid):
        """
        Remove a service instance from one of the host's processes.

        :param puuid: UUID of process to remove service instance from
        :type puuid: UUID
        :param suuid: UUID of service instance to be removed
        :type suuid: UUID
        """
        self.rawElement.removeMeta(
            WPHDB_HOST_PROCESSES_KEY, puuid,
            WPHDB_HOST_SERVICE_MAPS_KEY, suuid)

    def getServiceMappingUuids(self, puuid):
        """
        Get all service instances attached to the given process.

        :param puuid: UUID of process to retrieve service instances for
        :type puuid: UUID
        :return: UUIDs of service instances attached to the given process
        :rtype: Iterable[UUID]
        """
        return self.rawElement.getMetaKeys(
            WPHDB_HOST_PROCESSES_KEY, puuid, WPHDB_HOST_SERVICE_MAPS_KEY)

    def addDistributionServiceSocket(self, puuid, service, label=None):
        """
        Add a distribution service instance to one of the host's processes.

        :param puuid: UUID of process to add service instance to
        :type puuid: UUID
        :param service: Code of distribution service
        :type service: str
        :param label: Name of service instance to be added
        :type label: str | None
        :return: UUID of distribution service registration
        :rtype: UUID
        """
        duuid = wdbs_makeUUID()
        if not label:
            label = '{} Instance {}'.format(service, duuid)
        self.rawElement.setMeta(
            {WPHDB_HOST_DISTRIBSVC_SCKT_LABEL_KEY: label,
             WPHDB_HOST_DISTRIBSVC_KEY: service},
            WPHDB_HOST_PROCESSES_KEY, puuid,
            WPHDB_HOST_DISTRIBSVC_SCKTS_KEY, duuid)
        return duuid

    def removeDistributionServiceSocket(self, puuid, suuid):
        """
        Remove a distribution service instance from one of the host's
        processes.

        :param puuid: UUID of process to remove distribution service instance
                      from
        :type puuid: UUID
        :param suuid: UUID of distribution service instance to be removed
        :type suuid: UUID
        """
        self.rawElement.removeMeta(
            WPHDB_HOST_PROCESSES_KEY, puuid,
            WPHDB_HOST_DISTRIBSVC_SCKTS_KEY, suuid)

    def getDistributionServiceSocketUuids(self, puuid):
        """
        Get all ditribution service instances linked to the given process.

        :param puuid: UUID of process to retrieve distribution service
                      instance UUIDs for
        :type puuid: UUID
        :return: Distribution service instance UUIDs
        :rtype: Iterable[UUID]
        """
        return self.rawElement.getMeta(
            WPHDB_HOST_PROCESSES_KEY, puuid,
            WPHDB_HOST_DISTRIBSVC_SCKTS_KEY)

    def _makeProcess(self, uuid):
        """
        Create a process object for a given process UUID.

        :param uuid: UUID of process UUID to create process object for
        :type uuid: UUID | None
        :return: Process object for the given UUID
        :rtype: WphdbProcess | None
        """
        if uuid is None:
            return None
        return WphdbProcess(uuid, self)

    def getCards(self):
        """
        Get the cards of the host.

        :return: Cards of the host
        :rtype: list[WphdbCard]
        """
        return map(self._makeCard,
                   self.rawElement.getMetaKeys(WPHDB_HOST_CARDS_KEY))

    def _makeCard(self, uuid):
        """
        Create a card object for the given card UUID.

        :param uuid: Card UUID to create card object for
        :type uuid: UUID | None
        :return: Card object for the given UUID
        :rtype: WphdbCard | None
        """
        if uuid is None:
            return None
        return WphdbCard(uuid, self)

    def getCPU(self):
        """
        Get the host's CPU.

        :return: CPU of the host
        :rtype: str
        """
        return self.rawElement.getMeta(WPHDB_HOST_CAP_CPU_KEY)

    def setCPU(self, cpu):
        """
        Set the host's CPU.

        :param cpu: CPU of the host
        :type cpu: str
        """
        self.rawElement.setMeta(cpu, WPHDB_HOST_CAP_CPU_KEY)

    def getCores(self):
        """
        Get the host's CPU.

        :return: CPU of the host
        :rtype: int
        """
        return self.rawElement.getMeta(WPHDB_HOST_CAP_CORES_KEY)

    def setCores(self, cores):
        """
        Set the host's cores.

        :param cores: Cores of the host
        :type cores: int
        """
        self.rawElement.setMeta(cores, WPHDB_HOST_CAP_CORES_KEY)

    def getGPU(self):
        """
        Get the host's GPU.

        :return: GPU of the host
        :rtype: str
        """
        return self.rawElement.getMeta(WPHDB_HOST_CAP_GPU_KEY)

    def setGPU(self, gpu):
        """
        Set the host's GPU.

        :param gpu: GPU of the host
        :type gpu: str
        """
        self.rawElement.setMeta(gpu, WPHDB_HOST_CAP_GPU_KEY)

    def getMemory(self):
        """
        Get the host's memory.

        :return: Memory of the host
        :rtype: str
        """
        return self.rawElement.getMeta(WPHDB_HOST_CAP_MEMORY_KEY)

    def setMemory(self, memory):
        """
        Set the host's memory.

        :param memory: Memory of the host
        :type memory: str
        """
        self.rawElement.setMeta(memory, WPHDB_HOST_CAP_MEMORY_KEY)

    def getDiskSpace(self):
        """
        Get the host's disk space.

        :return: Disk space of the host
        :rtype: str
        """
        return self.rawElement.getMeta(WPHDB_HOST_CAP_DISK_KEY)

    def setDiskSpace(self, disk):
        """
        Set the host's disk space.

        :param disk: Disk space of the host
        :type disk: str
        """
        self.rawElement.setMeta(disk, WPHDB_HOST_CAP_DISK_KEY)


class _WphdbHostElement(WphdbHostElementP):
    """
    Base class for host sub-elements.
    """

    __slots__ = ('__host', '__uuid')

    def __init__(self, uuid, host):
        """
        :param uuid: UUID of process UUID
        :type uuid: UUID
        :param host: Host the process belongs to
            """
        self.__uuid = uuid
        self.__host = host

    @property
    def uuid(self):
        return self.__uuid

    @property
    def host(self):
        return self.__host

    def _getMeta(self, *keys):
        return self.__host.rawElement.getMeta(*keys)

    def _setMeta(self, value, *keys):
        self.__host.rawElement.setMeta(value, *keys)

    def __eq__(self, other):
        return isinstance(other, WphdbHostElementP) and \
            other.uuid == self.uuid

    def __hash__(self):
        return hash(self.__uuid)


class WphdbProcess(_WphdbHostElement, WphdbProcessP):
    __slots__ = ()

    def __init__(self, uuid, host):
        """
        :param uuid: UUID of process UUID
        :type uuid: UUID
        :param host: Host the process belongs to
        :type host: WphdbHostInstance
        """
        super().__init__(uuid, host)

    ##### PROPERTIES #####

    def getName(self):
        return self._getMeta(WPHDB_HOST_PROCESS_NAME_KEY)

    def setName(self, name):
        self._setMeta(name, WPHDB_HOST_PROCESS_NAME_KEY)

    # #### MODELS #####

    def addModelMapping(self, modelUuid, name):
        muuid = self.host.addModelMapping(self.uuid, modelUuid, name)
        return WphdbModelMapping(muuid, self)

    def removeModelMapping(self, model):
        self.host.removeModelMapping(self.uuid, _wphdb_getUuid(model))

    def getModelMappings(self):
        return [WphdbModelMapping(m, self) for m
                in self.host.getModelUuids(self.uuid)]

    def getModelMapping(self, modelUuid):
        return WphdbModelMapping(modelUuid, self)

    ##### SERVICES #####

    def addServiceMapping(self, service, label=None):
        suuid = self.host.addServiceMapping(
            self.uuid, service=service, label=label)
        return WphdbServiceMapping(suuid, self)

    def removeServiceMapping(self, service):
        self.host.removeServiceMapping(self.uuid, _wphdb_getUuid(service))

    def getServiceMappings(self):
        return [WphdbServiceMapping(s, self) for s
                in self.host.getServiceMappingUuids(self.uuid)]

    def getServiceMapping(self, mappingUuid):
        return WphdbServiceMapping(mappingUuid, self)

    ##### DISTRIBUTION SEVICES #####

    def addDistributionServiceSocket(self, service, label=None):
        duuid = self.host.addDistributionServiceSocket(
            self.uuid, service=service, label=label)
        return WphdbDistributionServiceSocket(duuid, self)

    def removeDistributionServiceSocket(self, service):
        self.host.removeDistributionServiceSocket(
            self.uuid, _wphdb_getUuid(service))

    def getDistributionServiceSockets(self):
        return [WphdbDistributionServiceSocket(s, self) for s
                in self.host.getDistributionServiceSocketUuids(self.uuid)]

    def getDistributionServiceSocket(self, socketUuid):
        return WphdbDistributionServiceSocket(socketUuid, self)

    def __str__(self):
        return 'Process {} @ {}'.format(self.uuid, self.host.getName())

    def _getMeta(self, *keys):
        return super()._getMeta(WPHDB_HOST_PROCESSES_KEY, self.uuid,
                                *keys)

    def _setMeta(self, value, *keys):
        super()._setMeta(value,
                         WPHDB_HOST_PROCESSES_KEY, self.uuid,
                         *keys)


class _WphdbProcessElement(_WphdbHostElement, WphdbProcessElementP):
    __slots__ = ('__proc',)

    def __init__(self, uuid, proc):
        """
        :param uuid: UUID
        :type uuid: UUID
        :param proc: Process the element belongs to
        :type proc: WphdbProcess
        """
        super().__init__(uuid, proc.host)

        self.__proc = proc

    @property
    def process(self):
        return self.__proc

    def _getMeta(self, *keys):
        return super()._getMeta(
            WPHDB_HOST_PROCESSES_KEY, self.__proc.uuid, *keys)

    def _setMeta(self, value, *keys):
        return super()._setMeta(
            value, WPHDB_HOST_PROCESSES_KEY, self.__proc.uuid, *keys)


class WphdbModelMapping(_WphdbProcessElement, WphdbModelMappingP):
    __slots__ = ()

    def __init__(self, uuid, proc):
        """
        :param uuid: UUID
        :type uuid: UUID
        :param proc: Process the element belongs to
        :type proc: WphdbProcess
        """
        _WphdbProcessElement.__init__(self, uuid, proc)

    def getModelName(self):
        return self._getMeta(WPHDB_HOST_MODEL_NAME_KEY)

    def _getMeta(self, *keys):
        return super()._getMeta(WPHDB_HOST_MODEL_MAPS_KEY, self.uuid, *keys)


class WphdbServiceMapping(_WphdbProcessElement, WphdbServiceMappingP):
    __slots__ = ()

    def __init__(self, uuid, proc):
        """
        :param uuid: UUID
        :type uuid: UUID
        :param proc: Process the element belongs to
        :type proc: WphdbProcess
        """
        _WphdbProcessElement.__init__(self, uuid, proc)

    def getService(self):
        return self._getMeta(WPHDB_HOST_SERVICE_KEY)

    def getMappingLabel(self):
        return self._getMeta(WPHDB_HOST_SERVICE_MAP_LABEL_KEY)

    def setMappingLabel(self, name):
        self._setMeta(name, WPHDB_HOST_SERVICE_MAP_LABEL_KEY)

    def getMappingParameterValue(self, param):
        return self._getMeta(
            WPHDB_HOST_SERVICE_MAP_PARAMS_KEY, param, WHDB_PARAMETER_VALUE_KEY)

    def setMappingParameterValue(self, value, param):
        self._setMeta(value, WPHDB_HOST_SERVICE_MAP_PARAMS_KEY, param,
                      WHDB_PARAMETER_VALUE_KEY)

    def _getMeta(self, *keys):
        return super()._getMeta(WPHDB_HOST_SERVICE_MAPS_KEY, self.uuid, *keys)

    def _setMeta(self, value, *keys):
        super()._setMeta(value, WPHDB_HOST_SERVICE_MAPS_KEY, self.uuid, *keys)


class WphdbDistributionServiceSocket(_WphdbProcessElement,
                                     WphdbDistributionServiceSocketP):
    __slots__ = ()

    def __init__(self, uuid, proc):
        """
        :param uuid: UUID
        :type uuid: UUID
        :param proc: Process the element belongs to
        :type proc: WphdbProcess
        """
        _WphdbProcessElement.__init__(self, uuid, proc)

    def getDistributionService(self):
        return self._getMeta(WPHDB_HOST_DISTRIBSVC_KEY)

    def getSocketLabel(self):
        return self._getMeta(WPHDB_HOST_DISTRIBSVC_SCKT_LABEL_KEY)

    def setSocketLabel(self, label):
        self._setMeta(label, WPHDB_HOST_DISTRIBSVC_SCKT_LABEL_KEY)

    def getSocketParameterValue(self, param):
        return self._getMeta(
            WPHDB_HOST_DISTRIBSVC_SCKT_PARAMS_KEY, param,
            WHDB_PARAMETER_VALUE_KEY)

    def setSocketParameterValue(self, value, param):
        self._setMeta(value, WPHDB_HOST_DISTRIBSVC_SCKT_PARAMS_KEY, param,
                      WHDB_PARAMETER_VALUE_KEY)

    def _getMeta(self, *keys):
        return super()._getMeta(
            WPHDB_HOST_DISTRIBSVC_SCKTS_KEY, self.uuid, *keys)

    def _setMeta(self, value, *keys):
        super()._setMeta(
            value, WPHDB_HOST_DISTRIBSVC_SCKTS_KEY, self.uuid, *keys)


class WphdbCard(_WphdbHostElement, WphdbCardP):
    __slots__ = ()

    def __init__(self, uuid, host):
        super().__init__(uuid, host)

    def getName(self):
        return self._getMeta(WPHDB_HOST_CARD_NAME_KEY)

    def setName(self, name):
        self._setMeta(name, WPHDB_HOST_CARD_NAME_KEY)

    def getType(self):
        return self._getMeta(WPHDB_HOST_CARD_TYPE_KEY)

    def setType(self, ctyp):
        self._setMeta(ctyp, WPHDB_HOST_CARD_TYPE_KEY)

    def getChannels(self):
        """
        Get the card's channels.

        :return: Card's channels [(UUID, Name, Type)]
        :rtype: list[tuple[UUID, str, str]]
        """
        channels = self._getMeta(WPHDB_HOST_CARD_CHANNELS_KEY)
        return ((k, c.get(WPHDB_HOST_CHANNEL_NAME_KEY),
                 c.get(WPHDB_HOST_CHANNEL_TYPE_KEY))
                for k, c in channels.items())

    def __str__(self):
        return 'Card {} @ {}'.format(self.getName(), self.__host.getName())

    def _getMeta(self, *keys):
        return super()._getMeta(WPHDB_HOST_CARDS_KEY, self.uuid, *keys)

    def _setMeta(self, value, *keys):
        return super()._setMeta(value, WPHDB_HOST_CARDS_KEY, self.uuid, *keys)


def _wphdb_getUuid(item):
    return item if isinstance(item, UUID) else item.uuid
