# ----------------------------------------------------------------------------
# - wphdb-host-instance-api                                                  -
# - wild cruncher: host instance api                      core / system / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wphdb_host_instance_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Host instance API

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod


class WphdbHostElementP(metaclass=ABCMeta):
    """
    Base class for host sub-elements.
    """

    __slots__ = ()

    @property
    @abstractmethod
    def uuid(self):
        """
        Get the process' UUID.

        :return: Process UUID
        :rtype: UUID
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def host(self):
        """
        Get the process' host.

        :return: Process host
        :rtype: WphdbHostInstance
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WphdbProcessP(WphdbHostElementP):
    __slots__ = ()

    @abstractmethod
    def getName(self):
        """
        Get the process' name.

        :return: Name of the process
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setName(self, name):
        """
        Set the process' name.

        :param name: Process name
        :type name: str
        """
        raise NotImplementedError()

    @abstractmethod
    def addModelMapping(self, modelUuid, name):
        """
        Map a model to the process.

        :param modelUuid: Model UUID
        :type modelUuid: UUID | None
        :param name: Model name
        :type name: str
        :return: Newly created model mapping
        :rtype: WphdbModelMappingP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeModelMapping(self, model):
        """
        Remove a model mapping from the process.

        :param model: Model mapping (or its UUID)
        :type model: WphdbModelMappingP | UUID
        """
        raise NotImplementedError()

    @abstractmethod
    def getModelMappings(self):
        """
        Get all models mapped to this process.

        :return: Models mappings for the process
        :rtype: list[WphdbModelMappingP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getModelMapping(self, modelUuid):
        """
        Get model mapping for the given model UUID.

        :param modelUuid: Model UUID
        :type modelUuid: UUID
        :return: Model mapping
        :rtype: WphdbModelMappingP
        """
        raise NotImplementedError()

    @abstractmethod
    def addServiceMapping(self, service, label=None):
        """
        Map a service to the process.

        :param service: Name of service
        :type service: str
        :param label: Name for service mapping
        :type label: str | None
        :return: Newly created service mapping
        :rtype: WphdbServiceMappingP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeServiceMapping(self, service):
        """
        Remove a service mapping from the process.

        :param service: Service mapping (or its UUID)
        :type service: WphdbServiceMappingP | UUID
        """
        raise NotImplementedError()

    @abstractmethod
    def getServiceMappings(self):
        """
        Get all services attached to the process.

        :return: List of service assignments
        :rtype: list[WphdbServiceMappingP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getServiceMapping(self, mappingUuid):
        """
        Get the service attached to the process with the given UUID.

        :param mappingUuid: Service assignment UUID
        :type mappingUuid: UUID
        :return: Service assignment with the given UUID
        :rtype: WphdbServiceMappingP
        """
        raise NotImplementedError()

    @abstractmethod
    def addDistributionServiceSocket(self, service, label=None):
        """
        Add a distribution service socket to the process.

        :param service: Name of distribution service
        :type service: str
        :param label: Socket label
        :type label: str | None
        :return: Newly created distribution service socket
        :rtype: WphdbDistributionServiceSocketP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeDistributionServiceSocket(self, service):
        """
        Remove a distribution service socket from the process.

        :param service: Distribution service socket (or its UUID)
        :type service: WphdbDistributionServiceSocketP | UUID
        """
        raise NotImplementedError()

    @abstractmethod
    def getDistributionServiceSockets(self):
        """
        Get all distribution service sockets attached to the process.

        :return: List of distribution service sockets
        :rtype: list[WphdbDistributionServiceSocketP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getDistributionServiceSocket(self, socketUuid):
        """
        Get the distribution service socket with the given UUID.

        :param socketUuid: Distribution service socket UUID
        :type socketUuid: UUID
        :return: Distribution service socket with the given UUID
        :rtype: WphdbDistributionServiceSocketP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WphdbProcessElementP(WphdbHostElementP):
    __slots__ = ()

    @property
    @abstractmethod
    def process(self):
        """
        Get the process the element belongs to.

        :return: Process
        :rtype: WphdbProcessP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WphdbModelMappingP(WphdbProcessElementP):
    __slots__ = ()

    @abstractmethod
    def getModelName(self):
        """
        Get the name of the model.

        :return: Model name
        :rtype: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WphdbServiceMappingP(WphdbProcessElementP):
    __slots__ = ()

    @abstractmethod
    def getService(self):
        """
        Get the name of the service.

        :return: Service name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getMappingLabel(self):
        """
        Get the label of the service mapping.

        :return: Service mapping label
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setMappingLabel(self, name):
        """
        Set the label of the service mapping.

        :param name: Service mapping label
        :type name: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getMappingParameterValue(self, param):
        """
        Get the value for the parameter with the given name.

        :param param: Parameter name
        :type param: str
        :return: Parameter value, if found
        :rtype: object | None
        """
        raise NotImplementedError()

    @abstractmethod
    def setMappingParameterValue(self, value, param):
        """
        Set the value for the parameter with the given name.

        :param value: New parameter value
        :type value: object | None
        :param param: Parameter name
        :type param: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WphdbDistributionServiceSocketP(WphdbProcessElementP):
    __slots__ = ()

    @abstractmethod
    def getDistributionService(self):
        """
        Get the name of the distribution service.

        :return: Distribution service name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getSocketLabel(self):
        """
        Get the label of the distribution service socket.

        :return: Distribution service socket label
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setSocketLabel(self, label):
        """
        Set the label of the distribution service socket.

        :param label: Distribution service socket name
        :type label: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getSocketParameterValue(self, param):
        """
        Get the value for the parameter with the given name.

        :param param: Parameter name
        :type param: str
        :return: Parameter value, if found
        :rtype: object | None
        """
        raise NotImplementedError()

    @abstractmethod
    def setSocketParameterValue(self, value, param):
        """
        Set the value for the parameter with the given name.

        :param value: New parameter value
        :type value: object | None
        :param param: Parameter name
        :type param: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WphdbCardP(WphdbHostElementP):
    __slots__ = ()

    @abstractmethod
    def getName(self):
        """
        Get the name of the card.

        :return: Card's name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getType(self):
        """
        Get the type of the card.

        :return: Card's type
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getChannels(self):
        """
        Get the card's channels.

        :return: Card's channels [(UUID, Name, Type)]
        :rtype: list[tuple[UUID, str, str]]
        """
        raise NotImplementedError()
