# ----------------------------------------------------------------------------
# - wphdb-verifier                                                           -
# - wild cruncher: database verifier                      core / system / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wphdb_verifier
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Database verifier

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
from wcr_core.common.wcmn_verifier_module import WcmnHierarchyVerificationError
from wcr_core.db.whdb_verifier import WhdbHierarchyVerifier, \
    WhdbVerifierModuleP
from wcr_core.modules.system.db.wphdb_block import WphdbTopBlock
from wcr_core.modules.system.db.wphdb_host_instance import WphdbHostInstance
from wcr_core.modules.system.db.wphdb_phybus_instance import \
    WphdbPhyBusInstance
from wcr_core.modules.system.wpmod_distrib_service_default_types import \
    wpmod_retrieve_conpt_ethdistsrv
from wcr_core.modules.system.wpmod_project_context import \
    wpmod_retrieve_lbus_obj


class WphdbOneHostPerModelVerifierModule(WhdbVerifierModuleP):
    """
    Verify if all models in the simulation have exactly one host set.
    If no simulation is set, no error should be produced from this verifier.
    """

    ERROR_ID = 'hierarchy.system.one-host-model'

    ERROR_MESSAGE_ZERO = "The model '{}' is not mapped to any host"

    ERROR_MESSAGE_MORE_ONE = "The model '{}' is mapped to more than one host"

    DESCRIPTION = 'All models must be mapped to exactly one host.'

    def __init__(self, pctx):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        """
        self.__pctx = pctx

    def verifyBlock(self, block, topLevel=False):
        """
        :type block: WhdbBlockP
        :type topLevel: bool
        :return: List of errors, one for each incorrectly mapped model.
            Offenders are the hosts to which the model is connected when it is
            mapped to more than one, otherwise there is only one offender in
            the case of no mapping: the topblock.
            Returns None if the block is not top (or said not to be using the
            flag) or if no simulation is loaded.
        :rtype: list[WcmnHierarchyVerificationError] | None
        """
        if not topLevel or not isinstance(block, WphdbTopBlock):
            return

        sim = self.__pctx.getLogicalSimulation()
        if sim is None:
            return

        modelsDict = {mInst.getUUID(): [mInst, 0, set()] for mInst in
                      sim.getModelInstances()}
        hosts = (inst for inst in block.getInstances()
                 if isinstance(inst, WphdbHostInstance))
        processes = [proc for hstProcs in (hst.getProcesses() for hst in hosts)
                     for proc in hstProcs]
        """:type: list[WphdbProcessP]"""

        # Iterate over all models in all hosts' processes and count their
        # relations to hosts, register them as host of model as well
        for model in (model for proc in processes
                      for model in proc.getModelMappings()):
            inst = modelsDict.get(model.uuid)
            assert (inst is not None and len(inst) == 3)
            inst[1] += 1
            inst[2].add(model.host)

        errList = []
        for mod in modelsDict.values():
            if mod[1] != 1:
                if mod[1] > 1:
                    msg = self.ERROR_MESSAGE_MORE_ONE
                    offenders = list(mod[2])
                else:
                    msg = self.ERROR_MESSAGE_ZERO
                    offenders = [block]
                msg = msg.format(mod[0].getName())
                errList.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID, msg, offenders, self))

        return errList

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        pass

    def description(self):
        return self.DESCRIPTION


class WphdbLogBusBoundToPhyBusVerifierModule(WhdbVerifierModuleP):
    """
    Verify that all physical buses with at least one connection point have
    at least one bound logical bus
    """

    ERROR_ID = 'hierarchy.system.logbus-bound-to-phybus'

    ERROR_MESSAGE = "The physical bus '{}' doesn't have any bound logical bus"

    DESCRIPTION = 'All physical bus with at least one connection point must ' \
                  'have at least one bound logical bus.'

    def verifyBlock(self, block, topLevel=False):
        """
        :type block: WhdbBlockP
        :type topLevel: bool
        :return: List of errors, one for each invalid physical buses.
            Returns None if the block is not top (or said not to be using the
            flag)
        :rtype: list[WcmnHierarchyVerificationError] | None
        """
        if not topLevel or not isinstance(block, WphdbTopBlock):
            return

        errList = []
        pbuses = [opInst for opInst in block.getOperatorInstances()
                  if isinstance(opInst, WphdbPhyBusInstance)]
        """:type: list[WphdbPhyBusInstance]"""

        for pbus in pbuses:
            if len(list(pbus.getConnectionPoints())) < 1:
                continue
            if len(pbus.getBoundLogicalBusesUUID()) == 0:
                errList.append(WcmnHierarchyVerificationError(
                    self.ERROR_ID, self.ERROR_MESSAGE.format(pbus.getName()),
                    [pbus], self))

        return errList

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        pass

    def description(self):
        return self.DESCRIPTION


class WphdbConsModelsAccessProdVerifierModule(WhdbVerifierModuleP):
    """
    Verify that all models consuming a variable have access to its consumer
    """

    ERROR_ID = 'hierarchy.system.cons-access-prod'

    ERROR_MSG_ACCESS = "Physical bus '{}' doesn't have access to "

    ERROR_MSG_ACCESS_BOTH = ERROR_MSG_ACCESS + \
        "model '{}' (consumer) and model '{}' (producer) for var. '{}'"

    ERROR_MSG_ACCESS_CONS = ERROR_MSG_ACCESS + "model '{}' (consumer) " \
                                               "for var. '{}'"

    ERROR_MSG_ACCESS_PROD = ERROR_MSG_ACCESS + "model '{}' (producer) " \
                                               "for var. '{}'"

    ERROR_MSG_LBUSBIND = "Model '{}' in '{}' is missing connection to a phy. "\
                         "bus with log. bus '{}' bound to it"

    DESCRIPTION = 'A model consuming a variable must have access to the ' \
                  'producer of the said variable.'

    def __init__(self, pctx):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        """
        self.__pctx = pctx

    def verifyBlock(self, block, topLevel=False):
        """
        :type block: WhdbBlockP
        :type topLevel: bool
        :return: List of errors, one for each consumer model with missing
            link to a producer for a variable.
            Returns None if the block is not top (or said not to be using the
            flag), if the simulation is not loaded or if there are unmapped
            models.
        :rtype: list[WcmnHierarchyVerificationError] | None
        """
        if not topLevel or not isinstance(block, WphdbTopBlock):
            return

        sim = self.__pctx.getLogicalSimulation()
        if sim is None:
            return

        if len(list(self.__pctx.getMappingManager().getUnmappedModels())) != 0:
            return

        errList = []

        pbuses = self.__pctx.getPhysicalBuses()
        lbusObjDict = wpmod_retrieve_lbus_obj(
            pbuses, [self.__pctx.getLogicalSimulation()])

        errList += self._getMissingLogicalBusError(
            sim, [lb for lbuses in lbusObjDict.values() for lb in lbuses])

        errList += self._getConsAccessProdError(block, sim, lbusObjDict)

        return errList

    def _getConsAccessProdError(self, sys, sim, lbusBinding):
        """
        :param sys: System mapper top block
        :type sys: WphdbTopBlock
        :param sim: Simulation top block
        :type sim: WshdbTopBlock
        :param lbusBinding: Dictionary representing the binding of physical
            buses with logical buses instances
        :type lbusBinding: dict[UUID, list[WshdbBusInstance]]
        :return: List of errors caused by a binding of a mapped logical bus
            to be unable to access its consumer, producer, or both
        :rtype: list[WcmnHierarchyVerificationError]
        """

        errList = []
        return errList

    def _getMissingLogicalBusError(self, sim, boundLogBuses):
        """
        :param sim: Simulation top block
        :type sim: WshdbTopBlock
        :param boundLogBuses: List of bound logical buses in the project
        :type boundLogBuses: list[WshdbBusInstance]
        :return: List of errors caused by a model being mapped to a host when
            it is present in a binding as a consumer in an unbound logical bus
        :rtype: list[WcmnHierarchyVerificationError]
        """
        errList = []

        return errList

    def _isModelAccessible(self, block, pbus, model):
        """
        Check if model is accessible by finding all the connections for the
        model's host that are connected to a connection point linked to a
        distributed service of the same process as the model.
        If a connection to an appropriate connection point is found both for
        the host and the physical bus, the model is accessible.

        :type block: WhdbBlockP
        :type pbus: WphdbPhyBusInstance
        :type model: WshdbModelInstance
        :return: True if the model if a host with mapped model is connected to
            the physical bus
        :rtype: bool
        """
        process = self.__pctx.getMappingManager().getModelProcess(
            model.getUUID())

        if process:
            distSrvSck = process.getDistributionServiceSockets()
            conPts = [conPt for conPt in (wpmod_retrieve_conpt_ethdistsrv(dss)
                      for dss in distSrvSck) if conPt is not None]

            host = process.host
            pbusConns = set(block.getConnectionsForObject(pbus))
            for cp in conPts:
                connInter = pbusConns & \
                    set(block.getConnectionsForObject(host, cp))
                """:type: set[WhdbConnectionP]"""

                if len(connInter) > 0:
                    return True
        return False

    def verifyConnection(self, transition):
        pass

    def verifyInstance(self, instance):
        pass

    def description(self):
        return self.DESCRIPTION


class WphdbHierarchyVerifier(WhdbHierarchyVerifier):
    """
    System mapper hierarchy verifier.
    """

    def __init__(self, pctx, modules=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param modules: Additional verification modules to use.
        :type modules: Iterable[WhdbVerifierModuleP] | None
        """
        mods = [WphdbOneHostPerModelVerifierModule(pctx),
                WphdbLogBusBoundToPhyBusVerifierModule(),
                WphdbConsModelsAccessProdVerifierModule(pctx),
                ]
        """:type: list[WhdbVerifierModuleP]"""

        if modules:
            mods.extend(modules)

        super().__init__(mods)
