# ----------------------------------------------------------------------------
# - wpmod-application-module                                                 -
# - wild cruncher: application module                          core / system -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpmod_application_module
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher system mapper application module

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.action.wact_action_api import WACT_WCR_ACTION_CODE
from wcr_core.db.whdb_top_level_factory import whdb_createTopLevelRawBlock
from wcr_core.modules.system.db.wphdb_meta_provider import WphdbMetaProvider
from wcr_core.modules.system.db.wphdb_verifier import WphdbHierarchyVerifier
from wcr_core.modules.system.db.wphdb_wrapper_registry import \
    WphdbWrapperRegistry
from wcr_core.modules.wmod_application_module import \
    WmodApplicationModuleBase, WmodApplicationModuleP, WmodApplicationModuleMP
from wcr_core.common.wcmn_mime_type import WCMN_PAYLOAD_BLOCK_SYSTEM
from wcr_core.modules.system.wpmod_project_context import \
    WpmodSystemProjectContext
from wcr_core.modules.system.db.wphdb_operator import WPHDB_OPER
from wcr_core.modules.system.db.wphdb_standard_operator import \
    WPHDB_DEFAULT_OPS


WPMOD_SYSTEM_APP_CODE = 'system'


# noinspection PyAbstractClass
class WpmodSystemApplicationModuleP(WmodApplicationModuleP):
    """
    System application module protocol.
    """

    @abstractmethod
    def getDistributionServiceTypes(self):
        """
        Get all known distribution service types.

        :return: Distribution service types
        :rtype: Iterable[WpmodDistributionServiceTypeP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getDistributionServiceType(self, code):
        """
        Get the distribution service type with the given code, if it exists.

        :param code: Code
        :type code: str
        :return: Distribution service type
        :rtype: WpmodDistributionServiceTypeP | None
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectionPointTypes(self):
        """
        Get all registered connection point types.

        :return: Connection point types
        :rtype: Iterable[WpmodConnectionPointTypeP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectionPointType(self, code):
        """
        Get the connection point type for the given code.

        :param code: Code
        :type code: str
        :return: Connection point type for the given code
        :rtype: WpmodConnectionPointTypeP | None
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WpmodSystemApplicationModuleMP(WmodApplicationModuleMP,
                                     WpmodSystemApplicationModuleP):
    @abstractmethod
    def addDistributionServiceType(self, dsvc):
        """
        Add a distribution service type.

        :param dsvc: Distribution service type to be added
        :type dsvc: WpmodDistributionServiceTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def addConnectionPointType(self, xpntType):
        """
        Add a connection point type.

        :param xpntType: New connection point type
        :type xpntType: WpmodConnectionPointTypeP
        """
        raise NotImplementedError()


class _WpmodSystemApplicationModule(WmodApplicationModuleBase,
                                    WpmodSystemApplicationModuleMP):
    def __init__(self, operRegistry=None):
        super().__init__(
            code=WPMOD_SYSTEM_APP_CODE,
            name='System Mapper',
            defaultOps=WPHDB_DEFAULT_OPS,
            operRegistry=operRegistry)

        #: Distribution service types
        self.__distribSvcTypes = {}
        """:type: dict[str, WpmodDistributionServiceTypeP]"""

        #: Connection point types
        self.__xpointTypes = {}
        """:type: dict[str, WpmodConnectionPointTypeP]"""

        self._createDefaultActions()

    def _createDefaultActions(self):
        actnSection = '{}.core.{}'.format(
            WACT_WCR_ACTION_CODE, WPMOD_SYSTEM_APP_CODE)

    def createHierarchyVerifier(self, pctx):
        return WphdbHierarchyVerifier(pctx, self._getDbVerifierModules())

    def getPayloadType(self):
        return WCMN_PAYLOAD_BLOCK_SYSTEM

    def createTopLevelBlock(self, name=None, info=None):
        return wpmod_create_top_block(self, name, info)

    def createProjectContext(self, gContext, projectDB, logger=None):
        return WpmodSystemProjectContext(gContext, self, projectDB, logger)

    def addDistributionServiceType(self, dsvc):
        self.__distribSvcTypes[dsvc.getCode()] = dsvc

    def getDistributionServiceType(self, code):
        return self.__distribSvcTypes.get(code)

    def getDistributionServiceTypes(self):
        return self.__distribSvcTypes.values()

    def addConnectionPointType(self, xpntType):
        self.__xpointTypes[xpntType.getCode()] = xpntType

    def getConnectionPointType(self, code):
        return self.__xpointTypes.get(code)

    def getConnectionPointTypes(self):
        return self.__xpointTypes.values()


def wpmod_create_top_block(module, name=None, info=None):
    """
    Create a new top-level block.

    :param module: Application module
    :type module: WpmodSystemApplicationModuleP
    :param name: Name for new top-level block, if any
    :type name: str | None
    :param info: Info for new top-level block, if any
    :type info: str | None
    :return: Newly created top-level block
    :rtype: WphdbTopBlock
    """
    xpntClzzs = {x.getCode(): x.getClass() for x
                 in module.getConnectionPointTypes()}
    """:type: dict[str, () -> WphdbConnectionPointBase]"""
    xpntCfgMetas = {x.getCode(): x.getConfigMeta() for x
                    in module.getConnectionPointTypes()}
    """:type: dict[str, WdbsMetaTypeP]"""
    rgty = WphdbWrapperRegistry(xpntClzzs)
    tblk = whdb_createTopLevelRawBlock(
        WphdbMetaProvider(xpntCfgMetas), name, info)
    return rgty.createWrapper(tblk)


#: Application module singleton
WPMOD_SYSTEM_APP_MODULE = _WpmodSystemApplicationModule(WPHDB_OPER)
