# ----------------------------------------------------------------------------
# - wpmod-mapping-manager                                                    -
# - wild cruncher - sys model mapping manager                  core / system -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpmod_mapping_manager
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher system mapper model mapping manager

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP
from wcr_core.modules.system.db.wphdb_host_instance import WphdbHostInstance


class WpmodMappingManager(object):
    """
    The mapping manager assists in finding models in the logical simulation,
    and maintains the reverse mapping from logical models to hosts/processes.

    All new mappings and un-mappings must be performed using this manager
    in order for the reverse mapping to be properly registered.
    """

    def __init__(self, tblk):
        """
        :param tblk: System mapper hierarchy top-level
        :type tblk: WphdbTopBlock
        """

        #: System mapper hierarchy top-level
        self.__topBlock = tblk

        #: Logical simulation (when loaded)
        self.__logicalSimulation = None
        #: Mapping from UUID to logical models
        self.__models = {}
        #: Mapping from UUID of a logical model to the mapped host/process
        self.__mappings = {}
        """:type: dict[UUID, WphdbProcess]"""

        self.rescanMappings()

    def setLogicalSimulation(self, lsim):
        """
        Set the logical simulation.

        :param lsim: Logical simulation
        :type lsim: WshdbTopBlock | None
        """
        self.__logicalSimulation = lsim

        if lsim:
            self.__models = self._readLogicalSimulation(lsim)
        else:
            self.__models = {}

    def getLogicalModels(self):
        """
        Get all logical models.

        :return: Logical models
        :rtype: Iterable[WshdbModelInstance]
        """
        return self.__models.values()

    def rescanMappings(self):
        """
        Rescan the current project for process-model mappings.
        """
        self.__mappings = self._readMappings(self.__topBlock)

    def getUnmappedModels(self):
        """
        Get all models from the logical simulation that are not mapped to any
        process.

        :return: Unmapped models
        :rtype: Iterable[WshdbModelInstance]
        """
        return (m for k, m in self.__models.items()
                if k not in self.__mappings)

    def getModelByUUID(self, uuid):
        """
        Get the logical model instance with the given UUID.

        :param uuid: Logical model UUID to look for
        :type uuid: UUID
        :return: Logical model instance with the given UUID, if any
        :rtype: WshdbModelInstance | None
        """
        return self.__models.get(uuid)

    def getModelProcess(self, model):
        """
        Get the process that the logical model with the given UUID is mapped
        to.

        :param model: Logical model UUID to look for
        :type model: UUID
        :return: Process the model with the given UUID is mapped to, if any
        :rtype: WphdbProcess | None
        """
        return self.__mappings.get(model)

    def mapModelToProcess(self, modelUuid, process):
        """
        Map the given model (UUID) to the given process.

        :param modelUuid: Model to map
        :type modelUuid: UUID
        :param process: Process to map model to
        :type process: WphdbProcess
        """
        pprc = self.__mappings.get(modelUuid)
        if pprc:
            pprc.removeModelMapping(modelUuid)
        minst = self.getModelByUUID(modelUuid)
        mname = minst.getName() if minst else None
        process.addModelMapping(modelUuid, mname)
        self.__mappings[modelUuid] = process

    def unmapModel(self, model):
        """
        Remove any mappings for the given model (UUID), if any.

        :param model: Model to remove mappings for
        :type model: UUID
        """
        pprc = self.__mappings.get(model)
        if pprc:
            pprc.removeModelMapping(model)
            del self.__mappings[model]

    def _readLogicalSimulation(self, lsim):
        """
        Read the UUID-model mappings from the given logical simulation.

        :param lsim: Logical simulation
        :type lsim: WshdbTopBlock
        :return: Mappings from UUIDs to models
        :rtype: dict[UUID, WshdbModelInstance]
        """

        def readBlock(block):
            for inst in block.getInstances():
                if isinstance(inst, WhdbBlockInstanceP):
                    readBlock(inst.getBlock())

        models = {}

        readBlock(lsim)

        return models

    def _readMappings(self, tblk):
        """
        Read the model-process mappings from the given system mapper hierarchy.

        :param tblk: System mapper hierarchy
        :type tblk: WphdbTopBlock
        :return: UUID-to-process mappings
        :rtype: dict[UUID, WphdbProcess]
        """

        def iterBlock(block):
            for inst in block.getInstances():
                if isinstance(inst, WhdbBlockInstanceP):
                    iterBlock(inst.getBlock())
                elif isinstance(inst, WphdbHostInstance):
                    for process in inst.getProcesses():
                        for model in process.getModelMappings():
                            mappings[model.uuid] = process

        mappings = {}

        iterBlock(tblk)

        return mappings
