# ----------------------------------------------------------------------------
# - wpmod-project-context                                                    -
# - wild cruncher - sys project context                       core / system  -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpmod_project_context
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher system mapper project context

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.common.wcmn_parameter_description import WcmnParameterDescription
from wcr_core.modules.system.controller.wpctl_controllers import \
    WpctlControllerFactory
from wcr_core.modules.system.db.wphdb_phybus_instance import \
    WphdbPhyBusInstance
from wcr_core.modules.system.wpmod_project_context_api import \
    WpmodSystemProjectContextMP
from wcr_core.project.wprj_project_context import WprjProjectContext
from wcr_core.modules.system.wpmod_mapping_manager import WpmodMappingManager


class WpmodSystemProjectContext(WprjProjectContext,
                                WpmodSystemProjectContextMP):
    """
    Default project context implementation.
    """

    def __init__(self, gContext, module, projectDB, logger=None):
        """
        :param gContext: Global context
        :type gContext: WcorGlobalContextP
        :param module: Application module
        :type module: WpmodSystemApplicationModuleP
        :param projectDB: Project database
        :type projectDB: WprjProjectDatabase
        :param logger: Logger
        :type logger: ScmnLoggerP
        """
        super().__init__(gContext, module, projectDB, logger)

        self.__ctrlFy = WpctlControllerFactory(module)

        #: Logical simulation, when set
        self.__simulation = None
        #: Mapping Manager
        # noinspection PyTypeChecker
        self.__mappingManager = WpmodMappingManager(
            projectDB.getTopHierarchy())

    def getLogicalSimulation(self):
        return self.__simulation

    def setLogicalSimulation(self, lsim):
        self.__simulation = lsim

    def getMappingManager(self):
        return self.__mappingManager

    def getServices(self):
        return [
            ('Service A', (
                WcmnParameterDescription(
                    'P-A1', 'P A1', 'P-A1 value', dval='Foo'),
                WcmnParameterDescription('P-A2', 'P A2', 'P-A2 value'))),
            ('Service B', (
                WcmnParameterDescription('P-B1', 'P B1', 'P-B1 value'),)),
            ('Service C', tuple())]

    def getPhysicalBuses(self):
        """
        :rtype: list[WphdbPhyBusInstance]
        """
        tpBlk = self.getProjectDB().getTopHierarchy()
        assert tpBlk
        return [inst for inst in tpBlk.getOperatorInstances()
                if isinstance(inst, WphdbPhyBusInstance)]

    def getControllerFactory(self):
        return self.__ctrlFy


def wpmod_retrieve_lbus_obj(phyBusList, simTopBlkList):
    """
    Retrieve the logical buses linked to all physical buses in a dictionary
    from a list of simulation projects

    :param phyBusList: List of physical buses
    :type phyBusList: Iterable[WphdbPhyBusInstance]
    :param simTopBlkList: Iterable of top blocks of simulation projects
        containing the logical buses bound
    :type simTopBlkList: Iterable[WshdbTopBlock]
    :return: Iterable of tuple of physical bus' UUID and iterable of logical
        buses bound to it
    :rtype: dict[UUID, list[WshdbBusInstance]]
    """
    logBuses = {lbus.getUUID(): lbus for simTopBlk in simTopBlkList
                for lbus in simTopBlk.getBusInstances()}

    res = {}
    """:type: dict[UUID, Iterable[WphdbPhyBusInstance]]"""

    for pb in phyBusList:
        boundLogBus = []
        """:type: list[WphdbPhyBusInstance]"""

        lbusUUIDList = pb.getBoundLogicalBusesUUID()
        for lbu in lbusUUIDList:
            lbusObj = logBuses.get(lbu)
            if lbusObj is not None:
                boundLogBus.append(lbusObj)

        res[pb.getUUID()] = boundLogBus

    return res