# ----------------------------------------------------------------------------
# - wvhdb-top-block                                                          -
# - wild cruncher:wsh top hierarchy object                 core / vlink / db -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 lambert philippe                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wvhdb_top_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher top hierarchy object

.. moduleauthor:: philippe lambert <lambert@silkan.com>
"""
from wcr_core.modules.vlink.db.wvhdb_block import WvhdbBlock
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaDescription, \
    WdbsMetaField
from wcr_core.dbstore.wdbs_meta_type import WDBS_PATH_META, WdbsListMetaType, \
    WdbsTupleMetaType, WDBS_UUID_META, WDBS_STRING_META, WdbsDictionaryMetaType
from silkan_common.scmn_file_path import scmn_path_rel2file, scmn_path_mkabs, \
    scmn_isfile, scmn_path_dir, scmn_relpath
from wcr_core.common.wcmn_file_handler import WcmnFile

WVHDB_ICD_PATH_KEY = "ICD_Path"
WVHDB_MESSAGES_BINDING_KEY = "MessageBinding_Vlink"
WVHDB_ALL_MESSAGES_KEY = "MessagesList"

MESSAGE_META = WdbsDictionaryMetaType(WDBS_STRING_META,
                                      WdbsListMetaType(
                                        WdbsTupleMetaType(WDBS_STRING_META,
                                                          WDBS_STRING_META)))


class WvhdbTopHierarchy(WvhdbBlock):
    """
    The WvhdbTopHierarchy is the top hierarchy model class. A top hierarchy
    is at the same time a model and an instance. The class provides several
    methods for creating the hierarchy.
    """
    META = WdbsMetaDescription(
        WvhdbBlock.META,
        WdbsMetaField(WVHDB_ICD_PATH_KEY, WDBS_PATH_META),
        WdbsMetaField(WVHDB_MESSAGES_BINDING_KEY,
                      WdbsListMetaType(
                          WdbsTupleMetaType(WDBS_UUID_META,
                                            MESSAGE_META))),
        WdbsMetaField(WVHDB_ALL_MESSAGES_KEY,
                      WdbsListMetaType(
                                    WDBS_STRING_META)),

    )

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        super().__init__(block, wreg)
        self._srdfilepath = None

    def setICDFile(self, bf):
        """
        Set the pass to bus file for bindings
        """
        self.rawElement.setMeta(bf, WVHDB_ICD_PATH_KEY)

    def getICDFile(self):
        """
        return the absolute path of the bus file for binding
        """
        relpathhard = self.rawElement.getMeta(WVHDB_ICD_PATH_KEY)
        #: no hardware file binded
        if relpathhard is None:
            return relpathhard
        abspathhard = scmn_path_mkabs(relpathhard.path,
                                      bdir=scmn_path_dir(self._srdfilepath))
        if not scmn_isfile(abspathhard):
            self.rawElement.setMeta(None, WVHDB_ICD_PATH_KEY)
            return None
        return WcmnFile(abspathhard)

    def setSourcePath(self, path):
        """
        set the absolute path to use as reference for relative path handling
        """
        if path is not None:
            self._srdfilepath = path

    def formatToSave(self, npath):
        ndir = scmn_path_dir(npath)
        icdf = self.getICDFile()
        if icdf:
            if scmn_isfile(icdf.path):
                icdrela = scmn_relpath(icdf.path, ndir)
                self.setICDFile(WcmnFile(icdrela))
        self.setSourcePath(npath)

    def addVlinkMessageBind(self, vlinkuuid, message):
        """
        add a tuple (vlinkuuid, message) to the list of binding
        :param vlinkuuid: uuid of the Vlink to bind
        :type vlinkuuid: uuid
        :param message: messages to bind
        :type message: dict
        """
        bndl = self.rawElement.getMeta(WVHDB_MESSAGES_BINDING_KEY)
        if (vlinkuuid, message) not in bndl:
            bndl.append((vlinkuuid, message))
        self.rawElement.setMeta(bndl, WVHDB_MESSAGES_BINDING_KEY)

    def getVlinkMessageBinds(self):
        """
        return a liste of tuple (vlinkuuid,message) to the list of
        bindings
        :return liste of vlinkuuid/message binding
        :rtype [(uuid, uuid)]
        """
        return self.rawElement.getMeta(WVHDB_MESSAGES_BINDING_KEY)

    def removeVlinkMessageBind(self, vlinkuuid, message):
        """
        remove the tuple (vlinkuuid, message) from the list of binding
        :param vlinkuuid: uuid of the Traffic Route to unbind
        :type vlinkuuid: uuid
        :param message: the message to unbind
        :type message: str
        """
        bndl = self.rawElement.getMeta(WVHDB_MESSAGES_BINDING_KEY)
        for vl, mess in bndl:
            if vl == vlinkuuid and mess == message:
                bndl.remove((vlinkuuid, message))
        self.rawElement.setMeta(bndl, WVHDB_MESSAGES_BINDING_KEY)

    def setAllMessages(self, messages):
        """
        set all messages
        :param messages: list of messages
        :type messages: list
        """
        self.rawElement.setMeta(messages, WVHDB_ALL_MESSAGES_KEY)

    def getAllMessages(self):
        """
        get all messages
        :return list messages
        :rtype [messages]
        """
        return self.rawElement.getMeta(WVHDB_ALL_MESSAGES_KEY)

    def checkMsgBindingCoherency(self):
        """
        Return list of messages that are not bound to a traffic route
        :return : Bool
        """
        allMessages = self.getAllMessages()
        allBindings = self.getVlinkMessageBinds()
        messageNotBound = []
        allBoundmMsgs = [list(bdmsg.keys())[0] for _, bdmsg in allBindings]
        for msg in allMessages:
            if msg not in allBoundmMsgs:
                messageNotBound.append(msg)

        return messageNotBound