# ----------------------------------------------------------------------------
# - wvhdb-top-block                                                          -
# - wild cruncher:wsh top hierarchy object                 core / vlink / db -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 lambert philippe                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wvhdb_top_block
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher top hierarchy object

.. moduleauthor:: philippe lambert <lambert@silkan.com>
"""
import tempfile
import os
from wcr_core.modules.vlink.db.wvhdb_block import WvhdbBlock
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaDescription, \
    WdbsMetaField
from wcr_core.dbstore.wdbs_meta_type import WDBS_PATH_META, WdbsListMetaType, \
    WdbsTupleMetaType, WDBS_UUID_META, WDBS_STRING_META, WdbsDictionaryMetaType
from silkan_common.scmn_file_path import scmn_path_rel2file, scmn_path_mkabs, \
    scmn_isfile, scmn_path_dir, scmn_relpath
from wcr_core.common.wcmn_file_handler import WcmnFile
from wcr_core.project.wprj_xmlio import WprjXml
from wcr_core.project.wprj_project import WprjProjectStructure

WVHDB_TMP_PATH_KEY = "TemporaryPath"
WVHDB_MESSAGES_BINDING_KEY = "MessageBinding_Vlink"
WVHDB_ALL_MESSAGES_KEY = "MessagesList"

MESSAGE_META = WdbsDictionaryMetaType(WDBS_STRING_META,
                                      WdbsListMetaType(
                                        WdbsTupleMetaType(WDBS_STRING_META,
                                                          WDBS_STRING_META)))


class WvhdbTopHierarchy(WvhdbBlock):
    """
    The WvhdbTopHierarchy is the top hierarchy model class. A top hierarchy
    is at the same time a model and an instance. The class provides several
    methods for creating the hierarchy.
    """
    META = WdbsMetaDescription(
        WvhdbBlock.META,
        WdbsMetaField(WVHDB_TMP_PATH_KEY, WDBS_PATH_META),
        WdbsMetaField(WVHDB_MESSAGES_BINDING_KEY,
                      WdbsListMetaType(
                          WdbsTupleMetaType(WDBS_UUID_META,
                                            MESSAGE_META))),
        WdbsMetaField(WVHDB_ALL_MESSAGES_KEY,
                      WdbsListMetaType(
                                    WDBS_STRING_META)),

    )

    def __init__(self, block, wreg):
        """
        :param block: Raw block
        :type block: WdbsBlockP
        :param wreg: Wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        super().__init__(block, wreg)
        self._srdfilepath = None
        self._tmpFilePath = None
        self._tmpFile = None
        self.isModified = False

    def createTmpFile(self):
        """
        Set the pass to bus file for bindings
        """
        if self._tmpFilePath and self._tmpFile:
            return
        tmpPath = tempfile.mkstemp(suffix='.tr.xml')[1]
        self._tmpFilePath = tmpPath
        self.saveTmpFile()

    def setTmpFile(self, tmpfile):
        self._tmpFile = tmpfile

    def setTmpFilePath(self, tmpfilePath):
        self._tmpFilePath = tmpfilePath

    def getTmpFile(self):
        """
        return the absolute path of the bus file for binding
        """
        return self._tmpFile

    def getTmpFilePath(self):
        """
        return the absolute path of the bus file for binding
        """
        return self._tmpFilePath

    def deleteTmpFile(self):
        if self._tmpFilePath:
            try:
                if self._tmpFile:
                    self._tmpFile.close()
                os.remove(self._tmpFilePath)
                self._tmpFilePath = None
                self._tmpFile = None
            except FileNotFoundError:
                pass
            except PermissionError:
                pass

    def saveTmpFile(self):
        wprjs = WprjProjectStructure()
        wprjs.setTopHierarchy(self)
        from wcr_core.modules.vlink.wvmod_application_module import \
            WVMOD_VLINK_APP_CODE
        out = WprjXml().serialize2(wprjs, WVMOD_VLINK_APP_CODE)
        if out and self._tmpFilePath:
            self._tmpFile = open(self._tmpFilePath, 'w+')
            self._tmpFile.write(out)
            self._tmpFile.close()

    def setSourcePath(self, path):
        """
        set the absolute path to use as reference for relative path handling
        """
        if path is not None:
            self._srdfilepath = path

    def formatToSave(self, npath):
        self.saveTmpFile()
        self.setSourcePath(npath)

    def getRouteFromUUID(self, uuid):
        for route in self.getInstances():
            if route.getUUID() == uuid:
                return route

    def addVlinkMessageBind(self, vlinkuuid, message):
        """
        add a tuple (vlinkuuid, message) to the list of binding
        :param vlinkuuid: uuid of the Vlink to bind
        :type vlinkuuid: uuid
        :param message: messages to bind
        :type message: dict
        """
        bndl = self.rawElement.getMeta(WVHDB_MESSAGES_BINDING_KEY)
        if (vlinkuuid, message) not in bndl:
            bndl.append((vlinkuuid, message))
        self.rawElement.setMeta(bndl, WVHDB_MESSAGES_BINDING_KEY)

    def getVlinkMessageBinds(self):
        """
        return a liste of tuple (vlinkuuid,message) to the list of
        bindings
        :return liste of vlinkuuid/message binding
        :rtype [(uuid, uuid)]
        """
        return self.rawElement.getMeta(WVHDB_MESSAGES_BINDING_KEY)

    def removeVlinkMessageBind(self, vlinkuuid, message):
        """
        remove the tuple (vlinkuuid, message) from the list of binding
        :param vlinkuuid: uuid of the Traffic Route to unbind
        :type vlinkuuid: uuid
        :param message: the message to unbind
        :type message: str
        """
        bndl = self.rawElement.getMeta(WVHDB_MESSAGES_BINDING_KEY)
        for vl, mess in bndl:
            if vl == vlinkuuid and mess == message:
                bndl.remove((vlinkuuid, message))
        self.rawElement.setMeta(bndl, WVHDB_MESSAGES_BINDING_KEY)

    def setAllMessages(self, messages):
        """
        set all messages
        :param messages: list of messages
        :type messages: list
        """
        self.rawElement.setMeta(messages, WVHDB_ALL_MESSAGES_KEY)

    def getAllMessages(self):
        """
        get all messages
        :return list messages
        :rtype [messages]
        """
        return self.rawElement.getMeta(WVHDB_ALL_MESSAGES_KEY)

    def checkMsgBindingCoherency(self):
        """
        Return list of messages that are not bound to a traffic route
        :return : Bool
        """
        allMessages = self.getAllMessages()
        allBindings = self.getVlinkMessageBinds()
        messageNotBound = []
        allBoundmMsgs = [list(bdmsg.keys())[0] for _, bdmsg in allBindings]
        for msg in allMessages:
            if msg not in allBoundmMsgs:
                messageNotBound.append(msg)

        return messageNotBound