# ----------------------------------------------------------------------------
# - wmod-application-module                                                  -
# - wild cruncher - application module                         core / module -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmod_application_module
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher application module

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import ABCMeta, abstractmethod

from silkan_common.scmn_utils import first
from wcr_core.db.whdb_meta_provider import WhdbMetaProvider
from wcr_core.db.whdb_operator import WhdbOperatorRegistry
from wcr_core.db.whdb_wrapper_registry import WhdbWrapperRegistry
from wcr_core.db.whdb_top_level_factory import whdb_createTopLevelRawBlock
from wcr_core.db.whdb_verifier import WhdbHierarchyVerifier
from silkan_common.scmn_mime_type import SCMN_MIME_EXTS, SCMN_MIMES, \
    SCMN_PAYLOAD_NONE
from wcr_core.db.whdb_block import WhdbBlock
from wcr_core.db_api.whdb_block_api import WHDB_DEFAULT_BLOCK_CODE
from wcr_core.modules.wmod_block_type import WmodBlockType


#: Whether to mute warnings about deprecated operators
WMOD_MUTE_DEPRECATED_OPS = False

#: Default blocok types
WMOD_DEFAULT_BLOCK_TYPES = (
    WmodBlockType(WHDB_DEFAULT_BLOCK_CODE, 'Block', WhdbBlock.META, WhdbBlock),
)

#: Dictionary of registered application modules
_WMOD_APPLICATION_MODULES = {}


WMOD_EVENT_COUNT = 2

# Application module notification event types
WMOD_NO_EVENT, \
    WMOD_ACTION_ADDED_EVENT = ((2 ** i) > 1 for i in range(WMOD_EVENT_COUNT))


class WmodApplicationModuleObserverP(metaclass=ABCMeta):
    """
    Protocol for application module observers.
    """
    @abstractmethod
    def appModuleNotify(self, module, event, *args):
        """
        Callback for application module notifications.

        :param module: Source module
        :type module: WmodApplicationModuleP
        :param event: Event type
        :type event: int
        :param args: Event-specific arguments
        :type args: tuple[object]
        """
        raise NotImplementedError()


class WmodApplicationModuleP(metaclass=ABCMeta):
    @property
    @abstractmethod
    def code(self):
        """
        Get the module's code/identifier.

        :return: Code of the application module
        :rtype: str
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self):
        """
        Get the module's name.

        :return: Name of the application module
        :rtype: str
        """
        raise NotImplementedError

    @abstractmethod
    def getLibraries(self):
        """
        Get the module's registered libraries.

        :return: Libraries for the application module
        :rtype: Iterable[WplgOperatorLibrary]
        """
        raise NotImplementedError()

    @abstractmethod
    def getLibrary(self, code):
        """
        Get the library with the given code.

        :param code: Library code to get library for
        :type code: str
        :return: Library for the given code, or `None` if no such lib exists
        :rtype: WplgOperatorLibrary
        """
        raise NotImplementedError()

    @abstractmethod
    def getDefaultOperators(self):
        """
        Get the default operators for the module.

        :return: Default operators
        :rtype: Iterable[WhdbOperatorP]
        """
        raise NotImplementedError()

    @abstractmethod
    def hasOperator(self, code):
        """
        Whether the module has an operator with the given code.

        :param code: Operator code to look for
        :type code: str
        :return: True iff an operator with the given code exists
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def getOperator(self, code):
        """
        Get the operator with the given code.

        :param code: Code of operator to retrieve
        :type code: str
        :return: Operator with the given code
        :rtype: WhdbOperatorP
        """
        raise NotImplementedError()

    @abstractmethod
    def getBlockTypes(self):
        """
        Get the possible block types.

        :return: Possible block types as list of code/description-tuples
        :rtype: Iterable[WmodBlockTypeP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getBlockType(self, code):
        """
        Get the block type for the given code (MIME-type), if any.

        :return: Block type for the given code, if it exists
        :rtype: WmodBlockTypeP | None
        """
        raise NotImplementedError()

    @abstractmethod
    def getModelMimeTypes(self):
        """
        Get the MIME-types for models that can be added to instances of this
        module.

        :return: MIME-types for known model types
        :rtype: Iterable[ScmnMimeType]
        """
        raise NotImplementedError()

    @abstractmethod
    def getMimeType(self, mime):
        """
        Get the MIME-type description for the given MIME identifier
        (e.g. "text/plain").

        :param mime: Identifier
        :type mime: str
        :return: MIME description for the given identifier
        :rtype: ScmnMimeType
        """
        raise NotImplementedError()

    @abstractmethod
    def getMimeTypeForExt(self, fext):
        """
        Get the MIME-type description for the given file name extension
        (e.g. "txt").

        :param fext: File name extension
        :type fext: str
        :return: MIME description for the given extension
        :rtype: ScmnMimeType
        """
        raise NotImplementedError()

    @abstractmethod
    def getProjectVerifierModules(self):
        """
        Get the additional project verifier modules.

        :return: Additional project verifier modules
        :rtype: Iterable[WprjVerifierModuleP]
        """
        raise NotImplementedError()

    @abstractmethod
    def createHierarchyVerifier(self, pctx):
        """
        Get the application module's database verifier.

        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :return: Database verifier for the module.
        :rtype: WmhdbHierarchyVerifier
        """
        raise NotImplementedError()

    @abstractmethod
    def getPayloadType(self):
        """
        Get the block payload type representing instances of this module.

        :return: Block payload type
        :rtype: int
        """
        raise NotImplementedError

    @abstractmethod
    def createTopLevelBlock(self, name=None, info=None):
        """
        Create a new top-level block for this module.

        :param name: Name of the new block
        :type name: str
        :param info: Info string for the new block
        :type info: str
        :return: Newly created block
        :rtype: WhdbBlockP
        """
        raise NotImplementedError()

    @abstractmethod
    def createProjectContext(self, gContext, projectDB, logger=None):
        """
        Create a new project context.

        :param gContext: Global context
        :type gContext: WcorGlobalContextP
        :param projectDB: Project database
        :type projectDB: WprjProjectDatabase
        :param logger: Logger
        :type logger: ScmnLoggerP
        :return: New project context
        :rtype: WcorProjectContextP
        """
        raise NotImplementedError()

    @abstractmethod
    def getActions(self):
        """
        Get the actions available for the module.

        :return: Module actions
        :rtype: list[(WactActionP, str, str, tuple[str])]
        """
        raise NotImplementedError()

    @abstractmethod
    def getProjectProperties(self):
        """
        Get the additional project properties.

        :return: Additional project properties
                 (source, source name, [(prop code, prop name)])
        :rtype: Iterable[(str, str, tuple[(str, str)])]
        """
        raise NotImplementedError()

    @abstractmethod
    def registerObserver(self, observer):
        """
        Register an observer for notifications.

        :param observer: Observer to register
        :type observer: WmodApplicationModuleObserverP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WmodApplicationModuleMP(WmodApplicationModuleP):
    @abstractmethod
    def addLibrary(self, lib):
        """
        Add a library to the application module.

        :param lib: Library to add
        :type lib: WplgOperatorLibrary
        """
        raise NotImplementedError

    @abstractmethod
    def addAction(self, action, code, source, *flags):
        """
        Add an action to the application module.

        :param action: Action to add
        :type action: WactActionP
        :param code: Action code (unique per source)
        :type code: str
        :param source: Source adding the action (e.g. plugin code)
        :type source: str
        :param flags: Flags
        :type flags: tuple[str]
        """
        raise NotImplementedError

    @abstractmethod
    def addBlockType(self, blockType):
        """
        Add a block type.

        :param blockType: Block type to be added
        :type blockType: WmodBlockTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def addOperator(self, operator, section):
        """
        Add an operator to the application module.

        :param operator: Operator to add
        :type operator: WhsbOperatorP
        :param section: Section
        :type section: str
        """
        raise NotImplementedError

    @abstractmethod
    def addOperators(self, operators, section):
        """
        Add an operator to the application module.

        :param operators: Operators to add
        :type operators: Iterable[WhsbOperatorP]
        :param section: Section
        :type section: str
        """
        raise NotImplementedError

    @abstractmethod
    def addModel(self, mime):
        """
        Add a model MIME-type, i.e. MIME-type that can serve as model within
        this application module.

        :param mime: Model MIME-type
        :type mime: ScmnMimeTypeP
        """
        raise NotImplementedError

    @abstractmethod
    def addProjectVerificationModule(self, vmod):
        """
        Add a project verifier module.

        :param vmod: Project verifier module
        :type vmod: WprjVerifierModuleP
        """
        raise NotImplementedError()

    @abstractmethod
    def addDbVerificationModule(self, vmod):
        """
        Add a database verifier module.

        :param vmod: Database verifier module
        :type vmod: WhdbVerifierModuleP
        """
        raise NotImplementedError()

    @abstractmethod
    def registerProjectProperties(self, source, srcName, props):
        """
        Register an additional project property.

        :param source: Source identifier
        :type source: str
        :param srcName: Source name
        :type srcName: str
        :param props: Properties [(code, name)]
        :type props: Iterable[(str, str)]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WmodApplicationModuleBase(WmodApplicationModuleMP):
    """
    TODO: Inherit OperatorRegistryP, ActionRegistryP, LibraryRegistryP
    TODO: Pass WmodApplicationModuleBase around instead of static registries
    """
    def __init__(self, code, name, blockTypes=None, modelMimes=None,
                 defaultOps=None, operRegistry=None):
        """
        :type code: str
        :type name: str
        :type blockTypes: Iterable[WmodBlockTypeP] | None
        :type modelMimes: Iterable[ScmnMimeTypeP] | None
        :type defaultOps: Iterable[WhdbOperatorP] | None
        :type operRegistry: WhdbOperatorRegistry | None
        """
        self.__code = code
        """:type: str"""
        self.__name = name
        """:type: str"""
        self.__blockTypes = list(blockTypes) if blockTypes else []
        """:type: list[WmodBlockTypeP]"""
        self.__modelMimes = list(modelMimes) if modelMimes else []
        """:type: list[ScmnMimeTypeP]"""
        self.__defaultOps = list(defaultOps) if defaultOps else []
        """:type: list[WhdbOperatorP]"""
        self.__operatorRegistry = operRegistry or WhdbOperatorRegistry()
        """:type: WhdbOperatorRegistry"""

        self.__libs = {}
        """:type: dict[str, WplgOperatorLibrary]"""

        self.__actions = []
        """:type: list[(WactActionP, str, str, tuple[str])]"""
        self.__observers = []
        """:type: list[WmodApplicationModuleObserverP]"""
        self.__dbVerifyMods = []
        """:type: list[WhdbVerifierModuleP]"""
        self.__prjVerifyMods = []
        """:type: list[WprjVerifierModuleP]"""
        self.__prjProps = []
        """:type: list[(str, str, tuple[str])]"""

    @property
    def code(self):
        return self.__code

    @property
    def name(self):
        return self.__name

    def addBlockType(self, blockType):
        self.__blockTypes.append(blockType)

    def getBlockTypes(self):
        return self.__blockTypes

    def getBlockType(self, code):
        return first(btyp for btyp in self.getBlockTypes()
                     if btyp.getCode() == code)

    def addModel(self, mime):
        self.__modelMimes.append(mime)

    def getModelMimeTypes(self):
        return iter(self.__modelMimes)

    def getMimeType(self, mime):
        return SCMN_MIMES.get(mime)

    def getMimeTypeForExt(self, fext):
        mimes = SCMN_MIME_EXTS.get(fext)
        return mimes[0] if mimes else None

    def addLibrary(self, library):
        self.__libs[library.code] = library
        self.__operatorRegistry.addOperators(library.operators)

    def getLibraries(self):
        return iter(self.__libs.values())

    def getLibrary(self, code):
        return self.__libs.get(code)

    def addAction(self, action, code, source, *flags):
        """
        :type action: WactActionP
        :type code: str
        :type source: str
        :type flags: tuple[str]
        """
        flags += (self.__code,)
        code = '{}.{}'.format(source, code)
        atpl = (action, code, source, flags)
        self.__actions.append(atpl)
        self._notifyObservers(WMOD_ACTION_ADDED_EVENT, atpl)

    def getActions(self):
        return iter(self.__actions)

    def addOperator(self, operator, section):
        self.__operatorRegistry.addOperator(operator)

        if section == 'default':
            self.__defaultOps.append(operator)

    def addOperators(self, operators, section):
        self.__operatorRegistry.addOperators(operators)

        if section == 'default':
            self.__defaultOps.extend(operators)

    def hasOperator(self, code):
        return code in self.__operatorRegistry

    def getOperator(self, code):
        if '.' not in code and not WMOD_MUTE_DEPRECATED_OPS:
            print('Use of not fully-qualified operator names is deprecated!')
        return self.__operatorRegistry.get(code)

    def getDefaultOperators(self):
        return iter(self.__defaultOps)

    def addProjectVerificationModule(self, vmod):
        self.__prjVerifyMods.append(vmod)

    def getProjectVerifierModules(self):
        return iter(self.__prjVerifyMods)

    def addDbVerificationModule(self, vmod):
        self.__dbVerifyMods.append(vmod)

    def _getDbVerifierModules(self):
        """
        Get the additional database verifier modules.

        :return: Additional database verifier modules
        :rtype: Iterable[WhdbVerifierModuleP]
        """
        return iter(self.__dbVerifyMods)

    def registerProjectProperties(self, source, srcName, props):
        self.__prjProps.append((source, srcName, tuple(props)))

    def getProjectProperties(self):
        return iter(self.__prjProps)

    def registerObserver(self, observer):
        self.__observers.append(observer)

    def _notifyObservers(self, event, *args):
        for obs in self.__observers:
            obs.appModuleNotify(self, event, *args)

    def getFileExtention(self):
        """
        Return string for extention to add at file when saved.
        :return: default extention dor Cetrac designer file.
        :return: str
        """
        return ''


def wmod_registerModule(module):
    """
    Register an application module globally.

    Note that application modules should be created as singletons only.

    :param module: Module to register
    :type module: WmodApplicationModuleP
    """
    _WMOD_APPLICATION_MODULES[module.code] = module


def wmod_getModules():
    """
    Get a list of all globally registered application modules.

    :return: Globally registered application modules
    :rtype: Iterable[WmodApplicationModuleP]
    """
    return iter(_WMOD_APPLICATION_MODULES.values())


def wmod_getModule(code):
    """
    Get the application module with the given code.

    :param code: Application module code to look for
    :type code: str
    :return: Application module with the given code
    :rtype: WmodApplicationModuleP
    """
    return _WMOD_APPLICATION_MODULES.get(code)


def wmod_getModuleForPdb(prjdb):
    """
    Get the application module for the given project DB.

    :param prjdb: Project DB to get module for
    :type prjdb: WprjProjectDatabase
    :return: Application module for the given DB
    :rtype: WmodApplicationModuleP
    """
    module = _WMOD_APPLICATION_MODULES.get(prjdb.getModule())
    if not module:
        raise ValueError(
            'Could not find application module with code {}'.format(
                prjdb.getModule()))
    return module


class _WmodEmptyApplicationModule(WmodApplicationModuleP):
    """
    Dummy application module.
    """

    @property
    def code(self):
        return 'default'

    @property
    def name(self):
        return 'Default'

    def addLibrary(self, library):
        raise NotImplementedError()

    def getLibraries(self):
        return []

    def getLibrary(self, code):
        return None

    def getDefaultOperators(self):
        return []

    def hasOperator(self, code):
        return False

    def getOperator(self, code):
        return None

    def getBlockTypes(self):
        return WMOD_DEFAULT_BLOCK_TYPES

    def getBlockType(self, code):
        return first(btyp for btyp in self.getBlockTypes()
                     if btyp.getCode() == code)

    def getModelMimeTypes(self):
        return []

    def getMimeType(self, mime):
        return SCMN_MIMES.get(mime)

    def getMimeTypeForExt(self, fext):
        mimes = SCMN_MIME_EXTS.get(fext)
        return mimes[0] if mimes else None

    def createHierarchyVerifier(self, pctx):
        return WhdbHierarchyVerifier()

    def getPayloadType(self):
        return SCMN_PAYLOAD_NONE

    def createTopLevelBlock(self, name=None, info=None):
        rgty = WhdbWrapperRegistry()
        tblk = whdb_createTopLevelRawBlock(WhdbMetaProvider(), name, info)
        return rgty.createWrapper(tblk)

    def createProjectContext(self, gContext, projectDB, logger=None):
        from wcr_core.project.wprj_project_context import WprjProjectContext

        return WprjProjectContext(gContext, self, projectDB, logger)

    def getActions(self):
        return {}

    def getProjectProperties(self):
        return []

    def getProjectVerifierModules(self):
        return []

    def registerObserver(self, observer):
        pass

    def getFileExtention(self):
        return ''


#: Empty application module singleton
WMOD_EMPTY_APPLICATION_MODULE = _WmodEmptyApplicationModule()

wmod_registerModule(WMOD_EMPTY_APPLICATION_MODULE)
