# ----------------------------------------------------------------------------
# - wmod-block-type                                                          -
# - wild cruncher: block type descriptor                              moduls -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmod_block_type
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Block type descriptor

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod


#: System-block flag (i.e. not instantiatable by user directly)
WMOD_SYSTEM_BLOCK_FLAG = 'system'


class WmodBlockTypeP(metaclass=ABCMeta):
    """
    Block type descriptor protocol.
    """

    @abstractmethod
    def getCode(self):
        """
        Get the code of the block type.

        :return: Block type code
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getName(self):
        """
        Get the name of the block type.

        :return: Block type name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getMeta(self):
        """
        Get the meta-data description for the block type.

        :return: Meta-data description
        :rtype: WdbsMetaDescriptionP
        """
        raise NotImplementedError()

    @abstractmethod
    def getClass(self):
        """
        Get the class/constructor for the block type.

        :return: Class/constructor
        :rtype: Class | () ->
        """
        raise NotImplementedError()

    @abstractmethod
    def hasFlag(self, flag):
        """
        Whether the block type carries the given flag.

        :param flag: Flag to check
        :type flag: str
        :return: True iff the block type has the given flag
        :rtype: bool
        """
        raise NotImplementedError()


class WmodBlockType(WmodBlockTypeP):
    def __init__(self, code, name, meta, clazz, *flags):
        self.__code = code
        self.__name = name
        self.__meta = meta
        self.__class = clazz
        self.__flags = set(flags)

    def getCode(self):
        return self.__code

    def getName(self):
        return self.__name

    def getMeta(self):
        return self.__meta

    def getClass(self):
        return self.__class

    def hasFlag(self, flag):
        return flag in self.__flags

    def __str__(self):
        return self.__code

    def __repr__(self):
        return self.__code
