# ----------------------------------------------------------------------------
# - wplg-core-install-context                                                -
# - wild cruncher: core install context                        core / plugin -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wplg_core_install_context
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Core install context

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.plugin.wplg_library import WplgOperatorLibrary
from wcr_core.plugin.wplg_install_context import \
    WplgPlatformInstallContextBase, WplgModuleInstallContextBase
from wcr_core.plugin.wplg_install_context_api import WplgInstallContextP, \
    WplgModuleInstallContextP, WplgPlatformInstallContextP


#: Core install code
WPLG_CORE_INSTALL_CODE = 'core'


# noinspection PyAbstractClass
class WplgCoreInstallContextMixin(WplgInstallContextP):
    def getComponentCode(self):
        return WPLG_CORE_INSTALL_CODE


# noinspection PyAbstractClass
class WplgCorePlatformInstallContextP(WplgPlatformInstallContextP):
    """
    Core platform install context protocol.
    """
    @abstractmethod
    def installAction(self, action):
        """
        Install a system-wide, i.e. non-module, action.

        :param action: Action to be installed
        :type action: WactActionP
        """
        raise NotImplementedError()


class WplgCorePlatformInstallContext(WplgPlatformInstallContextBase,
                                     WplgCoreInstallContextMixin,
                                     WplgCorePlatformInstallContextP):
    def __init__(self, plgCode, plgName, plgDesc):
        """
        :param plgCode: Plugin code
        :type plgCode: str
        :param plgName: Plugin name
        :type plgName: str
        :param plgDesc: Plugin description
        :type plgDesc: str
        """
        super().__init__(plgCode, plgName, plgDesc)

        self.__actions = []
        """:type: list[WactActionP]"""

    def commit(self):
        #for action in self.__actions:
        #    pass
        #self.__actions = []
        pass

    def installAction(self, action):
        #self.__actions.append(action)
        raise NotImplementedError()


# noinspection PyAbstractClass
class WplgCoreModuleInstallContextP(WplgModuleInstallContextP):
    """
    Core module install context protocol.
    """
    @abstractmethod
    def installAction(self, action, code, *flags):
        """
        Install a module core action.

        :param action: Action to be installed
        :type action: WactActionP
        :param code: Action code
        :type code: str
        :param flags:
        :type flags: tuple[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def registerProjectProperty(self, propCode, propName):
        """
        Register a new project property.

        :param propCode: Property code
        :type propCode: str
        :param propName: Property name
        :type propName: str
        """
        raise NotImplementedError()

    @abstractmethod
    def installOperator(self, operator):
        """
        Install a module operator.

        :param operator: Operator to be installed
        :type operator: WhdbOperatorP
        """
        raise NotImplementedError()

    @abstractmethod
    def installOperatorFile(self, file, libCode=None):
        """
        Install a module operator from the given description file.

        :param file: Operator description file path
        :type file: str
        :param libCode: Library code
        :type libCode: str
        """
        raise NotImplementedError()

    @abstractmethod
    def addBlockType(self, blockType):
        """
        Add a block type.

        :param blockType: Block type to be added
        :type blockType: WmodBlockTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def addModelType(self, mime):
        """
        Add a model MIME-type, i.e. MIME-type that can serve as model within
        this application module.

        :param mime: Model MIME-type
        :type mime: ScmnMimeTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def addDatabaseVerifierModule(self, vrfm):
        """
        Add a database verifier module.

        :param vrfm: Database verifier module to be added
        :type vrfm: WhdbVerifierModuleP
        """
        raise NotImplementedError()


class WplgCoreModuleInstallContext(WplgModuleInstallContextBase,
                                   WplgCoreInstallContextMixin,
                                   WplgCoreModuleInstallContextP):

    def __init__(self, plgCode, plgName, plgDesc, module):
        """
        :param plgCode: Plugin code
        :type plgCode: str
        :param plgName: Plugin name
        :type plgName: str
        :param plgDesc: Plugin description
        :type plgDesc: str
        :param module:
        :type module: WmodApplicationModuleMP
        """
        super().__init__(plgCode, plgName, plgDesc)

        self.__module = module

        self.__actions = []
        """:type: list[(WactActionP, str, tuple[str])]"""
        self.__operators = []
        """:type: list[WhdbOperatorP]"""
        self.__blockTypes = []
        """:type: list[WmodBlockTypeP]"""
        self.__models = []
        """:type: list[ScmnMimeTypeP]"""
        self.__prjProps = []
        """:type: list[(str, str)]"""
        self.__dbVrfMods = []
        """:type: list[WhdbVerifierModuleP]"""

    def commit(self):
        pcod = self.getPluginCode()
        pnam = self.getPluginName()
        pdsc = self.getPluginDescription()

        for action, code, flst in self.__actions:
            acod = '{}.{}'.format(pcod, code)
            self.__module.addAction(action, acod, pcod, *flst)
        self.__actions = []

        if self.__operators:
            opLib = WplgOperatorLibrary(
                pcod, self.__module.code, pnam, pdsc, self.__operators)
            self.__module.addLibrary(opLib)
            self.__operators = []

        for btyp in self.__blockTypes:
            self.__module.addBlockType(btyp)

        for model in self.__models:
            self.__module.addModel(model)

        if self.__prjProps:
            self.__module.registerProjectProperties(
                pcod, pnam, self.__prjProps)

        for vrfm in self.__dbVrfMods:
            self.__module.addDbVerificationModule(vrfm)

    def installAction(self, action, code, *flags):
        """
        :type action: WactActionP
        :type code: str
        :type flags: tuple[str]
        """
        if 'plugin' not in flags:
            flags += ('plugin',)
        self.__actions.append((action, code, flags))

    def installOperator(self, operator):
        self.__operators.append(operator)

    def installOperatorFile(self, file, libCode=None):
        raise NotImplementedError()

    def addBlockType(self, blockType):
        self.__blockTypes.append(blockType)

    def addModelType(self, mime):
        self.__models.append(mime)

    def registerProjectProperty(self, propCode, propName):
        self.__prjProps.append((propCode, propName))

    def addDatabaseVerifierModule(self, vrfm):
        self.__dbVrfMods.append(vrfm)
