# ----------------------------------------------------------------------------
# - wplg-installer                                                           -
# - wild cruncher: installer                                   core / plugin -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wplg_installer
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Installer

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
import sys


class WplgPluginDescriptor(object):
    """
    Plugin descriptor.
    """

    def __init__(self, code, name, description, installs, paths=None):
        """
        :param code: Plugin code
        :type code: str
        :param name: Plugin name
        :type name: str
        :param description: Plugin description
        :type description: str
        :param installs: Install steps [(comp, module | None, inst-func)]
        :type installs: Iterable[(str, str | None, (WplgInstallContextP) -> _)]
        :param paths: Code paths
        :type paths: Iterable[str] | None
        """
        self.__code = code
        self.__name = name
        self.__description = description
        self.__installs = installs
        self.__paths = paths

    @property
    def code(self):
        """
        Get the plugin's code.

        :return: Plugin code
        :rtype: str
        """
        return self.__code

    @property
    def name(self):
        """
        Get the plugin's name.

        :return: Plugin name
        :rtype: str
        """
        return self.__name

    @property
    def description(self):
        """
        Get the plugin's description.

        :return: Plugin description
        :rtype: str
        """
        return self.__description

    @property
    def installs(self):
        """
        Get the plugin's install steps.

        :return: Instal steps
        :rtype: Iterable[(str, str, (WplgInstallContextP) -> _)]
        """
        return self.__installs

    @property
    def paths(self):
        """
        Get the paths for the plugin's code (i.e. PYTHONPATH).

        :return: Code paths
        :rtype: Iterable[str] | None
        """
        return self.__paths

    def __str__(self):
        return 'Plugin-Descriptor {}'.format(self.__code)


class WplgInstaller(object):
    """
    Plugin installer.
    """

    def __init__(self, imrg):
        """
        :param imrg: Install manager registry
        :type imrg: WplgInstallContextRegistry
        """
        self.__imrg = imrg

    def installPlugin(self, plgDesc):
        """
        Install a plugin.

        :param plgDesc: Description of the plugin to be installed
        :type plgDesc: WplgPluginDescriptor
        """
        ctxs = []
        """:type: list[WplgInstallContextBaseP]"""

        newPaths = None
        """:type: set[str] | None"""

        if plgDesc.paths:
            newPaths = set(plgDesc.paths).difference(sys.path)
            sys.path.extend(newPaths)

        try:
            for comp, module, ifnc, *args in plgDesc.installs:
                ictf = self.__imrg.getInstallContextFactory(comp, module)
                if not ictf:
                    print(
                        'COULD NOT FIND INSTALL CONTEXT {}.{} FOR {}'.format(
                            module, comp, plgDesc.name))
                    continue
                ictx = ictf.createInstallContext(
                    plgDesc.code, plgDesc.name, plgDesc.description)
                ctxs.append(ictx)
                ifnc(ictx, *args)
            for ictx in ctxs:
                ictx.commit()
        except Exception as e:
            print('FAILED TO INSTALL PLUGIN {}: {} ({})'.format(
                plgDesc.name, e, type(e)))
            if newPaths:
                for p in newPaths:
                    sys.path.remove(p)
