# ----------------------------------------------------------------------------
# - wprj-project                                                             -
# - wild cruncher:wdb project database                        core / project -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2012-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wprj_project
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher project database

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

# noinspection PyStatementEffect
'''
#![Read Silkan XML Database File]
client:Object[a]
/fhandler:MimeFileHandler[a]
mime:MimeType[a]
reader:SilkanXMLFileReader[a]
phandler:PDBPayloadHandler[a]

client:fhandler.new(payloadHandler, mimeTypes)
client:fhandler.handleFile(file)
fhandler:types=phandler.getPayloadTypes()

[c:loop foreach mime in mimeTypes]
  fhandler:readers=mime.getReaders()
  [c:loop foreach reader in readers]
    [c:alt types != 0]
      fhandler:readTypes=reader.readFile(file, payloadHandler, types)
      reader:readTypes=0
      [c:loop foreach type in {P, I, M, S}]
        reader:payload=getPayload(type)
        [c:alt type in file and type in types]
          reader:phandler.handlePayload(type, payload)
          reader:readTypes |= type
        [/c]
      [/c]
      fhandler:types &= ~readTypes
    [/c]
  [/c]
[/c]
client:fhandler.destroy()
'''

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------
from wcr_core.modules.wmod_application_module import wmod_getModules, \
    wmod_getModule
from wcr_core.project.wprj_project_info import WprjProjectInfo
from silkan_common.scmn_cfg_table import ScmnConfigTable
from wcr_core.common.wcmn_file_handler import WcmnReadPayloadHandler, \
    WcmnMimeFileHandler, WcmnDefaultFileHandler, WcmnFile
from wcr_core.common.wcmn_mime_type import WCMN_PAYLOAD_PROJECT_INFO, \
    WCMN_PAYLOAD_CONFIG_TABLE, WCMN_PAYLOAD_BLOCK, WCMN_PAYLOAD_RESOURCES
from wcr_core.project.wprj_xmlio import WprjXml, \
    WPRJ_SILKAN_XML_MIME_TYPE, WPRJ_SILKAN_XML_MIME_NAME
from silkan_common.scmn_mime_type import scmn_mime_types
from wcr_frontend import showErrorWindow
import re

# ----------------------------------------------------------------------------
# - structure class
# -
# ----------------------------------------------------------------------------

class WprjProjectStructure:
    """
    The WprjProjectStructure class is a container class which is used for
    the serialization of the complete database into a single entity.
    """

    def __init__(self):
        self.lsct = ScmnConfigTable()
        self.__projectInfo = WprjProjectInfo()
        self.__toph = None
        self.__resources = {}

    def getLocalConfiguration(self):
        """
        Get the local configuration table.

        :return: Local configuration table
        :rtype: ScmnConfigTable
        """
        return self.lsct

    def getTopHierarchy(self):
        """
        Get the project hierarchy.

        :return: The new hierarchy top level
        :rtype: WhdbBlock
        """
        return self.__toph

    def setTopHierarchy(self, toph):
        """
        Set the structure top hierarchy.

        :param toph: The new hierarchy top level
        :type toph: WhdbBlock
        """
        self.__toph = toph

    @property
    def projectInfo(self):
        """
        Get the meta-information structure for the project.

        :return: The projects meta-information.
        :rtype: WprjProjectInfo
        """
        return self.__projectInfo

    def getResource(self, key):
        """
        Get the resource for the given key.

        :param key: Resource key
        :type key: str
        :return: Resource for the given key, or `None`
        :rtype: bytes
        """
        return self.__resources.get(key)

    def getResources(self):
        """
        Get all available resources and their keys.

        :return: Resources (key, data)
        :rtype: (str, bytes)
        """
        return self.__resources.items()

    def setResource(self, key, resource):
        """
        Set the resource for the given key.

        :param key: Resource key
        :type key: str
        :param resource: Resource for the given key
        :type resource: bytes
        """
        self.__resources[key] = resource


# ----------------------------------------------------------------------------
# - project class                                                            -
# ----------------------------------------------------------------------------

class WprjProjectDatabase:
    """
    The WprjProjectDatabase class is the top level project class which is
    designed to hold both the local configuration and the hierarchy.
    """

    def __init__(self, path=None):
        """
        Create a default project database.

        :param path: the project path
        :type path: str
        """
        # the project path
        self.path = path
        # the top structure
        self.tops = WprjProjectStructure()
        self.__module = None

    def getModule(self):
        """
        Get the module code for this database.

        :return: Module code
        :rtype: str
        """
        return self.__module

    def setModule(self, module):
        """
        Set the module code for this database.

        :param module: Module code
        :type module: str
        """
        self.__module = module

    def getProjectName(self):
        """
        Get the project name

        :return: Project name
        :rtype: str
        """
        return self.tops.projectInfo.projectName

    def getTopHierarchy(self):
        """
        Get the project hierarchy

        :return: Project hierarchy
        :rtype: WhdbBlock
        """
        return self.tops.getTopHierarchy()

    def getPath(self):
        """
        Get the path for the project.

        :return: Project path, or None if no path has been set yet.
        :rtype: str
        """
        return self.path

    @property
    def projectInfo(self):
        """
        Get the meta-information structure for the project.

        :return: The projects meta-information.
        :rtype: WprjProjectInfo
        """
        return self.tops.projectInfo

    # TODO: Move to ProjectDBWriter?
    def saveAs(self, path, mime=None):
        """
        Save a top structure by media and path

        :param path: the file path
        :type path: str
        :param mime: the suggested media type
        :type mime: str
        """
        ext = ""
        for mod in wmod_getModules():
            if not mod.getFileExtention() == "":
                ext += mod.getFileExtention() + "|"
        if len(ext) > 1:
            rexp = r"(" + ext[:-1] + ").xml$"
        else:
            rexp = r"().xml$"
        """rexp = r"(.hw|.log|.msg|.map|.vl).xml$" |  r"().xml$" """
        extmod = re.compile(rexp)
        test = re.findall(extmod, path)
        if test:
            # if the ewtentions is for wrong module
            if not wmod_getModule(self.getModule()).getFileExtention() == \
                    test[0]:
                path = path[:-(4+len(test[0]))] + \
                       wmod_getModule(self.getModule()).getFileExtention() + \
                       path[-4:]
        else:
            #reformat path
            path = path[:-4] + wmod_getModule(self.getModule())\
                .getFileExtention() + path[-4:]
        # check for a valid mime type
        if mime:
            assert mime == WPRJ_SILKAN_XML_MIME_NAME
        if not mime:
            mimes = scmn_mime_types(path)
            """:type: list[ScmnMimeTypeP]"""
            if not mimes:
                raise IOError('Cannot determine MIME-type for ' + path)
            assert WPRJ_SILKAN_XML_MIME_TYPE in mimes
            mime = WPRJ_SILKAN_XML_MIME_NAME
        if hasattr(self.tops.getTopHierarchy(), 'formatToSave') \
                and callable(getattr(self.tops.getTopHierarchy(),
                                     'formatToSave')):
            self.tops.getTopHierarchy().formatToSave(npath=path)

        hasError = self.checkCoherencyBeforeSaving()
        if hasError:
            return

        out = WprjXml().serialize(self)
        if out:
            with open(path, 'w') as f:
                f.write(out)
            self.path = path
        else:
            raise IOError('Cannot save with MIME-type \'' + mime + '\'')

    def checkCoherencyBeforeSaving(self):
        hasError = False
        from wcr_core.modules.icd.wimod_application_module import \
            WIMOD_ICD_APP_CODE
        from wcr_core.modules.vlink.wvmod_application_module import \
            WVMOD_VLINK_APP_CODE

        if self.getModule() == WIMOD_ICD_APP_CODE:
            top = self.getTopHierarchy()
            messageWoProducer = top.checkMsgProducerCoherency()
            if messageWoProducer:
                if len(messageWoProducer) > 1:
                    be = 'Messages [{}] are'.format(', '.join(messageWoProducer))
                elif len(messageWoProducer) == 1:
                    be = 'Message {} is'.format(', '.join(messageWoProducer))
                showErrorWindow('Warning message Error',
                                '{} consumed but not produced,'
                                ' this will cause problems in the future.',
                                be)
                hasError = True
        elif self.getModule() == WVMOD_VLINK_APP_CODE:
            top = self.getTopHierarchy()
            messagesNotBound = top.checkMsgBindingCoherency()
            if messagesNotBound:
                msgs = '+ {}'.format('\n+ '.join(messagesNotBound))
                showErrorWindow('Message binding Error',
                                'All message imported needs to be bound to at least one traffic route.\n'
                                'List of message that are not bound:\n{}',
                                msgs)
                hasError = True

        return hasError

    # TODO: Move to ProjectDBWriter?
    def save(self):
        """
        Save the project database to its associated file path, if any.
        """
        if not self.path:
            raise IOError('Cannot save without path')

        self.saveAs(path=self.path)


class WprjProjectPayloadHandler(WcmnReadPayloadHandler):
    """
    Import payload handler for importing projects.
    """

    #: Payload types handled
    __PAYLOAD_TYPES = \
        WCMN_PAYLOAD_PROJECT_INFO | \
        WCMN_PAYLOAD_CONFIG_TABLE | \
        WCMN_PAYLOAD_RESOURCES | \
        WCMN_PAYLOAD_BLOCK

    def __init__(self, projectDB):
        """
        :param projectDB: Project database to add imported data to
        :type projectDB: WprjProjectDatabase
        """
        #: Project database
        self.__projectDB = projectDB

    @property
    def handledPayloadTypes(self):
        return self.__PAYLOAD_TYPES

    def handlePayload(self, payloadType, mimeType, payloadFunc, file=None):
        pdb = self.__projectDB
        if payloadType == WCMN_PAYLOAD_PROJECT_INFO:
            payloadFunc(pdb.tops.projectInfo)
            toph = pdb.getTopHierarchy()
            if toph:
                toph.setName(pdb.tops.projectInfo.projectName)
        elif payloadType == WCMN_PAYLOAD_CONFIG_TABLE:
            payloadFunc(pdb.tops.getLocalConfiguration())
        elif payloadType == WCMN_PAYLOAD_RESOURCES:
            payloadFunc(pdb.tops)
        elif payloadType & WCMN_PAYLOAD_BLOCK:
            for module in wmod_getModules():
                if module.getPayloadType() == payloadType:
                    toph = module.createTopLevelBlock(
                        name=pdb.tops.projectInfo.projectName)
                    payloadFunc(toph)
                    # noinspection PyTypeChecker
                    pdb.tops.setTopHierarchy(toph)
                    pdb.setModule(module.code)
                    break
            else:
                raise NotImplementedError(
                    'No module found to read payload-type {}'.format(
                        payloadType))

    def handlePayloadGetConfigtable(self):
        """
        return the local configuration value
        """
        return self.__projectDB.tops.getLocalConfiguration()


def wprj_load_project(file, xctx):
    """
    Load a project from a Silkan XML file and return the new project.

    :param file: File to read project from
    :type file: WcmnFile
    :param xctx: context
    :type xctx: WcorContextP
    :return: Project read from the given file
    :rtype: WprjProjectDatabase
    """
    projectDB = WprjProjectDatabase(path=file.path)
    handler = WcmnMimeFileHandler(WprjProjectPayloadHandler(projectDB),
                                  WPRJ_SILKAN_XML_MIME_TYPE)
    if handler.handleFile(file, xctx):
        return projectDB
    else:
        return None


def wprj_import_project(file, xctx):
    """
    Import a project from a file containing a model block and return the
    new project.

    :param file: File to import project from
    :type file: WcmnFile
    :param xctx: Context
    :type xctx: WcorContextP
    :return: Project imported from the given file
    :rtype: WprjProjectDatabase
    """
    projectDB = WprjProjectDatabase()
    handler = WcmnDefaultFileHandler(
        WprjProjectPayloadHandler(projectDB),
        WCMN_PAYLOAD_BLOCK | WCMN_PAYLOAD_PROJECT_INFO)
    if handler.handleFile(file, xctx):
        return projectDB
    else:
        return None


def wprj_createProjectDB(module):
    """
    Create an empty project database.

    :param module: Application module
    :type module: WmodApplicationModuleP
    :return: New project database
    :rtype: WprjProjectDatabase
    """
    projectDB = WprjProjectDatabase()
    projectDB.setModule(module.code)
    # noinspection PyTypeChecker
    projectDB.tops.setTopHierarchy(module.createTopLevelBlock())
    return projectDB
