# ----------------------------------------------------------------------------
# - wprj-project-verifier                                                    -
# - wild cruncher:project verifier                            core / project -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wprj_project_verifier
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher project verifier

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod

from wcr_core.project.wprj_project import WprjProjectDatabase
from wcr_core.common.wcmn_verifier_module import WcmnVerificationError


class WprjVerificationError(WcmnVerificationError):
    """
    Single error found during a hierarchy verification run.
    """

    def __init__(self, errorId, errorMsg, offended,
                 reVerifyF=None, autoFixF=None):
        """
        :param errorId: Identifier for the error type
        :type errorId: str
        :param errorMsg: Error message
        :type errorMsg: str
        :param offended: Verification module raising this error
        :type offended: WprjVerifierModuleP
        :param reVerifyF: Function to re-verify, i.e. test whether the error
                          still persists, if possible (TBD/NYI)
        :type reVerifyF: () -> WcmnVerificationError
        :param autoFixF: Function to automatically correct the error,
                         if possible (TBD/NYI)
        :type autoFixF: () -> _
        """
        super().__init__(errorId, errorMsg, reVerifyF, autoFixF)

        #: Offended verification module
        self.__offended = offended

    @property
    def offendedVerifierModule(self):
        """
        Get the verification module that raised this error.

        :return: Verification module that raised the error
        :rtype: WprjVerifierModuleP
        """
        return self.__offended


class WprjVerifierModuleP(metaclass=ABCMeta):
    # noinspection PyPropertyDefinition
    @property
    @abstractmethod
    def description(self):
        """
        Get a description for the verification performed by this module.

        The description can be used to provide further information about
        verification errors to the user.

        :return: Module description
        :rtype: str
        """
        pass

    @abstractmethod
    def verifyProjectDatabase(self, projectDB):
        """
        Verify a project database.

        :param projectDB: Project database to verify
        :type projectDB: WprjProjectDatabase
        :return: List of verification errors, if any
        :rtype: list[WprjVerificationError]
        """
        pass


class WprjHasNameVerifierModule(WprjVerifierModuleP):
    """
    Verify that a project has a non-empty name.
    """

    ERROR_ID = 'pdb.main.missing-name'

    ERROR_MESSAGE = 'Project has no name'

    DESCRIPTION = 'A project must have a name'

    def verifyProjectDatabase(self, projectDB:WprjProjectDatabase):
        if not projectDB.getProjectName():
            return [WprjVerificationError(
                self.ERROR_ID, self.ERROR_MESSAGE, self)]
        else:
            return []

    def description(self):
        return WprjHasNameVerifierModule.DESCRIPTION


class WprjHasNoWireLoopVerifierModule(WprjVerifierModuleP):
    """
    Verify that a project have't empty wire loop.
    """

    ERROR_ID = 'pdb.main.flatten_failure'

    ERROR_MESSAGE = 'Model flattening failed : '

    DESCRIPTION = 'A project must can be flattened'

    def verifyProjectDatabase(self, projectDB:WprjProjectDatabase):
        try:
            projectDB.getTopHierarchy().flatten()
        except Exception as e:
            return [WprjVerificationError(self.ERROR_ID,
                                          self.ERROR_MESSAGE + str(e), self)]
        return []

    def description(self):
        return WprjHasNoWireLoopVerifierModule.DESCRIPTION


class WprjVerificationResult(object):
    """
    Result of a DB verification run.
    """

    def __init__(self, dbErrors, hierarchyResult):
        """
        :param dbErrors: Project DB verification errors
        :type dbErrors: list[WprjVerificationError]
        :param hierarchyResult: Result of the verification of the database's
                                hierarchy
        :type hierarchyResult: WcmnVerificationResult
        """
        self.__dbErrors = dbErrors
        self.__hierarchyResult = hierarchyResult

    @property
    def allErrors(self):
        """
        Get all errors found during the database and the hierarchy
        verification.

        :return: All errors
        :rtype: list[WcmnVerificationError]
        """
        # noinspection PyTypeChecker
        return self.__dbErrors + self.__hierarchyResult.allErrors

    @property
    def projectDbErrors(self):
        """
        Get all errors found during the database verification.
        The returned list of errors does not include any errors from the
        hierarchy verification.

        :return: All database errors
        :rtype: Iterable[WprjVerificationError]
        """
        return self.__dbErrors

    @property
    def hierarchyVerificationResult(self):
        """
        Get the result of the verification of the database's hierarchy.

        :return: Hierarchy verification result
        :rtype: WcmnVerificationResult
        """
        return self.__hierarchyResult

    def hasErrors(self):
        """
        Whether the result set contains any errors.

        :return: True iff the result set contains any errors
        :rtype: bool
        """
        return bool(self.__dbErrors) or self.__hierarchyResult.hasErrors()


class WprjProjectVerifier(object):
    def __init__(self, hvrfy, modules=None):
        """
        :param hvrfy: Hierarchy verifier
        :type hvrfy: WhdbHierarchyVerifier
        :param modules: Additional project verification modules
        :type modules: Iterable[WprjVerifierModuleP] | None
        """
        self.__modules = [WprjHasNameVerifierModule(),
                          WprjHasNoWireLoopVerifierModule()]
        """:type: list[WprjVerifierModuleP]"""
        if modules:
            self.__modules.extend(modules)
        self.__hierarchyVerifier = hvrfy

    def verifyProjectDatabase(self, projectDB):
        """
        Verify the given database and its hierarchy.

        :param projectDB:
        :type projectDB: WprjProjectDatabase
        :return: Result of the database/hierarchy verification
        :rtype: WprjVerificationResult
        """
        errors = []

        for module in self.__modules:
            errors.extend(module.verifyProjectDatabase(projectDB))

        hierarchyResult = self.__hierarchyVerifier.verifyBlock(
            projectDB.getTopHierarchy(), topLevel=True)

        return WprjVerificationResult(errors, hierarchyResult)
