# ----------------------------------------------------------------------------
# - wprj-xmlio                                                               -
# - wild cruncher:wdb xml serialization/deserialization       core / project -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2013-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wprj_xmlio
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher project DB xml serialization/deserialization

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from base64 import b64decode, b64encode
from functools import partial
from xml.dom.minidom import Document
from xml.etree import ElementTree

from wcr_core.modules.wmod_application_module import wmod_getModules, \
    WMOD_EMPTY_APPLICATION_MODULE, wmod_getModule
from wcr_core.xmlio.wxio_xmlio import wxio_serializeXml, \
    wxio_deserializeXml
from wcr_core.project.wprj_project_info import WprjProjectInfo
from silkan_common.scmn_cfg_table import ScmnConfigTable
from wcr_core.common.wcmn_file_handler import *
from wcr_core.common.wcmn_mime_type import *


#: Silkan XML MIME identifier
WPRJ_SILKAN_XML_MIME_NAME = 'application/xml+silkan'
#: Silkan XML MIME type
WPRJ_SILKAN_XML_MIME_TYPE = ScmnXmlMimeType(
    WPRJ_SILKAN_XML_MIME_NAME, 'Silkan Project', 'project',
    auxExtensions=['slx'])
scmn_register_mime_type(WPRJ_SILKAN_XML_MIME_TYPE)

#: Silkan XML MIME identifier for Hardware file
WPRJ_HARDWARE_SILKAN_XML_MIME_NAME = 'application/xml+silkan/hardware'
#: Silkan XML MIME type for Hardware file
WPRJ_HARDWARE_SILKAN_XML_MIME_TYPE = ScmnXmlMimeType(
    WPRJ_HARDWARE_SILKAN_XML_MIME_NAME, 'Silkan HardWare', 'project',
    extension='hw.xml')
scmn_register_mime_type(WPRJ_HARDWARE_SILKAN_XML_MIME_TYPE)

#: Silkan XML MIME identifier for Message file
WPRJ_MESSAGE_SILKAN_XML_MIME_NAME = 'application/xml+silkan/message'
#: Silkan XML MIME type for Message file
WPRJ_MESSAGE_SILKAN_XML_MIME_TYPE = ScmnXmlMimeType(
    WPRJ_MESSAGE_SILKAN_XML_MIME_NAME, 'Silkan Messages', 'project',
    extension='msg.xml')
scmn_register_mime_type(WPRJ_MESSAGE_SILKAN_XML_MIME_TYPE)

#: Silkan XML MIME identifier for Logical file
WPRJ_LOGICAL_SILKAN_XML_MIME_NAME = 'application/xml+silkan/logical'
#: Silkan XML MIME type for Logical file
WPRJ_LOGICAL_SILKAN_XML_MIME_TYPE = ScmnXmlMimeType(
    WPRJ_LOGICAL_SILKAN_XML_MIME_NAME, 'Silkan Logical', 'project',
    extension='bus.xml')
scmn_register_mime_type(WPRJ_LOGICAL_SILKAN_XML_MIME_TYPE)

#: Silkan XML MIME identifier for System_mapper file
WPRJ_SYS_MAP_SILKAN_XML_MIME_NAME = 'application/xml+silkan/sys_map'
#: Silkan XML MIME type for System_mapper file
WPRJ_SYS_MAP_SILKAN_XML_MIME_TYPE = ScmnXmlMimeType(
    WPRJ_SYS_MAP_SILKAN_XML_MIME_NAME, 'Silkan System Mapper', 'project',
    extension='map.xml')
scmn_register_mime_type(WPRJ_SYS_MAP_SILKAN_XML_MIME_TYPE)

#: Silkan XML MIME identifier for System_mapper file
WPRJ_VLINK_SILKAN_XML_MIME_NAME = 'application/xml+silkan/vlink'
#: Silkan XML MIME type for System_mapper file
WPRJ_VLINK_SILKAN_XML_MIME_TYPE = ScmnXmlMimeType(
    WPRJ_VLINK_SILKAN_XML_MIME_NAME, 'Silkan Traffic Route', 'project',
    extension='tr.xml')
scmn_register_mime_type(WPRJ_VLINK_SILKAN_XML_MIME_TYPE)

#: Silkan CSV MIME identifier for CSV file
WPRJ_SILKAN_CSV_MIME_NAME = 'application/csv+silkan'
#: Silkan CSV MIME type for System_mapper file
WPRJ_SILKAN_CSV_MIME_TYPE = ScmnMimeType(
    WPRJ_SILKAN_CSV_MIME_NAME, 'CSV file',
    extensions=['csv'])
scmn_register_mime_type(WPRJ_SILKAN_CSV_MIME_TYPE)

#: Silkan XML MIME identifier for ICD file
WPRJ_ICD_SILKAN_XML_MIME_NAME = 'application/xml+silkan/icd'
#: Silkan XML MIME type for System_mapper file
WPRJ_ICD_SILKAN_XML_MIME_TYPE = ScmnXmlMimeType(
    WPRJ_ICD_SILKAN_XML_MIME_NAME, 'Silkan ICD', 'project',
    extension='icd.xml')
scmn_register_mime_type(WPRJ_ICD_SILKAN_XML_MIME_TYPE)


class WprjSilkanXmlFileReader(WcmnFileReaderBase):
    """
    File reader for Silkan XML files.
    """
    #: Readable MIME types
    __MIME_TYPES = [WPRJ_SILKAN_XML_MIME_TYPE]

    #: Payload types generated
    __PAYLOAD_TYPES = \
        WCMN_PAYLOAD_PROJECT_INFO | \
        WCMN_PAYLOAD_CONFIG_TABLE | \
        WCMN_PAYLOAD_RESOURCES | \
        WCMN_PAYLOAD_BLOCK

    def __init__(self, auxmtype=None, auxpayloadtype=None):
        if auxmtype:
            self.__MIME_TYPES=auxmtype
        if auxpayloadtype:
            self.__PAYLOAD_TYPES = auxpayloadtype
        super().__init__(self.__MIME_TYPES, self.__PAYLOAD_TYPES)

        #: Payload functions
        self.__PAYLOAD_FUNCS = [
            (WCMN_PAYLOAD_PROJECT_INFO, self._getProject,
             self._readProjectInfo, WCMN_PAYLOAD_PROJECT_INFO),
            (WCMN_PAYLOAD_CONFIG_TABLE, lambda t: t.find('local-config'),
             self._readConfigTable, WCMN_PAYLOAD_CONFIG_TABLE),
            (WCMN_PAYLOAD_RESOURCES, lambda t: t.find('resources'),
             self._readResources, WCMN_PAYLOAD_RESOURCES)]

    @property
    def name(self):
        return 'Silkan XML Database'

    def readFile(self, file, payloadHandler, xctx, types=None):
        types = types if types is not None else payloadHandler.payloadTypes
        typesRead = SCMN_PAYLOAD_NONE
        tree = ElementTree.ElementTree(file=file.path)
        for ctyp, xmlf, func, rtyp in self._getPayloadFunctions():
            payload = xmlf(tree)
            if types & ctyp and payload is not None:
                payloadHandler.handlePayload(
                    ctyp, WPRJ_SILKAN_XML_MIME_TYPE, partial(func, tree))
                typesRead |= rtyp
        return typesRead | WCMN_PAYLOAD_RESOURCES

    def _getProject(self, tree):
        root = tree.getroot()
        if root is not None and root.tag == 'project':
            return root

    def _readProjectInfo(self, tree, info):
        """
        Read the project info from the XML tree and add the read data to the
        given info structure.

        :param tree: XML tree
        :type tree: ElementTree
        :param info: Info structure to add data to
        :type info: WprjProjectInfo
        """
        # noinspection PyUnresolvedReferences
        projectNode = tree.getroot()
        if projectNode.tag != 'project':
            raise Exception('XML document must start with project element')
        name = projectNode.get('name')
        if name:
            info.projectName = name
        author = projectNode.get('author')
        if author:
            info.author = author
        version = projectNode.get('version')
        if version:
            info.version = version
        company = projectNode.get('company')
        if company:
            info.company = company
        # noinspection PyUnresolvedReferences
        descriptionNode = tree.find('description')
        if descriptionNode is not None and descriptionNode.text:
            info.projectInfo = descriptionNode.text

    def _readConfigTable(self, tree, table):
        """
        Read the configuration data from the XML tree and add the read data to
        the given config table.

        :param tree: XML tree
        :type tree: xml.etree.ElementTree.ElementTree
        :param table: Config table to add data to
        :type table: ScmnConfigTable
        """
        for e in tree.find('local-config'):
            table.newRecord(e.get('name'),
                            e.get('info'),
                            e.get('value'),
                            e.text)

    def _readResources(self, tree, tops):
        """

        :param tree: XML tree
        :type tree: xml.etree.ElementTree.ElementTree
        :param tops: Destination for resources read
        :type tops: WprjProjectStructure
        """
        for r in tree.find('resources'):
            rsrc = r.text
            if rsrc:
                tops.setResource(
                    r.get('name'), b64decode(rsrc.encode('UTF-8')))

    def _readBlockModel(self, tree, block):
        """
        Read the model elements from the XML tree and add the read data to the
        given model block.

        :param tree: XML tree
        :type tree: xml.etree.ElementTree.ElementTree
        :param block: Model block to add data to
        :type block: WhdbBlockP
        """
        # noinspection PyTypeChecker
        wxio_deserializeXml(
            tree, 'tophm', block.rawElement)
        return block

    @staticmethod
    def _getHierarchyNode(module, tree):
        node = tree.find('hierarchy')
        if node is not None and node.get('module') == module:
            return node
        else:
            return None

    def _getPayloadFunctions(self):
        funcs = list(self.__PAYLOAD_FUNCS)
        for m in wmod_getModules():
            if m is WMOD_EMPTY_APPLICATION_MODULE:
                continue
            funcs.append(
                (m.getPayloadType(),
                 partial(self._getHierarchyNode, m.code),
                 partial(self._readBlockModel),
                 WCMN_PAYLOAD_BLOCK))
        return funcs


WPRJ_SILKAN_XML_MIME_TYPE.addReader(WprjSilkanXmlFileReader())

class WprjSilkanModuleFileReader(WprjSilkanXmlFileReader):
    """
    File reader for Silkan Module specific files.
    """
    #: Readable MIME types
    # __MIME_TYPES = [WPRJ_HARDWARE_SILKAN_XML_MIME_TYPE]

    #: Payload types generated
    __PAYLOAD_TYPES = \
        WCMN_PAYLOAD_PROJECT_INFO | \
        WCMN_PAYLOAD_CONFIG_TABLE | \
        WCMN_PAYLOAD_RESOURCES

    def __init__(self, mimetype, modulepayload):
        super().__init__(auxmtype=mimetype,#self.__MIME_TYPES,
                         auxpayloadtype=self.__PAYLOAD_TYPES | modulepayload)

#Reader for hardware module files
WPRJ_HARDWARE_SILKAN_XML_MIME_TYPE.addReader(
    WprjSilkanModuleFileReader([WPRJ_HARDWARE_SILKAN_XML_MIME_TYPE],
                               WCMN_PAYLOAD_BLOCK_HARDWARE))

#Reader for message module files
WPRJ_MESSAGE_SILKAN_XML_MIME_TYPE.addReader(
    WprjSilkanModuleFileReader([WPRJ_MESSAGE_SILKAN_XML_MIME_TYPE],
                               WCMN_PAYLOAD_BLOCK_MESSAGE))

#Reader for Logical module files (Bus)
WPRJ_LOGICAL_SILKAN_XML_MIME_TYPE.addReader(
    WprjSilkanModuleFileReader([WPRJ_LOGICAL_SILKAN_XML_MIME_TYPE],
                               WCMN_PAYLOAD_BLOCK_BUS))

#Reader for System Mapping module
WPRJ_SYS_MAP_SILKAN_XML_MIME_TYPE.addReader(
    WprjSilkanModuleFileReader([WPRJ_SYS_MAP_SILKAN_XML_MIME_TYPE],
                               WCMN_PAYLOAD_BLOCK_SYS_MAP))

#Reader for Traffic Route module
WPRJ_VLINK_SILKAN_XML_MIME_TYPE.addReader(
    WprjSilkanModuleFileReader([WPRJ_VLINK_SILKAN_XML_MIME_TYPE],
                               WCMN_PAYLOAD_BLOCK_VLINK))

#Reader for Traffic Route module
WPRJ_SILKAN_CSV_MIME_TYPE.addReader(
    WprjSilkanModuleFileReader([WPRJ_SILKAN_CSV_MIME_TYPE],
                               WCMN_PAYLOAD_BLOCK_CSV))

#Reader for ICD module
WPRJ_ICD_SILKAN_XML_MIME_TYPE.addReader(
    WprjSilkanModuleFileReader([WPRJ_ICD_SILKAN_XML_MIME_TYPE],
                                WCMN_PAYLOAD_BLOCK_ICD))


class WprjXml(object):
    """
    Serialization/deserialization of project databases from/to XML.
    """

    def __init__(self):
        pass

    def serialize(self, pdb):
        """
        Serialize the content of the given configuration table and the model
        hierarchy as XML to the specified file.

        :param pdb: Project DB
        :type pdb: WprjProjectDatabase
        :return: Serialization as XML-string
        :rtype: str
        """

        xml = Document()

        pinf = pdb.tops.projectInfo
        rtbl = pdb.tops.getLocalConfiguration()
        tphm = pdb.tops.getTopHierarchy()

        projectNode = xml.createElement('project')
        if pinf.projectName:
            projectNode.setAttribute('name', pinf.projectName)
        if pinf.projectInfo:
            descNode = xml.createElement('description')
            descNode.appendChild(xml.createTextNode(pinf.projectInfo))
            projectNode.appendChild(descNode)
        if pinf.author:
            projectNode.setAttribute('author', pinf.author)
        if pinf.company:
            projectNode.setAttribute('company', pinf.company)
        if pinf.version:
            projectNode.setAttribute('version', pinf.version)

        projectNode.appendChild(
            wxio_serializeXml(
                tphm.rawElement, xml, 'tophm', pdb.getModule()))
        projectNode.appendChild(self._serializeLocalConfiguration(rtbl, xml))
        projectNode.appendChild(self._serializeResources(pdb.tops, xml))

        xml.appendChild(projectNode)

        return xml.toprettyxml(indent='  ')

    def _serializeLocalConfiguration(self, rtbl, xml):
        """
        :param rtbl: Configuration table
        :type rtbl: ScmnConfigTable
        :param xml: XML Document
        :type xml: Document
        :return: Created XML elemnet
        :rtype: Element
        """
        tableNode = xml.createElement('local-config')

        for k, r in rtbl:
            if not r.getValue():
                continue

            recordNode = xml.createElement('record')

            if r.getName():
                recordNode.setAttribute('name', r.getName())
            if r.getInfo():
                recordNode.setAttribute('info', r.getInfo())
            if r.getValue():
                recordNode.setAttribute('value', r.getValue())
            if r.getText():
                recordNode.appendChild(xml.createTextNode(r.getText()))

            tableNode.appendChild(recordNode)

        return tableNode

    def _serializeResources(self, tops, xml):
        """
        :param tops: Project structure
        :type tops: WprjProjectStructure
        :param xml: XML Document
        :type xml: Document
        :return: Created XML elemnet
        :rtype: Element
        """
        node = xml.createElement('resources')

        for key, rsrc in tops.getResources():
            if not rsrc:
                continue
            rnod = xml.createElement('resource')
            rnod.setAttribute('name', key)
            rraw = rsrc if isinstance(rsrc, bytes) else rsrc.encode('UTF-8')
            renc = (b64encode(rraw) if rraw else None).decode('UTF-8')

            rnod.appendChild(xml.createTextNode(renc))
            node.appendChild(rnod)

        return node
