# ----------------------------------------------------------------------------
# - wcor-context                                                             -
# - wild cruncher - context procotols                                   core -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcor_context
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher context procotols

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod
import os

from silkan_common.scmn_log import ScmnMessageTargetP


# noinspection PyAbstractClass
class WcorContextP(ScmnMessageTargetP):
    """
    Protocol for contexts in general.
    """

    @abstractmethod
    def getConfigurationValues(self):
        """
        Get the available configuration name/value pairs.

        :return: Available configuration name/value pairs
        :rtype: dict[str, str]
        """
        raise NotImplementedError()

    @abstractmethod
    def getConfigurationValue(self, key):
        """
        Get the configuration value for the given key, if any.

        :param key: Key to retrieve value for
        :type key: str
        :return: Value for the given key, or `None`
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getMessageSource(self):
        """
        Get the context's message source.

        :return: Message source
        :rtype: ScmnMessageSourceP
        """
        raise NotImplementedError()

    @abstractmethod
    def handleFile(self, path):
        """
        Handle the file with the given path with the matching registered
        file handler, if any.

        :param path: Path of the file to handle
        :type path: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getRecentlyOpenedFiles(self):
        """
        Get the list of recently opened files.

        :return: Recently opened files
        :rtype: Iterable[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def addRecentlyOpenedFile(self, path):
        """
        Add a path to the list of recently opened files.

        :param path: Path to add
        :type path: str
        """
        # TODO: Use handleFile(...) and pass file handlers?
        raise NotImplementedError()

    @abstractmethod
    def createContextWith(self, msgTarget=None):
        """
        Create a new context as a copy of the current one, with certain
        functionalities amended/replaced.

        :param msgTarget: Additional message target to use
        :type msgTarget: ScmnMessageTargetP
        :return: New context with the given functionalities amended/replaced
        :rtype: WcorContextP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WcorGlobalContextP(WcorContextP):
    """
    Protocol for global contexts.
    """

    @abstractmethod
    def getWorkspaceManager(self):
        """
        Get the workspace manager.

        :return: Global workspace manager
        :rtype: WcorWorkspaceManagerP
        """
        raise NotImplementedError()

    @abstractmethod
    def createContextWith(self, msgTarget=None):
        """
        Create a new global context as a copy of this one, with certain
        functionalities amended/replaced.

        :param msgTarget: Additional message target to use
        :type msgTarget: ScmnMessageTargetP
        :return: New global context with the given functionalities amended/
                 replaced
        :rtype: WcorGlobalContextP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WcorProjectContextP(WcorContextP):
    """
    Protocol for project contexts.
    """

    @abstractmethod
    def getModule(self):
        """
        Get the module of the project.

        :return: The module
        :rtype: WmodApplicationModuleP
        """
        raise NotImplementedError()

    @abstractmethod
    def getGlobalContext(self):
        """
        Get the global context.

        :return: Global context
        :rtype: WcorGlobalContextP
        """
        raise NotImplementedError()

    @abstractmethod
    def getProjectName(self):
        """
        Get the project's name.

        :return: Project name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getProjectDB(self):
        """
        Get the project database.

        :return: Project database
        :rtype: WprjProjectDatabase
        """
        raise NotImplementedError()

    @abstractmethod
    def getProjectWorkspace(self):
        """
        Get the project's workspace.

        :return: Workspace for the project
        :rtype: WcorProjectWorkspaceP
        """
        raise NotImplementedError()

    @abstractmethod
    def createContextWith(self, msgTarget=None):
        """
        Create a new project context as a copy of this one, with certain
        functionalities amended/replaced.

        :param msgTarget: Additional message target to use
        :type msgTarget: ScmnMessageTargetP
        :return: New project context with the given functionalities amended/
                 replaced
        :rtype: WcorProjectContextP
        """
        raise NotImplementedError()

    @abstractmethod
    def getControllerFactory(self):
        """
        Get the controller factory for the project.

        :return: Controller factory
        :rtype: WctlControllerFactoryP
        """
        raise NotImplementedError()


#: Global configuration defaults registry
_WCOR_CFG_DEFAULTS = {}


def wcor_registerConfigKey(name, info, default):
    """
    Register a new configuration default.

    :param name: Name of the configuration entry
    :type name: str
    :param info: Description of the configuration entry
    :type info: str
    :param default: Default value
    :type default: str
    """
    _WCOR_CFG_DEFAULTS[name] = (info, default)


def wcor_getConfigDefaults():
    """
    Get all configuration default values.

    :return: Configuration key/default-value pairs
    :rtype: dict[str, str]
    """
    # noinspection PyShadowingNames
    return {k: v[1] for k, v in _WCOR_CFG_DEFAULTS.items()}


def wcor_getConfigDefault(name):
    """
    Get a configuration default value, if any.

    :param name: Configuration entry to get default value for
    :type name: str
    :return: Default value for the given name, or `None`
    :rtype: str
    """
    entry = _WCOR_CFG_DEFAULTS.get(name)
    return entry[1] if entry is not None else None


def wcor_getConfigDescription(name):
    """
    Get a configuration entry description.

    :param name: Configuration entry to get description for
    :type name: str
    :return: Description for the given configuration entry, if any
    :rtype: str
    """
    entry = _WCOR_CFG_DEFAULTS.get(name)
    return entry[0] if entry is not None else ''


for k, v in os.environ.items():
    if k.startswith('WCR_'):
        wcor_registerConfigKey(k, '', v)

