# ----------------------------------------------------------------------------
# - wcor-global-context                                                      -
# - wild cruncher:wac shell context                                     core -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2013 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcor_global_context
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher global context

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------
import traceback

from silkan_common import scmn_log
from wcr_core.wcor_context import WcorGlobalContextP, \
    wcor_getConfigDefaults, wcor_getConfigDefault
from wcr_core.common.wcmn_file_path import wcmn_path_data
from silkan_common.scmn_cfg_table import ScmnConfigTable
from wcr_core.common.wcmn_file_handler import WcmnFile
from wcr_core.wcor_workspace import WcorWorkspaceManager
from silkan_common.scmn_log import ScmnLogger
from silkan_common.scmn_file_path import scmn_path_join, scmn_isdir, \
    scmn_mkdir, scmn_abspath, scmn_isfile


# ----------------------------------------------------------------------------
# - context class                                                            -
# ----------------------------------------------------------------------------

#: Recently-opened-files storage
_RECENTS_FILE = 'recent.txt'

#: Number of files in recently-opened list
_RECENTS_FILE_LIMIT = 20


class WcorGlobalContext(WcorGlobalContextP):
    """
    The WcorGlobalContext class is the global application context which
    holds the project independent services. The application context
    provides the global configuration table, the system workspace and the
    system logger. Extra launcher for the graphical application are also
    part of this class.
    """

    def __init__(self, wspcMngr=None):
        """
        :param wspcMngr: Workspace manager
        :type wspcMngr: WcorWorkspaceManagerP | None
        """
        #: Configuration
        self.__configTable = ScmnConfigTable()
        if not wspcMngr:
            wspcMngr = WcorWorkspaceManager(self.__configTable)
        #: Workspace
        self.__workspace = wspcMngr

        logSize = self.getConfigurationValue('WCR_LOG_SIZE')
        logToStdout = self.getConfigurationValue('WCR_LOG_STDOUT')
        #: System logger
        self.__logger = ScmnLogger(
            int(logSize) if logSize else None,
            bool(logToStdout) if logToStdout else False)
        #: Registered file handlers
        self.__fileHandlers = []
        """:type: list[WcmnFileHandler]"""

    def getConfigurationService(self):
        """
        Get the configuration service.

        :return: the configuration service
        :rtype: ScmnConfigTable
        """
        return self.__configTable

    def getConfigurationValues(self):
        vals = dict(wcor_getConfigDefaults())
        vals.update({k: r.getValue() for k, r
                     in self.__configTable if r.getValue()})
        return vals

    def getConfigurationValue(self, key):
        gval = self.__configTable.getRecordValue(key)
        return gval if gval else wcor_getConfigDefault(key)

    def registerFileHandler(self, handler):
        """
        Register a file handler.

        :param handler: File handler to be registered
        :type handler: WcmnFileHandler
        """
        self.__fileHandlers.append(handler)

    def handleFile(self, path):
        if not path:
            self.addMessage(
                'Requested file handling for empty path',
                'appctx', scmn_log.SCMN_LOG_INFO)
            return

        file = WcmnFile(path)

        if not scmn_isfile(file.path):
            self.addMessage(
                'No such file: ' + path,
                'appctx', scmn_log.SCMN_LOG_ERROR)
            return

        for handler in self.__fileHandlers:
            try:
                if handler.canHandleFile(file):
                    handler.handleFile(file, self)
                    return
            except Exception as e:
                traceback.print_exc()
                self.addMessage(
                    'File handler {} failed to handle {} ({})'.format(
                        handler, path, e),
                    'appctx', scmn_log.SCMN_LOG_ERROR,
                    trce=traceback.format_exc())
        self.addMessage(
            'Could not find suitable handler for ' + path,
            'appctx', scmn_log.SCMN_LOG_WARN)

    def addMessage(self, mtxt, mcmp=None, mcat=None, mlvl=None, trce=None):
        return self.__logger.addMessage(mtxt, mcmp, mcat, mlvl, trce)

    def getMessageSource(self):
        return self.__logger

    def getRecentlyOpenedFiles(self):
        path = scmn_path_join(wcmn_path_data(), _RECENTS_FILE)
        try:
            with open(path, 'r') as rfile:
                return list(
                    filter(lambda l: bool(l),
                           [l.strip() for l in rfile.readlines()]))
        except IOError:
            return []

    def addRecentlyOpenedFile(self, path):
        if not path:
            return

        recents = self.getRecentlyOpenedFiles()
        path = scmn_abspath(path)
        if path in recents:
            recents.remove(path)
        if len(recents) >= _RECENTS_FILE_LIMIT:
            recents = recents[len(recents) - _RECENTS_FILE_LIMIT + 1:]
        recents.append(path)

        bdir = wcmn_path_data()
        path = scmn_path_join(bdir, _RECENTS_FILE)
        try:
            if not scmn_isdir(bdir):
                scmn_mkdir(bdir)

            with open(path, 'w') as rfile:
                for line in recents:
                    rfile.write(line + '\n')
        except IOError:
            pass

    def getWorkspaceManager(self):
        return self.__workspace

    def createContextWith(self, msgTarget=None):
        return _WcorGlobalContextWrapper(self, msgTarget)


# noinspection PyAbstractClass
class _WcorGlobalContextWrapper(WcorGlobalContextP):
    """
    Global context implementation for amended contexts.
    """

    def __init__(self, delegate, msgTarget=None):
        """
        :param delegate: Delegate context for default functionality
        :type delegate: WcorGlobalContextP
        :param msgTarget: Additional message target
        :type msgTarget: ScmnMessageTargetP
        """
        #: Delegate context
        self.__delegate = delegate

        #: Additional message target
        self.__msgTarget = msgTarget

    def addMessage(self, mtxt, mcmp=None, mcat=None, mlvl=None, trce=None):
        if self.__msgTarget is not None:
            self.__msgTarget.addMessage(mtxt, mcmp, mcat, mlvl, trce)

        return self.__delegate.addMessage(mtxt, mcmp, mcat, mlvl, trce)

    def getRecentlyOpenedFiles(self):
        return self.__delegate.getRecentlyOpenedFiles()

    def addRecentlyOpenedFile(self, path):
        return self.__delegate.addRecentlyOpenedFile(path)

    def getConfigurationValue(self, key):
        return self.__delegate.getConfigurationValue(key)

    def getMessageSource(self):
        return self.__delegate.getMessageSource()

    def handleFile(self, path):
        return self.__delegate.handleFile(path)

    def getConfigurationValues(self):
        return self.__delegate.getConfigurationValues()

    def getWorkspaceManager(self):
        return self.__delegate.getWorkspaceManager()

    def createContextWith(self, msgTarget=None):
        return self.__delegate.createContextWith(msgTarget)

    def getConfigurationService(self):
        return self.__delegate.getConfigurationService()
