# ----------------------------------------------------------------------------
# - wcor-workspace                                                           -
# - wild cruncher:wwm workspace object                                  core -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 PiV                                              -
# - author    (c) 2013-2013 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcor_workspace
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher application workspace

.. moduleauthor:: Amaury Darsch <amaury.darsch@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from abc import ABCMeta
from tempfile import mkdtemp

from silkan_common.scmn_cfg_table import ScmnConfigTable
from silkan_common.scmn_file_path import *
from wcr_core.common.wcmn_file_path import wcmn_path_data


# ----------------------------------------------------------------------------
# - workspace class                                                          -
# ----------------------------------------------------------------------------


class WcorWorkspaceManagerP(metaclass=ABCMeta):
    """
    Protocol for workspace managers.
    Workspace managers provide workspaces for projects.
    """

    def getProjectWorkspace(self, projectName):
        """
        Get the workspace for the project with the given name.

        :return: Project workspace for the given project
        :rtype: WcorProjectWorkspaceP
        """
        raise NotImplementedError()


class WcorProjectWorkspaceP(metaclass=ABCMeta):
    """
    Protocol for project workspaces.
    """

    def getProjectPath(self):
        """
        Get the root path of the project workspace.

        :return: Project workspace's root path
        :rtype: str
        """
        raise NotImplementedError()

    def getDirectoryPath(self, prof=None, sgrp=None):
        """
        Get the path of the directory for the given scheduling group in the
        given profile.

        :param prof: Profile
        :type prof: str
        :param sgrp: Scheduling group
        :type sgrp: str
        :return: Directory path
        :rtype: str
        """
        raise NotImplementedError()

    def getFilePath(self, name, prof=None, sgrp=None):
        """
        Get the path of the file with the given name for the given
        scheduling group in the given profile.

        :param name: File name
        :type name: str
        :param prof: Profile
        :type prof: str
        :param sgrp: Scheduling group
        :type sgrp: str
        :return: File path
        :rtype: str
        """
        raise NotImplementedError()

    def importText(self, name, prof=None, sgrp=None, text=None):
        """
        Import a text into the workspace.

        :param name: the text file name
        :param prof: the profile name
        :param sgrp: the scheduling group
        :param text: the text to import
        """
        raise NotImplementedError()

    def importFile(self, ifnm, ofnm, prof=None, sgrp=None):
        """
        Import a file into the workspace.

        :param ifnm: the input file name
        :param ofnm: the output file name
        :param prof: the profile name
        :param sgrp: the scheduling group
        """
        raise NotImplementedError()


class WcorWorkspaceManager(WcorWorkspaceManagerP):
    # default root path
    __DEFAULT_ROOT_PATH = 'workspace'

    def __init__(self, wcst=None):
        """
        Create a default workspace

        :param wcst: the configuration service
        :type wcst: ScmnConfigTable
        """
        # get the root path
        self.__root = self._getRootPath(wcst)
        # make sure the workspace exists
        scmn_mkdir(self.__root)

    def _getRootPath(self, wcst):
        """
        Get the workspace root path

        :param wcst: the configuration service
        """
        if wcst is not None and wcst.isRecord('BAS_WSP_ROOT'):
            return wcst.getRecordValue('BAS_WSP_ROOT')
        # get the default os workspace
        name = WcorWorkspaceManager.__DEFAULT_ROOT_PATH
        # preset the path
        return scmn_path_join(wcmn_path_data(), name)

    def getProjectWorkspace(self, projectName):
        return WcorProjectWorkspace(
            scmn_path_join(self.__root, projectName))


class WcorTempWorkspaceManager(WcorWorkspaceManagerP):
    """
    Workspace manager that creates temporary directories as workspaces
    for projects.

    It is the client's responsibility to delete those directories after use
    (see `getAllWorkspaces()`).
    """
    def __init__(self):
        self.__workspaces = {}

    def getProjectWorkspace(self, projectName):
        wspc = self.__workspaces.get(projectName)
        if not wspc:
            wspc = mkdtemp()
            self.__workspaces[projectName] = wspc
        return WcorProjectWorkspace(wspc)

    def getAllWorkspaces(self):
        """
        :return: All created workspaces (projectName, dir)
        :rtype: Iterable[(str, str)]
        """
        return self.__workspaces.items()


class WcorProjectWorkspace(WcorProjectWorkspaceP):
    def __init__(self, projectRoot):
        self.__root = projectRoot
        # create the root
        scmn_mkdir(self.__root)

    def getProjectPath(self):
        """
        Get a project path by name

        :return:     the path for a given project name
        """
        return self.__root

    def getDirectoryPath(self, prof='default', sgrp='default'):
        # add the profile
        path = scmn_path_join(self.__root,
                              prof or 'default',
                              sgrp or 'default')
        # create the path
        scmn_mkdir(path)
        # here it is
        return path

    def getFilePath(self, name, prof=None, sgrp=None):
        # get the project path
        path = self.getDirectoryPath(prof, sgrp)
        # add the instance name
        return scmn_path_join(path, name)

    def importText(self, name, prof=None, sgrp=None, text=None):
        # do nothing if none
        if text is None: return
        # get the path
        path = self.getFilePath(name, prof, sgrp)
        # open a file for writing
        with open(path, 'w') as f:
            f.write(text)

    def importFile(self, ifnm, ofnm, prof=None, sgrp=None):
        # get the input path
        ipth = self.getFilePath(ifnm, prof, sgrp)
        if not scmn_isfile(ipth): return
        # get the input text
        with open(ipth, 'r') as f:
            text = f.read()
        # import with text
        self.importText(ofnm, prof, sgrp, text)
