# ----------------------------------------------------------------------------
# - wxio_xml-0-1-0                                                           -
# - wild cruncher:xml serialization v1.0                        core / xmlio -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wxio_xml_0_1_0
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher xml serialization/deserialization V1.0

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db.whdb_connection_point import WHDB_DEFAULT_XPOINT_CODE
from wcr_core.dbstore_api.wdbs_meta_type_api import WdbsAtomicMetaTypeP

from wcr_core.xmlio.wxio_xml_base import *


class WxioXmlVer0_1_0(WxioXmlBase):
    """
    XML-IO Version 1.
    """

    #: XML-IO Version
    _VERSION = '1.0'

    def __init__(self, ignoreUUID=False):
        """
        :param ignoreUUID: Whether to ignore UUIDs during serialization
                           (and generate new UUIDs for all entities instead)
        :type ignoreUUID: bool
        """
        super().__init__(ignoreUUID)

    @staticmethod
    def getVersion():
        """
        Get the XML-IO version.

        :return: XML-IO version
        :rtype: str
        """
        return WxioXmlVer0_1_0._VERSION

    def serialize(self, block, xml, topName, module):
        raise NotImplementedError()

    @staticmethod
    def _fix_field_name(name):
        """
        :type name: str
        :rtype: str
        """
        return name.replace('_', '-')

    def setMetaForItem(self, node, item):
        if node is None:
            return
        data = []
        for mf in item.metaDescription.fields:
            if mf.is_temporary:
                continue
            rdrf, _ = mf.getRWFuncs(self._CODE, self._VERSION)
            mtyp = mf.metaType

            if rdrf:
                data.append((mf.name, rdrf(node)))
            elif isinstance(mtyp, WdbsAtomicMetaTypeP):
                rval = node.get(self._fix_field_name(mf.name))
                if rval:
                    data.append((mf.name, mtyp.fromString(rval)))
            else:
                raise Exception(
                    '{s}-Serialization V{v} cannot handle '
                    '{c}.{m} ({t})'.format(
                        s=self._CODE, v=self._VERSION, t=mtyp,
                        c=item.__class__.__name__, m=mf.name))
        for k, v in data:
            item.setMeta(v, k)

    def getBlockNodes(self, node):
        modelNodes = node.find('models')
        if modelNodes is not None:
            return (bnod for bnod in modelNodes if bnod.tag == 'mbloc')
        return []

    def readBlockData(self, node):
        return WxioXmlBlock(
            uuid=node.get('uuid') or node.get('sid'),
            type=node.get('blockType'),
            name=node.get('name'),
            info=node.get('info'))

    def getBlockInstanceNodes(self, node, blocks):
        instsNode = node.find('minsts')
        if instsNode is not None:
            return (node for node in instsNode if node.get('imdl') in blocks)
        return []

    def getModelInstanceNodes(self, node, models):
        instsNode = node.find('minsts')
        if instsNode is not None:
            return (node for node in instsNode if node.get('imdl') in models)
        return []

    def getOperatorInstanceNodes(self, node):
        instsNode = node.find('mfuncs')
        if instsNode is not None:
            return iter(instsNode)
        return []

    def getConnectionNodes(self, node):
        connsNode = node.find('mwires')
        if connsNode is not None:
            return iter(connsNode)
        return []

    def getShapeNodes(self, node):
        shapesNode = node.find('shapes')
        if shapesNode is not None:
            return iter(shapesNode)
        return []

    def getButtonNodes(self, node):
        buttonsNode = node.find('interacts')
        if buttonsNode is not None:
            return iter(buttonsNode)
        return []

    def getModelNodes(self, node):
        modelNodes = node.find('models')
        if modelNodes is not None:
            return (mnod for mnod in modelNodes if mnod.tag == 'mtext')
        return []

    def readModelData(self, node):
        return WxioXmlModel(
            uuid=node.get('uuid') or node.get('sid'),
            type=node.get('modelType'),
            name=node.get('name'),
            info=node.get('info'),
            path=node.get('path'),
            entry=node.get('fnam'))

    def readXPointData(self, node):
        ctyp = WxioXmlBase.XPOINT_DIR_INT[node.get('ctyp')]

        return WxioXmlXPoint(
            uuid=node.get('uuid') or node.get('sid'),
            type=WHDB_DEFAULT_XPOINT_CODE,
            direction=ctyp)

    def getXPointNodes(self, node):
        return node.findall('xconn')

    def readBlockInstanceData(self, node):
        return WxioXmlBlockInst(
            uuid=node.get('uuid') or node.get('sid'),
            name=node.get('name'),
            info=node.get('info'),
            block=node.get('imdl'))

    def readModelInstanceData(self, node):
        return WxioXmlModelInst(
            uuid=node.get('uuid') or node.get('sid'),
            name=node.get('name'),
            info=node.get('info'),
            model=node.get('imdl'))

    def readOperatorInstanceData(self, node):
        return WxioXmlOpInst(
            uuid=node.get('uuid') or node.get('sid'),
            name=node.get('name'),
            info=node.get('info'),
            status=node.get('status'),
            statusMask=node.get('statusMask'),
            oper=node.get('oper'))

    def readConnectionData(self, node):
        pathNode = node.find('path')
        path = None
        if pathNode is not None:
            path = [(float(pn.get('x')), float(pn.get('y')))
                    for pn in pathNode.findall('point')]

        return WxioXmlConnection(
            uuid=node.get('uuid') or node.get('sid'),
            sobj=node.get('sobj'), sxpnt=node.get('scon'),
            eobj=node.get('eobj'), expnt=node.get('econ'),
            name=node.get('name'), info=node.get('info'), path=path)

    def readShapeData(self, node):
        return WxioXmlShape(
            uuid=node.get('uuid') or node.get('sid'),
            stype=node.tag)

    def readButtonData(self, node):
        return WxioXmlButton(
            uuid=node.get('uuid') or node.get('sid'),
            title=node.get('title'),
            command=node.get('command'))

    def readParameters(self, nodes):
        return [(node.get('name'), node.get('value')) for node in nodes
                if node.get('value') != 'None']

    def readOperatorInstanceParameters(self, node):
        return self.readParameters(node.findall('mparam'))

    def readButtonParameters(self, node):
        return self.readParameters(node.findall('cmd-param'))
