# ----------------------------------------------------------------------------
# - wxio-xml-0-2-0                                                           -
# - wild cruncher:xml serialization v2.0                        core / xmlio -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wxio_xml_0_2_0
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher xml serialization/deserialization V2.0

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from operator import attrgetter
from xml.dom.minidom import Document

from wcr_core.dbstore_api.wdbs_meta_type_api import WdbsAtomicMetaTypeP
from wcr_core.xmlio.wxio_xml_base import *
from wcr_core.xmlio.wxio_nested_xml_base import WxioNestedXml, _META_DATA_TAG


_META_ITEM_TAG = 'meta-item'


class WxioXmlVer0_2_0(WxioNestedXml):
    """
    XML-IO Version 2.
    """

    #: XML-IO Version
    _VERSION = '2.0'

    def __init__(self, ignoreUUID=False):
        """
        :param ignoreUUID: Whether to ignore UUIDs during serialization
                           (and generate new UUIDs for all entities instead)
        :type ignoreUUID: bool
        """
        super().__init__(ignoreUUID)

    @staticmethod
    def getVersion():
        return WxioXmlVer0_2_0._VERSION

    def _createMetaNode(self, name, value, xml):
        """
        Create a meta-data node for the given name-value pair.

        :param name: Meta-data name
        :type name: str
        :param value: Meta-data value
        :type value: str
        :param xml: XML Document
        :type xml: Document
        :return: Created node, holding the given value
        :rtype: xml.dom.minidom.Element
        """
        node = xml.createElement(_META_ITEM_TAG)
        node.setAttribute('name', name)
        if value is not None:
            self._setNodeValue(str(value), node, xml)
        return node

    def _writeMetaValue(self, field, value, node, xml):
        """
        Add the given meta-data to the given node.

        :param field:Meta-data field descriptor
        :type field: tuple | str
        :param value: Meta-data value
        :type value: str
        :param node: Node to add meta-data to
        :type node: xml.dom.minidom.Element
        :param xml: XML Document
        :type xml: Document
        """
        if not value:
            return
        if isinstance(field, tuple):
            self._writeMetaValue(field[-1], value, node, xml)
        else:
            node.appendChild(self._createMetaNode(
                self.fixFieldName(field), value, xml))

    def _writeMetaData(self, item, node, xml):
        metaNode = xml.createElement(_META_DATA_TAG)

        if isinstance(item, WdbsNamedEntityP):
            metaNode.appendChild(self._createMetaNode(
                'name', item.getName(), xml))
            metaNode.appendChild(self._createMetaNode(
                'info', item.getInfo(), xml))

        for mf in sorted(item.metaDescription.fields, key=attrgetter('name')):
            if mf.is_temporary:
                continue
            _, wrtF = mf.getRWFuncs(self._CODE, self._VERSION)
            mtyp = mf.metaType
            rval = item.getMeta(mf.name)

            if rval is None:
                continue

            if wrtF:
                vals = wrtF(rval)
                for path, val in vals:
                    self._writeMetaValue(path, val, metaNode, xml)
            elif isinstance(mf.metaType, WdbsAtomicMetaTypeP):
                self._writeMetaValue(
                    mf.name, mtyp.toString(rval), metaNode, xml)
            else:
                raise Exception(
                    '{s}-Serialization V{v} cannot handle '
                    '{c}.{m} ({t})'.format(
                        s=self._CODE, v=self._VERSION, t=mf.metaType,
                        c=item.__class__.__name__, m=mf.name))

        node.appendChild(metaNode)

    def _readValueNode(self, tag, node):
        """
        Read the inner value from the element with the given tag.

        :param tag: Element tag to read value from
        :type tag: str
        :param node: Parent node
        :type node: xml.dom.minidom.Node
        :return: Read value, or `None`
        :rtype: str
        """
        valueNode = node.find("./{t}".format(t=tag))
        if valueNode is None:
            return None
        text = valueNode.text
        return text.strip() if text else None

    @staticmethod
    def _readMetaNode(name, node):
        """
        Read the meta-data value for the given name.

        :param name: Meta-data name to read
        :type name: str
        :param node: Parent node
        :type node: xml.dom.minidom.Node
        :return: Meta-data value, or `None`
        :rtype: str
        """
        valueNode = node.find("./{m}/{i}[@name='{n}']".format(
            m=_META_DATA_TAG, i=_META_ITEM_TAG,
            n=WxioXmlVer0_2_0.fixFieldName(name)))
        text = valueNode.text if valueNode is not None else None
        return text.strip() if text else None

    def setMetaForItem(self, node, item):
        if node is None or item is None:
            return

        data = []
        for mf in item.metaDescription.fields:
            if mf.is_temporary:
                continue
            rdrf, _ = mf.getRWFuncs(self._CODE, self._VERSION)
            mtyp = mf.metaType

            if rdrf:
                data.append((mf.name, rdrf(node)))
            elif isinstance(mf.metaType, WdbsAtomicMetaTypeP):
                rval = self._readMetaNode(mf.name, node)
                if rval:
                    data.append((mf.name, mtyp.fromString(rval)))
            else:
                raise Exception(
                    '{s}-Serialization V{v} cannot handle '
                    '{c}.{m} ({t})'.format(
                        s=self._CODE, v=self._VERSION, t=mf.metaType,
                        c=item.__class__.__name__, m=mf.name))
        for k, v in data:
            item.setMeta(v, k)

    def readName(self, node):
        return self._readMetaNode('name', node)

    def readInfo(self, node):
        return self._readMetaNode('info', node)

    def readStatus(self, node):
        return self._readMetaNode('status', node)

    def readStatusMask(self, node):
        return self._readMetaNode('statusMask', node)

    def readButtonData(self, node):
        # noinspection PyTypeChecker
        return WxioXmlButton(
            uuid=node.get('uuid') or node.get('sid'),
            title=self._readValueNode('title', node),
            command=self._readValueNode('command', node))
