# ----------------------------------------------------------------------------
# - wxio-xml-0-3-0                                                           -
# - wild cruncher:xml serialization v3.0                        core / xmlio -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wxio_xml_0_3_0
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher xml serialization/deserialization V3.0

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import OrderedDict
from xml.dom.minidom import Document

from wcr_core.dbstore_api.wdbs_meta_data_api import WdbsAtomicMetaP, \
    WdbsListMetaP, WdbsDictionaryMetaP, WdbsTupleMetaP, WdbsRecordMetaP, \
    WdbsUnionMetaP, WdbsMetaP
from wcr_core.dbstore_api.wdbs_meta_type_api import WdbsAtomicMetaTypeP, \
    WdbsUnionMetaTypeP, WdbsRecordMetaTypeP, WdbsTupleMetaTypeP, \
    WdbsListMetaTypeP, WdbsDictionaryMetaTypeP, WdbsOrderedDictionaryMetaTypeP
from wcr_core.dbstore_api.wdbs_entity_api import WdbsNamedEntityP
from wcr_core.dbstore.wdbs_meta_type import WdbsTupleMetaType, WdbsListMetaType
from wcr_core.xmlio.wxio_xml_base import WxioXmlButton
from wcr_core.xmlio.wxio_nested_xml_base import WxioNestedXml, _META_DATA_TAG
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaField
from wcr_core.dbstore.wdbs_meta_type import WDBS_STRING_META


class WxioXmlVer0_3_0(WxioNestedXml):
    """
    XML-IO Version 3.
    """

    #: XML-IO Version
    _VERSION = '3.0'

    def __init__(self, ignoreUUID=False):
        """
        :param ignoreUUID: Whether to ignore UUIDs during serialization
                           (and generate new UUIDs for all entities instead)
        :type ignoreUUID: bool
        """
        super().__init__(ignoreUUID)

    @staticmethod
    def getVersion():
        """
        Get the XML-IO version.

        :return: XML-IO version
        :rtype: str
        """
        return WxioXmlVer0_3_0._VERSION

    def _createVanillaMetaNode(self, meta, xml):
        """
        Create an empty meta-data node for the given meta-data item.

        :param meta: Meta-data
        :type meta: WdbsMetaP
        :param xml: XML Document
        :type xml: Document
        :return: Newly created node element
        :rtype: xml.dom.minidom.Element
        """
        mType = meta.getType()
        node = xml.createElement(mType.getStructName())
        node.setAttribute('type', mType.getName())
        return node

    def _createAtomicMetaNode(self, aMeta, xml, force=True):
        """
        Create a meta-data node for the given atomic meta-data item.

        :param aMeta: Meta-data
        :type aMeta: WdbsAtomicMetaP
        :param xml: XML Document
        :type xml: Document
        :param force: Whether to force node-creation even if the meta-data
                      item holds no value
        :type force: bool
        :return: Newly created node element
        :rtype: xml.dom.minidom.Element
        """
        mval = aMeta.toString()
        if not mval and not force:
            return None
        node = self._createVanillaMetaNode(aMeta, xml)
        if mval:
            self._setNodeValue(mval, node, xml)
        return node

    def _createListMetaNode(self, lMeta, xml, force=True):
        """
        Create a meta-data node for the given list meta-data item.

        :param lMeta: Meta-data
        :type lMeta: WdbsListMetaP
        :param xml: XML Document
        :type xml: Document
        :param force: Whether to force node-creation even if the meta-data
                      item holds no value
        :type force: bool
        :return: Newly created node element
        :rtype: xml.dom.minidom.Element
        """
        node = self._createVanillaMetaNode(lMeta, xml)
        for idx, meta in enumerate(lMeta.serializationIterator()):
            eNode = self._createMetaNode(meta, xml)
            eNode.setAttribute('index', str(idx))
            node.appendChild(eNode)
        return node if node or force else None

    def _createDictionaryMetaNode(self, dMeta, xml, force=True):
        """
        Create a meta-data node for the given dictionary meta-data item.

        :param dMeta: Meta-data
        :type dMeta: WdbsDictionaryMetaP
        :param xml: XML Document
        :type xml: Document
        :param force: Whether to force node-creation even if the meta-data
                      item holds no value
        :type force: bool
        :return: Newly created node element
        :rtype: xml.dom.minidom.Element
        """
        node = self._createVanillaMetaNode(dMeta, xml)
        for key, meta in sorted(dMeta.serializationIterator(),
                                key=lambda k: k[0].asValue()):
            kNode = xml.createElement('key')
            kNode.appendChild(self._createMetaNode(key, xml))
            vNode = xml.createElement('value')
            vNode.appendChild(self._createMetaNode(meta, xml))
            eNode = xml.createElement('item')
            eNode.appendChild(kNode)
            eNode.appendChild(vNode)
            node.appendChild(eNode)
        return node if node or force else None

    def _createTupleMetaNode(self, tMeta, xml):
        """
        Create a meta-data node for the given tuple meta-data item.

        :param tMeta: Meta-data
        :type tMeta: WdbsTupleMetaP
        :param xml: XML Document
        :type xml: Document
        :return: Newly created node element
        :rtype: xml.dom.minidom.Element
        """
        node = self._createVanillaMetaNode(tMeta, xml)
        for idx, meta in enumerate(tMeta.serializationIterator()):
            eNode = self._createMetaNode(meta, xml)
            eNode.setAttribute('field', str(idx))
            node.appendChild(eNode)
        return node

    def _createRecordMetaNode(self, rMeta, xml):
        """
        Create a meta-data node for the given record meta-data item.

        :param rMeta: Meta-data
        :type rMeta: WdbsRecordMetaP
        :param xml: XML Document
        :type xml: Document
        :return: Newly created node element
        :rtype: xml.dom.minidom.Element
        """
        node = self._createVanillaMetaNode(rMeta, xml)
        for field, meta in sorted(rMeta.serializationIterator()):
            eNode = self._createMetaNode(meta, xml)
            eNode.setAttribute('field', field)
            node.appendChild(eNode)
        return node

    def _createUnionMetaNode(self, uMeta, xml):
        """
        Create a meta-data node for the given union meta-data item.

        :param uMeta: Meta-data
        :type uMeta: WdbsUnionMetaP
        :param xml: XML Document
        :type xml: Document
        :return: Newly created node element
        :rtype: xml.dom.minidom.Element
        """
        node = self._createVanillaMetaNode(uMeta, xml)
        for vtyp, value in uMeta.serializationIterator():
            node.setAttribute('union-type', vtyp)
            eNode = self._createMetaNode(value, xml)
            node.appendChild(eNode)
        return node

    def _createMetaNode(self, meta, xml, force=True):
        """
        Create a meta-data node for the given meta-data item.

        :param meta: Meta-data
        :type meta: WdbsMetaP
        :param xml: XML Document
        :type xml: Document
        :param force: Whether to force node-creation even if the meta-data
                      item holds no value
        :type force: bool
        :return: Newly created node element
        :rtype: xml.dom.minidom.Element
        """
        if isinstance(meta, WdbsAtomicMetaP):
            return self._createAtomicMetaNode(meta, xml, force)
        elif isinstance(meta, WdbsListMetaP):
            return self._createListMetaNode(meta, xml, force)
        elif isinstance(meta, WdbsDictionaryMetaP):
            return self._createDictionaryMetaNode(meta, xml, force)
        elif isinstance(meta, WdbsTupleMetaP):
            return self._createTupleMetaNode(meta, xml)
        elif isinstance(meta, WdbsRecordMetaP):
            return self._createRecordMetaNode(meta, xml)
        elif isinstance(meta, WdbsUnionMetaP):
            return self._createUnionMetaNode(meta, xml)
        else:
            raise ValueError('Unknown meta-data type: {} ({})'.format(
                meta,
                meta.getType() if isinstance(meta, WdbsMetaP) else type(meta)
            ))

    def _writeMetaData(self, item, node, xml):
        if isinstance(item, WdbsNamedEntityP):
            nNode = xml.createElement('name')
            if item.getName():
                nNode.appendChild(xml.createTextNode(item.getName()))
            node.appendChild(nNode)
            iNode = xml.createElement('info')
            if item.getInfo():
                iNode.appendChild(xml.createTextNode(item.getInfo()))
            node.appendChild(iNode)
            status = getattr(item, 'getStatus', None)
            if status:
                sNode = xml.createElement('status')
                sNode.appendChild(xml.createTextNode(status))
                node.appendChild(sNode)
            statusMask = getattr(item, 'getStatusMask', None)
            if statusMask:
                smNode = xml.createElement('statusMask')
                smNode.appendChild(xml.createTextNode(statusMask))
                node.appendChild(smNode)

        metaNode = xml.createElement(_META_DATA_TAG)

        keys = set(f.name for f in item.metaDescription.fields
                   if not f.is_temporary)
        metas = tuple((k, v) for k, v in item.getMetas() if k in keys)

        for name, meta in sorted(metas):
            mNode = self._createMetaNode(meta, xml, force=False)
            if mNode:
                mNode.setAttribute('name', name)
                metaNode.appendChild(mNode)

        node.appendChild(metaNode)

    def _readNodeText(self, node):
        """
        Read the contained text from the given node.

        :param node: Node
        :type node: xml.etree.ElementTree.Element
        :return: Text
        :rtype: str
        """
        if node is None:
            return None
        text = node.text
        if not text:
            return None
        return text.strip()

    def _readAtomicMetaNode(self, mtyp, node):
        """
        Read the atomic value from the given meta-data node.

        :param mtyp: Meta-data type
        :type mtyp: WdbsAtomicMetaTypeP
        :param node: Nde
        :type node: xml.etree.ElementTree.Element
        :return: Atomic value
        :rtype: object
        """
        text = self._readNodeText(node)
        if not text:
            return None
        return mtyp.fromString(text)

    def _readListMetaNode(self, mtyp, node):
        """
        Read the list value from the given meta-data node.

        TODO: Index-based access.

        :param mtyp: Meta-data type
        :type mtyp: WdbsListMetaTypeP
        :param node: Node
        :type node: xml.etree.ElementTree.Element
        :return: List value
        :rtype: list[object]
        """
        if node is None:
            return []
        vtyp = mtyp.getValueType()
        return [self._readMetaNode(vtyp, eNode) for eNode in node]

    def _readDictionaryMetaNode(self, mtyp, node):
        """
        Read the dictionary/map value from the given meta-data node.

        :param mtyp: Meta-data type
        :type mtyp: WdbsDictionaryMetaTypeP
        :param node: Node
        :type node: xml.etree.ElementTree.Element
        :return: Dictionary/map value
        :rtype: dict[object, object]
        """
        if node is None:
            return {}
        vtyp = mtyp.getValueType()
        ktyp = mtyp.getKeyType()
        if isinstance(mtyp, WdbsOrderedDictionaryMetaTypeP):
            return OrderedDict(
                (self._readMetaNode(ktyp, eNode.find('key/*')),
                 self._readMetaNode(vtyp, eNode.find('value/*')))
                for eNode in node)
        else:
            return {self._readMetaNode(ktyp, eNode.find('key/*')):
                        self._readMetaNode(vtyp, eNode.find('value/*'))
                    for eNode in node}

    def _readTupleMetaNode(self, mtyp, node):
        """
        Read the tuple value from the given meta-data node.

        TODO: Index-based access.

        :param mtyp: Meta-data type
        :type mtyp: WdbsTupleMetaTypeP
        :param node: Node
        :type node: xml.etree.ElementTree.Element
        :return: Tuple value
        :rtype: tuple
        """
        if node is None:
            return None
        return tuple(self._readMetaNode(vtyp, eNode)
                     for (_, vtyp), eNode in zip(mtyp, node))

    def _readRecordMetaNode(self, mtyp, node):
        """
        Read the record value from the given meta-data node.

        :param mtyp: Meta-data value
        :type mtyp: WdbsRecordMetaTypeP
        :param node: Node
        :type node: xml.etree.ElementTree.Element
        :return: Record value
        :rtype: dict[str, object]
        """
        if node is None:
            return None
        return {field:
                    self._readMetaNode(
                        vtyp, node.find("./*[@field='{f}']".format(f=field)))
                for field, vtyp in mtyp}

    def _readUnionMetaNode(self, mtyp, node):
        """
        Read the union value from the given meta-data node.

        :param mtyp: Meta-data value
        :type mtyp: WdbsUnionMetaTypeP
        :param node: Node
        :type node: xml.etree.ElementTree.Element
        :return: Record value
        :rtype: (str, object)
        """
        if node is None:
            return None
        vnod = next(iter(node), None)
        if vnod is None:
            return None

        unam = node.get('union-type')
        utyp = mtyp.getType(unam)
        return unam, self._readMetaNode(utyp, vnod)

    def _readMetaNode(self, mtyp, node):
        """
        Read the value from the given meta-data node.

        :param mtyp: Meta-data value
        :type mtyp: WdbsMetaTypeP
        :param node: Node
        :type node: xml.etree.ElementTree.Element
        :return: Value
        :rtype: object
        """
        if isinstance(mtyp, WdbsAtomicMetaTypeP):
            return self._readAtomicMetaNode(mtyp, node)
        elif isinstance(mtyp, WdbsListMetaTypeP):
            return self._readListMetaNode(mtyp, node)
        elif isinstance(mtyp, WdbsDictionaryMetaTypeP):
            return self._readDictionaryMetaNode(mtyp, node)
        elif isinstance(mtyp, WdbsTupleMetaTypeP):
            return self._readTupleMetaNode(mtyp, node)
        elif isinstance(mtyp, WdbsRecordMetaTypeP):
            return self._readRecordMetaNode(mtyp, node)
        elif isinstance(mtyp, WdbsUnionMetaTypeP):
            return self._readUnionMetaNode(mtyp, node)
        else:
            raise ValueError('Unknown meta-data type:', mtyp)

    def setMetaForItem(self, node, item):
        metaNode = node.find('meta-data')
        if metaNode is None:
            return
        for mf in item.metaDescription.fields:
            if mf.is_temporary:
                continue
            mNode = metaNode.find("./*[@name='{n}']".format(n=mf.name))
            if mNode is None:
                continue
            meta = self._readMetaNode(mf.metaType, mNode)
            if meta is not None:
                item.setMeta(meta, mf.name)

    def readName(self, node):
        return self._readNodeText(node.find('name'))

    def readInfo(self, node):
        return self._readNodeText(node.find('info'))

    def readStatus(self, node):
        return self._readNodeText(node.find('status'))

    def readStatusMask(self, node):
        return self._readNodeText(node.find('statusMask'))

    def readType(self, node):
        return self._readNodeText(node.find('type'))

    def readChannels(self, node):
        params = {}
        for channel in node.find('channels'):
            key = 'channel{}'.format(channel.get('number'))
            params[key] = []
            for param in channel:
                params[key].append((param.get('name'), param.get('value')))
        return params

    def readButtonData(self, node):
        # noinspection PyTypeChecker
        return WxioXmlButton(
            uuid=node.get('uuid') or node.get('sid'),
            title=self._readNodeText(node.find('title')),
            command=self._readNodeText(node.find('command')))
