# ----------------------------------------------------------------------------
# - wxio-xml-base                                                            -
# - wild cruncher:wsh xml serialization/deserialization         core / xmlio -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wxio_xml_base
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher xml serialization/deserialization base classes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta
from collections import deque, OrderedDict, namedtuple
from itertools import chain
from xml.dom.minidom import Node
import re

from wcr_core.dbstore.mutable.wdbm_instance import WdbmOperatorInstance
from wcr_core.db_api.whdb_block_api import WHDB_DEFAULT_BLOCK_CODE, \
    WHDB_TOPLEVEL_BLOCK_CODE
from wcr_core.db_api.whdb_model_api import WHDB_DEFAULT_MODEL_CODE
from wcr_core.dbstore.wdbs_util import wdbs_makeUUID
from wcr_core.dbstore_api.wdbs_block_api import WdbsBlockP
from wcr_core.dbstore_api.wdbs_connection_api import WdbsConnectionP
from wcr_core.dbstore_api.wdbs_instance_api import WdbsBlockInstanceP, \
    WdbsModelInstanceP
from wcr_core.dbstore_api.wdbs_entity_api import WdbsNamedEntityP
from wcr_core.dbstore_api.wdbs_connection_point_api import *


#: Block DTO
WxioXmlBlock = namedtuple(
    'WxioXmlBlock',
    ['uuid', 'type', 'name', 'info'])

#: Model DTO
WxioXmlModel = namedtuple(
    'WxioXmlModel',
    ['uuid', 'type', 'name', 'info', 'path', 'entry'])

#: Connection point DTO
WxioXmlXPoint = namedtuple(
    'WxioXmlXPoint',
    ['uuid', 'type', 'direction'])

#: Block instance DTO
WxioXmlBlockInst = namedtuple(
    'WxioXmlBlockInst',
    ['uuid', 'name', 'info', 'block'])

#: Model instance DTO
WxioXmlModelInst = namedtuple(
    'WxioXmlModelInst',
    ['uuid', 'name', 'info', 'model'])

#: Operator instance DTO
WxioXmlOpInst = namedtuple(
    'WxioXmlOpInst',
    ['uuid', 'name', 'info', 'status', 'statusMask', 'oper'])

#: Card instance DTO
WxioXmlIOCardInst = namedtuple(
    'WxioXmlIOCardInst',
    ['uuid', 'name', 'idtstatus', 'channels', 'oper'])

#: Connection DTO
WxioXmlConnection = namedtuple(
    'WxioXmlConnection',
    ['uuid', 'name', 'info', 'sobj', 'sxpnt', 'eobj', 'expnt', 'path'])

#: Shape DTO
WxioXmlShape = namedtuple(
    'WxioXmlShape',
    ['uuid', 'stype'])

#: Button DTO
WxioXmlButton = namedtuple(
    'WxioXmlButton',
    ['uuid', 'title', 'command'])


def wxio_readXmlFloat(tag, attributesOrNode, default=None):
    """
    Reads a given attribute as floating point value.

    :param tag: Attribute name to read
    :type tag: str
    :param attributesOrNode: Node to read attribute from
    :param default: Default value if the attribute cannot be found
    :type default: float
    :return: Value of the attribute as floating point number
    :rtype: float
    """
    v = attributesOrNode.get(tag)
    return float(v) if v else default


def wxio_readXmlInt(tag, attributesOrNode, default=None):
    """
    Reads a given attribute as integer value.

    :param tag: Attribute name to read
    :type tag: str
    :param attributesOrNode: Node to read attribute from
    :param default: Default value if the attribute cannot be found
    :type default: int
    :return: Value of the attribute as integer number
    :rtype: int
    """
    v = attributesOrNode.get(tag)
    return int(v) if v else default


def wxio_readXmlBool(tag, attributesOrNode, default=None):
    """
    Reads a given attribute as boolean value.

    :param tag: Attribute name to read
    :type tag: str
    :param attributesOrNode: Node to read attribute from
    :param default: Default value if the attribute cannot be found
    :type default: bool
    :return: Value of the attribute as boolean number
    :rtype: bool
    """
    v = attributesOrNode.get(tag)
    return v.lower() == 'true' if v else default


def wxio_readXmlUUID(node):
    """
    Read the serialization ID from the given node.
    The serialization ID is usually the value of the 'uuid'-attribute;
    for backwards compatibility, the value of the 'name'-attribute will be
    returned if the node has no 'uuid'-attribute.

    :param node: Node to read serialization ID from
    :return: Serialization ID
    :rtype: str
    """
    return node.get('uuid') or node.get('sid')


class WxioXmlP(metaclass=ABCMeta):
    """
    Protocol for XML-IO serializer/deserializers.
    """

    @abstractmethod
    def serialize(self, block, xml, topName, module):
        """
        Serialize a model hierarchy to XML, starting at the top TreeElement.
        The serialization is written to the given file.

        :param block: Top level model/instance of the hierarchy to be
                      serialized.
        :type block: WdbsContainerP
        :param xml: XML document that the serialization will be added to.
        :type xml: xml.dom.minidom.Document
        :param module: Name of the application module the block originates from
        :type module: str
        :return: XML node containing the data of the block
        :rtype: Node
        """
        raise NotImplementedError()

    def deserialize(self, tree, topName, block):
        """
        Deserialize an XML serialization of a model hierarchy.
        The XML serialization is read from the given file.

        :param tree: XML tree to obtain serialization from.
        :type tree: xml.etree.ElementTree.ElementTree
        :param topName: Name of the top-element
        :type topName: str
        :param block: Entity container to add read elements to
        :type block: WdbsBlockP
        :return: Model hierarchy via its top-level model/instance.
        :rtype: WdbsBlockP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WxioXmlBase(WxioXmlP):
    """
    Abstract XML-IO serializer/deserializer base class.
    """

    #: Reader code
    _CODE = 'xmlio'

    NUMBERED_MODEL_NAME = re.compile('^\s*model\s*(\d+)\s*$', re.IGNORECASE)

    NUMBERED_FUNC_NAME = re.compile('^\s*\w+\s*(\d+)\s*$', re.IGNORECASE)

    NUMBERED_INSTANCE_NAME = re.compile('^\s*minst(?:ance)?\s*(\d+)\s*$',
                                        re.IGNORECASE)

    # Translation xpoint type model -> XML
    XPOINT_DIR_EXT = {WDBS_XPNT_NONE: 'none',
                      WDBS_XPNT_IN: 'input',
                      WDBS_XPNT_OUT: 'output',
                      WDBS_XPNT_BI: 'bidir'}
    # Translation xpoint type XML -> model
    XPOINT_DIR_INT = {v: k for k, v in XPOINT_DIR_EXT.items()}

    @staticmethod
    def getCode():
        """
        Get the serialization type code.

        :return: Serialization type code
        :rtype: str
        """
        return WxioXmlBase._CODE

    def __init__(self, ignoreUUID=False):
        """
        :param ignoreUUID: Whether to ignore UUIDs during serialization
                           (and generate new UUIDs for all entities instead)
        :type ignoreUUID: bool
        """
        self.__useUUID = not ignoreUUID

    def deserialize(self, tree, topName, block):
        instBlockFs = {}
        instModelFs = {}
        entities = {}
        metas = []

        # noinspection PyUnresolvedReferences
        hnod = tree.find('hierarchy')

        for node in chain(self.getModelNodes(hnod)):
            instModelF, uuid = self._readModelElement(node, entities, metas)
            if instModelF is not None and uuid is not None:
                instModelFs[uuid] = instModelF

        for node in chain(self.getBlockNodes(hnod)):
            instBlockF, uuid = self._readBlockElement(
                node, instBlockFs, instModelFs, entities, metas)
            if instBlockF is not None and uuid is not None:
                instBlockFs[uuid] = instBlockF

        tophmNode = hnod.find(topName)
        self._addBlockElements(
            tophmNode, block, instBlockFs, instModelFs, entities, metas)
        if isinstance(block, WdbsNamedEntityP):
            data = self.readBlockData(tophmNode)
            block.setName(data.name)
            if data.info:
                block.setInfo(data.info)

        metas.append((self.setBlockMeta, tophmNode, block))

        for mfnc, node, entity in metas:
            meta = mfnc(node, entity)
            if meta:
                for k, v in meta:
                    entity.setMeta(v, k)

        return block

    def collectModels(self, block):
        """
        Return all blocks and models contained in the given block, recursively.

        :param block: Top level model block containing blocks/models to collect
        :type block: WdbsContainerP
        :return: List of blocks and set of models contained in the given model
                 block
        :rtype: (list[WdbsBlockP], set[WdbsModelP])
        """
        models = set()
        blocks = OrderedDict()
        blocks[id(block)] = block

        todo = deque([])
        todo.appendleft(block)

        while len(todo) > 0:
            block = todo.popleft()

            for inst in block.getInstances():
                if isinstance(inst, WdbsBlockInstanceP):
                    block = inst.getBlock()
                    todo.append(block)
                    blockId = id(block)
                    if blockId in blocks:
                        blocks.move_to_end(blockId)
                    else:
                        blocks[blockId] = block
                elif isinstance(inst, WdbsModelInstanceP):
                    models.add(inst.getModel())

        return list(reversed(list(blocks.values())[1:])), models

    def _uuid(self, data):
        """
        :rtype: UUID
        """
        return wdbs_makeUUID(data.uuid) if self.__useUUID else None

    def _readModelElement(self, node, entities, metas):
        """
        Read a model entity from a XML element.

        :param node: XML element to read the entity from
        :type node: TreeElement
        :return: Text model read from the XML element, and its UUID
        :rtype: ((WdbsBlockP) -> WdbsModelInstanceP, str)
        """

        def modelInstF(parent:WdbsBlockP, *args, **kwargs):
            nonlocal model
            if model is None:
                data = self.readModelData(node)
                kwargs['muuid'] = self._uuid(data)
                kwargs['mtyp'] = data.type or WHDB_DEFAULT_MODEL_CODE
                # noinspection PyTypeChecker
                inst = parent.createModelInstance(*args, **kwargs)
                model = inst.getModel()
                model.setPath(data.path)
                model.setEntryPoint(data.entry)
                model.setName(data.name)
                model.setInfo(data.info)
                metas.append((self.setModelMeta, node, model))
                entities[data.uuid] = model
            else:
                raise Exception('Models cannot be reused')
            return inst

        model = None

        return modelInstF, wxio_readXmlUUID(node)

    def _fixBlockType(self, btyp):
        if not btyp:
            return WHDB_DEFAULT_BLOCK_CODE
        if '.block-' in btyp:
            return btyp
        elif btyp == 'top':
            return WHDB_TOPLEVEL_BLOCK_CODE
        elif btyp == 'default':
            return WHDB_TOPLEVEL_BLOCK_CODE
        else:
            return 'wcr.{}.block-{}'.format('', btyp)

    def _readBlockElement(self, node, blocks, models, entities, metas):
        """
        Read a block from a XML element.

        :param node: XML element to read the entity from
        :type node: TreeElement
        :param blocks: Dictionary of all known blocks
        :type blocks: dict[str, (WdbsBlockP) -> WdbsBlockInstanceP]
        :param models: Dictionary of all known models
        :type models: dict[str, (WdbsBlockP) -> WdbsModelInstanceP]
        :return: Model block read from the XML element, and its UUID
        :rtype: ((WdbsBlockP) -> WdbsBlockInstanceP, str)
        """

        def modelInstF(parent:WdbsBlockP, *args, **kwargs):
            nonlocal block
            if block is None:
                data = self.readBlockData(node)
                kwargs['buuid'] = self._uuid(data)
                kwargs['btyp'] = self._fixBlockType(data.type)
                # noinspection PyTypeChecker
                inst = parent.createBlockInstance(*args, **kwargs)
                block = inst.getBlock()
                block.setName(data.name)
                block.setInfo(data.info)
                metas.append((self.setBlockMeta, node, block))
                entities[data.uuid] = block
                self._addBlockElements(
                    node, block, blocks, models, entities, metas)
            else:
                raise Exception('Blocks cannot be reused')
            return inst

        block = None

        return modelInstF, wxio_readXmlUUID(node)

    def _readBlockInstanceElement(self, node, block, instBlockFs, entities,
                                  metas):
        """
        Read a model instance entity from a XML element.

        :param node: XML element to read the entity from
        :type node: TreeElement
        :param block: Parent block to create the instance in
        :type block: WdbsBlockP
        :param instBlockFs: Dictionary of all known models
        :type instBlockFs: dict[str, (WdbsBlockP) -> WdbsBlockInstanceP]
        :param entities: UUID-entity mapping dictionary
        :type entities: dict[int, WdbsEntityP]
        :return: Model instance read from the XML element, and its UUID
        :rtype: (WdbsBlockInstanceP, str)
        """
        xpnds = list(self.getXPointNodes(node))
        xdats = [self.readXPointData(xpnd) for xpnd in xpnds]
        data = self.readBlockInstanceData(node)
        instModelF = instBlockFs[data.block]
        inst = instModelF(
            block,
            name=data.name,
            info=data.info,
            clst=((xdat.type, xdat.uuid if self.__useUUID else None,
                   {'cdir': xdat.direction})
                  for xdat in xdats),
            uuid=self._uuid(data))
        for xdat, xpnt in zip(xdats, inst.getConnectionPoints()):
            entities[xdat.uuid] = xpnt
        metas.append((self.setBlockInstanceMeta, node, inst))
        metas.extend((self.setXPointMeta, xpnd, xpnt) for xpnd, xpnt
                     in zip(xpnds, inst.getConnectionPoints()))
        return inst, data.uuid

    def _readModelInstanceElement(self, node, block, instModelFs, entities,
                                  metas):
        """
        Read a model instance entity from a XML element.

        :param node: XML element to read the entity from
        :type node: TreeElement
        :param block: Parent block to create the instance in
        :type block: WdbsBlockP
        :param instModelFs: Dictionary of all known models
        :type instModelFs: dict[str, (WdbsBlockP) -> WdbsModelInstanceP]
        :param entities: UUID-entity mapping dictionary
        :type entities: dict[int, WdbsEntityP]
        :return: Model instance read from the XML element, and its UUID
        :rtype: (WdbsModelInstanceP, str)
        """
        xpnds = list(self.getXPointNodes(node))
        xdats = [self.readXPointData(xpnd) for xpnd in xpnds]
        data = self.readModelInstanceData(node)
        instModelF = instModelFs[data.model]
        inst = instModelF(
            block,
            name=data.name,
            info=data.info,
            clst=((xdat.type, xdat.uuid if self.__useUUID else None,
                   {'cdir': xdat.direction})
                  for xdat in xdats),
            uuid=self._uuid(data))
        for xdat, xpnt in zip(xdats, inst.getConnectionPoints()):
            entities[xdat.uuid] = xpnt
        metas.append((self.setModelInstanceMeta, node, inst))
        metas.extend((self.setXPointMeta, xpnd, xpnt) for xpnd, xpnt
                     in zip(xpnds, inst.getConnectionPoints()))
        return inst, data.uuid

    def _readOperatorInstanceElement(self, node, block, entities, metas):
        """
        Read a operator instance from a XML element.

        :param node: XML element to read the entity from
        :type node: TreeElement
        :param block: Parent block to create the instance in
        :type block: WdbsBlockP
        :param entities: UUID-entity mapping dictionary
        :type entities: dict[int, WdbsEntityP]
        :return: Operator instance read from the XML element, and its UUID
        :rtype: (WdbsOperatorInstanceP, str)
        """
        xpnds = list(self.getXPointNodes(node))
        xdats = [self.readXPointData(xpnd) for xpnd in xpnds]
        data = self.readOperatorInstanceData(node)
        cardsRead = self.readIoCards(node)
        cards = {}
        # create cards Operator
        for uuid, name, idtstatus, channels, oper in cardsRead:
            cardInst = block.createOperatorInstance(oper, name=name)
            cards[uuid] = (cardInst, [name, idtstatus, channels])
        func = block.createOperatorInstance(
            data.oper,
            name=data.name,
            info=data.info,
            clst=((xdat.type, xdat.uuid if self.__useUUID else None,
                   {'cdir': xdat.direction})
                  for xdat in xdats),
            params=self.readOperatorInstanceParameters(node),
            uuid=self._uuid(data),
            status=data.status,
            statusMask=data.statusMask,
            cards=cards)
        for xdat, xpnt in zip(xdats, func.getConnectionPoints()):
            entities[xdat.uuid] = xpnt
        metas.append((self.setOperatorInstanceMeta, node, func))
        metas.extend((self.setXPointMeta, xpnd, xpnt) for xpnd, xpnt
                     in zip(xpnds, func.getConnectionPoints()))
        return func, data.uuid

    def _readConnectionElement(self, node, block, entities, metas):
        """
        Read a connection from a XML element.

        :param node: XML element to read the entity from
        :type node: TreeElement
        :param block: Parent block to create the connection in
        :type block: WdbsBlockP
        :param entities: Dictionary of all known instances
        :type entities: dict[str, WdbsEntityP]
        :return: Connection read from the XML element, and its UUID
        :rtype: (WdbsConnectionP, str)
        """

        data = self.readConnectionData(node)

        sobj = entities[data.sobj]
        """:type: WdbsInstanceP"""
        sxpt = entities[data.sxpnt]
        """:type: WdbsConnectionPointP"""
        eobj = entities[data.eobj]
        """:type: WdbsInstanceP"""
        expt = entities[data.expnt]
        """:type: WdbsConnectionPointP"""

        conn = block.createConnection(
            sobj=sobj, sxpt=sxpt, eobj=eobj, expt=expt,
            name=data.name, info=data.info, path=data.path,
            uuid=self._uuid(data))
        metas.append((self.setConnectionMeta, node, conn))

        return conn, data.uuid

    def _readShapeElement(self, node, block, metas):
        """
        Read a shape entity from a XML element.

        :param node: XML element to read the entity from
        :type node: TreeElement
        :param block: Parent block to create the shape in
        :type block: WdbsBlockP
        :return: Shape read from the XML element, and its UUID
        :rtype: (WdbsShapeP, str)
        """
        data = self.readShapeData(node)
        shape = block.createShape(data.stype, uuid=self._uuid(data))
        metas.append((self.setShapeMeta, node, shape))
        return shape, data.uuid

    def _readButtonElement(self, node, block, metas):
        """
        Read a button entity from a XML element.

        :param node: XML element to read the entity from
        :type node: TreeElement
        :param block: Parent block to create the button in
        :type block: WdbsBlockP
        :return: Button read from the XML element, and its UUID
        :rtype: (WdbsButtonP, str)
        """
        data = self.readButtonData(node)
        button = block.createButton(
            title=data.title,
            command=data.command,
            params=self.readButtonParameters(node),
            uuid=self._uuid(data))
        metas.append((self.setButtonMeta, node, button))
        return button, data.uuid

    def _addBlockElements(self, node, block, blocks, models, entities, metas):
        """
        Read all child elements and add them to the given model block.

        :param node: XML element to read the child elements from
        :type node: TreeElement
        :param block: Model block to add elements to
        :type block: WdbsBlockP
        :param blocks: Dictionary of all known blocks
        :type blocks: dict[str, (WdbsBlockP) -> WdbsBlockInstanceP]
        :param models: Dictionary of all known models
        :type models: dict[str, (WdbsBlockP) -> WdbsModelInstanceP]
        """

        def e(tpl):
            entity, uuid = tpl
            if entity and uuid:
                entities[uuid] = entity

        for elementNode in self.getBlockInstanceNodes(node, blocks):
            e(self._readBlockInstanceElement(
                elementNode, block, blocks, entities, metas))

        for elementNode in self.getModelInstanceNodes(node, models):
            e(self._readModelInstanceElement(
                elementNode, block, models, entities, metas))

        for elementNode in self.getOperatorInstanceNodes(node):
            e(self._readOperatorInstanceElement(
                elementNode, block, entities, metas))

        for elementNode in self.getConnectionNodes(node):
            e(self._readConnectionElement(elementNode, block, entities, metas))

        for shapeNode in self.getShapeNodes(node):
            e(self._readShapeElement(shapeNode, block, metas))

        for interactNode in self.getButtonNodes(node):
            e(self._readButtonElement(interactNode, block, metas))

    @abstractmethod
    def setMetaForItem(self, node, item):
        """
        Read the meta-data from the given node, based on the provided
        meta-data description for the given item.

        :param node: Node to read meta-data from
        :type node: xml.dom.minidom.Element
        :param item: Hierarchy entity to use for meta-data description
        :type item: WdbsEntityP
        :return: Meta-data as name-value pairs
        :rtype: Iterable[(str, object)]
        """
        raise NotImplementedError()

    @abstractmethod
    def getBlockNodes(self, node):
        """
        Get all block nodes contained in the given node.

        :param node: Parent node
        :type node: xml.dom.minidom.Element
        :return: Iterable of block nodes
        :rtype: Iterable[xml.dom.minidom.Element]
        """
        raise NotImplementedError()

    @abstractmethod
    def readBlockData(self, node):
        """
        Get a block DTO with the data from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Block DTO
        :rtype: WdbsXmlBlock
        """
        raise NotImplementedError()

    def setBlockMeta(self, node, block):
        """
        Set the meta-data for the given block from the given block node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :param block: Block object to retrieve meta-data for
        :type block: WdbsBlockP
        """
        self.setMetaForItem(node, block)

    @abstractmethod
    def getBlockInstanceNodes(self, node, blocks):
        """
        Get all block instance nodes contained in the given node.

        :param node: Parent node
        :type node: xml.dom.minidom.Element
        :param blocks: Look-up for block UUID
        :type blocks: set[str] | dict[str:object]
        :return: Iterable of block nodes
        :rtype: Iterable[xml.dom.minidom.Element]
        """
        raise NotImplementedError()

    @abstractmethod
    def getModelInstanceNodes(self, node, models):
        """
        Get all model instance nodes contained in the given node.

        :param node: Parent node
        :type node: xml.dom.minidom.Element
        :param models: Look-up for model UUID
        :type models: set[str] | dict[str, object]
        :return: Iterable of block nodes
        :rtype: Iterable[xml.dom.minidom.Element]
        """
        raise NotImplementedError()

    @abstractmethod
    def getOperatorInstanceNodes(self, node):
        """
        Get all operator instance nodes contained in the given node.

        :param node: Parent node
        :type node: xml.dom.minidom.Element
        :return: Iterable of block nodes
        :rtype: Iterable[xml.dom.minidom.Element]
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectionNodes(self, node):
        """
        Get all connection nodes contained in the given node.

        :param node: Parent node
        :type node: xml.dom.minidom.Element
        :return: Iterable of block nodes
        :rtype: Iterable[xml.dom.minidom.Element]
        """
        raise NotImplementedError()

    @abstractmethod
    def getShapeNodes(self, node):
        """
        Get all shape nodes contained in the given node.

        :param node: Parent node
        :type node: xml.dom.minidom.Element
        :return: Iterable of block nodes
        :rtype: Iterable[xml.dom.minidom.Element]
        """
        raise NotImplementedError()

    @abstractmethod
    def getButtonNodes(self, node):
        """
        Get all button nodes contained in the given node.

        :param node: Parent node
        :type node: xml.dom.minidom.Element
        :return: Iterable of block nodes
        :rtype: Iterable[xml.dom.minidom.Element]
        """
        raise NotImplementedError()

    @abstractmethod
    def getModelNodes(self, node):
        """
        Get all model nodes contained in the given node.

        :param node: Parent node
        :type node: xml.dom.minidom.Element
        :return: Iterable of model nodes
        :rtype: Iterable[xml.dom.minidom.Element]
        """
        raise NotImplementedError()

    @abstractmethod
    def readModelData(self, node):
        """
        Get a model DTO with the data from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Model DTO
        :rtype: WdbsXmlModel
        """
        raise NotImplementedError()

    def setModelMeta(self, node, model):
        """
        Set the meta-data for the given model from the given model node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :param model: Model object to retrieve meta-data for
        :type model: WdbsModelP
        """
        self.setMetaForItem(node, model)

    @abstractmethod
    def readXPointData(self, node):
        """
        Get a connection point DTO with the data from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Connection point DTO
        :rtype: WdbsXmlXPoint
        """
        raise NotImplementedError()

    def setXPointMeta(self, node, xpnt):
        """
        Set the meta-data for the given connection point from the given
        connection point node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :param xpnt: Connection point
        :type xpnt: WdbsConnectionPointP
        """
        self.setMetaForItem(node, xpnt)

    @abstractmethod
    def getXPointNodes(self, node):
        """
        Get all connection point DTOs with the data from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Connection point DTOs
        :rtype: Iterable[xml.dom.minidom.Element]
        """
        raise NotImplementedError()

    @abstractmethod
    def readBlockInstanceData(self, node):
        """
        Get a block instance DTO with the data from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Block instance DTO
        :rtype: WdbsXmlBlockInst
        """
        raise NotImplementedError()

    def setBlockInstanceMeta(self, node, instance):
        """
        Set the meta-data for the given block instance from the given
        block instance node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :param instance: Block instance object to retrieve meta-data for
        :type instance: WdbsBlockInstanceP
        """
        self.setMetaForItem(node, instance)

    @abstractmethod
    def readModelInstanceData(self, node):
        """
        Get a model instance DTO with the data from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Model instance DTO
        :rtype: WdbsXmlModelInst
        """
        raise NotImplementedError()

    def setModelInstanceMeta(self, node, instance):
        """
        Set the meta-data for the given model instance from the given
        model instance node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :param instance: Model instance object to retrieve meta-data for
        :type instance: WdbsModelInstanceP
        """
        self.setMetaForItem(node, instance)

    @abstractmethod
    def readOperatorInstanceData(self, node):
        """
        Get a operator instance DTO with the data from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Operator instance DTO
        :rtype: WdbsXmlOpInst
        """
        raise NotImplementedError()

    def setOperatorInstanceMeta(self, node, instance):
        """
        Set the meta-data for the given operator instance from the given
        operator instance node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :param instance: Operator instance object to retrieve meta-data for
        :type instance: WdbsOperatorInstanceP
        """
        self.setMetaForItem(node, instance)

    def readIoCards(self, node):
        """
        Get the cards with given node

        :param node: Node
        :type node: xml.dom.minidom.Element
        """
        return []

    @abstractmethod
    def readConnectionData(self, node):
        """
        Get a connection DTO with the data from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Connection DTO
        :rtype: WdbsXmlConnection
        """
        raise NotImplementedError()

    def setConnectionMeta(self, node, connection):
        """
        Set the meta-data for the given connection from the given
        connection node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :param connection: Connection object to retrieve meta-data for
        :type connection: WdbsConnectionP
        """
        self.setMetaForItem(node, connection)

    @abstractmethod
    def readShapeData(self, node):
        """
        Get a shape DTO with the data from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Shape DTO
        :rtype: WdbsXmlShape
        """
        raise NotImplementedError()

    def setShapeMeta(self, node, shape):
        """
        Set the meta-data for the given shape from the given shape node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :param shape: Shape object to retrieve meta-data for
        :type shape: WdbsShapeP
        """
        self.setMetaForItem(node, shape)

    @abstractmethod
    def readButtonData(self, node):
        """
        Get a button DTO with the data from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Button instance DTO
        :rtype: WdbsXmlButton
        """
        raise NotImplementedError()

    def setButtonMeta(self, node, button):
        """
        Set the meta-data for the given shape from the given shape node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :param button: Button object to retrieve meta-data for
        :type button: WdbsButtonP
        """
        self.setMetaForItem(node, button)

    @abstractmethod
    def readOperatorInstanceParameters(self, node):
        """
        Get the parameters for the given operator instance node.

        :param node: Operator instance Node
        :type node: xml.dom.minidom.Element
        :return: Parameters as name-value pairs
        :rtype: Iterable[(str, str)]
        """
        raise NotImplementedError()

    @abstractmethod
    def readButtonParameters(self, node):
        """
        Get the parameters for the given button node.

        :param node: Button Node
        :type node: xml.dom.minidom.Element
        :return: Parameters as name-value pairs
        :rtype: Iterable[(str, str)]
        """
        raise NotImplementedError()
