# ----------------------------------------------------------------------------
# - wxio-xmlio                                                               -
# - wild cruncher:xml serialization/deserialization             core / xmlio -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wxio_xmlio
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher xml serialization/deserialization

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.xmlio.wxio_xml_0_1_0 import WxioXmlVer0_1_0
from wcr_core.xmlio.wxio_xml_0_2_0 import WxioXmlVer0_2_0
from wcr_core.xmlio.wxio_xml_0_3_0 import WxioXmlVer0_3_0


_WXIO_XML_IOS = [WxioXmlVer0_1_0, WxioXmlVer0_2_0, WxioXmlVer0_3_0]

_WXIO_DEFAULT_XML_IO = WxioXmlVer0_3_0


def _wdbs_getXmlIO(version):
    """
    Get the XmlIO handler class for the given version.

    :param version: XML-IO version
    :type version: str
    :return: XmlIO handler for the given version string, or `None`
    :rtype: class
    """

    def sanitize(v):
        return v.lower().strip('.0')

    for xmlio in _WXIO_XML_IOS:
        if sanitize(xmlio.getVersion()) == sanitize(version):
            return xmlio
    return None


def wxio_deserializeXml(tree, topName, block, ignoreUUID=False):
    """
    Deserialize the given XML tree into the given block.

    :param tree: XML tree with the hierarchy to be deserialized
    :type tree: xml.etree.ElementTree.ElementTree
    :param topName: Name of the top-element
    :type topName: str
    :param block: Entity container to add read elements to
    :type block: WdbsBlockP
    :param ignoreUUID: Whether to ignore UUIDs during serialization
                       (and generate new UUIDs for all entities instead)
    :type ignoreUUID: bool
    :return: Block containing the deserialized hierarchy (same as `block`)
    :rtype: WdbsBlockP
    """
    prjNode = tree.find('hierarchy')

    if prjNode is None:
        raise ValueError('Given XML tree is not a valid Silkan XML tree')

    version = prjNode.get('xmlio-version')

    if not version:
        return WxioXmlVer0_1_0(ignoreUUID).deserialize(
            tree, topName, block)

    xmlio = _wdbs_getXmlIO(version)
    if xmlio:
        return xmlio(ignoreUUID).deserialize(tree, topName, block)

    raise ValueError('Could not find XML-IO for version', version)


def wxio_serializeXml(block, xml, topName, module, version=None):
    """
    Serialize the given block (and its children).

    :param block: Hierarchy block to be serialized
    :type block: WdbsContainerP
    :param xml: XML document that the serialization will be added to,
                used for creating new XML elements
    :type xml: xml.dom.minidom.Document
    :param topName: Tag of the top-level XML element
    :type topName: str
    :param module: Name of the application module the block originates from
    :type module: str
    :param version: XML-IO version to use (default: latest)
    :type version: str
    :return: XML node containing the serialization of the block
    :rtype: Node
    """
    xmlio = None

    if version:
        xmlio = _wdbs_getXmlIO(version)
    if xmlio is None:
        xmlio = _WXIO_DEFAULT_XML_IO

    return xmlio().serialize(block, xml, topName, module)
