# ----------------------------------------------------------------------------
# - wxio_xtr01                                                               -
# - wild cruncher: meta-data adapters for XMLIO-V1.0          core / xmlio   -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wxio_xtr01
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher meta-data adapters for XMLIO-V1.0

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from functools import partial

from wcr_core.db.whdb_connection import WhdbConnection, WHDB_KEEP_PATH_KEY
from wcr_core.db.whdb_button import WhdbButton
from wcr_core.db.whdb_model import WhdbModel, WHDB_MODEL_PORTS_KEY
from wcr_core.db.whdb_protocols import WHDB_POSITION_KEY, WhdbHasLocation, \
    WHDB_BBOX_KEY, WHDB_DIMENSION_KEY, WhdbHasDimension
from wcr_core.db.whdb_shape import WhdbTextBasedShapeMixin, \
    WHDB_TEXT_HALIGN_KEY, WHDB_TEXT_VALIGN_KEY, WhdbLineShape, \
    WHDB_LINE_POINTA_KEY, WHDB_LINE_POINTB_KEY, WhdbArrowShape, \
    WhdbRectangleShape, WhdbOvalShape, WHDB_AREA_WIDTH_KEY, \
    WHDB_AREA_HEIGHT_KEY, WhdbTextShape, WHDB_TEXT_KEY
from wcr_core.db.whdb_connection_point import WhdbConnectionPoint, \
    WHDB_XPOINT_POSITION
# from wcr_core.modules.model.db.wmhdb_top_block import WmhdbTopHierarchy, \
#     WMHDB_META_CALLBACKS_KEY
from wcr_core.xmlio.wxio_xml_0_1_0 import WxioXmlVer0_1_0
from wcr_core.modules.model.db.wmhdb_text_model import WmhdbModelText, \
    WMHDB_MODEL_TEXT_KEY
from wcr_core.modules.model.db.wmhdb_wire import WmhdbWire, \
    WMHDB_WIRE_UTYPE_KEY
from wcr_core.modules.model.db.wmhdb_rti_type import WMHDB_RTI_DIM_BY_NAME, \
    WmhdbRtiType


def _asFloat(inStr):
    return float(inStr) if inStr else 0.0


def _asInt(inStr):
    return int(inStr) if inStr else 0.0


def _asBool(inStr):
    return True if (inStr and inStr.lower()[0] == 't') else False


def _readFTuple(keys, node):
    values = [node.get(k) for k in keys]
    if all(values):
        return tuple(_asFloat(v) for v in values)
    else:
        return None


def _readPosition(node):
    return _readFTuple(('xpos', 'ypos'), node)


def _addReader(clazz, meta, func):
    clazz.META[meta].addRWFuncs(
        WxioXmlVer0_1_0.getCode(), WxioXmlVer0_1_0.getVersion(), func)


_addReader(WhdbConnection, WHDB_KEEP_PATH_KEY,
           lambda n: _asBool(n.get('keep-path')))

_addReader(WhdbButton, WHDB_POSITION_KEY, _readPosition)

_addReader(WhdbHasLocation, WHDB_POSITION_KEY, _readPosition)

_addReader(WhdbHasDimension, WHDB_DIMENSION_KEY,
           partial(_readFTuple, ('width', 'height')))

_addReader(WhdbTextBasedShapeMixin, WHDB_TEXT_HALIGN_KEY,
           lambda n: _asInt(n.get('horizontal-alignment')))

_addReader(WhdbTextBasedShapeMixin, WHDB_TEXT_VALIGN_KEY,
           lambda n: _asInt(n.get('vertical-alignment')))

_addReader(WhdbLineShape, WHDB_LINE_POINTA_KEY,
           partial(_readFTuple, ('x1', 'y1')))

_addReader(WhdbLineShape, WHDB_LINE_POINTB_KEY,
           partial(_readFTuple, ('x2', 'y2')))

_addReader(WhdbArrowShape, WHDB_LINE_POINTA_KEY,
           partial(_readFTuple, ('tipx', 'tipy')))

_addReader(WhdbArrowShape, WHDB_LINE_POINTB_KEY,
           partial(_readFTuple, ('tailx', 'taily')))

_addReader(WhdbRectangleShape, WHDB_POSITION_KEY, _readPosition)

_addReader(WhdbRectangleShape, WHDB_AREA_WIDTH_KEY,
           lambda n: _asFloat(n.get('width')))

_addReader(WhdbRectangleShape, WHDB_AREA_HEIGHT_KEY,
           lambda n: _asFloat(n.get('height')))

_addReader(WhdbOvalShape, WHDB_POSITION_KEY, _readPosition)

_addReader(WhdbOvalShape, WHDB_AREA_WIDTH_KEY,
           lambda n: _asFloat(n.get('width')))

_addReader(WhdbOvalShape, WHDB_AREA_HEIGHT_KEY,
           lambda n: _asFloat(n.get('height')))

_addReader(WhdbTextShape, WHDB_POSITION_KEY, _readPosition)

_addReader(WhdbTextShape, WHDB_TEXT_KEY, lambda n: n.text)

_addReader(WhdbTextShape, WHDB_BBOX_KEY,
           partial(_readFTuple, ('box-x', 'box-y', 'box-width', 'box-height')))

_addReader(WhdbConnectionPoint, WHDB_XPOINT_POSITION, _readPosition)

_addReader(WmhdbModelText, WMHDB_MODEL_TEXT_KEY, lambda n: n.text)


def wxio_rtiTypeDataFromTuple(xs):
    tdim = WMHDB_RTI_DIM_BY_NAME.get(xs[1]) if xs[1] else None
    dimx = int(xs[2]) if xs[2] else None
    dimy = int(xs[3]) if xs[3] else None
    if tdim is None:
        return WmhdbRtiType.getTypeForName(xs[0]), None, \
            dimx if (dimx and dimx > 1) else None, \
            dimy if (dimy and dimy > 1) else None
    return WmhdbRtiType.getTypeForName(xs[0]), tdim, dimx, dimy


_addReader(WmhdbWire, WMHDB_WIRE_UTYPE_KEY,
           lambda n: WmhdbRtiType(*wxio_rtiTypeDataFromTuple(
               (n.get('rtype'), n.get('rdim'), n.get('dimx'), n.get('dimy')))))


def _readModelPorts(inStr):
    ports = {}
    if inStr:
        for c in inStr.split(','):
            d, pn, dn = c.split(':')
            ports.setdefault(int(d), []).append((pn, dn))
    return ports


_addReader(WhdbModel, WHDB_MODEL_PORTS_KEY,
           lambda n: _readModelPorts(n.get(WHDB_MODEL_PORTS_KEY)))


# _addReader(WmhdbTopHierarchy, WMHDB_META_CALLBACKS_KEY,
#            lambda _: {})
