# ----------------------------------------------------------------------------
# - wxio_xtr02                                                               -
# - wild cruncher: meta-data adapters for XMLIO-V2.0           core / xmlio  -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       app_xio_xtr02
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher meta-data adapters for XMLIO-V2.0

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from functools import partial
from uuid import UUID
# from wcr_core.modules.model.db.wmhdb_top_block import WmhdbTopHierarchy, \
#     WMHDB_META_CALLBACKS_KEY

from wcr_core.xmlio.wxio_xtr01 import wxio_rtiTypeDataFromTuple
from wcr_core.db.whdb_connection import WhdbConnection, WHDB_KEEP_PATH_KEY
from wcr_core.db.whdb_button import WhdbButton
from wcr_core.db.whdb_model import WhdbModel, WHDB_MODEL_PORTS_KEY
from wcr_core.db.whdb_protocols import WHDB_POSITION_KEY, WhdbHasLocation, \
    WHDB_BBOX_KEY, WHDB_DIMENSION_KEY, WhdbHasDimension
from wcr_core.db.whdb_shape import WhdbTextBasedShapeMixin, \
    WHDB_TEXT_HALIGN_KEY, WHDB_TEXT_VALIGN_KEY, WhdbLineShape, \
    WHDB_LINE_POINTA_KEY, WHDB_LINE_POINTB_KEY, WhdbArrowShape, \
    WhdbRectangleShape, WhdbOvalShape, WHDB_AREA_WIDTH_KEY, \
    WHDB_AREA_HEIGHT_KEY, WhdbTextShape
from wcr_core.db.whdb_connection_point import WhdbConnectionPoint, \
    WHDB_XPOINT_POSITION
from wcr_core.xmlio.wxio_xml_0_2_0 import WxioXmlVer0_2_0
from wcr_core.modules.statemachine.db.wfhdb_block import WfhdbTopBlock
from wcr_core.modules.statemachine.db.wfhdb_datamodel_support import \
    WFHDB_DATAMODEL_KEY
from wcr_core.modules.model.db.wmhdb_wire import WmhdbWire, \
    WMHDB_WIRE_UTYPE_KEY
from wcr_core.modules.model.db.wmhdb_rti_type import WmhdbRtiType, \
    WMHDB_RTI_DIM_NAMES


def _asFloat(inStr):
    return float(inStr) if inStr else 0.0


def _asInt(inStr):
    return int(inStr) if inStr else 0.0


def _asBool(inStr):
    return True if (inStr and inStr.lower()[0] == 't') else False


def _toStr(val):
    return str(val) if val is not None else ''


def _read(name, node):
    # noinspection PyProtectedMember
    return WxioXmlVer0_2_0._readMetaNode(name, node)


def _readFTuple(keys, node):
    values = [_read(k, node) for k in keys]
    if all(values):
        return tuple(_asFloat(v) for v in values)
    else:
        return None


def _writeFTuple(keys, tpl):
    return list(zip(keys, map(_toStr, tpl)))


def _readPosition(node):
    return _readFTuple(('xpos', 'ypos'), node)


def _writePosition(tpl):
    return _writeFTuple(('xpos', 'ypos'), tpl)


def _addReader(clazz, meta, readF, writeF):
    clazz.META[meta].addRWFuncs(
        WxioXmlVer0_2_0.getCode(), WxioXmlVer0_2_0.getVersion(), readF, writeF)


def _addPosFuncs(clazz, meta):
    _addReader(clazz, meta, _readPosition, _writePosition)


def _addFTplFuncs(clazz, meta, tpl):
    _addReader(clazz, meta,
               partial(_readFTuple, tpl),
               partial(_writeFTuple, tpl))


def _addIntFuncs(clazz, meta, key):
    _addReader(clazz, meta,
               lambda n: _asInt(_read(key, n)),
               lambda n: [(key, _toStr(n))])


def _addFloatFuncs(clazz, meta, key):
    _addReader(clazz, meta,
               lambda n: _asFloat(_read(key, n)),
               lambda n: [(key, _toStr(n))])


_addReader(WhdbConnection, WHDB_KEEP_PATH_KEY,
           lambda n: _asBool(_read('keep', n)),
           lambda v: [('keep', _toStr(v))])

_addReader(WhdbButton, WHDB_POSITION_KEY,
           lambda n: (_asFloat(n.find('position').get('xpos')),
                      _asFloat(n.find('position').get('ypos'))),
           lambda v: [(('position', 'xpos'), _toStr(v.getPosition()[0])),
                      (('position', 'ypos'), _toStr(v.getPosition()[1]))])

_addPosFuncs(WhdbHasLocation, WHDB_POSITION_KEY)

_addFTplFuncs(WhdbHasDimension, WHDB_DIMENSION_KEY, ('width', 'height'))

_addIntFuncs(
    WhdbTextBasedShapeMixin, WHDB_TEXT_HALIGN_KEY, 'horizontal-alignment')

_addIntFuncs(
    WhdbTextBasedShapeMixin, WHDB_TEXT_VALIGN_KEY, 'vertical-alignment')

_addFTplFuncs(WhdbLineShape, WHDB_LINE_POINTA_KEY, ('x1', 'y1'))

_addFTplFuncs(WhdbLineShape, WHDB_LINE_POINTB_KEY, ('x2', 'y2'))

_addFTplFuncs(WhdbArrowShape, WHDB_LINE_POINTA_KEY, ('tipx', 'tipy'))

_addFTplFuncs(WhdbArrowShape, WHDB_LINE_POINTB_KEY, ('tailx', 'taily'))

_addPosFuncs(WhdbRectangleShape, WHDB_POSITION_KEY)

_addFloatFuncs(WhdbRectangleShape, WHDB_AREA_WIDTH_KEY, 'width')

_addFloatFuncs(WhdbRectangleShape, WHDB_AREA_HEIGHT_KEY, 'height')

_addPosFuncs(WhdbOvalShape, WHDB_POSITION_KEY)

_addFloatFuncs(WhdbOvalShape, WHDB_AREA_WIDTH_KEY, 'width')

_addFloatFuncs(WhdbOvalShape, WHDB_AREA_HEIGHT_KEY, 'height')

_addPosFuncs(WhdbTextShape, WHDB_POSITION_KEY)

_addFTplFuncs(WhdbTextShape, WHDB_BBOX_KEY,
              ('box-x', 'box-y', 'box-width', 'box-height'))

_addPosFuncs(WhdbConnectionPoint, WHDB_XPOINT_POSITION)


def _readRtiType(node):
    if node is None or not _read('rtype', node):
        return None

    return WmhdbRtiType(
        *wxio_rtiTypeDataFromTuple(
            (_read('rtype', node), _read('rdim', node),
             _read('dimx', node), _read('dimy', node))))


def _writeRtiType(rtyp):
    if not rtyp:
        return []

    return [('rtype', WmhdbRtiType.getTypeName(rtyp.getType())),
            ('rdim', WMHDB_RTI_DIM_NAMES.get(rtyp.getDimension())),
            ('dimx', _toStr(rtyp.getXdim())),
            ('dimy', _toStr(rtyp.getYdim()))]


_addReader(WmhdbWire, WMHDB_WIRE_UTYPE_KEY, _readRtiType, _writeRtiType)


def _readModelPorts(inStr):
    ports = {}
    if inStr:
        for c in inStr.split(','):
            d, pn, dn = c.split(':')
            ports.setdefault(int(d), []).append((pn, dn))
    return ports


_addReader(WhdbModel, WHDB_MODEL_PORTS_KEY,
           lambda n: _readModelPorts(_read(WHDB_MODEL_PORTS_KEY, n)),
           lambda v: ','.join(':'.join((str(d), pn or '', dn or ''))
                              for d, l in v.items() for pn, dn in l))


def _readDatamodelTuples(s):
    r = {}
    for x in s.split(';'):
        n, t, k, e, s = x.split('|')
        r[n] = (t, k, e, s)
    return r


_addReader(WfhdbTopBlock, WFHDB_DATAMODEL_KEY,
           lambda n: _readDatamodelTuples(_read(WFHDB_DATAMODEL_KEY, n)),
           lambda v: ';'.join('{}|{}|{}|{}|{}'.format(n, t or '', k or '',
                                                      e or '', s or '')
                              for n, (t, k, e, s) in v.items()))


# def _readCallbacks(node):
#     cstr = _read(WMHDB_META_CALLBACKS_KEY, node)
#
#     if cstr:
#         return {k: f for kf in cstr.split(';') for k, f in kf.split(':')}
#     else:
#         return {}

# _addReader(WmhdbTopHierarchy, WMHDB_META_CALLBACKS_KEY,
#            _readCallbacks,
#            lambda v: ';'.join('{}:{}'.format(k, f) for k, f in v.items()))
