# ----------------------------------------------------------------------------
# - wcli-actions                                                             -
# - wild cruncher:cli commands                                frontend / cli -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcli_actions
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher console commands

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from silkan_common.scmn_mime_type import scmn_mime_types
from silkan_common.scmn_utils import first
from wcr_frontend.cli.wcli_command import WcliCommand
from wcr_frontend.cli.wcli_context import WcliContext
from wcr_frontend.cli.wcli_parameter import WcliParameter, WCLI_PTYPES, \
    WcliOption
from wcr_core.modules.wmod_application_module import wmod_getModuleForPdb
from wcr_core.project.wprj_project_verifier import WprjProjectVerifier
from wcr_core.project.wprj_project import wprj_load_project, \
    wprj_import_project
from wcr_backend.modules.model.aob.wmaob_builder import WmaobBuilder
from wcr_core.common.wcmn_file_handler import WcmnFile
from silkan_common.scmn_file_path import scmn_expshenv, scmn_listdir, \
    scmn_isdir, scmn_path_join
from silkan_common.scmn_process import ScmnProcStreamHandlerP
from wcr_backend.wbke_make_execute import WbkeMakeExecute
from wcr_backend.modules.model.runtime.wmrun_execute import WmrunExecute
from wcr_backend.modules.model.actions.wmbke_generate_code_action import \
    wmbke_generateCode
from wcr_backend.modules.model.aob.wmaob_callback_gene import \
    WmaobCallbacksGenerator
from wplg_cold.wcld_variables_parser import WcldColdVarsParser
from wcr_core.modules.model.db.wmhdb_type_propagation import \
    WmhdbTypePropagator
from wplg_cold.wcld_builder import WcldColdBuilder
import os


class WcliLoadCommand(WcliCommand):
    """
    Console command to load a file supported by Wild Cruncher.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('load', 'Load a project')

        self.__context = context

        self.setRequiredParameters(
            WcliParameter('file-path', context, WCLI_PTYPES['file'],
                          'Path of file to load'))

    def execute(self, options, parameters):
        context = self.__context
        path = scmn_expshenv(parameters[0])
        file = WcmnFile(path)
        gctx = context.getGlobalContext()

        projectDB = wprj_load_project(file, gctx)

        if projectDB is None:
            raise Exception('Could not read block model from ' + path)
        context.setProjectDB(projectDB)
        gctx.addRecentlyOpenedFile(path)


class WcliLoadRecentCommand(WcliCommand):
    """
    Console command to load a recently opened file.
    """

    def __init__(self, loadAction, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('load-recent', 'Load a recently opened project')

        self.__loadAction = loadAction

        self.__context = context

    def execute(self, options, parameters):
        recents = self.__context.getGlobalContext().getRecentlyOpenedFiles()

        for i, r in enumerate(recents):
            print('({i: 2d}) {r}'.format(i=i, r=r))

        idx = input('Please the enter index of file to load,'
                    ' or leave empty to abort: ')

        if idx:
            try:
                self.__loadAction.execute(options, [recents[int(idx)]])
            except (ValueError, IndexError):
                print('Invalid index')


class WcliImportCommand(WcliCommand):
    """
    Console command to import a model.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('import', 'Import an MDL file')

        self.__context = context

        self.setRequiredParameters(
            WcliParameter('file-path', context, WCLI_PTYPES['file'],
                          'Path of file to import'))

    def execute(self, options, parameters):
        context = self.__context
        path = scmn_expshenv(parameters[0])
        file = WcmnFile(path)

        projectDB = wprj_import_project(file, context.getGlobalContext())

        if projectDB is None:
            raise Exception('Could not import model from ' + path)
        context.setProjectDB(projectDB)


class WcliSaveCommand(WcliCommand):
    """
    Console command to save the current project to disk.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('save', 'Save a project file')

        self.__context = context

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        if pjdb.path:
            pjdb.save()
        else:
            path = input('Please enter new path (leave empty to abort): ')
            path = path.strip()
            if path:
                pjdb.saveAs(scmn_expshenv(path))
                self.__context.getGlobalContext().addRecentlyOpenedFile(path)


class WcliSaveAsCommand(WcliCommand):
    """
    Console command to save the current project at a new location.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('saveas', 'Save a project file under a new path')

        self.__context = context

        self.setRequiredParameters(
            WcliParameter('file-path', context, WCLI_PTYPES['file'],
                          'Path to save project file to'))

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        path = scmn_expshenv(parameters[0])
        if path:
            pjdb.saveAs(path)
            self.__context.getGlobalContext().addRecentlyOpenedFile(path)


class WcliGenerateCommand(WcliCommand):
    """
    Console command to generate code for a model instance.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('generate',
                         'Generate C code for an instance or top level')

        self.__context = context

        self.addOptions(WcliOption('profile', context, scode='-p',
                                   ptype=WCLI_PTYPES['string'],
                                   hlpm='Generation profile to use'))

        self.setOptionalParameters(
            WcliParameter('instance-or-function', context,
                          WCLI_PTYPES['instance'],
                          'Model instance or function to generate code for '
                          '(default: top level instance)'))

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        model = parameters[0] if parameters else pjdb.getTopHierarchy()

        result = _wcli_verifyDatabase(pjdb)
        if result is None:
            print('Could not generate code, could not verify database')
            return
        elif result.hasErrors():
            print('Could not generate code:')
            for error in result.allErrors:
                print('  ' + error.errorMessage)
            return

        fbld = WmaobBuilder(self.__context)
        fbld.generate(model, _WcliProcStreamHandler())


class WcliInitCallbacksCommand(WcliCommand):
    """
    Console command to generate code for a model callbacks.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('initcallback',
                         'Initialises callback for top level')

        self.__context = context

        self.addOptions(WcliOption('profile', context, scode='-p',
                                   ptype=WCLI_PTYPES['string'],
                                   hlpm='Generation profile to use'))

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return None
        prof = 'default'
        if 'profile' in options:
            prof = options.get('profile')
        mdlcallbackinit = WmaobCallbacksGenerator(self.__context, prof)
        return


class WcliPropagateTypeCommand(WcliCommand):
    """
    Console command to propagate types for a model.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('propagate_type',
                         'Propagate types for top level')

        self.setOptionalParameters(
            WcliParameter('sched-group', context,
                          WCLI_PTYPES['schedgrp'],
                          'Scheduling group to compile '
                          '(default: all scheduling groups)'))

        self.__context = context

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        sset = pjdb.getTopHierarchy().getSchedulingGroupSet()
        canuse = True
        if len(sset) > 1:
            for v in pjdb.getTopHierarchy().getCallbackDict().values():
                if v != '':
                    canuse = False
        file = None
        if canuse:
            sched = parameters[0] if parameters else None
            if sched is None:
                for name in sset:
                    if name != 'none':
                        sched = name
            file = os.path.join(self.__context.getProjectWorkspace().
                                getProjectPath(), 'default', sched,
                                'CldGlobals.vars')
        if os.path.isfile(file):
            coldparser = WcldColdVarsParser(file)
        else:
            coldparser = None
        coldacces = WcldColdBuilder(self.__context)
        tblk = self.__context.getProjectDB().getTopHierarchy()
        tprp = WmhdbTypePropagator()
        tprp.propagateTypes(tblk, coldparser, coldacces)


class WcliGUIGenerateCommand(WcliCommand):
    """
    Console command to generate code for a model like the GUI do.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('gui_generate',
                         'Generate C code for the top level like GUI does')

        self.__context = context

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        wmbke_generateCode(self.__context, _WcliProcStreamHandler())


class WcliCompileCommand(WcliCommand):
    """
    Console command to compile code.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('compile',
                         'Compile previously generated C code')

        self.__context = context

        self.addOptions(WcliOption('profile', context, scode='-p',
                                   ptype=WCLI_PTYPES['string'],
                                   hlpm='Generation profile to use'))

        self.addOptions(WcliOption('debug', context, scode='-g',
                                   hlpm='Compile in debug mode'))

        self.setOptionalParameters(
            WcliParameter('sched-group', context,
                          WCLI_PTYPES['schedgrp'],
                          'Scheduling group to compile '
                          '(default: all scheduling groups)'))

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        sgrp = parameters[0] if parameters else None
        make = WbkeMakeExecute(self.__context)

        if 'debug' in options:
            make.addEnvironment('SGRPRJCOMPILE', 'debug')
        strm = _WcliProcStreamHandler()

        if sgrp:
            make.doAllRule(sgrp, strm)
        else:
            for sgrp in pjdb.getTopHierarchy().getSchedulingGroupSet():
                if sgrp != 'none':
                    make.doAllRule(sgrp, strm)


class WcliExecuteCommand(WcliCommand):
    """
    Console command to execute complied code.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('execute',
                         'Execute previously generated/compiled code')

        self.__context = context

        self.addOptions(WcliOption('profile', context, scode='-p',
                                   ptype=WCLI_PTYPES['string'],
                                   hlpm='Generation profile to use'))
        self.addOptions(WcliOption('bus', context, scode='-b',
                                   ptype=WCLI_PTYPES['string'],
                                   hlpm='Name of bus to use'))
        self.addOptions(WcliOption('period', context, scode='-t',
                                   ptype=WCLI_PTYPES['float'],
                                   hlpm='Bus period'))
        self.addOptions(WcliOption('cycles', context, scode='-c',
                                   ptype=WCLI_PTYPES['int'],
                                   hlpm='Maximum number of bus cycles'))

        self.setRequiredParameters(
            WcliParameter('sched-group', context,
                          WCLI_PTYPES['schedgrp']))

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        execute = WmrunExecute(self.__context, options.get('profile'))

        if 'bus' in options:
            execute.setBusName(options.get('bus'))
        if 'period' in options:
            execute.setBusPeriod(options.get('period'))
        if 'cycles' in options:
            execute.setMaximumCycleCount(options.get('cycles'))

        sgrp = parameters[0]
        execute.execute(sgrp, _WcliProcStreamHandler())


class WcliCleanCommand(WcliCommand):
    """
    Console command to clean the workspace.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('clean',
                         'Clean the workspace')

        self.__context = context

        self.addOptions(WcliOption('profile', context, scode='-p',
                                   ptype=WCLI_PTYPES['string'],
                                   hlpm='Generation profile to use'))

        self.setOptionalParameters(
            WcliParameter('sched-group', context,
                          WCLI_PTYPES['schedgrp'],
                          'Scheduling group to compile '
                          '(default: all scheduling groups)'))

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        sgrp = parameters[0] if parameters else None

        make = WbkeMakeExecute(self.__context)

        strm = _WcliProcStreamHandler()
        if sgrp:
            print('Cleaning ' + sgrp)
            make.doCleanRule(sgrp, strm)
        else:
            print('Cleaning project ' + pjdb.getProjectName())
            make.doCleanProject(strm)


class WcliDbVerifyCommand(WcliCommand):
    """
    Console command to verify a database's consistency.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('verify',
                         'Verify the current database\'s consistency')

        self.__context = context

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        result = _wcli_verifyDatabase(pjdb)

        if result is None:
            print('Could not verify database')
        elif result.allErrors:
            for error in result.allErrors:
                print(error.errorMessage)
        else:
            print('No errors found')


class WcliDbInfoCommand(WcliCommand):
    """
    Console command to show information about the current project.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('info',
                         'Display information about the current project')

        self.__context = context

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        toph = pjdb.getTopHierarchy()
        if pjdb.getPath():
            print('Path: ' + pjdb.getPath())
        if pjdb.getProjectName():
            print('Name: ' + pjdb.getProjectName())
        if toph.getName():
            print('Top-Level Name: ' + toph.getName())
        if toph.getInfo():
            print('Top-Level Info: ' + toph.getInfo())
        print()

        instances = list(toph.getBlockInstances())
        if instances:
            print('[BLOCK INSTANCES]')
            for inst in instances:
                m = inst.getBlock()
                x, y = inst.getPosition()
                print('  %s (%s) #%s @%d/%d' %
                      (inst.getName(), m.getName(),
                       inst.getSchedulingGroup(), x, y))
            print()

        instances = list(toph.getModelInstances())
        if instances:
            print('[MODEL INSTANCES]')
            for inst in instances:
                m = inst.getModel()
                x, y = inst.getPosition()
                print('  %s (%s - %s) #%s @%d/%d' %
                      (inst.getName(), m.getName(), m.getMedia(),
                       inst.getSchedulingGroup(), x, y))
            print()

        instances = list(toph.getOperatorInstances())
        if instances:
            print('[OPERATOR INSTANCES]')
            for inst in instances:
                m = inst.getOperator()
                x, y = inst.getPosition()
                print('  %s (%s) #%s @%d/%d' %
                      (inst.getName(), m.getName(),
                       inst.getSchedulingGroup(), x, y))
            print()


class WcliLogCommand(WcliCommand):
    """
    Console command to display log entries.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('log', 'Display log messages')

        #: Context
        self.__context = context

        #: Last log entry shown
        self.__lastShown = -1

        self.setOptionalParameters(
            WcliParameter('number-of-entries', context,
                          WCLI_PTYPES['string'],
                          'Number of log entries to show, or \'all\''))

    def execute(self, options, parameters):
        logger = self.__context.getGlobalContext().getMessageSource()

        since = self.__lastShown if not parameters else 0

        msgs = logger.messages(since=since)

        if parameters:
            p = parameters[0]

            if p != 'all':
                try:
                    entries = int(p)
                    msgs = msgs[-entries:]
                except:
                    raise ValueError('Invalid value for number of log '
                                     'entries')

        for msg in msgs:
            print(msg.formatted())
            self.__lastShown = msg.counter


class WcliWorkspaceCommand(WcliCommand):
    """
    Console command to show the contents of the workspace.
    """

    def __init__(self, context):
        """
        :param context: The console execution context
        :type context: WcliContext
        """
        super().__init__('workspace',
                         'Display the contents of the workspace')

        #: Context
        self.__context = context

    def execute(self, options, parameters):
        pjdb = self.__context.getProjectDB()
        if pjdb is None:
            print('No open project')
            return

        wspc = self.__context.getProjectWorkspace().getProjectPath()

        self._printDir(wspc)

    def _printDir(self, path, level=0):
        def dirFilter(name):
            return scmn_isdir(scmn_path_join(path, name))

        entries = set(scmn_listdir(path))
        dirs = list(filter(dirFilter, entries))

        for sub in dirs:
            dirType = 'profile' if level == 0 else 'sched. group'
            print((level * 4 * ' ') + sub + ' (' + dirType + ')')
            self._printDir(scmn_path_join(path, sub), level + 1)

        for file in entries.difference(dirs):
            _mime = first(scmn_mime_types(path))
            """:type: ScmnMimeTypeP | None"""
            mime = _mime.mimeType if _mime else None
            mime = ' (' + mime + ')' if mime else ''

            print((level * 4 * ' ') + file + mime)


class _WcliProcStreamHandler(ScmnProcStreamHandlerP):
    def handleControlMsg(self, text, msgType):
        if text:
            print('- ' + text)

    def handleOutput(self, text):
        if text:
            print('* ' + text)

    def handleWarning(self, text):
        if text:
            print('W ' + text)

    def handleError(self, text):
        if text:
            print('! ' + text)


def _wcli_verifyDatabase(pjdb):
    try:
        module = wmod_getModuleForPdb(pjdb)
    except:
        print('Could not find module for type {}'.format(
            pjdb.getAppModule()))
        return None

    verifier = WprjProjectVerifier(module.createHierarchyVerifier(None),
                                   module.getProjectVerifierModules())
    return verifier.verifyProjectDatabase(pjdb)
