# ----------------------------------------------------------------------------
# - wcli-command                                                             -
# - wild cruncher:cli command registration object             frontend / cli -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcli_parameter
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher console command

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import ABCMeta, abstractmethod


class WcliCommand(metaclass=ABCMeta):
    """
    The WcliCommand is a command registration class. A command is defined
    by a name, an optional help message, and a set of options/parameters.
    """

    def __init__(self, name, hlpm=None):
        """
        Create a command
        :param name: the command name
        :type name: str
        :param hlpm: the command help
        :type hlpm: str
        """
        # base initialization
        self.__name = name
        self.__help = hlpm
        self.__options = []
        self.__optionDict = {}
        self.__reqParams = []
        self.__optParams = []
        self.__params = []

    @property
    def name(self):
        """
        Get the command's name.

        :return: Name of the command
        :rtype: str
        """
        return self.__name

    @property
    def helpMessage(self):
        """
        Get the command's help message.

        :return: Help message for the command
        :rtype: str
        """
        return self.__help

    @property
    def options(self):
        """
        Get the command's options.

        :return: Options of the command
        :rtype: list[WcliOption]
        """
        return self.__options

    def getOption(self, name):
        """
        Get one of the command's options by name.

        :param name: Name of the command to retrieve
        :type name: str
        :return: Option with the given name, or None
        :rtype: WcliOption
        """
        return self.__optionDict.get(name, None)

    @property
    def parameters(self):
        """
        Get the command's parameters, required and optional ones.

        :return: All parameters of the command
        :rtype: list[WcliParameter]
        """
        return self.__params

    @property
    def requiredParameters(self):
        """
        Get the command's required parameters.

        :return: Required parameters of the command
        :rtype: list[WcliParameter]
        """
        return self.__reqParams

    @property
    def requiredParameterCount(self):
        """
        Get the number of required parameters for the command.

        :return: Number of required parameters
        :rtype: int
        """
        return len(self.__reqParams)

    @property
    def optionalParameters(self):
        """
        Get the command's optional parameters.

        :return: Optional parameters of the command
        :rtype: list[WcliParameter]
        """
        return self.__optParams

    def addOptions(self, *options):
        """
        Add options for the command.

        :param options: Options to add
        :type options: Iterable[WcliOption]
        """
        for option in options:
            self.__options.append(option)
            if option.shortCode:
                self.__optionDict[option.shortCode] = option
            if option.longCode:
                self.__optionDict[option.longCode] = option

    def setRequiredParameters(self, *parameters):
        """
        Set the command's required parameters.
        Previously set required parameters will be replaced.

        :param parameters: Required parameters for the command
        :type parameters: Iterable[WcliParameter]
        """
        self.__reqParams.extend(parameters)
        self.__params = self.__reqParams + self.__optParams

    def setOptionalParameters(self, *parameters):
        """
        Set the command's optional parameters.
        Previously set optional parameters will be replaced.

        :param parameters: Optional parameters for the command
        :type parameters: Iterable[WcliParameter]
        """
        self.__optParams.extend(parameters)
        self.__params = self.__reqParams + self.__optParams

    @abstractmethod
    def execute(self, options, parameters):
        """
        Execute the command with the given options and parameters.

        :param options: Option values
        :type options: dict[str, object]
        :param parameters: Ordered parameter values
        :type parameters: list[object]
        """
        pass
