# ----------------------------------------------------------------------------
# - wcli-parameter                                                           -
# - wild cruncher:cli command option                          frontend / cli -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcli_parameter
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher command option class

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import ABCMeta, abstractmethod

from wcr_frontend.cli.wcli_context import WcliContext
from wcr_core.modules.model.db.wmhdb_instance_base import WmhdbInstanceBaseP
from wcr_core.modules.model.db.wmhdb_instance import WmhdbInstance
from silkan_common.scmn_file_path import scmn_expshenv, scmn_path_dir_base, \
    scmn_listdir, scmn_path_join, scmn_isdir


class WcliParamType(metaclass=ABCMeta):
    """
    Value type of an option/parameter.
    """

    # noinspection PyPropertyDefinition
    @property
    @abstractmethod
    def name(self):
        """
        Get the type's name.

        :return: Name of the type
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def parse(self, value, context):
        """
        Parse an input value into an object of the described type.

        :param value: Input value to parse
        :type value: str
        :param context: Console context
        :type context: WcliContext
        :return: Parsed value
        """
        raise NotImplementedError()

    @abstractmethod
    def format(self, value, context):
        """
        Format an object of the described type into a string

        :param value: Input value to format
        :param context: Console context
        :type context: WcliContext
        :return: Formatted string
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getCompletions(self, text, context):
        """
        Get possible textual completions for a given input text, if any.
        Any returned values are parseable by :py:meth:`parse`.

        :param text: Input text to be completed
        :type text: str
        :param context: Console context
        :type context: WcliContext
        :return: Possible completions for `text`
        :rtype: list[str]
        """
        raise NotImplementedError()


class _WcliMultiType(WcliParamType):
    """
    Multi-value option/parameter type.
    """

    def __init__(self, stype):
        """
        :param stype: Type for single values
        :type stype: WcliParamType
        """
        self.__singleType = stype

    def parse(self, value, context):
        return [self.__singleType.parse(v, context) for v in value.split(',')]

    def format(self, value, context):
        return ','.join([self.__singleType.format(v, context) for v in value])

    def getCompletions(self, text, context):
        values = text.split(',')
        if values:
            comps = self.__singleType.getCompletions(values[-1], context)
            if comps:
                others = ','.join(values[:-1])
                return [others + ',' + c for c in comps]
        return []

    @property
    def name(self):
        return self.__singleType.name + 's'


class _WcliStringType(WcliParamType):
    """
    String option/parameter type.
    """

    def parse(self, value, context):
        return value

    def format(self, value, context):
        return value

    def getCompletions(self, text, context):
        return []

    @property
    def name(self):
        return 'string'


class _WcliIntType(WcliParamType):
    """
    Integer option/parameter type.
    """

    def parse(self, value, context):
        return int(value)

    def format(self, value, context):
        return str(value)

    def getCompletions(self, text, context):
        return []

    @property
    def name(self):
        return 'integer'


class _WcliFloatType(WcliParamType):
    """
    Float option/parameter type.
    """

    def parse(self, value, context):
        return float(value)

    def format(self, value, context):
        return str(value)

    def getCompletions(self, text, context):
        return []

    @property
    def name(self):
        return 'float'


class _WcliFilepathType(WcliParamType):
    """
    File path option/parameter type.
    """

    def parse(self, value, context):
        return value

    def format(self, value, context):
        return value

    def getCompletions(self, text, context):
        path = scmn_expshenv(text or './')
        dn, fn = scmn_path_dir_base(path)
        pattern = fn + '.*' if fn else None
        entries = scmn_listdir(dn, fullpath=True, pattern=pattern)
        if len(entries) == 1 and scmn_isdir(entries[0]):
            return [scmn_path_join(entries[0], '')]
        else:
            return entries

    @property
    def name(self):
        return 'file path'


class _WcliInstanceType(WcliParamType):
    """
    Model instance option/parameter type.
    """

    def parse(self, value, context) -> WmhdbInstanceBaseP:
        top = context.getProjectDB().getTopHierarchy()
        for inst in top.getInstances():
            if inst.getName() == value:
                return inst
        raise ValueError(
            'No instance/function with name {n} found'.format(n=value))

    def format(self, value, context):
        """
        :type value: WhdbInstanceP
        """
        return value.getName()

    def getCompletions(self, text, context):
        top = context.getProjectDB().getTopHierarchy()
        return [inst.getName() for inst in top.getInstances()
                if inst.getName().startswith(text)]

    @property
    def name(self):
        return 'model instance or function'


class _WcliMimeInstanceType(WcliParamType):
    """
    Model instance option/parameter type of a specified mime-type.
    """

    def __init__(self, name, mime):
        """
        :param name: Name of the type
        :type name: str
        :param mime: Mime-type
        :type mime: str
        """
        self.__name = name
        self.__mime = mime

    def parse(self, value:str, context:WcliContext) -> WmhdbInstance:
        top = context.getProjectDB().getTopHierarchy()
        for inst in top.getModelInstances():
            if inst.getMedia() == self.__mime and inst.getName() == value:
                return inst
        raise ValueError(
            'No instance with name {n} and model type {m} found'.format(
                n=value, m=self.__mime))

    def format(self, value, context):
        """
        :type value: WhdbInstanceP
        """
        return value.getName()

    def getCompletions(self, text, context):
        top = context.getProjectDB().getTopHierarchy()
        return [inst.getName() for inst in top.getModelInstances()
                if inst.getMedia() == self.__mime and
                inst.getName().startswith(text)]

    @property
    def name(self):
        return self.__name


class _WcliSchedulingGroupType(WcliParamType):
    """
    Scheduling group option/parameter type.
    """

    def parse(self, value, context) -> str:
        if context.getProjectDB() is None or \
                context.getProjectDB().getTopHierarchy() is None:
            raise Exception(
                'No project open, cannot parse scheduling group')
        top = context.getProjectDB().getTopHierarchy()
        if value in top.getSchedulingGroupSet():
            return value
        raise ValueError(
            'No scheduling group \'{n}\' found'.format(n=value))

    def format(self, value, context):
        return value

    def getCompletions(self, text, context):
        pdb = context.getProjectDB()
        if pdb is None or pdb.getTopHierarchy() is None:
            return []
        sgrps = pdb.getTopHierarchy().getSchedulingGroupSet()
        sgrps.discard('none')
        return [sgrp for sgrp in sgrps if sgrp.startswith(text)]

    @property
    def name(self):
        return 'scheduling group'


WCLI_PTYPES = {
    'string': _WcliStringType(),
    'float': _WcliFloatType(),
    'int': _WcliIntType(),
    'file': _WcliFilepathType(),
    'instance': _WcliInstanceType(),
    'schedgrp': _WcliSchedulingGroupType(),
}


class WcliOption(object):
    """
    A command option is optional when executing a command.
    It can usually be specified by a short code (`-x`) or long code
    (`--option-x`).
    Options can have an optional argument.
    """

    # option constructor
    def __init__(self, name, context, ptype=None,
                 scode=None, lcode=None, hlpm=None):
        """
        :param name: the option's name
        :type name: str
        :param context: Execution context
        :type context: WcliContext
        :param ptype: the option's parameter type
        :type ptype: WcliParamType
        :param scode: the option character
        :type scode: str
        :param lcode: the option string
        :type lcode: str
        :param hlpm: the option help
        :type hlpm: str
        """
        self.__name = name
        self.__context = context
        self.__parameterType = ptype
        self.__shortCode = scode
        self.__longCode = lcode if lcode else '--' + name
        self.__help = hlpm

    @property
    def name(self):
        """
        Get the option's name.

        :return: The name of the option
        :rtype: str
        """
        return self.__name

    @property
    def typeName(self):
        """
        Get the option's argument's type name.
        Returns `None` if the option takes no argument.

        :return: The type name of the option's argument
        :rtype: str
        """
        if self.__parameterType is not None:
            return self.__parameterType.name
        else:
            return None

    @property
    def requiresParameter(self):
        """
        Whether the option requires an argument.

        :return: True iff the option requires an argument
        :rtype: bool
        """
        return self.__parameterType is not None

    @property
    def shortCode(self):
        """
        Get the option's short code, if any.

        :return: The short code of the option, or `None`
        :rtype: str
        """
        return self.__shortCode

    @property
    def longCode(self):
        """
        Get the option's long code.

        :return: The long code of the option
        :rtype: str
        """
        return self.__longCode

    @property
    def helpMessage(self):
        """
        Get the option's help message.

        :return: The help message of the option
        :rtype: str
        """
        return self.__help

    def parseArgument(self, arg):
        """
        Parse a value for the option's argument, given as string.

        :param arg: Argument value
        :type arg: str
        :return: Given argument value converted to its proper type
        """
        if self.__parameterType is not None:
            return self.__parameterType.parse(arg, self.__context)
        else:
            return None

    def getCompletions(self, text):
        """
        Get possible textual completions for a given input text, if any.
        Any returned values are parseable by :py:meth:`parseArgument`.

        If the option does usually take an argument, but no completions could
        be found, an empty list will be returned.
        If the option does not take arguments, `None` will be returned.

        :param text: Input text to provide completions for
        :type text: str
        :return: Possible completions for the input text, or `None`
        :rtype: list[str]
        """
        if self.__parameterType is not None:
            return self.__parameterType.getCompletions(text, self.__context)
        else:
            return None


class WcliParameter(object):
    """
    Command parameter.
    """

    def __init__(self, name, context, ptype, hlpm=None):
        """
        Create a command option
        :param name: the option's name
        :type name: str
        :param context: Execution context
        :type context: WcliContext,None
        :param ptype: the option's parameter type
        :type ptype: WcliParamType
        :param hlpm: the option help
        :type hlpm: str
        """
        # base initialization
        self.__name = name
        self.__context = context
        self.__parameterType = ptype
        self.__help = hlpm

    @property
    def name(self):
        """
        Get the parameter's name.

        :return: The name of the parameter
        :rtype: str
        """
        return self.__name

    @property
    def parameterType(self):
        """
        Get the parameter's type name.

        :return: The type name of the parameter
        :rtype: str
        """
        return self.__parameterType

    @property
    def helpMessage(self):
        """
        Get the parameter's help message.

        :return: The help message of the parameter
        :rtype: str
        """
        return self.__help

    def parseArgument(self, arg):
        """
        Parse a value for the parameter's argument, given as string.

        :param arg: Argument value
        :type arg: str
        :return: Given argument value converted to its proper type
        """
        return self.__parameterType.parse(arg, self.__context)

    def getCompletions(self, text):
        """
        Get possible textual completions for a given input text, if any.
        Any returned values are parseable by :py:meth:`parseArgument`.

        :param text: Input text to provide completions for
        :type text: str
        :return: Possible completions for the input text
        :rtype: list[str]
        """
        return self.__parameterType.getCompletions(text, self.__context)
