# ----------------------------------------------------------------------------
# - wcli-repl                                                                -
# - wild cruncher:cli REPL / interpreter                      frontend / cli -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcli_parameter
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher REPL / command line

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------
from collections import OrderedDict

try:
    import readline
except ImportError:
    readline = None

import shlex
import sys
import traceback

from wcr_frontend.cli.wcli_command import WcliCommand
from wcr_frontend.cli.wcli_parameter import WcliParameter, WCLI_PTYPES
from silkan_common.scmn_file_path import *
from silkan_common.scmn_memoization import memoize1


# ----------------------------------------------------------------------------
# - model class                                                              -
# ----------------------------------------------------------------------------

class _WcliException(Exception):
    pass


class WcliCommandLine(object):
    """
    REPL / Command line interpreter.
    Can be extended by adding :py:class:`WcliCommand` instances.
    """

    def __init__(self, prompt='% ', histFile=None, histLen=100, strace=False):
        """
        :param prompt: Prompt to display at the beginning of the input line
        :type prompt: str
        :param histFile: History file path name
        :type histFile: str
        :param histLen: Maximum length of the history file when saving it
        :type histLen: int
        """
        #: Command line prompt
        self.__prompt = prompt
        #: History file path
        self.__histFile = scmn_expshenv(histFile)
        #: Show stack trace on errors
        self.__stacktrace = strace

        #: Available commands
        self.__commands = OrderedDict()
        #: Command line parser
        self.__parser = _WcliParser(self.__commands)
        #: Whether the repl is still running, will stop looping if `False`
        self.__running = True

        if readline:
            completer = _WcliCompleter(self.__commands, self.__prompt)
            readline.set_completer(completer.complete)
            readline.set_completion_display_matches_hook(
                completer.displayCompletions)

            readline.set_history_length(histLen)
            if self.__histFile and scmn_isfile(self.__histFile):
                readline.read_history_file(self.__histFile)

        self.addCommand(_WcliHelpCommand(self.__commands))
        self.addCommand(_WcliExitCommand(self._exit))

    def repl(self):
        """
        Main read/eval/print-loop.
        """
        while self.__running:
            try:
                # read a line
                line = input(self.__prompt)
                # parse and execute the line
                try:
                    self._parseAndExecute(line)
                except Exception as e:
                    if self.__stacktrace:
                        traceback.print_exc()
                    else:
                        print(str(e) or e.__class__.__name__)
            except KeyboardInterrupt:
                print('^C')
            except (SystemExit, EOFError):
                self._exit()

    def evalAll(self, lines):
        """
        Evaluate all given lines of input.

        :param lines: Lines of input
        :type lines: Iterable[str]
        """
        for line in lines:
            # parse and execute the line
            try:
                line = line.split('#')[0].strip()
                if line:
                    print('> ' + line)
                    self._parseAndExecute(line)
                    print('--> Ok')
            except Exception as e:
                if self.__stacktrace:
                    traceback.print_exc()
                else:
                    print('--> Failed / ' + (str(e) or e.__class__.__name__))
                sys.exit(-1)

    def _exit(self):
        """
        Exit the console's main loop.
        """
        print('\nGood bye!')
        self.__running = False
        if readline and self.__histFile:
            scmn_mkdir(scmn_path_dir(self.__histFile))
            readline.write_history_file(self.__histFile)

    def _parseAndExecute(self, line):
        """
        Parse and execute a line of input.

        :param line: Line to parse and execute
        :type line: str
        """
        cmd, opts, params = self.__parser.parse(line)
        if cmd is not None:
            cmd.execute(opts, params)

    def addCommand(self, command):
        """
        Add a new command .

        :param command: Command to be added
        :type command: WcliCommand
        :return: The command line object itself, for chaining
        :rtype: WcliCommandLine
        """
        name = command.name
        # try to find the command
        if name in self.__commands:
            raise Exception('Command ' + name + ' is already defined')
        # add the command in the list
        self.__commands[name] = command
        # here it is
        return self


class _WcliCompleter(object):
    """
    Command line completion for WildCruncher console.
    """

    def __init__(self, commands, prompt):
        """
        :param commands: Dictionary of available commands
        :type commands: dict[str, WcliCommand]
        :param prompt: Command line prompt
        :type prompt: str
        """
        #: Available commands
        self.__commands = commands
        #: Command line prompt
        self.__prompt = prompt

        readline.set_completer_delims(' \t\n;')
        # Turn on tab completion for readline-based systems
        readline.parse_and_bind('tab: complete')
        # Turn on tab completion for libedit-based systems
        #readline.parse_and_bind("bind '\t' rl_complete")
        # Complete on first tab
        readline.parse_and_bind('set show-all-if-ambiguous on')

    def complete(self, text, state):
        """
        Complete a partial command line token `text` and return the `state`th
        completion.

        :param text: Partial token to complete
        :type text: str
        :param state: Index of requested completion (0, 1, 2, ...)
        :type state: int
        :return: `State`th completion, or `None` if no more completions are
                 available
        :rtype: str
        """
        comps = self._getCompletions(text, readline.get_line_buffer())

        return comps[state] if state < len(comps) else None

    @memoize1
    def _getCompletions(self, text, line):
        """
        Complete a partial command line token `text`.

        Since function is called once for every possible completion for `text`,
        and therefore memoized.

        :param text: Partial token to complete
        :type text: str
        :param line: Current command line content
        :type line: str
        :return: Possible completions for `text`
        :rtype: list[str]
        """
        try:
            line = line[:readline.get_begidx()]
            lparts = shlex.split(line)
        except ValueError:
            return None

        if len(lparts) == 0:
            # Partial command name on command line, give cmd name completions
            comps = [c for c in self._commandNames if c.startswith(text)]
        else:
            # Past command name, either options, option args or parameters
            cmd = self._getCommand(lparts[0])
            if cmd is None:
                # Unknown command, don't know how to complete arguments
                return None
            if text.startswith('--'):
                # Give long option completions
                comps = [opt.longCode for opt in cmd.options
                         if opt.longCode.startswith(text)]
            elif text.startswith('-'):
                # Give short option completions
                comps = [opt.shortCode for opt in cmd.options
                         if opt.shortCode.startswith(text)]
            elif lparts[-1].startswith('-') and \
                            cmd.getOption(lparts[-1]) is not None and \
                    cmd.getOption(lparts[-1]).requiresParameter:
                # Attempt option parameter completion
                comps = cmd.getOption(lparts[-1]).getCompletions(text)
            else:
                paramIdx = 0
                if len(lparts) > 1:
                    parts = list(reversed(lparts[1:]))
                    while parts:
                        if parts[0].startswith('-'):
                            break
                        if len(parts) > 1 and parts[1].startswith('-'):
                            opt = cmd.getOption(parts[1])
                            if opt is None:
                                return []
                            elif opt.requiresParameter:
                                break
                        paramIdx += 1
                        parts = parts[1:] if len(parts) > 1 else []
                comps = cmd.parameters[paramIdx].getCompletions(text)

        return comps

    def displayCompletions(self, subst, matches, maxlen):
        """
        Display the given possible completions to the user.

        :param subst: Text to be completed
        :type subst: str
        :param matches: Matches found by :py:meth:complete`
        :type matches: list[str]
        :param maxlen: Length of longest match
        :type maxlen: int
        """
        # Get current command line content as far as user has entered it
        line = readline.get_line_buffer()[0:readline.get_endidx()]

        try:
            lparts = shlex.split(line)
        except ValueError:
            return None

        lprtsLen = len(lparts)
        completed = False
        if lprtsLen == 0 or (lprtsLen == 1 and not line.endswith(' ')):
            print()
            self._printCommandCompletions(matches, maxlen)
            completed = True
        elif lprtsLen > 1 and subst.startswith('-'):
            cmd = self.__commands.get(lparts[0], None)
            if cmd is None: return
            print()
            self._printOptionCompletions(cmd, matches, maxlen)
            completed = True
        elif len(matches) > 0:
            print('\n' + ' '.join(matches))
            completed = True
        if completed:
            self._redisplayPrompt(line)

    def _printCommandCompletions(self, matches, maxlen):
        """
        Print the given possible command completions.

        :param matches: Matching command names found by :py:meth:complete`
        :type matches: list[str]
        :param maxlen: Length of longest match
        :type maxlen: int
        """
        for c in matches:
            cmd = self.__commands.get(c, None)
            if cmd is None: continue
            hlpm = cmd.helpMessage if cmd.helpMessage else ''
            print('    {c:<{l}}    {h}'.format(c=c, l=maxlen, h=hlpm))

    def _printOptionCompletions(self, cmd, matches, maxlen):
        """
        Print the given possible option completions.

        :param cmd: Command to print options for
        :type cmd: WcliCommand
        :param matches: Matching options found by :py:meth:complete`
        :type matches: list[str]
        :param maxlen: Length of longest match
        :type maxlen: int
        """
        for m in matches:
            opt = cmd.getOption(m)
            if opt is None: continue
            hlpm = opt.helpMessage if opt.helpMessage else ''
            print('    {c:<{l}}    {h}'.format(c=m, l=maxlen, h=hlpm))

    def _redisplayPrompt(self, line):
        """
        Display the prompt with a given initial content.

        :param line: Command line content
        :type line: str
        """
        print(self.__prompt, line, end='', sep='')
        sys.stdout.flush()
        readline.redisplay()

    @property
    def _commandNames(self):
        """
        Get a list of the names of all available commands.

        :return: Names of available commands
        :rtype: list[str]
        """
        return self.__commands.keys()

    def _getCommand(self, name):
        """
        Get the command with the given name.

        :param name: Name of the command to retrieve
        :type name: str
        :return: Command for the given name
        :rtype: WcliCommand
        """
        return self.__commands.get(name, None)


class _WcliParser(object):
    """
    Command line parser.

    Parses a line of user input into command, options and parameters.
    """

    def __init__(self, commands):
        """
        :param commands: Available commands
        :type commands: dict[str, WcliCommand]
        """
        #: Available commands
        self.__commands = commands

    def parse(self, line):
        """
        Parse a line of input and return the contained command, options and
        parameters.

        :param line: the line to parse
        :type line: str
        :return: Command, options and parameters
        :rtype: (WcliCommand, dict[str, object], list[object])
        """
        # check for none
        if not line: return None, None, None
        # split the line
        cl = line.split(' ', 1)
        # loop in the command list
        cmd = self._getCommand(cl[0])
        if cmd is not None:
            opts = {}
            params = []
            if len(cl) > 1:
                opts, params = self._parseArguments(cl[1], cmd)
            elif cmd.requiredParameterCount > 0:
                raise _WcliException('%s requires parameter(s)' % cmd.name)
            return cmd, opts, params
        else:
            raise _WcliException('No such command: \'' + cl[0] + '\'')

    def _parseArguments(self, args, command):
        """
        Parse an argument string into a dict of options and a list of
        parameters.

        :param args: The arguments
        :type args: str
        :param command: Command to parse options for
        :type command: WcliCommand
        :return: The parsed options and parameters
        :rtype: (dict[str, object], list[object])
        """
        try:
            args = shlex.split(args)
        except ValueError:
            raise _WcliException('Could not parse argument list')
        opts, args = self._parseOptions(args, command)
        params = self._parseParameters(args, command)

        return opts, params

    def _parseOptions(self, args, command):
        """
        Parse the options contained in a list of arguments.

        :param args: List of arguments
        :type args: [str]
        :return: Command to parse options for
        :rtype: WcliCommand
        :return: The parsed options and any arguments that have not been parsed
        :rtype: (dict[str, object], list[str])
        """
        opts = {}
        while args and args[0][0] == '-':
            code, *args = args
            opt = command.getOption(code)
            if opt is None:
                raise _WcliException('Unknown option: ' + code)
            arg = None
            if opt.requiresParameter:
                if not args or args[0][0] == '-':
                    raise _WcliException('Option ' + code +
                                         ' requires an argument')
                raw, *args = args
                try:
                    arg = opt.parseArgument(raw)
                except ValueError:
                    raise _WcliException(
                        'Option %s requires a(n) %s argument' %
                        (code, opt.typeName))
            opts[opt.name] = arg

        return opts, args

    def _parseParameters(self, args, command:WcliCommand):
        """
        Parse the parameters contained in a list of arguments.

        :param args: List of arguments
        :type args: list[str]
        :param command: Command to parse parameters for
        :type command: WcliCommand
        :return: The parsed parameters in the correct order
        :rtype: list[object]
        """
        params = []
        paramCount = len(command.parameters)
        if paramCount > 0:
            reqParamCount = command.requiredParameterCount
            if len(args) < reqParamCount:
                paramCount = len(command.parameters)
                if paramCount == reqParamCount:
                    msg = '%s requires %d parameter(s)'
                else:
                    msg = '%s requires at least %d parameter(s)'
                raise _WcliException(msg % (command.name, reqParamCount))
            for param, value in zip(command.parameters, args):
                try:
                    params.append(param.parseArgument(value))
                except ValueError:
                    raise _WcliException('Parameter %d must be a(n) %s' %
                                         (command.parameters.index(param),
                                          param.parameterType.name))
        elif args:
            raise _WcliException(command.name + ' does not take parameters')

        return params

    def _getCommand(self, name):
        """
        Get the command with the given name.

        :param name: Name of the command to retrieve
        :type name: str
        :return: Command for the given name
        :rtype: WcliCommand
        """
        return self.__commands.get(name, None)


class _WcliHelpCommand(WcliCommand):
    """
    Console command to display the console help information.

    Displays a list of all available commands if invoked without arguments,
    or the help information for the command given as argument.
    """

    def __init__(self, cmds):
        """
        :param cmds: Available commands
        :type cmds: dict[str, WcliCommand]
        """
        super().__init__('help', 'Display this help text')

        self.setOptionalParameters(WcliParameter('command', None,
                                                 WCLI_PTYPES['string'],
                                                 'Show help for command'))

        #: Available commands
        self.__commands = cmds

    def execute(self, options, parameters):
        if parameters:
            for p in parameters:
                cmd = self._getCommand(p)
                if cmd is not None:
                    print(self._getCommandOneLiner(cmd))
                    if cmd.helpMessage:
                        print(cmd.helpMessage)
                    if cmd.options:
                        print('  Options:')
                        self._printOptions(cmd.options)
                    if cmd.requiredParameters:
                        print('  Required Parameters:')
                        self._printParameters(cmd.requiredParameters)
                    if cmd.optionalParameters:
                        print('  Optional Parameters:')
                        self._printParameters(cmd.optionalParameters)
                else:
                    print('No such command: ' + p)
        else:
            for cmd in self._commands:
                print('   %s: %s' % (cmd.name, cmd.helpMessage))

    def _getCommand(self, name):
        """
        Get the command with the given name.

        :param name: Name of the command to retrieve
        :type name: str
        :return: Command for the given name
        :rtype: WcliCommand
        """
        return self.__commands.get(name, None)

    @property
    def _commands(self):
        """
        Get a list of all available commands.

        :return: Available commands
        :rtype: list[WcliCommand]
        """
        return self.__commands.values()

    def _getCommandOneLiner(self, cmd):
        """
        Get a one-line usage hint for the given command.

        :param cmd: Command to get usage hint for
        :type cmd: WcliCommand
        :return: Usage hint for the given command
        :rtype: str
        """
        options = self._getOptionsOneLiner(cmd)
        params = self._getParametersOneLiner(cmd)

        s = cmd.name
        if options: s += ' ' + options
        if params: s += ' ' + params

        return s

    def _getOptionsOneLiner(self, cmd):
        """
        Get a one-line usage hint for the given command's options.

        :param cmd: Command to get usage hint for
        :type cmd: WcliCommand
        :return: Usage hint for the given command's options
        :rtype: str
        """

        def fmtopt(o):
            s = '['
            if o.shortCode:
                s += o.shortCode
            if o.longCode:
                if len(s) > 1: s += '|'
                s += o.longCode
            if o.requiresParameter: s += ' ' + o.typeName
            s += ']'
            return s

        return ' '.join([fmtopt(o) for o in cmd.options])

    def _getParametersOneLiner(self, cmd):
        """
        Get a one-line usage hint for the given command's parameters.

        :param cmd: Command to get usage hint for
        :type cmd: WcliCommand
        :return: Usage hint for the given command's parameters
        :rtype: str
        """
        params = [p.name for p in cmd.requiredParameters] + \
                 ['[' + p.name + ']' for p in cmd.optionalParameters]
        return ' '.join(params)

    def _printOptions(self, options):
        """
        Print the help information for the given options.

        :param options: Options to print help information for
        :type options: list[WcliOption]
        """
        maxlen = max([len(o.longCode) for o in options if o.longCode])
        for opt in options:
            scod = opt.shortCode if opt.shortCode else ''
            lcod = opt.longCode if opt.longCode else ''
            hlpm = opt.helpMessage if opt.helpMessage else ''
            print('    {s:<2}  {c:<{l}}    {h}'.format(
                s=scod, c=lcod, l=maxlen, h=hlpm))

    def _printParameters(self, parameters):
        """
        Print the help information for the given parameters.

        :param parameters: Parameters to print help information for
        :type parameters: list[WcliParameter]
        """
        maxlen = max([len(p.name) for p in parameters if p.name])
        for p in parameters:
            name = p.name if p.name else ''
            hlpm = p.helpMessage if p.helpMessage else ''
            print('    {n:<{l}}    {h}'.format(
                n=name, h=hlpm, l=maxlen))


class _WcliExitCommand(WcliCommand):
    """
    Console command to quit the console.
    """

    def __init__(self, exitCallback):
        """
        :param exitCallback: Callback function to exit the console application
        :type exitCallback: () -> None
        """
        super().__init__('exit', 'Exit the Wild Cruncher 2.0 (TM)'
                                 ' command line')

        #: Exit callback function
        self.__exitCallback = exitCallback

    def execute(self, options, parameters):
        self.__exitCallback()
