# ----------------------------------------------------------------------------
# - wgui-button                                                              -
# - wild cruncher - interactive scene elements                               -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 clemens oertel                                   -
# - copyright (c) 2013-2013 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_button
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher interactive scene elements

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt4.QtCore import Qt, QPointF
from PyQt4.QtGui import QGraphicsItem, QGraphicsProxyWidget, QPushButton, \
    QGraphicsRectItem, QPainter, QPainterPath, QPainterPathStroker, QMatrix

from wcr_frontend.gui.wgui_scene_item import WguiSceneItemP
from wcr_frontend.gui.wgui_protocols import QGraphicsItem_Meta_M, \
    WguiZoomAwareP
from wcr_frontend.gui.wgui_utils import WGUI_SELECTION_COLOUR, wgui_snap_coord


class WguiButton(QGraphicsItem, WguiSceneItemP, WguiZoomAwareP,
                 metaclass=QGraphicsItem_Meta_M):
    """
    UI element for action buttons.
    """

    #: Width of highlighting outline
    HIGHLIGHT_WIDTH = 3

    def __init__(self, button, actions, actx, parent=None):
        """
        :param button: Button to be represented
        :type button: WhdbButtonP
        :param actions: Action dictionary
        :type actions: dict[str, WactActionP]
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Button to be represented
        self.__button = button
        #: Action dictionary
        self.__actions = actions
        #: Action execution context
        self.__actx = actx

        #: Proxy element
        self.__proxy = QGraphicsProxyWidget(self)
        #: Button widget
        self.__qbutton = QPushButton()
        self.__qbutton.setStyleSheet('QPushButton {'
                                     ' border: 2px solid gray;'
                                     ' border-radius: 5px;'
                                     ' border-style: outset;'
                                     ' padding: 3px; }'
                                     'QPushButton:pressed {'
                                     ' border-style: inset; }')
        self.__qbutton.clicked.connect(self._clicked)

        self.__proxy.setMinimumWidth(1)
        self.__proxy.setAutoFillBackground(False)
        self.__proxy.setWidget(self.__qbutton)
        self.setZ(self.__button.getZ())

        #: Handle for moving
        self.__rect = _WguiHandle(-10, -10, 10, 10, self)
        self.__rect.setVisible(False)
        self.__qbutton.setAutoFillBackground(False)

        self.__outlineStroke = None
        self.__minButtonWidth = self.__qbutton.width()
        self.__zoom = 1.0

        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges)
        self.setAcceptsHoverEvents(True)

        self.updateItem()

    def zoomChanged(self, zoom):
        self.__zoom = zoom
        self.__outlineStroke = None
        self.__rect.setMatrix(QMatrix().scale(1.0 / zoom, 1.0 / zoom))

    def updateItem(self):
        self.prepareGeometryChange()

        btn = self.__button
        qbtn = self.__qbutton
        qbtn.setText(btn.title)
        qbtn.setMinimumWidth(max(qbtn.minimumWidth(), self.__minButtonWidth))
        self.setPos(*btn.getPosition())
        qbtn.adjustSize()

        if btn.command:
            action = self.__actions.get(btn.command)
            qbtn.setEnabled(action.isEnabled(self.__actx))
        else:
            qbtn.setEnabled(False)

        self.updateBoundingBox()
        self.__outlineStroke = None

    def updateBoundingBox(self):
        """
        Persist the button's current bounding box.
        """
        brect = self.mapRectToScene(self.__proxy.boundingRect())
        self.__button.setBoundingBox(brect.x(), brect.y(),
                                     brect.width(), brect.height())

    def getZ(self):
        return self.zValue()

    def setZValue(self, zval):
        self.setZ(zval)

    def setZ(self, zval):
        QGraphicsItem.setZValue(self, zval)
        self.__button.setZ(zval)

    def getDataObject(self):
        return self.__button

    def mousePressEvent(self, event):
        QGraphicsItem.mousePressEvent(self, event)

    def _clicked(self):
        """
        Callback function for button clicks.
        """
        command = self.__button.command
        action = self.__actions.get(command)

        if action is not None:
            action.execute(self.__actx, **self.__button.parameters)

    def hoverEnterEvent(self, event):
        super().hoverEnterEvent(event)
        self.__rect.setVisible(True)

    def hoverLeaveEvent(self, event):
        super().hoverLeaveEvent(event)
        self.__rect.setVisible(False)

    def paint(self, painter:QPainter, options, widget=None):
        if self.isSelected():
            painter.fillPath(self._getOutlineStroke(),
                             WGUI_SELECTION_COLOUR)

    def boundingRect(self):
        adj = 5 / self.__zoom
        return self.__proxy.boundingRect() \
            .adjusted(-adj, -adj, adj, adj) \
            .united(self.__rect.boundingRect())

    def shape(self):
        path = QPainterPath()
        path.addRect(self.__proxy.boundingRect())
        path.addRect(self.__rect.boundingRect())
        return path

    def _getOutlineStroke(self):
        """
        Get the outline stroke of the button for rendering highlighting.
        The calculated stroke will be stored for future use.

        :return: Outline stroke
        :rtype: QPainterPath
        """
        outlineStroke = self.__outlineStroke
        if outlineStroke is not None:
            return outlineStroke

        path = QPainterPath()
        path.addRect(self.__proxy.boundingRect())

        ps = QPainterPathStroker()
        ps.setCapStyle(Qt.RoundCap)
        ps.setJoinStyle(Qt.RoundJoin)
        if self.__zoom != 1:
            ps.setWidth((2 * self.HIGHLIGHT_WIDTH + 1) / self.__zoom)
        else:
            ps.setWidth(2 * self.HIGHLIGHT_WIDTH + 1)
        outlineStroke = ps.createStroke(path)
        outlineStroke = outlineStroke.subtracted(path)
        self.__outlineStroke = outlineStroke
        return outlineStroke

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            # get the graphic scene and snap flag
            scene = self.scene()
            if scene is None: return value
            sflg = scene.getSnapFlag()
            xsnp = scene.getSnapX()
            ysnp = scene.getSnapY()
            # set position value
            x = value.x()
            y = value.y()
            if sflg:
                x = wgui_snap_coord(x, xsnp)
                y = wgui_snap_coord(y, ysnp)
                # get the data object as a function
            return QPointF(x, y)
        return value


class _WguiHandle(QGraphicsRectItem):
    """
    Rectangular handle.
    """

    def __init__(self, *__args):
        super().__init__(*__args)

    def paint(self, painter:QPainter, options, widget=None):
        r = self.rect()
        painter.fillRect(r.x(), r.y(), r.width(), r.height(), Qt.lightGray)
