# ----------------------------------------------------------------------------
# - wgui-button-props-dialog                                                 -
# - wild cruncher - action button properties widget                          -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 clemens oertel                                   -
# - copyright (c) 2013-2013 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_button_props_dialog
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher action button properties widget

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt4.QtGui import QLineEdit, QVBoxLayout, QLabel, QComboBox, QMessageBox

from wcr_core.action.wact_action_parameter_api import \
    WactFixedSelectionActionParameterP
from wcr_core.action.wact_action_parameter_type_api import \
    WactStringableActionParameterTypeP
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesDialog, \
    WguiPropertiesWidgetP
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from silkan_common.scmn_log import SCMN_LOG_WARN


class WguiButtonPropertiesDialog(WguiPropertiesDialog):
    """
    The WguiButtonPropertiesDialog class is a dialog design to access
    and manipulates action button data.
    """

    def __init__(self, button, actions, xctx, updateF, parent=None):
        """
        :param button: Button to display/modify
        :type button: WhdbButtonP
        :param actions: Action dictionary
        :type actions: dict[str, WactActionP]
        :param xctx: Context
        :type xctx: WcorContextP
        :param updateF: Callback function to notify about saves
        :type updateF: _ -> _
        :param parent: Parent widget
        """
        super().__init__('Button Properties', updateF, parent=parent)

        self.setPropertyWidget(
            _WguiButtonProperties(button, actions, xctx))


class _WguiButtonProperties(WguiPropertiesWidgetP):
    """
    Property editing widget for action buttons.
    """

    #: Actions available for selection
    __AVAILABLE_ACTIONS = ['main.open',
                           'editor.save',
                           'tabw.activate',
                           'project.build',
                           'project.generate',
                           'project.comp',
                           'project.clean',
                           'project.execute']

    def __init__(self, button, actions, xctx, parent=None):
        """
        :param button: Button to display/modify
        :type button: WhdbButtonP
        :param actions: Action dictionary
        :type actions: dict[str, WactActionP]
        :param xctx: Context
        :type xctx: WcorContextP
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Button object
        self.__button = button
        #: Actions dictionary
        self.__actions = actions
        #: Application context
        self.__context = xctx

        #: Label field
        self.__labelF = QLineEdit()
        #: Command field
        self.__commandF = QComboBox()
        #: Layout for command parameters
        self.__paramLayout = None

        self.__commandF.currentIndexChanged.connect(self._actionChanged)

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setMargin(0)
        layout.addWidget(QLabel('Button Label:'))
        layout.addWidget(self.__labelF)
        layout.addWidget(QLabel('Command:'))
        layout.addWidget(self.__commandF)
        self.__paramLayout = QVBoxLayout()
        layout.addLayout(self.__paramLayout)
        layout.addStretch(1)
        return layout

    def updateFields(self):
        cmdCombo = self.__commandF

        cmdCombo.clear()
        btnAction = self.__button.command
        for actCode in self.__AVAILABLE_ACTIONS:
            action = self.__actions.get(actCode)
            if action is not None:
                cmdCombo.addItem(action.getName(), (action, actCode))
                if btnAction == actCode:
                    cmdCombo.setCurrentIndex(self.__commandF.count() - 1)

        button = self.__button
        self.__labelF.setText(button.title)

    def _actionChanged(self):
        """
        Callback function for changes in action selection, to adjust action
        parameter fields.
        """
        layout = self.__paramLayout
        if layout is None:
            return

        for idx in range(layout.count()):
            layout.itemAt(idx).widget().hide()
            layout.itemAt(idx).widget().close()

        parameters = self.__button.parameters
        actionT = wgui_qcb_get_data(self.__commandF)
        if actionT is not None:
            action = actionT[0]
            """:type: WactActionP"""
            for p in action.getParameters():
                if not isinstance(p.getParamType(),
                                  WactStringableActionParameterTypeP):
                    raise Exception(
                        'Cannot create button for action with not-string-'
                        'compatible parameters')
                layout.addWidget(QLabel(p.getName()))
                if isinstance(p, WactFixedSelectionActionParameterP):
                    paramF = QComboBox()
                    paramF.parameter = p
                    value = parameters.get(p.getCode())
                    for v, t, _ in p.getPossibleValues():
                        paramF.addItem(t, v)
                        if value == v:
                            paramF.setCurrentIndex(paramF.count() - 1)
                    layout.addWidget(paramF)
                else:
                    paramF = QLineEdit()
                    paramF.parameter = p
                    value = parameters.get(p.getCode())
                    if value:
                        paramF.setText(value)
                    layout.addWidget(paramF)

            self.__labelF.setPlaceholderText(action.getName())

        self.adjustSize()

    def save(self):
        button = self.__button
        label = self.__labelF.text()
        actionT = wgui_qcb_get_data(self.__commandF)
        if actionT is not None:
            button.command = actionT[1]
            if not label:
                label = actionT[0].name
            params = {}
            for idx in range(self.__paramLayout.count()):
                c = self.__paramLayout.itemAt(idx).widget()
                if not c.isVisible(): continue

                if isinstance(c, QLineEdit):
                    param = c.parameter
                    """:type: WactActionParameterP"""
                    ptyp = param.getParamType()
                    if not isinstance(ptyp,
                                      WactStringableActionParameterTypeP):
                        continue
                    text = c.text()
                    try:
                        ptyp.parse(text)
                        params[param.getCode()] = text
                    except ValueError:
                        name = param.getName() if param is not None else '?'
                        msg = 'The value for parameter \'' + name + '\' ' + \
                              'is invalid and could not be saved.'
                        # noinspection PyCallByClass,PyTypeChecker
                        # noinspection PyArgumentList
                        QMessageBox.warning(
                            self, 'Parameter Value Invalid', msg)
                        self.__context.addMessage(
                            msg, 'cockpit', SCMN_LOG_WARN)
                elif isinstance(c, QComboBox):
                    param = c.parameter
                    params[param.code] = wgui_qcb_get_data(c)
                else:
                    continue

            button.parameters = params
        else:
            button.command = None
            button.parameters = None

        button.title = label
