# ----------------------------------------------------------------------------
# - wgui-clipboard                                                           -
# - wild cruncher: clipboard handler                                         -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_clipboard
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher clipboard handler

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from functools import partial
import traceback
from xml.dom.minidom import Document
from xml.etree import ElementTree

from PyQt4.QtCore import QMimeData, QVariant, QByteArray, QPoint
from PyQt4.QtGui import QApplication

from wcr_core.db.whdb_utils import whdb_ensureUniqueName
from wcr_core.db_api.whdb_button_api import WhdbButtonP
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP
from wcr_core.db_api.whdb_shape_api import WhdbShapeP
from wcr_core.db_api.whdb_connection_api import WhdbConnectionP
from wcr_core.dbstore.wdbs_entity import WdbsEntityBase
from wcr_core.dbstore.wdbs_meta_description import WdbsMetaDescription
from wcr_core.dbstore.wdbs_util import wdbs_makeUUID
from wcr_core.dbstore_api.wdbs_block_api import WdbsContainerP, \
    WdbsEntityTargetP
from wcr_core.modules.hardware.db.whhdb_rack_instance import \
    WhhdbRackInstance
from wcr_core.modules.hardware.db.whhdb_Cetrac_switch_instance import \
    WhhdbCetracSwitchInstance
from wcr_core.modules.hardware.db.whhdb_Host_instance import \
    WhhdbHostInstance
from wcr_core.modules.hardware.db.whhdb_kazymir_switch_instance import \
    WhhdbKazymirSwitchInstance
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_MODULES_IO_NAMES, WHHDB_XPOINT_NAMES
from wcr_core.common.wcmn_file_handler import WcmnReadPayloadHandler, \
    WcmnFileReaderBase
from wcr_core.common.wcmn_mime_type import WCMN_PAYLOAD_BLOCK
from wcr_core.xmlio.wxio_xmlio import wxio_serializeXml, wxio_deserializeXml
from wcr_core.project.wprj_xmlio import WPRJ_SILKAN_XML_MIME_TYPE, \
    WPRJ_SILKAN_XML_MIME_NAME
from wcr_frontend.gui.wgui_scene_item import WguiSceneItemP
from silkan_common.scmn_memoization import memoize1


class WguiClipboardHandler(object):
    """
    Provides copy/paste functionality for a graphics view/scene.
    """

    def __init__(self, scene, module):
        """
        :param scene: Graphics scene to copy selection from/paste to
        :type scene: WguiGraphicsScene
        :param module: Application module
        :type module: WmodApplicationModuleP
        """
        #: Scene to copy selection from/paste to
        self.__scene = scene
        #: Application module
        self.__module = module

        #: View to use as position reference
        self.__centerF = None

    def setCenterF(self, centerF):
        """
        Set the function that provides the current view center.

        :param centerF: View-center function
        :type centerF: _ -> QPoint
        """
        self.__centerF = centerF

    def _getCenter(self):
        """
        Get the function that provides the current view center.

        :return: View-center function
        :rtype: _ -> QPoint
        """
        centerF = self.__centerF
        if centerF is not None:
            return centerF()
        else:
            return QPoint(0, 0)

    def copy(self):
        """
        Copy the scene's current selection to the clipboard.
        """
        self._copyItems(
            [s.getDataObject() for s in self.__scene.selectedItems()
             if isinstance(s, WguiSceneItemP)])

    def copyItems(self, *items):
        """
        Copy the given items to the clipboard.
        """
        self._copyItems(items)

    def _copyItems(self, items):
        """
        Copy the given items to the clipboard.
        """
        if items:
            self._copyBlock(
                _WguiEntityContainer(items[0].rawElement.context,
                                     items))
        else:
            # noinspection PyArgumentList
            QApplication.clipboard().clear()

    def _copyBlock(self, block):
        """
        Copy the given block to the clipboard.
        """
        xml = Document()
        docNode = xml.createElement('selection')
        docNode.setAttribute('module', self.__module.code)

        blockNode = wxio_serializeXml(
            block, xml, 'temp-block', self.__module.code)
        docNode.appendChild(blockNode)

        viewNode = xml.createElement('view')
        center = self._getCenter()
        viewNode.setAttribute('view-center-x', str(center.x()))
        viewNode.setAttribute('view-center-y', str(center.y()))
        docNode.appendChild(viewNode)

        xml.appendChild(docNode)
        # noinspection PyArgumentList
        clipboard = QApplication.clipboard()
        clipboard.setMimeData(_WguiCnPMime(xml.toprettyxml()))

    def canPaste(self):
        """
        Whether the current content of the clipboard can be pasted into the
        scene.

        :return: True iff the current clipboard content can be pasted
        :rtype: bool
        """
        # noinspection PyArgumentList
        clipboard = QApplication.clipboard()
        return _wgui_isValidMimeData(clipboard.mimeData(), self.__module)

    def paste(self, centerPos=None):
        """
        Paste the clipboard contents into the scene.

        :param centerPos: Center position for pasted content
        :type centerPos: QPoint
        """
        # noinspection PyArgumentList
        clipboard = QApplication.clipboard()
        content = clipboard.mimeData()

        if centerPos is None:
            centerPos = self._getCenter()

        if content.hasText():
            payloadType = self.__module.getPayloadType()
            WguiSilkanXmlReader(payloadType, self.__module).readText(
                content.text(),
                WguiSilkanPasteHandler(self.__scene, centerPos),
                None)


class WguiSilkanPasteHandler(WcmnReadPayloadHandler):
    """
    Import payload handler to add block entities to the current scene.

    Note: This payload handler is not necessarily limited to clipboard
    operations.
    """

    def __init__(self, scene, centerPos):
        """
        :param scene: Graphic scene to add elements to
        :type scene: WguiGraphicsScene
        :param centerPos: Center position for pasted content
        :type centerPos: QPoint
        """
        #: Graphics scene
        self.__scene = scene
        #: Graphics view
        self.__centerPos = centerPos

        #: Current model block in scene
        self.__top = scene.topLevelModelBlock

    def handledPayloadTypes(self):
        return WCMN_PAYLOAD_BLOCK

    def handlePayload(self, payloadType, mimeType, payloadFunc):
        if payloadType & WCMN_PAYLOAD_BLOCK:
            try:
                # noinspection PyUnresolvedReferences
                tempBlock = _WguiPasteEntityTarget(
                    self.__top.rawElement.context, self.__top)
                payloadFunc(tempBlock)

                self._recenter(tempBlock)
                self._addEntities(tempBlock)
            except:
                traceback.print_exc()

    # noinspection PyUnresolvedReferences
    def _recenter(self, container):
        """
        Adjust the position for all elements in the given container so that
        their cumulative bounding box is centered in the current view.

        :param container: Element container
        :type container: _WguiPasteEntityTarget
        """
        brect = container.boundingBox
        if brect is None:
            return

        dx = self.__centerPos.x() - brect.x - brect.width / 2
        dy = self.__centerPos.y() - brect.y - brect.height / 2

        for inst in container.instances:
            x, y = inst.getPosition()
            inst.setPosition(x + dx, y + dy)
        for func in container.functions:
            x, y = func.getPosition()
            func.setPosition(x + dx, y + dy)
        for cnctn in container.connections:
            cnctn.getPath().translate(dx, dy)
        for shape in container.shapes:
            x, y = shape.location
            shape.location = (x + dx, y + dy)
        for element in container.interacts:
            x, y = element.getPosition()
            element.location = (x + dx, y + dy)

    # noinspection PyTypeChecker
    def _addEntities(self, container):
        """
        Add the entities from the given container to the associated scene.

        :param container: Element container
        :type container: _WguiPasteEntityTarget
        """
        self.__scene.clearSelection()

        added = []

        for inst in container.instances:
            added.append(self.__scene.addInstanceItem(inst))
        for func in container.functions:
            if (isinstance(func, WhhdbHostInstance) 
                    or isinstance(func, WhhdbCetracSwitchInstance)):
                for xpnt in func.getConnectionPoints():
                    if xpnt.getCode() in WHHDB_XPOINT_NAMES and not isinstance(func, WhhdbCetracSwitchInstance):
                        func.addPortUUIDToCard(xpnt.getCard(), xpnt.getUUID())
                    elif xpnt.getCode() in WHHDB_MODULES_IO_NAMES:
                        ioModule = xpnt.getIoCard()
                        func.addPortUUIDToIoCard(ioModule, xpnt.getUUID())
            elif isinstance(func, WhhdbKazymirSwitchInstance):
                for xpnt in func.getConnectionPoints():
                    if xpnt.getCode() in WHHDB_MODULES_IO_NAMES:
                        ioModule = xpnt.getIoCard()
                        func.addPortUUIDToIoCard(ioModule, xpnt.getUUID())

            added.append(self.__scene.addFunctionItem(func))
        for cnctn in container.connections:
            added.append(self.__scene.addConnectionItem(cnctn))
        for shape in container.shapes:
            added.append(self.__scene.addShapeItem(shape))
        for element in container.interacts:
            added.append(self.__scene.addActionButton(element))

        for z, item in enumerate(sorted(added, key=lambda i: i.getZ()),
                                 self.__scene.getTopZValue() + 1):
            item.setSelected(True)
            item.setZ(z)


class WguiSilkanXmlReader(WcmnFileReaderBase):
    """
    Import reader for Silkan-XML clipboard content.

    Todo: Consider merging with the general Silkan XML reader.
    """

    def __init__(self, payloadType, module):
        """
        :param payloadType: Handled payload type
        :type payloadType: int
        :param module: Application module to check for
        :type module: WmodApplicationModuleP
        """
        super().__init__([WPRJ_SILKAN_XML_MIME_TYPE], payloadType)

        self.__payloadType = payloadType
        self.__module = module

    @property
    def name(self):
        return 'Silkan XML'

    def readText(self, text, callback, xctx, types=None):
        if not _wgui_isValidClipboardText(text, self.__module):
            return
        tree = ElementTree.fromstring(text)

        callback.handlePayload(self.__payloadType,
                               WPRJ_SILKAN_XML_MIME_NAME,
                               partial(self._readText, tree))

    def _readText(self, tree, block):
        wxio_deserializeXml(
            tree, 'temp-block', block, ignoreUUID=True)


class _WguiPasteEntityTarget(WdbsEntityBase, WdbsEntityTargetP):
    """
    Entity target for pasting.
    """

    def __init__(self, dctx, target):
        """
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param target: Paste target
        :type target: WhdbBlockP
        """
        super().__init__(dctx)

        self.__target = target
        self.__wrapperRegistry = target.wrapperRegistry
        self.__rawTarget = target.rawElement

        self.__uuid = wdbs_makeUUID()

        #: Added instances
        self.__instances = []
        #: Added functions
        self.__functions = []
        #: Added connections
        self.__connections = []
        #: Added shapes
        self.__shapes = []
        #: Added interactive elements
        self.__interacts = []

    def getUUID(self):
        return self.__uuid

    @property
    def metaDescription(self):
        return WdbsMetaDescription()

    def getMeta(self, key):
        return None

    def setMeta(self, value, *keys):
        pass

    def removeMeta(self, *keys):
        pass

    def getMetaCount(self, *keys):
        return 0

    def getMetaKeys(self, *keys):
        return []

    def getMetas(self, *keys):
        return []

    def getAllMeta(self):
        return {}

    def createBlockInstance(self, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        inst = self.__rawTarget.createBlockInstance(*args, **kwargs)
        self.__instances.append(inst)
        return inst

    def createModelInstance(self, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        inst = self.__rawTarget.createModelInstance(*args, **kwargs)
        self.__instances.append(inst)
        return inst

    def createOperatorInstance(self, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        inst = self.__rawTarget.createOperatorInstance(*args, **kwargs)
        self.__functions.append(inst)
        return inst

    def createShape(self, styp, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        shape = self.__rawTarget.createShape(styp, *args, **kwargs)
        self.__shapes.append(shape)
        return shape

    def createConnection(self, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        cnctn = self.__rawTarget.createConnection(*args, **kwargs)
        self.__connections.append(cnctn)
        return cnctn

    def createButton(self, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        btn = self.__rawTarget.createButton(*args, **kwargs)
        self.__interacts.append(btn)
        return btn

    @property
    @memoize1
    def instances(self):
        """
        :rtype: list[WhdbInstanceP]
        """
        objs = [self.__wrapperRegistry.wrap(i) for i in self.__instances]
        for obj in objs:
            # noinspection PyTypeChecker
            whdb_ensureUniqueName(obj, self.__target)
        return objs

    @property
    @memoize1
    def functions(self):
        """
        :rtype: list[WhdbOperatorInstanceP]
        """
        objs = [self.__wrapperRegistry.wrap(i) for i in self.__functions]
        for obj in objs:
            # noinspection PyTypeChecker
            whdb_ensureUniqueName(obj, self.__target)
        return objs

    @property
    @memoize1
    def connections(self):
        """
        :rtype: list[WhdbConnectionP]
        """
        return [self.__wrapperRegistry.wrap(i) for i in self.__connections]

    @property
    @memoize1
    def shapes(self):
        """
        :rtype: list[WhdbShapeP]
        """
        return [self.__wrapperRegistry.wrap(i) for i in self.__shapes]

    @property
    @memoize1
    def interacts(self):
        """
        :rtype: list[WhdbButtonP]
        """
        return [self.__wrapperRegistry.wrap(i) for i in self.__interacts]

    # noinspection PyUnresolvedReferences
    @property
    @memoize1
    def boundingBox(self):
        def grow(rect):
            nonlocal bbox
            if bbox is None:
                bbox = rect
            elif rect is not None:
                bbox = bbox.united(rect)

        bbox = None
        for i in self.instances:
            grow(i.boundingBox)
        for f in self.functions:
            grow(f.boundingBox)
        for s in self.shapes:
            grow(s.boundingBox)
        for i in self.interacts:
            grow(i.boundingBox)
        return bbox


class _WguiEntityContainer(WdbsEntityBase, WdbsContainerP):
    """
    Entity container wrapper for a list of HSB entities.
    """

    def __init__(self, dctx, items):
        """
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param items: Represented items
        :type items: list[WhdbBaseP]
        """
        super().__init__(dctx)

        self.__dctx = dctx
        self.__items = items
        self.__uuid = wdbs_makeUUID()

    def getUUID(self):
        return self.__uuid

    @property
    def metaDescription(self):
        return WdbsMetaDescription()

    def getMeta(self, key):
        return None

    def setMeta(self, value, *keys):
        pass

    def removeMeta(self, *keys):
        pass

    def getMetaCount(self, *keys):
        return 0

    def getMetaKeys(self, *keys):
        return []

    def getMetas(self, *keys):
        return []

    def getAllMeta(self):
        return {}

    def getInstances(self):
        return [s.rawElement for s in self.__items
                if isinstance(s, WhdbInstanceP)]

    def getInstanceCount(self):
        return len(self.getInstances())

    def getConnections(self):
        # noinspection PyUnresolvedReferences
        return [s.rawElement for s in self.__items
                if isinstance(s, WhdbConnectionP) and
                s.getStartObject() in self.__items and
                s.getEndObject() in self.__items]

    def getConnectionCount(self):
        return len(self.getConnections())

    def getShapes(self):
        return [s.rawElement for s in self.__items
                if isinstance(s, WhdbShapeP)]

    def getShapeCount(self):
        return len(self.getShapes())

    def getButtons(self):
        return [s.rawElement for s in self.__items
                if isinstance(s, WhdbButtonP)]

    def getButtonCount(self):
        return len(self.getButtons())


class _WguiCnPMime(QMimeData):
    """
    Mime Data for copy/paste operations.
    """

    # 'application/x-qt-mime-type-name' is required for the Mac system
    # clipboard
    #: Supported mime types
    FORMATS = [WPRJ_SILKAN_XML_MIME_NAME,
               'application/xml', 'text/plain',
               'application/x-qt-mime-type-name']

    def __init__(self, data):
        """
        :param data: Mime data
        :type data: str
        """
        super().__init__()

        self.__data = data

    def formats(self):
        return self.FORMATS

    def hasFormat(self, fmt):
        return fmt in self.FORMATS

    def retrieveData(self, fmt, vType):
        # QVariant.Invalid is required for the Mac system clipboard
        if not self.hasFormat(fmt):
            raise ValueError('Can only provide data as one of ' +
                             ', '.join(self.FORMATS) + '; requested: ' + fmt)
        if vType == QVariant.String or vType == QVariant.Invalid:
            return self.__data
        elif vType == QVariant.ByteArray:
            return QByteArray().append(self.__data)
        else:
            raise ValueError('Cannot provide data as %d' % vType)


def _wgui_isValidMimeData(data, module):
    """
    Whether the MIME data contains valid contents for pasting.

    :param data: MIME data to check
    :type data: QMimeData
    :param module: Application module to check for
    :type module: WmodApplicationModuleP
    :return: True iff the MIME data can be pasted
    :rtype: bool
    """
    return data.hasText() and _wgui_isValidClipboardText(data.text(), module)


def _wgui_isValidClipboardText(text, module):
    """
    Whether the given string contains data that can be pasted.

    :param text: String to check
    :type text: str
    :param module: Application module to check for
    :type module: WmodApplicationModuleP
    :return: True iff the data encoded in the string can be pasted
    :rtype: bool
    """
    return text.startswith(
        '<?xml version="1.0" ?>\n<selection module="{}">'.format(module.code))
