# ----------------------------------------------------------------------------
# - wgui-db-elements-dialog                                                  -
# - wild cruncher - cockpit: element popup base classes                      -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_db_elements_dialog
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher element popup base classes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QComboBox, QDialog, QHBoxLayout, QVBoxLayout, \
	QPushButton, QListWidget, QListWidgetItem, QDialogButtonBox, QLabel


class WguiElementsDialogBase(QDialog):
    """
    Database verification dialog.
    """

    def __init__(self, projectDB, scene, runBtnLbl=None, closeIfOk=False,
                 items=None, parent=None):
        """
        :param projectDB: Project Database
        :type projectDB: WprjProjectDatabase
        :param scene: Graphic scene
        :type scene: WguiGraphicsScene
        :param runBtnLbl: Run button label
        :type runBtnLbl: str
        :param closeIfOk: Whether to close the verification dialog once
                          verification succeeds
        :type closeIfOk: bool
        :param items: Initial elements to display: list of
                      error message, affected elements, path, and add'l data
        :type items: list[(str, list[WdbsEntityP], list[WdbsBlockInstanceP],
                           object)]
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: The current project database
        self.__projectDB = projectDB
        #: The graphics scene
        self.__scene = scene
        #: Whether to close dialog once verification succeeds
        self.__closeIfOk = closeIfOk

        self.__resultList = QListWidget()
        if runBtnLbl:
            #: Verify button
            self.__runBtn = QPushButton(runBtnLbl)
            self.__runBtn.clicked.connect(self._run)
        else:
            self.__runBtn=None
        #: Close button
        self.__closeBtn = QPushButton('Close')
        self.__closeBtn.clicked.connect(self.accept)

        self.setLayout(self._createLayout())

        self.setMinimumWidth(400)

        if items:
            self._setItems(items)

    def _createLayout(self):
        """
        Create layout.
        Should be called only once.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()

        if self.__runBtn:
            layout.addWidget(self.__runBtn)

        self.__resultList.itemClicked.connect(self._itemClicked)
        layout.addWidget(self.__resultList)
        layout.setStretch(1, 1)

        hbox = QHBoxLayout()
        hbox.addStretch(1)
        hbox.addWidget(self.__closeBtn)
        layout.addLayout(hbox)

        return layout

    def _cleanUp(self):
        self.__scene.update()

    def _run(self):
        self._setItems(self._doRun())

    def _doRun(self):
        raise NotImplementedError()

    def accept(self):
        self._cleanUp()
        super().accept()

    def reject(self):
        self._cleanUp()
        super().reject()

    def _setItems(self, items):
        self.__resultList.clear()

        if not items and self.__closeIfOk:
            self.accept()

        for msg, items, path, data in items:
            listItem = QListWidgetItem(msg)
            listItem.setData(Qt.UserRole, (items, path, data))
            self.__resultList.addItem(listItem)

    def _itemClicked(self, item):
        qdat = item.data(Qt.UserRole)
        if not qdat:
            self.__scene.clearSelection()
            return

        (items, path, data) = qdat

        self.__scene.setHierarchy(path)
        self.__scene.clearSelection()

        for entity in items:
            item = self.__scene.getSceneItem(entity)
            if item is not None:
                # noinspection PyUnresolvedReferences
                item.setSelected(True)
                # noinspection PyUnresolvedReferences
                item.ensureVisible()

        self.__scene.update()


class WguiComboSelectionAsk(QDialog):
    def __init__(self, datas, formatF, parent=None, mess=None):
        """
        :param datas: datas in with the user chose
        :type datas: Iterable[Object(data)]
        :param formatF: Format function for presenting datas in combobox
        :type formatF: Function(data)->str
        :param parent: Parent widget
        :param mess: The message to display over the combobox
        :type mess: str
        """

        super().__init__(parent=parent)
        self.__msg = mess
        self.__parent = parent
        self.__selection=None

        #: Combobox for selection
        self.__cbbox = QComboBox()

        self.setLayout(self._createLayout())
        self._initElements(datas, formatF)

    def _createLayout(self):
        """
        Create the layout.

        :return: Layout
        :rtype: QLayout
        """

        layout = QVBoxLayout()

        if self.__msg:
            layout.addWidget(QLabel(self.__msg))
        layout.addWidget(self.__cbbox)

        buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

        layout.addWidget(buttons)

        return layout

    def _initElements(self, datas, formatF):
        """
        Initialize the UI elements, as needed.

        :param datas: Datas for selection
        :type datas: Iterable[object]
        :param formatF: Format function for presenting datas in combobox
        :type formatF: Function(data)->str
        """
        self.__cbbox.addItem('',None)
        for elem in datas:
            self.__cbbox.addItem(formatF(elem), elem)

    def selection(self):
        """
        return the data selected when accepted button was presed
        """
        return self.__selection

    def accept(self):
        self.__selection = self.__cbbox.itemData(self.__cbbox.currentIndex())

        super().accept()