# ----------------------------------------------------------------------------
# - wgui-db-tree                                                             -
# - wild cruncher: database tree display                                     -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_db_tree
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher database tree display

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from _weakref import ref
from itertools import chain

from PyQt4.QtCore import Qt
from PyQt4.QtGui import QFont, QIcon

from wcr_core.db_api.whdb_button_api import WhdbButtonP
from wcr_core.db_api.whdb_connection_api import WhdbConnectionP
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP, \
    WhdbModelInstanceP, WhdbOperatorInstanceP
from wcr_frontend.gui.wgui_toolbox_item import WGUI_ICON_PNG32
from wcr_frontend.gui.wgui_scene_item import WguiSceneItemP
from wcr_frontend.gui.wgui_tree_model import WguiTreeItemBase, \
    WguiTreeItemFactoryP, WguiFilterModel
from wcr_frontend.gui.wgui_tree import WguiTreeModelManager, \
    WGUI_TREE_ITEM_ROLE, WGUI_TREE_PATH_ROLE, WguiTree, WGUI_TREE_NOLEAF_ROLE
from wcr_core.db.whdb_connection_utils import whdb_getConnectionDescription
from wcr_core.db_api.whdb_shape_api import *


#: Number of columns in standard DB tree
WGUI_DB_TREE_COLUMNS = 2


class WguiDbGroupBlockItem(WguiTreeItemBase):
    """
    Database-tree item for grouping entities (e.g. "Operator Instances").
    """

    _FONT = None

    @staticmethod
    def _getFont():
        """
        Get the font to be used for grouping items.

        :return: Font to be used for grouping items
        :rtype: QFont
        """
        if WguiDbGroupBlockItem._FONT is None:
            font = QFont()
            font.setItalic(True)
            WguiDbGroupBlockItem._FONT = font

        return WguiDbGroupBlockItem._FONT

    def __init__(self, name, parentItem):
        """
        :param name: Group name
        :type name: str
        :param parentItem: Parent tree item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(parentItem)

        #: Group name
        self.__name = name

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            if column == 0:
                return self.__name
        elif role == Qt.FontRole:
            return WguiDbGroupBlockItem._getFont()
        elif role == WGUI_TREE_ITEM_ROLE:
            return 'GROUP ' + self.__name
        elif role == WGUI_TREE_NOLEAF_ROLE:
            return True

    def setData(self, data, column, role=None):
        raise NotImplementedError()

    def getColumnCount(self):
        return WGUI_DB_TREE_COLUMNS

    @property
    def flags(self):
        return Qt.ItemIsEnabled


_WGUI_BLOCK_GROUPS = [
    'Block-Instances', 'Model-Instances', 'Equipments', 'Connections',
    'Shapes']


class WguiDbTreeRootItem(WguiTreeItemBase):
    """
    Root-item for database trees.
    """

    def __init__(self, parentItem):
        """
        :param parentItem: Parent tree item
        :type parentItem: WguiTreeItemP
        """
        children = [WguiDbGroupBlockItem(s, self)
                    for s in _WGUI_BLOCK_GROUPS]

        super().__init__(parentItem, children)

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            if column == 0:
                return 'Item'
            elif column == 1:
                return 'Type'
            else:
                return 'ROOT'

    def getColumnCount(self):
        return WGUI_DB_TREE_COLUMNS

    @property
    def flags(self):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def addChild(self, child, row=None):
        raise NotImplementedError()

    def removeChild(self, row):
        raise NotImplementedError()

    def removeChildren(self, row, count):
        raise NotImplementedError()

    def setData(self, data, column, role=None):
        raise NotImplementedError()


class WguiDbTreeItemBase(WguiTreeItemBase):
    """
    Abstract base class for entity-items in a database tree.
    """

    def __init__(self, entity, parentItem, icon=None, path=None,
                 children=None):
        """
        :param entity: Represented database entity
        :type entity: WhdbBaseP
        :param parentItem: Parent tree item
        :type parentItem: WguiTreeItemP
        :param icon: Icon
        :type icon: QIcon | None
        :param path: Entity path
        :type path: list[WhdbBlockInstanceP]
        :param children: Pre-set child items
        :type children: list[WguiTreeItemP]
        """
        super().__init__(parentItem, children)

        #: Database entity
        self.__entity = entity
        #: Icon
        self.__icon = icon
        #: Entity path
        self.__path = path

    @property
    def flags(self):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def getColumnCount(self):
        return WGUI_DB_TREE_COLUMNS

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            if column == 0:
                return self._getName()
            elif column == 1:
                return self._getType()
        elif role == Qt.DecorationRole:
            if column == 0:
                return self.__icon
        elif role == WGUI_TREE_ITEM_ROLE:
            return self.__entity
        elif role == WGUI_TREE_PATH_ROLE:
            return self.__path

    def setData(self, data, column, role=None):
        raise NotImplementedError()

    @abstractmethod
    def _getName(self):
        """
        Get the entity's name to be displayed in the name column.

        :return: Entity's name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def _getType(self):
        """
        Get the entity's type description, to be displayed in the type column.

        :return: Entity's type description
        :rtype: str
        """
        raise NotImplementedError


class WguiDbTreeBlockInstanceItem(WguiDbTreeItemBase):
    def __init__(self, instance, parentItem, icon=None, path=None):
        """
        :param instance: Represented database entity
        :type instance: WhdbBlockInstanceP
        :param parentItem: Parent tree item
        :type parentItem: WguiTreeItemP
        :param icon: Icon
        :type icon: QIcon | None
        :param path: Entity path
        :type path: list[WhdbBlockInstanceP]
        """
        children = [WguiDbGroupBlockItem(s, self)
                    for s in _WGUI_BLOCK_GROUPS]

        super().__init__(instance, parentItem, icon, path, children)

        self.__instance = instance

    def _getName(self):
        return self.__instance.getName()

    def _getType(self):
        return self.__instance.getTypeName()

    def addChild(self, child, row=None):
        raise NotImplementedError()

    def removeChild(self, row):
        raise NotImplementedError()

    def removeChildren(self, row, count):
        raise NotImplementedError()


class WguiDbTreeModelInstanceItem(WguiDbTreeItemBase):
    def __init__(self, instance, parentItem, icon=None, path=None):
        """
        :param instance: Represented database entity
        :type instance: WhdbModelInstanceP
        :param parentItem: Parent tree item
        :type parentItem: WguiTreeItemP
        :param icon: Icon
        :type icon: QIcon | None
        :param path: Entity path
        :type path: list[WhdbBlockInstanceP]
        """
        super().__init__(instance, parentItem, icon, path)

        self.__instance = instance

    def _getName(self):
        return self.__instance.getName()

    def _getType(self):
        return self.__instance.getMedia()


class WguiDbTreeOperatorInstanceItem(WguiDbTreeItemBase):
    def __init__(self, instance, parentItem, icon=None, path=None):
        """
        :param instance: Represented database entity
        :type instance: WhdbOperatorInstanceP
        :param parentItem: Parent tree item
        :type parentItem: WguiTreeItemP
        :param icon: Icon
        :type icon: QIcon | None
        :param path: Entity path
        :type path: list[WhdbBlockInstanceP]
        """
        super().__init__(instance, parentItem, icon, path)

        self.__instance = instance

    def _getName(self):
        inst = self.__instance
        name = inst.getName()
        return name if name else inst.getOperator().getName()

    def _getType(self):
        op = self.__instance.getOperator()
        return '{} ({})'.format(op.getName(), op.getCode())


class WguiDbTreeConnectionItem(WguiDbTreeItemBase):
    def __init__(self, connection, parentItem, icon=None, path=None):
        """
        :param connection: Represented database entity
        :type connection: WhdbConnectionP
        :param parentItem: Parent tree item
        :type parentItem: WguiTreeItemP
        :param icon: Icon
        :type icon: QIcon | None
        :param path: Entity path
        :type path: list[WhdbBlockInstanceP]
        """
        super().__init__(connection, parentItem, icon, path)

        self.__connection = connection

    def _getName(self):
        return whdb_getConnectionDescription(self.__connection)

    def _getType(self):
        return self.__connection.getConnectionTypeDescription()


class WguiDbTreeShapeItem(WguiDbTreeItemBase):
    def __init__(self, shape, parentItem, icon=None, path=None):
        """
        :param shape: Represented database entity
        :type shape: WhdbShapeP
        :param parentItem: Parent tree item
        :type parentItem: WguiTreeItemP
        :param icon: Icon
        :type icon: QIcon | None
        :param path: Entity path
        :type path: list[WhdbBlockInstanceP]
        """
        super().__init__(shape, parentItem, icon, path)

        self.__shape = shape

    def _getName(self):
        return self.__shape.shapeName.title()

    def _getType(self):
        return None


class WguiDbTreeButtonItem(WguiDbTreeItemBase):
    def __init__(self, button, parentItem, icon=None, path=None):
        """
        :param button: Represented database entity
        :type button: WhdbButtonP
        :param parentItem: Parent tree item
        :type parentItem: WguiTreeItemP
        :param icon: Icon
        :type icon: QIcon | None
        :param path: Entity path
        :type path: list[WhdbBlockInstanceP]
        """
        super().__init__(button, parentItem, icon, path)

        self.__button = button

    def _getName(self):
        return self.__button.title or 'Button'

    def _getType(self):
        return self.__button.command or 'no command'


class WguiDbTreeItemFactory(WguiTreeItemFactoryP):
    """
    Tree-item factory for database trees.
    """

    #: Indices of block item children
    BLOCK_INST_IDX, MODEL_INST_IDX, OPERATOR_INST_IDX, CONNECTION_IDX, \
        SHAPE_IDX = range(0, 5)

    def __init__(self, iconPathFy):
        """
        :param iconPathFy: Icon path factory
        :type iconPathFy: WguiIconPathFactoryP
        """
        #: Icon path factory
        self.__iconPathFy = iconPathFy

        #: Icon dictionary {str:QIcon}
        self.__icons = {}
        """:type: dict[WhdbBaseP, QIcon]"""

    def _getIcon(self, item):
        """
        Get the icon for the given item.

        :param item: Item to retrieve icon for
        :type item: WhdbBaseP
        :return: Icon for given item
        :rtype: QIcon | None
        """
        icon = self.__icons.get(item)
        if icon:
            return icon
        path = self.__iconPathFy.getIconPath(item, WGUI_ICON_PNG32)
        if path:
            icon = QIcon(path)
            self.__icons[item] = icon
            return icon
        return None

    def createTreeItem(self, itemData, parentItem, **kwargs):
        """
        :type itemData: WhdbBaseP
        :type parentItem: WguiTreeItemP | None
        :rtype: (WguiTreeItemP, WguiTreeItemP, Iterable[object] | None)
        """
        if parentItem is None:
            return WguiDbTreeRootItem(parentItem), None, None

        return self._createTreeNode(itemData, parentItem, **kwargs)

    def _createTreeNode(self, itemData, parentItem, **kwargs):
        item = None
        icon = self._getIcon(itemData)
        path = kwargs.get('path')
        children = None

        if isinstance(itemData, WhdbBlockInstanceP):
            parentItem = parentItem.getChild(self.BLOCK_INST_IDX)
            item = WguiDbTreeBlockInstanceItem(
                itemData, parentItem, icon, path)
            kwargs['path'] = path + [itemData] if path else [itemData]
            blck = itemData.getBlock()
            children = (chain(blck.getInstances(), blck.getConnections(),
                              blck.getShapes(), blck.getButtons()),
                        kwargs)
        elif isinstance(itemData, WhdbModelInstanceP):
            parentItem = parentItem.getChild(self.MODEL_INST_IDX)
            item = WguiDbTreeModelInstanceItem(
                itemData, parentItem, icon, path)
        elif isinstance(itemData, WhdbOperatorInstanceP):
            parentItem = parentItem.getChild(self.OPERATOR_INST_IDX)
            item = WguiDbTreeOperatorInstanceItem(
                itemData, parentItem, icon, path)
        elif isinstance(itemData, WhdbConnectionP):
            parentItem = parentItem.getChild(self.CONNECTION_IDX)
            item = WguiDbTreeConnectionItem(
                itemData, parentItem, icon, path)
        elif isinstance(itemData, WhdbShapeP):
            parentItem = parentItem.getChild(self.SHAPE_IDX)
            item = WguiDbTreeShapeItem(
                itemData, parentItem, icon, path)
        elif isinstance(itemData, WhdbButtonP):
            parentItem = parentItem.getChild(self.SHAPE_IDX)
            item = WguiDbTreeButtonItem(
                itemData, parentItem, icon, path)

        return item, parentItem, children


class WguiDbTreeModelManager(WguiTreeModelManager):
    """
    Tree-model manager for database trees.
    """

    def __init__(self, tblk, itemFy):
        """
        :param tblk: Database top-level block
        :type tblk: WhdbEntityContainerP
        :param itemFy: Tree-item factory
        :type itemFy: WguiTreeItemFactoryP
        """
        super().__init__(
            lambda i: i.getUUID() if i is not None else None, itemFy)

        self.__tblk = tblk

    def populateModel(self, fromScratch=True):
        """
        :type fromScratch: bool
        """
        def addBlock(block, path):
            for inst in block.getInstances():
                self.itemAdded(inst, path)
            for conn in block.getConnections():
                self.itemAdded(conn, path)
            for shape in block.getShapes():
                self.itemAdded(shape, path)

        if fromScratch:
            self.itemClear()
        addBlock(self.__tblk, [])


class WguiDbSyncedTreeModelManager(WguiDbTreeModelManager):
    """
    Tree-model manager for database trees with item synchronization.
    """

    def __init__(self, tblk, scene, itemFy):
        """
        :param tblk: Database top-level block
        :type tblk: WhdbEntityContainerP
        :param scene: Graphics scene
        :type scene: WguiGraphicsScene
        :param itemFy: Tree-item factory
        :type itemFy: WguiTreeItemFactoryP
        """
        super().__init__(tblk, itemFy)

        #: Graphics scene
        self.__scene = scene

        scene.sceneItemAdded.connect(self.itemAdded)
        scene.sceneItemRemoved.connect(self.itemRemoved)
        scene.sceneItemChanged.connect(self.itemChanged)


class WguiDbTreeSceneSelectionSynchronizer(object):
    """
    Synchronization between the selection of the graphics scene and a database
    tree.
    """

    def __init__(self, tree, scene):
        """
        :param tree: Database tree
        :type tree: WguiTree
        :param scene: Graphics scene
        :type scene: WguiGraphicsScene
        """
        #: Database tree reference
        self.__tree = ref(tree)
        #: Graphics scene
        self.__scene = scene

        # Store reference to synchronizer to avoid garbage collection
        tree.sceneSelectionSynchronizer = self

        #: Whether an update is ongoing
        self.__updating = False

        tree.itemSelected.connect(self._treeSelectionChanged)
        scene.selectionChanged.connect(self._sceneSelectionChanged)
        scene.scenePathChanged.connect(tree.hierarchyLevelChanged)

    def _treeSelectionChanged(self, items):
        """
        Callback function to handle tree selection changes.

        :param items: Database items currently selected in the tree
        :type items: list[(WhdbBaseP, list[WhdbBlockInstanceP])]
        """
        if self.__updating:
            return
        curSel = [i.getUUID() for i
                  in [i.getDataObject() for i
                      in self.__scene.selectedItems()
                      if isinstance(i, WguiSceneItemP)]]
        newSel = [i.getUUID() for i, _ in items]
        if curSel != newSel:
            if items:
                self.__scene.setHierarchy(items[0][1])
            self.__scene.clearSelection()
            self.__scene.selectEntities(i for i, _ in items)

    def _sceneSelectionChanged(self):
        """
        Callback function to handle scene selection changes.
        """
        tree = self.__tree()
        if not tree:
            return
        self.__updating = True
        selected = self.__scene.selectedItems()
        selItems = [i.getDataObject() for i in selected
                    if isinstance(i, WguiSceneItemP)]
        tree.selectDbItems(selItems)
        self.__updating = False


class WguiDbTreeSortFilter(WguiFilterModel):
    """
    Base sort/filter proxy for DB tree models.
    """

    def __init__(self, model=None, parent=None):
        """
        :param model: Model to proxy
        :type model: QAbstractItemModel
        :param parent: Parent object
        """
        super().__init__(parent)

        #: Whether the proxy needs to revalidated due to inserts/removals
        self.__needsInvalidation = False

        if model:
            self.setSourceModel(model)

    def setSourceModel(self, model):
        """
        :type model: QAbstractItemModel
        """
        _model = self.sourceModel()
        if _model:
            _model.rowsAboutToBeInserted.disconnect(
                self._sourceRowsToBeInserted)
            _model.rowsInserted.connect(self._invalidateIfNeed)
            _model.rowsAboutToBeRemoved.connect(self._sourceRowsToBeRemoved)
            _model.rowsRemoved.connect(self._invalidateIfNeed)

        model.rowsAboutToBeInserted.connect(self._sourceRowsToBeInserted)
        model.rowsInserted.connect(self._invalidateIfNeed)
        model.rowsAboutToBeRemoved.connect(self._sourceRowsToBeRemoved)
        model.rowsRemoved.connect(self._invalidateIfNeed)
        super().setSourceModel(model)

    def _sourceRowsToBeInserted(self, parent, *_):
        """
        Callback function for pending row-insertion into source model.

        :param parent: Parent index of rows to be inserted
        :type parent: QModelIndex
        """
        if parent.parent() and not self.filterAcceptsRow(parent.row(),
                                                         parent.parent()):
            self.__needsInvalidation = True

    def _sourceRowsToBeRemoved(self, parent, row, last):
        """
        Callback function for pending row-removal from source model.

        :param parent: Parent index of rows to be removed
        :type parent: QModelIndex
        :param row: Number of first row to be removed
        :type row: int
        :param last: Number of last row to be removed
        :type last: int
        """
        if row == 0 and last == self.sourceModel().rowCount(parent) - 1:
            self.__needsInvalidation = True

    def _invalidateIfNeed(self):
        """
        Callback function for completed row-insertion/removal into/from
        source model.
        """
        if self.__needsInvalidation:
            self.invalidate()
            self.__needsInvalidation = False

    def filterAcceptsRow(self, srow, pidx):
        """
        :type srow: int
        :type pidx: QModelIndex
        :rtype: bool
        """
        model = self.sourceModel()
        """:type: QAbstractItemModel"""
        if not model:
            return False

        index = model.index(srow, 0, pidx)
        noleaf = model.data(index, WGUI_TREE_NOLEAF_ROLE)

        if noleaf and model.rowCount(index) == 0:
            return False
        return super().filterAcceptsRow(srow, pidx)


def wgui_createDbTree(tblk, scene, itemFy, menuFy):
    """
    Create a database tree and all required objects.

    :param tblk: Database top-level block
    :type tblk: WhdbBlock
    :param scene: Graphics scene
    :type scene: WguiGraphicsScene
    :param itemFy: Tree-item factory
    :type itemFy: WguiTreeItemFactoryP
    :param menuFy: Context-menu factory
    :type menuFy: WguiContextMenuFactoryP
    :return: Newly created database tree
    :rtype: WguiTree
    """
    modelMngr = WguiDbSyncedTreeModelManager(tblk, scene, itemFy)
    model = WguiDbTreeSortFilter(modelMngr.model)
    tree = WguiTree(model, modelMngr, menuFy)
    WguiDbTreeSceneSelectionSynchronizer(tree, scene)
    return tree
