# ----------------------------------------------------------------------------
# - wgui-editor-tab-widget                                                   -
# - wild cruncher - cockpit : editor tabs window                             -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_editor_tab_widget
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher tab widget

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
import re
from functools import partial

from PyQt4 import QtCore
from PyQt4.QtCore import Qt, QEvent
from PyQt4.QtGui import QTabWidget, QMessageBox, QIcon, QTabBar, QLineEdit,\
     QColor, QPainter, QStyleOptionTab, QStyle
from wcr_core.action.wact_action import WactActionBase
from wcr_core.action.wact_action_parameter import WactActionParameter
from wcr_core.action.wact_action_parameter_types import \
    WactActionStringParameterType
from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_frontend import WGUI_WINDOW_TITLE
from wcr_frontend.gui.wgui_editor import WguiEditorHasSaveP, WguiEditorBase
from silkan_common.scmn_file_path import scmn_path_base


def wgui_create_wtabs_actions(tabw):
    """
    Create a set of tab window functions.

    :param tabw: Tab window
    :type tabw: WguiTabsWindow
    :return: Project tab window actions
    :rtype: dict[str, WactActionP]
    """
    return {
        'tabw.activate': _WguiSwitchTabAction(tabw)
    }


class _WguiSwitchTabAction(WactActionBase):
    """
    Action to activate a given tab.
    """

    #: Action tab parameter
    __PARAMETER = WactActionParameter(
        'tab', 'Tab name', 'Tab name to switch to',
        WactActionStringParameterType())

    def __init__(self, tabw):
        """
        :param tabw: Tab widget
        :type tabw: WguiTabsWindow
        """
        super().__init__('Switch tab', 'Switch to another tab',
                         self.__PARAMETER)

        #: Tab widget
        self.__tabWindow = tabw

    def _doExecute(self, actx, paramValuePairs):
        tab = self._getParamValue(self.__PARAMETER, actx, paramValuePairs)
        if tab:
            self.__tabWindow.activateTab(tab)


class WguiTabsWindow(QTabWidget):
    #: Signal emitting whenever an editor is made active
    activeEditorChanged = QtCore.pyqtSignal(object)

    #: Signal emitting whenever an editor changes its name
    editorNameChanged = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        """
        :type parent: Parent widget
        """
        super().__init__(parent)
        self.colorIndexes = {
            1: QColor(20, 255, 204),
            2: QColor(10, 10, 204),
            3: QColor(204, 255, 204),
            4: QColor(240, 240, 240)
        }

        self.setTabBar(ProjectTab(self, parent))
        self.setDocumentMode(True)
        self.setMovable(True)
        self.setTabsClosable(True)
        self.currentChanged.connect(self._currentTabChanged)
        self.tabCloseRequested.connect(self._tabCloseRequested)

    def openEditor(self, ident, editorF, isEditableName=True):
        """
        Add the given editor in a new tab and make it the active one.
        If the given editor already exists, its tab we be made the active one.

        :param ident: Ident string for the editor to open
        :type ident: str
        :param editorF: Editor creation function, called if editor with the
                        given `ident` does not yet exist
        :type editorF: WguiEditor | _ -> WguiEditor
        """
        editor = None

        for _editor in self.editors:
            if _editor.getEditorIdent() == ident:
                editor = _editor

        if editor:
            pidx = self.indexOf(editor)
        else:
            editor = editorF() if callable(editorF) else editorF
            pidx = self.addTab(editor, editor.getTitle() or 'Untitled')
            self.tabBar().setEditor(editor)
            self.tabBar().isEditable = isEditableName
            editor.nameChanged.connect(partial(self._editorNameChanged,
                                               editor))
            icon = editor.getIcon()
            if icon:
                self.setTabIcon(pidx, QIcon(icon))
        self.setCurrentIndex(pidx)

    def activateTab(self, name):
        for idx in range(self.count()):
            if self.tabText(idx) == name:
                self.setCurrentIndex(idx)

    def _editorNameChanged(self, editor, title):
        """
        Callback slot for whenever an editor changes its name.

        :param editor: Editor that changed its name
        :type editor: WguiEditor
        :param title: New title for the editor's tab
        :type title: str
        """
        pidx = self.indexOf(editor)
        if pidx < 0: return

        self.setTabText(pidx, title or 'Untitled')
        self.parent().setWindowTitle(
            '{}: {} '.format(WGUI_WINDOW_TITLE,
                               title or 'Untitled'))
        self.clearFocus()
        self.parentWidget().setFocus(Qt.OtherFocusReason)
        self.editorNameChanged.emit()

    def _tabCloseRequested(self, tab):
        """
        Callback slot to intercept tab closing.

        :param tab: Index of tab to be closed
        :type tab: int
        """
        editor = self.widget(tab)
        if editor is None:
            return

        if isinstance(editor, WguiEditorHasSaveP) and editor.isDirty():
            cnfm = self._showConfirmTabCloseMessageBox(tab)
            if cnfm == QMessageBox.Save:
                try:
                    editor.save()
                except:
                    QMessageBox(QMessageBox.Warning,
                                'Error saving to file',
                                'The contents could not be saved, '
                                'please try to save manually.',
                                QMessageBox.Ok,
                                self).exec()
                    return
            elif cnfm == QMessageBox.Cancel:
                return

        if isinstance(editor, WcmnReleasableP):
            editor.release()

        self.removeTab(tab)

    def _showConfirmTabCloseMessageBox(self, tab):
        """
        Display a confirmation dialog for tab closing.

        :param tab: Index of tab to be closed
        :type tab: int
        :return: Dialog response (Save, Discard, Cancel)
        :rtype: int
        """
        diaw = QMessageBox(self)
        diaw.setWindowTitle('Confirm Close')
        diaw.setText('Do you want to save changes made to \'{n}\' before '
                     'closing its tab?'.format(n=self.tabText(tab)))
        diaw.setInformativeText(
            'All unsaved changes will be lost when the tab is closed.')
        diaw.addButton(QMessageBox.Save)
        diaw.addButton(QMessageBox.Discard)
        diaw.addButton(QMessageBox.Cancel)
        diaw.setDefaultButton(QMessageBox.Cancel)
        diaw.setIcon(QMessageBox.Question)
        diaw.setEscapeButton(QMessageBox.Cancel)

        return diaw.exec()

    def _currentTabChanged(self, *_):
        """
        Callback slot for when the current editor changes.
        """
        wwin = self.currentEditor
        self.tabBar().setEditor(wwin)
        if wwin is not None:
            if wwin.isICDProject():
                wwin.initialize()
            wwin.activated()
        self.activeEditorChanged.emit(wwin)

    @property
    def currentEditor(self):
        """
        Get the current editor.

        :return: Current editor
        :rtype: WguiEditor
        """
        return self.currentWidget()

    @property
    def editors(self):
        """
        Get all editors.

        :return: List of editors
        :rtype: list[WguiEditor]
        """
        from wcr_frontend.modules.vlink.gui.wvgui_scene_protocol_table import \
            WvguiSceneProtocolTable

        return [self.widget(i) for i in range(self.count())
                if (isinstance(self.widget(i), WguiEditorBase) or
                    isinstance(self.widget(i), WvguiSceneProtocolTable))]


class TabBar(QTabBar):
    """
       Editable Tab
    """
    def __init__(self, parent):
        """
            :type parent: Parent widget
        """
        QTabBar.__init__(self, parent)
        self._editor = QLineEdit(self)
        self._editor.setWindowFlags(Qt.Popup)
        self._editor.setFocusProxy(self)
        self._editor.editingFinished.connect(self.handleEditingFinished)
        self._editor.installEventFilter(self)

    def isEditable(self):
        return True

    def eventFilter(self, widget, event):
        if ((event.type() == QEvent.MouseButtonPress and
             not self._editor.geometry().contains(event.globalPos())) or
            (event.type() == QEvent.KeyPress and
             event.key() == Qt.Key_Escape)):
            self._editor.hide()
            return True
        return QTabBar.eventFilter(self, widget, event)

    def mouseDoubleClickEvent(self, event):
        """
            Get the index of tab.
        :param event:
        """
        if not self.isEditable:
            return
        index = self.tabAt(event.pos())
        if index >= 0:
            self.editTab(index)

    def editTab(self, index):
        """
            Use index to get tab data when editing
            :param index:
            :type int
        """
        rect = self.tabRect(index)
        self._editor.setFixedSize(rect.size())
        self._editor.move(self.parent().mapToGlobal(rect.topLeft()))
        self._editor.setText(self.tabText(index))
        if not self._editor.isVisible():
            self._editor.show()

    def handleEditingFinished(self):
        index = self.currentIndex()
        if index >= 0:
            self._editor.hide()
            self.setTabText(index, self._editor.text())


class ProjectTab(TabBar):
    def __init__(self, parent, window):
        TabBar.__init__(self, parent)
        self.colorIndexes = parent.colorIndexes
        self._tabEditor = None
        self.filename = None
        self.window = window

    def setEditor(self, editor):
        """
        :param editor:
        :type WguiProjectEditor:
        :return:
        """
        self._tabEditor = editor

    def editTab(self, index):
        """
        Edition of the tab, remove the filename before editing to avoid
        repetition.
        :param index:
        :return:
        """
        rect = self.tabRect(index)
        self._editor.setFixedSize(rect.size())
        self._editor.move(self.parent().mapToGlobal(rect.topLeft()))
        # take the projectdb
        projectDb = getattr(self._tabEditor, 'getProjectDB', None)
        if projectDb:
            filename = scmn_path_base(projectDb().path)
        else:
            # if no project (text or log) then filename set to Untitled
            filename = 'Untitled'
        self.filename = '({})'.format(filename)
        projectName = self.tabText(index).replace(self.filename, '')
        self._editor.setText(projectName)
        if not self._editor.isVisible():
            self._editor.show()

    def handleEditingFinished(self):
        """
            Get the text that is in editor use it to set projectName,
            setWindowTitle with format: prjName (filename)
        """
        index = self.currentIndex()
        if index >= 0:
            self._editor.hide()
            self.setTabText(index, '{} {}'.format(self._editor.text(),
                                                  self.filename))
            if self._tabEditor:
                projectDb = getattr(self._tabEditor, 'getProjectDB', None)
                if projectDb:
                    projectInfo = projectDb().projectInfo
                    projectInfo.projectName = self._editor.text()
                    self.window.setWindowTitle(
                        '{}: {} {}'.format(WGUI_WINDOW_TITLE,
                                           projectInfo.projectName or
                                           'Untitled',
                                           self.filename))
