# ----------------------------------------------------------------------------
# - wgui-editor-tab-widget                                                   -
# - wild cruncher - cockpit : editor tabs window                             -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_editor_tab_widget
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher tab widget

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
import re
from functools import partial

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt, QEvent, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QStyle, QMessageBox, QApplication, \
	QTabWidget, QLineEdit, QTabBar, QStyleOptionTab, QMainWindow
from PyQt5.QtGui import QIcon, QColor, QPainter, QCursor, QMouseEvent, QDrag, QPixmap, QScreen
from wcr_core.action.wact_action import WactActionBase
from wcr_core.action.wact_action_parameter import WactActionParameter
from wcr_core.action.wact_action_parameter_types import \
    WactActionStringParameterType
from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_frontend import WGUI_WINDOW_TITLE
from wcr_frontend.gui.wgui_editor import WguiEditorHasSaveP, WguiEditorBase
from silkan_common.scmn_file_path import scmn_path_base
from wcr_frontend.gui.wgui_text_editor import WguiTextEditor
from wcr_frontend.gui.wgui_logger import WguiLoggerWindow


def wgui_create_wtabs_actions(tabw):
    """
    Create a set of tab window functions.

    :param tabw: Tab window
    :type tabw: WguiTabsWindow
    :return: Project tab window actions
    :rtype: dict[str, WactActionP]
    """
    return {
        'tabw.activate': _WguiSwitchTabAction(tabw)
    }


##
# The DetachableTabWidget adds additional functionality to Qt's QTabWidget that allows it
# to detach and re-attach tabs.
#
# Additional Features:
#   Detach tabs by
#     dragging the tabs away from the tab bar
#     double clicking the tab
#   Re-attach tabs by
#     dragging the detached tab's window into the tab bar
#     closing the detached tab's window
#   Remove tab (attached or detached) by name
#
# Modified Features:
#   Re-ordering (moving) tabs by dragging was re-implemented
#
class DetachableTabWidget(QTabWidget):
    def __init__(self, parent=None):
        QTabWidget.__init__(self, parent)

        self.tabBar = self.TabBar(self)
        self.tabBar.onDetachTabSignal.connect(self.detachTab)
        self.tabBar.onMoveTabSignal.connect(self.moveTab)
        self.tabBar.detachedTabDropSignal.connect(self.detachedTabDrop)

        self.setTabBar(self.tabBar)

        # Used to keep a reference to detached tabs since their QMainWindow
        # does not have a parent
        self.detachedTabs = {}
        self.__editorIdent = ''
        # Close all detached tabs if the application is closed explicitly
        QtWidgets.qApp.aboutToQuit.connect(self.closeDetachedTabs) # @UndefinedVariable

    def setMovable(self, movable):
        """
         The default movable functionality of QTabWidget must remain disabled
         so as not to conflict with the added features
        """
        pass

    def moveTab(self, fromIndex, toIndex):
        """
            Move a tab from one position (index) to another
        :param fromIndex: original position of teh tab
        :param toIndex: final position of the tab
        :return:
        """
        widget = self.widget(fromIndex)
        icon = self.tabIcon(fromIndex)
        text = self.tabText(fromIndex)

        self.removeTab(fromIndex)
        self.insertTab(toIndex, widget, icon, text)
        self.setCurrentIndex(toIndex)

    def detachTab(self, index, point):
        """
           Detach the tab by removing it's contents and placing them in
        :param index: index of the tab in the tabBar
        :param point: position in the window
        :return:
        """
        # Get the tab content
        name = self.tabText(index)
        icon = self.tabIcon(index)
        if icon.isNull():
            icon = self.window().windowIcon()
        contentWidget = self.widget(index)

        try:
            contentWidgetRect = contentWidget.frameGeometry()
        except AttributeError:
            return
        from wcr_frontend.modules.vlink.gui.wvgui_scene_protocol_table import \
            WvguiSceneProtocolTable
        if isinstance(contentWidget, WvguiSceneProtocolTable):
            return
        contentWidget.scene.detachedEditor = contentWidget
        # Create a new detached tab window
        detachedTab = self.DetachedTab(name, contentWidget, self)
        detachedTab.setWindowModality(QtCore.Qt.NonModal)
        detachedTab.setWindowIcon(icon)
        detachedTab.setGeometry(contentWidgetRect)
        detachedTab.onCloseSignal.connect(self.attachTab)
        detachedTab.onDropSignal.connect(self.tabBar.detachedTabDrop)
        detachedTab.setEditorIdent(contentWidget.getEditorIdent())
        detachedTab.move(point)
        detachedTab.show()

        # Create a reference to maintain access to the detached tab
        self.detachedTabs[name] = detachedTab

    def attachTab(self, contentWidget, name, icon, insertAt=None):

        """
        Re-attach the tab by removing the content from the DetachedTab window,
        closing it, and placing the content back into the DetachableTabWidget
        :param contentWidget: attached widget
        :param name: name of the tab
        :param icon: icon of the tab
        :param insertAt: the position to attache the tab in tabBar
        :return:
        """

        # Make the content widget a child of this widget
        contentWidget.setParent(self)
        # Remove the reference
        self.detachedTabs[name].deleteLater()
        del self.detachedTabs[name]
        # Create an image from the given icon (for comparison)
        if not icon.isNull():
            try:
                tabIconPixmap = icon.pixmap(icon.availableSizes()[0])
                tabIconImage = tabIconPixmap.toImage()
            except IndexError:
                tabIconImage = None
        else:
            tabIconImage = None

        # Create an image of the main window icon (for comparison)
        if not icon.isNull():
            try:
                windowIconPixmap = self.window().windowIcon().pixmap(icon.availableSizes()[0])
                windowIconImage = windowIconPixmap.toImage()
            except IndexError:
                windowIconImage = None
        else:
            windowIconImage = None

        # Determine if the given image and the main window icon are the same.
        # If they are, then do not add the icon to the tab
        if tabIconImage == windowIconImage:
            if insertAt == None:
                contentWidget.reattached = True
                index = self.addTab(contentWidget, name)
                contentWidget.reattached = False
            else:
                index = self.insertTab(insertAt, contentWidget, name)
        else:
            if insertAt == None:
                index = self.addTab(contentWidget, icon, name)
            else:
                index = self.insertTab(insertAt, contentWidget, icon, name)

        # Make this tab the current tab
        if index > -1:
            self.setCurrentIndex(index)

    def removeTabByName(self, name):
        """
            Remove the tab with the given name, even if it is detached
        :param name: name of the name to remove
        :return:
        """

        # Remove the tab if it is attached
        attached = False
        for index in range(self.count()):
            if str(name) == str(self.tabText(index)):
                self.removeTab(index)
                attached = True
                break

        # If the tab is not attached, close it's window and
        # remove the reference to it
        if not attached:
            for key in self.detachedTabs:
                if str(name) == str(key):
                    self.detachedTabs[key].onCloseSignal.disconnect()
                    self.detachedTabs[key].close()
                    del self.detachedTabs[key]
                    break

    def detachedTabDrop(self, name, index, dropPos):
        """
            Handle dropping of a detached tab inside the DetachableTabWidget
        :param name: name of the tab
        :param index: position of the tab
        :param dropPos: position point to drop the detached the tab
        :return:
        """
        # If the drop occurred on an existing tab, insert the detached
        # tab at the existing tab's location
        if index > -1:

            # Create references to the detached tab's content and icon
            contentWidget = self.detachedTabs[name].contentWidget
            icon = self.detachedTabs[name].windowIcon()

            # Disconnect the detached tab's onCloseSignal so that it
            # does not try to re-attach automatically
            self.detachedTabs[name].onCloseSignal.disconnect()

            # Close the detached
            self.detachedTabs[name].close()

            # Re-attach the tab at the given index
            self.attachTab(contentWidget, name, icon, index)

        # If the drop did not occur on an existing tab, determine if the drop
        # occurred in the tab bar area (the area to the side of the QTabBar)
        else:

            # Find the drop position relative to the DetachableTabWidget
            tabDropPos = self.mapFromGlobal(dropPos)

            # If the drop position is inside the DetachableTabWidget...
            if self.rect().contains(tabDropPos):

                # If the drop position is inside the tab bar area (the
                # area to the side of the QTabBar) or there are not tabs
                # currently attached...
                if tabDropPos.y() < self.tabBar.height() or self.count() == 0:

                    # Close the detached tab and allow it to re-attach
                    # automatically
                    self.detachedTabs[name].close()

    def closeDetachedTabs(self):
        """
            Close all tabs that are currently detached.
        :return:
        """
        listOfDetachedTabs = []

        for key in self.detachedTabs:
            listOfDetachedTabs.append(self.detachedTabs[key])

        for detachedTab in listOfDetachedTabs:
            detachedTab.close()

    #  When a tab is detached, the contents are placed into this QMainWindow.  The tab
    #  can be re-attached by closing the dialog or by dragging the window into the tab bar
    class DetachedTab(QMainWindow):
        onCloseSignal = pyqtSignal(QtWidgets.QWidget, str, QIcon)
        onDropSignal = pyqtSignal(str, QtCore.QPoint)

        def __init__(self, name, contentWidget, parent):
            QMainWindow.__init__(self, None)
            self.__name = name
            self.__parent = parent
            self.setObjectName(name)
            self.setWindowTitle(name)

            self.contentWidget = contentWidget
            self.setCentralWidget(self.contentWidget)
            self.contentWidget.show()
            self.__editorIdent = ''

            self.windowDropFilter = self.WindowDropFilter()
            self.installEventFilter(self.windowDropFilter)
            # self.windowDropFilter.onDropSignal.connect(self.windowDropSlot)

        def windowDropSlot(self, dropPos):
            """
            Handle a window drop event
            :param dropPos:
            :return:
            """
            self.onDropSignal.emit(self.objectName(), dropPos)

        def closeEvent(self, event):
            """
            If the window is closed, emit the onCloseSignal and give the
            content widget back to the DetachableTabWidget
            :param event:
            :return:
            """
            self.onCloseSignal.emit(self.contentWidget, self.objectName(), self.windowIcon())

        def tabCloseRequested(self):
            """
            Callback slot to intercept tab closing.

            :param tab: Index of tab to be closed
            :type tab: int
            """

            # editor = self.contentWidget
            # if editor is None:
            #     return
            #
            # cnfm = self._showConfirmTabCloseMessageBox(self.contentWidget)
            # if cnfm == QMessageBox.Save:
            #     try:
            #         editor.save()
            #         event.accept()
            #     except:
            #         QMessageBox(QMessageBox.Warning,
            #                     'Error saving to file',
            #                     'The contents could not be saved, '
            #                     'please try to save manually.',
            #                     QMessageBox.Ok,
            #                     self).exec()
            #         event.ignore()
            # elif cnfm == QMessageBox.Cancel:
            #     event.ignore()


        def _showConfirmTabCloseMessageBox(self, tab):
            """
            Display a confirmation dialog for tab closing.

            :param tab: Index of tab to be closed
            :type tab: int
            :return: Dialog response (Save, Discard, Cancel)
            :rtype: int
            """
            diaw = QMessageBox(self)
            diaw.setWindowTitle('Confirm Close')
            diaw.setText('Do you want to save changes made before '
                         'closing this tab?')
            diaw.setInformativeText(
                'All unsaved changes will be lost when the tab is closed.')
            diaw.addButton(QMessageBox.Save)
            diaw.addButton(QMessageBox.Discard)
            diaw.addButton(QMessageBox.Cancel)
            diaw.setDefaultButton(QMessageBox.Cancel)
            diaw.setIcon(QMessageBox.Question)
            diaw.setEscapeButton(QMessageBox.Cancel)

            return diaw.exec()

        def setEditorIdent(self, ident):
            self.__editorIdent = ident

        def getEditorIdent(self):
            return self.__editorIdent

        #  An event filter class to detect a QMainWindow drop event
        class WindowDropFilter(QtCore.QObject):
            onDropSignal = pyqtSignal(QtCore.QPoint)

            def __init__(self):
                QtCore.QObject.__init__(self)
                self.lastEvent = None

            def eventFilter(self, obj, event):
                """
                    Detect a QMainWindow drop event by looking for a NonClientAreaMouseMove (173)
                    event that immediately follows a Move event
                :param obj:
                :param event:
                :return:
                """

                if self.lastEvent == QtCore.QEvent.Close:
                    return True
                # If a NonClientAreaMouseMove (173) event immediately follows a Move event...
                elif self.lastEvent == QtCore.QEvent.Move and event.type() == 173:
                    # Determine the position of the mouse cursor and emit it with the
                    # onDropSignal
                    mouseCursor = QCursor()
                    dropPos = mouseCursor.pos()
                    self.onDropSignal.emit(dropPos)
                    self.lastEvent = event.type()
                    return True

                else:
                    self.lastEvent = event.type()
                    return False

    ##
    #  The TabBar class re-implements some of the functionality of the QTabBar widget
    class TabBar(QtWidgets.QTabBar):
        onDetachTabSignal = pyqtSignal(int, QtCore.QPoint)
        onMoveTabSignal = pyqtSignal(int, int)
        detachedTabDropSignal = pyqtSignal(str, int, QtCore.QPoint)

        def __init__(self, parent=None):
            QtWidgets.QTabBar.__init__(self, parent)

            self.setAcceptDrops(True)
            self.setElideMode(QtCore.Qt.ElideRight)
            self.setSelectionBehaviorOnRemove(QtWidgets.QTabBar.SelectLeftTab)
            self._editor = QLineEdit(self)
            self._editor.setWindowFlags(Qt.Popup)
            self._editor.setFocusProxy(self)
            self._editor.editingFinished.connect(self.handleEditingFinished)
            self._editor.installEventFilter(self)

            self.dragStartPos = QtCore.QPoint()
            self.dragDropedPos = QtCore.QPoint()
            self.mouseCursor = QCursor()
            self.dragInitiated = False

        def setEditor(self, editor):
            """
            :param editor:
            :type WguiProjectEditor:
            :return:
            """
            self._tabEditor = editor

        def isEditable(self):
            return True

        def mouseDoubleClickEvent(self, event):
            """
                Get the index of tab.
            :param event:
            """
            if not self.isEditable:
                return
            index = self.tabAt(event.pos())
            if index >= 0:
                self.editTab(index)

        def editTab(self, index):
            """
                Use index to get tab data when editing
                :param index:
                :type int
            """
            rect = self.tabRect(index)
            self._editor.setFixedSize(rect.size())
            self._editor.move(self.parent().mapToGlobal(rect.topLeft()))
            self._editor.setText(self.tabText(index))
            if not self._editor.isVisible():
                self._editor.show()

        def handleEditingFinished(self):
            index = self.currentIndex()
            if index >= 0:
                self._editor.hide()
                self.setTabText(index, self._editor.text())

        def mousePressEvent(self, event):
            """
                Set the starting position for a drag event
                when the mouse button is pressed
            :param event:
            :return:
            """
            if event.button() == QtCore.Qt.LeftButton:
                self.dragStartPos = event.pos()

            self.dragDropedPos.setX(0)
            self.dragDropedPos.setY(0)

            self.dragInitiated = False

            QtWidgets.QTabBar.mousePressEvent(self, event)

        def mouseMoveEvent(self, event):
            """
                Determine if the current movement is a drag.
                If it is, convert it into a QDrag. If the drag ends inside
                the tab bar, emit an onMoveTabSignal.
                If the drag ends outside the tab bar, emit an onDetachTabSignal.
            :param event:
            :return:
            """
            # Determine if the current movement is detected as a drag
            isDrag = ((event.pos() - self.dragStartPos).manhattanLength() <
                                        QtWidgets.QApplication.startDragDistance())
            if not self.dragStartPos.isNull() and isDrag:
                self.dragInitiated = True

            # If the current movement is a drag initiated by the left button
            if (((event.buttons() & QtCore.Qt.LeftButton)) and
                    self.dragInitiated):

                # Stop the move event
                finishMoveEvent = QMouseEvent(QtCore.QEvent.MouseMove,
                                                    event.pos(),
                                                    QtCore.Qt.NoButton,
                                                    QtCore.Qt.NoButton,
                                                    QtCore.Qt.NoModifier)
                QtWidgets.QTabBar.mouseMoveEvent(self, finishMoveEvent)

                # Convert the move event into a drag
                drag = QDrag(self)
                mimeData = QtCore.QMimeData()
                mimeData.setData('action', 'application/tab-detach'.encode('utf-8'))
                drag.setMimeData(mimeData)

                # Create the appearance of dragging the tab content
                pixmap = self.parentWidget().currentWidget().grab()
                targetPixmap = QPixmap(pixmap.size())
                targetPixmap.fill(QtCore.Qt.transparent)
                painter = QPainter(targetPixmap)
                painter.setOpacity(0.85)
                painter.drawPixmap(0, 0, pixmap)
                painter.end()
                drag.setPixmap(targetPixmap)

                # Initiate the drag
                dropAction = drag.exec_(QtCore.Qt.MoveAction | QtCore.Qt.CopyAction)

                # For Linux:  Here, drag.exec_() will not return MoveAction on Linux.  So it
                #             must be set manually
                if self.dragDropedPos.x() != 0 and self.dragDropedPos.y() != 0:
                    dropAction = QtCore.Qt.MoveAction

                # If the drag completed outside of the tab bar, detach the tab and move
                # the content to the current cursor position
                if dropAction == QtCore.Qt.IgnoreAction:
                    event.accept()
                    self.onDetachTabSignal.emit(self.tabAt(self.dragStartPos),
                                                self.mouseCursor.pos())

                # Else if the drag completed inside the tab bar
                # move the selected tab to the new position
                elif dropAction == QtCore.Qt.MoveAction:
                    if not self.dragDropedPos.isNull():
                        event.accept()
                        self.onMoveTabSignal.emit(self.tabAt(self.dragStartPos),
                                                  self.tabAt(self.dragDropedPos))
            else:
                QtWidgets.QTabBar.mouseMoveEvent(self, event)

        def dragEnterEvent(self, event):
            """
                Determine if the drag has entered a tab position from another tab position
            :param event:
            :return:
            """
            mimeData = event.mimeData()
            # formats = mimeData.formats()

            if mimeData.data('action') == 'application/tab-detach':
                event.acceptProposedAction()

            QtWidgets.QTabBar.dragMoveEvent(self, event)

        def dropEvent(self, event):
            """
                Get the position of the end of the drag
            :param event:
            :return:
            """
            self.dragDropedPos = event.pos()
            QtWidgets.QTabBar.dropEvent(self, event)

        def detachedTabDrop(self, name, dropPos):
            """
            Determine if the detached tab drop event occurred on an existing tab,
            then send the event to the DetachableTabWidget
            :param name:
            :param dropPos:
            :return:
            """
            tabDropPos = self.mapFromGlobal(dropPos)

            index = self.tabAt(tabDropPos)

            self.detachedTabDropSignal.emit(name, index, dropPos)


class _WguiSwitchTabAction(WactActionBase):
    """
    Action to activate a given tab.
    """

    #: Action tab parameter
    __PARAMETER = WactActionParameter(
        'tab', 'Tab name', 'Tab name to switch to',
        WactActionStringParameterType())

    def __init__(self, tabw):
        """
        :param tabw: Tab widget
        :type tabw: WguiTabsWindow
        """
        super().__init__('Switch tab', 'Switch to another tab',
                         self.__PARAMETER)

        #: Tab widget
        self.__tabWindow = tabw

    def _doExecute(self, actx, paramValuePairs):
        tab = self._getParamValue(self.__PARAMETER, actx, paramValuePairs)
        if tab:
            self.__tabWindow.activateTab(tab)


class WguiTabsWindow(DetachableTabWidget):
    #: Signal emitting whenever an editor is made active
    activeEditorChanged = QtCore.pyqtSignal(object)

    #: Signal emitting whenever an editor changes its name
    editorNameChanged = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        """
        :type parent: Parent widget
        """
        super().__init__(parent)
        self.colorIndexes = {
            1: QColor(20, 255, 204),
            2: QColor(10, 10, 204),
            3: QColor(204, 255, 204),
            4: QColor(240, 240, 240)
        }

        # self.setTabBar(ProjectTab(self, parent))
        self.setDocumentMode(True)
        self.setMovable(True)
        self.setTabsClosable(True)
        self.currentChanged.connect(self._currentTabChanged)
        self.tabCloseRequested.connect(self._tabCloseRequested)
        self.previousTab = None

    def openEditor(self, ident, editorF, isEditableName=True):
        """
        Add the given editor in a new tab and make it the active one.
        If the given editor already exists, its tab we be made the active one.

        :param ident: Ident string for the editor to open
        :type ident: str
        :param editorF: Editor creation function, called if editor with the
                        given `ident` does not yet exist
        :type editorF: WguiEditor | _ -> WguiEditor
        """
        editor = None

        for _editor in self.editors:
            if _editor.getEditorIdent() == ident:
                editor = _editor

        if editor:
            pidx = self.indexOf(editor)
        else:
            editor = editorF() if callable(editorF) else editorF
            pidx = self.addTab(editor, editor.getTitle() or 'Untitled')
            self.tabBar.setEditor(editor)
            self.tabBar.isEditable = isEditableName
            editor.nameChanged.connect(partial(self._editorNameChanged,
                                               editor))
            icon = editor.getIcon()
            if icon:
                self.setTabIcon(pidx, QIcon(icon))
        if self.isEditorNeedTmpFile(editor):
            editor.createTmpFile(self)
        self.setCurrentIndex(pidx)
        self.previousTab = editor

    def isEditorNeedTmpFile(self, editor):
        from wcr_frontend.modules.vlink.gui.wvgui_scene_protocol_table import \
            WvguiSceneProtocolTable
        if (not isinstance(editor, WvguiSceneProtocolTable) and
                not isinstance(editor, WguiTextEditor) and
                not isinstance(editor, WguiLoggerWindow)and
                not isinstance(editor, DetachableTabWidget.DetachedTab)):
            return True

        return False

    def activateTab(self, name):
        for idx in range(self.count()):
            if self.tabText(idx) == name:
                self.setCurrentIndex(idx)

    def _editorNameChanged(self, editor, title):
        """
        Callback slot for whenever an editor changes its name.

        :param editor: Editor that changed its name
        :type editor: WguiEditor
        :param title: New title for the editor's tab
        :type title: str
        """
        pidx = self.indexOf(editor)
        if pidx < 0: return

        self.setTabText(pidx, title or 'Untitled')
        self.parent().setWindowTitle(
            '{}: {} '.format(WGUI_WINDOW_TITLE,
                               title or 'Untitled'))
        self.clearFocus()
        self.parentWidget().setFocus(Qt.OtherFocusReason)
        self.editorNameChanged.emit()

    def _tabCloseRequested(self, tab):
        """
        Callback slot to intercept tab closing.

        :param tab: Index of tab to be closed
        :type tab: int
        """
        editor = self.widget(tab)
        if editor is None:
            return

        if isinstance(editor, WguiEditorHasSaveP) and editor.isDirty():
            cnfm = self._showConfirmTabCloseMessageBox(tab)
            if cnfm == QMessageBox.Save:
                try:
                    editor.save()
                except:
                    QMessageBox(QMessageBox.Warning,
                                'Error saving to file',
                                'The contents could not be saved, '
                                'please try to save manually.',
                                QMessageBox.Ok,
                                self).exec()
                    return
            elif cnfm == QMessageBox.Cancel:
                return

        if isinstance(editor, WcmnReleasableP):
            editor.release(self)

        self.removeTab(tab)

    def _showConfirmTabCloseMessageBox(self, tab):
        """
        Display a confirmation dialog for tab closing.

        :param tab: Index of tab to be closed
        :type tab: int
        :return: Dialog response (Save, Discard, Cancel)
        :rtype: int
        """
        diaw = QMessageBox(self)
        diaw.setWindowTitle('Confirm Close')
        diaw.setText('Do you want to save changes made to \'{n}\' before '
                     'closing its tab?'.format(n=self.tabText(tab)))
        diaw.setInformativeText(
            'All unsaved changes will be lost when the tab is closed.')
        diaw.addButton(QMessageBox.Save)
        diaw.addButton(QMessageBox.Discard)
        diaw.addButton(QMessageBox.Cancel)
        diaw.setDefaultButton(QMessageBox.Cancel)
        diaw.setIcon(QMessageBox.Question)
        diaw.setEscapeButton(QMessageBox.Cancel)

        return diaw.exec()

    def _currentTabChanged(self, *_):
        """
        Callback slot for when the current editor changes.
        """
        wwin = self.currentEditor

        if self.previousTab:
            if getattr(self.previousTab, "isDirty", None):
                if self.previousTab.isDirty():
                   self.previousTab.saveTmpFile()

        if wwin is not None:
            # wwin.initialize()
            # if wwin.isICDProject():
            wwin.activated()
        self.tabBar.setEditor(wwin)
        self.previousTab = wwin
        self.activeEditorChanged.emit(wwin)

    @property
    def currentEditor(self):
        """
        Get the current editor.

        :return: Current editor
        :rtype: WguiEditor
        """
        return self.currentWidget()

    @property
    def editors(self):
        """
        Get all editors.

        :return: List of editors
        :rtype: list[WguiEditor]
        """
        from wcr_frontend.modules.vlink.gui.wvgui_scene_protocol_table import \
            WvguiSceneProtocolTable

        attachedEditors = [self.widget(i) for i in range(self.count())
                if (isinstance(self.widget(i), WguiEditorBase) or
                    isinstance(self.widget(i), WvguiSceneProtocolTable))]
        detachedEditors = list(self.detachedTabs.values())
        return attachedEditors + detachedEditors

    def getEditorFromPath(self, topPath):
        for editor in self.editors:
            if (isinstance(editor, WguiLoggerWindow)
                    or isinstance(editor, WguiTextEditor)):
                continue
            db = getattr(editor, 'getProjectDB', None)
            if db:
                if db().getPath() == topPath:
                    return editor
        # if no match found
        return None

    def getEditorMapRouteFromPath(self, topPath):
        for editor in self.editors:
            if (isinstance(editor, WguiLoggerWindow)
                    or isinstance(editor, WguiTextEditor)):
                continue
            top = getattr(editor, 'getTopHierarchy', None)
            if top:
                if getattr(top(), "getVLinkFiles", None):
                    vlBlock = top().getVLinkFiles()
                    if vlBlock:
                        vlMapPath = vlBlock.path
                        if vlMapPath == topPath:
                            return editor
        # if no match found
        return None



class TabBar(QTabBar):
    """
       Editable Tab
    """
    def __init__(self, parent):
        """
            :type parent: Parent widget
        """
        QTabBar.__init__(self, parent)
        self._editor = QLineEdit(self)
        self._editor.setWindowFlags(Qt.Popup)
        self._editor.setFocusProxy(self)
        self._editor.editingFinished.connect(self.handleEditingFinished)
        self._editor.installEventFilter(self)

    def isEditable(self):
        return True

    def eventFilter(self, widget, event):
        if ((event.type() == QEvent.MouseButtonPress and
             not self._editor.geometry().contains(event.globalPos())) or
            (event.type() == QEvent.KeyPress and
             event.key() == Qt.Key_Escape)):
            self._editor.hide()
            return True
        return QTabBar.eventFilter(self, widget, event)

    def mouseDoubleClickEvent(self, event):
        """
            Get the index of tab.
        :param event:
        """
        if not self.isEditable:
            return
        index = self.tabAt(event.pos())
        if index >= 0:
            self.editTab(index)

    def editTab(self, index):
        """
            Use index to get tab data when editing
            :param index:
            :type int
        """
        rect = self.tabRect(index)
        self._editor.setFixedSize(rect.size())
        self._editor.move(self.parent().mapToGlobal(rect.topLeft()))
        self._editor.setText(self.tabText(index))
        if not self._editor.isVisible():
            self._editor.show()

    def handleEditingFinished(self):
        index = self.currentIndex()
        if index >= 0:
            self._editor.hide()
            self.setTabText(index, self._editor.text())


class ProjectTab(TabBar):
    def __init__(self, parent, window):
        TabBar.__init__(self, parent)
        self.colorIndexes = parent.colorIndexes
        self._tabEditor = None
        self.filename = None
        self.window = window

    def setEditor(self, editor):
        """
        :param editor:
        :type WguiProjectEditor:
        :return:
        """
        self._tabEditor = editor

    def editTab(self, index):
        """
        Edition of the tab, remove the filename before editing to avoid
        repetition.
        :param index:
        :return:
        """
        rect = self.tabRect(index)
        self._editor.setFixedSize(rect.size())
        self._editor.move(self.parent().mapToGlobal(rect.topLeft()))
        # take the projectdb
        projectDb = getattr(self._tabEditor, 'getProjectDB', None)
        if projectDb:
            filename = scmn_path_base(projectDb().path)
        else:
            # if no project (text or log) then filename set to Untitled
            filename = 'Untitled'
        self.filename = '({})'.format(filename)
        projectName = self.tabText(index).replace(self.filename, '')
        self._editor.setText(projectName)
        if not self._editor.isVisible():
            self._editor.show()

    def handleEditingFinished(self):
        """
            Get the text that is in editor use it to set projectName,
            setWindowTitle with format: prjName (filename)
        """
        index = self.currentIndex()
        if index >= 0:
            self._editor.hide()
            self.setTabText(index, '{} {}'.format(self._editor.text(),
                                                  self.filename))
            if self._tabEditor:
                projectDb = getattr(self._tabEditor, 'getProjectDB', None)
                if projectDb:
                    projectInfo = projectDb().projectInfo
                    projectInfo.projectName = self._editor.text()
                    self.window.setWindowTitle(
                        '{}: {} {}'.format(WGUI_WINDOW_TITLE,
                                           projectInfo.projectName or
                                           'Untitled',
                                           self.filename))


