# ----------------------------------------------------------------------------
# - wgui-gui-module                                                          -
# - wild cruncher:gcs cockpit system contex                                  -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_gui_module
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher gcs cockpit system contex

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from itertools import chain

from wcr_core.action.wact_action_api import WACT_ACTION_W_DB_VERIFY, \
    WACT_ACTION_W_DB_MODIFY
from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_core.controller.wctl_controller_value_provider import \
    WCTL_CONTROLLER_VALUE_PROVIDER
from wcr_core.db.whdb_shape import WhdbShapeP
from wcr_core.db_api.whdb_block_api import WhdbBlockP
from wcr_core.db_api.whdb_connection_api import WhdbConnectionP
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP, \
    WhdbModelInstanceP, WhdbOperatorInstanceP
from wcr_core.db_api.whdb_operator_api import WhdbOperatorP
from wcr_frontend.gui.wgui_clipboard import WguiClipboardHandler
from wcr_frontend.gui.wgui_gui_action import WguiProjectEditorRefreshHandler
from wcr_frontend.gui.wgui_text_editor import WGUI_SYNTAX_HIGHLIGHTERS
from wcr_frontend.gui.wgui_toolbox_item import WGUI_ICONS, WGUI_ICON_PNG32
from wcr_frontend.gui.wgui_highlight_module import WguiHighlightModule
from wcr_frontend.gui.wgui_scene_item_factory import \
    WGUI_DEFAULT_SCENE_ITEM_FACTORY
from wcr_frontend.gui.wgui_props_dialog_factory import \
    WguiPropertiesDialogFactory
from wcr_frontend.gui.wgui_scene import WguiGraphicsScene
from wcr_frontend.gui.wgui_scene_context_menu_factory import \
    WguiSceneContextMenuFactory
from wcr_core.db_api.whdb_block_api import WHDB_DEFAULT_BLOCK_CODE
from wcr_core.modules.wmod_application_module import \
    WMOD_EMPTY_APPLICATION_MODULE, WmodApplicationModuleObserverP, \
    WMOD_ACTION_ADDED_EVENT
from silkan_common.scmn_mime_type import ScmnMimeTypeP
from wcr_frontend.gui.wgui_scene_item_ids import *
from wcr_frontend.gui.wgui_verifier import WguiDbVerificationFailureHandler


#: Registered cockpit contexts
_WGUI_COCKPIT_CONTEXTS = OrderedDict()


#: Section identifier for the 'project' action section in the menu
WGUI_ACTION_SECTION_PROJECT = 'project'

#: Section identifier for the 'project' action section in the menu
WGUI_ACTION_SECTION_IMPORT = 'import'

#: Section identifier for the toolbar action section
WGUI_ACTION_SECTION_TOOLBAR = 'toolbar'


# Icon kinds
WGUI_NO_ICON_KIND, \
    WGUI_ICON_MIME_KIND, \
    WGUI_ICON_ACTION_KIND, \
    WGUI_ICON_OPERATOR_KIND = ((2 ** i >> 1) for i in range(4))


# TODO: Fallback for models, operators, ...
WGUI_DEFAULT_ICON_CODES = {
    WhdbBlockP: WHDB_DEFAULT_BLOCK_CODE,
    WhdbBlockInstanceP: lambda x: (x.getBlock().getEntityType() or
                                   WHDB_DEFAULT_BLOCK_CODE),
    WhdbModelInstanceP: lambda x: x.getModel().getMedia(),
    WhdbOperatorInstanceP: lambda x: x.getOperator().getCode(),
    WhdbConnectionP: CNCT_ITEM_ID,
    WhdbOperatorP: lambda x: x.getCode(),
    WhdbShapeP: lambda x: 'wcr.shape.' + x.shapeName,
    ScmnMimeTypeP: lambda x: x.mimeType,
    WguiHighlightModule: lambda x: 'wcr.highlighter.' + x.identifier
}
# TODO: Icons provided by operators


class WguiIconPathFactoryP(metaclass=ABCMeta):
    @abstractmethod
    def getIconPath(self, entity, iconType):
        """
        Get the icon image for the given code. The code can be, among others,
        be an operator code or a MIME-type identifier

        :param entity: Code to obtain icon image path for
        :type entity: object
        :param iconType: Type of icon requested
        :param iconType: int
        :return: Icon image path for the given code
        :rtype: str
        """
        raise NotImplementedError()


class WguiEmptyIconPathFactory(WguiIconPathFactoryP):
    """
    Dummy icon-path factory returning `None` for every request.
    """

    def getIconPath(self, entity, iconType):
        return None


# noinspection PyAbstractClass
class WguiCockpitSystemModuleP(WguiIconPathFactoryP):
    @property
    @abstractmethod
    def code(self):
        """
        Get the context's code/identifier.

        This function should usually return the same value as
        `self.getAppModule().code`.

        :return: Identifier of the context
        :rtype: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def name(self):
        """
        Get the context's name.

        This function should usually return the same value as
        `self.getAppModule().name`.

        :return: Name of the context
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getAppModule(self):
        """
        Get the application module.

        :return: Application module
        :rtype: WmodApplicationModuleP
        """
        raise NotImplementedError()

    @abstractmethod
    def getActions(self):
        """
        Create a new set of editor actions for this context using the given
        action context.

        :return: Dictionary of editor actions for this context, with their
                 names
        :rtype: Iterable[(WactActionP, str, str, tuple[str])]
        """
        raise NotImplementedError()

    @abstractmethod
    def createContext(self, pctx, pcfg, actions, actx, parent):
        """
        Create a new cockpit system context.

        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param pcfg: Project editor configuration
        :type pcfg: WguiProjectEditorCfg
        :param actions: GUI actions
        :type actions: dict[str, WactActionP]
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param parent: Parent widget
        :return: Newly created cockpit system context
        :rtype: WguiCockpitSystemContextP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WguiCockpitSystemModuleMP(WguiCockpitSystemModuleP):
    """
    Protocol for mutable cockpit system modules.
    """
    @abstractmethod
    def addSyntaxHighlighter(self, shlg, code):
        """
        Register a new syntax highlighter for text editors.

        :param shlg: Syntax highlighter (constructor)
        :type shlg: () -> QSyntaxHighlighter
        :param code: MIME-type of compatible files
        :type code: str
        """
        raise NotImplementedError()

    @abstractmethod
    def addAction(self, action, code, source, *flags):
        """
        Add an action.

        :param action: Action to be added
        :type action: WactActionP
        :param code: Action code
        :type code: str
        :param source: Action source, e.g. code of plugin
        :type source: str
        :param flags: Action flags
        :type flags: tuple[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def addIcon(self, icon, code, source, kind):
        """
        Add an icon.

        :param icon: Icon path
        :type icon: str
        :param code: Icon code (e.g. MIME-type, operator-code, ...)
        :type code: str
        :param source: Icon source, e.g. code of plugin
        :type source: str
        :param kind: Icon kind, e.g. `WGUI_ICON_MIME_KIND`
        :type kind: int
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WguiCockpitSystemModuleBase(WguiCockpitSystemModuleMP,
                                  WmodApplicationModuleObserverP):
    """
    Abstract base class for cockpit system modules.
    """

    #: Default icon codes
    _ICON_CODES = WGUI_DEFAULT_ICON_CODES

    def __init__(self, appModule):
        """
        :param appModule:
        :type appModule: WmodApplicationModuleP
        """
        self.__appModule = appModule

        self.__actions = []
        """:type: list[(WactActionP, str, str, tuple[str)]"""

        self.__observers = self._createActionObservers()
        """:type: dict[class,
                    list[(WactActionObserverP,
                      list[(WactActionP, str, str, tuple[str]) -> bool)]]]"""

        self.__icons = {}
        """:type: dict[str, str]"""

        appModule.registerObserver(self)

        for atpl in appModule.getActions():
            self._registerActionObservers(*atpl)

    def _createActionObservers(self):
        """
        Create action observers.

        The observers are given together with test functions that determine
        to which actions the observer should be registered to.

        :return: Action observers and test functions, by observer class
        :rtype: dict[class,
                     (WactActionObserverP,
                      list[(WactActionP, str, str, tuple[str]) -> bool)]]
        """
        return {
            WguiDbVerificationFailureHandler: (
                WguiDbVerificationFailureHandler(),
                [lambda a, c, s, fs: WACT_ACTION_W_DB_VERIFY in fs]),
            WguiProjectEditorRefreshHandler: (
                WguiProjectEditorRefreshHandler(),
                [lambda a, c, s, fs: WACT_ACTION_W_DB_MODIFY in fs])}

    def _registerActionObservers(self, action, code, src, flgs):
        """
        Register the appropriate observers to the given action.

        Which observers are registered is primarily determined by the given
        action flags.

        :param action: Action to register observers to
        :type action: WactActionP
        :param code: Action code
        :type code: str
        :param src: Action source
        :type src: str
        :param flgs: Action flags
        :type flgs: tuple[str]
        """
        for obsrv, tstfs in self.__observers.values():
            for tstf in tstfs:
                if tstf(action, code, src, flgs):
                    action.registerActionObserver(obsrv)
                    break

    @property
    def code(self):
        return self.__appModule.code

    @property
    def name(self):
        return self.__appModule.name

    def getAppModule(self):
        return self.__appModule

    def addSyntaxHighlighter(self, shlg, code):
        WGUI_SYNTAX_HIGHLIGHTERS[code] = shlg

    def addAction(self, action, code, source, *flags):
        code = '{}.{}'.format(source, code)
        flags += ('gui',)
        self.__actions.append((action, code, source, flags))

        self._registerActionObservers(action, code, source, flags)

    def getActions(self):
        return chain(self.__appModule.getActions(), self.__actions)

    def getIconPath(self, entity, iconType):
        if isinstance(entity, WhdbOperatorInstanceP):
            entity = entity.getOperator()
        if isinstance(entity, WhdbOperatorP):
            icon = entity.getIcon(iconType)
            if icon:
                return icon

        if iconType == WGUI_ICON_PNG32:
            if isinstance(entity, str):
                code = entity
            else:
                rawc = self._getRawIconCode(entity)
                code = self._resolveIconCode(rawc, entity)
            return self.getIconForCode(code, iconType)
        else:
            return None

    def _getRawIconCode(self, entity):
        return wgui_getClassEntry(entity, self._ICON_CODES)

    # noinspection PyUnusedLocal
    def getIconForCode(self, code, iconType):
        """
        Get the icon path for the given icon code.

        :param code: Icon code
        :type code: str
        :param iconType: Type of icon requested
        :param iconType: int
        :return: Icon image path for the given code
        :rtype: str
        """
        return self.__icons.get(code) or WGUI_ICONS.get(code)

    def addIcon(self, icon, code, source, kind):
        if kind != WGUI_ICON_MIME_KIND:
            code = '{}.{}'.format(source, code)
        self.__icons[code] = icon

    def _resolveIconCode(self, code, entity):
        if not code:
            return None
        elif isinstance(code, str):
            return code
        elif callable(code):
            return code(entity)
        elif isinstance(code, list) or isinstance(code, tuple):
            return list(map(lambda p: self._resolveIconCode(p, entity), code))
        else:
            return None

    def appModuleNotify(self, module, event, *args):
        if event == WMOD_ACTION_ADDED_EVENT:
            atpl = args[0]
            self._registerActionObservers(*atpl)


# noinspection PyAbstractClass
class WguiCockpitSystemContextP(WguiIconPathFactoryP):
    """
    Protocol for contexts for the cockpit system.
    """

    # TODO: Turn the context into WguiCockpitSystemModuleP, and use
    # WguiCockpitSystemContextP for instances of the module

    @property
    @abstractmethod
    def code(self):
        """
        Get the context's code/identifier.

        This function should usually return the same value as
        `self.getModule().code`.

        :return: Identifier of the context
        :rtype: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def name(self):
        """
        Get the context's name.

        This function should usually return the same value as
        `self.getModule().name`.

        :return: Name of the context
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getAppModule(self):
        """
        Get the application module this cockpit context is associated with.

        :return: Associated application module
        :rtype: WmodApplicationModuleP
        """
        raise NotImplementedError()

    @abstractmethod
    def getCockpitModule(self):
        """
        Get the cockpit module this cockpit context is associated with.

        :return: Associated application module
        :rtype: WguiCockpitSystemModuleP
        """
        raise NotImplementedError()

    @abstractmethod
    def getExcludedSystemToolboxItems(self):
        """
        Get a list of system toolbox items that should not be made available
        to the user.

        :return: List of excluded system toolbox items
        :rtype: Iterable[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def getHighlightingModules(self):
        """
        Get the scene highlighting modules for the context.

        :return: Highlighting modules
        :rtype: Iterable[WguiHighlightModule]
        """
        raise NotImplementedError()

    @abstractmethod
    def getDirectoryDescription(self, level):
        """
        Get the description for directories at the given level in the
        workspace.

        :param level:
        :type level:
        :return: Description for directories at the given level
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getScene(self):
        """
        Get the graphics scene.

        :return: Graphics scene
        :rtype: WguiGraphicsScene
        """
        raise NotImplementedError()

    @abstractmethod
    def getSceneViewMouseHandler(self):
        """
        Get the scene view mouse handler.

        :return: Scene view mouse handler
        :rtype: WguiSceneViewMouseHanderP
        """
        raise NotImplementedError()

    @abstractmethod
    def getClipboardHandler(self):
        """
        Get the clipboard handler for graphic scene items.

        :return: Clipboard handler
        :rtype: WguiClipboardHandler
        """
        raise NotImplementedError()

    @abstractmethod
    def getSceneItemFactory(self):
        """
        Get the scene item factory for this context.

        :return: Scene item factory
        :rtype: WguiSceneItemFactoryP
        """
        raise NotImplementedError()

    @abstractmethod
    def getContextMenuFactory(self):
        """
        Get the context menu factory for this context.

        :return: Context menu factory
        :rtype: WguiSceneContextMenuFactoryP
        """
        raise NotImplementedError()

    @abstractmethod
    def getSidebarModules(self):
        """
        Get the sidebar modules/widgets for this context.

        :return: Sidebar widgets, and their identifiers and titles
        :rtype: Iterable[(str, str, QWidget)]
        """
        raise NotImplementedError()

    @abstractmethod
    def getPropertiesDialogFactory(self):
        """
        Get the properties dialog factory for the context.

        :return: Properties dialog factory
        :rtype: WguiPropertiesDialogFactoryP
        """
        raise NotImplementedError()

    @abstractmethod
    def getControllerValueProvider(self):
        """
        Get the controller value provider.

        :return: Controller value provider
        :rtype: WctlControllerValueProviderP
        """
        raise NotImplementedError()


def wgui_registerCockpitModule(module):
    """
    Register a cockpit context.

    :param module: Cockpit context to register.
    :type module: WguiCockpitSystemModuleP
    """
    _WGUI_COCKPIT_CONTEXTS[module.getAppModule().code] = module


def wgui_getCockpitModules():
    """
    Get a list of all registered cockpit contexts.

    :return: Registered cockpit contexts
    :rtype: Iterable[WguiCockpitSystemModuleP]
    """
    return iter(_WGUI_COCKPIT_CONTEXTS.values())


def wgui_getCockpitModule(code):
    """
    Get the cockpit context for the given code, if it exists.

    :param code: Code of cockpit context to return
    :type code: str
    :return: Cockpit module for the given code
    :rtype: WguiCockpitSystemModuleP | None
    """
    return _WGUI_COCKPIT_CONTEXTS.get(code)


def wgui_getClassEntry(item, cdct):
    """
    Retrieve dictionary value for the class of the given item.

    :param item: Item to use as key
    :type item: object
    :param cdct: Dictionary
    :type cdct: dict[class, object]
    :return: Value for the item's class, if any
    :rtype: object | None
    """
    clazz = item.__class__
    code = cdct.get(clazz)
    if code:
        return code

    # noinspection PyArgumentList
    for p in item.__class__.mro():
        code = cdct.get(p)
        if code:
            cdct[clazz] = code
            break

    return code


# noinspection PyAbstractClass
class WguiCockpitSystemContextBase(WguiCockpitSystemContextP, WcmnReleasableP):
    _ICON_CODES = WGUI_DEFAULT_ICON_CODES

    def __init__(self, cmod):
        """
        :param cmod: Cockpit system module
        :type cmod: WguiCockpitSystemModuleP
        """
        self.__module = cmod

    def release(self):
        pass

    @property
    def code(self):
        return self.__module.code

    @property
    def name(self):
        return self.__module.name

    def getAppModule(self):
        return self.__module.getAppModule()

    def getCockpitModule(self):
        return self.__module

    def getIconPath(self, entity, iconType):
        return self.__module.getIconPath(entity, iconType)

    def getControllerValueProvider(self):
        return WCTL_CONTROLLER_VALUE_PROVIDER


# noinspection PyAbstractClass
class _WguiCockpitSystemDefaultContext(WguiCockpitSystemContextBase):
    """
    Dummy cockpit context.
    """

    def __init__(self, cmod, pctx, pcfg, actions, actx, parent):
        """
        :param cmod: Cockpit system module
        :type cmod: WguiCockpitSystemModuleP
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param pcfg: Project editor configuration
        :type pcfg: WguiProjectEditorCfgP
        :param actions: Actions
        :type actions: dict[str, WactActionP]
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(cmod)

        self.__scene = WguiGraphicsScene(
            pctx.getProjectDB(), pcfg, actions, pctx, self, actx, parent)
        self.__propDiagF = WguiPropertiesDialogFactory(
            pctx, pcfg, self.__scene, actions, parent)
        self.__clpbrdHandler = WguiClipboardHandler(
            self.__scene, cmod.getAppModule())
        self.__cntxtMenuF = WguiSceneContextMenuFactory(
            self.__scene, actions, self.__clpbrdHandler, pctx,
            self.__propDiagF)

    def getExcludedSystemToolboxItems(self):
        return None

    def getDirectoryDescription(self, level):
        return 'directory'

    def getScene(self):
        return self.__scene

    def getClipboardHandler(self):
        return self.__clpbrdHandler

    def getSceneItemFactory(self):
        return WGUI_DEFAULT_SCENE_ITEM_FACTORY

    def getContextMenuFactory(self):
        return self.__cntxtMenuF

    def getHighlightingModules(self):
        return []

    def getSidebarModules(self):
        return []

    def getPropertiesDialogFactory(self):
        return self.__propDiagF


class _WguiCockpitDefaultModule(WguiCockpitSystemModuleBase):
    def __init__(self):
        super().__init__(WMOD_EMPTY_APPLICATION_MODULE)

    def createContext(self, pctx, pcfg, actions, actx, parent):
        return _WguiCockpitSystemDefaultContext(
            self, pctx, pcfg, actions, actx, parent)


#: Default cockpit system context instance
WGUI_COCKPIT_DEFAULT_MODULE = _WguiCockpitDefaultModule()

wgui_registerCockpitModule(WGUI_COCKPIT_DEFAULT_MODULE)
