# ----------------------------------------------------------------------------
# - wgui-highlight-module                                                    -
# - wild cruncher - cockpit: database element highlighting                   -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 clemens oertel                                   -
# - copyright (c) 2013-2013 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_highlight_module
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher cockpit database element highlighting

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod


class WguiHighlightModule(metaclass=ABCMeta):
    """
    Base class for highlighting modules.
    """

    #: 'Highlight includes line colour' flag
    HLGHT_LINE_COLOUR = 1

    #: 'Highlight includes line width' flag
    HLGHT_LINE_WIDTH = 2

    #: 'Highlight includes fill colour' flag
    HLGHT_FILL_COLOUR = 4

    @abstractmethod
    def providesHighlight(self, hType):
        """
        Whether this highlighting modules includes the given highlighting
        types.

        `hType` should be one or multiple of
        :py:data:`HLGHT_LINE_COLOUR`, :py:data:`HLGHT_LINE_WIDTH`, and
        :py:data:`HLGHT_FILL_COLOUR`.

        :param hType: Highlighting type in question
        :type hType: int
        :return: Whether this module provides highlighting information for the
                 given highlighting type
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def highlight(self, item):
        """
        Provide the highlighting for the given database item.

        :param item: Database item to provide highlighting hints for
        :type item: WhdbBaseP
        :return: Highlighting hints for the given item: line colour, line
                 width and fill colour
        :rtype: int, float, int
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def identifier(self):
        """
        Get this highlighting module's identifier.

        May return `None` if the module is inactive.

        :return: Identifier
        :rtype: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def description(self):
        """
        Get this highlighting module's description.

        :return: Description
        :rtype: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WguiToggleHighlightModule(WguiHighlightModule):
    @abstractmethod
    def enableHighlighting(self, enable):
        """
        Enable the highlighting of this module.

        :param enable: Whether to enable the module
        :type enable: bool
        """
        raise NotImplementedError()


class WguiItemHighlighter(object):
    """
    Database element highlighter.

    Provides highlighting hints for rendering database elements, based on
    the output of registered highlighting modules.
    """

    def __init__(self, order=None):
        """
        :param order: Module identifiers in their proper order
        :type order: [str]
        """
        #: Highlighting modules
        self.__modules = []
        #: Highlighting order
        self.__order = order
        #: Stack of primary highlighting modules
        self.__primaryModuleStack = []

    def _getModule(self, identifier):
        """
        Get the module with the given identifier, if any.

        :param identifier: Module identifier to look for
        :type identifier: str
        :return: Module with the given identifier, if any
        :rtype: WguiHighlightModule
        """
        return next(filter(lambda x: x.identifier == identifier,
                           self.__modules),
                    None)

    def addModule(self, module):
        """
        Add a highlighting module.

        :param module: Module to be added
        :type module: WguiHighlightModule
        """
        self.__modules.append(module)

        if self.__order:
            def keyFunc(x):
                try:
                    self.__order.index(x.identifier)
                except ValueError:
                    return len(self.__order) + 1

            self.__modules.sort(key=keyFunc)

    def removeModule(self, identifier):
        """
        Remove a highlighting module.

        :param identifier: Identifier of module to be removed
        :type identifier: str
        """
        module = self._getModule(identifier)

        if module is not None:
            self.__modules.remove(module)
            self.__primaryModuleStack.remove(module)

    def setPrimaryModule(self, primary):
        module = self._getModule(primary)
        if module is not None:
            self.__primaryModuleStack.append(module)

    def unsetPrimaryModule(self, primary):
        module = self._getModule(primary)
        if module is not None:
            self.__primaryModuleStack.remove(module)

    @property
    def _primaryModule(self):
        """
        Get the current primary module, if any.

        :return: Current primary module, if any
        :rtype: WguiHighlightModule
        """
        if self.__primaryModuleStack:
            return self.__primaryModuleStack[-1]
        else:
            return None

    def highlight(self, item):
        """
        Provide the highlighting for the given database item.

        :param item: Database item to provide highlighting hints for
        :type item: WhdbBaseP
        :return: Highlighting hints for the given item: line colour, line
                 width and fill colour
        :rtype: int, float, int
        """

        def anyMissing():
            return (lineColour is None and needLineWidth) or \
                   (lineWidth is None and needLineWidth) or \
                   (fillColour is None and needFillColour)

        lineColour = None
        needLineColour = True
        lineWidth = None
        needLineWidth = True
        fillColour = None
        needFillColour = True

        p = self._primaryModule
        if p is not None:
            hlghts = p.highlight(item)

            if hlghts is not None:
                lineColour, lineWidth, fillColour = hlghts
                needLineColour = not p.providesHighlight(p.HLGHT_LINE_COLOUR)
                needLineWidth = not p.providesHighlight(p.HLGHT_LINE_WIDTH)
                needFillColour = not p.providesHighlight(p.HLGHT_FILL_COLOUR)

        if anyMissing():
            for module in self.__modules:
                hlghts = module.highlight(item)

                if hlghts is not None:
                    lc, lw, fc = hlghts
                    if lineColour is None and needLineColour: lineColour = lc
                    if lineWidth is None and needLineWidth: lineWidth = lw
                    if fillColour is None and needFillColour: fillColour = fc

                    if not anyMissing():
                        break

        return lineColour, lineWidth, fillColour
