# ----------------------------------------------------------------------------
# - wgui_icon_provider                                                       -
# - wild cruncher: icon provider                              frontend / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 samuel longchamps                                -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_icon_provider
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Icon provider

.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
from abc import ABCMeta, abstractmethod


class IconPrvP(metaclass=ABCMeta):
    """
    Protocol for an icon provider
    """

    @abstractmethod
    def getIcon(self, code):
        """
        Retrieve an icon linked to the passed code

        :param code: Key value linked to the icon
        :type code: str
        :return: Icon if found, None otherwise
        :rtype: QIcon | None
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class IconPrvMP(IconPrvP):
    """
    Mutable protocol for an icon provider
    """

    @abstractmethod
    def registerIcon(self, code, icon):
        """
        Register an icon with the passed code

        :param code: Key value linked to the icon
        :type code: str
        :param icon: Icon to be linked with the passed code
        :type icon: QIcon
        """
        raise NotImplementedError()

    @abstractmethod
    def unregisterIcon(self, code):
        """
        Unregister the icon linked to the passed code

        :param code: Key value linked to the icon
        :type code: str
        """
        raise NotImplementedError()


class IconPrvCached(IconPrvMP):
    """
    Icon provider class which allows to retrieve QIcon objects using registered
    code effectively using a dictionary as data structure (used as a cache).
    """

    def __init__(self):
        self._cachedIcon = {}
        """:type: dict[str, QIcon]"""

    def getIcon(self, code):
        """
        Retrieve an icon linked to the passed code.

        :param code: Key value linked to the icon
        :type code: str
        :return: Icon if found, None otherwise
        :rtype: QIcon | None
        """
        return self._cachedIcon.get(code)

    def registerIcon(self, code, icon):
        """
        Register an icon with the passed code

        :param code: Key value linked to the icon
        :type code: str
        :param icon: Icon to be linked with the passed code
        :type icon: QIcon
        """
        self._cachedIcon[code] = icon

    def unregisterIcon(self, code):
        """
        Unregister the icon linked to the passed code

        :param code: Key value linked to the icon
        :type code: str
        """
        self._cachedIcon.pop(code, None)
