# ----------------------------------------------------------------------------
# - wgui-instance-props-dialog                                               -
# - wild cruncher - cockpit : instance properties dialog widgets             -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_instance_props_dialog
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher instance properties dialog

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod
import traceback

from PyQt4.QtGui import QLabel, QLineEdit, QVBoxLayout, QHBoxLayout, \
    QPushButton, QIcon, QComboBox, QFormLayout, QWidget

from wcr_core.common.wcmn_file_handler import WcmnFile
from wcr_core.db.whdb_model import whdb_getMimeForModelFile
from wcr_core.db.whdb_utils import whdb_getEntryPoints
from silkan_common.scmn_mime_type import SCMN_MIMES
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP
from wcr_core.db_api.whdb_protocols_api import WhdbHasCustomIconP
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP, \
    WguiFlipRotationProperties, WguiInstanceBaseProperties, \
    WguiPropertiesDialog, WguiXPointProperties, \
    WguiMultiPropertiesWidgetP, wgui_setActiveXPoint, \
    WguiXPointPropertiesBase, WguiCustomIconProperties, \
    WguiGroupedMultiPropertiesWidgetP
from wcr_frontend.gui.wgui_utils import wgui_getOpenFile
from silkan_common.scmn_file_path import scmn_path_dir, scmn_expshenv
from silkan_common.scmn_log import SCMN_LOG_ERROR


class WguiInstancePropertiesDialog(WguiPropertiesDialog):
    """
    The WguiInstancePropertiesDialog class is a dialog design to access
    and manipulate instance data. This include name, info and connectors.
    """

    def __init__(self, inst, pctx, updateF, selectXpnt=None, parent=None):
        """
        :param inst: the instance object
        :type inst: WhdbInstanceP
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param updateF: Callback function to notify about saves
        :type updateF: () -> _
        :param parent: Parent widget
        """
        name = None
        if isinstance(inst, WhdbBlockInstanceP):
            block = inst.getBlock()
            btyp = pctx.getModule().getBlockType(block.getEntityType())
            if btyp:
                name = '{} Instance Properties'.format(btyp.getName().title())
        if not name:
            name = 'Instance Properties'

        super().__init__(name, updateF, parent=parent)

        # noinspection PyListCreation
        widgets = []

        tophm = pctx.getProjectDB().getTopHierarchy()

        instWdgt = WguiInstanceBaseProperties('Instance', inst, tophm)
        addiWdgts = self._createAdditionalInstancePropertiesWidget(inst, tophm)
        if addiWdgts:
            generalWdgt = WguiMultiPropertiesWidgetP(instWdgt, *addiWdgts)
        else:
            generalWdgt = instWdgt

        widgets.append(('General', generalWdgt))
        for wdgt in self._createAdditionalPropertiesWidget(
                inst, tophm, pctx):
            widgets.append(wdgt)
        if isinstance(inst, WhdbHasCustomIconP):
            # noinspection PyTypeChecker
            appw = WguiGroupedMultiPropertiesWidgetP(
                ('Rotation / Orientation', WguiFlipRotationProperties(inst)),
                ('Custom Icon',
                 WguiCustomIconProperties(inst, pctx.getProjectDB().tops)))
        else:
            appw = WguiFlipRotationProperties(inst)
        widgets.append(('Appearance', appw))

        activate = None
        if selectXpnt:
            for _, wdgt in widgets:
                if isinstance(wdgt, WguiXPointPropertiesBase):
                    activate = wdgt
                    wgui_setActiveXPoint(selectXpnt, wdgt)

        self.setTabbedPropertyWidgets(*widgets, activate=activate)
        self.setMinimumWidth(500)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _createAdditionalInstancePropertiesWidget(self, inst, tophm):
        return []

    # noinspection PyMethodMayBeStatic
    def _createAdditionalPropertiesWidget(self, inst, tophm, pctx):
        return [('Connection Points', WguiXPointProperties(inst))]


class WguiXPointPropsEntryPointProviderP(metaclass=ABCMeta):
    """
    Model path/entry-point provider.
    """

    @property
    @abstractmethod
    def entryPointChanged(self):
        """
        Signal for entry-ponit changes.

        :rtype: pyqtSignal
        """
        raise NotImplementedError()

    @abstractmethod
    def getPathEntryPoint(self):
        """
        Get the current model path and entry point.

        :return: Model part and entry point
        :rtype: (str, str)
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WguiModelPathEPointProperties(WguiPropertiesWidgetP,
                                    WguiXPointPropsEntryPointProviderP):
    """
    Property editing widget for an instance's model path and entry-point.
    """

    def __init__(self, inst, tophm, pctx, parent=None):
        """
        :param inst: Model instance
        :type inst: WhdbModelInstanceP
        :param tophm: Hierarchy top level
        :type tophm: WhdbBlock
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param parent: Parent Widget
        """
        super().__init__(parent)

        #: Instance object
        self.__instance = inst
        #: Current model block
        self.__tophm = tophm
        #: Application context
        self.__prjContext = pctx

        #: Path field
        self.__pathF = QLineEdit(inst.getModel().getPath())
        #: Function name field
        self.__entryPointF = QComboBox()
        #: File button
        self.__fileButton = QPushButton()

        self._initElements()
        self.setLayout(self._createLayout())
        self.updateFields()

    def _initElements(self):
        self.__entryPointF.setAutoCompletion(True)
        self.__entryPointF.setEditable(True)

        self.__pathF.textChanged.connect(self._updateEntryPoints)

        self.__fileButton.setIcon(QIcon(':/img/app_was_open.png'))
        self.__fileButton.clicked.connect(self._selectPath)

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QFormLayout
        """
        layout = QFormLayout()
        layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)

        pathLayout = QHBoxLayout()
        pathLayout.setMargin(0)
        pathLayout.addWidget(self.__pathF)
        pathLayout.addWidget(self.__fileButton)
        wdgt = QWidget()
        wdgt.setLayout(pathLayout)
        layout.addRow(QLabel('Model Path:'), wdgt)

        layout.addRow(QLabel('Entry-Point Name:'), self.__entryPointF)

        return layout

    @property
    def entryPointChanged(self):
        return self.__entryPointF.editTextChanged

    def getPathEntryPoint(self):
        return self.__pathF.text(), self.__entryPointF.currentText()

    def _selectPath(self):
        # noinspection PyTypeChecker
        path = wgui_getOpenFile(
            'Select Model File', self,
            *list(filter(
                None, (SCMN_MIMES.get(m) for m in self._getModelMimeTypes()))),
            idir=scmn_path_dir(scmn_expshenv(self.__pathF.text())))
        if path:
            self.__pathF.setText(path)

    def updateFields(self):
        path = self._getModelPath()
        self.__pathF.setText(path.path if isinstance(path, WcmnFile) else path)
        self._updateEntryPoints()

    def _updateEntryPoints(self, path=None):
        path = scmn_expshenv(path)
        self.__entryPointF.clear()
        self.__entryPointF.addItem('')

        try:
            epnt = self._getModelEntryPoint()
            if not path:
                path = scmn_expshenv(self._getModelPath())
            mime = whdb_getMimeForModelFile(path)

            for e in whdb_getEntryPoints(path, mime):
                self.__entryPointF.addItem(e)
                if e == epnt:
                    self.__entryPointF.setCurrentIndex(
                        self.__entryPointF.count() - 1)

            if self.__entryPointF.currentIndex() <= 0 and epnt:
                self.__entryPointF.addItem(epnt)
                self.__entryPointF.setCurrentIndex(
                    self.__entryPointF.count() - 1)
        except Exception as e:
            self.__prjContext.addMessage(
                mtxt='Could not determine entry points for {} ({})'.format(
                    self._getModelPath(), e),
                mcmp='cockpit', mcat=SCMN_LOG_ERROR,
                trce=traceback.format_exc())

    def save(self):
        self._setModelPathEntryPoint(
            self.__pathF.text(), self.__entryPointF.currentText())

    def _getModelPath(self):
        return self.__instance.getModel().getPath()

    def _getModelEntryPoint(self):
        return self.__instance.getModel().getEntryPoint()

    def _getModelMimeTypes(self):
        return self.__instance.getModel().getMedia(),

    def _setModelPathEntryPoint(self, path, epnt):
        model = self.__instance.getModel()
        model.setPath(path)
        model.setEntryPoint(epnt)


# noinspection PyAbstractClass
class WguiModelProperties(WguiModelPathEPointProperties):
    """
    Property editing widget for an instance's model information.
    """

    def __init__(self, inst, tophm, pctx, parent=None):
        """
        :param inst: Model instance
        :type inst: WhdbModelInstanceP
        :param tophm: Hierarchy top level
        :type tophm: WhdbBlock
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param parent: Parent Widget
        """
        super().__init__(inst, tophm, pctx, parent)

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        warnLabel = QLabel(
            '<b>Changing the model path and/or entry point name may cause '
            'all current connectors of the instance and all wires attached to '
            'it to be removed.</b>')
        warnLabel.setWordWrap(True)
        layout.addWidget(warnLabel)
        layout.addSpacing(10)

        modelLayout = super()._createLayout()
        # noinspection PyUnresolvedReferences
        modelLayout.setMargin(0)
        layout.addLayout(modelLayout)

        layout.addStretch(1)

        return layout
