# ----------------------------------------------------------------------------
# - wgui-main-window                                                         -
# - wild cruncher - cockpit : main window                                    -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2012-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_main_window
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher main window

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------
from functools import partial

from PyQt4.QtGui import QMainWindow, QKeySequence, QIcon, QAction, \
    QCloseEvent, QMessageBox, QDialog, QVBoxLayout, QComboBox, QLabel, \
    QDialogButtonBox

from wcr_core.action.wact_action_parameter import WactActionParameter, \
    WactFixedMultiValueActionParameter
from wcr_core.action.wact_action_parameter_api import \
    WactFixedSelectionActionParameterP
from wcr_core.action.wact_action_parameter_type_api import \
    WactStringableActionParameterTypeP
from wcr_core.actions.wact_actions import WACT_PROJECT_ACTIONS
from wcr_frontend.gui.wgui_icon_provider import IconPrvCached
from wcr_frontend.gui.wgui_param_value_source import WguiParameterValueSource
from wcr_frontend.gui.wgui_action_execution_context import \
    WguiActionExecutionContextBase
from wcr_frontend.gui.wgui_gui_module import wgui_getCockpitModules, \
    wgui_getCockpitModule, WGUI_COCKPIT_DEFAULT_MODULE, WguiIconPathFactoryP
from wcr_core.modules.wmod_application_module import wmod_getModuleForPdb
from wcr_frontend.gui.wgui_gui_action import WguiActionAdapter, \
    wgui_wrap_action
from wcr_frontend.gui.wgui_about_dialog import WguiAboutDialog
from wcr_frontend.gui.wgui_toolbox_item import WGUI_ICON_PNG32
from wcr_frontend.gui.wgui_editor import wgui_create_editor_actions, \
    WguiEditorHasSaveP
from wcr_frontend.gui.wgui_global_props_dialog import \
    WguiGlobalPropertiesDialog
from wcr_frontend.gui.wgui_project_actions import wgui_create_pedit_actions
from wcr_frontend.gui.wgui_project_editor_config import *
from wcr_frontend.gui.wgui_utils import wgui_selectAndHandleFile, \
    wgui_qcb_get_data
from wcr_frontend.gui.wgui_logger import WguiLoggerWindow, WGUI_LOGGER_IDENT
from wcr_frontend.gui.wgui_editor_tab_widget import WguiTabsWindow, \
    wgui_create_wtabs_actions
from wcr_core.action.wact_action import WactActionBase, WactSimpleAction, \
    WACT_ACTION_EXIT_CANCELLED
from wcr_core.project.wprj_project import wprj_load_project, \
    wprj_import_project, wprj_createProjectDB
from wcr_core.project.wprj_xmlio import WPRJ_SILKAN_XML_MIME_TYPE, \
    WPRJ_HARDWARE_SILKAN_XML_MIME_TYPE, WPRJ_MESSAGE_SILKAN_XML_MIME_TYPE, \
    WPRJ_LOGICAL_SILKAN_XML_MIME_TYPE, WPRJ_SYS_MAP_SILKAN_XML_MIME_TYPE, \
    WPRJ_VLINK_SILKAN_XML_MIME_TYPE, WPRJ_ICD_SILKAN_XML_MIME_TYPE
from wcr_core.common.wcmn_file_handler import WcmnMimeFileHandlerBase, \
    SCMN_MIMES, WcmnFile, wcmn_handleFile
from wcr_core.common.wcmn_mime_type import WCMN_PAYLOAD_MODEL, \
    WCMN_PAYLOAD_BLOCK
from silkan_common.scmn_mime_type import SCMN_PAYLOAD_TEXT
from wcr_frontend.gui.wgui_project_editor import WguiProjectEditor
from wcr_frontend.gui.wgui_text_editor import WguiTextEditor
from silkan_common.scmn_file_path import scmn_path_base, scmn_isfile
from silkan_common.scmn_log import ScmnLogListener, ScmnLogMessage, \
    SCMN_LOG_ERROR, ScmnLogger
from wcr_core.action.wact_action_parameter_types import \
    WactActionOpenFileByHandlerParameterType
from wcr_frontend.gui.wgui_verifier import WguiDbVerificationFailureHandler
from wcr_frontend import WGUI_WINDOW_TITLE


# ----------------------------------------------------------------------------
# - class section                                                            -
# ----------------------------------------------------------------------------

class WguiMainWindow(QMainWindow):
    #: Action key sequences
    _ACTION_KEY_SEQUENCES = None

    #: Action icons
    _ACTION_ICONS = None

    def initActionAttributes(self):
        """
        Initialize action sequences and icons.

        Cannot be done statically because the Qt application must be
        initialized prior.
        """
        if self._ACTION_KEY_SEQUENCES is None:
            self._ACTION_KEY_SEQUENCES = {
                # TODO: 'main.newp': QKeySequence.New,
                'main.open': QKeySequence.Open,
                'main.prefs': QKeySequence.Preferences,
                'main.quit': QKeySequence.Quit,
                'editor.save': QKeySequence.Save,
                'editor.saveas': QKeySequence.SaveAs,
                'editor.selAll': QKeySequence.SelectAll,
                'editor.cut': QKeySequence.Cut,
                'editor.copy': QKeySequence.Copy,
                'editor.paste': QKeySequence.Paste,
                'editor.delete': QKeySequence.Delete,
                # 'editor.print': QKeySequence.Print,
                'editor.find': QKeySequence.Find,
                'editor.findagain': QKeySequence.FindNext,
                'editor.refresh': QKeySequence('F5'),
                'project.history.backward': QKeySequence('Alt+Shift+Left'),
                'project.history.forward': QKeySequence('Alt+Shift+Right'),
                'project.zoomin': QKeySequence('Ctrl++'),
                'project.zoomout': QKeySequence('Ctrl+-'),
                'project.zoomfit': QKeySequence('Ctrl+_'),
                'project.zoom100': QKeySequence('Ctrl+='),
                'project.dump': QKeySequence('Alt+Shift+D'),
            }

        if self._ACTION_ICONS is None:
            self._ACTION_ICONS = {
                'main.newp': QIcon(':/img/app_was_xnew.png'),
                'main.open': QIcon(':/img/app_was_open.png'),
                'editor.save': QIcon(':/img/app_was_save.png'),
                'editor.cut': QIcon(':/img/app_was_xcut.png'),
                'editor.copy': QIcon(':/img/app_was_copy.png'),
                'editor.paste': QIcon(':/img/app_was_xpst.png'),
                'editor.delete': QIcon(':/img/app_was_xdel.png'),
                # 'editor.print': QIcon(':/img/app_was_prnt.png'),
                'wcr.core.System_Mapper.generateconffile':
                    QIcon(':/img/app_was_comp.png'),
            }

    def __init__(self, appContext, ignoreCloseEvent=False):
        """
        :param appContext: Application context
        :type appContext: WcorGlobalContext
        """
        super().__init__()
        self.initActionAttributes()
        #: Global application context
        self.__appContext = appContext
        self.__ignoreCloseEvent = ignoreCloseEvent

        #: All available action
        self.__actions = {}
        #: GUI action wrappers (stored for enabling/disabling)
        self.__guiActions = {}

        # Icon Provider
        self._iconPrv = IconPrvCached()

        #: Global configuration for project editors
        cfgStore = WguiProjectEditorDbCfgStore(
            appContext.getConfigurationService())
        self.__globalProjectEditorCfg = WguiProjectEditorCfg(
            WguiProjectEditorCfgDefaults(), store=cfgStore)

        # create a new tabs window
        self.__tabs = WguiTabsWindow(self)
        self.__tabs.editorNameChanged.connect(self._updateRecentsMenu)
        self.__loggerWindow = WguiLoggerWindow(
            self.__appContext.getMessageSource())

        # noinspection PyTypeChecker
        self.__paramValueSrc = WguiParameterValueSource(self, self._iconPrv)
        self.__actx = WguiGlobalActionExecutionContext(
            self.__appContext, self.__paramValueSrc, self.__tabs, self)

        #: File handler to open Silkan XML files
        self.__openSilkanFileHandler = WguiOpenSilkanFileHandler(
            self._openEditor, self.__globalProjectEditorCfg,
            self.__appContext, self.__actions, self.__actx,
            WPRJ_SILKAN_XML_MIME_TYPE,
            WPRJ_ICD_SILKAN_XML_MIME_TYPE,
            WPRJ_MESSAGE_SILKAN_XML_MIME_TYPE,
            WPRJ_LOGICAL_SILKAN_XML_MIME_TYPE,
            WPRJ_VLINK_SILKAN_XML_MIME_TYPE,
            WPRJ_HARDWARE_SILKAN_XML_MIME_TYPE,
            WPRJ_SYS_MAP_SILKAN_XML_MIME_TYPE,
        )
        appContext.registerFileHandler(self.__openSilkanFileHandler)

        #: File handler to open text model files
        textModelType = WCMN_PAYLOAD_MODEL | SCMN_PAYLOAD_TEXT
        self.__openTextFileHandler = WguiOpenTextFileHandler(
            self._openEditor,
            *[m for m in SCMN_MIMES.values()
              if m.payloadTypes & textModelType
              or m.mimeType.startswith('text/')])
        appContext.registerFileHandler(self.__openTextFileHandler)

        #: File handler to import block model files
        # noinspection PyTypeChecker
        self.__importFileHandler = WguiImportFileHandler(
            self._openEditor, self.__globalProjectEditorCfg,
            self.__appContext, self.__actions, self.__actx,
            *[m for m in SCMN_MIMES.values()
              if m.payloadTypes & WCMN_PAYLOAD_BLOCK])
        appContext.registerFileHandler(self.__importFileHandler)

        #: Recent files
        self.__recentsMenu = None

        self._createActions()

        self._createMenus()
        self._createToolbar()
        self.setCentralWidget(self.__tabs)
        self.statusBar().setStyleSheet(
            'QStatusBar { border-top: 2px groove gray; }'
        )

        self.__tabs.activeEditorChanged.connect(self._activeEditorChanged)
        self._activeEditorChanged(None)

        appContext.getMessageSource().addMessageListener(
            WguiLogListener(self.statusBar()))

        self.setWindowTitle(WGUI_WINDOW_TITLE)
        self.setWindowIcon(QIcon(':/img/app_was_slki.png'))

        self._updateRecentsMenu()

    @property
    def tabs(self):
        return self.__tabs

    def _activeEditorChanged(self, *_):
        """
        Callback for whenever the active editor changes, to update
        actions accordingly.
        """

        def updateActions(actions):
            for a in actions:
                menu = a.menu()
                if menu is not None:
                    updateActions(menu.actions())
                if isinstance(a, WguiActionAdapter):
                    a.update()

        updateActions(self.menuBar().actions())

        current = self.__tabs.currentEditor
        if (current and not isinstance(current, WguiLoggerWindow) and
                not isinstance(current, WguiTextEditor)):
            self.synchronizeBetweenRouteMapping(current)

        self.setWindowTitle(
            '{}: {}'.format(WGUI_WINDOW_TITLE,
                            current.getTitle() or 'Untitled')
            if current else WGUI_WINDOW_TITLE)

    def synchronizeBetweenRouteMapping(self, editor):
        from wcr_frontend.modules.sys_map.gui.wcgui_scene import \
            WcguiGraphicsScene
        from wcr_core.modules.vlink.wvmod_application_module import \
            WVMOD_VLINK_APP_CODE
        from wcr_core.modules.sys_map.wcmod_bus_file_handler \
            import WcmodVLinkFileHandler

        if (editor.getProjectDB().getModule() == WVMOD_VLINK_APP_CODE
            and not editor.reattached):
            top = editor.getTopHierarchy()
            tmpFilePath = top.getTmpFilePath()
            vlDB = editor.getProjectDB()
            handler = WcmodVLinkFileHandler(vlDB.getModule())
            if tmpFilePath and editor.isRelatedEditorDirty(self.tabs):
                loadingFile = tmpFilePath
            else:
                loadingFile = vlDB.path
            if loadingFile:
                newVlinkBlock = handler.handleFile(loadingFile,
                                                   self.__appContext)
                newInstances = tuple(newVlinkBlock.getOperatorInstances())
                # remove all the instance currently loaded
                editor.scene.removeItems()
                for inst in newInstances:
                    oper = inst.getOperator()
                    itemScene = editor.scene.createFunctionItem(oper)
                    newInst = itemScene.getDataObject()
                    newInst.setName(inst.getName())
                    params = inst.getVLParametersDict(inst.getVLType())
                    newInst.setVLType(inst.getVLType())
                    newInst.setVLParameters(params)
                    msgs = inst.getVLMessages()
                    newInst.setUUID(inst.getUUID())
                    for msg in msgs:
                        newInst.addVLMessage(msg)
                        editor.updateTab()
            editor.scene.populateTable()
        elif getattr(editor, 'scene', None):
            if isinstance(editor.scene, WcguiGraphicsScene):
                # get TableLink object from mapping view
                tableVl = editor.sideBarInst.modules[0][2]
                top = editor.getTopHierarchy()
                vlBlock = top.getVLinkBlock()
                tmpFilePath = None
                if vlBlock and editor.isRelatedEditorDirty(self.tabs):
                    tmpFilePath = vlBlock.getTmpFilePath()
                tableVl.updateTable(tmpFilePath)

    def _createActions(self):
        """
        Create all action instances.
        """
        self.__actions.update({
            'main.newp': WactSimpleAction(
                'New Project', 'Create a new Project',
                self._openNewProjectEditor),
            'main.open': _WguiOpenAction(self.__appContext,
                                         self.__openSilkanFileHandler,
                                         self.__openTextFileHandler),
            'main.pmdl': WactSimpleAction(
                'Import Model as Project',
                'Import a model as project',
                self._importProjectMdl),
            'main.prefs': WactSimpleAction(
                'P&references', 'Global preferences',
                self._showGlobalPreferencesDialog),
            'main.quit': WactSimpleAction(
                '&Quit', 'Quit the application', self.close),
            'main.fscrn': WactSimpleAction(
                '&Full Screen', 'View in full screen mode',
                self._toggleFullScreen),
            'main.newt': WactSimpleAction(
                'Text Editor', 'New text editor', self._openNewTextEditor),
            'main.sysl': WactSimpleAction(
                'System Logger', 'System logger window',
                self._activateLoggerWindow),
            'main.about': WactSimpleAction(
                '&About', 'About Wild Cruncher 2.0', self._showAboutDialog)
        })

        self.__actions.update(WACT_PROJECT_ACTIONS)
        self.__actions.update(wgui_create_editor_actions())
        self.__actions.update(wgui_create_pedit_actions())
        self.__actions.update(wgui_create_wtabs_actions(self.__tabs))

        for module in wgui_getCockpitModules():
            actions = module.getActions()
            self.__actions.update((code, actn) for actn, code, *_ in actions)

        self.__actions.get('project.vrfydb').registerActionObserver(
            WguiDbVerificationFailureHandler())

    def _openNewProjectEditor(self, *_):
        """
        Open a new project editor tab.
        """

        def mkEditor(cmod):
            """
            :type cmod: WguiCockpitSystemModuleP
            """
            module = cmod.getAppModule()
            module.actx = self.__actx
            pctx = module.createProjectContext(
                self.__appContext, wprj_createProjectDB(module))

            cfgTable = pctx.getProjectDB().tops.getLocalConfiguration()
            pcfg = WguiProjectEditorCfg(
                self.__globalProjectEditorCfg,
                WguiProjectEditorDbCfgStore(cfgTable))
            cctx = cmod.createContext(
                pctx, pcfg, self.__actions, self.__actx, self)
            if cctx:
                editor = WguiProjectEditor(pcfg, pctx, cctx)
                self.__tabs.openEditor(editor.getEditorIdent(), editor)

        dialog = WguiProjectTypeSelectionDialog(mkEditor, self)
        dialog.show()

    def _openNewTextEditor(self, *_):
        """
        Add a new text editor tab..
        """
        editor = WguiTextEditor(self.__appContext)
        self.__tabs.openEditor(editor.getEditorIdent(), editor)

    def _activateLoggerWindow(self, *_):
        """
        Activate the logger tab.
        """
        self.__tabs.openEditor(WGUI_LOGGER_IDENT, self.__loggerWindow)

    def _openFile(self, *_):
        """
        Open a file with the help of a file dialog. Depending on the
        extension, the appropriate handler is called.
        """
        wgui_selectAndHandleFile('Open File', self, self.__appContext,
                                 self.__openSilkanFileHandler,
                                 self.__openTextFileHandler)

    def _importProjectMdl(self, *_):
        """
        Import a block model as a project.
        """
        wgui_selectAndHandleFile('Import File', self, self.__appContext,
                                 self.__importFileHandler)

    def _showGlobalPreferencesDialog(self, *_):
        """
        Display the meta-information dialog for the current project.
        """
        diaw = WguiGlobalPropertiesDialog(
            self.__appContext.getConfigurationService(),
            self.__globalProjectEditorCfg,
            parent=self.parentWidget())
        diaw.show()
        diaw.exec_()

    def _toggleFullScreen(self, *_):
        """
        Toggle the fullscreen mode.
        """
        action = self.__guiActions.get('main.fscrn')

        if not self.isFullScreen():
            self.showFullScreen()
            if action is not None:
                action.setText('End &Full Screen')
                action.setToolTip('Switch to normal screen mode')
        else:
            self.showNormal()
            if action is not None:
                action.setText('&Full Screen')
                action.setToolTip('Switch to full screen mode')

    def _showAboutDialog(self, *_):
        """
        Display the application's "About" dialog.
        """
        dialog = WguiAboutDialog(self)
        dialog.show()
        dialog.exec_()

    def _findIcon(self, actionCode):
        """
        Get a Qt icon from the main window's defined action icons or, if not
        found, search through the modules' icons.

        :param actionCode: Code related to the action
        :type actionCode: str
        :return: A Qt icon related to the action code, None if cannot be found
        :rtype: QIcon | None
        """
        icon = self._ACTION_ICONS.get(actionCode)
        if icon is None:
            for module in wgui_getCockpitModules():
                if isinstance(module, WguiIconPathFactoryP):
                    iconPath = module.getIconPath(actionCode, WGUI_ICON_PNG32)
                    if iconPath is not None:
                        return QIcon(iconPath)
        return icon

    # menu creation
    def _createMenus(self):
        """
        Create the global application menu.
        """
        def registerMultiValueIcon(param, guiCode):
            """
            Register each choice in the multi value parameter, retrieving
            the icon using the convention "parametercode-choicecode"

            :param param: Mutli value parameter
            :type param: WactFixedSelectionActionParameterP
            :param guiCode: GUI-related action code
            :type guiCode: str
            """
            for choice in param.getPossibleValues():
                choiceStr = '-' + str(choice[0])

                choiceIcon = self._findIcon(
                    '{}{}'.format(guiCode, choiceStr))

                if choiceIcon is not None:
                    self._iconPrv.registerIcon(
                        '{}{}'.format(param.getCode(), choiceStr), choiceIcon)

        # noinspection PyShadowingNames
        def decorate(guiAction, iconPath=None):
            """
            Set an icon and the shortcut to the action.

            :param guiAction: gui related action
            :type guiAction: WguiActionAdapter
            :param iconPath: Specific icon path (optional)
            :type iconPath: str | None
            :return gui related action
            :rtype: WguiActionAdapter
            """
            if iconPath:
                icon = QIcon(iconPath)
            else:
                icon = self._findIcon(guiAction.code)

            # Find and register icons of multi value parameters
            for param in guiAction.action.getParameters():
                if isinstance(param,
                              WactFixedMultiValueActionParameter) and \
                   isinstance(param.getParamType(),
                              WactStringableActionParameterTypeP):
                    registerMultiValueIcon(param, guiAction.code)

            if icon is not None:
                guiAction.setIcon(icon)
                guiAction.setIconVisibleInMenu(True)
                self._iconPrv.registerIcon(guiAction.code, icon)

            keySeq = self._ACTION_KEY_SEQUENCES.get(guiAction.code)
            if keySeq is not None:
                guiAction.setShortcut(keySeq)

            return guiAction

        # noinspection PyShadowingNames
        def addAction(menu, actName, iconPath=None, module=None):
            """
            Add an action to the gui, wrapping it and decorating it as well.

            :param menu: Menu to which the action is to be added
            :type menu: QMenu
            :param actName: Action name used as a key to get the object
            :type actName: str
            :param iconPath: Path of the icon (Optional)
            :type iconPath: str
            """
            action = self.__actions.get(actName)
            if action is None:
                return
            guiActions = wgui_wrap_action(
                action, actName, self.__actx, self, module=module)
            for guiAction in guiActions:
                guiAction = decorate(guiAction, iconPath)
                menu.addAction(guiAction)
                self.__guiActions[guiAction.code] = guiAction

        importActions = []
        projectActions = []
        for module in wgui_getCockpitModules():
            for action, code, section, flags in module.getActions():
                if 'import' in flags:
                    importActions.append((code, module))
                else:
                    projectActions.append((code, module))

        # file menu
        fileMenu = self.menuBar().addMenu('&File')
        addAction(fileMenu, 'main.newp')
        addAction(fileMenu, 'main.open')
        self.__recentsMenu = fileMenu.addMenu('&Recent Files')
        self.__recentsMenu.aboutToShow.connect(self._updateRecentsMenu)
        addAction(self.__recentsMenu, 'main.rldrcnts')
        addAction(fileMenu, 'editor.save')
        addAction(fileMenu, 'editor.saveas')
        fileMenu.addSeparator()
        ximpMenu = fileMenu.addMenu('&Import')
        fileMenu.addSeparator()
        # addAction(fileMenu, 'editor.print')
        fileMenu.addSeparator()
        addAction(fileMenu, 'main.prefs')
        fileMenu.addSeparator()
        addAction(fileMenu, 'main.quit')

        # import menu
        addAction(ximpMenu, 'main.pmdl')
        for acod, module in importActions:
            addAction(
                ximpMenu, acod, module.getIconPath(acod, WGUI_ICON_PNG32),
                module=module.code)

        # edit menu
        editMenu = self.menuBar().addMenu('&Edit')
        addAction(editMenu, 'editor.cut')
        addAction(editMenu, 'editor.copy')
        addAction(editMenu, 'editor.paste')
        editMenu.addSeparator()
        addAction(editMenu, 'editor.find')
        addAction(editMenu, 'editor.findagain')
        editMenu.addSeparator()
        addAction(editMenu, 'project.selmode')
        addAction(editMenu, 'editor.selAll')
        editMenu.addSeparator()
        addAction(editMenu, 'editor.delete')

        # view menu
        viewMenu = self.menuBar().addMenu('&View')
        addAction(viewMenu, 'project.poph')
        addAction(viewMenu, 'project.popt')
        viewMenu.addSeparator()
        addAction(viewMenu, 'project.zoomin')
        addAction(viewMenu, 'project.zoomout')
        addAction(viewMenu, 'project.zoomfit')
        addAction(viewMenu, 'project.zoom100')
        addAction(viewMenu, 'editor.refresh')
        viewMenu.addSeparator()
        addAction(viewMenu, 'project.history.backward')
        addAction(viewMenu, 'project.history.forward')
        viewMenu.addSeparator()
        addAction(viewMenu, 'main.fscrn')

        # project menu
        projMenu = self.menuBar().addMenu('&Project')
        for acod, module in projectActions:
            addAction(
                projMenu, acod, module.getIconPath(acod, WGUI_ICON_PNG32),
                module=module.code)

        projMenu.addSeparator()
        addAction(projMenu, 'project.pinf')
        addAction(projMenu, 'project.vrfydb')
        addAction(projMenu, 'project.log')
        addAction(projMenu, 'project.prefresh')
        if self.__appContext.getConfigurationValue('WCR_DEBUG'):
            addAction(projMenu, 'project.dump')

        # tools menu
        toolMenu = self.menuBar().addMenu('&Tools')
        addAction(toolMenu, 'main.newt')

        # windows menu
        winsMenu = self.menuBar().addMenu('&Windows')
        addAction(winsMenu, 'main.sysl')

        # help menu
        self.menuBar().addSeparator()
        helpMenu = self.menuBar().addMenu('&Help')
        addAction(helpMenu, 'main.about')

    def _createToolbar(self):
        """
        Create the application toolbar.
        """

        # noinspection PyShadowingNames
        def addAction(toolbar, actName):
            action = self.__guiActions.get(actName)
            if action is None: return
            toolbar.addAction(action)

        # file toolbar
        self.fileToolBar = self.addToolBar('File')
        addAction(self.fileToolBar, 'main.newp')
        addAction(self.fileToolBar, 'main.open')
        addAction(self.fileToolBar, 'editor.save')
        # addAction(self.fileToolBar, 'editor.print')

        # edit toolbar
        self.editToolBar = self.addToolBar('Edit')
        addAction(self.editToolBar, 'editor.cut')
        addAction(self.editToolBar, 'editor.copy')
        addAction(self.editToolBar, 'editor.paste')
        addAction(self.editToolBar, 'editor.delete')

        # project toolbar
        projectToolBar = self.addToolBar('Project')
        for module in wgui_getCockpitModules():
            for action, code, section, flags in module.getActions():
                if 'toolbar' in flags:
                    addAction(projectToolBar, code)

    def _openEditor(self, ident, editorF, path=None):
        self.__tabs.openEditor(ident, editorF)
        if path:
            self.__appContext.addRecentlyOpenedFile(path)
            self._updateRecentsMenu()

    def _updateRecentsMenu(self):
        def openFile(path):
            self.__appContext.handleFile(path)
            self.__appContext.addRecentlyOpenedFile(path)

        menu = self.__recentsMenu
        if menu is None: return

        menu.clear()
        for recent in reversed(self.__appContext.getRecentlyOpenedFiles()):
            # noinspection PyArgumentList
            menu.addAction(QAction(
                scmn_path_base(recent), self,
                statusTip='Open ' + recent,
                triggered=partial(openFile, recent)))

    def closeEvent(self, event:QCloseEvent):
        if self.__ignoreCloseEvent: return
        # noinspection PyUnresolvedReferences
        editors = [e for e in self.__tabs.editors
                   if isinstance(e, WguiEditorHasSaveP) and e.isDirty()]
        if editors:
            diaw = QMessageBox(self)
            diaw.setWindowTitle('Confirm Quit')
            diaw.setText('Do you want to save changes made to any '
                         'project/file before closing the application?')
            diaw.setInformativeText(
                'All unsaved changes will be lost when the application '
                'is closed.')
            diaw.addButton(QMessageBox.Save)
            diaw.addButton(QMessageBox.Discard)
            diaw.addButton(QMessageBox.Cancel)
            diaw.setDefaultButton(QMessageBox.Cancel)
            diaw.setIcon(QMessageBox.Question)
            diaw.setEscapeButton(QMessageBox.Cancel)
            cnfm = diaw.exec()
            if cnfm == QMessageBox.Save:
                for idx, editor in enumerate(editors):
                    self.__tabs.setCurrentIndex(idx)
                    try:
                        # noinspection PyUnresolvedReferences
                        editor.save()
                    except:
                        QMessageBox(QMessageBox.Warning,
                                    'Error saving to file',
                                    'The contents could not be saved, '
                                    'please try to save manually.',
                                    QMessageBox.Ok,
                                    self).exec()
                        event.ignore()
                        return
            elif cnfm == QMessageBox.Cancel:
                event.ignore()
                return
        event.accept()


class _WguiOpenAction(WactActionBase):
    """
    Main action to open new editors.
    """

    def __init__(self, gctx, *handlers):
        """
        :param gctx: Application context
        :type gctx: WcorGlobalContextP
        :param handlers: File handlers
        :type handlers: list[WcmnFileHandler]
        """
        self.__parameter = WactActionParameter(
            'file', 'Project or text file', 'Project or text file to open',
            WactActionOpenFileByHandlerParameterType(*handlers))

        super().__init__(
            'Open project...', 'Open a project (or text file)',
            self.__parameter)

        #: Application context
        self.__context = gctx
        #: File handlers
        self.__handlers = handlers

    def _doExecute(self, actx, paramValuePairs):
        file = self._getParamValue(self.__parameter, actx, paramValuePairs)
        """:type: WcmnFile | None"""
        if file is None:
            return WACT_ACTION_EXIT_CANCELLED

        wcmn_handleFile(file, self.__context, *self.__handlers)
        self.__context.addRecentlyOpenedFile(file.path)


class WguiGlobalActionExecutionContext(WguiActionExecutionContextBase):
    def __init__(self, xctx, prmValSrc, tabWdgt, parent):
        """
        :param xctx: Context
        :type xctx: WcorContextP
        :param prmValSrc: Parameter value source
        :type prmValSrc: WactParameterValueSourceP
        :param tabWdgt: Editor container widget
        :type tabWdgt: WguiTabsWindow
        """
        super().__init__(xctx, prmValSrc, parent)

        self.__tabWdgt = tabWdgt

    def getProjectContext(self):
        editor = self.__tabWdgt.currentEditor
        if isinstance(editor, WguiProjectEditor):
            return editor.projectContext
        return None

    def getEditor(self):
        return self.__tabWdgt.currentEditor


class _WguiReadFileHandler(WcmnMimeFileHandlerBase):
    def __init__(self, addEditorF, cfg, gContext, actions, actx, *mimeTypes):
        """
        :param addEditorF: Function to add project editor to application
        :type addEditorF: (WguiProjectEditor) -> _
        :param cfg: Global project editor default configuration
        :type cfg: WguiProjectEditorCfgP
        :param gContext: Global context
        :type gContext: WcorGlobalContextP
        :param actions: Action dictionary
        :type actions: dict[str, WactActionP]
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param mimeTypes: Mime types handled by the file handler
        :type mimeTypes: list[ScmnMimeTypeP]
        """
        super().__init__(*mimeTypes)

        #: Function to add project editor to application
        self.__addEditorF = addEditorF
        #: Global project editor default configuration
        self.__cfg = cfg
        #: Global context
        self.__gContext = gContext
        #: Action dictionary
        self.__actions = actions
        #: Action execution context
        self.__actx = actx

    def handleFile(self, file, xctx):
        if not scmn_isfile(file.path):
            raise ValueError('No such file: ' + file.path)

        logger = ScmnLogger()
        projectDB = self._createProjectDB(
            file, xctx.createContextWith(msgTarget=logger))
        if projectDB is not None:
            def createEditor():
                nonlocal editor
                cfgTable = projectDB.tops.getLocalConfiguration()
                pcfg = WguiProjectEditorCfg(
                    self.__cfg, WguiProjectEditorDbCfgStore(cfgTable))
                cctx = cmod.createContext(
                    pctx, pcfg, self.__actions, self.__actx, parent=None)
                editor = WguiProjectEditor(pcfg, pctx, cctx)
                return editor

            editor = None

            module = wmod_getModuleForPdb(projectDB)
            if not module:
                xctx.addMessage(
                    'Could not find module for {}'.format(module.code))
                self._failure(file, xctx)
                return False
            module.actx = self.__actx
            pctx = module.createProjectContext(
                self.__gContext, projectDB=projectDB, logger=logger)
            cmod = wgui_getCockpitModule(module.code)
            if not cmod:
                xctx.addMessage(
                    'Could not find cockpit module for {}'.format(module.code))
                self._failure(file, xctx)
                return False
            #adding the file to recently oppened type...
            self.__addEditorF('file:' + file.path, createEditor)
            if editor:
                editor.zoomToFit()
                if logger.lastMessage() is not None:
                    editor.showProjectLog()
            self._success(file, xctx)
            return True
        else:
            xctx.addMessage(
                'Could not read project from file {}'.format(file))
            self._failure(file, xctx)
            return False

    @abstractmethod
    def _createProjectDB(self, file, xctx):
        raise NotImplementedError()

    def _success(self, file, xctx):
        pass

    def _failure(self, file, xctx):
        pass


class WguiOpenSilkanFileHandler(_WguiReadFileHandler):
    """
    File handler to open projects from Silkan XML files.
    """

    def __init__(self, addEditorF, cfg, gContext, actions, actx, *mimeTypes):
        """
        :param addEditorF: Function to add project editor to application
        :type addEditorF: WguiProjectEditor -> _
        :param cfg: Global project editor default configuration
        :type cfg: WguiProjectEditorCfgP
        :param gContext: Global context
        :type gContext: WcorGlobalContextP
        :param actions: Action dictionary
        :type actions: dict[str, WactActionP]
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param mimeTypes: Mime types handled by the file handler
        :type mimeTypes: list[ScmnMimeTypeP]
        """
        super().__init__(addEditorF, cfg, gContext, actions, actx, *mimeTypes)

    def _createProjectDB(self, file, xctx):
        return wprj_load_project(file, xctx)

    def _success(self, file, xctx):
        xctx.addMessage('Opened project ' + file.path, 'cockpit')

    def _failure(self, file, xctx):
        xctx.addMessage('Could not open project ' + file.path,
                        'cockpit', SCMN_LOG_ERROR)


class WguiImportFileHandler(_WguiReadFileHandler):
    """
    File handler to import projects from files containing model blocks.
    """

    def __init__(self, addEditorF, cfg, gContext, actions, actx, *mimeTypes):
        """
        :param addEditorF: Function to add project editor to application
        :type addEditorF: (WguiProjectEditor) -> _
        :param cfg: Global project editor default configuration
        :type cfg: WguiProjectEditorCfgP
        :param gContext: Global context
        :type gContext: WcorGlobalContextP
        :param actions: Action dictionary
        :type actions: dict[str, WactActionP]
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param mimeTypes: Mime types handled by the file handler
        :type mimeTypes: list[ScmnMimeTypeP]
        """
        super().__init__(addEditorF, cfg, gContext, actions, actx, *mimeTypes)

    def _createProjectDB(self, file, xctx):
        return wprj_import_project(file, xctx)

    def _success(self, file, xctx):
        xctx.addMessage('Imported project ' + file.path, 'cockpit')

    def _failure(self, file, xctx):
        xctx.addMessage('Could not import project ' + file.path,
                        'cockpit', SCMN_LOG_ERROR)


class WguiOpenTextFileHandler(WcmnMimeFileHandlerBase):
    def __init__(self, openEditorF, *mimeTypes):
        """
        :param addEditorF: Function to add text editor to application
        :type addEditorF: (WguiTextEditor) -> _
        :param mimeTypes: Mime types handled by the file handler
        :type mimeTypes: list[ScmnMimeTypeP]
        """
        super().__init__(*mimeTypes)

        self.__addEditorF = openEditorF

    def handleFile(self, file:WcmnFile, xctx):
        if not scmn_isfile(file.path):
            raise ValueError('No such file: ' + file.path)

        def createEditor():
            return WguiTextEditor(xctx, path=file.path,
                                  mime=next(iter(file.mimeTypes)))

        self.__addEditorF('file:' + file.path, createEditor, file.path)
        xctx.addMessage(
            'Opened file ' + file.path, 'cockpit')
        return True


class WguiLogListener(ScmnLogListener):
    def __init__(self, statusBar):
        """
        :param statusBar: Status bar
        :type statusBar: QStatusBar
        """
        self.__statusBar = statusBar

    def messageAdded(self, msg:ScmnLogMessage):
        if msg.category == SCMN_LOG_ERROR or msg.component == 'cockpit':
            self.__statusBar.showMessage(msg.text, 10 * 1000)


class WguiProjectTypeSelectionDialog(QDialog):
    """
    Dialog to select project type.
    """

    def __init__(self, createF, parent):
        """
        :param createF: Project creation function
        :type createF: (WguiCockpitSystemModuleP) -> _
        :param parent: Parent widget
        """
        super().__init__(parent)

        self.__createF = createF

        self.__projectSelection = QComboBox()

        self.setLayout(self._createLayout())
        self._updateFields()

    def _createLayout(self):
        layout = QVBoxLayout()

        layout.addWidget(QLabel('Please choose type of project to create:'))
        layout.addWidget(self.__projectSelection)

        buttons = QDialogButtonBox()
        buttons.addButton('Create', QDialogButtonBox.AcceptRole)
        buttons.addButton(QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

        layout.addWidget(buttons)

        return layout

    def _updateFields(self):
        for ctxt in wgui_getCockpitModules():
            if ctxt is WGUI_COCKPIT_DEFAULT_MODULE:
                continue
            module = ctxt.getAppModule()
            self.__projectSelection.addItem(module.name, ctxt)

    def accept(self):
        ptyp = wgui_qcb_get_data(self.__projectSelection)
        if ptyp is not None:
            self.__createF(ptyp)

        super().accept()
