# ----------------------------------------------------------------------------
# - wgui-navigation-history-ctrl                                             -
# - wild cruncher: navigation history control                 frontend / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_navigation_history_ctrl
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Navigation history control

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod
from collections import deque
from PyQt5.QtCore import pyqtSignal, QObject

from wcr_frontend.gui.wgui_protocols import QObject_Meta_M


#: Navigation history size
WGUI_NAV_HIST_SIZE = 50


class WguiNavigationHistoryControllerP(metaclass=ABCMeta):
    """
    Protocol for navigation history controllers.
    """

    @property
    @abstractmethod
    def historyChanged(self):
        """
        Signal indicating that the history state has changed.

        :return: Signal
        :rtype: pyqtBoundSignal
        """
        raise NotImplementedError()

    @abstractmethod
    def record(self):
        """
        Record the current state in the navigation history.
        """
        raise NotImplementedError()

    @abstractmethod
    def backward(self):
        """
        Go backward in the navigation history, if possible.
        """
        raise NotImplementedError()

    @abstractmethod
    def forward(self):
        """
        Go forward in the navigation history, if possible.
        """
        raise NotImplementedError()

    @abstractmethod
    def canGoBackward(self):
        """
        Whether it is currently possible to go backward in the navigation
        history.

        :return: True iff it is possible to go backward
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def canGoForward(self):
        """
        Whether it is currently possible to go forward in the navigation
        history.

        :return: True iff it is possible to go forward
        :rtype: bool
        """
        raise NotImplementedError()


class WguiNavigationHistoryController(QObject,
                                      WguiNavigationHistoryControllerP,
                                      metaclass=QObject_Meta_M):
    _historyChanged = pyqtSignal()
    """:type: pyqtBoundSignal"""

    def __init__(self, scene, view, parent=None):
        """
        :param scene: Graphics scene
        :type scene: WguiGraphicsScene
        :param view: Graphics view
        :type view: WguiGraphicsView
        """
        QObject.__init__(self, parent)

        self.__scene = scene
        self.__view = view

        self.__backward = deque(maxlen=WGUI_NAV_HIST_SIZE)
        self.__forward = deque(maxlen=WGUI_NAV_HIST_SIZE)

        self.__settingState = False
        self.__current = None

    @property
    def historyChanged(self):
        return self._historyChanged

    def _getState(self):
        path = tuple(self.__scene.blockInstancePath)
        _cntr = self.__view.viewCenter()
        cntr = _cntr.x(), _cntr.y()
        zoom = self.__view.zoom

        return path, cntr, zoom

    def record(self):
        if self.__settingState:
            return
        self.__backward.append(self._getState())
        self.__forward.clear()
        self.__current = None
        self._historyChanged.emit()

    def backward(self):
        if not self.__backward:
            return

        state = self.__backward.pop()
        self.__forward.append(
            self.__current if self.__current else self._getState())
        self.__current = state

        self._setState(state)
        self._historyChanged.emit()

    def forward(self):
        if not self.__forward or not self.__current:
            return

        state = self.__forward.pop()
        self.__backward.append(self.__current)
        self.__current = state

        self._setState(state)
        self._historyChanged.emit()

    def _setState(self, state):
        self.__settingState = True
        path, cntr, zoom = state

        self.__scene.setHierarchy(path)
        self.__view.centerOn(*cntr)
        self.__view.zoom = zoom
        self.__settingState = False

    def canGoBackward(self):
        return bool(self.__backward)

    def canGoForward(self):
        return bool(self.__forward)
