# ----------------------------------------------------------------------------
# - wgui-operator-instance-props-dialog                                      -
# - wild cruncher - cockpit : operator instance properties dialog widgets    -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_operator_instance_props_dialog
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher operator instance properties dialog

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from wcr_frontend.gui.wgui_props_dialog_widget import *


class _WguiParametersProperties(WguiPropertiesWidgetP):
    """
    The _WguiParametersProperties is a tab widget for the instance
    properties dialog. It displays the instance parameters properties.
    """

    def __init__(self, inst, pctx, parent=None):
        """
        :param inst: instance
        :type inst: WhdbOperatorInstanceP
        :param parent: Parent widget
        """
        # base initialization
        super().__init__(parent)

        #: Instance object
        self.__instance = inst
        #: Instance object
        self.__pctx = pctx
        #: Parameter table
        self.__table = QTableWidget()

        self.setLayout(self._createLayout())
        self.updateFields()

    @property
    def instance(self):
        return self.__instance

    @property
    def table(self):
        return self.__table

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QLayout
        """
        table = self.__table
        table.setSortingEnabled(True)
        layout = QVBoxLayout()
        layout.addWidget(table)
        return layout

    def _addParameter(self, param, row):
        """
        Add a new parameter line in the parameter table.

        :param param: The parameter to add
        :type param: WhdbParameterP
        :param row: Row number of the new row
        :type row: int
        """
        table = self.__table

        item = self._dis(QTableWidgetItem(param.getName()))
        item.setData(Qt.UserRole, param)
        table.setItem(row, 0, item)

        table.setItem(row, 1, self._dis(QTableWidgetItem(param.getInfo())))

        value = param.getValue()
        item = QTableWidgetItem(str(value) if value is not None else '')
        item.setFlags(item.flags() | Qt.ItemIsEditable)
        table.setItem(row, 2, item)

    def updateFields(self):
        table = self.__table
        params = list(self.__instance.getParameters())

        table.clear()
        table.setColumnCount(3)
        table.setRowCount(len(params))
        table.setHorizontalHeaderLabels(['Name', 'Info', 'Value'])
        for row, param in enumerate(params):
            self._addParameter(param, row)
        table.resizeColumnsToContents()
        table.horizontalHeader().setStretchLastSection(True)

    def save(self):
        table = self.__table
        for row in range(table.rowCount()):
            param = table.item(row, 0).data(Qt.UserRole)
            value = table.item(row, 2).text()
            param.setValue(value)


class _WguiOperatorProperties(WguiPropertiesWidgetP):
    """
    Property editing widget for a instance's operator information.
    """

    def __init__(self, inst, parent=None):
        """
        :param inst: Instance
        :type inst: WhdbOperatorInstanceP
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Instance object
        self.__instance = inst

        #: Operator name field
        self.__opNameF = QLineEdit()
        self.__opNameF.setEnabled(False)
        #: Operator info field
        self.__opInfoF = QLineEdit()
        self.__opInfoF.setEnabled(False)

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.addWidget(QLabel('Operator Name:'))
        layout.addWidget(self.__opNameF)
        layout.addWidget(QLabel('Operator Information:'))
        layout.addWidget(self.__opInfoF)
        layout.addStretch(1)
        return layout

    def updateFields(self):
        inst = self.__instance
        self.__opNameF.setText(inst.getOperator().getName())
        self.__opInfoF.setText(inst.getOperator().getInfo())

    def save(self):
        pass


class WguiOperatorInstancePropertiesDialog(WguiPropertiesDialog):
    """
    The WguiOperatorInstancePropertiesDialog class is a dialog design to
    access and manipulates operator instance data. This include name,
    info and connection points.
    """

    _TYPE_NAME = 'Operator Instance'

    _SHOW_OPERATOR_WIDGET = True

    PARAM_TAB = _WguiParametersProperties

    GENERAL_TAB = WguiInstanceBaseProperties

    # noinspection PyTypeChecker
    def __init__(self, inst, pctx, updateF, selectXpnt=None, parent=None):
        """
        :param inst: Instance
        :type inst: WhdbOperatorInstanceP
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param updateF: Callback function to notify about saves
        :type updateF: _ -> _
        :param selectXpnt: Initially selected connection point, if any
        :type selectXpnt: WhdbConnectionPointP
        :param parent: Parent widget
        """
        super().__init__('{} Properties ({})'.format(self._TYPE_NAME,
                                                     inst.getName()),
                         updateF, parent=parent)

        tophm = pctx.getProjectDB().getTopHierarchy()

        instWdgt = self.GENERAL_TAB(self._TYPE_NAME, inst, tophm)
        addiWdgts = self._createAdditionalInstancePropertiesWidgets(inst, pctx)
        if addiWdgts:
            # noinspection PyTypeChecker
            generalWdgts = [instWdgt] + addiWdgts
        else:
            generalWdgts = [instWdgt]
        if self._SHOW_OPERATOR_WIDGET:
            generalWdgts.append(_WguiOperatorProperties(inst))

        widgets = [('General', WguiMultiPropertiesWidgetP(*generalWdgts))]

        widgets.extend(self._createAdditionalPropertiesWidgets(inst, pctx))

        xpntWdgt = self._createConnectionPointsPropertiesWidget(inst, pctx)
        if xpntWdgt:
            widgets.append(xpntWdgt)
        if getattr(self, "_createModuleIOPropertiesWidget", None):
            modulesWdgt = self._createModuleIOPropertiesWidget(inst, pctx)
            if modulesWdgt:
                widgets.append(modulesWdgt)

        widgets.append(('Parameters', self.PARAM_TAB(inst, pctx)))
        widgets.append(('Appearance', WguiFlipRotationProperties(inst)))

        # noinspection PyUnresolvedReferences
        self.setTabbedPropertyWidgets(
            *widgets,
            activate=xpntWdgt[1] if selectXpnt and xpntWdgt else None)

        if selectXpnt and xpntWdgt:
            wgui_setActiveXPoint(selectXpnt, xpntWdgt[1])

    def showOperatorWidget(self):
        return True

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _createAdditionalInstancePropertiesWidgets(self, inst, pctx):
        return []

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _createAdditionalPropertiesWidgets(self, inst, pctx):
        return []

    # noinspection PyMethodMayBeStatic
    def _createConnectionPointsPropertiesWidget(self, inst, pctx):
        return 'Connectors', WguiXPointProperties(inst)
