# ----------------------------------------------------------------------------
# - wgui-param-value-source                                                  -
# - wild cruncher: parameter value source                     frontend / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_param_value_source
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - GUI parameter value source

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt4.QtGui import QFileDialog, QMessageBox, QIcon

# noinspection PyUnresolvedReferences
from wcr_core.action.wact_action_parameter import \
    WactFixedMultiValueActionParameter
from wcr_core.action.wact_action_parameter_type_api import \
    WactStringableActionParameterTypeP
from wcr_frontend.gui.wgui_icon_provider import IconPrvP
from wcr_frontend.gui.wgui_object_selection_dialog import \
    WguiObjectSelectionDialog
from wcr_core.action.wact_action_parameter_types import \
    WactActionSaveFileByMimeParameterType, WactActionOpenFileParameterTypeP, \
    WactActionEmptyDirectoryParameterType, WactActionDirectoryParameterType
from wcr_core.action.wact_parameter_value_source_api import \
    WactParameterValueSourceP
from wcr_core.common.wcmn_file_handler import WcmnFile
from wcr_frontend.gui.wgui_utils import wgui_getSaveFile, wgui_getOpenFile, \
    WGUI_LAST_DIR


class WguiParameterValueSource(WactParameterValueSourceP):
    def __init__(self, parent, iconPrv):
        """
        :param parent: Parent widget
        :type parent: QWidget
        :param iconPrv: Icons provider which allows to retrieve QIcons
        :type iconPrv: IconPrvP
        """
        self.__parent = parent
        self.__iconPrv = iconPrv

    def getValue(self, parameter):
        paramType = parameter.getParamType()

        if isinstance(parameter, WactFixedMultiValueActionParameter):
            return self.getChooseFixedMultiValue(parameter)
        elif isinstance(paramType, WactActionSaveFileByMimeParameterType):
            return self.getSaveFile(parameter)
        elif isinstance(paramType, WactActionOpenFileParameterTypeP):
            return self.getOpenFile(parameter)
        elif isinstance(paramType, WactActionEmptyDirectoryParameterType):
            return self.getEmptyDirectory(parameter)
        elif isinstance(paramType, WactActionDirectoryParameterType):
            return self.getDirectory(parameter)
        else:
            raise Exception('Unknown parameter type {}'.format(paramType))

    def getChooseFixedMultiValue(self, parameter):
        """
        Create a dialog listing the options as string labels in a combo-box
        with their gui icon if defined in its correspondent module

        :param parameter: Parameter for the action
        :type parameter: WactFixedMultiValueActionParameter
        :return: Chosen value
        :rtype: object | None
        """
        if isinstance(parameter.getParamType(),
                      WactStringableActionParameterTypeP):
            choiceDlg = WguiObjectSelectionDialog(
                self.__parent,
                parameter.getChoiceMessage(),
                parameter.getPossibleValues(),
                parameter.getName(),
                _IconPrvFixedMultiValueActionParameter(parameter,
                                                       self.__iconPrv))
            if choiceDlg.exec():
                return choiceDlg.getChosenOption()

    def getOpenFile(self, parameter):
        paramType = parameter.getParamType()
        """:type: WactActionOpenFileParameterTypeP"""
        path = wgui_getOpenFile(
            'Select ' + parameter.getName(), self.__parent,
            *paramType.mimeTypes)
        if not path:
            return None
        return WcmnFile(path)

    def getSaveFile(self, parameter):
        """
        :type parameter: WactActionParameterP
        :rtype: WcmnFile | None
        """
        paramType = parameter.getParamType()
        """:type: WactActionSaveFileByMimeParameterType"""
        path, _ = wgui_getSaveFile(
            'Save {} as ...'.format(parameter.getName()),
            self.__parent, *paramType.mimeTypes)
        return WcmnFile(path) if path else None

    def getEmptyDirectory(self, parameter):
        global WGUI_LAST_DIR

        paramType = parameter.getParamType()
        """:type: WactActionEmptyDirectoryParameterType"""
        dirPath = None
        while not dirPath:
            dirPath = QFileDialog.getExistingDirectory(
                self.__parent, 'Select ' + parameter.getName(), WGUI_LAST_DIR)
            if not dirPath:
                return
            errs = paramType.validate(dirPath)
            if errs:
                #err = any(e for e, _ in errs)
                msg = ' '.join(m for _, m in errs)
                msg += 'Would you like to choose a different directory?'
                # noinspection PyArgumentList
                chgDir = QMessageBox.warning(
                    self.__parent, 'Warning', msg,
                    buttons=QMessageBox.Yes | QMessageBox.No,
                    defaultButton=QMessageBox.Yes)
                if chgDir == QMessageBox.Yes:
                    dirPath = None

        WGUI_LAST_DIR = dirPath
        return dirPath

    def getDirectory(self, parameter):
        global WGUI_LAST_DIR
        """:type: WactActionDirectoryParameterType"""
        dirPath = None
        while dirPath is None:
            sdlg = QFileDialog(self.__parent, 'Select ' + parameter.getName(), WGUI_LAST_DIR, None)
            options = QFileDialog.DontUseNativeDialog
            dirPath = sdlg.getExistingDirectory(options=options)
            if not dirPath:
                return
        WGUI_LAST_DIR = dirPath
        return dirPath


class _IconPrvFixedMultiValueActionParameter(IconPrvP):
    """
    FixedMultiValueActionParameter-specific icon provider
    """

    def __init__(self, parameter, delegate):
        """
        :param parameter: Multi value action parameter
        :type parameter: WactFixedMultiValueActionParameter
        :param delegate: Icon Provider to which the get action is delegated
        :type delegate: IconPrvP
        """
        self.__paramCode = parameter.getCode()
        self.__delegate = delegate

    def getIcon(self, code):
        """
        Retrieve an icon linked to the passed code

        :param code: Key value linked to the icon
        :type code: str
        :return: Icon if found, None otherwise
        :rtype: QIcon | None
        """
        return self.__delegate.getIcon('{}-{}'.format(self.__paramCode, code))
