# ----------------------------------------------------------------------------
# - wgui-project-actions                                                     -
# - wild cruncher - project editor actions                                   -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_project_actions
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher project editor actions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt4.QtCore import Qt

from wcr_core.action.wact_action import *
from wcr_core.action.wact_action_parameter import \
    WactFixedMultiValueActionParameter
from wcr_core.action.wact_action_parameter_types import \
    WactActionIntParameterType
from wcr_frontend.gui.wgui_project_editor import WguiProjectEditor
from wcr_core.db.whdb_utils import whdb_printBlock


def wgui_create_pedit_actions():
    """
    Create a set of project editor functions.

    :return: Project editor actions
    :rtype: dict[str, WactActionP]
    """

    return {
        'project.log': _WguiEditorShowLogAction(),
        'project.poph': _WguiHierarchyUpAction(),
        'project.popt': _WguiHierarchyTopAction(),
        'project.zoomin': _WguiZoomInAction(),
        'project.zoomout': _WguiZoomOutAction(),
        'project.zoomfit': _WguiZoomToFitAction(),
        'project.zoom100': _WguiZoomToOriginalAction(),
        'project.pinf': _WguiDisplayProjectInfoAction(),
        'project.prefresh': _WguiRefreshProjectAction(),
        'project.selmode': _WguiSetSelectionTypeAction(),
        'project.history.backward': _WguiNavHistoryBackwardAction(),
        'project.history.forward': _WguiNavHistoryForwardAction(),
        'project.dump': _WguiDumpAction(),
    }


# noinspection PyAbstractClass
class WguiEditorActionMixin(WactActionBaseP):
    """
    Base class for project editor actions.
    """

    def isEnabled(self, actx):
        """
        :type actx: WguiActionExecutionContextP
        """
        return isinstance(actx.getEditor(), WguiProjectEditor)

    def _doExecute(self, actx, paramValuePairs):
        editor = actx.getEditor()
        if isinstance(editor, WguiProjectEditor):
            self._executeForEditor(editor, actx, paramValuePairs)

    @abstractmethod
    def _executeForEditor(self, editor, actx, paramValuePairs):
        """
        Execute the action on the given editor.

        :param actx: Action execution context
        :type actx: WactActionExecutionContextP
        :param editor: Editor to execute action for
        :type editor: WguiProjectEditor
        :param paramValuePairs: Parameter-value pairs
        """
        pass


# noinspection PyAbstractClass
class WguiEditorActionBase(WactActionBase, WguiEditorActionMixin):
    """
    Base class for project editor actions.
    """

    def __init__(self, name, description, *parameters):
        """
        :param name: Action name
        :type name: str
        :param description: Action description
        :type description: str
        :param parameters: Action parameter descriptions
        :type parameters: list[WactActionParameterP]
        """
        WactActionBase.__init__(self, name, description, *parameters)
        WguiEditorActionMixin.__init__(self)

    def isEnabled(self, actx):
        return WguiEditorActionMixin.isEnabled(self, actx)


class _WguiEditorShowLogAction(WguiEditorActionBase):
    """
    Project editor action to show the project log window.
    """

    def __init__(self):
        super().__init__('Project log', 'Show project log')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.showProjectLog()


class _WguiHierarchyUpAction(WguiEditorActionBase):
    """
    Project editor action to go up one level in the hierarchy.
    """

    def __init__(self):
        super().__init__('Go to parent level', 'Go to parent level')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.goUpHierarchy()


class _WguiHierarchyTopAction(WguiEditorActionBase):
    """
    Project editor action to go to the top of the hierarchy.
    """

    def __init__(self):
        super().__init__('Go to top level', 'Go to top level')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.goUpHierarchyTop()


class _WguiZoomInAction(WguiEditorActionBase):
    """
    Project editor action to zoom in.
    """

    def __init__(self):
        super().__init__('Zoom in', 'Zoom in')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.zoomIn()


class _WguiZoomOutAction(WguiEditorActionBase):
    """
    Project editor action to zoom out.
    """

    def __init__(self):
        super().__init__('Zoom out', 'Zoom out')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.zoomOut()


class _WguiZoomToFitAction(WguiEditorActionBase):
    """
    Project editor action to zoom to fit the entire model block into the view.
    """

    def __init__(self):
        super().__init__(
            'Zoom to Fit',
            'Zoom to fit the entire project into the current view')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.zoomToFit()


class _WguiZoomToOriginalAction(WguiEditorActionBase):
    """
    Project editor action to zoom to the original magnification.
    """

    def __init__(self):
        super().__init__('Zoom to 100%', 'Zoom to the original magnification')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.zoomToOriginal()


class _WguiDisplayProjectInfoAction(WguiEditorActionBase):
    """
    Project editor action to display the project info dialog.
    """

    def __init__(self):
        super().__init__('Project Info', 'Display project info dialog')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.showProjectInfoDialog()


class _WguiRefreshProjectAction(WguiEditorActionBase):
    """
    Project editor action to display the project info dialog.
    """

    def __init__(self):
        super().__init__('Refresh Project ', 'refresh the project view')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.initialize()


class _WguiSetSelectionTypeAction(WactStateSelectionActionBase,
                                  WguiEditorActionMixin):
    """
    Project editor action to switch the rubber band selection mode.
    """

    __POSSIBLE_VALUES = [
        (Qt.IntersectsItemShape, 'Intersection Selection',
         'Rubber band selection only includes fully contained items'),
        (Qt.ContainsItemShape, 'Inclusion Selection',
         'Rubber band selection includes any intersected items')]

    _PARAMETER = WactFixedMultiValueActionParameter(
        'seltype', 'Selection Mode', 'Rubberband selection mode',
        WactActionIntParameterType(), possibleValues=__POSSIBLE_VALUES)

    def __init__(self):
        WactStateSelectionActionBase.__init__(
            self, 'Selection Mode', 'Selection Mode', self._PARAMETER)
        WguiEditorActionMixin.__init__(self)

    def _executeForEditor(self, editor, actx, paramValuePairs):
        styp = self._getParamValue(self._PARAMETER, actx, paramValuePairs)
        editor.config.selectionMode = styp

    def getCurrentValue(self, actx):
        editor = actx.getEditor()
        if isinstance(editor, WguiProjectEditor):
            return editor.config.selectionMode
        else:
            return None


class _WguiNavHistoryBackwardAction(WguiEditorActionBase):
    """
    Project editor action to go backward in the navigation history.
    """

    def __init__(self):
        super().__init__('Go Back', 'Go back if possible')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.history.backward()


class _WguiNavHistoryForwardAction(WguiEditorActionBase):
    """
    Project editor action to go forward in the navigation history.
    """

    def __init__(self):
        super().__init__('Go Forward', 'Go forward if possible')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        editor.history.forward()


class _WguiDumpAction(WguiEditorActionBase):
    """
    Project editor action to dump the current database to STDOUT.
    """

    def __init__(self):
        super().__init__('Dump DB', 'Dump database to stdout')

    def _executeForEditor(self, editor, actx, paramValuePairs):
        db = editor.projectContext.getProjectDB()
        whdb_printBlock(db.getTopHierarchy())
