# ----------------------------------------------------------------------------
# - wgui-project-editor                                                      -
# - wild cruncher - cockpit: project editor widget                           -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_project_editor
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher project editor widget

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from functools import partial
from uuid import UUID
import math
import weakref

from PyQt4.QtCore import Qt, pyqtSignal, QMimeData, QByteArray, QSize
from PyQt4.QtGui import QWidget, QHBoxLayout, QSlider, QVBoxLayout, QLabel, \
    QFont, QSplitter, QButtonGroup, QPrintDialog, QPrinter, QDialog, \
    QPainter, QAbstractPrintDialog, QLayout, QSizePolicy, QTabWidget, QIcon, \
    QToolButton, QToolBar, QGridLayout, QToolBox, QApplication, QDrag, \
    QAbstractButton,QPushButton, QGraphicsScene

from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_core.modules.wmod_block_type import WMOD_SYSTEM_BLOCK_FLAG
from wcr_core.project.wprj_xmlio import WPRJ_SILKAN_XML_MIME_TYPE
from wcr_frontend.gui.wgui_action_execution_context import \
    WguiActionExecutionContextBase
from wcr_frontend.gui.wgui_editor_api import WGUI_EDITOR_LEFT_SIDE, \
    WguiEditorInstTargetP
from wcr_frontend.gui.wgui_editor_side_button_bar import WguiSideButtonBar
from wcr_frontend.gui.wgui_editor_sidebar import \
    WguiEditorSidebarInstaller
from wcr_frontend.gui.wgui_highlight_module import WguiToggleHighlightModule
from wcr_frontend.gui.wgui_navigation_history_ctrl import \
    WguiNavigationHistoryController
from wcr_frontend.gui.wgui_toolbox_item import *
from wcr_frontend.gui.wgui_editor import WguiEditorHasSaveP, WguiEditorBase, \
    WguiEditorHasRefreshP, WguiEditorHasDeleteP, WguiEditor
from wcr_frontend.gui.wgui_scene_view import WguiGraphicsView
from wcr_frontend.gui.wgui_button import WguiButton
from wcr_frontend.gui.wgui_project_editor_config import WguiProjectEditorCfg
from wcr_frontend.gui.wgui_protocols import WguiChangeNotifyP, WguiHasTitleP
from wcr_frontend.modules.vlink.gui.wvgui_scene import WvguiGraphicsScene
from wcr_frontend.gui.wgui_thumbnail import WguiThumbnail
from wcr_frontend.gui.wgui_utils import wgui_getSaveFile, \
    WGUI_INST_DND_MIME_TYPE
from wcr_frontend.gui.wgui_logger import WguiLoggerWindow
from wcr_core.db.whdb_operator import WHDB_NO_UI_OPERATOR_FLAG
from wcr_core.project.wprj_project import WprjProjectDatabase
from silkan_common.scmn_file_path import scmn_path_base


WGUI_PROJECT_BROWSER_TITLE = 'Project Database'


class WguiProjectEditor(WguiEditorBase, WcmnReleasableP, WguiEditorHasSaveP,
                        WguiEditorHasRefreshP, WguiEditorHasDeleteP):
    sideButtonToggled = pyqtSignal(str, bool)
    """:type: pyqtBoundSignal[str, bool]"""

    # graphic window constructor
    def __init__(self, pcfg, pctx, cctx, parent=None):
        """
        :param pcfg: Project editor configuration
        :type pcfg: WguiProjectEditorCfgP
        :param pctx: project context
        :type pctx: WcorProjectContextP
        :param cctx: Cockpit system context
        :type cctx: WguiCockpitSystemContextP
        :param parent: the parent widget
        """
        super().__init__(parent)

        assert pctx
        assert cctx

        #: Application context
        self.__projectContext = pctx
        #: Cockpit system context
        self.__cctx = cctx
        #: Project database
        self.__projectDB = pctx.getProjectDB()
        #: Global project editor configuration
        self.__config = pcfg
        self.mainLayout = None
        self.__buttonGroup = None
        self.__scene = None
        self.__clipboard = None
        self.__pdgf = None
        self.__view = None
        self.__histCtrl = None
        self.__bottomDrawer = None
        self.__leftButtonBar = None
        self.__sideBarInst = None
        self.__highlighters = None
        self.__toolBar = None
        self.__thumbnail = None
        self.parent = parent
        self.initialize()

    def initialize(self):
        if self.mainLayout is not None:
            QWidget().setLayout(self.mainLayout)

        # create the toolbox and add it
        self.__buttonGroup = _WguiToolButtonGroup(self.__cctx)
        # noinspection PyUnresolvedReferences
        self.__buttonGroup.buttonClicked[QAbstractButton].connect(
            self._toolButtonClicked)
        self.__toolBox = None
        self._createToolsbox()
        #: Graphic scene
        self.__scene = self.__cctx.getScene()
        self.__scene.importModules()
        """:type: WguiGraphicsScene"""
        #: Clipboard handler
        self.__clipboard = self.__cctx.getClipboardHandler()
        #: Properties dialog factory
        self.__pdgf = self.__cctx.getPropertiesDialogFactory()

        menuFactory = self.__cctx.getContextMenuFactory()
        #: Graphic view
        self.__view = WguiGraphicsView(
            self.__projectContext, self.__cctx,
            self.__cctx.getSceneViewMouseHandler(), self.__scene)
        self.__histCtrl = WguiNavigationHistoryController(
            self.__scene, self.__view, parent=self)
        self.__scene.scenePathChanging.connect(self.__histCtrl.record)
        # TODO: View context (getViewCenterProvider)
        self.__clipboard.setCenterF(self.__view.viewCenter)
        # TODO: View context (getViewCenterProvider)
        menuFactory.setCenterF(self.__view.viewCenter)
        #: Bottom drawer
        self.__bottomDrawer = _WguiBottomDrawer()
        self.__bottomDrawer.minimumSizeRequested.connect(
            self._bottomDrawerResize)
        widgets = self.__cctx._createSidebarModules(self.__projectContext,
                                                    self.__scene,
                                                     menuFactory, [],
                                                    self)
        #: Left side button bar
        self.__leftButtonBar = WguiSideButtonBar(WGUI_EDITOR_LEFT_SIDE)
        #: Sidebar
        self.__sideBarInst = WguiEditorSidebarInstaller(widgets)
        self.__highlighters = list(self.__cctx.getHighlightingModules())
        for hlgt in self.__highlighters:
            self.__scene.viewContext.highlighter.addModule(hlgt)
        #: Toolbar
        self.__toolBar = QToolBar('Editor')
        #: Macro thumbnail
        if isinstance(self.__scene, QGraphicsScene):
            self.__thumbnail = WguiThumbnail(self.__scene, self.__view)
        else:
            self.__thumbnail = self.__scene
        #: Main vertical splitter
        self.__vSplitter = QSplitter(Qt.Vertical)
        #: Project log display
        self.__projectLog = None
        #: Path field
        self.__pathF = QLabel()

        # create the bottom toolbar
        self._createToolBar()
        self.mainLayout = self._createLayout()
        self.setLayout(self.mainLayout)

        # connect all signal emitters
        self.__scene.selectModeSignal.connect(self._setButtonSelection)
        self.__scene.scenePathChanged.connect(self._updatePath)
        self.__cctx.getPropertiesDialogFactory().projectPropertiesChangedS. \
            connect(self._prjPropsChanged)
        self.__pathF.linkActivated.connect(self._pathElementClicked)
        self.__config.addObserver(self._configChanged)
        self._configChanged()
        self._updatePath()

    def isICDProject(self):
        return getattr(self.__cctx, "isICDProject", False)

    def release(self):
        self.__config.removeObserver(self._configChanged)

        for item in [self.__cctx, self.__projectContext]:
            if isinstance(item, WcmnReleasableP):
                item.release()

        super().release()

    def _createLayout(self):
        """
        Create the layout for the project editor.

        :return: Layout for the project editor
        :rtype: QLayout
        """
        self.__toolBox.setSizePolicy(QSizePolicy(QSizePolicy.Fixed,
                                                 QSizePolicy.Expanding,
                                                 QSizePolicy.ButtonBox))

        editorLayout = QVBoxLayout()
        editorLayout.setMargin(0)
        editorLayout.setSpacing(2)

        hbox = QHBoxLayout()
        hbox.setMargin(0)
        hbox.setSpacing(2)

        vbox = QVBoxLayout()
        vbox.setMargin(0)
        vbox.setSpacing(2)

        vbox.addWidget(self.__thumbnail)
        if not isinstance(self.__thumbnail, WvguiGraphicsScene):
            vbox.addWidget(self.__toolBox)
        hbox.addLayout(vbox)

        hbox.addWidget(self.__view)
        editorLayout.addLayout(hbox)
        editorLayout.addWidget(self.__toolBar)

        view = QWidget()
        view.setLayout(editorLayout)
        hSplit = QSplitter(Qt.Horizontal)
        hSplit.setChildrenCollapsible(False)
        hSplit.addWidget(view)

        topHBox = QHBoxLayout()
        topHBox.setMargin(0)
        topHBox.setSpacing(3)
        topHBox.addWidget(hSplit)
        sideButtonVBox = QVBoxLayout()
        sideButtonVBox.addWidget(self.__leftButtonBar)
        sideButtonVBox.addStretch(1)
        topHBox.addLayout(sideButtonVBox)

        instTarget = _WguiProjectEditorInstTarget(
            hSplit, self.__leftButtonBar, self.sideButtonToggled)
        self.__sideBarInst.install(instTarget)

        vSplit = self.__vSplitter
        vSplit.setChildrenCollapsible(False)
        view = QWidget()
        view.setLayout(topHBox)
        vSplit.addWidget(view)
        vSplit.addWidget(self.__bottomDrawer)

        mainLayout = QVBoxLayout()
        mainLayout.setMargin(5)
        mainLayout.addWidget(self.__pathF)
        mainLayout.setSpacing(0)
        mainLayout.addWidget(vSplit)

        return mainLayout

    def _updatePath(self):
        uuidPath = []
        insts = list(self.__scene.blockInstancePath)
        topName = self.__projectDB.getTopHierarchy().getName() or '[TOP-LEVEL]'
        if insts:
            links = ['<a href="#">{}</a>'.format(topName)]
            for inst in insts[:-1]:
                uuidPath.append(str(inst.getUUID()))
                links.append(
                    '<a href="#{}">{}</a>'.format('/'.join(
                        uuidPath), inst.getName()))
            links.append(insts[-1].getName())
        else:
            links = [topName]
        path = ' > '.join(links)

        self.__pathF.setText(path)

    def _pathElementClicked(self, link):
        """
        :type link: str
        """
        path = [UUID(u) for u in link.strip('#').split('/') if u]
        self.__scene.setHierarchyUUIDs(*path)

    def _prjPropsChanged(self):
        self.nameChanged.emit(self.getTitle())
        self._updatePath()

    def _bottomDrawerResize(self, size):
        """
        Callback to handle resize requests from the bottom drawer.

        :param size: Requested height
        :type size: int
        """
        height = self.__vSplitter.height()
        size = min(size, height / 2)
        self.__vSplitter.setSizes([height - size, size])

    def _configChanged(self):
        """
        Callback function for configuration changes.
        """
        self.__view.setRubberBandSelectionMode(self.__config.selectionMode)

    @property
    def projectContext(self):
        """
        Get the project context.

        :return: Project context
        :rtype: WcorProjectContextP
        """
        return self.__projectContext

    @property
    def cockpitSystemContext(self):
        """
        Get the cockpit system context.

        :return: Cockpit system context
        :rtype: WguiCockpitSystemContextP
        """
        return self.__cctx

    @property
    def config(self):
        """
        Get the project editor's configuration.

        :return: The editor's configuration
        :rtype: WguiProjectEditorCfg
        """
        return self.__config

    @property
    def scene(self):
        """
        Get the project editor's graphics scene.

        :return: Graphic scene
        :rtype: WguiGraphicsScene
        """
        return self.__scene

    @property
    def view(self):
        """
        Get the project editor's graphics view.

        :return: Graphics view
        :rtype: WguiGraphicsView
        """
        return self.__view

    @property
    def history(self):
        """
        Get the navigation history controller.

        :return: History controller
        :rtype: WguiNavigationHistoryControllerP
        """
        return self.__histCtrl

    @property
    def sidebarModules(self):
        """
        Get the sidebar's additional module widgets.

        :return: List of sidebar widgets and their identifiers and names
        :rtype: Iterable[(str, str, QWidget)]
        """
        return self.__cctx.getSidebarModules()

    @property
    def buttonGroup(self):
        """
        Get the group of button of toolbox.

        :return: buttonGroup
        :rtype: _WguiToolButtonGroup
        """
        return self.__buttonGroup

    def setToolsBox(self, toolBox):
        """
        set the toolbox.
        """
        self.__toolBox = toolBox

    @property
    def toolsBox(self):
        """
        getter the toolbox.
        """
        return self.__toolBox

    def _createToolsbox(self):
        """
        Create the window toolbox.

        :return: Toolbox
        :rtype: QToolBox
        """

        def makeTab(items, name):
            pntg = QGridLayout()
            pntg.setMargin(2)
            pntg.setSpacing(6)

            cnt = len(items)

            idxc = -1
            for idx, item in enumerate(items):
                idxc += 1
                if isinstance(item, QPushButton):
                    idxc += 3-(idxc % 3)
                pntg.addWidget(item, idx // 3, idx % 3)

            pntg.setRowStretch((cnt - 1) // 3 + 1, 1)
            pntg.setColumnStretch(4, 1)

            fwgt = QWidget()
            fwgt.setLayout(pntg)
            self.__toolBox.addItem(fwgt, name)

        module = self.__cctx.getAppModule()
        self.__toolBox = QToolBox()

        dflts = []

        btyps = self.cockpitSystemContext.getAppModule().getBlockTypes()

        dflts.extend(self.__buttonGroup.addBlockButton(btyp.getCode(),
                                                       btyp.getName())
                     for btyp in btyps
                     if not btyp.hasFlag(WMOD_SYSTEM_BLOCK_FLAG))
        exls = self.cockpitSystemContext.getExcludedSystemToolboxItems()
        dflts.extend(self.__buttonGroup.addToolButton(key, drag) for key, drag
                     in [(XPNT_ITEM_ID, False), (CNCT_ITEM_ID, False)]
                     if not (exls and key in exls))
        dflts.extend(self.__buttonGroup.addModelButton(mime) for mime
                     in module.getModelMimeTypes())
        dflts.extend(self.__buttonGroup.addOperatorButton(oper) for oper
                     in module.getDefaultOperators())
        makeTab(dflts, 'Default')

        for lib in module.getLibraries():
            makeTab([self.__buttonGroup.addOperatorButton(o) for o
                     in lib.operators if
                     not o.hasFlag(WHDB_NO_UI_OPERATOR_FLAG)],
                    lib.name)

        makeTab([self.__buttonGroup.addToolButton(shape, False) for shape
                 in [LINE_ITEM_ID, AROW_ITEM_ID, RECT_ITEM_ID,
                     ELPS_ITEM_ID, TEXT_ITEM_ID, ACTN_ITEM_ID]],
                'Shapes')

        self.__toolBox.setFixedWidth(160)


    def _createToolBar(self):
        tbar = self.__toolBar
        tbar.layout().setMargin(0)
        tbar.setContentsMargins(0, 0, 0, 0)
        tbar.addWidget(self.__buttonGroup.addToolButton(SLCT_ITEM_ID, False))

        # new group
        tbar.addSeparator()

        # the grid button
        icon = QIcon(':/img/app_was_grid.png')
        self.gbut = QToolButton()
        self.gbut.setIcon(icon)
        self.gbut.setIconSize(QSize(32, 32))
        self.gbut.setCheckable(True)
        self.gbut.setToolTip('Enable grid display')
        self.gbut.setChecked(self.__scene.getGridFlag())
        self.gbut.toggled.connect(self.grid)
        tbar.addWidget(self.gbut)
        # the snap button
        icon = QIcon(':/img/app_was_snap.png')
        self.sbut = QToolButton()
        self.sbut.setIcon(icon)
        self.sbut.setIconSize(QSize(32, 32))
        self.sbut.setCheckable(True)
        self.sbut.setToolTip('Enable grid snapping')
        self.sbut.setChecked(self.__scene.getSnapFlag())
        self.sbut.toggled.connect(self.snap)
        tbar.addWidget(self.sbut)

        for hlgt in self.__highlighters:
            if not isinstance(hlgt, WguiToggleHighlightModule):
                continue
            icon = QIcon(self.__cctx.getIconPath(hlgt, WGUI_ICON_PNG32))
            self.sbut = QToolButton()
            self.sbut.setIcon(icon)
            self.sbut.setIconSize(QSize(32, 32))
            self.sbut.setCheckable(True)
            self.sbut.setToolTip(hlgt.description)
            #: TODO: Handle enabled-state outside of plugin (list-of-enableds)
            self.sbut.toggled.connect(hlgt.enableHighlighting)
            tbar.addWidget(self.sbut)

        # new group
        tbar.addSeparator()

        # the pop hierachy button
        icon = QIcon(':/img/app_was_poph.png')
        self.phbt = QToolButton()
        self.phbt.setIcon(icon)
        self.phbt.setIconSize(QSize(32, 32))
        self.phbt.setToolTip('Pop one level of hierarchy')
        self.phbt.clicked.connect(self.goUpHierarchy)
        tbar.addWidget(self.phbt)
        # the pop top hierachy button
        icon = QIcon(':/img/app_was_popt.png')
        self.ptbt = QToolButton()
        self.ptbt.setIcon(icon)
        self.ptbt.setIconSize(QSize(32, 32))
        self.ptbt.setToolTip('Pop the top level hierarchy')
        self.ptbt.clicked.connect(self.goUpHierarchyTop)
        tbar.addWidget(self.ptbt)

        right = QWidget()
        rightLayout = QHBoxLayout()
        right.setLayout(rightLayout)
        rightLayout.setMargin(0)
        rightLayout.addStretch(1)

        rightLayout.addWidget(_WguiZoomSlider(self.__view))
        rightLayout.addSpacing(3)
        tbar.addWidget(right)

        #todo: set a proper icon for resize
        icon = QIcon(':/img/app_was_xpan.png')
        self.rvbt = QToolButton()
        self.rvbt.setIcon(icon)
        self.rvbt.setIconSize(QSize(32, 32))
        self.rvbt.setToolTip('Resize view to see all model')
        self.rvbt.clicked.connect(self.resettViewZoom)
        tbar.addWidget(self.rvbt)

    def _addHighlightingModuleButtons(self):
        pass

    def _toolButtonClicked(self, button):
        """
        Set the mode when a button is clicked in the group. This method
        is a handler which is called with a button id.

        :param button: the button id in the group
        :type button: _WguiToolButton
        """
        # clear button
        buttons = self.__buttonGroup.buttons()
        for _button in buttons:
            if _button != button: _button.setChecked(False)
        # set the mode
        self.__scene.setMode(button.code)

    def _setButtonSelection(self):
        """
        Set the selection button, even by notification
        """
        self.__buttonGroup.getToolButton(SLCT_ITEM_ID).setChecked(True)
        self.__scene.setMode(SLCT_ITEM_ID)

    def getProjectName(self):
        """
        Get the project name.

        :return: Project name
        :rtype: str
        """
        return self.__projectDB.getProjectName()

    def getProjectDB(self):
        """
        Get the project database.

        :return: Project database
        :rtype: WprjProjectDatabase
        """
        return self.__projectDB

    def getTitle(self):
        prjName = self.getProjectName()
        file = scmn_path_base(self.__projectDB.path)

        if not file:
            return prjName
        elif not prjName:
            return file
        else:
            return '{} ({})'.format(prjName, file)

    def getIcon(self):
        return self.__cctx.getIconPath('editor-tab', WGUI_ICON_PNG32)

    def getEditorIdent(self):
        path = self.__projectDB.path
        return 'file:' + path if path else 'wcr:project-{}'.format(id(self))

    def isReadOnly(self):
        return False

    def isDirty(self):
        return True

    def save(self):
        """
        Save the project using its known file path.
        If no file path has been set yet, either because the project has been
        loaded from a file or from a previous save-as operation, save() will
        behave like saveAs().
        """
        if self.__projectDB.path:
            self.__projectDB.save()
        else:
            self.saveCopy()

    def saveCopy(self):
        path, _ = wgui_getSaveFile(
            'Save as File', self.parentWidget(),
            WPRJ_SILKAN_XML_MIME_TYPE)

        if not path:
            return None, None
        self.__projectDB.saveAs(path)
        self.__projectContext.addRecentlyOpenedFile(path)
        self.nameChanged.emit(self.getTitle())
        return path, WPRJ_SILKAN_XML_MIME_TYPE.mimeType

    def activated(self):
        # Ensure buttons' enabled-status gets updated
        for i in self.__scene.getSceneItems():
            if isinstance(i, WguiButton):
                i.updateItem()

    def refresh(self):
        self.__scene.refresh()

    def goUpHierarchy(self):
        self.__scene.popHierarchy()

    def goUpHierarchyTop(self):
        self.__scene.popTopHierarchy()

    def resettViewZoom(self):
        self.__view.zoomToFit()

    def delete(self):
        self.__scene.delSelection()

    def copy(self):
        if self.__scene.selectedItems():
            self.__clipboard.copy()

    def cut(self):
        if self.__scene.selectedItems():
            self.__clipboard.copy()
            self.__scene.delSelection()

    def paste(self):
        self.__clipboard.paste()

    def selectAll(self):
        items = self.__scene.items()
        if not items:
            return
        self.__scene.blockSignals(True)
        for item in items[:-1]:
            item.setSelected(True)
        items[-1].setSelected(False)
        self.__scene.blockSignals(False)
        items[-1].setSelected(True)

    def print(self):
        printer = QPrinter(QPrinter.HighResolution)
        dialog = QPrintDialog(printer, self.parentWidget())
        dialog.setWindowTitle('Print Project')
        dialog.setEnabledOptions(dialog.enabledOptions() &
                                 ~QAbstractPrintDialog.PrintPageRange)
        dialog.addEnabledOption(QAbstractPrintDialog.PrintSelection)
        if dialog.exec_() != QDialog.Accepted:
            return
        painter = QPainter()
        painter.begin(printer)
        if dialog.printRange() == QAbstractPrintDialog.Selection:
            self.__view.render(painter)
        else:
            self.__scene.render(painter,
                                source=self.__scene.itemsBoundingRect())
        painter.end()

    def zoomIn(self):
        """
        Zoom in.
        """
        self.__view.zoomIn()

    def zoomOut(self):
        """
        Zoom out.
        """
        self.__view.zoomOut()

    def zoomToFit(self):
        """
        Adjust the zoom level so that the entire project fits into the view.
        """
        self.__view.zoomToFit()

    def zoomToOriginal(self):
        """
        Adjust the zoom level to 100%.
        """
        self.__view.zoomToOriginal()

    def grid(self, gflg):
        """
        Handle a grid button action.

        :param gflg: New snap flag value
        :type gflg: bool
        """
        # set the grid flag
        self.__scene.setGridFlag(gflg)

    def snap(self, sflg):
        """
        Handle a grid snapping action.

        :param sflg: New snap flag value
        :type sflg: bool
        """
        # set the snap flag
        self.__scene.setSnapFlag(sflg)

    def showProjectInfoDialog(self):
        """
        Display the meta-information dialog for the current project.
        """
        diaw = self.__pdgf.createProjectPropertiesDialog()
        diaw.show()
        diaw.exec_()

    def getOrCreateBottomWidget(self, code, generatorF):
        """
        Get the widget in the bottom drawer with the given code.
        If the widget does not exist yet/any more, it will be (re-)created
        using the given generator function.

        :param code: Code of the widget to look for
        :type code: str
        :param generatorF: Generator function to create the widget should it
                           not exist, and its title
        :type generatorF: _ -> (QWidget, str)
        :return: Either the existing or the newly created widget
        :rtype: QWidget
        """
        return self.__bottomDrawer.getOrCreateWidget(code, generatorF)

    def addBottomWidget(self, widget, title):
        """
        Add a widget with the given title to the bottom drawer.

        :param widget: Widget to add
        :type widget: QWidget
        :param title: Title of the widget
        :type title: str
        """
        self.__bottomDrawer.addWidget(widget, title)

    def showProjectLog(self):
        """
        Display the project's log window.
        """

        def mkProjectLog():
            if self.__projectLog is None:
                self.__projectLog = WguiLoggerWindow(
                    self.__projectContext.getMessageSource())
            return self.__projectLog, 'Log'

        self.getOrCreateBottomWidget('log', mkProjectLog)


class _WguiBottomDrawer(QTabWidget):
    """
    Project editor bottom drawer.
    """

    #: Signal emitted whenever the bottom drawer requests a different height
    minimumSizeRequested = pyqtSignal(int)

    #: Minimum height
    MIN_VISIBLE_HEIGHT = 50

    #: Initial default height
    DEFAULT_VISIBLE_HEIGHT = 200

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Last height prior to hiding
        self.__lastHeight = self.DEFAULT_VISIBLE_HEIGHT

        self.setDocumentMode(True)
        self.setMovable(True)
        self.setTabsClosable(True)
        self.setTabPosition(QTabWidget.South)
        self.currentChanged.connect(self._currentTabChanged)
        self.tabCloseRequested.connect(self._tabCloseRequested)

        self.setHidden(True)

    def getOrCreateWidget(self, code, generatorF):
        """
        Get the widget in the bottom drawer with the given code.
        If the widget does not exist yet/any more, it will be (re-)created
        using the given generator function.

        :param code: Code of the widget to look for
        :type code: str
        :param generatorF: Generator function to create the widget should it
                           not exist, and its title
        :type generatorF: _ -> (QWidget, str)
        :return: Either the existing or the newly created widget
        :rtype: QWidget
        """
        for idx in range(self.count()):
            widget = self.widget(idx)
            if widget.property('WCR_CODE') == code:
                self.setCurrentIndex(idx)
                return widget

        widget, title = generatorF()
        widget.setProperty('WCR_CODE', code)
        self.addWidget(widget, title)
        return widget

    def addWidget(self, widget, title):
        """
        Add a widget with the given title to the bottom drawer.

        :param widget: Widget to add
        :type widget: QWidget
        :param title: Title of the widget
        :type title: str
        """
        self.setHidden(False)
        if isinstance(widget, WguiChangeNotifyP):
            widget.contentChanged.connect(
                partial(self._contentChanged, widget))
        if isinstance(widget, WguiHasTitleP):
            title = widget.title
            widget.titleChanged.connect(partial(self._titleChanged, widget))
        idx = self.addTab(widget, title)
        self.setCurrentIndex(idx)
        if self.height() < self.MIN_VISIBLE_HEIGHT:
            self.minimumSizeRequested.emit(
                max(self.__lastHeight, self.MIN_VISIBLE_HEIGHT))

    def _currentTabChanged(self, index):
        """
        Callback slot for when the current editor changes.
        """
        self.tabBar().setTabTextColor(
            index, self.tabBar().palette().windowText().color())
        wdgt = self.currentWidget()
        if isinstance(wdgt, WguiEditor):
            wdgt.activated()

    def _tabCloseRequested(self, tab):
        """
        Callback slot to intercept tab closing.

        :param tab: Index of tab to be closed
        :type tab: int
        """
        self.removeTab(tab)

        if self.count() == 0:
            self.__lastHeight = self.height()
            self.setHidden(True)

    def _contentChanged(self, widget):
        """
        Callback function to handle content changes in a widget.

        :param widget: Widget with changed contents
        :type widget: QWidget
        """
        if not self.currentWidget() is widget:
            idx = self.indexOf(widget)
            if idx >= 0:
                self.tabBar().setTabTextColor(idx, Qt.red)

    def _titleChanged(self, widget, title):
        """
        Callback function to handle a widget's title change.

        :param widget: Widget with changed title
        :type widget: QWidget
        :param title: The widget's new title
        :type title: str
        """
        idx = self.indexOf(widget)
        if idx >= 0:
            self.tabBar().setTabText(idx, title)


class _WguiZoomSlider(QWidget):
    """
    UI element to allow zooming of a project view.
    """

    #: Zoom level display format
    LABEL_FMT = 'Zoom: %.2fX'

    #: Default minimum slider value (log scale)
    DEFAULT_MIN = math.floor(math.log(0.25, 2)) * 10

    def __init__(self, view, parent=None):
        """
        :param view: View to control zoom of
        :type view: WguiGraphicsView
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: View to control zoom of
        self.__view = view

        #: Whether the slider is updating the zoom level
        self.__sliderUpdating = False

        #: Zoom slider
        self.__slider = QSlider()
        #: Current zoom level label
        self.__zoomLabel = QLabel()

        view.scene().sceneRectChanged.connect(self._updateZoomSliderMinMax)
        self.__view.sizeChanged.connect(self._updateZoomSliderMinMax)
        self.__view.zoomChanged.connect(self._updateZoomSlider)
        self.__slider.valueChanged.connect(self._setZoom)

        self.setLayout(self._createLayout())

    def _createLayout(self):
        """
        Create the layout for the zoom slider.
        """
        layout = QHBoxLayout()
        layout.setMargin(2)
        self.__slider.setFixedWidth(200)
        self.__slider.setMaximum(10)
        self.__slider.setMinimum(self.DEFAULT_MIN)
        self.__slider.setValue(0)
        self.__slider.setTickInterval(1)
        self.__slider.setOrientation(Qt.Horizontal)
        self.__slider.setTickPosition(QSlider.TicksBelow)
        vbox = QVBoxLayout()
        vbox.setMargin(0)
        vbox.setSpacing(0)
        self.__zoomLabel.setText(self.LABEL_FMT % 1)
        self.__zoomLabel.setMargin(0)
        self.__zoomLabel.setFont(QFont(None, 8))
        vbox.addWidget(self.__zoomLabel)
        vbox.addWidget(self.__slider)
        layout.addLayout(vbox)
        return layout

    def _updateZoomSlider(self, zoom):
        """
        Update the zoom slider value to the given zoom level.

        :param zoom: New zoom level (1 = 100%)
        :type zoom: float
        """
        if self.__sliderUpdating: return
        self.__sliderUpdating = True
        self.__slider.setValue(math.log(zoom, 2) * 10)
        self.__zoomLabel.setText(self.LABEL_FMT % zoom)
        self.__sliderUpdating = False

    def _updateZoomSliderMinMax(self, *_):
        """
        Adjust the zoom slider's minimum and maximum value based on the view
        and scene size.
        """
        slider = self.__slider
        view = self.__view
        sr = view.scene().sceneRect()
        if not sr:
            return
        # noinspection PyUnresolvedReferences
        vw = view.width()
        # noinspection PyUnresolvedReferences
        vh = view.height()

        if sr.width() == 0 or sr.height == 0:
            slider.setMinimum(0)
        elif sr.width() > vw or sr.height() > vh or slider.minimum() < 0:
            wf = vw / sr.width()
            hf = vh / sr.height()
            f = min(wf, hf)
            if f > 0:
                slider.setMinimum(min(math.floor(math.log(f, 2)) * 10,
                                      self.DEFAULT_MIN))

    def _setZoom(self, value):
        """
        Callback function for when the zoom slider position changes.

        :param value: Current zoom slider position
        :type value: int
        """
        if self.__sliderUpdating: return
        self.__sliderUpdating = True
        zoom = math.pow(2, value / 10)
        self.__view.zoom = zoom
        self.__zoomLabel.setText(self.LABEL_FMT % zoom)
        self.__sliderUpdating = False


class _WguiToolButton(QToolButton):
    """
    Tool button.
    """

    def __init__(self, code, icon, tooltip):
        """
        :param code: Code of the toolbox item
        :type code: str
        :param icon: Button icon
        :type icon: QIcon
        :param tooltip: Tooltip
        :type tooltip: str
        """
        super().__init__()

        #: Item code
        self.__code = code

        if icon is not None:
            self.setIcon(icon)
        self.setIconSize(QSize(32, 32))
        self.setCheckable(True)
        self.setAutoRaise(True)
        self.setToolTip(tooltip)

    @property
    def code(self):
        """
        Get the code of the item associated with this button.

        :return: Item code
        :rtype: str
        """
        return self.__code


class _WguiDragToolButton(_WguiToolButton):
    """
    Draggable tool button.
    """

    def __init__(self, code, icon, tooltip):
        """
        :param code: Code of the toolbox item
        :type code: str
        :param icon: Button icon
        :type icon: QIcon
        :param tooltip: Tooltip
        :type tooltip: str
        """
        super().__init__(code, icon, tooltip)

        #: Item code
        self.__code = code

        #: Position at drag start
        self.__dragStartPos = None

    def mousePressEvent(self, event):
        self.__dragStartPos = event.pos()

        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        self.__dragStartPos = None
        super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event):

        if self.__dragStartPos is None:
            return
        if not event.buttons() & Qt.LeftButton:
            self.__dragStartPos = None
            return
        dist = (event.pos() - self.__dragStartPos).manhattanLength()
        # noinspection PyArgumentList
        if dist < QApplication.startDragDistance():
            return

        drag = QDrag(self)
        mimeData = QMimeData()

        mimeData.setData(WGUI_INST_DND_MIME_TYPE,
                         QByteArray(self.__code))
        drag.setMimeData(mimeData)
        drag.setPixmap(self.icon().pixmap(32, 32))

        self.__dragStartPos = None
        self.setDown(False)

        drag.exec(Qt.CopyAction | Qt.MoveAction)


class _WguiToolButtonGroup(QButtonGroup):
    """
    Toolbox button group.
    """

    def __init__(self, cctx, parent=None):
        """
        :param cctx: Cockpit context
        :type cctx: WguiCockpitSystemContextP
        :param parent: Parent widget
        """
        super().__init__(parent)

        self.__cctx = cctx

        #: Buttons by key
        self.__buttons = {}

        self.setExclusive(True)

    def addToolButton(self, key, drag):
        item = WGUI_TOOLBOX_ITEMS[key]
        icon = self.__cctx.getIconPath(key, WGUI_ICON_PNG32) or item.icon
        return self._addButton(
            item.key, QIcon(icon), item.description, drag)

    def addBlockButton(self, code, description):
        """
        Add a button for a toolbox item.

        :param code: Block code
        :type code: str
        :param description: Block description
        :type description: str
        """
        iconPath = self.__cctx.getIconPath(code, WGUI_ICON_PNG32)
        icon = QIcon(iconPath) if iconPath else None
        return self._addButton(
            code, icon, description, True)

    def addModelButton(self, mime):
        """
        Add a button for a toolbox item.

        :param mime: Toolbox item to add
        :type mime: ScmnMimeType
        """
        iconPath = self.__cctx.getIconPath(mime, WGUI_ICON_PNG32)
        icon = QIcon(iconPath) if iconPath else None
        return self._addButton(
            mime.mimeType, icon, mime.description, True)

    def addOperatorButton(self, oper):
        """
        Add a button for an operator.

        :param oper: Operator item to add
        :type oper: WhdbOperatorP
        """
        opcode = oper.getCode()
        iconPath = self.__cctx.getIconPath(oper, WGUI_ICON_PNG32)
        icon = QIcon(iconPath) if iconPath else None
        tooltip = oper.getInfo()
        # noinspection PyTypeChecker
        return self._addButton(opcode, icon, tooltip, True)

    def addButton(self, button, p_int=None):
        raise NotImplementedError()

    def _addButton(self, code, icon, tooltip, drag):
        """
        Add a button.

        :param code: Code of the toolbox item
        :type code: str
        :param icon: Button icon
        :type icon: QIcon
        :param tooltip: Tooltip
        :type tooltip: str
        """
        if drag:
            ibut = _WguiDragToolButton(code, icon, tooltip)
        else:
            ibut = _WguiToolButton(code, icon, tooltip)
            # add it to the button group
        super().addButton(ibut)
        self.__buttons[code] = weakref.ref(ibut)
        return ibut

    def getToolButton(self, code):
        """
        Get the button for the item with the given code.

        :param code: Code to return button for
        :type code: str
        :return: Button for the item with the given code
        :rtype: _WguiToolButton
        """
        ref = self.__buttons.get(code)

        if ref is not None:
            return ref()
        else:
            return None


class _WguiProjectEditorInstTarget(WguiEditorInstTargetP):
    def __init__(self, mainWdgtTrgt, sideBtnBar, sideBtnToggled):
        """
        :param mainWdgtTrgt: Main widget target
        :type mainWdgtTrgt: QLayout | QSplitter
        :param sideBtnBar: Side button bar
        :type sideBtnBar: WguiSideButtonBar
        :param sideBtnToggled: Side button toggled signal
        :type sideBtnToggled: pyqtBoundSignal[str, bool]
        """
        self.__mainWdgtTrgt = mainWdgtTrgt
        self.__sideBtnBar = sideBtnBar
        self.__sideButtonToggled = sideBtnToggled

    @property
    def sideButtonToggled(self):
        return self.__sideButtonToggled

    def insertMainWidget(self, code, widget):
        self.__mainWdgtTrgt.addWidget(widget)

    def installSideButton(self, code, label, side, toggle=False):
        assert side == WGUI_EDITOR_LEFT_SIDE
        assert toggle

        button = self.__sideBtnBar.addToggleButton(code, label)
        button.toggled.connect(partial(self.__sideButtonToggled.emit, code))

        return button


class _WguiProjectEditorActionExecutionContext(WguiActionExecutionContextBase):
    def __init__(self, editor, prmValSrc):
        """
        :param editor: Editor
        :type editor: WguiProjectEditor
        :param prmValSrc: Parameter value source
        :type prmValSrc: WactParameterValueSourceP
        """
        super().__init__(editor.projectContext, prmValSrc, editor)

        self.__editor = editor

    def getEditor(self):
        return self.__editor

    def getProjectContext(self):
        return self.__editor.projectContext
