# ----------------------------------------------------------------------------
# - wgui-project-editor-config                                               -
# - wild cruncher - cockpit: project editor configuration                    -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_project_editor_config
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher project editor configuration

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import abstractmethod, ABCMeta

from PyQt4 import QtCore
from PyQt4.QtCore import QObject, Qt

from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_frontend.gui.wgui_protocols import QObject_Meta_M


class WguiProjectEditorCfgStoreP(metaclass=ABCMeta):
    """
    Protocol for editor configuration stores.
    """

    @abstractmethod
    def __getitem__(self, key):
        """
        Get a configuration value.

        :param key: Configuration value to retrieve
        :type key: str
        :return: Configuration value for the given key, if any
        :rtype: str
        """
        pass

    @abstractmethod
    def __setitem__(self, key, value):
        """
        Set a configuration value.

        :param key: Configuration value to retrieve
        :type key: str
        :param value: Configuration value for the given key
        :type value: str
        """
        pass


class WguiProjectEditorDbCfgStore(WguiProjectEditorCfgStoreP):
    """
    Configuration table based editor configuration store.
    """

    def __init__(self, configTable):
        """
        :param configTable: Local configuration table
        :type configTable: ScmnConfigTable
        """
        self.__configTable = configTable

    def __setitem__(self, key, value):
        self.__configTable.setRecordValue(key, value)

    def __getitem__(self, key):
        return self.__configTable.getRecordValue(key)


class WguiProjectEditorCfgP(metaclass=ABCMeta):
    """
    Configuration for project editors.
    """

    @abstractmethod
    def addObserver(self, func):
        """
        Add a callback function to be called on configuration changes.

        :param func: Callback function
        :return: () -> _
        """
        raise NotImplementedError()

    @abstractmethod
    def removeObserver(self, func):
        """
        Remove a callback function to be called on configuration changes.

        :param func: Callback function
        :return: () -> _
        """
        raise NotImplementedError()

    # noinspection PyPropertyDefinition
    @property
    @abstractmethod
    def selectionMode(self):
        """
        Get the current selection mode.

        :return: The selection mode
        :rtype: int
        """
        raise NotImplementedError()

    # noinspection PyPropertyDefinition
    @property
    @abstractmethod
    def snapGridX(self):
        """
        Get the current horizontal snap grid value.

        :return: Horizontal snap grid value
        :rtype: float
        """
        raise NotImplementedError()

    # noinspection PyPropertyDefinition
    @property
    @abstractmethod
    def snapGridY(self):
        """
        Get the current vertical snap grid value.

        :return: Vertical snap grid value
        :rtype: float
        """
        raise NotImplementedError()

    # noinspection PyPropertyDefinition
    @property
    @abstractmethod
    def gridSizeX(self):
        """
        Get the current horizontal grid size value.

        :return: Horizontal grid size value
        :rtype: float
        """
        raise NotImplementedError()

    # noinspection PyPropertyDefinition
    @property
    @abstractmethod
    def gridSizeY(self):
        """
        Get the current vertical grid size value.

        :return: Vertical grid size value
        :rtype: float
        """
        raise NotImplementedError()

    # noinspection PyPropertyDefinition
    @property
    @abstractmethod
    def displayName(self):
        """
        Whether the name of entities should be displayed in the scene.

        :return: Whether to display entity names
        :rtype: bool
        """
        raise NotImplementedError()

    # noinspection PyPropertyDefinition
    @property
    @abstractmethod
    def displayInfo(self):
        """
        Whether the info string of entities should be displayed in the scene.

        :return: Whether to display entity info strings
        :rtype: bool
        """
        raise NotImplementedError()


class WguiProjectEditorCfgDefaults(WguiProjectEditorCfgP):
    """
    Editor configuration default values.
    """

    def addObserver(self, func):
        pass

    def removeObserver(self, func):
        pass

    @property
    def selectionMode(self):
        return Qt.IntersectsItemShape

    @property
    def snapGridX(self):
        return 5.0

    @property
    def snapGridY(self):
        return 5.0

    @property
    def gridSizeY(self):
        return 50.0

    @property
    def gridSizeX(self):
        return 50.0

    @property
    def displayName(self):
        return True

    @property
    def displayInfo(self):
        return True


class WguiProjectEditorCfg(QObject, WguiProjectEditorCfgP, WcmnReleasableP,
                           metaclass=QObject_Meta_M):
    """
    Configuration for project editors.
    """

    #: Signal to notify about configuration changes
    _configChanged = QtCore.pyqtSignal()
    """:type: pyqtBoundSignal"""

    #: Selection mode serialization-to-code map
    SEL_MODES = {
        'intersect': Qt.IntersectsItemShape,
        'contain': Qt.ContainsItemShape
    }

    #: Selection mode code-to-serialization map
    SEL_MODES_R = dict((v, k) for k, v in SEL_MODES.items())

    def __init__(self, fallback, store=None):
        """
        :param fallback: Fallback configuration if this configuration has no
                         value set for a requested setting
        :type fallback: WguiProjectEditorCfgP
        :param store: Configuration persistence store
        :type store: WguiProjectEditorCfgStoreP
        """
        super().__init__()

        #: Global configuration
        self.__fallback = fallback
        #: Configuration persistence store
        self.__store = store

        #: Selection mode, e.g. for rubber band selection
        self.__selectionMode = None
        #: Snap grid spacing X
        self.__snapGridX = None
        #: Snap grid spacing Y
        self.__snapGridY = None
        #: Grid size X
        self.__gridSizeX = None
        #: Grid size Y
        self.__gridSizeY = None
        #: Display name
        self.__displayName = None
        #: Display info
        self.__displayInfo = None

        if store is not None:
            self._readConfig()

        fallback.addObserver(self._globalConfigChanged)

    def release(self):
        self.__fallback.removeObserver(self._globalConfigChanged)
        self._configChanged.disconnect()

    def _readConfig(self):
        """
        Read configuration values from the persistence store.
        """

        def readBool(key):
            v = self.__store[key]
            if v:
                return v.lower()[0] == 't'
            else:
                return None

        def readFloat(key):
            v = self.__store[key]
            return float(v) if v else None

        try:
            selMode = self.__store['gui.selectionmode']
            self.__selectionMode = self.SEL_MODES.get(selMode)
            self.__snapGridX = readFloat('gui.snapx')
            self.__snapGridY = readFloat('gui.snapy')
            self.__gridSizeX = readFloat('gui.gridx')
            self.__gridSizeY = readFloat('gui.gridy')
            self.__displayName = readBool('gui.displayname')
            self.__displayInfo = readBool('gui.displayinfo')
        except AttributeError:
            pass

    def _globalConfigChanged(self):
        self._configChanged.emit()

    def addObserver(self, func):
        self._configChanged.connect(func)

    def removeObserver(self, func):
        try:
            self._configChanged.disconnect(func)
        except:
            pass

    @property
    def selectionMode(self):
        if self.__selectionMode is not None:
            return self.__selectionMode
        elif self.__fallback is not None:
            return self.__fallback.selectionMode
        else:
            return WguiProjectEditorCfg.DEFAULT_SEL_MODE

    @property
    def rawSelectionMode(self):
        """
        Get the current selection mode from this configuration,
        i.e. without querying its fallback configuration should no value be
        available.

        :return: Selection mode
        :rtype: int
        """
        return self.__selectionMode

    @selectionMode.setter
    def selectionMode(self, mode):
        """
        Set the selection mode.

        :param mode: New selection mode
        :type mode: int
        """
        if mode != self.__selectionMode:
            self.__selectionMode = mode
            self.__store['gui.selectionmode'] = self.SEL_MODES_R.get(mode, '')
            self._configChanged.emit()

    @selectionMode.deleter
    def selectionMode(self):
        """
        Unset the selection mode.
        The value provided by the fallback configuration will be used hence
        forward.
        """
        self.selectionMode = None

    @property
    def snapGridX(self):
        if self.__snapGridX is not None:
            return self.__snapGridX
        else:
            return self.__fallback.snapGridX

    @property
    def rawSnapGridX(self):
        """
        Get the current horizontal snap grid value from this configuration,
        i.e. without querying its fallback configuration should no value be
        available.

        :return: Horizontal snap grid value
        :rtype: float
        """
        return self.__snapGridX

    @snapGridX.setter
    def snapGridX(self, grid):
        """
        Set the horizontal snap grid value.

        :param grid: New horizontal snap grid value
        :type grid: float
        """
        if grid != self.__snapGridX:
            self.__snapGridX = grid
            self.__store['gui.snapx'] = str(grid) if grid is not None else ''
            self._configChanged.emit()

    @snapGridX.deleter
    def snapGridX(self):
        """
        Unset the horizontal snap grid value.
        The value provided by the fallback configuration will be used hence
        forward.
        """
        self.snapGridX = None

    @property
    def snapGridY(self):
        if self.__snapGridY is not None:
            return self.__snapGridY
        else:
            return self.__fallback.snapGridY

    @property
    def rawSnapGridY(self):
        """
        Get the current vertical snap grid value from this configuration,
        i.e. without querying its fallback configuration should no value be
        available.

        :return: Vertical snap grid value
        :rtype: float
        """
        return self.__snapGridY

    @snapGridY.setter
    def snapGridY(self, grid):
        """
        Set the vertical snap grid value.

        :param grid: New vertical snap grid value
        :type grid: float
        """
        if grid != self.__snapGridY:
            self.__snapGridY = grid
            self.__store['gui.snapy'] = str(grid) if grid is not None else ''
            self._configChanged.emit()

    @snapGridY.deleter
    def snapGridY(self):
        """
        Unset the vertical snap grid value.
        The value provided by the fallback configuration will be used hence
        forward.
        """
        self.snapGridY = None

    @property
    def gridSizeX(self):
        if self.__gridSizeX is not None:
            return self.__gridSizeX
        else:
            return self.__fallback.gridSizeX

    @property
    def rawGridSizeX(self):
        """
        Get the current horizontal grid size value from this configuration,
        i.e. without querying its fallback configuration should no value be
        available.

        :return: Horizontal grid size value
        :rtype: float
        """
        return self.__gridSizeX

    @gridSizeX.setter
    def gridSizeX(self, grid):
        """
        Set the horizontal grid size value.

        :param grid: New horizontal grid size value
        :type grid: float
        """
        if grid != self.__gridSizeX:
            self.__gridSizeX = grid
            self.__store['gui.gridx'] = str(grid) if grid is not None else ''
            self._configChanged.emit()

    @gridSizeX.deleter
    def gridSizeX(self):
        """
        Unset the horizontal grid size value.
        The value provided by the fallback configuration will be used hence
        forward.
        """
        self.gridSizeX = None

    @property
    def gridSizeY(self):
        if self.__gridSizeY is not None:
            return self.__gridSizeY
        else:
            return self.__fallback.gridSizeY

    @property
    def rawGridSizeY(self):
        """
        Get the current vertical grid size value from this configuration,
        i.e. without querying its fallback configuration should no value be
        available.

        :return: Vertical grid size value
        :rtype: float
        """
        return self.__gridSizeY

    @gridSizeY.setter
    def gridSizeY(self, grid):
        """
        Set the vertical grid size value.

        :param grid: New vertical grid size value
        :type grid: float
        """
        if grid != self.__gridSizeY:
            self.__gridSizeY = grid
            self.__store['gui.gridy'] = str(grid) if grid is not None else ''
            self._configChanged.emit()

    @gridSizeY.deleter
    def gridSizeY(self):
        """
        Unset the vertical grid size value.
        The value provided by the fallback configuration will be used hence
        forward.
        """
        self.gridSizeY = None

    @property
    def displayName(self):
        if self.__displayName is not None:
            return self.__displayName
        else:
            return self.__fallback.displayName

    @property
    def rawDisplayName(self):
        """
        Whether the name of entities should be displayed in the scene;
        obtains the setting from this configuration, i.e. without querying its
        fallback configuration should no value be available.

        :return: Whether to display entity names
        :rtype: bool
        """
        return self.__displayName

    @displayName.setter
    def displayName(self, display):
        """
        Set whether to display entity names in the scene.

        :param display: Whether to display entity names in the scene
        :type display: bool
        """
        if display != self.__displayName:
            self.__displayName = display
            if display is not None:
                self.__store['gui.displayname'] = str(display)
            else:
                self.__store['gui.displayname'] = ''
            self._configChanged.emit()

    @displayName.deleter
    def displayName(self):
        """
        Unset the display-entity-name setting.
        The value provided by the fallback configuration will be used hence
        forward.
        """
        self.displayName = None

    @property
    def displayInfo(self):
        if self.__displayInfo is not None:
            return self.__displayInfo
        else:
            return self.__fallback.displayInfo

    @property
    def rawDisplayInfo(self):
        """
        Whether the info string of entities should be displayed in the scene;
        obtains the setting from this configuration, i.e. without querying its
        fallback configuration should no value be available.

        :return: Whether to display entity info strings
        :rtype: bool
        """
        return self.__displayInfo

    @displayInfo.setter
    def displayInfo(self, display):
        """
        Set whether to display entity info strings in the scene.

        :param display: Whether to display entity info strings in the scene
        :type display: bool
        """
        if display != self.__displayInfo:
            self.__displayInfo = display
            if display is not None:
                self.__store['gui.displayinfo'] = str(display)
            else:
                self.__store['gui.displayinfo'] = ''
            self._configChanged.emit()

    @displayInfo.deleter
    def displayInfo(self):
        """
        Unset the display-entity-info-string setting.
        The value provided by the fallback configuration will be used hence
        forward.
        """
        self.displayInfo = None
