# ----------------------------------------------------------------------------
# - wgui-project-props-dialog                                                -
# - wild cruncher - cockpit: project properties dialog widgets               -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_project_props_dialog
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher project properties dialog

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QComboBox, QVBoxLayout, QLabel, QLineEdit, QFormLayout

from wcr_frontend.gui.wgui_props_dialog_widget import \
    WguiConfigTableProperties, WguiPropertiesWidgetP, WguiPropertiesDialog, \
    WguiGroupedMultiPropertiesWidgetP
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data, wgui_qcb_set_data
from wcr_frontend.gui.wgui_widgets import WguiRadioBoxGroup, \
    WguiDefaultDoubleSpinBox
from wcr_core.project.wprj_project_info import WprjProjectInfo


class WguiProjectPropertiesDialog(WguiPropertiesDialog):
    """
    The WguiProjectPropertiesDialog class is a dialog design to access
    and manipulate project metadata, project editor preferences and project
    configuration settings.
    """

    def __init__(self, prjContext, config, updateF, parent=None):
        """
        :param prjContext: Project context
        :type prjContext: WcorProjectContextP
        :param config: Project editor configuration
        :type config: WguiProjectEditorCfg
        :param updateF: Callback function to notify about saves
        :type updateF: _ -> _
        :param parent: Parent widget
        """
        super().__init__('Project Properties', updateF, parent=parent)

        pInfo = prjContext.getProjectDB().projectInfo
        pCfg = prjContext.getProjectDB().tops.getLocalConfiguration()
        defs = prjContext.getGlobalContext().getConfigurationValues()

        pwds = self._createProjectPropertiesWidgets(prjContext)
        if pwds:
            gwdg = WguiGroupedMultiPropertiesWidgetP(
                ('', _WguiProjectProperties(pInfo)),
                *pwds)
        else:
            gwdg = _WguiProjectProperties(pInfo)

        allTabs = [('General', gwdg),
                   ('Config Table', WguiConfigTableProperties(
                       pCfg, defaults=defs)),
                   ('Editor', WguiEditorConfigProperties(config))]
        tabs = self._createProjectPropertiesTabs(prjContext)
        if tabs:
            allTabs.extend(tabs)

        props = prjContext.getModule().getProjectProperties()
        if props:
            for psrc, pnam, prps in props:
                wdgt = self._createProjectModulePropertiesWidget(
                    psrc, prps, prjContext)
                allTabs.append((pnam, wdgt))

        self.setTabbedPropertyWidgets(*allTabs)

    # noinspection PyUnusedLocal
    def _createProjectPropertiesTabs(self, prjContext):
        """
        :rtype: list[(str, WguiPropertiesWidgetP)] | None
        """
        return None

    # noinspection PyUnusedLocal
    def _createProjectPropertiesWidgets(self, prjContext):
        """
        :rtype: list[WguiPropertiesWidgetP] | None
        """
        return None

    def _createProjectModulePropertiesWidget(self, prsc, prps, pctx):
        return _WguiModuleProjectProperties(prsc, prps, pctx.getProjectDB())


class _WguiProjectProperties(WguiPropertiesWidgetP):
    """
    Configuration widget for project properties.
    """

    def __init__(self, projectInfo, parent=None):
        """
        :param projectInfo: the project object
        :type projectInfo: WprjProjectInfo
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project info properties
        self.__projectInfo = projectInfo

        #: Name field
        self.__nameF = QLineEdit()
        #: Info field
        self.__infoF = QLineEdit()
        #: Author field
        self.__authorF = QLineEdit()
        #: Company field
        self.__companyF = QLineEdit()
        #: Version field
        self.__versionF = QLineEdit()

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.addWidget(QLabel("Project Name:"))
        layout.addWidget(self.__nameF)
        layout.addWidget(QLabel("Project Description:"))
        layout.addWidget(self.__infoF)
        layout.addWidget(QLabel("Author:"))
        layout.addWidget(self.__authorF)
        layout.addWidget(QLabel("Company:"))
        layout.addWidget(self.__companyF)
        layout.addWidget(QLabel("Version:"))
        layout.addWidget(self.__versionF)

        return layout

    def updateFields(self):
        """
        Update all fields based on the current values in the represented
        data structures.
        """
        project = self.__projectInfo
        self.__nameF.setText(project.projectName)
        self.__infoF.setText(project.projectInfo)
        self.__authorF.setText(project.author)
        self.__companyF.setText(project.company)
        self.__versionF.setText(project.version)

    def save(self):
        project = self.__projectInfo
        project.projectName = self.__nameF.text()
        project.projectInfo = self.__infoF.text()
        project.author = self.__authorF.text()
        project.company = self.__companyF.text()
        project.version = self.__versionF.text()


class WguiEditorConfigProperties(WguiPropertiesWidgetP):
    """
    Project editor settings configuration widget.
    """

    def __init__(self, editorCfg, parent=None):
        """
        :param editorCfg: Project editor settings
        :type editorCfg: WguiProjectEditorCfg
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Project editor config
        self.__editorCfg = editorCfg

        #: Selection mode widget
        self.__selModeF = QComboBox()
        #: Name display selection
        self.__dspNameF = WguiRadioBoxGroup.mkTrueFalseDefaultBoxes()
        #: Info display selection
        self.__dspInfoF = WguiRadioBoxGroup.mkTrueFalseDefaultBoxes()
        #: Horizontal grid snap field
        self.__gridXF = WguiDefaultDoubleSpinBox()
        #: Vertical grid snap field
        self.__gridYF = WguiDefaultDoubleSpinBox()
        #: Horizontal grid size field
        self.__snapXF = WguiDefaultDoubleSpinBox()
        #: Vertical grid size field
        self.__snapYF = WguiDefaultDoubleSpinBox()

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QLayout
        """
        self.__selModeF.addItem('Intersection', Qt.IntersectsItemShape)
        self.__selModeF.addItem('Inclusion', Qt.ContainsItemShape)
        self.__selModeF.addItem('Default', None)

        # add the elements
        layout = QFormLayout()
        layout.addRow(QLabel('Selection mode:'), self.__selModeF)
        layout.addRow(QLabel('Display name:'), self.__dspNameF)
        layout.addRow(QLabel('Display info:'), self.__dspInfoF)
        layout.addRow(QLabel('Horizontal Grid Size:'), self.__gridXF)
        layout.addRow(QLabel('Vertical Grid Size:'), self.__gridYF)
        layout.addRow(QLabel('Horizontal Snap Grid:'), self.__snapXF)
        layout.addRow(QLabel('Vertical Snap Grid:'), self.__snapYF)

        return layout

    def updateFields(self):
        """
        Update all fields based on the current values in the represented
        data structures.
        """
        config = self.__editorCfg

        wgui_qcb_set_data(config.rawSelectionMode, self.__selModeF)
        self.__dspNameF.setState(config.rawDisplayName)
        self.__dspInfoF.setState(config.rawDisplayInfo)
        self.__gridXF.setValue(config.gridSizeX,
                               config.rawGridSizeX is not None)
        self.__gridYF.setValue(config.gridSizeY,
                               config.rawGridSizeY is not None)
        self.__snapXF.setValue(config.snapGridX,
                               config.rawSnapGridX is not None)
        self.__snapYF.setValue(config.snapGridY,
                               config.rawSnapGridY is not None)

    def save(self):
        config = self.__editorCfg

        config.selectionMode = wgui_qcb_get_data(self.__selModeF)
        config.displayName = self.__dspNameF.state()
        config.displayInfo = self.__dspInfoF.state()
        config.gridSizeX = self.__gridXF.value()
        config.gridSizeY = self.__gridYF.value()
        config.snapGridX = self.__snapXF.value()
        config.snapGridY = self.__snapYF.value()


class _WguiModuleProjectProperties(WguiPropertiesWidgetP):
    def __init__(self, source, properties, projectDb, parent=None):
        """

        :param source: Source identifier
        :type source: str
        :param properties: Properties [(code, name)]
        :type properties: Iterable[(str, str)]
        :param projectDb: Project database
        :type projectDb: WprjProjectDatabase
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__source = source
        self.__properties = properties
        self.__projectDb = projectDb

        self.__propFields = {}
        """:type: dict[str, QLineEdit]"""

        for prpc, _ in self.__properties:
            self.__propFields[prpc] = QLineEdit()

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        """
        Create the layout.

        :return: Layout
        :rtype: QLayout
        """
        layout = QFormLayout()
        layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)

        for prpc, prpn in self.__properties:
            layout.addRow(QLabel(prpn + ':'), self.__propFields[prpc])

        return layout

    def updateFields(self):
        cfg = self.__projectDb.tops.getLocalConfiguration()

        for prpc, _ in self.__properties:
            val = cfg.getRecordValue('{}.{}'.format(self.__source, prpc))
            self.__propFields[prpc].setText(val or '')

    def save(self):
        cfg = self.__projectDb.tops.getLocalConfiguration()

        for prpc, _ in self.__properties:
            cfg.setRecordValue('{}.{}'.format(self.__source, prpc),
                               self.__propFields[prpc].text())
