# ----------------------------------------------------------------------------
# - wgui-props-dialog-factory                                                -
# - wild cruncher - cockpit: properties dialog box factory                   -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_props_dialog_factory
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher properties dialog box factory

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt4.QtCore import QObject, pyqtSignal

from wcr_frontend.gui.wgui_button_props_dialog import \
    WguiButtonPropertiesDialog
from wcr_frontend.gui.wgui_operator_instance_props_dialog import \
    WguiOperatorInstancePropertiesDialog
from wcr_frontend.gui.wgui_instance_props_dialog import \
    WguiInstancePropertiesDialog
from wcr_frontend.gui.wgui_props_dialog_factory_api import \
    WguiPropertiesDialogFactoryP
from wcr_frontend.gui.wgui_project_props_dialog import \
    WguiProjectPropertiesDialog
from wcr_frontend.gui.wgui_protocols import QObject_Meta_M
from wcr_frontend.gui.wgui_shape_props_dialog import WguiShapePropertiesDialog
from wcr_frontend.gui.wgui_connection_props_dialog import \
    WguiConnectionPropertiesDialog


class WguiPropertiesDialogFactory(QObject, WguiPropertiesDialogFactoryP,
                                  metaclass=QObject_Meta_M):
    """
    Default properties dialag factory.
    """

    #: Project-properties changed signal
    projectPropertiesChanged = pyqtSignal()

    def __init__(self, prjContext, config, scene, actions, parent=None):
        """
        :param prjContext: Project context
        :type prjContext: WcorProjectContextP
        :param config: Project editor configuration
        :type config: WguiProjectEditorCfgP
        :param scene: Graphic scene adapter
        :type scene: WguiSceneAdapterP
        :param actions: GUI actions
        :type actions: dict[str, WactActionP]
        :param parent: Parent widget
        """
        QObject.__init__(self)

        #: Project context
        self.__prjContext = prjContext
        #: Editor configuration
        self.__config = config
        #: Graphics scene
        self.__scene = scene
        #: Actions
        self.__actions = actions
        #: Parent widget
        self.__parent = parent

    @property
    def projectPropertiesChangedS(self):
        return self.projectPropertiesChanged

    def createProjectPropertiesDialog(self, parent=None):
        def updateF():
            pdb = self.__prjContext.getProjectDB()
            pdb.getTopHierarchy().setName(pdb.projectInfo.projectName)
            self.projectPropertiesChanged.emit()

        return self._createProjectPropertiesDialog(
            self.__prjContext, self.__config, updateF,
            parent or self.__parent)

    def _createProjectPropertiesDialog(self, prjContext, config, updateF,
                                       parent):
        return WguiProjectPropertiesDialog(
            prjContext, config, updateF, parent=parent)

    def createBlockInstancePropertiesDialog(self, inst, selectXpnt=None,
                                            parent=None):
        return self._createBlockInstancePropertiesDialog(
            inst, self.__prjContext,
            self._createInstanceSaveFunction(inst, parent),
            selectXpnt, parent or self.__parent)

    def _createBlockInstancePropertiesDialog(self, inst, pctx, updateF,
                                             selectXpnt, parent):
        return WguiInstancePropertiesDialog(
            inst, pctx, updateF, selectXpnt=selectXpnt, parent=parent)

    def createModelInstancePropertiesDialog(self, inst, selectXpnt=None,
                                            parent=None):
        return self._createModelInstancePropertiesDialog(
            inst, self.__prjContext,
            self._createInstanceSaveFunction(inst, parent),
            selectXpnt, parent or self.__parent)

    def _createModelInstancePropertiesDialog(self, inst, pctx, updateF,
                                             selectXpnt, parent):
        return WguiInstancePropertiesDialog(
            inst, pctx, updateF, selectXpnt=selectXpnt, parent=parent)

    def createOperatorInstancePropertiesDialog(self, inst, selectXpnt=None,
                                               parent=None):
        return self._createOperatorInstancePropertiesDialog(
            inst, self.__prjContext,
            self._createInstanceSaveFunction(inst, parent),
            selectXpnt, parent or self.__parent)

    def _createOperatorInstancePropertiesDialog(self, inst, pctx, updateF,
                                                selectXpnt, parent):
        """
        Create the actual operator instance properties dialog.

        This function may/should be overwritted by sub-classes as needed.

        :param inst: Operator instance to create properties dialog for
        :type inst: WhdbOperatorInstanceP
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param updateF: Callback function to notify about saves
        :type updateF: _ -> _
        :param selectXpnt: Connection point to show initially, if any
        :type selectXpnt: WhdbConnectionPointP
        :param parent: Parent widget
        :return: Newly created operator instance properties dialog
        :rtype: WguiPropertiesDialog
        """
        return WguiOperatorInstancePropertiesDialog(
            inst, pctx, updateF, selectXpnt=selectXpnt, parent=parent)

    def createConnectionPropertiesDialog(self, conn, parent=None):
        def save():
            if hook is not None:
                # noinspection PyCallingNonCallable
                hook(conn, self.__prjContext)
            self.__scene.getSceneItem(conn).updateItem()
            self.__scene.sceneItemChangedS.emit(conn)

        # noinspection PyNoneFunctionAssignment
        hook = self._createConnectionPostSaveHook(conn, self.__prjContext)

        return self._createConnectionPropertiesDialog(
            conn, save, parent)

    def _createConnectionPropertiesDialog(self, conn, updateF, parent):
        return WguiConnectionPropertiesDialog(conn, updateF, parent=parent)

    def createShapePropertiesDialog(self, shape, parent=None):
        def save():
            self.__scene.getSceneItem(shape).updateItem()
            self.__scene.sceneItemChangedS.emit(shape)

        return self._createShapePropertiesDialog(
            shape, save, parent or self.__parent)

    def _createShapePropertiesDialog(self, shape, updateF, parent):
        return WguiShapePropertiesDialog(shape, updateF, parent)

    def createButtonPropertiesDialog(self, button, parent=None):
        def save():
            self.__scene.getSceneItem(button).updateItem()
            self.__scene.sceneItemChangedS.emit(button)

        return self._createButtonPropertiesDialog(
            button, self.__actions, self.__prjContext, save,
            parent or self.__parent)

    def _createButtonPropertiesDialog(self, button, actions, pctx, updateF,
                                      parent):
        return WguiButtonPropertiesDialog(
            button, actions, pctx, updateF, parent)

    def _createInstanceSaveFunction(self, inst, parent=None):
        def save():
            if hook is not None:
                # noinspection PyCallingNonCallable
                hook(inst, parent)
            self.__scene.removeInvalidConnections()
            news = inst.getRotation(), inst.isFlippedVertically(), \
                   inst.isFlippedHorizontally()
            sceneItem = self.__scene.getSceneItem(inst)
            if sceneItem:
                if prvs == news:
                    self.__scene.updateConnectionsForInstance(sceneItem)
                else:
                    self.__scene.resetConnectionsForInstance(sceneItem)
                sceneItem.updateItem()
                self.__scene.sceneItemChangedS.emit(inst)

        prvs = inst.getRotation(), inst.isFlippedVertically(), \
               inst.isFlippedHorizontally()
        # noinspection PyNoneFunctionAssignment
        hook = self._createInstancePostSaveHook(inst, self.__prjContext)

        return save

    def _createInstancePostSaveHook(self, inst, msgTarget):
        return None

    def _createConnectionPostSaveHook(self, conn, msgTarge):
        return None
