# ----------------------------------------------------------------------------
# - wgui-record-table-widgets                                                -
# - wild cruncher: record table widgets                       frontend / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_record_table_widgets
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher - Record table UI widgets

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections.abc import Iterable
from abc import ABCMeta
from PyQt4.QtCore import Qt, QAbstractTableModel, QModelIndex, pyqtSignal
from PyQt4.QtGui import QTableView, QWidget, QPushButton, QVBoxLayout, \
    QHBoxLayout, QAbstractItemView, QHeaderView, QSortFilterProxyModel, \
    QItemDelegate, QComboBox

from wcr_core.db.whdb_parameter import WHDB_PARAMETER_VALUE_KEY
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data


_WGUI_FIELD_LABEL = 0

_WGUI_FIELD_KEY = 1

WGUI_MODEL_ROW_KEY_ROLE = Qt.UserRole

WGUI_MODEL_ENTRY_ROLE = Qt.UserRole + 1


class WguiRecordTableModelBase(QAbstractTableModel):
    editabilityChanged = pyqtSignal(bool)
    """:type: pyqtBoundSignal[bool]"""

    def __init__(self, fields, parent=None):
        """
        :param fields: Field names and keys [(Name, Key)]
        :type fields: list[tuple[str]]
        :param parent: Parent widget
        """
        super().__init__(parent)

        self.__fields = fields[:]

        #: Current entries by row key
        self.__entries = {}
        """:type: dict[object, dict[str, object]]"""
        #: Current item keys in row-order
        self.__keys = []
        """:type: list[object]"""

    @property
    def keysBase(self):
        """
        :rtype: list[object]
        """
        return self.__keys

    def _setDataEntries(self, entries):
        """
        Set the model data entries.

        :param entries: Model data
        :type entries: dict[object, object] | None
        """
        if entries is None:
            entries = {}
        self.beginResetModel()
        self.__entries = entries
        self.__keys = list(self.__entries.keys())
        self.endResetModel()
        self.editabilityChanged.emit(self.isEditable())

    def headerData(self, section, orientation, role=None):
        if role != Qt.DisplayRole:
            return None

        if orientation == Qt.Horizontal:
            return self.__fields[section][0]
        else:
            return str(section + 1)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.__fields)

    def flags(self, index):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsEditable

    def data(self, index, role=None):
        if not (index and index.isValid()):
            return None

        rkey = self.__keys[index.row()]
        entry = self.__entries[rkey]

        col = index.column()
        if 0 <= col < len(self.__fields):
            if role == Qt.DisplayRole or role == Qt.EditRole:
                key = self.__fields[col][_WGUI_FIELD_KEY]
                return self._getFieldData(key, rkey, entry)
            if role == Qt.ToolTipRole:
                key = self.__fields[col][_WGUI_FIELD_KEY]
                return self._getFieldTooltip(key, rkey, entry)
            elif role == WGUI_MODEL_ROW_KEY_ROLE:
                return rkey
            elif role == WGUI_MODEL_ENTRY_ROLE:
                return entry

    def _getFieldTooltip(self, field, rkey, entry):
        """
        Get the field tooltip for the given key.

        :param field: Field key to return value for
        :type field: str
        :param rkey: Row key
        :type rkey: object
        :param entry: Data to return field value for
        :type entry: dict[str, object]
        :return: Field value for the given key
        :rtype: str
        """
        return None

    def _getFieldData(self, field, rkey, entry):
        """
        Get the field data for the given key.

        :param field: Field key to return value for
        :type field: str
        :param rkey: Row key
        :type rkey: object
        :param entry: Data to return field value for
        :type entry: dict[str, object]
        :return: Field value for the given key
        :rtype: str
        """
        return entry.get(field)

    def isEditable(self):
        """
        Whether the model is currently editable.

        :return: True iff the model is currently editable
        :rtype: bool
        """
        raise NotImplementedError()

    def setData(self, index, data, role=None):
        if not role == Qt.EditRole:
            return False
        if not self.isEditable():
            return

        rkey = self.__keys[index.row()]
        entry = self.__entries[rkey]

        col = index.column()
        if 0 <= col < len(self.__fields):
            entry[self.__fields[col][_WGUI_FIELD_KEY]] = data
            return True
        return False

    def addEntry(self):
        """
        Insert a new (empty) entry into the data model.

        :return: Row key for the new entry, if addition successful
        :rtype: object | None
        """
        if not self.isEditable():
            return None
        rows = len(self.__entries)
        self.beginInsertRows(QModelIndex(), rows, rows)
        rkey, entry = self._createEmptyEntry()
        if entry is None:
            return None
        if rkey in self.__entries:
            raise Exception('Row key already exists')
        self.__entries[rkey] = entry
        self.__keys.append(rkey)
        self.endInsertRows()
        return rkey

    def _createEmptyEntry(self):
        """
        Create an empty entry for insertion into the data model.

        :return: Newly created entry
        :rtype: (object, dict[str, str])
        """
        raise NotImplementedError()

    def removeEntry(self, index):
        """
        Remove the entry at the given index from the data model.

        :param index: Index of entry to be removed
        :type index: QModelIndex
        :return: Row key of the removed entry, if removal successful
        :rtype: object | None
        """
        if not self.isEditable():
            return None
        row = index.row()
        self.beginRemoveRows(QModelIndex(), row, row)
        rkey = self.__keys[row]
        del self.__entries[rkey]
        del self.__keys[row]
        self.endRemoveRows()
        return rkey

    def externalDataChange(self, rkey):
        """
        Notify model of an external data change for the record with the given
        row key.

        :param rkey: Row key of changed record
        :type rkey: object
        """
        try:
            idx = self.__keys.index(rkey)

            # noinspection PyUnresolvedReferences
            self.dataChanged.emit(
                self.index(idx, 0), self.index(idx, len(self.__fields) - 1))
        except ValueError:
            pass


class WguiRecordTableModel(WguiRecordTableModelBase):
    #: Signal for changes in child data, to notify parent, with row key
    anyDataChanged = pyqtSignal(object)
    """:type: pyqtBoundSignal[object]"""

    def __init__(self, fields, pfld=None, parent=None):
        """
        :param fields: Field names and keys [(Name, Key)]
        :type fields: list[tuple[str]]
        :param pfld: Field in parent data to extract child record data
        :type pfld: str | None
        :param parent: Parent widget
        """
        super().__init__(fields, parent)

        self.__parentDataField = pfld
        self.__data = None
        self.__key = None
        """:type: object | None"""

        self.dataChanged.connect(self._anyDataChange)
        self.rowsInserted.connect(self._anyDataChange)
        self.rowsRemoved.connect(self._anyDataChange)

    def _anyDataChange(self, *_):
        """
        Internal slot for changes in child data, e.g. to notify parent.
        """
        self.anyDataChanged.emit(self.__key)

    def getModelData(self):
        """
        Get the model data.

        :return: Model data
        :rtype: object
        """
        return self.__data

    def setModelData(self, data, key=None):
        """
        Set the data.

        :param key: Row key of child data record
        :type key: object | None
        :param data: Child data record
        :type data: dict | None
        """
        if data is not None:
            self.__key = key
            if self.__parentDataField:
                data = data.get(self.__parentDataField)
            self.__data = data
            self._setDataEntries(data)
        else:
            self.__key = key
            self._setDataEntries(None)

    def isEditable(self):
        return self.__key is not None or self.__data is not None


class WguiParameterTableModel(WguiRecordTableModel):
    #: Channel fields
    _FIELDS = [
        ('Parameter', 'p'),
        ('Code', 'c'),
        ('Value', WHDB_PARAMETER_VALUE_KEY)]

    def __init__(self, pfld=None, paramDescs=None, paramDescF=None,
                 parent=None):
        """
        :type pfld: str | None
        :type paramDescs: dict[str, WcmnParameterDescriptionP] | None
        :type paramDescF: (dict, str | None) ->
                          dict[str, WcmnParameterDescriptionP] | None
        :type parent: QWidget
        """
        super().__init__(self._FIELDS, pfld, parent)

        self.__paramDescF = paramDescF

        self.__paramDescs = paramDescs
        """:type: dict[str, WcmnParameterDescriptionP] | None"""

        self.__userObj = {}
        """:type: dict[object, object]"""

    def data(self, index, role=None):
        """
        :param index: Model index
        :type index: QModelIndex
        :param role: Role from which to store
        :type role: int | None
        :return: Object to be returned
        :rtype: object
        """
        if role == Qt.UserRole:
            rkey = self.keysBase[index.row()]
            return self.__userObj.get(rkey)
        else:
            return super().data(index, role)

    def setData(self, index, data, role=None):
        """
        :param index: Model index
        :type index: QModelIndex
        :param data: Object to store
        :type data: object
        :param role: Role from which to store
        :type role: int | None
        """
        if role == Qt.UserRole:
            rkey = self.keysBase[index.row()]
            self.__userObj[rkey] = data
            return True
        else:
            return super().setData(index, data, role)

    def setModelData(self, data, key=None):
        if self.__paramDescF:
            self.__paramDescs = self.__paramDescF(data, key)

        super().setModelData(data, key)

    def _getFieldData(self, field, rkey, entry):
        if field == 'p' or field == 'c':
            pdscs = self.__paramDescs
            if pdscs:
                pdsc = pdscs.get(rkey)
                if pdsc:
                    if field == 'p':
                        return pdsc.getLabel()
                    elif field == 'c':
                        return pdsc.getName()
            return rkey
        return super()._getFieldData(field, rkey, entry)

    def _getFieldTooltip(self, field, rkey, entry):
        pdscs = self.__paramDescs
        if pdscs:
            pdsc = pdscs.get(rkey)
            if pdsc:
                return pdsc.getInfo()

    def flags(self, index):
        flgs = super().flags(index)

        if index.column() != 2:
            flgs &= ~Qt.ItemIsEditable

        return flgs

    def _createEmptyEntry(self):
        raise NotImplementedError()


class WguiRecordTableWidgetBase(QWidget):
    #: Signal for changes in current/selected record
    currentRecordChanged = pyqtSignal(object, object)
    """:type: pyqtBoundSignal[dict[str, object] | None, object | None]"""

    def __init__(self, model, parent=None):
        """
        :param model: Model to display
        :type model: WguiRecordTableModelBase
        :param parent: Parent widget
        """
        super().__init__(parent)

        self.__model = model

        self.__tableW = QTableView()
        self.__modelProxy = QSortFilterProxyModel(self)

    @property
    def _model(self):
        """
        :rtype: WguiRecordTableModelBase
        """
        return self.__model

    @property
    def _tableW(self):
        """
        :rtype: QTableView
        """
        return self.__tableW

    def _initElements(self):
        """
        Initialize UI elements.
        """
        self.__modelProxy.setSourceModel(self.__model)
        self.__tableW.setModel(self.__modelProxy)
        self.__tableW.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.__tableW.setSelectionMode(QAbstractItemView.SingleSelection)
        self.__tableW.setSortingEnabled(True)
        self._setTableLayout(self.__tableW, self.__model)
        self._relayoutTable(self.__tableW)

        self.__tableW.selectionModel().selectionChanged.connect(
            self._currentRecordChange)
        self.__model.modelReset.connect(self._modelReset)

    def _createLayout(self):
        """
        Create the layout.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setMargin(0)
        layout.addWidget(self.__tableW)

        return layout

    def _setTableLayout(self, tableW, model):
        """
        Set the initial table layout, such as column sizes.

        :type tableW: QTableView
        :type model: QAbstractTableModel
        """
        hdrView = tableW.horizontalHeader()
        """:type: QHeaderView"""
        for idx in range(model.columnCount()):
            hdrView.setResizeMode(idx, QHeaderView.Interactive)
        hdrView.setStretchLastSection(True)

    def _modelReset(self):
        """
        Slot for model reset notifications.
        """
        self._relayoutTable(self.__tableW)
        self.__tableW.sortByColumn(0, Qt.AscendingOrder)
        self._currentRecordChange()

    def _relayoutTable(self, tableW):
        """
        Re-layout the table widget, e.g. reset column sizes.

        :type tableW: QTableView
        """
        tableW.resizeColumnsToContents()

    def _currentRecordChange(self):
        idxs = self.__tableW.selectedIndexes()
        if idxs:
            idx = self.__modelProxy.mapToSource(idxs[0])
            rkey = self.__model.data(idx, WGUI_MODEL_ROW_KEY_ROLE)
            entry = self.__model.data(idx, WGUI_MODEL_ENTRY_ROLE)
            self.currentRecordChanged.emit(entry, rkey)
        else:
            self.currentRecordChanged.emit(None, None)


class WguiFixedRecordTableWidget(WguiRecordTableWidgetBase):
    def __init__(self, model, parent=None):
        """
        :param model: Model to display
        :type model: WguiRecordTableModelBase
        :param parent: Parent widget
        """
        super().__init__(model, parent)

        self._initElements()
        self.setLayout(self._createLayout())


class WguiRecordTableWidget(WguiRecordTableWidgetBase):
    def __init__(self, model, itemName=None, addWdgts=None, parent=None):
        """
        :param model: Model to display
        :type model: WguiRecordTableModelBase
        :param itemName: Item name (for buttons)
        :type itemName: str | None
        :param addWdgts: Additional widgets to display next to 'Add'-button
        :type addWdgts: Iterable[QWidget] | None
        :param parent: Parent widget
        """
        super().__init__(model, parent)

        self.__itemName = itemName or 'item'
        self.__addWdgts = addWdgts

        self.__addBttn = QPushButton('Add {}'.format(self.__itemName))
        self.__delBttn = QPushButton(
            'Remove selected {}'.format(self.__itemName))

        self._initElements()
        self.setLayout(self._createLayout())

    def _initElements(self):
        """
        Initialize UI elements.
        """
        super()._initElements()

        self._model.editabilityChanged.connect(self._modelEditabilityChanged)

        self.__addBttn.clicked.connect(self._addItem)
        self.__delBttn.clicked.connect(self._removeItem)

        self._modelEditabilityChanged(self._model.isEditable())

    def _createLayout(self):
        """
        Create the layout.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setMargin(0)

        btnLayout = QHBoxLayout()
        btnLayout.setMargin(0)
        if self.__addWdgts:
            wdgts = self.__addWdgts
            """:type: Iterable[QWidget]"""
            for wdgt in wdgts:
                btnLayout.addWidget(wdgt)
        btnLayout.addWidget(self.__addBttn)
        btnLayout.addStretch(1)
        btnLayout.addWidget(self.__delBttn)

        layout.addWidget(self._tableW)
        layout.addLayout(btnLayout)

        return layout

    def _currentRecordChange(self):
        super()._currentRecordChange()
        idxs = self._tableW.selectedIndexes()
        self.__delBttn.setEnabled(self._model.isEditable() and bool(idxs))

    def _modelEditabilityChanged(self, editable):
        """
        Slot for changes in the editable of the model.

        :param editable: Whether the model is currently editable
        :type editable: bool
        """
        if self.__addWdgts:
            wdgts = self.__addWdgts
            """:type: Iterable[QWidget]"""
            for wdgt in wdgts:
                wdgt.setEnabled(editable)
        self.__addBttn.setEnabled(editable)
        self.__delBttn.setEnabled(
            editable and bool(self._tableW.selectedIndexes()))

    def _addItem(self):
        """
        Slot for adding a new item to the model.

        :return: Row key for the new entry, if addition successful
        :rtype: object | None
        """
        rkey = self._model.addEntry()
        rcnt = self._model.rowCount()
        if rcnt == 1:
            self._tableW.resizeColumnsToContents()
        self._tableW.selectRow(rcnt - 1)
        return rkey

    def _removeItem(self):
        """
        Slot for removing the current item from the model.

        :return: Row key of the removed entry, if removal successful
        :rtype: object | None
        """
        idxs = self._tableW.selectedIndexes()
        if not idxs:
            return None
        return self._model.removeEntry(idxs[0])

################################################################
# Variable delegated editors for use in table widget
################################################################


class WguiRecordTableItemDelegate(QItemDelegate):
    """
    Generic delegate which creates editors based on a dictionary joining codes
    that can be found in a column of a table where the delegate is set with
    delegate editor descriptors
    """

    def __init__(self, delegDct, codeCol, parent=None):
        """
        :param delegDct: Dictionary containing the codes of columns as keys
            and as values tuples of the type of delegate to use and the
            appropriate parameters necessary to set them
        :type delegDct: dict[str, _WpguiPDSParamValDescP]
        :param codeCol: Column id which entries are used for code matching
        :type codeCol: int
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)
        self.__delegDict = delegDct
        self.__codeCol = codeCol
        assert(codeCol >= 0)

    def createEditor(self, parent, option, mIdx):
        """
        :param parent: Parent widget
        :type parent: QWidget
        :param option: Option
        :type option: QStyleOptionViewItem
        :param mIdx: Model index
        :type mIdx: QModelIndex
        :return: Editor widget
        :rtype: QWidget
        """
        mdl = mIdx.model()
        """:type: QAbstractItemModel"""
        assert mdl

        if self.__codeCol >= mdl.columnCount():
            raise Exception('Code column used is out of the model bounds')

        curDataSel = mdl.itemData(mIdx).get(Qt.EditRole)

        dataCode = mdl.itemData(mdl.index(mIdx.row(), self.__codeCol))
        codeStr = dataCode.get(Qt.EditRole)
        if not dataCode:
            raise Exception('Cannot create editor, entry code is undefined')
        assert codeStr

        delegate = None
        delegDesc = self.__delegDict.get(codeStr)
        """:type: _WpguiPDSParamValDescP | None"""
        if delegDesc is not None:

            if isinstance(delegDesc, WguiRecordTableDelegateDescCmbBox):
                delegate = self._createCmbBoxDelegate(parent, delegDesc)
                assert(delegate is not None)
                if curDataSel is not None:
                    foundIndex = delegate.findData(curDataSel, Qt.EditRole)
                    if foundIndex != -1:
                        delegate.setCurrentIndex(foundIndex)

        return delegate or super().createEditor(parent, option, mIdx)

    # noinspection PyMethodMayBeStatic
    def _createCmbBoxDelegate(self, parent, delegDesc):
        """
        :param parent: Parent widget
        :type parent: QWidget
        :param delegDesc: Combo box delegate descriptor
        :type delegDesc: _WpguiPDSParamValComboBoxDesc
        :return: Combo box widget
        :rtype: QComboBox
        """
        editor = QComboBox(parent)

        for chStr, chObj in delegDesc.getChoices():
            editor.addItem(chStr, chObj)

        return editor

    def setEditorData(self, editor, index):
        """
        :param editor: Editor widget
        :type editor: QWidget
        :param index: Model index
        :type index: QModelIndex
        """
        return super().setEditorData(editor, index)

    def setModelData(self, editor, model, index):
        """
        :param editor: Editor widget
        :type editor: QWidget
        :param model: Item model
        :type model: QAbstractItemModel
        :param index: Model index
        :type index: QModelIndex
        """
        if isinstance(editor, QComboBox):
            model.setData(index, editor.currentText(), Qt.EditRole)
            model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)
        else:
            super().setModelData(editor, model, index)

# -- Delegate descriptors


class WguiRecordTableDelegateDescP(metaclass=ABCMeta):
    """
    Generic delegate descriptor protocol for which all implementing classes
    indicate a type of editor delegate, using this class as a common ancestor
    """

    def getDelegateType(self):
        """
        :return: A string representation of the delegate type
        :rtype: str
        """
        raise NotImplementedError()


class WguiRecordTableDelegateDescBase(WguiRecordTableDelegateDescP):
    """
    Base class of generic delegate descriptor
    """

    TYPE_DESCRIPTION_PREFIX = '(RecTbl Deleg): '

    def __init__(self, dType):
        """
        :param dType: String representation of the editor delegate type
        :type dType: str
        """
        self.__dType = dType

    def getDelegateType(self):
        """
        :rtype: str
        """
        return self.__dType

    def __str__(self):
        """
        :rtype: str
        """
        return self.getDelegateType()

    def __repr__(self):
        """
        :rtype: str
        """
        return self.TYPE_DESCRIPTION_PREFIX + self.getDelegateType()


class WguiRecordTableDelegateDescCmbBox(WguiRecordTableDelegateDescBase):
    """
    Implementation of combo box parameter value descriptor
    """

    TYPE_DESCRIPTION = 'Combo Box'

    def __init__(self, provider):
        """
        :param provider: Choices provider
        :type provider: (self: _, None) -> list[str, object]
        """
        super().__init__(self.TYPE_DESCRIPTION)
        self.__chPrv = provider

    def getChoices(self):
        """
        :return: Combo box choices as a list of stringable objects
        :rtype: list[str, object]
        """
        choices = self.__chPrv()
        assert(isinstance(choices, list))
        return choices