# ----------------------------------------------------------------------------
# - wgui-resizer                                                             -
# - wild cruncher: graphical element resizing                                -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 clemens oertel                                   -
# - copyright (c) 2013-2013 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_resizer
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graphical element resizing

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import abstractmethod, ABCMeta
from math import copysign

from PyQt5 import QtCore
from PyQt5.QtCore import QRectF, QPointF, Qt
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsRectItem, QStyle
from PyQt5.QtGui import QPainter, QColor, QBrush, QPen, \
    QPainterPath, QTransform

from wcr_frontend.gui.wgui_protocols import QGraphicsItem_Meta_M, \
    WguiZoomAwareP
from wcr_frontend.gui.wgui_utils import WGUI_SELECTION_COLOUR, \
    WGUI_HOVER_COLOUR


# noinspection PyAbstractClass
class WguiResizer(WguiZoomAwareP):
    """
    Base class for resizers.

    Resizers are responsible for providing the resizing functionality for
    graphical objects/shapes.

    A resizer holds one or more resize handles.
    A resizer instance can only be used for the resizing of a single shape.
    """

    @abstractmethod
    def shapeChanged(self):
        """
        Notify the resizer that the associated graphical element has changed.
        """
        raise NotImplementedError()

    @abstractmethod
    def setZValue(self, zval):
        raise NotImplementedError()


class WguiResizeHandle(metaclass=ABCMeta):
    """
    Base class for resizing handles.

    A resizing handle allows a user to resize a shape by changing the position
    of the corner/end the handle is linked to. It is displayed as a small
    draggable shape on or next to the shape's corner/end it is linked to.

    A resizing handle is usually given two functions, one positioning function
    that provides its current position (e.g. based on the shape's bounding
    box), and a callback function that it will call every time the user
    drags the handle.
    """
    #: Size of the handle (diameter/side)
    SIZE = 8
    #: Half-size of the handle (radius/half-side)
    HSIZ = SIZE / 2

    @abstractmethod
    def _setResizer(self, resizer:WguiResizer):
        """
        Set the resizer that this handle belongs to.

        :param resizer: Resizer that this handle belongs to.
        :type resizer: WguiResizer
        """
        pass

    @abstractmethod
    def updatePosition(self):
        """
        Notify the handle that its position may have changed.
        The handle will reposition itself to the new position as indicated by
        its positioning function.
        """
        pass

    @abstractmethod
    def boundingRect(self) -> QRectF:
        """
        .. seealso:: :py:meth:`QGraphicsItem.boundingRect`
        """
        pass

    @abstractmethod
    def setVisible(self, visible:bool):
        """
        .. seealso:: :py:meth:`QGraphicsItem.setVisible`
        """
        pass


# noinspection PyAbstractClass
class WguiAbstractResizeHandle(QGraphicsRectItem, WguiResizeHandle,
                               metaclass=QGraphicsItem_Meta_M):
    """
    Abstract base class for resize handles, providing some common
    functionality.
    """

    def __init__(self):
        # noinspection PyArgumentList
        QGraphicsItem.__init__(self, -self.HSIZ, -self.HSIZ,
                               self.SIZE, self.SIZE)

        #: Resizer this handle belongs to
        self.__resizer = None

        self.setBrush(QColor(127, 127, 127, 127))
        self.setPen(QPen(QtCore.Qt.NoPen))

    def _setResizer(self, resizer:WguiResizer):
        self.__resizer = resizer
        if isinstance(resizer, QGraphicsItem):
            self.setParentItem(resizer)

    def _updateSiblingHandles(self):
        """
        Update the positions of all the handles that belong to the handle's
        resizer.
        """
        if self.__resizer is not None:
            self.__resizer.shapeChanged()

    def boundingRect(self):
        return QGraphicsRectItem.boundingRect(self)

    def setVisible(self, visible:bool):
        QGraphicsRectItem.setVisible(self, visible)


class WguiRelativeResizeHandle(WguiAbstractResizeHandle):
    """
    Resize handle that works with relative positions and uses Qt's dragging
    mechanism.

    Note that due to Qt's internal workings, this handle type should not be
    used with selectable shapes, because it will have the same selection state
    as its parent and thus not separately be movable if the parent is selected.
    """

    def __init__(self, posF, moveF, basePosF):
        """
        :param posF: Positioning function providing the handle's current
                     position
        :type posF: None -> (float, float)
        :param moveF: Position update callback function
        :type moveF: callable (float, float) -> None
        :param basePosF: Function providing the resized shape's position
        :type basePosF: None -> (float, float)
        """
        super().__init__()

        self.__moveF = moveF
        self.__posF = posF
        self.__basePosF = basePosF

        self.__basePos = None

        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges)

        self.updatePosition()

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            if self.__basePos is not None:
                basePos = self.__basePosF()
                xDiff = basePos.x() - self.__basePos.x()
                yDiff = basePos.y() - self.__basePos.y()
                value = QPointF(value.x() - xDiff, value.y() - yDiff)
        if change == QGraphicsItem.ItemPositionChange:
            self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges, False)
            self.__moveF((value.x(), value.y()))
            self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges, True)
            self._updateSiblingHandles()
        return value

    def mousePressEvent(self, event):
        self.__basePos = self.__basePosF()
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        self.__basePos = None
        super().mouseReleaseEvent(event)

    def updatePosition(self):
        self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges, False)
        self.setPos(*self.__posF())
        self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges, True)


class WguiAbsoluteResizeHandle(WguiAbstractResizeHandle):
    """
    Resize handle that works with absolute/scene positions and handles dragging
    manually.
    """

    def __init__(self, posF, moveF, xoff=0, yoff=0):
        """
        :param posF: Positioning function providing the handle's current
                     position in scene coordinates
        :type posF: None -> (float, float)
        :param moveF: Position update callback function (scene coordinates)
        :type moveF: (float, float) -> None
        :param xoff: X-offset against the resizing rectangle
        :type xoff: float
        :param yoff: Y-offset against the resizing rectangle
        :type yoff: float
        """
        super().__init__()

        #: Positioning function
        self.__posF = posF
        #: Position update callback function
        self.__moveF = moveF
        #: X-offset against the resizing rectangle
        self.__xoffr = xoff
        #: Y-offset against the resizing rectangle
        self.__yoffr = yoff

        #: X-offset against the resizing rectangle
        self.__xoff = 0
        #: Y-offset against the resizing rectangle
        self.__yoff = 0

        #: Whether the handle is currently being dragged
        self.__isDragged = False
        #: Whether the handle is currently being hovered over
        self.__isHovered = False

        self.setAcceptedMouseButtons(Qt.LeftButton)
        self.setAcceptHoverEvents(True)

        self._updateOffsets()
        self.updatePosition()

    def _updateBrush(self):
        if self.__isDragged:
            self.setBrush(WGUI_SELECTION_COLOUR)
        elif self.__isHovered:
            self.setBrush(WGUI_HOVER_COLOUR)
        else:
            self.setBrush(QColor(127, 127, 127, 127))

    def mousePressEvent(self, event):
        self.__isDragged = True
        self._updateSiblingHandles()
        self._updateBrush()

    def mouseReleaseEvent(self, event):
        self.__isDragged = False
        self._updateBrush()

    def mouseMoveEvent(self, event):
        if self.__isDragged:
            pos = event.scenePos()
            self.__moveF((pos.x() - self.__xoff, pos.y() - self.__yoff))
            self._updateSiblingHandles()
        super().mouseMoveEvent(event)

    def hoverEnterEvent(self, event):
        self.__isHovered = True
        self._updateBrush()

    def hoverLeaveEvent(self, event):
        self.__isHovered = False
        self._updateBrush()

    def updatePosition(self):
        if self.parentItem() is not None:
            x, y = self.__posF()
            self.setPos(self.parentItem().mapFromScene(x + self.__xoff,
                                                       y + self.__yoff))

    def setTransform(self, trans, combine=False):
        super().setTransform(trans, combine)

        self._updateOffsets(trans.m11())

    def resetTransform(self):
        super().resetTransform()

        self._updateOffsets()

    def _updateOffsets(self, zoom=1.0):
        xoffr = self.__xoffr
        if xoffr:
            self.__xoff = (xoffr + copysign(self.HSIZ, xoffr)) * zoom
        else:
            self.__xoff = 0

        yoffr = self.__yoffr
        if yoffr:
            self.__yoff = (yoffr + copysign(self.HSIZ, yoffr)) * zoom
        else:
            self.__yoff = 0


# noinspection PyAbstractClass
class WguiAbstractResizer(QGraphicsItem, WguiResizer, WguiZoomAwareP,
                          metaclass=QGraphicsItem_Meta_M):
    """
    Abstract base class for resizer, providing some common functionality.
    """

    def __init__(self, parent=None, scene=None):
        """
        :param parent: Parent graphic item
        :param scene: Graphic scene the resizer belongs to
        """
        super().__init__(parent)

        #: Handles belonging to the resizer
        self.__handles = []
        #: Current zoom level, used for adjusting handle sizes
        self.__zoom = 1.0

        self.setAcceptHoverEvents(True)

    def addHandle(self, handle):
        """
        Add a handle to the resizer.

        Note that a resize handle can only belong to one resizer handler.

        :param handle: Resizing handle to add
        :type handle: WguiResizeHandle
        """
        self.__handles.append(handle)
        handle.setVisible(False)
        # noinspection PyProtectedMember
        handle._setResizer(self)
        handle.updatePosition()

    def boundingRect(self):
        brct = QRectF()
        for handle in self.__handles:
            hrct = handle.boundingRect()
            brct |= hrct.translated(handle.pos())
        return brct

    def paint(self, painter, options, widget=None):
        pass

    def _updateHandlePositions(self):
        for handle in self.__handles:
            handle.updatePosition()

    def hoverEnterEvent(self, event):
        self.shapeChanged()
        for handle in self.__handles:
            handle.setVisible(True)
        self.update()

    def hoverLeaveEvent(self, event):
        self.shapeChanged()
        for handle in self.__handles:
            handle.setVisible(False)

    def zoomChanged(self, zoom):
        self.__zoom = zoom
        if zoom != 1.0 and zoom > 0:
            trns = QTransform().scale(1.0 / zoom, 1.0 / zoom)
            for handle in self._handles:
                handle.setTransform(trns)
        else:
            for handle in self._handles:
                handle.resetTransform()

    @property
    def _handles(self):
        return self.__handles

    @property
    def _zoom(self):
        return self.__zoom


class WguiGenericShapeResizer(WguiAbstractResizer):
    """
    Resizer holding an arbitrary number of resizing handles provided by its
    clients.

    It requires a shape providing function to determine the resized object's
    shape path for hovering/selection detection.
    """

    def __init__(self, shapeF, parent=None, scene=None):
        """
        :param shapeF: Function providing the shape path of the resized shape
                       (in the shape's coordinates), taking an additional-width
                       parameter as argument
        :type shapeF: (float) -> QPainterPath
        :param parent: Parent graphic item
        :param scene: Graphic scene the resizer belongs to
        """
        super().__init__(parent, scene)

        #: Shape path providing function
        self.__shapeF = shapeF
        self.__shape = None

        self.shapeChanged()

    def shape(self):
        return self.__shape if self.__shape is not None else super().shape()

    def shapeChanged(self):
        self.__shape = self.__shapeF(WguiResizeHandle.HSIZ)
        self._updateHandlePositions()


# noinspection PyTupleAssignmentBalance
class WguiBoundingRectResizerBase(WguiAbstractResizer):
    X1 = 1
    Y1 = 2
    X2 = 4
    Y2 = 8

    # noinspection PyTypeChecker
    def __init__(self, shapeF, brctCBF, adjustF, margin=0, drawRect=False,
                 parent=None, scene=None):
        """
        :param shapeF: Function providing the resized shape's current bounding
                       rect in scene coordinates and its path in its own
                       coordinates
        :type shapeF: None -> (QRectF, QPainterPath)
        :param brctCBF: Bounding rect change callback function, called by the
                        resizer whenever the bounding rect is modified through
                        one of its handles
        :type brctCBF: ((float, float), (float, float)) -> None
        :param adjustF: Point coordinates adjustment function, e.g. for grid-
                        snapping
        :type adjustF: callable ((float, float), (float, float)) ->
                         ((float, float), (float, float))
        :param margin: Margin between handles and resizing rect
        :type margin: float
        :param drawRect: Whether to draw a bounding rect,
                         e.g. for non-rectangular shapes
        :type drawRect: bool
        :param parent: Parent graphic item
        :param scene: Graphic scene the resizer belongs to
        """
        super().__init__(parent, scene)

        #: Bounding rect/Shape path providing function
        self.__shapeF = shapeF
        #: Bounding rect change callback function
        self.__brctCBF = brctCBF
        #: Point coordinates adjustment function
        self.__adjustF = adjustF
        #: Whether to draw a bounding rect
        self.__drawRect = drawRect

        #: The current bounding rect
        self.__rect = (0, 0), (0, 0)
        #: The shape's current path
        self.__shapePath = None

        self.addHandle(WguiAbsoluteResizeHandle(
            self._cornerA, self._setCornerA, -margin, -margin))
        self.addHandle(WguiAbsoluteResizeHandle(
            self._cornerAB, self._setCornerAB, -margin, margin))
        self.addHandle(WguiAbsoluteResizeHandle(
            self._cornerB, self._setCornerB, margin, margin))
        self.addHandle(WguiAbsoluteResizeHandle(
            self._cornerBA, self._setCornerBA, margin, -margin))
        self.addHandle(WguiAbsoluteResizeHandle(
            self._lineA, self._setLineA, -margin, 0))
        self.addHandle(WguiAbsoluteResizeHandle(
            self._lineB, self._setLineB, 0, -margin))
        self.addHandle(WguiAbsoluteResizeHandle(
            self._lineC, self._setLineC, margin, 0))
        self.addHandle(WguiAbsoluteResizeHandle(
            self._lineD, self._setLineD, 0, margin))

        self.shapeChanged()

    @property
    def _rect(self):
        """
        Get the current rectangle.

        :return: Current rectangle
        :rtype: ((float, float), (float, float))
        """
        return self.__rect

    def _setRect(self, p1, p2, dirs):
        """
        Set the rectangle.

        :param p1: First corner point of new rectangle
        :type p1: (float, float)
        :param p2: Second corner point of new rectangle
        :type p2: (float, float)
        """
        self.__rect = p1, p2
        self.__brctCBF(self.__rect)

    def shapeChanged(self):
        rect, self.__shapePath = self.__shapeF()
        rc = rect.center()
        rw = abs(rect.width()) / 2
        rh = abs(rect.height()) / 2
        (x1, y1), (x2, y2) = self._rect
        fx = x1 > x2
        fy = y1 > y2
        self.__rect = ((rc.x() + rw if fx else rc.x() - rw,
                        rc.y() + rh if fy else rc.y() - rh),
                       (rc.x() - rw if fx else rc.x() + rw,
                        rc.y() - rh if fy else rc.y() + rh))
        self._updateHandlePositions()

    def _cornerA(self):
        return self._rect[0]

    def _setCornerA(self, point:(float, float)):
        self._setRect(self.__adjustF(point), self._rect[1], self.X1 | self.Y1)

    def _cornerAB(self):
        (x1, _), (_, y2) = self._rect
        return x1, y2

    def _setCornerAB(self, point:(float, float)):
        x, y = self.__adjustF(point)
        (_, y1), (x2, _) = self._rect
        self._setRect((x, y1), (x2, y), self.X1 | self.Y2)

    def _cornerB(self):
        return self.__rect[1]

    def _setCornerB(self, point:(float, float)):
        self._setRect(self._rect[0], self.__adjustF(point), self.X2 | self.Y2)

    def _cornerBA(self):
        (_, y1), (x2, _) = self._rect
        return x2, y1

    def _setCornerBA(self, point:(float, float)):
        x, y = self.__adjustF(point)
        (x1, _), (_, y2) = self._rect
        self._setRect((x1, y), (x, y2), self.X2 | self.Y1)

    def _lineA(self):
        (x1, y1), (_, y2) = self._rect
        return x1, y1 + (y2 - y1) / 2

    def _setLineA(self, point:(float, float)):
        x, _ = self.__adjustF(point)
        (_, y1), (x2, y2) = self._rect
        self._setRect((x, y1), (x2, y2), self.X1)

    def _lineB(self):
        (x1, y1), (x2, _) = self._rect
        return x1 + (x2 - x1) / 2, y1

    def _setLineB(self, point:(float, float)):
        _, y = self.__adjustF(point)
        (x1, _), (x2, y2) = self._rect
        self._setRect((x1, y), (x2, y2), self.Y1)

    def _lineC(self):
        (_, y1), (x2, y2) = self._rect
        return x2, y1 + (y2 - y1) / 2

    def _setLineC(self, point:(float, float)):
        x, _ = self.__adjustF(point)
        (x1, y1), (_, y2) = self._rect
        self._setRect((x1, y1), (x, y2), self.X2)

    def _lineD(self):
        (x1, _), (x2, y2) = self._rect
        return x1 + (x2 - x1) / 2, y2

    def _setLineD(self, point:(float, float)):
        _, y = self.__adjustF(point)
        (x1, y1), (x2, _) = self._rect
        self._setRect((x1, y1), (x2, y), self.Y2)

    def paint(self, painter:QPainter, options, widget=None):
        super().paint(painter, options, widget)
        if self.__drawRect and options.state & QStyle.State_MouseOver:
            painter.save()
            painter.setPen(QPen(QBrush(Qt.gray), 1 / self._zoom, Qt.DashLine))
            adj = WguiResizeHandle.HSIZ
            rect = self.boundingRect().adjusted(adj, adj, -adj, -adj)
            painter.drawRect(rect)
            painter.restore()

    def shape(self):
        path = QPainterPath()
        path.setFillRule(Qt.WindingFill)
        path.addRect(self.boundingRect())
        if self.__shapePath is not None:
            path = path.subtracted(self.__shapePath)
        path2 = QPainterPath()
        for handle in self._handles:
            path2.addRect(self.mapRectFromItem(handle, handle.boundingRect()))
        return path.united(path2)


# noinspection PyTupleAssignmentBalance
class WguiBoundingRectResizer(WguiBoundingRectResizerBase):
    """
    Resizer using 8 resizing handles around the resized shape's bounding box.
    """

    def __init__(self, shapeF, brctCBF, adjustF, drawRect=False, parent=None,
                 scene=None):
        """
        :param shapeF: Function providing the resized shape's current bounding
                       rect in scene coordinates and its path in its own
                       coordinates
        :type shapeF: None -> (QRectF, QPainterPath)
        :param brctCBF: Bounding rect change callback function, called by the
                        resizer whenever the bounding rect is modified through
                        one of its handles
        :type brctCBF: ((float, float), (float, float)) -> None
        :param adjustF: Point coordinates adjustment function, e.g. for grid-
                        snapping
        :type adjustF: callable ((float, float), (float, float)) ->
                         ((float, float), (float, float))
        :param drawRect: Whether to draw a bounding rect,
                         e.g. for non-rectangular shapes
        :type drawRect: bool
        :param parent: Parent graphic item
        :param scene: Graphic scene the resizer belongs to
        """
        super().__init__(shapeF, brctCBF, adjustF, 0, drawRect, parent, scene)


# noinspection PyTupleAssignmentBalance
class WguiNoFlipBoundingRectResizer(WguiBoundingRectResizerBase):
    """
    Resizer using 8 resizing handles around the resized shape's bounding box.
    """

    def __init__(self, shapeF, brctCBF, adjustF, minSize=1, margin=0,
                 drawRect=False, parent=None, scene=None):
        """
        :param shapeF: Function providing the resized shape's current bounding
                       rect in scene coordinates and its path in its own
                       coordinates
        :type shapeF: None -> (QRectF, QPainterPath)
        :param brctCBF: Bounding rect change callback function, called by the
                        resizer whenever the bounding rect is modified through
                        one of its handles
        :type brctCBF: ((float, float), (float, float)) -> _
        :param adjustF: Point coordinates adjustment function, e.g. for
                        grid- snapping
        :type adjustF: callable ((float, float), (float, float)) ->
                         ((float, float), (float, float))
        :param minSize: Minimum size of bounding rectangle
        :param minSize: int
        :param margin: Margin between handles and resizing rect
        :type margin: float
        :param drawRect: Whether to draw a bounding rect,
                         e.g. for non-rectangular shapes
        :type drawRect: bool
        :param parent: Parent graphic item
        :param scene: Graphic scene the resizer belongs to
        """
        self.__minSize = minSize

        super().__init__(shapeF, brctCBF, adjustF, margin, drawRect,
                         parent, scene)

    def _setRect(self, p1, p2, dirs):
        x1, y1 = p1
        x2, y2 = p2

        if x2 - x1 < self.__minSize:
            if x1 == self._rect[0][0]:
                x2 = x1 + self.__minSize
            else:
                x1 = x2 - self.__minSize
        if y2 - y1 < self.__minSize:
            if y1 == self._rect[0][1]:
                y2 = y1 + self.__minSize
            else:
                y1 = y2 - self.__minSize

        super()._setRect((x1, y1), (x2, y2), dirs)


# noinspection PyTupleAssignmentBalance
class WguiSquareBoundingRectResizer(WguiBoundingRectResizerBase):
    """
    Resizer using 8 resizing handles around the resized shape's bounding box,
    mainting a squary bounding box.
    """

    def __init__(self, shapeF, brctCBF, adjustF, minSize=1, margin=0,
                 drawRect=False, parent=None, scene=None):
        """
        :param shapeF: Function providing the resized shape's current bounding
                       rect in scene coordinates and its path in its own
                       coordinates
        :type shapeF: None -> (QRectF, QPainterPath)
        :param brctCBF: Bounding rect change callback function, called by the
                        resizer whenever the bounding rect is modified through
                        one of its handles
        :type brctCBF: ((float, float), (float, float)) -> None
        :param adjustF: Point coordinates adjustment function, e.g. for
                        grid- snapping
        :type adjustF: callable ((float, float), (float, float)) ->
                         ((float, float), (float, float))
        :param minSize: Minimum size of bounding rectangle
        :param minSize: int
        :param margin: Margin between handles and resizing rect
        :type margin: float
        :param drawRect: Whether to draw a bounding rect,
                         e.g. for non-rectangular shapes
        :type drawRect: bool
        :param parent: Parent graphic item
        :param scene: Graphic scene the resizer belongs to
        """
        self.__minSize = minSize

        super().__init__(shapeF, brctCBF, adjustF, margin, drawRect,
                         parent, scene)

    def _setRect(self, p1, p2, dirs):
        x1, y1 = p1
        x2, y2 = p2

        if x2 - x1 < self.__minSize:
            if x1 == self._rect[0][0]:
                x2 = x1 + self.__minSize
            else:
                x1 = x2 - self.__minSize
        if y2 - y1 < self.__minSize:
            if y1 == self._rect[0][1]:
                y2 = y1 + self.__minSize
            else:
                y1 = y2 - self.__minSize

        w = abs(x2 - x1)
        h = abs(y2 - y1)
        if w != h:
            if dirs == self.X1 or dirs == self.X2:
                adj = (w - h) / 2
                y1 += adj if y1 > y2 else -adj
                y2 += adj if y2 > y1 else -adj
            elif dirs == self.Y1 or dirs == self.Y2:
                adj = (h - w) / 2
                x1 += adj if x1 > x2 else -adj
                x2 += adj if x2 > x1 else -adj
            else:
                if w > h:
                    if dirs & self.Y1:
                        y1 = y2 + w if y1 > y2 else y2 - w
                    elif dirs & self.Y2:
                        y2 = y1 + w if y2 > y1 else y1 - w
                else:
                    if dirs & self.X1:
                        x1 = x2 + h if x1 > x2 else x2 - h
                    elif dirs & self.X2:
                        x2 = x1 + h if x2 > x1 else x1 - h

        super()._setRect((x1, y1), (x2, y2), dirs)
