# ----------------------------------------------------------------------------
# - wgui-scene                                                               -
# - wild cruncher - cockpit : graphic scene widget                           -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_scene
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graphics scene editor

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import QRectF
from PyQt4.QtGui import QCursor, QGraphicsScene

from silkan_common.scmn_utils import last, partition
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP, \
    WhdbModelInstanceP
from wcr_core.db_api.whdb_protocols_api import WhdbBaseP
from wcr_frontend.gui.wgui_scene_adapter import WguiSceneAdapterP
from wcr_core.db.whdb_utils import whdb_getNextInstanceName, \
    whdb_getNextFunctionName
from wcr_frontend.gui.wgui_highlight_module import WguiItemHighlighter
from wcr_frontend.gui.wgui_button import WguiButton
from wcr_frontend.gui.wgui_protocols import ViewContext
from wcr_core.db.whdb_shape import *
from wcr_core.modules.hardware.db.whhdb_rack_instance import WhhdbRackInstance
from wcr_frontend.gui.wgui_scene_item_ids import *
from wcr_frontend.gui.wgui_scene_item import *
from silkan_common.scmn_log import SCMN_LOG_INFO
from wcr_core.modules.hardware.db.whhdb_iocard_instance import \
    WhhdbIoCardInstance
from wcr_core.modules.statemachine.db.wfhdb_instance import \
    WfhdbBlockInstance, WfhdbStateInstance

# ----------------------------------------------------------------------------
# - graphics scene                                                           -
# ----------------------------------------------------------------------------

class WguiGraphicsScene(QGraphicsScene, WguiSceneAdapterP, WcmnReleasableP,
                        metaclass=QObject_Meta_M):
    """
    The WguiGraphicsScene class is the class responsible to manage
    the collection of instance objects used by Wild Cruncher. Each base
    objects is stored in the Silkan hierarchical representation and
    mapped to graphical items which are added to the scene.
    """

    RUBBER_SHAPE_IDS = [LINE_ITEM_ID, RECT_ITEM_ID, AROW_ITEM_ID, ELPS_ITEM_ID]

    # the selection mode signal
    selectModeSignal = pyqtSignal()
    """:type: pyqtBoundSignal"""

    #: scene change signal
    sceneChanged = pyqtSignal(bool)
    """:type: pyqtBoundSignal[bool]"""

    #: scene path about to change signal
    scenePathChanging = pyqtSignal([object])
    """:type: pyqtBoundSignal[list[WhdbBlockInstanceP]]"""

    #: scene path change signal
    scenePathChanged = pyqtSignal([object])
    """:type: pyqtBoundSignal[list[WhdbBlockInstanceP]]"""

    #: scene item added signal
    sceneItemAdded = pyqtSignal(WhdbBaseP, list)
    """:type: pyqtBoundSignal[WhdbBaseP, list[WhdbBlockInstanceP]]"""

    #: scene item removed signal
    sceneItemRemoved = pyqtSignal(WhdbBaseP)
    """:type: pyqtBoundSignal[WhdbBaseP]"""

    #: scene item changed signal
    sceneItemChanged = pyqtSignal(WhdbBaseP)

    #: scene items changed signal
    sceneItemsChanged = pyqtSignal(list)

    #: scene item selected signal
    sceneItemSelected = pyqtSignal(list)

    #: project info change signal
    projectInfoChanged = pyqtSignal()

    """:type: pyqtBoundSignal[list[WhdbBlockInstanceP]]"""

    def __init__(self, pjdb, config, actions, pctx, cctx, actx,
                 parent=None):
        """
        :param pjdb: the project database
        :type pjdb: WprjProjectDatabase
        :param config: Editor config
        :type config: WguiProjectEditorCfgP
        :param actions: Actions dictionary
        :type actions: dict[str, WactActionP]
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param cctx: Cockpit context
        :type cctx: WguiCockpitSystemContextP
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param parent: the parent object
        """
        super(WguiGraphicsScene, self).__init__(parent)

        # the project database
        self.__pjdb = pjdb
        #: Project editor configuration
        self.__config = config
        #: Action dictionary
        self.__actions = actions
        #: Context
        self.__projectCtx = pctx
        #: Cockpit context
        self.__cctx = cctx
        #: Action execution context
        self.__actx = actx

        #: Application module
        self.__module = cctx.getAppModule()
        #: Scene item factory
        self.__sitmF = cctx.getSceneItemFactory()

        #: Scene items database
        self.__itemWRefs = {}
        """:type: dict[UUID, ReferenceType[WguiSceneItemP]]"""
        #: Whether to show the grid
        self.__gridFlag = False
        #: Whether to snap
        self.__snapFlag = True
        #: Pen for the grid
        self.__gridPen = QtGui.QPen()
        self.__gridPen.setColor(QtCore.Qt.lightGray)
        self.__gridPen.setStyle(QtCore.Qt.DashDotLine)
        #: Object insert mode, if any
        self.__insertMode = None
        #: Currently created line, if any
        self.__line = None
        #: The currently highlighted connector, if any
        self.__highlightedConnector = None
        self.__viewContext = WguiSceneViewContext(self, pjdb, config)
        self.__topLevelProxy = pjdb.getTopHierarchy().createProxy()
        self.__muted = False

        # Set scene rect to a(ny) initial value to assist widget layouting
        self.setSceneRect(-1000, -1000, 2000, 2000)

        config.addObserver(self._configChanged)
        self._configChanged()
        self.refresh()

    def release(self):
        self.clear()
        self.__config.removeObserver(self._configChanged())

    def _configChanged(self):
        self.update()

    @property
    def sitmF(self):
        return self.__sitmF

    @property
    def config(self):
        return self.__config

    @property
    def sceneItemChangedS(self):
        self.sceneItemsChanged.emit(self.getSceneItems())
        return self.sceneItemChanged

    @property
    def pjdb(self):
        """
        Get the project database represented by this scene.

        :return: Project database
        :rtype: WprjProjectDatabase
        """
        return self.__pjdb

    @property
    def topLevelModelBlock(self):
        """
        :return: The current top-level block
        :rtype: WhdbBlockP
        """
        return self.__topLevelProxy

    @property
    def blockInstancePath(self):
        """
        :return: The path to the current top-level block
        :rtype: Iterable[WhdbBlockInstanceP]
        """
        return self.__topLevelProxy.instanceStack()

    @property
    def viewContext(self):
        """
        :return: Scene's view context
        :rtype: ViewContext
        """
        return self.__viewContext

    def importModules(self):
        """
        Generic function to improt modules
        must be overrode by project
        :return:
        """
        pass


    def getSceneItems(self):
        """
        Get all the scene's items that represent database entities.

        :return: List of database scene items
        :rtype: list[WguiSceneItemP]
        """
        return [item for item in self.items()
                if isinstance(item, WguiSceneItemP)]

    def getBottomZValue(self):
        """
        Get the bottom (lowest) z-value of all items in the scene.

        :return: Bottom z-value
        :rtype: int
        """
        items = self.getSceneItems()
        return math.floor(items[-1].getZ()) if items else 0

    def getTopZValue(self):
        """
        Get the top (greatest) z-value of all items in the scene.

        :return: Top z-value
        :rtype: int
        """
        items = self.getSceneItems()
        return math.ceil(items[0].getZ()) if items else 0

    def getSceneItem(self, dbItem):
        """
        Get the scene item for the given database item, if it exists.

        :param dbItem: Database item to retrieve scene item for
        :type dbItem: WhdbBaseP
        :return: Scene item for the given database item
        :rtype: WguiSceneItemP
        """
        ref = self.__itemWRefs.get(dbItem.getUUID())
        if ref is not None:
            item = ref()
            if item is not None:
                return item

        for item in self.items():
            if isinstance(item, WguiSceneItemP):
                dataItem = item.getDataObject()

                if dataItem is dbItem:
                    self.__itemWRefs[dbItem.getUUID()] = weakref.ref(item)
                    return item

    def setMode(self, mode):
        """
        Set the operating mode by id. This functions is generally called
        when a toolbox button is pressed.

        :param mode: the operating mode
        """
        # set the scene mode
        self.__insertMode = mode

    def getMode(self):
        """
        Get the operating mode
        """
        return self.__insertMode

    def getMediaMode(self):
        """
        Get the media associated with the operating mode if any.
        """
        imod = self.__insertMode
        if self.__module.getBlockType(imod):
            return imod

        for mime in self.__module.getModelMimeTypes():
            if mime.mimeType == imod:
                return imod

        return None

    def getOperatorMode(self):
        """
        Get the operator associated with the operating mode if any.

        :return: Operator associated with the operating mode, or `None`
        :rtype: WhdbOperatorP
        """
        # noinspection PyTypeChecker
        return self.__module.getOperator(self.__insertMode)

    @property
    def insertMode(self):
        """
        Get the currently active insert mode, if any.

        :return: Current insert mode, if any
        :rtype: int
        """
        return self.__insertMode

    def isSelectionMode(self):
        """
        Check if the operating mode is a scene mode.
        """
        return self.__insertMode == SLCT_ITEM_ID

    def isLineMode(self):
        """
        Check if the operating mode is a line editing mode.
        """
        return self.__insertMode == CNCT_ITEM_ID

    def _getCurrentLine(self):
        return self.__line

    def isRubberBandCreationMode(self):
        """
        Check if the operating mode is a rubber mode.
        """
        # check for none
        if self.__insertMode is None: return False
        # check for valid media
        if self.getMediaMode() is not None: return True
        if self.__insertMode in self.RUBBER_SHAPE_IDS:
            return True
            # default is false
        return False

    def isFunctionMode(self):
        """
        Check if the operating mode is a function mode
        """
        # noinspection PyTypeChecker
        return self.__module.hasOperator(self.__insertMode)

    def getGridFlag(self):
        """
        Get the scene grid flag
        """
        return self.__gridFlag

    def setGridFlag(self, gflg):
        """
        Set the grid display flag

        :param gflg: the grid flag to set
        """
        self.__gridFlag = gflg
        self.update()

    def setSnapFlag(self, sflg):
        """
        Set the grid snap flag

        :param sflg: the snap grid flag to set
        """
        self.__snapFlag = sflg

    def getSnapFlag(self):
        """
        Get the scene grid snap flag
        """
        return self.__snapFlag

    def getSnapX(self):
        """
        Get the scene x snapping grid
        """
        return self.__config.snapGridX

    def getSnapY(self):
        """
        Get the scene y snapping grid
        """
        return self.__config.snapGridY

    def updateItems(self):
        for item in self.items():
            if isinstance(item, WguiSceneItemP):
                item.updateItem()

    def refresh(self):
        """
        Refresh the graphics scene from the database.
        """
        self.clear()
        self.__muted = True

        block = self.topLevelModelBlock
        for shape in block.getShapes():
            self.addShapeItem(shape)

        for inst in block.getBlockInstances():
            self.addInstanceItem(inst)

        for inst in block.getModelInstances():
            self.addInstanceItem(inst)

        for func in block.getOperatorInstances():
            self.addFunctionItem(func)

        for connection in block.getConnections():
            self.addConnectionItem(connection)

        for element in block.getButtons():
            self.addActionButton(element)

        for z, i in enumerate(reversed(self.getSceneItems()), 1):
            i.setZ(z)

        self.__muted = False
        self._resetSceneRect()
        self.sceneChanged.emit(True)

    def delSelection(self):
        """
        Delete the selected items.

        """
        selected = self.selectedItems()
        if not selected: return
        # loop in the selected items
        conns, items = partition(
            lambda x: isinstance(x, WguiConnectionSceneItem),
            (i for i in self.selectedItems() if isinstance(i, WguiSceneItemP)))
        for conn in conns:
            self.deleteItem(conn)
        sceneCards = self.getScenCards()
        for item in items:
            data = item.getDataObject()
            # get cards if item has any and delete them
            cards = getattr(data, 'cards', {})
            for card in cards.values():
                if card in sceneCards.keys():
                    self.deleteItem(sceneCards[card])
            self.deleteItem(item)
        # clear the selection
        self.clearSelection()

    def getScenCards(self):
        """
        Get all Io Cards in the scene in the form of a dictionary
        with operator instances as keys and scene item as values
        :return dict
        """
        return {s.getDataObject(): s for s in self.getSceneItems()
                    if isinstance(s.getDataObject(), WhhdbIoCardInstance)}

    def selectedItems(self):
        items = super().selectedItems()
        self.sceneItemSelected.emit(items)
        return items

    def delDbEntities(self, entities):
        """
        Delete the entities items.

        """
        for item in [self.getSceneItem(e) for e in entities]:
            self.deleteItem(item)

    def deleteItem(self, item):
        """
        Delete the given scene item.
        Will also:
         - delete the associated database item from the database,
         - delete any connections attached to the item,
         - update any other affected items.

        :param item: Scene item to be deleted
        :type item: WguiSceneItemP
        """
        if item is None:
            return
        # get the associated object
        data = item.getDataObject()
        ctrl = self.__projectCtx.getControllerFactory().getItemController(data)
        delItms, updItms = ctrl.deleteItem(data, self.topLevelModelBlock)
        if delItms:
            for _itm in delItms:
                sitm = self.getSceneItem(_itm)
                if sitm is not None:
                    self.removeItem(sitm)
        if updItms:
            for _itm in updItms:
                sitm = self.getSceneItem(_itm)
                if sitm is not None:
                    sitm.updateItem()

    def _sceneBlockChanging(self):
        """
        Notify views about a new model block to be loaded.
        """
        for view in self.views():
            if isinstance(view, WguiSceneBlockChangeAware):
                view.sceneBlockChanging()

    def _sceneBlockChanged(self):
        """
        Notify views that a new model block has been loaded loaded.
        """
        for view in self.views():
            if isinstance(view, WguiSceneBlockChangeAware):
                view.sceneBlockChanged()

    def pushHierarchy(self, item):
        """
        Push an instance scene item's instance onto the current-model-stack.

        :param item: the instance item to push
        :type item: WguiInstanceBaseSceneItem
        """
        if item is None:
            return

        item = item.topLevelItem()
        if not isinstance(item, WguiInstanceSceneItem):
            return

        self.scenePathChanging.emit(self.__topLevelProxy.instanceStack())
        self._sceneBlockChanging()
        self.__topLevelProxy.pushInstanceBlock(item.getDataObject())
        self.refresh()
        self._sceneBlockChanged()
        self.scenePathChanged.emit(self.__topLevelProxy.instanceStack())

    def popHierarchy(self):
        """
        Pop one level from the current-model-stack.
        """
        stack = self.__topLevelProxy.instanceStack()
        if not stack:
            return

        self.scenePathChanging.emit(stack)
        self._sceneBlockChanging()
        self.__topLevelProxy.popModelBlock()
        self.refresh()
        self._sceneBlockChanged()
        self.scenePathChanged.emit(self.__topLevelProxy.instanceStack())

    def popTopHierarchy(self):
        """
        Pop the top level hierarchy
        """
        stack = self.__topLevelProxy.instanceStack()
        if not stack:
            return

        self.scenePathChanging.emit(stack)
        self._sceneBlockChanging()
        self.__topLevelProxy.popTopModelBlock()
        self.refresh()
        self._sceneBlockChanged()
        self.scenePathChanged.emit(None)

    def enterBlockInstance(self, item):
        if not item:
            return

        sitm = self.getSceneItem(item)
        if not isinstance(sitm, WguiInstanceBaseSceneItem):
            return

        self.pushHierarchy(sitm)

    def setHierarchy(self, path):
        if not path:
            if self.__topLevelProxy.instanceStack():
                self.popTopHierarchy()
            return

        if path[-1].getBlock() is self.topLevelModelBlock:
            return

        assert all(isinstance(p, WhdbBlockInstanceP) for p in path)

        self.scenePathChanging.emit(self.__topLevelProxy.instanceStack())
        self._sceneBlockChanging()
        self.__topLevelProxy.popTopModelBlock()
        for inst in path[:-1]:
            self.__topLevelProxy.pushInstanceBlock(inst)

        self.__topLevelProxy.pushInstanceBlock(path[-1])
        self.refresh()
        self._sceneBlockChanged()
        self.scenePathChanged.emit(self.__topLevelProxy.instanceStack())

    def setHierarchyUUIDs(self, *uuids):
        """
        Set the current-model-stack to the given path.

        :param path: List of model instances with the top-level at the front
                     and the current model instance last.
        :type path: list[UUID]
        """
        # noinspection PyShadowingNames
        def findInstance(uuid, blck):
            for inst in blck.getBlockInstances():
                if inst.getUUID() == uuid:
                    return inst
            raise ValueError(
                'Could not find instance with UUID {}'.format(uuid))

        if not uuids:
            if self.__topLevelProxy.instanceStack():
                self.popTopHierarchy()
            return

        current = last(self.blockInstancePath)
        """:type: WhdbBlockInstanceP"""
        if current and uuids[-1] == current.getUUID():
            return

        self.scenePathChanging.emit(self.__topLevelProxy.instanceStack())
        self._sceneBlockChanging()
        self.__topLevelProxy.popTopModelBlock()
        blck = self.__pjdb.getTopHierarchy()
        for uuid in uuids[:-1]:
            inst = findInstance(uuid, blck)
            self.__topLevelProxy.pushInstanceBlock(inst)
            blck = inst.getBlock()

        self.__topLevelProxy.pushInstanceBlock(findInstance(uuids[-1], blck))
        self.refresh()
        self._sceneBlockChanged()
        self.scenePathChanged.emit(self.__topLevelProxy.instanceStack())

    def _resetSceneRect(self):
        if not self.__muted:
            if self.items():
                self.setSceneRect(self.itemsBoundingRect())
            else:
                self.setSceneRect(QRectF(-1, -1, 2, 2))

    def addInstanceItem(self, inst):
        """
        Create an instance graphics item for the given instance database
        object.

        :param inst: The instance object
        :type inst: WhdbInstanceP
        :return: The instance scene item created for the given model instance
        :rtype: WguiInstanceSceneItem
        """
        if inst is None: return None
        if inst.getZ() is None: inst.setZ(self.getTopZValue() + 1)
        if isinstance(inst, WhdbBlockInstanceP):
            item = self.__sitmF.createBlockInstanceSceneItem(
                inst, self.viewContext)
        elif isinstance(inst, WhdbModelInstanceP):
            item = self.__sitmF.createModelInstanceSceneItem(
                inst, self.viewContext)
        else:
            raise ValueError('Unknown instance type {}'.format(inst))
        self.addItem(item)
        return item

    def addFunctionItem(self, func):
        """
        Create a function graphics item for the given function database
        object.

        :param func: The function object
        :type func: WhdbOperatorInstanceP
        :return: The newly created connection scene item
        :rtype: WguiOperatorInstanceSceneItem
        """
        if func is None: return None
        if func.getZ() is None: func.setZ(self.getTopZValue() + 1)
        item = self.__sitmF.createOperatorInstanceSceneItem(
            func, self.viewContext)
        self.addItem(item)
        return item

    def addConnectionItem(self, cnctn):
        """
        Create a connection graphics item for the given connection
        database object.

        :param cnctn: the connection object
        :type cnctn: WhdbConnectionP
        :return: The newly created connection scene item
        :rtype: WguiConnectionSceneItem
        """
        if cnctn is None: return None
        if cnctn.getZ() is None: cnctn.setZ(self.getTopZValue() + 1)
        item = self.__sitmF.createConnectionSceneItem(cnctn, self.viewContext)
        self.addItem(item)
        for obj, conn in [(cnctn.getStartObject(),
                           cnctn.getStartConnectionPoint()),
                          (cnctn.getEndObject(),
                           cnctn.getEndConnectionPoint())]:
            sItem = self.getSceneItem(obj)
            if isinstance(sItem, WguiInstanceBaseSceneItem):
                sConn = sItem.getConnector(conn)
                if sConn is not None:
                    sConn.updateConnectionStatus()
        return item

    def addShapeItem(self, shape):
        """
        Create a shape graphics item for the given database shape item.

        :param shape: Shape to add to the scene
        :type shape: WhdbShapeP
        :return: Newly created scene item
        :rtype: WguiShape
        """
        if shape is None: return None
        if shape.getZ() is None: shape.setZ(self.getTopZValue() + 1)
        item = self.__sitmF.createShapeSceneItem(shape, self.viewContext)
        # noinspection PyTypeChecker
        self.addItem(item)
        if hasattr(item, 'updateBoundingBox'):
            item.updateBoundingBox()
        return item

    def mousePressEvent(self, event):
        """
        Process a scene mouse pressed button event.

        :param event: the event to process
        :type event: QGraphicsSceneMouseEvent
        """
        if self.isLineMode():
            # line mode
            self._startConnectionAt(event.scenePos())
            event.accept()
        elif self.isFunctionMode():
            # function mode
            self.createFunctionItem(event.scenePos())
            event.accept()
        elif self.__insertMode == XPNT_ITEM_ID:
            # connector mode
            self.addInstanceConnector(event.scenePos())
            event.accept()
        elif self.__insertMode == TEXT_ITEM_ID:
            # text shape mode
            self.createTextShape(event.scenePos(), 15)
            event.accept()
        elif self.__insertMode == ACTN_ITEM_ID:
            # action button mode
            self.createActionButton(event.scenePos())
            event.accept()
        elif event.modifiers() & Qt.AltModifier:
            if self._handleAltPressAt(event.scenePos()):
                event.accept()
        elif self.isSelectionMode():
            if event.button() != Qt.LeftButton:
                # Avoid deselection on right-clicks
                event.accept()
                return
            QtGui.QGraphicsScene.mousePressEvent(self, event)

    def _startConnectionAt(self, spos):
        """
        Start a new connection at the given scene position, if possible.

        :param spos: Scene position
        :type spos: QPointF
        :return: True iff a new connection was started
        :rtype: bool
        """
        inst, xpnt = self._getConnectionPointData(spos)
        ctrlFy = self.__projectCtx.getControllerFactory()
        ctrl = ctrlFy.getConnectionController()
        if ctrl.canStartConnectionAt(self.topLevelModelBlock, inst, xpnt):
            self._startLineAt(spos)
            return True
        return False

    def _handleAltPressAt(self, spos):
        """
        Hande a shift-click at the given scene position, if possible.

        :param spos: Scene position
        :type spos: QPointF
        :return: True iff the shift-click was handled.
        :rtype: bool
        """
        inst, conn = self._getConnectionPointData(spos)

        ctrlFy = self.__projectCtx.getControllerFactory()
        ctrl = ctrlFy.getConnectionController()
        if ctrl.canStartConnectionAt(self.topLevelModelBlock, inst, conn):
            self._startLineAt(spos)
            return True

        return self.addInstanceConnector(spos)

    def _startLineAt(self, spos):
        """
        Set the start of the current line to the given scene position.

        :param spos: Scene position
        :type spos: QPointF
        """
        self.setMode(CNCT_ITEM_ID)
        line = QtCore.QLineF(spos, spos)
        self.__line = QtGui.QGraphicsLineItem(line)
        self.__line.setZValue(self.getTopZValue() + 1)
        self.addItem(self.__line)

    def mouseMoveEvent(self, event):
        """
        Process a scene mouse move button event.

        :param event: the event to process
        """
        if self.isLineMode():
            self._highlightClosestConnector(event.scenePos())
            if self.__line is not None:
                # line mode
                line = QtCore.QLineF(self.__line.line().p1(), event.scenePos())
                self.__line.setLine(line)
        elif event.modifiers() & Qt.AltModifier:
            conn = getinstance(self.items(event.scenePos()),
                               WguiConnectionPointSceneItem)
            if conn is not None:
                self._highlightConnector(conn)
            else:
                self._unhighlightConnector()
        elif self.isSelectionMode():
            # check for selection mode
            QtGui.QGraphicsScene.mouseMoveEvent(self, event)

            bdwn = (event.buttons() & QtCore.Qt.LeftButton) == \
                   QtCore.Qt.LeftButton

            if not bdwn: return
            for s in self.selectedItems():
                if s.contains(event.pos()):
                    # reset selected instance connections
                    self.resetConnectionsForSelectedInstances()
                    break

    def mouseReleaseEvent(self, event):
        """
        Process a scene mouse release button event.

        :param event: the event to process
        """
        if self.isLineMode():
            # line mode
            if self.__line is not None:
                # check for connection mode
                self.connectItems()
                # clean item and line
                self.removeItem(self.__line)
                self.__line = None
                self.selectModeSignal.emit()
            self._unhighlightConnector()
            return
        elif self.isSelectionMode():
            QtGui.QGraphicsScene.mouseReleaseEvent(self, event)
            # update selected instance connections
            self.updateConnectionsForSelectedItems()
            self.selectModeSignal.emit()
            return
        else:
            # default mode deselect
            if event.button() == Qt.LeftButton:
                self.clearSelection()
            self.selectModeSignal.emit()

    def _getMousePos(self):
        if self.views:
            v = self.views()[0]
            # noinspection PyArgumentList
            return v.mapToScene(v.mapFromGlobal(QCursor.pos()))
        else:
            return None

    def keyPressEvent(self, event:QKeyEvent):
        pos = self._getMousePos()
        if pos is None: return

        conn = getinstance(self.items(pos),
                           WguiConnectionPointSceneItem)
        if conn is not None:
            self._highlightConnector(conn)

        super().keyPressEvent(event)

    def keyReleaseEvent(self, event:QKeyEvent):
        if event.key() == Qt.Key_Shift:
            self._unhighlightConnector()

        super().keyReleaseEvent(event)

    def drawBackground(self, painter, rect):
        """
        Draw a scene background. This is mostly used to draw the
        scene grid if any.

        :param painter: the painter object
        :param    rect: the scene view rectangle
        """
        # check if the grid display is active
        if self.__gridFlag:
            # compute x,y min/max grid
            xgrd = self.__config.gridSizeX
            ygrd = self.__config.gridSizeY
            lbox = rect.x()
            rbox = lbox + rect.width()
            tbox = rect.y()
            bbox = tbox + rect.height()
            xmin = wgui_snap_coord(lbox, xgrd)
            xmax = wgui_snap_coord(rbox, xgrd)
            ymin = wgui_snap_coord(tbox, ygrd)
            ymax = wgui_snap_coord(bbox, ygrd)
            # set painter
            painter.setPen(self.__gridPen)
            # draw the x lines
            # noinspection PyShadowingNames
            x = xmin
            while x <= xmax:
                painter.drawLine(x, tbox, x, bbox)
                x += xgrd
                # draw the y lines
            y = ymin
            while y <= ymax:
                painter.drawLine(lbox, y, rbox, y)
                y += ygrd

    def createInstanceItem(self, pos, mime=None, rect=None):
        """
        Add an instance object as a rectangle. The rectangle is generally
        obtained from the graphics view with the help of a rubber drawer.
        The instance media type is obtained from the operating mode.

        :param rect: the instance rectangle geometry
        :type rect: QRectF | None
        :param pos: the mouse position
        :type pos: QPointF
        """
        if mime is None:
            mime = self.getMediaMode()
        if mime is None: return
        name = whdb_getNextInstanceName(self.topLevelModelBlock)
        # create an instance by mime
        if self.__module.getBlockType(mime):
            inst = self.topLevelModelBlock.createBlockInstance(
                name=name, btyp=mime)
        else:
            inst = self.topLevelModelBlock.createModelInstance(
                mime=mime, name=name)
        # snap the rectangle if needed
        if rect is not None and rect.width() > 3 and rect.height() > 3:
            if self.__snapFlag:
                cfg = self.__config
                rect = wgui_snap_rect(rect, cfg.snapGridX, cfg.snapGridY)
            inst.setPosition(rect.center().x(), rect.center().y())
            inst.setDimension(rect.width(), rect.height())
        else:
            if self.__snapFlag:
                cfg = self.__config
                pos = wgui_snap_point(pos, cfg.snapGridX, cfg.snapGridY)
            inst.setPosition(pos.x(), pos.y())
        # add instance to the scene
        self.addInstanceItem(inst).focusNameField()

    def createFunctionItem(self, pos, oper=None):
        """
        Add a function item by position

        :param pos: the function position
        :type pos: QPointF
        """
        # select the operator by mode
        if oper is None:
            oper = self.getOperatorMode()
        if oper is None: return
        name = whdb_getNextFunctionName(oper, self.topLevelModelBlock)
        # get the function by operator
        func = self.topLevelModelBlock.createOperatorInstance(oper, name=name)
        # set the function position
        if self.__snapFlag:
            cfg = self.__config
            pos = wgui_snap_point(pos, cfg.snapGridX, cfg.snapGridY)
        func.setPosition(pos.x(), pos.y())
        # add function to the scene
        self.addFunctionItem(func).focusNameField()

    def addInstanceConnector(self, pos):
        """
        Add a new instance connector by position.

        :param pos: the mouse position in scene coordinate
        """
        # get the item at the position
        itms = self.items(pos)
        item = itms[0].topLevelItem() if itms else None

        if not isinstance(item, WguiInstanceBaseSceneItem):
            return False
        inst = item.getDataObject()

        ctrlFy = self.__projectCtx.getControllerFactory()
        ctrl = ctrlFy.getInstanceController()
        if ctrl.canCreateConnectionPoint(inst):
            ptpl = item.getConnectionPointPosition(pos)
            if ptpl:
                cpos, fcng = ptpl
                ctrl.createConnectionPoint(
                    inst, self.__cctx.getControllerValueProvider(),
                    cpos=cpos, fcng=fcng)
                item.updateItem()
                return True
        return False

    def createLineShape(self, pointA, pointB):
        """
        Create a new line shape and add it to the scene and the scene's
        project database.

        :param pointA: First point of the new line
        :type pointA: QPointF
        :param pointB: Second point of the new line
        :type pointB: QPointF
        """
        ctrlFy = self.__projectCtx.getControllerFactory()
        ctrl = ctrlFy.getShapeController()
        shape = ctrl.createShape(
            WhdbLineShape.NAME, self.topLevelModelBlock,
            **{WHDB_LINE_POINTA_KEY: (pointA.x(), pointA.y()),
               WHDB_LINE_POINTB_KEY: (pointB.x(), pointB.y())})
        self.addShapeItem(shape)

    def createArrowShape(self, tip, end):
        """
        Create a new arrow shape and add it to the scene and the scene's
        project database.

        :param tip: Tip point of the new line
        :type tip: QPointF
        :param end: End point of the new line
        :type end: QPointF
        """
        ctrlFy = self.__projectCtx.getControllerFactory()
        ctrl = ctrlFy.getShapeController()
        shape = ctrl.createShape(
            WhdbArrowShape.NAME, self.topLevelModelBlock,
            **{WHDB_LINE_POINTA_KEY: (tip.x(), tip.y()),
               WHDB_LINE_POINTB_KEY: (end.x(), end.y())})
        self.addShapeItem(shape)

    def createRectShape(self, rect):
        """
        Create a new rectangle shape and add it to the scene and the scene's
        project database.

        :param rect: Coordinates and dimension of the new rectangle
        :type rect: QRectF
        """
        ctrlFy = self.__projectCtx.getControllerFactory()
        ctrl = ctrlFy.getShapeController()
        shape = ctrl.createShape(
            WhdbRectangleShape.NAME, self.topLevelModelBlock,
            **{WHDB_AREA_WIDTH_KEY: rect.width(),
               WHDB_AREA_HEIGHT_KEY: rect.height(),
               WHDB_POSITION_KEY: (rect.x(), rect.y())})
        self.addShapeItem(shape)

    def createOvalShape(self, rect):
        """
        Create a new oval shape and add it to the scene and the scene's
        project database.

        :param rect: Bounding box of the new oval
        :type rect: QRectF
        """
        ctrlFy = self.__projectCtx.getControllerFactory()
        ctrl = ctrlFy.getShapeController()
        center = rect.center()
        shape = ctrl.createShape(
            WhdbOvalShape.NAME, self.topLevelModelBlock,
            **{WHDB_AREA_WIDTH_KEY: rect.width(),
               WHDB_AREA_HEIGHT_KEY: rect.height(),
               WHDB_POSITION_KEY: (center.x(), center.y())})
        self.addShapeItem(shape)

    def createTextShape(self, location, size=15, text=''):
        """
        Create a new text shape and add it to the scene and the scene's
        project database.

        :param location: Location of the new text shape (upper-left corner)
        :type location: QPointF
        :param size: Font size
        :type size: float
        :param text: Text content of the new text shape
        :type text: str
        """
        ctrlFy = self.__projectCtx.getControllerFactory()
        ctrl = ctrlFy.getShapeController()
        shape = ctrl.createShape(
            WhdbTextShape.NAME, self.topLevelModelBlock,
            **{WHDB_POSITION_KEY: (location.x(), location.y()),
               WHDB_TEXT_KEY: text,
               WHDB_FONT_SIZE_KEY: size})
        item = self.addShapeItem(shape)
        # noinspection PyUnresolvedReferences
        item.setFocus(Qt.OtherFocusReason)

    def createActionButton(self, location):
        """
        Create a new action button and add it to the scene and the scene's
        project database.

        :param location: Location of the new action button
        :type location: QPointF
        """
        button = self.topLevelModelBlock.createButton(
            bpos=(location.x(), location.y()))
        self.addActionButton(button)

    def addActionButton(self, button):
        """
        Add an action button to the scene.

        :param button: Action button to add
        :type button: WhdbButtonP
        :return: Scene action button item
        :rtype: WguiButton
        """
        if button is None: return None
        if button.getZ() is None: button.setZ(self.getTopZValue() + 1)
        item = WguiButton(button, self.__actions, self.__actx)
        self.addItem(item)
        item.updateBoundingBox()
        return item

    def _getConnectionPoint(self, pos):
        """
        Get the connection end-point for the given position, if any.

        :param pos: Position to look for item/connector at
        :type pos: QPointF
        :return: Item and connector at the given position, or `None`
        :rtype: (WguiInstanceBaseSceneItem | None,
                 WguiConnectionPointSceneItem | None)
        """
        itms = self.items(pos)
        item = getinstance(itms, WguiInstanceBaseSceneItem)
        if item is None:
            return None, None
        return item, getinstance(itms, WguiConnectionPointSceneItem)

    def _getConnectionPointData(self, pos):
        """
        Get the connection end-point for the given position, if any.

        :param pos: Position to look for item/connector at
        :type pos: QPointF
        :return: Item and connector at the given position, or `None`
        :rtype: (WhdbInstanceP | None, WhdbConnectionPointP | None)
        """
        # get the start item
        item, xpnt = self._getConnectionPoint(pos)
        return \
            item.getDataObject() if item else None,\
            xpnt.getDataObject() if xpnt else None

    def _highlightClosestConnector(self, pos):
        """
        Highlight the connector closest to the given position, if any.

        :param pos: Position to highlight connector at
        :type pos: QPointF
        """
        # check if we are in connection mode
        if self.__insertMode != CNCT_ITEM_ID:
            return
        _, xpnt = self._getConnectionPoint(pos)
        if xpnt is None:
            self._unhighlightConnector()
            return
        self._highlightConnector(xpnt)

    def _highlightConnector(self, conn):
        """
        Highlight the given connector.

        :param conn: Connector to highlight
        :type conn: WguiConnectionPointSceneItem
        """
        if conn is not self.__highlightedConnector:
            self._unhighlightConnector()
            conn.setHighlighted(True)
            self.__highlightedConnector = conn

    def _unhighlightConnector(self):
        """
        Remove any connector highlighting.
        """
        if self.__highlightedConnector is not None:
            self.__highlightedConnector.setHighlighted(False)
            self.__highlightedConnector = None

    def connectItems(self):
        """
        Connect two instances with a connection. The connection must be close
        to the instance connectors.
        """
        line = self.__line
        # check if we are in connection mode
        if self.__insertMode != CNCT_ITEM_ID or line is None:
            return

        sitm, scon = self._getConnectionPoint(line.line().p1())
        sobj = sitm.getDataObject() if sitm else None
        sxpn = scon.getDataObject() if scon else None
        eitm, econ = self._getConnectionPoint(line.line().p2())
        eobj = eitm.getDataObject() if eitm else None
        expn = econ.getDataObject() if econ else None

        try:
            tblk = self.topLevelModelBlock
            cfct = self.__projectCtx.getControllerFactory()
            ctrl = cfct.getConnectionController()
            canCreate = ctrl.canCreateConnection(
                tblk, sobj, eobj, sxpnt=sxpn, expnt=expn)
            if canCreate:
                isConnected = self.connectProjectItems(ctrl, tblk, eobj, sobj,
                                                       expn, sxpn)
                if not isConnected:
                    return
                conn, upds = ctrl.createConnection(
                    tblk, sobj, eobj, sxpnt=sxpn, expnt=expn)
                assert conn
                citm = self.addConnectionItem(conn)

                if not scon:
                    stpl = sitm.getConnectionPointPosition(line.line().p1())
                    if stpl:
                        spos, sfcn = stpl
                        conn.getStartConnectionPoint().setPosition(*spos)
                        conn.getStartConnectionPoint().setGravity(sfcn)
                if not econ:
                    etpl = eitm.getConnectionPointPosition(line.line().p2())
                    if etpl:
                        epos, efcn = etpl
                        conn.getEndConnectionPoint().setPosition(*epos)
                        conn.getEndConnectionPoint().setGravity(efcn)

                for item in upds:
                    sitm = self.getSceneItem(item)
                    if sitm:
                        sitm.updateItem()

                self.clearSelection()
                citm.setSelected(True)
        except ValueError as e:
            self.__projectCtx.addMessage(
                'Could not create connection between {so}.{sc} and {eo}.{ec}: '
                '{ex}'.format(
                    so=sobj.getName() if sobj else '?',
                    sc=sxpn.getName() if sxpn else '?',
                    eo=eobj.getName() if eobj else '?',
                    ec=expn.getName() if expn else '?',
                    ex=e),
                'cockpit', SCMN_LOG_INFO)

    def connectProjectItems(self, ctrl, tblk, eobj, sobj, expn, sxpn):
        """
            Function that should be called by the Child class
            to update information
            if return False no connection should be created
        :return: bool
        """
        return True

    def getConnectionsForItems(self, ilst):
        """
        :rtype: set[WguiConnectionSceneItem]
        """
        # get the selected items
        if not ilst: return None
        # create the connection list
        wlst = set()
        # get all items and loop
        itms = self.items()
        for item in itms:
            # check for a connection
            if not isinstance(item, WguiConnectionSceneItem): continue
            # check if the connection is connected to one of the instance
            for inst in ilst:
                if not isinstance(inst, WguiInstanceBaseSceneItem): continue
                # check for one instance
                if item.isConnected(inst):
                    wlst.add(item)
                    # success
        return wlst

    def getConnectionsForSelectedItems(self):
        """
        Returns all connections connected to at least one currently selected
        item.

        :return: All connections connected to at least one currently selected
                 item.
        :rtype: set[WguiConnectionSceneItem]
        """
        return self.getConnectionsForItems(self.selectedItems())

    def updateConnectionsForSelectedItems(self):
        """
        This function update the item connection by calling the connection
        update path method. This method is generally called when a connection
        move has been updated
        """
        # get the selected instance connection
        wlst = self.getConnectionsForSelectedItems()
        if wlst:
            # loop in the list
            slst = self.getSelectedInstances()
            for item in wlst:
                item.moveSelectedPathEndPoints(slst)
                item.finishPathUpdate()

    def resetConnectionsForSelectedInstances(self):
        """
        Reset all connection graphics items connected to at least on currently
        selected item.
        This function is generally called during an instance move operation.
        """
        # get the selected instance connection
        wlst = self.getConnectionsForSelectedItems()
        if wlst:
            slst = self.getSelectedInstances()
            # loop in the list
            for item in wlst:
                # update the connection path
                item.moveSelectedPathEndPoints(slst)

    def updateConnectionsForInstance(self, inst):
        """
        Reset all connection graphic items connected to the given
        instance/function.
        """
        # get the selected instance connection
        wlst = self.getConnectionsForItems([inst])
        if wlst:
            # loop in the list
            for item in wlst:
                # update the connection path
                item.moveSelectedPathEndPoints([inst.getDataObject()])

    def resetConnectionsForInstance(self, inst):
        """
        Reset all connection graphic items connected to the given
        instance/function.
        """
        # get the selected instance connection
        wlst = self.getConnectionsForItems([inst])
        if wlst:
            # loop in the list
            for item in wlst:
                # update the connection path
                item.resetPath()

    def resetConnectionsForXPoint(self, inst, conn, reset):
        """
        Reset all connection graphic items connected to the given
        instance/function.
        """
        # get the selected instance connection
        wlst = self.getConnectionsForItems([inst])
        if wlst:
            # loop in the list
            for item in wlst:
                # update the connection path
                item.moveConnector(conn, reset)

    def removeInvalidConnections(self):
        """
        Remove all connections that are no longer connected to valid connectors
        on both ends.
        """
        for cnctn in list(self.topLevelModelBlock.getConnections()):
            if not cnctn.isValid():
                self.deleteItem(self.getSceneItem(cnctn))

    def getSelectedInstances(self):
        return [i.getDataObject() for i in self.selectedItems()
                if (isinstance(i, WguiInstanceSceneItem) or
                    isinstance(i, WguiOperatorInstanceSceneItem))]

    def getSelectedEntities(self):
        return [i.getDataObject() for i in self.selectedItems()
                if (isinstance(i, WguiSceneItemP))]

    def selectEntities(self, entities):
        self.clearSelection()
        for e in entities:
            item = self.getSceneItem(e)
            """:type: QGraphicsItem | None"""
            if item is not None:
                item.setSelected(True)
                item.ensureVisible()

    def selectEntity(self, entity):
        self.clearSelection()
        item = self.getSceneItem(entity)
        """:type: QGraphicsItem | None"""
        if item is not None:
            item.setSelected(True)
            item.ensureVisible()

    def addItem(self, item:QGraphicsItem):
        super().addItem(item)
        if isinstance(item, WguiSceneItemP):
            self.__itemWRefs[item.getDataObject().getUUID()] = \
                weakref.ref(item)
        if isinstance(item, WguiZoomAwareP) and self.views():
            view = self.views()[0]
            item.zoomChanged(view.zoom)
        self.sceneItemsChanged.emit(self.getSceneItems())
        if not self.__muted and isinstance(item, WguiSceneItemP):
            self.sceneItemAdded.emit(item.getDataObject(),
                                     self.__topLevelProxy.instanceStack())

    def removeItem(self, item):
        if item is None or not item.scene() is self:
            return
        if isinstance(item, WguiSceneItemP):
            itemID = item.getDataObject().getUUID()
            if itemID in self.__itemWRefs:
                del self.__itemWRefs[itemID]
        super().removeItem(item)
        self.sceneItemsChanged.emit(self.getSceneItems())
        if not self.__muted and isinstance(item, WguiSceneItemP):
            self.sceneItemRemoved.emit(item.getDataObject())
        if isinstance(item, WcmnReleasableP):
            item.release()

    def clear(self):
        items = self.items()
        super().clear()
        self.__itemWRefs.clear()
        for item in items:
            if isinstance(item, WcmnReleasableP):
                item.release()
        if not self.__muted:
            self.sceneChanged.emit(False)

    def moveSelectedBy(self, dx, dy):
        """
        Move all currently selected items by the given amount (in pixels).

        :param dx: How much to move items in x-direction (+/-)
        :type dx: int
        :param dy: How much to move items in y-direction (+/-)
        :type dy: int
        """
        moved = False
        for item in self.selectedItems():
            if self.isItemMovable(item):
                pos = item.pos()
                item.setPos(pos.x() + dx, pos.y() + dy)
                moved = True
        if moved:
            self.resetConnectionsForSelectedInstances()

    @staticmethod
    def isItemMovable(item):
        return isinstance(item, WguiSceneItemP) and \
            not isinstance(item, WguiConnectionSceneItem)


class WguiSceneViewContext(ViewContext):
    def __init__(self, scene, projectdb, config):
        """
        :param scene: Scene
        :type scene: WguiGraphicsScene
        :param projectdb: Project Database
        :type projectdb: WprjProjectDatabase
        :param config: Editor config
        :type config: WguiProjectEditorCfgP
        """
        #: Graphics scene
        self.__scene = scene
        #: Project Database
        self.__projectDB = projectdb
        #: Editor config
        self.__config = config

        #: Database element rendering highlights
        self.__highlighter = WguiItemHighlighter()

    # noinspection PyShadowingNames
    def snapPoint(self, x, y):
        if self.__scene.getSnapFlag():
            xs = self.__scene.getSnapX()
            ys = self.__scene.getSnapY()

            return wgui_snap_coord(x, xs), wgui_snap_coord(y, ys)
        else:
            return x, y

    @property
    def topLevelModelBlock(self):
        return self.__scene.topLevelModelBlock

    def getConnections(self, inst, conn):
        top = self.topLevelModelBlock
        if top is None: return None
        return top.getConnectionsForObject(ix=inst, cx=conn)

    @property
    def highlighter(self):
        return self.__highlighter

    @property
    def config(self):
        return self.__config

    def getResource(self, rkey):
        return self.__projectDB.tops.getResource(rkey)


class WguiSceneBlockChangeAware(metaclass=ABCMeta):
    @abstractmethod
    def sceneBlockChanging(self):
        """
        Callback function for when the scene is about to load a new model
        block.
        """
        raise NotImplementedError()

    @abstractmethod
    def sceneBlockChanged(self):
        """
        Callback function for when the scene has loaded load a new model
        block.
        """
        raise NotImplementedError()
