# ----------------------------------------------------------------------------
# - wgui-scene-adapter                                                       -
# - wild cruncher: graphical scene adapter protocol                          -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_scene_adapter
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graphical scene adapter protocol

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod


class WguiSceneAdapterP(metaclass=ABCMeta):
    @property
    @abstractmethod
    def sceneItemChangedS(self):
        """
        Get the signal to indicate scene-item changes.

        :return: Signal to indicate scene-item changes
        :rtype: pyqtBoundSignal
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def topLevelModelBlock(self):
        """
        :return: The current top-level block
        :rtype: WhdbBlockP
        """
        raise NotImplementedError()

    @abstractmethod
    def setHierarchy(self, path):
        """
        Set the current-model-stack to the given path.

        :param path: List of model instances with the top-level at the front
                     and the current model instance last.
        :type path: list[WhdbBlockInstanceP]
        """
        raise NotImplementedError()

    @abstractmethod
    def enterBlockInstance(self, item):
        """
        Push an instance scene item's instance onto the current-model-stack.

        :param item: the instance item to push
        :type item: WhdbBlockInstanceP
        """
        raise NotImplementedError()

    @abstractmethod
    def getSceneItem(self, dbItem):
        """
        Get the scene item for the given database item, if it exists.

        :param dbItem: Database item to retrieve scene item for
        :type dbItem: WhdbBaseP
        :return: Scene item for the given database item
        :rtype: WguiSceneItemP
        """
        raise NotImplementedError()

    @abstractmethod
    def getSceneItems(self):
        """
        Get all the scene's items that represent database entities.

        :return: List of database scene items
        :rtype: Iterable[WguiSceneItemP]
        """
        raise NotImplementedError()

    @abstractmethod
    def deleteItem(self, item):
        """
        Delete the given scene item.
        Will also:
         - delete the associated database item from the database,
         - delete any connections attached to the item,
         - update any other affected items.

        :param item: Scene item to be deleted
        :type item: WguiSceneItemP
        """
        raise NotImplementedError()

    @abstractmethod
    def removeItem(self, item):
        raise NotImplementedError()

    @abstractmethod
    def selectedItems(self):
        """
        Get the currently selected items.

        :return: Selected graphics items
        :rtype: Iterable[QGraphicsItem]
        """
        raise NotImplementedError()

    @abstractmethod
    def selectedItems(self):
        """
        Get the currently selected items.

        :return: Selected graphics items
        :rtype: Iterable[QGraphicsItem]
        """
        raise NotImplementedError()

    @abstractmethod
    def selectEntities(self, entities):
        """
        Select the given entities.

        :param entities: Entities to select
        :type entities: Iterable[WhdbBaseP]
        """

    @abstractmethod
    def selectEntity(self, entity):
        """
        Select the given entity.

        :param entity: Entity to select
        :type entity: WhdbBaseP
        """

    @abstractmethod
    def getSelectedEntities(self):
        raise NotImplementedError()

    @abstractmethod
    def delSelection(self):
        """
        Delete the selected items.
        """
        raise NotImplementedError()

    @abstractmethod
    def addInstanceItem(self, inst):
        """
        Create an instance graphics item for the given instance database
        object.

        :param inst: The instance object
        :type inst: WhdbInstanceP
        :return: The instance scene item created for the given model instance
        :rtype: WguiInstanceSceneItem
        """
        raise NotImplementedError()

    @abstractmethod
    def addConnectionItem(self, cnctn):
        """
        Create a connection graphics item for the given connection
        database object.

        :param cnctn: the connection object
        :type cnctn: WhdbConnectionP
        :return: The newly created connection scene item
        :rtype: WguiConnectionSceneItem
        """
        raise NotImplementedError()

    @abstractmethod
    def delDbEntities(self, entities):
        """
        Delete the entities items.

        """
        raise NotImplementedError()

    @abstractmethod
    def getTopZValue(self):
        """
        Get the top (greatest) z-value of all items in the scene.

        :return: Top z-value
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getBottomZValue(self):
        """
        Get the bottom (lowest) z-value of all items in the scene.

        :return: Bottom z-value
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def updateConnectionsForInstance(self, inst):
        """
        Reset all connection graphic items connected to the given
        instance/function.
        """
        raise NotImplementedError()

    @abstractmethod
    def resetConnectionsForInstance(self, inst):
        """
        Reset all connection graphic items connected to the given
        instance/function.
        """
        raise NotImplementedError()

    @abstractmethod
    def removeInvalidConnections(self):
        """
        Remove all connections that are no longer connected to valid connectors
        on both ends.
        """
        raise NotImplementedError()

    @abstractmethod
    def update(self):
        raise NotImplementedError()
