# ----------------------------------------------------------------------------
# - wgui-scene                                                               -
# - wild cruncher - cockpit: scene-based context menu factory                -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_scene_context_menu_factory
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher scene-based context menu factory

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from functools import partial

from PyQt5.QtWidgets import QMessageBox, QAction, QMenu
from wcr_core.db_api.whdb_button_api import WhdbButtonP

from wcr_core.db_api.whdb_connection_api import WhdbConnectionP
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP, \
    WhdbOperatorInstanceP, WhdbBlockInstanceP, WhdbModelInstanceP
from wcr_core.db_api.whdb_shape_api import WhdbShapeP
from wcr_frontend.gui.wgui_protocols import WguiContextMenuFactoryP
from wcr_frontend.gui.wgui_scene_item import *
from wcr_core.modules.hardware.db.whhdb_iocard_instance import WhhdbIoCardInstance


class WguiSceneContextMenuFactory(WguiContextMenuFactoryP):
    """
    Context menu generate based on the WCR graphics scene.
    """

    def __init__(self, scene, actions, clpbrd, pctx, pdgf):
        """
        :param scene: Graphics scene
        :type scene: WguiSceneAdapterP
        :param actions: Action dictionary
        :type actions: dict[str, WactActionP]
        :param clpbrd: Clipboard handler
        :type clpbrd: WguiClipboardHandler
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param pdgf: Properties dialog factory
        :type pdgf: WguiPropertiesDialogFactoryP
        """
        #: Graphics scene
        self.__scene = scene
        #: Action dictionary
        self.__actions = actions
        #: Clipboard handler
        self.__clipboard = clpbrd
        #: Project context
        self.__prjContext = pctx
        #: Properties dialog factory
        self.__propDiagFy = pdgf
        #: View to use as position reference
        self.__centerF = None

    def setCenterF(self, centerF):
        """
        Set the function that provides the current view center.

        :param centerF: View-center function
        :type centerF: _ -> QPoint
        """
        self.__centerF = centerF

    def _propertiesDialogAction(self, dialogF, *args, **kwargs):
        """
        Handle an instance context properties dialog

        :param dialogF: Dialog creation function
        :type dialogF: *args, **kwargs->QDialog
        """

        def show():
            diaw = dialogF(*args, **kwargs)
            diaw.show()
            diaw.exec_()

        return show

    def _resetPathAction(self, cnctn):
        """
        Reset the given connection's path.

        :param cnctn: Connection to reset path on
        :type cnctn: WhdbConnectionP
        """
        item = self.__scene.getSceneItem(cnctn)
        if isinstance(item, WguiConnectionSceneItem):
            item.resetPath()

    def _resetConnectorPositionsAction(self, instance):
        """
        Reset the given function's connector positions.

        :param instance: Function to reset the connectors for
        :type instance: WhdbInstanceP
        """
        item = self.__scene.getSceneItem(instance)
        if item is not None:
            instance.resetConnectionPointPositions()
            item.updateItem()

    def _addArrangeEntries(self, menu, item):
        def moveBack():
            sItem = self.__scene.getSceneItem(item)
            if sItem is not None:
                sItem.setZ(self.__scene.getBottomZValue() - 1)
                self.__scene.update()

        def moveFront():
            sItem = self.__scene.getSceneItem(item)
            if sItem is not None:
                sItem.setZ(self.__scene.getTopZValue() + 1)
                self.__scene.update()

        def moveBackward():
            sItem = self.__scene.getSceneItem(item)
            if sItem is not None:
                diff = None
                prv = None
                for itm in self.__scene.getSceneItems():
                    d = sItem.getZ() - itm.getZ()
                    if d > 0 and (diff is None or d < diff):
                        diff = d
                        prv = itm
                if prv is not None:
                    z = prv.getZ()
                    # noinspection PyUnresolvedReferences
                    prv.setZValue(sItem.getZ())
                    sItem.setZ(z)
                    self.__scene.update()

        def moveForward():
            sItem = self.__scene.getSceneItem(item)
            if sItem is not None:
                diff = None
                nxt = None
                for itm in self.__scene.getSceneItems():
                    d = itm.getZ() - sItem.getZ()
                    if d > 0 and (diff is None or d < diff):
                        diff = d
                        nxt = itm
                if nxt is not None:
                    z = nxt.getZ()
                    # noinspection PyUnresolvedReferences
                    nxt.setZValue(sItem.getZ())
                    sItem.setZ(z)
                    self.__scene.update()

        # noinspection PyArgumentList
        menu.addAction(
            QAction('Move to back', menu, triggered=moveBack))
        # noinspection PyArgumentList
        menu.addAction(
            QAction('Move backward', menu, triggered=moveBackward))
        # noinspection PyArgumentList
        menu.addAction(
            QAction('Move forward', menu, triggered=moveForward))
        # noinspection PyArgumentList
        menu.addAction(
            QAction('Move to front', menu, triggered=moveFront))

    def _addCopyItemEntries(self, menu, item):
        def copy():
            self.__clipboard.copyItems(item)

        def cut():
            self.__clipboard.copyItems(item)
            #: TODO: Use scene.deleteItem?
            self.__scene.removeItem(self.__scene.getSceneItem(item))

        # noinspection PyArgumentList
        menu.addAction(QAction('Copy', menu, triggered=copy))
        # noinspection PyArgumentList
        menu.addAction(QAction('Cut', menu, triggered=cut))

    def _addDeleteAction(self, menu, item, parent):
        """
        Add action to delete the given item to the given menu.

        :param menu: Menu to add action to
        :type menu: QMenu
        :param item: Item to delete on execution
        :type item: WhdbNamedEntity
        :param parent: Parent widget
        """

        def confirmDelete():
            btn = QMessageBox.question(
                parent, 'Confirm Deletion',
                'Are you sure you want to delete {i}'.format(
                    i=item.getName() or 'this ' + item.getTypeName()),
                QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes)
            if btn == QMessageBox.Yes:
                self.__scene.deleteItem(self.__scene.getSceneItem(item))

        # noinspection PyArgumentList,PyUnresolvedReferences
        menu.addAction(QAction('Delete', menu, triggered=confirmDelete))

    def createViewContextMenu(self, parent, scenePos=None):
        def paste():
            self.__clipboard.paste(scenePos)

        menu = QMenu(parent)
        # noinspection PyNoneFunctionAssignment
        acts = self._createViewContextMenuActions(menu, parent)
        if acts:
            # noinspection PyTypeChecker
            for act in acts:
                # noinspection PyUnresolvedReferences
                menu.addAction(act)
        menu.addSeparator()
        # noinspection PyArgumentList
        menu.addAction(QAction('Paste', menu, triggered=paste,
                               enabled=bool(self.__clipboard.canPaste())))
        return menu

    def createItemContextMenu(self, entity, parent):
        menu = QMenu(parent)
        menu.addSeparator()
        if isinstance(entity, WhdbOperatorInstanceP):
            self._createOperatorInstanceContextMenu(entity, menu, parent)
        elif isinstance(entity, WhdbInstanceP):
            self._createInstanceContextMenu(entity, menu, parent)
        elif isinstance(entity, WhdbConnectionP):
            self._createConnectionContextMenu(entity, menu, parent)
        elif isinstance(entity, WhdbShapeP):
            self._createShapeContextMenu(entity, menu, parent)
        elif isinstance(entity, WhdbButtonP):
            self._createButtonContextMenu(entity, menu, parent)
        else:
            return menu

        menu.addSeparator()
        if self._canCopyCutPaste(entity):
            self._addCopyItemEntries(menu, entity)
        menu.addSeparator()
        self._addArrangeEntries(menu, entity)
        menu.addSeparator()
        # noinspection PyTypeChecker
        if not isinstance(entity, WhhdbIoCardInstance):
            self._addDeleteAction(menu, entity, parent)

        return menu

    def _canCopyCutPaste(self, entity):
        return True

    def createSelectionContextMenu(self, parent, scenePos=None):
        def cut():
            self.__clipboard.copy()
            self.__scene.delSelection()

        menu = QMenu(parent)
        menu.addSeparator()
        canCopy = bool(self.__scene.selectedItems())

        if scenePos is None and self.__centerF is not None:
            # noinspection PyCallingNonCallable
            scenePos = self.__centerF()

        # noinspection PyNoneFunctionAssignment
        acts = self._createSelectionContextMenuActions(
            self.__scene.selectedItems(), menu, scenePos, parent)
        if acts:
            # noinspection PyTypeChecker
            for act in acts:
                menu.addAction(act)
            menu.addSeparator()

        # noinspection PyArgumentList
        menu.addAction(QAction('Copy', menu, triggered=self.__clipboard.copy,
                               enabled=canCopy))
        # noinspection PyArgumentList
        menu.addAction(QAction('Cut', menu, triggered=cut, enabled=canCopy))

        return menu

    def _createSelectionContextMenuActions(self, slct, menu, spos, parent):
        return None

    def createConnectorContextMenu(self, connector, entity, parent):
        if isinstance(entity, WhdbOperatorInstanceP):
            dialogF = self.__propDiagFy.createOperatorInstancePropertiesDialog
        elif isinstance(entity, WhdbBlockInstanceP):
            dialogF = self.__propDiagFy.createBlockInstancePropertiesDialog
        elif isinstance(entity, WhdbModelInstanceP):
            dialogF = self.__propDiagFy.createModelInstancePropertiesDialog
        else:
            raise ValueError('Invalid instance type: {}'.format(entity))

        def remove():
            sitm = self.__scene.getSceneItem(entity)
            if sitm is not None:
                entity.removeConnectionPoint(connector)
                sitm.updateItem()
                self.__scene.removeInvalidConnections()

        menu = QMenu(parent)
        menu.addSeparator()

        # noinspection PyArgumentList
        menu.addAction(
            QAction(
                'Connector Properties', menu,
                triggered=self._propertiesDialogAction(
                    dialogF, entity, selectXpnt=connector, parent=parent)))

        canRemove = entity.canRemoveConnectionPoint(connector)
        # noinspection PyArgumentList
        menu.addAction(
            QAction(
                'Remove Connector', menu, triggered=remove, enabled=canRemove))

        for actn in self._createConnectionPointContextMenuActions(
                connector, entity, menu, parent):
            menu.addAction(actn)

        return menu

    def _createConnectionPointContextMenuActions(self, xpnt, entity,
                                                 menu, parent):
        return []

    def _createInstanceContextMenu(self, instance, menu, parent):
        """
        Create a context menu for a model instance.

        :param instance: Model instance to create menu for
        :type instance: WhdbInstanceP
        :param menu: Menu to add menu items to
        :type menu: QMenu
        :param parent: Parent widget
        """
        if isinstance(instance, WhdbBlockInstanceP):
            dialogF = self.__propDiagFy.createBlockInstancePropertiesDialog
            # noinspection PyArgumentList,PyUnresolvedReferences
            menu.addAction(
                QAction(
                    'Enter', menu,
                    statusTip='Enter the instance\'s block',
                    triggered=partial(
                        self.__scene.enterBlockInstance, instance)))
        elif isinstance(instance, WhdbModelInstanceP):
            dialogF = self.__propDiagFy.createModelInstancePropertiesDialog
        else:
            raise ValueError('Invalid instance type: {}'.format(instance))
        # noinspection PyArgumentList,PyUnresolvedReferences
        menu.addAction(
            QAction(
                '&Properties...', menu,
                statusTip='Manage instance properties',
                triggered=self._propertiesDialogAction(
                    dialogF, instance, parent=parent)))
        # noinspection PyArgumentList,PyUnresolvedReferences
        menu.addAction(
            QAction(
                'Reset Connector Positions', menu,
                statusTip='Reset all connectors to their default positions',
                triggered=partial(self._resetConnectorPositionsAction,
                                  instance)))
        # noinspection PyNoneFunctionAssignment
        acts = self._createInstanceContextMenuActions(instance, menu, parent)
        if acts:
            # noinspection PyTypeChecker
            for act in acts:
                # noinspection PyTypeChecker,PyUnresolvedReferences
                menu.addAction(act)

    # noinspection PyMethodMayBeStatic
    def _createInstanceContextMenuActions(self, inst, menu, parent):
        return None

    def _createOperatorInstanceContextMenu(self, inst, menu, parent):
        """
        Create a context menu for a function.

        :param inst: Function to create menu for
        :type inst: WhdbOperatorInstanceP
        :param menu: Menu to add menu items to
        :type menu: QMenu
        :param parent: Parent widget
        """
        # noinspection PyArgumentList,PyUnresolvedReferences
        menu.addAction(
            QAction(
                '&Properties...', menu,
                statusTip='Manage function properties',
                triggered=self._propertiesDialogAction(
                    self.__propDiagFy.createOperatorInstancePropertiesDialog,
                    inst, parent=parent)))
        # noinspection PyArgumentList,PyUnresolvedReferences
        menu.addAction(
            QAction(
                'Reset Connector Positions', menu,
                statusTip='Reset all connectors to their default positions',
                triggered=partial(self._resetConnectorPositionsAction,
                                  inst)))
        # noinspection PyNoneFunctionAssignment
        acts = self._createOperatorInstanceContextMenuActions(
            inst, menu, parent)
        if acts:
            # noinspection PyTypeChecker
            for act in acts:
                # noinspection PyUnresolvedReferences
                menu.addAction(act)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _createOperatorInstanceContextMenuActions(self, inst, menu, parent):
        """
        :param inst: Operator-instance to create additional menu actions for
        :type inst: WhdbOperatorInstanceP
        :param menu: Menu the action will be added to
        :type menu: QMenu
        :param parent: Parent widget
        :type parent: QWidget
        :return: Additional menu actions for the given instance
        :rtype: list[QAction] | None
        """
        return None

    def _createConnectionContextMenu(self, cnctn, menu, parent):
        """
        Create a context menu for a connection.

        :param cnctn: Connection to create menu for
        :type cnctn: WhdbConnectionP
        :param menu: Menu to add menu items to
        :type menu: QMenu
        :param parent: Parent widget
        """
        # noinspection PyArgumentList,PyUnresolvedReferences
        menu.addAction(
            QAction(
                '&Properties...', menu,
                statusTip='Manage connection properties',
                triggered=self._propertiesDialogAction(
                    self.__propDiagFy.createConnectionPropertiesDialog,
                    cnctn, parent=parent)))
        # noinspection PyUnresolvedReferences
        menu.addSeparator()
        # noinspection PyArgumentList,PyUnresolvedReferences
        menu.addAction(
            QAction('&Reset path', menu,
                    statusTip='Reset the path',
                    triggered=partial(self._resetPathAction, cnctn)))

    def _createShapeContextMenu(self, shape, menu, parent):
        """
        Create a context menu for a shape.

        :param shape: Shape to create menu for
        :type shape: WhdbShapeP
        :param menu: Menu to add menu items to
        :type menu: QMenu
        :param parent: Parent widget
        """
        # noinspection PyArgumentList,PyUnresolvedReferences
        menu.addAction(
            QAction(
                '&Properties...', menu,
                statusTip='Manage shape properties',
                triggered=self._propertiesDialogAction(
                    self.__propDiagFy.createShapePropertiesDialog, shape,
                    parent=parent)))

    def _createButtonContextMenu(self, button, menu, parent):
        """
        Create a context menu for an action button.

        :param button: Action button to create menu for
        :type button: WhdbButtonP
        :param menu: Menu to add menu items to
        :type menu: QMenu
        :param parent: Parent widget
        """
        # noinspection PyArgumentList,PyUnresolvedReferences
        menu.addAction(
            QAction(
                '&Properties...', menu,
                statusTip='Manage shape properties',
                triggered=self._propertiesDialogAction(
                    self.__propDiagFy.createButtonPropertiesDialog, button,
                    parent=parent)))

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _createViewContextMenuActions(self, menu, parent):
        """
        :param menu: Menu the action will be added to
        :type menu: QMenu
        :param parent: Parent widget
        :type parent: QWidget
        :return: Additional menu actions for the given instance
        :rtype: list[QAction] | None
        """
        return None
