# ----------------------------------------------------------------------------
# - wgui-scene-item                                                          -
# - wild cruncher - cockpit : diagram scene item widgets                     -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - author    (c) 2016-2016 philippe lambert                                   -
# - copyright (c) 2012-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_scene_item
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graphics scene diagram items

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

import math
import weakref

from PyQt4.QtCore import QRectF, QPointF, QLineF, Qt, QEvent, QByteArray
from PyQt4.QtGui import QPainter, QGraphicsItem, QGraphicsRectItem, \
    QPolygonF, QTransform, QColor, QPen, QMatrix, QFont, QFontMetrics,\
    QPainterPath, QGraphicsPathItem, QBrush, QKeyEvent, \
    QGraphicsSimpleTextItem, QPainterPathStroker, QGraphicsPixmapItem, QPixmap
from PyQt4.QtSvg import QGraphicsSvgItem, QSvgRenderer
from silkan_common.scmn_file_path import scmn_expshenv, scmn_isfile
from silkan_common.scmn_trigonometry import SCMN_QUARTER_CIRCLE_RAD, \
    scmn_degToRad, scmn_radToDeg
from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_core.db.whdb_connection_path_arc import whdb_getArcEndPoints
from wcr_core.db.whdb_connection_point import WHDB_XPNT_SHOW_TARGET
from wcr_core.db.whdb_instance import whdb_isFlipRotation
from wcr_core.db_api.whdb_connection_point_api import WhdbConnectionPointP
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP
from wcr_core.db_api.whdb_protocols_api import WHDB_VALIGN_TOP, \
    WHDB_VALIGN_BOTTOM, WHDB_HALIGN_CENTER
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_OUT
from wcr_core.modules.bus.db.wbhdb_external_module_instance import \
    WbhdbExtModuleInstance
from wcr_core.modules.bus.db.wbhdb_logibus_instance import WbhdbLogiBusInstance
from wcr_core.modules.hardware.db.whhdb_generic_Host_instance import \
    WhhdbGenericHostInstance
from wcr_core.modules.hardware.db.whhdb_operator_instance import \
    WhhdbOperatorInstance
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_MODULES_IO_NAMES, WHHDB_XPOINT_DISCRET_CODE, WHHDB_XPOINT_ARINC825_CODE,\
    WHHDB_XPOINT_FIBER_CHANNEL_CODE, WHHDB_XPOINT_1553_CODE
from wcr_core.modules.sys_map.db.wchdb_Host_lure_instance import \
    WchdbHostLureInstance
from wcr_core.modules.sys_map.db.wchdb_Switch_lure_instance import \
    WchdbSwitchLureInstance
from wcr_frontend.gui.wgui_protocols import WguiZoomAwareP, QGraphicsItem_Meta_M
from wcr_frontend.gui.wgui_resizer import WguiNoFlipBoundingRectResizer
from wcr_frontend.gui.wgui_toolbox_item import WGUI_ICON_SVG
from wcr_frontend.gui.wgui_utils import *
from wcr_frontend.gui.wgui_widgets import WguiEditableText

# ----------------------------------------------------------------------------
# - private section                                                          -
# ----------------------------------------------------------------------------

#: the connector pen color in Silkan green
_WGUI_CONN_BGRND = Qt.lightGray
_WGUI_CONN_FGRND = QColor(95, 129, 77)
_WGUI_CONN_FGRND_NC = QColor(255, 0, 0)
_WGUI_CONN_IO = QColor(0, 0, 0)

_WGUI_RAD_CONN = 7.0
_WGUI_BOX_CONN = 2.0 * _WGUI_RAD_CONN

WGUI_SHAPE_COLOUR = QColor(26, 26, 26)

WGUI_CONNECTION_COLOUR = QColor(78, 78, 78)
WGUI_CONNECTION_STYLE = Qt.SolidLine

WGUI_TRANS_FILL = QColor(255, 255, 255, 196)

_WGUI_SVG_RENDERERS = {}

IO_TRIGRAMES = {WHHDB_XPOINT_DISCRET_CODE: 'DIS',
                WHHDB_XPOINT_FIBER_CHANNEL_CODE: 'FBC',
                WHHDB_XPOINT_ARINC825_CODE: 'CAN',
                WHHDB_XPOINT_1553_CODE: 'MLB'}

# ----------------------------------------------------------------------------
# - scene items section                                                      -
# ----------------------------------------------------------------------------

class WguiSceneItemP(metaclass=ABCMeta):
    """
    Protocol for scene items that represent database objects.
    """

    @abstractmethod
    def getDataObject(self):
        """
        Get the database object represented by this scene item.

        :return: Database object represented by this scene item
        :rtype: WhdbBaseP
        """
        pass

    @abstractmethod
    def updateItem(self):
        """
        Update the scene item to the current version of the represented
        database object.
        """
        pass

    @abstractmethod
    def getZ(self):
        """
        Get the item's z-value.

        :return: The item's z-value
        :rtype: float
        """
        raise NotImplementedError()

    @abstractmethod
    def setZ(self, zval):
        """
        Set the item's z-value.

        :param zval: The new z-value
        :type zval: float
        """
        raise NotImplementedError()

    @abstractmethod
    def pos(self):
        """
        Get the item's position.

        :return: The item's position
        :rtype: QPointF
        """
        raise NotImplementedError()

    @abstractmethod
    def setPos(self, x, y):
        """
        Set the item's position.

        :param x: New x-coordinate
        :type x: float
        :param y: New y-coordinate
        :type y: float
        """
        raise NotImplementedError()


class WguiConnectionPointSceneItem(QGraphicsItem,
                                   metaclass=QGraphicsItem_Meta_M):
    """
    The WguiConnectorSceneItem is the class that materializes an object
    connector. It can be used with an instance or a function.
    """

    def __init__(self, conn, item, posp, vctx, parent=None, scene=None):
        """
        Create a graphical connector by database connector object

        :param conn: the database connector
        :type conn: WhdbConnectionPointP
        :param item: database item the connector belongs to
        :type item: WhdbInstanceP
        :param posp: Connection point position provider
        :type posp: WguiXPointPositionProviderP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the object scene
        """
        # base initialization
        super().__init__(parent, scene)
        #: The displayed connector
        self.__conn = conn
        #: The item the connector belongs to
        self.__item = item
        #: View context
        self.__viewContext = vctx
        #: WguiXPointPositionProviderP
        self.__positionProvider = posp
        #: Whether the connector should be displayed as highlighted
        self.__highlighted = False
        #: Whether the connector is currently attached to any connections
        self.__connected = False
        #: defining colors for item
        self.__bgColor = _WGUI_CONN_BGRND

        #: Arrow indicating input/output
        self.__rect, self.__icon = self._createBackgroundAndIcon(conn, item)

        self._updateItem()
        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges)
        self.setAcceptedMouseButtons(Qt.LeftButton)

    def setbackgroundcolor(self, color):
        self.__bgColor = color
        self._updateItem()

    def boundingRect(self):
        rect = self.__rect
        return rect.boundingRect() if rect else QRectF(0, 0, 0, 0)

    def paint(self, painter, option, widget=None):
        pass

    def _createBackgroundAndIcon(self, conn, inst):
        rect = QGraphicsRectItem(self)
        rect.setPen(QPen(self.__bgColor))
        rect.setBrush(QBrush(self.__bgColor))
        rect.setRect(-_WGUI_RAD_CONN, -_WGUI_RAD_CONN,
                     _WGUI_BOX_CONN, _WGUI_BOX_CONN)
        HwIconSpeed = (isinstance(inst, WhhdbGenericHostInstance) or
                       isinstance(inst, WhhdbOperatorInstance) or
                       isinstance(inst, WchdbHostLureInstance) or
                       isinstance(inst, WchdbSwitchLureInstance))

        if HwIconSpeed:
            if conn.getCode() in WHHDB_MODULES_IO_NAMES:
                tri = IO_TRIGRAMES[conn.getCode()]
                textItem = self._IoIcon(tri, rect)
                itemFont = QFont()
                itemFont.setPointSize(5)
                textItem.setBrush(QColor(0, 0, 0))
                textItem.setFont(itemFont)
                textItem.setPos(-6, -3)
                icon = textItem
            else:
                icon = self._IconSpeed(speed=conn.getSpeed(),
                                       direction=conn.getDirection(),
                                       parent=self)
        elif (isinstance(inst, WbhdbExtModuleInstance) or
              isinstance(inst, WbhdbLogiBusInstance)):
            icon = self._BiIcon(self)
        else:
            icon = self._UniIcon(self)
            icon.updateDirection(conn)

        return rect, icon

    def updatePos(self):
        """
        Update the position of the connector graphics item based on the
        current database values.
        """
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, False)
        w, h = self.__item.getDimension()
        x, y = self.__conn.getPosition()
        self.setPos(x * w / 2, y * h / 2)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)

    def updateConnectionStatus(self):
        self.__connected = bool(
            next(self.__viewContext.getConnections(self.__item, self.__conn),
                 None))
        self._updateColours()

    def _updateColours(self):
        if self.__connected:
            fg = _WGUI_CONN_FGRND
        else:
            fg = _WGUI_CONN_FGRND_NC

        rect = self.__rect
        icon = self.__icon

        if self.__highlighted:
            if rect:
                rect.setPen(QPen(fg))
                rect.setBrush(QBrush(fg))
            if icon:
                icon.setPen(QPen(self.__bgColor))
                icon.setBrush(QBrush(self.__bgColor))
        else:
            if rect:
                rect.setPen(QPen(self.__bgColor))
                rect.setBrush(QBrush(self.__bgColor))
            if icon:
                text = getattr(icon, 'text', None)
                if text:
                    fg = _WGUI_CONN_IO
                icon.setPen(QPen(fg))
                icon.setBrush(QBrush(fg))
        self.update()

    def getDataObject(self) -> WhdbConnectionPointP:
        return self.__conn

    def _updateItem(self):
        """
        Update the item's position and status to reflect the current database
        state.
        """
        self.setToolTip(
            self.__item.getConnectionPointDescription(
                self.__conn, **{WHDB_XPNT_SHOW_TARGET: True}))
        self.updatePos()
        self.updateConnectionStatus()
        if self.__icon:
            self.__icon.updateDirection(self.__conn)
        self._appSpecificItemUpdate(self.__conn, self.__item)

    def _appSpecificItemUpdate(self, xpnt, item):
        pass

    def distance(self, pos):
        """
        Compute the distance to a point in scene coordinate.
        :param pos: the distance to this position
        :type pos: QPointF
        :return: Distance between the given position and the connector
        :rtype: float
        """
        # check for valid point
        if pos is None: return None
        # get the connector point
        p = self.pos()
        # create a vector line
        l = QLineF(p, pos)
        # get the vector length
        return l.length()

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            (x, y), g = self.__positionProvider.getXPointPosition(
                (value.x(), value.y()), self.__item, self.parentItem(),
                self.__conn.getGravity())
            self.__conn.setPosition(x, y)
            gvtyChange = g != self.__conn.getGravity()

            if gvtyChange:
                self.__conn.setGravity(g)
                if self.__icon:
                    self.__icon.updateDirection(self.__conn)
                self._appSpecificItemChange(self.__conn, self.__item)
            # Todo: Callback or notification from DB connector item
            self.scene().resetConnectionsForXPoint(
                self.parentItem(), self.__conn, gvtyChange)
            w, h = self.__item.getDimension()
            return QPointF(x * w / 2, y * h / 2)
        return value

    def _appSpecificItemChange(self, xpnt, item):
        pass

    def mousePressEvent(self, event):
        self.scene().clearSelection()

    def setHighlighted(self, highlighted):
        """
        Set the display of the connector to highlighted mode.

        :param highlighted: Whether to highlight
        :type highlighted: bool
        """
        self.__highlighted = highlighted
        self._updateColours()

    class _IoIcon(QGraphicsSimpleTextItem):
        """
        Arrow head for the connector display.
        """

        def __init__(self, tri, rect, parent=None):
            super().__init__(tri, rect, parent)

        def updateDirection(self, conn):
            """
            Update the arrow head's direction based on the connector gravity.

            :param conn: Connector
            :type conn: WhdbConnectionPointP
            """
            pass

    class _UniIcon(QGraphicsPathItem):
        """
        Arrow head for the connector display.
        """

        def __init__(self, parent=None):
            super().__init__(parent)
            path = QPainterPath()
            path.moveTo(-4, -3)
            path.lineTo(4, 0)
            path.lineTo(-4, 3)
            path.lineTo(-4, -3)
            self.setPath(path)
            self.setPen(QPen(_WGUI_CONN_FGRND))
            self.setBrush(QBrush(_WGUI_CONN_FGRND))

        def updateDirection(self, conn):
            """
            Update the arrow head's direction based on the connector gravity.

            :param conn: Connector
            :type conn: WhdbConnectionPointP
            """
            m = QMatrix()
            if conn.getDirection() == WDBS_XPNT_OUT:
                m = m.rotate(180)
            if conn.getGravity() == WhdbConnectionPointP.North:
                m = m.rotate(90)
            elif conn.getGravity() == WhdbConnectionPointP.East:
                m = m.rotate(180)
            elif conn.getGravity() == WhdbConnectionPointP.South:
                m = m.rotate(270)
            self.setMatrix(m)

        def paint(self, painter:QPainter, options, widget=None):
            painter.save()
            painter.setRenderHint(QPainter.Antialiasing, True)
            QGraphicsPathItem.paint(self, painter, options, widget)
            painter.restore()

    class _IconSpeed(QGraphicsPathItem):
        """
        Icon for the connector speed display.
        """

        def __init__(self, speed, direction=None, parent=None):
            super().__init__(parent)
            path = QPainterPath()

            from wcr_frontend.modules.hardware.gui.whgui_scene_item_factory import \
                BANDWIDTH_XPOINT_SPEED_1, BANDWIDTH_XPOINT_SPEED_0_1, \
                BANDWIDTH_XPOINT_SPEED_10, BANDWIDTH_XPOINT_SPEED_0_0_1

            if speed == BANDWIDTH_XPOINT_SPEED_0_1:
                path.addEllipse(-2, -2, 4, 4)
            elif speed == BANDWIDTH_XPOINT_SPEED_1:
                path.moveTo(0, 4)
                path.lineTo(4, 0)
                path.lineTo(0, -4)
                path.lineTo(-4, 0)
                path.lineTo(0, 4)
            elif speed == BANDWIDTH_XPOINT_SPEED_10:
                if direction in (3, 4):
                    path.addRect(-7, -3, 13.9, 6)
                elif direction in (1, 2):
                    path.addRect(-3, -7, 6, 13.9)
            elif speed == BANDWIDTH_XPOINT_SPEED_0_0_1:
                path.addEllipse(-1, -1, 2, 2)
            else:
                # default
                path.addEllipse(-1, -1, 1, 1)
            self.setPath(path)
            self.setPen(QPen(_WGUI_CONN_FGRND))
            self.setBrush(QBrush(_WGUI_CONN_FGRND))

        def updateDirection(self, conn):
            """
            Update the arrow head's direction based on the connector gravity.

            :param conn: Connector
            :type conn: WhdbConnectionPointP
            """
            m = QMatrix()
            if conn.getDirection() == WDBS_XPNT_OUT:
                m = m.rotate(180)
            if conn.getGravity() == WhdbConnectionPointP.North:
                m = m.rotate(90)
            elif conn.getGravity() == WhdbConnectionPointP.East:
                m = m.rotate(180)
            elif conn.getGravity() == WhdbConnectionPointP.South:
                m = m.rotate(270)
            self.setMatrix(m)

        def paint(self, painter:QPainter, options, widget=None):
            painter.save()
            painter.setRenderHint(QPainter.Antialiasing, True)
            QGraphicsPathItem.paint(self, painter, options, widget)
            painter.restore()

    class _BiIcon(QGraphicsPathItem):
        """
        Arrow head for the connector display.
        """

        def __init__(self, parent=None):
            super().__init__(parent)
            path = QPainterPath()
            path.moveTo(0, 4)
            path.lineTo(4, 0)
            path.lineTo(0, -4)
            path.lineTo(-4, 0)
            path.lineTo(0, 4)
            self.setPath(path)
            self.setPen(QPen(_WGUI_CONN_FGRND))
            self.setBrush(QBrush(_WGUI_CONN_FGRND))

        def updateDirection(self, conn):
            pass

        def paint(self, painter:QPainter, options, widget=None):
            painter.save()
            painter.setRenderHint(QPainter.Antialiasing, True)
            QGraphicsPathItem.paint(self, painter, options, widget)
            painter.restore()

    def getDirection(self):
        """
        Get the connector type (input, output, bi-dir).

        :return: The connector's type
        :rtype: int
        """
        return self.__conn.getDirection()

class WguiCardSceneItem(QGraphicsItem, metaclass=QGraphicsItem_Meta_M):
    """
    The WguiCardSceneItem is the class that materializes an object
    card. It can be used with an instance or a function.
    """

    def __init__(self, card, item, capp, vctx, parent=None, scene=None):
        """
        Create a graphical connector by database connector object

        :param card: the database card name
        :type card: str
        :param item: database item the connector belongs to
        :type item: WhdbInstanceP
        :param capp: card point position provider
        :type capp: WguiXPointPositionProviderP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the object scene
        """
        # base initialization
        super().__init__(parent, scene)
        #: The displayed card
        self.__card = card
        #: The item the card belongs to
        self.__item = item
        #: View context
        self.__viewContext = vctx
        #: WguiXPointPositionProviderP
        self.__positionProvider = capp
        #: Whether the card should be displayed as highlighted
        self.__highlighted = False
        #: Whether the card is currently attached to any connections
        self.__connected = False
        #:defining colors for item
        self.__bgColor = _WGUI_CONN_BGRND


        #: Arrow indicating input/output
        self.__rect, self.__icon = self._createBackgroundAndIcon(card, item)

        self._updateItem()
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges)
        self.setAcceptedMouseButtons(Qt.LeftButton)

    def setbackgroundcolor(self, color):
        self.__bgColor = color
        self._updateItem()

    def boundingRect(self):
        rect = self.__rect
        return rect.boundingRect() if rect else QRectF(0, 0, 0, 0)

    def paint(self, painter, option, widget=None):
        pass

    def _createBackgroundAndIcon(self, conn, inst):

        rect = QGraphicsRectItem(self)
        rect.setPen(QPen(self.__bgColor))
        rect.setBrush(QBrush(self.__bgColor))
        w, h = inst.getDimension()
        x, y = inst.getPosition()
        rect.setRect(x, y, w, h*2)
        icon = self._UniIcon(self)
        return rect, icon

    def updatePos(self):
        """
        Update the position of the connector graphics item based on the
        current database values.
        """
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, False)
        w, h = self.__item.getDimension()
        x, y = self.__item.getPosition()
        self.setPos(x * w / 2, y * h / 2)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)

    def updateConnectionStatus(self):
        self._updateColours()

    def _updateColours(self):
        if self.__connected:
            fg = _WGUI_CONN_FGRND
        else:
            fg = _WGUI_CONN_FGRND_NC

        rect = self.__rect
        icon = self.__icon

        if self.__highlighted:
            if rect:
                rect.setPen(QPen(fg))
                rect.setBrush(QBrush(fg))
            if icon:
                icon.setPen(QPen(self.__bgColor))
                icon.setBrush(QBrush(self.__bgColor))
        else:
            if rect:
                rect.setPen(QPen(self.__bgColor))
                rect.setBrush(QBrush(self.__bgColor))
            if icon:
                icon.setPen(QPen(fg))
                icon.setBrush(QBrush(fg))
        self.update()

    def getDataObject(self):
        return self.__card

    def _updateItem(self):
        """
        Update the item's position and status to reflect the current database
        state.
        """
        self.setToolTip(
            'Cards:' + '\n'.join(self.__item.getCards()))
        self.updatePos()
        self.updateConnectionStatus()

    def _appSpecificItemUpdate(self, xpnt, item):
        pass

    def distance(self, pos):
        """
        Compute the distance to a point in scene coordinate.
        :param pos: the distance to this position
        :type pos: QPointF
        :return: Distance between the given position and the connector
        :rtype: float
        """
        # check for valid point
        if pos is None: return None
        # get the connector point
        p = self.pos()
        # create a vector line
        l = QLineF(p, pos)
        # get the vector length
        return l.length()

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            (x, y), g = self.__positionProvider.getXPointPosition(
                (value.x(), value.y()), self.__item, self.parentItem(),
                self.__conn.getGravity())
            self.__conn.setPosition(x, y)
            gvtyChange = g != self.__conn.getGravity()

            if gvtyChange:
                self.__conn.setGravity(g)
                if self.__icon:
                    self.__icon.updateDirection(self.__conn)
                self._appSpecificItemChange(self.__conn, self.__item)
            # Todo: Callback or notification from DB connector item
            self.scene().resetConnectionsForXPoint(
                self.parentItem(), self.__conn, gvtyChange)
            w, h = self.__item.getDimension()
            return QPointF(x * w / 2, y * h / 2)
        return value

    def _appSpecificItemChange(self, xpnt, item):
        pass

    def mousePressEvent(self, event):
        self.scene().clearSelection()

    def setHighlighted(self, highlighted):
        """
        Set the display of the connector to highlighted mode.

        :param highlighted: Whether to highlight
        :type highlighted: bool
        """
        self.__highlighted = highlighted
        self._updateColours()

    class _UniIcon(QGraphicsPathItem):
        """
        Arrow head for the connector display.
        """

        def __init__(self, parent=None):
            super().__init__(parent)
            path = QPainterPath()
            path.moveTo(-4, -3)
            path.lineTo(4, 0)
            path.lineTo(-4, 3)
            path.lineTo(-4, -3)
            self.setPath(path)
            self.setPen(QPen(_WGUI_CONN_FGRND))
            self.setBrush(QBrush(_WGUI_CONN_FGRND))

        def paint(self, painter:QPainter, options, widget=None):
            painter.save()
            painter.setRenderHint(QPainter.Antialiasing, True)
            QGraphicsPathItem.paint(self, painter, options, widget)
            painter.restore()

    def getDirection(self):
        """
        Get the connector type (input, output, bi-dir).

        :return: The connector's type
        :rtype: int
        """
        return self.__conn.getDirection()


def wgui_points2qpoly(points):
    """
    Convert a list of points into a corresponding QPolygonF.

    :param points: The path to convert.
    :type points: list[(float, float)]
    :return: QPolygonF corresponding to the given path item
    :rtype: QPolygonF
    """
    # create a polygon
    poly = QPolygonF()
    # loop in the path list
    for x, y in points:
        poly.append(QPointF(x, y))
    if points[0][0] != points[-1][0] or points[0][1] != points[-1][1]:
        for x, y in reversed(points[1:-1]):
            poly.append(QPointF(x, y))
    return poly


class WguiIconBase(metaclass=ABCMeta):
    def __init__(self):
        #: Current icon scaling
        self.__scaling = 1.0
        #: Base-matrix set explicitely from outside
        self.__baseMatrix = QMatrix()
        #: Bounding box
        self.__boundingBox = self.boundingRect()

        self._updateMatrix()

    @abstractmethod
    def boundingRect(self):
        raise NotImplementedError()

    @abstractmethod
    def setMatrixSuper(self, matrix):
        raise NotImplementedError()

    def scaleToBox(self, bbox):
        """
        Scale the icon to fit into the given box.

        :param bbox: Box constraining the new icon size
        :type bbox: QRectF
        """
        ibbx = self.__boundingBox
        adj = _WGUI_RAD_CONN
        bbox = bbox.adjusted(adj, adj, -adj, -adj)
        sclx = bbox.width() / ibbx.width()
        scly = bbox.height() / ibbx.height()
        self.__scaling = min(sclx, scly)
        self._updateMatrix()

    def setMatrix(self, matrix):
        self.__baseMatrix = matrix
        self._updateMatrix()

    def _updateMatrix(self):
        sclf = self.__scaling
        brct = self.boundingRect()
        matrix = QMatrix(self.__baseMatrix)
        matrix.scale(sclf, sclf)
        matrix.translate(-brct.width() / 2.0, -brct.height() / 2.0)
        self.setMatrixSuper(matrix)


# noinspection PyAbstractClass
class WguiImageIconBase(WguiIconBase):
    def __init__(self):
        super().__init__()


class WguiPixmapIcon(QGraphicsPixmapItem, WguiImageIconBase,
                     metaclass=QGraphicsItem_Meta_M):
    """
    Pixmap-based icon.
    """

    def __init__(self, pixmap, parent=None):
        """
        :param pixmap: Pixmap
        :type pixmap: QPixmap
        :param parent: Optional parent item
        """
        QGraphicsPixmapItem.__init__(self, pixmap, parent)
        WguiIconBase.__init__(self)

    def boundingRect(self):
        return QGraphicsPixmapItem.boundingRect(self)

    def setMatrix(self, matrix, combine=False):
        WguiIconBase.setMatrix(self, matrix)

    def setMatrixSuper(self, matrix):
        QGraphicsPixmapItem.setMatrix(self, matrix)


class WguiSvgIcon(QGraphicsSvgItem, WguiImageIconBase,
                  metaclass=QGraphicsItem_Meta_M):
    """
    SVG-based icon.
    """

    def __init__(self, svgKey, svgSrc, parent=None):
        """
        :param svgKey: Shared SVG renderer key
        :type svgKey: str
        :param svgSrc: Path to SVG file or SVG byte data
        :type svgSrc: str | bytes
        :param parent: Optional parent item
        """
        QGraphicsSvgItem.__init__(self, parent)
        renderer = _WGUI_SVG_RENDERERS.get(svgKey)
        if renderer is None:
            renderer = QSvgRenderer(svgSrc)
            _WGUI_SVG_RENDERERS[svgKey] = renderer
        self.setSharedRenderer(renderer)
        WguiIconBase.__init__(self)

    def boundingRect(self):
        return QGraphicsSvgItem.boundingRect(self)

    def setMatrix(self, matrix, combine=False):
        WguiIconBase.setMatrix(self, matrix)

    def setMatrixSuper(self, matrix):
        QGraphicsSvgItem.setMatrix(self, matrix)


class WguiTextIcon(QGraphicsSimpleTextItem, WguiIconBase,
                   metaclass=QGraphicsItem_Meta_M):
    """
    Text-based icon.
    """

    def __init__(self, text, parent=None):
        """
        :param text: Text to display in icon
        :type text: str
        :param parent: Optional parent item
        """
        QGraphicsSimpleTextItem.__init__(self, text, parent)
        WguiIconBase.__init__(self)

    def boundingRect(self):
        return QGraphicsSimpleTextItem.boundingRect(self)

    def setMatrix(self, matrix, combine=False):
        WguiIconBase.setMatrix(self, matrix)

    def setMatrixSuper(self, matrix):
        QGraphicsSimpleTextItem.setMatrix(self, matrix)


def wgui_getFunctionIcon(defaultDim, text=None,
                         iconPath=None, iconData=None,
                         svgPath=None, shape=None, parent=None):
    """
    Get an icon-widget for the given function.

    :param text: Fallback/default icon text
    :type text: str
    :param iconPath: Path to pixmap/SVG file, if any
    :type iconPath: str
    :param iconData: Persisted binary data of file at iconPath
    :type iconData: bytes
    :param svgPath: Path to SVG file, if any
    :type svgPath: str
    :param shape: Icon shape, if any
    :type shape: (list[(float, float)], (float, float))
    :param parent: Parent graphics item
    :return: Icon-widget for the function, its bounding box and its outline
    :rtype: (QGraphicsItem, (float, float), QPainterPath)
    """
    path = QPainterPath()
    iconPath = scmn_expshenv(iconPath)
    if iconPath:
        if iconData:
            qba = QByteArray(iconData)
            if iconPath.lower().endswith('.svg'):
                # noinspection PyTypeChecker
                icon = WguiSvgIcon(iconPath, qba, parent)
            else:
                pixmap = QPixmap()
                pixmap.loadFromData(qba)
                icon = WguiPixmapIcon(pixmap, parent)
            w, h = defaultDim
            path.addRect(QRectF(-w / 2.0, -h / 2.0, w, h))
            return icon, (w, h), path
        if iconPath.startswith(':') or scmn_isfile(iconPath):
            if iconPath.lower().endswith('.svg'):
                icon = WguiSvgIcon(iconPath, iconPath, parent)
            else:
                icon = WguiPixmapIcon(QPixmap(iconPath), parent)
            w, h = defaultDim
            path.addRect(QRectF(-w / 2.0, -h / 2.0, w, h))
            return icon, (w, h), path

    svgPath = scmn_expshenv(svgPath)
    if svgPath and (svgPath.startswith(':') or scmn_isfile(svgPath)):
        svg = WguiSvgIcon(svgPath, svgPath, parent)
        w = svg.boundingRect().width()
        h = svg.boundingRect().height()
        path.addRect(QRectF(-w / 2.0, -h / 2.0, w, h))
        return svg, (w, h), path

    if shape:
        icon, dims = shape
        if icon and dims:
            path.addPolygon(wgui_points2qpoly(icon))
            return None, dims, path

    w, h = defaultDim
    icon = WguiTextIcon(text, parent) if text else None
    path.addRect(QRectF(-w / 2.0, -h / 2.0, w, h))
    return icon, (w, h), path


class WguiIconText(QGraphicsItem):
    """
    Display an item's name and info above/below its icon.
    """

    def __init__(self, data, align=WHDB_HALIGN_CENTER,
                 parent=None, scene=None):
        """
        :param data: Data
        :type data: WhdbNamedEntityP
        :param align: Text alignment
        :type align: int
        :param parent: the parent scene item
        :param scene: the scene object
        """
        super().__init__(parent, scene)

        #: Source data item
        self.__data = data
        #: Text alignment
        self.__align = align

        #: Top text with name
        self.__topText = WguiEditableText(
            placeHolder='Name', fixRefPoint=True, parent=self, scene=scene)
        self.__topText.horizontalAlign = align
        self.__topText.verticalAlign = WHDB_VALIGN_BOTTOM
        self.__topText.textChanged.connect(data.setName)

        #: Bottom text with info
        self.__bottomText = WguiEditableText(
            allowEmpty=True, fixRefPoint=True, parent=self, scene=scene)
        self.__bottomText.horizontalAlign = align
        self.__bottomText.verticalAlign = WHDB_VALIGN_TOP
        self.__bottomText.hideIfEmpty = True
        self.__bottomText.textChanged.connect(data.setInfo)

        self.updateItem()

    @property
    def nameChanged(self):
        """
        Get the signal that emits whenever the name text changes.

        :return: Signal emitting whenever the name text changes
        :rtype: pyqtSignal
        """
        return self.__topText.textChanged

    @property
    def infoTextChanged(self):
        """
        Get the signal that emits whenever the info text changes.

        :return: Signal emitting whenever the info text changes
        :rtype: pyqtSignal
        """
        return self.__bottomText.textChanged

    def sceneEventFilter(self, item:QGraphicsItem, event:QEvent):
        """
        Capture events on the two text fields to catch TAB key presses.
        """
        # noinspection PyUnresolvedReferences
        isTab = isinstance(event, QKeyEvent) and event.key() == Qt.Key_Tab
        if wgui_qgi_isany(item, lambda i: i is self.__topText) and isTab:
            if event.type() == QEvent.KeyRelease:
                self.__bottomText.setFocus(Qt.TabFocusReason)
            return True
        if wgui_qgi_isany(item, lambda i: i is self.__bottomText) and isTab:
            if event.type() == QEvent.KeyRelease:
                self.__topText.setFocus(Qt.TabFocusReason)
            return True

        return False

    def scaleToBox(self, bbox):
        """
        Adjust the text fields' positions to be placed around the given box.

        :param bbox: Bounding box
        :type bbox: QRectF
        """
        if self.__align < 0:
            x = bbox.left()
        elif self.__align > 0:
            x = bbox.right()
        else:
            x = bbox.center().x()

        self.__topText.setPos(x, bbox.top())
        self.__bottomText.setPos(x, bbox.bottom())

    def paint(self, painter, options, widget=None):
        pass

    def boundingRect(self):
        top = self.__topText
        btm = self.__bottomText
        return self.mapRectFromItem(top, top.boundingRect()).united(
            self.mapRectFromItem(btm, btm.boundingRect()))

    def hasFocus(self):
        return self.__topText.hasFocus() or self.__bottomText.hasFocus()

    def focusNameField(self):
        """
        Give the focus to the name field to start editing.
        """
        self.__topText.setFocus(Qt.OtherFocusReason)

    def setNameVisible(self, visible):
        """
        Set the name field visible or hidden.

        :param visible: Whether to show the name field
        :type visible: bool
        """
        self.__topText.setVisible(visible)

    def setInfoVisible(self, visible):
        """
        Set the info field visible or hidden.

        :param visible: Whether to show the info field
        :type visible: bool
        """
        self.__bottomText.setVisible(visible)

    def updateItem(self):
        """
        Update the text field contents to their current corresponding
        database values.
        """
        name, info = self._getNameInfo(self.__data)

        self.__topText.text = name
        self.__bottomText.text = info

    def _getNameInfo(self, item):
        """
        Get the text field contents to their current corresponding
        database values.

        :return: Current name and info values
        :rtype: (str, str)
        """
        return item.getName(), item.getInfo()

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemSceneHasChanged:
            if self.scene() is not None:
                self.__topText.installSceneEventFilter(self)
                self.__bottomText.installSceneEventFilter(self)
            else:
                self.__topText.removeSceneEventFilter(self)
                self.__bottomText.removeSceneEventFilter(self)
        return super().itemChange(change, value)


class WguiInstanceBaseSceneItem(QGraphicsItem, WguiZoomAwareP, WguiSceneItemP,
                                WcmnReleasableP,
                                metaclass=QGraphicsItem_Meta_M):
    """
    The WguiIconSceneItem is the base class for scene items representing
    instances (function or model).
    """
    #: Minimum size
    MINIMUM_SIZE = 2 * _WGUI_RAD_CONN + 1

    #: Width of highlighting outline
    HIGHLIGHT_WIDTH = 3

    #: Z-Levels for various child elements
    CONNECTOR_Z, NAMEINFO_Z, RESIZE_Z = range(1, 4)

    def __init__(self, data, citmF, catmF, posp, vctx, parent=None, scene=None):
        """
        Create a default scene icon.
        :param data: the associated instance
        :type data: WhdbInstanceP
        :param citmF: Connection point scene item creation function
        :type citmF: (WguiConnectionPoint, WguiInstance, ViewContext)
                     -> WguiConnectionPointSceneItem
        :param posp: Connection point position provider
        :type posp: WguiXPointPositionProviderP
        :param vctx: View context
        :type vctx: ViewContext
        :param scene: the scene object
        :type scene: WguiGraphicsScene
        :param parent: the parent object
        """
        # base initialization
        super(WguiInstanceBaseSceneItem, self).__init__(parent, scene)

        self.outline = None
        self.outlineStroke = None

        #: Source data item
        self.__dataItem = data
        #: Connection point scene item creation function
        self.__citmF = citmF
        #: Connection point scene item creation function
        self.__catmF = catmF
        #: Connection point position provider
        self.__positionProvider = posp
        #: View context
        self.__viewContext = vctx
        #: Connector items
        self.__connectorWRefs = {}
        #: Card items
        self.__cardWRefs = {}
        #: Current zoom level (TODO: Get value from context when needed)
        self.__zoom = 1.0
        #: Name/info fields
        self.__nameInfoText = self._createNameInfoFields(data)
        #: Icon
        self.__icon = None
        #: Icon bounding box
        self.__iconBbox = None
        #: Icon outline
        self.__iconOutline = None
        #: Tool-tip function
        self.__tooltipF = wgui_instance_tooltip

        #data.registerObserver(self)

        self._updateIcon()

        #: Item resizing
        # noinspection PyTypeChecker
        self.__resizer = self._createResizer(vctx, scene)

        self.__resizer.setZValue(self.RESIZE_Z)

        # add the function connector list
        self._addConnectorList(data.getConnectionPoints())

        self._updateTooltip()
        self._updateMatrix()
        self._scaleChildrenToBox()

        # set the item position
        self.setPos(*data.getPosition())
        self.setZValue(data.getZ() or 0)
        # make the item selectable and movable
        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges)
        self.setAcceptHoverEvents(True)
        self.hovering = False

        vctx.config.addObserver(self._configChanged)
        self._configChanged()

    def release(self):
        self.__viewContext.config.removeObserver(self._configChanged)
        #self.__dataItem.unregisterObserver(self)

    #def dbNotify(self, obj, event, *args, **kwargs):
    #    pass

    def setTooltipF(self, ttpf):
        """
        Set the tooltip function.

        :param ttpf: Tooltip function
        :type ttpf: (WhdbInstanceP) -> str
        """
        self.__tooltipF = ttpf
        self._updateTooltip()

    def _updateTooltip(self):
        if self.__tooltipF:
            self.setToolTip(self.__tooltipF(self.getDataObject()))

    def _createResizer(self, vctx, scene):
        """
        Create a resizer for this scene item

        :param vctx: View context
        :type vctx: ViewContext
        :param scene: the scene object
        :type scene: WguiGraphicsScene
        :return: Resizer
        :rtype: WguiResizer
        """
        # noinspection PyTypeChecker
        return WguiNoFlipBoundingRectResizer(
            self._getResizingRectAndPath, self._fitToRect,
            lambda t: vctx.snapPoint(*t), margin=_WGUI_RAD_CONN,
            minSize=self.MINIMUM_SIZE, parent=self, scene=scene)

    @property
    def viewContext(self):
        return self.__viewContext

    def _updateIcon(self):
        if self.__icon:
            self.__icon.setVisible(False)
            self.__icon.setParentItem(None)

        self.__icon, self.__iconBbox, self.__iconOutline = self._createIcon()
        self._updateOutline()

    def _createNameInfoFields(self, data):
        """
        Create the name/info elements.

        :param data: Instance
        :type data: WhdbInstanceP
        :return: The name/info fields
        :rtype: WguiIconText
        """
        # noinspection PyUnusedLocal
        def textChanged(text):
            scene = self.scene()
            if scene is not None:
                scene.sceneItemChanged.emit(self.__dataItem)
                scene.sceneItemsChanged.emit(scene.getSceneItems())


        data = self.__dataItem
        nameInfoText = WguiIconText(data, data.getLabelAlignment(),
                                    self, self.scene())
        nameInfoText.nameChanged.connect(textChanged)
        nameInfoText.infoTextChanged.connect(textChanged)
        nameInfoText.setZValue(self.NAMEINFO_Z)
        return nameInfoText

    def _createIcon(self):
        """
        Create the icon.

        :return: The icon, its bounding box and its outline
        :rtype: (QGraphicsItem, (float, float), QPainterPath)
        """
        item = self.getDataObject()
        return wgui_getFunctionIcon(
            item.getDefaultDimension(),
            text=item.getCustomText(),
            iconPath=item.getCustomIcon(),
            iconData=self.viewContext.getResource(item.getCustomIcon()),
            parent=self)

    def _updateMatrix(self):
        """
        Update the scene item's matrix to perform the transformations for the
        current rotation/flipping settings.
        """
        inst = self.__dataItem
        m = QMatrix()
        if inst.isFlippedHorizontally():
            m.setMatrix(-m.m11(), -m.m12(), m.m21(), m.m22(), m.dx(), m.dy())
        if inst.isFlippedVertically():
            m.setMatrix(m.m11(), m.m12(), -m.m21(), -m.m22(), m.dx(), m.dy())
        r = inst.getRotation()
        if r != 0:
            m = m.rotate(r * 360 / (2 * math.pi))
        self.setMatrix(m)

    def setMatrix(self, matrix, combine=False):
        super().setMatrix(matrix, combine)
        inv = self.matrix().inverted()[0]
        if self.__icon is not None:
            self.__icon.setMatrix(inv)
        self.__nameInfoText.setMatrix(inv)

    def _configChanged(self):
        """
        Callback function to handle editor configuration changes.
        """
        cfg = self.__viewContext.config

        self.__nameInfoText.setNameVisible(cfg.displayName)
        self.__nameInfoText.setInfoVisible(cfg.displayInfo)

    def focusNameField(self):
        """
        Give the keyboard focus to the name field.
        """
        self.__nameInfoText.focusNameField()

    def boundingRect(self):
        """
        Get the displayed bounding rectangle

        :return: Instance bounding box, incl. outline, icon, connectors.
        :rtype: QRectF
        """
        # get the enclosing rectangle
        rect = self.outlineBoundingRect()
        # adjust for normal frame
        adj = (self.HIGHLIGHT_WIDTH + 1) / self.__zoom
        return rect.normalized().adjusted(-adj, -adj, adj, adj)

    def outlineBoundingRect(self):
        """
        Get the instance bounding box without the connectors

        :return: Instance outline bounding box
        :rtype: QRectF
        """
        # get the instance object
        wdth, hght = self.__dataItem.getDimension()
        return QRectF(-wdth / 2, -hght / 2, wdth, hght)

    def getDataObject(self):
        return self.__dataItem

    def updateItem(self):
        for ref in self.__connectorWRefs.values():
            item = ref()
            if item is not None:
                item.setVisible(False)
                item.setParentItem(None)
        self.__connectorWRefs.clear()
        self._addConnectorList(self.__dataItem.getConnectionPoints())
        self.__nameInfoText.updateItem()
        self._updateTooltip()
        self._updateIcon()
        self._updateMatrix()
        self._sizeChanged()

    def getZ(self):
        return self.zValue()

    def setZValue(self, zval):
        self.setZ(zval)

    def setZ(self, zval):
        QGraphicsItem.setZValue(self, zval)
        self.__dataItem.setZ(zval)

    def paintHighlight(self, painter, colour):
        """
        Paint the highlight indicators.

        :param painter: Painter to use for painting
        :type painter: QPainter
        :param colour: Highlight colour
        :type colour: QColor
        """
        if self.outlineStroke is None:
            self._setOutlineStroke()

        painter.fillPath(self.outlineStroke, colour)

    def paint(self, painter:QPainter, option, widget=None):
        lc, lw, fc = self.__viewContext.highlighter.highlight(self.__dataItem)

        inst = self.getDataObject()
        hasOutline = not (isinstance(self.__icon, WguiImageIconBase) and
                          inst.isBoundingBoxHidden()) or lc or lw
        outline = self.outline if hasOutline else None

        if outline is not None:
            if fc is not None:
                painter.fillPath(outline, QBrush(fc))
            else:
                painter.fillPath(outline, QBrush(WGUI_TRANS_FILL))

        if self.isSelected():
            # draw the highlighting outline for selection
            self.paintHighlight(painter, WGUI_SELECTION_COLOUR)
        elif self.hovering:
            # draw the highlighting outline for hovering
            self.paintHighlight(painter, WGUI_HOVER_COLOUR)
        if outline is not None:
            self._paintOutline(painter, outline, lc, lw)

    def _paintOutline(self, painter, outline, lc, lw):
        painter.save()
        painter.setPen(QPen(
            lc if lc else WGUI_SHAPE_COLOUR,
            lw if lw is not None else 1))
        painter.drawPath(outline)
        painter.restore()

    def zoomChanged(self, zoom):
        self.__zoom = zoom
        self.outlineStroke = None
        self.__resizer.zoomChanged(zoom)

    def _setOutlineStroke(self):
        ps = QPainterPathStroker()
        ps.setCapStyle(Qt.RoundCap)
        ps.setJoinStyle(Qt.RoundJoin)
        if self.__zoom != 1:
            ps.setWidth((2 * self.HIGHLIGHT_WIDTH) / self.__zoom + 1)
        else:
            ps.setWidth(2 * self.HIGHLIGHT_WIDTH + 1)
        self.outlineStroke = ps.createStroke(self.outline)
        self.outlineStroke = self.outlineStroke.subtracted(self.outline)

    def _updateOutline(self):
        """
        Update the item's outline to the instance/function's icon's outline.
        """
        self.outlineStroke = None

        if self.__iconOutline is None:
            self.outline = QPainterPath()
            self.outline.setFillRule(Qt.WindingFill)
            return

        itemW, itemH = self.__dataItem.getDimension()
        iconW, iconH = self.__iconBbox
        sclx = itemW / iconW
        scly = itemH / iconH
        if sclx != 1.0 or scly != 1.0:
            trns = QTransform().scale(sclx, scly)
            self.outline = trns.map(self.__iconOutline)
        else:
            self.outline = QPainterPath(self.__iconOutline)

    def _scaleChildrenToBox(self):
        """
        Scale all scalable child items to the current bounding box.
        """
        box = self._rotationAdjustedBoundingRect()
        if self.__icon is not None:
            self.__icon.scaleToBox(box)
        self.__nameInfoText.scaleToBox(box)

    def _addConnectorList(self, clst):
        """
        Add a connector list to the item

        :param clst: The connector list to add
        :type clst: Iterable[WhdbConnectionPointP]
        """
        # check for none
        if clst is None: return
        # loop in the list
        for c in clst:
            self._addConnectorGraphicItem(c)

    def _addCardList(self, cardlst):
        """
        Add a card list to the item

        :param cardlst: The card list to add
        :type cardlst: Iterable[str]
        """
        # check for none
        if not cardlst: return
        # loop in the list
        for c in cardlst:
            self._addCardGraphicItem(c)

    def _addConnectorGraphicItem(self, citm):
        """
        Add a connector graphic item

        :param citm: The connector to add
        :type citm: WhdbConnectionPointP
        :return Created connection point, if any
        :rtype: WguiConnectionPointSceneItem | None
        """
        # make sure it is ok
        if not isinstance(citm, WhdbConnectionPointP):
            return
        # add the connector item
        item = self.__citmF(
            citm, self.__dataItem, self.__viewContext, parent=self)
        item.setZValue(self.CONNECTOR_Z)
        self.__connectorWRefs[citm.getUUID()] = weakref.ref(item)
        return item

    def _addCardGraphicItem(self, card):
        """
        Add a card graphic item

        :param card: The card to add
        :type card: str
        :return Created card point, if any
        :rtype: item
        """
        if self.__catmF:
            # add the connector item
            item = self.__catmF(
                card, self.__dataItem, self.__viewContext, parent=self)
            item.setZValue(self.CONNECTOR_Z)
            self.__cardWRefs[card] = weakref.ref(item)
            return item

    def hoverEnterEvent(self, event):
        self.hovering = True
        self.update()

    def hoverLeaveEvent(self, event):
        self.hovering = False
        self.update()

    def _fitToRect(self, rect):
        """
        Callback function handling resizing initiated by the resizing handles.

        :param rect: Current bounding rectangle of the resizing handles
        :type rect: ((float, float), (float, float))
        """
        (x1, y1), (x2, y2) = rect

        if whdb_isFlipRotation(self.__dataItem.getRotation()):
            self.__dataItem.setDimension(abs(y2 - y1), abs(x2 - x1))
        else:
            self.__dataItem.setDimension(abs(x2 - x1), abs(y2 - y1))
        self.__dataItem.setPosition(x1 + (x2 - x1) / 2, y1 + (y2 - y1) / 2)

        self._sizeChanged()

    def _getResizingRectAndPath(self):
        """
        Get the resizing rectangle and path.

        :return: Resizing recangle and path
        :rtype: (QRectF, QPainterPath)
        """
        rect = self.mapRectToScene(self.outlineBoundingRect())
        path = QPainterPath()
        path.addRect(rect)

        return rect, path

    def _sizeChanged(self):
        """
        Adjust all sizes to the size properties of the database item.
        """
        self.prepareGeometryChange()
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, False)
        self.setPos(*self.__dataItem.getPosition())
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self._updateOutline()
        self._scaleChildrenToBox()

        # update the connector list
        for item in self.childItems():
            if isinstance(item, WguiConnectionPointSceneItem):
                item.updatePos()
        self.scene().updateConnectionsForInstance(self)

    def _rotationAdjustedBoundingRect(self):
        """
        Get the bounding rectangle adjusted for rotation.
        For 90/270 degree rotations, width and hight will be swapped.

        :return: Bounding rectangle adjusted for rotation
        :rtype: QRectF
        """
        if self.matrix().m12() != 0:
            brct = self.outlineBoundingRect()
            cntr = brct.center()
            brct = QRectF(-brct.height() / 2, -brct.width() / 2,
                          brct.height(), brct.width())
            brct.translate(cntr)
            return brct
        else:
            return self.outlineBoundingRect()

    def itemChange(self, change, value):
        result = value
        if change == QGraphicsItem.ItemPositionChange:
            # get the graphic scene and snap flag
            scene = self.scene()
            sflg = scene.getSnapFlag()
            xsnp = scene.getSnapX()
            ysnp = scene.getSnapY()
            # set position value
            x = value.x()
            y = value.y()
            if sflg:
                x = wgui_snap_coord(x, xsnp)
                y = wgui_snap_coord(y, ysnp)
                # get the data object as a function
            inst = self.__dataItem
            inst.setPosition(x, y)
            result = QPointF(x, y)
        return result

    def getConnectionPointPosition(self, pos):
        """
        Add a new instance connector by position.

        :param pos: the connector position in scene coordinate
        :type pos: QPointF
        :return: New connection point, if any
        :rtype: WguiConnectionPointSceneItem | None
        """
        # get the instance object
        inst = self.getDataObject()
        # get the instance bounding box (with connectors)
        bbox = self.matrix().mapRect(self.boundingRect())
        bbox.translate(*inst.getPosition())
        # check if the position is inside the box
        if not bbox.contains(pos):
            return None
        spos = self.mapFromScene(pos)
        return self.__positionProvider.getXPointPosition(
            (spos.x(), spos.y()), inst, self)

    def getConnector(self, conn):
        ref = self.__connectorWRefs.get(conn.getUUID())
        return ref() if ref is not None else None


class WguiInstanceSceneItem(WguiInstanceBaseSceneItem):
    """
    The WguiInstanceSceneItem is the instance scene item which is
    constructed as an icon scene item.
    """

    def __init__(self, inst, citmF, catmF, posp, vctx, parent=None, scene=None):
        """
        :param inst: the the database instance
        :type inst: WhdbInstanceP
        :param citmF: Connection point scene item creation function
        :type citmF: (WguiConnectionPoint, WguiInstance, ViewContext)
                     -> WguiConnectionPointSceneItem
        :param posp: Connection point position provider
        :type posp: WguiXPointPositionProviderP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the scene object
        """
        super().__init__(inst, citmF, catmF, posp, vctx, parent, scene)


class WguiOperatorInstanceSceneItem(WguiInstanceBaseSceneItem):
    """
    The WguiFunctionSceneItem is the function scene item which is
    constructed as an icon scene item.
    """

    def __init__(self, inst, citmF, catmF, posp, vctx, parent=None, scene=None):
        """
        :param inst: the the database function
        :type inst: WhdbOperatorInstanceP
        :param citmF: Connection point scene item creation function
        :type citmF: (WguiConnectionPoint, WguiInstance, ViewContext)
                     -> WguiConnectionPointSceneItem
        :param posp: Connection point position provider
        :type posp: WguiXPointPositionProviderP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the scene object
        """
        super().__init__(inst, citmF, catmF, posp, vctx, parent, scene)

    # noinspection PyUnresolvedReferences
    def _createIcon(self):
        """
        Create the icon.

        :return: The icon, its bounding box and its outline
        :rtype: (QGraphicsItem, (float, float), QPainterPath)
        """
        inst = self.getDataObject()
        op = inst.getOperator()
        return wgui_getFunctionIcon(
            inst.getDefaultDimension(),
            text=inst.getCustomText() or inst.getSymbol() or op.getName(),
            iconPath=self.getDataObject().getCustomIcon(),
            iconData=self.viewContext.getResource(inst.getCustomIcon()),
            svgPath=op.getIcon(WGUI_ICON_SVG),
            shape=op.getShape(),
            parent=self)


class WguiConnectionSceneItem(QGraphicsPathItem, WguiSceneItemP,
                              metaclass=QGraphicsItem_Meta_M):
    """
    The WguiConnectionSceneItem is the connection scene item which is
    constructed as polygon scene item.
    """
    #: Shape width
    SHAPE_PATH_WIDTH = 10

    def __init__(self, conn, vctx, parent=None, scene=None):
        """
        :param conn: the database connection
        :type conn: WhdbConnectionP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the scene object
        """
        # base initialization
        super().__init__(parent, scene)
        #: Connection datase item
        self.__connection = conn
        #: Connection path
        self.__path = conn.getPath()
        #: View context
        self.__viewContext = vctx
        #: Whether the connection is currently hovered over
        self.__hovering = False
        #: Line currently being dragged, if any
        self.__editedLine = None
        #: Shape path
        self._shapePath = None
        #: Tool-tip function
        self.__tooltipF = wgui_directed_connection_tooltip

        self.__color = WGUI_CONNECTION_COLOUR
        self.__lineStyle = WGUI_CONNECTION_STYLE

        super().setZValue(conn.getZ() or 0)

        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.setAcceptHoverEvents(True)
        self.setAcceptedMouseButtons(Qt.LeftButton)

        # update the graphic path
        self.updateGraphicPath()
        self._updateTooltip()

    def setTooltipF(self, ttpf):
        """
        Set the tooltip function.

        :param ttpf: Tooltip function
        :type ttpf: (WhdbConnectionP) -> str
        """
        self.__tooltipF = ttpf
        self._updateTooltip()

    def _updateTooltip(self):
        if self.__tooltipF:
            self.setToolTip(self.__tooltipF(self.getDataObject()))

    def getDataObject(self):
        return self.__connection

    def updateItem(self):
        self.updateGraphicPath()
        self._updateTooltip()
        self.update()

    def getZ(self):
        return self.zValue()

    def setZValue(self, zval):
        self.setZ(zval)

    def setZ(self, zval):
        QGraphicsItem.setZValue(self, zval)
        self.__connection.setZ(zval)

    def isConnected(self, sceneItem):
        """
        Whether this connection is connected to the given scene item.

        :param sceneItem: Scene item to check for connection
        :type sceneItem: WguiSceneItemP
        """
        item = sceneItem.getDataObject()
        # noinspection PyTypeChecker
        return isinstance(item, WhdbInstanceP) and \
               self.__connection.isConnected(item)

    def updateGraphicPath(self):
        """
        Update automatically the connection path
        """
        cnctn = self.__connection
        sx, sy = cnctn.getStartObject().getEffectiveConnectionPointPosition(
            cnctn.getStartConnectionPoint())
        ex, ey = cnctn.getEndObject().getEffectiveConnectionPointPosition(
            cnctn.getEndConnectionPoint())

        path = QPainterPath()
        path.moveTo(sx, sy)
        path.lineTo(ex, ey)

        self._shapePath = QPainterPath(path)
        self.setPath(path)
        self.update()

    def getConnectionEndOverlaps(self, conn):
        return _WGUI_RAD_CONN, _WGUI_RAD_CONN

    def getShapePathWidth(self):
        return self.SHAPE_PATH_WIDTH

    def getPathWidth(self):
        return 1

    def boundingRect(self):
        rect = QGraphicsPathItem.boundingRect(self)

        # Extend the default bounding rectangle to accommodate an enlarged
        # selection area
        return rect.adjusted(-5, -5, 5, 5)

    def paint(self, painter, option, widget=None):
        painter.save()
        colour = None
        if self.isSelected():
            colour = WGUI_SELECTION_COLOUR
        elif self.__hovering:
            colour = WGUI_HOVER_COLOUR
        if colour is not None:
            ps = QPainterPathStroker()
            ps.setCapStyle(Qt.FlatCap)
            ps.setWidth(self.getPathWidth() + 4)
            painter.fillPath(ps.createStroke(self.path()), colour)
        lc, lw, _ = self.__viewContext.highlighter.highlight(self.__connection)
        self._paintPath(painter, lc, lw)
        painter.restore()

    def _paintPath(self, painter, lc, lw):
        # noinspection PyArgumentList
        painter.setPen(QPen(
            lc if lc is not None else self.__color,
            (lw if lw is not None else 1) + self.getPathWidth(),
            style=self.__lineStyle, cap=Qt.FlatCap))
        painter.drawPath(self.path())

    def resetPath(self):
        """
        Reset the connection path. This will undo all user-modifications done
        to the path.
        """
        # get the connection path
        self.__path.resetPath()
        self.updateGraphicPath()

    def moveSelectedPathEndPoints(self, slst):
        """
        Move the connection path end points according to the locations
        of the connected objects/connectors..
        This function is intended to be used when the end point(s) has/have
        been selected and the selection is being moved subsequently.

        If only one end point is selected, the adjacent path segment will
        be adjusted.
        If both end points are selected, the entire path will be translated.
        """
        cnctn = self.__connection
        # reset the connection path
        if cnctn.getStartObject() in slst and cnctn.getEndObject() in slst:
            self.__path.relocatePath()
        elif cnctn.getStartObject() in slst:
            self.__path.resetPathSrcEnd()
        elif cnctn.getEndObject() in slst:
            self.__path.resetPathDestEnd()
        else:
            self.__path.resetPath()
            # set the graphic path
        self.updateGraphicPath()

    def moveConnector(self, conn, reset):
        """
        Move the given connector to its latest position.

        :param conn: Connector to move
        :type conn: WhdbConnectionPointP
        :param reset: Whether the gravity of the connector has changed
        :type reset: bool
        """
        cnctn = self.__connection
        if not (conn is cnctn.getStartConnectionPoint() or
                        conn is cnctn.getEndConnectionPoint()):
            return
        if reset:
            self.__path.resetPath()
        elif conn is cnctn.getStartConnectionPoint():
            self.__path.resetPathSrcEnd()
        elif conn is cnctn.getEndConnectionPoint():
            self.__path.resetPathDestEnd()
        self.updateGraphicPath()

    def finishPathUpdate(self):
        """
        Indicates that a sequence of movement/editing operations
        has been completed.
        """
        pass

    def shape(self):
        # Limit the selection area to the outline of the path.
        ps = QPainterPathStroker()
        ps.setWidth(self.getShapePathWidth())
        return ps.createStroke(self._shapePath or self.path())

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        x = event.scenePos().x()
        y = event.scenePos().y()
        self._setEditedLine(x, y)

    def mouseReleaseEvent(self, event):
        self._unsetEditedLine()

    def mouseMoveEvent(self, event):
        x = event.scenePos().x()
        y = event.scenePos().y()
        self._updateEditedLine(x, y)

    def hoverEnterEvent(self, event):
        self.__hovering = True
        self.update()

    def hoverLeaveEvent(self, event):
        self.__hovering = False
        self.update()

    def _setEditedLine(self, x, y):
        self.__editedLine = self.__path.getPathSegment(x, y, pad=5, mthr=3)

        if self.__editedLine is not None:
            # Segment affected by event was found
            self.setSelected(True)

    def _updateEditedLine(self, x, y):
        if self.__editedLine is None: return

        scene = self.scene()
        sflg = scene.getSnapFlag()
        if sflg:
            x = wgui_snap_coord(x, scene.getSnapX())
            y = wgui_snap_coord(y, scene.getSnapY())

        self.__editedLine.move(x, y)
        self.updateGraphicPath()
        self.update()

    def _unsetEditedLine(self):
        if self.__editedLine is not None:
            self.__editedLine.finish()
            self.__editedLine = None

            self.updateGraphicPath()
            self.update()

    def setConnColour(self, color):
        self.__color = color

    def setConnStyle(self, style):
        self.__lineStyle = style


class WguiManhattanConnectionSceneItem(WguiConnectionSceneItem):
    def __init__(self, conn, vctx, parent=None, scene=None):
        """
        :param conn: the database connection
        :type conn: WhdbConnectionP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the scene object
        """
        #: View context
        self.__viewContext = vctx
        self.__path = conn.getPath()
        """:type: WhdbManhattanConnectionPath"""

        super().__init__(conn, vctx, parent, scene)

    def finishPathUpdate(self):
        path = self.getDataObject().getPath()
        """:type: WhdbManhattanConnectionPath"""
        path.simplify()

    def updateGraphicPath(self):
        def shorten(p, n, s, sx):
            if p.getX() == n.getX():
                if p.getY() < n.getY():
                    return p.getX(), p.getY() + sx + s
                else:
                    return p.getX(), p.getY() - sx - s
            else:
                if p.getX() < n.getX():
                    return p.getX() + sx + s, p.getY()
                else:
                    return p.getX() - sx - s, p.getY()

        def mkPath(s=0):
            s1, s2 = self.getConnectionEndOverlaps(self.getDataObject())
            qpath = QPainterPath()
            px, py = shorten(points[0], points[1], s, s1)
            qpath.moveTo(px, py)
            # loop in the path list
            for p in points[1:-1]:
                # set the start/end points
                qpath.lineTo(p.getX(), p.getY())
            px, py = shorten(points[-1], points[-2], s, s2)
            qpath.lineTo(px, py)
            return qpath

        # get the connection points
        points = self.__path.getPoints()
        # preset start/end points
        if len(points) < 2: return

        self._shapePath = mkPath(self.getShapePathWidth() / 2)
        self.setPath(mkPath())
        self.update()


class WguiArcConnectionSceneItem(WguiConnectionSceneItem):
    def __init__(self, conn, vctx, parent=None, scene=None):
        """
        :param conn: the database connection
        :type conn: WhdbConnectionP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the scene object
        """
        #: View context
        self.__viewContext = vctx
        self.__arrowHead = None

        super().__init__(conn, vctx, parent, scene)

    def updateGraphicPath(self):
        # Transition start/end points on state shapes and bezier point
        asx, asy, aex, aey, px, py, _ = \
            whdb_getArcEndPoints(self.getDataObject())

        path = QPainterPath()
        path.moveTo(asx, asy)
        path.quadTo(px, py, aex, aey)

        fang = scmn_degToRad(path.angleAtPercent(1.0))
        wgui_addArrowHead(aex, aey, fang, path)

        arrowHead = QPainterPath()
        arrowHead.moveTo(aex, aey)
        wgui_addArrowHead(aex, aey, fang, arrowHead)

        self.__arrowHead = arrowHead
        self._shapePath = path
        self.setPath(path)
        self.update()

    def paint(self, painter:QPainter, option, widget=None):
        painter.save()
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setRenderHint(QPainter.HighQualityAntialiasing)
        super().paint(painter, option, widget)
        painter.restore()

    def _paintPath(self, painter, lc, lw):
        super()._paintPath(painter, lc, lw)
        if self.__arrowHead:
            brush = QBrush(
                lc if lc is not None else WGUI_CONNECTION_COLOUR)
            painter.fillPath(self.__arrowHead, brush)


class WguiCircularConnectionSceneItem(WguiConnectionSceneItem):
    def __init__(self, conn, vctx, parent=None, scene=None):
        """
        :param conn: the database connection
        :type conn: WhdbConnectionP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the scene object
        """
        #: View context
        self.__viewContext = vctx
        self.__arrowHead = None

        super().__init__(conn, vctx, parent, scene)

    def updateGraphicPath(self):
        # Transition start/end points on state shapes and bezier point
        path = self.getDataObject().getPath()
        """:type: WhdbCircularConnectionPath"""
        cx, cy = path.getCenter()
        radius = path.getRadius()
        a1, a2 = path.getArcAngles()

        qpath = QPainterPath()
        qpath.arcMoveTo(cx - radius, cy - radius, 2 * radius, 2 * radius,
                        scmn_radToDeg(a1))
        qpath.arcTo(cx - radius, cy - radius, 2 * radius, 2 * radius,
                    scmn_radToDeg(a1), -scmn_radToDeg(a2))

        fang = scmn_degToRad(qpath.angleAtPercent(1.0))
        epnt = qpath.pointAtPercent(1.0)
        wgui_addArrowHead(epnt.x(), epnt.y(), fang, qpath)

        arrowHead = QPainterPath()
        arrowHead.moveTo(epnt.x(), epnt.y())
        wgui_addArrowHead(epnt.x(), epnt.y(), fang, arrowHead)

        self.__arrowHead = arrowHead
        self._shapePath = qpath
        self.setPath(qpath)
        self.update()

    def paint(self, painter:QPainter, option, widget=None):
        painter.save()
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setRenderHint(QPainter.HighQualityAntialiasing)
        super().paint(painter, option, widget)
        painter.restore()

    def _paintPath(self, painter, lc, lw):
        super()._paintPath(painter, lc, lw)
        if self.__arrowHead:
            brush = QBrush(
                lc if lc is not None else WGUI_CONNECTION_COLOUR)
            painter.fillPath(self.__arrowHead, brush)


class WguiXPointPositionProviderP(metaclass=ABCMeta):
    """
    Protocol for connection point position providers.

    Connection point position providers are used to determine the position of
    a connection point, given a position on the scene (e.g. from a mouse click
    or drag) for connection point creation and re-positioning.
    """

    @abstractmethod
    def getXPointPosition(self, pos, inst, sitm, fcng=None, **kwargs):
        """
        For the given scene position, return the corresponding/closest
        connection point position in relative coordinates.

        :param pos: Scene position
        :type pos: (float, float)
        :param inst: Parent instance of the connection point
        :type inst: wcr_core.db_api.whdb_protocols_api.WhdbHasDimensionP
        :param sitm: Scene instance item
        :type sitm: WguiSceneItemP
        :param fcng: Current facing/gravity value, if any
        :type fcng: int | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Adjusted position and new facing/gravity value
        :rtype: (float, float), int
        """
        raise NotImplementedError()


class WguiXPointRectPositionProvider(WguiXPointPositionProviderP):
    """
    Connection point position provider for instances with rectangular icons.
    """

    def getXPointPosition(self, pos, inst, sitm, fcng=None, **kwargs):
        # Todo: Get snap function from context
        w, h = inst.getDimension()
        x, y = pos
        x1, y1 = -w / 2, -h / 2
        x2, y2 = w / 2, h / 2
        g = fcng

        scene = sitm.scene()

        if abs(x * h / w) > abs(y):
            if x > 0 and (g is None or x > x2 - _WGUI_RAD_CONN):
                g = WhdbConnectionPointP.East
            elif g is None or x < x1 + _WGUI_RAD_CONN:
                g = WhdbConnectionPointP.West
        else:
            if y > 0 and (g is None or y > y2 - _WGUI_RAD_CONN):
                g = WhdbConnectionPointP.South
            elif g is None or y < y1 + _WGUI_RAD_CONN:
                g = WhdbConnectionPointP.North

        if g == WhdbConnectionPointP.East or \
                        g == WhdbConnectionPointP.West:
            x = x2 if g == WhdbConnectionPointP.East else x1
            if scene.getSnapFlag() and sitm is not None:
                p = sitm.mapToScene(x, y)
                sy = wgui_snap_coord(p.y(), scene.getSnapY())
                y = sitm.mapFromScene(p.x(), sy).y()
            y = min(max(y, y1), y2)
        else:
            y = y2 if g == WhdbConnectionPointP.South else y1
            if scene.getSnapFlag() and sitm is not None:
                p = sitm.mapToScene(x, y)
                sx = wgui_snap_coord(p.x(), scene.getSnapX())
                x = sitm.mapFromScene(sx, p.y()).x()
            x = min(max(x, x1), x2)

        return (x / (w / 2), y / (h / 2)), g


class WguiXPointCenterPositionProvider(WguiXPointPositionProviderP):
    """
    Connection point position provider placing all connection points
    on the center of the instance.
    """

    def getXPointPosition(self, pos, inst, sitm, fcng=None, **kwargs):
        return (0.0, 0.0), WhdbConnectionPointP.North


class WguiXPointCirclePositionProvider(WguiXPointPositionProviderP):
    """
    Connection point position provider for instances with round icons.
    """

    def getXPointPosition(self, pos, inst, sitm, fcng=None, **kwargs):
        # Todo: Get snap function from context
        w, h = inst.getDimension()
        x, y = pos
        x1, y1 = -w / 2, -h / 2
        x2, y2 = w / 2, h / 2
        g = fcng

        if abs(x * h / w) > abs(y):
            if x > 0 and (g is None or x > x2 - _WGUI_RAD_CONN):
                g = WhdbConnectionPointP.East
            elif g is None or x < x1 + _WGUI_RAD_CONN:
                g = WhdbConnectionPointP.West
        else:
            if y > 0 and (g is None or y > y2 - _WGUI_RAD_CONN):
                g = WhdbConnectionPointP.South
            elif g is None or y < y1 + _WGUI_RAD_CONN:
                g = WhdbConnectionPointP.North

        l = math.sqrt(pow(x, 2) + pow(y, 2))
        cx = x / l
        cy = y / l

        return (cx, cy), g


def wgui_addArrowHead(aex, aey, fang, path):
    fang -= SCMN_QUARTER_CIRCLE_RAD
    ang1 = fang + math.pi / 6
    path.lineTo(aex + math.sin(ang1) * 10, aey + math.cos(ang1) * 10)
    ang1 = fang - math.pi / 6
    path.lineTo(aex + math.sin(ang1) * 10, aey + math.cos(ang1) * 10)
    path.lineTo(aex, aey)
