# ----------------------------------------------------------------------------
# - wgui-scene-item-factory                                                  -
# - wild cruncher: scene item factory                                        -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_scene_item_factory
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher scene item factory

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod, ABCMeta
from wcr_core.db_api.whdb_shape_api import WhdbShapeVisitorP

from wcr_frontend.gui.wgui_scene_item import WguiConnectionPointSceneItem, \
    WguiInstanceSceneItem, WguiOperatorInstanceSceneItem, \
    WguiConnectionSceneItem, WguiXPointRectPositionProvider, \
    WguiArcConnectionSceneItem, WguiManhattanConnectionSceneItem, \
    WguiCircularConnectionSceneItem, WguiCardSceneItem
from wcr_frontend.gui.wgui_shape import WguiRectangle, WguiLine, WguiOval, \
    WguiText, WguiArrow
from wcr_core.db.whdb_connection_path_arc import WhdbArcConnectionPath
from wcr_core.db.whdb_connection_path_circular import \
    WhdbCircularConnectionPath
from wcr_core.db.whdb_connection_path_manhattan import \
    WhdbManhattanConnectionPath


class WguiSceneItemFactoryP(metaclass=ABCMeta):
    @abstractmethod
    def createBlockInstanceSceneItem(self, obj, *args, **kwargs):
        """
        Create a scene item for a block instance.

        :param obj: HDS block instance to create scene item for
        :type obj: WhdbBlockInstanceP
        :return: Created scene item
        :rtype: WguiInstanceSceneItem
        """
        raise NotImplementedError()

    @abstractmethod
    def createModelInstanceSceneItem(self, obj, *args, **kwargs):
        """
        Create a scene item for a model instance.

        :param obj: HDS model instance to create scene item for
        :type obj: WhdbModelInstanceP
        :return: Created scene item
        :rtype: WguiInstanceSceneItem
        """
        raise NotImplementedError()

    @abstractmethod
    def createOperatorInstanceSceneItem(self, obj, *args, **kwargs):
        """
        Create a scene item for an operator instance.

        :param obj: HDS operator instance to create scene item for
        :type obj: WhdbOperatorInstanceP
        :return: Created scene item
        :rtype: WguiOperatorInstanceSceneItem
        """
        raise NotImplementedError()

    @abstractmethod
    def createConnectionSceneItem(self, obj, *args, **kwargs):
        """
        Create a scene item for a connection.

        :param obj: HDS block instance to create scene item for
        :type obj: WhdbConnectionP
        :return: Created scene item
        :rtype: WguiConnectionSceneItem
        """
        raise NotImplementedError()

    @abstractmethod
    def createShapeSceneItem(self, obj, *args, **kwargs):
        """
        Create a scene item for a shape.

        :param obj: HDS shape to create scene item for
        :type obj: WhdbShapeP
        :return: Created scene item
        :rtype: WguiShape
        """
        raise NotImplementedError()


class WguiDefaultSceneItemFactoy(WguiSceneItemFactoryP):
    def __init__(self, posp):
        """
        :param posp: Connection point position provider
        :type posp: WguiXPointPositionProviderP
        """
        self.__positionProvider = posp

    @property
    def positionProvider(self):
        """
        :return: Connection point position provider
        :rtype: WguiXPointPositionProviderP
        """
        return self.__positionProvider

    # noinspection PyMethodMayBeStatic
    def createConnectionPoint(self, obj, item, *args, **kwargs):
        return WguiConnectionPointSceneItem(
            obj, item, self.__positionProvider, *args, **kwargs)

    # noinspection PyMethodMayBeStatic
    def createCardPoint(self, obj, item, *args, **kwargs):
            return WguiCardSceneItem(
                obj, item, self.__positionProvider, *args, **kwargs)

    def createBlockInstanceSceneItem(self, obj, *args, **kwargs):
        # noinspection PyTypeChecker
        return WguiInstanceSceneItem(
            obj, self.createConnectionPoint, self.createCardPoint,
            self.__positionProvider, *args, **kwargs)

    def createModelInstanceSceneItem(self, obj, *args, **kwargs):
        # noinspection PyTypeChecker
        return WguiInstanceSceneItem(
            obj, self.createConnectionPoint, self.createCardPoint,
            self.__positionProvider, *args, **kwargs)

    def createOperatorInstanceSceneItem(self, obj, *args, **kwargs):
        # noinspection PyTypeChecker
        return WguiOperatorInstanceSceneItem(
            obj, self.createConnectionPoint, self.createCardPoint,
            self.__positionProvider, *args, **kwargs)

    def createConnectionSceneItem(self, obj, *args, **kwargs):
        path = obj.getPath()
        if isinstance(path, WhdbArcConnectionPath):
            return WguiArcConnectionSceneItem(obj, *args, **kwargs)
        elif isinstance(path, WhdbCircularConnectionPath):
            return WguiCircularConnectionSceneItem(obj, *args, **kwargs)
        elif isinstance(path, WhdbManhattanConnectionPath):
            return WguiManhattanConnectionSceneItem(obj, *args, **kwargs)
        return WguiConnectionSceneItem(obj, *args, **kwargs)

    def createShapeSceneItem(self, obj, *args, **kwargs):
        # noinspection PyTypeChecker
        return obj.acceptVisitor(
            _WGUI_SCENE_ITEM_SHAPE_VISITOR, *args, **kwargs)


class WguiSceneItemShapeVisitor(WhdbShapeVisitorP):
    def visitRectangle(self, rectangle, *args, **kwargs):
        return WguiRectangle(rectangle, *args, **kwargs)

    def visitLine(self, line, *args, **kwargs):
        return WguiLine(line, *args, **kwargs)

    def visitOval(self, oval, *args, **kwargs):
        return WguiOval(oval, *args, **kwargs)

    def visitText(self, text, *args, **kwargs):
        return WguiText(text, *args, **kwargs)

    def visitArrow(self, arrow, *args, **kwargs):
        return WguiArrow(arrow, *args, **kwargs)


_WGUI_SCENE_ITEM_SHAPE_VISITOR = WguiSceneItemShapeVisitor()

#: Default scene item factory
WGUI_DEFAULT_SCENE_ITEM_FACTORY = WguiDefaultSceneItemFactoy(
    WguiXPointRectPositionProvider())
