# ----------------------------------------------------------------------------
# - wgui-scene                                                               -
# - wild cruncher - cockpit: graphics view                                   -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_scene_view
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graphics view

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from math import floor, log10

from PyQt5 import QtCore
from PyQt5.QtWidgets import QGraphicsView, QRubberBand

from wcr_frontend.gui.wgui_scene import WguiSceneBlockChangeAware
from wcr_frontend.gui.wgui_scene_item_ids import *
from wcr_frontend.gui.wgui_scene_item import *
from wcr_frontend.modules.vlink.gui.wvgui_scene import WvguiGraphicsScene


class WguiGraphicsView(QGraphicsView, WguiSceneBlockChangeAware,
                       metaclass=QObject_Meta_M):
    """
    The WguiGraphicsView is the class which acts as a scene viewer. The
    class binds a graphic widget to the scene. Such view is built in
    reference to the application context.
    """
    #: Signal triggering whenever the size of the view changes
    sizeChanged = QtCore.pyqtSignal()

    #: Signal triggering whenever the zoom-factor changes
    zoomChanged = QtCore.pyqtSignal(float)

    def __init__(self, pctx, cctx, mouseHandler, scene):
        """
        Create a new graphic view by parent and context
        :param pctx: Application context
        :type pctx: WcorProjectContextP
        :param cctx: Cockpit system context
        :type cctx: WguiCockpitSystemContextP
        :param mouseHandler: Mouse event handler
        :type mouseHandler: WguiSceneViewMouseHanderP
        :param scene: the parent widget
        :type scene: WguiGraphicsScene
        """
        # base initialization
        super(WguiGraphicsView, self).__init__(scene)
        #: Application context
        self.__pctx = pctx
        #: Cockpit system context
        self.__cctx = cctx
        #: Mouse event handler
        self.__mouseHandler = mouseHandler
        #: Parent scene
        self.__gfxs = scene
        # in Traffic Route view don't show view
        self.isTrafficRouteScene(scene)
        if self.__isTRScene:
            self.setVisible(False)
        # the rubber band origin
        self.__rbop = None
        # the rubber band
        self.__band = QRubberBand(QRubberBand.Rectangle, self)
        # the context position
        self.__cpos = None
        #: Whether the mouse is currently dragging
        self.__dragging = False
        self.__blockPositions = {}
        self.setDragMode(QGraphicsView.RubberBandDrag)
        if not self.__isTRScene:
            self.__gfxs.sceneItemAdded.connect(self._resetSceneRect)
            self.__gfxs.sceneItemRemoved.connect(self._resetSceneRect)
            self.__gfxs.sceneChanged.connect(self._resetSceneRect)

        self.setAcceptDrops(True)

    def getCctx(self):
        return self.__cctx

    def isTrafficRouteScene(self, scene):
        if isinstance(scene, WvguiGraphicsScene):
            self.__isTRScene = True
        else:
            self.__isTRScene = False

    def scene(self):
        if self.__isTRScene:
            scene = self.__gfxs
        else:
            scene = super().scene()
        return scene

    def sceneBlockChanging(self):
        """
        Callback function for when the scene is about to load a new model
        block.
        """
        uuid = self.__gfxs.topLevelModelBlock.getUUID()
        self.__blockPositions[uuid] = (self.viewCenter(), self.zoom)

    def sceneBlockChanged(self):
        """
        Callback function for when the scene has loaded load a new model
        block.
        """
        uuid = self.__gfxs.topLevelModelBlock.getUUID()
        pos = self.__blockPositions.get(uuid)
        if pos is not None:
            center, zoom = pos
            self.centerOn(center)
            self.zoom = zoom
        else:
            self.zoomToFit()

    def _roundZoom(self, zoom):
        z = round(zoom, -int(floor(log10(zoom))) + 1)
        return max(min(z, 100), 0.01)

    def zoomIn(self):
        """
        Zoom in.
        """
        self.zoom = self._roundZoom(self.zoom * 1.1)

    def zoomOut(self):
        """
        Zoom out.
        """
        self.zoom = self._roundZoom(self.zoom / 1.1)

    def zoomToFit(self):
        """
        Zoom to a level that allows the entire scene to be scene.
        """
        rect = self.__gfxs.itemsBoundingRect()
        self.fitInView(rect, Qt.KeepAspectRatio)

    def zoomToOriginal(self):
        """
        Zoom to 100%.
        """
        self.zoom = 1.0

    # noinspection PyUnusedLocal
    def _resetSceneRect(self, *args):
        """
        Callback function to adjust scene rect when new items are added.
        """
        rect = self.__gfxs.itemsBoundingRect()
        w = self.width() / self.zoom
        h = self.height() / self.zoom
        adj = 20 / self.zoom
        if rect.width() < w or rect.height() < h:
            vc = self.viewCenter()
            vrct = QRectF(vc.x() - w / 2, vc.y() - h / 2, w, h)
            rect = rect.adjusted(-adj, -adj, adj, adj).united(vrct)
        else:
            rect = rect.adjusted(-adj, -adj, adj, adj)
        self.__gfxs.setSceneRect(rect)

    def _zoomChanged(self):
        """
        Notify all clients and listeners about a change in zoom level.
        """
        zoom = self.zoom

        for c in self.items():
            if isinstance(c, WguiZoomAwareP):
                c.zoomChanged(zoom)

        self.zoomChanged.emit(zoom)

    @property
    def zoom(self):
        """
        Get the current zoom level.

        :return: Current zoom level
        :rtype: float
        """
        return self.transform().m11()

    @zoom.setter
    def zoom(self, value):
        """
        Set the zoom level.

        :param value: New zoom level
        :type value: float
        """
        self.setTransform(QTransform().scale(value, value))

    def fitInView(self, *__args):
        if self.__isTRScene:
            return
        self._resetSceneRect()
        super().fitInView(*__args)
        self._zoomChanged()

    def setTransform(self, transform, bool_combine=False):
        super().setTransform(transform, bool_combine)
        self._zoomChanged()

    def setMatrix(self, matrix, bool_combine=False):
        super().setMatrix(matrix, bool_combine)
        self._zoomChanged()

    def resizeEvent(self, QResizeEvent):
        super().resizeEvent(QResizeEvent)
        self.sizeChanged.emit()

    def _snapRubberBandCorner(self, point:QPointF) -> QPointF:
        scne = self.scene()
        snap = scne.getSnapFlag() and self.__gfxs.isRubberBandCreationMode()
        if snap:
            return self.mapFromScene(
                wgui_snap_point(
                    self.mapToScene(point),
                    scne.getSnapX(), scne.getSnapY()))
        else:
            return point

    def _rubberBandLinePoints(self, rect:QRectF):
        rbop = self.mapToScene(self.__rbop)
        center = rect.center().toPoint()
        ex = rect.left() if rbop.x() > center.x() else rect.right()
        ey = rect.top() if rbop.y() > center.y() else rect.bottom()
        return rbop, QPointF(ex, ey)

    def mousePressEvent(self, event):
        """
        Process a mouse press event.

        :param event: the event to process
        :type event: QMouseEvent
        """
        if event.modifiers() & Qt.AltModifier != Qt.NoModifier:
            self.setDragMode(QGraphicsView.NoDrag)
            QGraphicsView.mousePressEvent(self, event)
            return

        self.__dragging = True
        if self.__gfxs.isRubberBandCreationMode():
            self.__rbop = self._snapRubberBandCorner(event.pos())
            self.__band.setGeometry(QtCore.QRect(self.__rbop, QtCore.QSize()))
            self.__band.show()
        else:
            QGraphicsView.mousePressEvent(self, event)

        if self.__gfxs.insertMode != SLCT_ITEM_ID:
            self.setDragMode(QGraphicsView.NoDrag)

    def mouseMoveEvent(self, event):
        """
        Process a mouse move event.

        :param event: the event to process
        :type event: QMouseEvent
        """
        if self.__dragging:
            self._resetSceneRect()
        # rubber band section
        if self.__band.isVisible():
            pos = self._snapRubberBandCorner(event.pos())
            self.__band.setGeometry(
                QtCore.QRect(self.__rbop, pos).normalized())
            return
            # default to graphics view event
        QGraphicsView.mouseMoveEvent(self, event)

    def mouseReleaseEvent(self, event):
        """
        Process a mouse release event

        :param event: the event to process
        :type event: QMouseEvent
        """
        self.__dragging = False
        self.setDragMode(QGraphicsView.RubberBandDrag)
        # process rubber mode
        if self.__gfxs.isRubberBandCreationMode():
            # rubber band section
            self.__band.hide()
            # map the band geometry into a scene rectangle
            rect = self.mapToScene(self.__band.geometry()).boundingRect()
            if self.__gfxs.getMediaMode() is not None:
                pos = self.mapToScene(event.pos())
                # add an object in the scene
                self.__gfxs.createInstanceItem(pos, rect=rect.normalized())
            elif self.__gfxs.insertMode == LINE_ITEM_ID:
                self.__gfxs.createLineShape(*self._rubberBandLinePoints(rect))
            elif self.__gfxs.insertMode == AROW_ITEM_ID:
                self.__gfxs.createArrowShape(*self._rubberBandLinePoints(rect))
            elif self.__gfxs.insertMode == RECT_ITEM_ID:
                self.__gfxs.createRectShape(rect)
            elif self.__gfxs.insertMode == ELPS_ITEM_ID:
                self.__gfxs.createOvalShape(rect)
            # update view port
            self.viewport().update()
            self.__gfxs.selectModeSignal.emit()
            return
            # default to graphics view event
        QGraphicsView.mouseReleaseEvent(self, event)

    def mouseDoubleClickEvent(self, event):
        """
        Process a mouse double click event
        :param event: the event to process
        """
        # noinspection PyTypeChecker
        if not self.__mouseHandler.handleDoubleClick(event, self):
            QGraphicsView.mouseDoubleClickEvent(self, event)

    def wheelEvent(self, event):
        degs = event.angleDelta().y() / 8
        stps = int(degs / 15)

        if stps != 0:
            func = self.zoomIn if stps < 0 else self.zoomOut
            for z in range(abs(stps)):
                func()

        event.accept()

    # handle a right button context menu event
    def contextMenuEvent(self, event):
        self.__cpos = event.pos()
        # noinspection PyTypeChecker
        self.__mouseHandler.handleRightClick(event, self)

    def viewCenter(self):
        """
        Provides the current center of the view in scene coordinates.

        :return: Center of the view
        :rtype: QPoint
        """
        return self.mapToScene(self.rect().center())

    def viewRectangle(self):
        """
        Provides the current view area in scene coordinates.

        :return: View area
        :rtype: QRectF
        """
        return self.mapToScene(self.viewport().geometry()).boundingRect()

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat(WGUI_INST_DND_MIME_TYPE):
            event.acceptProposedAction()

    def dragLeaveEvent(self, event):
        """
        :type event: QDragLeaveEvent
        """
        pass

    def dragMoveEvent(self, event):
        if event.mimeData().hasFormat(WGUI_INST_DND_MIME_TYPE):
            event.acceptProposedAction()

    def dropEvent(self, event):
        data = event.mimeData().data(WGUI_INST_DND_MIME_TYPE)
        if not data or data.isEmpty():
            return

        mod = self.__cctx.getAppModule()
        pos = self.mapToScene(event.pos())
        # noinspection PyArgumentList
        code = str(data, 'ASCII')
        op = mod.getOperator(code)
        if op is not None:
            self.__gfxs.createFunctionItem(pos, oper=op)
        else:
            if mod.getMimeType(code) or mod.getBlockType(code):
                self.__gfxs.createInstanceItem(pos, mime=code)

        self.__gfxs.selectModeSignal.emit()
        event.acceptProposedAction()
        return

    def keyPressEvent(self, event):
        slct = self.__gfxs.selectedItems()

        if slct and not any(i.hasFocus() for i in slct):
            dx = self.__gfxs.getSnapX() if self.__gfxs.getSnapFlag() else 1
            dy = self.__gfxs.getSnapY() if self.__gfxs.getSnapFlag() else 1
            if event.modifiers() & Qt.ShiftModifier != Qt.NoModifier:
                dx *= 10
                dy *= 10

            if event.key() == Qt.Key_Up:
                self.__gfxs.moveSelectedBy(0, -dy)
                event.accept()
            elif event.key() == Qt.Key_Right:
                self.__gfxs.moveSelectedBy(dx, 0)
                event.accept()
            elif event.key() == Qt.Key_Down:
                self.__gfxs.moveSelectedBy(0, dy)
                event.accept()
            elif event.key() == Qt.Key_Left:
                self.__gfxs.moveSelectedBy(-dx, 0)
                event.accept()
            else:
                super().keyPressEvent(event)
        else:
            super().keyPressEvent(event)
