# ----------------------------------------------------------------------------
# - wgui-scene-view-mouse-handler                                            -
# - wild cruncher: mouse event handler                        frontend / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfgui_scene_view_mouse_handler
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Mouse event handler

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod
from PyQt5.QtCore import Qt

from wcr_core.db_api.whdb_button_api import WhdbButtonP
from wcr_core.db_api.whdb_connection_api import WhdbConnectionP
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP, \
    WhdbModelInstanceP, WhdbOperatorInstanceP
from wcr_core.db_api.whdb_shape_api import WhdbShapeP
from wcr_frontend.gui.wgui_scene_item import WguiConnectionPointSceneItem, \
    WguiSceneItemP
from wcr_frontend.gui.wgui_utils import isanyinstance, getinstance
from wcr_frontend.gui.wgui_widgets import WguiEditableText


class WguiSceneViewMouseHanderP(metaclass=ABCMeta):
    """
    Mouse event handler protocol.
    """

    @abstractmethod
    def handleDoubleClick(self, event, view):
        """
        Handle a double click event.

        :param event: Double click event
        :type event: QMouseEvent
        :param view: View the event originates from
        :type view: QGraphicsView
        :return: True iff the event has been handled
        :rtype: bool
        """
        raise NotImplementedError

    @abstractmethod
    def handleRightClick(self, event, view):
        """
        Handle a right click event.

        :param event: Right click event
        :type event: QMouseEvent
        :param view: View the event originates from
        :type view: QGraphicsView
        :return: True iff the event has been handled
        :rtype: bool
        """
        raise NotImplementedError


class WguiDefaultSceneViewMouseHandler(WguiSceneViewMouseHanderP):
    def __init__(self, ctxt, propDlgFy, scene, cntxtMenuFy):
        """
        :param ctxt: Context
        :type ctxt: WcorContextP
        :param propDlgFy: Property dialog factory
        :type propDlgFy: WguiPropertiesDialogFactoryP
        :param scene: Graphics scene
        :type scene: WguiGraphicsScene
        :param cntxtMenuFy: Context menu factory
        :type cntxtMenuFy: WguiContextMenuFactoryP
        """
        self.__ctxt = ctxt
        self.__propDlgFy = propDlgFy
        self.__scene = scene
        self.__cntxtMenuFy = cntxtMenuFy

    def handleDoubleClick(self, event, view):
        # get the item at position
        items = view.items(event.pos())

        if isanyinstance(items, WguiEditableText):
            return False

        if event.modifiers() == Qt.NoModifier:
            return self._handleDoubleClick(event, view, items)
        elif event.modifiers() & Qt.ShiftModifier != Qt.NoModifier:
            return self._handleShiftDoubleClick(event, view, items)

        return False

    def handleRightClick(self, event, view):
        items = view.items(event.pos())
        spos = view.mapToScene(event.pos())
        mfac = self.__cntxtMenuFy
        if not items:
            menu = mfac.createViewContextMenu(view, spos)
        else:
            sitm = getinstance(items, WguiSceneItemP)
            if sitm is None: return

            if sitm.isSelected() and \
                    len(self.__scene.getSelectedEntities()) > 1:
                menu = mfac.createSelectionContextMenu(view, spos)
            else:
                item = sitm.getDataObject()
                if item is None: return

                citm = getinstance(items, WguiConnectionPointSceneItem)
                if citm is not None:
                    conn = citm.getDataObject()
                    if conn is None: return
                    menu = mfac.createConnectorContextMenu(
                        conn, item, view.parentWidget())
                else:
                    menu = mfac.createItemContextMenu(item, view)

        if menu is not None:
            menu.exec_(event.globalPos())

    # noinspection PyUnusedLocal
    def _handleDoubleClick(self, event, view, items):
        if not items:
            return False

        item = items[0].topLevelItem()
        inst = item.getDataObject()
        xitm = getinstance(items, WguiConnectionPointSceneItem)
        xpnt = xitm.getDataObject() if xitm else None

        pdlg = None

        if isinstance(inst, WhdbBlockInstanceP):
            pdlg = self.__propDlgFy.createBlockInstancePropertiesDialog(
                inst, selectXpnt=xpnt, parent=view.parentWidget())
        elif isinstance(inst, WhdbModelInstanceP):
            pdlg = self.__propDlgFy.createModelInstancePropertiesDialog(
                inst, selectXpnt=xpnt, parent=view.parentWidget())
        elif isinstance(inst, WhdbOperatorInstanceP):
            pdlg = self.__propDlgFy.createOperatorInstancePropertiesDialog(
                inst, selectXpnt=xpnt, parent=view.parentWidget())
        elif isinstance(inst, WhdbConnectionP):
            pdlg = self.__propDlgFy.createConnectionPropertiesDialog(
                inst, parent=view.parentWidget())
        elif isinstance(inst, WhdbShapeP):
            pdlg = self.__propDlgFy.createShapePropertiesDialog(
                inst, parent=view.parentWidget())
        elif isinstance(inst, WhdbButtonP):
            pdlg = self.__propDlgFy.createButtonPropertiesDialog(
                inst, parent=view.parentWidget())

        if pdlg:
            pdlg.show()
            pdlg.exec_()
            return True

        return False

    # noinspection PyUnusedLocal
    def _handleShiftDoubleClick(self, event, view, items):
        if not items:
            self.__scene.popHierarchy()
            return True

        item = items[0].topLevelItem()
        inst = item.getDataObject()

        if isinstance(inst, WhdbBlockInstanceP):
            self.__scene.pushHierarchy(item)
            return True
        elif isinstance(inst, WhdbModelInstanceP):
            self.__ctxt.handleFile(inst.getModel().getPath())
            return True

        return False
