# ----------------------------------------------------------------------------
# - wgui-shape                                                               -
# - wild cruncher: graphical representations of shapes                       -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 clemens oertel                                   -
# - copyright (c) 2013-2013 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_shape
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graphical shape representations

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import abstractmethod
import math

from PyQt4.QtCore import QRectF, QPointF, Qt
from PyQt4.QtGui import QGraphicsItem, QPainterPath, \
    QPainterPathStroker, QPainter, QMatrix, QFont, \
    QStyleOptionGraphicsItem, QStyle, QColor, QBrush, QPen
from wcr_core.db_api.whdb_shape_api import WHDB_LINE_STYLE, \
    WhdbLineBasedShapeP, WhdbAreaBasedShapeP

from wcr_frontend.gui.wgui_scene_item import WguiSceneItemP
from wcr_frontend.gui.wgui_protocols import ViewContext, WguiZoomAwareP, \
    QGraphicsItem_Meta_M
from wcr_frontend.gui.wgui_resizer import WguiGenericShapeResizer, \
    WguiAbsoluteResizeHandle, WguiBoundingRectResizer
from wcr_frontend.gui.wgui_utils import WGUI_SELECTION_COLOUR, \
    WGUI_HOVER_COLOUR
from wcr_frontend.gui.wgui_widgets import WguiEditableText
from silkan_common.scmn_trigonometry import scmn_getSlope, scmn_radToDeg


# noinspection PyAbstractClass
class WguiShape(WguiSceneItemP):
    """
    Base class for graphical shape representations.
    """
    #: Width of outline stroke (used for selection/hover-highlighting)
    OUTLINE_STROKE_WIDTH = 7

    #: Minimum width of shape's shape()-path (for hover/selection detection)
    MINIMUM_SHAPE_PATH_WIDTH = 5

    def __init__(self):
        pass

    @abstractmethod
    def getDataObject(self):
        """
        Returns the associated shape.

        :rtype: WhdbShapeP
        """
        raise NotImplementedError()

    def _intToLineStyle(self, code):
        """
        Convert a shape line style code to a Qt line style.
        """
        if code == WHDB_LINE_STYLE.DASHED:
            return Qt.DashLine
        elif code == WHDB_LINE_STYLE.DOTTED:
            return Qt.DotLine
        else:
            return Qt.SolidLine

    @abstractmethod
    def setSelected(self, selected):
        """
        .. seealso:: Method :py:meth:`QtGui.QWidget.setSelected`
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WguiShapeBase(QGraphicsItem, WguiShape, WguiZoomAwareP,
                    metaclass=QGraphicsItem_Meta_M):
    """
    Base class for simple graphical shapes representations.
    """

    def __init__(self, shape, vctx, parent=None, scene=None):
        """
        :param shape: Shape to be represented
        :type shape: WhdbShapeP
        :param vctx: Viewing context
        :type vctx: ViewContext
        :parent: Parent graphics item
        :scene: Scene this item will be added to
        """
        super().__init__(parent, scene)

        #: The represented shape
        self.__shape = shape
        #: Viewing context
        self.__vctx = vctx

        self.setZ(shape.getZ() or 0)

        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setAcceptHoverEvents(True)

    def getDataObject(self):
        return self.__shape

    def updateItem(self):
        self.prepareGeometryChange()
        self._update()

    def getZ(self):
        return self.zValue()

    def setZValue(self, zval):
        self.setZ(zval)

    def setZ(self, zval):
        QGraphicsItem.setZValue(self, zval)
        self.__shape.setZ(zval)

    @abstractmethod
    def _update(self):
        """
        Update the graphical representation to reflect the latest version of
        the shape.
        """
        pass

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            x, y = self.__vctx.snapPoint(value.x(), value.y())
            value = QPointF(x, y)
        elif change == QGraphicsItem.ItemPositionHasChanged:
            self.__shape.location = (value.x(), value.y())
        return value

    def hoverEnterEvent(self, event):
        self.update()

    def hoverLeaveEvent(self, event):
        self.update()

    def _snap(self, point:(float, float)) -> (float, float):
        return self.__vctx.snapPoint(*point)

    @property
    def _shape(self):
        return self.__shape


class WguiLineShapeBase(WguiShapeBase, WguiZoomAwareP):
    """
    Base class for simple graphical shapes representations, such as lines or
    rectangles.
    """

    def __init__(self, shape, vctx, parent=None, scene=None):
        """
        :param shape: Shape to be represented
        :type shape: WhdbShapeP
        :param vctx: Viewing context
        :type vctx: ViewContext
        :parent: Parent graphics item
        :scene: Scene this item will be added to
        """
        super().__init__(shape, vctx, parent, scene)

        #: The shape's path (updated whenever shape changes)
        self.__path = None
        #: The shape's outline stroke (updated whenever shape changes)
        self.__outlineStroke = None
        #: The shape's shape()-stroke (updated whenever shape changes)
        self.__shapeStroke = None
        #: The shape's bounding box (updated whenever shape changes)
        self.__boundingBox = None
        #: Current zoom level
        self.__zoom = 1.0

    def zoomChanged(self, zoom:float):
        self.__zoom = zoom
        self._update()

    def _mkStroke(self, width:float):
        if self._path is None: return None
        stroker = QPainterPathStroker()
        stroker.setCapStyle(Qt.RoundCap)
        stroker.setJoinStyle(Qt.MiterJoin)
        stroker.setWidth(self._effectiveLineWidth + width)
        return stroker.createStroke(self._path)

    def __mkStrokes(self):
        """
        Create the outline and shape strokes for the associated shape using the
        given path.

        :return: Outline and shape strokes
        :rtype: (QPainterPath, QPainterPath)
        """
        shape = self._shape
        path = self._path
        stroker = QPainterPathStroker()
        stroker.setCapStyle(Qt.RoundCap)
        fullWidthPath = path
        shapeStroke = path
        if isinstance(shape, WhdbLineBasedShapeP):
            lineWidth = self._effectiveLineWidth
            # Create widened path iff line has width greater than 1
            if lineWidth > 1:
                stroker.setJoinStyle(Qt.MiterJoin)
                stroker.setWidth(lineWidth)
                fullWidthPath = stroker.createStroke(path)
            minShapePathWidth = self.MINIMUM_SHAPE_PATH_WIDTH / self.__zoom
            # Create stroke for shape()
            if lineWidth > minShapePathWidth:
                shapeStroke = fullWidthPath
            else:
                stroker.setJoinStyle(Qt.RoundJoin)
                stroker.setWidth(minShapePathWidth)
                shapeStroke = stroker.createStroke(path)
        # Create stroke for outline highlighting
        stroker.setJoinStyle(Qt.RoundJoin)
        stroker.setWidth(WguiShape.OUTLINE_STROKE_WIDTH / self.__zoom)
        outlineStroke = stroker.createStroke(fullWidthPath)
        if isinstance(shape, WhdbAreaBasedShapeP):
            outlineStroke -= fullWidthPath
            outlineStroke -= path
        return outlineStroke, shapeStroke

    def _update(self):
        self.prepareGeometryChange()
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, False)
        self.setPos(*self._shape.location)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)

        self.__path = self._getItemPath()
        self.__outlineStroke, shapeStroke = self.__mkStrokes()
        self.__boundingBox = self.__outlineStroke.boundingRect()
        if self._isOpaque():
            path = QPainterPath()
            path.addRect(self.__boundingBox)
            self.__shapeStroke = self.__path
        else:
            self.__shapeStroke = shapeStroke
        self.update()

    @abstractmethod
    def _getItemPath(self) -> QPainterPath:
        pass

    def _paintItem(self, painter:QPainter):
        """
        Paint the shape.

        :param painter: Painter to use for painting
        :type painter: QPainter
        """
        if self.__path is None: return
        shape = self._shape
        if isinstance(shape, WhdbAreaBasedShapeP):
            if shape.fillColour is not None:
                colour = QColor(shape.fillColour)
                colour.setAlphaF(shape.fillOpacity)
                brush = QBrush(colour)
                painter.fillPath(self.__path, brush)
        if isinstance(shape, WhdbLineBasedShapeP):
            lineWidth = self._effectiveLineWidth
            if lineWidth > 0:
                pen = QPen(QBrush(QColor(shape.lineColour)),
                           lineWidth,
                           self._intToLineStyle(shape.lineStyle),
                           Qt.RoundCap,
                           Qt.MiterJoin)
                painter.setPen(pen)
                painter.drawPath(self.__path)

    def paint(self, painter:QPainter, options:QStyleOptionGraphicsItem,
              widget=None):
        if self.__outlineStroke is None: return
        painter.save()
        painter.setRenderHint(QPainter.Antialiasing, True)
        # noinspection PyTypeChecker
        if options.state & QStyle.State_Selected:
            painter.fillPath(self.__outlineStroke, WGUI_SELECTION_COLOUR)
        elif options.state & QStyle.State_MouseOver:
            painter.fillPath(self.__outlineStroke, WGUI_HOVER_COLOUR)
        options.state &= ~(QStyle.State_Selected | QStyle.State_HasFocus)
        self._paintItem(painter)
        painter.restore()

    @property
    def _effectiveLineWidth(self):
        shape = self._shape
        if isinstance(shape, WhdbLineBasedShapeP):
            return shape.lineWidth
        else:
            return 0

    def boundingRect(self):
        if self.__boundingBox is not None:
            return self.__boundingBox
        else:
            return QRectF()

    def _isOpaque(self) -> bool:
        """
        Whether the shape has spatial extend and is opaque.

        :return: True iff the shape is area based and has a filling
        :rtype: bool
        """
        shape = self._shape
        if isinstance(shape, WhdbAreaBasedShapeP):
            return shape.fillColour is not None and shape.fillOpacity > 0.0
        else:
            return False

    def shape(self):
        if self.__shapeStroke is not None:
            return self.__shapeStroke
        else:
            return QPainterPath()

    @property
    def _path(self):
        return self.__path


class WguiLine(WguiLineShapeBase):
    """
    Graphical representation of a line shape.
    """

    def __init__(self, line, vctx, parent=None, scene=None):
        """
        :param line: Line shape to represent
        :type line: WhdbLineShape
        :param vctx: Viewing context
        :type vctx: ViewContext
        :parent: Parent graphics item
        :scene: Scene this item will be added to
        """
        super().__init__(line, vctx, parent, scene)

        #: Line shape represented by this graphics item
        self.__line = line

        # noinspection PyTypeChecker
        self.__resizer = WguiGenericShapeResizer(self._mkStroke, self, scene)
        for handle in self.__createResizingHandles(line):
            self.__resizer.addHandle(handle)

        self._update()
        self.__resizer.shapeChanged()

    def __createResizingHandles(self, line):
        def setPointA(point:(float, float)):
            line.pointA = self._snap(point)
            self._update()

        def setPointB(point:(float, float)):
            line.pointB = self._snap(point)
            self._update()

        return [WguiAbsoluteResizeHandle(lambda: line.pointA, setPointA),
                WguiAbsoluteResizeHandle(lambda: line.pointB, setPointB)]

    def _getItemPath(self):
        x1, y1 = self.__line.pointARel
        x2, y2 = self.__line.pointBRel

        # Update outline stroke
        path = QPainterPath()
        path.moveTo(x1, y1)
        path.lineTo(x2, y2)
        return path


class WguiArrow(WguiLineShapeBase):
    """
    Graphical representation of an arrow shape.
    """

    def __init__(self, arrow, vctx, parent=None, scene=None):
        """
        :param arrow: Arrow shape to represent
        :type arrow: WhdbArrowShape
        :param vctx: Viewing context
        :type vctx: ViewContext
        :parent: Parent graphics item
        :scene: Scene this item will be added to
        """
        super().__init__(arrow, vctx, parent, scene)

        #: Arrow shape represented by this graphics item
        self.__arrow = arrow

        # noinspection PyTypeChecker
        self.__resizer = WguiGenericShapeResizer(self._mkStroke, self, scene)
        for handle in self.__createResizingHandles(arrow):
            self.__resizer.addHandle(handle)

        self._update()
        self.__resizer.shapeChanged()

    def __createResizingHandles(self, arrow):
        def setTip(point:(float, float)):
            arrow.tip = self._snap(point)
            self._update()

        def setEnd(point:(float, float)):
            arrow.endPoint = self._snap(point)
            self._update()

        return [WguiAbsoluteResizeHandle(lambda: arrow.tip, setTip),
                WguiAbsoluteResizeHandle(lambda: arrow.endPoint, setEnd)]

    def _getItemPath(self):
        x1, y1 = self.__arrow.tipRel
        x2, y2 = self.__arrow.endPointRel
        w = x1 - x2
        h = y1 - y2
        l = math.sqrt(math.pow(w, 2) + math.pow(h, 2))
        lh = l / 2
        tx = lh - self.__arrow.lineWidth / 2
        headLength = max(self.__arrow.lineWidth * 4, 20)

        # Update outline stroke & arrow shape
        path = QPainterPath()
        path.moveTo(-tx + headLength, -headLength / 2)
        path.lineTo(-tx, 0)
        path.lineTo(-tx + headLength, headLength / 2)
        path.moveTo(-tx, 0)
        path.lineTo(lh, 0)
        matrix = QMatrix().rotate(
            scmn_radToDeg(scmn_getSlope(w, h)))
        return matrix.map(path)


class WguiRectangle(WguiLineShapeBase):
    """
    Graphical representation of a rectangle shape.
    """

    def __init__(self, rectangle, vctx, parent=None, scene=None):
        """
        :param rectangle: Rectangle shape to represent
        :type rectangle: WhdbRectangleShape
        :param vctx: Viewing context
        :type vctx: ViewContext
        :parent: Parent graphics item
        :scene: Scene this item will be added to
        """
        super().__init__(rectangle, vctx, parent, scene)

        #: Rectangle shape represented by this graphics item
        self.__rectangle = rectangle

        # noinspection PyTypeChecker
        self.__resizer = \
            WguiBoundingRectResizer(
                self.__getResizingRectAndPath, self.__fitToRect, self._snap,
                parent=self, scene=scene)

        self._update()
        self.__resizer.shapeChanged()

    def __fitToRect(self, rect:((float, float), (float, float))):
        (x1, y1), (x2, y2) = rect
        lw = self._effectiveLineWidth
        lwh = lw // 2
        self.__rectangle.corner = (min(x1, x2) + lwh, min(y1, y2) + lwh)
        self.__rectangle.width = max(abs(x1 - x2) - lw + 1, 0)
        self.__rectangle.height = max(abs(y1 - y2) - lw + 1, 0)
        self._update()

    def __getResizingRectAndPath(self):
        x, y = self.__rectangle.corner
        w = self.__rectangle.width
        h = self.__rectangle.height
        lw = self._effectiveLineWidth
        lwh = lw // 2
        return QRectF(x - lwh, y - lwh, w + lw - 1, h + lw - 1), self._path

    def _getItemPath(self):
        self.prepareGeometryChange()
        self.setPos(*self.__rectangle.location)

        w = self.__rectangle.width
        h = self.__rectangle.height
        path = QPainterPath()
        rect = QRectF(-w / 2, -h / 2, w, h)
        path.addRect(rect)
        return path

    @property
    def _effectiveLineWidth(self):
        rect = self.__rectangle
        return max(min(rect.width, rect.height, rect.lineWidth),
                   min(rect.lineWidth, 1))


class WguiOval(WguiLineShapeBase):
    """
    Graphical representation of an oval shape.
    """

    def __init__(self, oval, vctx, parent=None, scene=None):
        """
        :param oval: Oval shape to represent
        :type oval: WhdbOvalShape
        :param vctx: Viewing context
        :type vctx: ViewContext
        :parent: Parent graphics item
        :scene: Scene this item will be added to
        """
        super().__init__(oval, vctx, parent, scene)

        #: Oval shape represented by this graphics item
        self.__oval = oval

        # noinspection PyTypeChecker
        self.__resizer = \
            WguiBoundingRectResizer(
                self.__getResizingRectAndPath, self.__fitToRect, self._snap,
                drawRect=True, parent=self, scene=scene)

        self._update()
        self.__resizer.shapeChanged()

    def __fitToRect(self, rect:((float, float), (float, float))):
        (x1, y1), (x2, y2) = rect
        lw = self._effectiveLineWidth
        self.__oval.center = (x1 + (x2 - x1) / 2, y1 + (y2 - y1) / 2)
        self.__oval.width = max(abs(x1 - x2) - lw + 1, 0)
        self.__oval.height = max(abs(y1 - y2) - lw + 1, 0)
        self._update()

    def __getResizingRectAndPath(self):
        x, y = self.__oval.center
        w = self.__oval.width
        h = self.__oval.height
        lw = self._effectiveLineWidth
        return QRectF(x - (w / 2) - (lw // 2), y - (h / 2) - (lw // 2),
                      w + lw - 1, h + lw - 1), \
               self._path

    def _getItemPath(self):
        w = self.__oval.width
        h = self.__oval.height
        path = QPainterPath()
        path.addEllipse(-w / 2, -h / 2, w, h)
        return path

    @property
    def _effectiveLineWidth(self):
        oval = self.__oval
        return max(min(oval.width, oval.height, oval.lineWidth),
                   min(oval.lineWidth, 1))


# noinspection PyUnresolvedReferences
class WguiText(WguiEditableText, WguiShape, WguiZoomAwareP,
               metaclass=QGraphicsItem_Meta_M):
    """
    Graphical representation of a text shape.
    """
    #: Placeholder text
    PLACEHOLDER_TEXT = 'Double click to edit'

    #: Placeholder text when editing
    EDIT_PHLDR_TEXT = 'Type to edit'

    #: Radius for corner rounding of outline box
    RECT_CORNER_ROUNDING = 10

    def __init__(self, text, vctx, parent=None, scene=None):
        """
        :param text: Text shape to represent
        :type text: WhdbTextShape
        :param vctx: Viewing context
        :type vctx: ViewContext
        :parent: Parent graphics item
        :scene: Scene this item will be added to
        """
        WguiEditableText.__init__(self, placeHolder=self.PLACEHOLDER_TEXT,
                                  editingPlaceHolder=self.EDIT_PHLDR_TEXT,
                                  multiline=True, parent=parent, scene=scene)
        WguiShape.__init__(self)

        #: Text shape represented by this item
        self.__text = text
        #: View context
        self.__vctx = vctx
        #: The shape's outline stroke (updated whenever shape changes)
        self.__outlineStroke = None
        #: Current zoom level
        self.__zoom = 1.0

        self.setZ(text.getZ() or 0)

        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges)
        self.setAcceptHoverEvents(True)

        self._update()

    def updateBoundingBox(self):
        """
        Update the associated text shape's bounding box.
        """
        brect = self.mapRectToScene(self.boundingRect())
        self.__text.setBoundingBox(brect.x(), brect.y(),
                                   brect.width(), brect.height())

    def zoomChanged(self, zoom:float):
        self.prepareGeometryChange()
        self.__zoom = zoom
        self.__outlineStroke = self._mkOutlineStroke(self._mkOutlinePath())

    def getDataObject(self):
        return self.__text

    def updateItem(self):
        self.prepareGeometryChange()
        self._update()
        self.__outlineStroke = self._mkOutlineStroke(self._mkOutlinePath())
        self.update()

    def getZ(self):
        return self.zValue()

    def setZValue(self, zval):
        self.setZ(zval)

    def setZ(self, zval):
        QGraphicsItem.setZValue(self, zval)
        self.__text.setZ(zval)

    def _mkOutlinePath(self) -> QPainterPath:
        """
        Generate an outline path for the text shape, to be used for
        selection/hover-highlighting.

        :return: Outline path
        :rtype: QPainterPath
        """
        path = QPainterPath()
        rect = super().boundingRect()
        lWidth = self.__text.lineWidth
        if lWidth > 0:
            rect = rect.adjusted(-lWidth, -lWidth, lWidth, lWidth)
        rounding = self.RECT_CORNER_ROUNDING + lWidth / 2
        path.addRoundedRect(rect, rounding, rounding)
        return path

    def _mkOutlineStroke(self, path:QPainterPath) -> QPainterPath:
        """
        Generate an outline stroke for the text shape, to be used for
        selection/hover-highlighting.

        :param path: Outline path to generate a stroke for
        :type: QPainterPath
        :return: Outline stroke
        :rtype: QPainterPath
        """
        strokePath = QPainterPath(path)
        strokePath.setFillRule(Qt.WindingFill)
        stroker = QPainterPathStroker()
        stroker.setCapStyle(Qt.RoundCap)
        stroker.setJoinStyle(Qt.RoundJoin)
        stroker.setWidth(WguiShape.OUTLINE_STROKE_WIDTH / self.__zoom)
        shapeStroke = stroker.createStroke(strokePath)
        return shapeStroke.subtracted(strokePath)

    def _contentChanged(self):
        self.__text.text = self.text
        self.__outlineStroke = self._mkOutlineStroke(self._mkOutlinePath())

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            # Adjust move to scene's snapping settings
            x, y = self.__vctx.snapPoint(value.x(), value.y())
            value = QPointF(x, y)
        elif change == QGraphicsItem.ItemPositionHasChanged:
            self.__text.location = (self.x(), self.y())
            self.updateBoundingBox()
        return super().itemChange(change, value)

    def paint(self, painter:QPainter, options:QStyleOptionGraphicsItem,
              widget=None):
        rounding = self.RECT_CORNER_ROUNDING
        painter.save()
        painter.setRenderHint(QPainter.Antialiasing, True)
        # noinspection PyTypeChecker
        if options.state & QStyle.State_Selected or self.isEditing:
            painter.fillPath(self.__outlineStroke, WGUI_SELECTION_COLOUR)
        elif options.state & QStyle.State_MouseOver:
            painter.fillPath(self.__outlineStroke, WGUI_HOVER_COLOUR)

        fill = self.__text.fillColour is not None
        line = self.__text.lineWidth > 0
        if fill or line:
            # Enlarge outline to not cover text
            radj = self.__text.lineWidth / 2 + 0.5
            rect = super().boundingRect().adjusted(-radj, -radj, radj, radj)
            if fill:
                # Fill background
                colour = QColor(self.__text.fillColour)
                colour.setAlphaF(self.__text.fillOpacity)
                brush = QBrush(colour)
                painter.setBrush(brush)
                painter.setPen(Qt.NoPen)
                painter.drawRoundedRect(rect, rounding, rounding)
                painter.setBrush(Qt.NoBrush)
            if line:
                # Draw outline
                pen = QPen(QBrush(QColor(self.__text.lineColour)),
                           self.__text.lineWidth,
                           self._intToLineStyle(self.__text.lineStyle))
                painter.setPen(pen)
                painter.drawRoundedRect(rect, rounding, rounding)

        painter.restore()

    def _update(self):
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, False)
        self.setPos(*self.__text.location)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)

        font = QFont()
        font.setPixelSize(self.__text.fontSize)
        self.font = font
        self.textColour = QColor(self.__text.textColour)
        self.horizontalAlign = self.__text.horizontalAlignment
        self.verticalAlign = self.__text.verticalAlignment

        self.text = self.__text.text

        self.updateBoundingBox()

    def boundingRect(self) -> QRectF:
        adj = self.__text.lineWidth + math.ceil(self.OUTLINE_STROKE_WIDTH / 2)
        return super().boundingRect().adjusted(-adj, -adj, adj, adj)

    def shape(self) -> QRectF:
        path = QPainterPath()
        path.addRect(self.boundingRect())
        return path
