# ----------------------------------------------------------------------------
# - wgui-shape-props-dialog                                                  -
# - wild cruncher: shape properties dialog                                   -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 clemens oertel                                   -
# - copyright (c) 2013-2013 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_shape_props_dialog
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher shape properties dialog

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QComboBox, QDoubleSpinBox, QPushButton, QLabel, \
	QColorDialog, QGridLayout
from PyQt5.QtGui import QColor

from wcr_core.db_api.whdb_shape_api import WhdbAreaBasedShapeP, \
    WhdbLineBasedShapeP, WhdbTextBasedShapeP, WHDB_LINE_STYLE
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP, \
    WguiPropertiesDialog
from wcr_frontend.gui.wgui_utils import wgui_qcb_set_data, wgui_qcb_get_data
from wcr_core.db_api.whdb_protocols_api import WHDB_HALIGN_LEFT, \
    WHDB_HALIGN_CENTER, WHDB_HALIGN_RIGHT, WHDB_VALIGN_TOP, \
    WHDB_VALIGN_MIDDLE, WHDB_VALIGN_BOTTOM


class WguiShapePropertiesDialog(WguiPropertiesDialog):
    """
    Properties dialog for shapes.
    """

    def __init__(self, shape, updateF, parent=None):
        """
        :param shape: Shape to display/change properties for
        :type shape: WhdbShapeP
        :param updateF: Callback function to notify about saves
        :type updateF: _ -> _
        :param parent: Parent widget
        """
        super().__init__('Shape Properties', updateF, parent=parent)

        widgets = []

        if isinstance(shape, WhdbTextBasedShapeP):
            widgets.append(('Text Properties',
                            _WguiTextProperties(shape)))
        if isinstance(shape, WhdbLineBasedShapeP):
            widgets.append(('Line Properties',
                            _WguiLineProperties(shape)))
        if isinstance(shape, WhdbAreaBasedShapeP):
            widgets.append(('Area Properties',
                            _WguiAreaProperties(shape)))

        self.setGroupedPropertyWidgets(*widgets)


class _WguiLineProperties(WguiPropertiesWidgetP):
    """
    Property editing widget for a line-based shape.
    """

    def __init__(self, line):
        """
        :param line: Line-based shape object
        :type line: WhdbLineBasedShapeP
        """
        super().__init__()

        #: Line-based shape object
        self.__line = line

        #: Line width field
        self.__lineWidthF = QDoubleSpinBox()
        #: Line colour field
        self.__lineColourF = WguiColorLabel()
        #: Line style field
        self.__lineStyleF = QComboBox()

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QLayout
        """
        mainLayout = QGridLayout()
        mainLayout.setColumnStretch(1, 1)
        row = -1

        row += 1
        label = QLabel('Line Width:')
        label.setBuddy(self.__lineWidthF)
        reset = QPushButton('Reset')
        reset.clicked.connect(self._resetLineWidth)
        mainLayout.addWidget(label, row, 0)
        mainLayout.addWidget(self.__lineWidthF, row, 1)
        mainLayout.addWidget(reset, row, 2)

        row += 1
        self.__lineColourF.clicked.connect(self._showLineColourDialog)
        label = QLabel('Line Color:')
        label.setBuddy(self.__lineColourF)
        reset = QPushButton('Reset')
        reset.clicked.connect(self._resetLineColour)
        mainLayout.addWidget(label, row, 0)
        mainLayout.addWidget(self.__lineColourF, row, 1)
        mainLayout.addWidget(reset, row, 2)

        row += 1
        for name, style in [('Solid', WHDB_LINE_STYLE.SOLID),
                            ('Dotted', WHDB_LINE_STYLE.DOTTED),
                            ('Dashed', WHDB_LINE_STYLE.DASHED)]:
            self.__lineStyleF.addItem(name, style)
        label = QLabel('Line Style:')
        label.setBuddy(self.__lineStyleF)
        reset = QPushButton('Reset')
        reset.clicked.connect(self._resetLineStyle)
        mainLayout.addWidget(label, row, 0)
        mainLayout.addWidget(self.__lineStyleF, row, 1)
        mainLayout.addWidget(reset, row, 2)

        return mainLayout

    def updateFields(self):
        line = self.__line
        self.__lineWidthF.setValue(line.lineWidth)
        self.__lineColourF.setColour(line.lineColour)
        wgui_qcb_set_data(line.lineStyle, self.__lineStyleF)

    def _resetLineWidth(self):
        self.__lineWidthF.setValue(WhdbLineBasedShapeP.DEFAULT_LINE_WIDTH)

    def _showLineColourDialog(self):
        color = QColorDialog.getColor(QColor(self.__line.lineColour),
                                      self, 'Choose Line Color')
        self.__lineColourF.setColour(color.name())

    def _resetLineColour(self):
        self.__lineColourF.setColour(WhdbLineBasedShapeP.DEFAULT_LINE_COLOUR)

    def _resetLineStyle(self):
        wgui_qcb_set_data(WhdbLineBasedShapeP.DEFAULT_LINE_STYLE,
                          self.__lineStyleF)

    def save(self):
        line = self.__line
        line.lineWidth = self.__lineWidthF.value()
        line.lineColour = self.__lineColourF.colour()
        line.lineStyle = wgui_qcb_get_data(self.__lineStyleF)


class _WguiAreaProperties(WguiPropertiesWidgetP):
    """
    Property editing widget for an area-based shape.
    """

    def __init__(self, area):
        """
        :param area: Area-based shape object
        :type area: WhdbAreaBasedShapeP
        """
        super().__init__()

        #: Area-based shape object
        self.__area = area

        #: Fill colour field
        self.__fillColourF = WguiColorLabel()
        #: Fill opacity field
        self.__fillOpacityF = QDoubleSpinBox()

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QLayout
        """
        mainLayout = QGridLayout()
        mainLayout.setColumnStretch(1, 1)
        row = -1

        row += 1
        colourF = self.__fillColourF
        colourF.clicked.connect(self._showFillColourDialog)
        label = QLabel('Fill Color:')
        label.setBuddy(colourF)
        reset = QPushButton('Reset')
        reset.clicked.connect(self._resetFillColour)
        mainLayout.addWidget(label, row, 0)
        mainLayout.addWidget(colourF, row, 1)
        mainLayout.addWidget(reset, row, 2)

        row += 1
        opacityF = self.__fillOpacityF
        opacityF.setMinimum(0.0)
        opacityF.setMaximum(1.0)
        opacityF.setSingleStep(0.1)
        label = QLabel('Fill Opacity:')
        label.setBuddy(opacityF)
        reset = QPushButton('Reset')
        reset.clicked.connect(self._resetFillOpacity)
        mainLayout.addWidget(label, row, 0)
        mainLayout.addWidget(opacityF, row, 1)
        mainLayout.addWidget(reset, row, 2)

        return mainLayout

    def updateFields(self):
        area = self.__area
        self.__fillColourF.setColour(area.fillColour)
        self.__fillOpacityF.setValue(area.fillOpacity)

    def _showFillColourDialog(self):
        color = QColorDialog.getColor(QColor(self.__area.fillColour),
                                      self, 'Choose Fill Color')
        self.__fillColourF.setColour(color.name())

    def _resetFillColour(self):
        self.__fillColourF.setColour(WguiColorLabel.NO_COLOUR)

    def _resetFillOpacity(self):
        self.__fillOpacityF.setValue(WhdbAreaBasedShapeP.DEFAULT_FILL_OPACITY)

    def save(self):
        area = self.__area
        area.fillColour = self.__fillColourF.colour()
        area.fillOpacity = self.__fillOpacityF.value()


class _WguiTextProperties(WguiPropertiesWidgetP):
    """
    Property editing widget for a text-based shape.
    """

    def __init__(self, text):
        """
        :param text: Text-based shape object
        :type text: WhdbTextBasedShapeP
        """
        super().__init__()

        #: Text-based shape object
        self.__text = text

        #: Text colour field
        self.__textColourF = WguiColorLabel(text.textColour)
        #: Font size field
        self.__fontSizeF = QDoubleSpinBox()
        #: Horizontal alignment field
        self.__hAlignmentF = QComboBox()
        #: Vertical alignment field
        self.__vAlignmentF = QComboBox()

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QLayout
        """
        mainLayout = QGridLayout()
        mainLayout.setColumnStretch(1, 1)
        row = -1

        row += 1
        colourF = self.__textColourF
        colourF.clicked.connect(self._showTextColourDialog)
        label = QLabel('Text Color:')
        label.setBuddy(colourF)
        reset = QPushButton('Reset')
        reset.clicked.connect(self._resetTextColour)
        mainLayout.addWidget(label, row, 0)
        mainLayout.addWidget(colourF, row, 1)
        mainLayout.addWidget(reset, row, 2)

        row += 1
        fontSizeF = self.__fontSizeF
        fontSizeF.setMinimum(6.0)
        fontSizeF.setMaximum(48.0)
        label = QLabel('Font Size:')
        label.setBuddy(fontSizeF)
        reset = QPushButton('Reset')
        reset.clicked.connect(self._resetFontSize)
        mainLayout.addWidget(label, row, 0)
        mainLayout.addWidget(fontSizeF, row, 1)
        mainLayout.addWidget(reset, row, 2)

        row += 1
        hAlignmentF = self.__hAlignmentF
        hAlignmentF.addItem('Left', WHDB_HALIGN_LEFT)
        hAlignmentF.addItem('Center', WHDB_HALIGN_CENTER)
        hAlignmentF.addItem('Right', WHDB_HALIGN_RIGHT)
        label = QLabel('Horizontal Alignment:')
        label.setBuddy(fontSizeF)
        reset = QPushButton('Reset')
        reset.clicked.connect(self._resetHorizontalAlignment)
        mainLayout.addWidget(label, row, 0)
        mainLayout.addWidget(hAlignmentF, row, 1)
        mainLayout.addWidget(reset, row, 2)

        row += 1
        vAlignmentF = self.__vAlignmentF
        vAlignmentF.addItem('Top', WHDB_VALIGN_TOP)
        vAlignmentF.addItem('Middle', WHDB_VALIGN_MIDDLE)
        vAlignmentF.addItem('Bottom', WHDB_VALIGN_BOTTOM)
        label = QLabel('Vertical Alignment:')
        label.setBuddy(fontSizeF)
        reset = QPushButton('Reset')
        reset.clicked.connect(self._resetVerticalAlignment)
        mainLayout.addWidget(label, row, 0)
        mainLayout.addWidget(vAlignmentF, row, 1)
        mainLayout.addWidget(reset, row, 2)

        return mainLayout

    def updateFields(self):
        text = self.__text
        self.__textColourF.setColour(text.textColour)
        self.__fontSizeF.setValue(text.fontSize)
        wgui_qcb_set_data(text.horizontalAlignment, self.__hAlignmentF)
        wgui_qcb_set_data(text.verticalAlignment, self.__vAlignmentF)

    def _showTextColourDialog(self):
        color = QColorDialog.getColor(QColor(self.__text.textColour),
                                      self, 'Choose Text Color')
        self.__textColourF.setColour(color.name())

    def _resetTextColour(self):
        self.__textColourF.setColour(WhdbTextBasedShapeP.DEFAULT_TEXT_COLOUR)

    def _resetFontSize(self):
        self.__fontSizeF.setValue(WhdbTextBasedShapeP.DEFAULT_TEXT_FONT_SIZE)

    def _resetHorizontalAlignment(self):
        wgui_qcb_set_data(WhdbTextBasedShapeP.DEFAULT_TEXT_H_ALIGNMENT,
                          self.__hAlignmentF)

    def _resetVerticalAlignment(self):
        wgui_qcb_set_data(WhdbTextBasedShapeP.DEFAULT_TEXT_V_ALIGNMENT,
                          self.__vAlignmentF)

    def save(self):
        text = self.__text
        text.textColour = self.__textColourF.colour()
        text.fontSize = self.__fontSizeF.value()
        text.horizontalAlignment = wgui_qcb_get_data(self.__hAlignmentF)
        text.verticalAlignment = wgui_qcb_get_data(self.__vAlignmentF)


class WguiClickableLabel(QLabel):
    """
    Label that emits a signal when clicked.
    """

    #: Signal emitted whenever the label is clicked.
    clicked = pyqtSignal()

    def __init__(self, text:str=None):
        """
        :param text: Initially displayed text
        :type text: str
        """
        super().__init__(text)

    def mouseReleaseEvent(self, QMouseEvent):
        self.clicked.emit()


class WguiColorLabel(WguiClickableLabel):
    """
    Clickable label with a settable background colour.
    """

    NO_COLOUR = ''

    def __init__(self, colour=None, text=None):
        """
        :param colour: Initial background colour
        :type colour: str
        :param text: Initially displayed text
        :type text: str
        """
        super().__init__(text)

        self.__colour = colour

        self.setColour(colour)

    def colour(self):
        """
        Get the current background colour.

        :return: Current background colour
        :rtype: str
        """
        return self.__colour

    def setColour(self, colour):
        """
        Set the current background colour.

        :param colour: New background colour
        :type colour: str
        """
        self.__colour = colour
        if colour:
            self.setStyleSheet('QLabel {background-color:%s}' % colour)
        else:
            self.setStyleSheet('QLabel {}')
