# ----------------------------------------------------------------------------
# - wgui-text-editor                                                         -
# - wild cruncher - cockpit : text editor window                             -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_text_editor
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher text editor

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from PyQt5.QtCore import QFile, Qt, QRegExp
from PyQt5.QtWidgets import QMessageBox, QTextEdit, QHBoxLayout, QVBoxLayout
from PyQt5.QtGui import QSyntaxHighlighter, QFont, QTextCharFormat

from wcr_frontend.gui.wgui_editor import WguiEditorHasSaveP, WguiEditorBase, \
    WguiEditorHasSearchP
from wcr_frontend.gui.wgui_utils import wgui_print_text, wgui_getSaveFile
from wcr_frontend.gui.wgui_widgets import WguiTextSearchWidget
from silkan_common.scmn_mime_type import SCMN_MIME_TYPE_C, \
    SCMN_MIME_TYPE_MAKEFILE, SCMN_MIMES, SCMN_PAYLOAD_TEXT
from silkan_common.scmn_file_path import scmn_path_extension, scmn_path_base


# ----------------------------------------------------------------------------
# - class section                                                            -
# ----------------------------------------------------------------------------

class _WguiCSyntaxHighlighter(QSyntaxHighlighter):
    def __init__(self, document):
        """
        :param document: Parent document
        """
        super().__init__(document)

        self.__rules = []

        defFmt = QTextCharFormat()
        defFmt.setForeground(Qt.darkMagenta)
        self.__rules.append((QRegExp('_*[A-Z][A-Z0-9_]*'), defFmt))

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.darkBlue)
        fmt.setFontWeight(QFont.Bold)
        pat = ['\\bchar\\b', '\\bclass\\b', '\\bconst\\b', '\\bdouble\\b',
               '\\benum\\b', '\\bexplicit\\b', '\\binline\\b', '\\bint\\b',
               '\\blong\\b', '\\bshort\\b', '\\bsigned\\b', '\\bstatic\\b',
               '\\bstruct\\b', '\\btypedef\\b', '\\btypename\\b',
               '\\bunion\\b', '\\bunsigned\\b', '\\bvoid\\b',
               '\\bvolatile\\b']
        self.__rules.extend([(QRegExp(pat), fmt) for pat in pat])

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.blue)
        fmt.setFontItalic(True)
        self.__rules.append((QRegExp('\\b[A-Za-z0-9_]+\\s*(?=\\()'), fmt))

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.darkBlue)
        fmt.setFontItalic(True)
        pat = ['\\bif\\b', '\\bwhen\\b', '\\bfor\\b', '\\bwhile\\b',
               '\\bdo\\b', '\\belse\\b', '\\bcase\\b', '\\bdefault\\b',
               '\\bswitch\\b']
        self.__rules.extend([(QRegExp(pat), fmt) for pat in pat])

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.magenta)
        fmt.setFontWeight(QFont.Bold)
        pat = ['\\breturn\\b', '\\bgoto\\b', '\\bbreak\\b', '\\bcontinue\\b']
        self.__rules.extend([(QRegExp(pat), fmt) for pat in pat])

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.darkGreen)
        fmt.setFontItalic(True)
        self.__rules.append((QRegExp('\".*\"'), fmt))

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.gray)
        self.__rules.append((QRegExp('^#[^\n]*'), fmt))
        self.__rules.append((QRegExp('^#define\\s+[^\\s]+'), defFmt))
        self.__rules.append((QRegExp('^#define\\s*'), fmt))

        self.__comment = QTextCharFormat()
        self.__comment.setForeground(Qt.red)
        self.__rules.append((QRegExp('//[^\n]*'), self.__comment))
        self.__commentStart = QRegExp('/\\*')
        self.__commentEnd = QRegExp('\\*/')

    def highlightBlock(self, text):
        for pat, fmt in self.__rules:
            expr = QRegExp(pat)
            tidx = expr.indexIn(text)
            while tidx >= 0:
                tlen = expr.matchedLength()
                self.setFormat(tidx, tlen, fmt)
                tidx = expr.indexIn(text, tidx + tlen)
        self.setCurrentBlockState(0)
        sidx = 0
        if self.previousBlockState() != 1:
            sidx = self.__commentStart.indexIn(text)
        while sidx >= 0:
            eidx = self.__commentEnd.indexIn(text, sidx)
            if eidx == -1:
                self.setCurrentBlockState(1)
                clen = text.length() - sidx
            else:
                clen = eidx - sidx + self.__commentEnd.matchedLength()
            self.setFormat(sidx, eidx, self.__comment)
            sidx = self.__commentStart.indexIn(text, sidx + clen)


class _WguiMakefileSyntaxHighlighter(QSyntaxHighlighter):
    def __init__(self, document):
        """
        :param document: Parent document
        """
        super().__init__(document)

        self.__rules = []

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.darkMagenta)
        self.__rules.append((QRegExp('\\b_*[A-Z][A-Z0-9_]*\\s*(?==)'), fmt))
        self.__rules.append((QRegExp('_*[A-Z][A-Z0-9_]*(?=[\\):])'),
                             fmt))

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.blue)
        fmt.setFontItalic(True)
        self.__rules.append((QRegExp('^ *[^:=]+ *(?=:)'), fmt))

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.darkBlue)
        fmt.setFontWeight(QFont.Bold)
        pat = ['\\.DEFAULT\\b', '\\.DELETE_ON_ERROR\\b',
               '\\.INTERMEDIATE\\b', '\\.PHONY\\b', '\\.POSIX\\b',
               '\\.PRECIOUS\\b', '\\.SECONDARY\\b', '\\.SILENT\\b',
               '\\.SUFFIXES\\b']
        self.__rules.extend([(QRegExp(pat), fmt) for pat in pat])

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.darkBlue)
        fmt.setFontItalic(True)
        pat = ['\\bifdef\\b', '\\bifeq\\b', '\\bfifdef\\b', '\\bifneq\\b',
               '\\belse\\b', '\\bendif\\b']
        self.__rules.extend([(QRegExp(pat), fmt) for pat in pat])

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.darkGreen)
        fmt.setFontItalic(True)
        self.__rules.append((QRegExp('\".*\"'), fmt))

        fmt = QTextCharFormat()
        fmt.setForeground(Qt.gray)
        self.__rules.append((QRegExp('^-(include)[^\n]*'), fmt))

        self.__comment = QTextCharFormat()
        self.__comment.setForeground(Qt.red)
        self.__rules.append((QRegExp('#[^\n]*'), self.__comment))

    def highlightBlock(self, text):
        for pat, fmt in self.__rules:
            expr = QRegExp(pat)
            tidx = expr.indexIn(text)
            while tidx >= 0:
                tlen = expr.matchedLength()
                self.setFormat(tidx, tlen, fmt)
                tidx = expr.indexIn(text, tidx + tlen)


WGUI_SYNTAX_HIGHLIGHTERS = {
    SCMN_MIME_TYPE_C: _WguiCSyntaxHighlighter,
    SCMN_MIME_TYPE_MAKEFILE: _WguiMakefileSyntaxHighlighter,
}


# ----------------------------------------------------------------------------
# - text editor                                                              -
# ----------------------------------------------------------------------------

class WguiTextEditor(WguiEditorBase, WguiEditorHasSaveP,
                     WguiEditorHasSearchP):
    """
    The WguiTextEditor class is the class responsible to display text data
    from a file or from the silkan database. The etxt editor operates with a
    mime type and optionnaly a file path or a text string.
    """

    def __init__(self, gContext, mime=None, path=None, text=None, parent=None):
        """
        Create a new text editor by media type, path or text data.

        :param gContext: the global context
        :type gContext: WcorGlobalContextP
        :param mime: Media type
        :type mime: ScmnMimeTypeP | None
        :param path: File path
        :type path: str | None
        :param text: Optional text string
        :type text: str | None
        :parm parent: Parent widget
        """
        super().__init__(parent)

        #: Application context
        self.__globalContext = gContext
        #: Mime type
        self.__mime = mime
        #: File path
        self.__path = path

        #: Text editor widget
        self.__textEdit = QTextEdit()
        #: Search widget
        self.__searchWidget = WguiTextSearchWidget(self.__textEdit)

        # set the text string by path
        if text is None and path is not None:
            text = self._read()
            if text is None:
                # noinspection PyArgumentList
                QMessageBox.critical(parent, 'File not found',
                                     'Could not open file ' + path +
                                     ', please ensure that it exists'
                                     ' and is readable.')
        # add the text to the widget
        #: Original text
        self.__original = text
        if text is not None: self.__textEdit.setPlainText(text)
        # set the syntax highlighter
        hlght = WGUI_SYNTAX_HIGHLIGHTERS.get(mime)
        if hlght is not None:
            self.__textHighlight = hlght(self.__textEdit.document())

        self._initComponoents()
        self.setLayout(self._createLayout())

    def _initComponoents(self):
        """
        Initialize all UI components as needed.
        """
        # attach a courier like font
        font = QFont()
        font.setFamily('Courier')
        font.setFixedPitch(True)
        font.setPointSize(10)
        self.__textEdit.setFont(font)

    def _createLayout(self):
        """
        Create the widget's layout.

        :return: Layout for the widget
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setContentsMargins(5,5,5,5)
        layout.setSpacing(2)
        layout.addWidget(self.__textEdit)
        layout.setStretch(0, 1)

        btmLayout = QHBoxLayout()
        btmLayout.setContentsMargins(0,0,0,0)
        btmLayout.setSpacing(0)
        btmLayout.addWidget(self.__searchWidget)
        layout.addLayout(btmLayout)

        return layout

    def _read(self):
        text = None
        if self.__path:
            f = QFile(self.__path)
            if f.open(QFile.ReadOnly | QFile.Text):
                text = f.readAll()
                # noinspection PyArgumentList
                text = str(text, 'utf-8')
                f.close()
        return text

    def getTitle(self):
        return scmn_path_base(self.__path) if self.__path else None

    def getIcon(self):
        return ':/img/app_was_txti.png'

    def getEditorIdent(self):
        path = self.__path
        return 'file:' + path if path else 'wcr:text-{}'.format(id(self))

    def isReadOnly(self):
        return False

    def isICDProject(self):
        return False

    def isDirty(self):
        orig = self.__original
        text = self.__textEdit.toPlainText()
        return (orig or text) and text != orig

    def save(self):
        """
        Save the text editor's contents using its known file path.
        If no file path has been set yet, either because the content has been
        loaded from a file or from a previous save-as operation, save() will
        behave like saveAs().
        """
        if self.__path is None:
            self.saveCopy()
            return

        f = QFile(self.__path)
        if f.open(QFile.WriteOnly | QFile.Truncate | QFile.Text):
            # noinspection PyArgumentList
            text = bytes(self.__textEdit.toPlainText(), 'utf-8')
            f.write(text)
            f.close()
        else:
            raise IOError('Could not write to file ' + self.__path)

    def saveCopy(self):
        """
        Save the text file by media type and path

        :return: The actual destination path and mime type, or (None,None) if
                 the file could not be saved to the selected location.
        :rtype: (str, str)
        """

        if self.__path:
            pext = scmn_path_extension(self.__path)
            mtyps = [m for m in SCMN_MIMES.values()
                     if m.extension == pext or pext in m.auxExtensions]
        else:
            mtyps = [m for m in SCMN_MIMES.values()
                     if m.payloadTypes & SCMN_PAYLOAD_TEXT]

        # call the file dialog
        path, mtyp = wgui_getSaveFile(
            'Save as File', self.parentWidget(), *mtyps)

        if not path:
            return None, None
        self.__path = path
        self.__mime = mtyp.mimeType
        self.save()
        self.__globalContext.addRecentlyOpenedFile(path)
        self.nameChanged.emit(self.getTitle())
        return path, mtyp.mimeType

    def activated(self):
        """
        Called whenever this window becomes in the active one.
        """
        pass

    def print(self):
        wgui_print_text(self.__textEdit, parent=self.parentWidget())

    def copy(self):
        self.__textEdit.copy()

    def cut(self):
        self.__textEdit.cut()

    def paste(self):
        self.__textEdit.paste()

    def selectAll(self):
        self.__textEdit.selectAll()

    def activateFind(self):
        self.__searchWidget.setFocus(Qt.OtherFocusReason)

    def findAgain(self):
        self.__searchWidget.findAgain()
