# ----------------------------------------------------------------------------
# - wgui-dtree                                                               -
# - wild cruncher: tree display                                              -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_dtree
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher database tree display

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod, ABCMeta

from PyQt4.QtCore import Qt, pyqtSignal
from PyQt4.QtGui import QContextMenuEvent, QWidget, QVBoxLayout, QHBoxLayout, \
    QPushButton, QTreeView, QAbstractItemView, QItemSelectionModel, \
    QItemSelection, QSortFilterProxyModel, QHeaderView, QLineEdit

from wcr_frontend.gui.wgui_tree_model import WguiTreeModel


#: Role ID for storing the associated DB item with a tree item
WGUI_TREE_ITEM_ROLE = Qt.UserRole

#: Role ID for storing the associated path with a tree item
WGUI_TREE_PATH_ROLE = Qt.UserRole + 1

#: Role ID for storing whether the item may be a leaf
WGUI_TREE_NOLEAF_ROLE = Qt.UserRole + 2


class WguiTree(QWidget):
    """
    Database tree widget.
    """

    #: Signal emitting whenever the selection in the tree changes, carrying
    #: the selected database items
    # list[(WhdbBaseP, list[WhdbBlockInstanceP])]
    itemSelected = pyqtSignal([object])

    def __init__(self, model, modelMngr, menuFactory, parent=None):
        """
        :param model: Tree-model
        :type model: QAbstractItemModel
        :param modelMngr: Tree-model manager
        :type modelMngr: WguiTreeModelManager
        :param menuFactory: Context menu factory
        :type menuFactory: WguiContextMenuFactoryP
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: The actual tree widget
        self.__tree = _WguiTreeView(menuFactory)
        #: Whether selection handling is currently ongoing
        self.__selecting = False

        #: Model manager
        self.__treeAdapter = modelMngr
        #: Tree model
        self.__treeModel = model
        #: Last-set path
        self.__lastPath = None

        self.__modelProxied = isinstance(self.__treeModel,
                                         QSortFilterProxyModel)

        #: Reset-view button
        self.__resetButton = QPushButton('Reset')
        #: Expand-all button
        self.__cllOthersButton = QPushButton('Coll. Others')
        #: Expand-all button
        self.__expAllButton = QPushButton('Exp. All')
        #: Collapse-all button
        self.__cllAllButton = QPushButton('Coll. All')

        #: Filter field
        self.__filterF = QLineEdit() if self.__modelProxied else None

        self.setLayout(self._createLayout())
        self._initElements()

    def _createLayout(self):
        """
        Create the layout.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(3)
        layout.addWidget(self.__tree)

        if self.__filterF:
            layout.addWidget(self.__filterF)

        btnLayout = QHBoxLayout()
        btnLayout.setMargin(0)
        btnLayout.setSpacing(3)
        btnLayout.addWidget(self.__resetButton)
        btnLayout.addWidget(self.__cllOthersButton)
        btnLayout.addStretch(0)
        btnLayout.addWidget(self.__expAllButton)
        btnLayout.addWidget(self.__cllAllButton)
        layout.addLayout(btnLayout)

        return layout

    def _initElements(self):
        """
        Initialize the UI elements as needed.
        """
        if self.__filterF:
            model = self.__treeModel
            """:type: QSortFilterProxyModel"""
            self.__filterF.setPlaceholderText(
                'Filter Tree (regular expression)')
            self.__filterF.textChanged.connect(model.setFilterRegExp)

        self.__resetButton.clicked.connect(self._resetView)
        self.__expAllButton.clicked.connect(self.__tree.expandAll)
        self.__cllAllButton.clicked.connect(self.__tree.collapseAll)
        self.__cllOthersButton.clicked.connect(self._collapseOthers)

        self.__treeModel.setSourceModel(self.__treeAdapter.model)
        self.__treeModel.setDynamicSortFilter(True)
        self.__treeModel.invalidate()

        self.updateTree()

    def sizeHint(self):
        return self.minimumSizeHint()

    @property
    def filterF(self):
        return self.__filterF

    def updateTree(self):
        """
        Perform a complete update of the tree, i.e. recreate the tree.
        A complete tree update might lead to the loss of the current expansion
        and selection status.
        """
        self.__tree.setModel(None)
        self.__treeAdapter.populateModel()
        self.__tree.setModel(self.__treeModel)

        self.__tree.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.__tree.selectionModel().selectionChanged.connect(
            self._selectionChanged)

        self.__tree.setAutoScroll(True)
        hdrv = self.__tree.header()
        """:type: QHeaderView"""
        hdrv.setStretchLastSection(True)
        hdrv.setResizeMode(0, QHeaderView.ResizeToContents)

        self.hierarchyLevelChanged(None)

    def _resetView(self):
        self.hierarchyLevelChanged(self.__lastPath)

    def _collapseOthers(self):
        exps = set()
        idcs = self.__tree.selectedIndexes()
        """:type: list[QModelIndex]"""
        for idx in (i for i in idcs if i.column() == 0):
            while idx and idx.isValid():
                exps.add(idx)
                idx = idx.parent()
        """:type: QModelIndex"""
        self.__tree.collapseAll()
        for idx in exps:
            self.__tree.expand(idx)
        self.__tree.scrollTo(idcs[0], QAbstractItemView.PositionAtCenter)

    def hierarchyLevelChanged(self, path):
        """
        Set a new hierarchy level.

        :param path: Path to hierarchy level
        :type path: list[object] | None
        """
        self.__lastPath = path[:] if path else []

        self.__tree.collapseAll()

        if path:
            idx = self.__treeAdapter.getItemIndex(path[-1])
        else:
            idx = self.__treeAdapter.getItemIndex(None)

        if self.__modelProxied:
            idx = self.__treeModel.mapFromSource(idx)

        if idx.isValid():
            for row in range(0, self.__treeModel.rowCount(idx)):
                self.__tree.expand(idx.child(row, 0))
        else:
            for row in range(0, self.__treeModel.rowCount(idx)):
                _idx = self.__treeModel.index(row, 0)
                self.__tree.expand(_idx)

        while idx and idx.isValid():
            self.__tree.expand(idx)
            idx = idx.parent()

    def selectDbItems(self, items):
        """
        Select the given list of database items in the tree.

        :param items: Items to be selected
        :type items: list[object]
        """
        if self.__selecting:
            return

        fst = None
        slct = QItemSelection()
        expn = set()
        for item in items:
            idx = self.__treeAdapter.getItemIndex(item)
            if idx:
                if self.__modelProxied:
                    idx = self.__treeModel.mapFromSource(idx)
                if fst is None:
                    fst = idx
                slct.select(idx, idx)
                idx = idx.parent()
                while idx and idx.isValid():
                    expn.add(idx)
                    idx = idx.parent()
        for idx in expn:
            if not self.__tree.isExpanded(idx):
                self.__tree.expand(idx)
        self.__tree.selectionModel().select(
            slct,
            QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows)
        if fst:
            self.__tree.scrollTo(fst)

    def selectedDbItems(self):
        """
        Get the list of database items currently selected in the tree.

        :return: List of selected database items and their paths
        :rtype: list[(WhdbBaseP, list[WhdbBlockInstanceP])]
        """
        return [(self.__treeModel.data(idx, WGUI_TREE_ITEM_ROLE),
                 self.__treeModel.data(idx, WGUI_TREE_PATH_ROLE))
                for idx in self.__tree.selectionModel().selectedIndexes()
                if idx.column() == 0
                and self.__treeModel.data(idx, WGUI_TREE_ITEM_ROLE)]

    def _selectionChanged(self, *_):
        """
        Callback function to handle tree selection changes.
        """
        self.__selecting = True
        self.itemSelected.emit(self.selectedDbItems())
        self.__selecting = False


class _WguiTreeView(QTreeView):
    """
    Database tree widget backend.
    """

    def __init__(self, menuFactory, parent=None):
        """
        :param menuFactory: Context menu factory
        :type menuFactory: WguiContextMenuFactoryP
        :param parent: Parent widget
        """
        super().__init__(parent)

        self.__menuFactory = menuFactory

    def contextMenuEvent(self, event:QContextMenuEvent):
        menu = None
        slcd = [idx for idx in self.selectedIndexes() if idx.column() == 0]
        if len(slcd) > 1:
            menu = self.__menuFactory.createSelectionContextMenu(self)
        else:
            idx = self.indexAt(event.pos())

            if idx is not None:
                dataItem = self.model().data(idx, WGUI_TREE_ITEM_ROLE)
                if dataItem is not None:
                    menu = self.__menuFactory.createItemContextMenu(dataItem,
                                                                    self)

        if menu is not None:
            menu.exec_(event.globalPos())


class WguiTreeModelManager(metaclass=ABCMeta):
    """
    Model manager base class.
    """

    def __init__(self, itemIdF, itemFactory):
        """
        :param itemIdF: Item-ID function
        :type itemIdF: (object) -> object
        :param itemFactory: Item factory
        :type itemFactory: WguiTreeItemFactoryP
        """
        #: Item-ID function
        self.__itemIdF = itemIdF

        #: Tree model
        self.__treeModel = WguiTreeModel(itemFactory)

        #: Item-TreeItem mapping
        self.__items = {}
        """:type: dict[object, WguiTreeItemP]"""

    @property
    def model(self):
        """
        Get the underlaying tree model.

        :return: Tree model
        :rtype: WguiTreeModel
        """
        return self.__treeModel

    @abstractmethod
    def populateModel(self, fromScratch=True):
        """
        Repopulate the model, possible from scratch after clearing it.

        :param fromScratch: Clear the items prior to populating, effectively
            overriding completely the current items
        :type fromScratch: bool
        """
        raise NotImplementedError()

    def itemAdded(self, item, path=None):
        """
        Add an item to the tree model.

        :param item: Item to be added
        :type item: object
        :param path: Item path
        :type path: list[WhdbHasIdP]
        """
        if path:
            parentItem = self.__items.get(path[-1].getUUID())
        else:
            parentItem = None

        tpls = self.__treeModel.insertItem(item, parentItem, path=path)
        for data, treeItem in tpls:
            self.__items[self.__itemIdF(data)] = treeItem

    def itemClear(self):
        """
        Remove all the items.
        """

        # when we try to delet leaf it explode.
        for itemKey, itemVal in self.__items.items():
            self.__treeModel.removeItem(itemVal)

        self.__items.clear()

    def itemRemoved(self, item):
        """
        Remove the given item.

        :param item: Item to be removed
        :type item: object
        """
        itemId = self.__itemIdF(item)
        treeItem = self.__items.get(itemId)

        if treeItem is not None:
            self.__treeModel.removeItem(treeItem)
            del self.__items[itemId]

    def itemChanged(self, item):
        """
        Refresh a changed item.

        :param item: Item that has changed
        :type item: object
        """
        treeItem = self.__items.get(self.__itemIdF(item))

        self.__treeModel.refreshItem(treeItem)

    def getItemIndex(self, item):
        """
        Get the index for the given item.

        :param item: Item to get index for
        :type item: object
        :return: Index for given item, if any
        :rtype: QModelIndex
        """
        treeItem = self.__items.get(self.__itemIdF(item))
        return self.__treeModel.getItemIndex(treeItem)

    def getmap(self):
        return self.__items
