# ----------------------------------------------------------------------------
# - wgui-tree                                                                -
# - wild cruncher: generic tree                                              -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_tree
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher generic tree

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod, ABCMeta
import re

from PyQt4.QtCore import Qt, QModelIndex, QAbstractItemModel, QRegExp
from PyQt4.QtGui import QSortFilterProxyModel


WGUI_TREE_DATA_ITEM_ROLE = Qt.UserRole + 999


class WguiTreeItemP(metaclass=ABCMeta):
    """
    Tree-item protocol.
    """

    @abstractmethod
    def addChild(self, child, row=None):
        """
        Add a child to the tree item.

        :param child: Child to be added
        :type child: WguiTreeItemP
        :param row: Row to add child in (append at end if `None`)
        :type row: int | None
        """
        raise NotImplementedError()

    @abstractmethod
    def removeChild(self, row):
        """
        Remove the child in the given row.

        :param row: Row of child to be removed
        :type row: int
        """
        raise NotImplementedError()

    @abstractmethod
    def removeChildren(self, row, count):
        """
        Remove `count` children starting in the given row.

        :param row: Row to begin deletion
        :type row: int
        :param count: Number of children to be deleted
        :type count: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getChildren(self):
        """
        Get the item's children.

        :return: Children
        :rtype: Iterable[WguiTreeItemP]
        """
        raise NotImplementedError()

    @abstractmethod
    def getChild(self, row):
        """
        Get the child in the given row.

        :param row: Row of child to be retrieved
        :type row: int
        :return: Child in the given row
        :rtype: WguiTreeItemP
        """
        raise NotImplementedError()

    @abstractmethod
    def getChildCount(self):
        """
        Get the number of children.

        :return: Number of children
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getRow(self):
        """
        Get the row number of the item in the parent's list of children.

        :return: Row number of item at parent
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getChildRow(self, child):
        """
        Get the row number for the given child.

        :param child: Child to get row number for
        :type child: WguiTreeItemP
        :return: Row number of given child
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getColumnCount(self):
        """
        Get the number of columns for this item.

        :return: Number of columns
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getData(self, column, role=None):
        """
        Get the data for the given column.

        :param column: Column to return data for
        :type column: int
        :param role: Data type/role
        :type role: int
        :return: Data for the given column and row
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def setData(self, data, column, role=None):
        """
        Set the data for the given column

        :param data: Data item to set
        :type data: object
        :param column: Column to set data for
        :type column: int
        :param role: Type/role of given data
        :type role: int
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def parentItem(self):
        """
        Get the tree item's parent (or `None` in case of a root item).

        :return: Parent item
        :rtype: WguiTreeItemP | None
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def flags(self):
        """
        Get the flags for the tree item.

        :return: Flags
        :rtype: int
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WguiTreeItemBase(WguiTreeItemP):
    """
    Abstract base class for tree items.
    """

    def __init__(self, parentItem=None, children=None):
        """
        :param parentItem: Parent item
        :type parentItem: WguiTreeItemP
        :param children: Pre-set child items
        :type children: list[WguiTreeItemP]
        """
        assert parentItem is None or isinstance(parentItem, WguiTreeItemP)

        #: Parent item
        self.__parent = parentItem

        #: Children
        self.__children = children if children is not None else []
        """:type: list[WguiTreeItemP]"""

    def addChild(self, child, row=None):
        if row is None:
            self.__children.append(child)
        else:
            self.__children.insert(row, child)

    def removeChild(self, row):
        del self.__children[row]

    def removeChildren(self, row, count):
        if count == 1:
            del self.__children[row]
        else:
            self.__children[row:(row + count)] = []

    def getChildren(self):
        return iter(self.__children)

    def getChild(self, row):
        return self.__children[row]

    def getChildRow(self, child):
        return self.__children.index(child)

    def getChildCount(self):
        return len(self.__children)

    def getRow(self):
        if self.__parent:
            return self.__parent.getChildRow(self)
        else:
            return 0

    @property
    def parentItem(self):
        return self.__parent


class WguiSimpleTreeItem(WguiTreeItemBase):
    """
    Simple default implementation for tree items.
    """

    def __init__(self, itemData, parentItem=None):
        """
        :param itemData: List of data items, by column
        :type itemData: list[object]
        :param parentItem: Parent item
        :type parentItem: WguiTreeItemP | None
        """
        if not isinstance(itemData, list) or not itemData:
            raise ValueError('Invalid item data:', itemData)

        super().__init__(parentItem)

        self.__data = itemData

    def getColumnCount(self):
        return len(self.__data)

    def getData(self, column, role=None):
        if role != Qt.DisplayRole:
            return None
        return self.__data[column] if 0 <= column < len(self.__data) else None

    def setData(self, data, column, role=None):
        self.__data[column] = data

    @property
    def flags(self):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def __str__(self):
        return 'TreeItem {} ({})'.format(self.__data, self.parentItem)


class WguiTreeItemFactoryP(object):
    """
    Tree-item factory protocol.
    """

    @abstractmethod
    def createTreeItem(self, itemData, parentItem, **kwargs):
        """
        Create a tree item for the given item data.

        Also provides the effective parent item under which the new tree item
        should be placed.

        :param itemData: Data to create tree item for
        :type itemData: object
        :param parentItem: Parent item
        :type parentItem: WguiTreeItemP | None
        :return: Newly created tree-item, effective parent item and children
        :rtype: (WguiTreeItemP, WguiTreeItemP, (Iterable[object], dict) | None)
        """
        raise NotImplementedError()


class WguiSimpleTreeItemFactory(WguiTreeItemFactoryP):
    """
    Simple tree-item factory.
    """

    def __init__(self, headers):
        """
        :param headers: Header labels
        :type headers: list[str]
        """
        self.__headers = headers

    def createTreeItem(self, itemData, parentItem, **kwargs):
        if not parentItem:
            itemData = self.__headers
        return WguiSimpleTreeItem(itemData, parentItem), parentItem, None


class WguiTreeModel(QAbstractItemModel):
    """
    Tree model.
    """

    def __init__(self, itemFy, **kwargs):
        """
        :param itemFy: Item factory
        :type itemFy: WguiTreeItemFactoryP
        """
        super().__init__()

        #: Item factory
        self.__itemFactory = itemFy
        #: Root item
        self.__root, _, clds = itemFy.createTreeItem(None, None, **kwargs)

        if clds:
            children, cargs = clds
            for child in children:
                self.insertItem(child, parentItem=self.__root, **cargs)

    def index(self, row, col, parentIdx=None, *args, **kwargs):
        if not self.hasIndex(row, col, parentIdx or QModelIndex()):
            return QModelIndex()

        parentItem = self._getItem(parentIdx)
        childItem = parentItem.getChild(row)
        if childItem:
            return self.createIndex(row, col, childItem)
        else:
            return QModelIndex()

    def parent(self, index=None):
        item = self._getItem(index)

        parentItem = item.parentItem

        if parentItem is None:
            return QModelIndex()
        else:
            return self.getItemIndex(parentItem)

    def rowCount(self, pidx=None, *args, **kwargs):
        """
        :type pidx: QModelIndex
        :rtype: int
        """
        if pidx and pidx.column() > 0:
            return 0

        return self._getItem(pidx).getChildCount()

    def columnCount(self, pidx=None, *args, **kwargs):
        """
        :type pidx: QModelIndex
        :rtype: int
        """
        return self._getItem(pidx).getColumnCount()

    def data(self, indx, role=None):
        """
        :type indx: QModelIndex
        :type role: int
        :rtype: object
        """
        item = self._getItem(indx)
        if role == WGUI_TREE_DATA_ITEM_ROLE:
            return item
        else:
            return item.getData(indx.column(), role)

    def item(self, index):
        """
        Get the item at the given index, if any.

        :type index: QModelIndex
        :rtype: WguiTreeItemP | None
        """
        return self._getItem(index)

    def flags(self, index):
        return self._getItem(index).flags

    def headerData(self, section, orientation, role=None):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.__root.getData(section, role)

        return None

    def removeRows(self, row, count, parentIdx=None, *args, **kwargs):
        if parentIdx is None:
            parentIdx = QModelIndex()
        parentItem = self._getItem(parentIdx)
        if not parentItem:
            return False

        self.beginRemoveRows(parentIdx, row, row + count - 1)
        parentItem.removeChildren(row, count)
        self.endRemoveRows()
        return True

    def insertItem(self, itemData, parentItem=None, row=None, **kwargs):
        """
        Insert a new item under the given parent.

        The created tree-item is also returned, for reference purposes.

        :param itemData: Item data for tree-item to be inserted
        :type itemData: object
        :param parentItem: Parent item
        :type parentItem: WguiTreeItemP | None
        :param row: Row for new item (at end, if `None`)
        :type row: int | None
        :param kwargs: Additional arguments for tree-item creation
        :return: Newly created tree-item
        :rtype: Iterable[(object, WguiTreeItemP)]
        """
        if parentItem is None:
            parentItem = self.__root

        item, effParentItem, clds = self.__itemFactory.createTreeItem(
            itemData, parentItem, **kwargs)
        if effParentItem is not parentItem:
            parentItem = effParentItem


        self.insertTreeItem(item, parentItem=parentItem, row=row)

        itms = [(itemData, item)]

        if clds:
            children, cargs = clds
            for child in children:
                itms.extend(self.insertItem(child, parentItem=item, **cargs))

        return itms

    def insertTreeItem(self, item, parentIdx=None, parentItem=None, row=None):
        """
        Insert an item into the tree. If neither parent item nor parent index
        are specified, the item is inserted below the root element.
        If both parent item and parent index are given, but are inconsistent,
        the result is not defined.

        :param item: Item to be inserted
        :type item: WguiTreeItemP
        :param parentIdx: Index of the parent item
        :type parentIdx: QModelIndex | None
        :param parentItem: Parent item
        :type parentItem: WguiTreeItemP
        :param row: Row to insert item in
        :type row: int
        :return: Inserted tree item (same as `item`)
        :rtype: WguiTreeItemP
        """
        if parentIdx is None:
            parentIdx = self.getItemIndex(parentItem)
        if parentItem is None:
            parentItem = self._getItem(parentIdx)

        if row is None:
            row = parentItem.getChildCount()

        self.beginInsertRows(parentIdx, row, row)
        parentItem.addChild(item, row)
        self.endInsertRows()
        return item

    def removeItem(self, dataItem):
        """
        Remove the given tree-item.

        :param dataItem: Tree item to be removed
        :type dataItem: WguiTreeItemP
        """
        row = dataItem.getRow()
        parentIdx = self.getItemIndex(dataItem.parentItem)
        self.beginRemoveRows(parentIdx, row, row)
        dataItem.parentItem.removeChild(row)
        self.endRemoveRows()

    def refreshItem(self, dataItem):
        """
        Refresh the given tree-item.

        :param dataItem: Tree-item to be refreshed
        :type dataItem: WguiTreeItemP
        """
        start = self.getItemIndex(dataItem)
        # noinspection PyUnresolvedReferences
        self.dataChanged.emit(
            start, start.sibling(start.row(), dataItem.getColumnCount() - 1))

    def _getItem(self, index=None, allowInvalid=True):
        """
        Get the tree-item with the given index.

        :param index: Index of tree-item to be returned (root-item if `None`)
        :type index: QModelIndex
        :param allowInvalid: Whether to return root-item in case of an invalid
                             index
        :type allowInvalid: bool
        :return: Tree-item for given index
        :rtype: WguiTreeItemP
        """
        if not index:
            return self.__root
        elif not index.isValid():
            return self.__root if allowInvalid else None
        else:
            return index.internalPointer()

    def getItemIndex(self, item):
        """
        Get the index of the given item.

        :param item: Item to return index for
        :type item: WguiTreeItemP | None
        :return: Index for given item
        :rtype: QModelIndex
        """
        if item is None or item is self.__root:
            return QModelIndex()
        return self.createIndex(item.getRow(), 0, item)


class WguiFilterModel(QSortFilterProxyModel):
    """
    Sort/Filter model proxy for trees.
    """

    def __init__(self, parent):
        super().__init__(parent)

        self.__regExp = None
        self.__parentCache = set()

    def setFilterRegExp(self, *args):
        self.__parentCache.clear()
        _rgxp = None
        """:type: str | None"""
        if args:
            arg = args[0]
            if isinstance(arg, QRegExp):
                _rgxp = arg.pattern()
            else:
                _rgxp = arg
        if _rgxp:
            try:
                self.__regExp = re.compile(_rgxp, re.IGNORECASE)
            except:
                self.__regExp = None
        else:
            self.__regExp = None
        super().setFilterRegExp(*args)

    def setFilterCaseSensitivity(self, sensitive):
        self.__parentCache.clear()
        super().setFilterCaseSensitivity(sensitive)

    def setFilterKeyColumn(self, kcol):
        self.__parentCache.clear()
        super().setFilterKeyColumn(kcol)

    def setFilterRole(self, role):
        self.__parentCache.clear()
        super().setFilterRole(role)

    def setFilterWildcard(self, wildcard):
        self.__parentCache.clear()
        super().setFilterWildcard(wildcard)

    def setFilterFixedString(self, fstr):
        self.__parentCache.clear()
        super().setFilterFixedString(fstr)

    def setDynamicSortFilter(self, dynamic):
        self.__parentCache.clear()
        super().setDynamicSortFilter(dynamic)

    def filterAcceptsRow(self, srow, pidx):
        rgxp = self.__regExp

        if rgxp:
            srcIdx = self.sourceModel().index(
                srow, self.filterKeyColumn(), pidx)
            """:type: QModelIndex"""

            if srcIdx in self.__parentCache:
                return True

            if srcIdx.isValid():
                rows = self.sourceModel().rowCount(srcIdx)
                for i in range(rows):
                    if self.filterAcceptsRow(i, srcIdx):
                        self.__parentCache.add(
                            self.sourceModel().index(
                                i, self.filterKeyColumn(), srcIdx))
                        return True

            data = self.sourceModel().data(srcIdx, self.filterRole())
            acpt = rgxp.search(data)
            if acpt:
                self.__parentCache.add(srcIdx)
            return bool(acpt)

        return super().filterAcceptsRow(srow, pidx)
