# ----------------------------------------------------------------------------
# - wgui-utils                                                               -
# - wild cruncher - cockpit system utilities                                 -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_utils
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graphic utilities

.. moduleauthor:: Amaury Darsch <amaury.darsch@silkan.com>
"""
from abc import ABCMeta, abstractmethod

import itertools
import traceback

from PyQt4.QtCore import QPoint, QPointF, QRectF, QThread, pyqtSignal, QObject
from PyQt4.QtGui import QColor, QComboBox, QPrinter, QPrintDialog, QTextEdit, \
    QAbstractPrintDialog, QDialog, QFileDialog, QApplication

from wcr_core.common.wcmn_file_handler import WcmnFile, wcmn_handleFile
from silkan_common.scmn_file_path import scmn_path_dir
from silkan_common.scmn_utils import concatWithoutDuplicates
from wcr_frontend.gui.wgui_protocols import QObject_Meta_M


#: Selection highlight colour
WGUI_SELECTION_COLOUR = QColor(102, 178, 255)

#: Hovering highlight color
WGUI_HOVER_COLOUR = QColor(204, 229, 255)

#: Mime type for dragging model/function instances into view
WGUI_INST_DND_MIME_TYPE = 'application/x-silkan-function-type'

#: Directory used last
WGUI_LAST_DIR = None

WGUI_CONNECTION_ARROW = '&#10142;'


# ----------------------------------------------------------------------------
# - grid snapping                                                            -
# ----------------------------------------------------------------------------

def wgui_snap_coord(x:float, xs:float) -> float:
    """
    Snap value to a grid

    :param  x: the coordinate to snap
    :param xs: the x snapping grid
    """
    return round(x / xs) * xs


def wgui_snap_point(p:'QPoint,QPointF',
                    xs:float, ys:float) -> 'QPoint,QPointF':
    """
    Snap a point to a grid

    :param  p: the point to snap
    :param xs: the x snapping grid
    :param ys: the y snapping grid
    """
    # get the point coordinate
    x = p.x()
    y = p.y()
    # snap to x
    x = wgui_snap_coord(x, xs)
    # snap to y
    y = wgui_snap_coord(y, ys)
    # rebuild result point
    if isinstance(p, QPoint):
        return QPoint(x, y)
    else:
        return QPointF(x, y)


def wgui_snap_rect(r, xs, ys):
    """
    Snap a rectangle to a grid

    :param r: the rectangle to snap
    :type r: QRectF
    :param xs: the x snapping grid
    :type xs: float
    :param ys: the y snapping grid
    :type ys: float
    :rtype: QRectF
    """
    # get the rectangle points
    tl = r.topLeft()
    br = r.bottomRight()
    # snap the rectangle points
    tl = wgui_snap_point(tl, xs, ys)
    br = wgui_snap_point(br, xs, ys)
    # rebuild result rectangle
    return QRectF(tl, br)


# ----------------------------------------------------------------------------
# - misc ui helpers                                                          -
# ----------------------------------------------------------------------------

def _gen_denoms():
    for d in itertools.count(): yield 2 ** d


def _gen_fracs():
    for i in _gen_denoms():
        for j in range(1, i, 2):
            yield j / i


def wgui_colour_dict(items):
    """
    Generate a dictionary of distinct colours for a set of items.

    :param items: Collection of items
    :type items: [object]
    :return: Colour dictionary for the given collection of items
    :rtype: {object:QColor}
    """
    # noinspection PyArgumentList
    return dict([(s, QColor.fromHsvF(v, 0.2, 1.0)) for s, v
                 in zip(items, itertools.islice(_gen_fracs(), len(items)))])


def wgui_qgi_parents(item):
    """
    Produce a list of parents for the given item, recursively up to the top
    element.

    :param item: Item to provide parents list for
    :type item: QGraphicsItem
    :return: List of parents for the given item
    :rtype: [QGraphicsItem]
    """
    parents = []
    parent = item.parentItem()
    while parent is not None:
        parents.append(parent)
        parent = parent.parentItem()
    return parents


def wgui_qgi_isany(item, pred):
    """
    Whether the given item or any of its parents satisfy the given predicate.

    :param item: Item to test
    :type item: QGraphicsItem
    :param pred: Test lambda
    :type pred: QGraphicsItem -> bool
    :return: True iff the item or any of its parents satisfy the given
             predicate
    :rtype: bool
    """
    itm = item
    while itm is not None:
        if pred(itm): return True
        itm = itm.parentItem()
    return False


# ----------------------------------------------------------------------------
# - qcombobox helpers                                                        -
# ----------------------------------------------------------------------------

def wgui_qcb_set_data(data, combobox:QComboBox):
    idx = combobox.findData(data)
    if idx >= 0:
        combobox.setCurrentIndex(idx)


def wgui_qcb_get_data(combobox:QComboBox):
    idx = combobox.currentIndex()
    if idx >= 0:
        return combobox.itemData(idx)


# ----------------------------------------------------------------------------
# - general helpers                                                          -
# ----------------------------------------------------------------------------

def isanyinstance(objs, clazz):
    if not objs: return False
    for o in objs:
        if isinstance(o, clazz):
            return True
    return False


def getinstance(objs, clazz):
    if not objs: return None
    for o in objs:
        if isinstance(o, clazz):
            return o
    return None


def wgui_print_text(text, title=None, parent=None):
    """
    Print text from a text widget.

    :param text: Text widget to print text from
    :type text: QTextEdit
    :param title: Title of the print dialog
    :type title: str
    :param parent: Parent widget
    """
    printer = QPrinter()
    dialog = QPrintDialog(printer, parent)
    dialog.setWindowTitle(title or 'Print Options')
    if text.textCursor().hasSelection():
        dialog.addEnabledOption(QAbstractPrintDialog.PrintSelection)
    if dialog.exec_() != QDialog.Accepted:
        return
    text.print_(printer)


def wgui_selectAndHandleFile(title, parent, gctx, *handlers):
    """
    Display a file selection dialog and pass the selected file to the given
    handlers.
    The list of selectable file types is determined by the types that can be
    handled by the given handlers.

    :param title: Title of the file dialog
    :type title: str
    :param parent: Parent widget
    :param gctx: Global context
    :type gctx: WcorGlobalContextP
    :param handlers: File handlers
    :type handlers: list[WcmnFileHandler]
    """
    mimeTypes = concatWithoutDuplicates(
        *[h.handledMimeTypes for h in handlers])
    path = wgui_getOpenFile(title, parent, *mimeTypes)
    if path:
        wcmn_handleFile(WcmnFile(path), gctx, *handlers)


def wgui_getMimeTypeFilter(*mimeTypes):
    """
    Get the filter-string for Qt file dialogs for the given MIME-types.

    :param mimeTypes: MIME-types to produce filter-string for
    :type mimeTypes: list[ScmnMimeTypeP]
    :return: Filter-string and filter-to-MIME dictionary
    :rtype: (str, dict[str, ScmnMimeTypeP])
    """
    if not mimeTypes:
        return None

    def filterStr(mime, exts):
        return '{0} ({1})'.format(mime.description, exts)

    def extStr(ext):
        return '*.' + ext

    def extFilter(mime):
        ext = mime.extension
        auxExt = mime.auxExtensions
        if auxExt:
            return filterStr(mime, ' '.join(map(extStr, [ext] + auxExt))), mime
        elif ext:
            return filterStr(mime, extStr(ext)), mime
        else:
            return None

    filters = list(filter(None, (extFilter(m) for m in mimeTypes)))

    return ';;'.join(f for f, _ in filters), dict(filters)


def wgui_getOpenFile(title, parent, *mimeTypes, filterStr=None, idir=None):
    """
    Displays a file-selection dialog for opening of a file of the given
    MIME-type(s).

    :param title: Title for the file-selection dialog
    :type title: str
    :param parent: Parent widget
    :type parent: QWidget
    :param mimeTypes: Possible MIME-types
    :type mimeTypes: list[ScmnMimeTypeP]
    :param filterStr: Additional filter-string(s) in Qt-format
    :type filterStr: str | None
    :param idir: Initial directory
    :type idir: str | None
    :return: Path of selected file to open, if any
    :rtype: str | None
    """
    global WGUI_LAST_DIR
    if not idir and WGUI_LAST_DIR:
        idir = WGUI_LAST_DIR

    if mimeTypes:
        _filterStr, _ = wgui_getMimeTypeFilter(*mimeTypes)

        if filterStr:
            filterStr = _filterStr + ';;' + filterStr
        else:
            filterStr = _filterStr

    sdlg = QFileDialog(parent, title, idir, filterStr or None)
    sdlg.setAcceptMode(QFileDialog.AcceptOpen)
    sdlg.setFileMode(QFileDialog.ExistingFile)
    sdlg.setViewMode(QFileDialog.Detail)
    sdlg.setOption(QFileDialog.DontUseNativeDialog, False)

    if sdlg.exec_():
        files = sdlg.selectedFiles()
        assert len(files) == 1
        path = files[0]
        if path:
            WGUI_LAST_DIR = scmn_path_dir(path)
        return path
    else:
        return None


def wgui_getSaveFile(title, parent, *mimeTypes, idir=None):
    """
    Displays a file-selection dialog for saving of a file of the given
    MIME-type(s).

    :param title: Title for the file-selection dialog
    :type title: str
    :param parent: Parent widget
    :type parent: QWidget
    :param mimeTypes: Possible MIME-type(s)
    :type mimeTypes: list[ScmnMimeTypeP]
    :param idir: Initial directory
    :type idir: str | None
    :return: Selected file path and MIME-type, if any
    :rtype: (str, ScmnMimeTypeP) | (None, None)
    """
    global WGUI_LAST_DIR

    def filterSelected(fstr):
        nonlocal mtyp
        mtyp = filters.get(fstr)
        sdlg.setDefaultSuffix(mtyp.extension)

    if not idir and WGUI_LAST_DIR:
        idir = WGUI_LAST_DIR

    filterStr, filters = wgui_getMimeTypeFilter(*mimeTypes)
    mtyp = mimeTypes[0] if mimeTypes else None

    sdlg = QFileDialog(parent, title, idir, filterStr or None)
    sdlg.setAcceptMode(QFileDialog.AcceptSave)
    sdlg.setConfirmOverwrite(True)
    sdlg.setDefaultSuffix(mimeTypes[0].extension if mimeTypes else None)
    sdlg.setFileMode(QFileDialog.AnyFile)
    sdlg.setViewMode(QFileDialog.List)
    sdlg.setOption(QFileDialog.DontUseNativeDialog, False)
    sdlg.filterSelected.connect(filterSelected)

    if sdlg.exec_():
        files = sdlg.selectedFiles()
        assert len(files) == 1
        path = files[0]
        if path:
            WGUI_LAST_DIR = scmn_path_dir(path)
        return path, mtyp
    else:
        return None, None


def isMainThread():
    """
    Whether the thread on which this function is called is the main
    application thread.

    :return: True iff the current thread is the main application thread
    :rtype: bool
    """
    # noinspection PyArgumentList
    return QApplication.instance().thread() is QThread.currentThread()


class WguiCommandRunResultHandlerP(metaclass=ABCMeta):
    @abstractmethod
    def success(self):
        """
        Handle successful command execution.
        """
        raise NotImplementedError()

    @abstractmethod
    def error(self, exception, trace):
        """
        Handle error during command execution.

        :param exception: Exception
        :type exception: Exception
        :param trace: Stack trace
        :type trace: str
        """
        raise NotImplementedError()


class WguiCommandThread(QThread):
    """
    Thread to execute external commands.
    """

    #: Signal emitted when an exception occurs on the thread
    exceptionThrown = pyqtSignal(object, str)
    """:type: pyqtBoundSignal[object, str]"""

    #: Signal emitted when the thread finishedwithout exception
    finishedSuccesfully = pyqtSignal()
    """:type: pyqtBoundSignal"""

    def __init__(self, func, strm, handler, parent=None):
        """
        :param func: Function to execute on the thread
        :type func: (ScmnProcStreamHandlerP) -> _
        :param strm: Stream output handler
        :type strm: ScmnProcStreamHandlerP
        :param handler: Result handler
        :type handler: WguiCommandRunResultHandlerP
        :param parent: Parent object
        """
        super().__init__(parent)
        self.__func = func
        self.__strm = strm
        self.__handler = handler

        self.finishedSuccesfully.connect(handler.success)
        self.exceptionThrown.connect(handler.error)

    def run(self):
        try:
            self.__func(self.__strm)
            self.finishedSuccesfully.emit()
        except BaseException as e:
            self.exceptionThrown.emit(e, traceback.format_exc())


class WguiSimpleCommandRunResultHandler(QObject, WguiCommandRunResultHandlerP,
                                        metaclass=QObject_Meta_M):
    def __init__(self, successF=None, errorF=None):
        """
        :param successF: Function to run when the thread finished
        :type successF: () -> _
        :param errorF: Function to handle exceptions that occur on the thread,
                       taking an exception and a stack trace as arguments
        :type errorF: (BaseException, str) -> _
        """
        QObject.__init__(self)

        self.__successF = successF
        self.__errorF = errorF

    def success(self):
        if self.__successF:
            self.__successF()

    def error(self, exception, trace):
        if self.__errorF:
            self.__errorF(exception, trace)


def wgui_runCommandOnThread(func, strm, ffnc=None, efnc=None):
    """
    Run a function on a command execution thread.

    :param func: Function to execute on the thread
    :type func: (ScmnProcStreamHandlerP) -> _
    :param strm: Stream output handler
    :type strm: ScmnProcStreamHandlerP
    :param ffnc: Function to run when the thread finished
    :type ffnc: () -> _
    :param efnc: Function to handle exceptions that occur on the thread,
                 taking an exception and a stack trace as arguments
    :type efnc: (BaseException, str) -> _
    """
    handler = WguiSimpleCommandRunResultHandler(ffnc, efnc)
    # noinspection PyArgumentList
    WguiCommandThread(func, strm, handler, QApplication.instance()).start()


def wgui_instance_tooltip(instance):
    """
    Generate the tooltip text for the given state.

    :param instance: State to generate tooltip text for
    :type instance: WhdbInstanceP
    :return: Tooltip text for the given state
    :rtype: str
    """
    ttip = '<b>{}</b>'.format(instance.getName())

    if instance.getInfo():
        ttip += '<br/>' + instance.getInfo()

    return ttip


def wgui_connection_tooltip(arrow, connection, sname=None, ename=None):
    """
    Generate the tooltip text for the given transition.

    :param arrow: Arrow to be displayed between start and end objects
    :type arrow: str
    :param connection: Transition to generate tooltip text for
    :type connection: WhdbConnectionP
    :return: Tooltip text for the given transition
    :rtype: str
    """
    name = connection.getName()
    frto = '<span style="white-space:nowrap;">{} {} {}</span>'.format(
        sname or connection.getStartConnectionPoint().getName() or 'UNNAMED', arrow,
        ename or connection.getEndConnectionPoint().getName() or 'UNNAMED')

    if name:
        ttip = '<b>{}</b> ({})</span>'.format(name, frto)
    else:
        ttip = '<b>{}</b>'.format(frto)

    if connection.getInfo():
        ttip += '<br/>' + connection.getInfo()

    return ttip


def wgui_undirected_connection_tooltip(connection, sname=None, ename=None):
    """
    Generate the tooltip text for the given transition.

    :param connection: Transition to generate tooltip text for
    :type connection: WhdbConnectionP
    :return: Tooltip text for the given transition
    :rtype: str
    """
    return wgui_connection_tooltip('&mdash;', connection, sname, ename)


def wgui_directed_connection_tooltip(connection, sname=None, ename=None):
    """
    Generate the tooltip text for the given transition.

    :param connection: Transition to generate tooltip text for
    :type connection: WhdbConnectionP
    :return: Tooltip text for the given transition
    :rtype: str
    """
    return wgui_connection_tooltip(
        WGUI_CONNECTION_ARROW, connection, sname, ename)


def wgui_format_tooltip_section(title, items):
    """
    Create the tooltip text for the given tooltip section.

    :param title: Section title
    :type title: str
    :param items: Section items
    :type items: Iterable[str]
    :return: Tooltip text for the given section
    :rtype: str
    """
    return '<hr/><b>{}:</b><br/>{}<br/>'.format(
        title,
        '<br/>'.join('&nbsp;&#8226; {}'.format(s) for s in items))
