# ----------------------------------------------------------------------------
# - wgui-verifier                                                            -
# - wild cruncher - cockpit: database verification dialog                    -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_verifier
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher database verification dialog

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt4.QtCore import Qt
from PyQt4.QtGui import QColor

from wcr_core.action.wact_action_api import WactActionObserverP, \
    WACT_ACTION_EXECUTION_FAILED_EVENT
from wcr_core.actions.wact_verify_db_action import \
    WactDatabaseVerificationException
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP
from wcr_frontend.gui.wgui_db_elements_dialog import WguiElementsDialogBase
from wcr_frontend.gui.wgui_highlight_module import WguiHighlightModule
from wcr_core.project.wprj_project_verifier import WprjProjectVerifier
from wcr_frontend.gui.wgui_project_editor import WguiProjectEditor


class WguiVerifyDatabaseDialog(WguiElementsDialogBase):
    """
    Database verification dialog.
    """

    def __init__(self, projectDB, scene, verifier,
                 closeIfOk=False, result=None, parent=None):
        """
        :param projectDB: Project Database
        :type projectDB: WprjProjectDatabase
        :param scene: Graphic scene
        :type scene: WguiGraphicsScene
        :param verifier: Database verifier to use
        :type verifier: WprjProjectVerifier
        :param closeIfOk: Whether to close the verification dialog once
                          verification succeeds
        :type closeIfOk: bool
        :param result: Initial result set to display
        :type result: WprjVerificationResult
        :param parent: Parent widget
        """
        super().__init__(projectDB, scene,
                         runBtnLbl='Verify Database',
                         closeIfOk=closeIfOk,
                         items=self._resultToItemList(result),
                         parent=parent)

        #: The current project database
        self.__projectDB = projectDB
        #: The graphics scene
        self.__scene = scene
        #: Database verifier
        self.__verifier = verifier

        #: Highlighting module for hierarchy errors
        self.__highlighting = _WguiErrorHighlightingModule(scene)
        if self.__highlighting:
            scene.viewContext.highlighter.addModule(self.__highlighting)
            scene.viewContext.highlighter.setPrimaryModule(
                self.__highlighting.identifier)

        self.setWindowTitle('Database Verification')

        if result:
            self.__highlighting.result = result.hierarchyVerificationResult
        else:
            self._run()

    def _cleanUp(self):
        self.__scene.viewContext.highlighter.removeModule(
            self.__highlighting.identifier)
        super()._cleanUp()

    def _doRun(self):
        result = self.__verifier.verifyProjectDatabase(self.__projectDB)
        self.__highlighting.result = result.hierarchyVerificationResult
        return self._resultToItemList(result)

    @staticmethod
    def _resultToItemList(result):
        """
        :param result: Initial result set to display
        :type result: WprjVerificationResult
        """
        if not result or not result.hasErrors():
            return []

        pdbErrs = [(e.errorMessage, [], None, e)
                   for e in result.projectDbErrors]
        hrchErrs = [(e.errorMessage, e.offendingEntities, e.path, e)
                    for e in result.hierarchyVerificationResult.allErrors]
        # noinspection PyTypeChecker
        return pdbErrs + hrchErrs


class _WguiErrorHighlightingModule(WguiHighlightModule):
    """
    Highlighting module for hierarchy errors.
    """

    #: Colour for entities without errors
    CLEAN_COLOUR = Qt.green

    #: Colour for entities with errors
    ERROR_COLOUR = Qt.red

    #: Colour for entities with children with errors
    DEEP_ERROR_COLOUR = QColor(255, 165, 0)

    def __init__(self, scene):
        """
        :param scene: Graphic scene
        :type scene: WguiGraphicsScene
        """
        #: Graphics scene
        self.__scene = scene
        #: Current hierarchy verification result
        self.__result = None

    @property
    def result(self):
        """
        Get the current hierarchy verification result.

        :return: The current hierarchy verification result
        :rtype: WcmnVerificationResult
        """
        return self.__result

    @result.setter
    def result(self, result):
        """
        Set the current hierarchy verification result.

        :param result: The current hierarchy verification result
        :type result: WcmnVerificationResult
        """
        self.__result = result
        self.__scene.update()

    @property
    def identifier(self):
        return 'hierarchy-verification-errors'

    @property
    def description(self):
        return 'Highlight errors found during database verification'

    def highlight(self, item):
        result = self.result
        if result is None: return None

        if result.errorsForEntity(item):
            return self.ERROR_COLOUR, 3, None
        elif isinstance(item, WhdbInstanceP) and result.hasDeepErrors(item):
            return self.DEEP_ERROR_COLOUR, 3, None
        else:
            return self.CLEAN_COLOUR, 1, None

    def providesHighlight(self, hType):
        return hType == self.HLGHT_LINE_COLOUR or \
            hType == self.HLGHT_LINE_WIDTH


class WguiDbVerificationFailureHandler(WactActionObserverP):
    """
    Project verification failure handler.
    """

    def handle(self, projectCtx, scene, verifier, result, parent):
        """
        Handle a project verification failure.

        :param projectCtx: Project context
        :type projectCtx: WcorProjectContext
        :param scene: Scene
        :type scene: WguiGraphicsScene
        :param verifier: Verifier
        :type verifier: WprjProjectVerifier
        :param result: Verification result
        :type result: WprjVerificationResult
        :param parent: Parent widget
        """
        dialog = WguiVerifyDatabaseDialog(
            projectCtx.getProjectDB(), scene, verifier, result=result,
            parent=parent)
        dialog.show()
        dialog.exec_()

    def actionNotify(self, actx, action, event, *args, **kwargs):
        """
        :type actx: WguiActionExecutionContextP
        """
        if event == WACT_ACTION_EXECUTION_FAILED_EVENT and args:
            excp = args[0]
            if isinstance(excp, WactDatabaseVerificationException):
                projectCtx = actx.getProjectContext()
                editor = actx.getEditor()

                if projectCtx and isinstance(editor, WguiProjectEditor):
                    self.handle(projectCtx, editor.scene, excp.getVerifier(),
                                excp.getResult(), editor)
