# ----------------------------------------------------------------------------
# - wgui-widgets                                                             -
# - wild cruncher: general UI widgets                                        -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 clemens oertel                                   -
# - copyright (c) 2013-2013 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_shape
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general UI widgets

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import Counter
from PyQt4 import QtCore
from PyQt4.QtCore import Qt, QPointF, pyqtSignal
from PyQt4.QtGui import QGraphicsTextItem, QGraphicsItem, QCursor, \
    QGraphicsSceneMouseEvent, QPainter, QStyle, \
    QStyleOptionGraphicsItem, QPen, QKeyEvent, QTextCursor, QTextBlockFormat, \
    QGraphicsSimpleTextItem, QBrush, QFocusEvent, QDoubleSpinBox, QCheckBox, \
    QHBoxLayout, QWidget, QRadioButton, QTextEdit, QFont, QVBoxLayout, \
    QPushButton, QLineEdit, QTextCharFormat, QColor, QLabel

from wcr_frontend.gui.wgui_protocols import QObject_Meta_M, WguiChangeNotifyP
from wcr_frontend.gui.wgui_utils import WGUI_SELECTION_COLOUR, isMainThread
from wcr_core.db_api.whdb_protocols_api import WHDB_VALIGN_BOTTOM, \
    WHDB_VALIGN_MIDDLE, WHDB_HALIGN_RIGHT, WHDB_HALIGN_CENTER, \
    WHDB_VALIGN_TOP, WHDB_HALIGN_LEFT
from silkan_common.scmn_process import ScmnProcStreamHandlerP
from silkan_common.scmn_log import ScmnMessageTargetP, SCMN_LOG_ERROR
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError


class _WguiEditableTextInputField(QGraphicsTextItem):
    """
    Editable text field for QGraphicView.
    """

    #: Signal emitted whenever editing is finished and the text has changed
    textChanged = QtCore.pyqtSignal(str)

    #: Signal emitted whenever the field enters/leaves editing mode
    editingStateChanged = QtCore.pyqtSignal(bool)

    def __init__(self, text='', hAlign=None, multiline=False, allowEmpty=False,
                 parent=None, scene=None):
        """
        :param text: Initial text
        :type text: str
        :param hAlign: Horizontal alignment
        :type hAlign: int
        :param multiline: Whether to allow multiple lines
        :type multiline: bool
        :param allowEmpty: Whether to allow empty text
        :type allowEmpty: bool
        :parent: Parent graphics item
        :scene: Scene this item will be added to
        """
        super().__init__(text, parent, scene)

        #: Horizontal alignment
        self.__hAlign = hAlign if hAlign is not None else WHDB_HALIGN_LEFT
        #: Whether line breaks are allowed
        self.__multiline = multiline
        #: Whether empty text is allowed
        self.__allowEmpty = allowEmpty

        self.__orgText = text
        self.__realigning = False

        self.setTextInteractionFlags(Qt.NoTextInteraction)
        self.setFlag(QGraphicsItem.ItemIsFocusable)

        self.document().contentsChange.connect(self._realignText)

        self.text = text

    @property
    def text(self):
        """
        Get the current text contents.

        :return: Text contents
        :rtype: str
        """
        return self.toPlainText()

    @text.setter
    def text(self, text):
        """
        Set the fields text content.

        :param text: The new text content
        :type text: str
        """
        self.setPlainText(text)
        self._realignText()

    @property
    def horizontalAlign(self):
        """
        Get the field's horizontal alignment.

        Return value should be one of
        :py:data:`whds_protocols.WHDS_HALIGN_LEFT`,
        :py:data:`whds_protocols.WHDS_HALIGN_CENTER` or
        :py:data:`whds_protocols.WHDS_HALIGN_RIGHT`.

        :return: The field's horizontal alignment.
        :rtype: int
        """
        return self.__hAlign

    @horizontalAlign.setter
    def horizontalAlign(self, hAlign):
        """
        Set the field's horizontal alignment.

        Value should be one of
        :py:data:`whds_protocols.WHDS_HALIGN_LEFT`,
        :py:data:`whds_protocols.WHDS_HALIGN_CENTER` or
        :py:data:`whds_protocols.WHDS_HALIGN_RIGHT`.

        :param hAlign: The new horizontal alignment
        :type hAlign: int
        """
        self.__hAlign = hAlign
        self._realignText()

    @property
    def margin(self):
        """
        Get the field's document margin.

        :return: Document margin
        :rtype: float
        """
        return self.document().documentMargin()

    def _realignText(self):
        if self.__realigning: return
        self.__realigning = True

        self.setTextWidth(-1)
        self.setTextWidth(self.boundingRect().width())

        if self.document().lineCount() <= 1:
            self.__realigning = False
            return

        txtFormat = QTextBlockFormat()
        if self.__hAlign == WHDB_HALIGN_RIGHT:
            txtFormat.setAlignment(Qt.AlignRight)
        elif self.__hAlign == WHDB_HALIGN_CENTER:
            txtFormat.setAlignment(Qt.AlignHCenter)
        else:
            txtFormat.setAlignment(Qt.AlignLeft)
        cursor = self.textCursor()
        position = cursor.position()
        cursor.select(QTextCursor.Document)
        cursor.mergeBlockFormat(txtFormat)
        cursor.clearSelection()
        cursor.setPosition(position)
        self.setTextCursor(cursor)

        self.__realigning = False

    @property
    def isEditing(self):
        """
        Whether this text item is currently in editing mode.

        :return: True iff the text item is currently in editing mode
        :rtype: bool
        """
        flags = self.textInteractionFlags()
        return flags == Qt.TextEditorInteraction

    def _toggleEditingMode(self, editing):
        """
        Turn the text item's editing mode on or off.
        In editing mode, the text item will absorb mouse clicks and be thus
        only movable when the outline stroke is being dragged.

        :param editing: Whether to turn the editing mode on or off
        :type editing: bool
        """
        if self.isEditing == editing: return
        from wcr_core.modules.sys_map.db.wchdb_top_block import \
            WchdbTopHierarchy
        # can't edit items in system mapper view
        if isinstance(self.scene().topLevelModelBlock(), WchdbTopHierarchy):
            return

        if editing:
            self.__orgText = self.text
            self.setTextInteractionFlags(Qt.TextEditorInteraction)
            if not self.hasFocus():
                self.setFocus(Qt.OtherFocusReason)
            self.setCursor(QCursor(Qt.IBeamCursor))
        else:
            self.setTextInteractionFlags(Qt.NoTextInteraction)
            self.setFlag(QGraphicsItem.ItemIsFocusable)
            cursor = self.textCursor()
            cursor.clearSelection()
            self.setTextCursor(cursor)
            self.unsetCursor()
            if self.hasFocus():
                self.clearFocus()
            if not (self.text or self.__allowEmpty):
                self.text = self.__orgText
            # verify validity
            if not isNameValid(self.text):
                showErrorWindow("Name Error", NameFormatMessageError, self.text)
                self.text = self.__orgText
            if self.text != self.__orgText:
                self.textChanged.emit(self.text)
            # verify unicity
            from wcr_frontend.gui.wgui_scene_item import WguiOperatorInstanceSceneItem
            allEqpNames = [item.getDataObject().getName() for item in self.scene().getSceneItems()
                           if isinstance(item, WguiOperatorInstanceSceneItem)]
            counter = Counter(allEqpNames)
            commons = counter.most_common()
            if commons[0][1] > 1:
                showErrorWindow("Duplicate Name Error",
                                "The name {} is already used for another equipement!",
                                self.__orgText)
                self.text = self.__orgText
                self.textChanged.emit(self.text)

        self.editingStateChanged.emit(editing)

    def _getCursorPos(self, pos):
        """
        Get the cursor position for a given position in item coordinates.

        :param pos: Position in item coordinates
        :type pos: QPointF
        :return: Cursor position corresponding to the given position
        :rtype: int
        """
        return self.document().documentLayout().hitTest(pos, Qt.FuzzyHit)

    def startEditingAt(self, pos):
        """
        Put the field in editing mode, and place the cursor near the given
        position.

        :param pos: Position in item coordinates to place the cursor
        :type pos: QPointF
        """
        self.setFocus(Qt.MouseFocusReason)
        self._toggleEditingMode(True)
        cursor = self.textCursor()
        cursor.setPosition(self._getCursorPos(pos))
        self.setTextCursor(cursor)

    def mouseDoubleClickEvent(self, event:QGraphicsSceneMouseEvent):
        """
        Put the field into editing mode on double clicks.
        """
        if self.isEditing:
            # In editing mode, pass double clicks on to the parent text item
            QGraphicsTextItem.mouseDoubleClickEvent(self, event)
        else:
            # When not in editing mode, use double click to trigger it
            self.startEditingAt(event.pos())

    def paint(self, painter:QPainter, options:QStyleOptionGraphicsItem,
              widget=None):
        painter.save()
        options.state &= ~(QStyle.State_Selected | QStyle.State_HasFocus)
        super().paint(painter, options, widget)
        painter.restore()

    def keyPressEvent(self, event:QKeyEvent):
        """
        Leave editing mode on ``Return`` presses unless the field is set to
        allow multiple lines.
        """
        if self.isEditing:
            if event.key() == Qt.Key_Return and \
                    (not self.__multiline or
                             event.modifiers() & Qt.ShiftModifier):
                self._toggleEditingMode(False)
                self.update()
                return
        super().keyPressEvent(event)

    def keyReleaseEvent(self, event:QKeyEvent):
        """
        Leave editing mode on ``ESC`` presses.
        """
        if self.isEditing:
            if event.key() == Qt.Key_Escape:
                self.text = self.__orgText
                self._toggleEditingMode(False)
                self.update()
                return
        super().keyReleaseEvent(event)

    def focusInEvent(self, event:QFocusEvent):
        """
        Enter the editing mode whenever the focus is gained explicitly.
        """
        if event.reason() not in [Qt.TabFocusReason, Qt.OtherFocusReason]:
            return
        self._toggleEditingMode(True)
        super().focusInEvent(event)

    def focusOutEvent(self, event:QFocusEvent):
        """
        Leave editing mode whenever the focus is lost.
        """
        self._toggleEditingMode(True)
        super().focusOutEvent(event)


class WguiEditableText(QGraphicsItem):
    """
    Editable text field widget.

    The widget supports horizontal (left, center, right) and vertical text
    alignment (top, middle, bottom).
    The widget can display a placeholder text if no content has been entered
    yet by the user.
    """

    def __init__(self, text='', vAlign=None, hAlign=None,
                 multiline=False, allowEmpty=False,
                 placeHolder='', editingPlaceHolder='',
                 fixRefPoint=False, parent=None, scene=None):
        """
        :param text: Initial text
        :type text: str
        :param vAlign: Vertical alignment
        :type vAlign: int
        :param hAlign: Horizontal alignment
        :type hAlign: int
        :param multiline: Whether to allow multiple lines
        :type multiline: bool
        :param allowEmpty: Whether to allow empty text
        :type allowEmpty: bool
        :param placeHolder: Placeholder text to show when text field is empty
        :type placeHolder: str
        :param editingPlaceHolder: Placeholder text to show when text field is
                                   empty and being edited (otherwise
                                   `placeHolder` is shown)
        :type editingPlaceHolder: str
        :param fixRefPoint: Whether the item's position applies to the text
                            only (`False`), or to the boundary of its visible
                            elements, including the editing box (`True`)
        :type fixRefPoint: bool
        :parent: Parent graphics item
        :scene: Scene this item will be added to
        """
        super().__init__(parent, scene)

        #: Vertical alignment
        self.__vAlign = vAlign if vAlign is not None else WHDB_VALIGN_TOP
        #: Horizontal alignment
        self.__hAlign = hAlign if hAlign is not None else WHDB_HALIGN_LEFT
        #: Use text or visible bounding box as reference point
        self.__fixRefPoint = fixRefPoint
        #: Placeholder text
        self.__placeHolderText = placeHolder
        #: Placeholder text when editing
        self.__editPlaceHolderText = editingPlaceHolder
        #: Hide if empty
        self.__hideIfEmpty = False
        #: Explicit visibility
        self.__visible = True

        #: Anchor point
        self.__anchor = (0, 0)

        #: Text field
        self.__text = _WguiEditableTextInputField(text, vAlign,
                                                  multiline, allowEmpty,
                                                  self, scene)
        self.__text.setZValue(1)
        self.__text.document().contentsChange.connect(self.__contentChange)
        self.__text.document().contentsChanged.connect(self.__contentChanged)
        self.__text.editingStateChanged.connect(self.__editingStateChanged)

        if placeHolder or editingPlaceHolder:
            #: Placeholder field
            self.__placeHolder = QGraphicsSimpleTextItem(placeHolder, self,
                                                         scene)
            self.__placeHolder.setPos(self.__text.document().documentMargin(),
                                      self.__text.document().documentMargin())
            self.__placeHolder.setBrush(QBrush(Qt.lightGray))
            self.__placeHolder.setVisible(not bool(self.__text.text))
            self.__swapPlaceHolders = bool(editingPlaceHolder) and \
                                      placeHolder != editingPlaceHolder
        else:
            self.__placeHolder = None
            self.__swapPlaceHolders = False

        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges)
        self.setAcceptedMouseButtons(Qt.LeftButton)
        self.setFlag(QGraphicsItem.ItemIsFocusable)
        self.setFocusProxy(self.__text)

    #
    # PROPERTIES & PUBLIC METHODS
    #

    @property
    def textChanged(self):
        """
        Signal emitted whenever the field's text has changed when it leaves
        editing mode.
        """
        return self.__text.textChanged

    @property
    def text(self):
        """
        Get the current text contents.

        :return: Text contents
        :rtype: str
        """
        return self.__text.text

    @text.setter
    def text(self, text):
        """
        Set the text contents.

        :param text: New text contents
        :type text: str
        """
        self.prepareGeometryChange()
        self.__text.text = text
        self.__contentChange()
        if self.__hideIfEmpty:
            self.updateVisibility()

    @property
    def verticalAlign(self):
        """
        Get the field's vertical alignment.

        Return value should be one of
        :py:data:`whds_protocols.WHDS_VALIGN_TOP`,
        :py:data:`whds_protocols.WHDS_VALIGN_MIDDLE` or
        :py:data:`whds_protocols.WHDS_VALIGN_BOTTOM`.

        :return: The field's vertical alignment.
        :rtype: int
        """
        return self.__vAlign

    @verticalAlign.setter
    def verticalAlign(self, vAlign):
        """
        Set the field's vertical alignment.

        Value should be one of
        :py:data:`whds_protocols.WHDS_VALIGN_TOP`,
        :py:data:`whds_protocols.WHDS_VALIGN_MIDDLE` or
        :py:data:`whds_protocols.WHDS_VALIGN_BOTTOM`.

        :param vAlign: The new vertical alignment
        :type vAlign: int
        """
        self.prepareGeometryChange()
        self.__vAlign = vAlign
        self._adjustGeometry()

    @property
    def horizontalAlign(self):
        """
        Get the field's horizontal alignment.

        Return value should be one of
        :py:data:`whds_protocols.WHDS_HALIGN_LEFT`,
        :py:data:`whds_protocols.WHDS_HALIGN_CENTER` or
        :py:data:`whds_protocols.WHDS_HALIGN_RIGHT`.

        :return: The field's horizontal alignment.
        :rtype: int
        """
        return self.__hAlign

    @horizontalAlign.setter
    def horizontalAlign(self, hAlign):
        """
        Set the field's horizontal alignment.

        Value should be one of
        :py:data:`whds_protocols.WHDS_HALIGN_LEFT`,
        :py:data:`whds_protocols.WHDS_HALIGN_CENTER` or
        :py:data:`whds_protocols.WHDS_HALIGN_RIGHT`.

        :param hAlign: The new horizontal alignment
        :type hAlign: int
        """
        self.prepareGeometryChange()
        self.__hAlign = hAlign
        self.__text.horizontalAlign = hAlign
        self._adjustGeometry()

    @property
    def font(self):
        """
        Get the text field's font.

        :return: Font
        :rtype: QFont
        """
        return self.__text.font()

    @font.setter
    def font(self, font):
        """
        Set the text field's font.

        :param font: New font
        :type font: QFont
        """
        self.prepareGeometryChange()
        self.__text.setFont(font)
        if self.__placeHolder is not None:
            self.__placeHolder.setFont(font)

    @property
    def textColour(self):
        """
        Get the text field's text colour.

        :return: Text colour
        :rtype: QColor
        """
        return self.__text.defaultTextColor()

    @textColour.setter
    def textColour(self, colour):
        """
        Set the text field's text colour.

        :param colour: New text colour
        :type colour: QColor
        """
        self.__text.setDefaultTextColor(colour)

    @property
    def hideIfEmpty(self):
        """
        Whether the text item should be hidden if it has not text.

        :return: Whether hidden if empty
        :rtype: bool
        """
        return self.__hideIfEmpty

    @hideIfEmpty.setter
    def hideIfEmpty(self, hide):
        """
        Set whether the text item should be hidden if it has not text.

        :return: Whether hidden if empty
        :rtype: bool
        """
        self.__hideIfEmpty = hide

    def setVisible(self, visible):
        self.__visible = visible
        self.updateVisibility()

    def updateVisibility(self):
        """
        Update the items visibility, based on the parent's-visibility setting,
        the hideIfEmpty state and the text.
        """
        super().setVisible(
            self.__visible and (not self.__hideIfEmpty or bool(self.text)))

    #
    # POSITIONING AND ANCHORING
    #

    def setPos(self, *args):
        if len(args) == 1 and isinstance(args[0], QPointF):
            self.__anchor = args[0].x(), args[0].y()
        elif len(args) == 2 and isinstance(args[0], float) and \
                isinstance(args[1], float):
            self.__anchor = args[0], args[1]
        self._adjustGeometry()

    def setX(self, x):
        _, y = self.__anchor
        self.__anchor = x, y
        self._adjustGeometry()

    def setY(self, y):
        x, _ = self.__anchor
        self.__anchor = x, y
        self._adjustGeometry()

    def translate(self, dx, dy):
        x, y = self.__anchor
        self.__anchor = x + dx, y + dy
        self._adjustGeometry()

    def moveBy(self, dx, dy):
        x, y = self.__anchor
        self.__anchor = x + dx, y + dy
        self._adjustGeometry()

    def pos(self):
        x, y = self.__anchor
        return QPointF(x, y)

    def x(self):
        return self.__anchor[0]

    def y(self):
        return self.__anchor[1]

    def _anchorAdjustments(self):
        """
        Calculate the position of the anchor relative to the items position.
        """
        brct = self.boundingRect()
        margin = self.__text.margin
        editing = self.__text.isEditing

        offset = 0 if self.__fixRefPoint and not editing else margin
        if self.__vAlign == WHDB_VALIGN_BOTTOM:
            dy = - brct.height() - offset
        elif self.__vAlign == WHDB_VALIGN_MIDDLE:
            dy = - brct.height() / 2
        else:
            dy = offset

        offset = 0 if self.__fixRefPoint and editing else margin - 1
        if self.__hAlign == WHDB_HALIGN_RIGHT:
            dx = - brct.width() + offset
        elif self.__hAlign == WHDB_HALIGN_CENTER:
            dx = - brct.width() / 2
        else:
            dx = - offset

        return dx, dy

    def _adjustGeometry(self):
        """
        To be called whenever the item's geometry has changed.

        Sets the anchor based on the text field's alignment.
        """
        self.prepareGeometryChange()

        ax, ay = self.__anchor
        dx, dy = self._anchorAdjustments()

        super().setPos(ax + dx, ay + dy)

    #
    # EVENT HANDLING
    #

    def mouseDoubleClickEvent(self, event:QGraphicsSceneMouseEvent):
        """
        Put the field into editing mode on double clicks.
        """
        self.__text.startEditingAt(event.pos())

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionHasChanged:
            dx, dy = self._anchorAdjustments()
            self.__anchor = value.x() - dx, value.y() - dy
        return super().itemChange(change, value)

    def installSceneEventFilter(self, filterItem):
        self.__text.installSceneEventFilter(filterItem)

    def removeSceneEventFilter(self, filterItem):
        self.__text.removeSceneEventFilter(filterItem)

    #
    # SIGNALS & SLOTS
    #

    # noinspection PyUnusedLocal
    def __editingStateChanged(self, editing):
        """
        Slot to be called whenever the delegate's editing state changes.
        """
        try:
            if editing:
                self.setSelected(True)
            if self.__swapPlaceHolders:
                self.prepareGeometryChange()
                if editing:
                    self.__placeHolder.setText(self.__editPlaceHolderText)
                else:
                    self.__placeHolder.setText(self.__placeHolderText)
                self._contentChanged()
            self._adjustGeometry()
            self.update()
        except RuntimeError:
            pass

    # noinspection PyUnusedLocal
    def __contentChange(self, *args):
        """
        Slot to be called whenever the delegate's text content changes.
        """
        if self.__placeHolder is not None:
            vis = not bool(self.__text.text)
            if vis != self.__placeHolder.isVisible():
                self.prepareGeometryChange()
            self.__placeHolder.setVisible(vis)

    def __contentChanged(self):
        """
        Slot to be called whenever the delegate's text content has changed.
        """
        try:
            self._adjustGeometry()
            self._contentChanged()
        except RuntimeError:
            pass

    def _contentChanged(self):
        """
        Slot to be called whenever the delegate's text content changes.
        """
        pass

    #
    # PAINTING
    #

    def paint(self, painter:QPainter, options, widget=None):
        if self.__text.isEditing:
            painter.save()
            painter.setPen(QPen(WGUI_SELECTION_COLOUR))
            painter.fillRect(self.boundingRect(), Qt.white)
            painter.drawRect(self.boundingRect())
            painter.restore()

    def boundingRect(self):
        dtxt = self.__placeHolder
        if dtxt is not None and dtxt.isVisible():
            dRect = self.__placeHolder.boundingRect()
            dRect = self.mapFromItem(self.__placeHolder, dRect).boundingRect()
            dRect.adjust(0, 0, 3, 0)
            return self.__text.boundingRect().united(dRect)
        else:
            return self.__text.boundingRect()

    @property
    def isEditing(self):
        """
        Whether the text widget is currently in editing mode.

        :return: True iff the widget is currently in editing mode
        :rtype: bool
        """
        return self.__text.isEditing

    def hasFocus(self):
        return self.isEditing


class WguiRadioBoxGroup(QWidget):
    """
    Multi-value radio box group.
    """

    def __init__(self, *statesLabels):
        """
        :param statesLabels: 2-tuples of state/label pairs
        :type statesLabels: (object, str)
        """
        super().__init__()

        #: Radio button/state pairs
        self.__radioButtons = []

        layout = QHBoxLayout()

        for state, label in statesLabels:
            rb = QRadioButton(label)
            layout.addWidget(rb)
            self.__radioButtons.append((rb, state))

        self.setLayout(layout)

    def setState(self, state):
        """
        Set the current state of the group, i.e. check the radio button
        associated with the given state.

        :param state: Current state
        :type state: object
        """
        for rb, s in self.__radioButtons:
            if s == state:
                rb.setChecked(True)
                break

    def state(self):
        """
        Get the current state, i.e. the state associated with the currently
        checked radio button.

        :return: Current state
        :rtype: object
        """
        for rb, s in self.__radioButtons:
            if rb.isChecked():
                return s

    @staticmethod
    def mkTrueFalseDefaultBoxes(trueLabel='True', falseLabel='False',
                                defaultLabel='Default'):
        """
        Generate a tri-state radio button group (True, False, Default/None).

        :param trueLabel: Label for 'True' radio box
        :type trueLabel: str
        :param falseLabel: Label for 'False' radio box
        :type falseLabel: str
        :param defaultLabel: Label for default radio box
        :type defaultLabel: str
        :return: Radio box group with 3 radio boxes
        :rtype: WguiRadioBoxGroup
        """
        return WguiRadioBoxGroup((True, trueLabel),
                                 (False, falseLabel),
                                 (None, defaultLabel))


class WguiDefaultDoubleSpinBox(QWidget):
    """
    Spin-box widget for double values with 'Default' checkbox next to it.
    Checking the 'Default' checkbox will disable the spin box.
    """

    def __init__(self):
        super().__init__()

        layout = QHBoxLayout()

        #: Input field
        self.__input = QDoubleSpinBox()
        #: 'Default' checkbox
        self.__default = QCheckBox('Default')
        self.__default.toggled.connect(self._toggleInputEnabled)

        layout.addWidget(self.__input)
        layout.addWidget(self.__default)

        self.setLayout(layout)

        self.setFocusPolicy(Qt.StrongFocus)
        self.setFocusProxy(self.__input)

    def _toggleInputEnabled(self, disabled):
        """
        Disable the input field whenever the 'Default' checkbox is checked.

        :param disabled: Whether to disable the input field
        :type disabled: bool
        """
        self.__input.setEnabled(not disabled)

    def setValue(self, value, enabled):
        """
        Set the current value.

        :param value: Current value
        :type value: float
        :param enabled: Whether the input should be enabled, i.e. the 'Default'
                        checkbox not be checked
        :type enabled: bool
        """
        self.__default.setChecked(not enabled)
        self.__input.setValue(value)

    def value(self):
        """
        Get the current value.
        If the input field is currently disabled, `None` will be returned.

        :return: Current value, or `None`
        :rtype: float
        """
        if self.__default.isChecked():
            return None
        else:
            return self.__input.value()


class WguiCommandOutput(QWidget, ScmnProcStreamHandlerP,
                        ScmnMessageTargetP, WguiChangeNotifyP,
                        metaclass=QObject_Meta_M):
    """
    Widget to display command output.
    """

    #: Signal emitted when the content has changed
    __contentChanged = pyqtSignal()

    #: Signal emitted when a new line is to be added
    __lineToAdd = pyqtSignal(str, QColor)

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Text display widget
        self.__textEdit = QTextEdit()
        #: Search widget
        self.__searchWidget = WguiTextSearchWidget(self.__textEdit)
        #: Search result label
        self.__searchResults = QLabel()

        self._initComponents()
        self.setLayout(self._createLayout())

        self.__lineToAdd.connect(self._addLine)

    def _initComponents(self):
        """
        Initialize all UI components as needed.
        """
        # attach a courier like font
        font = QFont()
        font.setFamily('Courier')
        font.setFixedPitch(True)
        font.setPointSize(10)
        self.__textEdit.setFont(font)
        self.__textEdit.setReadOnly(True)

        self.__searchWidget.highlightingChanged.connect(
            self._highlightingChanged)

    def _createLayout(self):
        """
        Create the layout for the widget.

        :return: Layout for the widget
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setMargin(5)
        layout.setSpacing(2)

        layout.addWidget(self.__textEdit)
        layout.setStretch(0, 1)

        btmLayout = QHBoxLayout()
        btmLayout.setMargin(0)
        btmLayout.setSpacing(0)
        btmLayout.setAlignment(Qt.AlignVCenter)
        btmLayout.addWidget(self.__searchWidget)
        btmLayout.addSpacing(10)
        btmLayout.addWidget(self.__searchResults)
        btmLayout.addStretch(1)

        layout.addLayout(btmLayout)

        return layout

    def _highlightingChanged(self, count, strm):
        """
        Callback to handle highlight changes.

        :param count: Number of elements highlighted/fonud
        :type count: int
        :param strm: Current search term
        :type strm: str
        """
        if count < 0:
            self.__searchResults.clear()
        elif count == 1:
            self.__searchResults.setText(
                'Found 1 occurrence of {s}'.format(c=count, s=strm))
        else:
            self.__searchResults.setText(
                'Found {c} occurrences of {s}'.format(c=count, s=strm))

    def _addLine(self, line, colour):
        """
        Add a line in the given colour.

        :param line: Line to add
        :type line: str
        :param colour: Colour
        :type colour: QColor, int
        """
        docLen = self.__textEdit.document().characterCount()

        self.__textEdit.setTextColor(colour)
        self.__textEdit.append(line)

        vscb = self.__textEdit.verticalScrollBar()
        vscb.setValue(vscb.maximum())

        cursor = QTextCursor(self.__textEdit.document())
        cursor.setPosition(docLen)
        self.__searchWidget.highlight(cursor)
        self.__contentChanged.emit()

    def addLine(self, line, colour):
        """
        Add a line in the given colour.

        Gateway function for _addLine to ensure all modifying access to the
        text editor happens on the right thread.

        :param line: Line to add
        :type line: str
        :param colour: Colour
        :type colour: QColor, int
        """
        if isMainThread():
            self._addLine(line, colour)
        else:
            self.__lineToAdd.emit(line, colour)

    def addMessage(self, mtxt, mcmp=None, mcat=None, mlvl=None, trce=None):
        if mcat == SCMN_LOG_ERROR:
            self.addLine(mtxt, Qt.red)
        elif mcat == SCMN_LOG_ERROR:
            self.addLine(mtxt, Qt.blue)
        else:
            self.addLine(mtxt, Qt.black)

    def handleControlMsg(self, text, msgType):
        if msgType == self.CTRL_FAILED:
            self.addLine(text, Qt.darkRed)
        else:
            self.addLine(text, Qt.darkGray)

    def handleOutput(self, text):
        """
        Add a line of regular output.

        :param text: Line to add
        :type text: str
        """
        self.addLine(text, Qt.black)

    def handleWarning(self, text):
        """
        Add a line of warning output.

        :param text: Line to add
        :type text: str
        """
        self.addLine(text, Qt.blue)

    def handleError(self, text):
        """
        Add a line of error output.

        :param text: Line to add
        :type text: str
        """
        self.addLine(text, Qt.red)

    def clear(self):
        """
        Clear all output currently displayed.
        """
        self.__textEdit.clear()

    @property
    def contentChanged(self):
        return self.__contentChanged


class WguiTextSearchWidget(QWidget):
    """
    Text-search widget for editors.
    """

    #: Signal to indicate number of current search term highlightings
    highlightingChanged = pyqtSignal(int, str)

    def __init__(self, textEdit, parent=None):
        """
        :param textEdit: Text edit widget to search
        :type textEdit: QTextEdit
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Text editor to search
        self.__textEdit = textEdit

        #: Current search term
        self.__searchTerm = None
        #: Previous search term
        self.__prevSearchTerm = None

        #: Search field
        self.__searchF = _WguiSearchField()
        #: Find button
        self.__findBtn = QPushButton()

        self._initComponents()
        self.setLayout(self._createLayout())

        self.setFocusProxy(self.__searchF)

    def _initComponents(self):
        """
        Initialize all UI components as needed.
        """
        self.__searchF.setPlaceholderText('Find text ...')
        self.__searchF.textChanged.connect(self._setSearchTerm)
        self.__searchF.returnPressed.connect(self.findNext)

        self.__findBtn.setText('Find')
        self.__findBtn.clicked.connect(self.findNext)

    def _createLayout(self):
        """
        Create the widget's layout.
        Should be called only once.

        :return: Layout for the widget
        :rtype: QLayout
        """
        layout = QHBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(5)
        layout.addWidget(self.__searchF)
        layout.addWidget(self.__findBtn)
        return layout

    def _setSearchTerm(self, strm):
        """
        Set the current search term.
        Highlights all occurrences of the search term in the editor.

        :param strm: New search term
        :type strm: str
        """
        ptrm = self.__searchTerm
        self.__searchTerm = strm

        if ptrm:
            self._clearHighlight()

        self.highlight()

    def findNext(self):
        """
        Find the first occurrence of the current search term in the editor's
        document, select it and give the focus to the editor.
        """
        if not self.__searchTerm: return
        edit = self.__textEdit
        if self.__searchTerm != self.__prevSearchTerm:
            edit.moveCursor(QTextCursor.Start)
            self.__prevSearchTerm = self.__searchTerm
        edit.find(self.__searchTerm)
        edit.setFocus(Qt.OtherFocusReason)

    def findAgain(self):
        """
        Find the next occurrence of the current search term in the editor's
        document and select it.
        """
        self.__prevSearchTerm = self.__searchTerm
        self.findNext()

    def _clearHighlight(self):
        """
        Clear all current highlighting.
        """
        document = self.__textEdit.document()
        cursor = QTextCursor(document)
        cursor.select(QTextCursor.Document)
        colorFormat = QTextCharFormat()
        colorFormat.setBackground(QBrush(Qt.white))
        cursor.mergeCharFormat(colorFormat)

    def highlight(self, cursor=None):
        """
        Highlight all occurrences of the current search term in the editor.
        """
        strm = self.__searchTerm
        if strm:
            count = 0
            document = self.__textEdit.document()
            if cursor is None:
                cursor = QTextCursor(document)
            colorFormat = QTextCharFormat()
            colorFormat.setBackground(QBrush(Qt.yellow))
            while not cursor.isNull() and not cursor.atEnd():
                cursor = document.find(strm, cursor)

                if not cursor.isNull():
                    cursor.mergeCharFormat(colorFormat)
                    count += 1

            self.highlightingChanged.emit(count, strm)
        else:
            self.highlightingChanged.emit(-1, '')


class _WguiSearchField(QLineEdit):
    """
    Text-search field for editors.

    Clears itself on `ESC`.
    """

    def __init__(self, *__args):
        super().__init__(*__args)

    def keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.clear()
        super().keyReleaseEvent(event)
