# ----------------------------------------------------------------------------
# - wgui-workspace-tree                                                      -
# - wild cruncher: workspace tree display                                    -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2013 clemens oertel                                   -
# - copyright (c) 2013-2013 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_workspace_tree
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher workspace tree display

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from PyQt5 import QtCore
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QHBoxLayout, QWidget, QVBoxLayout, QTreeWidget, \
	QTreeWidgetItem, QPushButton

from silkan_common.scmn_mime_type import scmn_mime_types
from silkan_common.scmn_file_path import scmn_isdir, scmn_listdir, \
    scmn_path_join
from wcr_core.action.wact_action_api import WactActionObserverP, \
    WACT_ACTION_EXECUTED_EVENT
from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_frontend.gui.wgui_protocols import QWidget_Meta_M


class WguiWorkspaceTree(QWidget, WactActionObserverP, WcmnReleasableP,
                        metaclass=QWidget_Meta_M):
    """
    Workspace tree widget.
    """

    #: Signal emitting whenever a file entry is double-clicked
    fileDoubleClicked = QtCore.pyqtSignal(str)

    def __init__(self, pctx, cctx, actions=None, parent=None):
        """
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param cctx: Cockpit system context
        :type cctx: WguiCockpitSystemContextP
        :param actions: Actions to listen on for refresh
        :type actions: Iterable[WactActionP] | None
        :param parent: Parent widget
        """
        super().__init__(parent)

        self.__prjContext = pctx
        self.__cctx = cctx
        self.__actions = list(actions) if actions else []

        #: Tree widget
        self.__tree = QTreeWidget()
        self.__tree.doubleClicked.connect(self._doubleClicked)

        self.setLayout(self._createLayout())

        for action in self.__actions:
            action.registerActionObserver(self)

        self.refresh()

    def release(self):
        for action in self.__actions:
            action.unregisterActionObserver(self)
        self.__actions = []
        self.__prjContext = None
        super().release()

    def _createLayout(self):
        """
        Create the layout for the workspace tree browser.

        :return: Workspace tree layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.addWidget(self.__tree)

        refreshButton = QPushButton('Refresh')
        refreshButton.clicked.connect(self.refresh)
        expAllButton = QPushButton('Expand All')
        expAllButton.clicked.connect(self.__tree.expandAll)
        cllAllButton = QPushButton('Collapse All')
        cllAllButton.clicked.connect(self.__tree.collapseAll)
        btnLayout = QHBoxLayout()
        btnLayout.addStretch(1)
        btnLayout.addWidget(refreshButton)
        btnLayout.addWidget(expAllButton)
        btnLayout.addWidget(cllAllButton)
        layout.addLayout(btnLayout)

        return layout

    def _doubleClicked(self, idx):
        """
        Callback for when an item was double-clicked.

        :param idx: Model index of item that was clicked on
        :type idx: QModelIndex
        """
        if idx is None: return

        path = idx.model().data(idx, Qt.UserRole)
        if path:
            self.fileDoubleClicked.emit(path)

    def refresh(self):
        """
        Clear and rebuild the tree.
        """
        pctx = self.__prjContext
        if pctx is None:
            return

        tree = self.__tree
        tree.clear()

        tree.setColumnCount(2)
        tree.setHeaderLabels(['Name', 'Type'])
        tree.setAlternatingRowColors(True)
        #self.__tree.headerItem().setFirstColumnSpanned(True)

        workspace = pctx.getProjectWorkspace()
        projectName = pctx.getProjectName()
        rootDir = workspace.getProjectPath()
        if not scmn_isdir(rootDir): return

        item = QTreeWidgetItem([projectName or 'Untitled', 'project'])
        tree.addTopLevelItem(item)
        self._addDirectory(rootDir, item)

        tree.expandAll()
        tree.resizeColumnToContents(0)

    def actionNotify(self, actx, action, event, *args, **kwargs):
        """
        :type actx: WguiActionExecutionContextP
        """
        pctx = self.__prjContext
        if pctx is None:
            return

        actionDb = actx.getProjectContext().getProjectDB()
        ownDb = pctx.getProjectDB()

        if actionDb is not ownDb:
            return
        if event & WACT_ACTION_EXECUTED_EVENT:
            print('WORKSPACE REFRESH')
            self.refresh()

    def _addDirectory(self, path, parent, level=0):
        """
        Add a directory to the tree, recursively.

        :param path: Path of the directory to add
        :type path: str
        :param parent: Parent tree item
        :type parent: QTreeWidgetItem
        :param level: Depth of the current directory
        :type level: int
        """

        def dirFilter(name):
            return scmn_isdir(scmn_path_join(path, name))

        entries = set(scmn_listdir(path))
        dirs = list(filter(dirFilter, entries))
        for sub in dirs:
            dirType = self.__cctx.getDirectoryDescription(level)
            item = QTreeWidgetItem([sub, dirType])
            parent.addChild(item)
            self._addDirectory(scmn_path_join(path, sub), item, level + 1)

        for file in entries.difference(dirs):
            mimes = scmn_mime_types(file)
            item = QTreeWidgetItem([file,
                                    mimes[0].description if mimes else ''])
            item.setData(0, Qt.UserRole, scmn_path_join(path, file))
            parent.addChild(item)
