# ----------------------------------------------------------------------------
# - wgui-gui-install-context                                                 -
# - wild cruncher: gui install context                 frontend / gui_plugin -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_gui_install_context
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - GUI install context

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.plugin.wplg_install_context import \
    WplgPlatformInstallContextBase, WplgModuleInstallContextBase
from wcr_core.plugin.wplg_install_context_api import WplgInstallContextP, \
    WplgModuleInstallContextP, WplgPlatformInstallContextP


#: GUI install context code
WGUI_GUI_INSTALL_CODE = 'gui'


# noinspection PyAbstractClass
class WguiGuiInstallContextMixin(WplgInstallContextP):
    def getComponentCode(self):
        return WGUI_GUI_INSTALL_CODE


# noinspection PyAbstractClass
class WguiGuiPlatformInstallContextP(WplgPlatformInstallContextP):
    """
    GUI platform install context protocol.
    """

    @abstractmethod
    def installAction(self, action):
        """
        Install an action.

        :param action: Action to be installed
        :type action: WactActionP
        """
        raise NotImplementedError()


class WguiGuiPlatformInstallContext(WplgPlatformInstallContextBase,
                                    WguiGuiInstallContextMixin,
                                    WguiGuiPlatformInstallContextP):
    def __init__(self, plgCode, plgName, plgDesc):
        """
        :param plgCode: Plugin code
        :type plgCode: str
        :param plgName: Plugin name
        :type plgName: str
        :param plgDesc: Plugin description
        :type plgDesc: str
        """
        super().__init__(plgCode, plgName, plgDesc)

        self.__actions = []
        """:type: list[WactActionP]"""

    def commit(self):
        #for action in self.__actions:
        #    raise NotImplementedError('Cannot instal')
        #self.__actions = []
        pass

    def installAction(self, action):
        #self.__actions.append(action)
        raise NotImplementedError('Cannot instal')


# noinspection PyAbstractClass
class WguiGuiModuleInstallContextP(WplgModuleInstallContextP):
    """
    GUI module install context protocol.
    """

    @abstractmethod
    def addSyntaxHighlighter(self, shlg, code):
        """
        Register a new syntax highlighter for text editors.

        :param shlg: Syntax highlighter (constructor)
        :type shlg: () -> QSyntaxHighlighter
        :param code: MIME-type of compatible files
        :type code: str
        """
        raise NotImplementedError()

    @abstractmethod
    def installAction(self, action, code, *flags):
        """
        Install an action.

        :param action: Action to be installed
        :type action: WactActionP
        :param code: Action code
        :type code: str
        :param flags: Action flags
        :type flags: tuple[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def registerIcon(self, icon, code, kind):
        """
        Register an icon.

        :param icon: Icon path
        :type icon: str
        :param code: Icon code, e.g. MIME-type or operator code
        :type code: str
        :param kind: Icon kind, e.g. `WGUI_ICON_MIME_KIND`
        :type kind: int
        """
        raise NotImplementedError()


class WguiGuiModuleInstallContext(WplgModuleInstallContextBase,
                                  WguiGuiInstallContextMixin,
                                  WguiGuiModuleInstallContextP):

    def __init__(self, plgCode, plgName, plgDesc, module):
        """
        :param plgCode: Plugin code
        :type plgCode: str
        :param plgName: Plugin name
        :type plgName: str
        :param plgDesc: Plugin description
        :type plgDesc: str
        :param module: GUI module
        :type module: WguiCockpitSystemModuleMP
        """
        super().__init__(plgCode, plgName, plgDesc)

        #: Module
        self.__module = module

        self.__stxHghlghtrs = []
        """:type: list[(() -> QSyntaxHighlighter, str)]"""
        self.__actions = []
        """:type: list[(WactActionP, str, list[str])]"""
        self.__icons = []
        """:type: list[(str, str, int)]"""

    def commit(self):
        pcod = self.getPluginCode()

        for action, code, flst in self.__actions:
            self.__module.addAction(action, code, pcod, *flst)
        self.__actions = []

        for icon, code, kind in self.__icons:
            self.__module.addIcon(icon, code, pcod, kind)

        for shlg, code in self.__stxHghlghtrs:
            self.__module.addSyntaxHighlighter(shlg, code)

    def addSyntaxHighlighter(self, shlg, code):
        self.__stxHghlghtrs.append((shlg, code))

    def installAction(self, action, code, *flags):
        _flgs = set(flags)
        _flgs.add('plugin')
        self.__actions.append((action, code, list(_flgs)))

    def registerIcon(self, icon, code, kind):
        self.__icons.append((icon, code, kind))
