# ----------------------------------------------------------------------------
# - wbgui-message-binding-table                                              -
# - wild cruncher:message port bindings widget                               -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wbgui_message_binding_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan messge message bindings widget

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""

from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5.QtWidgets import QHBoxLayout, QWidget, QVBoxLayout, QPushButton, \
	QItemDelegate, QComboBox, QTableView, QHeaderView, QAbstractItemView
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_core.modules.bus.db.wbhdb_external_module_instance import \
    WbhdbExtModuleInstance
from wcr_core.modules.bus.db.wbhdb_connection_point_base import \
    WbhdbConnectionPointBase
from wcr_core.db.whdb_connection import WhdbConnection
from wcr_core.modules.bus.wbmod_project_context import \
    wbmod_retrieve_mes_obj


PROD_STR = "Producer"
CONS_STR = "Consumer"


class WbguiMessageBindingTable(QWidget):
    """
    Logical bus binding to physical buses table widget.
    """

    BTN_PRINT_BINDING_VISIBLE = False

    def __init__(self, pctx, scene, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project context
        self.__pctx = pctx
        #: Graphical scene
        self.__scene = scene

        #: Binded data
        self.__bindedPorts = []
        self.__bindedMsgs = {}
        self.__bindedMsgs['Prod'] = []
        self.__bindedMsgs['Cons'] = []

        #: lambdas
        self.__portF = lambda: self.getNonBindedPorts()
        self.__messF = lambda: self.getNonBindedMsg()
        self.__formatF = lambda o, c: self.formatF(o, c)

        #: Bus binding table model
        self.__bndgTblM = _WbguiMessageBindingTableModel(self.__formatF,
                                                         self.__bindedPorts,
                                                         self.__bindedMsgs)
        self.__bndgItmDlgt = _WbguiPortMessageBindDelegate(self.__portF,
                                                           self.__messF,
                                                           self.__formatF)
        #: Bus binding table widget
        self.__bndgTblW = QTableView()

        #: Add entry button
        self.__addButton = QPushButton('Add')
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')
        self._initElements()
        self.setLayout(self._createLayout())
        self._loadFromData()

    def _initElements(self):
        self.__bndgTblW.setModel(self.__bndgTblM)
        self.__bndgTblW.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.Stretch)
        self.__bndgTblW.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch)
        self.__bndgTblW.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.__bndgTblW.setItemDelegate(self.__bndgItmDlgt)

        self.__addButton.clicked.connect(self._addEntry)
        self.__deleteButton.clicked.connect(self._deleteSelEntries)

        # to do to automaticly refresh binding tab when editing principalview
        self.__scene.sceneItemRemoved.connect(self._updateDeletedPort)


    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)

        layout.addWidget(self.__bndgTblW)

        btnLayout = QHBoxLayout()

        btnLayout.addWidget(self.__addButton)
        btnLayout.addWidget(self.__deleteButton)
        layout.addLayout(btnLayout)

        return layout

    def _loadFromData(self):
        """
        Fill the table from the project's data
        """
        ports = self.__pctx.getProjectDB(). \
            getTopHierarchy().getModulesPorts()
        bindlist=[]
        for pb in ports:
            prod, cons = pb[1].getAllTransitMessages()
            """(filepath, module, messagename) (idem)"""

            for m in self.__pctx.getProjectDB(). \
            getTopHierarchy().getMessagesListe():
                if (m[0],m[2],m[3]) in prod:
                    bindlist.append((pb,m,True))
                if (m[0],m[2],m[3]) in cons:
                    bindlist.append((pb, m, False))

        for elem in bindlist:
            self.__bndgTblM.addRow(elem[0],elem[1],elem[2])
        pass

    def _addEntry(self):
        """
        Add entry in the model
        """
        self.__bndgTblM.addRow()

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.__bndgTblW.selectedIndexes()
        """:type: list[QModelIndex]"""

        self.__bndgTblM.deleteRows([idx.row() for idx in selIdx])

    def formatF(self, o, col):

        if o is None:
            return ''
        if col == 2:
            #producer/consomator data : is boolean
            if o:
                return PROD_STR
            else:
                return CONS_STR
        elif col == 1:
            """o=[Filename, Module, Name] => Message description"""
            return o[1]+' : '+o[3]
        elif col == 0:
            """o=[ModuleInstance, PortInstance]"""
            return o[0].getName()+' - '+o[1].getName()
        else:
            return 'No formating available'

    def _updateDeletedPort(self, delItem):
        if isinstance(delItem, WhdbConnection):
            self._deletInvalidMessageBinding(delItem)

    def _deletInvalidMessageBinding(self, item):
        def isValide(port, mess):
            if not port in self.__portF():
                return False
            else:
                return True
        self.__bndgTblM.deleteInvalidEntries(isValide)

    def getNonBindedPorts(self):
        ports = self.__pctx.getProjectDB(). \
            getTopHierarchy().getModulesPorts()
        return [port for port in ports if port[1] not in self.__bindedPorts]

    def getNonBindedMsg(self):
        msgs = self.__pctx.getProjectDB(). \
            getTopHierarchy().getMessagesListe()
        prodMsg = [msg for msg in msgs if msg not in self.__bindedMsgs['Prod']]
        consMsg = [msg for msg in msgs if msg not in self.__bindedMsgs['Cons']]
        return set(prodMsg).union(consMsg)


class _WbguiMessageBindingTableModel(QAbstractTableModel):
    """
    Bus Binding table model.
    """

    TITLE_MESSAGE = 'Message'
    TITLE_PORT = 'Port'
    TITLE_STATUS = 'Status'

    HEADERS = [TITLE_PORT, TITLE_MESSAGE, TITLE_STATUS]

    def __init__(self, formatF, bindedPorts, bindedMsgs, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Buses formatting function
        self.__formatF = formatF
        self.__bindedPorts = bindedPorts
        self.__bindedMsgs = bindedMsgs
        #: Entries as list of tuples of
        #  port, message and type
        self.__entries = []
        """:type: list[((Operatorinstance,WbhdbConnectionPointBase),
                        (Path, Filename, Modulename, Messagename),
                        Bool)]"""

        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, port=None, message=None, prod=None):
        """
        Add a row at the end of the list (append)

        :param port: Message to be set
        :type port: (Operatorinstance,WbhdbConnectionPointBase)| None
        :param message: Data to be set
        :type message:  (Path, Filename, Modulename, Messagename)| None
        :param prod: size to be set
        :type prod: Bool | None
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append((port, message, prod))

        if port is not None and message is not None and prod is not None:
            # call binding function of message
            port[1].addMessageBind(message, prod)
            self.__bindedPorts.append((port[1]))
            if prod:
                self.__bindedMsgs['Prod'].append(message)
            else:
                self.__bindedMsgs['Cons'].append(message)
            pass

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        port, message, product = self.__entries[rowIndex]
        if port is not None and message is not None and product is not None:
            try:
                #Unbound function of message
                port[1].removeMessageBind(message, product)
                self.__bindedPorts.remove((port[1]))
                if product:
                    self.__bindedMsgs['Prod'].remove(message)
                else:
                    self.__bindedMsgs['Cons'].remove(message)
            except ValueError:
                pass

        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def deleteInvalidEntries(self, validEntryF):
        """
        Delete invalid entries based on a passed function to determine if they
        are valid

        :param validEntryF: Function used to validate the entries
        :type validEntryF: (WphdbPhyBusInstance, WshdbBusInstance) -> bool
        """
        rowsToDelete = []
        """:type: list[int]"""

        for row in range(self.rowCount()):
            port, mess, prod = self.__entries[row]

            if not validEntryF(port, mess):
                rowsToDelete.append(row)

        self.deleteRows(rowsToDelete)

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column <= 2:
            info = self.__entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return self.__formatF(info, column)
            elif role == Qt.UserRole:
                return info
        return None

    def setData(self, index, data, role=None):
        """
        :type index: QModelIndex
        :type data: object
        """
        assert role == Qt.UserRole
        self._setDataEntry(index, data)

    def _setDataEntry(self, index, data, role=None):

        port, msg, prod = self.__entries[index.row()]
        if port is not None and msg is not None and prod is not None:
            #a binding could have been done by that line before
            # it have to be deleted.
            try:
                port[1].removeMessageBind(msg,prod)
                self.__bindedPorts.remove((port[1]))
                if prod:
                    self.__bindedMsgs['Prod'].remove(msg)
                else:
                    self.__bindedMsgs['Cons'].remove(msg)

            except ValueError:
                pass

        nraw = ()
        col = index.column()
        if col == 0:
            nraw = (data, msg, prod)
        elif col == 1:
            nraw = (port, data, prod)
        elif col == 2:
            nraw = (port, msg, data)

        self.__entries[index.row()] = nraw

        port, msg, prod = self.__entries[index.row()]
        if port is not None and msg is not None and prod is not None:
            #addbinding.
            port[1].addMessageBind(msg, prod)
            self.__bindedPorts.append((port[1]))
            if prod:
                self.__bindedMsgs['Prod'].append(msg)
            else:
                self.__bindedMsgs['Cons'].append(msg)



    def flags(self, index):
        flgs = super().flags(index)
        flgs |= Qt.ItemIsEditable
        return flgs

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return 3



class _WbguiPortMessageBindDelegate(QItemDelegate):
    """
    Delegate for bus items (both physical and logical)
    """

    def __init__(self, portsListPrv, messagesListPrv, formatF, parent=None):
        """
        :type bbmessListPrv: () -> list[blackboxmessage]
        :type formatF: (blackboxmessage) -> str
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)

        self.formatF = formatF
        self.__portsListPrv = portsListPrv
        self.__messagesListPrv = messagesListPrv

    def createEditor(self, parent, option, index):
        """
        Create a combobox containing adequate choices for the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """
        def save(*_):
            self.setModelData(editor, index.model(), index)


        editor = QComboBox(parent)
        col = index.column()
        editor.addItem(None, None)
        # Iterate through the messages using the adequate provider
        # and add items to the combo box
        if col == 0:
            busPrvF = self.__portsListPrv
        elif col == 1:
            busPrvF = self.__messagesListPrv
        elif col ==2:
            busPrvF = lambda : [True, False]

        if col in (0,1):
            currentData = index.model().data(index.model().index(index.row(), col),
                                             role=Qt.UserRole)
            if not currentData is None:
                editor.addItem(self.formatF(currentData, col), currentData)

        for bus in busPrvF():
            editor.addItem(self.formatF(bus, col), bus)
        editor.currentIndexChanged.connect(save)

        return editor


    def setEditorData(self, editor, index):
        """
        :type editor: QComboBox
        :type index: QModelIndex
        """
        data = index.model().data(index, Qt.UserRole)
        for index in range(editor.count()):
            if editor.itemData(index, Qt.UserRole) == data:
                editor.setCurrentIndex(index)
                break
        else:
            editor.setCurrentIndex(0)


    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)