# ----------------------------------------------------------------------------
# - wbgui-message-library-tree                                               -
# - wild cruncher: message library database tree browser                     -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpgui_logical_simulation_tree
   :platform:     Unix
   :synopsis:     Wild Cruncher bus messages tree browser

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""
import os
from collections import OrderedDict
from math import cos, sin

from PyQt4.QtCore import Qt, QPointF
from PyQt4.QtGui import QHBoxLayout, QPushButton
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError
from wcr_core.action.wact_action_api import WactActionObserverP, \
    WACT_ACTION_EXECUTED_EVENT
from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_core.db.whdb_block import WhdbBlock
from wcr_core.db.whdb_block_proxy import WhdbEntityContainerProxy
from wcr_core.modules.bus.db.wbhdb_top_block import BBM_MODULE
from wcr_core.project.wprj_xmlio import WPRJ_ICD_SILKAN_XML_MIME_TYPE
from wcr_core.modules.bus.db.wbhdb_wrapper_registry import \
    wbhdb_create_and_instanciate_module
from wcr_frontend.gui.wgui_protocols import WguiEmptyContextMenuFactory, \
    QWidget_Meta_M
from wcr_frontend.gui.wgui_tree import WguiTree, WguiTreeModelManager
from wcr_frontend.gui.wgui_tree_model import WguiFilterModel, \
    WguiTreeItemFactoryP, WguiTreeItemBase, WguiTreeModel
from wcr_frontend.gui.wgui_utils import wgui_getOpenFile
from wcr_frontend.modules.bus.gui.wbgui_import_message_file_handler import \
    WbguiMessageImportFileHandler


class WbguiMessageLibraryTree(WguiTree, WactActionObserverP,
                              WcmnReleasableP, metaclass=QWidget_Meta_M):
    def __init__(self, pctx, cctx, scene, actions=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param cctx: Cockpit system context
        :type cctx: WguiCockpitSystemContextP
        :param actions: Action to observe for update
        :type actions: Iterable[WactActionP] | None
        """
        self.__topLevel = WhdbEntityContainerProxy()
        self.__pctx = pctx
        self.__actions = list(actions) if actions else []
        self.__scene = scene
        self.instances = OrderedDict()
        # item factory
        self.__itemFc = _WbguiMessageTreeItemFactory()

        # WguiTreeModelManager
        modelMngr = WbguiMessageTreeModelManager(
            self.__pctx.getProjectDB().getTopHierarchy().getMessages(),
            self.__itemFc)

        #QAbstractItemModel
        model =WguiFilterModel(modelMngr.model)

        #: Collapse-all button
        self.__impButton = QPushButton('Import')

        # self.__updateButton = QPushButton('Update')
        # self.__updateButton.setEnabled(False)

        # self.__updateAllButton = QPushButton('Update All')
        # self.__updateAllButton.setEnabled(False)

        #tree init
        super().__init__(model, modelMngr, WguiEmptyContextMenuFactory())

        for action in self.__actions:
            if action:
                action.registerActionObserver(self)

    def _createLayout(self):
        """
        Create the layout.

        :return: Layout
        :rtype: QLayout
        """
        layout = super()._createLayout()

        btnLayout = QHBoxLayout()
        btnLayout.addWidget(self.__impButton)
        # btnLayout.addWidget(self.__updateButton)
        # btnLayout.addWidget(self.__updateAllButton)
        layout.addLayout(btnLayout)

        return layout

    def _initElements(self):
        """
        Initialize the UI elements as needed.
        """
        super()._initElements()
        #link to import action
        self.__impButton.clicked.connect(self._importMessages)
        # link to update action
        # self.__updateButton.clicked.connect()
        # link to update all action
        # self.__updateAllButton.clicked.connect(self.drawImportedModules)

    def release(self):
        for action in self.__actions:
            if action:
                action.unregisterActionObserver(self)
        self.__actions = []
        self.__topLevel.release()
        self.__pctx = None
        super().release()

    def actionNotify(self, actx, action, event, *args, **kwargs):
        if event & WACT_ACTION_EXECUTED_EVENT:
            self.updateTree()

    def updateTree(self):
        pctx = self.__pctx
        if pctx is None:
            return
        if self.__pctx.getProjectDB().getTopHierarchy().getMessages() == {}:
            for f in self.__pctx.getProjectDB().getTopHierarchy().\
                    getMessagesFiles():
                self._updatemessages(f)

        super().updateTree()

    def _updatemessages(self, path, importMsg=False):
        # Recuperation des messages
        handler = WbguiMessageImportFileHandler(
            *self.__pctx.getProjectDB().getModule())
        itop, name = handler.handleFile(path, self.__pctx)
        if not itop:
            raise Exception("No message to be found in {}".format(path))
        if name == '':
            name = 'file:' + os.path.basename(path)
        msgs = self.__pctx.getProjectDB().getTopHierarchy().getMessages()

        # we only want messages that are used in modules,
        # so we iter over all connection and get the start connection name
        # which correspond to the message name
        fmsgs = []
        for conn in itop.getConnections():
            m = conn.getStartConnectionPoint()
            fmsgs.append(m)
            #get icd module instances
            startIns = conn.getStartObject()
            endIns = conn.getEndObject()
            startUUID = startIns.getUUID()
            endUUID = endIns.getUUID()
            if startUUID not in self.instances:
                self.instances[startUUID] = [startIns, False]
            if endUUID not in self.instances:
                self.instances[endUUID] = [endIns, False]
        # fill msgs by file dictionary
        msgs[path] = (os.path.basename(path),
                      (itop, name),
                      {BBM_MODULE: fmsgs})

        if importMsg:
            self.drawImportedModules()

        # saving messages at right place
        self.__pctx.getProjectDB().getTopHierarchy().setMessages(msgs)
        return name

    def drawImportedModules(self):
        count = len(self.instances)
        theta = 0  # angle that will be increased each loop
        h = 12  # x coordinate of circle center
        k = 10  # y coordinate of circle center
        step = 30  # amount to add to theta each time(degrees)
        r = 200
        # create list of point to form a circle
        points = []
        for i in range(count):
            x = h + r * cos(theta)
            y = k + r * sin(theta)
            points.append((x, y))
            theta += step
        # draw each non-drawn instances
        for (uuid, (inst, isDrawn)), (xpos, ypos) in zip(self.instances.items(),
                                                         points):
            # create bus module operators
            oper = wbhdb_create_and_instanciate_module(inst)
            # create bus module instances
            if not isDrawn:
                self.__scene.createFunctionItem(QPointF(xpos, ypos), oper, inst)
                # set isDrawn at true
                self.instances[uuid][1] = True

    def _importMessages(self):
        #Selection d'un fichier
        selected_path = wgui_getOpenFile("Import Message from icd", self,
                                         WPRJ_ICD_SILKAN_XML_MIME_TYPE)
        btop = self.__pctx.getProjectDB().getTopHierarchy()
        if selected_path in btop.getMessagesFiles():
            filename = os.path.basename(selected_path)
            showErrorWindow("Import Error",
                            "ICD file '{}' was already imported", filename)
            return
        if selected_path:
            name = self._updatemessages(selected_path, importMsg=True)
            btop.addMessagesFile(selected_path,name)
            # super().updateTree()
            self.updateTree()
            #update tree to make new messages apear


class WbguiTreeModel(WguiTreeModel):

    def __init__(self, itemFc):
        super().__init__(itemFc)

    def setSourceModel(self, model):
        None

    def setDynamicSortFilter(self, stat):
        None

    def invalidate(self):
        None


class WbguiMessageTreeModelManager(WguiTreeModelManager):
    """
    Tree-model manager for database trees.
    """

    def __init__(self, msgs, itemFy):
        """
        :param tblk: Database top-level block
        :type tblk: WhdbEntityContainerP
        :param itemFy: Tree-item factory
        :type itemFy: WguiTreeItemFactoryP
        """
        def itemIdF(i):
            if i is None:
                return None
            return i.getUUID()
        super(WbguiMessageTreeModelManager, self).__init__(itemIdF, itemFy)

        self.__dtb = msgs

    def populateModel(self, fromScratch = True):
        """
        :type fromScratch: bool
        """
        def addFiles(block, path):
            for k in block.keys():
                path1 = list(path)
                f = block[k]
                msgs = f[2]
                f[1][0].setName(f[1][1])
                self.itemAdded(f[1][0], path)
                path1.append(f[1][0])
                # will be modified at the integration of whitebox messages
                for bbm in msgs['Blackboxmessages']:
                    self.itemAdded(bbm, path1)

        if fromScratch:
            self.itemClear()
        addFiles(self.__dtb, [])

    def itemClear(self):
        """
        Remove all the items. respecting leaforder
        """
        for itemKey, itemVal in self.getmap().items():
            #we delet only "_WbguiMessageFileTreeItem" constituing the first
            # level of the tree.
            if isinstance(itemVal, _WbguiMessageFileTreeItem):
                self.model.removeItem(itemVal)
                del itemVal
        self.getmap().clear()


# noinspection PyAbstractClass
class _WbguiTreeItemBase(WguiTreeItemBase):
    """
    State-machine Datamodel tree item.
    """

    def __init__(self, parentItem=None):
        super().__init__(parentItem)

    def setData(self, data, column, role=None):
        pass

    def getColumnCount(self):
        return 3

    @property
    def flags(self):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    # @abstractmethod
    def createDetailUI(self):
        """
        Create the detail-view UI.

        :return: Detail-view UI
        :rtype: QWidget
        """
        raise NotImplementedError()


class _WbguiMessageRootTreeItem(_WbguiTreeItemBase):
    def __init__(self):
        super().__init__()

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            if column == 0:
                return 'Type'
            elif column == 1:
                return 'Name'
            elif column == 2:
                return 'Size'
        return None

    def createDetailUI(self):
        return None


class _WbguiMessageFileTreeItem(_WbguiTreeItemBase):
    def __init__(self, itemData, parentItem=None):

        self.data = itemData

        super().__init__(parentItem)

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            if column == 0:
                return 'File'
            elif column == 1:
                return self.data.getName()
            elif column == 2:
                return None
        return None

    def createDetailUI(self):
        return None


class _WbguiMessageTreeItem(_WbguiTreeItemBase):
    def __init__(self, itemData, parentItem=None):

        self.data = itemData

        super().__init__(parentItem)

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            if column == 0:
                return 'Message'
            elif column == 1:
                return self.data.getName()
            elif column == 2:
                return '' #self.data.getSize()
        return None

    def createDetailUI(self):
        return None


class _WbguiMessageTreeItemFactory(WguiTreeItemFactoryP):

    def createTreeItem(self, itemData, parentItem, **kwargs):
        if not parentItem:
            return _WbguiMessageRootTreeItem(), None, None
        if not itemData:
            return None, None, None
        #WARNING whdbBlock are used cause it is the type of topblock of
        # message... it coul'd be more specified.
        if isinstance(itemData, WhdbBlock):
            return _WbguiMessageFileTreeItem(itemData, parentItem), \
                   parentItem, None
        return _WbguiMessageTreeItem(itemData, parentItem), parentItem, None
