# ----------------------------------------------------------------------------
# - whgui-switch-connection-points-props-widget                              -
# - wild cruncher - cockpit : connection point properties dialog widgets     -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whgui_host_connection_points_props_widget
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher connection point properties dialog widgets

.. moduleauthor:: Philippe lambert <lambert@silkan.com>
"""
from collections import OrderedDict
from functools import partial

from PyQt5.QtCore import Qt, QSortFilterProxyModel
from PyQt5.QtWidgets import QMessageBox, QComboBox, QWidget, QVBoxLayout, \
	QPushButton, QLabel, QTableView, QHeaderView, QSizePolicy, QLineEdit, \
	QAbstractItemView, QSpacerItem

from wcr_frontend.gui.wgui_props_dialog_widget import \
    WguiXPointPropertiesBase, WguiXPointPropsTextField, \
    WguiXPointPropsDeleteButton, WGUI_NAME_PLACEHOLDER, WGUI_XPNT_CODE, \
    WguiXPointPropsROLookupField, WguiXPointPropsFieldInstance, \
    WguiXPointPropsField
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_NAME, \
    WHDB_XPOINT_TYPE
from wcr_core.modules.hardware.db.whhdb_connection_point_base_meta import *
from wcr_core.modules.hardware.db.whhdb_connection_point_base import \
    WHHDB_XPOINT_CONFIG_KEY
from wcr_core.modules.hardware.db.whhdb_connection_point_base_meta import \
    WHHDB_TYPE_KEY
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_XPOINT_NAMES, WHHDB_XPOINT_DEFAULT_CODE, WHHDB_XPOINT_DISCRET_CODE,\
    WHHDB_XPOINT_FIBER_CHANNEL_CODE, WHHDB_XPOINT_AFDX_CODE, WHHDB_MODULES_IO_NAMES, \
    WHHDB_XPOINT_CETRAC_CODE
from wcr_frontend.modules.hardware.gui.whgui_connection_parameter_edition_delegate \
    import WhguiConnectionParameterBindDelegate
from wcr_core.db.whdb_utils import whdb_getNextName
from wcr_frontend.modules.hardware.gui.whgui_Operator_connection_points_props_widget import \
    _WHGUI_XPOINT_INFO, WguiXPointPropsComboField, _WHGUI_WGUI_CONF_OPTIONS, _WHGUI_COL_CONF,\
    _WhguiXPointConfigDataAdapter, WHHDB_XPOINT_DISCRET_KEY, WHHDB_DISCRET_KEYS, WHHDB_FIBER_KEYS,\
    WHHDB_AFDX_KEYS, WHHDB_XPOINT_FIBER_CHANNEL_KEY, WHHDB_XPOINT_AFDX_KEY, WHHDB_1553_KEYS, \
    _WhguiXPointConfigTableModel, _WHGUI_COL_STATUS


_WHGUI_XPNT_CONFIG = '__config'
WHDB_STATUS_MASK = 'status mask'
_WHGUI_CETRAC_BASE_HIDDEN_OPTION = [WHHDB_PORT_CARD_KEY, WHHDB_TYPE_KEY,
                                    WHHDB_PORT_STATUS_MONITOR_KEY,
                                    WHHDB_PORT_ACTIVE_KEY,
                                    WHHDB_PORT_STATUS_MASK_KEY]

_WHGUI_AFDX_HIDDEN_OPTION = [WHHDB_PORT_AFDX_NET_ID_KEY,
                             WHHDB_PORT_AFDX_INTERFACE_ID_KEY,
                             WHHDB_PORT_AFDX_SIDE_ID_KEY,
                             WHHDB_PORT_AFDX_LOCATION_ID_KEY]

_WHGUI_CETRAC_DEFAULT_HIDDEN_OPTIONS = _WHGUI_CETRAC_BASE_HIDDEN_OPTION + WHHDB_FIBER_KEYS + \
                                       WHHDB_DISCRET_KEYS + WHHDB_AFDX_KEYS + WHHDB_1553_KEYS

# noinspection PyAbstractClass
class WhguiCetracSwitchXPointProperties(WguiXPointPropertiesBase):
    def __init__(self, instance, pctx, parent=None, ConverifF=None):
        """
        :param instance: The instance object
        :type instance: WhhdbHostInstance
        :param pctx: Project context
        :type pctx: WhmodHardwareProjectContextP
        :param parent: Parent widget
        """
        self.__module = pctx.getModule()
        """:type: WpmodSystemApplicationModuleP"""
        self.__xpntTypeNames = {
            x.getCode(): x.getName() for x in self.__module.getConnectionPointTypes()}
        self.__xpntCfgMetas = self.getCfgMeta()

        super(WhguiCetracSwitchXPointProperties, self).\
            __init__(instance, parent=parent)

        self.__ConnectionUpdateF = ConverifF
        self.table = None
        self._configStatus = QComboBox()
        self._configStatus.addItem('No', 'No')
        self._configStatus.addItem('Yes', 'Yes')
        if self.__class__ is WhguiCetracSwitchXPointProperties:
            self._finishInitialization()
            self.table = self.ui.xpntTableF
            self._loadFromData()

    def _loadFromData(self):
        """
        Import information from Switch Status
        """
        # need to initialise tab with instance info
        status = self.instance.getStatus()
        if status is None:
            status = 'No'
        statusIdx = self._configStatus.findText(status)
        self._configStatus.setCurrentIndex(statusIdx)
        # init
        if self.table is None:
            return
        for row in range(self.table.rowCount()):
            comboConf = self.table.cellWidget(row, _WHGUI_COL_CONF)
            metaData = self.model.getValue(row+1, _WHGUI_XPNT_CONFIG)
            connType = metaData['type']
            idx = comboConf.findText(connType)
            comboConf.setCurrentIndex(idx)
            comboConf.currentIndexChanged.connect(partial(self.updatePortField,
                                                          row+1, comboConf))
            editStatus = self.table.cellWidget(row, _WHGUI_COL_STATUS)
            statusMask = metaData[WHHDB_PORT_STATUS_MASK_KEY]
            editStatus.setText(statusMask)

    def getCfgMeta(self):
        dflt = {}
        for x in self.__module.getConnectionPointTypes():
            cfgMeta = x.getConfigMeta()
            afdxMeta = x.getConfigAfdxMeta()
            dflt[x.getCode()] = cfgMeta + afdxMeta
        return dflt

    def createFields(self):
        return {
            WHDB_XPOINT_NAME: WguiXPointPropsTextField(
                WHDB_XPOINT_NAME, 'Conn. Point', WGUI_NAME_PLACEHOLDER),
            _WHGUI_XPNT_CONFIG: WguiXPointPropsField(
                _WHGUI_XPNT_CONFIG, 'Information',
                WhguiXPointConfigFieldInstance, self.__xpntCfgMetas),
            _WHGUI_XPOINT_INFO: WguiXPointPropsComboField(
                _WHGUI_XPOINT_INFO, 'Config', _WHGUI_WGUI_CONF_OPTIONS),
            WHDB_STATUS_MASK: WguiXPointPropsTextField(
                WHDB_STATUS_MASK, 'status mask')
        }

    def getFieldOrder(self):
        return [WHDB_XPOINT_NAME, _WHGUI_XPNT_CONFIG, _WHGUI_XPOINT_INFO]

    def getStretchFields(self):
        return {_WHGUI_XPNT_CONFIG}

    def getButtons(self):
        def add():
            ctyp = WHHDB_XPOINT_CETRAC_CODE
            if ctyp:
                meta = self.__xpntCfgMetas.get(ctyp)
                confbase=meta.getEmptyValue()
                confbase[WHHDB_TYPE_KEY] = ctyp
                name = 'Port_{}'.format(ctyp)
                names = [ptx[WHDB_XPOINT_NAME] for ptx in self.model.getValues()]
                connName = whdb_getNextName(name, names)
                portNumbers = [int(ptx[_WHGUI_XPNT_CONFIG][WHHDB_PORT_NUM_KEY])
                               for ptx in self.model.getValues()]
                portNum = max(portNumbers) + 1 if portNumbers else 1
                confbase[WHHDB_PORT_NUM_KEY] = portNum
                self.model.addXPoint(**{
                    WHDB_XPOINT_NAME: connName,
                    WHDB_XPOINT_TYPE: ctyp,
                    _WHGUI_XPNT_CONFIG: confbase})
            idx = tuple(self.model.getIndices())[-1] - 1
            if idx < 0:
                idx = 0
            comboConf = self.table.cellWidget(idx, _WHGUI_COL_CONF)
            comboIdx = comboConf.findText(ctyp)
            comboConf.setCurrentIndex(comboIdx)
            self.save()
            self.updatePortField(idx+1, comboConf)
            comboConf.currentIndexChanged.connect(partial(self.updatePortField,
                                                          idx +1, comboConf))

        def deleteRow():
            xpntTable = self.fields[_WHGUI_XPNT_CONFIG].configTblM
            idx = xpntTable.getLastIndex()
            if idx <= 0:
                return None
            self.model.removeXPoint(idx)
            # self.save()

        addBtn = QPushButton('Add')
        delBtn = QPushButton('Delete')
        addBtn.clicked.connect(add)
        delBtn.clicked.connect(deleteRow)

        label = QLabel('Ptp GrandMaster: ')
        spacer = QSpacerItem(1000, 0, QSizePolicy.Minimum, QSizePolicy.Minimum)

        return [label, self._configStatus, spacer, addBtn, delBtn]

    def resetModel(self, model, instance):
        model.setXPoints(
            {WGUI_XPNT_CODE: x,
             WHDB_XPOINT_NAME: x.getName(),
             WHDB_XPOINT_TYPE: x.getEntityType(),
             _WHGUI_XPNT_CONFIG:
                 OrderedDict(x.rawElement.getMeta(WHHDB_XPOINT_CONFIG_KEY))}
            for x in instance.getNonCardConnectionPoints())
        self._loadFromData()

    def createXPoint(self, idx, model, instance):
        return instance.createConnectionPoint(
            ctyp=model.getValue(idx, WHDB_XPOINT_TYPE),
            name=model.getValue(idx, WHDB_XPOINT_NAME),)

    def updateHiddenOption(self, opt):
        if opt == WHHDB_XPOINT_AFDX_CODE:
            hiddenOptions = _WHGUI_CETRAC_BASE_HIDDEN_OPTION + _WHGUI_AFDX_HIDDEN_OPTION
        else:
            hiddenOptions = _WHGUI_CETRAC_DEFAULT_HIDDEN_OPTIONS

        return hiddenOptions

    def updatePortField(self, row, comboConf):
        connType = comboConf.currentText()
        metaData = self.model.getValue(row, _WHGUI_XPNT_CONFIG)
        metaData['type'] = connType
        self.model.setValue(row, _WHGUI_XPNT_CONFIG, metaData)
        self.fields[_WHGUI_XPNT_CONFIG].configTblM.activeXPointChanged(row)

    def updateXPoint(self, idx, model, xpnt):
        if not xpnt:
            return None
        name = model.getValue(idx, WHDB_XPOINT_NAME)
        if name:
            xpnt.setName(name)

        metaKey = None
        confCombo = self.table.cellWidget(idx-1, _WHGUI_COL_CONF)
        statusEdit = self.table.cellWidget(idx - 1, _WHGUI_COL_STATUS)
        if confCombo is None or statusEdit is None:
            return None
        statusMask = statusEdit.text()
        confValue = confCombo.currentText()
        allMetaData = model.getValue(idx, _WHGUI_XPNT_CONFIG)

        def removeMetaFromKeys(meta, rmKeys):
            tmpMeta = OrderedDict()
            for k in rmKeys:
                if k in meta.keys():
                    tmpMeta[k] = meta[k]
            return tmpMeta

        metaData = removeMetaFromKeys(allMetaData, CONF_KEYS)
        metaData2 = None

        if confValue == WHHDB_XPOINT_AFDX_CODE:
            metaKey = WHHDB_XPOINT_AFDX_KEY
            metaData2 = removeMetaFromKeys(allMetaData, WHHDB_AFDX_KEYS)

        if metaData2:
            # update discret or fiber data
            xpnt.rawElement.setMeta(metaData2, metaKey)
        # update Normal data
        xpnt.rawElement.setMeta(metaData,
                                WHHDB_XPOINT_CONFIG_KEY)
        # update the status mask
        xpnt.setStatusMask(statusMask)

    def save(self):
        contoverif=[]
        contabelem=self.model.getValues()
        for con in contabelem:
            if '__xpnt' in con.keys():
                if con['__xpnt'] in self.instance.getConnectionPoints():
                    contoverif.append(con['__xpnt'])
        status = self._configStatus.currentText()

        self.instance.setStatus(status)
        #update ports
        self.saveData()

        isOk, errMsg = self.instance.checkPortNumberConsistency()
        if not isOk:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Information)
            msg.setText(errMsg)
            msg.setWindowTitle("Save Ports Warning")
        self.resetModel(self.model, self.instance)

        if self.__ConnectionUpdateF is not None:
            self.__ConnectionUpdateF(contoverif)

    def saveData(self):
        model = self.model
        xpnts = set(model.getValuesForCode(WGUI_XPNT_CODE))
        mark = []
        for xpnt in list(self.instance.getNonCardConnectionPoints()):
            if xpnt not in xpnts:
                mark.append(xpnt)
                # self.removeXPoint(xpnt)
        xpntAdded = False
        for idx in model.getIndices():
            xpnt = model.getValue(idx, WGUI_XPNT_CODE)
            if not xpnt:
                xpnt = self.createXPoint(idx, model, self.instance)
                xpntAdded = True
            self.updateXPoint(idx, model, xpnt)
        for xpnt in mark:
            self.removeXPoint(xpnt)
        if xpntAdded:
            self.instance.resetConnectionPointPositions()


class _WhguiXPointCetracConfigTableModel(_WhguiXPointConfigTableModel):
    def __init__(self, model, context, cfgAdapter, parent=None):
        """
        :param model: Connection points data model
        :type model: WguiXPointPropsModelP
        :param cfgAdapter: Configuration data adapter
        :type cfgAdapter: _WhguiXPointConfigDataAdapter
        :param parent: Parent item
        """
        super().__init__(model, context, cfgAdapter, parent)

    def _hideOptions(self, index):
        """
        function that hide the option that is not necessary and update the data
        :param index:
        :return:
        """
        connType = None
        conf = self.model.getValuesForIndex(index)[_WHGUI_XPNT_CONFIG]
        xpnt = self.model.getValuesForIndex(index)[WGUI_XPNT_CODE]
        confValues = None
        if self.context.table:
            # update data on selecting row
            comboConf = self.context.table.cellWidget(index - 1, _WHGUI_COL_CONF)
            if not comboConf:
                self.endResetModel()
                return
            connType = comboConf.currentText()
        else:
            metaData = self.model.getValue(index, _WHGUI_XPNT_CONFIG)
            connType = metaData['type']
        # update data on opening the window
        if connType == WHHDB_XPOINT_AFDX_CODE:
            confValues = xpnt._getConfigAfdx()

        if confValues:
            conf.update(confValues)

        hidOptions = self.context.updateHiddenOption(connType)
        # hide options depending on port type
        for hiddenOpt in hidOptions:
            if hiddenOpt in self.cfgKeys:
                self.cfgKeys.remove(hiddenOpt)

    def getFlags(self, flgs, index):
        if index.column() == 1:
            return flgs | Qt.ItemIsEditable
        else:
            return flgs & ~Qt.ItemIsEditable


class WhguiXPointConfigFieldInstance(WguiXPointPropsFieldInstance):
    def __init__(self, code, label, context, xpntCfgMetas):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        """
        super().__init__(code, label)

        self.__context = context

        #: Configuration data adapter
        self.__cfgAdapter = _WhguiXPointConfigDataAdapter(
            context.model, xpntCfgMetas)

        #: Configuration

        #: Configuration table view
        self.__configTblF = QTableView()
        #: Configuration table model
        self.__configTblM = _WhguiXPointCetracConfigTableModel(
            context.model, context, self.__cfgAdapter)
        self.__configTblDelegate = \
            WhguiConnectionParameterBindDelegate(sorted(WHHDB_XPOINT_NAMES))

        #: Configuration table filter proxy model
        self.__ctblFilter = QSortFilterProxyModel(self.__configTblM)
        self.__ctblFilter.setFilterCaseSensitivity(Qt.CaseInsensitive)
        #: Configuration table filter input
        self.__configFilterF = QLineEdit()

        self.__auxWidget = None

    def _createAuxWidget(self):
        self.__ctblFilter.setSourceModel(self.__configTblM)
        self.__configTblF.setModel(self.__ctblFilter)
        self.__configTblF.setSelectionMode(QAbstractItemView.NoSelection)
        self.__configTblF.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.ResizeToContents)
        self.__configTblF.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch)
        self.__configTblF.setItemDelegate(self.__configTblDelegate)
        self.__configFilterF.setPlaceholderText('Filter configuration keys')

        self.__context.activeXPointChanged.connect(
            self.__configTblM.activeXPointChanged)
        self.__configFilterF.textChanged.connect(
            self.__ctblFilter.setFilterFixedString)

        wdgt = QWidget()
        layout = QVBoxLayout()
        layout.addWidget(
            QLabel('Configuration values for the selected connection point:'))
        layout.addWidget(self.__configTblF)
        layout.addWidget(self.__configFilterF)
        wdgt.setLayout(layout)
        return wdgt

    @property
    def configTblM(self):
        return self.__configTblM

    def installWidget(self, index, widgetF, textF):
        # noinspection PyShadowingNames
        def formatConfig():
            cfg = self.__context.model.getValue(index, _WHGUI_XPNT_CONFIG)
            """:type: dict[str, str]"""
            if not cfg:
                return ''
            else:
                # set default Values
                if not cfg[WHHDB_PORT_BBONE_KEY]:
                    cfg[WHHDB_PORT_BBONE_KEY] = 'No'
                if not cfg[WHHDB_PORT_SPEED_KEY]:
                    cfg[WHHDB_PORT_SPEED_KEY] = '1G'
                if not cfg[WHHDB_PORT_NUM_KEY]:
                    cfg[WHHDB_PORT_NUM_KEY] = '1'
                if not cfg[WHHDB_PORT_PTP_ACTIVE_KEY]:
                    cfg[WHHDB_PORT_PTP_ACTIVE_KEY] = 'Off'
            if cfg['type'] in WHHDB_MODULES_IO_NAMES:
                return ''
            return '; '.join(
                '{}: {}'.format(self.__cfgAdapter.getLabel(index, k) or k,
                                cfg.get(k))
                for k in self.__cfgAdapter.getKeys(index) if cfg.get(k))

        def configChanged(idx, code):
            try:
                if idx == index and code == self.code:
                    textUpdateF(formatConfig())
            except:
                pass

        textUpdateF = textF(formatConfig())
        self.__context.model.valueChanged.connect(configChanged)

    def getAuxWidget(self):
        if not self.__auxWidget:
            self.__auxWidget = self._createAuxWidget()
        return self.__auxWidget
