# ----------------------------------------------------------------------------
# - whgui-operator-connection-points-props-widget                            -
# - wild cruncher - cockpit : connection point properties dialog widgets     -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whgui_operator_connection_points_props_widget
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher connection point properties dialog widgets

.. moduleauthor:: Philippe lambert <lambert@silkan.com>
"""
from collections import OrderedDict
from functools import partial

from PyQt5.QtCore import Qt, QAbstractTableModel, QSortFilterProxyModel
from PyQt5.QtWidgets import QMessageBox, QComboBox, QHBoxLayout, \
	QWidget, QVBoxLayout, QPushButton, QLabel, QTableView, QHeaderView, \
	QLineEdit, QSplitter, QItemDelegate, QAbstractItemView
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_NAME, \
    WHDB_XPOINT_TYPE
from wcr_core.db.whdb_utils import whdb_getNextName
from wcr_core.modules.hardware.db.whhdb_connection_point_base import \
    WHHDB_XPOINT_CONFIG_KEY, WHHDB_XPOINT_DISCRET_KEY, WHHDB_XPOINT_ARINC825_KEY,\
    WHHDB_XPOINT_FIBER_CHANNEL_KEY, WHHDB_XPOINT_1553_KEY, WHHDB_XPOINT_ARINC429_KEY
from wcr_core.modules.hardware.db.whhdb_connection_point_base_meta import *
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_XPOINT_NAMES, WHHDB_XPOINT_AFDX_CODE, WHHDB_XPOINT_CETRAC_CODE, WHHDB_XPOINT_DISCRET_CODE, \
    WHHDB_XPOINT_DEFAULT_CODE, WHHDB_XPOINT_1553_CODE, WHHDB_XPOINT_ARINC825_CODE, \
    WHHDB_XPOINT_FIBER_CHANNEL_CODE, WHHDB_MODULES_IO_NAMES, WHHDB_XPOINT_ARINC429_CODE
from wcr_core.modules.hardware.db.whhdb_Cetrac_switch_instance import \
    WhhdbCetracSwitchInstance
from wcr_core.modules.hardware.db.whhdb_kazymir_switch_instance import \
    WhhdbKazymirSwitchInstance
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError
from wcr_frontend.gui.wgui_props_dialog_widget import \
    WguiXPointPropsTextField, \
    WGUI_NAME_PLACEHOLDER, WGUI_XPNT_CODE, \
    WguiXPointPropsFieldInstance, \
    WguiXPointPropsField, WguiXPointPropsComboField, WguiModuleIOPointPropertiesBase
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_frontend.modules.hardware.gui.whgui_Host_connection_points_props_widget \
    import _WhguiPCICardTableInstance, _WhguiPCICardTableInstanceModel
from wcr_frontend.modules.hardware.gui.whgui_connection_parameter_edition_delegate \
    import WhguiConnectionParameterBindDelegate

_WHGUI_XPNT_CONFIG = '__config'
_WHGUI_XPOINT_INFO = '__info'
_WHGUI_XPOINT_ACTIVE = '__active'
WHDB_STATUS_MASK = 'status mask'
_WHGUI_WGUI_CONF_OPTIONS = [(x, y) for x, y in zip(WHHDB_MODULES_IO_NAMES,
                                                    WHHDB_MODULES_IO_NAMES)
                            ]
_WHGUI_IO_ACTIVE_OPTIONS = [('On', 'On'), ('Off', 'Off')]
_WHGUI_COL_CONF = 2
_WHGUI_COL_ACTIVE = 3
_WHGUI_COL_STATUS = 4

_WHGUI_BASE_HIDDEN_OPTION = [WHHDB_PORT_CARD_KEY, WHHDB_TYPE_KEY,
                             WHHDB_PORT_STATUS_MONITOR_KEY,
                             WHHDB_PORT_ACTIVE_KEY,
                             WHHDB_PORT_STATUS_MASK_KEY]
_WHGUI_DEFAULT_HIDDEN_OPTIONS = _WHGUI_BASE_HIDDEN_OPTION + WHHDB_FIBER_KEYS + \
                                WHHDB_DISCRET_KEYS + WHHDB_AFDX_KEYS + WHHDB_1553_KEYS, \
                                WHHDB_ARINC429_KEYS


_WHGUI_ALL_PORT_OPTIONS = {'base': _WHGUI_BASE_HIDDEN_OPTION,
                           WHHDB_XPOINT_FIBER_CHANNEL_CODE: WHHDB_FIBER_KEYS,
                           WHHDB_XPOINT_AFDX_CODE: WHHDB_AFDX_KEYS,
                           WHHDB_XPOINT_1553_CODE: WHHDB_1553_KEYS,
                           WHHDB_XPOINT_DISCRET_CODE: WHHDB_DISCRET_KEYS,
                           WHHDB_XPOINT_ARINC825_CODE: WHHDB_ARINC825_KEYS,
                           WHHDB_XPOINT_ARINC429_CODE: WHHDB_ARINC429_KEYS}

_WHGUI_SPECIAL_SPEED = [WHHDB_XPOINT_ARINC825_CODE, WHHDB_XPOINT_ARINC429_CODE]

def getAllHiddenOptions(ioCode):
    tmpKeys = []
    for code, keys in _WHGUI_ALL_PORT_OPTIONS.items():
        if ioCode != code:
            tmpKeys.extend(keys)
            # remove generic speed
        if ioCode in _WHGUI_SPECIAL_SPEED:
            tmpKeys.append(WHHDB_PORT_SPEED_KEY)
    return tmpKeys


# noinspection PyAbstractClass
class WhguiOperatorModulesIoProperties(WguiModuleIOPointPropertiesBase):
    def __init__(self, instance, pctx, parent=None, ConverifF=None):
        """
        :param instance: The instance object
        :type instance: WhhdbHostInstance
        :param pctx: Project context
        :type pctx: WhmodHardwareProjectContextP
        :param parent: Parent widget
        """
        self.__module = pctx.getModule()
        """:type: WpmodSystemApplicationModuleP"""
        self.__xpntTypeNames = {
            x.getCode(): x.getName() for x in self.__module.getConnectionPointTypes()}
        self.__xpntCfgMetas = self.getCfgMeta()

        # init table necessary for
        self.table = None
        super(WhguiOperatorModulesIoProperties, self). \
            __init__(instance, parent=parent)

        self.__card_tab = _WhguiIoCardTableInstance(pctx, instance, parent=self)
        #: Configuration table view
        if self.__class__ is WhguiOperatorModulesIoProperties:
            self._finishInitialization()
        self.table = self.ui.xpntTableF
        self.__ConnectionUpdateF = ConverifF

    def getCfgMeta(self):
        dflt = {}
        for x in self.__module.getConnectionPointTypes():
            cfgMeta = x.getConfigMeta()
            disMeta = x.getConfigDisMeta()
            fiberMeta = x.getConfigFiberMeta()
            c1553Meta = x.getConfig1553Meta()
            ARINC825Meta = x.getConfigARINC825Meta()
            ARINC429Meta = x.getConfigARINC429Meta()
            dflt[x.getCode()] = cfgMeta + disMeta + fiberMeta + c1553Meta +\
                                ARINC825Meta + ARINC429Meta
        return dflt

    def _loadData(self):
        """
        Import information from Switch Status
        """
        #need to initialise tab with instance info
        if self.table:
            for row in range(self.table.rowCount()):
                comboConf = self.table.cellWidget(row, _WHGUI_COL_CONF)
                metaData = self.model.getValue(row+1, _WHGUI_XPNT_CONFIG)
                connType = metaData['type']
                idx = comboConf.findText(connType)
                comboConf.setCurrentIndex(idx)
                comboConf.currentIndexChanged.connect(partial(self.updatePortField,
                                                              row+1, comboConf,
                                                              'type'))
                comboActive = self.table.cellWidget(row, _WHGUI_COL_ACTIVE)
                metaData = self.model.getValue(row + 1, _WHGUI_XPNT_CONFIG)
                connActive = metaData['active']
                idx = comboActive.findText(connActive)
                comboActive.setCurrentIndex(idx)
                comboActive.currentIndexChanged.connect(partial(self.updatePortField,
                                                              row + 1, comboActive,
                                                              'active'))
                editStatus = self.table.cellWidget(row, _WHGUI_COL_STATUS)
                statusMask = metaData[WHHDB_PORT_STATUS_MASK_KEY]
                editStatus.setText(statusMask)
                editStatus.textChanged.connect(partial(self.updatePortStatusMask,
                                                   row + 1, editStatus))
            # update view of first row when card selected
            comboConf = self.table.cellWidget(0, _WHGUI_COL_CONF)
            self.updatePortField(1, comboConf, 'type')
            comboActive = self.table.cellWidget(0, _WHGUI_COL_ACTIVE)
            self.updatePortField(1, comboActive, 'active')

    def updateFields(self):
        self.__card_tab.resettab()
        super().updateFields()

    def createFields(self):
        return {
            WHDB_XPOINT_NAME: WguiXPointPropsTextField(
                WHDB_XPOINT_NAME, 'Name', WGUI_NAME_PLACEHOLDER),
            _WHGUI_XPNT_CONFIG: WguiXPointPropsField(
                _WHGUI_XPNT_CONFIG, 'Information',
                WhguiXPointConfigFieldInstance, self.__xpntCfgMetas),
            _WHGUI_XPOINT_INFO: WguiXPointPropsComboField(
                _WHGUI_XPOINT_INFO, 'Config', _WHGUI_WGUI_CONF_OPTIONS),
            _WHGUI_XPOINT_ACTIVE: WguiXPointPropsComboField(
                _WHGUI_XPOINT_ACTIVE, 'Active', _WHGUI_IO_ACTIVE_OPTIONS),
            WHDB_STATUS_MASK: WguiXPointPropsTextField(
                WHDB_STATUS_MASK, 'status mask')
        }

    def getFieldOrder(self):
        return [WHDB_XPOINT_NAME, _WHGUI_XPNT_CONFIG, _WHGUI_XPOINT_INFO, _WHGUI_XPOINT_ACTIVE]

    def getStretchFields(self):
        return {_WHGUI_XPNT_CONFIG}

    def getButtons(self):
        def add():
            curmodule = self.__card_tab.selectedCard(role=Qt.DisplayRole)
            curcard = self.__card_tab.bindedCard(role=Qt.DisplayRole)
            if curmodule is None:
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Information)
                msg.setText("Select a card to add a port")
                msg.setWindowTitle("Add port Warning")
                msg.exec_()
                return
            selIdx = self.__card_tab.pcicTblW.selectedIndexes()[0]
            # ctyp = wgui_qcb_get_data(typeF)
            ctyp = WHHDB_XPOINT_DISCRET_CODE
            if ctyp:
                meta = self.__xpntCfgMetas.get(ctyp)
                confbase = meta.getEmptyValue()
                confbase[WHHDB_TYPE_KEY] = ctyp
                confbase[WHHDB_PORT_CARD_KEY] = '{}.{}'.format(curcard,
                                                               curmodule)
                name = 'Port_{}'.format(ctyp)
                names = [ptx[WHDB_XPOINT_NAME] for ptx in self.model.getValues()]
                connName = whdb_getNextName(name, names)
                portNumbers = [int(ptx.getPortNum())
                               for ptx in self.instance.getConnectionPoints()]
                portNum = max(portNumbers) + 1 if portNumbers else 1
                confbase[WHHDB_PORT_NUM_KEY] = portNum
                self.model.addXPoint(**{
                    WHDB_XPOINT_NAME: connName,
                    WHDB_XPOINT_TYPE: ctyp,
                    _WHGUI_XPNT_CONFIG: confbase})

            self.save()
            self.__card_tab.pcicTblW.setCurrentIndex(selIdx)

        def deleteRow():
            curcard = self.__card_tab.selectedCard(role=Qt.DisplayRole)
            if curcard is None:
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Information)
                msg.setText("Select a card to delete a port")
                msg.setWindowTitle("Delete port Warning")
                msg.exec_()
                return
            selIdx = self.__card_tab.pcicTblW.selectedIndexes()[0]
            xpntTable = self.fields[_WHGUI_XPNT_CONFIG].configTblM
            idx = xpntTable.getLastIndex()
            if idx <= 0:
                return None
            self.model.removeXPoint(idx)
            # self.save()
            self.__card_tab.pcicTblW.setCurrentIndex(selIdx)

        addBtn = QPushButton('Add')
        delBtn = QPushButton('Delete')
        addBtn.clicked.connect(add)
        delBtn.clicked.connect(deleteRow)

        return [addBtn, delBtn]

    def resetModel(self, model, instance):
        model.setXPoints(
            {WGUI_XPNT_CODE: x,
             WHDB_XPOINT_NAME: x.getName(),
             WHDB_XPOINT_TYPE: x.getEntityType(),
             _WHGUI_XPNT_CONFIG:
                 OrderedDict(x.rawElement.getMeta(WHHDB_XPOINT_CONFIG_KEY))}
            for x in instance.getIoCardConnectionPoints(
                self.__card_tab.selectedCard()))
        self._loadData()

    def createXPoint(self, idx, model, instance):
        return instance.createConnectionPoint(
            ctyp=model.getValue(idx, WHDB_XPOINT_TYPE),
            name=model.getValue(idx, WHDB_XPOINT_NAME), )

    def updateXPoint(self, idx, model, xpnt):
        pass

    def updateIOPoint(self, idx, model, xpnt):
        if self.__class__ != WhguiOperatorModulesIoProperties:
            return None
        if not xpnt:
            return None
        name = model.getValue(idx, WHDB_XPOINT_NAME)
        if name:
            xpnt.setName(name)

        metaKey = None
        # get port type
        confValue = self.model.getValue(idx, _WHGUI_XPNT_CONFIG)['type']

        # get port state (active of not)
        activeValue = self.model.getValue(idx, _WHGUI_XPNT_CONFIG)['active']
        allMetaData = model.getValue(idx, _WHGUI_XPNT_CONFIG)

        def removeMetaFromKeys(meta, rmKeys):
            tmpMeta = OrderedDict()
            for k in rmKeys:
                if k in meta.keys():
                    tmpMeta[k] = meta[k]
            return tmpMeta

        metaData = removeMetaFromKeys(allMetaData, CONF_KEYS)
        metaData2 = None

        if confValue == WHHDB_XPOINT_DISCRET_CODE:
            metaKey = WHHDB_XPOINT_DISCRET_KEY
            metaData2 = removeMetaFromKeys(allMetaData, WHHDB_DISCRET_KEYS)
        elif confValue == WHHDB_XPOINT_FIBER_CHANNEL_CODE:
            metaKey = WHHDB_XPOINT_FIBER_CHANNEL_KEY
            metaData2 = removeMetaFromKeys(allMetaData, WHHDB_FIBER_KEYS)
        elif confValue == WHHDB_XPOINT_1553_CODE:
            metaKey = WHHDB_XPOINT_1553_KEY
            metaData2 = removeMetaFromKeys(allMetaData, WHHDB_1553_KEYS)
        elif confValue == WHHDB_XPOINT_ARINC825_CODE:
            metaKey = WHHDB_XPOINT_ARINC825_KEY
            metaData2 = removeMetaFromKeys(allMetaData, WHHDB_ARINC825_KEYS)
        elif confValue == WHHDB_XPOINT_ARINC429_CODE:
            metaKey = WHHDB_XPOINT_ARINC429_KEY
            metaData2 = removeMetaFromKeys(allMetaData, WHHDB_ARINC429_KEYS)

        if metaData2:
            # update io data
            xpnt.rawElement.setMeta(metaData2, metaKey)
        # update Normal data
        xpnt.rawElement.setMeta(metaData,
                                WHHDB_XPOINT_CONFIG_KEY)
        # update the status mask
        xpnt.setActive(activeValue)
        self.instance.addPortUUIDToIoCard(xpnt.getIoCard(),
                                          xpnt.getUUID())

    def updateHiddenOption(self, opt):
        if opt in WHHDB_MODULES_IO_NAMES:
            hiddenOptions = getAllHiddenOptions(opt)
        else:
            hiddenOptions = _WHGUI_DEFAULT_HIDDEN_OPTIONS

        return hiddenOptions

    def updatePortField(self, row, comboConf, metaName):
        res = self.fields[_WHGUI_XPNT_CONFIG].configTblM.activeXPointChanged(row)
        if res:
            connType = comboConf.currentText()
            metaData = self.model.getValue(row, _WHGUI_XPNT_CONFIG)
            metaData[metaName] = connType
            self.model.setValue(row, _WHGUI_XPNT_CONFIG, metaData)

    def updatePortStatusMask(self, row, statusEdit):
        statusMask = statusEdit.text()
        metaData = self.model.getValue(row, _WHGUI_XPNT_CONFIG)
        metaData[WHHDB_PORT_STATUS_MASK_KEY] = statusMask
        self.model.setValue(row, _WHGUI_XPNT_CONFIG, metaData)

    def _createLayout(self):
        container = QWidget()
        container_Layout = QHBoxLayout()
        splitprinc = QSplitter()
        splitprinc.setOrientation(Qt.Vertical)

        left=QWidget()
        left.setLayout(self.__card_tab.createLayout())
        splitprinc.addWidget(left)

        right=QWidget()
        right.setLayout(super()._createLayout())
        container_Layout.addWidget(right)
        container.setLayout(container_Layout)
        container.setMinimumWidth(700)
        splitprinc.addWidget(container)
        layout=QHBoxLayout()
        layout.addWidget(splitprinc)
        layout.setContentsMargins(2,2,2,2)

        return layout

    def save(self):
        self.__card_tab.syncCards()

        curcard = self.__card_tab.selectedCard(role=Qt.DisplayRole)
        cardportuuidl = [x.getUUID() for x in \
                         self.instance.getIoCardConnectionPoints(curcard)]
        contoverif = []
        contabelem = self.model.getValues()
        for con in contabelem:
            if '__xpnt' in con.keys():
                if con['__xpnt'] in self.instance.getConnectionPoints():
                    contoverif.append(con['__xpnt'])

        for x in self.instance.getAllIoConnectionPoints():
            if not x.getUUID() in cardportuuidl:
                self.model.addXPoint(
                    **{WGUI_XPNT_CODE: x,
                       WHDB_XPOINT_NAME: x.getName(),
                       WHDB_XPOINT_TYPE: x.getEntityType(),
                       _WHGUI_XPNT_CONFIG:
                           OrderedDict(x.rawElement.getMeta(
                               WHHDB_XPOINT_CONFIG_KEY))})
        # update ports
        super(WhguiOperatorModulesIoProperties, self).save()
        # Now all port not binded with a card are the new port so we binding
        # them with the selected card.
        self.instance.delIoCardandPorts()
        # assign cardName to each port
        for moduleName in self.instance.getIoCards():
            for conn in self.instance.getIoCardConnectionPoints(moduleName):
                cardName = self.instance.getIoCardBind(moduleName)
                conn.setCard('.'.join((cardName, moduleName)))

        self.resetModel(self.model, self.instance)

        self.__card_tab.resettab()

        if self.__ConnectionUpdateF is not None:
            self.__ConnectionUpdateF(contoverif)

        isOk, msgContent = self.instance.checkPortNumberConsistency()

        if not isOk:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Information)
            msg.setText(msgContent)
            msg.setWindowTitle("Save Ports Warning")
            msg.exec_()
            return True

        isOk, msgContent = self.checkIoDataConsistency()

        if not isOk:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Information)
            msg.setText(msgContent)
            msg.setWindowTitle("Save Ports Warning")
            msg.exec_()
            return True

    def checkIoDataConsistency(self):
        for point in self.instance.getAllIoConnectionPoints():
            if point.getCode() == WHHDB_XPOINT_ARINC825_CODE:
                canType = point._getConfigARINC825(WHHDB_PORT_ARINC825_TYPE_KEY)
                baudRate = point._getConfigARINC825(WHHDB_PORT_ARINC825_SPEED_KEY)
                userBaudRate = point._getConfigARINC825(WHHDB_PORT_ARINC825_USER_SPEED_KEY)
                arbiterBaudRate = point._getConfigARINC825(WHHDB_PORT_ARINC825_ARBITRATION_SPEED_KEY)
                dataBaudRate = point._getConfigARINC825(WHHDB_PORT_ARINC825_DATA_SPEED_KEY)
                if canType == 'ISO':
                    if dataBaudRate != '' or arbiterBaudRate != '':
                        msg = "Port consistency error, " \
                              "dataBaudRate and arbiterBaudRate are specific to FD type.\n" \
                              "Please check port {}.".format(point.getName())
                        return False, msg
                elif canType == 'ISO FD':
                    if baudRate != '' or userBaudRate != '':
                        msg = "Port consistency error, " \
                              "BaudDate and User BaudRate are specific to ISO.\n" \
                              "Please check port {}.".format(point.getName())
                        return False, msg
                if baudRate and userBaudRate != '':
                    msg = "Port consistency error, " \
                          "BaudDate and User BaudRate can't be both specified.\n" \
                          "Please check port {}.".format(point.getName())
                    return False, msg
                if baudRate and arbiterBaudRate != '':
                    msg = "Port consistency error, " \
                          "arbiterBaudRate is specific to FD type and " \
                          "can't be specified with baudRate.\n" \
                          "Please check port {}.".format(point.getName())
                    return False, msg
                elif dataBaudRate == '' and arbiterBaudRate != '':
                    msg = "Port consistency error, " \
                          "arbiterBaudRate is specific to FD type and " \
                          "need to be specified with dataBaudRate.\n" \
                          "Please check port {}.".format(point.getName())
                    return False, msg
                elif dataBaudRate != '' and arbiterBaudRate == '':
                    msg = "Port consistency error, " \
                          "dataBaudRate is specific to FD type and " \
                          "need to be specified with arbiterBaudRate.\n" \
                          "Please check port {}.".format(point.getName())
                    return False, msg

        return True, None


class _WhguiXPointConfigDataAdapter(object):
    def __init__(self, model, xpntCfgMetas):
        self.__model = model
        self.__xpntCfgMetas = xpntCfgMetas

    def _getMeta(self, index):
        """
        Get the configuration meta-data for the connection point at the given
        index.

        :param index: Index
        :type index: int
        :return: Configuration meta-data
        :rtype: WdbsRecordMetaTypeP
        """
        ctyp = self.__model.getValue(index, WHDB_XPOINT_TYPE)
        """:type: str"""
        return self.__xpntCfgMetas.get(ctyp)

    def getKeys(self, index):
        """
        Get the configuration keys for the connection point at the given index.

        :param index: Index
        :type index: int
        :return: Configuration keys
        :rtype: Iterable[str]
        """
        return self._getMeta(index).getKeys()

    def getLabel(self, index, key):
        """
        Get the label for the configuration key for the connection point
        at the given index.

        :param index: Index
        :type index: int
        :param key: Configuration key
        :type key: str
        :return: Label
        :rtype: str
        """
        return self._getMeta(index).getLabel(key)


class _WhguiIoCardTableInstance(_WhguiPCICardTableInstance):

    def __init__(self, pctx, instance, parent=None):
        """
        :param pctx: Project context
        :type pctx:
        :param instance: The instance edited
        :param parent: Parent widget
        """
        super().__init__(pctx, instance, parent)
        self.__parent = parent

    def _initModels(self):
        self._CardTypesF = WHHDB_MODULES_IO_NAMES
        self._xpntTypeNames = WHHDB_MODULES_IO_NAMES
        self.__pcicTblM = _WhguiIoCardTableInstanceModel(self._xpntTypeNames,
                                                         self.formatcards,
                                                         parent=self)
        instance = self.parent().instance
        cards = instance.getCards()
        self.__pcicItmDlgt = _WhguiPCICardBindDelegate(cards,
                                                       self._pcictypF,
                                                       instance=instance,
                                                       parent=self)
        self.__deleteButton = QPushButton('Delete')
        self.__deleteButton.clicked.connect(self._deleteSelEntries)

    @property
    def pcicTblM(self):
        return self.__pcicTblM

    @property
    def pcicItmDlgt(self):
        return self.__pcicItmDlgt

    def resettab(self):
        self.pcicTblM.deleteRows(range(0, self.pcicTblM.rowCount()))
        self._Inst.IOCARD_TO_REMOVE = []
        self._loadFromData()

    def _loadFromData(self):
        """
        Import information from HostInst to fill pcicTblM
        """
        #need to initialise tab with instanceinfo
        fullinfo = self._Inst.getIoCardsDescription()
        for card, typ, bind in fullinfo:
            self.pcicTblM.addRow(Cardid=card, Card=card, Bind=typ, Ports=bind)

    def formatcards(self, data, column, colhead):
        """
        :param data: data to format
        :type data: Cardname | Cardtype | [PortUUID]
        """
        header_size = len(_WhguiIoCardTableInstanceModel.HEADERS)
        if data is None:
            return ''
        if column == 0:
            return data
        elif column == 1:
            return data
        elif 1 < column < (header_size + len(self._xpntTypeNames)):
            dtplist = self._Inst.getIoConnectionDistributionTypeByUUID(data)
            for dpt in dtplist.keys():
                if colhead in dpt:
                    return str(dtplist[dpt])
            return '0'
        else:
            return 'No formating available'

    def syncCards(self):
        datas = []
        datatab = self.pcicTblM.getFullCards()
        for data in datatab:
            cardname, cardbind, bindings, carid = data
            """data = (cardname, cardtype, bindings, cardid)"""
            if data is not None and carid is '' and cardname is not None \
                    and (cardbind is not None or bindings is not None):
                self._Inst.setIoCard(cardname)
                self._Inst.setIoCardBind(cardname, cardbind)
            elif carid is not None:
                datas.append(data)
        hcards = [(n, t, b, n) for n, t, b, in
                  self._Inst.getIoCardsDescription()]
        for dat in datas:
            # name edition
            if dat[3] in [cid for _,_,_,cid in hcards]:
                self._Inst.replaceIoCard(dat[3], [dat[0], dat[1], dat[2],
                                                  dat[3]])

    def removecard(self, cardn):
        """
        add card with name cardn sit the card to delete when save
        """
        self._Inst.markIoCard(cardn)

    def createLayout(self):
        layout = QVBoxLayout()
        layout.addWidget(self.pcicTblW)

        btnLayout = QHBoxLayout()
        btnLayout.addStretch(1)
        btnLayout.addWidget(self.__deleteButton)
        layout.addLayout(btnLayout)

        return layout

    def bindedCard(self, selectignor=False, role=Qt.UserRole):
        lrow = []
        for index in self.getSelectedIndexes():
            lrow.append(index.row())
        ulrow = set(lrow)
        if len(ulrow) == 1 and not selectignor:
            return self.pcicTblM.data(self.pcicTblM.index(lrow[0], 1),
                                      role=role)
        else:
            return None

    def selecChanged(self, newselec, oldselect):
        """
        Signal handler : when selection change reset connection editor with
        the param of the new card selected
        """
        for idx in self.__parent.model.getIndices():
            xpnt = self.__parent.model.getValue(idx, WGUI_XPNT_CODE)
            self.__parent.updateIOPoint(idx, self.__parent.model, xpnt)
        self.__parent.resetModel(self.__parent.model,
                                 self.__parent.instance)


class _WhguiIoCardTableInstanceModel(_WhguiPCICardTableInstanceModel):
    """
    Bus Binding table model.
    """
    TITLE_NAME = 'Io Name'
    TITLE_TYPE = 'Card Bind'
    HEADERS = [TITLE_NAME, TITLE_TYPE]

    def __init__(self, xpntTypeNames, formatF, parent=None):
        """
        :param XpntTypeNames: The Names of Xpoint of the mule
        :param parent: Parent widget
        """
        self.__parent = parent
        self.__formatF = formatF
        super().__init__(xpntTypeNames, formatF, parent)

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column <= self.columnCount():
            if column < 2:
                info = self.entries[row][column]
            else:
                info = self.entries[row][2]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return self.__formatF(info, column, self.HEADERS[column])
            elif role == Qt.UserRole:
                return self.entries[row][3]#info
        return None

    def addRow(self, Cardid='', Card=None, Bind=None, Ports=None, Status='No'):
        """
        Add a row at the end of the list (append)

        :param Card: The card name inside the host
        :type Card: str | None
        :param Ctyp: Type of the card
        :type Ctyp: str | None
        :param Ports: ports uuid binded to card
        :type Ports: [uuid] | None
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.entries.append((Card, Bind, Ports, Cardid))

        self.endInsertRows()
        self._update()

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        card, bind, ports, cardid = self.entries[rowIndex]
        if card is not None and (bind is not None or ports is not None):
            try:
                # Unbound function of message
                self.__parent.removecard(self.entries[rowIndex][0])
                pass
            except ValueError:
                pass

        self.entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def _setCardEntry(self, index, data):
        card, ctype, bind, cardid = self.entries[index.row()]
        if index.column() == 0:
            actualentry = (data, ctype, bind, cardid)
        elif index.column() == 1:
            actualentry = (card, data, bind, cardid)
        self.entries[index.row()] = actualentry


class _WhguiPCICardBindDelegate(QItemDelegate):
    """
    Delegate for bus items (both physical and logical)
    """

    def __init__(self, CardNames, formatF, instance, parent=None):
        """
        :type CardTyp: () -> list[CardTyp]
        :type formatF: (elem,column) -> str
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)
        self.instance = instance
        self.formatF = formatF
        self.__CardNames = CardNames

    def createEditor(self, parent, option, index):
        """
        Create a editors for card table the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """

        def save(*_):
            self.setModelData(editor, index.model(), index)

        editor = None
        col = index.column()
        # id col = 0 is str for cardname
        if col == 0:
            editor = QLineEdit(parent)
        # if col = 1 combobox with CardNames
        elif col == 1:
            if (not isinstance(self.instance, WhhdbCetracSwitchInstance) and
                    not isinstance(self.instance, WhhdbKazymirSwitchInstance)):
                editor = QComboBox(parent)
                cardPrvF = self.__CardNames
                for type in cardPrvF:
                    editor.addItem(self.formatF(type, col), type)

        return editor

    def setEditorData(self, editor, index):
        """
        :type editor: QComboBox | QLineEdit
        :type index: QModelIndex
        """
        data = index.model().data(index, Qt.EditRole)
        if isinstance(editor,QComboBox):
            for index in range(editor.count()):
                if editor.itemData(index, Qt.UserRole) == data:
                    editor.setCurrentIndex(index)
                    break
            else:
                editor.setCurrentIndex(0)
        elif isinstance(editor,QLineEdit):
            editor.setText(data)
            None

    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox | QLineEdit
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        if isinstance(editor, QComboBox):
            model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)
        elif isinstance(editor, QLineEdit):
            if isNameValid(editor.text()):
                model.setData(index, editor.text(), Qt.UserRole)
            else:
                showErrorWindow("Name Error", NameFormatMessageError,
                                editor.text())


class _WhguiXPointConfigTableModel(QAbstractTableModel):
    HEADERS = ['field', 'data']

    def __init__(self, model, context, cfgAdapter, parent=None):
        """
        :param model: Connection points data model
        :type model: WguiXPointPropsModelP
        :param cfgAdapter: Configuration data adapter
        :type cfgAdapter: _WhguiXPointConfigDataAdapter
        :param parent: Parent item
        """
        super().__init__(parent)

        #: Connection points data model
        self.__model = model
        #: Connection points data context
        self.__context = context
        #: Configuration data adapter
        self.__cfgAdapter = cfgAdapter

        #: Current connection point index
        self.__index = -1
        #: Current configuration keys
        self.__cfgKeys = []

    def _getCfgDict(self, idx=None):
        """
        Get the configuration dictionary of the connection point at the given
        index.

        :param idx: Index of the connection point (current if `None`)
        :type idx: int | None
        :return: Configuration dictionary
        :rtype: dict[str, str]
        """
        if idx is None:
            idx = self.__index
        if idx is None or idx < 0:
            return {}

        return self.__model.getValue(idx, _WHGUI_XPNT_CONFIG)

    def _getCfgValue(self, key):
        """
        Get the configuration value for the given key for the current
        connection point.

        :param key: Key
        :type key: str
        :return: Configuration value
        :rtype: str
        """
        return self._getCfgDict().get(key)

    def _setCfgValue(self, key, value):
        """
        Set the configuration value for the current connection point for the
        given key to the given value.

        :param key: Key
        :type key: str
        :param value: Value
        :type value: str
        """
        idx = self.__index
        if idx is None or idx < 0:
            return

        cdct = self.__model.getValue(idx, _WHGUI_XPNT_CONFIG)
        """:type: dict[str, str]"""
        cdct[key] = value
        self.__model.setValue(idx, _WHGUI_XPNT_CONFIG, cdct)

    def activeXPointChanged(self, index):
        """
        Callback for changes in the currently active connection point.

        :param index: Index of the currently active connection point
        :type index: int
        """
        self.beginResetModel()
        if self.__index > 0:
            try:
                xpnt = self.__model.getValuesForIndex(self.__index)[WGUI_XPNT_CODE]
                self.__context.updateIOPoint(self.__index, self.__model, xpnt)
            except KeyError:
                pass
        self.__index = index
        if index > 0:
            try:
                self.__cfgKeys = list(self.__cfgAdapter.getKeys(index))
            except KeyError:
                self.endResetModel()
                return False
            self._hideOptions(index)
        else:
            self.__cfgKeys = []

        self.endResetModel()
        return True

    def _hideOptions(self, index):
        """
        function that hide the option that is not necessary and update the data
        :param index:
        :return:
        """
        connType = None
        conf = self.__model.getValuesForIndex(index)[_WHGUI_XPNT_CONFIG]
        xpnt = self.__model.getValuesForIndex(index)[WGUI_XPNT_CODE]
        confValues = None
        if self.__context.table:
            # update data on selecting row
            comboConf = self.__context.table.cellWidget(index - 1, _WHGUI_COL_CONF)
            if not comboConf:
                self.endResetModel()
                return
            connType = comboConf.currentText()
        else:
            metaData = self.__model.getValue(index, _WHGUI_XPNT_CONFIG)
            connType = metaData['type']
        # update data on opening the window
        if connType == WHHDB_XPOINT_DISCRET_CODE:
            confValues = xpnt._getConfigDiscret()
        elif connType == WHHDB_XPOINT_FIBER_CHANNEL_CODE:
            confValues = xpnt._getConfigFiberChannel()
        elif connType == WHHDB_XPOINT_1553_CODE:
            confValues = xpnt._getConfig1553()
        elif connType == WHHDB_XPOINT_AFDX_CODE:
            confValues = xpnt._getConfigAfdx()
        elif connType == WHHDB_XPOINT_ARINC825_CODE:
            confValues = xpnt._getConfigARINC825()
        elif connType == WHHDB_XPOINT_ARINC429_CODE:
            confValues = xpnt._getConfigARINC429()

        if confValues:
            conf.update(confValues)
        hidOptions = self.__context.updateHiddenOption(connType)
        if connType == WHHDB_XPOINT_ARINC825_CODE:
            if confValues[WHHDB_PORT_ARINC825_TYPE_KEY] == 'ISO':
                hidOptions.append('dataBaudRate')
                hidOptions.append('arbitBaudRate')
            elif confValues[WHHDB_PORT_ARINC825_TYPE_KEY] == 'ISO FD':
                hidOptions.append('baudRate')
                hidOptions.append('userBaudRate')
        # hide options depending on port type
        for hiddenOpt in hidOptions:
            if hiddenOpt in self.__cfgKeys:
                self.__cfgKeys.remove(hiddenOpt)

    def getLastIndex(self):
        return self.__index

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__cfgKeys)

    def columnCount(self, parent=None, *args, **kwargs):
        return 2

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def data(self, index, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            key = self.__cfgKeys[index.row()]
            if index.column() == 0:
                return self.__cfgAdapter.getLabel(self.__index, key)
            else:
                return self._getCfgValue(key)
        return None

    def setData(self, index, value, role=None):
        if role == Qt.EditRole:
            if index.column() == 1:
                key = self.__cfgKeys[index.row()]
                self._setCfgValue(key, value)
                if key == WHHDB_PORT_ARINC825_TYPE_KEY:
                    self.activeXPointChanged(self.__index)
                return True
        return False

    def flags(self, index):
        flgs = super().flags(index)
        return self.getFlags(flgs, index)

    def getFlags(self, flgs, index):
        idx = index.column()
        # fix the row CetraC conf as non editable
        if self.__cfgKeys[index.row()] == WHHDB_PORT_CETRAC_CONF_KEY:
            return flgs & Qt.ItemIsEditable
        if idx == 1:
            return flgs | Qt.ItemIsEditable
        elif idx == 0:
            return flgs & ~Qt.ItemIsEditable


class WhguiXPointConfigFieldInstance(WguiXPointPropsFieldInstance):
    def __init__(self, code, label, context, xpntCfgMetas):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        """
        super().__init__(code, label)

        self.__context = context

        #: Configuration data adapter
        self.__cfgAdapter = _WhguiXPointConfigDataAdapter(
            context.model, xpntCfgMetas)

        #: Configuration

        #: Configuration table view
        self.__configTblF = QTableView()
        #: Configuration table model
        self.__configTblM = _WhguiXPointConfigTableModel(
            context.model, context, self.__cfgAdapter)
        self.__configTblDelegate = \
            WhguiConnectionParameterBindDelegate(sorted(WHHDB_MODULES_IO_NAMES))
        #: Configuration table filter proxy model
        self.__ctblFilter = QSortFilterProxyModel(self.__configTblM)
        self.__ctblFilter.setFilterCaseSensitivity(Qt.CaseInsensitive)
        #: Configuration table filter input
        self.__configFilterF = QLineEdit()

        self.__auxWidget = None

    def _createAuxWidget(self):
        self.__ctblFilter.setSourceModel(self.__configTblM)
        self.__configTblF.setModel(self.__ctblFilter)
        self.__configTblF.setSelectionMode(QAbstractItemView.NoSelection)
        self.__configTblF.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.ResizeToContents)
        self.__configTblF.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch)
        self.__configTblF.setItemDelegate(self.__configTblDelegate)
        self.__configFilterF.setPlaceholderText('Filter configuration keys')

        self.__context.activeXPointChanged.connect(
            self.__configTblM.activeXPointChanged)
        self.__configFilterF.textChanged.connect(
            self.__ctblFilter.setFilterFixedString)

        wdgt = QWidget()
        layout = QVBoxLayout()
        layout.addWidget(
            QLabel('Configuration values for the selected IO:'))
        layout.addWidget(self.__configTblF)
        layout.addWidget(self.__configFilterF)
        wdgt.setLayout(layout)
        return wdgt

    @property
    def configTblM(self):
        return self.__configTblM

    def installWidget(self, index, widgetF, textF):
        # noinspection PyShadowingNames
        def formatConfig():
            cfg = self.__context.model.getValue(index, _WHGUI_XPNT_CONFIG)
            """:type: dict[str, str]"""
            if not cfg:
                return ''
            else:
                # set default Values
                if not cfg[WHHDB_PORT_BBONE_KEY]:
                    cfg[WHHDB_PORT_BBONE_KEY] = 'No'
                if not cfg[WHHDB_PORT_SPEED_KEY]:
                    cfg[WHHDB_PORT_SPEED_KEY] = '1G'
                if not cfg[WHHDB_PORT_NUM_KEY]:
                    cfg[WHHDB_PORT_NUM_KEY] = '1'
                if not cfg[WHHDB_PORT_ACTIVE_KEY]:
                    cfg[WHHDB_PORT_ACTIVE_KEY] = 'On'
                if not cfg[WHHDB_PORT_PTP_ACTIVE_KEY]:
                    cfg[WHHDB_PORT_PTP_ACTIVE_KEY] = 'Off'

                point = self.__context.model.getValue(index, WGUI_XPNT_CODE)
                if cfg[WHHDB_TYPE_KEY] == WHHDB_XPOINT_ARINC825_CODE:
                    data = point._getConfigARINC825()
                    if not data.get(WHHDB_PORT_ARINC825_CYCLIC_KEY, None):
                        cfg[WHHDB_PORT_ARINC825_CYCLIC_KEY] = 'No'
                    if not data.get(WHHDB_PORT_ARINC825_LISTEN_MODE_KEY, None):
                        cfg[WHHDB_PORT_ARINC825_LISTEN_MODE_KEY] = 'No'
                    if not data.get(WHHDB_PORT_ARINC825_TYPE_KEY, None):
                        cfg[WHHDB_PORT_ARINC825_TYPE_KEY] = 'ISO'

                if cfg[WHHDB_TYPE_KEY] == WHHDB_XPOINT_ARINC429_CODE:
                    data = point._getConfigARINC429()
                    if not data.get(WHHDB_PORT_ARINC429_MODE_KEY, None):
                        cfg[WHHDB_PORT_ARINC429_MODE_KEY] = 'Periodic'
                    if not data.get(WHHDB_PORT_ARINC429_IS50K_KEY, None):
                        cfg[WHHDB_PORT_ARINC429_IS50K_KEY] = 'No'
                    if not data.get(WHHDB_PORT_ARINC429_TX_RATE_KEY, None):
                        cfg[WHHDB_PORT_ARINC429_TX_RATE_KEY] = 'highSpeed'
                    if not data.get(WHHDB_PORT_ARINC429_RX_RATE_KEY, None):
                        cfg[WHHDB_PORT_ARINC429_RX_RATE_KEY] = 'highSpeed'
                    if not data.get(WHHDB_PORT_ARINC429_TX_USAGE_KEY, None):
                        cfg[WHHDB_PORT_ARINC429_TX_USAGE_KEY] = 'parity'
                    if not data.get(WHHDB_PORT_ARINC429_RX_USAGE_KEY, None):
                        cfg[WHHDB_PORT_ARINC429_RX_USAGE_KEY] = 'parity'
                    if not data.get(WHHDB_PORT_ARINC429_TX_PARITY_KEY, None):
                        cfg[WHHDB_PORT_ARINC429_TX_PARITY_KEY] = 'even'
                    if not data.get(WHHDB_PORT_ARINC429_RX_PARITY_KEY, None):
                        cfg[WHHDB_PORT_ARINC429_RX_PARITY_KEY] = 'even'
                    if not data.get(WHHDB_PORT_ARINC429_SDI_COMPARE_KEY, None):
                        cfg[WHHDB_PORT_ARINC429_SDI_COMPARE_KEY] = 'No'

            return '; '.join(
                '{}: {}'.format(self.__cfgAdapter.getLabel(index, k) or k,
                                cfg.get(k))
                for k in self.__cfgAdapter.getKeys(index) if cfg.get(k))

        def configChanged(idx, code):
            try:
                if idx == index and code == self.code:
                    textUpdateF(formatConfig())
            except:
                pass

        textUpdateF = textF(formatConfig())
        self.__context.model.valueChanged.connect(configChanged)

    def getAuxWidget(self):
        if not self.__auxWidget:
            self.__auxWidget = self._createAuxWidget()
        return self.__auxWidget
