# ----------------------------------------------------------------------------
# - whgui-scene                                                              -
# - wild cruncher - cockpit : graphic scene widget                           -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_scene
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher hardware graphics scene editor

.. moduleauthor:: Philippe Lambert <lamnert@silkan.com>
"""

from PyQt5 import QtWidgets
from wcr_core.db.whdb_utils import whdb_getNextFunctionName, whdb_getNextName
from wcr_core.modules.hardware.db.whhdb_Host_instance import \
    WhhdbHostInstance
from wcr_core.modules.hardware.db.whhdb_connection_point_base import \
    WHHDB_XPOINT_CONFIG_KEY
from wcr_core.modules.hardware.db.whhdb_iocard_instance import \
    WhhdbIoCardInstance
from wcr_core.modules.hardware.db.whhdb_operator_instance import \
    WhhdbOperatorInstance, WHHDB_IO_SWITCH_CARD_KEY
from wcr_core.modules.hardware.db.whhdb_rack_instance import \
    WhhdbRackInstance
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_MODULES_IO_NAMES, WHHDB_XPOINT_CETRAC_CODE
from wcr_core.modules.hardware.db.whhdb_standard_operators import \
    WHHDB_RACK_IO_CODE, DEFAULT_CARDS, WHHDB_CETRAC_SWITCH_CODE, \
    WHHDB_HOST_CODE, WHHDB_KAZYMIR_CETRAC_CODE, WHHDB_KAZYMIR_FIBER_CODE,\
    WHHDB_KAZYMIR_CODE, WHHDB_WHALES_CODE
from wcr_frontend import showErrorWindow
from wcr_frontend.gui.wgui_scene import WguiGraphicsScene
from wcr_frontend.gui.wgui_utils import wgui_snap_point
from wcr_frontend.modules.hardware.gui.whgui_scene_item_factory import \
    WHGUI_ERROR_CON_COLOR, WHGUI_STYLE_CON_DIC, WHGUI_COLOR_CON_POINT_DIC
from wcr_core.modules.hardware.db.whhdb_Host_instance_meta import \
    WHHDB_META_IOCTYPE_KEY

_WHGUI_SCENE_CARD_HOLDERS = {WHHDB_CETRAC_SWITCH_CODE,
                             WHHDB_HOST_CODE, WHHDB_RACK_IO_CODE,
                             WHHDB_KAZYMIR_CETRAC_CODE,
                             WHHDB_KAZYMIR_FIBER_CODE,
                             WHHDB_KAZYMIR_CODE,
                             WHHDB_WHALES_CODE}


class WhguiGraphicsScene(WguiGraphicsScene):
    """
    The WhguiGraphicsScene class is the class responsible to manage
    specific scene action in hardware view that WguiGraphicsScene don't handle
    """

    def __init__(self, pjdb, config, actions, pctx, cctx, actx, mod,
                 parent=None):
        # todo: initialise the right library to consider as cards
        self.__cardlib = mod.getLibrary("wcr.cetc")
        self.__iocardlib = mod.getLibrary("wcr.ioc")
        self._defaultCard = DEFAULT_CARDS
        super().__init__(pjdb, config, actions, pctx, cctx, actx, parent=parent)
        self.__projectCtx = pctx
        self.__maxRackId = 0

    def release(self):
        pass

    def createFunctionItem(self, pos, oper=None):
        """
        Handle the fact tha we want to add a card directly to an host
        and not to create it in the scene view.
        Else add a function item by position it not a card

        :param pos: the function position
        :type pos: QPointF
        """
        if self.__cardlib is None:
            return super().createFunctionItem(pos, oper)

        if oper is None:
            oper = self.getOperatorMode()
            if oper is None: return
        isCard = False
        if oper in self.__cardlib.operators or oper in self._defaultCard:
            isCard = True

        isIoCard = False
        if oper in self.__iocardlib.operators:
            isIoCard = True
        # if object created is not a card
        if not (isCard or isIoCard):
            # Rack special case (handle id)
            if oper.getCode() == WHHDB_RACK_IO_CODE:
                return self.createRackFunctionItem(pos, oper)
            # other
            else:
                return super().createFunctionItem(pos, oper)

        inst, conn = self._getConnectionPointData(pos)
        if inst is None:
            showErrorWindow('Card Error',
                            'Cards need to be inserted in equipment.',
                            None)
            return None
        if inst.getOperator().getCode() not in _WHGUI_SCENE_CARD_HOLDERS:
            return None
        name = whdb_getNextFunctionName(oper, self.topLevelModelBlock)
        cardName = whdb_getNextName(name, inst.getCards())
        # get the function by operator
        card = self.topLevelModelBlock.createOperatorInstance(oper, name=name)
        cardType = card.getOperator().getMeta('type')
        if cardType is None:
            cardType = card.getType()
        if isinstance(inst, WhhdbHostInstance):
            if not self.addCardToHost(cardType, inst, cardName, card):
                return

        elif isinstance(inst, WhhdbRackInstance):
            if not self.addCardToRack(cardType, inst, cardName, card):
                return

        elif isinstance(inst, WhhdbOperatorInstance):
            if not self.addCardToSwitch(cardType, inst, cardName, card):
                return

        # self.topLevelModelBlock.removeInstance(card)
        inst.resetConnectionPointPositions()
        itms = self.items(pos)
        item = itms[0].topLevelItem() if itms else None
        item.updateItem()

    def addCardToHost(self, cardType, inst, cardName, card):
        if cardType in WhhdbHostInstance.TYPES:
            inst.setPCICard(cardName)
            inst.setPCICardType(cardName, cardType)
            self.addFunctionItem(card)
            for p in card.getConnectionPoints():
                p.setCard(cardName)
                xpnt = inst.createConnectionPoint(
                    ctyp=p.getCode(),
                    name=p.getName(),
                    gvty=p.getGravity(),
                    cpos=p.getPosition())
                xpnt.rawElement.setMeta(p.rawElement.getMeta(
                    WHHDB_XPOINT_CONFIG_KEY), WHHDB_XPOINT_CONFIG_KEY)
                inst.addPortUUIDToCard(cardName, xpnt.getUUID())

        elif cardType in WhhdbHostInstance.IO_TYPES:
            # we take the first card to bind the module
            cards = inst.getCards()
            if len(cards) == 0:
                showErrorWindow('Module Error',
                                'A module needs a card to be installed.',
                                None)
                return False
            inst.setIoCard(cardName)
            bindedCardName = cards[0]
            inst.setIoCardBind(cardName, bindedCardName)
            self.addFunctionItem(card)
            for p in card.getConnectionPoints():
                p.setCard('.'.join((cardName, bindedCardName)))
                portNumbers = [ptx.getPortNum()
                               for ptx in inst.getConnectionPoints()]
                portNum = max(portNumbers) + 1 if portNumbers else 1
                p.setPortNum(portNum)
                xpnt = inst.createConnectionPoint(
                    ctyp=p.getCode(),
                    name='{}_{}'.format(cardName, p.getName()),
                    gvty=p.getGravity(),
                    cpos=p.getPosition())
                xpnt.rawElement.setMeta(p.rawElement.getMeta(
                    WHHDB_XPOINT_CONFIG_KEY), WHHDB_XPOINT_CONFIG_KEY)
                inst.addPortUUIDToIoCard(cardName, xpnt.getUUID())


        else:
            showErrorWindow('Card Error',
                            'A host can\'t have a card/module of type {}',
                            cardType)
            return False

        return True

    def addCardToRack(self, cardType, inst, cardName, card):
        if cardType in WhhdbRackInstance.TYPES:
            card.setName(cardName)
            card.linkedEquipement = inst
            card.setEquipement(inst)
            ok = inst.setIOCard(card.getUUID(), card)
            if ok:
                card.InitChannelsId()
                self.addFunctionItem(card)
                return True
        else:
            showErrorWindow('Card Error',
                            'A RackIO can\'t have a card/module of type {}',
                            cardType)
            return False

    def addCardToSwitch(self, cardType, inst, cardName, card):
        if cardType in WhhdbOperatorInstance.IO_TYPES:
            inst.setIoCard(cardName)
            inst.setIoCardBind(cardName, WHHDB_IO_SWITCH_CARD_KEY)
            self.addFunctionItem(card)
            for p in card.getConnectionPoints():
                p.setCard('.'.join((WHHDB_META_IOCTYPE_KEY, cardName)))
                portNumbers = [ptx.getPortNum()
                               for ptx in inst.getConnectionPoints()]
                portNum = max(portNumbers) + 1 if portNumbers else 1
                p.setPortNum(portNum)
                xpnt = inst.createConnectionPoint(
                    ctyp=p.getCode(),
                    name='{}_{}'.format(cardName, p.getName()),
                    gvty=p.getGravity(),
                    cpos=p.getPosition())
                xpnt.rawElement.setMeta(p.rawElement.getMeta(
                    WHHDB_XPOINT_CONFIG_KEY), WHHDB_XPOINT_CONFIG_KEY)
                inst.addPortUUIDToIoCard(cardName, xpnt.getUUID())
            return True
        else:
            showErrorWindow('Card Error',
                            'A Switch can\'t have a card/module of type {}',
                            cardType)
            return False

    def verifyConnectionbyPorts(self, lports):
        """
        verify every connection linked to a port in the port list given are
        correct
        """
        connl = self.topLevelModelBlock.getConnections()

        for con in connl:
            if con.getStartConnectionPoint() not in lports and \
                            con.getEndConnectionPoint() not in lports:
                continue

            ctrl = self.__projectCtx.getControllerFactory().\
                getItemController(con)
            scon = con.getStartConnectionPoint()
            econ = con.getEndConnectionPoint()
            if ctrl.canCreateConnection(self.topLevelModelBlock,
                                        con.getStartObject(),
                                        con.getEndObject(),
                                        sxpnt=scon,
                                        expnt=econ):
                self.getSceneItem(con).setConnColour(
                    WHGUI_COLOR_CON_POINT_DIC.get(scon.getCode()))
                self.getSceneItem(con).setConnStyle(
                    WHGUI_STYLE_CON_DIC.get(econ.getSpeed()))
                continue
            self.getSceneItem(con).setConnColour(WHGUI_ERROR_CON_COLOR)

    def createRackFunctionItem(self, pos, oper):
        """
            Add a rack function item by position
            :param pos: the function position
            :type pos: QPointF
        """

        # select the operator by mode
        if oper is None:
            oper = self.getOperatorMode()
        if oper is None: return
        name = whdb_getNextFunctionName(oper, self.topLevelModelBlock)
        # get the function by operator
        racks = [inst for inst in self.topLevelModelBlock.getOperatorInstances()
                 if isinstance(inst, WhhdbRackInstance)]
        racksIds = [int(rack.getParameter('id').getValue()) for rack in racks]
        func = self.topLevelModelBlock.createOperatorInstance(oper, name=name)
        # default value is 0
        if racksIds:
            maxId = max(racksIds)
            func.getParameter('id').setValue(maxId+1)
        # set the function position
        if self.getSnapFlag:
            cfg = self.config
            pos = wgui_snap_point(pos, cfg.snapGridX, cfg.snapGridY)
        func.setPosition(pos.x(), pos.y())
        # add function to the scene
        self.addFunctionItem(func).focusNameField()

    def addFunctionItem(self, func):
        """
        Create a function graphics item for the given function database
        object.

        :param func: The function object
        :type func: WhdbOperatorInstanceP
        :return: The newly created connection scene item
        :rtype: WguiOperatorInstanceSceneItem
        """
        if func is None: return None
        if ('wcr.cetc' in func.getOperator().getCode() or
                    func.getOperator() in self._defaultCard):
            return None
        if func.getZ() is None:
            func.setZ(self.getTopZValue() + 1)

        item = self.sitmF.createOperatorInstanceSceneItem(
            func, self.viewContext)
        # don't display card as block in scene
        if isinstance(func, WhhdbIoCardInstance):
            item.setVisible(False)
        self.addItem(item)
        return item

    def connectProjectItems(self, ctrl, tblk, eobj, sobj, expn, sxpn):
        """
            Function that to update port (code and speed) information
            if return False no connection should be created
        :return: bool
        """
        hasConns, points = ctrl.hasEndPortMultipleConnection(tblk,
                                                             sxpn,
                                                             expn)
        if (sxpn.getCode() in WHHDB_MODULES_IO_NAMES or
                    expn.getCode() in WHHDB_MODULES_IO_NAMES):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setText("Port Io can't be connected they are independent of the CetraC system")
            msg.setWindowTitle("Port Info Warning !")
            msg.exec_()
            return False

        if sxpn.getActive() == "Off" or expn.getActive() == "Off":
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setText("Port is disable, it cannot be connected")
            msg.setWindowTitle("Port Info Warning !")
            msg.exec_()
            return False

        if hasConns:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setText("Port {} already connected".format(
                ' and '.join(points)))
            msg.setWindowTitle("Port Info Warning !")
            msg.exec_()
            return False

        isRack = isinstance(eobj, WhhdbRackInstance)
        # Convert EndPoint Port using StartPoint Port
        if sxpn.getCode() != expn.getCode() and not isRack:
            if (sxpn.getCode() == WHHDB_XPOINT_CETRAC_CODE
                    or expn.getCode() == WHHDB_XPOINT_CETRAC_CODE):
                return True
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setText("Are you sure you want to change Port Type "
                        "and Speed ?")
            msg.setWindowTitle("Port Info Warning")
            msg.addButton(QtWidgets.QMessageBox.Yes)
            msg.addButton(QtWidgets.QMessageBox.No)
            ret = msg.exec_()
            if ret == msg.Yes:
                expn.setCode(sxpn.getCode())
                expn.setSpeed(sxpn.getSpeed())
            else:
                return False
        elif sxpn.getSpeed() != expn.getSpeed() and not isRack:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setText("Are you sure you want to change the Speed ?")
            msg.setWindowTitle("Port Info Warning")
            msg.addButton(QtWidgets.QMessageBox.Yes)
            msg.addButton(QtWidgets.QMessageBox.No)
            ret = msg.exec_()
            if ret == msg.Yes:
                expn.setSpeed(sxpn.getSpeed())
            else:
                return False
        return True