# ----------------------------------------------------------------------------
# - wigui-binding-table                                                       -
# - wild cruncher:icd binding dictionary widget                               -
# ----------------------------------------------------------------------------
# - author    (c) 2017-2017 achraf mahti                                     -
# - copyright (c) 2017-2017 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wigui_binding_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan icd binding dictionary widget

.. moduleauthor:: achraf mahti <mahti@silkan.net>
"""

from PyQt4.QtCore import QAbstractTableModel, Qt
from PyQt4.QtGui import QWidget, QItemDelegate, QLabel, QHBoxLayout, \
    QHeaderView, QAbstractItemView
from wcr_frontend.modules.icd.gui.wigui_table import WiguiTable
from wcr_core.modules.icd.db.wihdb_module_instance import WihdbModuleInstance


class WiguiBindingTable(WiguiTable):
    """
    Binding dictionary module/global table widget.
    """

    def __init__(self, pctx, scene, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param parent: Parent widget
        """
        super().__init__(pctx, scene, parent)
        self.__topBlock = pctx.getProjectDB().getTopHierarchy()
        #: Bus binding table model
        self.gblTblM = _WiguiBindingTableModel(scene, top=self.__topBlock,
                                               parent=self)
        self.__scene = scene
        #: Creating editor for element in table
        self.gblDlgt = _WiguiBindingDelegate()

        self._initElements()
        self.setLayout(self._createLayout())

        self._loadFromData()

    def _initElements(self):
        # super()._initElements()
        self.gblTblW.setModel(self.gblTblM)
        self.gblTblW.horizontalHeader().setResizeMode(0, QHeaderView.ResizeToContents)
        self.gblTblW.horizontalHeader().setResizeMode(1, QHeaderView.ResizeToContents)
        self.gblTblW.horizontalHeader().setResizeMode(2, QHeaderView.Stretch)
        self.gblTblW.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.gblTblW.setItemDelegate(self.gblDlgt)
        #: delegate classe for edition in table

    def _loadFromData(self):
        self.gblTblM.populateTable(self.__scene.getSceneItems())
        super()._loadFromData()

    def _createLayout(self):
        layout = super()._createLayout()
        btnLayout = QHBoxLayout()
        layout.addLayout(btnLayout)
        return layout

    def _bindEntry(self):
        """
        Add entry in the model
        """
        self.gblTblM.addRow()

    def _unbindEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.__gblTblW.selectedIndexes()
        """:type: list[QModelIndex]"""

        self.gblTblM.deleteRows([idx.row() for idx in selIdx])


class _WiguiBindingTableModel(QAbstractTableModel):
    """
    Global dictionary table model.
    """
    TITLE_GLOBAL_MSG = 'Messages'
    TITLE_PROD_MSG = 'Produced by'
    TITLE_CONS_MSG = 'Consumed by'

    HEADERS = [TITLE_GLOBAL_MSG, TITLE_PROD_MSG, TITLE_CONS_MSG]

    def __init__(self, scene, top, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent = parent
        self.__top = top
        #: Entries as list of tuples of
        #  global variable and module variable
        self.__entries = []

        scene.sceneItemsChanged.connect(self.populateTable)
        scene.sceneItemRemoved.connect(self.populateTable)
        scene.GlobalMsgChanged.connect(self.populateTable)
        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def populateTable(self, instances=None):
        self.deleteRows(range(self.rowCount()))
        messages = self.__top.getAllMessageDictBindings()

        for glbMsg in messages:
            producedMsg = messages[glbMsg].get('P', None)
            condumedMsg = '[{}]'.format(', '.join(messages[glbMsg].get('C', [])))
            self.addRow(glbMsg, producedMsg, condumedMsg)

    def addRow(self, globalVar=None, prodMsg=None, consMsg=None):
        """
        Add a row at the end of the list (append)

        :param globalVar: global variable name
        :type globalVar: Str
        :param prodMsg: variable type
        :type prodMsg: str
        :param consMsg: variable type
        :type consMsg: str
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append((globalVar, prodMsg, consMsg))

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices, Unbind=True):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row, Unbind=Unbind)

    def deleteRow(self, rowIndex, Unbind=True):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)
        self.__entries.pop(rowIndex)
        self.endRemoveRows()
        self._update()

    def data(self, index, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            row = index.row()
            column = index.column()
            info = self.__entries[row][column]
            return info
        return None

    def _setDataEntry(self, index, data):
        glbMsg, prodMsg, consMsg = self.__entries[index.row()]
        if glbMsg is not None and prodMsg is not None and consMsg is not None:
            #a binding could have been done by that line before
            # it have to be deleted.
            try:
                self.__parent.unbindPortProd(glbMsg, prodMsg, consMsg)
            except ValueError:
                pass

        #creating the new representation line
        nl = ()
        if index.column() == 0:
            nl = (data, prodMsg, consMsg)
        elif index.column() == 1:
            nl = (glbMsg, data, consMsg)
        elif index.column() == 2:
            nl = (glbMsg, prodMsg, data)

        self.__entries[index.row()] = nl

        glbMsg, prodMsg, consMsg = self.__entries[index.row()]
        if glbMsg is not None and prodMsg is not None and consMsg is not None:
            #line is full a binding have to be done
            try:
                self.__parent.bindPortProd(glbMsg, prodMsg, consMsg)
            except ValueError:
                self.__entries[index.row()] = (None, None, None)
                self._update()
                pass

    def flags(self, index):
        """
        set flags fot able
        """
        flgs = super().flags(index)
        #allowing editing
        flgs |= Qt.ItemIsEditable
        return flgs

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)


class _WiguiBindingDelegate(QItemDelegate):
    """
    Delegate for module dictionary table
    """

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        """
        Create a label containing for the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """
        editor = QLabel(parent)
        return editor