# ----------------------------------------------------------------------------
# - wigui-creation-type-props-dialog                                       -
# - wild cruncher - cockpit : graphic scene widget                           -
# ----------------------------------------------------------------------------
# - author    (c) 2018-2018 achraf mahti                                     -
# - copyright (c) 2018-2018 silkan                                           -
# ----------------------------------------------------------------------------

"""
.. module::       wigui_creation_type_props_dialog
   :platform:     Unix, Windows
   :synopsis:     Dezygner wigui graphics scene editor

.. moduleauthor:: achraf mahti <mahti@silkan.net>
"""

from PyQt5.QtCore import Qt, QAbstractTableModel
from PyQt5.QtWidgets import QMessageBox, QComboBox, QHBoxLayout, \
	QWidget, QVBoxLayout, QPushButton, QDialogButtonBox, QLineEdit, \
	QTableView, QItemDelegate, QSplitter, QAbstractItemView
from wcr_core.db.whdb_utils import whdb_getNextName

from wcr_core.modules.icd.db.wihdb_top_block import GENERIC_TYPES

from wcr_frontend import showErrorWindow, isNameValid, NameFormatMessageError
from wcr_frontend.gui.wgui_props_dialog_widget import \
    WguiXPointPropertiesBase, WguiXPointPropsTextField, \
    WguiXPointPropsFieldInstance, WguiXPointPropsField,\
    WGUI_NAME_PLACEHOLDER
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP, \
    WguiFlipRotationProperties, \
    WguiPropertiesDialog, WguiXPointProperties, WguiMultiPropertiesWidgetP

_WIDB_VARIABLE_NAME = 'name'
_WIDB_OFFSET_SIZE = 'offset (Byte)'
_WIDB_VARIABLE_TYPE = 'type'
_WIDB_VARIABLE_MSG = 'MessageBind'


class WiguiTypeInstanciateDialog(WguiPropertiesDialog):
    """
    The WguiOperatorInstancePropertiesDialog class is a dialog design to
    access and manipulates operator instance data. This include name,
    info and connection points.
    """

    _TYPE_NAME = 'Type Creation'

    _SHOW_OPERATOR_WIDGET = True

    # noinspection PyTypeChecker
    def __init__(self, pctx, updateF=None, selectXpnt=None, parent=None):
        """
        :param inst: Instance
        :type inst: WhdbOperatorInstanceP
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param updateF: Callback function to notify about saves
        :type updateF: _ -> _
        :param selectXpnt: Initially selected connection point, if any
        :type selectXpnt: WhdbConnectionPointP
        :param parent: Parent widget
        """
        super().__init__(self._TYPE_NAME, updateF, parent=parent)

        instWdgt = _WiguiMessageTypeInstanciate(pctx)
        self.instWdgt = instWdgt
        addiWdgts = self._createAdditionalInstancePropertiesWidgets(pctx)
        if addiWdgts:
            # noinspection PyTypeChecker
            generalWdgts = [instWdgt] + addiWdgts
        else:
            generalWdgts = [instWdgt]

        widgets = [('User Type', WguiMultiPropertiesWidgetP(*generalWdgts))]

        widgets.extend(self._createAdditionalPropertiesWidgets(pctx))

        xpntWdgt = self._createConnectionPointsPropertiesWidget(pctx)
        if xpntWdgt:
            widgets.append(xpntWdgt)


        # noinspection PyUnresolvedReferences
        self.setTabbedPropertyWidgets(
            *widgets,
            activate=xpntWdgt[1] if selectXpnt and xpntWdgt else None)


    def showOperatorWidget(self):
        return True

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _createAdditionalInstancePropertiesWidgets(self, pctx):
        return []

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _createAdditionalPropertiesWidgets(self, pctx):
        return []

    # noinspection PyMethodMayBeStatic
    def _createConnectionPointsPropertiesWidget(self, pctx):
        pass

    def _setMainLayout(self, widgets, widget):
        """
        Set the dialog's property widgets.

        :param widgets: Property widgets, and their titles
        :type widgets: Iterable[(str, WguiPropertiesWidgetP)]
        :param widget: Wrapper widget containing the property widgets
        :type widget: QWidget
        """
        self.__widgets = [widget for _, widget in widgets]

        layout = QVBoxLayout()
        layout.addWidget(widget)
        buttons = QDialogButtonBox(QDialogButtonBox.Ok |
                                   QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        self.okStateChanged.connect(
            buttons.button(QDialogButtonBox.Ok).setEnabled)
        layout.addWidget(buttons)
        # create the main layout
        self.setLayout(layout)

    def accept(self):
        hasErr = self._save()
        if not hasErr:
            super().accept()

    def _apply(self):
        """
        Apply the current changes, and leave the dialog window open.
        """
        self._save()
        self._updateFields()

    def _reset(self):
        """
        Reset the dialog window.
        """
        self._updateFields()

    def _save(self):
        """
        Save all changes.
        """
        hasErr = False
        for widget in self.__widgets:
            widgErr = widget.save()
            if widgErr:
                hasErr = True

        return hasErr

    def _updateFields(self):
        """
        Reset all fields to their original state.
        """
        for widget in self.__widgets:
            widget.updateFields()


# noinspection PyAbstractClass
class _WiguiMessageTypeInstanciate(WguiXPointPropertiesBase):
    GLB_COL = 0
    LOC_COL = 1

    def __init__(self, pctx, instance=None, parent=None, ConverifF=None):
        """
        :param instance: The instance object
        :type instance: WhhdbHostInstance
        :param pctx: Project context
        :type pctx: WhmodHardwareProjectContextP
        :param parent: Parent widget
        """
        self.__module = pctx.getModule()
        """:type: WpmodSystemApplicationModuleP"""
        self.__xpntTypeNames = ()
        self.msg2variable = {}
        self.__globrecords = []
        self.__newlocalrecords = []
        self.__topBlock = pctx.getProjectDB().getTopHierarchy()
        # init table necessary for
        self.table = None
        self.__index = -1

        super(_WiguiMessageTypeInstanciate, self). \
            __init__(instance, parent=parent)

        self.__message_tab = _WhguiMessageTableInstance(pctx, self.__topBlock,
                                                        parent=self)

        #: Configuration table view
        if self.__class__ is _WiguiMessageTypeInstanciate:
            self._finishInitialization()
        self.table = self.ui.xpntTableF

        self.__ConnectionUpdateF = ConverifF

    def updateFields(self):
        pass

    def createFields(self):
        return {
            _WIDB_VARIABLE_NAME: WguiXPointPropsTextField(
                _WIDB_VARIABLE_NAME, 'Name', _WIDB_VARIABLE_NAME),
            _WIDB_OFFSET_SIZE: WguiXPointPropsTextField(
                _WIDB_OFFSET_SIZE, 'Offset (Byte)', _WIDB_OFFSET_SIZE),
            _WIDB_VARIABLE_TYPE: WguiXPointPropsTextField(
                _WIDB_VARIABLE_TYPE, 'type', _WIDB_VARIABLE_TYPE),
        }

    def getFieldOrder(self):
        return [_WIDB_VARIABLE_NAME, _WIDB_OFFSET_SIZE, _WIDB_VARIABLE_TYPE]

    def getStretchFields(self):
        return {_WIDB_VARIABLE_NAME, _WIDB_OFFSET_SIZE, _WIDB_VARIABLE_TYPE}

    def getButtons(self):

        def add():
            curMsg = self.__message_tab.selectedMsg(role=Qt.DisplayRole)

            if curMsg is None:
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Information)
                msg.setText("Select a message before addding a variable.")
                msg.setWindowTitle("Add variable warning")
                msg.exec_()
                return
            varName = 'variable'

            if self.msg2variable.get(curMsg, None):
                allMsgVar = [var for (var, _, _, _) in self.msg2variable.get(curMsg, [])]
                varName = whdb_getNextName(varName, allMsgVar)

            vars = [varName, '0', 'int', curMsg]
            self.model.addXPoint(**{
                _WIDB_VARIABLE_NAME: vars[0],
                _WIDB_OFFSET_SIZE: vars[1],
                _WIDB_VARIABLE_TYPE: vars[2],
                _WIDB_VARIABLE_MSG: vars[3]})
            self.msg2variable.setdefault(curMsg, []).append(vars)

        def deleteRow():
            # remove the last index
            curMsg = self.__message_tab.selectedMsg(role=Qt.DisplayRole)
            if curMsg is None:
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Information)
                msg.setText("Select a message before deleting a variable.")
                msg.setWindowTitle("Delete variable warning")
                msg.exec_()
                return
            selIdx = self.__message_tab.pcicTblW.selectedIndexes()[0]
            # get last index
            indexes = self.model.getIndices()
            if len(indexes) == 0:
                return
            lastIdx = tuple(indexes)[-1]
            self.model.removeXPoint(lastIdx)
            self.model.setLastIndex(self.model.lastIndex - 1)
            del self.msg2variable[curMsg][-1]
            self.__message_tab.pcicTblW.setCurrentIndex(selIdx)

        addBtn = QPushButton('Add')
        delBtn = QPushButton('Delete')
        addBtn.clicked.connect(add)
        delBtn.clicked.connect(deleteRow)

        return [addBtn, delBtn]

    def createXPoint(self, idx, model, instance):
        pass

    def resetModel(self, model, selectedMsg):
        if not selectedMsg:
            model.setXPoints({})
            return
        model.setXPoints(
            {_WIDB_VARIABLE_NAME: name,
             _WIDB_OFFSET_SIZE: offset,
             _WIDB_VARIABLE_TYPE: type_,
             _WIDB_VARIABLE_MSG: msg}
            for name, offset, type_, msg in self.msg2variable.get(selectedMsg, [])
        )

    def updateXPoint(self, idx=None, model=None):
        for idx in self.model.getIndices():
            name = self.model.getValue(idx, _WIDB_VARIABLE_NAME)
            offset = self.model.getValue(idx, _WIDB_OFFSET_SIZE)
            type_ = self.model.getValue(idx, _WIDB_VARIABLE_TYPE)
            msg = self.model.getValue(idx, _WIDB_VARIABLE_MSG)
            if msg:
                self.msg2variable[msg][idx - 1] = [name, offset, type_, msg]

    def _createLayout(self):
        container = QWidget()
        container_Layout = QHBoxLayout()
        splitprinc = QSplitter()
        splitprinc.setOrientation(Qt.Horizontal)

        left = QWidget()
        left.setLayout(self.__message_tab.createLayout())
        splitprinc.addWidget(left)

        right = QWidget()
        right.setLayout(super()._createLayout())
        container_Layout.addWidget(right)
        container.setLayout(container_Layout)
        splitprinc.addWidget(container)
        layout = QHBoxLayout()
        layout.addWidget(splitprinc)
        layout.setContentsMargins(2,2,2,2)

        return layout

    def save(self):
        return self.saveUserTypes()

    def saveUserTypes(self):

        hasError = self.check_before_saving()
        if hasError:
            if getattr(self, 'instance', None):
                self.instance.setOperator(None)
            return hasError

        indexes = self.__message_tab.getSelectedIndexes()
        if len(indexes) > 0:
            idx = indexes[0]
            self.__message_tab.updateVariables(idx)

        # init with default types
        self.__topBlock.iniGlobalType()
        # add all new user Type
        for userType in self.__message_tab.getAllTypes():
            self.__topBlock.addType(userType, self.msg2variable[userType])
        return hasError

    def check_before_saving(self):
        # return True if an error was found
        if len(self.__message_tab.getSelectedIndexes()) > 0:
            idx = self.__message_tab.getSelectedIndexes()[0]
            self.__message_tab.updateVariables(idx)

        # logical checks
        for userType in self.__message_tab.getAllTypes():
            # check that all messages have valid name
            if not isNameValid(userType):
                showErrorWindow('Type Error',
                                NameFormatMessageError, userType)
                return True
            # check that all messages have variable
            if userType not in self.msg2variable:
                showErrorWindow('Type Error',
                                'Type {} has no variables declared.', '')
                return True
            # check that all variables have valid name
            for varName, _, _, _ in self.msg2variable[userType]:
                if not isNameValid(varName):
                    showErrorWindow('Variable Error',
                                    NameFormatMessageError, varName)
                    return True
            for varName, _, type_, _ in self.msg2variable[userType]:
                if type_ not in GENERIC_TYPES:
                    showErrorWindow('Variable Error',
                                    'Variable {} has type unknown ('+type_+').', varName)
                    return True
        # check that message name is not empty
        for userType in self.__message_tab.getAllTypes():
            if userType == '':
                showErrorWindow('Type Error',
                                'Type name can\'t be empty ', '')
                return True

        return False


class _WhguiMessageTableInstance(QWidget):
    def __init__(self, pctx, top, parent=None):
        """
        :param pctx: Project context
        :type pctx:
        :param instance: The instance edited
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent = parent
        self.__top = top

        """:type: WpmodSystemApplicationModuleP"""
        # : Project context
        self.__pctx = pctx
        self.__pcicTblW = QTableView(self)
        self.__tempPreviousModel = []

        self._initModels()
        self._initElements()
        self._loadFromData()

    @property
    def pcicTblM(self):
        return self.__pcicTblM

    @property
    def pcicTblW(self):
        return self.__pcicTblW

    @property
    def pcicItmDlgt(self):
        return self.__pcicItmDlgt

    def _initModels(self):
        self.__pcicTblM = _WhguiMessageTableInstanceModel(self.formatMsgs,
                                                          parent=self)
        self.__pcicItmDlgt = _WhguiMessageBindDelegate(parent=self)

    def _initElements(self):
        #: Add entry button
        self.__addButton = QPushButton('Add')
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')
        self.__saveButton = QPushButton('Save')
        self.__pcicTblW.setModel(self.pcicTblM)
        self.__pcicTblW.horizontalHeader().setSectionResizeMode(1)
        self.__pcicTblW.setMaximumWidth(300)
        self.__pcicTblW.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.__pcicTblW.setItemDelegate(self.pcicItmDlgt)

        self.__addButton.clicked.connect(self._addEntry)
        self.__deleteButton.clicked.connect(self._deleteSelEntries)
        self.__saveButton.clicked.connect(self._saveUserTypes)
        self.__pcicTblW.selectionModel(). \
            selectionChanged.connect(self.selecChanged)

    def _loadFromData(self):
        for type_, variables in self.__top.getGlobalTypes():
            if type_ not in GENERIC_TYPES:
                self.pcicTblM.addRow(type_)
                self.__parent.msg2variable[type_] = variables

    def _saveUserTypes(self):
        self.__parent.saveUserTypes()

    def resettab(self):
        """
        reset self.__pcicTblM
        """
        self.pcicTblM.deleteRows(range(0, self.pcicTblM.rowCount()))

    def getAllTypes(self):
        """
        return the description of all card describes in the Table
        """
        return self.pcicTblM.getAllUserTypes()

    def formatMsgs(self, data, column):
        """
        :param data: data to format
        :type data: Cardname | Cardtype | [PortUUID]
        """
        if data is None:
            return ''
        if column == 0:
            return data
        elif column == 1:
            return data
        else:
            return 'No formating available'

    def createLayout(self):
        layout = QVBoxLayout()
        layout.addWidget(self.__pcicTblW)

        btnLayout = QHBoxLayout()
        btnLayout.addStretch(1)
        btnLayout.addWidget(self.__addButton)
        btnLayout.addWidget(self.__deleteButton)
        btnLayout.addWidget(self.__saveButton)
        layout.addLayout(btnLayout)

        return layout

    def _addEntry(self):
        """
        Add entry in the model
        """
        index = self.pcicTblM.addRow()
        self.__pcicTblW.setCurrentIndex(index)

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.__pcicTblW.selectedIndexes()
        """:type: list[QModelIndex]"""
        self.pcicTblM.deleteRows([idx.row() for idx in selIdx])
        rl = []
        for idx in selIdx:
            rl.append(idx.row())
        rowlist = sorted(set(rl))
        if self.pcicTblM.rowCount() > 0 and rowlist:
            if rowlist[0] >= self.pcicTblM.rowCount():
                self.__pcicTblW.setCurrentIndex(self.pcicTblM.
                                                index(self.pcicTblM.rowCount() - 1, 0))
            else:
                self.__pcicTblW.setCurrentIndex(self.pcicTblM.
                                                index(rowlist[0], 0))

    def removeMsg(self, glbMsg):
        """
        remove message and variables associated to it
        """
        if glbMsg in self.__parent.msg2variable:
            del self.__parent.msg2variable[glbMsg]
        self.__parent.resetModel(self.__parent.model, glbMsg)

    def selectedMsg(self, selectignor=False, role=Qt.UserRole):
        lrow = []
        for index in self.getSelectedIndexes():
            lrow.append(index.row())
        ulrow = set(lrow)
        if len(ulrow) == 1 and not selectignor:
            return self.pcicTblM.data(self.pcicTblM.index(lrow[0], 0),
                                      role=role)
        else:
            return None

    def getSelectedIndexes(self):
        return self.__pcicTblW.selectionModel().selectedIndexes()

    def selecChanged(self, newselec, oldselect):
        """
        Signal handler : when selection change reset connection editor with
        the param of the new card selected
        """
        if len(newselec.indexes()) == 0:
            self.updateLastSelection()
            self.__parent.resetModel(self.__parent.model, None)
            return
        idx = newselec.indexes()[0]
        self.updateVariables(idx)

    def updateVariables(self, idx):
        row = idx.row()
        glbMsg = self.getAllTypes()[row]
        self.__parent.updateXPoint()
        self.__parent.resetModel(self.__parent.model, glbMsg)

    def updateLastSelection(self):
        self.__parent.updateXPoint()


class _WhguiMessageTableInstanceModel(QAbstractTableModel):
    """
    icd Binding table model.
    """

    TITLE_TYPE_NAME = 'User type name'
    HEADERS = [TITLE_TYPE_NAME]

    def __init__(self, formatF, parent=None):
        """
        :param XpntTypeNames: The Names of Xpoint of the mule
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent = parent
        self.oldParent = parent.parent()

        self.__formatF = formatF

        #: Entries as list of tuples of
        #  port, message and type
        self.__entries = []
        self._update()

    @property
    def entries(self):
        return self.__entries

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
                                    0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column <= self.columnCount():
            info = self.__entries[row]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return self.__formatF(info, column)
        return None

    def setData(self, index, data, role=None):
        assert role == Qt.UserRole
        self._setMessageEntry(index, data)

    def _setMessageEntry(self, index, data):
        oldType = self.__entries[index.row()]
        if not self.checkUnicityMsg(data, index.row()):
            return
        if index.column() == 0:
            if oldType in self.oldParent.msg2variable:
                if oldType != data:
                    # update the dictionary variable
                    self.oldParent.msg2variable[data] = self.oldParent.msg2variable[oldType]
                    for i in range(len(self.oldParent.msg2variable[data])):
                        self.oldParent.msg2variable[data][i][-1] = data
                    self.oldParent.resetModel(self.oldParent.model, data)
                    del self.oldParent.msg2variable[oldType]
                self.__entries[index.row()] = data

    def checkUnicityMsg(self, msg, row):
        for i, entry in enumerate(self.__entries):
            if i != row:
                if msg == entry:
                    return False
        return True

    def getAllUserTypes(self):
        return self.__entries

    def addRow(self, TypeName='newType'):
        """
        Add a row at the end of the list (append)

        :param TypeName: The type name
        :type TypeName: str | None
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())
        # help uniqueness by auto increment name
        allUserTypes = self.getAllUserTypes()
        nextType = whdb_getNextName(TypeName, allUserTypes)
        self.__entries.append(nextType)
        self.oldParent.msg2variable[nextType] = []

        self.endInsertRows()
        self._update()
        return index

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        userType = self.__entries[rowIndex]
        if userType is not None:
            try:
                # Unbound function of message
                self.__parent.removeMsg(userType)
                pass
            except ValueError:
                pass
        self.__entries.pop(rowIndex)
        self.endRemoveRows()
        self._update()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def flags(self, index):
        flgs = super().flags(index)
        flgs |= Qt.ItemIsEditable
        return flgs


class _WhguiMessageBindDelegate(QItemDelegate):
    """
    Delegate for message items
    """

    def __init__(self, parent=None):
        """
        :type CardTyp: () -> list[CardTyp]
        :type formatF: (elem,column) -> str
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        """
        Create a editors for message table the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """

        def save(*_):
            self.setModelData(editor, index.model(), index)

        editor = QLineEdit(parent)
        editor.editingFinished.connect(save)

        return editor

    def setEditorData(self, editor, index):
        """
        :type editor: QComboBox | QLineEdit
        :type index: QModelIndex
        """
        data = index.model().data(index, Qt.EditRole)
        editor.setText(data)

    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox | QLineEdit
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        if editor.text():
            model.setData(index, editor.text(), Qt.UserRole)
