# ----------------------------------------------------------------------------
# - wigui-global-table                                                       -
# - wild cruncher:icd global dictionary widget                               -
# ----------------------------------------------------------------------------
# - author    (c) 2017-2017 achraf mahti                                     -
# - copyright (c) 2017-2017 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wigui_global_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan icd global dictionary widget

.. moduleauthor:: achraf mahti <mahti@silkan.net>
"""

import re
from PyQt4.QtCore import QAbstractTableModel, Qt
from PyQt4.QtGui import QWidget, QItemDelegate, QPushButton, \
    QHBoxLayout, QLineEdit, QComboBox, QBrush, QColor, \
    QHeaderView, QAbstractItemView
from wcr_frontend.modules.icd.gui.wigui_table import WiguiTable
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_frontend.modules.icd.gui.wigui_creation_type_props_dialog import \
    WiguiTypeInstanciateDialog
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError


class WiguiGlobalTable(WiguiTable):
    """
    Global dictionary binding to module table widget.
    """

    def __init__(self, pctx, scene, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param parent: Parent widget
        """
        super().__init__(pctx, scene, parent)
        self.__scene = scene
        self.__pctx = pctx
        #: formating function for element ot table
        self.__topBlock = pctx.getProjectDB().getTopHierarchy()
        #: Bus binding table model
        self.gblTblM = _WiguiGlobalTableModel(self.__topBlock, scene,
                                              parent=self)

        #: Creating function geting all available
        self.__msgType = self.getMessagesType()
        #: Creating editor for element in table
        self.gblDlgt = _WiguiGlobalDelegate(self.__topBlock)

        #: Add entry button
        self.__addButton = QPushButton('Add')
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')
        #: Create type button
        self.__createTypeButton = QPushButton('Create Type')

        self._initElements()
        self.setLayout(self._createLayout())

        self._loadFromData()

    def _initElements(self):
        self.gblTblW.setModel(self.gblTblM)
        self.gblTblW.horizontalHeader().setResizeMode(0, QHeaderView.ResizeToContents)
        self.gblTblW.horizontalHeader().setResizeMode(1, QHeaderView.ResizeToContents)
        self.gblTblW.horizontalHeader().setResizeMode(2, QHeaderView.Stretch)
        self.gblTblW.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.gblTblW.setItemDelegate(self.gblDlgt)
        self.__addButton.clicked.connect(self._addEntry)
        self.__deleteButton.clicked.connect(self._deleteSelEntries)
        self.__createTypeButton.clicked.connect(self._openTypeCreationWindow)

    def _loadFromData(self):
        self.gblTblM.populateTable()
        super()._loadFromData()

    def _createLayout(self):
        layout = super()._createLayout()
        btnLayout = QHBoxLayout()
        layout.addLayout(btnLayout)
        btnLayout.addWidget(self.__addButton)
        btnLayout.addWidget(self.__deleteButton)
        btnLayout.addWidget(self.__createTypeButton)
        return layout

    def _addEntry(self):
        """
        Add entry in the model
        """
        self.gblTblM.addRow()

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.gblTblW.selectedIndexes()
        """:type: list[QModelIndex]"""

        self.gblTblM.deleteRows([idx.row() for idx in selIdx])

    def _openTypeCreationWindow(self):
        instanciate = WiguiTypeInstanciateDialog(pctx=self.__pctx)
        instanciate.show()
        instanciate.exec_()

    def getGlobalDict(self):
        """
        Get all global messages with their type.
        :return: messages type
        """
        return self.__topBlock.getGlobalTypes()

    def getMessagesType(self):
        """
        Get all messages type default + user.
        :return: messages type
        """
        return self.__topBlock.getAllMessages()


class _WiguiGlobalTableModel(QAbstractTableModel):
    """
    Global dictionary table model.
    """
    TITLE_VARIABLE = 'Message'
    TITLE_VAR_TYPE = 'Type'
    TITLE_VAR_SIZE = 'Payload Size (byte)'

    HEADERS = [TITLE_VARIABLE, TITLE_VAR_TYPE, TITLE_VAR_SIZE]

    def __init__(self, topBlock, scene, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent = parent
        self.__scene = scene

        #: Entries as list of tuples of
        #  Variable and Type
        self.__entries = []
        """:type: list[]"""
        self.__topBlock = topBlock
        self.color_back = Qt.red

        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def populateTable(self):
        self.deleteRows(range(self.rowCount()))
        datas = self.__parent.getMessagesType()
        for msg, (type_, size) in datas.items():
            self.addRow(msg, type_, size)

    def addRow(self, message=None, varType=None, varSize=None):
        """
        Add a row at the end of the list (append)

        :param variable: global variable name
        :type variable: Str
        :param varType: variable type
        :type varType: str
        :param varSize: message size
        :type varSize: str
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append((message, varType, varSize))

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices, Unbind=True):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row, Unbind=Unbind)

    def deleteRow(self, rowIndex, Unbind=True):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)
        msg = self.__entries[rowIndex][0]
        if self.__topBlock.getAllMessageBindings().get(msg, []):
            showErrorWindow('Global message Error',
                            'Message {} is binded to a module message'
                            ' so it can\'t be deleted.', msg)
            return
        self.__topBlock.removeMessageType(msg)
        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def data(self, index, role=None):
        row = index.row()
        col = index.column()
        info = self.__entries[row][col]
        if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
            return info
        elif role == Qt.UserRole:
            return info
        if role == Qt.BackgroundRole:
            self.isEditable(index, row)
            return QBrush(self.color_back)
        return None

    def setData(self, index, data, role=None):
        """
        :type index: QModelIndex
        :type data: object
        """
        assert role == Qt.UserRole
        self._setDataEntry(index, data)

    def _setDataEntry(self, index, data):
        msg, varType, size = self.__entries[index.row()]
        if not self.checkUnicityMsg(data, index.row()):
            return
        entry = ()
        if index.column() == 0:
            entry = (data, varType, size)
        elif index.column() == 1:
            entry = (msg, data, size)
        elif index.column() == 2:
            entry = (msg, varType, data)
        self.__entries[index.row()] = entry
        self.__topBlock.addMsgType(entry[0], entry[1], entry[2])

    def checkUnicityMsg(self, msg, row):
        for i, (entry) in enumerate(self.__entries):
            if i != row:
                if msg == entry[0]:
                    return False
        return True

    def isEditable(self, index, row):
        msg = index.sibling(row, 0).data()
        type_ = index.sibling(row, 1).data()
        size = index.sibling(row, 2).data()
        if not msg or not type_ or not size:
            self.color_back = QColor(255, 180, 180)
            return True
        else:
            self.color_back = Qt.white
            return False

    def flags(self, index):
        """
        set flags fot able
        """
        flgs = super().flags(index)
        row = index.row()
        if self.isEditable(index, row):
            #allowing editing
            flgs |= Qt.ItemIsEditable
            # self.setColortoRow(row, 'red')
        else:
            flgs &= ~Qt.ItemIsEditable
        return flgs

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)

    def setColortoRow(self, rowIndex, color):
        for j in range(self.columnCount()):
            self.item(rowIndex, j).setBackground(color)


class _WiguiGlobalDelegate(QItemDelegate):
    """
    Delegate for global dictionary table
    """

    def __init__(self, top, parent=None):
        """
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)
        self.__top = top

    def createEditor(self, parent, option, index):
        """
        Create a editors for message table the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """

        def save(*_):
            self.setModelData(editor, index.model(), index)

        col = index.column()
        # global msg name
        if col == 0:
            editor = QLineEdit(parent)
        # message type
        elif col == 1:
            editor = QComboBox(parent)
            for type_, _ in self.__top.getGlobalTypes():
                editor.addItem(type_, type_)
        if col == 2:
            editor = QLineEdit(parent)
        return editor

    def setEditorData(self, editor, index):
        """
        :type editor: QComboBox | QLineEdit
        :type index: QModelIndex
        """
        data = index.model().data(index, Qt.EditRole)
        if isinstance(editor, QComboBox):
            for index in range(editor.count()):
                if editor.itemData(index, Qt.UserRole) == data:
                    editor.setCurrentIndex(index)
                    break
            else:
                editor.setCurrentIndex(0)
        elif isinstance(editor, QLineEdit):
            editor.setText(data)

    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox | QLineEdit
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        if isinstance(editor, QComboBox):
            model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)
        elif isinstance(editor, QLineEdit):
            if index.column() == 0:
                if editor.text():
                    if isNameValid(editor.text()):
                        model.setData(index, editor.text(), Qt.UserRole)
                    else:
                        showErrorWindow("Name Error", NameFormatMessageError,
                                        editor.text())
            elif index.column() == 2:
                if editor.text():
                    if re.match("^[0-9]+$", editor.text()) is None:
                        showErrorWindow("Payload size Error", "Value '{}' is incorrect,"
                                        " needs to be a whole value.",
                                        editor.text())
                    else:
                        model.setData(index, editor.text(), Qt.UserRole)
