# ----------------------------------------------------------------------------
# - wbgui-blbxmessage-instance-props-dialog                                  -
# - wild cruncher - cockpit :                                                -
# -         Black box message properties dialog widgets                      -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wbgui_blbxmessage_instance_props_dialog
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher instance properties dialog

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""

from collections import Counter
from PyQt5.QtCore import Qt, QAbstractTableModel
from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QWidget, QVBoxLayout, \
	 QLabel, QLineEdit, QTableView, QItemDelegate, QDialogButtonBox

from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_frontend.gui.wgui_operator_instance_props_dialog import \
    WguiOperatorInstancePropertiesDialog
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP
from wcr_frontend.modules.icd.gui.wigui_operator_props_dialog import \
    WiguiOperatorInstanciateDialog
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError

_WIDB_VARIABLE_NAME = 'name'
_WIDB_OFFSET_SIZE = 'offset (Byte)'
_WIDB_VARIABLE_TYPE = 'type'
_WIDB_VARIABLE_MSG = 'MessageBind'


class WiguiInstanceModuleProperties(WguiPropertiesWidgetP):
    """
        Configuration widget for a  data.
        """

    def __init__(self, typeName, inst, tophm, parent=None):
        """
        :param typeName: Type name of the displayed element
        :type typeName: str
        :param inst: the instance object
        :type inst: WhdbInstanceP
        :param tophm: Hierarchy top level
        :type tophm: WhdbBlock
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Type name
        self.__typeName = typeName
        #: instance object
        self.__instance = inst
        #: Hierarchy top level
        self.__tophm = tophm

        #: Name field
        self.__nameF = QLineEdit()
        self.__nameF.setFixedWidth(200)
        #: Info field
        self.__moduleType = QLineEdit()
        self.__moduleType.setReadOnly(True)
        self.__moduleType.setFixedWidth(100)

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        container = QWidget()
        layout = QVBoxLayout()
        layout.addWidget(self.__nameF)
        layout.addWidget(container)
        layout.addWidget(QLabel('Module type:'))
        layout.addWidget(self.__moduleType)
        layout.addStretch(1)
        return layout

    def updateFields(self):
        inst = self.__instance
        self.__nameF.setText(inst.getName())
        self.__moduleType.setText(inst.getModuleType())

    def save(self):
        inst = self.__instance
        previousName = inst.getName()
        if isNameValid(self.__nameF.text()):
            inst.setName(self.__nameF.text())
        else:
            showErrorWindow("Name Error", NameFormatMessageError,
                            self.__nameF.text())

        allEqpNames = tuple(e.getName() for e in self.__tophm.getOperatorInstances())
        counter = Counter(allEqpNames)
        commons = counter.most_common()
        if commons[0][1] > 1:
            showErrorWindow("Duplicate Name Error", "The name {} is already used for another equipement!",
                            commons[0][0])
            # keep previous name
            inst.setName(previousName)
            return True


class WiguiModulePropertiesDialog(WiguiOperatorInstanciateDialog):
    _TYPE_NAME = 'Module '

    _SHOW_OPERATOR_WIDGET = False
    GENERAL_TAB = WiguiInstanceModuleProperties

    def __init__(self, inst, pctx, selectXpnt=None, parent=None):
        """
        :param inst: the instance object
        :type inst: WhdbInstanceP
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param updateF: Callback function to notify about saves
        :type updateF: () -> _
        :param parent: Parent widget
        """
        self.oper = inst
        super().__init__(pctx, inst, selectXpnt=selectXpnt, oper=inst)
        self.setMinimumWidth(600)

    def _createAdditionalInstancePropertiesWidgets(self, pctx):
        return []

    # noinspection PyMethodMayBeStatic
    def _createAdditionalPropertiesWidgets(self, pctx):
        return [("Module dictionary", _WiguiModuleMessageProperties(self.oper, pctx))]

    def _createConnectionPointsPropertiesWidget(self, pctx):
        return None

    def _setMainLayout(self, widgets, widget):
        """
        Set the dialog's property widgets.

        :param widgets: Property widgets, and their titles
        :type widgets: Iterable[(str, WguiPropertiesWidgetP)]
        :param widget: Wrapper widget containing the property widgets
        :type widget: QWidget
        """
        self.__widgets = [widget for _, widget in widgets]

        layout = QVBoxLayout()
        layout.addWidget(widget)
        buttons = QDialogButtonBox(QDialogButtonBox.Ok)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        self.okStateChanged.connect(
            buttons.button(QDialogButtonBox.Ok).setEnabled)
        layout.addWidget(buttons)
        # create the main layout
        self.setLayout(layout)

    def accept(self):
        hasErr = self._save()
        if not hasErr:
            super().accept()

    def _apply(self):
        """
        Apply the current changes, and leave the dialog window open.
        """
        self._save()
        self._updateFields()

    def _reset(self):
        """
        Reset the dialog window.
        """
        self._updateFields()

    def _save(self):
        """
        Save all changes.
        """
        pass

    def _updateFields(self):
        """
        Reset all fields to their original state.
        """
        for widget in self.__widgets:
            widget.updateFields()


class WiguiModuleInstancePropertiesDialog(WguiOperatorInstancePropertiesDialog):
        _TYPE_NAME = 'Module Instance '

        _SHOW_OPERATOR_WIDGET = False
        GENERAL_TAB = WiguiInstanceModuleProperties

        def __init__(self, inst, pctx, updateF, selectXpnt=None, parent=None):
            """
            :param inst: the instance object
            :type inst: WhdbInstanceP
            :param pctx: Project context
            :type pctx: WcorProjectContextP
            :param updateF: Callback function to notify about saves
            :type updateF: () -> _
            :param parent: Parent widget
            """
            super().__init__(inst, pctx, updateF, selectXpnt, parent)
            self.setMinimumWidth(600)

        def _createAdditionalInstancePropertiesWidgets(self, inst, pctx):
            return []

        # noinspection PyMethodMayBeStatic
        def _createAdditionalPropertiesWidgets(self, inst, pctx):
            return [("Module dictionary", _WiguiModuleMessageInstanciate(inst, pctx))]

        def _createConnectionPointsPropertiesWidget(self, inst, pctx):
            return None


# noinspection PyAbstractClass
class _WiguiModuleMessageInstanciate(QWidget):
    GLB_COL = 0
    LOC_COL = 1

    def __init__(self, instance, pctx, parent=None, ConverifF=None):
        """
        :param instance: The instance object
        :type instance: WhhdbHostInstance
        :param pctx: Project context
        :type pctx: WhmodHardwareProjectContextP
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.instance = instance
        self.__module = pctx.getModule()
        """:type: WpmodSystemApplicationModuleP"""
        self.__xpntTypeNames = ()
        self.messages = self.instance.getData()
        self.__topBlock = pctx.getProjectDB().getTopHierarchy()
        self.hasError = True
        # init table necessary for
        self._msgTblW = QTableView(self)
        self._initModels()
        self._initElements()
        self.createLayout()

        self._loadFromData()

    @property
    def moduleName(self):
        return self.instance.getName()

    @property
    def msgTblM(self):
        return self.__msgTblM

    @property
    def msgTblW(self):
        return self._msgTblW

    @property
    def msgDlgt(self):
        return self.__msgDlgt

    def _initModels(self):
        self.__msgTblM = _WhguiMessageTableInstanceModel(self.formatMsgs,
                                                         parent=self)
        self.__msgDlgt = _WhguiMessageBindDelegate(self.__topBlock, parent=self)

    def _initElements(self):
        self.msgTblW.setModel(self.msgTblM)
        self.msgTblW.horizontalHeader().setSectionResizeMode(1)
        # self.__msgTblW.setMaximumWidth(600)
        self.msgTblW.setItemDelegate(self.__msgDlgt)

    def _loadFromData(self):
        modName = self.instance.getName()
        for msg, (type_, consprod, storage, period) in self.messages.items():
            glbMsg = self.__topBlock.getGlbMsgByLocalMsg(modName, msg)
            self.msgTblM.addRow(glbMsg, type_, msg, consprod, storage, period)

    def updateFields(self):
        pass

    def resettab(self):
        """
        reset self.__pcicTblM
        """
        self.msgTblM.deleteRows(range(0, self.msgTblM.rowCount()))

    def getMsgs(self):
        """
        return the description of all card describes in the Table
        """
        return self.msgTblM.getFullMessages()

    def formatMsgs(self, data, column):
        """
        :param data: data to format
        :type data: Cardname | Cardtype | [PortUUID]
        """
        if data is None:
            return ''
        return data

    def createLayout(self):
        layout = QVBoxLayout()
        layout.addWidget(self.msgTblW)

        btnLayout = QHBoxLayout()
        btnLayout.addStretch()
        layout.addLayout(btnLayout)

        self.setLayout(layout)

        return layout

    def _addEntry(self):
        """
        Add entry in the model
        """
        self.msgTblM.addRow()

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.msgTblW.selectedIndexes()
        """:type: list[QModelIndex]"""
        self.msgTblM.deleteRows([idx.row() for idx in selIdx])
        rl = []
        for idx in selIdx:
            rl.append(idx.row())
        rowlist = sorted(set(rl))
        if self.msgTblM.rowCount() > 0 and rowlist:
            if rowlist[0] >= self.pcicTblM.rowCount():
                self.msgTblW.setCurrentIndex(self.msgTblM.
                                             index(self.msgTblM.rowCount() - 1, 0))
            else:
                self.msgTblW.setCurrentIndex(self.msgTblM.
                                             index(rowlist[0], 0))

    def removeMsg(self, glbMsg):
        """
        remove message and variables associated to it
        """
        if glbMsg in self.__parent.msg2variable:
            del self.__parent.msg2variable[glbMsg]
        self.__parent.resetModel(self.__parent.model, glbMsg)

    def selectedMsg(self, selectignor=False, role=Qt.UserRole):
        lrow = []
        for index in self.getSelectedIndexes():
            lrow.append(index.row())
        ulrow = set(lrow)
        if len(ulrow) == 1 and not selectignor:
            return self.msgTblM.data(self.msgTblM.index(lrow[0], 0),
                                     role=role)
        else:
            return None

    def getSelectedIndexes(self):
        return self.msgTblW.selectionModel().selectedIndexes()

    def save(self):
        self.check_before_saving()
        if self.hasError:
            return self.hasError
        self.removeCurrentModuleData()
        for glbMsg, msgType, msg, consprod, _, _ in self.getMsgs():
            self.__topBlock.bindMsg(msg, glbMsg, self.moduleName, consprod, msgType)

    def removeCurrentModuleData(self):
        for lclmsg, (_, state, _, _) in self.instance.getData().items():
            modMsg = ('{}.{}'.format(self.moduleName, lclmsg), state)
            for glbMsg in self.__topBlock.getAllMessageBindings():
                try:
                    self.__topBlock.removeMessageBinding(modMsg, glbMsg)
                except ValueError:
                    pass

    def check_before_saving(self):
        self.hasError = False
        # logical checks
        # for glbMsg, type_, msg, consprod, storage, period in self.getMsgs():

        allBindings = self.__topBlock.getAllMessageDictBindings()
        for glbMsg, typeMsg, lclMsg, consprod, _, _ in self.getMsgs():
            # state = self.messages[modMsg][1]
            moduleMsg = '{}.{}'.format(self.moduleName, lclMsg)
            isUnique = self.__topBlock.checkMsgProducerUnicity(glbMsg, moduleMsg,
                                                               consprod, allBindings)
            if not isUnique:
                showErrorWindow('Message Error',
                                'Message {} is already assigned'
                                ' to a producer.', glbMsg)
                self.hasError = True
                return
            glbMsgData = self.__topBlock.getAllMessages()[glbMsg]
            glbMsgType = glbMsgData[0]
            if glbMsgType != typeMsg:
                showErrorWindow('Message Error',
                                'The message type of "{}" is not the same as in the module message.',
                                glbMsg)
                self.hasError = True
                return


class _WhguiMessageTableInstanceModel(QAbstractTableModel):
    """
    icd Binding table model.
    """
    TITLE_GLOBAL = 'Global msg'
    TITLE_TYPE = 'Type'
    TITLE_NAME = 'Name'
    TITLE_CONSPROD = 'Con/Prod'
    TITLE_STORAGE = 'Storage'
    TITLE_PERIOD = 'Period (ms)'
    HEADERS = [TITLE_GLOBAL, TITLE_TYPE, TITLE_NAME, TITLE_CONSPROD,
               TITLE_STORAGE, TITLE_PERIOD]

    def __init__(self, formatF, parent=None):
        """
        :param XpntTypeNames: The Names of Xpoint of the mule
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent = parent
        self.__formatF = formatF

        #: Entries as list of tuples of
        #  port, message and type
        self._entries = []
        self._update()

    @property
    def entries(self):
        return self._entries

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
                                    0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self._entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column <= self.columnCount():
            info = self._entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return self.__formatF(info, column)
        return None

    def setData(self, index, data, role=None):
        assert role == Qt.UserRole
        self._setMessageEntry(index, data)

    def _setMessageEntry(self, index, data):
        glb, type_, msg, consprod, storage, period = self._entries[index.row()]
        if index.column() == 0:
            actualentry = (data, type_, msg, consprod, storage, period)
        if index.column() == 1:
            actualentry = (glb, data, msg, consprod, storage, period)
        elif index.column() == 2:
            actualentry = (glb, type_, data, consprod, storage, period)
        elif index.column() == 3:
            actualentry = (glb, type_, msg, data, storage, period)
        elif index.column() == 4:
            actualentry = (glb, type_, msg, consprod, data, period)
        elif index.column() == 5:
            actualentry = (glb, type_, msg, consprod, storage, data)
        self._entries[index.row()] = actualentry

    def checkUnicityMsg(self, msg, row):
        for i, (entry) in enumerate(self._entries):
            if i != row:
                # check uniqueness on msg name only
                if msg == entry[0]:
                    return False
        return True

    def getFullMessages(self):
        return self._entries

    def addRow(self, glbMsg, type_, localGbl, consprod, storage, period):
        """
        Add a row at the end of the list (append)

        :param msgGbl: The message name global
        :type msgGbl: str | None
        :param msgLcl: The message name local
        :type msgLcl: str | None
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())
        self._entries.append((glbMsg, type_, localGbl, consprod, storage, period))

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        msgGbl, msgLcl = self._entries[rowIndex]
        if msgGbl is not None or msgLcl is not None:
            try:
                # Unbound function of message
                self.__parent.removeMsg(msgGbl)
                pass
            except ValueError:
                pass
        self._entries.pop(rowIndex)
        self.endRemoveRows()
        self._update()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def flags(self, index):
        flgs = super().flags(index)
        flgs |= Qt.ItemIsEditable
        return flgs


class _WhguiMessageBindDelegate(QItemDelegate):
    """
    Delegate for module message data
    """

    def __init__(self, top, parent=None):
        """
        :param top: icd topBlock
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__top = top

    def createEditor(self, parent, option, index):
        """
        Create a editors for card table the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """

        def save(*_):
            self.setModelData(editor, index.model(), index)

        col = index.column()
        if col == 0:
            editor = QComboBox(parent)
            for glbMsg in self.__top.getAllMessages():
                editor.addItem(glbMsg, glbMsg)
            editor.currentIndexChanged.connect(save)
        else:
            editor = QLabel(parent)

        return editor

    def setEditorData(self, editor, index):
        """
        :type editor: QComboBox | QLineEdit
        :type index: QModelIndex
        """
        data = index.model().data(index, Qt.EditRole)
        if isinstance(editor, QComboBox):
            for index in range(editor.count()):
                if editor.itemData(index, Qt.UserRole) == data:
                    editor.setCurrentIndex(index)
                    break
            else:
                editor.setCurrentIndex(0)
        elif isinstance(editor, QLineEdit):
            editor.setText(data)

    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox | QLineEdit
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        if isinstance(editor, QComboBox):
            model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)
        elif isinstance(editor, QLineEdit):
            if editor.text():
                model.setData(index, editor.text(), Qt.UserRole)


class _WiguiModuleMessageProperties(_WiguiModuleMessageInstanciate):

    def __init__(self, oper, pctx, parent=None):
        """
            :param instance: The instance object
            :type instance: WhhdbHostInstance
            :param pctx: Project context
            :type pctx: WhmodHardwareProjectContextP
            :param parent: Parent widget
        """
        super().__init__(oper, pctx, parent)

    @property
    def msgTblM(self):
        return self.__msgTblM

    @property
    def msgTblW(self):
        return self._msgTblW

    def _initModels(self):
        self.__msgTblM = WhguiMessageTablePropertiesModel(self.formatMsgs,
                                                         parent=self)

    def _initElements(self):
        self.msgTblW.setModel(self.msgTblM)
        self.msgTblW.horizontalHeader().setSectionResizeMode(1)

    def save(self):
        pass


class WhguiMessageTablePropertiesModel(_WhguiMessageTableInstanceModel):
    """
        icd Binding table model.
    """
    TITLE_TYPE = 'Type'
    TITLE_NAME = 'Name'
    TITLE_CONSPROD = 'Con/Prod'
    TITLE_STORAGE = 'Storage'
    TITLE_PERIOD = 'Period (ms)'
    HEADERS = [TITLE_TYPE, TITLE_NAME, TITLE_CONSPROD,
               TITLE_STORAGE, TITLE_PERIOD]

    def __init__(self, formatF, parent=None):
        """
        :param XpntTypeNames: The Names of Xpoint of the mule
        :param parent: Parent widget
        """
        super().__init__(formatF, parent)

    def addRow(self, glbMsg, type_, localGbl, consprod, storage, period):
        """
        Add a row at the end of the list (append)

        :param msgGbl: The message name global
        :type msgGbl: str | None
        :param msgLcl: The message name local
        :type msgLcl: str | None
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())
        self._entries.append((type_, localGbl, consprod, storage, period))

        self.endInsertRows()
        self._update()

    def flags(self, index):
        flgs = super().flags(index)
        flgs &= ~Qt.ItemIsEditable
        return flgs