# ----------------------------------------------------------------------------
# - wigui-project-props-dialog                                               -
# - wild cruncher - cockpit: project properties dialog widgets               -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wigui_project_props_dialog
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher project properties dialog

.. moduleauthor:: philippe lambert <lambert@silkan.com>
"""
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QCheckBox, QHBoxLayout, QVBoxLayout, QLineEdit, \
    QPushButton, QTableWidget, QSizePolicy

from wcr_frontend.gui.wgui_project_props_dialog import \
    WguiProjectPropertiesDialog
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP


class WiguiProjectPropertiesDialog(WguiProjectPropertiesDialog):
    """
    The WiguiProjectPropertiesDialog class is a dialog design to access
    and manipulate project metadata, project editor preferences and project
    configuration settings.
    """

    def __init__(self, prjContext, config, updateF, scene, parent=None):
        """
        :param prjContext: Project context
        :type prjContext: WcorProjectContextP
        :param config: Project editor configuration
        :type config: WguiProjectEditorCfg
        :param updateF: Callback function to notify about saves
        :type updateF: _ -> _
        :param parent: Parent widget
        """
        self.scene = scene
        super().__init__(prjContext, config, updateF, parent=parent)

    def _createProjectPropertiesTabs(self, prjContext):
        return [('Global Dictionnary',
                 _WiguiGlobalDictionnaryWidget(prjContext, self.scene))]


class _WiguiGlobalDictionnaryWidget(WguiPropertiesWidgetP):
    """
    Configuration widget for project global dictionnary variables
    """

    DEFAULT_COL_WIDTH=50

    def __init__(self, prjContext, scene, parent=None):
        """
        :param prjContext: Project context
        :type prjContext: WpmodSystemProjectContextP
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__scene = scene
        #: Project info properties
        self.__prjContext = prjContext
        #: Hierarchy top-level
        self.__topBlock = prjContext.getProjectDB().getTopHierarchy()
        """:type: WihdbTopHierarchy"""

        self.__table = QTableWidget()

        # self._initElements()
        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        """
        Create the layout for the widget.
        Should only be called once.

        :return: Layout
        :rtype: QLayout
        """
        mainLayout = QVBoxLayout()

        table = self.__table
        table.setColumnCount(4)
        table.setColumnWidth(0, self.DEFAULT_COL_WIDTH)

        table.verticalHeader().hide()
        table.horizontalHeader().setStretchLastSection(True)
        table.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        table.setSortingEnabled(True)
        table.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        mainLayout.addWidget(table)

        addbutton = QPushButton('Add new')
        addbutton.clicked.connect(self._addRow)
        delbutton = QPushButton('Delete')
        delbutton.clicked.connect(self._delRow)
        btnLayout = QHBoxLayout()
        btnLayout.addWidget(delbutton)
        btnLayout.addStretch(1)
        btnLayout.addWidget(addbutton)
        mainLayout.addLayout(btnLayout)

        return mainLayout

    def updateFields(self):
        table = self.__table
        table.clear()
        table.setHorizontalHeaderLabels(['To Del', 'Value', 'Type', 'Aliases'])
        records = self.__topBlock.getGlobalDic()
        table.setRowCount(len(records))

        row = 0
        for value, type, aliaces in records:
            cbwgt = QCheckBox()
            table.setCellWidget(row, 0, cbwgt)
            vwgt = QLineEdit()
            vwgt.setText(value)
            table.setCellWidget(row, 1, vwgt)
            twgt = QLineEdit()
            twgt.setText(type)
            table.setCellWidget(row, 2, twgt)
            als = ''
            for a in aliaces:
                als += a+', '
            awgt = QLineEdit()
            awgt.setText(als[:-2])
            table.setCellWidget(row, 3, awgt)
            row += 1

    def _addRow(self):
        """
        Slot function to create new rows.
        """
        table = self.__table
        table.setRowCount(table.rowCount() + 1)
        row = table.rowCount() - 1
        cbwgt = QCheckBox()
        table.setCellWidget(row, 0, cbwgt)
        vwgt = QLineEdit()
        vwgt.setPlaceholderText('Value')
        table.setCellWidget(row, 1, vwgt)
        vwgt = QLineEdit()
        vwgt.setPlaceholderText('Type')
        table.setCellWidget(row, 2, vwgt)
        vwgt = QLineEdit()
        vwgt.setPlaceholderText('Alias1, Alias2,...')
        table.setCellWidget(row, 3, vwgt)

    def _delRow(self):
        """
        Slot function to remove a row.
        """
        tableView = self.__table
        """:type: list[QModelIndex]"""

        tab=[]
        for row in range(tableView.rowCount()):
            if isinstance(tableView.cellWidget(row, 0), QCheckBox):
                if tableView.cellWidget(row, 0).checkState() == 0:
                    tab.append((tableView.cellWidget(row, 1).text(),
                                tableView.cellWidget(row, 2).text(),
                                tableView.cellWidget(row, 3).text()
                                ))

        tableView.setRowCount(len(tab))
        row = 0
        for value, type, alliaces in tab:
            cbwgt = QCheckBox()
            tableView.setCellWidget(row, 0, cbwgt)
            vwgt = QLineEdit()
            vwgt.setText(value)
            tableView.setCellWidget(row, 1, vwgt)
            twgt = QLineEdit()
            twgt.setText(type)
            tableView.setCellWidget(row, 2, twgt)
            awgt = QLineEdit()
            awgt.setText(alliaces)
            tableView.setCellWidget(row, 3, awgt)

            row += 1

    def save(self):
        # noinspection PyShadowingNames
        def getText(row, col):
            if isinstance(tableView.cellWidget(row, col), QLineEdit):
                return tableView.cellWidget(row, col).text()
            elif tableView.item(row, col) is not None:
                return tableView.item(row, col).text()
            else:
                raise Exception('Cannot retrieve text for r%d c%d' %
                                (row, col))

        tableView = self.__table
        newdic=[]
        for row in range(tableView.rowCount()):
            if not getText(row, 1) is '':
                v=getText(row, 1)
                vtype=getText(row, 2)
                a=[]
                alias = getText(row, 3)
                for al in alias.split(','):
                    a.append(al.strip())
                newdic.append((v,vtype,a))
        self.__topBlock.setGlobalDic(newdic)
        self.__scene.projectInfoChanged.emit()
