# ----------------------------------------------------------------------------
# - wigui-scene                                                              -
# - wild cruncher - cockpit : graphic scene widget                           -
# ----------------------------------------------------------------------------
# - author    (c) 2017-2017 achraf mahti                                     -
# - copyright (c) 2017-2017 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wigui_scene
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher wigui graphics scene editor

.. moduleauthor:: achraf mahti <mahti@silkan.net>
"""

from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtCore import QPointF
from wcr_core.db.whdb_utils import whdb_getNextFunctionName
from wcr_core.modules.icd.db.wihdb_meta_provider import IMPORTED_MODULE_CODE
from wcr_core.modules.icd.db.wihdb_operator import WIHDB_OPER
from wcr_core.modules.icd.db.wihdb_module_instance import WihdbModuleInstance
from wcr_core.plugin.wplg_library import WplgOperatorLibrary
from wcr_core.modules.icd.db.wihdb_wrapper_registry import create_new_imported_module
from wcr_frontend import showErrorWindow
from wcr_frontend.gui.wgui_scene import WguiGraphicsScene
from wcr_frontend.gui.wgui_utils import wgui_snap_point
from wcr_frontend.modules.icd.gui.wigui_message_binding_props_dialog import \
    WiguiMessageInstanciateDialog
from wcr_frontend.modules.icd.gui.wigui_module_creation_props_dialog \
    import WiguiModuleInstanciateDialog
from wcr_frontend.gui.wgui_scene_item import *
from wcr_frontend.modules.icd.gui.wigui_module_instance_props_dialog import \
    WiguiModulePropertiesDialog

LIB_NAME = 'Modules'


class WiguiGraphicsScene(WguiGraphicsScene):
    """
    The WhguiGraphicsScene class is the class responsible to manage
    specific scene action in hardware view that WguiGraphicsScene don't handle
    """

    GlobalMsgChanged = pyqtSignal(str, str)
    GlobalBoundMsgChanged = pyqtSignal(str, str)
    GlobalMsgAdded = pyqtSignal(str, str, str)

    def __init__(self, pjdb, config, actions, pctx, cctx, actx, mod,
                 parent=None):

        # todo: initialise the right library to consider as cards
        super().__init__(pjdb, config, actions, pctx, cctx, actx, parent=parent)
        self.__projectCtx = pctx
        self.__top = pctx.getProjectDB().getTopHierarchy()
        self.__actx = actx
        self.projOperators = {}
        self.projModule = mod
        self.moduleToolBox = None
        self.tool = None
        self.x = 0
        self.y = 0
        self.importModules()
        self.instanciatedOperators = {}
        self.sceneItemChanged.connect(self.updateBindingMessages)
        self.GlobalMsgChanged.connect(self.updateGlobalMessages)
        self.GlobalBoundMsgChanged.connect(self.updateGlobalMessagesBinding)

    def release(self):
        pass
    
    def importModules(self):
        importedModules = WIHDB_OPER.getItems().items()
        for cod, mod in importedModules:
            if IMPORTED_MODULE_CODE in cod:
                self.projOperators[mod.getName()] = mod
        if self.projOperators:
            lib = WplgOperatorLibrary(
                code=LIB_NAME, module='icd', name=LIB_NAME,
                description='modules created by user',
                operators=[o for o in self.projOperators.values()])
            self.__actx.addMessage('New modules imported from project file',
                                   'icdView')
            self.projModule.addLibrary(lib)

    @property
    def projectCtx(self):
        return self.__projectCtx

    def createFunctionItem(self, pos, oper=None, userInstance=True):
        """
        Handle the fact tha we want to add a card directly to an host
        and not to create it in the scene view.
        Else add a function item by position it not a card

        :param pos: the function position
        :type pos: QPointF
        """
        # select the operator by mode
        if oper is None:
            oper = self.getOperatorMode()
        if oper is None: return
        name = whdb_getNextFunctionName(oper, self.topLevelModelBlock)
        # get the function by operator
        func = self.topLevelModelBlock.createOperatorInstance(oper, name=name)
        moduledat = oper.getEntries()
        # feed messages to instance module without type
        func.setData(moduledat)
        if userInstance:
            instanciate = WiguiMessageInstanciateDialog(pctx=self.__projectCtx,
                                                        instance=func)
            instanciate.show()
            instanciate.exec_()
            # feed messages to instance module with type
            if not instanciate.instWdgt.hasError:
                func.setData(moduledat, self.__top)
        # set the function position
        cfg = self.config
        pos = wgui_snap_point(pos, cfg.snapGridX, cfg.snapGridY)
        func.setPosition(pos.x(), pos.y())
        # add function to the scene
        self.addFunctionItem(func).focusNameField()
        if userInstance:
            # delete creating item since binding was canceled
            if instanciate.instWdgt.hasError:
                item = self.getSceneItem(func)
                super().deleteItem(item)

    def addFunctionItem(self, func):
        """
        Create a function graphics item for the given function database
        object.

        :param func: The function object
        :type func: WhdbOperatorInstanceP
        :return: The newly created connection scene item
        :rtype: WguiOperatorInstanceSceneItem
        """
        if func is None: return None
        if 'wbmessage' in func.getOperator().getCode():
            return None
        if func.getZ() is None:
            func.setZ(self.getTopZValue() + 1)

        item = self.sitmF.createOperatorInstanceSceneItem(
            func, self.viewContext)
        # don't display card as block in scene
        self.addItem(item)
        return item

    def createOperator(self):
        """
        Add a function to create operator

        :param pos: the function position
        :type pos: QPointF
        """

        instanciate = WiguiModuleInstanciateDialog(pctx=self.__projectCtx,
                                                   projOper=self.projOperators,
                                                   parent=self)
        instanciate.show()
        instanciate.exec_()
        oper = instanciate.instWdgt.operator
        moduleType = instanciate.instWdgt.moduleType
        self.addModuleToLibrary(oper, moduleType)

    def showOperator(self, operName):
        """
            Add a function to edit an operator that has not been instantiated

            :param pos: the function position
            :type pos: QPointF
        """
        previousOper = self.projOperators[operName]
        instanciate = WiguiModulePropertiesDialog(inst=previousOper,
                                                  pctx=self.__projectCtx)
        instanciate.show()
        instanciate.exec_()
        if instanciate.instWdgt.hasError:
            return

        oper = instanciate.instWdgt.operator
        self.editModuleToLibrary(operName, oper)

    def createAndSetMsgs(self, modName, modType, localMsgName, msgData):
        library = self.projModule.getLibrary(LIB_NAME)
        if not library:
            library = WplgOperatorLibrary(
                code=LIB_NAME, module='icd', name=LIB_NAME,
                description='modules created by user',
                operators=[])
        operators = library.operators
        opByNames = {op.getName(): op for op in operators}
        if modName not in opByNames.keys():
            WIHDB_MODULE_CODE = 'wcr.icd.module.{}'.format(modName)
            oper = create_new_imported_module(WIHDB_MODULE_CODE, moduleType=modType)
            self.addModuleToLibrary(oper, modType)
        else:
            oper = opByNames[modName]
        oper.setMsgEntry(localMsgName, msgData)
        return oper

    def importOperator(self, data):
        moduleToInstantiate = set()
        filteredModulesName = set()
        for id, line in enumerate(data[1:]):
            if len(line) != 9:
                showErrorWindow('Import Error',
                                'There should be 9 value per row but {} detected in line {}.',
                                len(line), id+1)
                continue
            modName = line[0]
            modType = line[1]
            localMsgName = line[2]
            localMsgType = line[3]
            ConsProd = line[4]
            Storage = line[5]
            Period = line[6]
            glbMsgName = line[7]
            payload = line[8]
            msgData = (localMsgType, ConsProd, Storage, Period)
            # module has already been decided to not be created
            if modName in filteredModulesName:
                continue
            # create module
            oper = self.createAndSetMsgs(modName, modType, localMsgName, msgData)
            sceneModulesName = {mod.getDataObject().getName(): mod
                                for mod in self.getSceneItems()}
            if modName in sceneModulesName:
                filteredModulesName.add(modName)
                showErrorWindow('Import Module Error',
                                'Module with name "{}" already exist!',
                                modName)
                continue
            moduleToInstantiate.add(oper)
            # create global msg
            self.__top.addMsgType(glbMsgName, localMsgType, payload)
            self.GlobalMsgAdded.emit(glbMsgName, localMsgType, payload)
            # binding msg
            isBound = self.__top.bindMsg(localMsgName, glbMsgName, modName,
                                         ConsProd, localMsgType)
            # if msg can't be bound warn the user
            if not isBound:
                showErrorWindow('Import Message Warning',
                                'Global message "{}" does not exist and thus message "{}" can\'t be bound.',
                                glbMsgName, localMsgName)

        self.instantiateMultipleModule(moduleToInstantiate)

    def instantiateMultipleModule(self, moduleToInstantiate):
        """
        function to instantiate multiple module and placing them
        automatically in the scene
        :param moduleToInstantiate: set(operator)
        :return:
        """
        for i, oper in enumerate(moduleToInstantiate):
            if i % 2 == 0:
                self.y += 100
                self.x = 0
            pos = QPointF(self.x, self.y)
            self.createFunctionItem(pos, oper, userInstance=False)
            self.x += 100

    def addModuleToLibrary(self, oper, moduleType):
        if oper is None:
            return
        oper.setModuleType(moduleType)
        self.projOperators[oper.getName()] = oper
        projLib = self.projModule.getLibrary(LIB_NAME)
        self.projEditor = self.__actx.getEditor()
        if self.projEditor is None:
            self.projEditor = self.detachedEditor

        if projLib:
            projLib.operators.append(oper)
            self.projModule.addOperator(oper, "lib_name")
        else:
            lib = WplgOperatorLibrary(
                code=LIB_NAME, module='icd', name=LIB_NAME,
                description='modules created by user',
                operators=[o for o in self.projOperators.values()])
            self.projModule.addLibrary(lib)
        self.projEditor.saveState()
        self.projEditor.initialize()
        self.projEditor.loadState()
        self.sceneModified.emit()

    def editModuleToLibrary(self, oldOperName, newOper):
        if oldOperName is None or newOper is None:
            return
        if self.checkRemoveOperator(oldOperName):
            QMessageBox.warning(
                self.parent(), 'Deletion attempt',
                'You are trying to remove a module from the library but'
                ' that module has been instantiated. \n'
                'First you need to remove it\'s instances. ')
            return
        self.projEditor = self.__actx.getEditor()
        self.projOperators[newOper.getName()] = newOper
        projLib = self.projModule.getLibrary(LIB_NAME)
        projLib.operators.append(newOper)
        self.projModule.addOperator(newOper, "lib_name")
        self.removeOperator(oldOperName, resetScene=True)
        self.projEditor.saveState()
        self.projEditor.initialize()
        self.projEditor.loadState()
        self.sceneModified.emit()

    def removeOperator(self, opName, resetScene=False):
        """
        remove an operator

        :param pos: the function position
        :type pos: QPointF
        """
        code = 'wcr.icd.module.{}'.format(opName)
        oper = self.projModule.getOperator(code)
        self.projModule.deleteOperator(oper, LIB_NAME)
        del self.projOperators[opName]
        self.projModule.emptyLibrary()
        lib = WplgOperatorLibrary(
            code=LIB_NAME, module='icd', name=LIB_NAME,
            description='modules created by user',
            operators=[o for o in self.projOperators.values()])
        self.projModule.addLibrary(lib)
        if not resetScene:
            self.projEditor = self.__actx.getEditor()
            if self.projEditor is None:
                self.projEditor = self.detachedEditor
            self.projEditor.saveState()
            self.projEditor.initialize()
            self.projEditor.loadState()
            self.sceneModified.emit()

    def connectProjectItems(self, ctrl, tblk, eobj, sobj, expn, sxpn):
        """
            Function that to update information (check consumer/producer)
            if return False no connection should be created
        :return: bool
        """
        hasConns, points = ctrl.hasEndPortMultipleConnection(tblk,
                                                             sxpn,
                                                             expn)
        if hasConns:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Information)
            msg.setText("Variable {} already connected!".format(
                ' and '.join(points)))
            msg.setWindowTitle("Variable Consumption Warning !")
            msg.exec_()
            return False

        isDiffMod = (eobj.getOperator().getCode() !=
                     sobj.getOperator().getCode())
        if isDiffMod:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Information)
            msg.setText("Variables come from different modules "
                        "({}, {})!".format(eobj.getName(),
                                           sobj.getName()))
            msg.setWindowTitle("Variable Modules Warning !")
            msg.exec_()
            return False

        isDiffMsg = expn.getName() != sxpn.getName()
        if isDiffMsg:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Information)
            msg.setText("Variables have different names "
                        "({}, {})!".format(expn.getName(),
                                           sxpn.getName()))
            msg.setWindowTitle("Variables Name Warning !")
            msg.exec_()
            return False

        if expn.getDirection() == sxpn.getDirection():
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Information)
            msg.setText("Variables are of same type, "
                        "one need to be receiver and the other consumer!")
            msg.setWindowTitle("Variables Direction Warning !")
            msg.exec_()
            return False

        return True

    def removeInvalidConnections(self):
        """
        Remove all connections that are no longer connected to valid connectors
        on both ends.
        """
        for cnctn in list(self.topLevelModelBlock.getConnections()):
            sxpn = cnctn.getStartConnectionPoint()
            expn = cnctn.getEndConnectionPoint()

            if expn.getDirection() == sxpn.getDirection():
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Information)
                msg.setText("Variables have are of same type, "
                            "one need to be receiver and the other consumer!")
                msg.setWindowTitle("Variables Direction Warning !")
                msg.exec_()
                self.deleteItem(self.getSceneItem(cnctn))

    def checkBeforeDeleteItem(self, itemName):
        diaw = QMessageBox()
        diaw.setWindowTitle('Confirm Delete')
        diaw.setText('Do you want to delete the module {} ?\n'.format(itemName))
        diaw.setInformativeText('All binding will be lost.')
        diaw.addButton(QMessageBox.Yes)
        diaw.addButton(QMessageBox.No)
        diaw.setDefaultButton(QMessageBox.No)
        diaw.setIcon(QMessageBox.Question)
        diaw.setEscapeButton(QMessageBox.No)
        cnfm = diaw.exec()
        if cnfm == QMessageBox.Yes:
            return True
        elif cnfm == QMessageBox.No:
            return False

    def deleteItem(self, item):
        if not getattr(item.getDataObject(), 'getName', None):
            super().deleteItem(item)
            return
        if (item is None or item.scene() != self or
                not self.checkBeforeDeleteItem(item.getDataObject().getName())):
            return
        module = item.getDataObject()
        if isinstance(module, WihdbModuleInstance):
            for msg, (_, state, _, _) in module.getData().items():
                modMsg = ('{}.{}'.format(module.getName(), msg), state)
                for glbMsg in self.__top.getAllMessageBindings():
                    try:
                        self.__top.removeMessageBinding(modMsg, glbMsg)
                    except ValueError:
                        pass

        super().deleteItem(item)

    def updateBindingMessages(self, item):
        """
            Update the message bindings when the module name is changed
            by detecting the non existing module in the message binding
            :param item:
            :return:
        """
        if not isinstance(item, WihdbModuleInstance):
            if item is None or not item.scene() is self:
                return
        newModuleName = item.getName()
        allModules = [itemS.getDataObject().getName()
                        for itemS in self.getSceneItems()]
        allBindings = self.__top.getAllMessageBindings()

        newBindings = {}
        for glbMsg in allBindings:
            glbMsgBinding = []
            for msg, _ in allBindings[glbMsg]:
                moduleName, lclMsg = msg.split('.')
                if moduleName in allModules:
                    glbMsgBinding.append((msg, _))
                else:
                    newMsg = '.'.join((newModuleName, lclMsg))
                    glbMsgBinding.append((newMsg, _))
            newBindings[glbMsg] = glbMsgBinding

        self.__top.setAllMessageBindings(newBindings)
        self.sceneModified.emit()

    def updateGlobalMessagesBinding(self, oldMsg, newMsg):
        # if same name do nothing
        if oldMsg == newMsg:
            return
        allBindings = self.__top.getAllMessageBindings()
        values = allBindings[oldMsg]
        allBindings[newMsg] = values
        del allBindings[oldMsg]
        # update the bindings
        self.__top.setAllMessageBindings(allBindings)
        self.sceneModified.emit()

    def updateGlobalMessages(self, oldMsg, newMsg):
        # if same name do nothing
        allMessages = self.__top.getAllMessages()
        if not oldMsg or oldMsg == newMsg or oldMsg not in allMessages:
            return
        # update the global message
        values = allMessages[oldMsg]
        allMessages[newMsg] = values
        del allMessages[oldMsg]
        self.__top.setAllMessages(allMessages)

    def checkRemoveOperator(self, opName):
        code = 'wcr.icd.module.{}'.format(opName)
        usedInstances = tuple(it.getOperator().getCode()
                              for it in self.__top.getOperatorInstances())
        if code in usedInstances:
            return True
        else:
            return False