# ----------------------------------------------------------------------------
# - wigui-scene                                                              -
# - wild cruncher - cockpit : graphic scene widget                           -
# ----------------------------------------------------------------------------
# - author    (c) 2017-2017 achraf mahti                                     -
# - copyright (c) 2017-2017 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wigui_scene
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher wigui graphics scene editor

.. moduleauthor:: achraf mahti <mahti@silkan.net>
"""

from PyQt4 import QtGui
from wcr_core.db.whdb_utils import whdb_getNextFunctionName
from wcr_core.modules.icd.db.wihdb_meta_provider import IMPORTED_MODULE_CODE
from wcr_core.modules.icd.db.wihdb_operator import WIHDB_OPER
from wcr_core.plugin.wplg_library import WplgOperatorLibrary
from wcr_frontend import showErrorWindow
from wcr_frontend.gui.wgui_scene import WguiGraphicsScene
from wcr_frontend.gui.wgui_utils import wgui_snap_point
from wcr_frontend.modules.icd.gui.wigui_message_binding_props_dialog import \
    WiguiMessageInstanciateDialog
from wcr_frontend.modules.icd.gui.wigui_module_creation_props_dialog \
    import WiguiModuleInstanciateDialog
from wcr_frontend.gui.wgui_scene_item import *

LIB_NAME = 'Imported Modules'


class WiguiGraphicsScene(WguiGraphicsScene):
    """
    The WhguiGraphicsScene class is the class responsible to manage
    specific scene action in hardware view that WguiGraphicsScene don't handle
    """

    GlobalMsgChanged = pyqtSignal()

    def __init__(self, pjdb, config, actions, pctx, cctx, actx, mod,
                 parent=None):

        # todo: initialise the right library to consider as cards
        super().__init__(pjdb, config, actions, pctx, cctx, actx, parent=parent)
        self.__projectCtx = pctx
        self.__top = pctx.getProjectDB().getTopHierarchy()
        self.__actx = actx
        self.projOperators = {}
        self.projModule = mod
        self.moduleToolBox = None
        self.tool = None
        self.importModules()
        self.instanciatedOperators = {}
        self.sceneItemChanged.connect(self.updateBindingMessages)

    def importModules(self):
        importedModules = WIHDB_OPER._WhdbOperatorRegistry__items.items()
        for cod, mod in importedModules:
            if IMPORTED_MODULE_CODE in cod:
                self.projOperators[mod.getName()] = mod
        if self.projOperators:
            lib = WplgOperatorLibrary(
                code=LIB_NAME, module='icd', name=LIB_NAME,
                description='modules created by user',
                operators=[o for o in self.projOperators.values()])
            self.__actx.addMessage('New modules imported from project file',
                                   'icdView')
            self.projModule.addLibrary(lib)

    @property
    def projectCtx(self):
        return self.__projectCtx

    def createFunctionItem(self, pos, oper=None):
        """
        Handle the fact tha we want to add a card directly to an host
        and not to create it in the scene view.
        Else add a function item by position it not a card

        :param pos: the function position
        :type pos: QPointF
        """
        # select the operator by mode
        if oper is None:
            oper = self.getOperatorMode()
        if oper is None: return
        name = whdb_getNextFunctionName(oper, self.topLevelModelBlock)
        # get the function by operator
        func = self.topLevelModelBlock.createOperatorInstance(oper, name=name)
        moduledat = oper.getEntries()
        # feed messages to instance module without type
        func.setData(moduledat)
        instanciate = WiguiMessageInstanciateDialog(pctx=self.__projectCtx,
                                                    instance=func)
        instanciate.show()
        instanciate.exec_()
        # feed messages to instance module with type
        if not instanciate.instWdgt.hasError:
            func.setData(moduledat, self.__top)
        # set the function position
        cfg = self.config
        pos = wgui_snap_point(pos, cfg.snapGridX, cfg.snapGridY)
        func.setPosition(pos.x(), pos.y())
        # add function to the scene
        self.addFunctionItem(func).focusNameField()
        # delete creating item since binding was canceled
        if instanciate.instWdgt.hasError:
            item = self.getSceneItem(func)
            self.deleteItem(item)

    def addFunctionItem(self, func):
        """
        Create a function graphics item for the given function database
        object.

        :param func: The function object
        :type func: WhdbOperatorInstanceP
        :return: The newly created connection scene item
        :rtype: WguiOperatorInstanceSceneItem
        """
        if func is None: return None
        if 'wbmessage' in func.getOperator().getCode():
            return None
        if func.getZ() is None:
            func.setZ(self.getTopZValue() + 1)

        item = self.sitmF.createOperatorInstanceSceneItem(
            func, self.viewContext)
        # don't display card as block in scene
        self.addItem(item)
        return item

    def createOperator(self):
        """
        Add a function to create operator

        :param pos: the function position
        :type pos: QPointF
        """

        instanciate = WiguiModuleInstanciateDialog(pctx=self.__projectCtx,
                                                   projOper=self.projOperators,
                                                   parent=self)
        instanciate.show()
        instanciate.exec_()
        oper = instanciate.instWdgt.operator
        if oper is None:
            return
        oper.setModuleType(instanciate.instWdgt.moduleType)
        self.projOperators[oper.getName()] = oper
        projLib = self.projModule.getLibrary(LIB_NAME)
        self.projEditor = self.__actx.getEditor()

        if projLib:
            projLib.operators.append(oper)
            self.projModule.addOperator(oper, "lib_name")
        else:
            lib = WplgOperatorLibrary(
                code=LIB_NAME, module='icd', name=LIB_NAME,
                description='modules created by user',
                operators=[o for o in self.projOperators.values()])
            self.projModule.addLibrary(lib)

        self.projEditor.initialize()

    def connectProjectItems(self, ctrl, tblk, eobj, sobj, expn, sxpn):
        """
            Function that to update information (check consumer/producer)
            if return False no connection should be created
        :return: bool
        """
        hasConns, points = ctrl.hasEndPortMultipleConnection(tblk,
                                                             sxpn,
                                                             expn)
        if hasConns:
            msg = QtGui.QMessageBox()
            msg.setIcon(QtGui.QMessageBox.Information)
            msg.setText("Variable {} already connected!".format(
                ' and '.join(points)))
            msg.setWindowTitle("Variable Consumption Warning !")
            msg.exec_()
            return False

        isDiffMod = (eobj.getOperator().getCode() !=
                     sobj.getOperator().getCode())
        if isDiffMod:
            msg = QtGui.QMessageBox()
            msg.setIcon(QtGui.QMessageBox.Information)
            msg.setText("Variables come from different modules "
                        "({}, {})!".format(eobj.getName(),
                                           sobj.getName()))
            msg.setWindowTitle("Variable Modules Warning !")
            msg.exec_()
            return False

        isDiffMsg = expn.getName() != sxpn.getName()
        if isDiffMsg:
            msg = QtGui.QMessageBox()
            msg.setIcon(QtGui.QMessageBox.Information)
            msg.setText("Variables have different names "
                        "({}, {})!".format(expn.getName(),
                                           sxpn.getName()))
            msg.setWindowTitle("Variables Name Warning !")
            msg.exec_()
            return False

        if expn.getDirection() == sxpn.getDirection():
            msg = QtGui.QMessageBox()
            msg.setIcon(QtGui.QMessageBox.Information)
            msg.setText("Variables are of same type, "
                        "one need to be receiver and the other consumer!")
            msg.setWindowTitle("Variables Direction Warning !")
            msg.exec_()
            return False

        return True

    def removeInvalidConnections(self):
        """
        Remove all connections that are no longer connected to valid connectors
        on both ends.
        """
        for cnctn in list(self.topLevelModelBlock.getConnections()):
            sxpn = cnctn.getStartConnectionPoint()
            expn = cnctn.getEndConnectionPoint()

            if expn.getDirection() == sxpn.getDirection():
                msg = QtGui.QMessageBox()
                msg.setIcon(QtGui.QMessageBox.Information)
                msg.setText("Variables have are of same type, "
                            "one need to be receiver and the other consumer!")
                msg.setWindowTitle("Variables Direction Warning !")
                msg.exec_()
                self.deleteItem(self.getSceneItem(cnctn))

    def removeItem(self, item):
        if item is None or not item.scene() is self:
            return
        module = item.getDataObject()
        for msg, (_, state, _, _) in module.getData().items():
            modMsg = ('{}.{}'.format(module.getName(), msg), state)
            for glbMsg in self.__top.getAllMessageBindings():
                try:
                    self.__top.removeMessageBinding(modMsg, glbMsg)
                except ValueError:
                    pass

        super().removeItem(item)

    def updateBindingMessages(self, item):
        """
            Update the message bindings when the module name is changed
            by detecting the non existing module in the message binding
            :param item:
            :return:
        """
        newModuleName = item.getName()
        allModules = [itemS.getDataObject().getName()
                        for itemS in self.getSceneItems()]
        allBindings = self.__top.getAllMessageBindings()

        newBindings = {}
        for glbMsg in allBindings:
            glbMsgBinding = []
            for msg, _ in allBindings[glbMsg]:
                moduleName, lclMsg = msg.split('.')
                if moduleName in allModules:
                    glbMsgBinding.append((msg, _))
                else:
                    newMsg = '.'.join((newModuleName, lclMsg))
                    glbMsgBinding.append((newMsg, _))
            newBindings[glbMsg] = glbMsgBinding

        self.__top.setAllMessageBindings(newBindings)
