# ----------------------------------------------------------------------------
# - wigui-table                                                              -
# - wild cruncher:icd table widget                                           -
# ----------------------------------------------------------------------------
# - author    (c) 2017-2017 achraf mahti                                     -
# - copyright (c) 2017-2017 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wigui_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan icd table widget

.. moduleauthor:: achraf mahti <mahti@silkan.net>
"""

from PyQt4.QtCore import QAbstractTableModel, Qt
from PyQt4.QtGui import QWidget, QTableView, QHeaderView, QAbstractItemView, \
    QVBoxLayout, QHBoxLayout, QPushButton, QItemDelegate, QComboBox
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data


class WiguiTable(QWidget):
    """
    Dictionary binding to module table widget.
    """

    def __init__(self, pctx, scene, filter, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param model: model object (for current table)
        :type model: QAbstractTableModel
        :param delegate: delegate object (for current table)
        :type delegate: QItemDelegate
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project context
        self.__pctx = pctx
        #: Graphical scene
        self.__scene = scene

        #: formating function for element or table
        self.__formatF = lambda o, c: self.formatF(o, c)
        #: Creating function to get all possible host
        self.__pportF = []
        #: Creating function geting all available modules
        self.__modF = []

        #: Creating model table
        self.gblTblM = _WiguiTableModel(self.__formatF, parent=self)

        #: Creating editor for element in table
        self.gblDlgt = _WiguiGlobalDelegate(self.__pportF, self.__modF,
                                            self.__formatF)

        #: table widget
        self.__gblTblW = QTableView()

    @property
    def gblTblW(self):
        return self.__gblTblW

    def getGblTblW(self):
        return self.__gblTblW

    def formatF(self, obj, col):
        if obj is None or obj is '':
            return ''
        elif col in obj:
            return obj[0]
        else:
            #: we asking a name for an object in column normaly not present
            return 'No formating available'

    def _initElements(self):
        self.__gblTblW.setModel(self.gblTblM)
        nb_col = self.gblTblM.columnCount()
        for col in range(nb_col):
            self.__gblTblW.horizontalHeader().setResizeMode(
                col, QHeaderView.Stretch)
        self.__gblTblW.setEditTriggers(QAbstractItemView.AllEditTriggers)
        #: delegate classe for edition in table
        self.__gblTblW.setItemDelegate(self.gblDlgt)

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setMargin(0)
        layout.addWidget(self.__gblTblW)

        return layout

    def reloadFromData(self):
        """
        remove all row of table and refill it with element from project data
        """
        self.gblTblM.deleteRows(range(0, self.gblTblM.rowCount()),
                                   Unbind=False)
        self._loadFromData()

    def _loadFromData(self):
        """
        Fill the table from the project's data
        """
        pass


class _WiguiTableModel(QAbstractTableModel):
    """
    Global dictionary table model.
    """

    TITLE_VARIABLE = 'Variable'
    TITLE_VAR_TYPE = 'Type'

    HEADERS = [TITLE_VARIABLE, TITLE_VAR_TYPE]

    def __init__(self, formatF, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent = parent


        #: Entries as list of tuples of
        #  Variable and Type
        self.__entries = []
        """:type: list[]"""
        self.__formatF = formatF

        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, variable=None, varType=None):
        """
        Add a row at the end of the list (append)

        :param variable: global variable name
        :type variable: Str
        :param varType: variable type
        :type varType: str
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append((variable, varType))

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices, Unbind=True):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row, Unbind=Unbind)

    def deleteRow(self, rowIndex, Unbind=True):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        variable, varType = self.__entries[rowIndex]
        if variable is not None and varType is not None and Unbind:
            try:
                self.__parent.unbindPortProd(variable, varType)
                pass
            except ValueError:
                pass

        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column < 2:
            info = self.__entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return self.__formatF(info, column)
            elif role == Qt.UserRole:
                return info
        return None

    def setData(self, index, data, role=None):
        """
        :type index: QModelIndex
        :type data: object
        """
        assert role == Qt.UserRole
        self._setDataEntry(index, data)

    def _setDataEntry(self, index, data):
        variable, varType = self.__entries[index.row()]
        if variable is not None and varType is not None:
            #a binding could have been done by that line before
            # it have to be deleted.
            try:
                self.__parent.unbindPortProd(variable, varType)
            except ValueError:
                pass

        #creating the new representation line
        nl = ()
        if index.column() == 0:
            nl = (data, variable)
        elif index.column() == 1:
            nl = (varType, data)

        self.__entries[index.row()]=nl

        variable, varType = self.__entries[index.row()]
        if variable is not None and varType is not None:
            #line is full a binding have to be done
            try:
                self.__parent.bindPortProd(variable, varType)
            except ValueError:
                self.__entries[index.row()] = (None,None)
                self._update()
                pass

    def flags(self, index):
        """
        set flags fot able
        """
        flgs = super().flags(index)
        #allowing editing
        flgs |= Qt.ItemIsEditable
        return flgs

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return 2


class _WiguiGlobalDelegate(QItemDelegate):
    """
    Delegate for global dictionary
    """

    def __init__(self, hostListF, typeListF, formatF, parent=None):
        """
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)

        self.formatF = formatF
        self.__hostListPrv = hostListF
        self.__typeListPrv = typeListF

    def createEditor(self, parent, option, index):
        """
        Create a combobox containing adequate choices for the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """

        def save(*_):
            self.setModelData(editor, index.model(), index)

        editor = QComboBox(parent)
        col = index.column()
        #start withe empty line to allowing unset
        editor.addItem('', None)

        # Iterate through the messages using the adequate provider
        # and add items to the combo box
        if col == 0:
            varPrvF = self.__hostListPrv
        elif col == 1:
            varPrvF = self.__typeListPrv
            # add the curent module to the list to conserve it
            # if a binding have been done
            cdata=index.model().data(index.model().index(index.row(), 1),
                                    role=Qt.UserRole)
            if not cdata is None:
                editor.addItem(self.formatF(cdata,1),
                               cdata)
        for elem in varPrvF:
            editor.addItem(self.formatF(elem, col), elem)
        editor.currentIndexChanged.connect(save)

        return editor

    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)

    def setEditorData(self, editor, index):
        """
        Initialising editor data with most
        :type editor: QComboBox
        :type index: QModelIndex
        """
        data = index.model().data(index, Qt.UserRole)
        for index in range(editor.count()):
            if editor.itemData(index, Qt.UserRole) == data:
                editor.setCurrentIndex(index)
                break
        else:
            editor.setCurrentIndex(0)
