# ----------------------------------------------------------------------------
# - wmegui-message-binding-table                                             -
# - wild cruncher:message message bindings widget                            -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmegui_message_binding_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan messge message bindings widget

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""

from PyQt4.QtCore import QAbstractTableModel, Qt
from PyQt4.QtGui import QWidget, QTableView, QHeaderView, QAbstractItemView, \
    QVBoxLayout, QHBoxLayout, QPushButton, QItemDelegate, QComboBox


class WmeguiMessageBindingTable(QWidget):
    """
    Logical bus binding to physical buses table widget.
    """

    BTN_PRINT_BINDING_VISIBLE = False

    def __init__(self, pctx, scene, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project context
        self.__pctx = pctx
        #: Graphical scene
        self.__scene = scene

        # !!! Will need update for handling of multiple logical simulations !!!
        self.__logSimLstPrvF = lambda: [pctx.getLogicalSimulation()] \
            if pctx.getLogicalSimulation() is not None else []
        """:type: () -> list[WshdbTopBlock]"""


        #: Bus binding table model
        self.__bndgTblM = _WmeguiMessageBindingTableModel()#self.__busFormatF)


        """:type: (WphdbPhyBusInstance) -> list[WshdbBusInstance]"""

        #: Bus binding table widget
        self.__bndgTblW = QTableView()

        #: Add entry button
        self.__addButton = QPushButton('Add')
        #Desable add buton waiting for data implementation
        self.__addButton.setEnabled(False)
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')

        self._initElements()
        self.setLayout(self._createLayout())
        self._loadFromData()

    def _initElements(self):
        self.__bndgTblW.setModel(self.__bndgTblM)
        self.__bndgTblW.horizontalHeader().setResizeMode(
            0, QHeaderView.Stretch)
        self.__bndgTblW.horizontalHeader().setResizeMode(
            1, QHeaderView.Stretch)
        self.__addButton.clicked.connect(self._addEntry)
        self.__deleteButton.clicked.connect(self._deleteSelEntries)
        pass

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setMargin(0)

        layout.addWidget(self.__bndgTblW)

        btnLayout = QHBoxLayout()

        btnLayout.addWidget(self.__addButton)
        btnLayout.addWidget(self.__deleteButton)
        layout.addLayout(btnLayout)

        return layout

    def _loadFromData(self):
        """
        Fill the table from the project's data
        """
        pass

    def _addEntry(self):
        """
        Add entry in the model
        """
        self.__bndgTblM.addRow()

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.__bndgTblW.selectedIndexes()
        """:type: list[QModelIndex]"""

        self.__bndgTblM.deleteRows([idx.row() for idx in selIdx])



class _WmeguiMessageBindingTableModel(QAbstractTableModel):
    """
    Bus Binding table model.
    """

    TITLE_MESSAGE = 'Message'
    TITLE_DATA = 'Data'
    TITLE_STATUS = 'Status'

    HEADERS = [TITLE_MESSAGE, TITLE_DATA, TITLE_STATUS]


    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)


        #: Entries as list of tuples of
        #  physical buses (pbus) and logical buses (lbus)
        self.__entries = []
        """:type: list[(WphdbPhyBusInstance, WshdbBusInstance)]"""

        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, msg=None, data=None, size=None):
        """
        Add a row at the end of the list (append)

        :param pbus: Message to be set
        :type pbus: | None
        :param lbus: Data to be set
        :type lbus:  | None
        :param lbus: size to be set
        :type lbus:  | None
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append((msg, data, size))

        if msg is not None and data is not None and size is not None:
            # call binding function of message
            pass


        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        msg, data, size = self.__entries[rowIndex]
        if msg is not None and data is not None and size is not None:
            try:
                #Unbound function of message
                #pbus.removeBoundLogicalBus(lbus)
                pass
            except ValueError:
                pass

        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return 3

