# ----------------------------------------------------------------------------
# - wfgui-Datamodel-props-widget                                             -
# - wild cruncher: exec. properties widget      state machine cockpit system -
# ----------------------------------------------------------------------------
# - author    (c) 2015-2015 Philippe lambert                                 -
# - copyright (c) 2015-2015 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfgui_datamodel_props_widget
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher Datamodel properties widget

.. moduleauthor:: Philippe Lambert <philippe.lambert@silkan.com>
"""
import traceback

from PyQt5.QtCore import Qt, QItemSelectionModel, QItemSelection
from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QWidget, QVBoxLayout, \
	QPushButton, QHeaderView, QLayout, QTreeView, QGroupBox, QSplitter, \
	QScrollArea, QFormLayout, QAbstractItemView

from wcr_frontend.gui.wgui_tree_model import WguiTreeModel, \
    WguiTreeItemFactoryP, WguiTreeItemBase
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_frontend.gui.wgui_form_widget_creator import \
    WguiSimpleWidgetCreator, WGUI_META_LONGDESC_KEY, WGUI_META_LABEL_KEY, \
    WGUI_META_ITEMS_KEY, WGUI_META_KEY
from wcr_core.modules.statemachine.db.wfhdb_protocols import *

#: List of available Data for drop-down selection
_WFGUI_DATAS = [
    ('Data', WFHDB_DATA_KEY)]

_WFGUI_DATA_EXTRAS = {
    WGUI_META_ITEMS_KEY: {
        WFHDB_DATA_KEY: {
            WGUI_META_LABEL_KEY: 'Data',
            WFHDB_DATA_ID_KEY: {
                WGUI_META_LABEL_KEY: 'ID',
                WGUI_META_LONGDESC_KEY: 'Data Identifier'
            },
            WFHDB_DATA_SRC_KEY: {
                WGUI_META_LABEL_KEY: 'Source',
                WGUI_META_LONGDESC_KEY: 'The location from which the '
                                        'data should be fetched'
            },
            WFHDB_DATA_EXPRE_KEY: {
                WGUI_META_LABEL_KEY: 'Expression',
                WGUI_META_LONGDESC_KEY: 'The expression to solve to '
                                        'initialise the data'
            }
        },
    }
}



class _WfguiWidgetCreator(WguiSimpleWidgetCreator):
    def __init__(self, data, updateF):
        """
        :param data: Meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItem) -> _
        """
        super().__init__(data)

        self.__updateF = updateF

    def _dataChanged(self, *keys, **kwargs):
        treeItem = kwargs.get('treeItem')
        if treeItem:
            self.__updateF(treeItem)


def _wfgui_truncate(istr):
    """
    Truncate the given string to one line.

    :param istr: String to be truncated
    :type istr: str
    :return: Truncated string
    :rtype: str
    """
    prts = istr.strip().split('\n')
    if len(prts) > 1:
        return prts[0] + ' ...'
    else:
        return istr


def _wfgui_getKeys(item, data):
    """
    Get the keys for the given item in the given meta-data set.

    :param item: Tree item to get keys for
    :type item: _WfguiDatasTreeItemBase
    :param data: Top-level meta-data set
    :type data: object
    :return: Keys for the given item
    :rtype: list[object]
    """
    if not isinstance(item, _WfguiDatasTreeItemBase):
        return tuple()
    path = item.getPath()
    mtyp = WFHDB_EXECUTABLE_META
    keys = []
    for elem in path:
        try:
            key = mtyp.getKey(elem, data)
        except:
            print('NO {}'.format(elem))
            print('  DATA {}'.format(data))
            print('  KEYS {}'.format(keys))
            print('  PATH {}'.format(path))
            traceback.print_exc()
            raise
        keys.append(key)
        # TODO mtyp.getValueType(key, data)
        if isinstance(mtyp, WdbsMultiMetaTypeP):
            mtyp = mtyp.getValueType()
        elif isinstance(mtyp, WdbsUnionMetaTypeP):
            mtyp = mtyp.getType(data[0])
        elif isinstance(mtyp, WdbsDictionaryMetaTypeP) or \
                isinstance(mtyp, WdbsTupleMetaTypeP) or \
                isinstance(mtyp, WdbsRecordMetaTypeP):
            mtyp = mtyp.getValueType(key)
        else:
            raise Exception('Unknown type: {}'.format(type(mtyp)))
        data = elem
    return tuple(keys)


def _wfgui_getData(data, *keys):
    """
    Get the data for the given keys.

    :param data: Top-level meta-data
    :type data: object
    :param keys: Keys
    :type keys: tuple
    :return: Data for the given keys
    :rtype: object
    """
    for key in keys:
        data = data[key]
    return data


def _wfgui_getPath(data, *keys):
    """
    Get the path for the given keys.

    :param data: Top-level meta-data
    :type data: object
    :param keys: Keys
    :type keys: tuple
    :return: Path for the given keys
    :rtype: [object]
    """
    path = []
    for key in keys:
        data = data[key]
        path.append(data)
    return path




class WfguiDatamodelProperties(WguiPropertiesWidgetP):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.__data = None
        """:type: list[(str, object)] | None"""

        #: Tree item factory
        self.__itemFc = _WfguiDataTreeItemFactory()
        #: Tree model
        self.__treeM = WguiTreeModel(self.__itemFc)

        #: Tree-view widget
        self.__treeW = QTreeView()
        #: Current item properties group
        self.__propertiesW = QGroupBox('Expression selected Data')
        #: Combo-box for type-selection for new items
        self.__newComboW = QComboBox()

        #: Add-item button
        self.__addBtn = QPushButton('Add')
        #: Remove button
        self.__delBtn = QPushButton('Remove')

        self.setLayout(self._createLayout())
        self._initElements()
        self.updateFields()

        self.setMinimumWidth(800)

    def _createLayout(self):
        """
        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()

        self.__propertiesW.setMinimumWidth(300)

        splitter = QSplitter(Qt.Horizontal)
        splitter.addWidget(self.__treeW)
        scroll = QScrollArea()
        scroll.setWidget(self.__propertiesW)
        scroll.setWidgetResizable(True)
        scroll.setMinimumWidth(350)
        splitter.addWidget(scroll)
        layout.addWidget(splitter)

        btnLayout = QHBoxLayout()
        btnLayout.addWidget(self.__newComboW)
        btnLayout.addWidget(self.__addBtn)
        btnLayout.addStretch(6)
        btnLayout.addWidget(self.__delBtn)
        layout.addLayout(btnLayout)

        return layout

    def _initElements(self):
        """
        Initialize UI elements.
        """
        self.__treeW.setModel(self.__treeM)
        self.__treeW.selectionModel().selectionChanged.connect(
            self._selectionChanged)
        self.__treeW.header().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.__treeW.header().setSectionResizeMode(1, QHeaderView.Stretch)
        self.__treeW.setSelectionMode(QAbstractItemView.SingleSelection)
        self.__treeW.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.__propertiesW.setLayout(QVBoxLayout())
        self.__propertiesW.setFlat(True)

        for item, code in _WFGUI_DATAS:
            self.__newComboW.addItem(item, code)

        self.__addBtn.clicked.connect(self._add)
        self.__delBtn.clicked.connect(self._removeSelected)

        self._updateButtonStates()

    # ------------------------------------------------------------------------

    def _item(self, idx):
        """
        Get the tree item at the given index.

        :param idx: Item index
        :type idx: QModelIndex
        :return: Tree item at the given index
        :rtype: _WfguiDatasTreeItemBase
        """
        return self.__treeM.item(idx)

    def _getSelectedIndex(self):
        """
        Get the index of the currently selected item, if any.

        :return: Currently selected item, if any
        :rtype: QModelIndex | None
        """
        indices = self.__treeW.selectedIndexes()
        if not indices:
            return None
        idx = indices[0]
        return idx

    def _newEmptyData(self):
        """
        Create an empty data structure for the currently chosen executable_getS
        type.

        :return: Empty data structure
        :rtype: (str, object)
        """
        code = wgui_qcb_get_data(self.__newComboW)
        return ScmnMTuple(
            code, WFHDB_DATA_META_UNION.getType(code).getEmptyValue())

    #--------------------------------------------------------------------------

    def _doAdd(self, data, pidx, row=None, insertF=None):
        """
        Add a new element.

        :param data: Data for the new element
        :type data: object
        :param pidx: Index of the parent item
        :type pidx: QModelIndex
        :param row: Row to insert the new element at
        :type row: int
        :param insertF: Insertion function, if any (Keys, parent-index, row)
        :type insertF: (tuple, _WfguiTreeItemBase, int) -> _WfguiTreeItemBase
        """
        pitm = self._item(pidx)
        keys = _wfgui_getKeys(pitm, self.__data)
        if isinstance(pitm, _WfguiDatasTreeNodeItemBase):
            keys += pitm.getChildContainerKeys()
        _data = _wfgui_getData(self.__data, *keys)
        assert isinstance(_data, list)

        if row is None:
            row = len(_data)
        _data.insert(row, data)
        keys += (row,)

        if insertF:
            item = insertF(keys, pitm, row)
        else:
            (_, item), *_ = self.__treeM.insertItem(
                (keys, self.__data, self._dataUpdated), pitm, row)
        idx = self.__treeM.getItemIndex(item)
        self.__treeW.expand(idx)
        self.__treeW.selectionModel().select(
            QItemSelection(
                idx, idx.sibling(idx.row(), self.__treeM.columnCount() - 1)),
            QItemSelectionModel.ClearAndSelect)
        self._updateButtonStates()

    def _add(self):
        """
        Add a new element of the currently chosen type below the currently
        selected element.
        """
        idx = self._getSelectedIndex()
        prnt = idx.parent() if idx else None
        row = idx.row() + 1 if idx else None
        self._doAdd(self._newEmptyData(), prnt, row)

    def _removeSelected(self):
        """
        Remove the currently selected element.
        """
        idx = self._getSelectedIndex()
        item = self._item(idx)
        pitm = item.parentItem
        if isinstance(pitm, _WfguiDatasTreeNodeItemBase):
            pkey = _wfgui_getKeys(pitm, self.__data)
            pkey += pitm.getChildContainerKeys()
        else:
            pkey = tuple()

        keys = _wfgui_getKeys(item, self.__data)
        _data = self.__data
        for key in pkey:
            _data = _data[key]
        assert isinstance(_data, list)

        self.__treeM.removeRow(idx.row(), idx.parent())
        # noinspection PyTypeChecker
        del _data[keys[len(pkey)]]

    #--------------------------------------------------------------------------

    def _dataUpdated(self, item):
        """
        Notify the tree-model about an updated item.

        :param item: Updated item
        :type item: WguiTreeItemP
        """
        self.__treeM.refreshItem(item)

    def _selectionChanged(self, selection, *_):
        """
        Callback-function for selection changes.

        :param selection: New selection
        :type selection: QItemSelection
        """
        layout = self.__propertiesW.layout()
        """:type: QVBoxLayout"""
        for widx in reversed(range(layout.count())):
            layout.itemAt(widx).widget().setParent(None)

        indeces = selection.indexes()

        self._updateButtonStates(indeces[0] if indeces else None)

        if not indeces:
            return

        item = self._item(indeces[0])
        if isinstance(item, _WfguiTreeItemBase):
            detailUI = item.createDetailUI()
            if detailUI is None:
                return
            if isinstance(detailUI, QLayout):
                widget = QWidget()
                widget.setLayout(detailUI)
                layout.addWidget(widget)
            else:
                layout.addWidget(detailUI)

    def _updateButtonStates(self, index=None):
        """
        Update the states of the buttons based on the given index or the
        currently selected item.

        :param index: Item index
        :type index: QModelIndex | None
        """
        if not index:
            index = self._getSelectedIndex()

        if not index:
            self.__delBtn.setEnabled(False)
            return

        self.__delBtn.setEnabled(True)
        self.__addBtn.setEnabled(True)

    #--------------------------------------------------------------------------

    def updateFields(self):
        self.__treeW.selectionModel().clearSelection()

        model = self.__treeM
        model.removeRows(0, model.rowCount())

        self.__data = self._getDatas()

        for idx, _ in enumerate(self.__data):
            model.insertItem(((idx,), self.__data, self._dataUpdated))

        self.__treeW.expandAll()

    def save(self):
        if self.__data is not None:
            self._setDatas(self.__data)

    #--------------------------------------------------------------------------

    def _getDatas(self):
        """
        Get the current Datas to display.

        :return: Current datas
        :rtype: list[(str, object)]
        """
        raise NotImplementedError()

    def _setDatas(self, execs):
        """
        Set the Datas.

        :param execs: datas
        :type execs: list[(str, object)]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class _WfguiTreeItemBase(WguiTreeItemBase):
    """
    State-machine Datamodel tree item.
    """

    def __init__(self, parentItem=None):
        super().__init__(parentItem)

    def setData(self, data, column, role=None):
        pass

    def getColumnCount(self):
        return 2

    @property
    def flags(self):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    @abstractmethod
    def createDetailUI(self):
        """
        Create the detail-view UI.

        :return: Detail-view UI
        :rtype: QWidget
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class _WfguiDatasTreeItemBase(_WfguiTreeItemBase):
    _LABEL = ''

    _CODE = None

    _TYPE = None

    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(parentItem)

        assert self._CODE
        assert self._TYPE

        self.__data = data
        self.__updateF = updateF

        self.__path = _wfgui_getPath(data, *keys)
        self.__ownData = _wfgui_getData(data, *keys)

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            return self._getLabel() if column == 0 else self._getDescription()
        return None

    def getPath(self):
        return self.__path

    def getField(self, *keys):
        return _wfgui_getData(self.__ownData, *keys)

    def _getLabel(self):
        """
        Get the label for the tree item.

        :return: Label
        :rtype: str
        """
        return self._LABEL

    @abstractmethod
    def _getDescription(self):
        """
        Get the description for the tree item.

        :return: Description
        :rtype: str
        """
        raise NotImplementedError()

    def _getMetaDescription(self):
        """
        Get the meta-description for the tree item.

        :return: Meta-description
        :rtype: dict[object, object]
        """
        meta = _WFGUI_DATA_EXTRAS.get(WGUI_META_ITEMS_KEY)

        code = self._CODE
        if isinstance(code, tuple):
            for c in code:
                meta = meta.get(c)
        else:
            meta = meta.get(code)

        return meta

    def createDetailUI(self):
        uiCreator = _WfguiWidgetCreator(self.__data, self.__updateF)
        keys = _wfgui_getKeys(self, self.__data)
        meta = self._getMetaDescription()

        widget = uiCreator.createWidget(
            self._TYPE, *keys, treeItem=self, meta=meta)
        if not widget:
            return None
        elif widget.layout():
            return widget
        else:
            wrap = QWidget()
            layout = QFormLayout()
            layout.setRowWrapPolicy(QFormLayout.WrapAllRows)
            layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
            layout.setContentsMargins(0,0,0,0)
            wmta = widget.property(WGUI_META_KEY)
            labl = wmta.get(WGUI_META_LABEL_KEY) if wmta else ''
            layout.addRow(labl, widget)
            wrap.setLayout(layout)
            return wrap


# noinspection PyAbstractClass
class _WfguiDatasTreeNodeItemBase(_WfguiDatasTreeItemBase):
    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

    @abstractmethod
    def getChildContainerKeys(self):
        raise NotImplementedError()


# noinspection PyAbstractClass
class _WfguiDatasTreeLeafItemBase(_WfguiDatasTreeItemBase):
    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

    def addChild(self, child, row=None):
        raise Exception('Unsupported operation')

    def removeChild(self, row):
        raise Exception('Unsupported operation')

    def removeChildren(self, row, count):
        raise Exception('Unsupported operation')


class _WfguiDataTreeItem(_WfguiDatasTreeLeafItemBase):
    _LABEL = 'Data'

    _CODE = WFHDB_DATA_KEY

    _TYPE = WFHDB_DATA_META

    def __init__(self, keys, data, updateF, factory, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
       :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)


    def _getDescription(self):
        data = self.getField(WFHDB_DATA_ID_KEY)
        return data


class _WfguiDataRootTreeItem(_WfguiTreeItemBase):
    def __init__(self):
        super().__init__()

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            return 'Type' if column == 0 else 'Name'
        return None

    def createDetailUI(self):
        return None


class _WfguiDataTreeItemFactory(WguiTreeItemFactoryP):
    def createTreeItem(self, itemData, parentItem, **kwargs):
        if not parentItem:
            return _WfguiDataRootTreeItem(), None, None
        if not itemData:
            return None, None, None

        keys, data, updateF = itemData
        """:type: (list[object], object, (WguiTreeItemP)) -> _"""
        keys = keys + (1,)
        return _WfguiDataTreeItem(
            keys, data, updateF, self, parentItem=parentItem), parentItem, None
