# ----------------------------------------------------------------------------
# - wfgui-event-table                                                        -
# - wild cruncher: event table                         state machine cockpit -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfgui_event_table
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher state machine cockpit event table

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod

from PyQt4.QtCore import QAbstractListModel, Qt, QObject, pyqtSignal
from PyQt4.QtGui import QWidget, QSortFilterProxyModel, QVBoxLayout, \
    QListView, QLineEdit, QAbstractItemView, QItemSelectionModel
from wcr_core.db_api.whdb_connection_api import WhdbConnectionP

from wcr_frontend.gui.wgui_scene_item import WguiSceneItemP
from wcr_frontend.gui.wgui_protocols import QWidget_Meta_M, QObject_Meta_M


class WfguiEventSelectionProviderP(metaclass=ABCMeta):
    """
    Protocol for providers of selected events.
    """

    @property
    @abstractmethod
    def selectedEventChanged(self):
        """
        Signal indicating a change in the event selection.

        :return: Signal indicating a change in the event selection
        :rtype: pyqtSignal[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def getSelectedEvent(self):
        """
        Get the selected event.

        :return: Selected event
        :rtype: str
        """
        raise NotImplementedError()


class WfguiItemSelectionProviderP(metaclass=ABCMeta):
    """
    Protocol for providers of selected items.
    """

    @property
    @abstractmethod
    def selectedItemsChanged(self):
        """
        Signal indicating a change in the items selection.

        :return: Signal indicating a change in the items selection
        :rtype: pyqtSignal
        """
        raise NotImplementedError()

    @abstractmethod
    def getSelectedItems(self):
        """
        Get the selected items.

        :return: Selected items, and their paths
        :rtype: list[(WhdbBaseP, list[WhdbBlockInstanceP])]
        """
        raise NotImplementedError()


class _WfguiEventSelectionProvider(QObject, WfguiEventSelectionProviderP,
                                   metaclass=QObject_Meta_M):
    _selectedEventChanged = pyqtSignal(str)

    def __init__(self, eventLstView, parent=None):
        """
        :param eventLstView: Event list view
        :type eventLstView: QListView
        :param parent: Parent object
        """
        super().__init__(parent)

        self.__eventLstView = eventLstView

        slctModel = eventLstView.selectionModel()
        assert slctModel
        slctModel.selectionChanged.connect(self._eventChanged)

    def _eventChanged(self, *_):
        self._selectedEventChanged.emit(self.getSelectedEvent())

    @property
    def selectedEventChanged(self):
        return self._selectedEventChanged

    def getSelectedEvent(self):
        slctd = self.__eventLstView.selectedIndexes()
        if not slctd:
            return None
        return self.__eventLstView.model().data(slctd[0], Qt.DisplayRole)


class _WfguiEventItemsProvider(QObject, WfguiItemSelectionProviderP,
                               metaclass=QObject_Meta_M):
    _selectedItemsChanged = pyqtSignal()

    def __init__(self, eventPrv, eventMngr, parent=None):
        """
        :param eventPrv: Event provider
        :type eventPrv: WfguiEventSelectionProviderP
        :param eventMngr: Event manager
        :type eventMngr: WfmodEventManager
        :param parent: Parent object
        """
        super().__init__(parent)

        self.__eventPrv = eventPrv
        self.__eventMngr = eventMngr

        eventPrv.selectedEventChanged.connect(self._selectionChanged)

    def _selectionChanged(self, *_):
        self._selectedItemsChanged.emit()

    @property
    def selectedItemsChanged(self):
        return self._selectedItemsChanged

    def getSelectedItems(self):
        event = self.__eventPrv.getSelectedEvent()
        wrtrs, rdrs = self.__eventMngr.getEventUsers(event)

        return list(wrtrs | rdrs)


class _WfguiItemProvider(QObject, WfguiItemSelectionProviderP,
                         metaclass=QObject_Meta_M):
    _selectedItemsChanged = pyqtSignal()

    def __init__(self, itemLstView, itemRole=None, pathRole=None,
                 itemPathRole=None, parent=None):
        """
        :param itemLstView: Item list view
        :type itemLstView: QListView
        :param itemRole: Role to retrieve item
        :type itemRole: int
        :param itemRole: Role to retrieve path
        :type itemRole: int
        :param itemPathRole: Role to retrieve item and path
        :type itemPathRole: int
        :param parent: Parent object
        """
        super().__init__(parent)

        assert bool(itemPathRole) != (itemRole and pathRole)

        self.__itemLstView = itemLstView
        self.__itemRole = itemRole
        self.__pathRole = pathRole
        self.__itemPathRole = itemPathRole

        slctModel = itemLstView.selectionModel()
        assert slctModel
        slctModel.selectionChanged.connect(self._selectionChanged)

    def _selectionChanged(self, *_):
        self._selectedItemsChanged.emit()

    @property
    def selectedItemsChanged(self):
        return self._selectedItemsChanged

    def getSelectedItems(self):
        slctd = self.__itemLstView.selectedIndexes()
        if not slctd:
            return []
        idx = slctd[0]
        model = self.__itemLstView.model()
        if self.__itemPathRole:
            item, path = model.data(idx, self.__itemPathRole)
        else:
            item = model.data(idx, self.__itemRole)
            path = model.data(idx, self.__pathRole)
        return [(item, path)]


class WfguiEventDisplay(QWidget, metaclass=QWidget_Meta_M):
    def __init__(self, evntMngr, parent=None):
        """
        :param evntMngr: Event manager
        :type evntMngr: WfmodEventManager
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Event model
        self.__model = _WfguiEventListModel(evntMngr)
        #: Event filter model
        self.__filterModel = QSortFilterProxyModel()
        #: Event list view
        self.__eventLstW = QListView()
        self.__eventLstW.setModel(self.__filterModel)
        #: Event filter input
        self.__filterW = QLineEdit()

        #: Event selection provider
        self.__eventPrv = _WfguiEventSelectionProvider(self.__eventLstW)
        #: Event selection items provider
        self.__eventItemsPrv = _WfguiEventItemsProvider(
            self.__eventPrv, evntMngr)

        #: Event items model
        self.__itemsModel = _WfguiEventItemsListModel(
            evntMngr, self.__eventPrv)
        #: Event items list view
        self.__itemLstW = QListView()
        self.__itemLstW.setModel(self.__itemsModel)
        #: Event items selection provider
        self.__selItemPrv = _WfguiItemProvider(
            self.__itemLstW, itemPathRole=Qt.UserRole)

        self.setLayout(self._createLayout())
        self._initElements()

    def _createLayout(self):
        """
        Create the layout.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(3)

        layout.addWidget(self.__eventLstW)
        layout.addWidget(self.__filterW)
        layout.addWidget(self.__itemLstW)

        return layout

    def _initElements(self):
        """
        Initialize all UI elements.
        """
        self.__filterW.setPlaceholderText('Event-Name Filter')
        self.__filterModel.setSourceModel(self.__model)
        self.__filterModel.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.__filterModel.setDynamicSortFilter(True)

        self.__filterW.textEdited.connect(
            self.__filterModel.setFilterFixedString)

        self.__eventLstW.setSelectionMode(QAbstractItemView.SingleSelection)

    def refresh(self):
        """
        Refresh the model.
        To be called when the event configuration has changed.
        """
        self.__model.refresh()

    def setSelectedItem(self, items):
        """
        Select any of the items in the given list that are currently being
        shown in the items list view.

        :param items: Selected items
        :type items: Iterable[WhdbBaseP]
        """
        idxs = list(
            filter(None, (self.__itemsModel.itemIndex(i) for i in items)))
        self.__itemLstW.clearSelection()
        if idxs:
            selModel = self.__itemLstW.selectionModel()
            for idx in idxs:
                selModel.select(idx, QItemSelectionModel.SelectCurrent)

    def getSelectedEventProvider(self):
        """
        Get the event selection provider.

        :return: Event selection provider
        :rtype: WfguiEventSelectionProviderP
        """
        return self.__eventPrv

    def getSelectedEventItemsProvider(self):
        """
        Get the item provider for all items of the currently selected event.

        :return: Event selection items provider
        :rtype: WfguiItemSelectionProviderP
        """
        return self.__eventItemsPrv

    def getSelectedItemProvider(self):
        """
        Get the item provider for currently selected item (of the currently
        selected event).

        :return: Event item selection provider
        :rtype: WfguiItemSelectionProviderP
        """
        return self.__selItemPrv


class _WfguiEventListModel(QAbstractListModel):
    def __init__(self, evntMngr, parent=None):
        """
        :param evntMngr: Event manager
        :type evntMngr: WfmodEventManager
        :param parent: Parent object
        """
        super().__init__(parent)

        #: Event manager
        self.__eventManager = evntMngr

        #: List of events, set from data from event manager
        self.__events = []

        self.refresh()

    def refresh(self):
        """
        Refresh the model from the current event manager's data.
        """
        evnts = list(self.__eventManager.getEvents())
        evnts.sort()
        self.beginResetModel()
        self.__events = evnts
        self.endResetModel()

    def rowCount(self, index=None, *args, **kwargs):
        return len(self.__events)

    def data(self, index, role=None):
        if role == Qt.DisplayRole:
            return self.__events[index.row()]

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole:
            return 'Event' if orientation == Qt.Horizontal else str(section)

    def flags(self, index):
        flgs = super().flags(index)
        flgs &= ~Qt.ItemIsEditable
        return flgs


class _WfguiEventItemsListModel(QAbstractListModel):
    def __init__(self, evntMngr, eventPrv, parent=None):
        """
        :param evntMngr: Event manager
        :type evntMngr: WfmodEventManager
        :param eventPrv: Event provider
        :type eventPrv: WfguiEventSelectionProviderP
        :param parent: Parent object
        """
        super().__init__(parent)

        #: Event manager
        self.__eventManager = evntMngr
        #: Event provider
        self.__eventPrv = eventPrv

        #: Items for the currently selected event
        self.__items = []

        eventPrv.selectedEventChanged.connect(self.refresh)

        self.refresh()

    def refresh(self, *_):
        """
        Refresh the list of items for the currently selected event.
        """
        event = self.__eventPrv.getSelectedEvent()
        wrtrs, rdrs = self.__eventManager.getEventUsers(event)

        items = list(wrtrs | rdrs)
        items.sort(key=lambda x: self._itemLabel(x[0]))

        self.beginResetModel()
        self.__items = items
        self.endResetModel()

    def rowCount(self, index=None, *args, **kwargs):
        return len(self.__items)

    def data(self, index, role=None):
        if role == Qt.DisplayRole:
            return self._itemLabel(self.__items[index.row()][0])
        elif role == Qt.UserRole:
            return self.__items[index.row()]

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole:
            return None

    def flags(self, index):
        flgs = super().flags(index)
        flgs &= ~Qt.ItemIsEditable
        return flgs

    def itemIndex(self, item):
        """
        Get the index of the given item, if any.

        :param item: Item to look for
        :type item: WhdbBaseP
        :return: Index of given item, or `None`
        :rtype: QModelIndex | None
        """
        for idx, (_item, _) in enumerate(self.__items):
            if item is _item:
                return self.index(idx)
        return None

    def _itemLabel(self, item):
        """
        Get the display label for the given item.

        :param item: Item
        :type item: WhdbNamedEntityP
        :return: Display label
        :rtype: str
        """
        if isinstance(item, WhdbConnectionP):
            return '{} -> {}'.format(item.getStartObject().getName(),
                                     item.getEndObject().getName())
        else:
            return item.getName() or '{} {}'.format(item.getTypeName(),
                                                    item.getUUID())


class WfguiItemSceneSelectionSynchronizer(object):
    """
    Synchronization between the selection of the graphics scene and an
    event display element.
    """

    def __init__(self, eventDisplay, scene):
        """
        :param eventDisplay:
        :type eventDisplay: WfguiEventDisplay
        :param scene:
        :type scene: WguiGraphicsScene
        """
        self.__eventDisplay = eventDisplay
        self.__scene = scene
        self.__itemPrv = eventDisplay.getSelectedItemProvider()

        self.__updating = False

        self.__itemPrv.selectedItemsChanged.connect(self._itemSelectionChanged)
        scene.selectionChanged.connect(self._sceneChanged)
        scene.scenePathChanged.connect(self._sceneChanged)

    def _itemSelectionChanged(self, *_):
        if self.__updating:
            return
        self.__updating = True
        items = self.__itemPrv.getSelectedItems()
        curSel = [i.getUUID() for i
                  in [i.getDataObject() for i
                      in self.__scene.selectedItems()
                      if isinstance(i, WguiSceneItemP)]]
        newSel = [i.getUUID() for i, _ in items]
        if curSel != newSel:
            if items:
                self.__scene.setHierarchy(items[0][1])
            self.__scene.clearSelection()
            for sel, _ in items:
                item = self.__scene.getSceneItem(sel)
                if item is not None:
                    # noinspection PyUnresolvedReferences
                    item.setSelected(True)
                    # noinspection PyUnresolvedReferences
                    item.ensureVisible()
        self.__updating = False

    def _sceneChanged(self):
        if self.__updating:
            return
        self.__updating = True
        selected = self.__scene.selectedItems()
        selItems = [i.getDataObject() for i in selected
                    if isinstance(i, WguiSceneItemP)]
        self.__eventDisplay.setSelectedItem(selItems)
        self.__updating = False
