# ----------------------------------------------------------------------------
# - wfgui-executables-props-widget                                           -
# - wild cruncher: exec. properties widget      state machine cockpit system -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfgui_executables_props_widget
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher executables properties widget

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
import traceback
from copy import deepcopy

from PyQt5.QtCore import Qt, QItemSelectionModel, QItemSelection
from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QWidget, QVBoxLayout, \
	QPushButton, QHeaderView, QLayout, QTreeView, QGroupBox, QSplitter, \
	QScrollArea, QFormLayout, QAbstractItemView

from wcr_frontend.gui.wgui_tree_model import WguiTreeModel, \
    WguiTreeItemFactoryP, WguiTreeItemBase
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_frontend.gui.wgui_form_widget_creator import \
    WguiSimpleWidgetCreator, WGUI_META_LONGDESC_KEY, WGUI_META_LABEL_KEY, \
    WGUI_META_SKIP_KEY, WGUI_META_WIDGET_ARGS_KEY, WGUI_META_WARNING_KEY, \
    WGUI_META_WIDGET_KEY, WGUI_META_HEADERS_KEY, \
    WGUI_META_FIELDS_KEY, WGUI_META_ITEMS_KEY, WGUI_META_KEY
from wcr_core.modules.statemachine.db.wfhdb_protocols import *


#: Key for if/else-if elements
_WFGUI_ELSEIF_KEY = WFHDB_EXEC_IF_KEY + '-elseif'

#: Key for else elements
_WFGUI_ELSE_KEY = WFHDB_EXEC_IF_KEY + '-else'

#: List of available executables for drop-down selection
_WFGUI_EXECUTABLES = [
    ('Assign', WFHDB_EXEC_ASSIGN_KEY),
    ('If', WFHDB_EXEC_IF_KEY),
    ('Invoke', WFHDB_EXEC_INVOKE_KEY),
    ('Log', WFHDB_EXEC_LOG_KEY),
    ('Raise', WFHDB_EXEC_RAISE_KEY),
    ('Send', WFHDB_EXEC_SEND_KEY),
    ('Script', WFHDB_EXEC_SCRIPT_KEY)]

#: Meta-data for content-elements
_WFGUI_DATA_CONTENT_EXTRAS = {
    WGUI_META_LABEL_KEY: 'Content',
    WFHDB_CONTENT_EXPR_KEY: {
        WGUI_META_LABEL_KEY: 'Content-Expression',
        WGUI_META_WARNING_KEY: 'Cannot be used together with "Content"'
    },
    WFHDB_CONTENT_TEXT_KEY: {
        WGUI_META_LABEL_KEY: 'Content',
        WGUI_META_WARNING_KEY: 'Cannot be used together with '
                               '"Content-Expression"'
    }
}

#: Meta-data for parameter-list-elements
_WFGUI_DATA_PARAMS_EXTRAS = {
    WGUI_META_LABEL_KEY: 'Parameters',
    WGUI_META_WIDGET_KEY: 'table',
    WGUI_META_WIDGET_ARGS_KEY: {
        WGUI_META_FIELDS_KEY: [WFHDB_PARAM_NAME_KEY,
                               WFHDB_PARAM_EXPR_KEY],
        WGUI_META_HEADERS_KEY: ['Name', 'Value / Expression']
    }
}

#: Meta-data for executables
_WFGUI_DATA_EXTRAS = {
    WGUI_META_ITEMS_KEY: {
        WFHDB_EXEC_ASSIGN_KEY: {
            WGUI_META_LABEL_KEY: 'Assignment',
            WFHDB_EXEC_ASSIGN_LOC_KEY: {
                WGUI_META_LABEL_KEY: 'Variable-Name',
                WGUI_META_LONGDESC_KEY: 'Name of the variable to be '
                                        'assigned'
            },
            WFHDB_EXEC_ASSIGN_EXPR_KEY: {
                WGUI_META_LABEL_KEY: 'Value',
                WGUI_META_LONGDESC_KEY: 'Value to assign to the variable '
                                        'on execution'
            }
        },
        WFHDB_EXEC_IF_KEY: {
            WGUI_META_LABEL_KEY: 'If',
            WGUI_META_ITEMS_KEY: {
                WFHDB_EXEC_IF_COND_KEY: {
                    WGUI_META_LABEL_KEY: 'Condition',
                    WGUI_META_LONGDESC_KEY: 'Condition'
                },
                WFHDB_EXEC_IF_EXECS_KEY: {
                    WGUI_META_LABEL_KEY: 'Executables',
                    WGUI_META_SKIP_KEY: True
                }
            },
        },
        WFHDB_EXEC_INVOKE_KEY: {
            WGUI_META_LABEL_KEY: 'Invoke',
            WFHDB_EXEC_INVOKE_TYPE_KEY: {
                WGUI_META_LABEL_KEY: 'Service Type',
                WGUI_META_LONGDESC_KEY: 'Type of the service to be invoked'
            },
            WFHDB_EXEC_INVOKE_SRC_KEY: {
                WGUI_META_LABEL_KEY: 'Source',
                WGUI_META_LONGDESC_KEY: 'URI to be passed to the external '
                                        'service'
            },
            WFHDB_EXEC_INVOKE_ID_KEY: {
                WGUI_META_LABEL_KEY: 'ID',
                WGUI_META_LONGDESC_KEY: 'Invoke-Instance Identifier'
            },
            WFHDB_EXEC_INVOKE_AUTOFWD_KEY: {
                WGUI_META_LABEL_KEY: 'Auto-Forward',
                WGUI_META_LONGDESC_KEY: 'Whether to forward events to the '
                                        'invoked process'
            },
            WFHDB_EXEC_INVOKE_CONTENT_KEY: _WFGUI_DATA_CONTENT_EXTRAS,
            WFHDB_EXEC_SEND_PARAMS_KEY: _WFGUI_DATA_PARAMS_EXTRAS,
            WFHDB_EXEC_INVOKE_FINALIZE_KEY: {
                WGUI_META_SKIP_KEY: True
            }
        },
        WFHDB_EXEC_LOG_KEY: {
            WGUI_META_LABEL_KEY: 'Log',
            WFHDB_EXEC_LOG_LABEL_KEY: {
                WGUI_META_LABEL_KEY: 'Label',
                WGUI_META_LONGDESC_KEY: 'Label describing the log message'
            },
            WFHDB_EXEC_LOG_EXPR_KEY: {
                WGUI_META_LABEL_KEY: 'Message',
                WGUI_META_LONGDESC_KEY: 'Message to be logged on execution'
            }
        },
        WFHDB_EXEC_RAISE_KEY: {
            WGUI_META_LABEL_KEY: 'Event',
            WGUI_META_LONGDESC_KEY: 'Event to raise on execution'
        },
        WFHDB_EXEC_SCRIPT_KEY: {
            WGUI_META_LABEL_KEY: 'Script',
            WGUI_META_LONGDESC_KEY: 'Script to execute'
        },
        WFHDB_EXEC_SEND_KEY: {
            WGUI_META_LABEL_KEY: 'Send',
            WFHDB_EXEC_SEND_EVENT_KEY: {
                WGUI_META_LABEL_KEY: 'Event',
                WGUI_META_LONGDESC_KEY: 'Name of the message being '
                                        'generated'
            },
            WFHDB_EXEC_SEND_TARGET_KEY: {
                WGUI_META_LABEL_KEY: 'Target',
                WGUI_META_LONGDESC_KEY: 'Message target to send event to'
            },
            WFHDB_EXEC_SEND_TYPE_KEY: {
                WGUI_META_LABEL_KEY: 'Transport-Type',
                WGUI_META_LONGDESC_KEY: 'URI that identifies the transport '
                                        'mechanism for the message'
            },
            WFHDB_EXEC_SEND_ID_KEY: {
                WGUI_META_LABEL_KEY: 'ID',
                WGUI_META_LONGDESC_KEY: 'Send-instance Identifier'
            },
            WFHDB_EXEC_SEND_DELAY_KEY: {
                WGUI_META_LABEL_KEY: 'Delay',
                WGUI_META_LONGDESC_KEY: 'Pre-dispatch delay'
            },
            WFHDB_EXEC_SEND_CONTENT_KEY: _WFGUI_DATA_CONTENT_EXTRAS,
            WFHDB_EXEC_SEND_PARAMS_KEY: _WFGUI_DATA_PARAMS_EXTRAS
        },
    }
}


class _WfguiWidgetCreator(WguiSimpleWidgetCreator):
    def __init__(self, data, updateF):
        """
        :param data: Meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItem) -> _
        """
        super().__init__(data)

        self.__updateF = updateF

    def _dataChanged(self, *keys, **kwargs):
        treeItem = kwargs.get('treeItem')
        if treeItem:
            self.__updateF(treeItem)


def _wfgui_truncate(istr):
    """
    Truncate the given string to one line.

    :param istr: String to be truncated
    :type istr: str
    :return: Truncated string
    :rtype: str
    """
    prts = istr.strip().split('\n')
    if len(prts) > 1:
        return prts[0] + ' ...'
    else:
        return istr


def _wfgui_getKeys(item, data):
    """
    Get the keys for the given item in the given meta-data set.

    :param item: Tree item to get keys for
    :type item: _WfguiExecutablesTreeItemBase
    :param data: Top-level meta-data set
    :type data: object
    :return: Keys for the given item
    :rtype: list[object]
    """
    if not isinstance(item, _WfguiExecutablesTreeItemBase):
        return tuple()
    path = item.getPath()
    mtyp = WFHDB_EXECUTABLE_META
    keys = []
    for elem in path:
        try:
            key = mtyp.getKey(elem, data)
        except:
            print('NO {}'.format(elem))
            print('  DATA {}'.format(data))
            print('  KEYS {}'.format(keys))
            print('  PATH {}'.format(path))
            traceback.print_exc()
            raise
        keys.append(key)
        # TODO mtyp.getValueType(key, data)
        if isinstance(mtyp, WdbsMultiMetaTypeP):
            mtyp = mtyp.getValueType()
        elif isinstance(mtyp, WdbsUnionMetaTypeP):
            mtyp = mtyp.getType(data[0])
        elif isinstance(mtyp, WdbsDictionaryMetaTypeP) or \
                isinstance(mtyp, WdbsTupleMetaTypeP) or \
                isinstance(mtyp, WdbsRecordMetaTypeP):
            mtyp = mtyp.getValueType(key)
        else:
            raise Exception('Unknown type: {}'.format(type(mtyp)))
        data = elem
    return tuple(keys)


def _wfgui_getData(data, *keys):
    """
    Get the data for the given keys.

    :param data: Top-level meta-data
    :type data: object
    :param keys: Keys
    :type keys: tuple
    :return: Data for the given keys
    :rtype: object
    """
    for key in keys:
        data = data[key]
    return data


def _wfgui_getPath(data, *keys):
    """
    Get the path for the given keys.

    :param data: Top-level meta-data
    :type data: object
    :param keys: Keys
    :type keys: tuple
    :return: Path for the given keys
    :rtype: [object]
    """
    path = []
    for key in keys:
        data = data[key]
        path.append(data)
    return path


class WfguiExecutableProperties(WguiPropertiesWidgetP):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.__data = None
        """:type: list[(str, object)] | None"""

        #: Tree item factory
        self.__itemFc = _WfguiExecutablesTreeItemFactory()
        #: Tree model
        self.__treeM = WguiTreeModel(self.__itemFc)

        #: Tree-view widget
        self.__treeW = QTreeView()
        #: Current item properties group
        self.__propertiesW = QGroupBox('Properties of selected item')
        #: Combo-box for type-selection for new items
        self.__newComboW = QComboBox()

        #: Add-item button
        self.__addBtn = QPushButton('Add')
        #: Add-item-as-child button
        self.__addChldBtn = QPushButton('Add as Child')
        #: Add-elseif button
        self.__addElseifBtn = QPushButton('Add Elseif')
        #: Add-else button
        self.__addElseBtn = QPushButton('Add Else')
        #: Remove button
        self.__delBtn = QPushButton('Remove')

        self.setLayout(self._createLayout())
        self._initElements()
        self.updateFields()

        self.setMinimumWidth(800)

    def _createLayout(self):
        """
        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()

        self.__propertiesW.setMinimumWidth(300)

        splitter = QSplitter(Qt.Horizontal)
        splitter.addWidget(self.__treeW)
        scroll = QScrollArea()
        scroll.setWidget(self.__propertiesW)
        scroll.setWidgetResizable(True)
        scroll.setMinimumWidth(350)
        splitter.addWidget(scroll)
        layout.addWidget(splitter)

        btnLayout = QHBoxLayout()
        btnLayout.addWidget(self.__newComboW)
        btnLayout.addWidget(self.__addBtn)
        btnLayout.addWidget(self.__addChldBtn)
        btnLayout.addStretch(1)
        btnLayout.addWidget(self.__addElseifBtn)
        btnLayout.addWidget(self.__addElseBtn)
        btnLayout.addStretch(2)
        btnLayout.addWidget(self.__delBtn)
        layout.addLayout(btnLayout)

        return layout

    def _initElements(self):
        """
        Initialize UI elements.
        """
        self.__treeW.setModel(self.__treeM)
        self.__treeW.selectionModel().selectionChanged.connect(
            self._selectionChanged)
        self.__treeW.header().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.__treeW.header().setSectionResizeMode(1, QHeaderView.Stretch)
        self.__treeW.setSelectionMode(QAbstractItemView.SingleSelection)
        self.__treeW.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.__propertiesW.setLayout(QVBoxLayout())
        self.__propertiesW.setFlat(True)

        for item, code in _WFGUI_EXECUTABLES:
            self.__newComboW.addItem(item, code)

        self.__addBtn.clicked.connect(self._add)
        self.__addChldBtn.clicked.connect(self._addChild)
        self.__addElseifBtn.clicked.connect(self._addElseIf)
        self.__addElseBtn.clicked.connect(self._addElse)
        self.__delBtn.clicked.connect(self._removeSelected)

        self._updateButtonStates()

    # ------------------------------------------------------------------------

    def _item(self, idx):
        """
        Get the tree item at the given index.

        :param idx: Item index
        :type idx: QModelIndex
        :return: Tree item at the given index
        :rtype: _WfguiExecutablesTreeItemBase
        """
        return self.__treeM.item(idx)

    def _getSelectedIndex(self):
        """
        Get the index of the currently selected item, if any.

        :return: Currently selected item, if any
        :rtype: QModelIndex | None
        """
        indices = self.__treeW.selectedIndexes()
        if not indices:
            return None
        idx = indices[0]
        return idx

    def _getIfParent(self, idx=None):
        """
        Get the closest "if"-element parent's index of the currently selected
        item, if any.

        :return: "if"-parent of currently selected item, if any
        :rtype: QModelIndex | None
        """
        if not idx:
            idx = self._getSelectedIndex()
        while idx and idx.isValid() and \
                not isinstance(self._item(idx), _WfguiExecutablesIfTreeItem):
            idx = idx.parent()
        return idx if (idx and idx.isValid()) else None

    def _newEmptyExecutableData(self):
        """
        Create an empty data structure for the currently chosen executable
        type.

        :return: Empty data structure
        :rtype: (str, object)
        """
        code = wgui_qcb_get_data(self.__newComboW)
        return ScmnMTuple(
            code, WFHDB_EXECUTABLE_META_UNION.getType(code).getEmptyValue())

    #--------------------------------------------------------------------------

    def _doAdd(self, data, pidx, row=None, insertF=None):
        """
        Add a new element.

        :param data: Data for the new element
        :type data: object
        :param pidx: Index of the parent item
        :type pidx: QModelIndex
        :param row: Row to insert the new element at
        :type row: int
        :param insertF: Insertion function, if any (Keys, parent-index, row)
        :type insertF: (tuple, _WfguiTreeItemBase, int) -> _WfguiTreeItemBase
        """
        pitm = self._item(pidx)
        keys = _wfgui_getKeys(pitm, self.__data)
        if isinstance(pitm, _WfguiExecutablesTreeNodeItemBase):
            keys += pitm.getChildContainerKeys()
        _data = _wfgui_getData(self.__data, *keys)
        assert isinstance(_data, list)

        if row is None:
            row = len(_data)
        _data.insert(row, data)
        keys += (row,)

        if insertF:
            item = insertF(keys, pitm, row)
        else:
            (_, item), *_ = self.__treeM.insertItem(
                (keys, self.__data, self._dataUpdated), pitm, row)
        idx = self.__treeM.getItemIndex(item)
        self.__treeW.expand(idx)
        self.__treeW.selectionModel().select(
            QItemSelection(
                idx, idx.sibling(idx.row(), self.__treeM.columnCount() - 1)),
            QItemSelectionModel.ClearAndSelect)
        self._updateButtonStates()

    def _add(self):
        """
        Add a new element of the currently chosen type below the currently
        selected element.
        """
        idx = self._getSelectedIndex()
        prnt = idx.parent() if idx else None
        row = idx.row() + 1 if idx else None
        self._doAdd(self._newEmptyExecutableData(), prnt, row)

    def _addChild(self):
        """
        Add a new element of the currently chosen type as child of the
        currently selected element, if possible.
        """
        idx = self._getSelectedIndex()
        if idx is None:
            return None
        item = self._item(idx)
        if not (isinstance(item, _WfguiExecutablesIfCondTreeItem) or
                    isinstance(item, _WfguiExecutablesElseTreeItem) or
                    isinstance(item, _WfguiExecutablesInvokeTreeItem)):
            return
        self._doAdd(self._newEmptyExecutableData(), idx)

    def _addElseIf(self):
        """
        Add an new elseif-element for the currently selected if-element, or for
        the closest "if"-parent of the currently selected element.
        """
        # noinspection PyShadowingNames
        def insert(keys, parent, row):
            """
            :type keys: tuple
            :type parent: _WfguiTreeItemBase
            :type row: int
            """
            # noinspection PyTypeChecker
            item = _WfguiExecutablesIfCondTreeItem(
                keys, self.__data, self._dataUpdated, parentItem=parent)
            self.__treeM.insertTreeItem(item, parentItem=parent, row=row)
            return item

        pidx = self._getIfParent()
        if not pidx:
            return
        pitm = self._item(pidx)
        assert isinstance(pitm, _WfguiExecutablesIfTreeItem)
        row = 0
        for chld in pitm.getChildren():
            if isinstance(chld, _WfguiExecutablesElseTreeItem):
                break
            row += 1
        data = WFHDB_EXECUTABLE_IF_BRANCH_META.getEmptyValue()
        self._doAdd(data, pidx, row=row, insertF=insert)
        self._updateButtonStates()

    def _addElse(self):
        """
        Add an new else-element for the currently selected if-element, or for
        the closest "if"-parent of the currently selected element.
        """
        # noinspection PyShadowingNames
        def insert(keys, parent, row):
            """
            :type keys: tuple
            :type parent: _WfguiTreeItemBase
            :type row: int
            """
            # noinspection PyTypeChecker
            item = _WfguiExecutablesElseTreeItem(
                keys, self.__data, self._dataUpdated, parentItem=parent)
            self.__treeM.insertTreeItem(item, parentItem=parent, row=row)
            return item

        pidx = self._getIfParent()
        if not pidx:
            return
        pitm = self._item(pidx)
        assert isinstance(pitm, _WfguiExecutablesIfTreeItem)
        for chld in pitm.getChildren():
            if isinstance(chld, _WfguiExecutablesElseTreeItem):
                return
        data = WFHDB_EXECUTABLE_IF_BRANCH_META.getEmptyValue()
        self._doAdd(data, pidx, insertF=insert)

    def _removeSelected(self):
        """
        Remove the currently selected element.
        """
        idx = self._getSelectedIndex()
        item = self._item(idx)
        pitm = item.parentItem
        if isinstance(pitm, _WfguiExecutablesTreeNodeItemBase):
            pkey = _wfgui_getKeys(pitm, self.__data)
            pkey += pitm.getChildContainerKeys()
        else:
            pkey = tuple()

        keys = _wfgui_getKeys(item, self.__data)
        _data = self.__data
        for key in pkey:
            _data = _data[key]
        assert isinstance(_data, list)

        self.__treeM.removeRow(idx.row(), idx.parent())
        # noinspection PyTypeChecker
        del _data[keys[len(pkey)]]

    #--------------------------------------------------------------------------

    def _dataUpdated(self, item):
        """
        Notify the tree-model about an updated item.

        :param item: Updated item
        :type item: WguiTreeItemP
        """
        self.__treeM.refreshItem(item)

    def _selectionChanged(self, selection, *_):
        """
        Callback-function for selection changes.

        :param selection: New selection
        :type selection: QItemSelection
        """
        layout = self.__propertiesW.layout()
        """:type: QVBoxLayout"""
        for widx in reversed(range(layout.count())):
            layout.itemAt(widx).widget().setParent(None)

        indeces = selection.indexes()

        self._updateButtonStates(indeces[0] if indeces else None)

        if not indeces:
            return

        item = self._item(indeces[0])
        if isinstance(item, _WfguiTreeItemBase):
            detailUI = item.createDetailUI()
            if detailUI is None:
                return
            if isinstance(detailUI, QLayout):
                widget = QWidget()
                widget.setLayout(detailUI)
                layout.addWidget(widget)
            else:
                layout.addWidget(detailUI)

    def _updateButtonStates(self, index=None):
        """
        Update the states of the buttons based on the given index or the
        currently selected item.

        :param index: Item index
        :type index: QModelIndex | None
        """
        if not index:
            index = self._getSelectedIndex()

        if not index:
            self.__delBtn.setEnabled(False)
            self.__addElseifBtn.setEnabled(False)
            self.__addElseBtn.setEnabled(False)
            self.__addChldBtn.setEnabled(False)
            return

        self.__delBtn.setEnabled(True)
        ifix = self._getIfParent(index)
        isIf = ifix is not None
        self.__addElseifBtn.setEnabled(isIf)
        self.__addElseBtn.setEnabled(
            isIf and not any(isinstance(c, _WfguiExecutablesElseTreeItem)
                             for c in self.__treeM.item(ifix).getChildren()))

        item = self._item(index)
        self.__addChldBtn.setEnabled(
            isinstance(item, _WfguiExecutablesIfCondTreeItem) or
            isinstance(item, _WfguiExecutablesElseTreeItem) or
            isinstance(item, _WfguiExecutablesInvokeTreeItem))
        self.__addBtn.setEnabled(
            not (isinstance(item, _WfguiExecutablesIfCondTreeItem) or
                 isinstance(item, _WfguiExecutablesElseTreeItem)))

    #--------------------------------------------------------------------------

    def updateFields(self):
        self.__treeW.selectionModel().clearSelection()

        model = self.__treeM
        model.removeRows(0, model.rowCount())

        self.__data = self._getExecutables()

        for idx, _ in enumerate(self.__data):
            model.insertItem(((idx,), self.__data, self._dataUpdated))

        self.__treeW.expandAll()

    def save(self):
        if self.__data is not None:
            self._setExecutables(self.__data)

    #--------------------------------------------------------------------------

    def _getExecutables(self):
        """
        Get the current executables to display.

        :return: Current executables
        :rtype: list[(str, object)]
        """
        raise NotImplementedError()

    def _setExecutables(self, execs):
        """
        Set the executables.

        :param execs: Executables
        :type execs: list[(str, object)]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class _WfguiTreeItemBase(WguiTreeItemBase):
    """
    State-machine executable tree item.
    """

    def __init__(self, parentItem=None):
        super().__init__(parentItem)

    def setData(self, data, column, role=None):
        pass

    def getColumnCount(self):
        return 2

    @property
    def flags(self):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    @abstractmethod
    def createDetailUI(self):
        """
        Create the detail-view UI.

        :return: Detail-view UI
        :rtype: QWidget
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class _WfguiExecutablesTreeItemBase(_WfguiTreeItemBase):
    _LABEL = ''

    _CODE = None

    _TYPE = None

    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(parentItem)

        assert self._CODE
        assert self._TYPE

        self.__data = data
        self.__updateF = updateF

        self.__path = _wfgui_getPath(data, *keys)
        self.__ownData = _wfgui_getData(data, *keys)

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            return self._getLabel() if column == 0 else self._getDescription()
        return None

    def getPath(self):
        return self.__path

    def getField(self, *keys):
        return _wfgui_getData(self.__ownData, *keys)

    def _getLabel(self):
        """
        Get the label for the tree item.

        :return: Label
        :rtype: str
        """
        return self._LABEL

    @abstractmethod
    def _getDescription(self):
        """
        Get the description for the tree item.

        :return: Description
        :rtype: str
        """
        raise NotImplementedError()

    def _getMetaDescription(self):
        """
        Get the meta-description for the tree item.

        :return: Meta-description
        :rtype: dict[object, object]
        """
        meta = _WFGUI_DATA_EXTRAS.get(WGUI_META_ITEMS_KEY)

        code = self._CODE
        if isinstance(code, tuple):
            for c in code:
                meta = meta.get(c)
        else:
            meta = meta.get(code)

        return meta

    def createDetailUI(self):
        uiCreator = _WfguiWidgetCreator(self.__data, self.__updateF)
        keys = _wfgui_getKeys(self, self.__data)
        meta = self._getMetaDescription()

        widget = uiCreator.createWidget(
            self._TYPE, *keys, treeItem=self, meta=meta)
        if not widget:
            return None
        elif widget.layout():
            return widget
        else:
            wrap = QWidget()
            layout = QFormLayout()
            layout.setRowWrapPolicy(QFormLayout.WrapAllRows)
            layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
            layout.setContentsMargins(0,0,0,0)
            wmta = widget.property(WGUI_META_KEY)
            labl = wmta.get(WGUI_META_LABEL_KEY) if wmta else ''
            layout.addRow(labl, widget)
            wrap.setLayout(layout)
            return wrap


# noinspection PyAbstractClass
class _WfguiExecutablesTreeNodeItemBase(_WfguiExecutablesTreeItemBase):
    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

    @abstractmethod
    def getChildContainerKeys(self):
        raise NotImplementedError()


# noinspection PyAbstractClass
class _WfguiExecutablesTreeLeafItemBase(_WfguiExecutablesTreeItemBase):
    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

    def addChild(self, child, row=None):
        raise Exception('Unsupported operation')

    def removeChild(self, row):
        raise Exception('Unsupported operation')

    def removeChildren(self, row, count):
        raise Exception('Unsupported operation')


class _WfguiExecutablesAssignTreeItem(_WfguiExecutablesTreeLeafItemBase):
    _LABEL = 'Assignment'

    _CODE = WFHDB_EXEC_ASSIGN_KEY

    _TYPE = WFHDB_EXECUTABLE_ASSIGN_META

    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

    def _getDescription(self):
        vrbl = self.getField(WFHDB_EXEC_ASSIGN_LOC_KEY)
        vval = self.getField(WFHDB_EXEC_ASSIGN_EXPR_KEY)

        if vrbl:
            return '{} = {}'.format(vrbl, vval or '')
        else:
            return ''


class _WfguiExecutablesLogTreeItem(_WfguiExecutablesTreeLeafItemBase):
    _LABEL = 'Log'

    _CODE = WFHDB_EXEC_LOG_KEY

    _TYPE = WFHDB_EXECUTABLE_LOG_META

    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

    def _getDescription(self):
        msg = self.getField(WFHDB_EXEC_LOG_EXPR_KEY)
        """:type: str | None"""
        lbl = self.getField(WFHDB_EXEC_LOG_LABEL_KEY)
        """:type: str | None"""
        if msg:
            mstr = _wfgui_truncate(msg)
            return '{}: {}'.format(lbl, mstr) if lbl else mstr
        else:
            return lbl or ''


class _WfguiExecutablesRaiseTreeItem(_WfguiExecutablesTreeLeafItemBase):
    _LABEL = 'Raise'

    _CODE = WFHDB_EXEC_RAISE_KEY

    _TYPE = WFHDB_EXECUTABLE_RAISE_META

    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

    def _getDescription(self):
        return self.getField() or None


class _WfguiExecutablesScriptTreeItem(_WfguiExecutablesTreeLeafItemBase):
    _LABEL = 'Script'

    _CODE = WFHDB_EXEC_SCRIPT_KEY

    _TYPE = WFHDB_EXECUTABLE_SCRIPT_META

    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

    def _getDescription(self):
        script = self.getField()
        """:type: str | None"""
        if script:
            return _wfgui_truncate(script)
        else:
            return ''


class _WfguiExecutablesSendTreeItem(_WfguiExecutablesTreeLeafItemBase):
    _LABEL = 'Send'

    _CODE = WFHDB_EXEC_SEND_KEY

    _TYPE = WFHDB_EXECUTABLE_SEND_META

    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

    def _getDescription(self):
        return self.getField(WFHDB_EXEC_SEND_EVENT_KEY)


class _WfguiExecutablesInvokeTreeItem(_WfguiExecutablesTreeNodeItemBase):
    _LABEL = 'Invoke'

    _CODE = WFHDB_EXEC_INVOKE_KEY

    _TYPE = WFHDB_EXECUTABLE_INVOKE_META

    def __init__(self, keys, data, updateF, factory, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

        fData = self.getField(WFHDB_EXEC_INVOKE_FINALIZE_KEY)
        """:type: list[object] | None"""
        if fData:
            for idx, _ in enumerate(fData):
                ckey = keys + (WFHDB_EXEC_INVOKE_FINALIZE_KEY, idx)
                cNode, _, _ = factory.createTreeItem(
                    (ckey, data, updateF), self)
                if cNode:
                    self.addChild(cNode)

    def _getDescription(self):
        return self.getField(WFHDB_EXEC_SEND_TYPE_KEY) or None

    def getChildContainerKeys(self):
        return WFHDB_EXEC_INVOKE_FINALIZE_KEY,


class _WfguiExecutablesIfTreeItem(_WfguiExecutablesTreeNodeItemBase):
    _LABEL = 'If-ElseIf-Else'

    _CODE = WFHDB_EXEC_IF_KEY

    _TYPE = WFHDB_EXECUTABLE_IF_META

    def __init__(self, keys, data, updateF, factory, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param factory: Item factory
        :type factory: WguiTreeItemFactoryP
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem)

        params = self.getField()
        """:type: list[dict[str, list]] | None"""
        if params:
            last = len(params) - 1
            for bidx, brnch in enumerate(params):
                cond = brnch.get(WFHDB_EXEC_IF_COND_KEY)
                excs = brnch.get(WFHDB_EXEC_IF_EXECS_KEY)
                bkey = keys + (bidx,)
                if bidx == 0 or bidx != last or cond:
                    bNode = _WfguiExecutablesIfCondTreeItem(
                        bkey, data, updateF, self)
                else:
                    bNode = _WfguiExecutablesElseTreeItem(
                        bkey, data, updateF, self)

                if excs:
                    for cidx, cData in enumerate(excs):
                        ckey = bkey + (WFHDB_EXEC_IF_EXECS_KEY, cidx)
                        cNode, _, _ = factory.createTreeItem(
                            (ckey, data, updateF), bNode)
                        if cNode:
                            bNode.addChild(cNode)

                self.addChild(bNode)

        if self.getChildCount() == 0:
            cdat = _wfgui_getData(data, *keys)
            assert isinstance(cdat, list)
            cdat.append(WFHDB_EXECUTABLE_IF_BRANCH_META.getEmptyValue())
            self.addChild(_WfguiExecutablesIfCondTreeItem(
                keys + (0,), data, updateF, parentItem=self))

    def _getDescription(self):
        return ''

    def _getMetaDescription(self):
        return {WGUI_META_ITEMS_KEY: {WGUI_META_SKIP_KEY: True}}

    def getChildContainerKeys(self):
        return tuple()


class _WfguiExecutablesIfCondTreeItem(_WfguiExecutablesTreeNodeItemBase):
    _LABEL = 'If'

    _CODE = (WFHDB_EXEC_IF_KEY, WGUI_META_ITEMS_KEY)

    _TYPE = WFHDB_EXECUTABLE_IF_BRANCH_META

    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem=parentItem)

    def _getLabel(self):
        return 'If' if self.getRow() == 0 else 'Elseif'

    def _getDescription(self):
        return self.getField(WFHDB_EXEC_IF_COND_KEY)

    def getChildContainerKeys(self):
        return WFHDB_EXEC_IF_EXECS_KEY,


class _WfguiExecutablesElseTreeItem(_WfguiExecutablesTreeNodeItemBase):
    _LABEL = 'Else'

    _CODE = (WFHDB_EXEC_IF_KEY, WGUI_META_ITEMS_KEY)

    _TYPE = WFHDB_EXECUTABLE_IF_BRANCH_META

    def __init__(self, keys, data, updateF, parentItem=None):
        """
        :param keys: Keys of the represented meta-data element
        :type keys: tuple
        :param data: Top-level meta-data
        :type data: object
        :param updateF: Update notification function
        :type updateF: (WguiTreeItemP) -> _
        :param parentItem: Parent tree-item
        :type parentItem: WguiTreeItemP
        """
        super().__init__(keys, data, updateF, parentItem=parentItem)

        meta = deepcopy(super()._getMetaDescription())
        meta[WFHDB_EXEC_IF_COND_KEY][WGUI_META_SKIP_KEY] = True
        self.__meta = meta

    def _getDescription(self):
        return ''

    def _getMetaDescription(self):
        return self.__meta

    def getChildContainerKeys(self):
        return WFHDB_EXEC_IF_EXECS_KEY,


class _WfguiExecutablesRootTreeItem(_WfguiTreeItemBase):
    def __init__(self):
        super().__init__()

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            return 'Type' if column == 0 else 'Description'
        return None

    def createDetailUI(self):
        return None


class _WfguiExecutablesTreeItemFactory(WguiTreeItemFactoryP):
    def createTreeItem(self, itemData, parentItem, **kwargs):
        if not parentItem:
            return _WfguiExecutablesRootTreeItem(), None, None
        if not itemData:
            return None, None, None

        keys, data, updateF = itemData
        """:type: (list[object], object, (WguiTreeItemP)) -> _"""
        idat = _wfgui_getData(data, *keys)
        """:type: tuple"""
        code, _ = idat
        keys = keys + (1,)

        if code == WFHDB_EXEC_ASSIGN_KEY:
            return _WfguiExecutablesAssignTreeItem(
                keys, data, updateF, parentItem=parentItem), parentItem, None
        if code == WFHDB_EXEC_LOG_KEY:
            return _WfguiExecutablesLogTreeItem(
                keys, data, updateF, parentItem=parentItem), parentItem, None
        if code == WFHDB_EXEC_SCRIPT_KEY:
            return _WfguiExecutablesScriptTreeItem(
                keys, data, updateF, parentItem=parentItem), parentItem, None
        if code == WFHDB_EXEC_RAISE_KEY:
            return _WfguiExecutablesRaiseTreeItem(
                keys, data, updateF, parentItem=parentItem), parentItem, None
        if code == WFHDB_EXEC_SEND_KEY:
            return _WfguiExecutablesSendTreeItem(
                keys, data, updateF, parentItem=parentItem), parentItem, None
        if code == WFHDB_EXEC_INVOKE_KEY:
            return _WfguiExecutablesInvokeTreeItem(
                keys, data, updateF, self, parentItem=parentItem), \
                parentItem, None
        if code == WFHDB_EXEC_IF_KEY:
            return _WfguiExecutablesIfTreeItem(
                keys, data, updateF, self, parentItem=parentItem), \
                parentItem, None
        else:
            raise Exception('Unknown executable code: {}'.format(code))
