# ----------------------------------------------------------------------------
# - wfgui-scene-item                                                         -
# - wild cruncher: scene items                         state machine cockpit -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfgui_scene_item
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher scene items

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod
import math

from PyQt4.QtCore import QRectF, Qt
from PyQt4.QtGui import QPainterPath, QPainter, QPen
from wcr_core.db_api.whdb_protocols_api import WHDB_HALIGN_CENTER

from wcr_core.modules.statemachine.db.wfhdb_protocols import \
    WfhdbMayBeFinalP, WfhdbMayBeInitialP
from wcr_core.modules.statemachine.db.wfhdb_xbt_utils import \
    WfhdbXbtException
# noinspection PyProtectedMember
from wcr_frontend.gui.wgui_scene_item import WguiOperatorInstanceSceneItem, \
    _WGUI_RAD_CONN, WguiInstanceSceneItem, WguiTextIcon, \
    WguiConnectionPointSceneItem, WGUI_SHAPE_COLOUR, WguiIconText, \
    WguiInstanceBaseSceneItem
from wcr_frontend.gui.wgui_resizer import WguiSquareBoundingRectResizer
from wcr_core.modules.statemachine.db.wfhdb_block import WfhdbParallelBlock
from wcr_core.modules.statemachine.db.wfhdb_instance import WfhdbBlockInstance


def _len(a, b):
    """
    :type a: float
    :type b: float
    :rtype: float
    """
    return math.sqrt(pow(a, 2.0) + pow(b, 2.0))


class WfguiStatePaintMixin(metaclass=ABCMeta):
    @abstractmethod
    def getDataObject(self):
        raise NotImplementedError()

    def paint(self, painter, option, widget=None):
        painter.save()
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setRenderHint(QPainter.HighQualityAntialiasing)
        # noinspection PyCallByClass,PyTypeChecker
        WguiInstanceBaseSceneItem.paint(self, painter, option, widget)
        painter.restore()

    def _paintOutline(self, painter, outline, lc, lw):
        inst = self.getDataObject()
        """:type: WfhdbStateInstance"""
        if isinstance(inst, WfhdbMayBeInitialP) and inst.isInitialState():
            wfct = 3
        else:
            wfct = 1
        painter.save()
        painter.setPen(QPen(
            lc if lc else WGUI_SHAPE_COLOUR,
            wfct * lw if lw is not None else wfct,
            Qt.SolidLine))
        painter.drawPath(outline)
        painter.restore()


class WfguiConnectionPointSceneItem(WguiConnectionPointSceneItem):
    def __init__(self, conn, item, posp, vctx, parent=None, scene=None):
        super().__init__(conn, item, posp, vctx, parent, scene)

    def _createBackgroundAndIcon(self, conn, inst):
        return None, None


class WfguiOperatorInstanceSceneItem(WguiOperatorInstanceSceneItem,
                                     WfguiStatePaintMixin):
    def __init__(self, inst, citmF, posp, vctx, parent=None, scene=None):
        """
        :param inst: the the database function
        :type inst: WfhdbOperatorInstance
        :param citmF: Connection point scene item creation function
        :type citmF: (WguiConnectionPoint, WguiInstance, ViewContext)
                     -> WguiConnectionPointSceneItem
        :param posp: Connection point position provider
        :type posp: WguiXPointPositionProviderP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the scene object
        """
        super().__init__(inst, citmF, posp, vctx, parent, scene)

    def _createResizer(self, vctx, scene):
        # noinspection PyTypeChecker
        return WguiSquareBoundingRectResizer(
            self._getResizingRectAndPath, self._fitToRect,
            lambda t: vctx.snapPoint(*t), margin=_WGUI_RAD_CONN,
            minSize=self.MINIMUM_SIZE, parent=self, scene=scene)

    # noinspection PyUnresolvedReferences
    def _createIcon(self):
        inst = self.getDataObject()
        """:type: WfhdbStateInstance"""
        w, h = inst.getDefaultDimension()
        path = QPainterPath()
        text = inst.getCustomText() or inst.getSymbol()
        icon = WguiTextIcon(text, self.parentWidget()) if text else None
        path.addEllipse(QRectF(-w / 2.0, -h / 2.0, w, h))
        if isinstance(inst, WfhdbMayBeFinalP) and inst.isFinalState():
            path.addEllipse(
                QRectF(-w / 2.0 + 5, -h / 2.0 + 5, w - 10, h - 10))
        return icon, (w, h), path

    def paint(self, painter, option, widget=None):
        WfguiStatePaintMixin.paint(self, painter, option, widget)

    def _paintOutline(self, painter, outline, lc, lw):
        WfguiStatePaintMixin._paintOutline(self, painter, outline, lc, lw)


class WfguiXbtIconText(WguiIconText):
    def __init__(self, data, pctx, align=WHDB_HALIGN_CENTER,
                 parent=None, scene=None):
        """
        :param data: Data
        :type data: WhdbNamedEntity
        :param pctx: Project context
        :type pctx: WfmodStatemachineProjectContextP
        :param align: Text alignment
        :type align: int
        :param parent: the parent scene item
        :param scene: the scene object
        """
        self.__projectCtx = pctx

        super().__init__(data, align, parent, scene)

    def _getNameInfo(self, item):
        try:
            xmng = self.__projectCtx.getXbtManager()
            sobj, strn, sstt, estt, _, eobj, *_ = xmng.getXbtElements(item)
        except WfhdbXbtException:
            return 'INVALID XBT TRANSITION', ''

        name = strn.getName()
        if not name:
            if item is sstt:
                name = '\u2197{} [XBT]'.format(eobj.getName())
            else:
                name = '{}\u2197 [XBT]'.format(sobj.getName())

        return name, item.getInfo()


class WfguiXbtInstanceSceneItem(WfguiOperatorInstanceSceneItem):
    def __init__(self, inst, citmF, posp, pctx, vctx, parent=None, scene=None):
        """
        :param inst: the the database function
        :type inst: WfhdbOperatorInstance
        :param citmF: Connection point scene item creation function
        :type citmF: (WguiConnectionPoint, WguiInstance, ViewContext)
                     -> WguiConnectionPointSceneItem
        :param posp: Connection point position provider
        :type posp: WguiXPointPositionProviderP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the scene object
        """
        self.__projectCtx = pctx

        super().__init__(inst, citmF, posp, vctx, parent, scene)

    def _paintOutline(self, painter, outline, lc, lw):
        painter.save()
        painter.setPen(QPen(
            lc if lc else WGUI_SHAPE_COLOUR,
            2 * lw if lw is not None else 2,
            Qt.DotLine))
        painter.drawPath(outline)
        painter.restore()

    def _createNameInfoFields(self, data):
        nameInfoText = WfguiXbtIconText(
            data, self.__projectCtx, data.getLabelAlignment(),
            self, self.scene())
        nameInfoText.setZValue(self.NAMEINFO_Z)
        return nameInfoText


class WfguiInstanceSceneItem(WguiInstanceSceneItem, WfguiStatePaintMixin):
    def __init__(self, inst, citmF, posp, vctx, parent=None, scene=None):
        """
        :param inst: the the database function
        :type inst: WfhdbStateInstance
        :param citmF: Connection point scene item creation function
        :type citmF: (WguiConnectionPoint, WguiInstance, ViewContext)
                     -> WguiConnectionPointSceneItem
        :param posp: Connection point position provider
        :type posp: WguiXPointPositionProviderP
        :param vctx: View context
        :type vctx: ViewContext
        :param parent: the parent object
        :param scene: the scene object
        """
        super().__init__(inst, citmF, posp, vctx, parent, scene)

    def _createResizer(self, vctx, scene):
        # noinspection PyTypeChecker
        return WguiSquareBoundingRectResizer(
            self._getResizingRectAndPath, self._fitToRect,
            lambda t: vctx.snapPoint(*t), margin=_WGUI_RAD_CONN,
            minSize=self.MINIMUM_SIZE, parent=self, scene=scene)

    # noinspection PyUnresolvedReferences
    def _createIcon(self):
        inst = self.getDataObject()
        w, h = inst.getDefaultDimension()
        path = QPainterPath()
        text = inst.getCustomText()
        if not text:
            if isinstance(inst, WfhdbBlockInstance):
                if isinstance(inst.getBlock(), WfhdbParallelBlock):
                    text = 'Parallel'
                else:
                    text = 'Nested'
        icon = WguiTextIcon(text, self) if text else None
        path.addEllipse(QRectF(-w / 2.0, -h / 2.0, w, h))
        return icon, (w, h), path

    def paint(self, painter, option, widget=None):
        WfguiStatePaintMixin.paint(self, painter, option, widget)

    def _paintOutline(self, painter, outline, lc, lw):
        WfguiStatePaintMixin._paintOutline(self, painter, outline, lc, lw)
