# ----------------------------------------------------------------------------
# - fsm-gui-cstcx                                                            -
# - wild cruncher:state machine cockpit system context                       -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfgui_cstcx
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher state machine cockpit system contex

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

# noinspection PyUnresolvedReferences
import wcr_core.modules.statemachine.wfmod_application_module
# noinspection PyUnresolvedReferences
import wcr_frontend.modules.statemachine.gui.wfgui_icons

from wcr_frontend.gui.wgui_scene import WguiGraphicsScene
from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_frontend.gui.wgui_scene_item_ids import XPNT_ITEM_ID
from wcr_frontend.gui.wgui_toolbox_item import CNCT_ITEM_ID
from wcr_frontend.gui.wgui_gui_module import \
    WguiCockpitSystemContextBase, wgui_getClassEntry, \
    WguiCockpitSystemModuleBase, WguiCockpitSystemContextP
from wcr_core.db_api.whdb_instance_api import WhdbBlockInstanceP
from wcr_frontend.modules.statemachine.gui.wfgui_clipboard import \
    WfguiClipboardHandler
from wcr_frontend.modules.statemachine.gui.wfgui_scene_context_menu_factory \
    import WfguiSceneContextMenuFactory
from wcr_frontend.modules.statemachine.gui.wfgui_scene_item_factory import \
    WfguiSceneItemFactory
from wcr_core.modules.statemachine.wfmod_application_module import \
    WFMOD_STATEMACHINE_APP_MODULE
from wcr_core.modules.statemachine.db.wfhdb_block import WfhdbNestedBlock, \
    WfhdbParallelBlock, WFHDB_NESTED_BLOCK_CODE, WFHDB_PARALLEL_BLOCK_CODE
from wcr_frontend.modules.statemachine.gui.wfgui_scene_view_mouse_handler \
    import WfguiSceneViewMouseHander
from wcr_frontend.modules.statemachine.gui.wfgui_variable_table import \
    WfguiVariableDictionary
from wcr_frontend.modules.statemachine.gui.wfgui_props_dialog_factory import \
    WfguiPropertiesDialogFactory
from wcr_frontend.gui.wgui_db_tree import wgui_createDbTree
from wcr_frontend.modules.statemachine.gui.wfgui_event_table import \
    WfguiEventDisplay, WfguiItemSceneSelectionSynchronizer
from wcr_frontend.modules.statemachine.gui.wfgui_highlight_module import \
    WfguiEventItemsHighlightModule, WfguiTriggerMissingHighlightModule
from wcr_frontend.modules.statemachine.gui.wfgui_db_tree import \
    WfguiDbTreeItemFactory


_WFGUI_ICONS = {
    WFHDB_NESTED_BLOCK_CODE: ':/img/wfgui_nsted.png',
    WFHDB_PARALLEL_BLOCK_CODE: ':/img/wfgui_prlll.png',
    CNCT_ITEM_ID: ':/img/wfgui_trans.png',
    'wcr.statemachine.std.state': ':/img/wfgui_state.png',
    'wcr.highlighter.event': ':/img/wfgui_event.png',
    'editor-tab': ':/img/wfgui_ticon.png',
    'wcr.core.statemachine.layout-spring': ':/img/wfgui_lyopt.png',
    'wcr.core.statemachine.layout-grid': ':/img/wfgui_lygrd.png',
    'wcr.core.statemachine.layout-circular': ':/img/wfgui_lycir.png',
}

WFGUI_DEFAULT_ICON_CODES = {
    WfhdbNestedBlock: WFHDB_NESTED_BLOCK_CODE,
    WfhdbParallelBlock: WFHDB_PARALLEL_BLOCK_CODE,
    WhdbBlockInstanceP: lambda x: WFGUI_DEFAULT_ICON_CODES.get(
        x.getBlock().__class__)
}


# noinspection PyAbstractClass
class WfguiCockpitSystemContextP(WguiCockpitSystemContextP):
    @abstractmethod
    def getEventItemsProvider(self):
        """
        Get the provider for items affected by the currently selected event.

        :return: Event items provider
        :rtype: WfguiItemSelectionProviderP
        """
        raise NotImplementedError()

    @abstractmethod
    def refreshEvents(self):
        """
        Refresh the list of events.
        """
        raise NotImplementedError()


class _WfguiCockpitSystemContext(WguiCockpitSystemContextBase,
                                 WfguiCockpitSystemContextP):
    def __init__(self, cmod, pctx, pcfg, actions, actx, parent):
        """
        :param cmod: Cockpit system module
        :type cmod: WguiCockpitSystemModuleP
        :param pctx: Project context
        :type pctx: WfmodStatemachineProjectContextP
        :param pcfg: Project editor configuration
        :type pcfg: WguiProjectEditorCfgP
        :param actions: Actions
        :type actions: dict[str, WactActionP]
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(cmod)

        #: Project context
        self.__projectCtx = pctx

        #: Event display widget
        self.__eventDisplay = None
        """:type: WfguiEventDisplay"""

        self.__sceneItmFy = WfguiSceneItemFactory(pctx)
        self.__scene = WguiGraphicsScene(
            pctx.getProjectDB(), pcfg, actions, pctx, self, actx, parent)
        self.__propDiagF = WfguiPropertiesDialogFactory(
            pctx, pcfg, self.__scene, actions, self, parent)
        self.__clpbrdHandler = WfguiClipboardHandler(
            self.__projectCtx, self.__scene, cmod.getAppModule())
        self.__cntxtMenuF = WfguiSceneContextMenuFactory(
            self.__scene, actions, self.__clpbrdHandler, pctx,
            self.__propDiagF)
        self.__mouseHandler = WfguiSceneViewMouseHander(
            pctx, self.__propDiagF, self.__scene, self.__cntxtMenuF)

        self.__sidebarModules = self._createSidebarModules(
            pctx, self.__scene, self.__cntxtMenuF)
        self.__hilightModules = self._createHighlighters(
            self.__eventDisplay, self.__scene)

    def release(self):
        for _, _, smod in self.__sidebarModules:
            if isinstance(smod, WcmnReleasableP):
                smod.release()

        for item in [self.__sceneItmFy, self.__scene, self.__propDiagF,
                     self.__clpbrdHandler, self.__cntxtMenuF,
                     self.__mouseHandler, self.__sidebarModules,
                     self.__hilightModules]:
            if isinstance(item, WcmnReleasableP):
                item.release()

        self.__sceneItmFy = None
        self.__scene = None
        self.__propDiagF = None
        self.__clpbrdHandler = None
        self.__cntxtMenuF = None
        self.__mouseHandler = None
        self.__sidebarModules = []
        self.__hilightModules = []

        super().release()

    def _createSidebarModules(self, pctx, scene, cntxtMenuF):
        modules = []

        dbTree = wgui_createDbTree(
            pctx.getProjectDB().getTopHierarchy(), scene,
            WfguiDbTreeItemFactory(self), cntxtMenuF)
        modules.append(('dbtree', 'Database', dbTree))

        evntLst = WfguiEventDisplay(pctx.getEventManager())
        sync = WfguiItemSceneSelectionSynchronizer(evntLst, scene)
        evntLst.sync = sync
        self.__eventDisplay = evntLst
        modules.append(('events', 'Events', evntLst))

        varBrowser = WfguiVariableDictionary(pctx)
        modules.append(('variables', 'Variables', varBrowser))

        return modules

    def _createHighlighters(self, eventDisplay, scene):
        hlghtrs = [WfguiTriggerMissingHighlightModule(
            self.__projectCtx.getXbtManager(), scene)]
        """:type: list[WguiHighlightModule]"""
        if eventDisplay is not None:
            hlghtrs.append(
                WfguiEventItemsHighlightModule(
                    eventDisplay.getSelectedEventItemsProvider(), scene))
        return hlghtrs

    def getExcludedSystemToolboxItems(self):
        return [XPNT_ITEM_ID]

    def getHighlightingModules(self):
        return iter(self.__hilightModules)

    def getDirectoryDescription(self, level):
        return 'directory'

    def getScene(self):
        return self.__scene

    def getSceneViewMouseHandler(self):
        return self.__mouseHandler

    def getClipboardHandler(self):
        return self.__clpbrdHandler

    def getSceneItemFactory(self):
        return self.__sceneItmFy

    def getContextMenuFactory(self):
        return self.__cntxtMenuF

    def getSidebarModules(self):
        return iter(self.__sidebarModules)

    def getPropertiesDialogFactory(self):
        return self.__propDiagF

    def getEventItemsProvider(self):
        return self.__eventDisplay.getSelectedEventItemsProvider()

    def refreshEvents(self):
        self.__projectCtx.getEventManager().refresh()
        self.__eventDisplay.refresh()


class _WfguiCockpitSystemModule(WguiCockpitSystemModuleBase):
    _ICON_CODES = dict(WguiCockpitSystemModuleBase._ICON_CODES)
    _ICON_CODES.update(WFGUI_DEFAULT_ICON_CODES)

    def __init__(self):
        super().__init__(WFMOD_STATEMACHINE_APP_MODULE)

    def createContext(self, pctx, pcfg, actions, actx, parent):
        return _WfguiCockpitSystemContext(
            self, pctx, pcfg, actions, actx, parent)

    def getIconForCode(self, code, iconType):
        return _WFGUI_ICONS.get(code) or \
            super().getIconForCode(code, iconType)

    def _getRawIconCode(self, entity):
        return wgui_getClassEntry(entity, self._ICON_CODES)


#: GUI Cockpit system context singleton
WFGUI_COCKPIT_SYSTEM_MODULE = _WfguiCockpitSystemModule()
