# ----------------------------------------------------------------------------
# - wfgui-utils                                                              -
# - wild cruncher: utility functions           frontend / statemachine / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfgui_utils
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Utility functions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.modules.statemachine.db.wfhdb_protocols import *
from wcr_frontend.gui.wgui_utils import wgui_directed_connection_tooltip, \
    wgui_format_tooltip_section


def wfgui_state_tooltip(state):
    """
    Generate the tooltip text for the given state.

    :param state: State to generate tooltip text for
    :type state: WhdbInstanceP | WfhdbStateP
    :return: Tooltip text for the given state
    :rtype: str
    """

    name = state.getName()
    stid = state.getStateId()

    if stid and stid != name:
        ttip = '<b>{} ({})</b>'.format(name, stid)
    else:
        ttip = '<b>{}</b>'.format(name)

    if state.getInfo():
        ttip += '<br/>' + state.getInfo()

    if isinstance(state, WfhdbExecutableP):
        ents = _wfgui_exes_formatter(state.getOnEntryExec())
        if ents:
            ttip += '<hr/><b>On Entry:</b><br/>' + ents
        exts = _wfgui_exes_formatter(state.getOnExitExec())
        if exts:
            ttip += '<hr/><b>On Exit:</b><br/>' + exts

    return ttip


def wfgui_block_state_tooltip(state):
    """
    Generate the tooltip text for the given state.

    :param state: State to generate tooltip text for
    :type state: WfhdbBlockInstance
    :return: Tooltip text for the given state
    :rtype: str
    """
    ttip = wfgui_state_tooltip(state)

    clds = list(state.getBlock().getInstances())
    if clds:
        ttip += wgui_format_tooltip_section(
            'Child-States', (c.getName() for c in clds))

    return ttip


def wfgui_transition_tooltip(transition, sname=None, ename=None):
    """
    Generate the tooltip text for the given transition.

    :param transition: Transition to generate tooltip text for
    :type transition: WfhdbTransition
    :return: Tooltip text for the given transition
    :rtype: str
    """
    ttip = wgui_directed_connection_tooltip(transition, sname, ename)

    if transition.getTriggerEvent():
        ttip += '<br/>Event: {}'.format(transition.getTriggerEvent())
    if transition.getTriggerCondition():
        ttip += '<br/>Condition: {}'.format(transition.getTriggerCondition())

    excs = _wfgui_exes_formatter(transition.getExecutable())
    if excs:
        ttip += '<hr/><b>Execution:</b><br/>' + excs

    return ttip


def _wfgui_exes_formatter(excs, prfx='&nbsp;&nbsp;&nbsp;', dpth=0):
    """
    Format the given list of executables, recursively.

    :param excs: Executables
    :type excs: list[(str, object)]
    :param prfx: Prefix
    :type prfx: str
    :param dpth: Recursion depth, i.e. prefix-multiplier
    :type dpth: int
    :return: Formatted executables
    :rtype: str
    """
    fmtd = []
    """:type: list[str]"""
    for code, _extb in excs:
        extb = _extb
        """:type: str | dict[str, object] | list[(str, dict)]"""
        if code == WFHDB_EXEC_RAISE_KEY:
            fmtd.append('RAISE {}'.format(extb if extb else '?'))
        elif code == WFHDB_EXEC_ASSIGN_KEY:
            fmtd.append('{} = {}'.format(
                extb[WFHDB_EXEC_ASSIGN_LOC_KEY],
                extb[WFHDB_EXEC_ASSIGN_EXPR_KEY]))
        elif code == WFHDB_EXEC_SEND_KEY:
            fmtd.append('SEND {}'.format(extb[WFHDB_EXEC_SEND_EVENT_KEY]))
        elif code == WFHDB_EXEC_INVOKE_KEY:
            fmtd.append('INVOKE {}'.format(extb[WFHDB_EXEC_INVOKE_ID_KEY]))
        elif code == WFHDB_EXEC_SCRIPT_KEY:
            if extb:
                fst, *rst = extb.split()
                fmtd.append('SCRIPT {}{}'.format(fst, ' ...' if rst else ''))
        elif code == WFHDB_EXEC_IF_KEY:
            fst, *rst = extb
            fcnd = fst[WFHDB_EXEC_IF_COND_KEY]
            fexs = fst[WFHDB_EXEC_IF_EXECS_KEY]
            fmtd.append('IF {}<br/>{}'.format(
                fcnd, _wfgui_exes_formatter(fexs, dpth=dpth + 1)))
            for nxt in rst:
                fcnd = nxt[WFHDB_EXEC_IF_COND_KEY]
                fexs = nxt[WFHDB_EXEC_IF_EXECS_KEY]
                fmtd.append('{} {}<br/>{}'.format(
                    'ELSE-IF' if fcnd else 'ELSE',
                    fcnd, _wfgui_exes_formatter(fexs, dpth=dpth + 1)))
        elif code == WFHDB_EXEC_LOG_KEY:
            fmtd.append('LOG {}: {}'.format(
                extb[WFHDB_EXEC_LOG_LABEL_KEY],
                extb[WFHDB_EXEC_LOG_EXPR_KEY]))
    _pfx = prfx * dpth
    return '<br/>'.join(
        '<span style="white-space:nowrap;">{}{}</span>'.format(_pfx, e)
        for e in fmtd)


def wfhdb_get_transition_description(transition):
    """
    Get the textual description for the given transition.

    :param transition: Transition to get name for
    :type transition: WfhdbTransition
    :return: Description for the given transition
    :rtype: str
    """
    name = transition.getName()
    so = transition.getStartObject()
    soName = (so.getName() or so.getTypeName()) if so is not None else 'None'
    do = transition.getEndObject()
    doName = (do.getName() or do.getTypeName()) if do is not None else 'None'

    n = '{so} \u279e {eo}'.format(so=soName, eo=doName)
    if name:
        n = '{0}: {1}'.format(name, n)

    return n
