# ----------------------------------------------------------------------------
# - wfgui-variable-table                                                     -
# - wild cruncher - cockpit: variable dictionary                             -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfgui_variable_table
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher state machine variable dictionary

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt5.QtCore import QAbstractTableModel, Qt, QSortFilterProxyModel
from PyQt5.QtWidgets import QHBoxLayout, QWidget, QVBoxLayout, QPushButton, \
	QLineEdit, QHeaderView, QTableView, QAbstractItemView

from silkan_common.scmn_log import SCMN_LOG_INFO
from wcr_core.modules.statemachine.db.wfhdb_datamodel_support import \
    WFHDB_DM_EXPRESSION_KEY, WFHDB_DM_KIND_KEY, WFHDB_DM_TYPE_KEY

class WfguiVariableDictionary(QWidget):
    def __init__(self, pctx, parent=None):
        """
        :name pctx: Project context
        :type pctx: WcorProjectContextP
        """
        super().__init__(parent)

        #: Project context
        self.__pctx = pctx

        self.__model = _WfguiVariablesTableModel(pctx)
        self.__proxyModel = QSortFilterProxyModel()
        self.__table = QTableView()

        self.__filterF = QLineEdit()
        self.__addBtn = QPushButton('Add')
        self.__delBtn = QPushButton('Delete')

        self._initElements()
        self.setLayout(self._createLayout())

    def _createLayout(self):
        """
        Create new layout.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.setSpacing(3)
        layout.addWidget(self.__table)

        layout.addWidget(self.__filterF)
        btnLayout = QHBoxLayout()
        btnLayout.addStretch(1)
        btnLayout.addWidget(self.__addBtn)
        btnLayout.addWidget(self.__delBtn)
        layout.addLayout(btnLayout)
        return layout

    def _initElements(self):
        """
        Initialize UI elements.
        """
        self.__proxyModel.setSourceModel(self.__model)
        self.__proxyModel.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.__proxyModel.setSortCaseSensitivity(Qt.CaseInsensitive)

        self.__table.setModel(self.__proxyModel)
        self.__table.horizontalHeader().setSectionResizeMode(0, QHeaderView.Stretch)
        self.__table.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.__table.horizontalHeader().setSectionResizeMode(2, QHeaderView.Stretch)
        self.__table.setEditTriggers(
            QAbstractItemView.DoubleClicked |
            QAbstractItemView.EditKeyPressed |
            QAbstractItemView.AnyKeyPressed)
        self.__table.setSelectionMode(QAbstractItemView.SingleSelection)
        self.__table.setSortingEnabled(True)
        self.__table.verticalHeader().setVisible(False)

        self.__filterF.setPlaceholderText('Filter')
        self.__filterF.textChanged.connect(
            self.__proxyModel.setFilterFixedString)

        self.__addBtn.clicked.connect(self._addRow)
        self.__delBtn.clicked.connect(self._deleteRow)

    def _deleteRow(self):
        """
        Delete the variable represented by the currently selected row, if any.
        """
        slct = self.__table.selectedIndexes()
        if slct:
            row = slct[0].row()
            self.__proxyModel.removeRows(row, 1)

    def _addRow(self):
        """
        Add a row for a new variable.
        """
        row = self.__proxyModel.rowCount()
        self.__proxyModel.insertRow(row)
        idx = self.__proxyModel.index(row, 0)
        self.__table.setCurrentIndex(idx)
        self.__table.edit(idx)


class _WfguiVariablesTableModel(QAbstractTableModel):
    """
    Variables table model.
    """

    #: Column headers
    _HEADERS = ['Name', 'Type', 'Kind', 'Default']

    def __init__(self, pctx, parent=None):
        """
        :param pctx:
        :type pctx: WcorProjectContextP
        :param parent: Parent object
        """
        super().__init__(parent)

        #: Project context
        self.__pctx = pctx

        #: Variables [(Name, Type, Kind, Default-value)]
        self.__vlist = []
        #: Set of all variable names
        self.__names = set()

        self._update()

    @property
    def _topLevelBlock(self):
        """
        Get the top-level block.

        :return: Top-level block
        :rtype: WfhdbTopBlock
        """
        return self.__pctx.getProjectDB().getTopHierarchy()

    def _update(self):
        """
        Update the model from the top-level block and notify observers.
        """
        self.beginResetModel()
        vdict = self._topLevelBlock.getDataModel(self._topLevelBlock.rawElement)
        self.__vlist = [[n,
                         d.get(WFHDB_DM_TYPE_KEY),
                         d.get(WFHDB_DM_KIND_KEY),
                         d.get(WFHDB_DM_EXPRESSION_KEY)]
                        for n, d in vdict.items()]
        self.__vlist.sort(key=lambda i: i[0])
        self.__names = set(vdict.keys())
        self.endResetModel()

    def rowCount(self, *_):
        return len(self.__vlist)

    def columnCount(self, *_):
        return 4

    def data(self, index, role=None):
        if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
            return self.__vlist[index.row()][index.column()] or ''
        return None

    def setData(self, index, data, role=None):
        if role != Qt.EditRole:
            return False
        row = index.row()
        col = index.column()
        item = self.__vlist[row]
        odat = item[col]

        if odat == data:
            return True

        if col == 0:
            if data in self.__names:
                self.__pctx.addMessage(
                    'Variable with name "{}" already exists.'.format(data),
                    'cockpit', SCMN_LOG_INFO)
                return False
            if odat:
                self._topLevelBlock.removeFromDataModel(self._topLevelBlock.rawElement, odat)
                self.__names.remove(odat)
            self.__names.add(data)

        item[col] = data
        self._topLevelBlock.addToDataModel(self._topLevelBlock.rawElement,
            name=item[0], dtyp=item[1], kind=item[2], expr=item[3])
        self.__vlist[index.row()] = item
        return True

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal:
                return self._HEADERS[section]
            else:
                return str(section)
        return None

    def flags(self, index):
        return super().flags(index) | Qt.ItemIsEditable

    # noinspection PyUnusedLocal
    def insertRow(self, row, parent=None, *args, **kwargs):
        return self.insertRows(row, 1, parent)

    # noinspection PyUnusedLocal
    def insertRows(self, row, count, parent=None, *args, **kwargs):
        self.beginInsertRows(parent, row, row + count - 1)
        for x in range(count):
            self.__vlist.append([None, None, None, None])
        self.endInsertRows()
        return True

    def removeRows(self, row, count, parent=None, *args, **kwargs):
        self.beginRemoveRows(parent, row, row + count - 1)
        for r in range(row, row + count):
            name = self.__vlist[row][0]
            self._topLevelBlock.removeFromDataModel(self._topLevelBlock.rawElement, name)
            self.__names.remove(name)
        self.__vlist[row:(row + count)] = []
        self.endRemoveRows()
        return True
