# ----------------------------------------------------------------------------
# - wigui-scene                                                              -
# - wild cruncher - cockpit : graphic scene widget                           -
# ----------------------------------------------------------------------------
# - author    (c) 2017-2017 achraf mahti                                     -
# - copyright (c) 2017-2017 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wigui_scene
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher wigui graphics scene editor

.. moduleauthor:: achraf mahti <mahti@silkan.net>
"""

from operator import methodcaller
from wcr_core.plugin.wplg_library import WplgOperatorLibrary
from wcr_core.modules.sys_map.db.wchdb_Host_lure_instance \
    import WchdbHostLureInstance
from wcr_core.modules.sys_map.db.wchdb_Switch_lure_instance import \
    WchdbSwitchLureInstance
from wcr_core.modules.sys_map.db.wchdb_standard_operators import *
from wcr_frontend import showErrorWindow
from wcr_frontend.gui.wgui_scene import WguiGraphicsScene
from wcr_frontend.gui.wgui_scene_item import *

_WCGUI_LIB_NAME = 'Modules'
_WCGUI_SCENE_MODULE_HOLDERS = [WCHDB_HOST_LURE_CODE,
                               WCHDB_SWITCH_LURE_CODE]


class WcguiGraphicsScene(WguiGraphicsScene):
    """
    The WcguiGraphicsScene class is the class responsible to manage
    specific scene action in Mapper view that WguiGraphicsScene don't handle
    """

    ModuleUnbound = pyqtSignal(str)

    def __init__(self, pjdb, config, actions, pctx, cctx, actx, mod,
                 parent=None):

        # todo: initialise the right library to consider as cards
        super().__init__(pjdb, config, actions, pctx, cctx, actx, parent=parent)
        self.__projectCtx = pctx
        self.__top = pctx.getProjectDB().getTopHierarchy()
        self.__cctx = cctx
        self.__actx = actx
        self.__mapp = pctx.getMappingManager()
        self.__listOfBoundModule = {}
        self.projOperators = {}
        mod.emptyLibrary()
        self.projModule = mod
        self.projectName = self.projectCtx.getProjectDB().getPath()
        self.importModules()
        self.instanciatedOperators = {}
        self.projEditor = None
        self.ModuleUnbound.connect(self.enableModuleFromLibrary)
        self.updateModuleBoundFromHosts()

    def release(self):
        pass

    def importModules(self, projectEditor=None):
        self.projEditor = projectEditor
        modules = self.__mapp.getAllModuleDesc()

        for mod in modules:
            self.projOperators.setdefault(self.projectName, {})[mod.getName()] = mod
        if self.projOperators.get(self.projectName, None):

            lib = WplgOperatorLibrary(
                code=_WCGUI_LIB_NAME, module='icd', name=_WCGUI_LIB_NAME,
                description='modules created by user',
                operators=[o for o in sorted(self.projOperators[self.projectName].values(),
                                             key=methodcaller('getName'))])
            self.__actx.addMessage('New modules imported from project file',
                                   'Mapping View')

            self.projModule.addLibrary(lib)
            boundModules = self.getModulesBound()
            for modName in boundModules:
                self.disableModuleFromLibrary(modName)

    @property
    def producerTable(self):
        return self.__cctx.getProducerTable()

    @property
    def consumerTable(self):
        return self.__cctx.getConsumerTable()

    @property
    def top(self):
        return self.__top

    @property
    def projectCtx(self):
        return self.__projectCtx

    def createProducerBind(self, pos, routes):
        """
        Handle the fact that we want to bind routes to a producer port

        :param pos: the function position
        :type pos: QPointF
        :param routes: list of producer routes to bind
        :type routes: [WvhdbVLinkInstance]
        """
        # select the connector
        inst, conn = self._getConnectionPointData(pos)
        if conn is None:
            showErrorWindow('Route binding error',
                            'Route need to be bound to a Port.',
                            None)
            return False
        portName = '{}.{}.{}'.format(inst.getName(), conn.getCard(), conn.getName())

        hconn = self.__mapp.getPhyPortByUUID(conn.getUUIDSource())
        for route in routes:
            self.producerTable._addEntry((portName, hconn),
                                         (route.getName(), route))
            self.producerTable.bindPortProd((portName, hconn),
                                            (route.getName(), route))
        return True

    def createConsumerBind(self, pos, routes):
        """
        Handle the fact that we want to bind a route to a consumer port

        :param pos: the function position
        :type pos: QPointF
        :param routes: consumed route to bind
        :type routes: WvhdbVLinkInstance
        """
        # select the connector
        inst, conn = self._getConnectionPointData(pos)
        if conn is None:
            showErrorWindow('Route binding error',
                            'No port found!\nRoute need to be bound to a Port.',
                            None)
            return False
        portName = '{}.{}.{}'.format(inst.getName(), conn.getCard(), conn.getName())

        hconn = self.__mapp.getPhyPortByUUID(conn.getUUIDSource())
        for route in routes:
            self.consumerTable._addEntry((portName, hconn),
                                         (route.getName(), route))
            self.consumerTable.bindPorts((portName, hconn),
                                            (route.getName(), route))
        return True

    def createFunctionItem(self, pos, oper=None):
        """
        Handle the fact tha we want to add a card directly to an host
        and not to create it in the scene view.
        Else add a function item by position it not a card

        :param pos: the function position
        :type pos: QPointF
        """
        # get the function by operator
        # select the operator by mode
        if oper is None:
            oper = self.getOperatorMode()
        if oper is None: return
        inst, conn = self._getConnectionPointData(pos)
        if inst is None:
            showErrorWindow('Module Error',
                            'Module needs to be bind to a Card.',
                            None)
            return None
        instCode = inst.getOperator().getCode()

        if instCode not in _WCGUI_SCENE_MODULE_HOLDERS:
            return None

        boundModules = self.getModulesBound()
        if oper.getName() in boundModules:
            showErrorWindow('Module Error',
                            'Module {} is already bound to host {}.'.format(oper.getName(),
                                                            boundModules[oper.getName()]),
                            None)
            return None
        self.addModuleToHost(inst, oper.getName())

        self.addModuleBound(oper.getName(), inst.getName())
        self.disableModuleFromLibrary(oper.getName())

    def addFunctionItem(self, func):
        """
        Create a function graphics item for the given function database
        object.

        :param func: The function object
        :type func: WhdbOperatorInstanceP
        :return: The newly created connection scene item
        :rtype: WguiOperatorInstanceSceneItem
        """
        if func is None: return None
        if 'wcr.vlink' in func.getOperator().getCode():
            return None
        if func.getZ() is None:
            func.setZ(self.getTopZValue() + 1)

        item = self.sitmF.createOperatorInstanceSceneItem(
            func, self.viewContext)
        self.addItem(item)
        return item

    def addModuleToHost(self, instance, modName):
        instance.addModule(modName)
        cards = instance.getCards()
        if len(cards) == 1:
            card = cards[0]
            self.__top.addModuleCardBind(instance.getName(), modName,
                                         card, period='1')

    def disableModuleFromLibrary(self, moduleName):
        if self.projEditor:
            # print('disableModuleFromLibrary')
            for but in self.projEditor.buttonGroup.buttons():
                if but.toolTip() == moduleName:
                    but.setEnabled(False)

    def enableModuleFromLibrary(self, moduleName):
        if self.projEditor:
            # print('enableModuleFromLibrary')
            for but in self.projEditor.buttonGroup.buttons():
                if but.toolTip() == moduleName:
                    but.setEnabled(True)

    def updateModuleBoundFromHosts(self):
        sceneEqps = self.getSceneItems()
        cardModuleBinding = self.__top.getModuleCardBind()
        for eqpName in cardModuleBinding:
            for moduleName, _, _ in cardModuleBinding[eqpName]:
                for sceneEqp in sceneEqps:
                    eqp = sceneEqp.getDataObject()
                    if eqp.getName() == eqpName:
                        eqp.addModule(moduleName)
        for sceneEqp in sceneEqps:
            eqp = sceneEqp.getDataObject()
            if isinstance(eqp, WchdbHostLureInstance) or isinstance(eqp, WchdbSwitchLureInstance) :
                modules = eqp.getModules()
                for modName in modules:
                    self.addModuleBound(modName,
                                        eqp.getName())

    def addModuleBound(self, modName, eqpName):
        if modName not in self.__listOfBoundModule:
            self.__listOfBoundModule[modName] = eqpName

    def removeModuleBound(self, modName):
        if modName in self.__listOfBoundModule:
            del self.__listOfBoundModule[modName]

    def getModulesBound(self):
        return self.__listOfBoundModule

    def resetModulesBound(self):
        self.__listOfBoundModule = {}
