# ----------------------------------------------------------------------------
# - wcgui-vlink-cons-binding-table                                           -
# - wild cruncher:sys_map ports bindings widget                              -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcgui_vlink_cons_binding_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan sys_map ports bindings widget

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""

from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QWidget, QVBoxLayout, \
	QPushButton, QTableView, QHeaderView, QItemDelegate, QAbstractItemView
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_core.modules.hardware.db.whhdb_Cetrac_switch_instance import \
WhhdbCetracSwitchInstance
from wcr_core.modules.hardware.db.whhdb_kazymir_switch_instance import \
WhhdbKazymirSwitchInstance
from wcr_core.modules.sys_map.plugin.wcplg_config_generator_params import \
    WcplgGetAllRouteOfSameCodeFromPort, WcplgGetAllPortOfSameCodeFromRoute
from silkan_common.scmn_utils import unique


class WcguiVLConsBindingTable(QWidget):
    """
    Virtual link consumers port binding to physical port table widget.
    """

    def __init__(self, pctx, scene, filter, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project context
        self.__pctx = pctx
        #: Graphical scene
        self.__scene = scene
        #: View filter
        self.__filter = filter

        self.__formatF = lambda o, c: self.formatF(o, c)
        #: Bus binding table model
        self.__bndgTblM = _WcguiConsBindingTableModel(self.__formatF,
                                                        parent=self)

        #: Bus binding table model
        self.__pportF = lambda: self.getphisicalports()
        self.__vlF = lambda : self.getNonConsBindedVLink()
        self.__bndgPortsDlgt = _WcguiPortConsBindDelegate(self.__pportF,
                                                          self.__vlF,
                                                          self.__formatF,
                                                          self.getModel())

        """:type: (WphdbPhyBusInstance) -> list[WshdbBusInstance]"""

        #: Bus binding table widget
        self.__bndgTblW = QTableView()

        #: Add entry button
        self.__addButton = QPushButton('Add')
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')

        self._initElements()
        self.setLayout(self._createLayout())
        self._loadFromData()

    def getModel(self):
        return self.__bndgTblM

    def getScene(self):
        return self.__scene

    def _initElements(self):
        self.__bndgTblW.setModel(self.__bndgTblM)
        self.__bndgTblW.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.Stretch)
        self.__bndgTblW.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch)
        self.__bndgTblW.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.__bndgTblW.setItemDelegate(self.__bndgPortsDlgt)

        self.__addButton.clicked.connect(self._addEntry)
        self.__deleteButton.clicked.connect(self._deleteSelEntries)
        pass

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)

        layout.addWidget(self.__bndgTblW)

        btnLayout = QHBoxLayout()

        btnLayout.addWidget(self.__addButton)
        btnLayout.addWidget(self.__deleteButton)
        layout.addLayout(btnLayout)

        return layout

    def reloadFromData(self):
        """
        remove all row of table and refill it with element from project data
        """
        self.__bndgTblM.deleteRows(range(0,self.__bndgTblM.rowCount()),
                                   Unbind=False)
        self._loadFromData()

    def _loadFromData(self):
        """
        Fill the table from the project's data
        """
        map=self.__pctx.getMappingManager()

        pportperml=[]
        for h in map.getHosts():
            if self.__filter.uuidpassfilter(h.getUUID()):
                for p in h.getConnectionPoints():
                    pportperml.append(p.getUUID())
        sportperml = []
        for sw in self.__pctx.getMappingManager().getSwitchs():
            if self.__filter.uuidpassfilter(sw.getUUID()):
                if isinstance(sw, WhhdbCetracSwitchInstance):
                        # isinstance(sw, WhhdbKazymirSwitchInstance):
                    for conIo in sw.getAllIoConnectionPoints():
                        sportperml.append(conIo.getUUID())
        for bnd in self.__pctx.getProjectDB().getTopHierarchy(). \
                getConsBinds():
            if bnd[0] in pportperml:
                port = map.getPhyPortByUUID(bnd[0])
                cardName = port.getCard()
                host = map.getHostFromPortUUID(bnd[0])
                self.__bndgTblM.addRow((
                    self.formatPhisicalPortName(port, cardName, host),
                    map.getPhyPortByUUID(bnd[0])),
                    (map.getVlinkNameByUUID(bnd[1]),
                     map.getVlinkByUUID(bnd[1])))
            elif bnd[0] in sportperml:
                port = map.getPhyPortByUUID(bnd[0])
                cardName = port.getCard()
                host = map.getSwitchByPortUUID(bnd[0])
                self.__bndgTblM.addRow((
                    self.formatPhisicalPortName(port, cardName, host),
                    map.getPhyPortByUUID(bnd[0])),
                    (map.getVlinkNameByUUID(bnd[1]),
                     map.getVlinkByUUID(bnd[1])))

    def _addEntry(self, port=None, vl=None):
        """
        Add entry in the model
        """
        self.__bndgTblM.addRow(port, vl)

    def getBindings(self):
        """
            Get entry in the model
        """
        return self.__bndgTblM.getDataRows()

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.__bndgTblW.selectedIndexes()
        """:type: list[QModelIndex]"""

        self.__bndgTblM.deleteRows([idx.row() for idx in selIdx])

    def getphisicalports(self):
        conl=[]
        for h in self.__pctx.getMappingManager().getHosts():
            if self.__filter.uuidpassfilter(h.getUUID()):
                for con in h.getConnectionPoints():
                    if con.getActive() == "Off":
                        continue
                    cardName = con.getIoCard()
                    conl.append((self.formatPhisicalPortName(con, cardName, h),
                                 con))
        # get Io connectors from switch
        for sw in self.__pctx.getMappingManager().getSwitchs():
            if self.__filter.uuidpassfilter(sw.getUUID()):
                if isinstance(sw, WhhdbCetracSwitchInstance) or \
                        isinstance(sw, WhhdbKazymirSwitchInstance):
                    for conIo in sw.getAllIoConnectionPoints():
                        if conIo.getActive() == "Off":
                            continue
                        cardName = conIo.getIoCard()
                        conl.append((self.formatPhisicalPortName(conIo, cardName, sw),
                                     conIo))
        return conl

    def getNonConsBindedVLink(self):
        """
        return the list of virtual links how are not binded to a phisical port
        as procuders
        """
        mngr = self.__pctx.getMappingManager()
        vlinkl = mngr.getVlList()
        if self.__scene.top.getVLinkBlock():
            autoVls = [(inst.getName(), inst) for inst in self.__scene.top.getVLinkBlock().getInstances()]
            vlinkl.extend(autoVls)
            vlinkl = unique(vlinkl)
        return vlinkl

    def formatPhisicalPortName(self, port, cardName, host):
        return host.getName() + "." + cardName + "." + port.getName()

    def getpossibleVirtiallink(self, phyportUUID):
        """
        return the list ot the virtual link how still can be bind to consumer
        """
        virtuallinkl = self.__pctx.getMappingManager().getVlList()
        """[(name,vl)]"""

        vliter={}
        existbind=[]
        #: counting number of consumer and saving all ready binded to port vl
        for pp, vl in self.__pctx.getProjectDB().getTopHierarchy(). \
                getConsBinds():
            if vl in vliter.keys():
                vliter[vl]+=1
            else:
                vliter[vl]=1
            if str(pp) == str(phyportUUID):
                existbind.append(vl)

        rmvllist=[]
        #: get all  vl how cant accept any more consumer
        for vln, vl in virtuallinkl:
            if vl.getUUID() in existbind:
                continue
            if not self.__pctx.getMappingManager().\
                    vlcanBeBinded(vl, vliter[vl.getUUID()] if \
                            vl.getUUID() in vliter.keys() else 0):
                rmvllist.append((vln, vl))
                continue
        #: removing from the proposal list all vl how have they consumer fully
        # binded
        for rmvl in rmvllist:
            virtuallinkl.remove(rmvl)
        return virtuallinkl

    def formatF(self, obj, col):
        if not obj:
            return ''
        elif col == 0:
            return obj[0] # obj.getName()
        elif col == 1:
            return obj[0]
        else:
            return 'No formating available'

    def unbindPorts(self, pport, vlinkd):
        """
        unbind the host and the module for the topblock
        :param : physical port instance
        :type : WhhdbHardwareXPoint
        :param : virtual link instance
        :type : WvhdbVLinkInstance
        """
        self.__pctx.getProjectDB().getTopHierarchy().removeConsBind(pport[1].getUUID(), vlinkd[1].getUUID())

    def bindPorts(self, pport, vlinkd):
        """
        bind the host and the module for the topblock
        :param pport: physical port instance
        :type pport: WhhdbHardwareXPoint
        :param vlinkd: Logic port instance
        :type vlinkd: WvhdbVLinkInstance
        """
        for pp, vl in self.__pctx.getProjectDB().getTopHierarchy().getConsBinds():
            if pp == pport[1].getUUID() and vl == vlinkd[1].getUUID():
                return
                # raise ValueError("Bind already exists")
        self.__pctx.getProjectDB().getTopHierarchy(). \
            addConsBind(pport[1].getUUID(), vlinkd[1].getUUID())



class _WcguiConsBindingTableModel(QAbstractTableModel):
    """
    Bus Binding table model.
    """

    TITLE_VLINK = 'Traffic Route'
    TITLE_PPORT = 'Phys Port'

    HEADERS = [TITLE_PPORT, TITLE_VLINK]

    def __init__(self, formatfunc, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)

        self.__parent=parent
        #: Entries as list of tuples of
        #  physical Port (pport) and logical Port (vl)
        self.__entries = []
        """:type: list[]"""

        self.__formatF=formatfunc
        self.rowsInserted.connect(self.dataHasChanged)
        self.rowsRemoved.connect(self.dataHasChanged)
        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, pport=None, vl=None):
        """
        Add a row at the end of the list (append)

        :param pport: Message to be set
        :type pport: WhhdbHardwareXPoint | None
        :param vl: Data to be set
        :type vl: WvhdbVLinkInstance | None
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append((pport, vl))

        if pport is not None and vl is not None:
            # call binding function of message
            pass


        self.endInsertRows()
        self._update()

    def getDataRows(self):
        return self.__entries

    def deleteRows(self, rowIndices, Unbind=True):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row, Unbind=Unbind)

    def deleteRow(self, rowIndex, Unbind=True):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        pport, lport = self.__entries[rowIndex]
        if pport is not None and lport is not None and Unbind:
            try:
                self.__parent.unbindPorts(pport, lport)
                pass
            except ValueError:
                pass

        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def getVlRowIndex(self, vlUUID):
        indexes = []
        for rowIndex in range(0, self.rowCount()):
            pport, vlinkd = self.__entries[rowIndex]
            if vlinkd:
                if vlUUID == vlinkd[1].getUUID():
                    indexes.append(rowIndex)
        indexes.sort(reverse=True)
        return indexes

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column < 2:
            info = self.__entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return self.__formatF(info, column)
            elif role == Qt.UserRole:
                return info
        return None

    def setData(self, index, data, role=None):
        """
        :type index: QModelIndex
        :type data: object
        """
        assert role == Qt.UserRole
        self._setDataEntry(index, data)

    def _setDataEntry(self, index, data):
        pport, lport = self.__entries[index.row()]
        if pport and lport :
            # a binding could have been done by that line before
            # it have to be deleted.
            try:
                #remove binding function
                self.__parent.unbindPorts(pport, lport)
                pass
            except ValueError:
                pass

        nl = ()
        if index.column() == 0:
            nl = (data, lport)
        elif index.column() == 1:
            nl = (pport, data)
        self.__entries[index.row()] = nl
        pport, lport = self.__entries[index.row()]
        if pport and lport:
            #: the line is complet a binding have to be done
            try:
                #binding function
                self.__parent.bindPorts(pport, lport)
                pass
            except ValueError:
                pass

    def flags(self, index):
        flgs = super().flags(index)
        #: set table as editable
        flgs |= Qt.ItemIsEditable
        return flgs

    def dataHasChanged(self):
        self.__parent.getScene().sceneModified.emit()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return 2


class _WcguiPortConsBindDelegate(QItemDelegate):
    """
    Delegate for physical and logical ports
    """

    def __init__(self, pportListF, vlListF, formatF, model, parent=None):
        """
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)
        self.formatF = formatF
        self.__phypListPrv = pportListF
        self.__vlListPrv = vlListF
        self.__model = model

    def createEditor(self, parent, option, index):
        """
        Create a combobox containing adequate choices for the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """

        def save(*_):
            self.setModelData(editor, index.model(), index)

        editor = QComboBox(parent)
        row = index.row()
        col = index.column()
        busPrvF= []
        # Iterate through the messages using the adequate provider
        # and add items to the combo box
        editor.addItem(None, None)
        entries = self.__model._WcguiConsBindingTableModel__entries
        # the available port and route depends partially of selected element
        if col == 0:
            routeData = entries[row][1]
            if not routeData:
                busPrvF = self.__phypListPrv
            else:
                busPrvF = lambda: WcplgGetAllPortOfSameCodeFromRoute(routeData[1], self.__phypListPrv())
        elif col == 1:
            portData = entries[row][0]
            if not portData:
                busPrvF = self.__vlListPrv
            else:
                busPrvF = lambda: WcplgGetAllRouteOfSameCodeFromPort(portData[1], self.__vlListPrv())
            cdata=index.model().data(index.model().index(index.row(), 1),
                                     role=Qt.UserRole)
            if not cdata is None:
                editor.addItem(self.formatF(cdata,1),
                               cdata)
        for elem in busPrvF():
            editor.addItem(self.formatF(elem, col), elem)

        editor.currentIndexChanged.connect(save)

        return editor

    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)

    def setEditorData(self, editor, index):
        """
        :type editor: QComboBox
        :type index: QModelIndex
        """
        data = index.model().data(index, Qt.UserRole)
        for index in range(editor.count()):
            if editor.itemData(index, Qt.UserRole) == data:
                editor.setCurrentIndex(index)
                break
        else:
            editor.setCurrentIndex(0)
