# ----------------------------------------------------------------------------
# - wcgui-virtual-link-table                                                 -
# - wild cruncher:sys_map virtual link widget                                -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcgui_virtual_link_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan sys_map virtuallinks widget

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""

import os
from PyQt4.QtCore import QAbstractTableModel, Qt
from PyQt4.QtGui import QWidget, QTableView, QHeaderView, \
    QVBoxLayout, QHBoxLayout, QPushButton
from wcr_frontend.gui.wgui_utils import wgui_getOpenFile
from wcr_core.project.wprj_xmlio import WPRJ_VLINK_SILKAN_XML_MIME_TYPE
from wcr_core.common.wcmn_file_handler import WcmnFile
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_MODULES_IO_NAMES
from wcr_core.modules.sys_map.plugin.wcplg_config_generator_params import \
    VL_NETWORK_TYPE, HARD_NETWORK_TYPE


class WcguiVLinkViewTable(QWidget):
    """
    Virtual link presentation table widget.
    """

    def __init__(self, pctx, scene, pBdtbl, cBdtbl, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param pBdtbl: producer VL binding table
        :type pBdtbl: WcguiVLProdBindingTable
        :param cBdtbl: consumer VL binding table
        :type cBdtbl: WcguiVLConsBindingTable
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project context
        self.__pctx = pctx
        self.__top = self.__pctx.getProjectDB().getTopHierarchy()
        #: Graphical scene
        self.__scene = scene
        #: Update binding tables removing invalids binds
        self.__PropagateUpdateFunc = None
        # producer binding table
        self.__pBdtbl = pBdtbl
        # consumer binding table
        self.__cBdtbl = cBdtbl


        #: Bus binding table model
        self.__modTblM = _WcguiVLinkTableModel()


        """:type: (WphdbPhyBusInstance) -> list[WshdbBusInstance]"""

        #: Bus binding table widget
        self.__modTblW = QTableView()

        #: Add entry button
        self.__impButton = QPushButton('Import')
        #: update entry button
        self.__updButton = QPushButton('Update')
        #: delete entry button
        self.__delButton = QPushButton('Delete')
        self.__autoBindButton = QPushButton('Auto Bind')
        self._initElements()
        self.setLayout(self._createLayout())
        self.updateTable()

    def _initElements(self):
        self.__modTblW.setModel(self.__modTblM)
        self.__modTblW.horizontalHeader().setResizeMode(
            0, QHeaderView.ResizeToContents)
        self.__modTblW.horizontalHeader().setResizeMode(
            1, QHeaderView.Stretch)
        self.__impButton.clicked.connect(self._importVirtualLink)
        self.__updButton.clicked.connect(self.updateTable)
        self.__delButton.clicked.connect(self.deleteVL)
        self.__autoBindButton.clicked.connect(self.autoBindFunction)

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setMargin(0)

        layout.addWidget(self.__modTblW)

        btnLayout = QHBoxLayout()

        btnLayout.addWidget(self.__impButton)
        btnLayout.addWidget(self.__updButton)
        btnLayout.addWidget(self.__delButton)
        btnLayout.addWidget(self.__autoBindButton)
        layout.addLayout(btnLayout)

        return layout

    def ResetWithData(self):
        """
        Fill the table from the project's data
        """
        self.__modTblM.deleteAll()
        map=self.__pctx.getMappingManager()
        vlinks=map.getAllVirtualLinks()
        if vlinks:
            l = 0
            for vl in vlinks:
                messages = map.getMessagesOfVlink(vl[1].getUUID())
                self.__modTblM.addelem((vl, messages), l, 0)
                l += 1
        pass

    def _importVirtualLink(self):
        """
        Import a new virtual link file and fill the mapper data and the table
        with is data
        """
        # Selection d'un fichier
        selected_path = wgui_getOpenFile("Import Tr", self,
                                         WPRJ_VLINK_SILKAN_XML_MIME_TYPE)
        if selected_path:
            if self.__top.addVLinkFile(WcmnFile(selected_path)):
                self.__pctx.updateVL(selected_path)
            self.ResetWithData()
        self._unbindinvalid()

    def updateTable(self):
        """
        Update the table
        """
        self.__pctx.getMappingManager().resetVLink()
        for vlf in self.__top.getVLinkFiles():
            self.__pctx.updateVL(vlf.path)
        self.ResetWithData()
        self._unbindinvalid()

    def deleteVL(self):
        """
        Delete a Traffic route
        """
        selIdx = self.__modTblW.selectedIndexes()
        for idx in sorted(selIdx, reverse=True):
            row = idx.row()
            # always get info of column 1 to check vlname
            idx = idx.sibling(row, 0)
            manager = self.__pctx.getMappingManager()
            vLinks = manager.virtualLink
            entries = self.__modTblM.getEntries()
            (vLinkName, vlInst), msgs = entries[row]
            assert vLinkName == idx.data()
            vlUuid = vlInst.getUUID()
            # unbind producer unbind vl
            producerModel = self.__pBdtbl.getModel()
            rowIndex = producerModel.getVlRowIndex(vlUuid)
            if rowIndex is not None:
                producerModel.deleteRow(rowIndex)
            # unbind consumer deleted vl
            consumerModel = self.__cBdtbl.getModel()
            rowIndexes = consumerModel.getVlRowIndex(vlUuid)
            if len(rowIndexes) > 0:
                for rowIndex in rowIndexes:
                    consumerModel.deleteRow(rowIndex)
            manager.removeVLink(vlInst)
            del vLinks[vlUuid]
            self.__modTblM.deleteRow(row)

    def autoBindFunction(self):
        moduleBindings = self.__top.getModuleCardBind()
        map = self.__pctx.getMappingManager()
        messageBindings = map.getVlMessageBinding()
        hosts = map.getHosts()
        allVls = {vlInst.getUUID(): vlInst
                  for vlName, vlInst in map.getAllVirtualLinks()}

        vlProdConsBinding = {}
        for vluuid, msg in messageBindings:
            vlProdConsBinding[vluuid] = {'C': [], 'P': None}
            consumerEqpCard = []
            for glbMsg in msg:
                for lclMsg, consprod in msg[glbMsg]:
                    # find producer
                    if consprod == 'P':
                        producerModule = lclMsg.split('.')[0]
                        producerEqpCard = self.__top.getEqpCardFromModule(producerModule)
                        if not vlProdConsBinding[vluuid]['P']:
                            vlProdConsBinding[vluuid]['P'] = producerEqpCard
                        else:
                            if vlProdConsBinding[vluuid]['P'] != producerEqpCard:
                                raise ValueError('At least 2 messages have producer bound '
                                                 'to different equipment!')
                    elif consprod == 'C':
                        consumerModule = lclMsg.split('.')[0]
                        consumerEqpCard.append(self.__top.getEqpCardFromModule(consumerModule))
            vlProdConsBinding[vluuid]['C'].extend(consumerEqpCard)

        for vluuid in vlProdConsBinding:
            vlInst = allVls[vluuid]
            vlName = vlInst.getName()
            allPorts = self.__pBdtbl.getAvailPorts()
            producer = vlProdConsBinding[vluuid]['P']
            for portName, portObj in allPorts():
                if not producer:
                    continue
                if producer in portName:
                    vlCode = vlInst.getVLType()
                    portCode = portObj.getCode()
                    hasSameCode = VL_NETWORK_TYPE[vlCode] == HARD_NETWORK_TYPE[portCode]
                    isConnected = self.__top.checkportIsConnected(portObj.getUUID())
                    existingRows = [(portName, vlName)
                                    for((portName, _), (vlName, _)) in self.__pBdtbl.getBindings()]
                    if hasSameCode:
                        if portObj.getCode() in WHHDB_MODULES_IO_NAMES:
                            if (portName, vlName) not in existingRows:
                                # print('For Vl {} the port {} was added as Io producer.'.format(vlName, portName))
                                self.__pBdtbl._addEntry((portName, portObj), (vlName, vlInst))
                                self.__pBdtbl.bindPortProd((portName, portObj), (vlName, vlInst))

                            break
                        else:
                            if isConnected:
                                if (portName, vlName) not in existingRows:
                                    # print('For Vl {} the port {} was added as producer.'.format(vlName, portName))
                                    self.__pBdtbl._addEntry((portName, portObj), (vlName, vlInst))
                                    self.__pBdtbl.bindPortProd((portName, portObj), (vlName, vlInst))
                                break
            consumers = vlProdConsBinding[vluuid]['C']
            for consumer in consumers:
                if not consumer:
                    continue
                for portName, portObj in allPorts():
                    if consumer in portName:
                        vlCode = vlInst.getVLType()
                        portCode = portObj.getCode()
                        hasSameCode = VL_NETWORK_TYPE[vlCode] == HARD_NETWORK_TYPE[portCode]
                        isConnected = self.__top.checkportIsConnected(portObj.getUUID())
                        if hasSameCode:
                            existingRows = [(portName, vlName)
                                            for ((portName, _), (vlName, _)) in self.__cBdtbl.getBindings()]
                            if portObj.getCode() in WHHDB_MODULES_IO_NAMES:
                                if (portName, vlName) not in existingRows:
                                    self.__cBdtbl._addEntry((portName, portObj), (vlName, vlInst))
                                    self.__cBdtbl.bindPorts((portName, portObj), (vlName, vlInst))
                                    # print('For Vl {} the port {} was added as Io consumer.'.format(vlName, portName))
                                break
                            else:
                                if isConnected:
                                    if (portName, vlName) not in existingRows:
                                        self.__cBdtbl._addEntry((portName, portObj), (vlName, vlInst))
                                        self.__cBdtbl.bindPorts((portName, portObj), (vlName, vlInst))
                                        # print('For Vl {} the port {} was added as consumer.'.format(vlName, portName))
                                    break

    def _unbindinvalid(self):
        """
        inbind invalide physical port/virtual link and connections
        """
        self.__pctx.unbindInvalidVLinkBinding()
        if not self.__PropagateUpdateFunc is None:
            self.__PropagateUpdateFunc()

    def setPropagationFunction(self, propfunc):
        """
        give logic table a function to propagate a module update effect on
        other tables
        """
        self.__PropagateUpdateFunc = propfunc


class _WcguiVLinkTableModel(QAbstractTableModel):
    """
    Bus Binding table model.
    """

    TITLE_VL = 'Traffic Routes'
    MESSAGES = 'Messages'
    HEADERS = [TITLE_VL, MESSAGES]

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)


        #: Entries as list of tuples of
        #  physical Port (pport) and logical Port (lport)
        self.__entries = []
        """:type: list[]"""

        self._update()

    def getEntries(self):
        return self.__entries

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, vl, messages):
        """
        Add a row at the end of the list (append)

        :param vl: traffic route name
        :type vl: | string
        :param messages: Messages to be set
        :type messages: | [string]
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())
        if len(messages) == 0:
            strMsg = 'No messages bound'
        elif len(messages) > 1:
            strMsg = '[{}]'.format(', '.join([list(msg.keys())[0] for msg in messages]))
        else:
            strMsg = list(messages[0].keys())[0]
        self.__entries.append((vl, strMsg))

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)
        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def deleteAll(self):
        self.deleteRows(range(0,self.rowCount()))

    def addelem(self, elem, line, column):
        """
        add the element at the line and column given : eventualy create new line
        """

        if line >= self.rowCount():
            if column == 0:
                self.addRow(elem[0], elem[1])
            return
        if column == 0:
            self.__entries[line] = elem
        self._update()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if column == 0:
            info = self.__entries[row][column]
            if info is None:
                return None
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return info[0]
            elif role == Qt.UserRole:
                return info[0]
        elif column == 1:
            info = self.__entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole, Qt.UserRole]:
                return info
        return None
