# ----------------------------------------------------------------------------
# - wcgui-virtual-link-table                                                 -
# - wild cruncher:sys_map virtual link widget                                -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcgui_virtual_link_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan sys_map virtuallinks widget

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""
import os, re

from PyQt5.QtCore import QAbstractTableModel, Qt, QMimeData, QByteArray, QSortFilterProxyModel
from PyQt5.QtWidgets import QMessageBox, QHBoxLayout, QWidget, QTableView, \
                            QHeaderView, QVBoxLayout, QPushButton, QSplitter, \
                            QAbstractItemView, QLineEdit

from PyQt5.QtGui import QDrag

from wcr_core.db.whdb_utils import whdb_getNextName
from wcr_frontend.gui.wgui_utils import wgui_getOpenFile
from wcr_core.project.wprj_xmlio import WPRJ_VLINK_SILKAN_XML_MIME_TYPE
from wcr_core.common.wcmn_file_handler import WcmnFile
from wcr_frontend.modules.sys_map.gui.wcgui_binding_module_dialog import \
    WcguiAutoBindModules
from wcr_frontend.modules.vlink.gui.wvgui_vlink_gui_module import \
    WVGUI_COCKPIT_SYSTEM_MODULE
from wcr_core.project.wprj_project import wprj_load_project, \
    wprj_import_project, wprj_createProjectDB
from wcr_frontend.gui.wgui_project_editor_config import WguiProjectEditorCfg, \
    WguiProjectEditorDbCfgStore, WguiProjectEditorCfgDefaults
from wcr_frontend.gui.wgui_project_editor import WguiProjectEditor
from wcr_core.wcor_global_context import WcorGlobalContext
from wcr_core.modules.sys_map.wcmod_bus_file_handler \
    import WcmodVLinkFileHandler
from wcr_frontend import showErrorWindow
from wcr_frontend.gui.wgui_utils import WGUI_VL_PRODUCER_MIME_TYPE, \
                                        WGUI_VL_CONSUMER_MIME_TYPE


class WcguiTableWidgetDragRows(QTableView):
    def __init__(self, scene, trColId, *args, **kwargs):
        QTableView.__init__(self, *args, **kwargs)

        self.__scene = scene
        self.__trColId = trColId
        self.setDragEnabled(True)
        self.setAcceptDrops(False)
        self.setDragDropOverwriteMode(False)
        self.setDropIndicatorShown(True)

        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setDragDropMode(QAbstractItemView.DragDrop)

    def mouseMoveEvent(self, event):
        routes = self.getSelectedRoutesFast()
        uuids = [str(route.getUUID()) for route in routes]
        res = self.sendMultipleRoutes(uuids, WGUI_VL_PRODUCER_MIME_TYPE)
        if res:
            if len(uuids) > 1:
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Question)
                msg.setText('We now need to bind consumers for each routes.\n'
                            'Do you want to bind them all to one port ? if not you will have to'
                            'choose a port for each routes ?\n')
                msg.setWindowTitle("Producer routes binding")
                msg.addButton(QMessageBox.Yes)
                msg.addButton(QMessageBox.No)
                ret = msg.exec_()
                strUuids = [str(route.getUUID()) for route in routes]
                uuidsRoutes = [(str(route.getUUID()), route) for route in routes]
                if ret == msg.No:
                    for uuid, route in uuidsRoutes:
                        self.sendOneRoute(uuid, route,
                                          WGUI_VL_CONSUMER_MIME_TYPE)
                elif ret == msg.Yes:
                    self.sendMultipleRoutes(strUuids,
                                            WGUI_VL_CONSUMER_MIME_TYPE)
            else:
                uuid, route = [(str(route.getUUID()), route) for route in routes][0]
                self.sendOneRoute(uuid, route, WGUI_VL_CONSUMER_MIME_TYPE)

        self.__scene._unhighlightsConnectedConnectorPoint()

    def sendMultipleRoutes(self, uuids, mode):
        strUuids = ', '.join(uuids)
        drag = QDrag(self)
        mimeData = QMimeData()

        mimeData.setData(mode,
                         QByteArray(strUuids.encode('utf-8')))
        drag.setMimeData(mimeData)
        return drag.exec_(Qt.CopyAction | Qt.MoveAction)

    def sendOneRoute(self, uuid, route, mode):
        showErrorWindow('Consumer routes binding',
                        'Select port to consume route {}',
                        route.getName())
        strUuid = ''.join(uuid)
        drag = QDrag(self)
        mimeData = QMimeData()
        mimeData.setData(mode,
                         QByteArray(strUuid.encode('utf-8')))
        drag.setMimeData(mimeData)
        return drag.exec_(Qt.CopyAction | Qt.MoveAction)

    def getSelectedRoutesFast(self):
        """
        get all the selected routes
        :return:
        """
        selRoutes = []
        model = self.model()
        for index in self.selectedIndexes():
            if index.column() == self.__trColId:
                _, route = model.data(index, Qt.UserRole, returnRoute=True)
                selRoutes.append(route)
        return selRoutes


class WcguiCustomFilterModel(QSortFilterProxyModel):

    def __init__(self, model):
        super().__init__()
        self.model = model
        self.__entries = model.getEntries()

    def getEntries(self):
        return self.__entries

    def setFilterFixedString(self, str):
        self.__entries = self.model.filterEntries(str)
        super().setFilterFixedString(str)

    def data(self, index, role=None, returnRoute=False):
        return self.sourceModel().data(index, role, returnRoute,
                                       filtered=self.__entries)


class WcguiFilterTwoModels:

    def __init__(self, model1, model2):
        self.__model1 = model1
        self.__model2 = model2
        self._initialize()

    def _initialize(self):
        self.__filter1 = WcguiCustomFilterModel(self.__model1)
        self.__filter2 = WcguiCustomFilterModel(self.__model2)
        self.__filter1.setDynamicSortFilter(True)
        self.__filter2.setDynamicSortFilter(True)
        self.__filter1.setFilterKeyColumn(1)
        self.__filter2.setFilterKeyColumn(0)
        self.__filter1.setSourceModel(self.__model1)
        self.__filter2.setSourceModel(self.__model2)
        self.__filter1.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.__filter2.setFilterCaseSensitivity(Qt.CaseInsensitive)

    def setFilter(self, text):
        self.__filter1.setFilterFixedString(text)
        self.__filter2.setFilterFixedString(text)

    def getFilter1(self):
        return self.__filter1

    def getFilter2(self):
        return self.__filter2


class WcguiVLinkViewTable(QWidget):
    """
    Virtual link presentation table widget.
    """

    def __init__(self, pctx, scene, pBdtbl, cBdtbl, actx, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param pBdtbl: producer VL binding table
        :type pBdtbl: WcguiVLProdBindingTable
        :param cBdtbl: consumer VL binding table
        :type cBdtbl: WcguiVLConsBindingTable
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project context
        self.__pctx = pctx
        self.__actx = actx
        self.__parent = parent
        self.__top = self.__pctx.getProjectDB().getTopHierarchy()
        #: Graphical scene
        self.__scene = scene
        #: Update binding tables removing invalids binds
        self.__PropagateUpdateFunc = None
        # producer binding table
        self.__pBdtbl = pBdtbl
        # consumer binding table
        self.__cBdtbl = cBdtbl

        #: Bus binding table model
        self.__modTblM = _WcguiVLinkTableModel()
        self.__modTblMAuto = _WcguiVLinkAutoTableModel()


        """:type: (WphdbPhyBusInstance) -> list[WshdbBusInstance]"""

        #: Bus binding table widget
        userTrColId = 1
        autoTrColId = 0
        self.__modTblW = WcguiTableWidgetDragRows(scene, userTrColId)
        self.__modTblWAuto = WcguiTableWidgetDragRows(scene, autoTrColId)

        #: Add entry button
        self.__linkButton = QPushButton('Link Route File')
        self.__impButton = QPushButton('Import Routes')
        self.__autoRouteButton = QPushButton('Generate routes')
        self._initElements()
        self.setLayout(self._createLayout())
        self.updateTable()

    @property
    def modTblM(self):
        return self.__modTblM

    @property
    def modTblMAuto(self):
        return self.__modTblMAuto

    def _initElements(self):
        self.__ctblFilter = WcguiFilterTwoModels(self.__modTblM, self.__modTblMAuto)
        fModel1 = self.__ctblFilter.getFilter1()
        fModel2 = self.__ctblFilter.getFilter2()

        self.__modTblW.setModel(fModel1)
        self.__modTblW.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.ResizeToContents)
        self.__modTblW.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch)

        self.__modTblWAuto.setModel(fModel2)
        self.__modTblWAuto.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.ResizeToContents)
        self.__modTblWAuto.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch)

        #: Configuration table filter input
        self.__configFilterF = QLineEdit()
        self.__configFilterF.setPlaceholderText('Filter route name')
        self.__configFilterF.textChanged.connect(self.__ctblFilter.setFilter)
        self.__modTblWAuto.doubleClicked.connect(self._openRoutesView)
        self.__modTblW.doubleClicked.connect(self._openRoutesView)
        self.__linkButton.clicked.connect(self._linkVirtualLink)
        self.__impButton.clicked.connect(self._importVirtualLink)
        self.__autoRouteButton.clicked.connect(self.autoRouteFunction)

    def _openRoutesView(self, idx):
        module = WVGUI_COCKPIT_SYSTEM_MODULE.getAppModule()
        gctx = WcorGlobalContext()
        pctx = module.createProjectContext(gctx, wprj_createProjectDB(module))

        cfgTable = pctx.getProjectDB().tops.getLocalConfiguration()
        cfgStore = WguiProjectEditorDbCfgStore(gctx.getConfigurationService())
        glblPrjtEdiCfg = WguiProjectEditorCfg(
            WguiProjectEditorCfgDefaults(), store=cfgStore)
        pcfg = WguiProjectEditorCfg(glblPrjtEdiCfg,
                                    WguiProjectEditorDbCfgStore(cfgTable))
        module.actx = self.__actx
        cctx = WVGUI_COCKPIT_SYSTEM_MODULE.createContext(
            pctx, pcfg, {}, self.__actx, self)
        if cctx:
            editor = WguiProjectEditor(pcfg, pctx, cctx)
            if self.__top.getVLinkFiles():
                filepath = self.__top.getVLinkFiles().path
            else:
                filepath = None
            mainWin = self.__actx.getParent()
            if filepath:
                top = editor.getProjectDB().getTopHierarchy()
                editor.getProjectDB().path = filepath
                top.setSourcePath(filepath)
            editor.scene.refresh()
            editor.scene.populateTable()
            mainWin.tabs.openEditor(editor.getEditorIdent(), editor)
            self.__openProtocolTab(idx, mainWin)

    def __openProtocolTab(self, idx, mainWin):
        """
        if protocol detected open the corresponding tab in traffic route view
        :param idx: Index Object
        :param mainWin: MainWindow application
        """
        model = idx.model()
        if isinstance(model, _WcguiVLinkAutoTableModel):
            idxRoute = 0
            vl = model.getEntries()[idx.row()][idxRoute]
        elif isinstance(model, _WcguiVLinkTableModel):
            idxRoute = 1
            vl = model.getEntries()[idx.row()][idxRoute][1]
        else:
            return
        protocol = vl.getVLType()
        mainWin.tabs.currentEditor.scene.openTrafficEditor(protocol)

    def removeNonExistantBoundVl(self):
        """
        Delete a Traffic routes that were bound but don't exist anymore
        """
        selIdx = self.__modTblWAuto.selectedIndexes()
        producerModel = self.__pBdtbl.getModel()
        consumerModel = self.__cBdtbl.getModel()
        messageVlsUUID = []
        for vlInst, msg in self.__modTblMAuto.getEntries():
            messageVlsUUID.append(vlInst.getUUID())
        producerRows = producerModel.getDataRows()
        for idx in sorted(range(producerModel.rowCount()), reverse=True):
            # [1] get column that contain vl data [1] get Vl instance
            vlProduced = producerRows[idx][1][1]
            if vlProduced.getUUID() not in messageVlsUUID:
                producerModel.deleteRow(idx)

        consumerRows = consumerModel.getDataRows()
        for idx in sorted(range(consumerModel.rowCount()), reverse=True):
            # [1] get column that contain vl data [1] get Vl instance
            vlConsumed = consumerRows[idx][1][1]
            if vlConsumed.getUUID() not in messageVlsUUID:
                consumerModel.deleteRow(idx)

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)

        layout.addWidget(self.__modTblW)

        btnLayout = QHBoxLayout()

        btnLayout.addWidget(self.__linkButton)
        btnLayout.addWidget(self.__impButton)
        layout.addLayout(btnLayout)

        layout.addWidget(QSplitter())

        layout.addWidget(self.__modTblWAuto)
        btnLayout = QHBoxLayout()
        btnLayout.addWidget(self.__autoRouteButton)
        layout.addLayout(btnLayout)
        layout.addWidget(self.__configFilterF)

        return layout

    def ResetWithData(self):
        """
        Fill the table from the project's data
        """
        if self.__top.getVLinkBlock() and self.__top.getVLinkFiles():
            oldRoutesUUIDs = self.getAllRoutesUUIDs()
            self.__modTblM.deleteAll()
            self.__modTblMAuto.deleteAll()
            routes = [inst for inst in self.__top.getVLinkBlock().getInstances()]
            vlFileName = os.path.basename(self.__top.getVLinkFiles().path)
            if len(routes) > 0:
                for vl in routes:
                    messages = vl.getVLMessagesName()
                    if messages:
                        self.__modTblMAuto.addRow(vl, messages)
                    else:
                        self.__modTblM.addRow(vlFileName, (vl.getName(), vl))
            # compare all the routes before and after deletion
            # to see if changes occurred
            newRoutesUUIDs = self.getAllRoutesUUIDs()
            if sorted(oldRoutesUUIDs) != sorted(newRoutesUUIDs):
                self.__scene.sceneModified.emit()

    def getAllRoutesUUIDs(self):
        routes = []
        for row in self.__modTblM.getEntries():
            routes.append(row[1][1].getUUID())
        for row in self.__modTblMAuto.getEntries():
            routes.append(row[0].getUUID())
        return routes

    def _linkVirtualLink(self):
        """
        Import a new virtual link file and fill the mapper data and the table
        with is data
        """
        # Selection d'un fichier
        selected_path = wgui_getOpenFile("Link route file", self,
                                         WPRJ_VLINK_SILKAN_XML_MIME_TYPE)
        if selected_path:
            if self.__top.getVLinkFiles():
                if self.__top.getVLinkFiles().path != selected_path:
                    self.__askLinkFileErase(selected_path)
            elif self.__top.setVLinkFile(WcmnFile(selected_path)):
                self.__pctx.updateVL(selected_path)
                self.ResetWithData()
        self._unbindinvalid()
        # for some reason need to call initialize twice ... need investigation
        self.__actx.getEditor().initialize()
        self.__actx.getEditor().initialize()

    def __askLinkFileErase(self, selected_path):
        diaw = QMessageBox(self)
        diaw.setWindowTitle('Confirm Link')
        diaw.setText('A Traffic Route file is already linked, '
                     ' do you want to link a new one and erase the existing one ?')
        diaw.setInformativeText(
            'All the existing routes will be lost.')
        diaw.addButton(QMessageBox.Yes)
        diaw.addButton(QMessageBox.No)
        diaw.setIcon(QMessageBox.Warning)
        cnfm = diaw.exec()
        if cnfm == QMessageBox.Yes:
            if self.__top.setVLinkFile(WcmnFile(selected_path)):
                self.__pctx.updateVL(selected_path)
                self.ResetWithData()
                return
        elif cnfm == QMessageBox.No:
            return

    def _importVirtualLink(self):
        # Selection d'un fichier
        selected_path = wgui_getOpenFile("Import routes", self,
                                         WPRJ_VLINK_SILKAN_XML_MIME_TYPE)
        if selected_path and self.__top.getVLinkFiles():
            if self.__top.getVLinkFiles().path != selected_path:
                # Recuperation des messages
                handler = WcmodVLinkFileHandler(self.__pctx.getProjectDB().getModule())
                vlink = handler.handleFile(selected_path, self.__pctx)
                if not vlink:
                    raise ValueError("Wrong source file for Traffic Route Source.")
                vlBlock = self.__top.getVLinkBlock()
                for inst in vlink.getOperatorInstances():
                    oper = inst.getOperator()
                    allRoutes = [inst.getName() for inst in vlBlock.getInstances()]
                    routeName = whdb_getNextName(inst.getName(),
                                                 allRoutes)
                    newInst = vlBlock.createOperatorInstance(oper, routeName)
                    params = inst.getVLParametersDict(inst.getVLType())
                    newInst.setVLType(inst.getVLType())
                    newInst.setVLParameters(params)
                self.ResetWithData()
                self._unbindinvalid()

    def updateTable(self, tmpPath=None):
        """
        Update the table
        """
        self.__pctx.getMappingManager().resetVLink()
        if tmpPath:
            self.__pctx.updateVL(tmpPath)
            self.__top.getVLinkBlock().setTmpFilePath(tmpPath)
        else:
            vlf = self.__top.getVLinkFiles()
            if vlf:
                self.__pctx.updateVL(vlf.path)
                editor = self.__parent
                tabs = self.__actx.getParent().tabs
                if editor:
                    editor.createTmpFile(tabs)
        self.ResetWithData()
        self._unbindinvalid()

    def autoRouteFunction(self):
        map = self.__pctx.getMappingManager()
        if not map.logicTop:
            showErrorWindow("Route Creation",
                            'No logical view is bound', None)
            return
        bindingModule = WcguiAutoBindModules(self.__pctx, self.__scene,
                                             self.__modTblWAuto,
                                             self.__modTblMAuto,
                                             self.__pBdtbl,
                                             self.__cBdtbl,
                                            )
        oldRoutesUUIDs = self.getAllRoutesUUIDs()
        bindingModule.computeMapping()
        # compare all the routes before and after deletion
        # to see if changes occurred
        newRoutesUUIDs = self.getAllRoutesUUIDs()
        if sorted(oldRoutesUUIDs) != sorted(newRoutesUUIDs):
            self.__scene.sceneModified.emit()

    def _unbindinvalid(self):
        """
        inbind invalide physical port/virtual link and connections
        """
        self.__pctx.unbindInvalidVLinkBinding()
        if not self.__PropagateUpdateFunc is None:
            self.__PropagateUpdateFunc()

    def setPropagationFunction(self, propfunc):
        """
        give logic table a function to propagate a module update effect on
        other tables
        """
        self.__PropagateUpdateFunc = propfunc


class _WcguiVLinkTableModel(QAbstractTableModel):
    """
    Bus Binding table model.
    """

    TITLE_FILE = 'File'
    TITLE_VL = 'Traffic Routes User'
    HEADERS = [TITLE_FILE, TITLE_VL]

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Entries as list of tuples of
        #  physical Port (pport) and logical Port (lport)
        self.__entries = []
        """:type: list[]"""

        self._update()

    def getEntries(self):
        return self.__entries

    def filterEntries(self, str):
        filteredEntries = []
        if str:
            for (file_, vl) in self.__entries:
                if str.lower() in vl[0].lower():
                    filteredEntries.append((file_, vl))
        else:
            return self.__entries
        return filteredEntries

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, file_, vl):
        """
        Add a row at the end of the list (append)

        :param vl: traffic route name
        :type vl: | string
        :param file_: traffic route file
        :type file_: | [string]
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())
        self.__entries.append((file_, vl))
        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)
        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def deleteAll(self):
        self.deleteRows(range(0, self.rowCount()))

    def addelem(self, elem, line, column):
        """
        add the element at the line and column given : eventualy create new line
        """

        if line >= self.rowCount():
            if column == 0:
                self.addRow(elem[0], elem[1])
            return
        if column == 0:
            self.__entries[line] = elem
        self._update()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)

    def data(self, index, role=None, returnRoute=False, filtered=None):
        row = index.row()
        column = index.column()
        if filtered:
            entries = filtered
        else:
            entries = self.__entries
        if column == 0:
            info = entries[row][column]
            if info is None:
                return None
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole, Qt.UserRole]:
                return info
        elif column == 1:
            info = entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole, Qt.UserRole]:
                # return route data if route is True
                if returnRoute:
                    return info[0], info[1]
                else:
                    return info[0]
        return None

    def flags(self, index):
        return Qt.ItemIsEditable | Qt.ItemIsDragEnabled | Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsDropEnabled


class _WcguiVLinkAutoTableModel(QAbstractTableModel):
    """
    Bus Binding table model.
    """

    TITLE_VL = 'Traffic Routes Auto'
    MESSAGES = 'Messages'
    HEADERS = [TITLE_VL, MESSAGES]

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)


        #: Entries as list of tuples of
        #  physical Port (pport) and logical Port (lport)
        self.__entries = []
        """:type: list[]"""

        self._update()

    def getEntries(self):
        return self.__entries

    def filterEntries(self, str):
        filteredEntries = []
        if str:
            for (vl, strMsg) in self.__entries:
                if str.lower() in vl.getName().lower():
                    filteredEntries.append((vl, strMsg))
        else:
            return self.__entries
        return filteredEntries

    def getRoutesName(self):
        return [vl.getName() for vl, _ in self.__entries]

    def getMessagesName(self):
        return [msg for _, msg in self.__entries]

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, vl, messages):
        """
        Add a row at the end of the list (append)

        :param vl: traffic route name
        :type vl: | string
        :param messages: Messages to be set
        :type messages: | [string]
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())
        if len(messages) == 0:
            strMsg = 'No messages bound'
        elif len(messages) > 1:
            strMsg = '[{}]'.format(', '.join([msg for msg in messages]))
        else:
            strMsg = messages[0]
        self.__entries.append((vl, strMsg))

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)
        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def deleteAll(self):
        self.deleteRows(range(0,self.rowCount()))

    def addelem(self, elem, line, column):
        """
        add the element at the line and column given : eventualy create new line
        """

        if line >= self.rowCount():
            if column == 0:
                self.addRow(elem[0], elem[1])
            return
        if column == 0:
            self.__entries[line] = elem
        self._update()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)

    def data(self, index, role=None, returnRoute=False, filtered=None):
        row = index.row()
        column = index.column()
        if filtered:
            entries = filtered
        else:
            entries = self.__entries
        if column == 0:
            info = entries[row][column]
            if info is None:
                return None
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole, role == Qt.UserRole]:
                # return route data if route is True
                if returnRoute:
                    return info.getName(), info
                else:
                    return info.getName()
        elif column == 1:
            info = entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole, Qt.UserRole]:
                return info
        return None

    def flags(self, index):
        return Qt.ItemIsEditable | Qt.ItemIsDragEnabled | Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsDropEnabled
