# ----------------------------------------------------------------------------
# - wpgui-host-connection-points-props-widget                                -
# - wild cruncher - cockpit : connection point properties dialog widgets     -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpgui_host_connection_points_props_widget
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher connection point properties dialog widgets

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import OrderedDict
from operator import itemgetter

from PyQt4.QtCore import Qt, QAbstractTableModel
from PyQt4.QtGui import QPushButton, QComboBox, QSortFilterProxyModel, \
    QLineEdit, QAbstractItemView, QWidget, QVBoxLayout, QLabel, \
    QTableView, QHeaderView

from wcr_frontend.gui.wgui_props_dialog_widget import \
    WguiXPointPropertiesBase, WguiXPointPropsTextField, \
    WguiXPointPropsDeleteButton, WGUI_NAME_PLACEHOLDER, WGUI_XPNT_CODE, \
    WguiXPointPropsROLookupField, WguiXPointPropsFieldInstance, \
    WguiXPointPropsField
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_NAME, \
    WHDB_XPOINT_TYPE
from wcr_core.modules.system.db.wphdb_connection_point_base import \
    WPHDB_XPOINT_CONFIG_KEY


_WPGUI_XPNT_CONFIG = '__config'


# noinspection PyAbstractClass
class WpguiHostXPointProperties(WguiXPointPropertiesBase):
    def __init__(self, instance, pctx, parent=None):
        """
        :param instance: The instance object
        :type instance: WphdbHostInstance
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param parent: Parent widget
        """
        module = pctx.getModule()
        """:type: WpmodSystemApplicationModuleP"""
        self.__xpntTypeNames = {
            x.getCode(): x.getName() for x in module.getConnectionPointTypes()}
        self.__xpntCfgMetas = {
            x.getCode(): x.getConfigMeta() for x
            in module.getConnectionPointTypes()}

        super().__init__(instance, parent=parent)

        if self.__class__ is WpguiHostXPointProperties:
            self._finishInitialization()

    def createFields(self):
        return {
            WHDB_XPOINT_NAME: WguiXPointPropsTextField(
                WHDB_XPOINT_NAME, 'Conn. Point', WGUI_NAME_PLACEHOLDER),
            WHDB_XPOINT_TYPE: WguiXPointPropsROLookupField(
                WHDB_XPOINT_TYPE, 'Type', self.__xpntTypeNames),
            _WPGUI_XPNT_CONFIG: WguiXPointPropsField(
                _WPGUI_XPNT_CONFIG, 'Configuration',
                WpguiXPointConfigFieldInstance, self.__xpntCfgMetas),
            'del': WguiXPointPropsDeleteButton('')}

    def getFieldOrder(self):
        return [WHDB_XPOINT_NAME, WHDB_XPOINT_TYPE,
                _WPGUI_XPNT_CONFIG, 'del']

    def getStretchFields(self):
        return {_WPGUI_XPNT_CONFIG}

    def getButtons(self):
        def add():
            ctyp = wgui_qcb_get_data(typeF)
            if ctyp:
                meta = self.__xpntCfgMetas.get(ctyp)
                self.model.addXPoint(**{
                    WHDB_XPOINT_TYPE: ctyp,
                    _WPGUI_XPNT_CONFIG: meta.getEmptyValue()})

        typeF = QComboBox()
        for code, name in sorted(self.__xpntTypeNames.items(),
                                 key=itemgetter(1)):
            typeF.addItem(name, code)

        addBtn = QPushButton('Add')
        addBtn.clicked.connect(add)

        return [typeF, addBtn]

    def resetModel(self, model, instance):
        model.setXPoints(
            {WGUI_XPNT_CODE: x,
             WHDB_XPOINT_NAME: x.getName(),
             WHDB_XPOINT_TYPE: x.getEntityType(),
             _WPGUI_XPNT_CONFIG:
                 OrderedDict(x.rawElement.getMeta(WPHDB_XPOINT_CONFIG_KEY))}
            for x in instance.getConnectionPoints())

    def createXPoint(self, idx, model, instance):
        return instance.createConnectionPoint(
            ctyp=model.getValue(idx, WHDB_XPOINT_TYPE),
            name=model.getValue(idx, WHDB_XPOINT_NAME))

    def updateXPoint(self, idx, model, xpnt):
        name = model.getValue(idx, WHDB_XPOINT_NAME)
        if name:
            xpnt.setName(name)
        xpnt.rawElement.setMeta(model.getValue(idx, _WPGUI_XPNT_CONFIG),
                                WPHDB_XPOINT_CONFIG_KEY)


class _WpguiXPointConfigDataAdapter(object):
    def __init__(self, model, xpntCfgMetas):
        self.__model = model
        self.__xpntCfgMetas = xpntCfgMetas

    def _getMeta(self, index):
        """
        Get the configuration meta-data for the connection point at the given
        index.

        :param index: Index
        :type index: int
        :return: Configuration meta-data
        :rtype: WdbsRecordMetaTypeP
        """
        ctyp = self.__model.getValue(index, WHDB_XPOINT_TYPE)
        """:type: str"""
        return self.__xpntCfgMetas.get(ctyp)

    def getKeys(self, index):
        """
        Get the configuration keys for the connection point at the given index.

        :param index: Index
        :type index: int
        :return: Configuration keys
        :rtype: Iterable[str]
        """
        return self._getMeta(index).getKeys()

    def getLabel(self, index, key):
        """
        Get the label for the configuration key for the connection point
        at the given index.

        :param index: Index
        :type index: int
        :param key: Configuration key
        :type key: str
        :return: Label
        :rtype: str
        """
        return self._getMeta(index).getLabel(key)


class _WpguiXPointConfigTableModel(QAbstractTableModel):
    def __init__(self, model, cfgAdapter, parent=None):
        """
        :param model: Connection points data model
        :type model: WguiXPointPropsModelP
        :param cfgAdapter: Configuration data adapter
        :type cfgAdapter: _WpguiXPointConfigDataAdapter
        :param parent: Parent item
        """
        super().__init__(parent)

        #: Connection points data model
        self.__model = model
        #: Configuration data adapter
        self.__cfgAdapter = cfgAdapter

        #: Current connection point index
        self.__index = -1
        #: Current configuration keys
        self.__cfgKeys = []

    def _getCfgDict(self, idx=None):
        """
        Get the configuration dictionary of the connection point at the given
        index.

        :param idx: Index of the connection point (current if `None`)
        :type idx: int | None
        :return: Configuration dictionary
        :rtype: dict[str, str]
        """
        if idx is None:
            idx = self.__index
        if idx is None or idx < 0:
            return {}

        return self.__model.getValue(idx, _WPGUI_XPNT_CONFIG)

    def _getCfgValue(self, key):
        """
        Get the configuration value for the given key for the current
        connection point.

        :param key: Key
        :type key: str
        :return: Configuration value
        :rtype: str
        """
        return self._getCfgDict().get(key)

    def _setCfgValue(self, key, value):
        """
        Set the configuration value for the current connection point for the
        given key to the given value.

        :param key: Key
        :type key: str
        :param value: Value
        :type value: str
        """
        idx = self.__index
        if idx is None or idx < 0:
            return

        cdct = self.__model.getValue(idx, _WPGUI_XPNT_CONFIG)
        """:type: dict[str, str]"""
        cdct[key] = value
        self.__model.setValue(idx, _WPGUI_XPNT_CONFIG, cdct)

    def activeXPointChanged(self, index):
        """
        Callback for changes in the currently active connection point.

        :param index: Index of the currently active connection point
        :type index: int
        """
        self.beginResetModel()
        self.__index = index
        if index > 0:
            self.__cfgKeys = list(self.__cfgAdapter.getKeys(index))
        else:
            self.__cfgKeys = []
        self.endResetModel()

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__cfgKeys)

    def columnCount(self, parent=None, *args, **kwargs):
        return 2

    def data(self, index, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            key = self.__cfgKeys[index.row()]
            if index.column() == 0:
                return self.__cfgAdapter.getLabel(self.__index, key)
            else:
                return self._getCfgValue(key)
        return None

    def setData(self, index, value, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if index.column() == 1:
                key = self.__cfgKeys[index.row()]
                self._setCfgValue(key, value)
                return True
        return False

    def flags(self, index):
        flgs = super().flags(index)

        if index.column() == 1:
            return flgs | Qt.ItemIsEditable
        else:
            return flgs & ~Qt.ItemIsEditable


class WpguiXPointConfigFieldInstance(WguiXPointPropsFieldInstance):
    def __init__(self, code, label, context, xpntCfgMetas):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        """
        super().__init__(code, label)

        self.__context = context

        #: Configuration data adapter
        self.__cfgAdapter = _WpguiXPointConfigDataAdapter(
            context.model, xpntCfgMetas)

        #: Configuration table view
        self.__configTblF = QTableView()
        #: Configuration table model
        self.__configTblM = _WpguiXPointConfigTableModel(
            context.model, self.__cfgAdapter)
        #: Configuration table filter proxy model
        self.__ctblFilter = QSortFilterProxyModel(self.__configTblM)
        self.__ctblFilter.setFilterCaseSensitivity(Qt.CaseInsensitive)
        #: Configuration table filter input
        self.__configFilterF = QLineEdit()

        self.__auxWidget = None

    def _createAuxWidget(self):
        self.__ctblFilter.setSourceModel(self.__configTblM)
        self.__configTblF.setModel(self.__ctblFilter)
        self.__configTblF.setSelectionMode(QAbstractItemView.NoSelection)
        self.__configTblF.horizontalHeader().setResizeMode(
            0, QHeaderView.ResizeToContents)
        self.__configTblF.horizontalHeader().setResizeMode(
            1, QHeaderView.Stretch)
        self.__configFilterF.setPlaceholderText('Filter configuration keys')

        self.__context.activeXPointChanged.connect(
            self.__configTblM.activeXPointChanged)
        self.__configFilterF.textChanged.connect(
            self.__ctblFilter.setFilterFixedString)

        wdgt = QWidget()
        layout = QVBoxLayout()
        layout.addWidget(
            QLabel('Configuration values for the selected connection point:'))
        layout.addWidget(self.__configTblF)
        layout.addWidget(self.__configFilterF)
        wdgt.setLayout(layout)
        return wdgt

    def installWidget(self, index, widgetF, textF):
        # noinspection PyShadowingNames
        def formatConfig():
            cfg = self.__context.model.getValue(index, _WPGUI_XPNT_CONFIG)
            """:type: dict[str, str]"""
            if not cfg:
                return ''

            return '; '.join(
                '{}: {}'.format(self.__cfgAdapter.getLabel(index, k) or k,
                                cfg.get(k))
                for k in self.__cfgAdapter.getKeys(index) if cfg.get(k))

        def configChanged(idx, code):
            if idx == index and code == self.code:
                textUpdateF(formatConfig())

        textUpdateF = textF(formatConfig())
        self.__context.model.valueChanged.connect(configChanged)

    def getAuxWidget(self):
        if not self.__auxWidget:
            self.__auxWidget = self._createAuxWidget()
        return self.__auxWidget
