# ----------------------------------------------------------------------------
# - wpgui-host-processes-props-widget                                        -
# - wild cruncher: processes properties dialog       frontend / system / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpgui_host_processes_props_widget
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher processes properties dialog

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from functools import partial
from itertools import chain
from PyQt4.QtCore import Qt, QAbstractListModel, QModelIndex
from PyQt4.QtGui import QVBoxLayout, QComboBox, QTabWidget, QWidget, QLabel

from silkan_common.scmn_utils import first
from wcr_core.modules.system.db.wphdb_connection_points import \
    WphdbEthernetXPoint
from wcr_core.modules.system.wpmod_distrib_service_default_types import \
    WPMOD_ETH_DISTSERV_PARAMCODE_INTERFACE
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP
from wcr_frontend.modules.system.gui.wpgui_selection_provider import \
    WpguiComboboxSelectionProvider
from wcr_frontend.gui.wgui_record_table_widgets import \
    WguiRecordTableWidget, WguiRecordTableModel, WguiFixedRecordTableWidget, \
    WguiParameterTableModel, WguiRecordTableWidgetBase, \
    WguiRecordTableDelegateDescCmbBox, WguiRecordTableItemDelegate
from wcr_core.modules.system.db.wphdb_host_instance import *


class WpguiHostProcessesProperties(WguiPropertiesWidgetP):
    """
    Properties widget for host processes.
    """

    def __init__(self, inst, pctx, parent=None):
        """
        :param inst: the instance object
        :type inst: WphdbHostInstance
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: instance object
        self.__instance = inst
        #: Project context
        self.__pctx = pctx

        #: Processes model
        self.__processM = _WpguiProcessTableModel(parent=self)

        #: Processes widget
        self.__processLstW = WguiRecordTableWidget(
            self.__processM, itemName='Process')
        #: Tab-widget for process child widgets
        self.__mdlSvcTabW = QTabWidget()

        #: Model table widget
        self.__modelsW = _WpguiProcessModelsWidget(pctx)

        #: Service table widget
        self.__servicesW = _WpguiProcessServicesWidget(pctx)

        #: Distribution service table widget
        self.__distribSvcsW = _WpguiProcessDistributionServicesWidget(
            pctx, inst)

        self._initElements()
        self.setLayout(self._createLayout())
        self.updateFields()

    def _initElements(self):
        """
        Initialize UI elements.
        """
        processChanged = self.__processLstW.currentRecordChanged
        processChildChange = self.__processM.externalDataChange

        # Link model items to current process
        childModel = self.__modelsW.getChildModel()
        processChanged.connect(childModel.setModelData)
        childModel.anyDataChanged.connect(processChildChange)

        # Link service items to current process
        childModel = self.__servicesW.getChildModel()
        processChanged.connect(childModel.setModelData)
        childModel.anyDataChanged.connect(processChildChange)

        # Link distribution service items to current process
        childModel = self.__distribSvcsW.getChildModel()
        processChanged.connect(childModel.setModelData)
        childModel.anyDataChanged.connect(processChildChange)

    def _createLayout(self):
        """
        Create the layout.

        :return: Newly created layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setMargin(0)

        self.__mdlSvcTabW.addTab(self.__modelsW, 'Models')
        self.__mdlSvcTabW.addTab(self.__servicesW, 'Services')
        self.__mdlSvcTabW.addTab(self.__distribSvcsW, 'Distrib. Services')

        layout.addWidget(self.__processLstW)
        layout.setStretch(0, 1)
        layout.addWidget(self.__mdlSvcTabW)
        layout.setStretch(1, 2)

        return layout

    def updateFields(self):
        self.__modelsW.updateFields()
        self.__servicesW.updateFields()
        self.__distribSvcsW.updateFields()
        self.__processM.setModelData(self.__instance.getRawProcesses())

    def save(self):
        data = self.__processM.getModelData()
        """:type: dict[UUID, dict[str, object]]"""
        self.__instance.setRawProcesses(data)
        self.__pctx.getMappingManager().rescanMappings()


class _WpguiUnmappedModelsModel(QAbstractListModel):
    def __init__(self, mmngr):
        """
        :param mmngr: Mapping manager
        :type mmngr: WpmodMappingManager
        """
        super().__init__()

        #: Mapping manager
        self.__mappingManager = mmngr

        #: List of unmapped models
        self.__unmapped = []

        self.resetUnmappedModels()

    def resetUnmappedModels(self):
        """
        Reload the list of unmapped models from the mapping manager.
        """
        self.__unmapped = list(
            sorted(((m.getUUID(), m.getName()) for m
                    in self.__mappingManager.getUnmappedModels()),
                   key=lambda t: t[1]))

    def markModelMapped(self, uuid):
        """
        Mark the given model as mapped, i.e. remove it from the list of
        unmapped models.

        :param uuid: UUID of the model to be marked as mapped
        :type uuid: UUID
        """
        idx = first(i for i, (u, _) in enumerate(self.__unmapped) if u == uuid)
        assert idx is not None

        self.beginRemoveRows(QModelIndex(), idx, idx)
        del self.__unmapped[idx]
        self.endRemoveRows()

    def markModelUnmapped(self, uuid):
        """
        Mark the given model as unmapped, i.e. add it to the list of unmapped
        models.

        :param uuid: UUID of the model to be marked as unmapped
        :type uuid: UUID
        """
        name = self.__mappingManager.getModelByUUID(uuid).getName()
        for idx, (_, n) in enumerate(self.__unmapped):
            if name < n:
                break
        else:
            idx = len(self.__unmapped)

        self.beginInsertRows(QModelIndex(), idx, idx)
        self.__unmapped.insert(idx, (uuid, name))
        self.endInsertRows()

    def data(self, idx, role=None):
        row = idx.row()

        if row < 0 or row > len(self.__unmapped):
            return None

        if role == Qt.UserRole:
            return self.__unmapped[row][0]
        else:
            return self.__unmapped[row][1]

    def headerData(self, col, orientation, role=None):
        return str(col) if orientation == Qt.Horizontal else 'Model'

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__unmapped)


class _WpguiProcessTableModel(WguiRecordTableModel):
    _MODELS_SERVICES_FIELD_KEY = '__ms'

    #: Channel fields
    _FIELDS = [
        ('Name', WPHDB_HOST_PROCESS_NAME_KEY),
        ('Models / Services', _MODELS_SERVICES_FIELD_KEY)]

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(_WpguiProcessTableModel._FIELDS, parent=parent)

    def flags(self, index):
        return super().flags(index) & ~Qt.ItemIsEditable

    def _getFieldData(self, field, rkey, entry):
        if field == self._MODELS_SERVICES_FIELD_KEY:
            return ', '.join(sorted(chain(
                (m[WPHDB_HOST_MODEL_NAME_KEY] for m
                 in entry[WPHDB_HOST_MODEL_MAPS_KEY].values()),
                (s[WPHDB_HOST_SERVICE_KEY] for s
                 in entry[WPHDB_HOST_SERVICE_MAPS_KEY].values()))))
        return super()._getFieldData(field, None, entry)

    def _createEmptyEntry(self):
        rkey = wdbs_makeUUID()
        entry = WPHDB_PROCESS_META_TYPE.getEmptyValue()
        entry[WPHDB_HOST_PROCESS_NAME_KEY] = 'Process {}'.format(rkey)
        return rkey, entry


#############################################################################
#
# MODELS
#
#############################################################################

class _WpguiProcessModelsWidget(QWidget):
    def __init__(self, pctx, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Unmapped models list model
        self.__unmappedModelsL = \
            _WpguiUnmappedModelsModel(pctx.getMappingManager())

        #: Unmapped models combobox
        self.__unmappedModelsW = QComboBox()

        #: Model-mapping model
        self.__modelM = _WpguiProcessModelTableModel(
            WpguiComboboxSelectionProvider(self.__unmappedModelsW),
            parent=self)

        #: Model table widget
        self.__modelsW = _WguiProcessModelTableWidget(
            self.__modelM, self.__unmappedModelsL, self.__unmappedModelsW)

        self._initElements()
        self.setLayout(self._createLayout())
        self.updateFields()

    def _initElements(self):
        """
        Initialize UI elements.
        """
        self.__unmappedModelsW.setModel(self.__unmappedModelsL)

    def _createLayout(self):
        """
        Create the layout.

        :return: Newly created layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setMargin(0)
        layout.addWidget(self.__modelsW)
        return layout

    def getChildModel(self):
        return self.__modelM

    def updateFields(self):
        self.__unmappedModelsL.reset()


class _WguiProcessModelTableWidget(WguiRecordTableWidget):
    def __init__(self, model, unmappedModelsL, unmappedModelsW, parent=None):
        super().__init__(
            model, itemName='Model', addWdgts=(unmappedModelsW,),
            parent=parent)

        self.__unmappedModelsL = unmappedModelsL
        """:type: _WpguiUnmappedModelsModel"""

    def _addItem(self):
        rkey = super()._addItem()
        """:type: UUID"""
        if rkey:
            self.__unmappedModelsL.markModelMapped(rkey)

    def _removeItem(self):
        rkey = super()._removeItem()
        """:type: UUID"""
        if rkey:
            self.__unmappedModelsL.markModelUnmapped(rkey)


class _WpguiProcessModelTableModel(WguiRecordTableModel):
    #: Channel fields
    _FIELDS = [
        ('Model', WPHDB_HOST_MODEL_NAME_KEY)]

    def __init__(self, mdlPrv, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(
            self._FIELDS, WPHDB_HOST_MODEL_MAPS_KEY, parent=parent)

        self.__modelProvider = mdlPrv

    def flags(self, index):
        return super().flags(index) & ~Qt.ItemIsEditable

    def _createEmptyEntry(self):
        mtpl = self.__modelProvider.getSelection()
        if not mtpl:
            return None, None
        modelName, rkey = mtpl

        entry = WPHDB_PROCESS_MODEL_META_TYPE.getEmptyValue()
        entry[WPHDB_HOST_MODEL_NAME_KEY] = modelName
        return rkey, entry


#############################################################################
#
# SERVICES
#
#############################################################################

def _wpgui_service_param_func(pctx):
    """

    :param pctx:
    :type pctx: WpmodSystemProjectContextP
    :return:
    :rtype: (dict, str | None) -> dict[str, WcmnParameterDescriptionP]
    """
    # noinspection PyUnusedLocal
    def getParam(record, rkey):
        if record is None:
            return None
        scod = record.get(WPHDB_HOST_SERVICE_KEY)
        if not scod:
            return None
        prms = first(ps for s, ps in pctx.getServices() if s == scod)
        if not prms:
            return None
        return {p.getName():p for p in prms}
    return getParam


class _WpguiProcessServicesWidget(QWidget):
    def __init__(self, pctx, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Known services combobox
        self.__availServicesW = QComboBox()

        #: Services model
        self.__servicesM = _WpguiProcessServiceTableModel(
            WpguiComboboxSelectionProvider(self.__availServicesW), parent=self)
        #: Service parameters model
        self.__srvParamM = WguiParameterTableModel(
            pfld=WPHDB_HOST_SERVICE_MAP_PARAMS_KEY, parent=self,
            paramDescF=_wpgui_service_param_func(pctx))

        #: Service table widget
        self.__servicesW = WguiRecordTableWidget(
            self.__servicesM, addWdgts=(self.__availServicesW,),
            itemName='Service')
        #: Service parameters table widget
        self.__svcParamsW = WguiFixedRecordTableWidget(self.__srvParamM)

        self._initElements(pctx)
        self.setLayout(self._createLayout())
        self.updateFields()

    def _initElements(self, pctx):
        """
        Initialize UI elements.
        """
        self.__servicesW.currentRecordChanged.connect(
            self.__srvParamM.setModelData)
        self.__srvParamM.anyDataChanged.connect(
            self.__servicesM.externalDataChange)

        for srvc, prms in pctx.getServices():
            self.__availServicesW.addItem(srvc, prms)

    def _createLayout(self):
        """
        Create the layout.

        :return: Newly created layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setMargin(0)
        layout.addWidget(self.__servicesW)
        layout.addSpacing(5)
        layout.addWidget(
            QLabel('Parameters for the currently selected service:'))
        layout.addWidget(self.__svcParamsW)
        return layout

    def getChildModel(self):
        return self.__servicesM

    def updateFields(self):
        pass


class _WpguiProcessServiceTableModel(WguiRecordTableModel):
    #: Channel fields
    _FIELDS = [
        ('Name', WPHDB_HOST_SERVICE_MAP_LABEL_KEY),
        ('Service', WPHDB_HOST_SERVICE_KEY)]

    def __init__(self, srvcPrv, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(
            self._FIELDS, WPHDB_HOST_SERVICE_MAPS_KEY, parent=parent)

        self.__serviceProvider = srvcPrv

    def flags(self, index):
        flgs = super().flags(index)

        if index.column() == 1:
            flgs &= ~Qt.ItemIsEditable

        return flgs

    def _createEmptyEntry(self):
        stpl = self.__serviceProvider.getSelection()
        """:type: (str, Iterable[WcmnParameterDescriptionP]) | None"""
        if not stpl:
            return None, None
        srvc, prms = stpl

        rkey = wdbs_makeUUID()
        entry = WPHDB_PROCESS_SERVICE_META_TYPE.getEmptyValue()

        name = '{}-Instance {}'.format(srvc, rkey)
        entry[WPHDB_HOST_SERVICE_MAP_LABEL_KEY] = name
        entry[WPHDB_HOST_SERVICE_KEY] = srvc

        for p in prms:
            prec = WHDB_PARAMETER_META_TYPE.getEmptyValue()
            prec[WHDB_PARAMETER_VALUE_KEY] = p.getDefaultValue()
            entry[WPHDB_HOST_SERVICE_MAP_PARAMS_KEY][p.getName()] = prec

        return rkey, entry


#############################################################################
#
# DISTRIBUTION SERVICES
#
#############################################################################


def _wpgui_distrib_service_param_func(module):
    """

    :param module:
    :type module: WpmodSystemApplicationModuleP
    :return:
    :rtype: (dict, str | None) -> dict[str, WcmnParameterDescriptionP]
    """
    # noinspection PyUnusedLocal
    def getParam(record, rkey):
        if record is None:
            return None
        scod = record.get(WPHDB_HOST_DISTRIBSVC_KEY)
        if not scod:
            return None
        dsvc = module.getDistributionServiceType(scod)
        if not dsvc:
            return None
        return {p.getName():p for p in dsvc.getSocketParameters()}
    return getParam


class _WpguiProcessDistributionServicesWidget(QWidget):
    def __init__(self, pctx, inst, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param inst: the instance object
        :type inst: WphdbHostInstance
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Known services combobox
        self.__availDistribServicesW = QComboBox()

        #: Distribution services model
        self.__distribSvcsM = _WpguiProcessDistribServiceTableModel(
            WpguiComboboxSelectionProvider(self.__availDistribServicesW),
            parent=self)
        #: Service parameters model
        module = pctx.getModule()
        """:type: WpmodSystemApplicationModuleP"""
        self.__dsvcParamM = WguiParameterTableModel(
            pfld=WPHDB_HOST_DISTRIBSVC_SCKT_PARAMS_KEY, parent=self,
            paramDescF=_wpgui_distrib_service_param_func(module))

        #: Distribution service table widget
        self.__distribSvcsW = WguiRecordTableWidget(
            self.__distribSvcsM, addWdgts=(self.__availDistribServicesW,),
            itemName='Distribution Service')
        #: Service parameters table widget
        self.__dsvcParamsW = _WpguiProcessDistServParamTableWidget(
            self.__dsvcParamM, inst)

        self._initElements(pctx)
        self.setLayout(self._createLayout())
        self.updateFields()

    def _initElements(self, pctx):
        """
        Initialize UI elements.
        """
        self.__distribSvcsW.currentRecordChanged.connect(
            self.__dsvcParamM.setModelData)
        self.__dsvcParamM.anyDataChanged.connect(
            self.__distribSvcsM.externalDataChange)

        module = pctx.getModule()
        svcs = list(module.getDistributionServiceTypes())
        svcs.sort(key=lambda s: s.getName())
        """:type: WpmodSystemApplicationModuleP"""
        for dsvc in svcs:
            self.__availDistribServicesW.addItem(dsvc.getName(), dsvc)

    def _createLayout(self):
        """
        Create the layout.

        :return: Newly created layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setMargin(0)
        layout.addWidget(self.__distribSvcsW)
        layout.addSpacing(5)
        layout.addWidget(
            QLabel(
                'Parameters for the currently selected distribution service:'))
        layout.addWidget(self.__dsvcParamsW)
        return layout

    def getChildModel(self):
        return self.__distribSvcsM

    def updateFields(self):
        pass


class _WpguiProcessDistribServiceTableModel(WguiRecordTableModel):
    #: Channel fields
    _FIELDS = [
        ('Name', WPHDB_HOST_DISTRIBSVC_SCKT_LABEL_KEY),
        ('Service', WPHDB_HOST_DISTRIBSVC_KEY)]

    def __init__(self, srvcPrv, parent=None):
        """
        :param srvcPrv: Current service selection provider for new instances
        :type srvcPrv: WpguiSelectionProvider
        :param parent: Parent widget
        """
        super().__init__(
            self._FIELDS, WPHDB_HOST_DISTRIBSVC_SCKTS_KEY, parent=parent)

        self.__serviceProvider = srvcPrv

    def flags(self, index):
        flgs = super().flags(index)

        if index.column() == 1:
            flgs &= ~Qt.ItemIsEditable

        return flgs

    def _createEmptyEntry(self):
        stpl = self.__serviceProvider.getSelection()
        """:type: (str, WpmodDistributionServiceTypeP)"""
        if not stpl:
            return None, None
        _, dsvc = stpl

        rkey = wdbs_makeUUID()
        entry = WPHDB_PROCESS_DISTRIB_SERVICE_META_TYPE.getEmptyValue()

        name = '{}-Instance {}'.format(dsvc.getName(), rkey)
        entry[WPHDB_HOST_DISTRIBSVC_SCKT_LABEL_KEY] = name
        entry[WPHDB_HOST_DISTRIBSVC_KEY] = dsvc.getCode()

        for p in dsvc.getSocketParameters():
            mtyp = WHDB_PARAMETER_META_TYPE
            prec = mtyp.getEmptyValue()
            prec[WHDB_PARAMETER_VALUE_KEY] = p.getDefaultValue()
            entry[WPHDB_HOST_DISTRIBSVC_SCKT_PARAMS_KEY][p.getName()] = prec

        return rkey, entry


class _WpguiProcessDistServParamTableWidget(WguiRecordTableWidgetBase):
    """
    Distributed service parameters table with delegated value column
    """

    def __init__(self, model, inst, parent=None):
        """
        :param model: Model to display
        :type model: WguiRecordTableModelBase
        :param inst: the instance object
        :type inst: WphdbHostInstance
        :param parent: Parent widget
        """
        def ethConCheck(pt):
            return isinstance(pt, WphdbEthernetXPoint)

        super().__init__(model, parent)

        self.__inst = inst

        delegDict = {
            WPMOD_ETH_DISTSERV_PARAMCODE_INTERFACE:
            WguiRecordTableDelegateDescCmbBox(
                partial(self._connectionPtPrv, ethConCheck))}

        self.__valueDelegate = WguiRecordTableItemDelegate(
            delegDict, 1, self)

        self._initElements()
        self._tableW.setItemDelegateForColumn(2, self.__valueDelegate)

        self.setLayout(self._createLayout())

    # noinspection PyMethodMayBeStatic
    def _connectionPtPrv(self, fltrFct):
        """
        :param fltrFct: Function used to filter the object
        :type fltrFct: (object) -> bool
        :rtype: list[str, object]
        """
        return [(pt.getName() + ' {' + str(pt.getUUID()) + '}', pt)
                for pt in self.__inst.getConnectionPoints() if fltrFct(pt)]
