# ----------------------------------------------------------------------------
# - wpgui-logical-simulation-tree                                            -
# - wild cruncher: logical simulation database tree browser                  -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpgui_logical_simulation_tree
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher logical simulation database tree browser

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.action.wact_action_api import WactActionObserverP, \
    WACT_ACTION_EXECUTED_EVENT
from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP
from wcr_frontend.gui.wgui_db_tree import WguiDbTreeItemFactory, \
    WguiDbTreeModelManager, WguiDbTreeSortFilter
from wcr_frontend.gui.wgui_tree import WguiTree
from wcr_frontend.gui.wgui_protocols import WguiEmptyContextMenuFactory, \
    QWidget_Meta_M
from wcr_core.db.whdb_block_proxy import WhdbEntityContainerProxy
from wcr_core.modules.system.db.wphdb_host_instance import WphdbHostInstance


class WpguiLogicalSimulationTree(WguiTree, WactActionObserverP,
                                 WcmnReleasableP, metaclass=QWidget_Meta_M):
    def __init__(self, pctx, cctx, actions=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param cctx: Cockpit system context
        :type cctx: WguiCockpitSystemContextP
        :param actions: Action to observe for update
        :type actions: Iterable[WactActionP] | None
        """
        self.__topLevel = WhdbEntityContainerProxy()
        self.__pctx = pctx
        self.__actions = list(actions) if actions else []

        modelMngr = WguiDbTreeModelManager(
            self.__topLevel, WguiDbTreeItemFactory(cctx))
        model = WguiDbTreeSortFilter(modelMngr.model)

        super().__init__(model, modelMngr, WguiEmptyContextMenuFactory())

        for action in self.__actions:
            if action:
                action.registerActionObserver(self)

    def release(self):
        for action in self.__actions:
            if action:
                action.unregisterActionObserver(self)
        self.__actions = []
        self.__topLevel.release()
        self.__pctx = None
        super().release()

    def actionNotify(self, actx, action, event, *args, **kwargs):
        if event & WACT_ACTION_EXECUTED_EVENT:
            print('LOGICAL SIM REFRESH')
            self.updateTree()

    def updateTree(self):
        pctx = self.__pctx
        if pctx is None:
            return

        self.__topLevel.setDelegate(pctx.getLogicalSimulation())
        super().updateTree()


class WpguiLogicalSimulationSelectionAdapter():
    """
    The `WpguiLogicalSimulationSelectionAdapter` links the current selection
    of physical items (in the graphics scene and the system-mapper database
    browser) and the selection of the logical items (in the logical simulation
    database browser), ensuring that they are in sync.

    When a physical item is selected, all logical items that are mapped to that
    physical item, if any, will be selected.
    Vice versa, when a logical item is selected, the physical item that this
    logical item is mapped to, if any, will be selected.
    """

    def __init__(self, scene, tree, mmngr):
        """
        :param scene: Graphics scene
        :type scene: WguiGraphicsScene
        :param tree: Logical simulation database tree
        :type tree: WpguiLogicalSimulationTree
        :param mmngr: Mapping manager
        :type mmngr: WpmodMappingManager
        """
        #: Graphics scene
        self.__scene = scene
        #: Logical simulation database tree
        self.__tree = tree
        #: Mapping manager
        self.__mappingManager = mmngr

        #: Whether a selection process is currently ongoing (to avoid loops)
        self.__selectionOngoing = False

        scene.selectionChanged.connect(self._sceneItemsSelected)
        tree.itemSelected.connect(self._logicalItemsSelected)

    def _sceneItemsSelected(self):
        if self.__selectionOngoing:
            return
        self.__selectionOngoing = True

        hosts = (h for h in self.__scene.getSelectedEntities()
                 if isinstance(h, WphdbHostInstance))
        models = filter(None,
                        (self.__mappingManager.getModelByUUID(m.uuid)
                         for h in hosts
                         for p in h.getProcesses()
                         for m in p.getModelMappings()))
        self.__tree.selectDbItems(list(models))

        self.__selectionOngoing = False

    def _logicalItemsSelected(self, items):
        if self.__selectionOngoing:
            return
        self.__selectionOngoing = True

        insts = (ip[0] for ip in items if isinstance(ip[0], WhdbInstanceP))
        procs = (self.__mappingManager.getModelProcess(i.getUUID())
                 for i in insts)
        hosts = set(p.host for p in procs if p)

        self.__scene.clearSelection()
        for host in hosts:
            item = self.__scene.getSceneItem(host)
            if item is not None:
                # noinspection PyUnresolvedReferences
                item.setSelected(True)
                # noinspection PyUnresolvedReferences
                item.ensureVisible()

        self.__selectionOngoing = False
