# ----------------------------------------------------------------------------
# - wvgui-bus-binding-table                                                  -
# - wild cruncher:Vlink Bus bindings widget                                  -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wvgui_bus_binding_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan Vlink Bus bindings widget

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""

from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QWidget, QTableView, \
	QHeaderView, QVBoxLayout, QPushButton, QItemDelegate, QAbstractItemView
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_core.modules.vlink.db.wvhdb_Vlink_instance import WvhdbVLinkInstance


class WvguiBusBindingTable(QWidget):
    """
    Traffic Route binding to Logical Bus table widget.
    """

    BTN_PRINT_BINDING_VISIBLE = False

    def __init__(self, pctx, scene, parent=None):
        """
        :param pctx: Project context
        :type pctx: WvmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project context
        self.__pctx = pctx
        self.__topBlock = self.__pctx.getProjectDB().getTopHierarchy()
        #: Graphical scene
        self.__scene = scene
        #: View filter
        self.__filter = filter


        #: formating function for element ot table
        self.__formatF = lambda o, c: self.formatF(o, c)
        #: message binding table model
        self.__msgTblM = _WvguiMessageBindingTableModel(self.__formatF,
                                                        parent=self)

        #: Creating function to get all possible host
        self.__vlinkF = lambda: self.getVlinks()
        #: Creating function geting all available messages for the given bus
        self.__messF = lambda: self.getMessages()
        #: Creating editor for element in table
        self.__msgDlgt = _WcguiModuleHostBindDelegate(self.__vlinkF,
                                                      self.__messF,
                                                      self.formatF,
                                                      self.__topBlock)
        # table
        self.__msgTblW = QTableView()
        #: Add entry button
        self.__addButton = QPushButton('Add')
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')

        self._initElements()
        self.setLayout(self._createLayout())
        self._loadFromData()

    def formatF(self, obj, col):
        if obj is None or obj is '':
            return ''
        elif col == 0:
            #: at col 0 we must have a WvhdbVLinkInstance
            return obj.getName()
        elif col == 1 and isinstance(obj, dict):
            return list(obj.keys())[0] if len(obj.keys()) > 0 else None
        else:
            #: we asking a name for an object in column normaly not present
            return 'No formating available'

    def _initElements(self):
        self.__msgTblW.setModel(self.__msgTblM)
        self.__msgTblW.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.Stretch)
        self.__msgTblW.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch)
        self.__msgTblW.setEditTriggers(QAbstractItemView.AllEditTriggers)
        #: delegate classe for edition in table
        self.__msgTblW.setItemDelegate(self.__msgDlgt)

        self.__addButton.clicked.connect(self._addEntry)
        self.__deleteButton.clicked.connect(self._deleteSelEntries)

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.addWidget(self.__msgTblW)

        btnLayout = QHBoxLayout()
        btnLayout.addWidget(self.__addButton)
        btnLayout.addWidget(self.__deleteButton)
        layout.addLayout(btnLayout)

        return layout

    def reloadFromData(self):
        """
        remove all row of table and refill it with element from project data
        """
        self.__msgTblM.deleteRows(range(0,self.__msgTblM.rowCount()),
                                   Unbind=False)
        self._loadFromData()

    def _loadFromData(self):
        """
        Fill the table from the project's data
        """
        tblk =  self.__topBlock
        map = self.__pctx.getMappingManager()
        #creating all lines coresponding to an allready existing binding
        for vluuid, message in tblk.getVlinkMessageBinds():
            """ bnd = (Vlink uuid, bus uuid, message uuid)"""""

            for oper in tblk.getOperatorInstances():
                if (isinstance(oper, WvhdbVLinkInstance) and
                            oper.getUUID() == vluuid):
                    selectoper = oper
                    break
            else:
                raise Exception("Binded operator Don't exist")

            self.__msgTblM.addRow(selectoper, message)

    def _addEntry(self):
        """
        Add entry in the model
        """
        self.__msgTblM.addRow()

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.__msgTblW.selectedIndexes()
        """:type: list[QModelIndex]"""

        self.__msgTblM.deleteRows([idx.row() for idx in selIdx])

    def unbindVlinkMsg(self, vlink, message):
        """
        unbind the vlink and the msg for the topblock
        :param : vlink
        :type : WvhdbVlinkInstance
        :param : msg
        :type : str
        """
        self.__topBlock.removeVlinkMessageBind(vlink.getUUID(), message)

    def bindVlinkMsg(self, vlink, message):
        """
        bind the vlink and the message for the topblock
        :param : vlink instance
        :type : WvhdbVlinkInstance
        :param : message
        :type : str
        """
        self.__topBlock.addVlinkMessageBind(vlink.getUUID(), message)

    def getMessages(self):
        """
        return the list of available bus
        """
        mngr = self.__pctx.getMappingManager()
        messages = mngr.getAllMessagesBinding()
        """:type :[message]"""
        return messages

    def getVlinks(self):
        """
        return the list of Virtual link
        """
        vlinkl = self.__topBlock.getOperatorInstances()
        return vlinkl


class _WvguiMessageBindingTableModel(QAbstractTableModel):
    """
    Bus Binding table model.
    """

    TITLE_VLINK = 'TRoute'
    TITLE_MESSAGE = 'Message'
    HEADERS = [TITLE_VLINK, TITLE_MESSAGE]

    def __init__(self, formatF, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent=parent


        #: Entries as list of tuples of
        #  physical Vlink (vlink) and logical Bus (bus)
        self.__entries = []
        """:type: list[]"""
        self.__formatF=formatF

        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, vlink=None, message=None):
        """
        Add a row at the end of the list (append)

        :param vlink: Vlink to be set
        :type vlink: WvhdbVlinkInstance | None
        :param message: message to be set
        :type message: string
        :return: The appended row index
        :rtype: int
        """

        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append((vlink, message))

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices, Unbind=True):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row, Unbind=Unbind)

    def deleteRow(self, rowIndex, Unbind=True):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        vlink, mess = self.__entries[rowIndex]
        if vlink is not None and mess is not None and Unbind:
            try:
                self.__parent.unbindVlinkMsg(vlink, mess)
                pass
            except ValueError:
                pass

        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column < 3:
            info = self.__entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return self.__formatF(info, column)
            elif role == Qt.UserRole:
                return info
        return None

    def setData(self, index, data, role=None):
        """
        :type index: QModelIndex
        :type data: object
        """
        assert role == Qt.UserRole
        self._setDataEntry(index, data)

    def _setDataEntry(self, index, data):
        vlink, mess = self.__entries[index.row()]
        if vlink is not None and mess is not None:
            #a binding could have been done by that line before
            # it have to be deleted.
            try:
                self.__parent.unbindVlinkMsg(vlink, mess)
            except ValueError:
                pass

        #creating the new representation line
        nl = ()
        if index.column() == 0:
            nl = (data, mess)
        elif index.column() == 1:
            nl = (vlink, data)

        self.__entries[index.row()] = nl

        vl, m = self.__entries[index.row()]
        if vl is not None and m is not None:
            #line is full a binding have to be done
            try:
                self.__parent.bindVlinkMsg(vl, m)
            except ValueError:
                self.__entries[index.row()] = (None, None)
                self._update()
                pass

    def flags(self, index):
        """
        set flags fot able
        """
        flgs = super().flags(index)
        #allowing editing
        flgs |= Qt.ItemIsEditable
        return flgs

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)


class _WcguiModuleHostBindDelegate(QItemDelegate):
    """
    Delegate for modules and host
    """

    def __init__(self, vlinkList, messList, formatF, topblk, parent=None):
        """
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)

        self.formatF = formatF
        self.__vlinkList = vlinkList
        self.__messList = messList
        self.__topBlock = topblk

    def createEditor(self, parent, option, index):
        """
        Create a combobox containing adequate choices for the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """

        def save(*_):
            self.setModelData(editor, index.model(), index)

        editor = QComboBox(parent)
        col = index.column()
        #start withe empty line to allowing unset
        editor.addItem('', None)

        # Iterate through the messages using the adequate provider
        # and add items to the combo box
        if col == 0:
            availData = [vl for vl in self.__vlinkList()]
        elif col == 1:
            availData = [{key: msg} for key, msg in self.__messList().items()]
        for data in availData:
            editor.addItem(self.formatF(data, col), data)
        editor.currentIndexChanged.connect(save)

        return editor

    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)

    def setEditorData(self, editor, index):
        """
        Initialising editor data with most
        :type editor: QComboBox
        :type index: QModelIndex
        """
        data = index.model().data(index, Qt.UserRole)
        for index in range(editor.count()):
            if editor.itemData(index, Qt.UserRole) == data:
                editor.setCurrentIndex(index)
                break
        else:
            editor.setCurrentIndex(0)
