# ----------------------------------------------------------------------------
# - wvgui-logic-table                                                        -
# - wild cruncher:vlink imported bus widget                                  -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wvgui_logic_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan vlink widget for bus view import

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""

from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5.QtWidgets import QHBoxLayout, QWidget, QTableView, QHeaderView, \
    QVBoxLayout, QPushButton
from wcr_frontend.gui.wgui_utils import wgui_getOpenFile
from wcr_core.project.wprj_xmlio import WPRJ_ICD_SILKAN_XML_MIME_TYPE

class WvguiLogicViewTable(QWidget):
    """
    Logical bus importing and updating table widget.
    """

    def __init__(self, pctx, scene, parent=None):
        """
        :param pctx: Project context
        :type pctx: WvmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project context
        self.__pctx = pctx
        #: Graphical scene
        self.__scene = scene
        #: Update binding tables removing invalids binds
        self.__PropagateUpdateFunc = None

        #: Message binding table model
        self.__msgTblM = _WvguiLogicTableModel()

        #: Message binding table widget
        self.__msgTblW = QTableView()

        #: Add entry button
        self.__impButton = QPushButton('Import')
        #: Delete entry button
        self.__updButton = QPushButton('Update')
        #: Reset entry button
        self.__resButton = QPushButton('Reset')

        self._initElements()
        self.setLayout(self._createLayout())
        self.updateTable()

    def _initElements(self):
        self.__msgTblW.setModel(self.__msgTblM)
        self.__msgTblW.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.Stretch)
        self.__impButton.clicked.connect(self._importModule)
        self.__updButton.clicked.connect(self.updateTable)
        self.__resButton.clicked.connect(self.resetTable)

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)

        layout.addWidget(self.__msgTblW)

        btnLayout = QHBoxLayout()

        btnLayout.addWidget(self.__impButton)
        btnLayout.addWidget(self.__updButton)
        btnLayout.addWidget(self.__resButton)
        layout.addLayout(btnLayout)

        return layout

    def resetTable(self):
        """
        empty the table from the project's data
        """
        self.__pctx.getMappingManager().resetMessages()
        top = self.__pctx.getProjectDB().getTopHierarchy()
        top.setAllMessages([])
        top.setICDFile('')
        self.__msgTblM.deleteAll()

    def ResetWithData(self):
        """
        Fill the table from the project's data
        """
        self.__msgTblM.deleteAll()
        top = self.__pctx.getProjectDB().getTopHierarchy()
        messages = self.__pctx.getMappingManager().getAllMessages()
        if messages:
            l = 0
            for i in messages:
                self.__msgTblM.addelem(i, l)
                l += 1
        top.setAllMessages(messages.keys())

    def _importModule(self):
        """
        Import a new ICD view and fill the mapper data and the table
        """
        # Selection d'un fichier
        selected_path = wgui_getOpenFile("Import Messages", self,
                                         WPRJ_ICD_SILKAN_XML_MIME_TYPE)
        if selected_path:
            self.__pctx.getProjectDB().getTopHierarchy(). \
                setICDFile(selected_path)
            self.__pctx.updateMessages(selected_path)
            self.ResetWithData()
        self._unbindinvalid()

    def updateTable(self):
        """
        Re-import the ICD view of mapping manager and fill the table
        with the updated information
        """
        self.__pctx.getMappingManager().resetMessages()
        icdFile = self.__pctx.getProjectDB().getTopHierarchy().getICDFile()
        if icdFile:
            self.__pctx.updateMessages(str(icdFile))
        self.ResetWithData()
        self._unbindinvalid()

    def _unbindinvalid(self):
        """
        inbind invalide vlink/bus and connections
        """
        pass

    def setPropagationFunction(self, propfunc):
        """
        give logic table a function to propagate a bus update effect on
        other tables
        """
        self.__PropagateUpdateFunc = propfunc


class _WvguiLogicTableModel(QAbstractTableModel):
    """
    Bus Binding table model.
    """

    TITLE_MESSAGES = 'Messages'
    HEADERS = [TITLE_MESSAGES]

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)

        self.__entries = []
        """:type: list[]"""

        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, bus):
        """
        Add a row at the end of the list (append)

        :param bus: the new value as entry
        :type bus: | None
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append(bus)

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)
        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def deleteAll(self):
        self.deleteRows(range(0,self.rowCount()))

    def addelem(self, elem, line):
        """
        add the element at the line and column given : eventualy create new line
        """
        if line < self.rowCount():
            self.__entries[line] = (elem)
        else:
            self.addRow(elem)
        self._update()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return 1

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column < 1:
            info = self.__entries[row]
            if info is None:
                return None
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return info
            elif role == Qt.UserRole:
                return info
        return None
