# ----------------------------------------------------------------------------
# - wvgui-project-actions                                                    -
# - wild cruncher - project editor actions                                   -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wvgui_project_actions
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher project editor actions

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""

import csv

from wcr_core.action.wact_action_parameter import WactActionParameter
from wcr_core.action.wact_action_parameter_types import \
    WactActionOpenFileByPayloadParameterType
from wcr_core.common.wcmn_mime_type import WCMN_PAYLOAD_BLOCK_CSV
from wcr_core.db.whdb_utils import whdb_ensureUniqueName
from wcr_core.modules.vlink.db.wvhdb_standard_operators import \
    WVHDB_VLINK_CODE
from wcr_core.modules.vlink.wvmod_application_module import \
    WVMOD_VLINK_APP_CODE
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_XPOINT_CETRAC_CODE, WHHDB_XPOINT_ETHERNET_CODE, \
    WHHDB_XPOINT_DISCRET_CODE, WHHDB_XPOINT_FIBER_CHANNEL_CODE, \
    WHHDB_XPOINT_1553_CODE, WHHDB_XPOINT_ARINC825_CODE, \
    WHHDB_XPOINT_AFDX_CODE
from wcr_frontend import showErrorWindow
from wcr_frontend.gui.wgui_project_actions import WguiEditorActionBase
from wcr_frontend.modules.vlink.gui.wvgui_scene_protocol_table import *
from wcr_core.modules.vlink.db.wvhdb_Vlink_instance import *

"""dic = {csvhead(str): vlinkinst_metakey(str)} """

WVGUI_CSV_VL_NAME_KEY = 'Name'
WVGUI_CSV_VL_TYPE_KEY = 'Type'

WVGUI_CSV_VL_CONFIG = { 'Arp': LOG_TR_ETH_ARP_TAG,
                        'isDeterminist': LOG_TR_ETH_DETER_TAG,
                        'StorageMethod': LOG_TR_STORAGE_TAG,
                        'NbBuffers': LOG_TR_BUFFER_TAG,
                        'FrameSize': LOG_TR_FRAMESIZE_TAG,
                        'VlNumber': LOG_TR_AFDX_NUMBER_TAG,
                        'Bag': LOG_TR_AFDX_BAG_TAG,
                        'Jitter': LOG_TR_AFDX_JITTER_TAG,
                        'Lmax': LOG_TR_AFDX_LMAX_TAG,
                        'Lmin': LOG_TR_AFDX_LMIN_TAG,
                        'snIntegrityCheck':LOG_TR_AFDX_SN_CHECK_TAG,
                        'RedundancyManagement': LOG_TR_AFDX_REDNT_MGNT_TAG,
                        'RedundantVl': LOG_TR_AFDX_REDNT_VL_TAG,
                        'Priority': LOG_TR_AFDX_PRIO_TAG,
                        'IpSrc': LOG_TR_ETH_IP_SRC_TAG,
                        'IpDest': LOG_TR_ETH_IP_DEST_TAG,
                        'MacSrc': LOG_TR_ETH_MAC_SRC_TAG,
                        'MacDest': LOG_TR_ETH_MAC_DEST_TAG,
                        'Protocol': LOG_TR_ETH_PROTOCOL_TAG,
                        }


class _WmguiEditorActionBase(WguiEditorActionBase):
    def __init__(self, name, description, *parameters):
        """
        :param name: Action name
        :type name: str
        :param description: Action description
        :type description: str
        :param parameters: Action parameter descriptions
        :type parameters: list[WactActionParameterP]
        """
        super().__init__(name, description, *parameters)

    def isEnabled(self, actx):
        if not super().isEnabled(actx):
            return False
        editor = actx.getEditor()
        """:type: WguiProjectEditor"""
        cctx = editor.cockpitSystemContext
        return cctx.code == WVMOD_VLINK_APP_CODE


class _WvguiEditorImportCSVAction(_WmguiEditorActionBase):
    """
    Project editor action to import a model.
    """

    #: Action parameters
    __PARAMETER = WactActionParameter(
        'model', 'Model file', 'Model file to import',
        WactActionOpenFileByPayloadParameterType(WCMN_PAYLOAD_BLOCK_CSV))

    def __init__(self):
        super().__init__(
            'Import Vlink from CSV...',
            'Import Traffic Route from CSV File',
            self.__PARAMETER)

    def _executeForEditor(self, editor, actx, paramValuePairs):
        # raise NotImplementedError("Import is not compatible with cetrac v3 format yet")
        file = self._getParamValue(self.__PARAMETER, actx, paramValuePairs)
        """:type: WcmnFile | None"""
        if not file:
            return
        pctx = actx.getProjectContext()
        tblk = pctx.getProjectDB().getTopHierarchy()
        fl = open(file.path, 'r')
        csvdesc = csv.reader(fl)
        headers = []
        csvparamdic = {}
        modules = {}
        modulesInst = editor.scene.getTabModule()
        modules[WVHDB_VLINK_TYPE_AFDX] = modulesInst[WHHDB_XPOINT_AFDX_CODE]
        modules[WVHDB_VLINK_TYPE_CETRAC] = modulesInst[WHHDB_XPOINT_CETRAC_CODE]
        modules[WVHDB_VLINK_TYPE_ETNERNET] = modulesInst[WHHDB_XPOINT_ETHERNET_CODE]
        modules[WVHDB_VLINK_TYPE_DISCRETE] = modulesInst[WHHDB_XPOINT_DISCRET_CODE]
        modules[WVHDB_VLINK_TYPE_FIBER] = modulesInst[WHHDB_XPOINT_FIBER_CHANNEL_CODE]
        modules[WVHDB_VLINK_TYPE_1553] = modulesInst[WHHDB_XPOINT_1553_CODE]
        modules[WVHDB_VLINK_TYPE_ARINC825] = modulesInst[WHHDB_XPOINT_ARINC825_CODE]
        for vlp in csvdesc:
            if headers == []:
                headers = vlp
                continue
            for i in range(0, min(len(headers), len(vlp))):
                csvparamdic[headers[i]] = vlp[i]
            newop=tblk.createOperatorInstance(WVHDB_VLINK_CODE)
            if WVGUI_CSV_VL_NAME_KEY in csvparamdic.keys():
                name = csvparamdic[WVGUI_CSV_VL_NAME_KEY]
            else:
                name = 'default'

            newop.setName(name)
            if WVGUI_CSV_VL_TYPE_KEY in csvparamdic.keys():
                VL_TYPE = csvparamdic[WVGUI_CSV_VL_TYPE_KEY]
                newop.setVLType(VL_TYPE)
                for hkey in csvparamdic.keys():
                    if hkey != WVGUI_CSV_VL_TYPE_KEY and \
                                    hkey != WVGUI_CSV_VL_NAME_KEY:
                        prmKey = WVGUI_CSV_VL_CONFIG.get(hkey, None)
                        if prmKey:
                            newop.setVLParameters({prmKey: csvparamdic[hkey]})
                        else:
                            showErrorWindow('Import Error',
                                            'Traffic Route key "{}" does not exist.', hkey)

                whdb_ensureUniqueName(newop, editor.scene.topLevelModelBlock)

                sceneItem = editor.scene.addFunctionItem(newop)
                modules[VL_TYPE].addRow(sceneItem)


#: Model csv import action
WVGUI_IMPORT_CSV_ACTION = _WvguiEditorImportCSVAction()
