# ----------------------------------------------------------------------------
# - whgui-scene                                                              -
# - wild cruncher - cockpit : graphic scene widget                           -
# ----------------------------------------------------------------------------
# - author    (c) 2018-2018 achraf mahti                                     -
# - copyright (c) 2017-2018 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wvgui_scene
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher traffic route graphics scene editor

.. moduleauthor:: Achraf Mahti <mahti@silkan.net>
"""
from PyQt4.QtGui import QWidget, QListWidget, QHBoxLayout
from wcr_core.db.whdb_utils import whdb_getNextFunctionName
from wcr_core.db_api.whdb_protocols_api import WhdbBaseP
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_XPOINT_CETRAC_CODE, WHHDB_XPOINT_ETHERNET_CODE, \
    WHHDB_XPOINT_DISCRET_CODE, WHHDB_XPOINT_FIBER_CHANNEL_CODE, \
    WHHDB_XPOINT_1553_CODE, WHHDB_XPOINT_ARINC825_CODE, \
    WHHDB_XPOINT_AFDX_CODE
from wcr_core.modules.vlink.db.wvhdb_Vlink_instance import *
from wcr_frontend.gui.wgui_editor_tab_widget import WguiTabsWindow, \
    wgui_create_wtabs_actions
from wcr_frontend.gui.wgui_scene import WguiSceneViewContext
from wcr_frontend.gui.wgui_scene_item import *
from wcr_frontend.gui.wgui_scene_item_ids import *
from wcr_frontend.modules.vlink.gui.wvgui_scene_protocol_table import *


class WvguiGraphicsScene(QWidget):
    """
    The WvguiGraphicsScene class is the class responsible to manage
    specific scene action in traffic route view that WguiGraphicsScene don't handle
    """
    """:type: pyqtBoundSignal[WhdbBaseP, list[WhdbBlockInstanceP]]"""

    # the selection mode signal
    selectModeSignal = pyqtSignal()
    """:type: pyqtBoundSignal"""

    #: scene change signal
    sceneChanged = pyqtSignal(bool)
    """:type: pyqtBoundSignal[bool]"""

    #: scene path about to change signal
    scenePathChanging = pyqtSignal([object])
    """:type: pyqtBoundSignal[list[WhdbBlockInstanceP]]"""

    #: scene path change signal
    scenePathChanged = pyqtSignal([object])
    """:type: pyqtBoundSignal[list[WhdbBlockInstanceP]]"""

    #: scene item added signal
    sceneItemAdded = pyqtSignal(WhdbBaseP, list)
    """:type: pyqtBoundSignal[WhdbBaseP, list[WhdbBlockInstanceP]]"""

    #: scene item removed signal
    sceneItemRemoved = pyqtSignal(WhdbBaseP)
    """:type: pyqtBoundSignal[WhdbBaseP]"""

    #: scene item changed signal
    sceneItemChanged = pyqtSignal(WhdbBaseP)

    #: scene item changed signal
    selectionChanged = pyqtSignal(list)

    #: scene item selected signal
    sceneItemSelected = pyqtSignal(list)
    """:type: pyqtBoundSignal[list[WhdbBlockInstanceP]]"""

    #: scene item selected signal
    sceneRectChanged = pyqtSignal()
    """:type: pyqtBoundSignal[list[WhdbBlockInstanceP]]"""

    sideButtonToggled = pyqtSignal(str, bool)

    def __init__(self, pjdb, config, actions, pctx, cctx, actx, mod,
                 parent=None):
        super().__init__(parent)
        # the project database
        self.__pjdb = pjdb
        #: Project editor configuration
        self.__config = config
        #: Action dictionary
        self.__actions = actions
        #: Context
        self.__projectCtx = pctx
        #: Cockpit context
        self.__cctx = cctx
        #: Action execution context
        self.__actx = actx

        # : Application module
        self.__module = cctx.getAppModule()
        # : Scene item factory
        self.__sitmF = cctx.getSceneItemFactory()

        #: Scene items database
        self.__itemWRefs = {}
        self.__items = []
        self.listWidget = QListWidget()
        self.listWidget.itemClicked.connect(self.openTrafficEditor)
        """:type: dict[UUID, ReferenceType[WguiSceneItemP]]"""
        #: Currently created line, if any
        self.__line = None
        self.__viewContext = WguiSceneViewContext(self, pjdb, config)
        self.__topLevelProxy = pjdb.getTopHierarchy().createProxy()
        self.__muted = False
        # create a new tabs window
        self.__tabs = WguiTabsWindow(self)
        self.__tabModules = self._createTabModules()
        self.refresh()
        self.populateTable()
        self.__tabs.activeEditorChanged.connect(self._activeEditorChanged)
        self.setLayout(self._createLayout())

    def _activeEditorChanged(self, *_):
        """
        Callback for whenever the active editor changes, to update
        actions accordingly.
        """
        current = self.__tabs.currentEditor
        if current:
            self.setWindowTitle('{}'.format(current.getTitle()))

    def populateTable(self):
        modules = {}
        # modulesInst = self.sideBarInst.modules
        modulesInst = self.__tabModules
        modules[WVHDB_VLINK_TYPE_AFDX] = modulesInst[WHHDB_XPOINT_AFDX_CODE]
        modules[WVHDB_VLINK_TYPE_CETRAC] = modulesInst[WHHDB_XPOINT_CETRAC_CODE]
        modules[WVHDB_VLINK_TYPE_ETNERNET] = modulesInst[WHHDB_XPOINT_ETHERNET_CODE]
        modules[WVHDB_VLINK_TYPE_DISCRETE] = modulesInst[WHHDB_XPOINT_DISCRET_CODE]
        modules[WVHDB_VLINK_TYPE_FIBER] = modulesInst[WHHDB_XPOINT_FIBER_CHANNEL_CODE]
        modules[WVHDB_VLINK_TYPE_1553] = modulesInst[WHHDB_XPOINT_1553_CODE]
        modules[WVHDB_VLINK_TYPE_ARINC825] = modulesInst[WHHDB_XPOINT_ARINC825_CODE]
        for sceneItem in self.items():
            item = sceneItem.getDataObject()
            vlType = item.getVLType()
            tableModule = modules.get(vlType)
            if tableModule:
                tableModule.addRow(sceneItem)

    def getTabModule(self):
        return self.__tabModules

    def openTrafficEditor(self, item):
        protocol = item.text()
        moduleRoute = self.__tabModules[protocol]

        self.__tabs.openEditor(moduleRoute.getEditorIdent(), moduleRoute, isEditableName=False)
        self.__tabs.editors.append(moduleRoute)

    def _createLayout(self):
        topHBox = QHBoxLayout()
        containerList = QWidget()
        container_Layout = QHBoxLayout()
        self.listWidget.show()
        container_Layout.addWidget(self.listWidget)
        containerList.setLayout(container_Layout)
        containerList.setFixedWidth(120)
        topHBox.addWidget(containerList)
        containerModules = QWidget()
        container2_Layout = QHBoxLayout()
        container2_Layout.addWidget(self.__tabs)
        containerModules.setLayout(container2_Layout)
        containerModules.setMinimumWidth(750)
        topHBox.addWidget(containerModules)

        return topHBox

    def _createSidebarModules(self, parent=None):
        modules = []

        afdx = WvguiSceneAFDXTable(self, parent)
        modules.append(('AFDX', 'AFDX', afdx, ('#c0c4c6', '#8d8d8e')))
        cetrac = WvguiSceneCetracTable(self, parent)
        modules.append((WHHDB_XPOINT_CETRAC_CODE, WHHDB_XPOINT_CETRAC_CODE,
                        cetrac, ('#73ef1c', '#5db221')))
        ethernet = WvguiSceneEthernetTable(self, parent)
        modules.append(('Ethernet', 'Ethernet', ethernet, ('#007FFF', '#1C86EE')))

        discrete = WvguiSceneDiscreteTable(self, parent)
        modules.append(('Discrete', 'Discrete', discrete, ('#EE9A00', '#EFA419')))

        fiber = WvguiSceneFiberChannelTable(self, parent)
        modules.append(('FiberChannel', 'FiberChannel', fiber, ('#EE9A00', '#EFA419')))

        mill1553 = WvguiScene1553Table(self, parent)
        modules.append(('1553', '1553', mill1553, ('#EE9A00', '#EFA419')))

        ARINC825 = WvguiSceneARINC825Table(self, parent)
        modules.append(('ARINC825', 'ARINC825', ARINC825, ('#EE9A00', '#EFA419')))

        ARINC825 = WvguiSceneARINC825Table(self, parent)
        modules.append(('CAN', 'CAN', ARINC825, ('#EE9A00', '#EFA419')))

        return modules

    def _createTabModules(self, parent=None):
        modules = {}

        mill1553 = WvguiScene1553Table(self, parent)
        modules[WHHDB_XPOINT_1553_CODE] = mill1553
        self.listWidget.addItem(WHHDB_XPOINT_1553_CODE)

        afdx = WvguiSceneAFDXTable(self, parent)
        modules[WHHDB_XPOINT_AFDX_CODE] = afdx
        self.listWidget.addItem(WHHDB_XPOINT_AFDX_CODE)

        ARINC825 = WvguiSceneARINC825Table(self, parent)
        modules[WHHDB_XPOINT_ARINC825_CODE] = ARINC825
        self.listWidget.addItem(WHHDB_XPOINT_ARINC825_CODE)

        cetrac = WvguiSceneCetracTable(self, parent)
        modules[WHHDB_XPOINT_CETRAC_CODE] = cetrac
        self.listWidget.addItem(WHHDB_XPOINT_CETRAC_CODE)

        discrete = WvguiSceneDiscreteTable(self, parent)
        modules[WHHDB_XPOINT_DISCRET_CODE] = discrete
        self.listWidget.addItem(WHHDB_XPOINT_DISCRET_CODE)

        ethernet = WvguiSceneEthernetTable(self, parent)
        modules[WHHDB_XPOINT_ETHERNET_CODE] = ethernet
        self.listWidget.addItem(WHHDB_XPOINT_ETHERNET_CODE)

        fiber = WvguiSceneFiberChannelTable(self, parent)
        modules[WHHDB_XPOINT_FIBER_CHANNEL_CODE] = fiber
        self.listWidget.addItem(WHHDB_XPOINT_FIBER_CHANNEL_CODE)

        return modules

    def getSidebarModules(self):
        return iter(self.__sidebarModules)

    def _deleteSelEntries(self):
        pass

    @property
    def sitmF(self):
        return self.__sitmF

    @property
    def config(self):
        return self.__config

    @property
    def sceneItemChangedS(self):
        return self.sceneItemChanged

    @property
    def pjdb(self):
        """
        Get the project database represented by this scene.

        :return: Project database
        :rtype: WprjProjectDatabase
        """
        return self.__pjdb

    @property
    def topLevelModelBlock(self):
        """
        :return: The current top-level block
        :rtype: WhdbBlockP
        """
        return self.__topLevelProxy

    @property
    def blockInstancePath(self):
        """
        :return: The path to the current top-level block
        :rtype: Iterable[WhdbBlockInstanceP]
        """
        return self.__topLevelProxy.instanceStack()

    @property
    def viewContext(self):
        """
        :return: Scene's view context
        :rtype: ViewContext
        """
        return self.__viewContext

    def items(self):
        return self.__items

    def getGridFlag(self):
        """
        Get the scene grid flag
        """
        return False

    def setGridFlag(self, gflg):
        """
        Set the scene grid flag
        """
        pass

    def getSnapFlag(self):
        """
        Get the scene grid snap flag
        """
        return False

    def setSnapFlag(self, _):
        """
        Set the scene grid snap flag
        """
        pass

    def setMode(self, _):
        pass

    def popHierarchy(self):
        pass

    def popTopHierarchy(self):
        pass

    def setHierarchy(self, path):
        pass

    def render(self, *args):
        pass

    def itemsBoundingRect(self):
        pass

    def sceneRect(self):
        pass

    def setSceneRect(self, _):
        pass

    def clearSelection(self):
        pass

    def selectEntities(self, entities):
        pass

    def importModules(self):
        """
        Generic function to improt modules
        must be overrode by project
        :return:
        """
        pass

    def getSceneItems(self):
        """
        Get all the scene's items that represent database entities.

        :return: List of database scene items
        :rtype: list[WguiSceneItemP]
        """
        return [item for item in self.items()
                if isinstance(item, WguiSceneItemP)]

    def getSceneItem(self, dbItem):
        """
        Get the scene item for the given database item, if it exists.

        :param dbItem: Database item to retrieve scene item for
        :type dbItem: WhdbBaseP
        :return: Scene item for the given database item
        :rtype: WguiSceneItemP
        """
        ref = self.__itemWRefs.get(dbItem.getUUID())
        if ref is not None:
            item = ref()
            if item is not None:
                return item

        for item in self.items():
            if isinstance(item, WguiSceneItemP):
                dataItem = item.getDataObject()

                if dataItem is dbItem:
                    self.__itemWRefs[dbItem.getUUID()] = weakref.ref(item)
                    return item

    def updateItems(self):
        for item in self.items():
            if isinstance(item, WguiSceneItemP):
                item.updateItem()

    def refresh(self):
        """
        Refresh the graphics scene from the database.
        """
        self.clear()
        self.__muted = True

        block = self.topLevelModelBlock
        for func in block.getOperatorInstances():
            self.addFunctionItem(func)

        for z, i in enumerate(reversed(self.getSceneItems()), 1):
            i.setZ(z)

        self.__muted = False
        self.sceneChanged.emit(True)

    def delSelection(self):
        """
        Delete the selected items.

        """
        selected = self.selectedItems()
        if not selected: return
        # loop in the selected items
        conns, items = partition(
            lambda x: isinstance(x, WguiConnectionSceneItem),
            (i for i in self.selectedItems() if isinstance(i, WguiSceneItemP)))
        for conn in conns:
            self.deleteItem(conn)
        sceneCards = self.getScenCards()
        for item in items:
            data = item.getDataObject()
            # get cards if item has any and delete them
            cards = getattr(data, 'cards', {})
            for card in cards.values():
                if card in sceneCards.keys():
                    self.deleteItem(sceneCards[card])
            self.deleteItem(item)
        # clear the selection
        self.clearSelection()

    def getScenCards(self):
        """
        Get all Io Cards in the scene in the form of a dictionary
        with operator instances as keys and scene item as values
        :return dict
        """
        return {s.getDataObject(): s for s in self.getSceneItems()
                if isinstance(s.getDataObject(), WhhdbIoCardInstance)}

    def selectedItems(self):
        # items = super().selectedItems()
        # self.sceneItemSelected.emit(items)
        # return items
        return []

    def delDbEntities(self, entities):
        """
        Delete the entities items.

        """
        for item in [self.getSceneItem(e) for e in entities]:
            self.deleteItem(item)

    def deleteItem(self, item):
        """
        Delete the given scene item.
        Will also:
         - delete the associated database item from the database,
         - delete any connections attached to the item,
         - update any other affected items.

        :param item: Scene item to be deleted
        :type item: WguiSceneItemP
        """
        if item is None:
            return
        # get the associated object
        data = item.getDataObject()
        ctrl = self.__projectCtx.getControllerFactory().getItemController(data)
        delItms, updItms = ctrl.deleteItem(data, self.topLevelModelBlock)
        if delItms:
            for _itm in delItms:
                sitm = self.getSceneItem(_itm)
                if sitm is not None:
                    self.removeItem(sitm)
        if updItms:
            for _itm in updItms:
                sitm = self.getSceneItem(_itm)
                if sitm is not None:
                    sitm.updateItem()

    def addFunctionItem(self, func):
        """
        Create a function graphics item for the given function database
        object.

        :param func: The function object
        :type func: WhdbOperatorInstanceP
        :return: The newly created connection scene item
        :rtype: WguiOperatorInstanceSceneItem
        """
        if func is None: return None
        if func.getZ() is None: func.setZ(self.getTopZValue() + 1)
        item = self.__sitmF.createOperatorInstanceSceneItem(
            func, self.viewContext)
        self.addItem(item)
        return item

    def addItem(self, item:QGraphicsItem):
        self.items().append(item)
        if isinstance(item, WguiSceneItemP):
            self.__itemWRefs[item.getDataObject().getUUID()] = \
                weakref.ref(item)
        self.sceneItemAdded.emit(item.getDataObject(),
                                 self.__topLevelProxy.instanceStack())

    def createFunctionItem(self, oper=None):
        """
        Add a function item by position

        :param pos: the function position
        :type pos: QPointF
        """
        # select the operator by mode
        if oper is None:
            oper = self.getOperatorMode()
        if oper is None: return
        name = whdb_getNextFunctionName(oper, self.topLevelModelBlock)
        # get the function by operator
        func = self.topLevelModelBlock.createOperatorInstance(oper, name=name)
        # add function to the scene
        sceneFunc = self.addFunctionItem(func)
        sceneFunc.focusNameField()
        return sceneFunc

    def removeItem(self, item):
        if isinstance(item, WguiSceneItemP):
            itemID = item.getDataObject().getUUID()
            if itemID in self.__itemWRefs:
                del self.__itemWRefs[itemID]
        if not self.__muted and isinstance(item, WguiSceneItemP):
            self.sceneItemRemoved.emit(item.getDataObject())
        if isinstance(item, WcmnReleasableP):
            item.release()
        self.items().remove(item)

    def clear(self):
        items = self.items()
        self.__itemWRefs.clear()
        for item in items:
            if isinstance(item, WcmnReleasableP):
                item.release()
        if not self.__muted:
            self.sceneChanged.emit(False)


