# ----------------------------------------------------------------------------
# - whgui-scene-protocol-table                                               -
# - wild cruncher - cockpit : protocol table widget                          -
# ----------------------------------------------------------------------------
# - author    (c) 2018-2018 achraf mahti                                     -
# - copyright (c) 2017-2018 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wvgui_scene_protocol_table
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher traffic route protocol table scene editor

.. moduleauthor:: Achraf Mahti <mahti@silkan.net>
"""
from PyQt4 import QtCore
from functools import partial

from PyQt4.QtGui import QWidget, QTableWidget, QPushButton,\
    QVBoxLayout, QHBoxLayout, QTableWidgetItem, QHeaderView, QComboBox
from silkan_common.scmn_utils import isHexa, isDec, isIpAdress, isMacAdress
from wcr_core.db.whdb_utils import whdb_getNextName
from wcr_core.modules.vlink.db.wvhdb_Vlink_instance import *
from wcr_core.modules.vlink.db.wvhdb_standard_operators import WVHDB_OPER, \
    WVHDB_VLINK_CODE
from wcr_core.modules.sys_map.plugin.wcplg_config_generator_params import *
from wcr_frontend.gui.wgui_scene_item_ids import *
from wcr_frontend.gui.wgui_scene_item import *
from wcr_frontend import showErrorWindow


class WvguiSceneProtocolTable(QWidget):
    """
    The WvguiSceneProtocolTable class is the class responsible to manage
    the protocol table scene action in traffic route view
    that WguiGraphicsScene don't handle
    """
    HEADER = []
    PROTOCOL = None
    COMBO_PARAMS = (LOG_TR_AFDX_REDNT_MGNT_TAG, LOG_TR_AFDX_REDNT_VL_TAG,
                    LOG_TR_STORAGE_TAG, LOG_TR_ETH_ARP_TAG,
                    LOG_TR_ETH_PROTOCOL_TAG, LOG_TR_AFDX_SN_CHECK_TAG)
    #: Signal emitting whenever an editor changes its name
    nameChanged = QtCore.pyqtSignal()
    signalMapper = QtCore.QSignalMapper()

    def __init__(self, scene, parent=None):
        super().__init__(parent)
        # the project database
        self.__scene = scene
        #: Scene items database
        self.sceneItems = {}
        #: table widget
        self.table = QTableWidget()
        #: Add entry button
        self.__addButton = QPushButton('Add')
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')
        self._initElements()
        self.setLayout(self._createLayout())
        self.currentRow = 0
        self.updateTable()

    def getTitle(self):
        return self.PROTOCOL

    def getIcon(self):
        pixmap = QPixmap(100, 100)
        pixmap.fill(QColor(self.TABCOLOR))
        return pixmap

    def isICDProject(self):
        return False

    def activated(self):
        return None

    def getEditorIdent(self):
        return self.PROTOCOL

    def _initElements(self):
        self.header = ['Name']
        self.header.extend([prm[1] for prm in self.HEADER])
        self.header.append('uuid')
        self.table.setColumnCount(len(self.header))
        self.uuidCol = len(self.header)-1
        # uuid col is used for tracking Vl instance only so it is invisible
        self.table.setColumnHidden(self.uuidCol, True)
        self.table.setHorizontalHeaderLabels(self.header)
        self.table.horizontalHeader().setResizeMode(QHeaderView.Stretch)
        self.__addButton.clicked.connect(self._addEntry)
        self.__deleteButton.clicked.connect(self._deleteSelEntries)
        self.table.itemChanged.connect(self.itemChanged)

    def updateTable(self):
        for item in self.__scene.items():
            self.addRow(item.getDataObject())

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.addWidget(self.table)
        Hlayout = QHBoxLayout()
        Hlayout.addWidget(self.__addButton)
        Hlayout.addWidget(self.__deleteButton)
        layout.addLayout(Hlayout)
        return layout

    def addRow(self, itemScene):
        item = itemScene.getDataObject()
        self.table.setRowCount(self.currentRow + 1)
        name = QTableWidgetItem(str(item.getName()) if
                                item.getName() is not None else '')
        vlType = item.getVLType()
        if vlType:
            self.table.setItem(self.currentRow, 0, name)
            for i, (name, value, key) in enumerate(item.getVLParameters(vlType)):
                self._createColumn(i, value, key)
            # add uuid as last row
            prmItem = QTableWidgetItem(str(item.getUUID()))
            self.table.setItem(self.currentRow, self.uuidCol, prmItem)
            self.currentRow += 1
            self.sceneItems[str(item.getUUID())] = itemScene

    def _createTrafficRoute(self):
        oper = WVHDB_OPER[WVHDB_VLINK_CODE]
        return self.__scene.createFunctionItem(oper)

    def _createColumn(self, i, value, key):
        if self.dataCheck(key, value):
            if key in self.COMBO_PARAMS:
                if key == LOG_TR_STORAGE_TAG:
                    values = ('Queue', 'Sample')
                elif key in (LOG_TR_ETH_ARP_TAG, LOG_TR_AFDX_REDNT_MGNT_TAG,
                             LOG_TR_AFDX_REDNT_VL_TAG, LOG_TR_AFDX_SN_CHECK_TAG):
                    values = ('Yes', 'No')
                elif key == LOG_TR_ETH_PROTOCOL_TAG:
                    values = ('TCP', 'UDP')

                prmItem = QComboBox()
                for name in sorted(values):
                    prmItem.addItem(name, name)
                idxCombo = prmItem.findText(value)
                prmItem.setCurrentIndex(idxCombo)
                prmItem.currentIndexChanged.connect(partial(
                                                    self.itemChanged, (self.currentRow, i + 1)))
                self.table.setCellWidget(self.currentRow, i + 1, prmItem)

            else:
                prmItem = QTableWidgetItem(str(value)
                                       if value is not None else '')
                self.table.setItem(self.currentRow, i + 1, prmItem)

    def _addEntry(self):
        itemScene = self._createTrafficRoute()
        item = itemScene.getDataObject()
        name = item.getName()
        protName = name.replace('Traffic', self.PROTOCOL)
        newName = whdb_getNextName(protName, [sc.getDataObject()
                                          for sc in self.sceneItems.values()])
        item.setName(newName)
        item.setVLType(self.PROTOCOL)
        self.addRow(itemScene)

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        """:type: list[QModelIndex]"""
        selIdx = self.table.selectedIndexes()
        deleteRows = []
        for idx in selIdx:
            row = idx.row()
            # Get name of the TrafficRoute wherever it is clicked on the row
            uuid = idx.sibling(row, self.uuidCol).data()
            sceneItem = self.sceneItems.get(uuid)
            if sceneItem is None:
                continue
            deleteRows.append((row, sceneItem))
        uniqueRows = set(deleteRows)
        # iter the item in reverse remove them without position issue
        for row, sceneItem in sorted(uniqueRows,
                                     key=lambda row: row[0],
                                     reverse=True):
            self.__scene.deleteItem(sceneItem)
            uuid = sceneItem.getDataObject().getUUID()
            del self.sceneItems[str(uuid)]
            self.table.removeRow(row)
            self.currentRow -= 1
            self.table.selectRow(min(selIdx).row())

    def itemChanged(self, selec):
        """
        Signal handler : when item is change update the traffic route
        parameter associated
        """
        if type(selec) == tuple:
            row, col = selec
        else:
            row = selec.row()
            col = selec.column()
        itemUuid = self.table.item(row, self.uuidCol)

        if itemUuid is None:
            return
        uuid = itemUuid.text()
        item = self.sceneItems.get(uuid)
        if not item:
            return
        vlInst = item.getDataObject()
        if type(selec) == tuple:
            data = self.table.cellWidget(row, col).currentText()
        else:
            data = selec.text()

        if col == 0:
            vlInst.setName(data)
        elif type(selec) == tuple:
            # don't check on combobox
            paramkey = self.HEADER[col - 1][0]
            vlInst.setVLParameter(self.PROTOCOL, paramkey, data)
        else:
            # sub 1 because we added Name col
            paramkey = self.HEADER[col-1][0]
            isOk = self.dataCheck(paramkey, data)
            if isOk:
                vlInst.setVLParameter(self.PROTOCOL, paramkey, data)
            else:
                selec.setText('')

    def dataCheck(self, paramkey, data):
        """
        Function that will check the data Type depending on the column
        :param paramkey: the name of the column
        :type paramkey: Str
        :param data: data to be checked
        :type data: Str
        :return: Bool
        """
        if not data:
            return True
        isOk = False
        if paramkey in (LOG_TR_ETH_IP_SRC_TAG, LOG_TR_ETH_IP_DEST_TAG):
            isOk, title, message = isIpAdress(data, returnMessages=True)
        elif paramkey in (LOG_TR_ETH_MAC_DEST_TAG, LOG_TR_ETH_MAC_SRC_TAG):
            isOk, title, message = isMacAdress(data, returnMessages=True)
        elif paramkey in self.COMBO_PARAMS:
            return True
        else:
            if isHexa(data) or isDec(data):
                isOk = True
            else:
                title = 'Data Type Error'
                message = 'Field "{}" must have integer or hexadecimal data!'
        if not isOk:
            showErrorWindow(title, message, paramkey)
        return isOk


class WvguiSceneAFDXTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_AFDX_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_AFDX
    TABCOLOR = "black"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)


class WvguiSceneCetracTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_CETRAC_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_CETRAC
    TABCOLOR = "green"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)


class WvguiSceneEthernetTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_ETNERNET_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_ETNERNET
    TABCOLOR = "blue"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)


class WvguiSceneDiscreteTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_DISCRETE_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_DISCRETE
    TABCOLOR = "orange"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)


class WvguiSceneFiberChannelTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_FIBER_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_FIBER
    TABCOLOR = "orange"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)


class WvguiScene1553Table(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_1553_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_1553
    TABCOLOR = "orange"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)

class WvguiSceneARINC825Table(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_ARINC825_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_ARINC825
    TABCOLOR = "orange"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)
